<?php


namespace Plugins\Ccm19CBImport\Controller;

use App\Component\Curl;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\LocaleService;
use App\Component\OnBoarding\ScanService;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\Embedding as EmbeddingModel;
use App\Model\EmbeddingAsset;
use App\Model\Index\DomainIndex;
use App\Model\LegalText;
use App\Model\LocaleBase;
use App\Model\Placeholder as PlaceholderModel;
use App\Model\PluginState;
use App\Model\Purpose;
use App\Model\User;
use App\Utils;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use Plugins\Ccm19CBImport\Model\CBImportModel;

/**
 * Controller for CBImport - domain dependent
 * @Route("/cbimport/", name="plugin_ccm19_cookieimport_cbbot_")
 */
class CookieBotImporter extends AbstractController
{

	/**
	 * @Route("cookiebot", name="index", methods={"HEAD", "GET"})
	 * @param PluginState $state
	 * @return Response
	 */
	public function index(PluginState $state): Response
	{
		if ($state->isAllowedForCurrentUser()) {

            //wenn KEIN Access aufgrund von Tarif besteht und Trial Expired ist
            if($state->trialIsExpired())
            {
                return $this->renderTrialExpired();
            }

			//Render template with all Data
			return $this->render('@plugin:Ccm19CBImport/internal/cookiebot.index.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'cookieBotData' => CBImportModel::byName("cookiebot"),
			]);
		}
		return $this->redirectToRoute('app_domains_index');
	}

	/**
	 * @Route("cookiebot/import_list", name="import_list", methods={"HEAD",
	 *     "GET"})
	 * @param PluginState $state
	 * @return Response
	 */
	public function showImported(PluginState $state): Response
	{
		if (!$state->isAllowedForCurrentUser()) {
			return $this->redirectToRoute('app_domains_index');
		}

        //wenn KEIN Access aufgrund von Tarif besteht und Trial Expired ist
        if($state->trialIsExpired())
        {
            return $this->renderTrialExpired();
        }

		//Nochmal die importierten Domains holen
		$importData = CBImportModel::byName("cookiebot");
		$imported = array();
		$emBedList = array();
		$cookies = array();

		//Alle durchloopen und erstellen...
		foreach ($importData->getDomainsArray() as $dom) {
			//Aktuelle Domain holen
			$domain = Domain::byName($dom);

			if(empty($domain->getId()))
			{
				continue;
			}
			//Die passenden Embeddings
			Embedding::switchRepositoryContext(User::loggedInUser(), $domain);
			$allEmbeddings = Embedding::all();

			//Von denen widerum die cookies
			foreach ($allEmbeddings as $ebd) {
				$cookies[$ebd->getId()] = $ebd->getAssets();
			}

			$imported[] = $domain;
			$emBedList[$domain->getId()] = Embedding::all();
		}

		//Render template with all Data
		return $this->render('@plugin:Ccm19CBImport/internal/cookiebot.imported_list.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'cookieBotDataImported' => $imported,
			'cookieBotDataImportedEmbeddingList' => $emBedList,
			'cookieBotDataImportedEmbeddingListAssets' => $cookies,
		]);

	}

	/**
	 * @Route("cookiebot/import", name="import", methods={"GET"})
	 * @param PluginState $state
	 * @param TranslatorInterface $translator
	 * @param ScanService $service
	 * @param Request $request
	 * @return Response
	 */
	public function import(PluginState $state, TranslatorInterface $translator, ScanService $service, Request $request): Response
	{
		if (!$state->isAllowedForCurrentUser()) {
			return $this->redirectToRoute('app_domains_index');
		}

        //wenn KEIN Access aufgrund von Tarif besteht und Trial Expired ist
        if($state->trialIsExpired())
        {
            return $this->renderTrialExpired();
        }

		//Connect Daten holen
		$importData = CBImportModel::byName("cookiebot");
		$curl = new Curl();

		//Vorlagen aus der zentralen DB holen - kann einige Sekunden dauern
		$embeddingsVorlagen = EmbeddingDb::fetchEmbeddingList("");

		//Alle durchloopen und erstellen...
		foreach ($importData->getDomainsArray() as $dom) {
			//Wenn nix ist - dann ist auch nix
			if (empty($dom)) {
				continue;
			}
			//Domain erstellen oder holen falls schon existiert
			$selectedDomain = $this->saveDomain(trim($dom), $translator, $importData->getCulture());

			//Daten aus der URL / API von CB holen
			$url = 'https://consent.cookiebot.com/api/v1/' . $importData->getApiKey() . '/json/domaingroup/' . $importData->getSerial() . '/' . $importData->getCulture() . '/domain/' . trim($dom) . '/cookies';
			$curl->get($url);
			$response = $curl->getResponse();

			//Antwort verarbeiten und mit Embeddings aus unserer DB abgleichen, Embeddingsarray mit Assets erstellen
			$embeddings = $this->createEmbeddings($response, $embeddingsVorlagen);

			//Embeddings erstellen und speichern in der Domain
			$this->saveEmbeddings($embeddings, $selectedDomain);

			//Jetzt die URL Daten Impresseum / Datenschutz und Platzhalter holen und in der Domain speichern
			$this->saveUrlUndCoDataFromDomain($selectedDomain, $service, $translator);

			//Huch schon fertig? dann die nächste...
			DomainIndex::rebuild();
		}
		DomainIndex::rebuild();

		return $this->redirectToRoute('plugin_ccm19_cookieimport_cbbot_import_list');
	}

	/**
	 * @param $selectedDomain
	 * @param ScanService $service
	 * @param TranslatorInterface $translator
	 * @return bool
	 */
	private function saveUrlUndCoDataFromDomain($selectedDomain, ScanService $service, TranslatorInterface $translator)
	{
		//Daten holen
		$resData = $service->fetchImprintData($selectedDomain->getUrl());

		//Daten speichern
		$this->saveTextData($resData, $translator);

		return true;
	}

	/**
	 * @param $resData
	 * @param TranslatorInterface $translator
	 * @return bool
	 */
	public function saveTextData($resData, TranslatorInterface $translator): bool
	{
		/** @var LocaleBase[] $locales */
		$locales = LocaleService::enabledLocales();

		foreach ($locales as $locale) {
			/** @var LegalText|null $legalText */
			$legalText = LegalText::byLocale($locale);

			if ($legalText === null) {
				$legalText = LegalText::create()->setLocale($locale);
			}

			$legalText
				->setImprintScheme('link')
				->setPrivacyScheme('link')
				->setImprintUrl($resData['impressum'] ?? '')
				->setPrivacyUrl($resData['datenschutz_url'] ?? '');
		}

		LegalText::flush();

		$placeholders = PlaceholderModel::all();

		// Platzhalter ersetzen wenn möglich...
		if (array_key_exists("FIRMA", $placeholders)) {
			if (!empty($resData['name'])) {
				$placeholders['FIRMA']->setValue($resData['name']);
			}
		}
		if (array_key_exists("Firma_Datenschutz", $placeholders)) {
			if (!empty($resData['datenschutz_url'])) {
				$placeholders['Firma_Datenschutz']->setValue($resData['datenschutz_url']);
			}
		}
		if (array_key_exists("Firma_Ort", $placeholders)) {
			if (!empty($resData['stadt'])) {
				$placeholders['Firma_Ort']->setValue($resData['plz'] . " " . $resData['stadt']);
			}
		}

		PlaceholderModel::flush();
		return true;
	}

	/**
	 * @param $embeddings
	 * @param $selectedDomain
	 */
	private function saveEmbeddings($embeddings, $selectedDomain)
	{
		Domain::select($selectedDomain->getId());
		//Alles zurücksetzen
		EmbeddingModel::clear();
		EmbeddingAsset::clear();

		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleName = $defaultLocale->getName();

		if(stristr($defaultLocaleName,"_"))
		{
			$df = explode("_",$defaultLocaleName);
			$defaultLocaleName = trim($df['0']);
		}

		foreach ($embeddings as $embeddingKey => $embeddingData) {
			$embedding = EmbeddingModel::create();

			// Script-Blocking
			$codeBlock = explode("\n", trim((string)($embeddingData['code_block'] ?? '')));
			$codeBlock = array_filter($codeBlock, function ($line) {
				return trim($line) !== '';
			});
			$codeBlock = array_values(array_map(function ($line) {
				return trim($line);
			}, $codeBlock));

			// Zweck
			$purpose = Purpose::find($embeddingData['purpose'] ?? '') ?? Purpose::find('15c61c3');
			assert($purpose !== null);

			// Allgemeines speichern
			$embedding
				->setName($embeddingData['name'] ?? '')
				->setActive((bool)($embeddingData['active'] ?? true))
				->setCodeSnippet($embeddingData['code_snippet'] ?? '')
				->setCodeBlock($codeBlock)
				->setPurpose($purpose)
				->setEmbeddingVendor($embeddingData['embedding_vendor'] ?? '')
				->setDescription($embeddingData['translations'][$defaultLocaleName]['description'] ?? '')
				->setPrivacyPolicyUrl($embeddingData['translations'][$defaultLocaleName]['privacy_policy_url'] ?? '')
				->setCollectedDataInfo($embeddingData['translations'][$defaultLocaleName]['collected_data_info'] ?? '')
				->setPurposeOfDataCollection($embeddingData['translations'][$defaultLocaleName]['purpose_of_data_collection'] ?? '')
				->setLegalBasis($embeddingData['translations'][$defaultLocaleName]['legal_basis'] ?? '')
				->setPlaceOfProcessing($embeddingData['translations'][$defaultLocaleName]['place_of_processing'] ?? '')
				->setAutoUpdated(true)
				;

			// Übersetzungen speichern
			foreach (LocaleService::enabledLocales() as $locale) {
				$localeName = $locale->getName();

				if(stristr($localeName,"_"))
				{
					$df = explode("_",$localeName);
					$localeName = trim($df['0']);
				}
				if (!isset($embeddingData['translations'][$localeName]['description'])) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				$embedding
					->setTranslatedDescription($locale, $embeddingData['translations'][$localeName]['description'] ?? '')
					->setTranslatedPrivacyPolicyUrl($locale, $embeddingData['translations'][$localeName]['privacy_policy_url'] ?? '')
					->setTranslatedCollectedDataInfo($locale, $embeddingData['translations'][$localeName]['collected_data_info'] ?? '')
					->setTranslatedPurposeOfDataCollection($locale, $embeddingData['translations'][$localeName]['purpose_of_data_collection'] ?? '')
					->setTranslatedLegalBasis($locale, $embeddingData['translations'][$localeName]['legal_basis'] ?? '')
					->setTranslatedPlaceOfProcessing($locale, $embeddingData['translations'][$localeName]['place_of_processing'] ?? '');
			}

			// Assets hinzufügen
			foreach ($embeddingData['assets'] as $assetKey => $assetData) {
				$assetName = (string)($assetData['name'] ?? '');

				if ($assetName == '') {
					continue;
				}

				$asset = EmbeddingAsset::create()
					->setName($assetName)
					->setDynamic((bool)($assetData['dynamic'] ?? false))
					->setStorageType((string)($assetData['storageType'] ?? ''))
					->setLifetime((string)($assetData['translations'][$defaultLocaleName]['lifetime'] ?? ''))
					->setDescription((string)($assetData['translations'][$defaultLocaleName]['description'] ?? ''));

				// Übersetzungen speichern
				foreach (LocaleService::enabledLocales() as $locale) {
					$localeName = $locale->getName();
					if (!isset($assetData['translations'][$localeName]['description'])) {
						continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
					}
					$asset
						->setTranslatedLifetime($locale, $assetData['translations'][$localeName]['lifetime'] ?? '')
						->setTranslatedDescription($locale, $assetData['translations'][$localeName]['description'] ?? '');
				}

				$embedding->addAsset($asset);
			}
			EmbeddingAsset::flush();

			$embedding->save();
		}
	}

	/**
	 * @param string $newDomain
	 * @param TranslatorInterface $translator
	 * @param string $country
	 * @return Domain|null
	 */
	private function saveDomain(string $newDomain, TranslatorInterface $translator, string $country = "de"): ?Domain
	{
		$domain = Domain::byName($newDomain);
		if (!empty($domain)) {
			return $domain;
		}

		/** @var User $user */
		$user = User::loggedInUser()->getAccountOwner();
		if ($user->isDomainLimitReached()) {
			$this->addFlash('danger', $translator->trans('You\'ve reached the domain limit. Please contact your provider to raise that limit.'));
			return null;
		}

		$domainInfo = parse_url(filter_var(
			preg_match('~^https?://~', $newDomain) ? $newDomain : 'https://' . $newDomain,
			FILTER_SANITIZE_URL
		));
		if (isset($domainInfo['host'])) {
			$newDomain = $domainInfo['host'];
			$urlPath = '/' . ltrim($domainInfo['path'] ?? '/', '/');
			$urlScheme = $domainInfo['scheme'] ?? 'https';
			$url = $urlScheme . $newDomain . $urlPath;
		} else {
			$urlPath = '/';
			$urlScheme = 'https';
			$url = $newDomain;
		}

		if ($newDomain) {
			$domain = Domain::createForUser($user);
			$domain
				->setName($newDomain)
				->setUrlPath($urlPath)
				->setUrlScheme($urlScheme)
				->setFrontendWidgetEnabled(false)
				->setWhitelabel(false)
				->setCountry(null)
				->setInlineScriptsBlocked(false)
				->setSameDomainScriptsBlocked(false)
				->setManagementStructurePreLegalConsent(true)
				->setManagementStructure(Domain::MANAGEMENT_STRUCTURE_EMBEDDING)
				->setCodeMinification(true)
				->setfrontendLegalEnabled(true)
				->setFrontendWidgetEnabled(true)
				->setOnBoarding(true)
				->save();

			// Index neu generieren
			DomainIndex::rebuild();
			Utils::getSession()->set('form_domain', $url);
			return $domain;
		} else {
			return null;
		}
	}

	/**
	 * Erstelle eine Domain aus den Daten
	 *
	 * @param $response
	 * @param $embeddingsVorlagen
	 * @return array()
	 */
	private function createEmbeddings($response, $embeddingsVorlagen)
	{
		//Unser Cookie Lookup
		$cookies = $this->getCookieArray($embeddingsVorlagen);
		$embeddings = array();

		//Cookies einlesen
		$data = json_decode($response, true);
		//print_R($data);exit();

		//Alle Cookies durchgehen und Embeddings daraus holen.
		if (!empty($data['cookies'])) {
			foreach ($data['cookies'] as $cbkey => $cookie) {
				if (!empty($cookies)) {
					foreach ($cookies as $cookieCCMName => $embediID) {
						if (!empty($cookieCCMName) && strlen($cookieCCMName) > 2) {
							//str_starts_with(string $haystack, string $needle): bool
							if (str_starts_with(strtolower($cookie['Name']), strtolower($cookieCCMName)) && strtolower($cookie['Name']) != "cookieconsent") //
							{
								//print_r($cookie['Name']." = ".$cookieCCMName." -> ".$embediID." \n");
								$embeddings[$embediID] = $embeddingsVorlagen[$embediID];
								unset($data['cookies'][$cbkey]);
							}
						}
					}
				}
			}
		}

		//nicht erkannte Cookies -> $data['cookies'] in sonstige einlesen
		if (!empty($data['cookies'])) {
			$assets = [];
			foreach ($data['cookies'] as $k => $cookieDat) {
				if(empty(trim($cookieDat['Name'])))
				{
					continue;
				}
				$assets = [
					'id' => -($k + 1),
					'name' => $cookieDat['Name'],
					'storageType' => 'cookie',
					'dynamic' => false,
					'translations' => [
						'de' => ['lifetime' => $cookieDat['ExpireDescription'], 'description' => $cookieDat['PurposeDescription']],
						'en' => ['lifetime' => $cookieDat['ExpireDescription'], 'description' => $cookieDat['PurposeDescription']],
					],
				];

			}

			if ($assets) {
				$embeddings[] = [
					'id' => 0,
					'name' => 'Sonstige',
					'purpose' => '15c61c3',
					'code_snippet' => '',
					'code_block' => '',
					'embedding_vendor' => 'N.A.',
					'translations' => [
						'de' => [
							'label' => '',
							'description' => '',
							'privacy_policy_url' => '',
							'collected_data_info' => '',
							'purpose_of_data_collection' => '',
							'legal_basis' => 'Einwilligung, Art. 6 Abs. 1 lit. a DSGVO',
							'place_of_processing' => '',
						],
						'en' => [
							'label' => 'Other',
							'description' => '',
							'privacy_policy_url' => '',
							'collected_data_info' => '',
							'purpose_of_data_collection' => '',
							'legal_basis' => 'Consent, Art. 6 para. 1 s. 1 lit. c GDPR',
							'place_of_processing' => '',
						]
					],
					'assets' => $assets,
				];
			}
		}

		//Nicht erkannte Cookies an DB melden UND als sonstige Eintragen...
		$ok = $this->reportUnknownCookies($data['cookies']);

		return $embeddings;
	}

	/**
	 * TODO Funktion mit Leben füllen... Report an API?
	 *
	 * @param array $cookies
	 * @return bool
	 */

	private function reportUnknownCookies($cookies = array()): bool
	{
		return true;
	}

	/**
	 * Erstellt ein einfaches Array mit cookiename => id des Embeddings
	 *
	 * @param $embeddingsVorlagen
	 * @return array
	 */
	private function getCookieArray($embeddingsVorlagen)
	{
		$cookieNamen = array();
		if (!empty($embeddingsVorlagen)) {
			foreach ($embeddingsVorlagen as $ebk => $embed) {
				//print_r($embed);
				foreach ($embed['assets'] as $ask => $assets) {
					$name = str_ireplace('*', "", $assets['name']);
					$cookieNamen[$name] = $ebk;
				}
			}
		}
		return $cookieNamen;
	}

	/**
	 * @Route("cookiebot/post", name="post", methods={"POST"})
	 * @param PluginState $state
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function savePost(PluginState $state, Request $request, TranslatorInterface $translator): Response
	{
		if (!$state->isAllowedForCurrentUser()) {
			return $this->redirectToRoute('app_domains_index');
		}
		if ($request->getRealMethod() === 'POST') {

			try {
				CsrfToken::get()->checkRequest($request);
			} catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('plugin_ccm19_cookieimport_cbbot_index');
			}

			//activatelegal
			if (empty($request->get('enableFrontendLegal'))) {
				$this->addFlash('danger', $translator->trans('Legal activation ist not set, please try again.'));
				return $this->redirectToRoute('plugin_ccm19_cookieimport_cbbot_index');
			}

			//Daten rausholen
			$submit = $request->request->all();

			//Check Anzahl
			if(!empty($submit['cookieBotData']['domains']))
			{
				$domainsDat = explode("\n", trim($submit['cookieBotData']['domains']));
				$domCount = count($domainsDat);

				if ($domCount > User::loggedInUser()->getMaxDomainCount())
				{
					$this->addFlash('danger', $translator->trans('To much domains - the max number of domains is reached.'));
					return $this->redirectToRoute('plugin_ccm19_cookieimport_cbbot_index');
				};
			}


			if (!empty($submit['cookieBotData'])) {
				$import = CBImportModel::byName("cookiebot");
				if (empty($import)) {
					$import = CBImportModel::create();
				}
				$domains = explode("\n", trim($submit['cookieBotData']['domains']));
				$import->setName("cookiebot");
				$import->setApiKey($submit['cookieBotData']['apikey']);
				$import->setCulture($submit['cookieBotData']['culture']);
				$import->setSerial($submit['cookieBotData']['serial']);
				$import->setDomains(($domains));
				$import->save();
			} else {
				$this->addFlash('info', $translator->trans('No data are send!'));
				return $this->redirectToRoute('plugin_ccm19_cookieimport_cbbot_index');
			}

			$this->addFlash('success', $translator->trans('Data were saved'));
			return $this->redirectToRoute('plugin_ccm19_cookieimport_cbbot_import');
		}


		//Render template with all Data
		return $this->render('@plugin:Ccm19CBImport/internal/cookiebot.index.html.twig', [
			'csrfToken' => CsrfToken::get()
		]);
	}

    /**
     * @return \Symfony\Component\HttpFoundation\Response
     */
    public function renderTrialExpired(): Response
    {
        return $this->render('@plugin:Ccm19CBImport/internal/trial_end.html.twig', [
            'csrfToken' => CsrfToken::get(),
        ]);
    }

}
