<?php

namespace Plugins\Ccm19Imex\Component;

use Plugins\Ccm19Imex\Event\ExportDomainEvent;
use Plugins\Ccm19Imex\Event\ExportClientEvent;
use App\Component\LocaleService;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Model\Cookie;
use App\Model\Domain;
use App\Model\EmbeddingBase;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingAssetBase;
use App\Model\LegalText;
use App\Model\Locale;
use App\Model\LocaleBase;
use App\Model\Model;
use App\Model\Placeholder;
use App\Model\Protocol;
use App\Model\Purpose;
use App\Model\SubGroups;
use App\Model\SystemTheme;
use App\Model\UserTheme;
use App\Model\Theme;
use App\Model\User;
use App\Model\UserEmbedding;
use App\Model\UserLocale;
use App\Utils;

/**
 * Class Export
 * @package Plugins\Ccm19Imex\Component
 */
class Export
{
	public function __construct(){ }

	/**
	 * Starting method for the export of data in a client
	 *
	 * @param User $client
	 * @param array $params
	 * @return array
	 */
	public static function exportClient(User $client, array $params=[]): array
	{
		$prevUser = User::getCurrentApiUser();
		User::setApiUser($client->getApiKey());
		Model::closeUserModels();
		Domain::switchToUser($client);

		try {
			$userLocales = array_map(['self', 'makeLocaleExportData'], UserLocale::all());
			$domains = self::exportDomains($client, Domain::all(), $params);

			// Strip locales exported per domain that are exactly the same as a user locale.
			foreach ($domains as $domainKey => $domain) {
				if (isset($domain['locales']) && $domain['locales'] && $userLocales) {
					$domains[$domainKey]['locales'] = array_udiff_assoc($domain['locales'], $userLocales, function (array $a, array $b) {
						return $a <=> $b;
					});
				}
			}

			$result = [
				'clientData' => self::makeClientExportData($client),
				'domains' => $domains,
				'subusers' => self::makeSubUsersExportData($client),
				'subuserGroups' => self::makeSubUserGroupsExportData($client),
				'userThemes' => array_values(array_map(['self', 'makeThemeExportData'], UserTheme::all())),
				'userEmbeddings' => self::buildEmbeddingExportData(UserEmbedding::all(), $client->getUserEmbeddingLocales()),
				'standardTexts' => $userLocales,
			];

			$event = new ExportClientEvent($client, $result);
			Utils::dispatchEvent(ExportClientEvent::NAME, $event);
			$result = $event->getReturn();
		} finally {
			User::setApiUser($prevUser ? $prevUser->getApiKey() : '');
			Model::closeUserModels();
		}
		return $result;
	}

	/**
	 * Starting method for the export of multiple clients
	 *
	 * @param User[] $clients
	 * @return array
	 */
	public static function exportClients(array $clients): array
	{
		return [
			'clients' => array_map(function (User $client) {
				return self::exportClient($client);
			}, $clients),
		];
	}

	/**
	 * @return array
	 */
	public static function exportSystemThemes(): array
	{
		return array_values(array_map(['self', 'makeThemeExportData'], SystemTheme::all()));
	}

	/**
	 * Starting method for the export of data in a domain
	 *
	 * @param Domain $domain
	 * @param string[] $exportData
	 * @return array
	 */
	public static function exportDomain(Domain $domain, array $exportData=[]): array
	{
		// If empty, everything will be exported
		$filteredExport = !empty($exportData);

		// Switch to domain
		$prevDomainId = Domain::getDomainIdForRequest();
		Domain::setDomainForRequest($domain->getId());
		Model::closeDomainModels();

		try {
			$activeLanguages = LocaleService::enabledLocales();
			$localesOfInterest = array_merge(Locale::all(), $activeLanguages);

			$activeLanguages = array_combine(
				array_map(function (LocaleBase $locale) { return $locale->getName(); }, $activeLanguages),
				$activeLanguages
			);
			$localesOfInterest = array_combine(
				array_map(function (LocaleBase $locale) { return $locale->getName(); }, $localesOfInterest),
				$localesOfInterest
			);

			$data = [];

			// Domain-specific
			if (isset($exportData['exportDomain']) && $exportData['exportDomain'] && $filteredExport || $filteredExport == false) {
				$data['domainData'] = self::makeDomainExportData($domain);
			}
			// Themes
			if (isset($exportData['exportThemes']) && $exportData['exportThemes'] && $filteredExport || $filteredExport == false) {
				$data['themes'] = array_values(array_map(['self', 'makeThemeExportData'], Theme::all()));
			}
			// Cookies Statement Table
			if (isset($exportData['exportCookieStatementTable'] ) && $exportData['exportCookieStatementTable'] && $filteredExport || $filteredExport == false) {
				$data['cookieStatementTable'] = self::makeCookieStatementTableExportData($domain);
			}
			// Texts & Language
			if (isset($exportData['exportLocales']) && $exportData['exportLocales'] && $filteredExport || $filteredExport == false) {
				$data['locales'] = array_map(['self', 'makeLocaleExportData'], $localesOfInterest);
			}
			// Imprint
			if (isset($exportData['exportImprint']) && $exportData['exportImprint'] && $filteredExport || $filteredExport == false) {
				$data['imprint'] = self::makeImprintExportData($domain, $activeLanguages);
			}
			// Data protection
			if (isset($exportData['exportDataProtection']) && $exportData['exportDataProtection'] && $filteredExport || $filteredExport == false) {
				$data['dataProtection'] = self::makeDataProtectionExportData($domain, $activeLanguages);
			}
			// Accessibility
			if (isset($exportData['exportAccessibility']) && $exportData['exportAccessibility'] && $filteredExport || $filteredExport == false) {
				$data['accessibility'] = self::makeAccessibilityExportData($domain, $activeLanguages);
			}
			// Placeholders
			if (isset($exportData['exportPlaceholders']) && $exportData['exportPlaceholders'] && $filteredExport || $filteredExport == false) {
				$data['placeholders'] = self::makePlaceholderExportData($domain);
			}
			// Embeddings
			if (isset($exportData['exportEmbeddings']) && $exportData['exportEmbeddings'] && $filteredExport || $filteredExport == false) {
				$data['embeddings'] = self::makeEmbeddingExportData($domain, $activeLanguages);
			}
			// Cookies (obsolete)
			if (isset($exportData['exportCookies']) && $exportData['exportCookies'] && $filteredExport || $filteredExport == false) {
				$data['cookies'] = self::makeCookieExportData($domain, $activeLanguages);
			}
			// Consents
			if (isset($exportData['exportConsents']) && $exportData['exportConsents'] && $filteredExport || $filteredExport == false) {
				$data['consents'] = self::makeConsentExportData($domain);
			}

			$event = new ExportDomainEvent($domain, $data);
			Utils::dispatchEvent(ExportDomainEvent::NAME, $event);
			$data = $event->getReturn();
		} finally {
			Domain::setDomainForRequest($prevDomainId);
			Model::closeDomainModels();
		}

		return $data ?? [];
	}

	/**
	 * Starting method for the export of multiple domains
	 *
	 * @param User $user
	 * @param Domain[] $domains
	 * @param array $params
	 * @return array
	 */
	public static function exportDomains(User $user, array $domains, array $params=[]): array
	{
		Domain::switchToUser($user);

		return array_map(function (Domain $domain) use ($params) {
			return self::exportDomain($domain, $params);
		}, $domains);
	}

	/**
	 * @param User $client
	 * @return array
	 */
	private static function makeClientExportData(User $client): array
	{
		$image = $client->getUserImage();
		$imageData = $image ? base64_encode(file_get_contents($image->getPathname())) : null;
		$tz = $client->getDefaultTimeZone();
		$result = [
			'username' => $client->getUsername(),
			'role' => $client->getRole(),
			'apiKey' => $client->getApiKey(),
			'generalApiKey' => $client->getGeneralApiKey(),
			'createTimeStamp' => $client->getCreateTimeStamp(),
			'lastLogin' => $client->getLastLogin(),
			'passwordHash' => $client->getPasswordHash(),
			'passwordLocked' => $client->isPasswordLocked(),
			'emailAddress' => $client->getEmailAddress(),
			'invoiceEmailAddress' => $client->getInvoiceEmailAddress(),
			'dunningEmailAddress' => $client->getDunningEmailAddress(),
			'firstName' => $client->getFirstName(),
			'lastName' => $client->getLastName(),
			'company' => $client->getCompany(),
			'onboarding' => $client->isOnBoarding(),
			'active' => $client->isActive(),
			'maxWLCount' => $client->getMaxWLCount(),
			'defaultLocale' => $client->getDefaultLocale()->getName(),
			'maxDomainCount' => $client->getMaxDomainCount(),
			'MaxCallCount' => $client->getMaxCallCount(),
			'autoAddDomains' => $client->isAutoAddDomains(),
			'vatId' => $client->getVatId(),
			'strNr' => $client->getStrNr(),
			'plz' => $client->getPlz(),
			'city' => $client->getCity(),
			'settings' => $client->getSettings(),
			'mailPref' => $client->getMailingPreferences(),
			'country' => $client->getCountryCode(),
			'defaultTimeZone' => $tz ? $tz->getName() : null,
			'image' => $imageData,
			'2fa' => $client->getAllTwoFactorAuthParameters(),
			'perferredColorScheme' => $client->getPreferredColorScheme(),
			'newDomainOnlyActiveOnMatchingDomain' => $client->isNewDomainOnlyActiveOnMatchingDomain(),
			'allowAutoUpdateOfEmbeddings' => $client->isAllowAutoUpdateOfEmbeddings(),
			'sendFullConsentAndRejectionAcrossDomains' => $client->isSendFullConsentAndRejectionAcrossDomains(),
		];

		$externalMetadata = $client->getExternalMetadata();
		if ($externalMetadata) {
			$result['externalMetadata'] = $externalMetadata;
		}

		// Sonderfall: Gleittarife in Cloud-Installationen
		$slidingTarif = $client->get("SlidingTarif");
		if ($slidingTarif !== null) {
			$result['slidingTarif'] = (bool)$slidingTarif;
		}

		return $result;
	}

	/**
	 * @param User $client
	 * @return array
	 */
	private static function makeSubUsersExportData(User $client): array
	{
		return array_map(function($subuser) {
			$subuserData = self::makeClientExportData($subuser);

			$subuserData['groups'] = array_map(function ($group) {
				return $group == "on" ?: NULL;
			}, $subuser->getGroupActive());
			$subuserData['domains'] = array_map(function ($domain) {
				return $domain == "on" ?: NULL;
			}, $subuser->getDomActive());

			return $subuserData;
		}, $client->getSubusers());
	}

	/**
	 * @param User $client
	 * @return array
	 */
	private static function makeSubUserGroupsExportData(User $client): array
	{
		SubGroups::switchRepositoryContext($client);

		return array_map(function($subGroup) {
			$subGroupRights = array_map(function ($permission) {
				return $permission == "on" ?: NULL;
			}, $subGroup->getGroupRights());

			return [
				'id' => $subGroup->getId(),
				'groupname' => $subGroup->getGroupname(),
				'isSubAdmin' => $subGroup->isSubAdmin(),
				'groupRights' => $subGroupRights,
			];
		}, SubGroups::all());
	}

	/**
	 * @param Domain $domain
	 * @return array
	 */
	private static function makeDomainExportData(Domain $domain): array
	{
		return [
			'id' => $domain->getId(),
			'name' => $domain->getName(),
			'urlPath' => $domain->getUrlPath(),
			'urlScheme' => $domain->getUrlScheme(),
			'url' => $domain->getUrl(),
			'country' => $domain->getCountryCode(),
			'timeZone' => $domain->getTimeZone() ? $domain->getTimeZone()->getName() : null,
			'whitelabel' => $domain->hasWhitelabel(),
			'frontendWidgetEnabled' => $domain->isFrontendWidgetEnabled(),
			'shareDomainNames' => $domain->getShareDomainNames(),
			'onBoarding' => $domain->isOnBoarding(),
			'iframesBlocked' => $domain->isIframesBlocked(),
			'rememberIframeConsentPerDomain' => $domain->isRememberIframeConsentPerDomain(),
			'appendConsentSwitchToIframes' => $domain->isAppendConsentSwitchToIframes(),
			'evalInlineScripts' => $domain->isEvalInlineScripts(),
			'scriptsTrack' => $domain->isScriptsTrack(),
			'newScriptsBlocked' => $domain->isNewScriptsBlocked(),
			'sameDomainScriptsBlocked' => $domain->isSameDomainScriptsBlocked(),
			'inlineScriptsBlocked' => $domain->isInlineScriptsBlocked(),
			'activatePermaScan' => $domain->isActivatePermaScan(),
			'allowedScriptMarkers' => $domain->getAllowedScriptMarkers(),
			'deniedScriptMarkers' => $domain->getDeniedScriptMarkers(),
			'iframeMarkers' => $domain->getIframeMarkers(),
			'iframeBlockMode' => $domain->getIframeBlockMode(),
			'frontendLegalEnabled' => $domain->isFrontendLegalEnabled(),
			'widgetCachingEnabled' => $domain->isWidgetCachingEnabled(),
			'ManagementStructurePreLegalConsent' => $domain->isManagementStructurePreLegalConsent(),
			'ManagementStructure' => $domain->getManagementStructure(),
			'deleteUnknownCookies' => $domain->getDeleteUnknownCookies(),
			'deleteUnknownCookiesForceReload' => $domain->getDeleteUnknownCookiesForceReload(),
			'loadTechnicallyNecessaryEmbeddingsWithoutConsent' => $domain->isLoadTechnicallyNecessaryEmbeddingsWithoutConsent(),
			'consentStorage' => $domain->getConsentStorage(),
			'consentStorageNamespaceUrlParameter' => $domain->getConsentStorageNamespaceUrlParameter(),
			'consentLifetime' => $domain->getConsentLifetime(),
			'consentLifetimeForPartialConsent' => $domain->getConsentLifetimeForPartialConsent(),
			'consentCookieLifetime' => $domain->getConsentCookieLifetime(),
			'crossDomainSharing' => $domain->hasCrossDomainSharing(),
			'consentCookieReset' => $domain->getconsentCookieReset(),
			'consentResetOnChange' => $domain->hasConsentResetOnRelevantChange(),
			'secureConsentCookie' => $domain->hasSecureConsentCookie(),
			'codeMinification' => $domain->hasCodeMinification(),
			'partOfAdAndTrackingFreeNetwork' => $domain->isPartOfAdAndTrackingFreeNetwork(),
			'consentTablePrependIntroText' => $domain->getConsentTablePrependIntroText(),
			'consentTablePrependConsentChangeLink' => $domain->getConsentTablePrependConsentChangeLink(),
			'consentTableHeadingStartLevel' => $domain->getConsentTableHeadingStartLevel(),
			'tcfVendorBlacklist' => array_map(function($vendor) {
					return (int)$vendor->getId();
				}, $domain->getTcfVendorBlacklist()),
			'tcfVendorsEnabled' => $domain->isTcfVendorsEnabled(),
			'newTcfVendorsActive' => $domain->isNewTcfVendorsActive(),
			'unusedTcfVendorIds' => $domain->getUnusedTcfVendorIds(),
			'explicitTcfSpecialFeatures' => array_map(function($specialFeature) {
				return (int)$specialFeature->getId();
			}, $domain->getExplicitTcfSpecialFeatures()),
			'googleAdditionalConsentMode' => $domain->hasGoogleAdditionalConsentMode(),
			'googleAdditionalConsentWhitelist' => array_map(function($adTechProvider) {
				return (int)$adTechProvider->getId();
			}, $domain->getGoogleAdditionalConsentWhitelist()),
			'tcfAdTagsEnabled' => $domain->isTcfAdTagsEnabled(),
			'tcfAdTagsAutoAppend' => $domain->hasTcfAdTagsAutoAppend(),
			'declineButtonEnforced' => $domain->isDeclineButtonEnforced(),
			'activatedPluginIds' => array_values($domain->getActivatedPluginIds()),
			'forcedCssImportant' => $domain->hasForcedCssImportant(),
			'lcpOptimizationEnabled' => $domain->isLcpOptimizationEnabled(),
			'onlyActiveOnMatchingDomain' => $domain->isOnlyActiveOnMatchingDomain(),
			'embeddingConsentViaJsApi' => $domain->isAllowEmbeddingConsentViaJsApi(),
			'triggeringExtGtmEvents' => $domain->isTriggeringExtGtmEvents(),
			'detectCloakedScripts' => $domain->isDetectCloakedScripts(),
			'prioritizeLanguageHeaderOverLangAttribute' => $domain->isPrioritizeLanguageHeaderOverLangAttribute(),
			'widgetDelayInMilliseconds' => $domain->getWidgetDelayInMilliseconds(),
			'widgetExclusionUrls' => $domain->getWidgetExclusionUrls(),
			'shareGroupName' => $domain->getShareGroupName(),
			'shareGroupUrl' => $domain->getShareGroupUrl(),
		];
	}

	/**
	 * @param Theme|SystemTheme|UserTheme $theme
	 * @return array
	 */
	private static function makeThemeExportData($theme): array
	{
		$themeData = [
			'id' => $theme->getId(),
			'isDefault' => $theme->isDefaultTheme(),

			'name' => $theme->getName(),
			'baseLayout' => $theme->getBaseLayout()->getId(),
			'position' => $theme->getPosition(),
			'selectType' => $theme->getSelectType(),
			'showPurposesInMainWindow' => $theme->isShowPurposesInMainWindow(),
			'blocking' => $theme->isBlocking(),
			'settingsIconEnabled' => $theme->isSettingsIconEnabled(),
			'settingsIconTarget' => $theme->getSettingsIconTarget(),
			'complyingDoNotTrack' => $theme->isComplyingDoNotTrack(),
			'onlyInEu' => $theme->getOnlyInEu(),
			'manipulationPrevention' => $theme->getManipulationPrevention(),

			'widgetButtons' => $theme->getWidgetButtons(),

			'showEmbeddingsInControlPanel' => $theme->isShowEmbeddingsInControlPanel(),
			'showAcceptAllButtonInControlPanel' => $theme->isShowAcceptAllButtonInControlPanel(),
			'showWithdrawAllButtonInControlPanel' => $theme->isShowWithdrawAllButtonInControlPanel(),
			'showCheckAndUncheckAllButtonInControlPanel' => $theme->isShowCheckAndUncheckAllButtonInControlPanel(),
			'showExtraImprintFooter' => $theme->isShowExtraImprintFooter(),
			'declineButtonRestrictions' => $theme->hasDeclineButtonRestrictions() ? $theme->getDeclineButtonRestrictions() : null,

			'saveButtonInDetailsWindow' => $theme->hasSaveButtonInDetailsWindow(),

			'primaryButtonBackgroundColor' => $theme->getOkButtonBackgroundColor(),
			'primaryButtonTextColor' => $theme->getOkButtonColor(),
			'primaryButtonFrameColor' => $theme->getOkButtonBorderColor(),
			'rejectButtonBackgroundColor' => $theme->getDeclineButtonBackgroundColor(),
			'rejectButtonTextColor' => $theme->getDeclineButtonTextColor(),
			'rejectButtonFrameColor' => $theme->getDeclineButtonBorderColor(),
			'otherButtonBackgroundColor' => $theme->getMoreInfoButtonBackgroundColor(),
			'otherButtonTextColor' => $theme->getMoreInfoButtonColor(),
			'otherButtonFrameColor' => $theme->getMoreInfoButtonBorderColor(),
			'windowsBackgroundColor' => $theme->getWindowBackgroundColor(),
			'windowsFontColor' => $theme->getInfotextColor(),
			'windowsFrameColor' => $theme->getWindowBorderColor(),
			'windowsLinkColor' => $theme->getLinkInInfotextColor(),
			'iFrameBackgroundColor' => $theme->getIframeBlockerBackgroundColor(),
			'iFrameFontColor' => $theme->getIframeBlockerForegroundColor(),
			'iFrameButtonBackgroundColor' => $theme->getIframeBlockerButtonBackgroundColor(),
			'iFrameButtonFontColor' => $theme->getIframeBlockerButtonForegroundColor(),

			'customCss' => $theme->getCustomCss(),
			'customCssTemplated' => $theme->isCustomCssTemplated(),
			'customIFrameCss' => $theme->getCustomCssForIframeBlocker(),
			'customIFrameCssTemplated' => $theme->isCustomCssForIframeBlockerTemplated(),
			'bannerHeightMobile' => $theme->getBannerHeightMobile(),
		];

		if ($theme instanceof Theme) {
			$themeData += [
				'logo' => $theme->getLogo() != NULL ? base64_encode(file_get_contents($theme->getLogo()->getRealPath())) : '',
				'logoDisplayed' => $theme->isLogoDisplayed(),
				'settingsIcon' => $theme->getSettingsIcon() != NULL ? base64_encode(file_get_contents($theme->getSettingsIcon()->getRealPath())) : '',
				'whitelabel' => $theme->getWhitelabel(),
			];

			if ($theme->getUserTheme() != NULL) {
				$themeData['id'] = $theme->getUserTheme()->getId();
			}
			if ($theme->getSystemTheme() != NULL) {
				$themeData['id'] = $theme->getSystemTheme()->getId();
			}
		}

		return $themeData;
	}

	/**
	 * @param Domain $domain
	 * @return array
	 */
	private static function makeCookieStatementTableExportData(Domain $domain): array
	{
		return [
			'prependIntroText' => $domain->getConsentTablePrependIntroText(),
			'prependConsentChangeLink' => $domain->getConsentTablePrependConsentChangeLink(),
			'headingStartLevel' => $domain->getConsentTableHeadingStartLevel(),
		];
	}

	/**
	 * @param LocaleBase $locale
	 * @return array
	 */
	private static function makeLocaleExportData(LocaleBase $locale): array
	{
		return [
			'enabled' => $locale->isEnabled(),
			'widgetTitle' => $locale->getWidgetTitle(),
			'widgetIntroduction' => $locale->getWidgetIntroduction(),
			'widgetConsentButtonText' => $locale->getWidgetConsentButtonText(),
			'widgetDeclineButtonText' => $locale->getWidgetDeclineButtonText(),
			'widgetSettingsButtonText' => $locale->getWidgetSettingsButtonText(),
			'controlPanelTitle' => $locale->getControlPanelTitle(),
			'controlPanelCancelButtonText' => $locale->getControlPanelCancelButtonText(),
			'controlPanelSaveButtonText' => $locale->getControlPanelSaveButtonText(),
			'imprintText' => $locale->getImprintText(),
			'privacyPolicyText' => $locale->getPrivacyPolicyText(),
			'accessibilityText' => $locale->getAccessibilityText(),
			'morePurposeDetailsButtonText' => $locale->getMorePurposeDetailsButtonText(),

			'categoriesAcceptAllButtonText' => $locale->getCategoriesAcceptAllButtonText(),
			'checkAllButtonText' => $locale->getCheckAllButtonText(),
			'uncheckAllButtonText' => $locale->getUncheckAllButtonText(),

			'statusActiveText' => $locale->getStatusActiveText(),
			'statusInactiveText' => $locale->getStatusInactiveText(),
			'statusPartialText' => $locale->getStatusPartialText(),

			'detailsTitle' => $locale->getDetailsTitle(),
			'detailsCloseButtonText' => $locale->getDetailsCloseButtonText(),
			'cookieVendorText' => $locale->getCookieVendorText(),
			'cookieDescriptionText' => $locale->getCookieDescriptionText(),
			'cookiePrivacyPolicyUrlText' => $locale->getCookiePrivacyPolicyUrlText(),
			'cookieLifetimeText' => $locale->getCookieLifetimeText(),
			'cookieCollectedDataInfoText' => $locale->getCookieCollectedDataInfoText(),
			'cookiePurposeOfDataCollectionText' => $locale->getCookiePurposeOfDataCollectionText(),
			'cookieLegalBasisText' => $locale->getCookieLegalBasisText(),
			'cookiePlaceOfProcessingText' => $locale->getCookiePlaceOfProcessingText(),

			'changeConsentLabel' => $locale->getChangeConsentLabel(),
			'assetNameLabel' => $locale->getAssetNameLabel(),
			'assetLifetimeLabel' => $locale->getAssetLifetimeLabel(),
			'assetDescriptionLabel' => $locale->getAssetDescriptionLabel(),

			'tcfWidgetIntroduction' => $locale->getTcfWidgetIntroduction(),
			'tcfStacksListText' => $locale->getTcfStacksListText(),
			'tcfVendorGrantedPermissionsText' => $locale->getTcfVendorGrantedPermissionsText(),
			'tcfPurposesLabel' => $locale->getTcfPurposesLabel(),
			'tcfSpecialPurposesLabel' => $locale->getTcfSpecialPurposesLabel(),
			'tcfFeaturesLabel' => $locale->getTcfFeaturesLabel(),
			'tcfSpecialFeaturesLabel' => $locale->getTcfSpecialFeaturesLabel(),
			'tcfOptedInRemark' => $locale->getTcfOptedInRemark(),
			'tcfThirdPartiesLabel' => $locale->getTcfThirdPartiesLabel(),
			'tcfConsentLabel' => $locale->getTcfConsentLabel(),
			'tcfLegitimateInterestLabel' => $locale->getTcfLegitimateInterestLabel(),
			'nonTcfPurposesLabel' => $locale->getNonTcfPurposesLabel(),
			'nonTcfVendorsLabel' => $locale->getNonTcfVendorsLabel(),
			'moreInformationLabel' => $locale->getMoreInformationLabel(),
			'tcfVendorUsesNonCookieStorageText' => $locale->getTcfVendorUsesNonCookieStorageText(),
			'tcfSpecialFeaturesHandledGloballyText' => $locale->getTcfSpecialFeaturesHandledGloballyText(),
			'tcfCookieRefreshText' => $locale->getTcfCookieRefreshText(),
			'tcfExtendedDisclosureText' => $locale->getTcfExtendedDisclosureText(),
			'tcfAssetTypeLabel' => $locale->getTcfAssetTypeLabel(),
			'tcfAssetPurposesLabel' => $locale->getTcfAssetPurposesLabel(),
			'tcfDataRetentionText' => $locale->getTcfDataRetentionText(),
			'tcfDataRetentionSessionText' => $locale->getTcfDataRetentionSessionText(),
			'tcfDataCategoriesLabel' => $locale->getTcfDataCategoriesLabel(),
			'tcfIllustrationsLabel' => $locale->getTcfIllustrationsLabel(),
			'tcfPurposeVendorListLabel' => $locale->getTcfPurposeVendorListLabel(),

			'blockedContentTitle' => $locale->getBlockedContentTitle(),
			'blockedContentText' => $locale->getBlockedContentText(),
			'blockedContentButtonText' => $locale->getBlockedContentButtonText(),
			'blockedContentAllowEmbedding' => $locale->getBlockedContentAllowEmbedding(),
			'blockedContentProviderDetails' => $locale->getBlockedContentProviderDetails(),
			'blockedContentShowExternalContent' => $locale->getBlockedContentShowExternalContent(),

			'withdrawAllText' => $locale->getWithdrawAllText(),
			'consentSharingText' => $locale->getConsentSharingText(),
			'consentNoSharingText' => $locale->getConsentNoSharingText(),

			'purposes' => array_map(function($purpose) use ($locale) {
				return [
					'name' => $locale->getPurposeName($purpose),
					'description' => $locale->getPurposeDescription($purpose),
				];
			}, Purpose::all()),
		];
	}

	/**
	 * @param Domain $domain
	 * @param LocaleBase[] $locales
	 * @return array
	 */
	private static function makeImprintExportData(Domain $domain, array $locales): array
	{
		LegalText::switchRepositoryContext($domain->getUser(), $domain);
		$data = [];

		foreach ($locales as $locale) {
			$legalText = LegalText::byLocale($locale);
			if (!$legalText) {
				continue;
			}

			$data[$locale->getName()] = [
				'imprintScheme' => $legalText->getImprintScheme(),
				'imprinText' => $legalText->getImprint(),
				'imprintLink' => $legalText->getImprintUrl(),
				'displayOnImprintPage' => $legalText->isShowWidgetOnImprintPage(),
				'targetBlank' => $legalText->istargetBlank()
			];
		}
		return $data;
	}

	/**
	 * @param Domain $domain
	 * @param LocaleBase[] $locales
	 * @return array
	 */
	private static function makeDataProtectionExportData(Domain $domain, array $locales): array
	{
		LegalText::switchRepositoryContext($domain->getUser(), $domain);
		$data = [];

		foreach ($locales as $locale) {
			$legalText = LegalText::byLocale($locale);
			if (!$legalText) {
				continue;
			}

			$data[$locale->getName()] = [
				'privacyScheme' => $legalText->getPrivacyScheme(),
				'privacyText' => $legalText->getPrivacyPolicy(),
				'privacyLink' => $legalText->getPrivacyUrl(),
				'displayOnPrivacyPage' => $legalText->isShowWidgetOnPrivacyPage(),
				'targetBlankPrivacy' => $legalText->istargetBlankPrivacy()
			];
		}
		return $data;
	}

	/**
	 * @param Domain $domain
	 * @param LocaleBase[] $locales
	 * @return array
	 */
	private static function makeAccessibilityExportData(Domain $domain, array $locales): array
	{
		LegalText::switchRepositoryContext($domain->getUser(), $domain);
		$data = [];

		foreach ($locales as $locale) {
			$legalText = LegalText::byLocale($locale);
			if (!$legalText) {
				continue;
			}

			$data[$locale->getName()] = [
				'accessibilityScheme' => $legalText->getAccessibilityScheme(),
				'accessibilityText' => $legalText->getAccessibility(),
				'accessibilityLink' => $legalText->getAccessibilityUrl(),
				'displayOnFrontend' => $legalText->isShowAccessibilityInFrontendWidget(),
			];
		}
		return $data;
	}

	/**
	 * @param Domain $domain
	 * @return array
	 */
	private static function makePlaceholderExportData(Domain $domain): array
	{
		Placeholder::switchRepositoryContext($domain->getUser(), $domain);

		return array_values(array_map(function (Placeholder $placeholder) {
			return [
				'placeholderName' => $placeholder->getName(),
				'placeholderValue' => $placeholder->getValue(),
			];
		}, Placeholder::all()));
	}

	/**
	 * @param EmbeddingBase[] $embeddings
	 * @param LocaleBase[] $locales
	 * @return array
	 */
	public static function buildEmbeddingExportData(array $embeddings, array $locales): array
	{
		return array_values(array_map(function (EmbeddingBase $embedding) use ($locales) {
			$translations = [];
			foreach ($locales as $locale) {
				$translations[$locale->getName()] = [
					'label' => $embedding->getTranslatedName($locale),
					'description' => $embedding->getTranslatedDescription($locale),
					'privacyPolicyUrl' => $embedding->getTranslatedPrivacyPolicyUrl($locale),
					'collectedDataInfo' => $embedding->getTranslatedCollectedDataInfo($locale),
					'purposeOfDataCollection' => $embedding->getTranslatedPurposeOfDataCollection($locale),
					'legalBasis' => $embedding->getTranslatedLegalBasis($locale),
					'placeOfProcessing' => $embedding->getTranslatedPlaceOfProcessing($locale),
					'blockedContentTitle' => $embedding->getTranslatedBlockedContentTitle($locale, false),
					'blockedContentText' => $embedding->getTranslatedBlockedContentText($locale, false),
					'blockedContentButtonText' => $embedding->getTranslatedBlockedContentButtonText($locale, false),
				];
			}

			$assets = array_values(array_map(function (EmbeddingAssetBase $embeddingsAsset) use ($locales) {
				$translations = [];
				foreach ($locales as $locale) {
					$translations[$locale->getName()] = [
						'lifetime' => $embeddingsAsset->getTranslatedLifetime($locale),
						'description' => $embeddingsAsset->getTranslatedDescription($locale),
					];
				}
				return [
					'name' => $embeddingsAsset->getName(),
					'isDynamic' => $embeddingsAsset->isDynamic(),
					'storageType' => $embeddingsAsset->getStorageType(),

					'translations' => $translations,
				];
 			}, $embedding->getAssets()));

			$sourceTimestamp = $embedding->getSourceTimestamp()
				? $embedding->getSourceTimestamp()->format('Y-m-d H:i:s')
				: null;
			$lastModified = $embedding->getLastModified()
				? $embedding->getLastModified()->format('Y-m-d H:i:s')
				: null;

			$tcfData = ($tcfVendor = $embedding->getTcfVendor())
				? [
					'vendor' => (int)$tcfVendor->getId(),
					'purposes' => array_values(array_map(function(TcfPurpose $purpose) {
						return (int)$purpose->getId();
					}, $embedding->getTcfPurposes())),
					'specialFeatures' => array_values(array_map(function(TcfSpecialFeature $specialFeature) {
						return (int)$specialFeature->getId();
					}, $embedding->getTcfSpecialFeatures())),
				]
				: null;

			return [
				'id' => $embedding->getId(),
				'name' => $embedding->getName(),
				'purpose' => $embedding->getPurpose() ? $embedding->getPurpose()->getId() : null,
				'isActive' => $embedding->isActive(),
				'updateParameterBlacklist' => $embedding->getUpdateParameterBlacklist(),
				'sourceCode' => $embedding->getCodeSnippet(),
				'provider' => $embedding->getEmbeddingVendor(),
				'blockedScripts' => $embedding->getCodeBlock(),
				'iframeBlockingStrings' => $embedding->getIframeBlockingStrings(),
				'hideIframesUntilConsent' => $embedding->isIframesHiddenUntilConsent(),
				'dispatchLoadEventForIframes' => $embedding->isDispatchLoadEventForIframes(),
				'excludeURLs' => $embedding->getExcludeUrls(),
				'excludeUrlsMode' => $embedding->getExcludeUrlsMode(),
				'scriptLoaderGroup' => $embedding->getScriptLoaderGroup(),

				'source' => $embedding->getSource(),
				'sourceId' => $embedding->getSourceId(),
				'autoUpdated' => $embedding->isAutoUpdated(),
				'sourceTimestamp' => $sourceTimestamp,
				'lastModified' => $lastModified,
				'structuredLegalBasis' => $embedding->getStructuredLegalBasis(),
				'legalBasisIsStructured' => $embedding->isLegalBasisStructured(),

				'translations' => $translations,

				'assets' => $assets,

				'tcf' => $tcfData,
			];

		}, $embeddings));
	}

	/**
	 * @param Domain $domain
	 * @param LocaleBase[] $locales
	 * @return array
	 */
	private static function makeEmbeddingExportData(Domain $domain, array $locales): array
	{
		Embedding::switchRepositoryContext($domain->getUser(), $domain);
		EmbeddingAsset::switchRepositoryContext($domain->getUser(), $domain);

		return self::buildEmbeddingExportData(Embedding::all(), $locales);
	}

	/**
	 * @param Domain $domain
	 * @param LocaleBase[] $locales
	 * @return array
	 */
	private static function makeCookieExportData(Domain $domain, array $locales): array
	{
		Cookie::switchRepositoryContext($domain->getUser(), $domain);

		return array_values(array_map(function (Cookie $cookie) use ($locales) {
			$translations = [];
			foreach ($locales as $locale) {
				$translations[$locale->getName()] = [
					'description' => $cookie->getTranslatedDescription($locale),
					'privacyPolicyUrl' => $cookie->getTranslatedPrivacyPolicyUrl($locale),
					'lifetime' => $cookie->getTranslatedLifetime($locale),
					'collectedDataInfo' => $cookie->getTranslatedCollectedDataInfo($locale),
					'purposeOfDataCollection' => $cookie->getTranslatedPurposeOfDataCollection($locale),
					'legalBasis' => $cookie->getTranslatedLegalBasis($locale),
					'placeOfProcessing' => $cookie->getTranslatedPlaceOfProcessing($locale),
				];
			}

			return [
				'name' => $cookie->getName(),
				'purpose' => $cookie->getPurpose() ? $cookie->getPurpose()->getId() : null,
				'isActive' => $cookie->isActive(),
				'httpCookieName' => $cookie->getHttpCookieName(),
				'sourceCode' => $cookie->getCodeSnippet(),
				'provider' => $cookie->getCookieVendor(),
				'blockedScripts' => $cookie->getCodeBlock(),

				'translations' => $translations,
			];
		}, Cookie::all()));
	}

	/**
	 * @param Domain $domain
	 * @return array
	 */
	private static function makeConsentExportData(Domain $domain): array
	{
		// TODO: Add archived Consents?
		return iterator_to_array(Protocol::byDomain($domain)->all(true));
	}
}
