<?php

namespace Plugins\Ccm19Imex\Component;

use DateTimeZone;
use Plugins\Ccm19Imex\Event\ImportDomainEvent;
use App\Component\LocaleService;
use App\Component\Tcf\v2\Model\Google\AdTechProvider;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Model\BaseLayout;
use App\Model\Cookie;
use App\Model\Country;
use App\Model\Domain;
use App\Model\Index\DomainIndex;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingBase;
use App\Model\LegalText;
use App\Model\Locale;
use App\Model\Language;
use App\Model\Model;
use App\Model\Placeholder;
use App\Model\Protocol;
use App\Model\Purpose;
use App\Model\SubGroups;
use App\Model\SystemTheme;
use App\Model\SystemEmbedding;
use App\Model\SystemEmbeddingAsset;
use App\Model\Theme;
use App\Model\User;
use App\Model\UserLocale;
use App\Model\UserTheme;
use App\Model\UserEmbedding;
use App\Model\UserEmbeddingAsset;
use App\Utils;
use DateTimeImmutable;
use LogicException;
use Exception;
use Symfony\Component\HttpFoundation\File\File;

/**
 * Class Import
 * @package Plugins\Ccm19Imex\Component
 */
class Import
{
	public function __construct(){ }

	/**
	 * Starting method for the import of a domain
	 *
	 * @param Domain $domain
	 * @param array $data data to import into the domain
	 * @param bool $isNewDomain
	 * @param array{systemTheme?:array,userTheme?:array,userEmbedding?:array,systemEmbedding?:array} $idMappings
	 * @throws Exception
	 */
	public static function importDomain(Domain $domain, array $data, bool $isNewDomain=false, array $idMappings=[])
	{
		$user = $domain->getUser();
		assert($user instanceof User);
		$systemThemeMap = $idMappings['systemTheme'] ?? [];
		$userThemeMap = $idMappings['userTheme'] ?? [];

		$activeDomain = Domain::activeDomain();
		if (!$activeDomain or $activeDomain->getId() !== $domain->getId()) {
			Model::closeDomainModels();
		}
		$lastActiveDomainId = Domain::getDomainIdForRequest();
		Domain::setDomainForRequest($domain->getId());

		// Domain
		if (isset($data['domainData']) && $data['domainData']) {
			try {
				self::makeDomainImport($domain, $data['domainData']);
			}
			catch (Exception $exception) {
				// Revert changes to the domain if an error occurs
				Domain::reload();

				if ($isNewDomain) {
					Domain::delete($domain->getId());
				}

				Domain::setDomainForRequest($lastActiveDomainId);
				Model::closeDomainModels();

				throw $exception;
			}
			$user->updateDomainInfo();
		}

		Locale::switchRepositoryContext($domain->getUser(), $domain);

		// Themes
		if (isset($data['themes']) && $data['themes']) {
			$themeMap = self::makeThemesImport('standard', $data['themes'], $domain, $isNewDomain, $systemThemeMap, $userThemeMap);
		}
		// Cookies Statement Table
		if (isset($data['cookieStatementTable']) && $data['cookieStatementTable']) {
			self::makeCookieStatementTableImport($domain, $data['cookieStatementTable']);
		}
		// Texts & Language
		if (isset($data['locales']) && $data['locales']) {
			self::makeLocalesImport('domain', $data['locales']);
		}
		// Imprint
		if (isset($data['imprint']) && $data['imprint']) {
			self::makeImprintImport($data['imprint']);
		}
		// Data protection
		if (isset($data['dataProtection']) && $data['dataProtection']) {
			self::makeDataProtectionImport($data['dataProtection']);
		}
		// Accessibility
		if (isset($data['accessibility']) && $data['accessibility']) {
			self::makeAccessibilityImport($data['accessibility']);
		}
		// Placeholders
		if (isset($data['placeholders']) && $data['placeholders']) {
			self::makePlaceholderImport($data['placeholders'], $isNewDomain);
		}

		$idMappings['theme'] = $themeMap ?? [];

		// Embeddings
		if (isset($data['embeddings']) && $data['embeddings']) {
			$embeddingMap = self::makeEmbeddingsImport('domain', $data['embeddings'], $idMappings);

			// Update TcF-Stacks for newly added embeddings if they are active
			$domain->updateTcfStacks();
		}
		// Cookies (obsolete)
		if (isset($data['cookies']) && $data['cookies']) {
			self::makeCookieImport($data['cookies']);
		}
		// Consents
		if (isset($data['consents']) && $data['consents']) {
			self::makeConsentImport($domain, $data['consents']);
		}

		$idMappings['embedding'] = $embeddingMap ?? [];

		$event = new ImportDomainEvent($domain, $data, $isNewDomain, $idMappings);
		Utils::dispatchEvent(ImportDomainEvent::NAME, $event);

		Domain::setDomainForRequest($lastActiveDomainId);
		Model::closeDomainModels();
	}

	/**
	 * @param User $client
	 * @param array $clientData
	 */
	public static function makeClientImport(User $client, array $clientData): void
	{
		$client
			->setCreateTimeStamp($clientData['createTimestamp'] ?? time())
			->setLastLogin($clientData['lastLogin'] ?? 0)
			->setPasswordHash($clientData['passwordHash'])
			->setPasswordLocked($clientData['passwordLocked'] ?? false)
			->setEmailAddress($clientData['emailAddress'] ?? '')
			->setInvoiceEmailAddress($clientData['invoiceEmailAddress'] ?? null)
			->setDunningEmailAddress($clientData['dunningEmailAddress'] ?? null)
			->setFirstName($clientData['firstName'] ?? '')
			->setLastName($clientData['lastName'] ?? '')
			->setCompany($clientData['company'] ?? '')
			->setMaxDomainCount($clientData['maxDomainCount'] ?? 0)
			->setRole($clientData['role'] ?? 'client')
			->setMaxWLCount($clientData['maxWLCount'] ?? 0)
			->setMaxCallCount($clientData['MaxCallCount'] ?? 0)
			->setAutoAddDomains($clientData['autoAddDomains'] ?? false)
			->setVatId($clientData['vatId'] ?? '')
			->setStrNr($clientData['strNr'] ?? '')
			->setPlz($clientData['plz'] ?? '')
			->setCity($clientData['city'] ?? '')
			->setActive($clientData['active'] ?? true)
			->setSettings($clientData['settings'] ?? [
				"moreThan2Lang" => false,
				"canUseAnalytics" => false,
				"canUseABTests" => false,
				"canUseScripts" => false,
				"canUseIframes" => false,
				"canUseConsentSharing" => false,
				"canUseOwnLogo" => false,
				"canUseIndCSS" => false,
				"canUseIframeCSS" => false,
				"canUseMultiDom" => false,
				"canUseIAB" => false,
				"canUseUserGroupRights" => false,
				"canUseDomAPI" => false,
				"canUseExternalCrawler" => false,
			])
			->setOnBoarding($clientData['onboarding'] ?? true)
			->setDefaultLocale(UserLocale::find($clientData['defaultLocale'] ?? 'de_DE'))
			->setMailingPreferences($clientData['mailPref'] ?? [])
			->setCountry(Country::find($clientData['country'] ?? 'DE'))
			->setPreferredColorScheme($clientData['perferredColorScheme'] ?? null)
			->setNewDomainOnlyActiveOnMatchingDomain((bool)($clientData['newDomainOnlyActiveOnMatchingDomain'] ?? false))
			->setAllowAutoUpdateOfEmbeddings((bool)($clientData['allowAutoUpdateOfEmbeddings'] ?? true))
			->setSendFullConsentAndRejectionAcrossDomains((bool)($clientData['sendFullConsentAndRejectionAcrossDomains'] ?? false))
		;

		try {
			if ($timeZone = new DateTimeZone((string)($clientData['defaultTimeZone'] ?? ''))) {
				$client->setDefaultTimeZone($timeZone);
			}
		} catch (Exception $exception) {}

		if (!empty($clientData['2fa'])) {
			$client->setAllTwoFactorAuthParameters($clientData['2fa']);
		}

		if (!empty($clientData['externalMetadata'])) {
			$client->setExternalMetadata($clientData['externalMetadata']);
		}

		// Sonderfall: Gleittarife in Cloud-Installationen
		if (isset($clientData['slidingTarif'])) {
			$client->set("SlidingTarif", (bool)$clientData['slidingTarif']);
		}

		if (isset($clientData['image'])) {
			if ($clientData['image'] === null) {
				$client->setUserImage(null);
			} else {
				$temp_file = tempnam(sys_get_temp_dir(), 'avatar');
				try {
					file_put_contents($temp_file, base64_decode($clientData['image']));
					$file = new File($temp_file);
					$client->setUserImage($file)->save();
				} finally {
					@unlink($temp_file);
				}
			}
		}

		$client->updateDomainInfo();
		User::flush();
	}

	/**
	 * @param array $subGroupData
	 * @return SubGroups
	 */
	public static function importSubUserGroup(array $subGroupData): SubGroups
	{
		$subGroupRights = array_map(function ($rights) {
			return (bool)$rights;
		}, $subGroupData['groupRights']);

		return SubGroups::create()
			->setGroupname($subGroupData['groupname'])
			->setSubAdmin($subGroupData['isSubAdmin'] ?? false)
			->setGroupRights($subGroupRights)
		;
	}

	/**
	 * @param Domain $domain
	 * @param array $domainData
	 * @throws Exception
	 */
	private static function makeDomainImport(Domain $domain, array $domainData): void
	{
		$user = $domain->getUser();

		$maxWhitelabelCount = $user->getMaxWLCount();
		$availableWhiteLabelCount =  $maxWhitelabelCount >= 0
			? max(0, $maxWhitelabelCount - Domain::whitelabelCountForAllDomains())
			: -1;

		$shouldWhitelabel = (bool)($domainData['whitelabel'] ?? false);
		if ($shouldWhitelabel && $domain->hasWhitelabel() == false && $availableWhiteLabelCount == 0) {
			throw new Exception('Whitelabel limit reached');
		}

		$shareDomains = (array)($domainData['shareDomainNames'] ?? []);
		$addedShareDomainCount = count(array_diff($shareDomains, $domain->getShareDomainNames()));
		$availableDomainCount = $user->getAvailableDomainCount();

		if ($availableDomainCount > -1 && $addedShareDomainCount > $availableDomainCount) {
			throw new Exception('Domain limit reached');
		}

		$shareDomains = array_unique(array_filter(array_map('trim', array_merge($shareDomains, $domain->getShareDomainNames()))));

		try {
			$timeZone = new DateTimeZone((string)($domainData['timeZone'] ?? ''));
		} catch (Exception $exception) {
			$timeZone = null;
		}

		$widgetExclusionUrls = array_values(array_filter(
			array_map('trim', (array)($domainData['widgetExclusionUrls'] ?? [])),
			function ($line) { return $line != ''; }
		));

		$domain
			->setUrlPath($domainData['urlPath'] ?? '')
			->setUrlScheme($domainData['urlScheme'] ?? '')
			->setCountry(Country::find($domainData['country'] ?? 'DE'))
			->setTimeZone($timeZone)
			->setWhitelabel($shouldWhitelabel)
			->setFrontendWidgetEnabled((bool)$domainData['frontendWidgetEnabled'] ?? false)
			->setShareDomainNames($shareDomains)
			->setOnBoarding((bool)$domainData['onBoarding'] ?? false)
			->setIframesBlocked((bool)$domainData['iframesBlocked'] ?? false)
			->setRememberIframeConsentPerDomain((bool)$domainData['rememberIframeConsentPerDomain'] ?? false)
			->setAppendConsentSwitchToIframes((bool)($domainData['appendConsentSwitchToIframes'] ?? false))
			->setEvalInlineScripts((bool)($domainData['evalInlineScripts'] ?? false))
			->setScriptsTrack((bool)($domainData['scriptsTrack'] ?? false))
			->setNewScriptsBlocked((bool)$domainData['newScriptsBlocked'] ?? false)
			->setSameDomainScriptsBlocked((bool)$domainData['sameDomainScriptsBlocked'] ?? false)
			->setInlineScriptsBlocked((bool)$domainData['inlineScriptsBlocked'] ?? false)
			->setActivatePermaScan((bool)($domainData['activatePermaScan'] ?? false))
			->setAllowedScriptMarkers($domainData['allowedScriptMarkers'] ?? [])
			->setDeniedScriptMarkers($domainData['deniedScriptMarkers'] ?? [])
			->setIframeMarkers($domainData['iframeMarkers'] ?? [])
			->setIframeBlockMode($domainData['iframeBlockMode'] ?? '')
			->setFrontendLegalEnabled((bool)$domainData['frontendLegalEnabled'] ?? false)
			->setWidgetCachingEnabled((bool)($domainData['widgetCachingEnabled'] ?? false))
			->setManagementStructurePreLegalConsent((bool)$domainData['ManagementStructurePreLegalConsent'] ?? false)
			->setManagementStructure($domainData['ManagementStructure'] ?? 'embedding')
			->setDeleteUnknownCookies((bool)$domainData['deleteUnknownCookies'] ?? false)
			->setDeleteUnknownCookiesForceReload((bool)$domainData['deleteUnknownCookiesForceReload'] ?? false)
			->setLoadTechnicallyNecessaryEmbeddingsWithoutConsent((bool)($domainData['loadTechnicallyNecessaryEmbeddingsWithoutConsent'] ?? false))
			->setConsentStorage($domainData['consentStorage'] ?? '')
			->setConsentStorageNamespaceUrlParameter((string)($domainData['consentStorageNamespaceUrlParameter'] ?? ''))
			->setConsentLifetime($domainData['consentLifetime'] ?? 0)
			->setConsentLifetimeForPartialConsent($domainData['consentLifetimeForPartialConsent'] ?? null)
			->setConsentCookieLifetime($domainData['consentCookieLifetime'] ?? 365)
			->setCrossDomainSharing((bool)($domainData['crossDomainSharing']))
			->setConsentCookieReset($domainData['consentCookieReset'] ?? '')
			->setConsentResetOnRelevantChange((bool)($domainData['consentResetOnChange'] ?? false))
			->setSecureConsentCookie((bool)$domainData['secureConsentCookie'] ?? false)
			->setCodeMinification((bool)$domainData['codeMinification'] ?? false)
			->setPartOfAdAndTrackingFreeNetwork((bool)($domainData['partOfAdAndTrackingFreeNetwork'] ?? false))
			->setConsentTablePrependIntroText((bool)($domainData['consentTablePrependIntroText'] ?? false))
			->setConsentTablePrependConsentChangeLink((bool)($domainData['consentTablePrependConsentChangeLink'] ?? false))
			->setConsentTableHeadingStartLevel((int)($domainData['consentTableHeadingStartLevel'] ?? 3))
			->setTcfVendorBlacklist(array_filter(array_map(function ($vendorId) {
				return TcfVendor::find((string)$vendorId);
			}, $domainData['tcfVendorBlacklist'])))
			->setTcfVendorsEnabled((bool)$domainData['tcfVendorsEnabled'] ?? false)
			->setNewTcfVendorsActive((bool)$domainData['newTcfVendorsActive'] ?? false)
			->setUnusedTcfVendorIds($domainData['unusedTcfVendorIds'] ?? [])
			->setExplicitTcfSpecialFeatures(array_filter(array_map(function ($specialFeatureId) {
				return TcfSpecialFeature::find((string)$specialFeatureId);
			}, $domainData['explicitTcfSpecialFeatures'])))
			->setGoogleAdditionalConsentMode((bool)$domainData['googleAdditionalConsentMode'] ?? false)
			->setGoogleAdditionalConsentWhitelist(array_filter(array_map(function ($adTechProviderId) {
				return AdTechProvider::find((string)$adTechProviderId);
			}, $domainData['googleAdditionalConsentWhitelist'])))
			->setTcfAdTagsEnabled((bool)$domainData['tcfAdTagsEnabled'] ?? false)
			->setTcfAdTagsAutoAppend((bool)$domainData['tcfAdTagsAutoAppend'] ?? false)
			->setDeclineButtonEnforced((bool)($domainData['declineButtonEnforced'] ?? true))
			->setForcedCssImportant((bool)($domainData['forcedCssImportant'] ?? false))
			->setLcpOptimizationEnabled((bool)($domainData['lcpOptimizationEnabled'] ?? false))
			->setOnlyActiveOnMatchingDomain((bool)($domainData['onlyActiveOnMatchingDomain'] ?? false))
			->setAllowEmbeddingConsentViaJsApi((bool)($domainData['embeddingConsentViaJsApi'] ?? false))
			->setTriggeringExtGtmEvents((bool)($domainData['triggeringExtGtmEvents'] ?? true))
			->setDetectCloakedScripts((bool)($domainData['detectCloakedScripts'] ?? false))
			->setPrioritizeLanguageHeaderOverLangAttribute((bool)($domainData['prioritizeLanguageHeaderOverLangAttribute'] ?? false))
			->setWidgetDelayInMilliseconds((int)($domainData['widgetDelayInMilliseconds'] ?? 0))
			->setWidgetExclusionUrls($widgetExclusionUrls)
			->setShareGroupName((string)($domainData['shareGroupName'] ?? ''))
			->setShareGroupUrl((string)($domainData['shareGroupUrl'] ?? ''))
		;

		if ($domainData['activatedPluginIds'] ?? null) {
			foreach ($domainData['activatedPluginIds'] as $pluginId) {
				$domain->activatePlugin($pluginId);
			}
		}

		Domain::flush();

		DomainIndex::switchRepositoryContext($user);
		DomainIndex::rebuild();
	}

	/**
	 * @param string $themeType
	 * @param array $themeData
	 * @param Domain|null $domain
	 * @param bool $isNewDomain
	 * @param array $systemThemeMap
	 * @param array $userThemeMap
	 * @return array
	 */
	public static function makeThemesImport(string $themeType, array $themeData, ?Domain $domain, bool $isNewDomain = false, array $systemThemeMap=[], array $userThemeMap=[])
	{
		if ($isNewDomain) {
			// Not doing this would result in the creation of extra Themes in the import if no Themes are present (e.g. new Domain)
			if ($themeType == 'standard') {
				Theme::beginWrite();
				Theme::clear(false);
			}
		}

		$themeMap = [];

		foreach ($themeData as $data) {
			$theme = ($themeType == 'system') ? SystemTheme::create() : (($themeType == 'user') ? UserTheme::create() : Theme::create());

			if ($themeType == 'standard') {
				if (isset($systemThemeMap[$data['id']]) and $systemTheme = SystemTheme::find($systemThemeMap[$data['id']])) {
					$theme
						->setSystemTheme($systemTheme)
						->setSettingsIconBackgroundColor($systemTheme->getPublishedSettingsIconBackgroundColor() ?? $systemTheme->getSettingsIconBackgroundColor())
						->setSettingsIconForegroundColor($systemTheme->getPublishedSettingsIconForegroundColor() ?? $systemTheme->getSettingsIconForegroundColor())
					;
				}
				elseif (isset($userThemeMap[$data['id']]) and $userTheme = UserTheme::find($userThemeMap[$data['id']])) {
					$theme
						->setUserTheme($userTheme)
						->setSettingsIconBackgroundColor($userTheme->getPublishedSettingsIconBackgroundColor() ?? $userTheme->getSettingsIconBackgroundColor())
						->setSettingsIconForegroundColor($userTheme->getPublishedSettingsIconForegroundColor() ?? $userTheme->getSettingsIconForegroundColor())
					;
				}
			}

			$theme
				->setName($data['name'] ?? '')
				->setBaseLayout(BaseLayout::find((string)$data['baseLayout']) ?? BaseLayout::default())
				->setPosition($data['position'] ?? 'center')
				->setSelectType($data['selectType'] ?? 'checkbox')
				->setShowPurposesInMainWindow((bool)($data['showPurposesInMainWindow'] ?? false))
				->setBlocking((bool)($data['blocking'] ?? false))
				->setSettingsIconEnabled((bool)($data['settingsIconEnabled'] ?? false))
				->setSettingsIconTarget((string)($data['settingsIconTarget'] ?? 'main'))
				->setComplyingDoNotTrack((bool)($data['complyingDoNotTrack'] ?? false))
				->setOnlyInEu((bool)($data['onlyInEu'] ?? false))
				->setManipulationPrevention((bool)($data['manipulationPrevention'] ?? false))

				->setWidgetButtons(($data['widgetButtons'] ?? []))

				->setShowEmbeddingsInControlPanel((bool)($data['showEmbeddingsInControlPanel'] ?? false))
				->setShowAcceptAllButtonInControlPanel((bool)($data['showAcceptAllButtonInControlPanel'] ?? false))
				->setShowWithdrawAllButtonInControlPanel((bool)($data['showWithdrawAllButtonInControlPanel'] ?? false))
				->setShowCheckAndUncheckAllButtonInControlPanel((bool)($data['showCheckAndUncheckAllButtonInControlPanel'] ?? false))
				->setShowExtraImprintFooter((bool)($data['showExtraImprintFooter'] ?? false))
				->setDeclineButtonRestrictions((array)($data['declineButtonRestrictions'] ?? []))

				->setSaveButtonInDetailsWindow((bool)($data['saveButtonInDetailsWindow'] ?? false))

				->setOkButtonBackgroundColor($data['primaryButtonBackgroundColor'] ?? '')
				->setOkButtonColor($data['primaryButtonTextColor'] ?? '')
				->setOkButtonBorderColor($data['primaryButtonFrameColor'] ?? '')
				->setDeclineButtonBackgroundColor($data['rejectButtonBackgroundColor'] ?? '')
				->setDeclineButtonTextColor($data['rejectButtonTextColor'] ?? '')
				->setDeclineButtonBorderColor($data['rejectButtonFrameColor'] ?? '')
				->setMoreInfoButtonBackgroundColor($data['otherButtonBackgroundColor'] ?? '')
				->setMoreInfoButtonColor($data['otherButtonTextColor'] ?? '')
				->setMoreInfoButtonBorderColor($data['otherButtonFrameColor'] ?? '')
				->setWindowBackgroundColor($data['windowsBackgroundColor'] ?? '')
				->setInfotextColor($data['windowsFontColor'] ?? '')
				->setWindowBorderColor($data['windowsFrameColor'] ?? '')
				->setLinkInInfotextColor($data['windowsLinkColor'] ?? '')
				->setIframeBlockerBackgroundColor($data['iFrameBackgroundColor'] ?? '')
				->setIframeBlockerForegroundColor($data['iFrameFontColor'] ?? '')
				->setIframeBlockerButtonBackgroundColor($data['iFrameButtonBackgroundColor'] ?? '')
				->setIframeBlockerButtonForegroundColor($data['iFrameButtonFontColor'] ?? '')

				->setCustomCss($data['customCss'] ?? '')
				->setCustomCssTemplated((bool)($data['customCssTemplated'] ?? false))
				->setCustomCssForIframeBlocker($data['customIFrameCss'] ?? '')
				->setCustomCssForIframeBlockerTemplated((bool)($data['customIFrameCssTemplated'] ?? false))
				->setBannerHeightMobile((int)($data['bannerHeightMobile'] ?? 100))
			;

			if ($themeType == 'standard') {
				$theme
					->setLogoDisplayed((bool)($data['logoDisplayed'] ?? false))
					->setWhitelabel($domain->hasWhitelabel() && ($data['whitelabel'] ?? false))
				;

				if (isset($data['logo']) && $data['logo']) {
					$destination = Utils::getVarDir();
					Utils::assertDirectoryWritable($destination);
					if (file_put_contents($destination . '/logo', base64_decode($data['logo'])) != false) {
						$theme->setLogo(new File($destination . '/logo'));
						unlink($destination . '/logo');
					}
				}

				if (isset($data['settingsIcon']) && $data['settingsIcon']) {
					$destination = Utils::getVarDir();
					Utils::assertDirectoryWritable($destination);
					if (file_put_contents($destination . '/settingsicon', base64_decode($data['settingsIcon'])) != false) {
						$theme->setSettingsIcon(new File($destination . '/settingsicon'));
						unlink($destination . '/settingsicon');
					}
				}
			}
			else {
				$themeMap[$data['id']] = $theme->getId();
			}

			if (isset($data['isDefault']) && $data['isDefault']) {
				$theme->makeDefault();
			}
		}

		if ($themeType == 'standard') {
			Theme::flush();
		}
		elseif ($themeType == 'user') {
			UserTheme::flush();
		}
		elseif ($themeType == 'system') {
			SystemTheme::flush();
		}
		return $themeMap;
	}

	/**
	 * @param Domain $domain
	 * @param array $cookieStatementTableData
	 */
	private static function makeCookieStatementTableImport(Domain $domain, array $cookieStatementTableData): void
	{
		$domain
			->setConsentTablePrependIntroText((bool)($cookieStatementTableData['prependIntroText'] ?? false))
			->setConsentTablePrependConsentChangeLink((bool)($cookieStatementTableData['prependConsentChangeLink'] ?? false))
			->setConsentTableHeadingStartLevel((int)($cookieStatementTableData['headingStartLevel'] ?? 1))
			->save();
	}

	/**
	 * @param string $type 'domain'|'user'
	 * @param array $localeData
	 */
	public static function makeLocalesImport(string $type, array $localeData): void
	{
		foreach ($localeData as $localeName => $data) {
			if ($type == 'user') {
				try {
					$locale = UserLocale::find($localeName) ?? UserLocale::create($localeName);
				} catch (LogicException $ignore) {
					// Standard-Locales mit ungültiger Sprache ignorieren
					continue;
				}
			} else {
				$locale = Locale::find($localeName) ?? Locale::create($localeName);
			}

			$locale
				->setEnabled($data['enabled'] ?? false)
				->setWidgetTitle($data['widgetTitle'] ?? '')
				->setWidgetIntroduction($data['widgetIntroduction'] ?? '')
				->setWidgetConsentButtonText($data['widgetConsentButtonText'] ?? '')
				->setWidgetDeclineButtonText($data['widgetDeclineButtonText'] ?? '')
				->setWidgetSettingsButtonText($data['widgetSettingsButtonText'] ?? '')
				->setControlPanelTitle($data['controlPanelTitle'] ?? '')
				->setControlPanelCancelButtonText($data['controlPanelCancelButtonText'] ?? '')
				->setControlPanelSaveButtonText($data['controlPanelSaveButtonText'] ?? '')
				->setImprintText($data['imprintText'] ?? '')
				->setPrivacyPolicyText($data['privacyPolicyText'] ?? '')
				->setAccessibilityText($data['accessibilityText'] ?? '')
				->setMorePurposeDetailsButtonText($data['morePurposeDetailsButtonText'] ?? '')
				->setCategoriesAcceptAllButtonText($data['categoriesAcceptAllButtonText'] ?? '')
				->setCheckAllButtonText($data['checkAllButtonText'] ?? '')
				->setUncheckAllButtonText($data['uncheckAllButtonText'] ?? '')
				->setStatusActiveText($data['statusActiveText'] ?? '')
				->setStatusInactiveText($data['statusInactiveText'] ?? '')
				->setStatusPartialText($data['statusPartialText'] ?? '')
				->setDetailsTitle($data['detailsTitle'] ?? '')
				->setDetailsCloseButtonText($data['detailsCloseButtonText'] ?? '')
				->setCookieVendorText($data['cookieVendorText'] ?? '')
				->setCookieDescriptionText($data['cookieDescriptionText'] ?? '')
				->setCookiePrivacyPolicyUrlText($data['cookiePrivacyPolicyUrlText'] ?? '')
				->setCookieLifetimeText($data['cookieLifetimeText'] ?? '')
				->setCookieCollectedDataInfoText($data['cookieCollectedDataInfoText'] ?? '')
				->setCookiePurposeOfDataCollectionText($data['cookiePurposeOfDataCollectionText'] ?? '')
				->setCookieLegalBasisText($data['cookieLegalBasisText'] ?? '')
				->setCookiePlaceOfProcessingText($data['cookiePlaceOfProcessingText'] ?? '')
				->setChangeConsentLabel($data['changeConsentLabel'] ?? '')
				->setAssetNameLabel($data['assetNameLabel'] ?? '')
				->setAssetLifetimeLabel($data['assetLifetimeLabel'] ?? '')
				->setAssetDescriptionLabel($data['assetDescriptionLabel'] ?? '')
				->setTcfWidgetIntroduction($data['tcfWidgetIntroduction'] ?? '')
				->setTcfStacksListText($data['tcfStacksListText'] ?? '')
				->setTcfVendorGrantedPermissionsText($data['tcfVendorGrantedPermissionsText'] ?? '')
				->setTcfPurposesLabel($data['tcfPurposesLabel'] ?? '')
				->setTcfSpecialPurposesLabel($data['tcfSpecialPurposesLabel'] ?? '')
				->setTcfFeaturesLabel($data['tcfFeaturesLabel'] ?? '')
				->setTcfSpecialFeaturesLabel($data['tcfSpecialFeaturesLabel'] ?? '')
				->setTcfOptedInRemark($data['tcfOptedInRemark'] ?? '')
				->setTcfThirdPartiesLabel($data['tcfThirdPartiesLabel'] ?? '')
				->setTcfConsentLabel($data['tcfConsentLabel'] ?? '')
				->setTcfLegitimateInterestLabel($data['tcfLegitimateInterestLabel'] ?? '')
				->setNonTcfPurposesLabel($data['nonTcfPurposesLabel'] ?? '')
				->setNonTcfVendorsLabel($data['nonTcfVendorsLabel'] ?? '')
				->setMoreInformationLabel($data['moreInformationLabel'] ?? '')
				->setTcfVendorUsesNonCookieStorageText($data['tcfVendorUsesNonCookieStorageText'] ?? '')
				->setTcfSpecialFeaturesHandledGloballyText($data['tcfSpecialFeaturesHandledGloballyText'] ?? '')
				->setTcfCookieRefreshText($data['tcfCookieRefreshText'] ?? '')
				->setTcfExtendedDisclosureText($data['tcfExtendedDisclosureText'] ?? '')
				->setTcfAssetTypeLabel($data['tcfAssetTypeLabel'] ?? '')
				->setTcfAssetPurposesLabel($data['tcfAssetPurposesLabel'] ?? '')
				->setTcfDataRetentionText($data['tcfDataRetentionText'] ?? '')
				->setTcfDataRetentionSessionText($data['tcfDataRetentionSessionText'] ?? '')
				->setTcfDataCategoriesLabel($data['tcfDataCategoriesLabel'] ?? '')
				->setTcfIllustrationsLabel($data['tcfIllustrationsLabel'] ?? '')
				->setTcfPurposeVendorListLabel($data['tcfPurposeVendorListLabel'] ?? '')
				->setBlockedContentTitle($data['blockedContentTitle'] ?? '')
				->setBlockedContentText($data['blockedContentText'] ?? '')
				->setBlockedContentButtonText($data['blockedContentButtonText'] ?? '')
				->setBlockedContentAllowEmbedding($data['blockedContentAllowEmbedding'] ?? '')
				->setBlockedContentProviderDetails($data['blockedContentProviderDetails'] ?? '')
				->setBlockedContentShowExternalContent($data['blockedContentShowExternalContent'] ?? '')
				->setWithdrawAllText($data['withdrawAllText'] ?? '')
				->setConsentSharingText($data['consentSharingText'] ?? '')
				->setConsentNoSharingText($data['consentNoSharingText'] ?? '')
			;

			foreach ($data['purposes'] as $purposeId => $purposeData) {
				$purpose = Purpose::find((string)$purposeId);
				if (!$purpose) {
					continue;
				}

				$locale
					->setPurposeName($purpose, $purposeData['name'] ?? '')
					->setPurposeDescription($purpose, $purposeData['description'] ?? '')
				;
			}
		}

		if ($type != 'user') {
			Locale::flush();
		}
		UserLocale::flush();
	}

	/**
	 * @param array $imprintData
	 */
	private static function makeImprintImport(array $imprintData): void
	{
		foreach ($imprintData as $localeName => $data) {
			$locale = LocaleService::find((string)$localeName);

			if (!$locale) {
				continue;
			}

			$legalText = LegalText::byLocale($locale) ?? LegalText::create()->setLocale($locale);

			$legalText
				->setImprintScheme($data['imprintScheme'] ?? '')
				->setImprint($data['imprinText'] ?? '')
				->setImprintUrl($data['imprintLink'] ?? '')
				->setShowWidgetOnImprintPage((bool)($data['displayOnImprintPage'] ?? false))
				->settargetBlank((bool)($data['targetBlank'] ?? false))
			;
		}
		LegalText::flush();
	}

	/**
	 * @param array $dataProtectionData
	 */
	private static function makeDataProtectionImport(array $dataProtectionData): void
	{
		foreach ($dataProtectionData as $localeName => $data) {
			$locale = LocaleService::find((string)$localeName);
			if (!$locale) {
				continue;
			}

			$legalText = LegalText::byLocale($locale) ?? LegalText::create()->setLocale($locale);

			$legalText
				->setPrivacyScheme($data['privacyScheme'] ?? '')
				->setPrivacyPolicy($data['privacyText'] ?? '')
				->setPrivacyUrl($data['privacyLink'] ?? '')
				->setShowWidgetOnPrivacyPage((bool)($data['displayOnPrivacyPage'] ?? false))
				->settargetBlankPrivacy((bool)($data['targetBlankPrivacy'] ?? false))
			;
		}
		LegalText::flush();
	}

	/**
	 * @param array $accessibilityData
	 */
	private static function makeAccessibilityImport(array $accessibilityData): void
	{
		foreach ($accessibilityData as $localeName => $data) {
			$locale = LocaleService::find((string)$localeName);
			if (!$locale) {
				continue;
			}

			$legalText = LegalText::byLocale($locale) ?? LegalText::create()->setLocale($locale);

			$legalText
				->setAccessibilityScheme($data['accessibilityScheme'] ?? '')
				->setAccessibility($data['accessibilityText'] ?? '')
				->setAccessibilityUrl($data['accessibilityLink'] ?? '')
				->setShowAccessibilityInFrontendWidget((bool)($data['displayOnFrontend'] ?? false))
			;
		}
		LegalText::flush();
	}

	/**
	 * @param array $placeholderData
	 */
	private static function makePlaceholderImport(array $placeholderData, bool $deleteOld = false): void
	{
		$prevExisting = Placeholder::all();

		foreach ($placeholderData as $data) {
			$placeholder = Placeholder::find($data['placeholderName']) != null ?
				Placeholder::find($data['placeholderName']) : Placeholder::create($data['placeholderName'])
			;
			$placeholder->setValue($data['placeholderValue'] ?? '');
			unset($prevExisting[$placeholder->getId()]);
		}

		if ($deleteOld) {
			foreach (array_keys($prevExisting) as $id) {
				Placeholder::delete($id);
			}
		}

		Placeholder::flush();
	}

	/**
	 * @param string $type 'domain'/'user'/'system'
	 * @param array[] $embeddingData
	 * @param array{systemEmbedding?:array, userEmbedding?:array}
	 * @return array $embeddingMap
	 */
	public static function makeEmbeddingsImport(string $type, array $embeddingData, array $idMappings): array
	{
		$embeddingMap = [];
		switch ($type) {
			case 'domain':
				$embeddingType = Embedding::class;
				$assetType = EmbeddingAsset::class;
				$localeLocator = LocaleService::class;
				break;
			case 'user':
				$embeddingType = UserEmbedding::class;
				$assetType = UserEmbeddingAsset::class;
				$localeLocator = UserLocale::class;
				break;
			case 'system':
				$embeddingType = SystemEmbedding::class;
				$assetType = SystemEmbeddingAsset::class;
				$localeLocator = Language::class;
				break;
			default:
				throw new LogicException('unsupported embedding type');
		}

		foreach ($embeddingData as $data) {
			$purpose = Purpose::find($data['purpose'] ?? '') ?? Purpose::find('15c61c3');
			assert($purpose !== null);

			$embedding = $embeddingType::create()
				->setName($data['name'] ?? '')
				->setPurpose($purpose)
				->setActive((bool)($data['isActive'] ?? false))
				->setCodeSnippet($data['sourceCode'] ?? '')
				->setEmbeddingVendor($data['provider'] ?? '')
				->setCodeBlock($data['blockedScripts'] ?? [])
				->setIframeBlockingStrings($data['iframeBlockingStrings'] ?? [])
				->setIframesHiddenUntilConsent((bool)($data['hideIframesUntilConsent'] ?? false))
				->setDispatchLoadEventForIframes((bool)($data['dispatchLoadEventForIframes'] ?? true))
				->setExcludeUrls($data['excludeURLs'] ?? [])
				->setExcludeUrlsMode($data['excludeUrlsMode'] ?? EmbeddingBase::EXCLUDE_URLS_MODE_BLACKLIST)
				->setSource($data['source'] ?? '')
				->setAutoUpdated((bool)($data['autoUpdated'] ?? false))
				->setStructuredLegalBasis((array)($data['structuredLegalBasis'] ?? []))
				->setLegalBasisStructured((bool)($data['legalBasisIsStructured'] ?? false))
				->setScriptLoaderGroup($data['scriptLoaderGroup'] ?? '')
			;

			// Aufgrund von Standardwerten nur setzen, wenn im Export definiert
			if (isset($data['updateParameterBlacklist']) && is_array($data['updateParameterBlacklist'])) {
				$embedding->setUpdateParameterBlacklist($data['updateParameterBlacklist']);
			}

			$sourceId = $data['sourceId'] ?? null;
			switch ($data['source'] ?? '') {
				case EmbeddingBase::SOURCE_SYSTEM_TEMPLATE:
					$embedding->setSourceId($idMappings['systemEmbedding'][$sourceId] ?? $sourceId);
					break;
				case EmbeddingBase::SOURCE_USER_TEMPLATE:
					$embedding->setSourceId($idMappings['userEmbedding'][$sourceId] ?? $sourceId);
					break;
				default:
					$embedding->setSourceId($sourceId);
			}

			$embeddingMap[$data['id'] ?? ''] = $embedding->getId();

			try {
				if (isset($data['lastModified'])) {
					$lastModified = ($data['lastModified'] === null) ? null : new DateTimeImmutable((string)$data['lastModified']);

					$embedding
						->setLastModified($lastModified)
					;
				}
			} catch (Exception $e) { }

			try {
				if (isset($data['sourceTimestamp'])) {
					$sourceTimestamp = ($data['sourceTimestamp'] === null) ? null : new DateTimeImmutable((string)$data['sourceTimestamp']);

					$embedding
						->setSourceTimestamp($sourceTimestamp)
					;
				}
			} catch (Exception $e) { }

			foreach ($data['translations'] as $localeName => $translation) {
				$locale = $localeLocator::find((string)$localeName);
				if (!$locale) {
					continue;
				}

				$embedding
					->setTranslatedName($locale, $translation['label'] ?? '')
					->setTranslatedDescription($locale, $translation['description'] ?? '')
					->setTranslatedPrivacyPolicyUrl($locale, $translation['privacyPolicyUrl'] ?? '')
					->setTranslatedCollectedDataInfo($locale, $translation['collectedDataInfo'] ?? '')
					->setTranslatedPurposeOfDataCollection($locale, $translation['purposeOfDataCollection'] ?? '')
					->setTranslatedLegalBasis($locale, $translation['legalBasis'] ?? '')
					->setTranslatedPlaceOfProcessing($locale, $translation['placeOfProcessing'] ?? '')
					->setTranslatedBlockedContentTitle($locale, $translation['blockedContentTitle'] ?? '')
					->setTranslatedBlockedContentText($locale, $translation['blockedContentText'] ?? '')
					->setTranslatedBlockedContentButtonText($locale, $translation['blockedContentButtonText'] ?? '')
				;
			}

			foreach ($data['assets'] as $asset) {
				$embeddingAsset = $assetType::create()
					->setName($asset['name'] ?? '')
					->setDynamic((bool)($asset['isDynamic'] ?? false))
					->setStorageType($asset['storageType'] ?? '')
				;

				foreach ($asset['translations'] as $localeName => $translation) {
					$locale = $localeLocator::find((string)$localeName);
					if (!$locale) {
						continue;
					}

					$embeddingAsset
						->setTranslatedLifetime($locale, $translation['lifetime'] ?? '')
						->setTranslatedDescription($locale, $translation['description'] ?? '')
					;
				}

				$embedding->addAsset($embeddingAsset);
			}

			if ($data['tcf']) {
				$tcfVendor = TcfVendor::find($data['tcf']['vendor']);

				$purposes = array_filter(array_map(function($purposeID) {
					return TcfPurpose::find($purposeID);
				}, $data['tcf']['purposes']));

				$specialFeatures = array_filter(array_map(function($featureID) {
					return TcfSpecialFeature::find($featureID);
				}, $data['tcf']['specialFeatures']));

				$embedding
					->setTcfVendor($tcfVendor)
					->setTcfPurposes($purposes)
					->setTcfSpecialFeatures($specialFeatures)
				;
			}
		}

		$assetType::flush();
		$embeddingType::flush();
		return $embeddingMap;
	}

	/**
	 * @param array $cookieData
	 */
	private static function makeCookieImport(array $cookieData): void
	{
		foreach ($cookieData as $data) {
			$cookie = Cookie::create()
				->setName($data['name'] ?? '')
				->setPurpose(Purpose::find($data['purpose']))
				->setActive((bool)($data['isActive'] ?? false))
				->setHttpCookieName($data['httpCookieName'] ?? '')
				->setCodeSnippet($data['sourceCode'] ?? '')
				->setCookieVendor($data['provider'] ?? '')
				->setCodeBlock($data['blockedScripts'] ?? [])
			;

			foreach ($data['translations'] as $localeName => $translation) {
				$locale = LocaleService::find((string)$localeName);
				if (!$locale) {
					continue;
				}

				$cookie
					->setTranslatedCollectedDataInfo($locale, $translation['collectedDataInfo'] ?? '')
					->setTranslatedDescription($locale, $translation['description'] ?? '')
					->setTranslatedLegalBasis($locale, $translation['legalBasis'] ?? '')
					->setTranslatedLifetime($locale, $translation['lifetime'] ?? '')
					->setTranslatedPlaceOfProcessing($locale, $translation['placeOfProcessing'] ?? '')
					->setTranslatedPrivacyPolicyUrl($locale, $translation['privacyPolicyUrl'] ?? '')
					->setTranslatedPurposeOfDataCollection($locale, $translation['purposeOfDataCollection'] ?? '')
				;
			}
		}

		Cookie::flush();
	}

	/**
	 * @param Domain $domain
	 * @param array $consentData
	 */
	private static function makeConsentImport(Domain $domain, array $consentData): void
	{
		foreach ($consentData as $consent) {
			Protocol::byDomain($domain)->append($consent);
		}
	}
}
