<?php

namespace Plugins\Ccm19Imex\Controller;

use Plugins\Ccm19Imex\Component\Export;
use Plugins\Ccm19Imex\Component\Import;
use Plugins\Ccm19Imex\Event\ImportClientEvent;
use Plugins\Ccm19Imex\Event\ImportClientPreDomainsEvent;

use App\Exception\ApiKeyAlreadyExistsException;
use App\Exception\UniqueConstraintException;
use App\Model\Model;
use InvalidArgumentException;
use App\Model\PluginState;
use App\Component\Menu\Annotation\Menu;
use App\Utils;
use App\Model\Domain;
use App\Model\SubGroups;
use App\Model\User;
use App\Model\UserEmbedding;
use App\Controller\HostingController;
use App\Exception\CsrfTokenException;
use App\Exception\UserAlreadyExistsException;
use App\Model\CsrfToken;
use Exception;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\ResponseHeaderBag;
use Symfony\Component\HttpFoundation\StreamedResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Controller for ImEx Plugin - hosting
 *
 *  @Route("/imex-hosting/", name="plugin_ccm19_imex_hosting_", methods={"GET", "HEAD"})
 */
class Hosting extends HostingController
{
	private const STANDARD_IMPORT_FILE_NAME = 'import_clients';

	/**
	 * @Menu("Import/Export", icon="fa-exchange", order=50, group="Plugins", editions={"extended"})
	 * @Route("", name="index_get", methods={"GET", "HEAD"})
	 * @return Response
	 */
	public function index(): Response
	{
		// Render template
		return $this->render('@plugin:Ccm19Imex/internal/hosting.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'clients' => User::getClients(),
			'specializedImportAllowed' => (bool)($_ENV['IMEX_SPECIAL_FUNCTIONS'] ?? false),
		]);
	}

	/**
	 * @Route("export/user-list.json", name="export-list", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function listJson(Request $request, TranslatorInterface $translator): Response
	{
		$list = array_map(function ($user) {
			return [
				'id' => $user->getId(),
				'username' => $user->getUsername(),
				'active' => $user->isActive(),
			];
		}, array_values(User::getClients()));

		$response = new JsonResponse([
			'data' => $list,
		]);
		$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );
		return $response;
	}

	/**
	 * @Route("export", name="export", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param PluginState $pluginState
	 * @return Response
	 */
	public function export(Request $request, TranslatorInterface $translator, PluginState $pluginState): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('plugin_ccm19_imex_hosting_index_get');
		}

		switch ($request->request->get('action')) {
			case 'export-all':
				$clients = User::getClients();
				break;
			case 'save':
				$clients = array_map(function ($client) {
					return User::byId($client);
				}, explode(',', $request->request->get('selected-clients')));
				break;
			default:
				$clients = [];
				break;
		}

		if (empty($clients)) {
			$this->addFlash('danger', $translator->trans('Nothing was selected to export'));
			return $this->redirectToRoute('plugin_ccm19_imex_hosting_index_get');
		}

		$exportData = [
			'exportDomain' => true,
			'exportThemes' => true,
			'exportCookieStatementTable' => true,
			'exportLocales' => true,
			'exportImprint' => true,
			'exportDataProtection' => true,
			'exportAccessibility' => true,
			'exportPlaceholders' => true,
			'exportEmbeddings' => true,
			'exportCookies' => true,
			'exportConsents' => isset($request->request->get('exportSettings')['consents']) && $request->request->get('exportSettings')['consents'],
		];

		$exportScheme = $request->request->get('exportScheme') ? $request->request->get('exportScheme') : "json";

		$response = new StreamedResponse(function () use ($exportData, $clients, $exportScheme, $pluginState) {
			$response = [
				'clients' => array_map(function (User $client) use ($exportData) {
					return Export::exportClient($client, $exportData);
				}, $clients),
				'systemThemes' => Export::exportSystemThemes(),
				'pluginVersion' => $pluginState->getVersion(),
			];

			echo $exportScheme == "gzip" ? gzencode(json_encode($response), 9) : json_encode($response);
		});

		$exportFileName = $exportScheme == "gzip" ? 'export_clients.json.gz' : 'export_clients.json';

		$disposition = $response->headers->makeDisposition(
			ResponseHeaderBag::DISPOSITION_ATTACHMENT,
			$exportFileName
		);
		$response->headers->set('Content-Disposition', $disposition);
		return $response;
	}

	/**
	 * @Route("import", name="import", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function import(Request $request, TranslatorInterface $translator): Response
	{
		/** @var UploadedFile $file */
		$file = $request->files->get('importfile');

		if (!$file) {
			$this->addFlash('danger', $translator->trans('No file was uploaded'));
			return $this->redirectToRoute('plugin_ccm19_imex_hosting_index_get');
		}
		elseif ($file->getError() != UPLOAD_ERR_OK) {
			$this->addFlash('danger', $translator->trans('An error occurred during upload'));
			return $this->redirectToRoute('plugin_ccm19_imex_hosting_index_get');
		}

		// SEE: https://www.rfc-editor.org/rfc/rfc1952#page-6
		$fileData = file_get_contents($file->getPathname(), false, NULL, 0, 2) === "\x1f\x8b"
			? gzdecode(file_get_contents($file->getPathname()))
			: file_get_contents($file->getPathname());
		$data = json_decode($fileData, true);

		if (!$data) {
			$this->addFlash('danger', $translator->trans('The file is broken'));
			return $this->redirectToRoute('plugin_ccm19_imex_hosting_index_get');
		}

		$specializedImport = !empty($request->request->get('specializedImport')) && $request->request->get('specializedImport') == 'on';

		if (is_array($data) && (isset($data['clients']) && is_array($data['clients']))) {
			$user = User::loggedInUser();
			assert($user instanceof User);

			$destination = Utils::getBundledVarDir() . "/";
			Utils::assertDirectoryWritable($destination);
			$file->move($destination, self::STANDARD_IMPORT_FILE_NAME . "_{$user->getId()}.json");

			$takenUsernames = array_filter($data['clients'], function($item) {
				return User::find($item['clientData']['username'] ?? '') != NULL;
			});
			foreach ($takenUsernames as $takenUsername) {
				$this->addFlash('warning', $translator->trans('The username %username% already exists and will not be imported as a new client', [
					'%username%' => $takenUsername['clientData']['username'],
				]));
			}

			if ($specializedImport) {
				$takenUserAPIKeys = array_filter($data['clients'], function($item) {
					return User::byApiKey($item['clientData']['apiKey'] ?? '') != NULL;
				});
				foreach ($takenUserAPIKeys as $takenUserAPIKey) {
					$this->addFlash('warning', $translator->trans('The user with APIKey %apikey% and username %username% already exists and will not be imported as a new client', [
						'%%apikey%' => $takenUserAPIKey['clientData']['apiKey'],
						'%username%' => $takenUserAPIKey['clientData']['username'],
					]));
				}
			}

			$clients = array_map(function ($item) {
				return User::createNonPersistent()
					->setUsername($item['clientData']['username'] ?? '') // FIXME: Should be rewritten, if the username is already taken
					;
			}, $data['clients']);

			return $this->render('@plugin:Ccm19Imex/internal/mapping-hosting.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'clients' => $clients,
				'_clients' => User::getClients(),
				'specializedImport' => $specializedImport,
				'specializedImportAllowed' => (bool)($_ENV['IMEX_SPECIAL_FUNCTIONS'] ?? false),
			]);
		}
		else {
			$this->addFlash('warning', $translator->trans('There was no client in the file'));
			return $this->redirectToRoute('plugin_ccm19_imex_hosting_index_get');
		}
	}
	/**
	 * @Route("import/mapping", name="import_mapped", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function finishMappedImport(Request $request, TranslatorInterface $translator): Response
	{
		$mappingEntries = $request->request->get('import');

		if ($mappingEntries === NULL) {
			$this->addFlash('danger', $translator->trans('Nothing was selected to import'));
			return $this->redirectToRoute('plugin_ccm19_imex_hosting_index_get');
		}

		$user = User::loggedInUser();
		assert($user instanceof User);

		$fileName = Utils::getBundledVarDir() . "/" . self::STANDARD_IMPORT_FILE_NAME . "_{$user->getId()}.json";

		// SEE: https://www.rfc-editor.org/rfc/rfc1952#page-6
		$file = file_get_contents($fileName, false, NULL, 0, 2) === "\x1f\x8b"
			? gzdecode(file_get_contents($fileName))
			: file_get_contents($fileName);
		$file = json_decode($file, true);

		$specializedImport = !empty($request->request->get('specializedImport')) && $request->request->get('specializedImport') == 1;

		$systemThemeMap = [];
		if (isset($file['systemThemes']) && $file['systemThemes']) {
			$systemThemeMap = Import::makeThemesImport('system', $file['systemThemes'], NULL, true);
		}

		foreach ($mappingEntries as $key => $mappingEntry) {
			$fileClient = $file['clients'][$key] ?? null;
			if (isset($fileClient['clientData']) && $fileClient['clientData']) {
				$domainMap = [];

				if ($mappingEntry == "_no_" || $fileClient == null || $mappingEntry == null) {
					continue;
				}
				elseif ($mappingEntry == "_new_") {
					try {
						$client = User::create($fileClient['clientData']['username']);

						if ($specializedImport) {
							$client->setApiKey($fileClient['clientData']['apiKey']);
							if (!empty($fileClient['clientData']['generalApiKey'])) {
								$client->setGeneralApiKey($fileClient['clientData']['generalApiKey']);
							}
						}
					} catch (UserAlreadyExistsException $e) {
						$this->addFlash('warning', $translator->trans('The username %username% already exists and was not imported', [
							'%username%' => $fileClient['clientData']['username'],
						]));
						continue;
					} catch (ApiKeyAlreadyExistsException $e) {
						$this->addFlash('warning', $translator->trans('The user with APIKey %apikey% and username %username% already exists and was not imported', [
							'%apikey%' => $fileClient['clientData']['apiKey'],
							'%username%' => $fileClient['clientData']['username'],
						]));
						continue;
					}

					// Writes changes of the client to prevent accidental deadlocks
					$client->save();
				}
				elseif (User::byId($mappingEntry) !== NULL) {
					$client = User::byId($mappingEntry);
				}
				else {
					continue;
				}

				Import::makeClientImport($client, $fileClient['clientData']);

				// Switch to newly created user
				User::setApiUser($client->getApiKey());
				Model::closeUserModels();
				Domain::switchToUser($client);

				$userThemeMap = Import::makeThemesImport('user', $fileClient['userThemes'], NULL, true);
				Import::makeLocalesImport('user', $fileClient['standardTexts']);

				$idMappings = [
					'systemTheme' => $systemThemeMap,
					'userTheme' => $userThemeMap,
					'systemEmbedding' => [],
					'userEmbedding' => [],
				];
				UserEmbedding::clear();
				$idMappings['userEmbedding'] = Import::makeEmbeddingsImport('user', $fileClient['userEmbeddings'] ?? [], $idMappings);

				Utils::dispatchEvent(ImportClientPreDomainsEvent::NAME, new ImportClientPreDomainsEvent($client, $fileClient, $idMappings));

				if (isset($fileClient['domains']) && is_array($fileClient['domains'])) {
					foreach ($fileClient['domains'] as $domainId => $fileDomain) {
						$newDomainID = $specializedImport ? Domain::canonicalizeId($client->getId(), $fileDomain['domainData']['id']) : NULL;

						try {
							$domain = Domain::createForUser($client, false, $newDomainID)->setName($fileDomain['domainData']['name']);
						} catch (UniqueConstraintException $e) {
							$this->addFlash('danger', $translator->trans('The ID for the domain %domain% and the user %user% already exists and was not imported', [
								'%domain%' => $fileDomain['domainData']['name'],
								'%user%' => $client->getUsername(),
							]));
							continue;
						} catch (InvalidArgumentException $e) {
							$this->addFlash('danger', $translator->trans('The ID for the domain %domain% and the user %user% is not usable in this repository and was not imported', [
								'%domain%' => $fileDomain['domainData']['name'],
								'%user%' => $client->getUsername(),
							]));
							continue;
						}

						// Writes changes of the domain to prevent accidental deadlocks
						$domain->save();

						// Grant access to newly created domain(s) to subuser
						if ($client->isSubUser()) {
							$client
								->grantDomainAccess($domain)
								->save();
						}

						try {
							Import::importDomain($domain, $fileDomain, true, $idMappings);
							$domainMap[$domainId] = $domain;
						}
						catch (Exception $e) {
							$this->addFlash('danger', $translator->trans("HOSTING: " . $e->getMessage(), [
								'%username%' => $client->getUsername(),
							]));
						}
						Model::closeDomainModels();
					}
				}
				$client->updateDomainInfo()->save();

				Utils::dispatchEvent(ImportClientEvent::NAME, new ImportClientEvent($client, $fileClient, $idMappings));

				if (isset($fileClient['subuserGroups']) && $fileClient['subuserGroups']) {
					$groupMap = [];

					SubGroups::switchRepositoryContext($client);
					foreach ($fileClient['subuserGroups'] as $groupId => $fileGroup) {
						$subgroup = Import::importSubUserGroup($fileGroup);
						$groupMap[$groupId] = $subgroup;
					}
					SubGroups::flush();

					if (isset($fileClient['subusers']) && $fileClient['subusers']) {
						foreach ($fileClient['subusers'] as $subuserData) {
							try {
								$subuser = User::create($subuserData['username'])
									->setSubUserOf($client->getId())
								;
							} catch (UserAlreadyExistsException $e) {
								$this->addFlash('warning', $translator->trans('The username %username% already exists and was not imported', [
									'%username%' => $fileClient['clientData']['username'],
								]));
								continue;
							}

							$realGroups = [];
							$realDomains = [];
							foreach ($subuserData['groups'] as $key => $group) {
								if (isset($groupMap[$key])) {
									$realGroups[$groupMap[$key]->getId()] = $group ? "on" : "off";
								}
							}
							foreach ($subuserData['domains'] as $key => $domain) {
								if (isset($domainMap[$key])) {
									$realDomains[$domainMap[$key]->getId()] = $domain ? "on" : "off";
								}
							}
							$subuser->setGroupActive($realGroups);
							$subuser->setDomActive($realDomains);

							Import::makeClientImport($subuser, $subuserData);
						}
						User::flush();
					}
				}
			}
		}

		@unlink($fileName);

		$this->addFlash('success', $translator->trans('The import has been completed'));
		return $this->redirectToRoute('plugin_ccm19_imex_hosting_index_get');
	}
}
