<?php

namespace Plugins\Ccm19Imex\Controller;

use Plugins\Ccm19Imex\Component\Export;
use Plugins\Ccm19Imex\Component\Import;

use App\Model\PluginState;
use App\Component\Menu\Annotation\Menu;
use App\Model\User;
use App\Utils;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Domain;
use Exception;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\ResponseHeaderBag;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Controller for ImEx Plugin - abstract
 *
 * TODO: Maybe extend to include User-Themes in the Import/Export? This would have to be tested alot and checked, how it would act in what scenario
 *
 * @Route("/imex/", name="plugin_ccm19_imex_multi_", methods={"GET", "HEAD"})
 */
class MultiDomain extends AbstractController
{
	private const STANDARD_IMPORT_FILE_NAME = 'import_domains';

	/**
	 * @Menu("Import/Export", icon="fa-exchange", order=50, group="Plugins", navigation="meta", editions={"extended", "multi"})
	 * @Route("", name="index_get", methods={"GET", "HEAD"})
	 * @return Response
	 */
	public function index(): Response
	{
		// Render template
		return $this->render('@plugin:Ccm19Imex/internal/multi.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'userDomains' => Domain::all(),
            'specializedImportAllowed' => (bool)($_ENV['IMEX_SPECIAL_FUNCTIONS'] ?? false),
		]);
	}

	/**
	 * @Route("export/domain-list.json", name="export-list", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function listJson(Request $request, TranslatorInterface $translator): Response
	{
		$list = array_map(function ($domain) {
			return [
				'id' => $domain->getId(),
				'domainname' => $domain->getName() . $domain->getUrlPath(),
			];
		}, array_values(Domain::all()));

		$response = new JsonResponse([
			'data' => $list,
		]);

		$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );
		return $response;
	}

	/**
	 * @Route("export", name="export", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param PluginState $pluginState
	 * @return Response
	 */
	public function export(Request $request, TranslatorInterface $translator, PluginState $pluginState): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}

		switch ($request->request->get('action')) {
			case 'export-all':
				$domains = Domain::all();
				break;
			case 'save':
				$domains = array_map(function ($domain) {
					return Domain::find($domain);
				}, explode(',', $request->request->get('selected-domains')));
				break;
			default:
				$domains = [];
				break;
		}

		if (empty($domains)) {
			$this->addFlash('danger', $translator->trans('Nothing was selected to export'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}

		$exportData = [
			'exportDomain' => true,
			'exportThemes' => true,
			'exportCookieStatementTable' => true,
			'exportLocales' => true,
			'exportImprint' => true,
			'exportDataProtection' => true,
			'exportAccessibility' => true,
			'exportPlaceholders' => true,
			'exportEmbeddings' => true,
			'exportCookies' => true,
			'exportConsents' => isset($request->request->get('exportSettings')['consents']) && $request->request->get('exportSettings')['consents'],
		];

		$returnedData['domains'] = Export::exportDomains(User::loggedInUser(), $domains, $exportData);

		if (empty($returnedData['domains'])) {
			$this->addFlash('danger', $translator->trans('Nothing was exported'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}

		$returnedData['pluginVersion'] = $pluginState->getVersion();

		$exportScheme = $request->request->get('exportScheme') ? $request->request->get('exportScheme') : "json";

		if ($exportScheme == "gzip") {
			$exportFileName = 'export_domains.json.gz';
			$fileContent = gzencode(json_encode($returnedData), 9);

			$response = new Response($fileContent, 200, [
				'Content-Type' => 'application/x-ccm19-export+gzip',
			]);
		}
		else {
			$exportFileName = 'export_domains.json';
			$fileContent = json_encode($returnedData);

			$response = new Response($fileContent, 200, [
				'Content-Type' => 'application/x-ccm19-export+json',
			]);
		}

		$disposition = $response->headers->makeDisposition(
			ResponseHeaderBag::DISPOSITION_ATTACHMENT,
			$exportFileName
		);
		$response->headers->set('Content-Disposition', $disposition);
		return $response;
	}

	/**
	 * @Route("import", name="import", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function import(Request $request, TranslatorInterface $translator): Response
	{
		/** @var UploadedFile $file */
		$file = $request->files->get('importfile');

		if (!$file) {
			$this->addFlash('danger', $translator->trans('No file was uploaded'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}
		elseif ($file->getError() != UPLOAD_ERR_OK) {
			$this->addFlash('danger', $translator->trans('An error occurred during upload'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}

		// SEE: https://www.rfc-editor.org/rfc/rfc1952#page-6
		$fileData = file_get_contents($file->getPathname(), false, NULL, 0, 2) === "\x1f\x8b"
			? gzdecode(file_get_contents($file->getPathname()))
			: file_get_contents($file->getPathname());
		$data = json_decode($fileData, true);

		if (!$data) {
			$this->addFlash('danger', $translator->trans('The file is broken'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}

		if (is_array($data)) {
			$user = User::loggedInUser();
			assert($user instanceof User);

			$destination = Utils::getBundledVarDir() . "/";
			Utils::assertDirectoryWritable($destination);
			$file->move($destination, self::STANDARD_IMPORT_FILE_NAME . "_{$user->getId()}.json");

			$domains = array_map(function ($item) {
				return Domain::createNonPersistent()
					->setName($item['domainData']['name'] ?? '')
                    ->setUrlPath($item['domainData']['urlPath'] ?? '')
					;
			}, $data['domains'] ?? []);

            $specializedImport = !empty($request->request->get('specializedImport')) && $request->request->get('specializedImport') == 'on';

			return $this->render('@plugin:Ccm19Imex/internal/mapping-multi.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'domains' => $domains,
                'specializedImport' => $specializedImport,
                'specializedImportAllowed' => (bool)($_ENV['IMEX_SPECIAL_FUNCTIONS'] ?? false),
			]);
		}
		else {
			$this->addFlash('warning', $translator->trans('There was no domain in the file'));
			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}
	}

	/**
	 * @Route("import/mapping", name="import_mapped", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function finishMappedImport(Request $request, TranslatorInterface $translator): Response
	{
		$mappingEntries = $request->request->get('import');

		if ($mappingEntries === NULL) {
			$this->addFlash('danger', $translator->trans('Nothing was selected to import'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}

		$user = User::loggedInUser();
		assert($user instanceof User);

		$fileName = Utils::getBundledVarDir() . "/" . self::STANDARD_IMPORT_FILE_NAME . "_{$user->getId()}.json";

		// SEE: https://www.rfc-editor.org/rfc/rfc1952#page-6
		$file = file_get_contents($fileName, false, NULL, 0, 2) === "\x1f\x8b"
			? gzdecode(file_get_contents($fileName))
			: file_get_contents($fileName);
		$file = json_decode($file, true);

        $specializedImport = !empty($request->request->get('specializedImport')) && $request->request->get('specializedImport') == '1';

        $importedDomains = [];

		$currentUser = User::loggedInUser();

		if ($currentUser->isDomainLimitReached()) {
			$this->addFlash('danger', $translator->trans('You\'ve reached the domain limit. Please contact your provider to raise that limit.'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}

		foreach ($mappingEntries as $key => $mappingEntry) {
			$fileDomain = $file['domains'][$key] ?? null;

			if ($mappingEntry == "_no_" || $fileDomain == null || $mappingEntry == null) {
				continue;
			}
			elseif ($mappingEntry == "_new_") {
                //Im Falle spezialisierter Import (d.h. kommt aus einer alten CCM19 Instanz) wird eine speziell DomId erzeugt damit die alte DomId genutzt werden kann
                $newDomainID = $specializedImport ? Domain::canonicalizeId($currentUser->getId(), $fileDomain['domainData']['id']) : NULL;

                //Domain erzeugen - falls vorhanden dann die connonicalizeId verwenden
				$domain = Domain::createForUser($currentUser, false,$newDomainID)->setName($fileDomain['domainData']['name']);
			}
			elseif (Domain::find($mappingEntry) !== NULL) {
				$domain = Domain::find($mappingEntry);
			}
			else {
				continue;
			}

			try {
				Import::importDomain($domain, $fileDomain, $mappingEntry == "_new_");
				$importedDomains[] = $domain;
			}
			catch (Exception $exception) {
				$this->addFlash('danger', $translator->trans($exception->getMessage()));
			}
		}

		// Grant access to newly created domain(s) to subuser
		foreach ($importedDomains as $importedDomain) {
			if ($currentUser->isSubUser()) {
				$currentUser
					->grantDomainAccess($importedDomain)
					->save();
			}
		}
		User::flush();

		@unlink($fileName);

		$this->addFlash('success', $translator->trans('The import has been completed'));
		return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
	}
}
