<?php

namespace Plugins\Ccm19Imex\Controller;

use Plugins\Ccm19Imex\Component\Export;
use Plugins\Ccm19Imex\Component\Import;

use App\Model\User;
use App\Model\PluginState;
use App\Component\Menu\Annotation\Menu;
use App\Utils;
use App\Model\Domain;
use App\Controller\DomainDependantController;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use Exception;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\ResponseHeaderBag;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Controller for ImEx Plugin - domain dependent
 *
 *  @Route("/domains/{_domainId}/plugins/imex/", name="plugin_ccm19_imex_single_")
 */
class SingleDomain extends DomainDependantController
{
	private const STANDARD_IMPORT_FILE_NAME = 'import';

	/**
	 * @Menu("Import/Export", icon="fa-exchange", order=50, group="Plugins", editions={"extended", "multi", "base"})
	 * @Route("", name="index_get", methods={"GET", "HEAD"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function index(Request $request, TranslatorInterface $translator): Response
	{
		$mainCategory = $translator->trans('Cookies & Settings');
		$textCategory = $translator->trans('Content & Texts');

		// Render template
		return $this->render('@plugin:Ccm19Imex/internal/single.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'categoryGroups' => [[
				'label' => $mainCategory,
				'categories' => [
					['name' => 'themes', 'label' => $translator->trans('Themes')],
					['name' => 'embeddings', 'label' => $translator->trans('Embeddings')],
					['name' => 'cookies', 'label' => $translator->trans('Cookies') . ' ' . $translator->trans('(deprecated)')],
					['name' => 'cookieeStatementTable', 'label' => $translator->trans('Cookie declaration table')],
					['name' => 'domainSettings', 'label' => $translator->trans('System')],
					['name' => 'consents', 'label' => $translator->trans('Protocol')],
				],
			], [
				'label' => $textCategory,
				'categories' => [
					['name' => 'locales', 'label' => $translator->trans('Texts')],
					['name' => 'imprint', 'label' => $translator->trans('Imprint')],
					['name' => 'dataProtection', 'label' => $translator->trans('Privacy protection')],
					['name' => 'accessibility', 'label' => $translator->trans('Accessibility')],
					['name' => 'placeholders', 'label' => $translator->trans('Placeholder')],
				],
			]],
		]);
	}

	/**
	 * @Route("export", name="export", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param PluginState $pluginState
	 * @return Response
	 */
	public function export(Request $request, TranslatorInterface $translator, PluginState $pluginState): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}

		$exportData = [
			"exportDomain" => (bool)$request->request->get('domainSettings'),
			"exportThemes" => (bool)$request->request->get('themes'),
			"exportCookieStatementTable" => (bool)$request->request->get('cookieeStatementTable'),
			"exportLocales" => (bool)$request->request->get('locales'),
			"exportImprint" => (bool)$request->request->get('imprint'),
			"exportDataProtection" => (bool)$request->request->get('dataProtection'),
			"exportAccessibility" => (bool)$request->request->get('accessibility'),
			"exportPlaceholders" => (bool)$request->request->get('placeholders'),
			"exportEmbeddings" => (bool)$request->request->get('embeddings'),
			"exportCookies" => (bool)$request->request->get('cookies'),
			"exportConsents" => (bool)$request->request->get('consents'),
		];

		// If no data is to be exported, throw an error
		if (array_sum($exportData) == 0) {
			$this->addFlash('danger', $translator->trans('Nothing was selected to export'));
			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}

		$returnedData = Export::exportDomain(Domain::activeDomain(), $exportData);

		if (empty($returnedData)) {
			$this->addFlash('danger', $translator->trans('Nothing was exported'));
			return $this->redirectToRoute('plugin_ccm19_imex_multi_index_get');
		}

		$returnedData['pluginVersion'] = $pluginState->getVersion();

		$exportScheme = $request->request->get('exportScheme') ? $request->request->get('exportScheme') : "json";

		if ($exportScheme == "gzip") {
			$exportFileName = 'export.json.gz';
			$fileContent = gzencode(json_encode($returnedData), 9);

			$response = new Response($fileContent, 200, [
				'Content-Type' => 'application/x-ccm19-export+gzip',
			]);
		}
		else {
			$exportFileName = 'export.json';
			$fileContent = json_encode($returnedData);

			$response = new Response($fileContent, 200, [
				'Content-Type' => 'application/x-ccm19-export+json',
			]);
		}

		$disposition = $response->headers->makeDisposition(
			ResponseHeaderBag::DISPOSITION_ATTACHMENT,
			$exportFileName
		);
		$response->headers->set('Content-Disposition', $disposition);
		return $response;
	}

	/**
	 * @Route("import", name="import", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function import(Request $request, TranslatorInterface $translator): Response
	{
		/** @var UploadedFile $file */
		$file = $request->files->get('importfile');

		if (!$file) {
			$this->addFlash('danger', $translator->trans('No file was uploaded'));
			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}
		elseif ($file->getError() != UPLOAD_ERR_OK) {
			$this->addFlash('danger', $translator->trans('An error occurred during upload'));
			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}

		// SEE: https://www.rfc-editor.org/rfc/rfc1952#page-6
		$fileData = file_get_contents($file->getPathname(), false, NULL, 0, 2) === "\x1f\x8b"
			? gzdecode(file_get_contents($file->getPathname()))
			: file_get_contents($file->getPathname());
		$data = json_decode($fileData, true);

		if (!$data) {
			$this->addFlash('danger', $translator->trans('The file is broken'));
			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}

		if (isset($data['domains']) && is_array($data['domains'])) {
			$user = User::loggedInUser();
			assert($user instanceof User);

			$destination = Utils::getBundledVarDir() . "/";
			Utils::assertDirectoryWritable($destination);
			$file->move($destination, self::STANDARD_IMPORT_FILE_NAME . "_{$user->getId()}.json");

			$domains = array_map(function ($item) {
				return Domain::createNonPersistent()
					->setName($item['domainData']['name'] ?? '')
					;
			}, $data['domains']);

			return $this->render('@plugin:Ccm19Imex/internal/mapping-single.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'domains' => $domains,
			]);
		}
		elseif (is_array($data)) {
			try {
				Import::importDomain(Domain::activeDomain(), $data);
				$this->addFlash('success', $translator->trans('The import has been completed'));
			}
			catch (Exception $exception) {
				$this->addFlash('danger', $translator->trans($exception->getMessage()));
			}

			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}
		else {
			$this->addFlash('warning', $translator->trans('There was no domain in the file'));
			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}
	}
	/**
	 * @Route("import/mapping", name="import_mapped", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function finishMappedImport(Request $request, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$fileName = Utils::getBundledVarDir() . "/" . self::STANDARD_IMPORT_FILE_NAME . "_{$user->getId()}.json";

		// SEE: https://www.rfc-editor.org/rfc/rfc1952#page-6
		$file = file_get_contents($fileName, false, NULL, 0, 2) === "\x1f\x8b"
			? gzdecode(file_get_contents($fileName))
			: file_get_contents($fileName);
		$file = json_decode($file, true);

		$domainKey = $request->request->get('importDomain');
		$domainToImport = $file['domains'][$domainKey] ?? null;

		if (empty($domainToImport)) {
			$this->addFlash('danger', $translator->trans('Invalid domain selected'));
			return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
		}

		try {
			Import::importDomain(Domain::activeDomain(), $domainToImport);
			$this->addFlash('success', $translator->trans('The import has been completed'));
		}
		catch (Exception $exception) {
			$this->addFlash('danger', $translator->trans($exception->getMessage()));
		}

		@unlink($fileName);

		return $this->redirectToRoute('plugin_ccm19_imex_single_index_get');
	}
}
