<?php

namespace Plugins\Ccm19ScriptPlaceholder\Component;

use App\Model\Domain;
use App\Model\Model;
use Plugins\Ccm19ScriptPlaceholder\Model\Placeholder;
use Plugins\Ccm19ScriptPlaceholder\Model\Snippet;
use Plugins\Ccm19ScriptPlaceholder\Model\UserPlaceholder;
use Plugins\Ccm19ScriptPlaceholder\Model\UserSnippet;

class Deployment
{
	/**
	 * This method distributes all `UserPlaceholder` objects to multiple domains.
	 *
	 * This method closes all domain model repositories, locks repositories
	 * of the models `Placeholder` and `Snippet` and releases the locks upon
	 * writing the changes to the database.
	 *
	 * Also, this method updates the published values after deploying to
	 * every domain if `$onlyPublished` is `false`.
	 *
	 * @param Domain[] $domains
	 * @param bool $onlyPublished whether to publish only already published objects using the last published values.
	 * @return void
	 */
	private static function publishUserPlaceholdersToDomains(array $domains, bool $onlyPublished=false): void
	{
		$userPlaceholders = UserPlaceholder::all();
		if (empty($userPlaceholders)) {
			return;
		}

		// Remember the current domain context
		$requestDomainId = Domain::getDomainIdForRequest();

		foreach ($domains as $domain) {
			Domain::setDomainForRequest($domain->getId());
			Model::closeDomainModels();

			// Keep the transaction open for the duration of processing a single domain
			Placeholder::beginWrite();
			Snippet::beginWrite();

			// Delete dangling placeholders
			foreach (Placeholder::all() as $placeholder) {
				if ($placeholder->isDangling()) {
					foreach ($placeholder->getSnippets() as $snippet) {
						Snippet::delete($snippet->getId(), false);
					}
					Placeholder::delete($placeholder->getId(), false);
				}
			}

			// Copy data from user placeholders
			foreach ($userPlaceholders as $userPlaceholder) {
				if ($onlyPublished && !$userPlaceholder->isPublished()) {
					continue;
				}

				/** @var Placeholder $placeholder */
				$placeholder = Placeholder::byUserPlaceholder($userPlaceholder) ?? Placeholder::create();

				// Delete existing snippets, they will be recreated if necessary
				foreach ($placeholder->getSnippets() as $snippet) {
					Snippet::delete($snippet->getId(), false);
				}

				// Create snippets
				$snippets = array_map(function (UserSnippet $snippet) {
					return Snippet::create()
						->setName($snippet->getName())
						->setTranslations($snippet->getTranslations());
				}, $userPlaceholder->getSnippets());

				// Copy values
				$placeholder
					->setName($userPlaceholder->getName())
					->setPlacement($userPlaceholder->getPlacement())
					->setHtmlCode($userPlaceholder->getHtmlCode())
					->setCssCode($userPlaceholder->getCssCode())
					->setJsCode($userPlaceholder->getJsCode())
					->setSnippets($snippets)
				;

				if ($placeholder->getPlacementCssSelector() == '' || !$onlyPublished && $placeholder->getPlacementCssSelector() === $userPlaceholder->getPublishedPlacementCssSelector()) {
					$placeholder->setPlacementCssSelector($onlyPublished ? (string)$userPlaceholder->getPublishedPlacementCssSelector() : $userPlaceholder->getPlacementCssSelector());
				}
				if ($placeholder->getPlacementScriptNeedle() == '' || !$onlyPublished && $placeholder->getPlacementScriptNeedle() === $userPlaceholder->getPublishedPlacementScriptNeedle()) {
					$placeholder->setPlacementScriptNeedle($onlyPublished ? (string)$userPlaceholder->getPublishedPlacementScriptNeedle() : $userPlaceholder->getPlacementScriptNeedle());
				}
				if (empty($placeholder->getTimeConstraints()) || !$onlyPublished && $placeholder->getTimeConstraints() === $userPlaceholder->getPublishedTimeConstraints()) {
					$placeholder->setTimeConstraints($onlyPublished ? (array)$userPlaceholder->getPublishedTimeConstraints() : $userPlaceholder->getTimeConstraints());
				}

				// Reference user placeholder
				$placeholder->setUserPlaceholder($userPlaceholder);
			}

			// Commit transactions
			Snippet::flush();
			Placeholder::flush();
		}

		// Restore the previous domain context
		Domain::setDomainForRequest($requestDomainId);
		Model::closeDomainModels();

		if (!$onlyPublished) {
			// Update published values
			foreach ($userPlaceholders as $userPlaceholder) {
				$userPlaceholder
					->setPublishedPlacementCssSelector($userPlaceholder->getPlacementCssSelector())
					->setPublishedPlacementScriptNeedle($userPlaceholder->getPlacementScriptNeedle())
					->setPublishedTimeConstraints($userPlaceholder->getTimeConstraints())
				;
			}
			UserPlaceholder::flush();
		}
	}

	/**
	 * This method distributes previously published `UserPlaceholder` objects
	 * to a single domain.
	 *
	 * This method closes all domain model repositories, locks repositories
	 * of the models `Placeholder` and `Snippet` and releases the locks upon
	 * writing the changes to the database.
	 *
	 * Due to data integrity, only published placeholders are updated.
	 *
	 * @param Domain $domain
	 * @return void
	 */
	public static function publishUserPlaceholdersToDomain(Domain $domain): void
	{
		self::publishUserPlaceholdersToDomains([$domain], true);
	}

	/**
	 * This method distributes all `UserPlaceholder` objects to all domains
	 * in the current `User` context.
	 *
	 * This method closes all domain model repositories, locks repositories
	 * of the models `Placeholder` and `Snippet` and releases the locks upon
	 * writing the changes to the database.
	 *
	 * Also, this method updates the published values after deploying to
	 * every domain. Therefore, this method is ideally called in the context
	 * of the account owner.
	 *
	 * @return void
	 */
	public static function publishUserPlaceholders(): void
	{
		self::publishUserPlaceholdersToDomains(Domain::all());
	}
}
