<?php

namespace Plugins\Ccm19ScriptPlaceholder\Controller;

use App\Component\Cache\Cache;
use App\Component\HttpFoundation\JavascriptResponse;
use App\Component\LocaleService;
use App\Controller\ExternalController;
use App\Model\Domain;
use App\Model\Theme;
use App\Model\User;
use App\Utils;
use DateTime;
use Plugins\Ccm19ScriptPlaceholder\Model\Placeholder;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/plugins/Ccm19ScriptPlaceholder", name="plugin_ccm19_script_placeholder_front_")
 */
class Frontend extends ExternalController
{
	/**
	 * @Route("/module.js", name="main_javascript_file", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @return JavascriptResponse
	 */
	public function mainJavascriptFile(Request $request): JavascriptResponse
	{
		// Benutzer anhand des verwendeten API-Schlüssels ermitteln - siehe AuthenticationHandler
		$user = User::loggedInUser();
		$domain = Domain::activeDomain();
		assert($user instanceof User);

		if (!$domain) {
			return new JavascriptResponse('console.error(\'[CCM19] Plugin error: '.addcslashes('Invalid domain.', '\'').'\');', 400);
		}

		// Ignore system and user themes, i.e. preview mode
		$theme = Theme::find((string)$request->query->get('theme'));

		if (!$theme) {
			return new JavascriptResponse('console.error(\'[CCM19] Plugin error: '.addcslashes('Invalid theme.', '\'').'\');', 400);
		}

		$locale = LocaleService::find((string)$request->query->get('lang'));

		if (!$locale) {
			return new JavascriptResponse('console.error(\'[CCM19] Plugin error: '.addcslashes('Invalid locale.', '\'').'\');', 400);
		}

		// Prepare template data
		$placeholders = Placeholder::enabledAndViable();
		$placeholdersJsObject = array_values(array_map(function (Placeholder $placeholder) use ($locale) {
			if ($timeConstraints = $placeholder->getTimeConstraints()) {
				$timeConstraints = array_map(function (array $constraint) {
					$timeFrom = explode(':', $constraint['timeFrom']);
					$timeTo = explode(':', $constraint['timeTo']);

					return [
						'days' => $constraint['dayFrom'] <= $constraint['dayTo']
							? range($constraint['dayFrom'], $constraint['dayTo'])
							: array_values(array_merge(range(1, $constraint['dayTo']), range($constraint['dayFrom'], 7))),
						'from' => [
							'hour' => max(0, min((int)$timeFrom[0], 23)),
							'minute' => max(0, min((int)($timeFrom[1] ?? 0), 59)),
						],
						'to' => [
							'hour' => max(0, min((int)$timeTo[0], 23)),
							'minute' => max(0, min((int)($timeTo[1] ?? 0), 59)),
						],
					];
				}, $timeConstraints);
			}

			return [
				'id' => $placeholder->getId(),
				'name' => $placeholder->getName(),
				'embedding' => $placeholder->getEmbedding()->getId(),
				'htmlCode' => trim($placeholder->getHtmlCodeWithSnippetsReplaced($locale)),
				'cssCode' => trim($placeholder->getCssCodeWithPlaceholderScope()),
				'jsCode' => trim($placeholder->getJsCode()),
				'placement' => $placeholder->getPlacement(),
				'placementCssSelector' => $placeholder->getPlacementCssSelector(),
				'placementScriptNeedle' => $placeholder->getPlacementScriptNeedle(),
				'timeConstraints' => $timeConstraints,
			];
		}, $placeholders));

		// Determine daylight savings transitions to properly offset local time on the client-side
		$timeZone = $domain->getTimeZone() ?? $user->getDefaultTimeZone();

		$earliestTimestamp = (new DateTime('first day of last year 01:00:00', new \DateTimeZone('UTC')))->getTimestamp();
		$dstTransitions = array_map(
			function (array $transition) {
				// Convert to JavaScript units
				$transition['ts'] *= 1000;
				$transition['offset'] = (int)($transition['offset'] / -60);
				return array_intersect_key($transition, array_flip(['ts', 'offset']));
			},
			$timeZone->getTransitions($earliestTimestamp) ?: [['ts'=>$earliestTimestamp, 'offset'=>0]]
		);

		$response = $this->render('@plugin:Ccm19ScriptPlaceholder/frontend/module.js.twig', [
			'placeholders' => $placeholdersJsObject,
			'dstTransitions' => $dstTransitions,
		], new JavascriptResponse());

		if ($request->query->has('v')) {
			// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
			Utils::resetCacheControl($response);

			// Lebensdauer auf ein Jahr setzen, da ein Zeitstempel an die Widget-Route angehängt wird
			$response->headers->remove('Cache-Control');
			$response->setCache([
				'max_age' => 31536000,
				'immutable' => true,
				'public' => true,
			]);
		}

		// Robots sollen das Widget nicht crawlen
		$response->headers->set('X-Robots-Tag', 'noindex');

		if ($domain->isWidgetCachingEnabled() && $request->query->has('preview') == false) {
			Cache::storeResponse($request, $response, $domain);
		}

		return $response;
	}
}
