<?php

namespace Plugins\Ccm19ScriptPlaceholder\Controller;

use App\Component\LocaleService;
use App\Component\Menu\Annotation\Menu;
use App\Controller\DomainDependantController;
use App\Model\Embedding;
use App\Model\EmbeddingBase;
use Plugins\Ccm19ScriptPlaceholder\Model\Placeholder;
use Plugins\Ccm19ScriptPlaceholder\Model\Snippet;
use Plugins\Ccm19ScriptPlaceholder\Model\Template;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;


/**
 * @Route("/domains/{_domainId}/plugins/ccm19/script_placeholder/placeholder", name="plugin_ccm19_script_placeholder_placeholder_")
 */
class PlaceholderController extends DomainDependantController
{
	/**
	 * @Menu("Script placeholder", icon="fa-spinner", group="Plugins")
	 * @Route("", name="index", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function index(): Response
	{
		$placeholders = Placeholder::all();

		// ORDER BY placeholder>userPlaceholder, name
		usort($placeholders, function ($a, $b) {
			/**
			 * @var Placeholder $a
			 * @var Placeholder $b
			 */
			if ($a->referencesUserPlaceholder() xor $b->referencesUserPlaceholder()) {
				return $a->referencesUserPlaceholder() ? 1 : -1;
			}
			else {
				return strnatcasecmp($a->getName(), $b->getName());
			}
		});

		return $this->render('@plugin:Ccm19ScriptPlaceholder/placeholder/index.html.twig', [
			'placeholders' => $placeholders,
		]);
	}

	/**
	 * @Route("/templates", name="templates", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function templates(): Response
	{
		$templates = Template::all();

		// Templates nach ihrem Namen sortieren
		usort($templates, function (Template $a, Template $b) {
			return strnatcasecmp($a->getName(), $b->getName());
		});

		return $this->render('@plugin:Ccm19ScriptPlaceholder/placeholder/templates.html.twig', [
			'templates' => $templates,
		]);
	}

	/**
	 * @Route("/create", name="create", methods={"HEAD", "GET", "POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function create(Request $request, TranslatorInterface $translator): Response
	{
		if ($request->getMethod() == 'POST') {
			$placeholder = Placeholder::create();
			self::feedFormDataIntoPlaceholder($placeholder, $request);

			if ($this->validatePlaceholder($placeholder, $translator)) {
				Snippet::flush();
				$placeholder->save();

				$this->addFlash('success', $translator->trans('Item has been created.'));
				return $this->redirectToRoute('plugin_ccm19_script_placeholder_placeholder_index', [], 303);
			}
		}
		else {
			$placeholder = Placeholder::createNonPersistent();

			if ($template = Template::find($request->query->getAlnum('template'))) {
				$snippets = array_map(function (array $snippet) {
					return Snippet::createNonPersistent()
						->setName((string)($snippet['name'] ?? ''))
						->setTranslations((array)($snippet['translations'] ?? []));
				}, $template->getSnippets());

				$placeholder
					->setName($template->getName())
					->setPlacement($template->getPlacement())
					->setPlacementCssSelector($template->getPlacementCssSelector())
					->setPlacementScriptNeedle($template->getPlacementScriptNeedle())
					->setHtmlCode($template->getHtmlCode())
					->setCssCode($template->getCssCode())
					->setJsCode($template->getJsCode())
					->setSnippets($snippets)
					->setTimeConstraints($template->getTimeConstraints())
				;
			}
		}

		return $this->render('@plugin:Ccm19ScriptPlaceholder/placeholder/create.html.twig', [
			'placeholder' => $placeholder,
			'embeddings' => Embedding::optional(),
			'locales' => LocaleService::enabledLocales(),
		]);
	}

	/**
	 * @Route("/{id}", name="read", methods={"HEAD", "GET"})
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function read(string $id, TranslatorInterface $translator): Response
	{
		$placeholder = Placeholder::find($id);
		if (!$placeholder) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_placeholder_index');
		}

		if ($placeholder->isActive()) {
			if ($embedding = $placeholder->getEmbedding()) {
				if (!$embedding->isActive()) {
					$this->addFlash('warning', $translator->trans('The placeholder won\'t be displayed because the embedding is inactive.'));
				}
				elseif ($embedding->getExcludeUrlsMode() == EmbeddingBase::EXCLUDE_URLS_MODE_WHITELIST && empty($embedding->getExcludeUrls())) {
					$this->addFlash('warning', $translator->trans(
						'The placeholder won\'t be displayed because the embedding has defined mode ":mode" in conjunction with an empty filter list for the option ":excludeUrls".',
						[':mode'=>$translator->trans('Whitelist'), ':excludeUrls'=>$translator->trans('Exclude URLs')]
					));
				}
			}
			else {
				$this->addFlash('warning', $translator->trans('The placeholder can\'t be displayed without an embedding linked to it.'));
			}
		}

		return $this->render('@plugin:Ccm19ScriptPlaceholder/placeholder/read.html.twig', [
			'placeholder' => $placeholder,
			'embeddings' => Embedding::optional(),
			'locales' => LocaleService::enabledLocales(),
			'reducedViewForDeployedPlaceholder' => $placeholder->referencesUserPlaceholder(),
		]);
	}

	/**
	 * @Route("/{id}", name="update", methods={"PUT"})
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function update(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$placeholder = Placeholder::find($id);
		if (!$placeholder) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_placeholder_index');
		}

		self::feedFormDataIntoPlaceholder($placeholder, $request);

		if ($this->validatePlaceholder($placeholder, $translator)) {
			Snippet::flush();
			$placeholder->save();

			$this->addFlash('success', $translator->trans('Item has been updated.'));
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_placeholder_read', [
				'id' => $id,
			], 303);
		}

		return $this->render('@plugin:Ccm19ScriptPlaceholder/placeholder/read.html.twig', [
			'placeholder' => $placeholder,
			'embeddings' => Embedding::optional(),
			'locales' => LocaleService::enabledLocales(),
		]);
	}

	/**
	 * @Route("/{id}/delete", name="delete_confirm", methods={"HEAD", "GET"})
	 * @param string $id
	 * @return Response
	 */
	public function deleteConfirmDialog(string $id): Response
	{
		$placeholder = Placeholder::find($id);
		if (!$placeholder || $placeholder->referencesUserPlaceholder()) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_placeholder_index');
		}

		return $this->render('@plugin:Ccm19ScriptPlaceholder/placeholder/delete.html.twig', [
			'placeholder' => $placeholder,
		]);
	}

	/**
	 * @Route("/{id}", name="delete", methods={"DELETE"})
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function delete(string $id, TranslatorInterface $translator): Response
	{
		$placeholder = Placeholder::find($id);
		if (!$placeholder || $placeholder->referencesUserPlaceholder()) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_placeholder_index');
		}

		// Delete own snippets
		foreach ($placeholder->getSnippets() as $snippet) {
			Snippet::delete($snippet->getId(), false);
		}
		Snippet::flush();

		Placeholder::delete($id);

		$this->addFlash('success', $translator->trans('Item has been deleted.'));
		return $this->redirectToRoute('plugin_ccm19_script_placeholder_placeholder_index', [], 303);
	}

	/**
	 * @param Placeholder $placeholder
	 * @param TranslatorInterface $translator
	 * @return bool
	 */
	private function validatePlaceholder(Placeholder $placeholder, TranslatorInterface $translator): bool
	{
		$valid = true;

		if ($placeholder->getEmbedding() == null) {
			$this->addFlash('danger', $translator->trans('Select a value from the drop-down list ":field".', [
				':field' => $translator->trans('Embedding'),
			]));
			$valid = false;
		}

		if (strlen($placeholder->getName()) == 0) {
			$this->addFlash('danger', $translator->trans('Enter a value for the field ":field".', [
				':field' => $translator->trans('Name'),
			]));
			$valid = false;
		}

		// TODO timeFrom < timeTo

		return $valid;
	}

	/**
	 * This method locks collections Placeholder and Snippet in this order; flush should occur in reverse.
	 * @param Placeholder $placeholder
	 * @param Request $request
	 * @return void
	 */
	private static function feedFormDataIntoPlaceholder(Placeholder $placeholder, Request $request): void
	{
		// Only update a selection of properties for deployed placeholders
		if ($placeholder->referencesUserPlaceholder()) {
			$placeholder
				->setEmbedding(Embedding::find((string)$request->request->get('embedding')))
				->setActive($request->request->getBoolean('active'))
				->setPlacementCssSelector((string)$request->request->get('placementCssSelector'))
				->setPlacementScriptNeedle((string)$request->request->get('placementScriptNeedle'))
				->setTimeConstraints((array)$request->request->get('timeConstraints'))
			;
			return;
		}

		$placeholder
			->setName((string)$request->request->get('name'))
			->setEmbedding(Embedding::find((string)$request->request->get('embedding')))
			->setActive($request->request->getBoolean('active'))
			->setPlacement((string)$request->request->get('placement'))
			->setPlacementCssSelector((string)$request->request->get('placementCssSelector'))
			->setPlacementScriptNeedle((string)$request->request->get('placementScriptNeedle'))
			->setHtmlCode((string)$request->request->get('htmlCode'))
			->setCssCode((string)$request->request->get('cssCode'))
			->setJsCode((string)$request->request->get('jsCode'))
			->setTimeConstraints((array)$request->request->get('timeConstraints'))
		;

		$enabledLocales = LocaleService::enabledLocales();

		$snippets = $placeholder->getSnippets();
		$snippets = array_combine(array_map(function (Snippet $snippet) { return $snippet->getId(); }, $snippets), $snippets);

		// Remove snippets
		foreach ($request->request->get('deleteSnippets', []) as $snippetId) {
			Snippet::delete((string)$snippetId, false);
			unset($snippets[$snippetId]);
		}

		// Update snippets
		foreach ($request->request->get('snippets', []) as $snippetId => $data) {
			if ($snippet = $snippets[$snippetId] ?? null) {
				$snippet->setName($data['name'] ?? '');

				foreach ($enabledLocales as $locale) {
					$snippet->setTranslation($locale, $data['text'][$locale->getName()] ?? '');
				}
			}
		}

		// Create snippets
		foreach ($request->request->get('addSnippets', []) as $data) {
			if (strlen(trim($data['name'] ?? '')) == 0) {
				break;
			}

			$snippet = Snippet::create()
				->setName((string)$data['name'])
			;

			foreach ($enabledLocales as $locale) {
				$snippet->setTranslation($locale, $data['text'][$locale->getName()] ?? '');
			}

			$snippets[] = $snippet;
		}

		$placeholder->setSnippets($snippets);
	}

	/**
	 * @Route("/{id}/copy", methods={"POST"}, name="copy")
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function copyEditForm(string $id, TranslatorInterface $translator): Response
	{
		$placeholder = Placeholder::find($id);
		$newPlaceholder = clone $placeholder;

		$this->addFlash('success', $translator->trans('Placeholder ":name" has been copied.', [
			':name' => $placeholder->getName(),
		]));

		return $this->redirectToRoute('plugin_ccm19_script_placeholder_placeholder_index', [], 303);
	}

}
