<?php

namespace Plugins\Ccm19ScriptPlaceholder\Controller;

use App\Component\LocaleService;
use App\Component\Menu\Annotation\Menu;
use App\Controller\AbstractController;
use App\Model\Domain;
use App\Model\Model;
use App\Model\User;
use App\Model\UserLocale;
use Plugins\Ccm19ScriptPlaceholder\Component\Deployment;
use Plugins\Ccm19ScriptPlaceholder\Model\Template;
use Plugins\Ccm19ScriptPlaceholder\Model\UserPlaceholder;
use Plugins\Ccm19ScriptPlaceholder\Model\UserSnippet;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;


/**
 * @Route("/plugins/ccm19/script_placeholder/user_placeholder", name="plugin_ccm19_script_placeholder_user_placeholder_")
 */
class UserPlaceholderController extends AbstractController
{
	/**
	 * @Menu("Script placeholder", icon="fa-spinner", group="Plugins")
	 * @Route("", name="index", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function index(): Response
	{
		return $this->render('@plugin:Ccm19ScriptPlaceholder/user_placeholder/index.html.twig', [
			'placeholders' => UserPlaceholder::all(),
		]);
	}

	/**
	 * @Route("/templates", name="templates", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function templates(): Response
	{
		$templates = Template::all();

		// Templates nach ihrem Namen sortieren
		usort($templates, function (Template $a, Template $b) {
			return strnatcasecmp($a->getName(), $b->getName());
		});

		return $this->render('@plugin:Ccm19ScriptPlaceholder/user_placeholder/templates.html.twig', [
			'templates' => $templates,
		]);
	}

	/**
	 * @Route("/create", name="create", methods={"HEAD", "GET", "POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function create(Request $request, TranslatorInterface $translator): Response
	{
		/** @var UserPlaceholder|null $placeholder */
		$placeholder = null;

		if ($request->getMethod() == 'POST') {
			$placeholder = UserPlaceholder::create();
			self::feedFormDataIntoPlaceholder($placeholder, $request);

			if ($this->validatePlaceholder($placeholder, $translator)) {
				UserSnippet::flush();
				$placeholder->save();

				$this->addFlash('success', $translator->trans('Item has been created.'));
				return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index', [], 303);
			}
		}
		elseif ($template = Template::find($request->query->getAlnum('template'))) {
			$snippets = array_map(function (array $snippet) {
				return UserSnippet::createNonPersistent()
					->setName((string)($snippet['name'] ?? ''))
					->setTranslations((array)($snippet['translations'] ?? []));
			}, $template->getSnippets());

			$placeholder = UserPlaceholder::createNonPersistent()
				->setName($template->getName())
				->setPlacement($template->getPlacement())
				->setPlacementCssSelector($template->getPlacementCssSelector())
				->setPlacementScriptNeedle($template->getPlacementScriptNeedle())
				->setHtmlCode($template->getHtmlCode())
				->setCssCode($template->getCssCode())
				->setJsCode($template->getJsCode())
				->setSnippets($snippets)
				->setTimeConstraints($template->getTimeConstraints())
			;
		}

		return $this->render('@plugin:Ccm19ScriptPlaceholder/user_placeholder/create.html.twig', [
			'placeholder' => $placeholder ?? UserPlaceholder::createNonPersistent(),
			'locales' => self::getEnabledUserLocalesOrFallback(),
		]);
	}

	/**
	 * @Route("/{id}", name="read", methods={"HEAD", "GET"})
	 * @param string $id
	 * @return Response
	 */
	public function read(string $id): Response
	{
		$placeholder = UserPlaceholder::find($id);
		if (!$placeholder) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index');
		}

		return $this->render('@plugin:Ccm19ScriptPlaceholder/user_placeholder/read.html.twig', [
			'placeholder' => $placeholder,
			'locales' => self::getEnabledUserLocalesOrFallback(),
		]);
	}

	/**
	 * @Route("/{id}", name="update", methods={"PUT"})
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function update(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$placeholder = UserPlaceholder::find($id);
		if (!$placeholder) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index');
		}

		self::feedFormDataIntoPlaceholder($placeholder, $request);

		if ($this->validatePlaceholder($placeholder, $translator)) {
			UserSnippet::flush();
			$placeholder->save();

			$this->addFlash('success', $translator->trans('Item has been updated.'));
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_read', [
				'id' => $id,
			], 303);
		}

		return $this->render('@plugin:Ccm19ScriptPlaceholder/user_placeholder/read.html.twig', [
			'placeholder' => $placeholder,
			'locales' => self::getEnabledUserLocalesOrFallback(),
		]);
	}

	/**
	 * @Route("/{id}/delete", name="delete_confirm", methods={"HEAD", "GET"})
	 * @param string $id
	 * @return Response
	 */
	public function deleteConfirmDialog(string $id): Response
	{
		$placeholder = UserPlaceholder::find($id);
		if (!$placeholder) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index');
		}

		return $this->render('@plugin:Ccm19ScriptPlaceholder/user_placeholder/delete.html.twig', [
			'placeholder' => $placeholder,
		]);
	}

	/**
	 * @Route("/{id}", name="delete", methods={"DELETE"})
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function delete(string $id, TranslatorInterface $translator): Response
	{
		$placeholder = UserPlaceholder::find($id);
		if (!$placeholder) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index');
		}

		// Delete own snippets
		foreach ($placeholder->getSnippets() as $snippet) {
			UserSnippet::delete($snippet->getId(), false);
		}
		UserSnippet::flush();

		UserPlaceholder::delete($id);

		$this->addFlash('success', $translator->trans('Item has been deleted.'));
		return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index', [], 303);
	}

	/**
	 * @Route("/deploy", name="deploy", methods={"POST"})
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deploy(TranslatorInterface $translator): Response
	{
		if (UserPlaceholder::isEmpty()) {
			return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index', [], 303);
		}

		$user = User::loggedInUser();
		assert($user instanceof User);
		$accountOwner = $user->getAccountOwner();

		// Switch to account owner context to process every domain
		User::setApiUser($accountOwner->getApiKey());
		Model::closeUserModels();
		Domain::switchToUser($accountOwner);

		// Deploy placeholders to each domain
		Deployment::publishUserPlaceholders();

		// Switch back to the original user context
		User::setApiUser($user->getApiKey());
		Model::closeUserModels();
		Domain::switchToUser($user);

		$this->addFlash('success', $translator->trans('Items have been deployed.'));
		return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index', [], 303);
	}

	/**
	 * @param UserPlaceholder $placeholder
	 * @param TranslatorInterface $translator
	 * @return bool
	 */
	private function validatePlaceholder(UserPlaceholder $placeholder, TranslatorInterface $translator): bool
	{
		$valid = true;

		if (strlen($placeholder->getName()) == 0) {
			$this->addFlash('danger', $translator->trans('Enter a value for the field ":field".', [
				':field' => $translator->trans('Name'),
			]));
			$valid = false;
		}

		// TODO timeFrom < timeTo

		return $valid;
	}

	/**
	 * This method locks collections UserPlaceholder and UserSnippet in this order; flush should occur in reverse.
	 * @param UserPlaceholder $placeholder
	 * @param Request $request
	 * @return void
	 */
	private static function feedFormDataIntoPlaceholder(UserPlaceholder $placeholder, Request $request): void
	{
		$placeholder
			->setName((string)$request->request->get('name'))
			->setPlacement((string)$request->request->get('placement'))
			->setPlacementCssSelector((string)$request->request->get('placementCssSelector'))
			->setPlacementScriptNeedle((string)$request->request->get('placementScriptNeedle'))
			->setHtmlCode((string)$request->request->get('htmlCode'))
			->setCssCode((string)$request->request->get('cssCode'))
			->setJsCode((string)$request->request->get('jsCode'))
			->setTimeConstraints((array)$request->request->get('timeConstraints'))
		;

		$enabledLocales = self::getEnabledUserLocalesOrFallback();

		$snippets = $placeholder->getSnippets();
		$snippets = array_combine(array_map(function (UserSnippet $snippet) { return $snippet->getId(); }, $snippets), $snippets);

		// Remove snippets
		foreach ($request->request->get('deleteSnippets', []) as $snippetId) {
			UserSnippet::delete((string)$snippetId, false);
			unset($snippets[$snippetId]);
		}

		// Update snippets
		foreach ($request->request->get('snippets', []) as $snippetId => $data) {
			if ($snippet = $snippets[$snippetId] ?? null) {
				$snippet->setName($data['name'] ?? '');

				foreach ($enabledLocales as $locale) {
					$snippet->setTranslation($locale, $data['text'][$locale->getName()] ?? '');
				}
			}
		}

		// Create snippets
		foreach ($request->request->get('addSnippets', []) as $data) {
			if (strlen(trim($data['name'] ?? '')) == 0) {
				break;
			}

			$snippet = UserSnippet::create()
				->setName((string)$data['name'])
			;

			foreach ($enabledLocales as $locale) {
				$snippet->setTranslation($locale, $data['text'][$locale->getName()] ?? '');
			}

			$snippets[] = $snippet;
		}

		$placeholder->setSnippets($snippets);
	}

	/**
	 * @return UserLocale[]
	 */
	private static function getEnabledUserLocalesOrFallback(): array
	{
		return UserLocale::enabledLocales() ?: [LocaleService::getUserDefault()];
	}

	/**
	 * @Route("/{id}/copy", methods={"POST"}, name="copy")
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function copyEditForm(string $id, TranslatorInterface $translator): Response
	{
		$userPlaceholder = UserPlaceholder::find($id);
		$newUserPlaceholder = clone $userPlaceholder;

		$this->addFlash('success', $translator->trans('Placeholder ":name" has been copied.', [
			':name' => $userPlaceholder->getName(),
		]));

		return $this->redirectToRoute('plugin_ccm19_script_placeholder_user_placeholder_index', [], 303);
	}
}
