<?php

namespace Plugins\Ccm19ScriptPlaceholder\EventSubscriber;

use App\Event\DomainClonePrepareEvent;
use App\Event\DomainCloneReferencesEvent;
use App\Event\DomainDeletePrepareEvent;
use App\Event\DomainInitializedEvent;
use App\Event\UserDeletePrepareEvent;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\Model;
use App\Model\PluginState;
use Plugins\Ccm19ScriptPlaceholder\Component\Deployment;
use Plugins\Ccm19ScriptPlaceholder\Model\Placeholder;
use Plugins\Ccm19ScriptPlaceholder\Model\Snippet;
use Plugins\Ccm19ScriptPlaceholder\Model\UserPlaceholder;
use Plugins\Ccm19ScriptPlaceholder\Model\UserSnippet;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

class DomainInitialization implements EventSubscriberInterface
{
	/** @var PluginState $pluginState */
	private $pluginState;

	public function __construct(PluginState $pluginState)
	{
		$this->pluginState = $pluginState;
	}

	public static function getSubscribedEvents(): array
	{
		return [
			DomainClonePrepareEvent::NAME => ['onDomainClonePrepare', 1000],
			DomainCloneReferencesEvent::NAME => ['onDomainCloneReferences', 1000],
			DomainDeletePrepareEvent::NAME => ['onDomainDeletePrepare', 1000],
			DomainInitializedEvent::NAME => ['onDomainInitialized', 1000],
			UserDeletePrepareEvent::NAME => ['onUserDeletePrepare', 1000],
		];
	}

	/**
	 * Causes the objects of certain models to be copied.
	 * @param DomainClonePrepareEvent $event
	 * @return void
	 */
	public function onDomainClonePrepare(DomainClonePrepareEvent $event): void
	{
		if ($this->pluginState->isAllowedForCurrentUser()) {
			$event
				->addDomainSpecificModel(Placeholder::class)
				->addDomainSpecificModel(Snippet::class)
			;
		}
	}

	/**
	 * Links placeholders with their respective embedding and snippets.
	 * @param DomainCloneReferencesEvent $event
	 * @return void
	 */
	public function onDomainCloneReferences(DomainCloneReferencesEvent $event): void
	{
		if (!$this->pluginState->isAllowedForCurrentUser()) {
			return;
		}

		// Remember the current domain context
		$requestDomainId = Domain::getDomainIdForRequest();

		// Switch to source's domain context
		Domain::setDomainForRequest($event->getSource()->getId());
		Model::closeDomainModels();

		$sourcePlaceholders = array_map(function (Placeholder $placeholder) {
			return [
				'embedding' => $placeholder->getEmbedding() ? $placeholder->getEmbedding()->getId() : null,
				'snippets' => array_map(function (Snippet $snippet) {
					return $snippet->getId();
				}, $placeholder->getSnippets()),
			];
		}, Placeholder::all());

		// Switch to target's domain context
		Domain::setDomainForRequest($event->getTarget()->getId());
		Model::closeDomainModels();

		$embeddings = Embedding::all();
		$placeholders = Placeholder::all();
		$snippets = Snippet::all();

		$embeddingMap = $event->getIdMapping(Embedding::class);
		$placeholderMap = $event->getIdMapping(Placeholder::class);
		$snippetMap = $event->getIdMapping(Snippet::class);

		// Link correct embedding and snippets to each placeholder
		foreach ($sourcePlaceholders as $sourcePlaceholderId => $sourcePlaceholder) {
			if ($placeholder = $placeholders[$placeholderMap[$sourcePlaceholderId] ?? ''] ?? null) {
				$placeholder
					->setEmbedding($embeddings[$embeddingMap[$sourcePlaceholder['embedding']] ?? ''] ?? null)
					->setSnippets(array_filter(array_map(function (string $id) use ($snippets, $snippetMap) {
						return $snippets[$snippetMap[$id] ?? ''] ?? null;
					}, $sourcePlaceholder['snippets'])))
				;
			}
		}

		// Save changes
		Placeholder::flush();

		// Restore the previous domain context
		Domain::setDomainForRequest($requestDomainId);
		Model::closeDomainModels();
	}

	/**
	 * Causes the repositories of certain models to be cleared.
	 * @param DomainDeletePrepareEvent $event
	 * @return void
	 */
	public function onDomainDeletePrepare(DomainDeletePrepareEvent $event): void
	{
		$event
			->addDomainSpecificModel(Placeholder::class)
			->addDomainSpecificModel(Snippet::class)
		;
	}

	/**
	 * Adds published placeholders to new domains.
	 * @param DomainInitializedEvent $event
	 * @return void
	 */
	public function onDomainInitialized(DomainInitializedEvent $event): void
	{
		if ($this->pluginState->isAllowedForCurrentUser()) {
			Deployment::publishUserPlaceholdersToDomain($event->getDomain());
		}
	}

	/**
	 * Causes the repositories of certain models to be cleared.
	 * @param UserDeletePrepareEvent $event
	 * @return void
	 */
	public function onUserDeletePrepare(UserDeletePrepareEvent $event): void
	{
		$event
			->addUserSpecificModel(UserPlaceholder::class)
			->addUserSpecificModel(UserSnippet::class)
		;
	}
}
