<?php

namespace Plugins\Ccm19ScriptPlaceholder\EventSubscriber;

use App\Event\FilterEvent;
use Symfony\Contracts\EventDispatcher\Event;
use App\Model\PluginState;
use App\Model\Embedding;
use Plugins\Ccm19ScriptPlaceholder\Model\Placeholder;
use Plugins\Ccm19ScriptPlaceholder\Model\PlaceholderBase;
use Plugins\Ccm19ScriptPlaceholder\Model\Snippet;
use Plugins\Ccm19ScriptPlaceholder\Model\UserPlaceholder;
use Plugins\Ccm19ScriptPlaceholder\Model\UserSnippet;
use Plugins\Ccm19Imex\Event\ExportDomainEvent;
use Plugins\Ccm19Imex\Event\ImportDomainEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

class ImExSubscriber implements EventSubscriberInterface
{
	private $pluginState;

	/** @var array<string,string> */
	private $userPlaceholderMap;

	public function __construct(PluginState $state) {
		$this->pluginState = $state;
		$this->userPlaceholderMap = [];
	}

	/**
	 * @return array<string,array|string>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			'ccm19.plugin.ccm19imex.exportclient' => ['onExportClient'],
			'ccm19.plugin.ccm19imex.importclientpredomains' => ['onImportClientPreDomains'],
			'ccm19.plugin.ccm19imex.exportdomain' => ['onExport'],
			'ccm19.plugin.ccm19imex.importdomain' => ['onImport'],
		];
	}

	/**
	 * Accountweite Platzhalter exportieren
	 *
	 * @param \Plugins\Ccm19Imex\Event\ExportClientEvent $event
	 */
	public function onExportClient(FilterEvent $event): void
	{
		$event->set('ccm19ScriptPlaceholder', [
			'placeholders' => array_values(array_map(function (UserPlaceholder $placeholder) {
				return [
					'id' => $placeholder->getId(),
					'name' => $placeholder->getName(),
					'placement' => $placeholder->getPlacement(),
					'placementCssSelector' => $placeholder->getPlacementCssSelector(),
					'placementScriptNeedle' => $placeholder->getPlacementScriptNeedle(),
					'htmlCode' => $placeholder->getHtmlCode(),
					'cssCode' => $placeholder->getCssCode(),
					'jsCode' => $placeholder->getJsCode(),
					'snippets' => array_map(function (UserSnippet $snippet) {
						return [
							'name' => $snippet->getName(),
							'translations' => $snippet->getTranslations(),
						];
					}, $placeholder->getSnippets()),
					'timeConstraints' => $placeholder->getTimeConstraints(),
					'publishedPlacementCssSelector' => $placeholder->getPublishedPlacementCssSelector(),
					'publishedPlacementScriptNeedle' => $placeholder->getPublishedPlacementScriptNeedle(),
					'publishedTimeConstraints' => $placeholder->getPublishedTimeConstraints(),
				];
			}, UserPlaceholder::all())),
		]);
	}

	/**
	 * Accountweite Platzhalter importieren
	 *
	 * @param \Plugins\Ccm19Imex\Event\ImportClientPreDomainsEvent $event
	 */
	public function onImportClientPreDomains(Event $event): void
	{
		$data = $event->get('ccm19ScriptPlaceholder');
		if (!$data) {
			return;
		}

		UserPlaceholder::clear(false);
		UserSnippet::clear(false);
		foreach ($data['placeholders'] ?? [] as $placeholderData) {
			$snippets = array_map(function (array $snippet) {
				return UserSnippet::create()
					->setName((string)($snippet['name'] ?? ''))
					->setTranslations((array)($snippet['translations'] ?? []));
			}, $placeholderData['snippets'] ?? []);

			$placeholder = UserPlaceholder::create()
				->setName((string)($placeholderData['name'] ?? ''))
				->setPlacement((string)($placeholderData['placement'] ?? ''))
				->setPlacementCssSelector((string)($placeholderData['placementCssSelector'] ?? ''))
				->setPlacementScriptNeedle((string)($placeholderData['placementScriptNeedle'] ?? ''))
				->setHtmlCode((string)($placeholderData['htmlCode'] ?? ''))
				->setCssCode((string)($placeholderData['cssCode'] ?? ''))
				->setJsCode((string)($placeholderData['jsCode'] ?? ''))
				->setSnippets($snippets)
				->setTimeConstraints((array)($placeholderData['timeConstraints'] ?? []))
				->setPublishedPlacementCssSelector($placeholderData['publishedPlacementCssSelector'] ?? null)
				->setPublishedPlacementScriptNeedle($placeholderData['publishedPlacementScriptNeedle'] ?? null)
				->setPublishedTimeConstraints($placeholderData['publishedTimeConstraints'] ?? null)
			;

			$this->userPlaceholderMap[$placeholderData['id'] ?? ''] = $placeholder->getId();
		}
		UserSnippet::flush();
		UserPlaceholder::flush();
	}

	/**
	 * Platzhalter der Domain exportieren
	 *
	 * @param ExportDomainEvent $event
	 */
	public function onExport(FilterEvent $event): void
	{
		if (!$event->has('embeddings')) {
			return;
		}

		$exportData = $event->getReturn();

		// Platzhalter nach Embeddings ordnen
		$placeholdersByEmbedding = [];
		foreach (Placeholder::all() as $placeholder) {
			$embedding = $placeholder->getEmbedding();
			if (!$embedding) {
				continue;
			}
			$embeddingId = $embedding->getId();

			if (!isset($placeholdersByEmbedding[$embeddingId])) {
				$placeholdersByEmbedding[$embeddingId] = [$placeholder];
			} else {
				$placeholdersByEmbedding[$embeddingId][] = $placeholder;
			}
		}

		// Platzhalter in den Embedding-Export einfügen
		foreach ($exportData['embeddings'] as &$embedding) {
			$placeholders = $placeholdersByEmbedding[$embedding['id']] ?? null;
			if (!$placeholders) {
				continue;
			}
			$embedding['scriptPlaceholders'] = array_map(function (Placeholder $placeholder) {
				return [
					'name' => $placeholder->getName(),
					'userPlaceholder' => $placeholder->getUserPlaceholder() ? $placeholder->getUserPlaceholder()->getId() : null,
					'active' => $placeholder->isActive(),
					'placement' => $placeholder->getPlacement(),
					'placementCssSelector' => $placeholder->getPlacementCssSelector(),
					'placementScriptNeedle' => $placeholder->getPlacementScriptNeedle(),
					'htmlCode' => $placeholder->getHtmlCode(),
					'cssCode' => $placeholder->getCssCode(),
					'jsCode' => $placeholder->getJsCode(),
					'snippets' => array_map(function (Snippet $snippet) {
						return [
							'name' => $snippet->getName(),
							'translations' => $snippet->getTranslations()
						];
					}, $placeholder->getSnippets()),
					'timeConstraints' => $placeholder->getTimeConstraints(),
				];
			}, $placeholders);
		}
		unset($embedding);

		$event->setReturn($exportData);
	}

	/**
	 * Platzhalter in die Domain importieren
	 *
	 * @param ImportDomainEvent $event
	 */
	public function onImport(Event $event): void
	{
		$embeddings = array_filter($event->get('embeddings'), function (array $embedding) {
			return isset($embedding['scriptPlaceholders']);
		});

		if (!$embeddings) {
			return;
		}
		$idMapping = $event->getEmbeddingMap();
		if (!$idMapping) {
			return;
		}

		if ($event->isNewDomain()) {
			Placeholder::clear();
			Snippet::clear();
		}

		$existingPlaceholders = Placeholder::all();

		foreach ($embeddings as $emData) {
			$embeddingId = $idMapping[$emData['id']] ?? '';
			$embedding = Embedding::find($embeddingId);
			if (!$embedding) {
				continue;
			}

			// Neue Einträge erstellen
			foreach ($emData['scriptPlaceholders'] as $phData) {
				// Ggf. alten Eintrag löschen
				$existingPlaceholders = array_filter($existingPlaceholders, function ($placeholder) use ($embeddingId, $phData) {
					$embedding = $placeholder->getEmbedding();
					if ($embedding and $embedding->getId() === $embeddingId) {
						if ($placeholder->getName() === ($phData['name'] ?? '')) {
							$snippets = $placeholder->getSnippets();
							Placeholder::delete($placeholder->getId());
							foreach ($snippets as $snippet) {
								Snippet::delete($snippet->getId());
							}
							return false;
						}
					}
					return true;
				});

				$placeholder = Placeholder::create()
					->setName($phData['name'] ?? '')
					->setEmbedding($embedding)
					->setActive((bool)($phData['active'] ?? false))
					->setPlacement($phData['placement'] ?? PlaceholderBase::PLACEMENT_BODY_END)
					->setPlacementCssSelector($phData['placementCssSelector'] ?? '')
					->setPlacementScriptNeedle($phData['placementScriptNeedle'] ?? '')
					->setHtmlCode($phData['htmlCode'] ?? '')
					->setCssCode($phData['cssCode'] ?? '')
					->setJsCode($phData['jsCode'] ?? '')
					->setTimeConstraints((array)($phData['timeConstraints'] ?? []))
				;

				$snippets = array_map(function (array $snData) {
					return Snippet::create()
						->setName($snData['name'] ?? '')
						->setTranslations((array)($snData['translations'] ?? []))
					;
				}, $phData['snippets'] ?? []);

				$placeholder->setSnippets($snippets);

				// Check if an imported user placeholder matches, alternatively look for an existing one.
				if ($userPlaceholder = UserPlaceholder::find($this->userPlaceholderMap[$phData['userPlaceholder'] ?? ''] ?? $phData['userPlaceholder'] ?? '')) {
					$placeholder->setUserPlaceholder($userPlaceholder);
				}
			}
		}
		Snippet::flush();
		Placeholder::flush();
	}
}
