<?php

namespace Plugins\Ccm19ScriptPlaceholder\Model;

use App\Component\Config\RepositoryInterface;
use App\Model\Embedding;
use App\Model\LocaleBase;

class Placeholder extends PlaceholderBase
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = true;

	/** @var Embedding|null $embedding */
	private $embedding;

	/** @var Snippet[] $snippets */
	private $snippets;

	/** @var UserPlaceholder|null $userPlaceholder */
	private $userPlaceholder;

	/**
	 * @return self[]
	 */
	public static function enabled(): array
	{
		return array_filter(self::all(), function (self $placeholder) {
			return $placeholder->isActive();
		});
	}

	/**
	 * @return self[]
	 */
	public static function enabledAndViable(): array
	{
		return array_filter(self::enabled(), function (self $placeholder) {
			return $placeholder->getEmbedding()
				&& $placeholder->getEmbedding()->isActive()
				&& $placeholder->getEmbedding()->isOptional()
				&& strlen($placeholder->getHtmlCode()) > 0;
		});
	}

	/**
	 * @param UserPlaceholder $placeholder
	 * @return static|null
	 */
	public static function byUserPlaceholder(UserPlaceholder $placeholder): ?self
	{
		return array_reduce(self::all(), function (?self $carry, self $item) use ($placeholder) {
			return $carry ?? ($item->getUserPlaceholder() && $item->getUserPlaceholder()->getId() == $placeholder->getId() ? $item : $carry);
		}, null);
	}

	/**
	 * @return UserPlaceholder|null
	 */
	public function getUserPlaceholder(): ?UserPlaceholder
	{
		if ($this->userPlaceholder === null) {
			$this->userPlaceholder = UserPlaceholder::find((string)$this->get('userPlaceholder'));
		}
		return $this->userPlaceholder;
	}

	/**
	 * @param UserPlaceholder|null $placeholder
	 * @return $this
	 */
	public function setUserPlaceholder(?UserPlaceholder $placeholder): self
	{
		$this->userPlaceholder = $placeholder;
		$this->set('userPlaceholder', $placeholder ? $placeholder->getId() : null);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function referencesUserPlaceholder(): bool
	{
		return (bool)$this->get('userPlaceholder');
	}

	/**
	 * @return bool Whether the property `userPlaceholder` points to a non-existing object.
	 */
	public function isDangling(): bool
	{
		return $this->referencesUserPlaceholder() && $this->getUserPlaceholder() === null;
	}

	/**
	 * @return Embedding|null
	 */
	public function getEmbedding(): ?Embedding
	{
		$this->embedding = $this->embedding ?? Embedding::find((string)$this->get('embedding'));

		// The purpose could have changed after the placeholder was last saved
		return $this->embedding && $this->embedding->isOptional()
			? $this->embedding
			: null;
	}

	/**
	 * @param Embedding|null $embedding
	 * @return $this
	 */
	public function setEmbedding(?Embedding $embedding): self
	{
		$this->embedding = $embedding;
		$this->set('embedding', $embedding ? $embedding->getId() : null);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isActive(): bool
	{
		return (bool)$this->get('active');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setActive(bool $state): self
	{
		$this->set('active', $state);
		return $this;
	}

	/**
	 * @return Snippet[]
	 */
	public function getSnippets(): array
	{
		if (empty($this->snippets)) {
			$this->snippets = array_values(array_filter(array_map(function ($id) {
				return Snippet::find($id);
			}, $this->get('snippets', []))));
		}

		return $this->snippets;
	}

	/**
	 * @param Snippet[] $snippets
	 * @return $this
	 */
	public function setSnippets(array $snippets): self
	{
		// Remove duplicates
		$snippets = array_values(array_reduce($snippets, function (array $snippets, Snippet $snippet) {
			$snippets[$snippet->getId()] = $snippet;
			return $snippets;
		}, []));

		$ids = array_map(function (Snippet $snippet) {
			return $snippet->getId();
		}, $this->snippets = $snippets);

		$this->set('snippets', $ids);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @return string
	 */
	public function getHtmlCodeWithSnippetsReplaced(LocaleBase $locale): string
	{
		return array_reduce($this->getSnippets(), function (string $html, Snippet $snippet) use ($locale) {
			return str_replace("{{$snippet->getName()}}", htmlspecialchars($snippet->getTranslation($locale)), $html);
		}, $this->getHtmlCode());
	}

	/**
	 * @return string
	 */
	public function getCssCodeWithPlaceholderScope(): string
	{
		$cssCode = $this->getCssCode();

		$hideout = [];

		// Mask css comments to facilitate further changes
		$cssCode = preg_replace_callback(
			'~/\*((?:[^*]*)\*(?(?=/)/|(?1)))~',
			function ($match) use (&$hideout) {
				$needle = '@comment--'.uniqid().';';
				$hideout[$needle] = $match[0];
				return $needle;
			},
			$cssCode
		);

		// Mask keyframes to prevent inner rules from being modified
		$cssCode = preg_replace_callback(
			'~@keyframes\s[^{]+(?<rule>(?>[^{}]*+)(?(?={){(?&rule)*}))~i',
			function ($match) use (&$hideout) {
				$needle = '@at-rule--'.uniqid().';';
				$hideout[$needle] = $match[0];
				return $needle;
			},
			$cssCode
		);

		// Modify every css selector to achieve a scope limited to the current placeholder
		$scope = '.ccm-plugin--script-placeholder[data-placeholder="'.addcslashes($this->getId(), '"').'"]';
		$cssCode = preg_replace_callback(
			'~(?<pre>(?<mq>@media\s[^{]++)?(?(mq){|(?:^|\*/|}))\s*+)(?<sel>(?(R&sel),|)\s*+(?!\*/)[a-z.#+\~:*[][^,{}]*+(?(?=})!|)(?&sel)*)~i',
			function ($match) use ($scope) {
				return $match['pre'].$scope.' '.str_replace(',', ", $scope ", $match['sel']);
			},
			$cssCode
		);

		// Restore masked code to the previous state
		return str_replace(array_reverse(array_keys($hideout)), array_reverse($hideout), $cssCode);
	}

	/**
	 * @param string $name
	 * @return self|null
	 * @phan-side-effect-free
	 */
	public static function byName(string $name): ?self
	{
		$name = mb_strtolower($name);
		return array_reduce(self::all(), function ($carry, $placeholder) use ($name) {
			return $carry ?? (mb_strtolower($placeholder->getName()) == $name ? $placeholder : $carry);
		}, null);
	}

	/**
	 * Übernimmt Eigenschaften des ursprünglichen Objekts aus dem Repository.
	 * Fordert exklusiven Lock an und gibt abschließend wieder frei.
	 */
	public function __clone()
	{
		// Generiere neue ID und stelle ursprünglichen Platzhalter in $this->source bereit
		parent::__clone();

		// Generiere aufsteigende Nummerierung für neuen Namen
		for ($i = 2; ; $i++) {
			if (!self::byName($newPlaceholderName = "{$this->source->getName()} #$i")) {
				break;
			}
		}

		// Klone Snippets für das neue Platzhalter-Objekt
		$newSnippets = array_map(static function (Snippet $snippet) {
			return Snippet::create()
				->setName($snippet->getName())
				->setTranslations($snippet->getTranslations());
		}, $this->source->getSnippets());
		Snippet::flush();

		// Neue Ausangswerte vorgeben und Platzhalter speichern
		$this
			->setName($newPlaceholderName)
			->setActive(false)
			->setEmbedding(null)
			->setUserPlaceholder(null)
			->setSnippets($newSnippets)
			->save();
	}
}
