<?php

namespace Plugins\Ccm19ScriptPlaceholder\Model;

use App\Model\Model;
use LogicException;

abstract class PlaceholderBase extends Model
{
	const PLACEMENT_BODY_END = 'body-end';
	const PLACEMENT_CSS_SELECTOR = 'css-selector';
	const PLACEMENT_CSS_SELECTOR_INSERT_BEFORE = 'css-selector-pre';
	const PLACEMENT_SCRIPT_TAGS = 'script-tags';

	const ALLOWED_PLACEMENT_VALUES = [
		self::PLACEMENT_BODY_END,
		self::PLACEMENT_CSS_SELECTOR,
		self::PLACEMENT_CSS_SELECTOR_INSERT_BEFORE,
		self::PLACEMENT_SCRIPT_TAGS,
	];

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'plugin-ccm19ScriptPlaceholder-placeholders';

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setName(string $name): self
	{
		$this->set('name', trim($name));
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPlacement(): string
	{
		return (string)$this->get('placement');
	}

	/**
	 * @param string $placement
	 * @return $this
	 * @throws LogicException
	 */
	public function setPlacement(string $placement): self
	{
		if (in_array($placement, self::ALLOWED_PLACEMENT_VALUES) == false) {
			throw new LogicException('Invalid placement, allowed values: ' . implode(', ', self::ALLOWED_PLACEMENT_VALUES));
		}

		$this->set('placement', $placement);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPlacementCssSelector(): string
	{
		return (string)$this->get('placementCssSelector');
	}

	/**
	 * @param string $selector
	 * @return $this
	 */
	public function setPlacementCssSelector(string $selector): self
	{
		$this->set('placementCssSelector', $selector);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPlacementScriptNeedle(): string
	{
		return (string)$this->get('placementScriptNeedle');
	}

	/**
	 * @param string $needle
	 * @return $this
	 */
	public function setPlacementScriptNeedle(string $needle): self
	{
		$this->set('placementScriptNeedle', $needle);
		return $this;
	}

	/**
	 * @return SnippetBase[]
	 */
	public abstract function getSnippets(): array;

	/**
	 * @param SnippetBase[] $snippets
	 * @return $this
	 */
	public abstract function setSnippets(array $snippets);

	/**
	 * @return string
	 */
	public function getHtmlCode(): string
	{
		return (string)$this->get('htmlCode');
	}

	/**
	 * @param string $htmlCode
	 * @return $this
	 */
	public function setHtmlCode(string $htmlCode): self
	{
		$this->set('htmlCode', $htmlCode);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCssCode(): string
	{
		return (string)$this->get('cssCode');
	}

	/**
	 * @param string $cssCode
	 * @return $this
	 */
	public function setCssCode(string $cssCode): self
	{
		$this->set('cssCode', $cssCode);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getJsCode(): string
	{
		return (string)$this->get('jsCode');
	}

	/**
	 * @param string $jsCode
	 * @return $this
	 */
	public function setJsCode(string $jsCode): self
	{
		$this->set('jsCode', $jsCode);
		return $this;
	}

	public function getTimeConstraints(): array
	{
		$constraints = $this->get('timeConstraints');
		return is_array($constraints) ? array_values($constraints) : [];
	}

	/**
	 * @param string[][] $constraints
	 * @return $this
	 */
	public function setTimeConstraints(array $constraints): self
	{
		$mandatoryKeys = [
			'dayFrom', 'dayTo', 'timeFrom', 'timeTo',
		];
		$validKeys = array_merge($mandatoryKeys, [
		]);

		// Strip unexpected data
		$constraints = array_map(function (array $constraint) use ($validKeys) {
			return array_intersect_key($constraint, array_flip($validKeys));
		}, $constraints);

		// Make sure that all mandatory keys exist
		$valid = array_reduce($constraints, function (bool $valid, array $constraint) use ($mandatoryKeys) {
			return $valid && count($mandatoryKeys) == count(array_intersect_key(array_flip($mandatoryKeys), $constraint));
		}, true);

		if (!$valid) {
			throw new LogicException(sprintf(
				"Invalid time constraints. (Mandatory keys: %s) (Allowed keys: %s)"
				, implode(', ', $mandatoryKeys)
				, implode(', ', $validKeys)
			));
		}

		array_walk($constraints, function (array &$constraint) {
			$constraint['dayFrom'] = max(1, min((int)$constraint['dayFrom'], 7));
			$constraint['dayTo'] = max(1, min((int)$constraint['dayTo'], 7));
			$constraint['timeFrom'] = preg_match('~^\d\d?:\d\d$~', $constraint['timeFrom']) ? $constraint['timeFrom'] : '00:00';
			$constraint['timeTo'] = preg_match('~^\d\d?:\d\d$~', $constraint['timeTo']) ? $constraint['timeTo'] : '23:59';
		});

		$this->set('timeConstraints', array_values($constraints));
		return $this;
	}
}
