<?php

namespace Plugins\Ccm19SingleSignOn\Component;

use App\Component\Curl;
use Plugins\Ccm19SingleSignOn\Exception\AuthException;
use Plugins\Ccm19SingleSignOn\Model\Provider;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;
use DateTimeImmutable;

class OpenId
{
	/** @var string|null $extraState */
	private $extraState = null;

	/** @var Provider|null $provider */
	private $provider = null;

	/** @var string|null $token */
	private $token = null;

	/** @var DateTimeImmutable|null $tokenExpires */
	private $tokenExpires = null;

	/** @var array|null $userInfo */
	private $userInfo = null;

	/** @var UrlGeneratorInterface $urlGenerator */
	private $urlGenerator;

	public function __construct(UrlGeneratorInterface $urlGenerator, ?Provider $provider = null)
	{
		$this->urlGenerator = $urlGenerator;
		if ($provider) {
			$this->setProvider($provider);
		}
	}

	/**
	 * Provider holen
	 * @return Provider|null
	 */
	public function getProvider(): Provider
	{
		return $this->provider;
	}

	/**
	 * Provider festlegen
	 *
	 * @param Provider $provider
	 * @return $this
	 */
	public function setProvider(Provider $provider): self
	{
		$this->provider = $provider;
		return $this;
	}

	/**
	 * Subject-ID (User-ID auf Seiten des Providers)
	 *
	 * Nur nach erfolgreichem `authenticate()` nutzbar
	 *
	 * @return string|null
	 */
	public function getSubjectId(): ?string
	{
		return ($this->userInfo ?? [])['sub'] ?? null;
	}

	/**
	 * E-Mail-Adresse
	 *
	 * Nur nach erfolgreichem `authenticate()` nutzbar
	 *
	 * @return string|null
	 */
	public function getEmailAddress(): ?string
	{
		return ($this->userInfo ?? [])['email'] ?? null;
	}

	/**
	 * Access-Token
	 *
	 * @return string|null
	 */
	public function getAccessToken(): ?string
	{
		return $this->token;
	}

	/**
	 * Access-Token
	 *
	 * @return DateTimeImmutable|null
	 */
	public function getAccessTokenExpiration(): ?DateTimeImmutable
	{
		return $this->tokenExpires;
	}

	/**
	 * Access-Token manuell setzen
	 *
	 * @param string|null $token
	 * @param DateTimeImmutable|null $expires
	 * @return $this
	 */
	public function setAccessToken(?string $token, ?DateTimeImmutable $expires = null): self
	{
		$this->token = $token;
		$this->tokenExpires = $expires;
		return $this;
	}

	public function reset(): self
	{
		$this->provider = null;
		$this->extraState = null;
		$this->token = null;
		$this->tokenExpires = null;
		$this->userInfo = null;
		return $this;
	}

	/**
	 * Extra-State nach der Anmeldung holen, der ursprünglich bei authenticate() angegeben wurde.
	 *
	 * @return string|null
	 */
	public function getExtraState(): ?string
	{
		return $this->extraState;
	}

	/**
	 * Führt eine Authentifizierung über OAuth/OpenID Connect durch
	 *
	 * Dabei wird der Browser zuerst redirected und dann die Redirect-URI aufgerufen.
	 * Bei der Redirect-URI muss wieder authenticate() mit dem exakt gleichen Parameter aufgerufen werden.
	 *
	 * Bei Erfolg kehrt die Funktion mit `null` zurück. Ist noch ein Authentifizierungsschritt
	 * nötig, wird eine RedirectResponse zurückgegeben.
	 *
	 * @param Request $request
	 * @param string $redirectUrl
	 * @param string $extraState
	 * @param string|null $prompt 'none'|'login'|'consent'|'select_account'
	 * @param string|null $loginHint Username/Mailadresse falls bekannt
	 * @throws AuthException
	 * @return RedirectResponse|null
	 */
	public function authenticate(Request $request, string $redirectUrl, string $extraState='', ?string $prompt = null, ?string $loginHint=null): ?RedirectResponse
	{
		$session = $request->getSession();

		// Schritt 1
		if ($request->query->get('state', '') === '') {
			// Zufalls-State generieren und in Session speichern
			$state = bin2hex(random_bytes(16));
			$states = $session->get('oauth2_state', []);
			$states[$this->provider->getId()] = $state;
			$session->set('oauth2_state', $states);

			$extraState = $this->provider->getId().'.'.$extraState;

			$data = [
				'response_type' => 'code',
				'client_id' => $this->provider->getClientId(),
				'redirect_uri' => $redirectUrl,
				'scope' => implode(' ', $this->provider->getScopes()),
				'state' => hash_hmac('sha256', $extraState, $state).'.'.$extraState,
			];
			if ($prompt) {
				$data['prompt'] = $prompt;
			}
			if ($loginHint) {
				$data['login_hint'] = $loginHint;
			}

			// Workaround für Microsoft...
			if (stripos($this->provider->getAuthEndpoint(), 'login.microsoftonline.com') !== false) {
				unset($data['login_hint']);
			}

			// Weiterleiten
			return new RedirectResponse($this->provider->getAuthEndpoint().'?'.http_build_query($data), 303);
		}

		// Schritt 2

		// State überprüfen
		$states = $session->get('oauth2_state', []);

		// State dekodieren
		$stateData = explode('.', $request->query->get('state'), 2);
		$stateHash = $stateData[0];
		$stateData = $stateData[1] ?? '';
		if (strpos($stateData, '.') === false) {
			throw new AuthException('invalid state', 422);
		}

		// Provider-ID rausholen
		list($providerId, $extraState) = explode('.', $stateData, 2);

		// State verifizieren
		$state = $states[$providerId] ?? '';
		if (!hash_equals(hash_hmac('sha256', $stateData, $state), $stateHash) or $state === '') {
			throw new AuthException('invalid state', 422);
		}

		// Provider holen
		$this->provider = Provider::find($providerId);
		if (!$this->provider) {
			throw new AuthException('provider unknown', 404);
		}
		$this->extraState = $extraState;

		// Token holen
		$url = $this->provider->getTokenEndpoint();
		$response = $this->post($url, [
			'grant_type'=>'authorization_code',
			'code' => $_GET['code'] ?? '',
			'client_id'=>$this->provider->getClientId(),
			'client_secret'=>$this->provider->getClientSecret(),
			'redirect_uri'=>$redirectUrl,
		]);

		$code = $response->getStatusCode();
		$json = @json_decode($response->getContent(), true, 16);
		if (!$json) {
			throw new AuthException('token exchange failed', $code);
		}
		if (isset($json['error'])) {
			$error = $json['error'];
			if (is_string($error)) {
				$errormsg = $error;
			} elseif (isset($error['message'])) {
				$errormsg = $error['message'];
				if (isset($error['code'])) {
					$code = $error['code'];
				}
			} else {
				$errormsg = (string)$error;
			}
			throw new AuthException($errormsg, $code);
		}
		if (!isset($json['access_token'])) {
			throw new AuthException('acquired no access token', 502);
		}

		$this->token = $json['access_token'];
		$this->tokenExpires = new DateTimeImmutable('@'.(time()+$json['expires_in']));

		// User-Daten holen
		$url = $this->provider->getProfileEndpoint();
		$response = $this->get($url);
		$json = @json_decode($response->getContent(), true, 16);
		if (!$response->isOk() or !$json or !isset($json['sub'])) {
			throw new AuthException('could not access user information', 502);
		}
		$this->userInfo = $json;

		return null;
	}

	private function post(string $endpoint, array $data = []): Response
	{
		$curl = new Curl();
		$curl->setOpt(CURLOPT_CONNECTTIMEOUT, 10);
		$curl->setOpt(CURLOPT_TIMEOUT, 11);
		$curl->setOpt(CURLOPT_SSL_VERIFYHOST, 2);
		$curl->setOpt(CURLOPT_USERAGENT, 'Mozilla/5.0 (compatible; CCM19; +http://ccm19.de)');
		if ($this->token !== null) {
			$curl->setHeader('Authorization', 'Bearer '.$this->token);
		}
		$curl->post($endpoint, $data);
		return $curl->getResponseObject();
	}

	private function get(string $endpoint, array $query = []): Response
	{
		if ($query) {
			$endpoint .= '?'.http_build_query($query);
		}
		$curl = new Curl();
		$curl->setOpt(CURLOPT_CONNECTTIMEOUT, 10);
		$curl->setOpt(CURLOPT_TIMEOUT, 11);
		$curl->setOpt(CURLOPT_USERAGENT, 'Mozilla/5.0 (compatible; CCM19; +http://ccm19.de)');
		if ($this->token !== null) {
			$curl->setHeader('Authorization', 'Bearer '.$this->token);
		}
		$curl->get($endpoint);
		return $curl->getResponseObject();
	}

	/**
	 * Beendet die Session eines Access-Tokens
	 *
	 * @return void
	 */
	public function revokeToken(string $accessToken)
	{
		$url = $this->provider->getEndSessionEndpoint();
		if ($url) {
			$this->post($url, ['token' => $accessToken]);
		}
	}

}
