<?php

namespace Plugins\Ccm19SingleSignOn\Controller;

use App\Component\Color;
use App\Component\Curl;
use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Controller\AdminController;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\PluginState;
use Plugins\Ccm19SingleSignOn\Model\Provider;
use Plugins\Ccm19SingleSignOn\Model\AuthLink;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use UnexpectedValueException;

/**
 * Controller for SSO Administration Backend
 * @Route("/plugins/sso/admin/", name="plugin_ccm19_singlesignon_admin_")
 */
class Admin extends AdminController
{
	/**
	 * @Route("", name="index", methods={"HEAD", "GET"})
	 * @Menu("Single Sign-On", icon="fa-sign-in", order=60)
	 * @return Response
	 */
	public function index(PluginState $state): Response
	{
		$providers = Provider::all();

		$counts = array_map(function (Provider $provider) {
			return count(AuthLink::byProvider($provider));
		}, $providers);

		return $this->render('@plugin:Ccm19SingleSignOn/internal/admin/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'backTo' => null,
			'providers' => $providers,
			'counts' => $counts,
		]);
	}

	/**
	 * @Route("provider/", name="provider_list", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function indexRedirect(): Response
	{
		return $this->redirectToRoute('plugin_ccm19_singlesignon_admin_index');
	}

	/**
	 * @Route("provider/new", name="provider_new_form", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function createProviderForm(): Response
	{
		$providers = Provider::all();

		$formData = [
			'name' => '',
			'scopes' => ['openid', 'email'],
			'buttonTextColor' => '#243140',
			'buttonBackgroundColor' => '#DBDBDB',
			'icon' => '',
		];

		return $this->render('@plugin:Ccm19SingleSignOn/internal/admin/new.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'backTo' => null,
			'formData' => $formData,
		]);
	}

	/**
	 * @Route("provider/new", name="provider_new", methods={"POST"})
	 * @return Response
	 */
	public function createProvider(Request $request): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
			if ($request->request->getAlnum('action') === 'create') {
				$provider = Provider::create();
				$active = $request->request->getBoolean('active');
				$scopes = $request->request->get('scopes');
				if (!is_array($scopes)) {
					$scopes = explode(' ', trim(preg_replace('/\s{2,}/', ' ', $scopes)));
					$request->request->set('scopes', $scopes);
				}

				try {
					$buttonTextColor = Color::fromCssString((string)$request->request->get('buttonTextColor'));
				} catch (UnexpectedValueException $e) {
					$buttonTextColor = Color::fromRGB(0x243140);
				}
				try {
					$buttonBackgroundColor = Color::fromCssString((string)$request->request->get('buttonBackgroundColor'));
				} catch (UnexpectedValueException $e) {
					$buttonBackgroundColor = Color::fromRGB(0xDBDBDB);
				}

				$provider
					->setName((string)$request->request->get('name', ''))
					->setActive($active)
					->setAuthEndpoint((string)$request->request->get('authEndpoint'))
					->setTokenEndpoint((string)$request->request->get('tokenEndpoint'))
					->setProfileEndpoint((string)$request->request->get('profileEndpoint'))
					->setClientId((string)$request->request->get('clientId'))
					->setClientSecret((string)$request->request->get('clientSecret'))
					->setIcon((string)$request->request->get('icon'))
					->setButtonTextColor($buttonTextColor)
					->setButtonBackgroundColor($buttonBackgroundColor)
					->setScopes($scopes);

				if ($active and ($provider->getClientId() === '' or $provider->getClientSecret() === '')) {
					$provider->setActive(false);
					$this->addFlash('danger', 'No credentials set. Provider cannot be set to active without credentials.');
				}

				$provider->save();
				$this->addFlash('success', 'Provider has been created.');
				return $this->redirectToRoute('plugin_ccm19_singlesignon_admin_index');
			}
		}
		catch (ConfigNotWritableException $exception) {
			$this->addFlash('danger', 'Changes could not be written to disk.');
		}
		catch (CsrfTokenException $exception) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		return $this->render('@plugin:Ccm19SingleSignOn/internal/admin/new.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'formData' => $request->request->all(),
		]);
	}

	/**
	 * @Route("provider/{provider}", name="provider_edit_form", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function editProviderForm(Provider $provider): Response
	{
		return $this->render('@plugin:Ccm19SingleSignOn/internal/admin/edit.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'provider' => $provider,
			'formData' => $provider,
		]);
	}

	/**
	 * @Route("provider/{provider}", name="provider_edit", methods={"PUT"})
	 * @return Response
	 */
	public function editProvider(Request $request, Provider $provider): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
			$active = $request->request->getBoolean('active');
			$scopes = $request->request->get('scopes');
			if (!is_array($scopes)) {
				$scopes = explode(' ', trim(preg_replace('/\s{2,}/', ' ', $scopes)));
				$request->request->set('scopes', $scopes);
			}

			try {
				$buttonTextColor = Color::fromCssString((string)$request->request->get('buttonTextColor'));
			} catch (UnexpectedValueException $e) {
				$buttonTextColor = $provider->getButtonTextColor();
			}
			try {
				$buttonBackgroundColor = Color::fromCssString((string)$request->request->get('buttonBackgroundColor'));
			} catch (UnexpectedValueException $e) {
				$buttonBackgroundColor = $provider->getButtonBackgroundColor();
			}

			$provider
				->setName((string)$request->request->get('name', ''))
				->setActive($active)
				->setAuthorityUrl((string)$request->request->get('authorityUrl'))
				->setAuthEndpoint((string)$request->request->get('authEndpoint'))
				->setTokenEndpoint((string)$request->request->get('tokenEndpoint'))
				->setProfileEndpoint((string)$request->request->get('profileEndpoint'))
				->setEndSessionEndpoint((string)$request->request->get('endSessionEndpoint'))
				->setClientId((string)$request->request->get('clientId'))
				->setClientSecret((string)$request->request->get('clientSecret'))
				->setIcon((string)$request->request->get('icon'))
				->setButtonTextColor($buttonTextColor)
				->setButtonBackgroundColor($buttonBackgroundColor)
				->setScopes($scopes);

			if ($active and ($provider->getClientId() === '' or $provider->getClientSecret() === '')) {
				$provider->setActive(false);
				$this->addFlash('danger', 'No credentials set. Provider cannot be set to active without credentials.');
			}

			$provider->save();
			$this->addFlash('success', 'Settings have been saved.');
			return $this->redirectToRoute('plugin_ccm19_singlesignon_admin_index');
		}
		catch (ConfigNotWritableException $exception) {
			$this->addFlash('danger', 'Changes could not be written to disk.');
		}
		catch (CsrfTokenException $exception) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		return $this->render('@plugin:Ccm19SingleSignOn/internal/admin/edit.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'provider' => $provider,
			'formData' => $provider,
		]);
	}

	/**
	 * @Route("provider/{provider}/delete", name="provider_delete_form", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function deleteProviderForm(Provider $provider): Response
	{
		$providers = Provider::all();

		return $this->render('@plugin:Ccm19SingleSignOn/internal/admin/delete.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'provider' => $provider,
		]);
	}

	/**
	 * @Route("provider/{provider}", name="provider_delete", methods={"DELETE"})
	 * @return Response
	 */
	public function deleteProvider(Request $request, Provider $provider): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);

			// Alle Verbindungen zu dem Provider löschen
			foreach (AuthLink::byProvider($provider) as $authLink) {
				AuthLink::delete($authLink->getId(), false);
			}
			AuthLink::flush();

			// Provider selbst löschen
			Provider::delete($provider->getId());

			$this->addFlash('success', 'Provider was successfully deleted.');
			return $this->redirectToRoute('plugin_ccm19_singlesignon_admin_index');
		}
		catch (ConfigNotWritableException $exception) {
			$this->addFlash('danger', 'Changes could not be written to disk.');
		}
		catch (CsrfTokenException $exception) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		return $this->redirectToRoute('plugin_ccm19_singlesignon_admin_index');
	}

	/**
	 * Kürzt den Pfad einer URL um ein Verzeichnis.
	 *
	 * Gibt einen leeren String bei absoluten Parse-Fehlern zurück,
	 * ansonsten die neue URL ohne / am Ende.
	 */
	private function urlDirname(string $url): string
	{
		$parts = parse_url($url);
		if (!$parts) {
			return '';
		}

		$path = $parts['path'] ?? '';
		$path = explode('/', trim($path, '/'));
		array_pop($path);
		if ($path) {
			$path = '/' . implode('/', $path);
		} else {
			$path = '';
		}

		if (isset($parts['host'])) {
			$result = ($parts['scheme'] ?? 'https').'://'.$parts['host'];
			if (!empty($parts['port'])) {
				$result .= ':'.$port;
			}
			return $result . $path;
		} else {
			return ($path === '') ? '/' : $path;
		}
	}

	/**
	 * @Route("oidc_discovery", name="oidc_discovery", methods={"POST"})
	 * @return JsonResponse
	 */
	public function autodiscovery(Request $request): JsonResponse
	{
		try {
			CsrfToken::get()->checkRequest($request);
		}
		catch (CsrfTokenException $exception) {
			return $this->json([
				'success' => false,
				'error' => 'CSRF token invalid',
			], 403);
		}

		$name = (string)$request->request->get('name');
		$baseUrl = (string)$request->request->get('authorityUrl');
		$authEndpoint = (string)$request->request->get('authEndpoint');

		// Authority-Base-URL raten, wenn nicht gegeben
		if ($baseUrl === '' and $authEndpoint) {
			$baseUrl = $this->urlDirname($authEndpoint);
		}
		if ($baseUrl === '' and $name) {
			if (stripos($name, 'google') !== false) {
				$baseUrl = 'https://accounts.google.com/';
			}
			elseif (stripos($name, 'azure') !== false or stripos($name, 'microsoft') !== false) {
				$baseUrl = 'https://login.microsoftonline.com/common/v2.0/';
			}
		}

		// Abbrechen, wenn nichts sinnvolles ermittelt werden kann
		if ($baseUrl === '') {
			return $this->json([
				'success' => false,
				'error' => 'no authority url provided or inferrable',
			], 422);
		}
		elseif (stripos($baseUrl, 'https://') !== 0 and stripos($baseUrl, 'http://') !== 0) {
			return $this->json([
				'success' => false,
				'error' => 'authority url invalid',
			], 422);
		}

		// Versuchen die Detect-URL zu laden, dabei bis zum Domain-Root runtergehen
		$curl = new Curl();
		$curl->setOpt(CURLOPT_CONNECTTIMEOUT, 5);
		$curl->setOpt(CURLOPT_TIMEOUT, 6);
		$curl->setOpt(CURLOPT_USERAGENT, 'Mozilla/5.0 (compatible; CCM19; +http://ccm19.de)');

		$baseUrl = rtrim($baseUrl, '/');
		for ($i = 0; $i < 8; ++$i) {
			$url = $baseUrl . '/.well-known/openid-configuration';

			$curl->get($url);

			if ($curl->http_status_code >= 200 and $curl->http_status_code <= 203) {
				// JSON verarbeiten
				$data = @json_decode($curl->response, true, 32);
				if (!is_array($data)) {
					continue;
				}
				if (!isset($data['token_endpoint'])) {
					continue;
				}

				$data['authorityUrl'] = $baseUrl;
				$data['success'] = true;
				return $this->json($data);
			}

			$newBaseUrl = $this->urlDirname($baseUrl);
			if ($newBaseUrl !== $baseUrl)	{
				$baseUrl = $newBaseUrl;
			} else {
				break;
			}
		}

		// Wenn nichts gefunden, Fehler zurückliefern
		return $this->json([
			'success' => false,
			'error' => 'no OpenId Connect Discovery file found',
		], 404);
	}
}
