<?php

namespace Plugins\Ccm19SingleSignOn\Controller;

use App\Component\LoggerInterface;
use App\Controller\Login as LoginController;
use App\Controller\UnrestrictedController;
use App\Exception\UserIsDeactivatedException;
use App\Exception\SecondFactorRequiredException;
use App\Event\LoginEvent;
use App\Model\CsrfToken;
use App\Model\LoginAttemptsLog;
use App\Model\PluginState;
use App\Model\User;
use App\Utils;
use Plugins\Ccm19SingleSignOn\Component\OpenId;
use Plugins\Ccm19SingleSignOn\Exception\AuthException;
use Plugins\Ccm19SingleSignOn\Model\Provider;
use Plugins\Ccm19SingleSignOn\Model\AuthLink;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Controller for SSO Login
 * @Route("/login/sso", name="app_login_plugin_ccm19_singlesignon")
 */
class Login extends UnrestrictedController
{
	// Maximale Zeit zum Eingeben des Passworts um eine OAuth-Verknüpfung zu bestätigen
	const CONNECT_TIMEOUT_SECONDS = 900;

	public function __construct(LoginController $mainLoginController, LoggerInterface $logger, TranslatorInterface $translator, PluginState $plugin)
	{
		$this->loginController = $mainLoginController;
		$this->logger = $logger;
		$this->translator = $translator;
		$this->plugin = $plugin;
	}

	/**
	 * Fortsetzung eines OAuth/OpenID-Login-Vorgangs
	 *
	 * Je nach Ergebnis wird ein Fehler angezeigt, die Anmeldung beendet oder wieder umgeleitet.
	 *
	 * @Route("/continue", name="_continue", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function continuationPoint(Request $request, OpenId $openId): Response
	{
		$openId->reset();
		$redirectUrl = $this->generateUrl('app_login_plugin_ccm19_singlesignon_continue', [], UrlGeneratorInterface::ABSOLUTE_URL);

		try {
			$redirect = $openId->authenticate($request, $redirectUrl, '');
		}
		catch (AuthException $e) {
			return $this->render('login/index.html.twig', [
				'error' => $this->translator->trans('Authentication failed: :error', [':error' => htmlspecialchars($e->getMessage())]),
				'username' => null,
				'backTo' => $request->query->get('backTo', null),
			], new Response('', 502));
		}

		if ($redirect) {
			return $redirect;
		} else {
			return $this->finishAuthentication($request, $openId);
		}
	}

	/**
	 * @Route("/relogin-done", name="_relogin_done", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function reloginDone(): Response
	{
		$response = $this->render('@plugin:Ccm19SingleSignOn/relogin_done.html.twig');
		Utils::resetCacheControl($response);
		$response->setCache(['max_age'=>180, 'public'=>true]);
		return $response;
	}

	/**
	 * Start-Punkt eines OAuth/OpenID-Login-Vorgangs
	 *
	 * Löst i.d.R. ein Redirect zum Anbieter aus,
	 * der dann später zum continuationPoint zurückleitet.
	 *
	 * @Route("/{provider}", name="", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function entryPoint(Request $request, Provider $provider, OpenId $openId): Response
	{
		$openId->reset();
		$openId->setProvider($provider);
		$redirectUrl = $this->generateUrl('app_login_plugin_ccm19_singlesignon_continue', [], UrlGeneratorInterface::ABSOLUTE_URL);
		$resumeToken = (string)$request->query->get('resumeToken', '');
		$extraData = ((string)$request->query->get('backTo', '')).'.'.$resumeToken.'..'.((string)$request->query->get('switchTo', ''));

		$prompt = ($resumeToken === '') ? null : 'login';
		try {
			$redirect = $openId->authenticate($request, $redirectUrl, $extraData, $prompt, $request->query->get('loginHint'));
		}
		catch (AuthException $e) {
			return $this->render('login/index.html.twig', [
				'error' => $this->translator->trans('Authentication failed: :error', [':error' => htmlspecialchars($e->getMessage())]),
				'username' => null,
				'backTo' => $request->query->get('backTo', null),
			], new Response('', 502));
		}
		if ($redirect) {
			return $redirect;
		} else {
			return $this->finishAuthentication($request, $openId);
		}
	}

	/**
	 * Bestätigungs-Formular, um einen User mit einem SSO-Account zu verknüpfen
	 *
	 * Wird aufgerufen, wenn die E-Mail-Adresse übereinstimmte, aber noch keine
	 * explizite Verknüpfung eingerichtet war.
	 *
	 * @Route("/{provider}/connect", name="_connect", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function connectAccountForm(Request $request, Provider $provider): Response
	{
		$session = $request->getSession();
		$ts = $session->get('oauthTimestamp');
		$token = $session->get('oauthToken');
		$sub = $session->get('oauthSubjectId');
		$providerId = $session->get('oauthProviderId');
		$email = $session->get('oauthEmail');
		$userId = $session->get('oauthUserId');

		// Wenn ohne gültige Session oder nach zu langer Zeit aufgerufen: abbrechen
		if ($ts === null or (time() - $ts) > self::CONNECT_TIMEOUT_SECONDS) {
			return $this->redirectToRoute('app_login', [], 302);
		}

		// Provider und User holen und ggf. abbrechen
		$user = User::byId($userId);
		if (!$provider or !$user or $provider->getId() !== $providerId) {
			return $this->redirectToRoute('app_login', [], 302);
		}
		// Wenn nicht für den User erlaubt, abbrechen
		if (!$this->plugin->isAllowedForUser($user)) {
			return $this->redirectToRoute('app_login', [], 302);
		}

		return $this->render('@plugin:Ccm19SingleSignOn/connect.html.twig', [
			'user' => $user,
			'provider' => $provider,
			'email' => $email,
			'backTo' => $request->query->get('backTo', null),
		]);
	}

	/**
	 * @Route("/{provider}/connect", name="_connect_cancel", methods={"DELETE"})
	 * @return Response
	 */
	public function connectAccountCancel(Request $request, Provider $provider, OpenId $openId): Response
	{
		$session = $request->getSession();

		$token = $session->get('oauthToken');
		$providerId = (string)$session->get('oauthProviderId');
		if ($provider->getId() == $providerId and $token) {
			$openId->reset();
			$openId->setProvider($provider);
			$openId->revokeToken($token);
		}

		$session->remove('oauthTimestamp');
		$session->remove('oauthToken');
		$session->remove('oauthSubjectId');
		$session->remove('oauthProviderId');
		$session->remove('oauthEmail');
		$session->remove('oauthUserId');

		return $this->redirectToRoute('app_login', [], 303);
	}

	/**
	 * @Route("/{provider}/connect", name="_connect_post", methods={"POST"})
	 * @return Response
	 */
	public function connectAccount(Request $request, Provider $provider, OpenId $openId): Response
	{
		$session = $request->getSession();
		$ts = $session->get('oauthTimestamp');
		$token = $session->get('oauthToken');
		$sub = $session->get('oauthSubjectId');
		$providerId = $session->get('oauthProviderId');
		$email = $session->get('oauthEmail');
		$userId = $session->get('oauthUserId');
		$error = null;

		// Wenn ohne gültige Session oder nach zu langer Zeit aufgerufen: abbrechen
		if ($ts === null or (time() - $ts) > self::CONNECT_TIMEOUT_SECONDS) {
			return $this->redirectToRoute('app_login', [], 302);
		}

		// Provider und User holen und ggf. abbrechen
		$user = User::byId($userId);
		if (!$provider or !$user or $provider->getId() !== $providerId) {
			return $this->redirectToRoute('app_login', [], 302);
		}
		// Wenn nicht für den User erlaubt, abbrechen
		if (!$this->plugin->isAllowedForUser($user)) {
			return $this->redirectToRoute('app_login', [], 302);
		}

		$password = (string)$request->request->get('password');

		if (($password === '' and !$user->hasPassword()) or $user->checkPassword((string)$request->request->get('password'))) {
			$authLink = AuthLink::createForSubject($provider, $sub);
			$authLink->setUser($user);
			$authLink->setEmailAddress($email);
			$authLink->setToken($token);
			$authLink->setActive(true);
			$authLink->touchCreationTimestamp()->touchLastUseTimestamp();
			$authLink->save();

			$session->remove('oauthTimestamp');
			$session->remove('oauthToken');
			$session->remove('oauthSubjectId');
			$session->remove('oauthProviderId');
			$session->remove('oauthEmail');
			$session->remove('oauthUserId');

			try {
				$this->doLogin($user);
				$session->set('loginViaSSO', $authLink->getId());
				$this->addFlash('success', $this->translator->trans('Your account at :provider was successfully connected to your CCM19 account.', [':provider' => htmlspecialchars($provider->getName())]));
				return $this->loginController->afterSuccessfulLogin($user, $request, $this->logger, $this->translator);
			}
			catch (UserIsDeactivatedException $exception) {
				$reason = $exception->getExtendedReason();
				if ($reason) {
					$error = $this->translator->trans($reason);
					$this->logger->logError("Login", 'Failed login due to a plugin rejecting the SSO login for user '.$user->getUsername().': '.$reason);
				} else {
					$error = $this->translator->trans('User is deactivated from logging into the system.');
					$this->logger->logError("Login", 'Failed SSO login due to deactivation for user '.$username);
				}
			}
			catch (SecondFactorRequiredException $exception) {
				$user = $exception->getUser();

				$error = $this->translator->trans('Two Factor Authentication required');
				$params = [
					'backTo' => $request->query->get('backTo')
				];

				$methods = $exception->getMethods();
				// Wenn nur eine Methode, dann direkt zu dieser springen
				if (count($methods) === 1) {
					$method = reset($methods);
					return $this->redirectToRoute($method['route'], array_merge($method['routeParams'] ?? [], $params), 303);
				} else {
					return $this->redirectToRoute('app_login_2fa', $params, 303);
				}
			}
		} else {
			$error = $this->translator->trans('Please check your credentials.');
		}

		return $this->render('@plugin:Ccm19SingleSignOn/connect.html.twig', [
			'error' => $error,
			'user' => $user,
			'provider' => $provider,
			'email' => $email,
			'backTo' => $request->query->get('backTo', null),
		]);
	}

	private function finishAuthentication(Request $request, OpenId $openId): Response
	{
		$provider = $openId->getProvider();
		$token = $openId->getAccessToken();
		$sub = $openId->getSubjectId();
		$email = $openId->getEmailAddress();

		$extraData = $openId->getExtraState();
		if ($extraData and strpos($extraData, '.') !== false) {
			$parts = explode('.', $extraData, 4);
			$backTo = $parts[0];
			$resumeToken = $parts[1];
			$connectUserId = $parts[2] ?? null;
			$switchTo = (string)($parts[3] ?? '');

			if ($backTo !== '') {
				$request->query->set('backTo', $backTo);
			}
			if ($resumeToken !== '') {
				$request->request->set('resumeToken', $resumeToken);
			}
			if ($switchTo !== '') {
				$request->request->set('switchTo', $switchTo);
			}
		} else {
			$resumeToken = '';
			$backTo = null;
			$connectUserId = null;
			$switchTo = '';
		}

		// Abbrechen, wenn mit ungültigem Zustand aufgerufen
		if ($provider === null or $token === null or $sub === null) {
			return $this->redirectToRoute('app_login', [], 303);
		}

		$authLink = AuthLink::bySubject($provider, $sub);
		$response = new Response();

		// Wenn neue Verknüpfung übers Backend
		if ($connectUserId !== null and $connectUserId !== '') {
			$user = User::byId($connectUserId);
			$loggedInUser = User::loggedInUser();
			if ($user and $loggedInUser and $connectUserId === $loggedInUser->getId()) {
				if (!$authLink or !$authLink->getUser() or $authLink->getUser()->getId() == $connectUserId) {
					$authLink = AuthLink::createForSubject($provider, $sub);
					$authLink->setUser($user);
					$authLink->setEmailAddress($email);
					$authLink->setToken($token);
					$authLink->setActive(true);
					$authLink->touchCreationTimestamp()->touchLastUseTimestamp();
					$authLink->save();
					$this->addFlash('success', 'Account was successfully connected.');
				} else {
					$this->addFlash('danger', 'This account is already connected to another CCM19 account.');
				}
			}
			else {
				$this->addFlash('danger', 'Account connection failed.');
			}
			return $this->loginController->redirectAfterLogin($request);
		}
		// Wenn User bestätigt verbunden, dann Anmeldung durchführen
		elseif ($authLink and $authLink->isActive() and $authLink->getUser() !== null) {
			$user = $authLink->getUser();
			if ($this->plugin->isAllowedForUser($user)) {
				try {
					$this->doLogin($user);
					$authLink->setToken($token)->touchLastUseTimestamp()->setEmailAddress($email)->save();
					$request->getSession()->set('loginViaSSO', $authLink->getId());
					return $this->loginController->afterSuccessfulLogin($user, $request, $this->logger, $this->translator);
				}
				catch (UserIsDeactivatedException $exception) {
					$reason = $exception->getExtendedReason();
					if ($reason) {
						$error = $this->translator->trans($reason);
						$this->logger->logError("Login", 'Failed login due to a plugin rejecting the SSO login for user '.$user->getUsername().': '.$reason);
						$response->setStatusCode(403);
					} else {
						$error = $this->translator->trans('User is deactivated from logging into the system.');
						$this->logger->logError("Login", 'Failed SSO login due to deactivation for user '.$username);
						$response->setStatusCode(403);
					}
				}
				catch (SecondFactorRequiredException $exception) {
					$user = $exception->getUser();

					// Session-Re-Login: auf 2FA verzichten
					if (hash_equals($user->getSessionResumeToken(), $resumeToken)) {
						$user->switchTo();
						return $this->loginController->afterSuccessfulLogin($user, $request, $this->logger, $this->translator);
					}

					$error = $this->translator->trans('Two Factor Authentication required');
					$params = [
						'backTo' => $backTo
					];

					$methods = $exception->getMethods();
					// Wenn nur eine Methode, dann direkt zu dieser springen
					if (count($methods) === 1) {
						$method = reset($methods);
						return $this->redirectToRoute($method['route'], array_merge($method['routeParams'] ?? [], $params), 303);
					} else {
						return $this->redirectToRoute('app_login_2fa', $params, 303);
					}
				}
			} else {
				$error = $this->translator->trans('User not allowed to use Single Sign-on.');
				$response->setStatusCode(403);
			}
		}
		// Matching per E-Mail-Adresse, um Verbinden zu erlauben, aber nur wenn kein Relogin
		else {
			$user = ($email) ? User::find($email) : null;
			if (!$user and $email) {
				$user = $this->findUserByMail($email);
			}
			if (($switchTo === '') and $user and $this->plugin->isAllowedForUser($user)) {
				$session = $request->getSession();
				$session->set('oauthTimestamp', (int)time());
				$session->set('oauthToken', $token);
				$session->set('oauthSubjectId', $sub);
				$session->set('oauthProviderId', $provider->getId());
				$session->set('oauthUserId', $user->getId());
				$session->set('oauthEmail', $email);
				return $this->redirectToRoute('app_login_plugin_ccm19_singlesignon_connect', ['provider'=>$provider->getId(), 'backTo'=>$backTo], 303);
			} else {
				$error = $this->translator->trans('User not found. Please connect your account with :provider (:email) in your account settings beforehand.', [':provider' => htmlspecialchars($provider->getName()), ':email'=>htmlspecialchars($email ?? '?')]);
				$response->setStatusCode(403);
			}
		}

		return $this->render('login/index.html.twig', [
			'error' => $error,
			'username' => ($user) ? $user->getUsername() : null,
			'backTo' => $request->query->get('backTo', null),
			'backToRoute' => $request->request->get('backToRoute', 'app_root'),
			'backToArgs' => $request->request->get('backToArgs', ''),
		], $response);
	}

	private function findUserByMail(string $email): ?User
	{
		$email = strtolower($email);
		foreach (User::all() as $user) {
			if ($user->getEmailAddress() and $email === strtolower($user->getEmailAddress())) {
				return $user;
			}
		}
		return null;
	}

	private function doLogin(User $user): void
	{
		$isBlocked = $user->isBlocked();
		//User ist inaktiv und ein CLient -> raus...
		if ($user->isActive() == false && $user->getRole() == User::ROLE_CLIENT) {
			throw new UserIsDeactivatedException($user->getUsername());
		}
		//Ist ein Subuser aber Hauptuser Account ist inaktiv - dann raus....
		elseif ($user->isSubUser() && $user->getAccountOwner()->isActive() == false) {
			throw new UserIsDeactivatedException($user->getUsername());
		}

		// Session-Resume-Token ggf. setzen
		if ($user->getSessionResumeToken() === null) {
			$user->generateSessionResumeToken();
		}

		$user
			->updateLastLogin()
			->save()
		;

		$event = new LoginEvent($user);
		Utils::dispatchEvent(LoginEvent::NAME, $event);

		if ($event->isDispositionAllow()) {
			$user->switchTo();
		} elseif ($event->isDispositionReject()) {
			throw new UserIsDeactivatedException($user->getUsername(), $event->getRejectReason());
		} else {
			$user->setLoginInProgress($event->getSecondFactorAuths());
			throw new SecondFactorRequiredException($user, $event->getSecondFactorAuths());
		}
	}
}
