<?php

namespace Plugins\Ccm19SingleSignOn\Model;

use App\Component\Config\RepositoryInterface;
use App\Model\Model;
use App\Model\User;
use DateTime;

class AuthLink extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'plugin-ccm19-sso-authlinks';

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = true;

	/**
	 * @param Provider $provider
	 * @param string $subjectId
	 * @return bool
	 */
	public static function subjectExists(Provider $provider, string $subjectId): bool
	{
		$id = $provider->getId().':'.$subjectId;
		return parent::exists($id);
	}

	/**
	 * @param Provider $provider
	 * @param string $subjectId
	 * @return static|null
	 */
	public static function bySubject(Provider $provider, string $subjectId): ?AuthLink
	{
		$id = $provider->getId().':'.$subjectId;
		return parent::find($id);
	}

	/**
	 * @param User $user
	 * @return self[]
	 */
	public static function byUser(User $user): array
	{
		$userId = $user->getId();
		return array_filter(self::all(), function ($authLink) use ($userId) {
			/** @var AuthLink $authLink */
			return $authLink->getUserId() === $userId;
		});
	}

	/**
	 * @param Provider $provider
	 * @return self[]
	 */
	public static function byProvider(Provider $provider): array
	{
		$providerId = $provider->getId();
		return array_filter(self::all(), function ($authLink) use ($providerId) {
			/** @var AuthLink $authLink */
			return $authLink->getProviderId() === $providerId;
		});
	}

	/**
	 * @param User $user
	 * @param User $user
	 * @return self[]
	 */
	public static function byProviderAndUser(Provider $provider, User $user): ?AuthLink
	{
		$providerId = $provider->getId();
		foreach (self::byUser($user) as $authLink) {
			if ($authLink->getProviderId() === $providerId) {
				return $authLink;
			}
		}
		return null;
	}

	/**
	 * @return static
	 * @override
	 * @phan-assert string $id
	 */
	public static function create(?string $id = null): Model
	{
		if ($id === null) {
			throw new LogicException('AuthLink ID may not be null.');
		}

		return new self($id);
	}

	/**
	 * @param ?string $localeName
	 * @return static
	 * @throws LocaleAlreadyExistsException
	 * @throws LogicException
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function createForSubject(Provider $provider, string $subjectId): AuthLink
	{
		$id = $provider->getId().':'.$subjectId;
		return static::create($id);
	}

	/**
	 * @return string
	 */
	public function getProviderId(): string
	{
		$id = $this->getId();
		return explode(':', $id, 2)[0];
	}

	/**
	 * @return Provider|null
	 */
	public function getProvider(): ?Provider
	{
		$id = $this->getProviderId();
		return Provider::find($id);
	}

	/**
	 * @return bool
	 */
	public function isActive(): bool
	{
		return (bool)$this->get('active');
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setActive(bool $value): self
	{
		$this->set('active', $value);
		return $this;
	}

	/**
	 * ID des Users beim Provider
	 * @return string
	 */
	public function getSubjectId(): string
	{
		$id = $this->getId();
		if (strpos($id, ':') === false) {
			return '';
		}
		return (string)explode(':', $id, 2)[1];
	}

	/**
	 * @return string
	 */
	public function getUserId(): string
	{
		return (string)$this->get('user');
	}

	/**
	 * @return User|null
	 */
	public function getUser(): ?User
	{
		return User::byId((string)$this->get('user'));
	}

	/**
	 * @param User $user
	 * @return $this
	 */
	public function setUser(User $user): self
	{
		$this->set('user', $user->getId());
		return $this;
	}

	/**
	 * @return DateTime
	 */
	public function getCreationTimestamp(): ?DateTime
	{
		$ts = $this->get('created');
		if ($ts) {
			return new DateTime((string)$ts);
		} else {
			return null;
		}
	}

	/**
	 * @return string|null
	 */
	public function getEmailAddress(): ?string
	{
		return $this->get('email');
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setEmailAddress(?string $value): self
	{
		$this->set('email', $value);
		return $this;
	}

	/**
	 * Zugriffs-Token.
	 *
	 * Kann mit "Authorization: Bearer $token" zur Authentifizierung
	 * verwendet werden.
	 *
	 * @return string
	 */
	public function getToken(): string
	{
		return (string)$this->get('token');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setToken(string $value): self
	{
		$this->set('token', $value);
		return $this;
	}

	/**
	 * @return $this
	 */
	public function touchCreationTimestamp(): self
	{
		$this->set('created', (new DateTime('now'))->format('Y-m-d H:i:s'));
		return $this;
	}

	/**
	 * @return DateTime
	 */
	public function getLastUseTimestamp(): ?DateTime
	{
		$ts = $this->get('lastUse');
		if ($ts) {
			return new DateTime((string)$ts);
		} else {
			return null;
		}
	}

	/**
	 * @return $this
	 */
	public function touchLastUseTimestamp(): self
	{
		$this->set('lastUse', (new DateTime('now'))->format('Y-m-d H:i:s'));
		return $this;
	}
}
