<?php

namespace Plugins\Ccm19Statistics\Controller;

use App\Component\HttpFoundation\JavascriptResponse;
use App\Controller\AbstractController;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Model;
use App\Model\StatisticsData;
use App\Utils;
use DateTime;
use Exception;
use Plugins\Ccm19Statistics\Model\StatisticsModell;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use App\Component\Menu\Annotation\Menu;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Controller for Statistics
 * @Route("/plugins/ccm19/statistics/user_statistics", name="plugin_ccm19_statistics_user_statistics_")
 */


class UserStatisticsController extends AbstractController
{
	/**
	 * @Menu("Erweiterte Statistik", icon="fa-bar-chart-o", group="Plugins")
	 * @Route("", name="index", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws Exception
	 */
	public function index(Request $request, TranslatorInterface $translator): Response
	{
		$barCharts = StatisticsModell::getBarCharts($translator);

		// Parameter bestimmen
		$valuemode = $request->query->getAlpha('valuemode', 'r');
		$showmode = $request->query->getAlpha('showmode', 'd');
		$range = preg_replace('/[^0-9. -]/', '', (string)$request->query->get('range', ''));
		if (!$range or strpos($range, ' - ') === false) {
			$start = new DateTime('- 2 months');
			$end = new DateTime('now');
			$range = $start->format('d.m.Y').' - '.$end->format('d.m.Y');
		}

		// Datums-Range aufspalten
		$dates = explode(" - ", $range, 2);

		// Render template
		return $this->render('@plugin:Ccm19Statistics/user_statistics/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'daterange' => $range,
			'startDate' => $dates[0],
			'endDate' => $dates[1],
			'maxDate' => date("d.m.Y"),
			'valuemode' => $valuemode,
			'showImportArchivLink'=>true,
			'showmode' => $showmode,
			'barCharts' => $barCharts,
			'statdat' => [
				'showmode' => $showmode,
				'valuemode' => $valuemode,
				'daterange' => $range,
			],
		]);
	}

	/**
	 * @Route("export", methods={"HEAD", "GET"}, name="export")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function jsonExport(Request $request, TranslatorInterface $translator): Response
	{
		$valuemode = trim((string)$request->query->get("valuemode", 'r'));
		$daterange = trim((string)$request->query->get("daterange", ''));
		$showmode = trim((string)$request->query->get("showmode", 'd'));
		$dates = explode(" - ", $daterange, 2);
		$startDate = $dates[0];
		$endDate = $dates[1];
		$startDate = date("Y-m-d", strtotime($startDate));
		$endDate = date("Y-m-d", strtotime($endDate));

		$valuemodeMapping = [
			"r" => 'absolute',
			"p" => 'percent',
		];
		$periodMapping = [
			"d" => 'days',
			"w" => 'weeks',
			"m" => 'months',
			"y" => 'years',
		];

		$statisticData = [
			'from' => $startDate,
			'to' => $endDate,
			'period' => $periodMapping[$showmode],
			'mode' => $valuemodeMapping[$valuemode],
		];

		// erstelle JSON für Liniendiagramm
		$timeline = $this->getJsonDataForTimeline($request, $translator);
		$statisticData['timeline'] = [];
		foreach ($timeline['x'] as $key => $date) {
			$dateFormat = "Y-m-d";
			if ($showmode === "w") {
				$dateFormat = "Y - W";
			} elseif ($showmode === "m") {
				$dateFormat = "Y-m";
			} elseif ($showmode === "y") {
				$dateFormat = "Y";
			}
			$date = date($dateFormat, strtotime($date));

			if ($valuemode == 'r') {
				$statisticData['timeline'][$date]['calls'] = $timeline[$translator->trans('Calls')][$key];
			}

			$statisticData['timeline'][$date][] = [
				'banner views' => $timeline[$translator->trans('Banner views')][$key],
				'acceptAll' => $timeline[$translator->trans('acceptAll')][$key],
				'decline' => $timeline[$translator->trans('decline')][$key],
				'leave' =>  $timeline[$translator->trans('leave')][$key]
			];
		}

		// erstelle JSON für Bar Charts
		$barCharts = StatisticsModell::getBarCharts($translator);
		foreach ($barCharts as $barChart) {
			$dataBarchart = $this->getJsonDataForBar($request, $translator, $barChart['name'], false);
			$statisticData[$barChart['name']] = $dataBarchart;
		}

		$response = new JsonResponse($statisticData);
		$response->headers->set('Content-Type', 'application/json');
		$response->headers->set('Content-Disposition','attachment; filename="jsonLog_'.time().'.json"');
		// Use the JSON_PRETTY_PRINT
		//$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );
		return $response;
	}

	/**
	 * @Route("statistics_basic.js", name="basic_chart_js"), methods={"HEAD","GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function makeBasicChart(Request $request, TranslatorInterface $translator): Response
	{
		$response = new JavascriptResponse();
		Utils::resetCacheControl($response);
		$response->setCache([
			'max_age'=>90,
			'public'=>false
		]);

		$statDat = array_filter($request->query->all(), function ($key) { return $key && $key[0] !== '_'  && $key !== 'chart'; }, ARRAY_FILTER_USE_KEY);

		if (!$statDat['showmode']) {
			$statDat['showmode'] = "d";
		}

		if ($statDat['showmode'] == "w") {
			$dateForm = "%Y - %W";
		}
		elseif ($statDat['showmode'] == "m") {
			$dateForm = "%Y-%m";
		}
		elseif($statDat['showmode'] =="y") {
			$dateForm = "%Y";
		}
		else /* "d" */ {
			$dateForm = "%Y-%m-%d";
		}

		if ($statDat['valuemode'] == "p") {
			$barHeadline = $translator->trans('Percental View of all Visits');
			$categories ="'".$translator->trans('Accept all')." %','".$translator->trans('Decline')." %','".$translator->trans('Leave')." %','".$translator->trans('All Visits')." %s'";
		}
		else {
			$barHeadline = $translator->trans('Real Counts of all visits');
			$categories = "'".$translator->trans('Accept all')."','".$translator->trans('Decline')."','".$translator->trans('Leave')."','".$translator->trans('All Visits')."'";
		}

		$responseArray = [
			"statdat" => $statDat,
			"chart" => $request->get("chart"),
			"dateForm" => $dateForm,
			"categories"=>$categories,
			"barHeadline" =>$barHeadline,
			"valuemode" => $statDat['valuemode']
		];

		return $this->render('@plugin:Ccm19Statistics/user_statistics/basic_chart.js.twig', $responseArray, $response);
	}


	/**
	 * @Route("statistics_basic2.js", name="basic_chart2_js"), methods={"HEAD","GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function makeBasicChart2(Request $request, TranslatorInterface $translator): Response
	{
		$response = new JavascriptResponse();
		Utils::resetCacheControl($response);

		$response->setCache([
			'max_age' => 90,
			'public' => false
		]);

		$statDat = array_filter($request->query->all(), function ($key) { return $key && $key[0] !== '_'  && $key !== 'chart'; }, ARRAY_FILTER_USE_KEY);

		if (!$statDat['showmode']) {
			$statDat['showmode'] = "d";
		}

		if ($statDat['showmode'] == "w") {
			$dateForm = "%Y - %W";
		}
		elseif ($statDat['showmode'] == "m") {
			$dateForm = "%Y-%m";
		}
		elseif($statDat['showmode'] =="y") {
			$dateForm = "%Y";
		}
		else /* "d" */ {
			$dateForm = "%Y-%m-%d";
		}

		if ($statDat['valuemode'] == "p") {
			$barHeadline = $translator->trans('Percental View of all Visits');
			$categories = "'".$translator->trans('Accept all')." %','".$translator->trans('Decline')." %','".$translator->trans('Leave')." %','".$translator->trans('All Visits')." %s'";
		}
		else{
			$barHeadline = $translator->trans('Real Counts of all visits');
			$categories = "'".$translator->trans('Accept all')."','".$translator->trans('Decline')."','".$translator->trans('Leave')."','".$translator->trans('All Visits')."'";
		}

		$responseArray = [
			"statdat" => $statDat,
			"chart" => $request->get("chart"),
			"dateForm" => $dateForm,
			"categories"=>$categories,
			"barHeadline" =>$barHeadline,
		];

		return $this->render('@plugin:Ccm19Statistics/user_statistics/basic_chart2.js.twig', $responseArray, $response);
	}

	/**
	 * @Route("basic_data_bar.json", name="basic_data_json_bar", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return JSONResponse
	 */
	public function showJsonDataBar(Request $request, TranslatorInterface $translator): JSONResponse
	{
		$all = $this->getJsonDataForBar($request, $translator);
		$response = new JsonResponse($all);

		//For Dev JSON_PRETTY_PRINT
		$response->setEncodingOptions($response->getEncodingOptions() | JSON_PRETTY_PRINT);

		//und raus...
		return $response;
	}

	/**
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return array
	 */
	public function getJsonDataForBar(Request $request, TranslatorInterface $translator, ?string $chartName = null, ?bool $getResultTranslated = true): array
	{
		//Daten aus Formular - enhält z.B. prozentual
		$daterange = (string)$request->query->get('daterange');
		$valuemode = $request->query->getAlpha('valuemode', 'r');

		$selDate = [
			'from' => time()-(60*60*24*60),
			'to' => time(),
		];
		if ($daterange) {
			$daDates = explode(' - ', $daterange, 2);
			if ($daDates) {
				$selDate['from'] = (int) strtotime($daDates[0]);
				$selDate['to'] = (int) (strtotime($daDates[1])+86600);
			}
		}

		//nur die Daten ab Mitte Februar - vorher nur crap...
		if ($selDate['from']<1613857795) {
			$selDate['from']=1613857795;
		}

		try {
			//die gesammentelten Daten kommen hierein
			$sumData = [
				'Calls' => 0
			];
			$buttonName = array("acceptAll","decline","opened");

			// Remember the current domain context
			$requestDomainId = Domain::getDomainIdForRequest();
			foreach (Domain::all() as $domain) {
				Domain::setDomainForRequest($domain->getId());
				Model::closeDomainModels();

				//Alle Datendurchgehen
				foreach (StatisticsData::all() as $dateKey => $stat) {
					//Wenn Datum eines Eintrages innerhalb des gewünschten Ranges liegt UND Im Testzeitraum...
					if(strtotime($stat->getId()) >= $selDate['from'] && strtotime($stat->getId()) <= $selDate['to']){

						//die einzelnen Themes durchgehen die im AB test vorliegen
						if (!empty($stat->getDetails())) {
							foreach ($stat->getDetails() as $themeID => $themeData) {
								if(in_array($themeID, $buttonName, true))
								{
									foreach($themeData as $type=> $singleDat){
										$type = trim($type);
										if (empty($sumData[$themeID][$type])) {
											$sumData[$themeID][$type] = 0;
										}
										$sumData[$themeID][$type] += $singleDat;
									}
								}
							}
						}
						$sumData['Calls'] += $stat->getCallCount();
					}
				}
			}
			// Restore the previous domain context
			Domain::setDomainForRequest($requestDomainId);
			Model::closeDomainModels();

		} catch (RuntimeException $exception) {}

		if (empty($sumData)) {
			$sumData = array(array("0"));
		}

		//Die einzelnen Typen rausholen...
		$barCharts = StatisticsModell::getBarCharts($translator);

		//Das Ausgabearray mit den ProzentDaten
		$sumDatanNeu = array();

		//prozentual wenn so gewünscht
		if ($valuemode == 'p') {
			foreach ($sumData as $theme => $buttons) {
				foreach ($barCharts as $key => $type) {

					//Anteil accept
					if (!empty($sumData['acceptAll'][$type['name']]) && !empty($sumData['opened'][$type['name']])) {
						$sumDatanNeu['acceptAll'][$type['name']] = round($sumData['acceptAll'][$type['name']] / $sumData['opened'][$type['name']] * 100, 2);
					} else {
						$sumData['acceptAll'][$type['name']] = 0;
						$sumDatanNeu['acceptAll'][$type['name']] = 0;
					}

					if (!empty($sumData['decline'][$type['name']]) && !empty($sumData['opened'][$type['name']])) {
						$sumDatanNeu['decline'][$type['name']] = round($sumData['decline'][$type['name']] / $sumData['opened'][$type['name']] * 100, 2);
					} else {
						$sumData['decline'][$type['name']] = 0;
						$sumDatanNeu['decline'][$type['name']] = 0;
					}

					if(!empty($sumData['opened'][$type['name']])) {
						//Anteil Leave - der Rest vom Schützenfest...
						$sumDatanNeu['leave'][$type['name']] = round(100 - $sumDatanNeu['acceptAll'][$type['name']] - $sumDatanNeu['decline'][$type['name']], 2);
						$sumDatanNeu['opened'][$type['name']] = round($sumData['opened'][$type['name']] / $sumData['Calls'] * 100, 2);
					} else{
						$sumDatanNeu['opened'][$type['name']] = 0;
					}
				}
			}
		}
		else {
			foreach ($sumData as $theme => $buttons) {
				foreach ($barCharts as $key => $type) {
					if(!empty($sumData['opened'][$type['name']]))
					{
						if(empty($sumData['acceptAll'][$type['name']]))
						{
							$sumData['acceptAll'][$type['name']] =0;
						}
						if(empty($sumData['decline'][$type['name']]))
						{
							$sumData['decline'][$type['name']] =0;
						}
						//alle Einblendungen
						$sumData['leave'][$type['name']] = $sumData['opened'][$type['name']]-$sumData['acceptAll'][$type['name']]-$sumData['decline'][$type['name']];
					}
					else{
						//alle Einblendungen
						$sumData['leave'][$type['name']] = "";
					}
				}
			}
			$sumDatanNeu=$sumData;
		}

		$chartName = $chartName ?? $request->query->getAlpha('chartName');

		if ($getResultTranslated) {
			$result = [
				// Die Reihenfolge wird in der Diagramm-Legende beibehalten
				$translator->trans('Banner views') => $sumDatanNeu['opened'][$chartName] ?? 0,
				$translator->trans('acceptAll') => $sumDatanNeu['acceptAll'][$chartName] ?? 0,
				$translator->trans('decline') => $sumDatanNeu['decline'][$chartName] ?? 0,
				$translator->trans('leave') => $sumDatanNeu['leave'][$chartName] ?? 0,
			];
		} else {
			$result = [
				// Die Reihenfolge wird in der Diagramm-Legende beibehalten
				'banner views' => $sumDatanNeu['opened'][$chartName] ?? 0,
				'acceptAll' => $sumDatanNeu['acceptAll'][$chartName] ?? 0,
				'decline' => $sumDatanNeu['decline'][$chartName] ?? 0,
				'leave' => $sumDatanNeu['leave'][$chartName] ?? 0,
			];
		}

		return $result;
	}


	/**
	 * @Route("basic_data.json", name="basic_data_json", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return JSONResponse
	 */
	public function showJsonData(Request $request, TranslatorInterface $translator): JSONResponse
	{
		$jsonDataTimeline = $this->getJsonDataForTimeline($request, $translator);

		// Session schließen
		$request->getSession()->save();

		$response = new JsonResponse($jsonDataTimeline);
		//For Dev JSON_PRETTY_PRINT
		$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );

		//und raus...
		return $response;
	}

	/**
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return array
	 */
	public function getJsonDataForTimeline(Request $request, TranslatorInterface $translator): array
	{
		$daterange = (string)$request->query->get('daterange');
		$valuemode = $request->query->getAlpha('valuemode', 'r');
		$showmode = $request->query->getAlpha('showmode', 'd');

		//Datum immer setzen
		$selDate = [
			'from' => time()-(60*60*24*60),
			'to' => time(),
		];
		if ($daterange) {
			$daDates = explode(" - ", $daterange);
			if ($daDates) {
				$selDate['from'] = (int) strtotime($daDates[0]);
				$selDate['to'] = (int) (strtotime($daDates[1]));
			}
		}

		//nur die Daten ab Mitte Februar - vorher nur crap...
		if ($selDate['from']<1613857795) {
			$selDate['from']=1613857795;
		}

		$statistics = [
			'opened' => [],
			'acceptAll' => [],
			'leave' => [],
			'decline' => [],
			'Calls' => [],
		];

		// Einträge für Datums-Range ziehen
		$fromDay = strftime('%Y-%m-%d', $selDate['from']);
		$toDay = strftime('%Y-%m-%d', $selDate['to']);

		// Remember the current domain context
		$requestDomainId = Domain::getDomainIdForRequest();
		foreach (Domain::all() as $domain) {
			Domain::setDomainForRequest($domain->getId());
			Model::closeDomainModels();

			$allData = array_filter(StatisticsData::all(), function ($dateKey) use ($fromDay, $toDay) {
				return ($dateKey >= $fromDay and $dateKey <= $toDay);
			}, ARRAY_FILTER_USE_KEY);
			// Nach Datum sortieren
			ksort($allData);

			try {
				foreach ($allData as $date => $stat) {
					$statistics["x"][$date] = $date;
				}
			}
			catch (RuntimeException $exception) {}

			try {
				foreach ($allData as $dateKey => $stat) {
					if (!array_key_exists($dateKey, $statistics["Calls"])) {
						$statistics["opened"][$dateKey] = 0;
						$statistics["acceptAll"][$dateKey] = 0;
						$statistics["leave"][$dateKey] = 0;
						$statistics["decline"][$dateKey] = 0;
						$statistics["Calls"][$dateKey] = 0;
					}

					$details = $stat->getDetails();

					//absolut count
					if (!empty($details)) {
						foreach ($details as $themeID => $themeData) {
							if (array_key_exists($themeID, $statistics) && array_key_exists($dateKey, $statistics[$themeID])) {
								$statistics[$themeID][$dateKey] += $themeData['all'] ?? 0;
							}
						}
						if(!empty($details['opened']['all']) && !empty($details['acceptAll']['all'])) {
							if(empty($details['decline']['all'])) {
								$details['decline']['all'] = 0;
								$statistics["decline"][$dateKey] += 0;
							}
							$decline = $details['opened']['all'] - $details['decline']['all'] - $details['acceptAll']['all'];
							//oha - da wurde vermutlich durch Bots Murks produziert... korrigieren.
							if ($decline < 0) {
								$decline = 0;
							}
							$statistics["leave"][$dateKey] += $decline;
						} else {
							$statistics["acceptAll"][$dateKey] += 0;
							$statistics["leave"][$dateKey] += 0;
							$statistics["decline"][$dateKey] += 0;
						}
					}
					$statistics["Calls"][$dateKey] += $stat->getCallCount();
				}
			}
			catch (RuntimeException $exception) {}
		}
		// Restore the previous domain context
		Domain::setDomainForRequest($requestDomainId);
		Model::closeDomainModels();

		// Datums-Keys entfernen
		$statistics = array_map(function ($subarray) {
			if (is_array($subarray)) {
				return array_values($subarray);
			} else {
				return $subarray;
			}
		}, $statistics);

		if ($showmode != "d") {
			if($showmode == "w") {
				$timePoints = "Y-W";
			}
			if($showmode == "m") {
				$timePoints = "Y-m";
			}
			if($showmode == "y") {
				$timePoints = "Y";
			}
			$newXCounter = [];

			if(!empty($statistics)) {
				foreach ($statistics as $cat => $dataPerDay) {
					if (!empty($dataPerDay)) {
						foreach ($dataPerDay as $k=>$aTimePoint) {
							if(!isset($newXCounter[$cat][$statistics['x'][$k]])) {
								$newXCounter[$cat][$statistics['x'][$k]]=0;
							}
							if (isset($statistics['x'][$k]) and isset($newXCounter[$cat][$statistics['x'][$k]])) {
								$newXCounter[$cat][$statistics['x'][$k]]=$aTimePoint;
							}
						}
					}
				}
			}
			$newXCounterReal = [];

			foreach ($newXCounter as $cat =>$catDat) {
				if($cat!="x") {
					if(!empty($catDat)) {
						foreach ($catDat as $aDate=>$aTimePoint) {
							$daTimeDat = date($timePoints, strtotime($aDate));
							if(empty($newXCounterReal[$cat][$daTimeDat])) {
								$newXCounterReal[$cat][$daTimeDat]['value'] = 0;
								$newXCounterReal[$cat][$daTimeDat]['counter'] = 0;
							}
							$newXCounterReal[$cat][$daTimeDat]['value'] += ($aTimePoint);
							$newXCounterReal[$cat][$daTimeDat]['counter'] += 1;
						}
					}
				} else {
					if(!empty($catDat)) {
						foreach ($catDat as $aDate=>$aTimePoint) {
							$daTimeDat = date($timePoints,strtotime($aDate));
							if(empty($newXCounterReal[$cat][$daTimeDat])) {
								$newXCounterReal[$cat][$daTimeDat] =0;
							}
							$newXCounterReal[$cat][$daTimeDat]=$aTimePoint;
						}
					}
				}
			}

			//wird nicht benutzt... oder? $newXCounterRealData = array();
			$retData = [];
			foreach ($newXCounterReal as $cat =>$catDat) {
				if($cat!="x") {
					foreach($catDat as $k=>$dataPoint) {
						if ($valuemode == "p") {
							$catDat[$k]=round($dataPoint['value']/$dataPoint['counter'],2);
						} else {
							$catDat[$k]=round($dataPoint['value'],2);
						}
					}
				}
				$retData[$cat] = array_values($catDat);
			}
			$statistics = $retData;
		}

		$statistics = [
			'x' => $statistics['x'] ?? [],
			// Die Reihenfolge wird in der Diagramm-Legende beibehalten
			$translator->trans('Calls') => $statistics['Calls'] ?? [],
			$translator->trans('Banner views') => $statistics['opened'] ?? [],
			$translator->trans('acceptAll') => $statistics['acceptAll'] ?? [],
			$translator->trans('decline') => $statistics['decline'] ?? [],
			$translator->trans('leave') => $statistics['leave'] ?? [],
		];

		if ($valuemode == 'p') {
			// prozentual - überführe absolute Werte in Prozentwerte (Bannereinblendungen in Relation zu den Aufrufen
			// und "Alle akzeptiert", "Abgelehnt" und "Absprung" in Relation zu den Banner-Einblendungen)
			for ($i = 0; $i < count($statistics['x']); $i++) {
				if ($statistics[$translator->trans('Banner views')][$i] > 0) {
					$statistics[$translator->trans('acceptAll')][$i] = round($statistics[$translator->trans('acceptAll')][$i] / $statistics[$translator->trans('Banner views')][$i] * 100, 2);
					$statistics[$translator->trans('decline')][$i] = round($statistics[$translator->trans('decline')][$i] / $statistics[$translator->trans('Banner views')][$i] * 100, 2);
					$statistics[$translator->trans('leave')][$i] = round($statistics[$translator->trans('leave')][$i] / $statistics[$translator->trans('Banner views')][$i] * 100, 2);
				} else {
					$statistics[$translator->trans('acceptAll')][$i] = 0;
					$statistics[$translator->trans('decline')][$i] = 0;
					$statistics[$translator->trans('leave')][$i] = 0;
				}
				$statistics[$translator->trans('Banner views')][$i] = round($statistics[$translator->trans('Banner views')][$i] / $statistics[$translator->trans('Calls')][$i] * 100, 2);
			}
			unset($statistics[$translator->trans('Calls')]);
		}

		return $statistics;
	}

	/**
	 * @return bool
	 */
	private function setCssAndJSForExternal()
	{
		//Set Plugin CSS for this Domain
		$domain = Domain::activeDomain();
		$cssData = array("plugin_ccm19_statistics_user_statistics_css");
		$domain->addExtraCSS($cssData);

		$jsData = array("plugin_ccm19_statistics_user_statistics_js");
		$domain->addExtraJS($jsData);
		$domain->save();

		return true;
	}
}
