(function () {
	"use strict";

	function bytesToBase64(bytes) {
		bytes = new Uint8Array(bytes);
		return btoa(Array.from(bytes, function (byte) { return String.fromCodePoint(byte); }).join(""));
	}
	function bytesFromHex(hexString) {
		if (hexString.length % 2 !== 0) {
			hexString = "0"+hexString;
		}
		return Uint8Array.from(hexString.match(/.{2}/g).map(function (byte) { return parseInt(byte, 16); }));
	}
	function bytesFromBase64(base64) {
		const binString = atob(base64.replaceAll('-', '+').replaceAll('_', '/'));
		return Uint8Array.from(binString, function (m) { return m.codePointAt(0); });
	}

	function startRename() {
		var span = this;
		var input = document.createElement('input');
		var url = this.getAttribute('data-rename-url');
		span.style.display = 'none';
		input.type = 'text';
		input.className = 'input-sm form-control';
		input.value = span.innerText;
		input.placeholder = input.value;
		if (this.id) {
			input.setAttribute('aria-controls', this.id);
		}
		this.parentNode.insertBefore(input, span);
		input.focus();
		input.select();

		var endRename = (async function (cancel) {
			var csrf_token = $('meta[name="X-CSRF-Token"]').prop('content');
			document.removeEventListener('blur', onBlur, { capture: true });
			input.removeEventListener('keyup', onKeyUp);
			var value = input.value;
			input.disabled = true;
			if (cancel !== true) {
				try {
					var data = new FormData();
					data.set('name', value);
					data.set('_method', 'PATCH');
					var result = await fetch(url, {
						credentials: 'include',
						headers: {"X-CSRF-Token": csrf_token},
						method: 'POST',
						body: data,
						keepalive: true,
					});
					if (result.ok) {
						span.innerText = value;
					}
				} catch (e) {
					console.error(e);
				}
			}
			span.style.display = null;
			input.parentNode.removeChild(input);

		}).bind(this);

		var onBlur = async function () { await endRename(false); };
		var onKeyUp = async function (evt) {
			if (evt.key == "Enter") {
				await endRename(false);
				span.focus();
			} else if (evt.key == "Escape") {
				await endRename(true);
				span.focus();
			}
		};

		document.addEventListener('blur', onBlur, { passive: true, capture: true, once: true });
		input.addEventListener('keyup', onKeyUp, { passive: true });
	}

	async function doPasskeyCheck(config, credentials) {
		var csrf_token = $('meta[name="X-CSRF-Token"]').prop('content');

		// Challenge erzeugen
		var challenge = await fetch(config.checkChallengeUrl, {
			cache: 'no-store',
			credentials: 'include',
			headers: {"X-CSRF-Token": csrf_token, 'Accept': 'application/octet-stream'}
		});
		if (!challenge.ok) {
			var e = new Error("could not generate challenge", { cause: challenge });
			e.name = "ChallengeError";
			throw e;
		}

		// WebAuthn-API anfragen
		var credential = await navigator.credentials.get({
			publicKey: {
				challenge: await challenge.arrayBuffer(),
				rpId: config.rpId,
				allowCredentials: credentials,
			}
		});

		return {
			type: credential.type,
			id: credential.id,
			clientData: bytesToBase64(credential.response.clientDataJSON),
			authenticatorData: bytesToBase64(credential.response.authenticatorData),
			signature: bytesToBase64(credential.response.signature),
		};
	}

	async function passkeyAutoForm() {
		try {
			var config = JSON.parse(document.getElementById('passkey-config').innerHTML);
			var csrf_token = $('meta[name="X-CSRF-Token"]').prop('content');

			if (!('registeredCredentials' in config) || this.tagName.toLowerCase() !== 'form') {
				return;
			}

			var messageContainer = document.getElementById('passkey-error-message');
			messageContainer.innerText = '';
			messageContainer.classList.add('hidden');

			var credentials = config.registeredCredentials.map(function (entry) {
				var result = {
					type: entry.type,
					id: bytesFromBase64(entry.id),
				};
				if (entry.transports) {
					result.transports = entry.transports;
				}
				return result;
			});

			var data = await doPasskeyCheck(config, credentials);
			document.getElementsByName('passkeyId')[0].value = data.id;
			document.getElementsByName('authenticatorData')[0].value = data.authenticatorData;
			document.getElementsByName('clientData')[0].value = data.clientData;
			document.getElementsByName('signature')[0].value = data.signature;
			var pwInput = document.getElementsByName('password')[0];
			pwInput.value = '';
			pwInput.disabled = true;
			this.submit();
		} catch (e) {
			if ('name' in e && config.messages.login[e.name] !== undefined) {
				if (e.name == 'NotAllowedError') {
					return;
				}
				messageContainer.innerText = config.messages.login[e.name];
			} else {
				messageContainer.innerText = config.messages.login.error.replace("%ERROR%", e);
			}
			messageContainer.classList.remove("hidden");
		}
	}

	async function testPasskey() {
		this.disabled = true;
		$(this).popover('hide');
		try {
			var config = JSON.parse(document.getElementById('passkey-config').innerHTML);
			var csrf_token = $('meta[name="X-CSRF-Token"]').prop('content');
			var id = this.getAttribute('data-id');
			var rawId = bytesFromBase64(id);
			var credential = { type: 'public-key', id: rawId };

			// Credential-Daten ausfüllen
			for (var item of config.registeredCredentials) {
				if (item.id == id) {
					credential.type = item.type;
					if ('transports' in item && item.transports && item.transports.length > 0) {
						credential.transports = item.transports;
					}
					break;
				}
			}

			var data = await doPasskeyCheck(config, [credential]);
			var result = await fetch(config.checkUrl, {
				cache: 'no-store',
				credentials: 'include',
				headers: {"X-CSRF-Token": csrf_token, 'Content-Type': 'application/json'},
				method: 'POST',
				body: JSON.stringify(data),
			});

			this.disabled = false;
			if (result.ok) {
				this.setAttribute('data-content', config.messages.login.success);
			} else {
				this.setAttribute('data-content', config.messages.login.fail);
			}
		} catch (e) {
			this.disabled = false;
			if ('name' in e && config.messages.login[e.name] !== undefined) {
				this.setAttribute('data-content', config.messages.login[e.name]);
			} else {
				this.setAttribute('data-content', config.messages.login.error.replace("%ERROR%", e));
			}
		}
		$(this).popover('show');
	}

	async function addNewPasskey() {
		var config = JSON.parse(document.getElementById('passkey-config').innerHTML);
		var csrf_token = $('meta[name="X-CSRF-Token"]').prop('content');
		var username = $('meta[name="X-Username"]').prop('content');
		var userid = bytesFromHex($('meta[name="X-UserID"]').prop('content'));

		var messageContainer = document.getElementById('passkey-error-message');
		messageContainer.innerText = '';
		messageContainer.classList.add('hidden');

		// Key-Parameter festlegen
		var params = config.algorithms.map(function (alg) { return {type: "public-key", alg: alg}; });

		// Challenge erzeugen
		var challenge = await fetch(config.registerChallengeUrl, {
			cache: 'no-store',
			credentials: 'include',
			headers: {"X-CSRF-Token": csrf_token, 'Accept': 'application/octet-stream'}
		});
		if (!challenge.ok) {
			messageContainer.innerText = config.messages.register.ChallengeError;
			messageContainer.classList.remove('hidden');
			return;
		}

		var excludeCredentials = config.registeredCredentials.map(function (entry) {
			var result = {
				type: entry.type,
				id: bytesFromBase64(entry.id),
			};
			if (entry.transports) {
				result.transports = entry.transports;
			}
			return result;
		});

		try {
			// WebAuthn-API anfragen
			var credential = await navigator.credentials.create({
				publicKey: {
					challenge: await challenge.arrayBuffer(),
					attestation: "indirect",
					rp: { id: config.rpId, name: document.title },
					user: {
						id: userid,
						name: username,
						displayName: username,
					},
					pubKeyCredParams: params,
					excludeCredentials: excludeCredentials,
				}
			});

			// Ergebnis an Backend senden
			var data = {
				type: credential.type,
				id: credential.id,
				attachment: credential.authenticatorAttachment,
				transports: credential.response.getTransports(),
				algorithm: credential.response.getPublicKeyAlgorithm(),
				publicKey: bytesToBase64(credential.response.getPublicKey()),
				clientData: bytesToBase64(credential.response.clientDataJSON),
				authenticatorData: bytesToBase64(credential.response.getAuthenticatorData()),
			};
			document.getElementsByName('passkeyData')[0].value = JSON.stringify(data);
			this.submit();
		} catch (e) {
			if ('name' in e && config.messages.register[e.name] !== undefined) {
				messageContainer.innerText = config.messages.register[e.name];
				messageContainer.classList.remove('hidden');
			} else {
				messageContainer.innerText = config.messages.register.error.replace("%ERROR%", e);
				messageContainer.classList.remove('hidden');
			}
			return;
		}

	}

	Array.prototype.forEach.call(document.getElementsByClassName("passkey-register-form"), function (el) {
		el.addEventListener('submit', function (evt) {
			evt.preventDefault();
			addNewPasskey.call(this);
		});
		if (!('credentials' in navigator) || !('AuthenticatorAssertionResponse' in window)) {
			var messageContainer = document.getElementById('passkey-error-message');
			if (messageContainer) {
				var config = JSON.parse(document.getElementById('passkey-config').innerHTML);
				messageContainer.innerText = config.messages.register.NotSupportedError;
				messageContainer.classList.remove('hidden');
			}
			return;
		}

	});
	Array.prototype.forEach.call(document.getElementsByClassName("test-passkey"), function (el) {
		el.addEventListener('click', testPasskey);
	});
	Array.prototype.some.call(document.getElementsByClassName("passkey-immediate-form"), function (el) {
		setTimeout(passkeyAutoForm.bind(el), 50);
		return true;
	});
	Array.prototype.forEach.call(document.getElementsByClassName("renamable"), function (el) {
		if (el.hasAttribute('data-rename-url')) {
			el.addEventListener('click', startRename);
			el.addEventListener('keyup', function (evt) {
				if (evt.key == "Enter") {
					startRename.call(this);
				}
			});
			el.style.cursor = 'pointer';
			el.tabIndex = 0;
		}
	});
})();
