<?php

namespace Plugins\Ccm19TwoFactorAuth\Component;

use App\Config;
use Plugins\Ccm19TwoFactorAuth\Exception\FIDO\RPIdMismatchException;
use Plugins\Ccm19TwoFactorAuth\Exception\FIDO\InvalidEncodingException;
use Plugins\Ccm19TwoFactorAuth\Exception\FIDO\ChallengeExpiredException;
use Plugins\Ccm19TwoFactorAuth\Exception\FIDO\UnsupportedAlgorithmException;

const ALGO_ED25519 = -8;
const ALGO_ES256 = -7;
const ALGO_RS256 = -257;

class FIDO {
	private $challengeKey;

	/**
	 * @return array<string, int> [ "AlgoName" => Algo-ID ]
	 */
	public function getAvailableAlgorithms(): array {
		$algos = [];
		if (function_exists('sodium_crypto_sign_verify_detached')) {
			$algos['Ed25519'] = ALGO_ED25519;
		}
		if (function_exists('openssl_get_curve_names') and defined('OPENSSL_KEYTYPE_EC')) {
			$curves = openssl_get_curve_names();
			if (array_search('prime256v1', $curves, true) !== false) {
				$algos['ES256'] = ALGO_ES256;
			}
		}
		if (function_exists('openssl_verify')) {
			$algos['RS256'] = ALGO_RS256;
		}
		return $algos;
	}

	public function getDeviceInfo($guid): ?array {
		$guid = strtr(strtolower($guid), ['-' => '']);
		$devices = json_decode(file_get_contents(dirname(dirname(__DIR__)).'/share/fido-devices.json'), true, 10);
		$key = $devices['aliases'][$guid] ?? $guid;
		return $devices['devices'][$key] ?? null;
	}

	/**
	 * @return string|null
	 */
	public function getRelyingPartyId(): ?string {
		$baseUrl = Config::getInstance()->getBaseUrl();
		if (!$baseUrl) {
			$baseUrl = 'https://'.$_SERVER['HTTP_HOST'];
		}
		$origin = parse_url($baseUrl, PHP_URL_HOST);
		if (!$origin) {
			return null;
		}
		return $origin;
	}

	public function __construct()
	{
		$this->challengeKey = hash_pbkdf2(
			'sha256',
			$_ENV['APP_SECRET'],
			'passkey',
			200,
			32,
			true
		);
	}

	private function makeChallenge(int $timestamp, string $bytes, string $salt): string {
		// Timestamp shifted um 5 Bits (auf 32s Genauigkeit gesenkt), dann auf 32 bit gekappt.
		$data = pack('N', $timestamp >> 5) . $bytes;
		$hash = $this->hashChallenge($data, $salt);
		return $data . $hash;
	}

	private function hashChallenge(string $bytes, string $salt): string {
		$hash = hash_hmac('sha256', $bytes, $this->challengeKey.$salt, true);
		return substr($hash, 2, 24);
	}

	public function generateChallenge(string $salt): string
	{
		return $this->makeChallenge((int)time(), random_bytes(20), $salt);
	}

	private function verifySignature(array $keyInfo, string $signature, string $authenticatorData, string $clientData): bool {
		$algorithms = $this->getAvailableAlgorithms();
		if (!in_array($keyInfo['algorithm'], $algorithms)) {
			throw new UnsupportedAlgorithmException();
		}
		$data = $authenticatorData . hash('sha256', $clientData, true);

		switch ($keyInfo['algorithm']) {
			case ALGO_ED25519:
				$key = base64_decode($keyInfo['publicKey']);
				$key = substr($key, -32); // Header entfernen.. Sodium will nur das reine Schlüsselmaterial.
				return @sodium_crypto_sign_verify_detached($signature, $data, $key);
			case ALGO_ES256:
			case ALGO_RS256:
				return @openssl_verify($data, $signature, "-----BEGIN PUBLIC KEY-----\n".$keyInfo['publicKey']."\n-----END PUBLIC KEY-----", OPENSSL_ALGO_SHA256) === 1;
		}
	}

	public function verify(array $keyInfo, string $salt, string $authenticatorData, string $clientData, string $signature, ?int $timestamp = null): bool {
		if ($timestamp === null) {
			$timestamp = (int)time();
		}

		// Check RPID
		$rpidHash = substr($authenticatorData, 0, 32);
		if (!hash_equals(hash('sha256', $this->getRelyingPartyId(), true), $rpidHash)) {
			throw new RPIdMismatchException();
		}

		// Decode client data
		$decodedClientData = json_decode($clientData, true, 8);
		if (!$decodedClientData or !is_array($decodedClientData)) {
			throw new InvalidEncodingException();
		}

		// Verify challenge timestamp (not older than 30 minutes)
		$timestamp = (($timestamp >> 5) & 0xFFFFFFFF);
		$challenge = base64_decode(strtr($decodedClientData['challenge'] ?? '', '-_', '+/'));
		$ts = unpack('N', $challenge)[1];
		if ($ts === false or abs($timestamp - $ts) > (1800 >> 5)) {
			throw new ChallengeExpiredException();
		}

		// Verify challenge is generated by this CCM19 instance
		$challengeHash = substr($challenge, -24);
		$challengeData = substr($challenge, 0, -24);
		$challengeExpectedHash = $this->hashChallenge($challengeData, $salt);
		if (!hash_equals($challengeExpectedHash, $challengeHash)) {
			return false;
		}

		// Check signature
		return $this->verifySignature($keyInfo, $signature, $authenticatorData, $clientData);
	}

	public function verifyRegistration(string $salt, string $authenticatorData, string $clientData, ?int $timestamp = null): bool {
		if ($timestamp === null) {
			$timestamp = (int)time();
		}

		// Check RPID
		$rpidHash = substr($authenticatorData, 0, 32);
		if (!hash_equals(hash('sha256', $this->getRelyingPartyId(), true), $rpidHash)) {
			throw new RPIdMismatchException();
		}

		// Decode client data
		$decodedClientData = json_decode($clientData, true, 8);
		if (!$decodedClientData or !is_array($decodedClientData)) {
			throw new InvalidEncodingException();
		}

		// Verify challenge timestamp (not older than 30 minutes)
		$timestamp = (($timestamp >> 5) & 0xFFFFFFFF);
		$challenge = base64_decode(strtr($decodedClientData['challenge'] ?? '', '-_', '+/'));
		$ts = unpack('N', $challenge)[1];
		if ($ts === false or abs($timestamp - $ts) > (1800 >> 5)) {
			throw new ChallengeExpiredException();
		}

		// Verify challenge is generated by this CCM19 instance
		$challengeHash = substr($challenge, -24);
		$challengeData = substr($challenge, 0, -24);
		$challengeExpectedHash = $this->hashChallenge($challengeData, $salt);
		if (!hash_equals($challengeExpectedHash, $challengeHash)) {
			return false;
		}

		return true;
	}
}
