<?php

namespace Plugins\Ccm19TwoFactorAuth\Component;

use Base2n;

class TOTP {
	private $algo;
	private $secret;
	private $digits;
	private $period;
	private static $BASE32;

	private static function getBase32(): Base2n
	{
		if (empty(self::$BASE32)) {
			self::$BASE32 = new Base2n(5, 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567', false, true, true);
		}
		return self::$BASE32;
	}

	public function __construct($parameters)
	{
		$this->algo = $parameters['algo'] ?? 'SHA-1';
		$this->secret = $parameters['secret'] ?? '';
		$this->digits = (int)($parameters['digits'] ?? 6);
		$this->period = (int)($parameters['period'] ?? 30);
	}

	public static function generate(int $secretBytes = 20, string $algo = 'SHA-1', int $digits = 6, int $period = 30)
	{
		$bytes = random_bytes($secretBytes);
		$secret = self::getBase32()->encode($bytes);
		return new static([
			'secret' => $secret,
			'algo' => $algo,
			'digits' => $digits,
			'period' => $period,
		]);
	}

	public function validate(string $code, ?int $timestamp = null): bool {
		if ($timestamp === null) {
			$timestamp = (int)time();
		}

		$algo = strtr(strtolower($this->algo), ['-' => '']);

		// Eingabe bereinigen
		$cleaned_code = @preg_replace("/\s/u", '', $code);
		if ($cleaned_code === null) {
			// Fallback für nicht-gültiges UTF-8
			$cleaned_code = @preg_replace("/\s/", '', $code);
		}
		$code = $cleaned_code;

		// Counter berechnen
		$counter = (int)($timestamp / $this->period);

		// Secret enpacken
		$secret = self::getBase32()->decode($this->secret);

		// Counter-Werte berechnen (+- eine Zeitperiode)
		$cts = [
			pack('J', $counter),
			pack('J', $counter-1),
			pack('J', $counter+1),
		];
		$time = str_pad( pack('N', $counter ), 8, "\x00", STR_PAD_LEFT);

		// Prüfen, ob der übergebene Code für einen der Counter passt
		$ok = false;
		foreach ($cts as $ct) {
			$hash = hash_hmac($algo, $ct, $secret, false);
			$number = (hexdec(substr($hash, hexdec($hash[39]) * 2, 8)) & 0x7FFFFFFF) % pow(10, $this->digits);
			$number_code = str_pad((string)$number, $this->digits, '0', STR_PAD_LEFT);
			if (hash_equals($number_code, $code)) {
				$ok = true;
			}
		}
		return $ok;
	}

	public function getParameters(): array
	{
		return [
			'type' => 'totp',
			'algo' => $this->algo,
			'secret' => $this->secret,
			'digits' => $this->digits,
			'period' => $this->period,
		];
	}

	public function generateQrCodeUrl(string $label, ?string $issuer = null): string
	{
		$params = [
			'secret' => $this->secret
		];

		if ($issuer !== null) {
			$params['issuer'] = $issuer;
		}

		if ($this->algo !== 'SHA-1') {
			$params['algorithm'] = $this->algo;
		}

		if ($this->digits !== 6) {
			$params['digits'] = $this->digits;
		}

		if ($this->period !== 30) {
			$params['period'] = $this->period;
		}

		return 'otpauth://totp/'.rawurlencode($label).'?'.http_build_query($params, "", '&');
	}
}
