<?php

namespace Plugins\Ccm19TwoFactorAuth\Controller;

use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Controller\UniversalController;
use App\Controller\Login as LoginController;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\PluginState;
use App\Model\User;
use App\Config;
use App\Utils;
use Plugins\Ccm19TwoFactorAuth\Component\TOTP;
use Plugins\Ccm19TwoFactorAuth\Component\FIDO;
use Plugins\Ccm19TwoFactorAuth\Exception\FIDO\Exception as FIDOException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;


/**
 * Controller for 2FA Backend
 * @Route("/plugins/2fa", name="plugin_ccm19_twofactorauth_")
 */
class Backend extends UniversalController
{
	/**
	 * @Menu("Two-Factor Authentication", group="Account Settings", icon="fa-unlock-alt", routeGroup="plugin_ccm19_twofactorauth_", navigation="meta", order=75)
	 * @Route("/", name="index", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function index(PluginState $state): Response
	{
		$user = User::loggedInUser();

		$totpActive = (bool)$user->getTwoFactorAuthParameters($state->getPluginUID());
		$passkeyActive = (bool)$user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey');

		return $this->render('@plugin:Ccm19TwoFactorAuth/internal/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'backTo' => null,
			'totpActive' => $totpActive,
			'passkeyActive' => $passkeyActive,
			'anyActive' => $user->hasTwoFactorAuthParameters(),
		]);
	}

	private function getIssuerName(): string {
		$config = Config::getInstance();
		$baseUrl = $config->getBaseUrl();
		if (!$baseUrl) {
			$baseUrl = 'http://'.$_SERVER['HTTP_HOST'];
		}
		$origin = parse_url($baseUrl, PHP_URL_HOST);
		if (function_exists('idn_to_utf8')) {
			$originPuny = idn_to_utf8($origin, IDNA_DEFAULT | IDNA_CHECK_BIDI | IDNA_NONTRANSITIONAL_TO_UNICODE);
			if ($originPuny) {
				$origin = $originPuny;
			}
		}
		switch (strtolower($origin)) {
			case "cloud.ccm19.de":
				return "CCM19 Cloud";
			case "mein.clickskeks.at":
				return "Clickskeks";
			default:
				return $config->getAppName()." (".$origin.")";
		}
	}

	/**
	 * @Route("/totp/", name="totp", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function totp(PluginState $state): Response
	{
		$user = User::loggedInUser();

		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID());
		$isActive = (bool)$parameters;
		if (!$parameters) {
			$totp = TOTP::generate();
			$parameters = $totp->getParameters();
			$issuerName = $this->getIssuerName();
			$qrCodeUrl = $totp->generateQrCodeUrl($user->getUsername(), $issuerName);
		} else {
			$qrCodeUrl = null;
		}

		return $this->render('@plugin:Ccm19TwoFactorAuth/internal/totp.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'backTo' => null,
			'isActive' => $isActive,
			'parameters' => $parameters,
			'qrCodeUrl' => $qrCodeUrl,
		]);
	}

	/**
	 * @Route("/totp/", name="totp_post", methods={"POST"})
	 * @return Response
	 */
	public function totpAction(Request $request, TranslatorInterface $translator, PluginState $state): Response
	{
		$user = User::loggedInUser();

		try {
			CsrfToken::get()->checkRequest($request);
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('plugin_ccm19_twofactorauth_totp_post');
		}

		$action = $request->request->getAlnum('action');
		$totp = null;

		switch ($action) {
			case 'activate':
				$totp = new TOTP($request->request->get('parameters'));
				if ($totp->validate($request->request->getAlnum('otp', ''))) {
					$user
						->setTwoFactorAuthParameters($state->getPluginUID(), $totp->getParameters())
						->generateSessionResumeToken()
						->save();

					$this->addFlash('success', $translator->trans('Two-factor authentication successfully activated.'));
					return $this->redirectToRoute('plugin_ccm19_twofactorauth_index', [], 303);
				} else {
					$this->addFlash('danger', $translator->trans('Invalid or expired login code. Please try again.'));
				}
				break;
			case 'deactivate':
				$totp = new TOTP($user->getTwoFactorAuthParameters($state->getPluginUID()));
				if ($totp->validate($request->request->getAlnum('otp', ''))) {
					$user->setTwoFactorAuthParameters($state->getPluginUID(), [])->save();
					$this->addFlash('success', $translator->trans('Two-factor authentication successfully deactivated.'));
					return $this->redirectToRoute('plugin_ccm19_twofactorauth_index', [], 303);
				} else {
					$this->addFlash('danger', $translator->trans('Invalid or expired login code. Please try again.'));
				}
				break;
		}

		if ($totp and $totp->getParameters()['secret'] !== '') {
			$parameters = $totp->getParameters();
		} else {
			$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID());
		}

		$isActive = (bool)$parameters;
		if (!$parameters) {
			$totp = TOTP::generate();
			$parameters = $totp->getParameters();
		}

		$issuerName = $this->getIssuerName();
		$qrCodeUrl = ($totp) ? $totp->generateQrCodeUrl($user->getUsername(), $issuerName) : null;

		return $this->render('@plugin:Ccm19TwoFactorAuth/internal/totp.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'backTo' => null,
			'isActive' => $isActive,
			'parameters' => $parameters,
			'qrCodeUrl' => $qrCodeUrl,
		]);
	}

	/**
	 * @Route("/email/", name="email", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function email(PluginState $state): Response
	{
		$user = User::loggedInUser();

		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--email');
		$isActive = (bool)$parameters;
		// TODO
		throw new NotFoundHttpException("Method not yet available");
	}

	/**
	 * @Route("/passkey/", name="passkey", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function passkey(PluginState $state, FIDO $fido, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey');

		return $this->render('@plugin:Ccm19TwoFactorAuth/internal/passkey.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'devices' => $parameters,
			'passkeyConfig' => [
				'rpId' => $fido->getRelyingPartyId(),
				'checkChallengeUrl' => $this->generateUrl('plugin_ccm19_twofactorauth_passkey_challenge', ['salt' => 'check']),
				'checkUrl' => $this->generateUrl('plugin_ccm19_twofactorauth_passkey_check'),
				'messages' => [
					'login' => [
						'ChallengeError' => $translator->trans('Security check failed. Please reload this page and try again.'),
						'NotSupportedError' => $translator->trans('Passkeys are not supported in this browser.'),
						'NotAllowedError' => $translator->trans('Access denied. Accessing the passkey failed.'),
						'SecurityError' => $translator->trans('Passkey access request denied by the browser. Please make sure you are accessing this page via HTTPS.'),
						'AbortError' => $translator->trans('Passkey access was aborted.'),
						'InvalidStateError' => $translator->trans('This device seems to be incompatible.'),
						'error' => $translator->trans('Accessing the passkey failed because of an unexpected error: %ERROR%'),
						'fail' => $translator->trans('This passkey did not pass verification.'),
						'success' => $translator->trans('Success! This passkey was verified to work correctly.'),
					],
				],
				'registeredCredentials' => array_map(function ($id, $entry) {
					return [
						'type' => $entry['type'] ?? 'public-key',
						'id' => $id,
						'transports' => $entry['transports'] ?? null,
					];
				}, array_keys($parameters), $parameters ?? []),
			],
		]);
	}

	/**
	 * @Route("/passkey/challenge/{salt}.bin", name="passkey_challenge", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function passkeyChallenge(string $salt, Request $request, FIDO $fido): Response
	{
		try {
			$csrfToken = CsrfToken::get();
			$csrfToken->checkRequest($request);
		} catch (CsrfTokenException $e) {
			return new JsonResponse('{"success":false,"message":"invalid CSRF token"}', 403, [
				'Content-Type' => 'application/json',
				'Cache-Control' => 'no-store, no-cache',
			], true);
		}
		$user = User::loggedInUser();
		$challenge = $fido->generateChallenge($user->getId().':'.$salt);

		return new Response($challenge, 200, [
			'Content-Type' => 'application/octet-stream',
			'Cache-Control' => 'no-store, no-cache',
		]);
	}

	/**
	 * @Route("/passkey/register", name="passkey_register_form", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function passkeyRegisterForm(PluginState $state, FIDO $fido, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey');

		return $this->render('@plugin:Ccm19TwoFactorAuth/internal/passkey_register.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'devices' => $parameters ?? [],
			'rpId' => $fido->getRelyingPartyId(),
			'passkeyConfig' => [
				'rpId' => $fido->getRelyingPartyId(),
				'algorithms' => array_values($fido->getAvailableAlgorithms()),
				'registerChallengeUrl' => $this->generateUrl('plugin_ccm19_twofactorauth_passkey_challenge', ['salt' => 'register']),
				'registerUrl' => $this->generateUrl('plugin_ccm19_twofactorauth_passkey_register'),
				'messages' => [
					'register' => [
						'ChallengeError' => $translator->trans('Security check failed. Please reload this page and try again.'),
						'NotSupportedError' => $translator->trans('Passkeys are not supported in this browser.'),
						'NotAllowedError' => $translator->trans('Access denied. Registering a new passkey failed.'),
						'SecurityError' => $translator->trans('Passkey access request denied by the browser. Please make sure you are accessing this page via HTTPS.'),
						'InvalidStateError' => $translator->trans('This device is already registered with this website or is incompatible.'),
						'AbortError' => $translator->trans('Registration was aborted.'),
						'error' => $translator->trans('Registering a new passkey failed because of an unexpected error: %ERROR%'),
					],
				],
				'registeredCredentials' => array_map(function ($id, $entry) {
					return [
						'type' => $entry['type'] ?? 'public-key',
						'id' => $id,
						'transports' => $entry['transports'] ?? null,
					];
				}, array_keys($parameters ?? []), $parameters ?? []),
			],
		]);
	}

	/**
	 * @Route("/passkey/register", name="passkey_register", methods={"POST"})
	 * @return Response
	 */
	public function passkeyRegister(PluginState $state, Request $request, FIDO $fido): Response
	{
		try {
			$csrfToken = CsrfToken::get();
			$csrfToken->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
			return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
		}
		$user = User::loggedInUser();
		$salt = $user->getId() . ':register';

		// Payload dekodieren
		$payload = $request->request->get('passkeyData');
		if (!$payload) {
			$this->addFlash('danger', 'No passkey information transmitted. Does your browser support passkeys and is javascript enabled?');
			return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
		}
		$payload = json_decode($payload, true, 5);
		if (!$payload or !is_array($payload) or !isset($payload['id']) or !is_string($payload['id'])) {
			$this->addFlash('danger', 'Invalid passkey information transmitted. Please try again.');
			return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
		}

		// Schlüsseldaten extrahieren
		$keyId = $payload['id'];
		$authenticatorData = base64_decode((string)($payload['authenticatorData'] ?? ''));
		$clientData = base64_decode((string)($payload['clientData'] ?? ''));
		$flags = ord($authenticatorData[32] ?? "\0");
		$signCount = unpack('N', $authenticatorData, 33)[1];
		$aaguid = bin2hex(substr($authenticatorData, 37, 16));
		$deviceInfo = $fido->getDeviceInfo($aaguid);

		// Art des Geräts so gut wie möglich bestimmen
		$transports = array_unique(array_merge($payload['transports'] ?? [], $deviceInfo ? ($deviceInfo['transportsHint'] ?? []) : []));
		$attachment = (string)($payload['attachment'] ?? 'unknown');
		if ($attachment == 'platform' or in_array('internal', $transports)) {
			$deviceKind = 'platform';
		}
		elseif (in_array('hybrid', $transports)) {
			$deviceKind = 'phone';
		}
		elseif (in_array('usb', $transports)) {
			$deviceKind = 'stick';
		}
		elseif (in_array('smart-card', $transports)) {
			$deviceKind = 'card';
		}
		elseif (in_array('nfc', $transports)) {
			$deviceKind = 'nfc';
		}
		elseif (in_array('ble', $transports)) {
			$deviceKind = 'ble';
		}
		else {
			$deviceKind = 'unknown';
		}

		$keyInfo = [
			'type' => $payload['type'] ?? 'public-key',
			'aaguid' => $aaguid,
			'deviceInfo' => $deviceInfo,
			'algorithm' => (int)($payload['algorithm'] ?? 0),
			'publicKey' => (string)($payload['publicKey'] ?? ''),
			'transports' => (array)($transports),
			'name' => ($deviceInfo) ? $deviceInfo['description'] : 'Passkey '.substr(hash('sha256', $keyId), 0, 8),
			'created' => time(),
			'deviceKindHint' => $deviceKind,
		];

		// Schlüssel verifizieren
		try {
			if (!$fido->verifyRegistration($salt, $authenticatorData, $clientData)) {
				return new JsonResponse([ "success" => false, "message" => "invalid signature" ], 422, []);
			}
		} catch (FIDOException $e) {
			$code = $e->getCode() ?: 422;
			return new JsonResponse([ "success" => false, "message" => $e->getMessage() ], $code, []);
		}

		// Daten speichern
		User::beginWrite();
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey');
		if (!$parameters) {
			$parameters = [];
		}
		$parameters[$keyId] = $keyInfo;
		$user->setTwoFactorAuthParameters($state->getPluginUID().'--passkey', $parameters);
		$user->save();

		$this->addFlash('success', 'Passkey was registered successfully.');
		return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
	}

	/**
	 * @Route("/passkey/check", name="passkey_check", methods={"POST"})
	 * @return Response
	 */
	public function passkeyCheck(PluginState $state, Request $request, FIDO $fido): JsonResponse
	{
		try {
			$csrfToken = CsrfToken::get();
			$csrfToken->checkRequest($request);
		} catch (CsrfTokenException $e) {
			return new JsonResponse('{"success":false,"message":"invalid CSRF token"}', 403, [
				'Content-Type' => 'application/json',
				'Cache-Control' => 'no-store, no-cache',
			], true);
		}
		$user = User::loggedInUser();
		$salt = $user->getId() . ':check';

		// Payload dekodieren
		$payload = null;
		if (Utils::decodeJsonRequest($request, $payload, true, 5) == false) {
			/** @var JsonResponse $payload */
			assert($payload instanceof JsonResponse);
			return $payload;
		}
		elseif (!is_array($payload)) {
			return new JsonResponse('{"success":false,"message":"invalid payload"}', 400, [], true);
		}
		if (!isset($payload['id']) or !is_string($payload['id'])) {
			return new JsonResponse('{"success":false,"message":"invalid payload"}', 400, [], true);
		}

		// Schlüsseldaten extrahieren
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey');
		$keyId = $payload['id'];
		$keyInfo = $parameters[$keyId] ?? null;
		if (!$keyInfo) {
			return new JsonResponse('{"success":false,"message":"unknown passkey"}', 404, [], true);
		}
		$authenticatorData = base64_decode((string)($payload['authenticatorData'] ?? ''));
		$clientData = base64_decode((string)($payload['clientData'] ?? ''));
		$signature = base64_decode((string)($payload['signature'] ?? ''));
		$flags = ord($authenticatorData[32] ?? "\0");
		$signCount = unpack('N', $authenticatorData, 33)[1];

		// Signatur verifizieren
		try {
			if (!$fido->verify($keyInfo, $salt, $authenticatorData, $clientData, $signature)) {
				return new JsonResponse([ "success" => false, "message" => "invalid signature" ], 422, []);
			}
		} catch (FIDOException $e) {
			$code = $e->getCode() ?: 422;
			return new JsonResponse([ "success" => false, "message" => $e->getMessage() ], $code, []);
		}

		return new JsonResponse('{"success":true}', 200, [], true);
	}

	/**
	 * @Route("/passkey/{id}", name="passkey_rename", methods={"PATCH"})
	 * @return Response
	 */
	public function passkeyRename(string $id, PluginState $state, Request $request): JsonResponse
	{
		try {
			$csrfToken = CsrfToken::get();
			$csrfToken->checkRequest($request);
		} catch (CsrfTokenException $e) {
			return new JsonResponse('{"success":false,"message":"invalid CSRF token"}', 403, [
				'Content-Type' => 'application/json',
				'Cache-Control' => 'no-store, no-cache',
			], true);
		}
		$user = User::loggedInUser();

		if ($request->request->has('name')) {
			$name = trim((string)$request->request->get('name'));
			if (strlen($name) == 0) {
				return new JsonResponse('{"success":false,"message":"empty name"}', 422, [], true);
			}
		} else {
			$name = null;
		}

		User::beginWrite();
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey') ?? [];
		if (!$parameters or !isset($parameters[$id])) {
			return new JsonResponse('{"success":false,"message":"not found"}', 404, [], true);
		}
		if ($name !== null) {
			$parameters[$id]['name'] = $name;
		}
		$user->setTwoFactorAuthParameters($state->getPluginUID().'--passkey', $parameters);
		$user->save();

		return new JsonResponse('{"success":true}', 200, [], true);
	}


	/**
	 * @Route("/passkey/{id}/delete", name="passkey_delete_form", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function passkeyDeleteForm(string $id, FIDO $fido, PluginState $state, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey') ?? [];

		return $this->render('@plugin:Ccm19TwoFactorAuth/internal/passkey_delete.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'id' => $id,
			'keyInfo' => $parameters[$id] ?? [],
			'passkeyConfig' => [
				'rpId' => $fido->getRelyingPartyId(),
				'checkChallengeUrl' => $this->generateUrl('plugin_ccm19_twofactorauth_passkey_challenge', ['salt' => 'delete']),
				'messages' => [
					'login' => [
						'ChallengeError' => $translator->trans('Security check failed. Please reload this page and try again.'),
						'NotSupportedError' => $translator->trans('Passkeys are not supported in this browser.'),
						'NotAllowedError' => $translator->trans('Access denied. Accessing the passkey failed.'),
						'SecurityError' => $translator->trans('Passkey access request denied by the browser. Please make sure you are accessing this page via HTTPS.'),
						'AbortError' => $translator->trans('Passkey access was aborted.'),
						'InvalidStateError' => $translator->trans('This device seems to be incompatible.'),
						'error' => $translator->trans('Accessing the passkey failed because of an unexpected error: %ERROR%'),
					],
				],
				'registeredCredentials' => array_map(function ($id, $entry) {
					return [
						'type' => $entry['type'] ?? 'public-key',
						'id' => $id,
						'transports' => $entry['transports'] ?? null,
					];
				}, array_keys($parameters ?? []), $parameters ?? []),
			],
		]);
	}

	/**
	 * @Route("/passkey/{id}", name="passkey_delete", methods={"DELETE"})
	 * @return Response
	 */
	public function passkeyDelete(string $id, FIDO $fido, Request $request, PluginState $state, TranslatorInterface $translator): Response
	{
		try {
			$csrfToken = CsrfToken::get();
			$csrfToken->checkRequest($request);

		} catch (CsrfTokenException $e) {
			return new JsonResponse('{"success":false,"message":"invalid CSRF token"}', 403, [
				'Content-Type' => 'application/json',
				'Cache-Control' => 'no-store, no-cache',
			], true);
		}

		$user = User::loggedInUser();
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey') ?? [];
		$keyName = $parameters[$id]['name'] ?? null;
		if ($keyName === null) {
			return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
		}

		// Wenn Passwort in Request, dieses zum Authentifizieren verwenden
		if ($request->request->has('password')) {
			if (!$user->checkPassword((string)$request->request->get('password'))) {
				$this->addFlash('danger', $translator->trans("The password you've entered is incorrect."));
				return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
			}
		}
		// Ansonsten Passkey-Authentifizierung verwenden
		else {
			$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey') ?? [];
			$salt = $user->getId() . ':delete';
			$keyInfo = $parameters[(string)$request->request->get('passkeyId')] ?? null;
			if (!$keyInfo) {
				$this->addFlash('danger', 'Passkey authentication failed.');
				return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
			}
			$authenticatorData = base64_decode((string)$request->request->get('authenticatorData'));
			$clientData = base64_decode((string)$request->request->get('clientData'));
			$signature = base64_decode((string)$request->request->get('signature'));
			$flags = ord($authenticatorData[32] ?? "\0");
			$signCount = unpack('N', $authenticatorData, 33)[1];

			try {
				if (!$fido->verify($keyInfo, $salt, $authenticatorData, $clientData, $signature)) {
					$this->addFlash('danger', 'Passkey authentication failed.');
					return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
				}
			} catch (FIDOException $e) {
				$this->addFlash('danger', $translator->trans($e->getMessage()));
				return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
			}
		}

		// Eintrag löschen
		User::beginWrite();
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey');
		if ($parameters and isset($parameters[$id])) {
			unset($parameters[$id]);
			$parameters = $user->setTwoFactorAuthParameters($state->getPluginUID().'--passkey', $parameters);
		}
		$user->save();

		$this->addFlash('success', $translator->trans("Passkey \"%name%\" was deleted successfully.", ['%name%' => $keyName]));
		return $this->redirectToRoute('plugin_ccm19_twofactorauth_passkey', [], 303);
	}
}
