<?php

namespace Plugins\Ccm19TwoFactorAuth\Controller;

use App\Component\LoggerInterface;
use App\Controller\Login as LoginController;
use App\Controller\UnrestrictedController;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\LoginAttemptsLog;
use App\Model\PluginState;
use App\Model\User;
use App\Utils;
use Plugins\Ccm19TwoFactorAuth\Component\FIDO;
use Plugins\Ccm19TwoFactorAuth\Component\TOTP;
use Plugins\Ccm19TwoFactorAuth\Exception\FIDO\Exception as FIDOException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Controller for TOTP 2FA Login
 * @Route("/login/2fa", name="app_login_plugin_ccm19_twofactorauth")
 */
class Login extends UnrestrictedController
{
	/**
	 * @Route("/totp", name="_totp", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function formTotp(Request $request, PluginState $state): Response
	{
		$user = User::loggingInUser();
		if (!$user) {
			return $this->redirectToRoute('app_login', [], 303);
		}

		return $this->render('@plugin:Ccm19TwoFactorAuth/2fa_totp.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'backTo' => $request->query->get('backTo', null),
		]);
	}

	/**
	 * @Route("/totp", name="_totp_post", methods={"POST"})
	 * @return Response
	 */
	public function checkTotp(Request $request, LoggerInterface $logger, PluginState $state, LoginController $login, TranslatorInterface $translator): Response
	{
		$user = User::loggingInUser();
		if (!$user) {
			return $this->redirectToRoute('app_login', [], 303);
		}

		$username = $user->getUsername();

		sleep(1);

		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID());
		if (!$parameters or !$state->isAllowedForUser($user)) {
			return $this->render('@plugin:Ccm19TwoFactorAuth/2fa_totp.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'backTo' => null,
				'error' => 'Not allowed. Please use another authentication method.'
			]);
		}
		$totp = new TOTP($parameters);

		if ($totp->validate($request->request->getAlnum('otp', ''))) {
			$user->switchTo();

			$logger->logInfo("Login","Successful 2FA login by user: ".$username);
			return $login->redirectAfterLogin($request);

		} else {
			$logger->logError("Login", 'Failed login due to wrong 2FA code for user '.$username);
		}

		return $this->render('@plugin:Ccm19TwoFactorAuth/2fa_totp.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'backTo' => $request->query->get('backTo', null),
			'error' => 'Invalid or expired login code. Please try again.'
		]);
	}

	/**
	 * @Route("/passkey/challenge.bin", name="_passkey_challenge", methods={"GET"})
	 * @return Response
	 */
	public function challengePasskey(Request $request, FIDO $fido): Response
	{
		usleep(25000);
		try {
			$csrfToken = CsrfToken::get();
			$csrfToken->checkRequest($request);
		} catch (CsrfTokenException $e) {
			return new JsonResponse('{"success":false,"message":"invalid CSRF token"}', 403, [
				'Content-Type' => 'application/json',
				'Cache-Control' => 'no-store, no-cache',
			], true);
		}
		$user = User::loggingInUser();
		if (!$user) {
			return new JsonResponse('{"success":false,"message":"login expired"}', 403, [
				'Content-Type' => 'application/json',
				'Cache-Control' => 'no-store, no-cache',
			], true);
		}

		$challenge = $fido->generateChallenge($user->getId().':login');

		return new Response($challenge, 200, [
			'Content-Type' => 'application/octet-stream',
			'Cache-Control' => 'no-store, no-cache',
		]);
	}

	/**
	 * @Route("/passkey", name="_passkey", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function formPasskey(Request $request, PluginState $state, FIDO $fido, TranslatorInterface $translator): Response
	{
		$user = User::loggingInUser();
		if (!$user) {
			return $this->redirectToRoute('app_login', [], 303);
		}

		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey');

		$error = $request->query->get('error', null);
		if ($error === 'csrf') {
			$error = 'CSRF token is missing or invalid. Please try again.';
		}
		elseif ($error !== null) {
			$error = 'Passkey validation failed. Please try again.';
		}

		return $this->render('@plugin:Ccm19TwoFactorAuth/2fa_passkey.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'backTo' => $request->query->get('backTo', null),
			'error' => $error,
			'passkeyConfig' => [
				'rpId' => $fido->getRelyingPartyId(),
				'loginChallengeUrl' => $this->generateUrl('app_login_plugin_ccm19_twofactorauth_passkey_challenge'),
				'messages' => [
					'login' => [
						'ChallengeError' => $translator->trans('Security check failed. Please reload this page and try again.'),
						'NotSupportedError' => $translator->trans('Passkeys are not supported in this browser.'),
						'NotAllowedError' => $translator->trans('Access denied. Accessing the passkey failed.'),
						'SecurityError' => $translator->trans('Passkey access request denied by the browser. Please make sure you are accessing this page via HTTPS.'),
						'AbortError' => $translator->trans('Passkey access was aborted.'),
						'InvalidStateError' => $translator->trans('This device seems to be incompatible.'),
						'error' => $translator->trans('Accessing the passkey failed because of an unexpected error: %ERROR%'),
						'fail' => $translator->trans('This passkey did not pass verification.'),
					],
				],
				'registeredCredentials' => array_map(function ($id, $entry) {
					return [
						'type' => $entry['type'] ?? 'public-key',
						'id' => $id,
						'transports' => $entry['transports'] ?? null,
					];
				}, array_keys($parameters), $parameters ?? []),
			],
		]);
	}

	/**
	 * @Route("/passkey", name="_passkey_post", methods={"POST"})
	 * @return Response
	 */
	public function checkPasskey(Request $request, FIDO $fido, LoggerInterface $logger, PluginState $state, LoginController $login): Response
	{
		usleep(100000);
		try {
			$csrfToken = CsrfToken::get();
			$csrfToken->checkRequest($request);
		} catch (CsrfTokenException $e) {
			return $this->redirectToRoute('app_login_plugin_ccm19_twofactorauth_passkey', [
				'error' => 'csrf'
			], 303);
		}

		$user = User::loggingInUser();
		if (!$user) {
			return $this->redirectToRoute('app_login', [], 303);
		}

		$username = $user->getUsername();
		$parameters = $user->getTwoFactorAuthParameters($state->getPluginUID().'--passkey') ?? [];
		$salt = $user->getId() . ':login';
		$keyInfo = $parameters[(string)$request->request->get('passkeyId')] ?? null;
		if (!$keyInfo) {
			return $this->redirectToRoute('app_login_plugin_ccm19_twofactorauth_passkey', [
				'error' => '1',
				'backTo' => $request->query->get('backTo', null),
			], 303);
		}
		$authenticatorData = base64_decode((string)$request->request->get('authenticatorData'));
		$clientData = base64_decode((string)$request->request->get('clientData'));
		$signature = base64_decode((string)$request->request->get('signature'));
		$flags = ord($authenticatorData[32] ?? "\0");
		$signCount = unpack('N', $authenticatorData, 33)[1];

		try {
			if (!$fido->verify($keyInfo, $salt, $authenticatorData, $clientData, $signature)) {
				$logger->logError("Login", 'Failed login due to invalid passkey signature for user '.$username);
				return $this->redirectToRoute('app_login_plugin_ccm19_twofactorauth_passkey', [
					'error' => '1',
					'backTo' => $request->query->get('backTo', null),
				], 303);
			}
		} catch (FIDOException $e) {
			$logger->logError("Login", 'Failed login due to passkey error "'.$e->getMessage().'" for user '.$username);
			return $this->redirectToRoute('app_login_plugin_ccm19_twofactorauth_passkey', [
				'error' => '1',
				'backTo' => $request->query->get('backTo', null),
			], 303);
		}

		$user->switchTo();

		$logger->logInfo("Login","Successful 2FA passkey login by user: ".$username);
		return $login->redirectAfterLogin($request);
	}
}
