(function () {
'use strict';

/* TCFv2 API
 * https://github.com/InteractiveAdvertisingBureau/GDPR-Transparency-and-Consent-Framework/blob/master/TCFv2/IAB%20Tech%20Lab%20-%20CMP%20API%20v2.md
 */
/**
 * @typedef TCData
 * @type {object}
 * @property {string} tcString
 * @property {number} tcfPolicyVersion
 * @property {number} cmpId
 * @property {number} cmpVersion
 * @property {(boolean|undefined)} gdprApplies
 * @property {string} eventStatus
 * @property {string} cmpStatus
 * @property {(number|undefined)} listenerId
 * @property {boolean} isServiceSpecific
 * @property {boolean} useNonStandardTexts
 * @property {string} publisherCC
 * @property {boolean} purposeOneTreatment
 * @property {?Object} outOfBand
 * @property {?Object.<string,(boolean|undefined)>} outOfBand.allowedVendors
 * @property {?Object.<string,(boolean|undefined)>} outOfBand.disclosedVendors
 * @property {Object} purpose
 * @property {Object.<string,(boolean|undefined)>} purpose.consents
 * @property {Object.<string,(boolean|undefined)>} purpose.legitimateInterests
 * @property {Object.<string,(boolean|undefined)>} vendor.consents
 * @property {Object.<string,(boolean|undefined)>} vendor.legitimateInterests
 * @property {Object.<string,(boolean|undefined)>} specialFeatureOptins
 * @property {Object} publisher
 * @property {Object.<string,(boolean|undefined)>} publisher.consents
 * @property {Object.<string,(boolean|undefined)>} publisher.legitimateInterests
 * @property {Object} publisher.customPurpose
 * @property {Object.<string,(boolean|undefined)>} publisher.customPurpose.consents
 * @property {Object.<string,(boolean|undefined)>} publisher.customPurpose.legitimateInterests
 * @property {Object.<string,Object.<String,number>>} publisher.restrictions
 *
 * @typedef PingReturn
 * @type {object}
 * @property {(boolean|undefined)} gdprApplies
 * @property {boolean} cmpLoaded
 * @property {string} cmpStatus ('stub'/'loading'/'loaded'/'error')
 * @property {string} displayStatus ('visible'/'hidden'/'disabled')
 * @property {string} apiVersion
 * @property {(number|undefined)} cmpVersion
 * @property {(number|undefined)} cmpId
 * @property {(gvlVersion|undefined)} gvlVersion
 * @property {(tcfPolicyVersion|undefined)} tcfPolicyVersion
 *
 * @callback TCDataCallback
 * @param {?TCData} tcData
 * @param {boolean} success
 *
 * @callback PingCallback
 * @param {?PingReturn} pingReturn
 *
 * @callback RemoveListenerCallback
 * @param {boolean} success
 *
 * @typedef {TCDataCallback|PingCallback|RemoveListenerCallback} TcfCallback
 */

var CmpId = 343;
var CmpVersion = encodeCMPVersion(CCM.version);
var GvlSpecVersion = CCM._tcfInfo.repository.gvlVersion;
var _CompatVersion = +CCM._tcfInfo.repository.compatVersion; // für Kompatibilitätsmodi reserviert
var VendorListVersion = CCM._tcfInfo.repository.vlVersion;
var PolicyVersion = 4;
var ApiVersion = '2.2';

/** @type {Object.<number,TCDataCallback>} */
var TcfEventListeners = {};
/** @type {TCData} */
var TcfStateData = nameObject('TCData', {
	tcString: 'CAAAAAAAAAAAAAAAAAZZAACwAAAAAAAAADLIAAAAAAAA',
	tcfPolicyVersion: PolicyVersion,
	cmpId: CmpId,
	cmpVersion: CmpVersion,
	gdprApplies: CCM.consentRequired,
	eventStatus: CCM.consent ? 'tcloaded' : 'cmpuishown',
	cmpStatus: 'loaded',
	isServiceSpecific: true,
	useNonStandardTexts: false,
	publisherCC: CCM.primaryCountry,
	purposeOneTreatment: false,
	purpose: {
		consents: new Set(),
		legitimateInterests: new Set(),
	},
	vendor: {
		consents: new Set(),
		legitimateInterests: new Set()
	},
	specialFeatureOptins: new Set(),
	publisher: {
		consents: new Set(),
		legitimateInterests: new Set(),
		customPurpose: {
			consents: new Set(),
			legitimateInterests: new Set(),
		},
		restrictions: {},
	},
	addtlConsent: undefined,
});

/**
 * Kodieren eines CCM-Versionsstrings als Integer
 *
 * @param {string} versionString
 * @return {number}
 */
function encodeCMPVersion(versionString)
{
	var versionData1 = versionString.split('-', 2);
	var versionData2 = versionData1[0].split('.',3);
	return ((versionData2[0]-2020) << 9) | ((versionData2[1] & 0xF) << 5) | (versionData2[2] & 0x1F)
}

/**
 * Hilfsfunktion: Setzt (soweit unterstützt) den Namen des Objekts bei toString()- und Debug-Ausgaben
 *
 * @param {string} name
 * @param {Object} obj
 * @return {Object} obj
 */
function nameObject(name, obj) {
	if (Symbol && Symbol.toStringTag) {
		Object.defineProperty(obj, Symbol.toStringTag, { value: name, configurable: true });
	}
	return obj;
}

/**
 * Hilfsfunktion: Klonen eines Objekts
 *
 * @param {Object} obj
 * @param {Object=} base
 * @return {Uint6Array}
 */
function deepClone(obj, base) {
	base = (base !== undefined) ? base : {};
	var result = Object.assign(base, obj);
	for (var key in result) {
		var value = result[key];
		if (typeof value == 'object') {
			result[key] = deepClone(value, ('constructor' in value) ? new value.constructor : {});
		}
	}
	return result;
}

/**
 * Datentyp für {1:true, 2:false}-Objekte.
 *
 * @class
 * @param {(Set|Array.<boolean>|Object.<number,boolean>})=} values
 */
function Set(values) {
	if (typeof this != 'object') {
		throw new Error('Set is a constructor');
	}
	if (typeof values == 'object') {
		if ('length' in values && !(values instanceof Set)) {
			for (var i = 0; i < values.length; ++i) {
				this[values[i]] = true;
			}
		} else {
			Object.assign(this, values);
		}
	}
}
Set.prototype = nameObject('Set', Object.create(null, {
	constructor: {
		value: Set,
	},
	length: {
		get: function () {
			return Math.max(0, Math.max.apply(null, Object.keys(this)));
		},
	},
	union: {
		value: function (otherSet) {
			var result = new Set(this);
			for (var item in otherSet) {
				if (otherSet[item]) {
					result[item] = true;
				}
			}
			return result;
		},
	},
	toString: {
		value: function () {
			return '{'+Object.keys(this).filter(function (k) { return !!this[k]; }.bind(this)).join(',')+'}';
		},
	},
}));

/* ========================= TC String Erzeugung ========================= */

/**
 * @typedef Uint8Array Uint6Array
 *
 * Uint6Array als Alias für Uint8Array mit Fallback-Implementierung
 * für ältere Browser.
 */
var Uint6Array;
if ('Uint8Array' in window) {
	Uint6Array = Uint8Array;
} else {
	Uint6Array = function (length_or_array) {
		var result;
		if (typeof length_or_array == 'object') {
			result = Array.apply(this, length_or_array);
		} else {
			result = Array.call(this, length_or_array);
			result.fill(0);
		}
		return result;
	}
	Uint6Array.prototype = Array.prototype;
}

/**
 * Aneinanderhängen von 6bit-Arrays
 *
 * Dabei wird die Property `overlapTailBits` berücksichtigt. Wenn diese z.B.
 * auf dem ersten Array 2 ist, werden die ersten zwei Bit des zweiten Arrays
 * in das letzte 6-bit-Byte des ersten Arrays gemerged.
 *
 * @param {...Uint6Array} arrays
 * @return {Uint6Array}
 */
function joinUint6Arrays(firstArray) {
	var bitLength = firstArray.length*6 - (firstArray.overlapTailBits | 0);
	var bitOffset = bitLength;
	var array = firstArray;
	for (var j = 1; j < arguments.length; ++j) {
		array = arguments[j];
		bitLength += (array.length)*6 - (array.overlapTailBits | 0);
	}
	var result = new Uint6Array(Math.ceil(bitLength/6));
	for (var i = 0; i < firstArray.length; ++i) {
		result[i] = firstArray[i];
	}
	for (var j = 1; j < arguments.length; ++j) {
		var array = arguments[j];

		var offset = (bitOffset / 6) | 0;
		if (offset * 6 === bitOffset) {
			for (var i = 0; i < array.length; ++i) {deepClone
				result[offset+i] = array[i];
			}
		} else {
			var bitShift = bitOffset - (offset * 6);

			for (var i = 0; i < array.length; ++i) {
				result[offset+i] |= (array[i] & 0x3F) >> bitShift;
				result[offset+i+1] |= (array[i] << (6-bitShift)) & 0x3F;
			}
		}
		bitOffset += (array.length*6) - (array.overlapTailBits | 0)
	}
	result.overlapTailBits = result.length*6 - bitLength;
	return result;
}

/**
 * Umwandlung von 6-bit-Arrays in einen Base64-String
 *
 * @param {Uint6Array} array
 * @return {string}
 */
function uint6ArraytoBase64(array) {
	var result = Array.prototype.map.call(array, function (i) {
		return "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_".charAt(i);
	}).join('');

	// Auf 8-bit-Grenzen vervollständigen. Einige Parser ignorieren ansonsten
	// die letzten Bits...
	var realBitLength = array.length * 6;
	var relevantBitLength = realBitLength - (array.overlapTailBits | 0);
	var protudingBits = (relevantBitLength % 8) - (realBitLength - relevantBitLength);
	if (protudingBits == 1) {
		result += 'AA';
	}
	else if (protudingBits > 1) {
		result += 'A';
	}
	// Base64-String auf ein Vielfaches von 4 auffüllen, um alle Parser glücklich zu machen.
	var mod4 = result.length % 4;
	if (mod4 == 1) {
		result += 'AAA';
	} else if (mod4 == 2) {
		result += 'AA';
	} else if (mod4 == 3) {
		result += 'A';
	}

	return result;
}

/**
 * Bitmap/Bitfield in 6-bit-Blöcken generieren
 *
 * @param {Object} obj
 * @param {number} length
 * @return {Uint6Array}
 */
function makeBitmap6(obj, length)
{
	var result = new Uint6Array(Math.ceil(length/6));
	for (var key in obj) {
		if (obj[key]) {
			key = parseInt(key) - 1;
			result[(key / 6) | 0] |= 1 << (5 - (key % 6));
		}
	}
	result.overlapTailBits = (6 - (length) % 6) % 6;
	return result;
}

/**
 * ID-Ranges sammeln (z.B. für PurposeConsents)
 *
 * Aus {1:true, 2: true, 4: true} wird z.B.
 * [{start:1, end:2}, {start:4, end:4}]
 *
 * @param {Object} obj
 * @return {Array}
 */
function collectRanges(obj)
{
	var ranges = [];
	var keys = Object.keys(obj).map(function (x) { return x|0; }).sort(function (a,b) { return a - b; });
	for (var i = 0; i < keys.length; ++i) {
		var key = keys[i];
		if (obj[key]) {
			key = parseInt(key);
			var lastRange = ranges[ranges.length-1];
			if (lastRange && lastRange.end == key-1) {
				lastRange.end = key;
			} else {
				ranges.push({start:key, end:key});
			}
		}
	}
	return ranges;
}

/**
 * Kodieren von Ranges für den TC-String
 *
 * @param {Array} ranges Range-Array von collectRanges()
 * @return {Uint6Array}
 */
function encodeRanges(ranges)
{
	// Ranges kodieren
	var encodedRanges = ranges.map(function (range) {
		var result;
		if (range.start == range.end) {
			result = new Uint6Array([
				(0 << 5) | ((range.start >> 11) & 0x1F),
				(range.start >> 5) & 0x3F,
				(range.start & 0x1F) << 1,
			]);
			result.overlapTailBits = 1;
			return result;
		} else {
			result = new Uint6Array([
				(1 << 5) | (range.start >> 11) & 0x1F,
				(range.start >> 5) & 0x3F,
				((range.start & 0x1F) << 1) | ((range.end >> 15) & 0x01),
				(range.end >> 9) & 0x3F,
				(range.end >> 3) & 0x3F,
				(range.end & 0x7) << 3,
			]);
			result.overlapTailBits = 3;
			return result;
		}
	});
	// Anzahl davor setzen
	encodedRanges.unshift(new Uint6Array([
		(ranges.length >> 6) & 0x3F,
		ranges.length & 0x3F,
	]));
	// Payload zusammensetzen
	return joinUint6Arrays.apply(null, encodedRanges);
}

/**
 * Generieren eines Bitfields/einer Range-Liste (je nachdem was kürzer ist)
 *
 * z.B. für PurposeConsents im TC-String
 *
 * @param {Object} obj
 * @return {Uint6Array}
 */
function makeBitfieldOrRange(obj)
{
	var payload, header;
	var ranges = collectRanges(obj);
	var maxVendorId = (ranges.length) ? ranges[ranges.length-1].end : 0;

	var rangesLength = 12;
	for (var i = 0; i < ranges.length; ++i) {
		if (ranges[i].start == ranges[i].end) {
			rangesLength += 17;
		} else {
			rangesLength += 33;
		}
	}

	var useRangeEncoding = ((rangesLength < maxVendorId) && (rangesLength < 4096));
	header = new Uint6Array([
		(maxVendorId >> 10) & 0x3F,
		(maxVendorId >> 4) & 0x3F,
		((maxVendorId & 0x0F) << 2) | (useRangeEncoding << 1),
	]);
	header.overlapTailBits = 1;

	if (ranges.length == 0) {
		return header;
	}
	else if (useRangeEncoding) {
		payload = encodeRanges(ranges);
	}
	else {
		payload = makeBitmap6(obj, maxVendorId);
	}

	// Header und Payload zusammensetzen und zurückgeben
	return joinUint6Arrays(header, payload);
}

/**
 * Vereinfachen von Restriction-Ranges
 *
 * Nimmt an, dass es bei allen IDs, die nicht in exceptionSet vorkommen, egal
 * ist, ob sie gesetzt sind oder nicht.
 *
 * @param {Array} ranges
 * @param {Set=} exceptionSet
 */
function simplifyRanges(ranges, exceptionSet)
{
	if (exceptionSet === undefined) {
		exceptionSet = new Set();
	}
	if (ranges.length == 0) {
		return [];
	}
	var first = ranges[0].start;
	var last = ranges[ranges.length-1].end;
	var result = [];
	var currentRange = null;
	for (var i = first; i <= last; ++i) {
		if (i in exceptionSet && exceptionSet[i]) {
			currentRange = null;
		}
		else if (!currentRange) {
			currentRange = { start: i, end: i };
			result.push(currentRange);
		}
		else {
			currentRange.end = i;
		}
	}
	return result;
}

/**
 * Generieren einer publisherRestrictions-Liste für den TC-String
 *
 * @param {Object} obj
 * @return {Uint6Array}
 */
function makePubRestrictionRanges(obj)
{
	var payloads = [], header;

	// Restrictions bauen
	for (var purposeId in obj) {
		var purposeData = obj[purposeId];
		var restrictions = [new Set(),new Set(),new Set()];
		for (var vendorId in purposeData) {
			restrictions[purposeData[vendorId]][vendorId] = true;
		}
		for (var restrictionType = 0; restrictionType < 3; ++restrictionType) {
			if (Object.keys(restrictions[restrictionType]).length) {
				var payloadHeader = new Uint6Array([
					purposeId & 0x3F,
					(restrictionType & 0x3) << 4
				]);
				payloadHeader.overlapTailBits = 4;

				var payloadRanges = collectRanges(restrictions[restrictionType]);
				if (restrictionType == 1) {
					// Optimieren des häufigsten Falls. Geht so gut, bis wir Legitimate
					// Interest vollständig unterstützen. Dann ist eine Überarbeitung nötig.
					var exceptionSet = restrictions[0].union(restrictions[2]);
					payloadRanges = simplifyRanges(payloadRanges, exceptionSet);
				}

				payloads.push(joinUint6Arrays(
					payloadHeader,
					encodeRanges(payloadRanges)
				));
			}
		}
	}

	// Header bauen
	var header = new Uint6Array([
		(payloads.length >> 6) & 0x3F,
		payloads.length & 0x3F,
	]);

	// Header und Payloads zusammensetzen und zurückgeben
	payloads.unshift(header);
	return joinUint6Arrays.apply(null, payloads);
}

/**
 * Nullen der Zeit in einem Timestamp
 *
 * @param {number} timestamp Timestamp in 1/10-Sekunden
 * @return {number} Zeitstempel in 1/10-Sekunden mit Tagesgenauigkeit
 */
function zeroTimeInTS(timestamp) {
	return ((timestamp/864000) | 0) * 864000;
}

/**
 * Generieren des Core-Segments des TC-Strings
 *
 * @param {TCData} data
 * @param {?number} created Timestamp in 1/10-Sekunden
 * @param {?number} updated Timestamp in 1/10-Sekunden
 * @param {string} consentLanguage Sprachcode in dem der Consent gegeben wurde
 * @return {string}
 */
function encodeTCCoreString(data, created, updated, consentLanguage) {
	updated = zeroTimeInTS((updated) ? updated : Math.round((new Date()).getTime()/100));
	created = (created) ? zeroTimeInTS(created) : updated;
	var consentScreen = 0;
	consentLanguage = (consentLanguage.length == 2) ? consentLanguage.toLowerCase() : 'zz';
	var publisherCC = (data.publisherCC.length == 2) ? data.publisherCC.toLowerCase() : 'zz';
	var codeA = 'a'.charCodeAt(0);

	var header = new Uint6Array([
		2, /* format version */
		(created / (1 << 30)) & 0x3F,
		(created >> 24) & 0x3F,
		(created >> 18) & 0x3F,
		(created >> 12) & 0x3F,
		(created >> 6) & 0x3F,
		created & 0x3F,
		(updated / (1 << 30)) & 0x3F,
		(updated >> 24) & 0x3F,
		(updated >> 18) & 0x3F,
		(updated >> 12) & 0x3F,
		(updated >> 6) & 0x3F,
		updated & 0x3F,
		(data.cmpId >> 6) & 0x3F,
		data.cmpId & 0x3F,
		(data.cmpVersion >> 6) & 0x3F,
		data.cmpVersion & 0x3F,
		consentScreen & 0x3F,
		consentLanguage.charCodeAt(0) - codeA,
		consentLanguage.charCodeAt(1) - codeA,
		(VendorListVersion >> 6) & 0x3F,
		VendorListVersion & 0x3F,
		data.tcfPolicyVersion & 0x3F,
		(data.isServiceSpecific ? 0x20 : 0) | (data.useNonStandardTexts ? 0x10 : 0)
	]);
	header.overlapTailBits = 4;

	var PublisherCCBlock = new Uint6Array([
		((data.purposeOneTreatment|0) << 5) | (((publisherCC.charCodeAt(0) - codeA) >> 1) & 0x1F),
		(((publisherCC.charCodeAt(0) - codeA) << 5) & 0x20) | (((publisherCC.charCodeAt(1) - codeA) >> 1) & 0x1F),
		((publisherCC.charCodeAt(1) - codeA) << 5) & 0x20,
	]);
	PublisherCCBlock.overlapTailBits = 5;

	var result = joinUint6Arrays(
		header,
		makeBitmap6(data.specialFeatureOptins, 12),
		makeBitmap6(data.purpose.consents, 24),
		makeBitmap6(data.purpose.legitimateInterests, 24),
		PublisherCCBlock,
		makeBitfieldOrRange(data.vendor.consents),
		makeBitfieldOrRange(data.vendor.legitimateInterests),
		makePubRestrictionRanges(data.publisher.restrictions)
	);

	return uint6ArraytoBase64(result);
}

/**
 * Generieren des Core-Segments des TC-Strings
 *
 * @param {TCData} data
 * @return {string}
 */
function encodeTCPublisherString(data) {
	var header = new Uint6Array([
		3<<3, /* segment type */
	]);
	header.overlapTailBits = 3;

	var maxCustomPurpose = Math.min(Math.max(
		Math.max.apply(null, Object.keys(data.publisher.customPurpose.consents)),
		Math.max.apply(null, Object.keys(data.publisher.customPurpose.legitimateInterests)),
		0
	), 63);

	var result = joinUint6Arrays(
		header,
		makeBitmap6(data.publisher.consents, 24),
		makeBitmap6(data.publisher.legitimateInterests, 24),
		[maxCustomPurpose & 0x3F],
		makeBitmap6(data.publisher.customPurpose.consents, maxCustomPurpose),
		makeBitmap6(data.publisher.customPurpose.legitimateInterests, maxCustomPurpose)
	);
	return uint6ArraytoBase64(result);
}


/**
 * Generieren des TC-Strings
 *
 * @param {TCData} data
 * @param {?number} timestamp Timestamp in 1/100-Sekunden
 * @param {string} consentLanguage Sprachcode in dem der Consent gegeben wurde
 * @return {string}
 */
function encodeTCString(data, timestamp, consentLanguage) {

	var coreString = encodeTCCoreString(data, timestamp, timestamp, consentLanguage);
	return coreString;

	// Publisher String wegen Problemen vorerst deaktiviert
	//var publisherString = encodeTCPublisherString(data);
	//return coreString+'.'+publisherString;
}

/**
 * Aktualisierem des TCF-Zustands
 */
function updateTCState()
{
	var useConsent = (TcfStateData.eventStatus !== 'cmpuishown');
	var tcfData = CCM._tcfInfo;
	var vendors = tcfData.repository.vendor || {};
	var embeddings = CCM.acceptedEmbeddings || [];
	var legIntEnabled = !!tcfData.legIntEnabled;

	// Special Feature Opt-In
	TcfStateData.specialFeatureOptins = new Set(useConsent ? tcfData.consent.sf : undefined);
	// Purposes
	TcfStateData.purpose.consents = new Set(useConsent ? tcfData.consent.p : undefined);
	TcfStateData.purpose.legitimateInterests = new Set(tcfData.consent.lip);

	// Restrictions bestimmen
	var restrictions = {};
	var acceptedVendors = new Set(useConsent ? tcfData.consent.v : undefined);
	var legIntVendors = new Set(tcfData.consent.liv);

	// TCF-Vendors aus Drittanbieter-Liste
	tcfData.consent.v.forEach(function (vendorId) {
		if (vendorId in vendors) {
			if (!legIntEnabled) {
				// Wenn LegInt-Funktion nicht aktiv:
				vendors[vendorId].pF.forEach(function (purposeId) {
					if (!(purposeId in restrictions)) {
						restrictions[purposeId] = {};
					}
					restrictions[purposeId][vendorId] = 1; // Nur Consent, kein LegInt
				});
			}
			vendors[vendorId].p.forEach(function (purposeId) {
				if (purposeId in restrictions) {
					delete restrictions[purposeId][vendorId];
				}
			});
		}
	});
	// Embeddings mit gekoppeltem TCF-Vendor
	embeddings.forEach(function (embedding) {
		if (embedding._tcf) {
			for (var vendorId in embedding._tcf) {
				var acceptedPurposes = embedding._tcf[vendorId].p;

				acceptedVendors[vendorId] = (useConsent);

				vendors[vendorId].pF.forEach(function (purposeId) {
					if (!(purposeId in restrictions)) {
						restrictions[purposeId] = {};
					}
					restrictions[purposeId][vendorId] = 0;
				});
				vendors[vendorId].p.forEach(function (purposeId) {
					if (!(purposeId in restrictions)) {
						restrictions[purposeId] = {};
					}
					restrictions[purposeId][vendorId] = 0;
				});
				vendors[vendorId].pL.forEach(function (purposeId) {
					if (!(purposeId in restrictions)) {
						restrictions[purposeId] = {};
					}
					restrictions[purposeId][vendorId] = 0;
				});

				acceptedPurposes.forEach(function (purposeId) {
					if (vendors[vendorId].p.indexOf(purposeId) > -1) {
						delete restrictions[purposeId][vendorId];
					} else {
						restrictions[purposeId][vendorId] = 1; // Nur Consent, kein LegInt
					}
				});

			}
		}
	});

	TcfStateData.vendor.consents = acceptedVendors;
	TcfStateData.vendor.legitimateInterests = legIntVendors;
	TcfStateData.publisher.restrictions = restrictions;

	// Google Additional Consent
	var googleAdditionalConsent = tcfData.consent.gad;
	if (tcfData.repository.googleACMode && useConsent) {
		TcfStateData.addtlConsent = (googleAdditionalConsent) ? '1~'+googleAdditionalConsent.join('.') : '1~';
	} else {
		TcfStateData.addtlConsent = undefined;
	}

	// TC-String aktualisieren
	TcfStateData.tcString = encodeTCString(
		TcfStateData,
		(useConsent)
			? ((tcfData.consent.updated) ? tcfData.consent.updated : tcfData.consent.created)
			: (('now' in Date) ? Date.now() : new Date().getTime()) / 100,
		CCM.consentLanguage.replace('-', '_').split('_')[0]
	);
}

/**
 * Aktualisiert Makros in Links, die vom Hauptscript gefunden wurden.
 */
function updateLinkMacros() {
	document.querySelectorAll('a[data-ccm-href][href]').forEach(function (node) {
		node.href = node.dataset.ccmHref.replace(
			/\$\{GDPR(_CONSENT_[0-9]{1,10})?\}/g,
			function (match, want_consent_string) {
				if (want_consent_string) {
					return TcfStateData.tcString;
				} else {
					return (TcfStateData.gdprApplies) ? '1' : '0';
				}
			}
		);
	});
}

/**
 * Aktualisiert Makros in Bildern, die vom Hauptscript gefunden wurden.
 */
function updateImageMacros() {
	document.querySelectorAll('img[data-ccm-src][src]').forEach(function (node) {
		node.src = node.dataset.ccmSrc.replace(
			/\$\{GDPR(_CONSENT_[0-9]{1,10})?\}/g,
			function (match, want_consent_string) {
				if (want_consent_string) {
					return TcfStateData.tcString;
				} else {
					return (TcfStateData.gdprApplies) ? '1' : '0';
				}
			}
		);
	});
}

/**
 * Prüft ob ein Wert `undefined`, `null` oder ein leeres Objekt ist.
 * @param {} value
 * @return {boolean}
 */
function isNone(value) {
	return (value === undefined || value === null || (typeof value === 'object' && Object.keys(value).length === 0));
}

/**
 * @return {void}
 */
function _noCallback() {}

/**
 * Ruft alle registrierten Event Listener auf
 */
function callEventListeners() {
	if (TcfStateData.eventStatus !== 'cmpuishown') {
		updateImageMacros();
		updateLinkMacros();
	}
	for (var listenerId in TcfEventListeners) {
		try {
			var callback = TcfEventListeners[listenerId];
			// Datentyp der Listener-ID korrigieren
			listenerId = parseInt(listenerId);
			// Callback auslösen
			TcfCommands.getTCData(2, callback, undefined, listenerId);
		} catch (e) {
			console.error('[CCM19] error in TCF event listener callback %d', listenerId, e, callback);
		}
	}
}

/**
 * Event-Listener für das (Wieder-)Öffnen des Widgets
 */
window.addEventListener('ccm19WidgetOpened', function () {
	if (TcfStateData.eventStatus !== 'cmpuishown') {
		TcfStateData.eventStatus = 'cmpuishown';
		updateTCState();
		callEventListeners();
	}
}, false);

/**
 * Event-Listener für das Schließen des Widgets
 */
window.addEventListener('ccm19WidgetClosed', function () {
	if (TcfStateData.eventStatus === 'cmpuishown') {
		TcfStateData.eventStatus = 'useractioncomplete';
		updateTCState();
		if ('adsbygoogle' in window) {
			window.adsbygoogle.pauseAdRequests=0;
		}
		callEventListeners();
	}
}, false);

/**
 * Implementierte TCF-Kommandos
 */
var TcfCommands = {
	/**
	 * @param {number} version
	 * @param {PingCallback} callback
	 */
	ping: function (version, callback)
	{
		if (version == 1 || version > 2) {
			callback(null, false);
			return;
		}

		var data = nameObject('PingReturn', {
			gdprApplies: CCM.consentRequired,
			cmpLoaded: true,
			cmpStatus: 'loaded',
			displayStatus: CCM.consentRequired ? (document.querySelector('.ccm-modal.ccm-show') ? 'visible' : 'hidden') : 'disabled',
			apiVersion: ApiVersion,
			cmpVersion: CmpVersion,
			cmpId: CmpId,
			gvlVersion: VendorListVersion,
			tcfPolicyVersion: PolicyVersion,
		});
		callback(data, true);
		return data;
	},

	/**
	 * @param {number} version
	 * @param {TCDataCallback} callback
	 * @param {number[]=} vendorIds
	 * @param {number=} listenerId
	 */
	getTCData: function (version, callback, vendorIds, listenerId)
	{
		if (version == 1 || version > 2) {
			callback(null, false);
			return;
		}

		var data = deepClone(TcfStateData, { listenerId: listenerId });
		callback(data, true);
		return data;
	},

	/**
	 * @param {number} version
	 * @param {TCDataCallback} callback
	 */
	addEventListener: function (version, callback)
	{
		if (version == 1 || version > 2) {
			callback(null, false);
			return;
		}

		// Zufällige Listener-ID erstellen
		var listenerId, i;
		for (i = 0; i < 30; ++i) {
			listenerId = Math.random()*2147483647 | 0;
			if (!(listenerId in TcfEventListeners)) {
				break;
			}
		}
		// Abbrechen, wenn eindeutige ID nicht erstellt werden konnte
		if (i >= 30) {
			callback(null, false);
			return;
		}

		TcfEventListeners[listenerId] = callback;
		return TcfCommands.getTCData(version, callback, undefined, listenerId);
	},

	/**
	 * @param {number} version
	 * @param {RemoveListenerCallback} callback
	 * @param {number} listenerId
	 */
	removeEventListener: function (version, callback, listenerId)
	{
		if (version == 1 || version > 2) {
			callback(false, false);
			return false;
		}

		if (listenerId in TcfEventListeners) {
			delete TcfEventListeners[listenerId];
			callback(true, true);
			return true;
		} else {
			callback(false, true);
			return false;
		}
	},
};

/**
 * @param {string} command
 * @param {number} version
 * @param {TcfCallback} callback
 * @param {*=} parameter
 */
function TCFAPI_CCM19(command, version, callback, parameter)
{
	if (isNone(callback)) {
		callback = _noCallback;
	}
	if (command in TcfCommands) {
		return TcfCommands[command](version, callback, parameter);
	}
	else {
		try {
			callback(null, false);
		} catch (e) {
			console.error('[CCM19] TCF command "%s" unknown/not implemented and invalid callback provided', command);
		}
	}
}

// TC-String initial generieren
updateTCState();

// Referenz auf ggf. vorher vorhandenen Stub holen
var stub = window.__tcfapi;

// TCF-Implementation verfügbar machen
window.__tcfapi = TCFAPI_CCM19;

// TCF-Iframe-Handler wird schon vom Stub erzeugt.

// Makros direkt bearbeiten und AdSense fortsetzen, wenn Consent schon erteilt
if (TcfStateData.eventStatus === 'tcloaded') {
	updateImageMacros();
	updateLinkMacros();
	if ('adsbygoogle' in window) {
		window.adsbygoogle.pauseAdRequests=0;
	}
}

// Wenn Stub vorhanden, am Ende Queue abarbeiten
if (stub) {
	var queue = stub();
	for (var i = 0; i < queue.length; ++i) {
		TCFAPI_CCM19.apply(null, queue[i]);
	}
	queue = i = stub = undefined;
}

})();
