<?php
namespace App\Command\Consent;

use App\Utils;
use App\Model\User;
use App\Model\Domain;
use App\Model\Protocol;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use DateTime;

class MigrateCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:consent:migrate';

	protected function configure(): void
	{
		$this
			->setDefinition([
				new InputOption('from', 'f', InputOption::VALUE_REQUIRED, 'Migrate from another archive storage (<info>json</info>, <info>mongodb</info> or <info>redis</info>; default: currently configured storage)'),
				new InputOption('to', 't', InputOption::VALUE_REQUIRED, 'Migrate to another archive storage (<info>json</info>, <info>mongodb</info> or <info>redis</info>; default: currently configured storage)'),
				new InputOption('user', 'u', InputOption::VALUE_REQUIRED|InputOption::VALUE_IS_ARRAY, 'User id or name'),
			])
			->setDescription('Migrates active consent protocols (EXPERIMENTAL)')
			->setHelp("Copies all non-archived consent protocols (or if <info>-u</info> is used, those of specific users) to another storage.\n\n<comment>Warning:</comment> running this multiple times can cause duplicates.")
		;
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$io = new SymfonyStyle($input, $output);
		$from = $input->getOption('from') ?? Utils::getArchiveRepositoryStore();
		$to = $input->getOption('to') ?? Utils::getArchiveRepositoryStore();

		// Check input
		$ok = true;

		$valid_storages = [Utils::REPOSITORY_STORE_FILE, Utils::REPOSITORY_STORE_MONGODB, Utils::REPOSITORY_STORE_REDIS];
		if (!in_array($from, $valid_storages, true)) {
			$io->error("--from has an invalid value.");
			$ok = false;
		}
		if (!in_array($to, $valid_storages, true)) {
			$io->error("--to has an invalid value.");
			$ok = false;
		}

		if ($to === $from) {
			$io->error("Cannot migrate to and from the same storage.");
			if ($input->getOption('from') === null and $input->getOption('to') === null)
			$io->comment("Please add at least <info>--to</info> or <info>--from</info> to do a migration.");
			$ok = false;
		}

		$usernames = $input->getOption('user');
		$users = array_combine($usernames, array_map(function ($username) {
			return User::byId($username) ?? User::find($username);
		}, $usernames));

		foreach ($users as $username => $user) {
			if ($user === null) {
				$io->error("User '".htmlspecialchars($username)."' not found.");
				$ok = false;
			}
		}

		// Abbrechen wenn Validierung fehlschlägt
		if (!$ok) {
			return 1;
		}

		// Benutzerliste fertig bestimmen
		if (!$users) {
			if (Utils::isExtendedEdition()) {
				$users = User::getClients();
			} else {
				$users = User::byRole(User::ROLE_ADMIN);
			}
		}

		// Migration durchführen
		foreach ($users as $user) {
			if (!User::setApiUser($user->getApiKey())) {
				continue;
			}
			$io->comment("Processing user '".htmlspecialchars($user->getUsername())."'");
			Domain::reloadRepository();
			foreach (Domain::all() as $domain) {
				$io->comment(" Processing domain '".htmlspecialchars($domain->getName())."'");
				Domain::setDomainForRequest($domain->getId());

				// Quell-Repo öffnen
				$_ENV['APP_PROTOCOL_STORE'] = $from;
				$sourceProtocol = Protocol::byDomain($domain);

				// Ziel-Repo öffnen
				$_ENV['APP_PROTOCOL_STORE'] = $to;
				$destProtocol = Protocol::byDomain($domain);

				// Einträge kopieren
				$source = $sourceProtocol->filter(new DateTime('2019-01-01'), null);
				$count = $destProtocol->writeBulk($source);

				$io->comment("  Copied $count log entries.");

				unset($source, $sourceProtocol, $destProtocol);
			}
		}

		$io->success("Done.");

		return 0;
	}

}
