<?php
namespace App\Command\Cron;

use App\Config;
use App\Utils;
use App\Model\User;
use App\Model\Domain;
use App\Model\CronState;
use App\Model\UserCronState;
use App\Component\Cron\Cron;
use App\Component\Cron\Job;
use App\Component\Cron\DomainJob;
use App\Component\Cron\UserJob;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;

class CronWorkerCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:cron:worker';

	/// @var Cron $cron
	private $cron;

	/// @var RequestStack $stack
	private $stack;

	/// @var string|null $currentUserId
	private $currentUserId = null;

	/// @var string|null $currentDomainId
	private $currentDomainId = null;

	public function __construct(Cron $cron, RequestStack $stack)
	{
		$this->stack = $stack;
		$this->cron = $cron;
		parent::__construct();
	}

	protected function configure(): void
	{
		$this
			->setHidden(true)
			->setDefinition([
				new InputOption('deadline', null, InputOption::VALUE_REQUIRED, 'Deadline', 0),
				new InputOption('domain', 'd', InputOption::VALUE_REQUIRED|InputOption::VALUE_IS_ARRAY, 'Domain id'),
				new InputOption('force', 'f', InputOption::VALUE_NONE, 'Even run cron jobs which aren\'t due yet'),
			])
			->setDescription('Internal: Worker for parallel cron job execution')
		;
	}

	public function prepareRequestContext(): ?Request
	{
		// Basis-URL bestimmen
		$config = Config::getInstance();
		$baseUrl = $config->getBaseUrl();
		if (!$baseUrl) {
			return null;
		}
		$baseUrlParts = parse_url($baseUrl);
		$path = $baseUrlParts['path'] ?? '';

		// HTTP-Request-Umgebung vorbereiten
		$_SERVER['REQUEST_URI'] = rtrim($path, '/').'/cron';
		$_SERVER['SCRIPT_NAME'] = $_SERVER['PHP_SELF'] = stripos($path, 'index.php') !== false ? $path : rtrim($path, '/').'/index.php';
		$_SERVER['SCRIPT_FILENAME'] = Utils::getBaseDir() . '/index.php';

		$_SERVER['HTTP_HOST'] = $baseUrlParts['host'];
		if (!empty($baseUrlParts['port'])) {
			$_SERVER['HTTP_HOST'] .= ':'.$baseUrlParts['port'];
			$_SERVER['SERVER_PORT'] = (int)$baseUrlParts['port'];
		} else {
			$_SERVER['SERVER_PORT'] = (strtolower($baseUrlParts['scheme']) === 'https') ? 443 : 80;
		}
		$_SERVER['REMOTE_ADDR'] = '127.0.0.1';
		$_SERVER['SERVER_ADDR'] = '127.0.0.1';
		$_SERVER['REQUEST_METHOD'] = 'POST';
		$_SERVER['REQUEST_SCHEME'] = $baseUrlParts['scheme'];
		if (strtolower($baseUrlParts['scheme']) === 'https') {
			$_SERVER['HTTPS'] = 'on';
		}
		$request = Request::createFromGlobals();
		$request->setSession($this->stack->getSession());
		Utils::getRouter()->getContext()->fromRequest($request);

		return $request;
	}

	public function setContext(User $user, ?string $domainId): bool
	{
		if ($user->getId() === $this->currentUserId) {
			if ($domainId === null or $domainId === $this->currentDomainId) {
				return true;
			}
		}
		if ($this->cron->setContext($user, $domainId)) {
			if ($domainId === null) {
				$firstDomain = Domain::getFirst() ?: null;
				if ($firstDomain) {
					$domainId = $firstDomain->getId();
					if (!$this->cron->setContext($user, $domainId)) {
						$domainId = null;
					}
				}
			}
			$this->currentUserId = $user->getId();
			$this->currentDomainId = $domainId;
			return true;
		} else {
			return false;
		}
	}

	public function runSystemJob(Job $job, Request $request, Config $config, SymfonyStyle $io)
	{
		$io->comment("Running '".htmlspecialchars($job->getName())."' ...");
		try {
			$job->run($request);
		} finally {
			// Konfiguration schreiben, damit sie im nächsten Cronjob direkt verfügbar ist
			$config->flush();
		}
		$io->success("'".htmlspecialchars($job->getName())."' completed successfully.");
	}

	public function runUserJob(UserJob $job, Request $request, Config $config, SymfonyStyle $io)
	{
		$io->comment("Running '".htmlspecialchars($job->getName())."' ...");
		try {
			$job->run($request);
		} finally {
			// Konfiguration schreiben, damit sie im nächsten Cronjob direkt verfügbar ist
			$config->flush();
			UserCronState::flush();
		}
		$io->success("'".htmlspecialchars($job->getName())."' completed successfully.");
	}

	public function runDomainJob(DomainJob $job, Request $request, Config $config, SymfonyStyle $io)
	{
		$io->comment("Running '".htmlspecialchars($job->getName())."' ...");
		try {
			$job->run($request);
		} finally {
			// Konfiguration schreiben, damit sie im nächsten Cronjob direkt verfügbar ist
			$config->flush();
			CronState::flush();
		}
		$io->success("'".htmlspecialchars($job->getName())."' completed successfully.");

	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$io = new SymfonyStyle($input, $output);
		$socket = fopen('php://fd/0', 'r+');

		try {
			$config = Config::getInstance();
			$request = $this->prepareRequestContext();
			if (!$request) {
				$io->error("Base URL not configured. Aborted.");
				return 1;
			}
			$this->stack->push($request);

			$adminUser = current(User::byRole(User::ROLE_ADMIN));
			$deadline = (float)$input->getOption('deadline');
			$force = $input->getOption('force');
			$domainIds = $input->getOption('domain');

			fwrite($socket, 'READY');
			while (true) {
				$read = [$socket];
				$write = [];
				$except = [];
				if ($deadline !== 0) {
					$timeout = (int)($deadline - microtime(true));
					if ($timeout <= 0) {
						break;
					}
				} else {
					$timeout = null;
				}
				$result = stream_select($read, $write, $except, $timeout);

				if ($result === 0 or ($deadline and $deadline <= microtime(true))) {
					// Deadline erreicht
					break;
				}
				if ($result === false) {
					continue;
				}

				$command = fread($socket, 4095);
				if ($command === false) {
					if (feof($socket)) {
						break;
					} else {
						continue;
					}
				}
				$command = explode(' ', rtrim($command));
				try {
					switch ($command[0]) {
						case 'EXIT':
						case '':
							break 2;

						case 'SYSTEM':
							$this->setContext($adminUser, null);
							$jobs = array_filter(array_map(function ($name) {
								$result = $this->cron->jobByName($name);
								assert(!$result or $result instanceof Job);
								return $result;
							}, array_slice($command, 1)));
							foreach ($jobs as $job) {
								if ($force or $job->isDue()) {
									$this->runSystemJob($job, $request, $config, $io);
									if ($deadline and $deadline <= microtime(true)) {
										break;
									}
								}
							}
							fwrite($socket, 'SUCCESS SYSTEM');
							break;

						case 'USER':
							$user = User::byId($command[1] ?? '');
							$jobs = array_filter(array_map(function ($name) { return $this->cron->jobByName($name); }, array_slice($command, 2)));
							$userJobs = array_filter($jobs, function ($job) { return $job instanceof UserJob; });
							$domainJobs = array_filter($jobs, function ($job) { return $job instanceof DomainJob; });
							if (!$user) {
								fwrite($socket, 'ERROR Unknown user: ' . $command[1]);
								break;
							}
							if (!$this->setContext($user, null)) {
								fwrite($socket, 'ERROR Context switch failed');
								break;
							}

							$io->comment(sprintf("User '%s':",
								htmlspecialchars($user->getUsername())
							));

							foreach ($userJobs as $job) {
								if ($force or $job->isDue()) {
									assert($job instanceof UserJob);
									$this->runUserJob($job, $request, $config, $io);
									if ($deadline and $deadline <= microtime(true)) {
										break;
									}
								}
							}

							if ($domainJobs) {
								$domains = Domain::all();
								shuffle($domains);
								if ($domainIds) {
									$domains = array_filter($domains, function ($domain) use ($domainIds) { return in_array($domain->getId(), $domainIds); });
								}

								foreach ($domains as $domain) {
									$io->comment(sprintf("Domain '%s' [%s] of '%s':",
										htmlspecialchars($domain->getName()),
										htmlspecialchars($domain->getId()),
										htmlspecialchars($user->getUsername())
									));
									if (!$this->setContext($user, $domain->getId())) { continue; }

									// Domain-Jobs ausführen
									foreach ($domainJobs as $job) {
										assert($job instanceof DomainJob);
										if ($force or $job->isDue()) {
											$this->runDomainJob($job, $request, $config, $io);

											if ($deadline and $deadline <= microtime(true)) {
												break;
											}
										}
									}
								}
							}

							fwrite($socket, 'SUCCESS USER');
							break;

						default:
							fwrite($socket, 'ERROR Unknown command ' . json_encode($command[0]));
							break;
					}
				}
				catch (\Throwable $e) {
					fwrite($socket, 'EXCEPTION ' . $e);
				}
			}

			$this->stack->pop();
		} finally {
			@fwrite($socket, "EXITED");
		}
		return 0;
	}

}
