<?php
namespace App\Command\Cron;

use App\Config;
use App\Utils;
use App\Model\User;
use App\Model\Domain;
use App\Model\CronState;
use App\Component\Cron\Cron;
use App\Component\Cron\Job;
use App\Component\Cron\DomainJob;
use App\Component\Cron\UserJob;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;

class RunCronCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:cron:run';

	/// @var Cron $cron
	private $cron;

	/// @var RequestStack $stack
	private $stack;

	/// @var CronWorkerCommand $stack
	private $worker;

	public function __construct(Cron $cron, RequestStack $stack, CronWorkerCommand $worker)
	{
		$this->stack = $stack;
		$this->cron = $cron;
		$this->worker = $worker;
		parent::__construct();
	}

	protected function configure(): void
	{
		$this
			->setDefinition([
				new InputOption('timeout', 't', InputOption::VALUE_REQUIRED, 'Timeout in seconds', -1.0),
				new InputOption('user', 'u', InputOption::VALUE_REQUIRED|InputOption::VALUE_IS_ARRAY, 'User id or name'),
				new InputOption('domain', 'd', InputOption::VALUE_REQUIRED|InputOption::VALUE_IS_ARRAY, 'Domain id'),
				new InputOption('force', 'f', InputOption::VALUE_NONE, 'Even run cron jobs which aren\'t due yet'),
				new InputOption('parallel', 'p', InputOption::VALUE_REQUIRED, 'The number of jobs to run in parallel (experimental)'),
			])
			->addArgument('job', InputArgument::OPTIONAL|InputArgument::IS_ARRAY, 'Job name')
			->setDescription('Runs cron jobs')
			->setHelp("Runs due cron jobs for all domains and users. If no job name is specified, all due cron jobs will be run in random order.")
		;
	}

	private function spawnWorker(bool $force, ?float $deadline, ?array $domainIds)
	{
		$sockets = stream_socket_pair(STREAM_PF_UNIX, STREAM_SOCK_DGRAM, 0);
		$_pipes = [];
		$env = $_SERVER;
		$env['CCM19_NO_SANITY_CHECK'] = '1';
		unset($env['argv']);
		unset($env['argc']);

		$args = [
			$_SERVER['argv'][0],
			'app:cron:worker',
		];
		if ($force) {
			$args[] = '-f';
		}
		if ($deadline !== null) {
			$args[] = '--deadline='.$deadline;
		}
		foreach ($domainIds ?? [] as $domainId) {
			$args[] = '--domain='.$domainId;
		}

		$proc = proc_open(
			$args,
			[
				0 => $sockets[1],
				1 => STDOUT,
				2 => STDERR,
			],
			$_pipes,
			null,
			$env
		);
		fclose($sockets[1]);
		return [$proc, $sockets[0]];
	}

	private function removeWorker(array &$workers, $socket) {
		foreach ($workers as $i => $worker) {
			if ($worker[1] == $socket) {
				fclose($worker[1]);
				proc_close($worker[0]);
				unset($workers[$i]);
				break;
			}
		}
	}

	private function runWithWorkers(int $workerCount, SymfonyStyle $io, float $deadline, bool $force, array $users, ?array $domainIds, array $systemJobs, array $userJobs, array $domainJobs): void
	{
		// Prozesse starten
		$workers = array_map(function ($i) use ($deadline, $force, $domainIds) {
			return $this->spawnWorker($force, $deadline, $domainIds);
		}, range(1, $workerCount));
		unset($domainIds);

		// Kommando-Warteschlange erstellen
		$queue = [];
		if ($systemJobs) {
			$queue[] = 'SYSTEM '.implode(' ', array_map(function ($job) { return $job->getName(); }, $systemJobs));
		}
		$jobString = implode(' ', array_map(function ($job) { return $job->getName(); }, array_merge($userJobs, $domainJobs)));
		unset($systemJobs, $userJobs, $domainJobs);
		foreach ($users as $user) {
			$queue[] = 'USER '.$user->getId().' '.$jobString;
		}
		unset($users);
		$queue = array_reverse($queue);

		// Main loop
		while ($workers) {
			// Timeout aus Deadline berechnen
			$timeout = (int)($deadline - microtime(true));
			// Wenn Timeout überschritten, hier abbrechen
			if ($timeout <= 0) {
				break;
			}
			// Liste der zu überwachenden Sockets bestimmen
			$read = array_map(function ($worker) { return $worker[1]; }, $workers);
			$write = $except = [];
			// Auf Ereignis warten (max. bis zum Timeout)
			$result = stream_select($read, $write, $except, $timeout);

			// Wenn Timeout oder Fehler, zurück an den Anfang der Schleife
			if (!$result) {
				continue;
			}
			// Ansonsten alle Sockets mit Ereignis auslesen,
			// und dann nächsten Auftrag aus Queue verteilen.
			foreach ($read as $socket) {
				$result = fread($socket, 4095);
				// Wenn Worker beendet ist, aus Liste entfernen
				if ($result === 'EXITED' or $result === '') {
					$this->removeWorker($workers, $socket);
				} else {
					// Fehlermeldungen der Worker anzeigen
					if (strpos($result, "EXCEPTION ") === 0) {
						$io->error("Exception: ".substr($result, 10));
					} else if (strpos($result, "ERROR ") === 0) {
						$io->error("Internal Error: ".substr($result, 6));
					}

					// Nächsten Auftrag senden
					if ($queue) {
						$item = array_pop($queue);
						$result = fwrite($socket, $item);
						//  Fehler beim Senden -> Worker beenden und löschen
						if ($result === false) {
							// Item zurück in Queue schieben
							array_push($queue, $item);
							@fwrite($socket, 'EXIT');
							$this->removeWorker($workers, $socket);
						}
					} else {
						// Wenn nichts mehr in der Queue: Worker beenden
						@fwrite($socket, 'EXIT');
					}
				}
			}
		}
		// Am Ende alle Worker beenden, die noch laufen
		foreach ($workers as $worker) {
			@fwrite($worker[1], 'EXIT');
		}
		// und abwarten, bis sie beendet sind.
		foreach ($workers as $worker) {
			fclose($worker[1]);
			proc_close($worker[0]);
		}
	}

	private function runDirectly(SymfonyStyle $io, float $deadline, bool $force, array $users, ?array $domainIds, array $systemJobs, array $userJobs, array $domainJobs, Request $request): void
	{
		$config = Config::getInstance();

		// System-Jobs ausführen
		foreach ($systemJobs as $job) {
			if ($force or $job->isDue()) {
				$this->worker->runSystemJob($job, $request, $config, $io);
				if (microtime(true) >= $deadline) {
					$io->text("Timeout exceeded.");
					return;
				}
			}
		}

		if (!$domainJobs and !$userJobs) {
			return;
		}
		// Für jede Domain
		foreach ($users as $user) {
			if (!$this->cron->setContext($user, null)) { continue; }

			$domains = Domain::all();
			shuffle($domains);
			if ($domainIds) {
				$domains = array_filter($domains, function ($domain) use ($domainIds) { return in_array($domain->getId(), $domainIds); });
			}

			$firstDomain = reset($domains) ?: null;
			$this->cron->setContext($user, ($firstDomain) ? $firstDomain->getId() : null);
			$io->comment(sprintf("User '%s':",
				htmlspecialchars($user->getUsername())
			));
			foreach ($userJobs as $job) {
				if ($force or $job->isDue()) {
					$this->worker->runUserJob($job, $request, $config, $io);
				}
				if (microtime(true) >= $deadline) {
					$io->text("Timeout exceeded.");
					return;
				}
			}

			if (!$domainJobs) {
				continue;
			}
			foreach ($domains as $domain) {

				$io->comment(sprintf("Domain '%s' [%s] of '%s':",
					htmlspecialchars($domain->getName()),
					htmlspecialchars($domain->getId()),
					htmlspecialchars($user->getUsername())
				));
				if (!$this->cron->setContext($user, $domain->getId())) { continue; }

				// Domain-Jobs ausführen
				foreach ($domainJobs as $job) {
					if ($force or $job->isDue()) {
						$this->worker->runDomainJob($job, $request, $config, $io);

						if (microtime(true) >= $deadline) {
							$io->text("Timeout exceeded.");
							return;
						}
					}
				}
			}
		}
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{

		$io = new SymfonyStyle($input, $output);
		$userNames = $input->getOption('user');
		$domainIds = $input->getOption('domain');
		$force = $input->getOption('force');
		$workerCount = (int)$input->getOption('parallel');
		$jobNames = $input->getArgument('job');

		if ($workerCount > 1 and PHP_VERSION_ID < 70400) {
			$workerCount = 1;
			$io->warning("Parallelization only supported for PHP >= 7.4");
		}

		// Check input
		$ok = true;

		$timeout = $input->getOption('timeout');
		if ($timeout == -1.0) {
			$timeout = 60*60.0;
		}
		elseif (!is_numeric($timeout) or ((float)$timeout) <= 0) {
			$io->error("Timeout must be a positive number.");
			$ok = false;
			$timeout = 0.0;
		} else {
			$timeout = (float)$timeout;
		}
		$deadline = microtime(true)+$timeout;

		$users = array_combine($userNames, array_map(function ($username) {
			return User::byId($username) ?? User::find($username);
		}, $userNames));

		foreach ($users as $username => $user) {
			if ($user === null) {
				$io->error("User '".htmlspecialchars($username)."' not found.");
				$ok = false;
			}
		}

		// Abbrechen wenn Validierung fehlschlägt
		if (!$ok) {
			return 1;
		}

		// Benutzerliste fertig bestimmen
		if (!$users) {
			if (Utils::isExtendedEdition()) {
				$users = User::getClients();
			} else {
				$users = User::byRole(User::ROLE_ADMIN);
			}
			shuffle($users);
		}

		// Job-Liste fertig bestimmen
		if ($jobNames) {
			$jobs = array_filter(array_map(function ($jobName) use ($io) {
				$result = $this->cron->jobByName($jobName);
				if (!$result){
					$io->note("Job '".htmlspecialchars($jobName)."' unknown. Skipped.");
				}
				return $result;
			}, $jobNames));
		} else {
			$jobs = $this->cron->jobs();
		}
		shuffle($jobs);
		$systemJobs = array_filter($jobs, function ($job) { return $job instanceof Job; });
		$userJobs = array_filter($jobs, function ($job) { return $job instanceof UserJob; });
		$domainJobs = array_filter($jobs, function ($job) { return $job instanceof DomainJob; });

		if ($workerCount > 1) {
			$success = $this->cron->withCronLock(function () use ($workerCount, $io, $force, $systemJobs, $userJobs, $domainJobs, $users, $domainIds, $deadline) {
				$this->runWithWorkers($workerCount, $io, $deadline, $force, $users, $domainIds, $systemJobs, $userJobs, $domainJobs);
			});
		} else {
			// HTTP-Request-Umgebung vorbereiten
			$request = $this->worker->prepareRequestContext();
			if (!$request) {
				$io->error("Base URL not configured. Aborted.");
				return 1;
			}
			$this->stack->push($request);

			// Zunächst in Admin-User wechseln
			$user = current(User::byRole(User::ROLE_ADMIN));
			$this->cron->setContext($user, null);

			$success = $this->cron->withCronLock(function () use ($request, $io, $force, $systemJobs, $userJobs, $domainJobs, $users, $domainIds, $deadline) {
				$this->runDirectly($io, $deadline, $force, $users, $domainIds, $systemJobs, $userJobs, $domainJobs, $request);
			});

			$this->stack->pop();
		}

		if (!$success) {
			$io->note("Cron job already running. Please try again later.");
			return 2;
		}

		if (microtime(true) >= $deadline) {
			$io->text("Timeout exceeded.");
			return 0;
		}

		$deferredJobs = $this->cron->deferredJobs();
		if ($deferredJobs) {
			$this->cron->withCronLock(function () use ($io, $deferredJobs, $deadline) {
				$request = Request::createFromGlobals();
				$request->setSession($this->stack->getSession());
				$this->stack->push($request);
				try {
					$config = Config::getInstance();

					foreach ($deferredJobs as $jobData) {
						list($job, $apiKey, $domainId) = $jobData;
						$user = User::byApiKey($apiKey);
						if (!$user or !$this->cron->setContext($user, $domainId)) { continue; }
						$io->comment("Continuing deferred job '".htmlspecialchars($job->getName())."' ...");
						try {
							$job->run($request);
						} finally {
							// Konfiguration schreiben, damit sie im nächsten Cronjob direkt verfügbar ist
							$config->flush();
							if (Domain::activeDomain()) {
								CronState::flush();
							}
						}
						$io->success("'".htmlspecialchars($job->getName())."' completed successfully.");

						if (microtime(true) >= $deadline) {
							$io->comment("Timeout exceeded.");
							break;
						}
					}
				}
				finally {
					$this->stack->pop();
				}
			});
		}

		$io->text("Done.");
		return 0;
	}

}
