<?php
namespace App\Command\Debug\Model;

use App\Component\Config\ModelInterface;
use App\Model\User;
use App\Model\Domain;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Helper\TableSeparator;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;

function array_union(array $a, array $b): array
{
	return array_merge(
		array_intersect($a, $b),
		array_diff($a, $b),
		array_diff($b, $a)
	);
}

function encode_value($value): string
{
	if ($value === null) {
		$value = '<comment>null</comment>';
	}
	else {
		$value = json_encode($value);
		if (!is_string($value)) {
			$value = '<comment>???</comment>';
		}
	}
	return $value;
}

class ListCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:debug:model:list';

	protected function configure(): void
	{
		$this
			->setDescription('List model contents.')
			->setDefinition([
				new InputOption('user', 'u', InputOption::VALUE_REQUIRED, 'User-ID in whose context the model should be opened'),
				new InputOption('domain', 'd', InputOption::VALUE_REQUIRED, 'Domain-ID for which the model should be opened'),
				new InputArgument('model', InputArgument::REQUIRED, 'Model name'),
			])
		;
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$io = new SymfonyStyle($input, $output);

		/** @var class-string $model */
		$model = $input->getArgument('model');
		$model = str_replace('/', '\\', $model);
		if ((!class_exists($model) or !is_subclass_of($model, ModelInterface::class)) and class_exists('App\\Model\\'.$model)) {
			$model = 'App\\Model\\'.$model;
		}

		if (!class_exists($model) or !is_subclass_of($model, ModelInterface::class)) {
			$io->error("Unknown model \"$model\"");
			return 1;
		}

		$username = $input->getOption("user");
		if ($username !== null) {
			$user = User::byId($username) ?? User::find($username);
			if ($user === null) {
				$io->error("User '".htmlspecialchars($username)."' not found.");
				return 2;
			}
			User::setApiUser($user->getApiKey());

			$domainId = $input->getOption('domain');
			if ($domainId !== null) {
				if (!Domain::setDomainForRequest($input->getOption('domain'))) {
					$io->error("Domain '".htmlspecialchars($domainId)."' not found.");
					return 2;
				}
			}
		}
		else {
			if ($input->getOption('domain') !== null) {
				$io->error("A user must be specified to select a domain.");
				return 2;
			}
		}

		if (!method_exists($model, '__debugInfo')) {
			$io->warning("Model \"$model\" cannot be listed here.");
			return 2;
		}

		try {
			$modelData = array_map(function ($entry) {
				return $entry->__debugInfo();
			}, $model::all());
		} catch (\LogicException $e) {
			$io->error($e->getMessage());
			return 3;
		}

		$keys = array_reduce($modelData, function (array $carry, array $entry) {
			return array_union($carry, array_keys($entry));
		}, []);

		if (count($keys) > 8) {
			$result = [];
			foreach ($modelData as $model) {
				foreach ($model as $key => $value) {
					$value = encode_value($value);
					if ($key and $key[0] == '.') {
						$key = substr($key, 1);
					}
					if ($key === 'passwordHash') {
						$value = '<comment>(redacted)</comment>';
					}
					$result[] = [$key => $value];
				}
				$result[] = new TableSeparator();
			}
			call_user_func_array([$io, 'definitionList'], $result);
		} else {
			$io->table(
				array_map(function ($key) {
					if ($key and $key[0] == '.') {
						$key = substr($key, 1);
					}
					return $key;
				}, $keys),
				array_map(function ($model) use ($keys) {
					return array_map(function ($key) use ($model) {
						$value = encode_value($model[$key] ?? null);
						if ($key === 'passwordHash') {
							$value = '<comment>(redacted)</comment>';
						}
						return $value;
					}, $keys);
				}, $modelData)
			);
		}

		return 0;
	}
}
