<?php
namespace App\Command;

use App\Component\Cron\Cron;
use App\Component\MongoDbService;
use App\Component\Plugin\Manager as PluginManager;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Utils;
use App\Model\Model;
use App\Model\User;
use App\Model\Domain;
use CallbackFilterIterator;
use FilesystemIterator;
use Iterator;
use MongoDB\BSON\ObjectId;
use RecursiveDirectoryIterator;
use RecursiveIteratorIterator;
use ReflectionClass;
use RuntimeException;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Console\Style\StyleInterface;

class Fsck extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:fsck';

	/** @var PluginManager $pluginManager */
	private $pluginManager;

	/** @var Cron $cron */
	private $cron;

	/** @var StyleInterface $io */
	private $io;

	/** @var int $problemsFound */
	private $problemsFound = 0;

	/** @var int $problemsFound */
	private $problemsFixed = 0;

	const USER_DIR_REGEX = '/^[a-f0-9]{7}(?:[a-f0-9]{17})?$/';
	const DOMAIN_DIR_REGEX = '/^[a-f0-9]{7}(?:[a-f0-9]{17})?$/';

	public function __construct(PluginManager $pluginManager, Cron $cron) {
		parent::__construct();
		$this->pluginManager = $pluginManager;
		$this->cron = $cron;
	}

	protected function configure(): void
	{
		$this
			->setDefinition([
				new InputOption('preen', 'p', InputOption::VALUE_NONE, 'Fix found problems, deleting data if neccessary.'),
				new InputOption('exhaustive', '', InputOption::VALUE_NONE, 'Fully remove files/collections that are no longer needed, or were created by uninstalled/disabled plugins (DANGEROUS if combined with <info>--preen</>)'),
			])
			->setDescription('Checks data consistency and cleans up unneeded/broken data.')
		;
	}

	public function setIo(StyleInterface $io) {
		$this->io = $io;
	}

	public function resetCounters() {
		$this->problemsFound = 0;
		$this->problemsFixed = 0;
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$this->resetCounters();
		$io = $this->io = new SymfonyStyle($input, $output);

		$io->comment("Waiting for cron jobs to finish...");

		// Blockiere alle Cronjobs solange die Checks laufen
		while (!$this->cron->withCronLock(function () use ($input) { $this->executeChecks($input); })) {}

		return ($this->problemsFound - $this->problemsFixed == 0) ? 0 : 1;
	}

	private function executeChecks(InputInterface $input): void
	{
		$io = $this->io;

		$allowChanges = (bool)$input->getOption('preen');
		$exhaustive = $input->getOption('exhaustive');

		$io->comment("Checking models...");
		$this->checkModels($allowChanges, $exhaustive);

		if ($allowChanges) {
			$this->io->comment("Found $this->problemsFound problems, of which $this->problemsFixed were fixed.");
		} else {
			$this->io->comment("Found $this->problemsFound problems.");
		}
	}

	private function checkModels($allowChanges, $exhaustive)
	{
		$problemsFound = 0;
		$problemsFixed = 0;

		// EmbeddingAssets ohne dazu passendes Embedding löschen
		foreach (User::all() as $user) {
			if (!$user->isSubUser()) {
				$userId = $user->getId();
				Model::closeUserModels();
				Domain::switchRepositoryContext($user);
				foreach (Domain::all() as $domain) {
					$domainId = $domain->getId();
					Embedding::switchRepositoryContext($user, $domain);
					EmbeddingAsset::switchRepositoryContext($user, $domain);
					$assetMap = array_map(function (Embedding $embedding) {
						return array_map(function (EmbeddingAsset $asset) { return $asset->getId(); }, $embedding->getAssets());
					}, Embedding::all());
					$wantedAssetIds = $assetMap ? call_user_func_array('array_merge', array_values($assetMap)) : [];

					$assetIds = array_keys(EmbeddingAsset::all());
					$unneededAssetIds = array_diff($assetIds, $wantedAssetIds);

					if ($unneededAssetIds) {
						$count = count($unneededAssetIds);
						$this->problemsFound += $count;
						$this->io->text("User <comment>$userId</> domain <comment>$domainId</>: Found <info>$count</> orphaned EmbeddingAsset(s).");

						if ($allowChanges) {
							foreach ($unneededAssetIds as $assetId) {
								EmbeddingAsset::delete($assetId, false);
								$this->problemsFixed++;
							}
							EmbeddingAsset::flush();
							$this->io->success("Deleted $count orphaned EmbeddingAssets");
						}
					}

					Model::closeDomainModels();
				}
				Model::closeUserModels();
			}
		}

		// Alle Models suchen
		$modelsMeta = [
			'cm-config' => ['global' => ['customIds' => true, 'store' => Utils::getRepositoryStore()]]
		];
		$models = [
			Utils::REPOSITORY_STORE_FILE => ['global' => [], 'user' => [], 'domain' => []],
			Utils::REPOSITORY_STORE_MONGODB => ['global' => [], 'user' => [], 'domain' => []],
			Utils::REPOSITORY_STORE_REDIS => ['global' => [], 'user' => [], 'domain' => []],
		];
		$models[Utils::getRepositoryStore()]['global']['App\Config'] = 'cm-config';

		foreach ($this->listModelClasses() as $class) {
			$repo = $class::getRepositoryName();
			$store = $class::getRepositoryStoreType();
			$meta =  [
				'customIds' => $class::isModelWithCustomIds(),
				'store' => $store,
			];
			if (!isset($modelsMeta[$repo])) {
				$modelsMeta[$repo] = [];
			}
			if ($class::isDomainSpecificModel()) {
				$models[$store]['domain'][$class] = $repo;
				$modelsMeta[$repo]['domain'] = $meta;
			}
			elseif ($class::isUserSpecificModel()) {
				$models[$store]['user'][$class] = $repo;
				$modelsMeta[$repo]['user'] = $meta;
			} else {
				$models[$store]['global'][$class] = $repo;
				$modelsMeta[$repo]['global'] = $meta;
			}
		}

		// Doppelte Repo-Namen suchen
		$globalModels = array_merge($models[Utils::REPOSITORY_STORE_FILE]['global'], $models[Utils::REPOSITORY_STORE_MONGODB]['global']);
		$globalDuplicates = array_diff_key(array_values($globalModels), array_unique(array_values($globalModels)));
		$userModels = array_merge($models[Utils::REPOSITORY_STORE_FILE]['user'], $models[Utils::REPOSITORY_STORE_MONGODB]['user']);
		$userDuplicates = array_diff_key(array_values($userModels), array_unique(array_values($userModels)));
		$domainModels = array_merge($models[Utils::REPOSITORY_STORE_FILE]['domain'], $models[Utils::REPOSITORY_STORE_MONGODB]['domain']);
		$domainDuplicates = array_diff_key(array_values($domainModels), array_unique(array_values($domainModels)));
		foreach ($globalDuplicates as $name) {
			$this->io->caution("Repository name '$name' is shared by multiple non-user-specific models!");
			$this->problemsFound++;
		}
		foreach ($userDuplicates as $name) {
			$this->io->caution("Repository name '$name' is shared by multiple user-specific models!");
			$this->problemsFound++;
		}
		foreach ($domainDuplicates as $name) {
			$this->io->caution("Repository name '$name' is shared by multiple domain-specific models!");
			$this->problemsFound++;
		}

		// == JSON-Datei-Models ==

		// Nach User-Verzeichnissen ohne dazugehörigen User suchen
		$dir = Utils::getVarDir().'/user_data/';
		$userIds = array_map(function ($user) { return $user->getId(); }, User::all());
		foreach (new FilesystemIterator($dir) as $subdir) {
			$subdirName = $subdir->getFilename();
			if (!$subdir->isDir()) {
				$this->io->text("Unexpected file <comment>var/user_data/$subdirName</>");
				$this->problemsFound++;
				if ($allowChanges) {
					$this->io->text("* Please delete <comment>var/user_data/$subdirName</> manually.");
				}
			}
			elseif ($subdir->getFilename() === 'consentarchiv') {
				// ignore for now
			}
			elseif (!isset($userIds[$subdir->getFilename()])) {
				$this->io->text("User <info>$subdirName</> unknown, but directory <comment>var/user_data/$subdirName</> exists.");
				$this->problemsFound++;
				if ($allowChanges and preg_match(self::USER_DIR_REGEX, $subdirName)) {
					Utils::removeDirectory($subdir->getPathname());
					$this->io->success('Deleted '.$subdir->getPathname());
					$this->problemsFixed++;
				}
			}
		}

		// Domain-Verzeichnisse
		foreach (User::all() as $user) {
			$dir = Utils::getVarDir().'/user_data/'.$user->getId().'/domains/';
			if (!file_exists($dir)) {
				continue;
			}
			if ($user->isSubUser()) {
				// Alle Domain-Verzeichnisse in Subusern bemängeln
				foreach (new FilesystemIterator($dir) as $subdir) {
					$domPath = 'var/user_data/'.$user->getId().'/domains/'.$subdir->getFilename();
					if ($subdir->isDir()) {
						$this->io->text("Unexpected domain directory in subuser: <comment>$domPath</>");
						$this->problemsFound++;
						if ($allowChanges and preg_match(self::DOMAIN_DIR_REGEX, $subdir->getFilename())) {
							Utils::removeDirectory($subdir->getPathname());
							$this->io->success('Deleted '.$subdir->getPathname());
							$this->problemsFixed++;
						}
					} else {
						$this->io->text("Unexpected file: <comment>$domPath</>");
						$this->problemsFound++;
						if ($allowChanges) {
							$this->io->text("* Please delete <comment>$domPath</> manually.");
						}
					}
				}
				if ($allowChanges) {
					@rmdir($dir);
				}
			} else {
				Domain::switchRepositoryContext($user);
				$domainIds = array_map(function ($domain) { return $domain->getId(); }, Domain::all());
				Model::closeUserModels();
				foreach (new FilesystemIterator($dir) as $subdir) {
					$domainId = $subdir->getFilename();
					$domPath = 'var/user_data/'.$user->getId().'/domains/'.$subdir->getFilename();
					if ($subdir->isDir()) {
						if (isset($domainIds[$domainId])) {
							$this->io->text("Directory for unknown domain exists: <comment>$domPath</>");
							$this->problemsFound++;
							if ($allowChanges and preg_match(self::DOMAIN_DIR_REGEX, $subdir->getFilename())) {
								Utils::removeDirectory($subdir->getPathname());
								$this->io->success('Deleted '.$subdir->getPathname());
								$this->problemsFixed++;
							}
						}
					} else {
						$this->io->text("Unexpected file: <comment>$domPath</>");
						$this->problemsFound++;
						if ($allowChanges) {
							$this->io->text("* Please delete <comment>$domPath</> manually.");
						}
					}
				}
			}
		}

		// Bei --exhaustive überall nach Repo-Dateien suchen, zu denen es kein Model gibt
		if ($exhaustive) {
			$this->purgeUnlistedJsonFiles(Utils::getVarDir(), $models[Utils::REPOSITORY_STORE_FILE]['global'], 'Repository file without associated model: <comment>var/%s</>', $allowChanges);

			foreach (User::all() as $user) {
				$userPath = Utils::getVarDir().'/user_data/'.$user->getId();
				if (!file_exists($userPath)) continue;
				$this->purgeUnlistedJsonFiles($userPath, $models[Utils::REPOSITORY_STORE_FILE]['user'], 'Repository file without associated model: <comment>var/user_data/'.$user->getId().'/%s</>', $allowChanges);

				if (!$user->isSubUser()) {
					Model::closeUserModels();
					Domain::switchRepositoryContext($user);
					foreach (Domain::all() as $domain) {
						$domainPath = $userPath.'/domains/'.$domain->getId();
						$this->purgeUnlistedJsonFiles($domainPath, $models[Utils::REPOSITORY_STORE_FILE]['domain'], 'Repository file without associated model: <comment>var/user_data/'.$user->getId().'/domains/'.$domain->getId().'/%s</>', $allowChanges);
					}
					Model::closeUserModels();
				}
			}
		}

		if (Utils::getRepositoryStore() === Utils::REPOSITORY_STORE_MONGODB) {
			$this->io->comment("Checking MongoDB...");

			$dbService = MongoDbService::getInstance();
			$client = $dbService->getClient();
			$db = $dbService->getDatabase();

			// Indexes checken
			$allModels = array_merge(
				array_values($models[Utils::REPOSITORY_STORE_MONGODB]['global']),
				array_values($models[Utils::REPOSITORY_STORE_MONGODB]['user']),
				array_values($models[Utils::REPOSITORY_STORE_MONGODB]['domain'])
			);
			foreach ($allModels as $collectionName) {

				// TODO: Liste nicht hier dopppelt führen, sondern aus Model-Metadaten auslesen
				$shouldHaveMtimeIndex = in_array($collectionName, ['cm-locales', 'cm-user-locales', 'cm-cookies', 'cm-legal-texts', 'cm-placeholders', 'cm-embeddings', 'cm-embedding-assets', 'cm-themes']);

				$expected = [];
				$meta = $modelsMeta[$collectionName];
				if (isset($meta['domain'])) {
					if ($shouldHaveMtimeIndex) {
						$expected['_domainId_1__mtime_-1'] = ['key'=>['_domainId'=>1, '_mtime'=>-1], 'unique'=>false];
					} elseif (!$meta['domain']['customIds']) {
						$expected['_domainId_1'] = ['key'=>['_domainId'=>1], 'unique'=>false];
					}
					if ($meta['domain']['customIds']) {
						$expected['_domainId_1__customId_1'] = ['key'=>['_domainId'=>1, '_customId'=>1], 'unique'=>true];
					}
				}
				if (isset($meta['user'])) {
					if ($shouldHaveMtimeIndex) {
						$expected['_userId_1__mtime_-1'] = ['key'=>['_userId'=>1, '_mtime'=>-1], 'unique'=>false];
					} elseif (!$meta['user']['customIds']) {
						$expected['_userId_1'] = ['key'=>['_userId'=>1], 'unique'=>false];
					}
					if ($meta['user']['customIds']) {
						$expected['_userId_1__customId_1'] = ['key'=>['_userId'=>1, '_customId'=>1], 'unique'=>true];
					}
				}
				if (isset($meta['global'])) {
					if ($shouldHaveMtimeIndex) {
						$expected['_mtime_-1'] = ['key'=>['_mtime'=>-1], 'unique'=>false];
					}
					if ($meta['global']['customIds']) {
						$expected['_customId_hashed'] = ['key'=>['_customId'=>'hashed'], 'unique'=>false];
					}
				}
				if ($collectionName == 'cm-users') {
					$expected['apiKey_hashed'] = ['key'=>['apiKey'=>'hashed'], 'unique'=>false];
				}

				$this->checkMongoDbIndexes($db, $collectionName, $expected, $allowChanges);
			}

			if (Utils::getProtocolRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB) {
				$this->checkMongoDbIndexes($db, 'cm-consent-protocol', [ ['key'=> [ '_domainId' => 1, '_id' => -1], 'unique'=>false] ], $allowChanges);
			}
			if (Utils::getArchiveRepositoryStore() === Utils::REPOSITORY_STORE_MONGODB) {
				// Archiv aktuell noch nicht behandeln
				//$this->checkMongoDbIndexes($db, 'cm-consent-protocol.archive', [ ['key'=> [ '_domainId' => 1, '_id' => -1], 'unique'=>false] ], $allowChanges);
			}
			if (Utils::getLoginAttemptsLogRepositoryStore() === Utils::REPOSITORY_STORE_MONGODB) {
				$this->checkMongoDbIndexes($db, 'cm-login-attempts', [ ['key'=> [ 'kind' => 1, 'ip' => 1], 'unique'=>false] ], $allowChanges);
			}
			$this->checkMongoDbIndexes($db, 'cm-system-log', [], $allowChanges);
			$this->checkMongoDbIndexes($db, 'cm-media', [
				['key'=> [ '_domainId' => 1, 'kind' => 1, 'name' => 1], 'unique'=>false],
				['key'=> [ '_userId' => 1, 'kind' => 1, 'name' => 1], 'unique'=>false],
				['key'=> [ 'kind' => 1, 'name' => 1], 'unique'=>false],
			], $allowChanges);

			// Globale Dokumente in rein User-/Domain-spezifischen Model-Collections löschen
			$nonGlobalModels = array_diff(array_merge(array_values($userModels), array_values($domainModels)), array_values($globalModels));
			$nonGlobalModels[] = 'cm-consent-protocol';
			foreach ($nonGlobalModels as $collectionName) {
				$collection = $db->selectCollection($collectionName);
				$filter = ['$and' => [
					['_domainId' => ['$exists' => false]],
					['_userId' => ['$exists' => false]],
				]];
				$this->purgeMatchingMongoDbDocuments($collection, $filter, "Non-global model <comment>%s</> contains document without user- or domain-id: <comment>%s</>", $allowChanges);
			}

			// User-spezifische Dokumente in rein globalen/Domain-spezifischen Model-Collections löschen
			$nonUserModels = array_diff(array_merge(array_values($globalModels), array_values($domainModels)), array_values($userModels));
			$nonUserModels[] = 'cm-system-log';
			$nonUserModels[] = 'cm-login-attempts';
			$nonUserModels[] = 'cm-translations';
			$nonUserModels[] = 'cm-consent-protocol';
			foreach ($nonUserModels as $collectionName) {
				$collection = $db->selectCollection($collectionName);
				$filter = ['$and' => [
					['_domainId' => ['$exists' => false]],
					['_userId' => ['$exists' => true]],
				]];
				$this->purgeMatchingMongoDbDocuments($collection, $filter, "Non-user model <comment>%s</> contains document with user-id: <comment>%s</>", $allowChanges);
			}

			// Domain-spezifische Dokumente in rein globalen/User-spezifischen Model-Collections löschen
			$nonDomainModels = array_diff(array_merge(array_values($globalModels), array_values($userModels)), array_values($domainModels));
			$nonDomainModels[] = 'cm-system-log';
			$nonDomainModels[] = 'cm-login-attempts';
			$nonDomainModels[] = 'cm-translations';
			foreach ($nonDomainModels as $collectionName) {
				$collection = $db->selectCollection($collectionName);
				$filter = ['$and' => [
					['_domainId' => ['$exists' => true]],
				]];
				$this->purgeMatchingMongoDbDocuments($collection, $filter, "Non-domain model <comment>%s</> contains document with domain-id: <comment>%s</>", $allowChanges);
			}

			// Models nach nicht mehr existenten User-Ids durchsuchen
			$userIds = array_map(function ($user) { return $user->getId(); }, User::all());
			$userModels2 = $userModels;
			$userModels2['App\\Component\\Media\\MediaFileService'] = 'cm-media';
			foreach ($userModels2 as $collectionName) {
				$collection = $db->selectCollection($collectionName);
				$filter = ['$and' => [
					['_userId' => ['$exists' => true]],
					['_userId' => ['$not' => ['$in' => array_map(function ($id) { return new ObjectId($id); }, array_values($userIds))]]],
				]];
				$foundItems = [];
				foreach ($collection->find($filter, ['projection' => ['_userId'=>true]]) as $document) {
					if (!isset($foundItems[(string)$document['_userId']])) {
						$foundItems[(string)$document['_userId']] = 1;
					} else {
						$foundItems[(string)$document['_userId']]++;
					}
				}
				foreach ($foundItems as $userId => $count) {
					$this->io->text(sprintf('In collection <comment>%1$s</>: found <info>%3$s</> document(s) for non-existant user <comment>%2$s</>', $collectionName, $userId, $count));
					$this->problemsFound += $count;
					if ($allowChanges) {
						$result = $collection->deleteMany(['_userId' => new ObjectId($userId)]);
						$delCount = $result->getDeletedCount();
						$this->io->success("Deleted $delCount documents from ".$collection->getCollectionName());
						$this->problemsFixed += min($delCount, $count);
					}
				}
			}

			// Models nach nicht mehr existenten Domain-Ids durchsuchen
			$domainIds = $this->listAllMongoDbDomainIds($db);
			$domainModels2 = $domainModels;
			$domainModels2['App\\Component\\Media\\MediaFileService'] = 'cm-media';
			$domainModels2['App\\Model\\Protocol'] = 'cm-consent-protocol';
			$domainModels2['App\\Model\\ProtocolArchive'] = 'cm-consent-protocol.archive';
			foreach ($domainModels2 as $collectionName) {
				$collection = $db->selectCollection($collectionName);
				$filter = ['$and' => [
					['_domainId' => ['$exists' => true]],
					['_domainId' => ['$not' => ['$in' => $domainIds]]],
				]];
				$foundItems = [];
				foreach ($collection->find($filter, ['projection' => ['_domainId'=>true]]) as $document) {
					if (!isset($foundItems[(string)$document['_domainId']])) {
						$foundItems[(string)$document['_domainId']] = 1;
					} else {
						$foundItems[(string)$document['_domainId']]++;
					}
				}
				foreach ($foundItems as $domainId => $count) {
					$this->io->text(sprintf('In collection <comment>%1$s</>: found <info>%3$s</> document(s) for non-existant domain <comment>%2$s</>', $collectionName, $domainId, $count));
					$this->problemsFound += $count;
					if ($allowChanges) {
						$result = $collection->deleteMany(['_domainId' => new ObjectId($domainId)]);
						$delCount = $result->getDeletedCount();
						$this->io->success("Deleted $delCount documents from ".$collection->getCollectionName());
						$this->problemsFixed += min($delCount, $count);
					}
				}
			}

			// Bei --exhaustive Collections suchen, zu denen es kein Model gibt
			if ($exhaustive) {
				$allModels = array_merge(
					array_values($models[Utils::REPOSITORY_STORE_MONGODB]['global']),
					array_values($models[Utils::REPOSITORY_STORE_MONGODB]['user']),
					array_values($models[Utils::REPOSITORY_STORE_MONGODB]['domain'])
				);
				$allModels[] = 'cm-media';
				$allModels[] = 'cm-translations';
				$allModels[] = 'cm-system-log';
				if (Utils::getProtocolRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB) {
					$allModels[] = 'cm-consent-protocol';
				}
				if (Utils::getArchiveRepositoryStore() === Utils::REPOSITORY_STORE_MONGODB) {
					$allModels[] = 'cm-consent-protocol.archive';
				}
				if (Utils::getLoginAttemptsLogRepositoryStore() === Utils::REPOSITORY_STORE_MONGODB) {
					$allModels[] = 'cm-login-attempts';
				}
				$collections = iterator_to_array($db->listCollections());
				foreach ($collections as $collection) {
					if (!in_array($collection->getName(), $allModels)) {
						$this->io->text('Database collection without associated model: <comment>'.$collection->getName().'</>');
						$this->problemsFound++;
						if ($allowChanges) {
							$db->dropCollection($collection->getName());
							$this->io->success('Dropped collection '.$collection->getName());
							$this->problemsFixed++;
						}
					}
				}
			}
		}

		// Redis: Einträge für nicht-mehr existierende Nutzer/Domains löschen
		if (Utils::getProtocolRepositoryStore() == Utils::REPOSITORY_STORE_REDIS) {
			$redis = $GLOBALS['kernel']->getContainer()->get('Redis');
			$prefix = 'cm-consent-protocol:';
			$keys = $redis->keys($prefix.'*');

			$users = array_filter(User::all(), function ($u) { return $u->isAccountOwner(); } );
			$users = array_combine(array_map(function ($u) { return $u->getApiKey(); }, $users), $users);
			$domains = [];
			foreach ($users as $user) {
				Domain::switchRepositoryContext($user);
				$domains[$user->getApiKey()] = array_map(function ($domain) { return $domain->getId(); }, Domain::all());
				Model::closeUserModels();
			}

			foreach ($keys as $key) {
				list($prefix_, $apiKey, $domainId) = explode(':', $key, 3);
				if (!isset($users[$apiKey])) {
					$this->io->text('Redis consent log for non-existant user: <comment>'.$key.'</>');
					$this->problemsFound++;
					if ($allowChanges) {
						if ($redis->del($key) > 0) {
							$this->io->success("Deleted $key from Redis");
							$this->problemsFixed++;
						}
					}
				} elseif (!isset($domains[$apiKey][$domainId])) {
					$this->io->text('Redis consent log for non-existant domain of user <info>'.$users[$apiKey]->getId().'</>: <comment>'.$key.'</>');
					$this->problemsFound++;
					if ($allowChanges) {
						if ($redis->del($key) > 0) {
							$this->io->success("Deleted $key from Redis");
							$this->problemsFixed++;
						}
					}
				}
			}
		}

		// Consent-Archiv in getrenntem Verzeichnisbaum
		$archiveRoot = Utils::getCustomArchiveRoot();
		if ($archiveRoot and Utils::getArchiveRepositoryStore() == Utils::REPOSITORY_STORE_FILE) {
			$users = array_filter(User::all(), function ($u) { return $u->isAccountOwner(); } );
			if (!file_exists($archiveRoot)) {
				$this->io->text("Consent archive root directoy missing: <comment>$archiveRoot</>");
			} else {
				foreach (new FilesystemIterator($archiveRoot) as $dir) {
					$dirName = $dir->getFilename();
					if (!$dir->isDir()) {
						$this->io->text("Unexpected file <comment>$archiveRoot/$dirName</>");
						$this->problemsFound++;
						if ($allowChanges) {
							$this->io->text("* Please delete <comment>$archiveRoot/$dirName</> manually.");
						}
					}
					elseif (!isset($users[$dirName])) {
						$this->io->text("User <info>$dirName</> unknown, but directory <comment>$archiveRoot/$dirName</> exists.");
						$this->problemsFound++;
						if ($allowChanges and preg_match(self::USER_DIR_REGEX, $dirName)) {
							Utils::removeDirectory($dir->getPathname());
							$this->io->success('Deleted '.$dir->getPathname());
							$this->problemsFixed++;
						}
					}
					else {
						$user = $users[$dirName];
						Domain::switchRepositoryContext($user);
						$domains = Domain::all();
						$domainsDir = $dir->getPathname().'/domains';

						if (file_exists($domainsDir)) {
							foreach (new FilesystemIterator($domainsDir) as $subdir) {
								$subdirName = $subdir->getFilename();
								if (!$subdir->isDir()) {
									$this->io->text("Unexpected file <comment>$domainsDir/$subdirName</>");
									$this->problemsFound++;
									if ($allowChanges) {
										$this->io->text("* Please delete <comment>$domainsDir/$subdirName</> manually.");
									}
								}
								elseif (!isset($domains[$subdirName])) {
									$this->io->text("Domain <info>$subdirName</> of user <info>$dirName</info> unknown, but directory <comment>$domainsDir/$subdirName</> exists.");
									$this->problemsFound++;
									if ($allowChanges and preg_match(self::DOMAIN_DIR_REGEX, $subdirName)) {
										Utils::removeDirectory($subdir->getPathname());
										$this->io->success('Deleted '.$subdir->getPathname());
										$this->problemsFixed++;
									}
								}
							}
						}
						Model::closeUserModels();
					}
				}
			}
		}
	}

	private function checkMongoDbIndexes($db, string $collectionName, array $expected, bool $allowChanges): void
	{
		$collection = $db->selectCollection($collectionName);
		$existingIndexes = iterator_to_array($collection->listIndexes());
		$superfluousIndexes = [];
		foreach ($existingIndexes as $index) {
			$found = false;
			foreach ($expected as $key => $wantedIndex) {
				if ($index->getKey() === $wantedIndex['key'] and ($wantedIndex['unique'] === null or $index->isUnique() === $wantedIndex['unique'])) {
					if (empty($expected[$key]['found'])) {
						$expected[$key]['found'] = $found = true;
						break;
					}
				}
			}
			if (!$found and $index->getKey() !== ['_id'=>1]) {
				$superfluousIndexes[] = $index;
			}
		}

		foreach ($superfluousIndexes as $index) {
			$this->io->text(sprintf(
				"For collection <comment>%s</>: Superfluous %sindex <info>%s</>",
				$collectionName,
				$index->isUnique() ? 'unique ' : 'non-unique ',
				json_encode($index->getKey())
			));
			$this->problemsFound++;
			if ($allowChanges) {
				$collection->dropIndex($index->getName());
				$this->problemsFixed++;
				$this->io->success("Dropped index");
			}
		}
		$create = [];
		foreach ($expected as $name => $index) {
			if (empty($index['found'])) {
				$this->io->text(sprintf(
					"For collection <comment>%s</>: Missing %sindex <info>%s</>",
					$collectionName,
					$index['unique'] ? 'unique ' : '',
					json_encode($index['key'])
				));
				$this->problemsFound++;
				$create[] = ['key' => $index['key'], 'unique' => $index['unique']];
			}
		}
		if ($create and $allowChanges) {
			$collection->createIndexes($create);
			$this->io->success("Created ".count($create)." indexes");
			$this->problemsFixed += count($create);
		}
	}

	private function listAllMongoDbDomainIds($db): array
	{
		$collection = $db->selectCollection('cm-domains');
		$filter = [ '_userId' => ['$exists' => true] ];
		$result = [];
		foreach ($collection->find($filter, ['projection' => ['_id'=>true]]) as $document) {
			$result[] = $document['_id'];
		}
		return $result;
	}

	private function purgeMatchingMongoDbDocuments($collection, array $filter, string $msgTemplate, bool $reallyDelete): void
	{
		$collectionName = $collection->getCollectionName();
		$result = $collection->find($filter, ['projection' => ['_id'=>true]]);
		foreach ($result as $document) {
			$documentId = $document['_id']->__toString();
			$this->io->text(sprintf($msgTemplate, $collectionName, $documentId));
			$this->problemsFound++;
			if ($reallyDelete) {
				$deleteFilter = $filter;
				$deleteFilter['$and'][] = ['_id' => $document['_id']];
				$collection->deleteOne($deleteFilter);
				$this->problemsFixed++;
				$this->io->success("Deleted document $documentId in collection $collectionName");
			}
		}
	}

	/**
	 * @param string $dir Verzeichnis-Pfad
	 * @param string[] $basenameList Liste von Namen (ohne Erweiterung) die erhalten bleiben sollen
	 * @param string $msgTemplate Text für die Ausgabe
	 * @param bool $reallyDelete Ob auch gelöscht werden soll
	 */
	private function purgeUnlistedJsonFiles(string $dir, array $basenameList, string $msgTemplate, bool $reallyDelete): void
	{
		if (!file_exists($dir)) {
			return;
		}
		foreach (new FilesystemIterator($dir) as $file) {
			if ($file->isFile() and $file->getExtension() == 'json') {
				$repoName = $file->getBasename('.json');
				if (!in_array($repoName, $basenameList)) {
					$this->io->text(sprintf($msgTemplate, $file->getFilename()));
					$this->problemsFound++;
					if ($reallyDelete) {
						if (unlink($file->getPathname())) {
							$this->io->success('Deleted '.$file->getPathname());
							$this->problemsFixed++;
						} else {
							$this->io->error('Delete failed: '.$file->getPathname());
						}
					}
				}
			}
		}
	}

	/**
	 * @return Iterator<\SplFileInfo>
	 */
	private function globFilesRecursive($basePath, $pattern): Iterator
	{
		$dirInodes = [];
		return new CallbackFilterIterator(
			new RecursiveIteratorIterator(
				new RecursiveDirectoryIterator($basePath, FilesystemIterator::KEY_AS_PATHNAME | FilesystemIterator::CURRENT_AS_FILEINFO | FilesystemIterator::SKIP_DOTS | FilesystemIterator::FOLLOW_SYMLINKS)
			),
			function ($item) use (&$dirInodes, $pattern) {
				$dirInfo = $item->getPathInfo();
				$dirInode = $dirInfo->getInode();
				if ($dirInode) {
					if (isset($dirInodes[$dirInode])) {
						if ($dirInodes[$dirInode][0] !== $dirInfo->getPathname() and $dirInodes[$dirInode][1] === disk_total_space($dirInfo->getPathname())) {
							throw new RuntimeException("filesystem loop detected at ".$dirInfo->getPathname());
						}
					} else {
						$dirInodes[$dirInode] = [$dirInfo->getPathname(), disk_total_space($dirInfo->getPathname())];
					}
				}
				if (fnmatch($pattern, $item->getFilename(), FNM_PATHNAME | FNM_PERIOD | FNM_CASEFOLD)) {
					return $item;
				}
			}
		);
	}

	/**
	 * @return Iterator<class-string>
	 */
	private function findPsr4Classes($basePath, $baseNamespace): Iterator
	{
		$basePath = rtrim($basePath, '/'.DIRECTORY_SEPARATOR).DIRECTORY_SEPARATOR;
		$baseNamespace = rtrim($baseNamespace, '\\').'\\';
		foreach ($this->globFilesRecursive($basePath, '*.php') as $path => $item) {
			yield $path => $baseNamespace.strtr(substr($path, strlen($basePath), -4), [ DIRECTORY_SEPARATOR => '\\', '/' => '\\' ]);
		}
	}

	/**
	 * @return class-string[]
	 */
	private function listModelClasses(): array
	{
		$result = [];
		foreach ($this->findPsr4Classes(Utils::getBaseDir().'/src/Model', 'App\\Model') as $className) {
			$cls = new ReflectionClass($className);
			if (!$cls->isAbstract() and !$cls->isInterface() and !$cls->isTrait() and $cls->isSubclassOf('\App\Model\Model') and $cls->hasProperty('repositoryFilename')) {
				$result[] = $className;
			}
		}

		foreach ($this->pluginManager->installedPlugins() as $plugin) {
			if ($plugin->getDirPath() and $plugin->isValidAndActive()) {
				foreach ($this->findPsr4Classes($plugin->getDirPath().'/src', 'Plugins\\'.$plugin->getDirname()) as $className) {
					$cls = new ReflectionClass($className);
					if (!$cls->isAbstract() and !$cls->isInterface() and !$cls->isTrait() and $cls->isSubclassOf('\App\Model\Model') and $cls->hasProperty('repositoryFilename')) {
						$result[] = $className;
					}
				}
			}
		}
		return $result;
	}
}
