<?php
namespace App\Command\Statistics;

use App\Model\User;
use App\Model\Domain;
use App\Model\Protocol;
use App\Model\ProtocolArchive;
use App\Model\Statistic;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\StyleInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use DateTime;
use DateTimeInterface;
use App\Utils;

class RebuildCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:statistics:rebuild';

	public function __construct()
	{
		parent::__construct();
	}

	protected function configure(): void
	{
		$this
			->setDefinition([
				new InputOption('domain', 'd', InputOption::VALUE_REQUIRED | InputOption::VALUE_IS_ARRAY, 'Domain(s) for which statistics should be rebuilt.'),
				new InputOption('user', 'u', InputOption::VALUE_REQUIRED | InputOption::VALUE_IS_ARRAY, 'Username(s) whose statistics should be rebuilt.'),
				new InputOption('since', 's', InputOption::VALUE_REQUIRED, 'Date to start at'),
				new InputOption('to', 't', InputOption::VALUE_REQUIRED, 'End date (exclusive)'),
			])
			->setDescription('Rebuild basic statistics from consent protocols.')
			->setHelp('Rebuild basic statistics from consent protocols.')
		;
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$usernames = (array)$input->getOption('user');
		$domainnames = (array)$input->getOption('domain');
		if ($input->getOption('since') === null) {
			$startDate = new DateTime('yesterday midnight');
			$startDate->modify('first day of this month');
		} else {
			$startDate = new DateTime($input->getOption('since'));
		}
		$endDate = new DateTime($input->getOption('to') ?? 'today');

		$io = new SymfonyStyle($input, $output);

		if ($startDate > $endDate) {
			$io->error("Invalid date range supplied");
			return 1;
		}

		if (Utils::isExtendedEdition() and !$usernames and !$domainnames) {
			$io->error("You must specify domains (-d) or users (-u) in agency editions.");
			return 1;
		}

		// Get users for usernames
		if ($usernames) {
			$users = array_map(function ($name) { return User::find($name); }, $usernames);
			$error = false;
			foreach ($users as $i => $user) {
				if ($user === null) {
					$io->error("User '".$usernames[$i]."' doesn't exist.");
					$error = true;
				}
			}
			if ($error) {
				return 1;
			}
		}
		// If no usernames are specified, get all users that have the given domains
		else {
			$users = [];
			foreach (User::all() as $user) {
				if ($user->isSubUser()) {
					continue;
				}
				foreach ($domainnames as $domain) {
					// Strip URL path to solely compare the domain name
					$userDomainList = array_map(function (string $domain) {
						return explode('/', $domain)[0];
					}, $user->getDomList());

					if (in_array($domain, $userDomainList)) {
						$users[$user->getId()] = $user;
						break;
					}
				}
			}
		}

		// If no domainnames are specified, get all domains for the given users
		if (!$domainnames) {
			foreach ($users as $user) {
				$domainnames = array_merge($domainnames, $user->getDomList());
			}
		}

		// Process the domains
		$usedDomainNames = [];
		foreach ($users as $user) {
			$io->comment("User '".$user->getUsername()."'");
			$user->switchTo();
			Domain::reloadRepository();

			foreach ($domainnames as $domainName) {
				$domain = Domain::byName($domainName);
				Domain::select($domain->getId());
				Statistic::reloadRepository();
				if ($domain) {
					$usedDomainNames[$domainName] = $domainName;
					$io->comment("Processing domain '".$domainName."'");

					$this->rebuildForOneDomain($domain, $startDate, $endDate, $io);
				}
			}
		}

		return 0;
	}

	private function saveStatistics(string $date, array $sums): void {
		// Statistiken speichern
		Statistic::beginWrite();
		if (!$sums and !Statistic::exists($date)) {
			return;
		}

		Statistic::delete($date, false);
		$stat = Statistic::findOrCreate($date);

		foreach ($sums['purposes'] ?? [] as $purposeId => $count) {
			$stat->setPurposeAcceptCount($purposeId, $count);
		}
		foreach ($sums['embeddings'] ?? [] as $embeddingId => $count) {
			$stat->setEmbeddingAcceptCount($embeddingId, $count);
		}
	}

	private function commitStatisticsRange(array &$sumsPerDay, bool $finish = true): void
	{
		if ($finish) {
			foreach ($sumsPerDay as $date => $sums) {
				$this->saveStatistics($date, $sums);
			}
			$sumsPerDay = [];
		} else {
			$firstDay = new DateTime(min(array_keys($sumsPerDay)));
			$lastDay = new DateTime(max(array_keys($sumsPerDay)));
			$preLastDay = (clone $lastDay)->modify('-1 day');
			if (isset($sumsPerDay[$preLastDay->format('Y-m-d')])) {
				$lastDay = $preLastDay;
			}
			$currentDay = $firstDay;
			while ($currentDay < $lastDay) {
				$dateStr = $currentDay->format('Y-m-d');
				$this->saveStatistics($dateStr, $sumsPerDay[$dateStr] ?? []);
				$currentDay->modify('+1 day');
				unset($sumsPerDay[$dateStr]);
			}
		}
		Statistic::flush();
	}

	/**
	 * @param array<string,mixed> $item
	 * @param array<string,array<string,array<string,int>>> $sumsPerDay
	 * @param array<string,array<string,mixed>> $byUcid
	 */
	private function processOneItem(array $item, array &$sumsPerDay, array &$byUcid): void
	{
		$date = date("Y-m-d", $item['timestamp'] ?? 0);
		$item['date'] = $date;
		$ucid = $item['ucid'];
		unset($item['ucid']);

		// Bei Tageswechsel UCID-Lookup-Liste leeren und Summen vorbereiten
		if (!isset($sumsPerDay[$date])) {
			$byUcid = [];
			$sumsPerDay[$date] = [
				'purposes' => [],
				'embeddings' => [],
			];
		}

		if (isset($byUcid[$ucid])) {
			$olditem = $byUcid[$ucid];
			$item['first_date'] = $olditem['first_date'] ?? $olditem['date'];
		} else {
			$olditem = null;
		}

		// Purposes zählen
		if (isset($item['purposes']) && is_array($item['purposes'])) {
			if ($olditem && isset($olditem['purposes'])) {
				foreach ($olditem['purposes'] as $purposeId) {
					if (isset($sumsPerDay[$date]['purposes'][$purposeId])) {
						$sumsPerDay[$date]['purposes'][$purposeId] = 0;
					}
					// @phan-suppress-next-line PhanTypeArraySuspiciousNullable
					$sumsPerDay[$date]['purposes'][$purposeId]--;
				}
			}
			foreach ($item['purposes'] as $purposeId) {
				if (!isset($sumsPerDay[$date]['purposes'][$purposeId])) {
					$sumsPerDay[$date]['purposes'][(string)$purposeId] = 0;
				}
				// @phan-suppress-next-line PhanTypeArraySuspiciousNullable
				$sumsPerDay[$date]['purposes'][(string)$purposeId]++;
			}
		}

		// Embeddings zählen
		if (isset($item['embeddings']) && is_array($item['embeddings'])) {
			if ($olditem && isset($olditem['embeddings'])) {
				foreach ($olditem['embeddings'] as $embeddingId) {
					if (!isset($sumsPerDay[$date]['embeddings'][$embeddingId])) {
						$sumsPerDay[$date]['embeddings'][$embeddingId] = 0;
					}
					// @phan-suppress-next-line PhanTypeArraySuspiciousNullable
					$sumsPerDay[$date]['embeddings'][$embeddingId]--;
				}
			}
			foreach ($item['embeddings'] as $embeddingId) {
				if (!isset($sumsPerDay[$date]['embeddings'][$embeddingId])) {
					$sumsPerDay[$date]['embeddings'][(string)$embeddingId] = 0;
				}
				// @phan-suppress-next-line PhanTypeArraySuspiciousNullable
				$sumsPerDay[$date]['embeddings'][(string)$embeddingId]++;
			}
		}
		$byUcid[$ucid] = $item;

	}

	private function rebuildForOneDomain(Domain $domain, DateTimeInterface $startDate, DateTimeInterface $endDate, StyleInterface $io): void
	{
		$archive = ProtocolArchive::byDomain($domain);
		$protocol = Protocol::byDomain($domain);
		$endTimestamp = $endDate->getTimestamp();

		$currentDate = null;
		// Nach Unique-ID gruppieren; immer neuesten Eintrag nehmen
		/** @var array<string,array> $byUcid */
		$byUcid = [];
		/** @var array<string,array<string,int>> $sums */
		$sumsPerDay = [];

		foreach ($archive->filter($startDate, $endDate) as $item) {
			if (count($sumsPerDay) > 2) {
				$this->commitStatisticsRange($sumsPerDay, false);
			}

			$this->processOneItem($item, $sumsPerDay, $byUcid);

		}
		foreach ($protocol->allFromDate($startDate->getTimestamp()) as $item) {
			if ($item['timestamp'] >= $endTimestamp) {
				break;
			}

			if (count($sumsPerDay) > 2) {
				$this->commitStatisticsRange($sumsPerDay, false);
			}

			$this->processOneItem($item, $sumsPerDay, $byUcid);

		}
		$this->commitStatisticsRange($sumsPerDay, true);
		$io->success("Rebuilt statistics for domain '".$domain->getName()."'");
	}
}
