<?php
namespace App\Command\Translate;

use App\Component\Update;
use App\Model\Language;
use App\Model\TransLate as TranslateModel;
use App\Utils;
use RuntimeException;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Yaml\Yaml;

class TranslateCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:trans:translatenow';

	/** @var Update $updater */
	private $updater;

	public function __construct(Update $updater)
	{
		$this->updater = $updater;
		parent::__construct();
	}

	protected function configure(): void
	{
		$this
			->setDescription('Translates the messages files all at once.')
		;
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$io = new SymfonyStyle($input, $output);
		$transCount =0;

		$allLangs = $this->getAvailableLangs();
		$activeLangs = Language::activeLanguages();

		$deeplApiKey = TranslateModel::getDeeplApiKey();
		$googleApiKey = TranslateModel::getGoogleApiKey();

		if(empty($deeplApiKey) && empty($googleApiKey)) {
			$io->error("Please check your Google and Deepl Api Keys!");
			return 0;
		}

		// Fehlende Informationen in /share/skeleton/clickskeks nachtragen
		$io->text('Complementing Clickskeks skeleton files');
		TranslateModel::complementClickskeksSkeleton();

		foreach($activeLangs as $lang)
		{
			if($allLangs[$lang->getShortCode()]<100)
			{
				$io->text("Translating ".$lang->getShortCode());
				$transCount++;

				try {
					//Cm-user-locale übersetzen
					TranslateModel::translateJsonsAndCo($deeplApiKey, $lang->getCode());

					//PLugins übersetzen
					TranslateModel::translateJPlugins($deeplApiKey, $lang->getCode());

					//Dann Sprachdateien übersetzen
					TranslateModel::translateNow($deeplApiKey, $lang->getCode());
				} catch (RuntimeException $ex) {
					$io->error("[ERROR] ({$ex->getCode()}) {$ex->getMessage()}");
					$io->error($ex->getTraceAsString());
					if ($io->confirm('Continue anyways?', false) == false) {
						return 0;
					}
				}
			}
			else{
				$io->text("100% translated: ".$lang->getShortCode());
			}
		}

		$io->success('Translation complete!!');

		if($transCount>0)
		{
			$allLangs = $this->getAvailableLangs();

			foreach($activeLangs as $lang)
			{
				if($allLangs[$lang->getShortCode()]<100)
				{
					$io->text("Please repeat translation for ".$lang->getShortCode());
				}
				else{
					$io->text("100% translated: ".$lang->getShortCode());
				}
			}
		}

		return 0;
	}

	/**
	 * @return array<string, float>
	 */
	private function getAvailableLangs(): array
	{
		$activeLangs = Language::activeLanguages();
		$countLang = [];
		$filesystem = new Filesystem();

		//base = DE
		$filename = Utils::getBaseDir()."/translations/"."messages.de.yaml";
		$yamlBase = Yaml::parseFile($filename);
		$baseCount = count($yamlBase);
		$baseDir = Utils::getBaseDir();

		// Für alle Sprachen Abdeckung berechnen
		foreach ($activeLangs as $language)
		{
			$langKey = $language->getShortCode();
			$langKeyfile = $langKey;

			$filename = $baseDir."/translations/"."messages.".$langKeyfile.".yaml";
			if($filesystem->exists($filename)) {
				$yamlTarget = Yaml::parseFile($filename);
			} else {
				$filesystem->touch($filename);
				$yamlTarget = Yaml::parseFile($filename);
			}

			/** @var string[] $yamlTarget */
			$yamlTarget = array_map('strval', is_array($yamlTarget) ? $yamlTarget : []);

			$diffCount = count(array_diff_key($yamlBase, (array)$yamlTarget));

			// Consider pending messages, which are indicated by two leading underscores
			$diffCount += count(array_filter(
				array_intersect_key($yamlTarget, $yamlBase),
				[TranslateModel::class, 'isTranslationMessagePending']
			));

			//$countLang[$langKey] = round((count($value)/$baseCount*100), 2);
			$countLang[$langKey] = round((100-(($diffCount*100.0)/$baseCount)), 4);
		}
		return $countLang;
	}
}
