<?php
namespace App\Command\Update;

use App\Component\Update;
use App\Exception\DownloaderException;
use App\Exception\TempFileNotWritableException;
use App\Exception\LicenseInvalidException;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;

class CheckCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:update:check';

	/** @var Update $updater */
	private $updater;

	public function __construct(Update $updater)
	{
		$this->updater = $updater;
		parent::__construct();
	}

	protected function configure(): void
	{
		$this
			->setDescription('Check for updates.')
			->setHelp('Checks if an update for CCM19 is available.')
		;
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$io = new SymfonyStyle($input, $output);

		// Prüfe Verzeichnisse rekursiv auf Schreibzugriff
		$notWriteableDirectories = Update::checkDirectoryAccess();
		if ($notWriteableDirectories) {
			$io->warning(array_merge(["Please check if you have write access to the following directories:"], array_slice($notWriteableDirectories, 0, 20)));
		}

		try {
			$result = $this->updater->checkForUpdate();
		} catch (DownloaderException $e) {
			$io->error("Download failed: ".$e->getMessage());
			return 1;
		} catch (TempFileNotWritableException $e) {
			$io->error("Writing temporary file failed: ".$e->getMessage());
			return 1;
		}
		catch (LicenseInvalidException $e) {
			$io->error("Invalid license");
			return 1;
		}

		if ($result === null) {
			$io->text('Your version is up to date. No update is necessary.');
			return 3;
		} elseif (($result['fileSize'] ?? -1) === 0 and ($result['nextGenAvailable'] ?? 0)) {
			$io->text([
				'Your version is up to date. Update only available in next generation.'
			]);
			return 0;
		} else {
			$io->text([
				'An update is available:',
				"ID:\t<info>".($result['versionId'] ?? '?').'</info>',
				"Size:\t<info>".($result['fileSize'] ?? '?').'</info> bytes',
				"Date:\t<info>".($result['lastModified']->format('Y-m-d H:i') ?? '?').'</info>',
			]);
			return 0;
		}
	}
}
