<?php
namespace App\Command\Update;

use App\Config;
use App\Component\Update;
use App\Exception\DownloaderException;
use App\Exception\TempFileNotWritableException;
use App\Exception\LicenseInvalidException;
use App\Exception\ZipArchiveException;
use App\Exception\ZipArchiveAccessException;
use App\Exception\ZipArchiveCorruptedException;
use App\Exception\ZipArchiveExtractException;
use App\Exception\ZipArchiveMemoryLimitException;
use App\Exception\ZipArchiveNoFileException;
use App\Utils;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Throwable;

class UpdateCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:update';

	/** @var Update $updater */
	private $updater;

	public function __construct(Update $updater)
	{
		$this->updater = $updater;
		parent::__construct();
	}

	protected function configure(): void
	{
		$this
			->setDefinition([
				new InputOption('offline', 'O', InputOption::VALUE_NONE, 'Do not download a new version. Use the update package in <info>var/update.zip</info>'),
				new InputOption('download-only', 'D', InputOption::VALUE_NONE, 'Don\'t install the update package. Only download it.'),
				new InputOption('force-version', '', InputOption::VALUE_REQUIRED, 'Request an explicit version id instead of using the latest version'),
			])
			->setDescription('Update to the latest version available')
			->setHelp("Updates the installation to another CCM19 version.\n\nBy default it updates to the latest available version, but you can change this by using <info>-O</info> or <info>--force-version</info>.")
		;
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$io = new SymfonyStyle($input, $output);
		$offline = (bool)$input->getOption('offline');
		$downloadOnly = (bool)$input->getOption('download-only');
		$forceVersion = $input->getOption('force-version') ?? null;

		return $this->perform($io, $forceVersion, $offline, $downloadOnly);
	}

	public function perform($io, $forceVersion, $offline, $downloadOnly): int
	{
		if ($offline and $downloadOnly) {
			$io->error("Cannot use both <info>--offline</info> and <info>--download-only</info>");
			return 2;
		}
		$doDownload = !$offline;

		// Prüfe Verzeichnisse rekursiv auf Schreibzugriff
		$notWriteableDirectories = Update::checkDirectoryAccess();
		if ($notWriteableDirectories) {
			$io->error(array_merge(["Please check if you have write access to the following directories:"], array_slice($notWriteableDirectories, 0, 20)));
			return 1;
		}

		$destination = Utils::getVarDir().'/update.zip';

		// DOWNLOAD
		if ($doDownload) {
			$io->comment("Downloading update");

			try {
				$progressBar = $io->createProgressBar(100);

				$result = $this->updater->downloadUpdate($destination, null, $forceVersion, function ($total, $current) use ($progressBar) {
					if ($total > 0) {
						$progressBar->setMaxSteps($total);
						$progressBar->setProgress($current);
					}
				});

				if ($result) {
					$progressBar->finish();
				} else {
					$progressBar->clear();
				}
			} catch (DownloaderException $e) {
				$io->error("Download failed: ".$e->getMessage());
				return 1;
			} catch (TempFileNotWritableException $e) {
				$io->error("Writing temporary file failed: ".$e->getMessage());
				return 1;
			}
			catch (LicenseInvalidException $e) {
				$progressBar->clear();
				$io->error("Invalid license");
				return 1;
			}

			if ($result) {
				$progressBar->finish();
				$io->newLine(2);
			} else {
				$progressBar->clear();
				$io->text('Your version is up to date. No update is necessary.');
				return 0;
			}
		}

		if ($downloadOnly) {
			$io->success('Update package downloaded into '.$destination);
			return 0;
		} else {
			$io->comment('Update package downloaded into '.$destination);
		}

		$config = Config::getInstance();
		if ($forceVersion and $config->get('enableAutoUpdate', false)) {
			$io->comment("Disabling auto updates because of downgrade/forced version id.");
			$config->set('enableAutoUpdate', false)->flush();
		}

		if (!file_exists($destination)) {
			$io->error($destination.' does not exist.');
			return 1;
		}
		elseif (!is_file($destination)) {
			$io->error($destination.' is not a file.');
			return 1;
		}
		elseif (!is_readable($destination)) {
			$io->error($destination.' is not readable.');
			return 1;
		}
		elseif ($offline) {
			$io->comment('Using update package in '.$destination);
		}

		// EXTRACTION PREPARATION
		try {
			$result = $this->updater->prepareUpdateExtraction($destination);
			$token = $result['token'];
			$fileCount = $result['fileCount'];
		}
		catch (ZipArchiveCorruptedException $e) {
			$io->error('Update archive corrupted. Please try again later.');
			return 1;
		}
		catch (ZipArchiveMemoryLimitException $e) {
			$io->error('Memory limit exceeded. Please increase the PHP memory limits.');
			return 1;
		}
		catch (ZipArchiveNoFileException $e) {
			$io->error('No update downloaded.');
			return 1;
		}
		catch (ZipArchiveAccessException $e) {
			$io->error('Cannot open downloaded archive. Please check the server file permissions.');
			return 1;
		}
		catch (ZipArchiveException $e) {
			$io->error('An error occured while opening the update archive. Please try again later.');
			return 1;
		}
		catch (TempFileNotWritableException $e) {
			$io->error('Cannot create temporary directory in var/cache/');
			return 1;
		}
		catch (Throwable $e) {
			$io->error($e->getMessage());
			return 1;
		}

		// EXTRACTION
		$io->comment('Unpacking update package');

		$progressBar = $io->createProgressBar($fileCount);
		$offset = 0;

		try {
			while (true) {
				$result = $this->updater->extractUpdate($destination, $token, $offset, 1);
				$offset = $result['nextOffset'];
				$progressBar->setProgress($offset);
				if ($result['done']) {
					break;
				}
			}
		}
		catch (ZipArchiveExtractException $e) {
			$io->error('Could not extract file '.$e->getMessage().' from archive. Unknown error.');
			return 1;
		}
		catch (TempFileNotWritableException $e) {
			$io->error('Could not write temporary file in var/cache/: '.$e->getMessage());
			return 1;
		}

		$progressBar->finish();
		$io->newLine(2);

		$io->comment('Cleaning up');
		$this->updater->finishUpdate($destination, $token);

		$io->success('Update completed successfully');

		return 0;
	}
}
