<?php
namespace App\Command\User;

use App\Model\User;
use App\Component\LoggerInterface;
use App\Event\AccountUpdateEvent;
use App\Utils;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;

class ResetPasswordCommand extends Command
{
	// the name of the command (the part after "bin/console")
	protected static $defaultName = 'app:user:reset-password';

	/** @var LoggerInterface $updater */
	private $logger;

	public function __construct(LoggerInterface $logger)
	{
		$this->logger = $logger;
		parent::__construct();
	}

	protected function configure(): void
	{
		$this
			->setDefinition([
				new InputArgument('user', InputArgument::OPTIONAL, 'user id or name'),
			])
			->setDescription('Set a new password for an user.')
		;
	}

	protected function execute(InputInterface $input, OutputInterface $output): int
	{
		$io = new SymfonyStyle($input, $output);

		$username = $input->getArgument('user');

		if ($username === null) {
			$users = ListCommand::makeSortedUserList();

			if (count($users) == 1) {
				$user = current($users);
			} else {
				$userLabels = array_map(function (User $v) {
					return $v->getUsername()."\t(".$v->getId().')';
				}, $users);

				$choice = $io->choice('Please select a user', $userLabels);
				$index = array_search($choice, $userLabels, true);
				$user = $users[$index];
			}
		}
		else {
			$user = User::byId($username) ?? User::find($username);
		}

		if (!$user) {
			$io->error('User doesn\'t exist.');
			return 2;
		}

		$io->comment(sprintf('Resetting password for user "%s" (%s).', $user->getUsername(), $user->getId()));

		for ($i = 3; $i > 0; --$i) {
			$password = $io->askHidden('New password');
			if ($password === '' or $password === null) {
				$io->comment('Aborted.');
				return 0;
			}

			$passwordRepeat = $io->askHidden('Repeat new password');

			if (!hash_equals($password, (string)$passwordRepeat)) {
				if ($i == 1) {
					$io->error('Passwords don\'t match. Aborted.');
				} else {
					$io->warning('Passwords don\'t match. Please try again.');
				}
			}
			else {
				$user->setPassword($password);
				$user->setPasswordRecoveryKey(null);
				$event = new AccountUpdateEvent($user, AccountUpdateEvent::CAUSE_CMDLINE);
				$event->setPasswordChanged();
				Utils::dispatchEvent(AccountUpdateEvent::NAME_PRE_SAVE, $event);
				$this->logger->logInfo('Console', sprintf('Password for user "%s" reset by command-line user.', $user->getUsername()));
				$user->save();
				Utils::dispatchEvent(AccountUpdateEvent::NAME_POST_SAVE, $event);

				$io->success('Password successfully set.');
				return 0;
			}
		}

		return 1;
	}
}
