<?php

namespace App\Component\Api;

use App\Component\Curl;
use App\Component\CurlResponse;
use App\Config;
use App\Utils;
use Dflydev\Hawk\Credentials\Credentials as HawkCredentials;
use Dflydev\Hawk\Server\Response as HawkResponse;
use Dflydev\Hawk\Server\Server as HawkServer;
use Dflydev\Hawk\Server\ServerBuilder as HawkServerBuilder;
use Dflydev\Hawk\Client\ClientBuilder as HawkClientBuilder;
use Dflydev\Hawk\Client\Client as HawkClient;
use Dflydev\Hawk\Server\UnauthorizedException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;
use UnexpectedValueException;

class ClusterApi
{
	public const PBKDF2_ALGORITHM = 'sha256';
	public const PBKDF2_ITERATIONS = 500;
	public const PBKDF2_KEY_LENGTH = 24;
	public const PBKDF2_SALT = 'cluster:hawk';

	/** @var self|null $instance */
	private static $instance = null;

	/** @var string $sharedSecret */
	private $sharedSecret;

	/** @var HawkServer $hawkServer */
	private $hawkServer;

	/** @var HawkClient $hawkClient */
	private $hawkClient;

	/** @var HawkCredentials $credentials */
	private $hawkCredentials;

	/** @var HawkResponse|null $hawkResponse */
	private $hawkResponse;

	/** @var UrlGeneratorInterface $router */
	private $router;

	private function __construct()
	{
		$config = Config::getInstance();
		$this->sharedSecret = (string)$_ENV['APP_SECRET'].'/'.(string)$config->get('licensekey');
		$this->router = Utils::getRouter();

		$this->initHawk();
	}

	/**
	 * @return self
	 */
	public static function getInstance(): self
	{
		if (self::$instance == null) {
			self::$instance = new self;
		}

		return self::$instance;
	}

	private function initHawk(): void
	{
		$key = hash_pbkdf2(
			self::PBKDF2_ALGORITHM,
			$this->sharedSecret,
			self::PBKDF2_SALT,
			self::PBKDF2_ITERATIONS,
			self::PBKDF2_KEY_LENGTH,
			true
		);
		$this->hawkCredentials = new HawkCredentials($key, 'sha256', null);

		$credentialsProvider = function ($id) {
			return $this->hawkCredentials;
		};

		$this->hawkServer = HawkServerBuilder::create($credentialsProvider)->build();
		$this->hawkClient = HawkClientBuilder::create()->build();
	}

	/**
	 * @param Request $request
	 * @throws UnauthorizedException Bei ungültigem API-Schlüssel oder sonstigen Hawk-Fehlern
	 */
	public function authorizeRequest(Request $request): void
	{
		$this->hawkResponse = $this->hawkServer->authenticate(
			$request->getRealMethod(),
			$request->getHttpHost(),
			$request->getPort(),
			$request->getRequestUri(),
			$request->headers->get('Content-Type'),
			$request->getContent() ?: null,
			$request->headers->get('Authorization')
		);
	}

	public function signResponse(Response $response): void
	{
		if ($this->hawkResponse) {
			$options = array_filter([
				'content_type' => (string)$response->headers->get('Content-Type'),
				'payload' => (string)$response->getContent(),
			]);

			$header = $this->hawkServer->createHeader(
				$this->hawkResponse->credentials(),
				$this->hawkResponse->artifacts(),
				$options
			);

			// Setzt den 'Server-Authorization' HTTP-Header
			$response->headers->set($header->fieldName(), $header->fieldValue());
		}
	}

	public static function listPeers(): array
	{
		$str = trim(preg_replace('/\s+/', ' ', $_ENV['APP_CLUSTER_PEERS'] ?? ''));
		if ($str === '') {
			return [];
		} else {
			return explode(' ', $str);
		}
	}

	/**
	 * Ruft einen Cluster-API-Endpunkt auf einem anderen Host des Clusters auf
	 *
	 * @param string $peer Basis-URL zur CCM19-Installation
	 * @param string $route Routenname des API-Endpunkts (z.B. "app_cluster_api_test")
	 * @param array|null $routeParams Routen-Parameter
	 * @param string $method HTTP-Methode
	 * @param array|null $data Daten (URL-Parameter bei HEAD/GET/DELETE, ansonsten JSON-Payload)
	 * @return CurlResponse
	 */
	public function callRemoteApi(string $peer, string $route, ?array $routeParams=[], string $method='GET', ?array $data=null): CurlResponse
	{
		$routePath = $this->router->generate($route, $routeParams ?? [], UrlGeneratorInterface::ABSOLUTE_PATH);
		// Pfad zur lokalen Installation aus generiertem Routen-Pfad entfernen
		$routePath = substr($routePath, strlen($this->router->getContext()->getBaseUrl()));

		$method = strtoupper($method);
		$curl = new Curl();

		// Payload/URL-Parameter vorbereiten
		switch ($method) {
			case 'HEAD':
			case 'DELETE':
			case 'GET':
				if ($data !== null and count($data) > 0) {
					if (strpos($routePath, '?') === false) {
						$routePath = $routePath.'?'.http_build_query($data);
					} else {
						$routePath = $routePath.'&'.http_build_query($data);
					}
					$data = null;
				}
				break;
			default:
				$curl->setHeader('Content-Type', 'application/json');
				$data = ($data === null) ? '{}' : json_encode($data,  JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES);
				break;
		}

		// URL zusammensetzen
		$url = rtrim($peer, '/') . $routePath;

		// Create signature
		$hawkRequest = $this->hawkClient->createRequest(
			$this->hawkCredentials,
			$url,
			$method,
			($data === null) ? [] : [
				'payload' => $data,
				'content_type' => 'application/json',
			]
		);
		$curl->setHeader($hawkRequest->header()->fieldName(), $hawkRequest->header()->fieldValue());

		switch ($method) {
			case 'HEAD':
				$curl->head($url);
				break;
			case 'GET':
				$curl->get($url);
				break;
			case 'DELETE':
				$curl->delete($url);
				break;
			case 'POST':
				assert(is_string($data));
				$curl->post($url, $data);
				break;
			case 'PUT':
				assert(is_string($data));
				// @phan-suppress-next-line PhanTypeMismatchArgument
				$curl->put($url, $data, true);
				break;
			case 'PATCH':
				assert(is_string($data));
				// @phan-suppress-next-line PhanTypeMismatchArgument
				$curl->patch($url, $data, true);
				break;
			default:
				throw new UnexpectedValueException('Unknown HTTP method: '.$method);
		}

		$response = $curl->getResponseObject();

		$isAuthenticatedResponse = $response->headers->has('Server-Authorization') ? $this->hawkClient->authenticate(
			$this->hawkCredentials,
			$hawkRequest,
			$response->headers->get('Server-Authorization') ?? '',
			($method === 'HEAD') ? [] : [
				'payload' => $response->getContent(),
				'content_type' => $response->headers->get('content-type'),
			]
		) : false;
		$response->setVerified($isAuthenticatedResponse);

		return $response;
	}

	/**
	 * Ruft einen Cluster-API-Endpunkt auf den restlichen Hosts (Peers) des Clusters auf
	 *
	 * @param string $route Routenname des API-Endpunkts (z.B. "app_cluster_api_test")
	 * @param array|null $routeParams Routen-Parameter
	 * @param string $method HTTP-Methode
	 * @param array|null $data Daten (URL-Parameter bei HEAD/GET/DELETE, ansonsten JSON-Payload)
	 * @return CurlResponse[]
	 */
	public function callRemoteApiBroadcast(string $route, ?array $routeParams=[], string $method='GET', ?array $data=null): array
	{
		$responses = [];

		foreach (self::listPeers() as $peer) {
			$responses[$peer] = $this->callRemoteApi($peer, $route, $routeParams, $method, $data);
		}

		return $responses;
	}
}
