<?php

namespace App\Component\Cache;

use App\Component\LocaleService;
use App\Model\Domain;
use App\Model\LocaleBase;
use App\Model\Theme;
use App\Model\User;
use App\Utils;
use LogicException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

class Cache
{
	public const EVENT_ALLOWED_QUERY_PARAMETERS = 'ccm19.cache.allowed_query_parameters';

	/** @var array $responseQueue */
	private static $responseQueue = [];

	/** @var (?array)[] $domainMetaCache */
	private static $domainMetaCache = [];

	/** @var int[][] $versionIdentifiers */
	private static $versionIdentifiers = [];

	/**
	 * @param Request $request
	 * @param array $salt
	 * @return string
	 */
	private static function hashRequest(Request $request, array $salt=[]): string
	{
		//Damit der Hash unabhängig von https/http ist - beim Loadbalancing kann es sonst zu Problemen kommen... Header Forwarding funktioniert nicht zuverlässig.
		$salt['host']=str_ireplace("https:",":",$salt['host']);
		$salt['host']=str_ireplace("http:",":",$salt['host']);

		return hash('sha1', $request->getRequestUri() . ($salt ? '&'.http_build_query($salt, '', '&') : ''));
	}

	/**
	 * @param string $userApiKey
	 * @return string
	 */
	private static function getUserCacheDir(string $userApiKey): string
	{
		return implode(DIRECTORY_SEPARATOR, [
			Utils::getCacheDir(),
			'frontend',
			substr($userApiKey, 0, 4),
			substr($userApiKey, 4),
		]);
	}

	/**
	 * @param string $userApiKey
	 * @param string $domainId
	 * @return string
	 */
	private static function getDomainCacheDir(string $userApiKey, string $domainId): string
	{
		return implode(DIRECTORY_SEPARATOR, [
			self::getUserCacheDir($userApiKey),
			$domainId,
		]);
	}

	/**
	 * @param Domain $domain
	 * @param string|null $domainId Domain-ID im Cache-Pfad, falls für Kompatibilitätszwecke nicht in kanonischer Form
	 * @return bool
	 */
	private static function storeDomainMetadata(Domain $domain, ?string $domainId=null): bool
	{
		$user = $domain->getUser();
		$domainId = $domainId ?? $domain->getId();
		assert($user instanceof User);

		$cachePathname = self::getDomainCacheDir($user->getApiKey(), $domainId) . DIRECTORY_SEPARATOR . 'domain.meta';

		if (
			is_dir(dirname($cachePathname)) == false && @mkdir(dirname($cachePathname), 0777, true) == false
			|| is_writable(dirname($cachePathname)) == false
			|| is_file($cachePathname) && is_writable($cachePathname) == false
		) {
			// Don't log to prevent log from flooding
			return false;
		}

		$oldState = self::loadDomainMetadata($user->getApiKey(), $domainId);

		// Versionsbezeichner vereinen
		$versionIdentifiers = array_values(array_unique(array_merge(
			$oldState['versionIdentifiers'] ?? [], self::$versionIdentifiers[$domainId] ?? []
		), SORT_NUMERIC));

		// Theme bestimmen
		$defaultTheme = Theme::activeTheme();

		// Locales bestimmen
		$localeNames = array_values(array_map(function (LocaleBase $locale) { return $locale->getName(); }, LocaleService::enabledLocales()));
		$defaultLocale = $user->getDefaultLocale();
		$defaultLocaleName = $defaultLocale->getName();
		if (!in_array($defaultLocaleName, $localeNames)) {
			$defaultLocaleName = 'de_DE';
		}
		// Standard-Locale an den Anfang der Liste schieben
		$pos = array_search($defaultLocaleName, $localeNames, true);
		if ($pos !== false) {
			unset($localeNames[$pos]);
			array_unshift($localeNames, $defaultLocaleName);
		}

		$metadata = self::$domainMetaCache["{$user->getApiKey()}/$domainId"] = [
			'apiKey' => $user->getApiKey(),
			'defaultThemeId' => $defaultTheme ? $defaultTheme->getId() : null,
			'domainId' => $domain->getId(),
			'euOnly' => array_map(function (Theme $theme) { return $theme->getOnlyInEu(); }, Theme::all()),
			'widgetVaryCountry' => array_map(function (Theme $theme) { return $theme->hasDeclineButtonRestrictions(); }, Theme::all()),
			'useCache' => $domain->isWidgetCachingEnabled(),
			'userId' => $user->getId(),
			'locales' => $localeNames,
			'versionIdentifiers' => $versionIdentifiers,
		];

		ksort($metadata['euOnly']);
		ksort($metadata['widgetVaryCountry']);

		return (bool)file_put_contents($cachePathname, json_encode($metadata), LOCK_EX);
	}

	/**
	 * @param string $userApiKey
	 * @param string $domainId
	 * @return array|null
	 */
	private static function loadDomainMetadata(string $userApiKey, string $domainId): ?array
	{
		$data = null;
		$cachePathname = self::getDomainCacheDir($userApiKey, $domainId) . DIRECTORY_SEPARATOR . 'domain.meta';
		$metaCacheKey = $userApiKey.'/'.$domainId;

		if (isset(self::$domainMetaCache[$metaCacheKey])) {
			return self::$domainMetaCache[$metaCacheKey];
		}
		else {
			// Read metadata for cached content
			if ($file = @fopen($cachePathname, 'rb') and (flock($file, LOCK_SH) or (fclose($file) and false))) {
				try {
					$data = json_decode((string)@stream_get_contents($file), true);
				}
				finally {
					flock($file, LOCK_UN);
					fclose($file);
				}
			}

			return (self::$domainMetaCache[$metaCacheKey] = is_array($data) ? $data : null);
		}
	}

	/**
	 * Diese Methode merkt einen gültigen Wert für den URL-Parameter `v` vor, um eine Flutung
	 * des Caches zu unterbinden.
	 * @param Domain $domain
	 * @param int $version
	 * @return void
	 */
	public static function registerVersionIdentifier(Domain $domain, int $version): void
	{
		if (array_key_exists($domain->getId(), self::$versionIdentifiers) == false) {
			self::$versionIdentifiers[$domain->getId()] = [];
		}

		self::$versionIdentifiers[$domain->getId()][$version] = $version;

		// Consider IDs of domains that were migrated from JSON store
		if (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB) {
			$user = $domain->getUser();
			assert($user instanceof User);
			$domainAlias = Domain::convertObjectIdToJsonId($user->getId(), $domain->getId());

			if ($domainAlias != $domain->getId()) {
				if (is_array(self::$versionIdentifiers[$domainAlias] ?? null) == false) {
					self::$versionIdentifiers[$domainAlias] = [];
				}

				self::$versionIdentifiers[$domainAlias][$version] = $version;
			}
		}
	}

	/**
	 * @param string $userApiKey
	 * @param string $domainId
	 * @param string|null $themeId
	 * @return bool
	 */
	public static function isDomainEuOnly(string $userApiKey, string $domainId, ?string $themeId=null): bool
	{
		$metadata = self::loadDomainMetadata($userApiKey, $domainId);
		$themeId = $themeId ?? $metadata['defaultThemeId'] ?? null;
		return $metadata['euOnly'][$themeId] ?? false;
	}

	/**
	 * Gibt zurück, ob das Widget je nach Land/Region unterschiedlich ist
	 *
	 * Die Abhängigkeit ist im Response-Header `X-Target-Region` kodiert.
	 *
	 * @param string $userApiKey
	 * @param string $domainId
	 * @param string|null $themeId
	 * @return bool
	 */
	public static function isDomainWidgetVaryCountry(string $userApiKey, string $domainId, ?string $themeId=null): bool
	{
		$metadata = self::loadDomainMetadata($userApiKey, $domainId);
		$themeId = $themeId ?? $metadata['defaultThemeId'] ?? null;
		return $metadata['widgetVaryCountry'][$themeId] ?? false;
	}

	/**
	 * @param string $userApiKey
	 * @param string $domainId
	 * @return string[]
	 */
	public static function getActiveLocales(string $userApiKey, string $domainId): array
	{
		$metadata = self::loadDomainMetadata($userApiKey, $domainId);
		return $metadata['locales'] ?? [];
	}

	/**
	 * @param Request $request
	 * @param Response $response
	 * @param Domain $domain
	 * @param array $salt
	 * @param int|null $lifetime
	 * @throws LogicException if request method isn't `GET` or `HEAD`.
	 */
	public static function storeResponseAddToQueue(Request $request, Response $response, Domain $domain, array $salt=[], ?int $lifetime=null): void
	{
		if (!$request->isMethodCacheable()) {
			throw new LogicException("Cannot cache response for request method '{$request->getMethod()}'.");
		}

		self::$responseQueue[] = [
			'request' => $request,
			'response' => $response,
			'domain' => $domain,
			'salt' => $salt,
			'lifetime' => $lifetime,
		];
	}

	public static function storeResponseExecuteQueue(): void
	{
		foreach (self::$responseQueue as $item) {
			self::storeResponse($item['request'], $item['response'], $item['domain'], $item['salt'], $item['lifetime']);
		}

		self::$responseQueue = [];
	}

	/**
	 * Gibt zurück, welche komprimierte Version für den Request verwendet werden kann.
	 *
	 * Wenn keine Kompression unterstützt wird, wird [$cachePathname, 'identity'] zurückgegeben.
	 *
	 * @param Request $request
	 * @param string $cachePathname Pfad zur unkomprimierten Cache-Datei
	 * @return string[] [$cachePathname, $encodingName]
	 */
	private static function evaluateAcceptedCompression(Request $request, string $cachePathname): array
	{
		// Check if Cloudflare. They only support gzip but don't neccessarily announce it.
		if ($request->headers->has('CF-Ray') and is_readable($cachePathname.'.gz')) {
			$cachePathname = $cachePathname.'.gz';
			$contentEncoding = 'gzip';
			return [$cachePathname.'.gz', 'gzip'];
		}
		else {
			// Check if a prepared compression can be applied
			$acceptedEncodings = $request->getEncodings();
			$contentEncoding = '';
			foreach (['.br'=>'br', '.gz'=>'gzip'] as $suffix => $encoding) {
				if (in_array($encoding, $acceptedEncodings) and is_readable($cachePathname.$suffix)) {
					return [$cachePathname.$suffix, $encoding];
				}
			}
		}
		return [$cachePathname, 'identity'];
	}

	/**
	 * @param Request $request
	 * @param Response $response
	 * @param Domain $domain
	 * @param array $salt
	 * @param int|null $lifetime Für wie viele Sekunden die Response als gültig angesehen werden soll
	 * @param bool $modifyResponse Wenn true, wird die aktuelle Response ggf. auf eine komprimierte Version umgeschrieben
	 * @throws LogicException if request method isn't `GET` or `HEAD`.
	 */
	public static function storeResponse(Request $request, Response $response, Domain $domain, array $salt=[], ?int $lifetime=null, bool $modifyResponse=true): void
	{
		if (!$request->isMethodCacheable()) {
			throw new LogicException("Cannot cache response for request method '{$request->getMethod()}'.");
		}

		$user = $domain->getUser();
		assert($user instanceof User);

		$metadata = self::loadDomainMetadata($user->getApiKey(), $domain->getId());

		$allowedQueryParameters = [
			'apiKey',
			'domain',
			'gen',
			'lang',
			'systemTheme',
			'theme',
			'userTheme',
			'v',
		];

		// Give other components a chance to add custom query parameters
		$allowedQueryParameters = array_unique(array_merge(
			$allowedQueryParameters,
			Utils::filterWithEvent(self::EVENT_ALLOWED_QUERY_PARAMETERS, [])
		));

		parse_str(parse_url($request->getRequestUri(), PHP_URL_QUERY), $query);

		// Don't cache for requests that specify illegal query parameters
		if (array_diff(array_keys($query), $allowedQueryParameters)) {
			return;
		}
		// Disallow illegal gen identifiers
		elseif ($request->query->has('gen') && in_array($request->query->getInt('gen'), [1, 2]) == false) {
			return;
		}
		// Disallow unregistered version identifiers
		elseif ($request->query->has('v') && in_array($request->query->getInt('v'), $metadata['versionIdentifiers'] ?? []) == false) {
			return;
		}

		// Prioritize the URL parameter to enable reading from cache for old URLs after a migration from JSON to MongoDB
		$domainId = $request->query->getAlnum('domain', $domain->getId());
		if ($domainId === '') {
			return;
		}

		// Store domain metadata for the canonical ID and aliases
		$domainAliases = array_values(array_unique([
			$domain->getId(),
			$domainId,
		]));

		$domainMetadataOk = array_reduce($domainAliases, static function (bool $success, string $domainId) use ($domain) {
			return $success && self::storeDomainMetadata($domain, $domainId);
		}, true);

		$salt += [
			'domain' => $domainId, // Explicitly specify the domain in case it was detected automatically
			'host' => $request->getSchemeAndHttpHost(), // Consider absolute URLs within the cached resource
		];

		$hash = self::hashRequest($request, $salt);
		$cachePathname = self::getDomainCacheDir($user->getApiKey(), $domainId) . DIRECTORY_SEPARATOR . $hash;

		if (
			$domainMetadataOk == false
			|| is_dir(dirname($cachePathname)) == false && @mkdir(dirname($cachePathname), 0777, true) == false
			|| is_writable(dirname($cachePathname)) == false
			|| is_file($cachePathname) && is_writable($cachePathname) == false
			|| is_file($cachePathname.'.meta') && is_writable($cachePathname.'.meta') == false
		) {
			// Don't log to prevent log from flooding
			return;
		}

		$content = $response->getContent();
		$could_compress = false;
		$compressedContent = [];
		file_put_contents($cachePathname, $content, LOCK_EX);
		if (function_exists('gzencode')) {
			$could_compress = true;
			$compressedContent['gzip'] = gzencode($content, 9);
			file_put_contents($cachePathname.'.gz', $compressedContent['gzip'], LOCK_EX);
		}
		if (function_exists('brotli_compress')) {
			$could_compress = true;
			$compressedContent['br'] = brotli_compress($content, 10);
			file_put_contents($cachePathname.'.br', $compressedContent['br'], LOCK_EX);
		}

		if ($could_compress) {
			$vary = $response->getVary();
			if (($vary[0] ?? '') !== '*') {
				$vary[] = 'Accept-Encoding';
				$response->setVary($vary);
			}
		}

		$f = fopen($cachePathname, 'rb');
		if (!$f) {
			return; // Wenn die gerade geschriebene Datei sich nicht öffnen lässt, abbrechen
		}
		// ETag aus Dateihash berechnen, während die Datei gesperrt ist
		$etag = null;
		if (flock($f, LOCK_SH)) {
			try {
				$headerHash = sha1($response->headers->get("Vary", "") . ';' . $response->headers->get("X-Target-Region", ""), true);
				$fileHash = hash_file('sha1', $cachePathname, true);
				$etag = base64_encode(substr($headerHash, 0, 4).$fileHash);
				$response->setEtag($etag);
			} finally {
				flock($f, LOCK_UN);
				fclose($f);
			}
		} else {
			fclose($f);
		}

		file_put_contents($cachePathname.'.meta', json_encode([
			'headers' => $response->headers->all(),
			'lifetime' => $lifetime,
			'uri' => $request->getRequestUri(),
		]), LOCK_EX);

		if ($modifyResponse) {
			// Check if compression can be applied to the current request
			list($cachePathname, $contentEncoding) = self::evaluateAcceptedCompression($request, $cachePathname);

			if ($contentEncoding !== 'identity' and isset($compressedContent[$contentEncoding])) {
				$content = $compressedContent[$contentEncoding];
				$response->setContent($content);
				$response->headers->set('Content-Length', (string)strlen($content));
				$response->headers->set('Content-Encoding', $contentEncoding);
				if ($etag !== null) {
					$response->setEtag($etag.'-'.$contentEncoding);
				}
			}
		}
	}

	/**
	 * @param Request $request
	 * @param string $userApiKey
	 * @param string $domainId
	 * @param array $salt
	 * @param callable|null $beforeSend
	 */
	public static function loadAndSendResponse(Request $request, string $userApiKey, string $domainId, array $salt=[], ?callable $beforeSend=null): void
	{
		// Also deliver cached resource for OPTIONS request
		if (!$request->isMethodCacheable() && $request->getMethod() != 'OPTIONS') {
			return;
		}

		$salt += [
			'domain' => $domainId, // Explicitly specify the domain in case it was detected automatically
			'host' => $request->getSchemeAndHttpHost(), // Consider absolute URLs within the cached resource
		];

		$hash = self::hashRequest($request, $salt);
		$cachePathname = self::getDomainCacheDir($userApiKey, $domainId) . DIRECTORY_SEPARATOR . $hash;
		$domainMeta = self::loadDomainMetadata($userApiKey, $domainId);

		//Das ist sauberer - dann muss man das fopen auch nicht mit @unterdrücken...
		if(!file_exists($cachePathname.'.meta'))
		{
			return;
		}

		// Read metadata for cached content
		$cacheMeta = null;
		if ($metaFile = @fopen($cachePathname.'.meta', 'rb') and (flock($metaFile, LOCK_SH) or (fclose($metaFile) and false))) {
			try {
				$cacheMeta = json_decode((string)@stream_get_contents($metaFile), true);
			}
			finally {
				flock($metaFile, LOCK_UN);
				fclose($metaFile);
			}
		}
		else {
			return;
		}

		// Abort if metadata is missing or the request hasn't been cached yet
		if (!$domainMeta || !$cacheMeta || !is_file($cachePathname) || !is_readable($cachePathname)) {
			return;
		}

		// Abort if cache is disabled
		if (!($domainMeta['useCache'] ?? false)) {
			return;
		}

		// Abort if cached resource has expired
		if (is_int($lifetime = $cacheMeta['lifetime'] ?? null) && filemtime($cachePathname) + $lifetime < time()) {
			return;
		}

		// OPTIONS-Shortcut: Nur Access-Control-*-Headers werden benötigt, kein Content
		if ($request->getMethod() == 'OPTIONS') {
			header_remove('Content-Type');
			header_remove('Cache-Control');
			header_remove('Expires');
			http_response_code(204);

			$headers = $cacheMeta['headers'] ?? [];
			array_walk($headers, function ($values, $name) use ($cachePathname) {
				$lcName = strtolower($name);
				if (strpos($lcName, 'access-control-') !== 0 and !in_array($lcName, ['allow'])) {
					return;
				}

				foreach ($values as $key => $value) {
					header("$name: $value", $key == 0);
				}
			});

			exit();
		}

		// 304-Shortcut
		$noneMatch = $request->headers->get('If-None-Match', '');
		$etag = $cacheMeta['headers']['etag'][0] ?? null;
		if ($etag !== null) {
			$etag = trim($etag, '"');
			$etagMatches = $noneMatch == '"'.$etag.'"';
			$gzipEtagMatches = $noneMatch == '"'.$etag.'-gzip"';
			$brotliEtagMatches = $noneMatch == '"'.$etag.'-br"';
			if ($etagMatches or $gzipEtagMatches or $brotliEtagMatches) {
				header_remove('Content-Type');
				header_remove('Cache-Control');
				header_remove('Expires');
				http_response_code(304);

				// Send neccessary cached response headers
				$headers = $cacheMeta['headers'] ?? [];
				array_walk($headers, function ($values, $name) use ($gzipEtagMatches, $brotliEtagMatches, $cachePathname) {
					$lcName = strtolower($name);
					if (!in_array($lcName, ['cache-control', 'content-location', 'etag', 'expires', 'vary'])) {
						return;
					}

					// Modify etag when compression was used
					if ($lcName == 'etag') {
						if ($brotliEtagMatches and file_exists($cachePathname.'.br')) {
							$values = array_map(function ($value) {
								return substr($value, 0, -1).'-br"';
							}, $values);
						}
						elseif ($gzipEtagMatches and file_exists($cachePathname.'.gz')) {
							$values = array_map(function ($value) {
								return substr($value, 0, -1).'-gzip"';
							}, $values);
						}
					}

					foreach ($values as $key => $value) {
						header("$name: $value", $key == 0);
					}
				});

				if (is_callable($beforeSend)) {
					$beforeSend($domainMeta);
				}

				exit();
			}
		}

		// Check if a prepared compression can be applied
		list($cachePathname, $contentEncoding) = self::evaluateAcceptedCompression($request, $cachePathname);

		// Send response
		if ($contentFile = @fopen($cachePathname, 'rb')) {
			$usingCache = false;
			try {
				if (flock($contentFile, LOCK_SH)) {
					// Send cached response headers
					$headers = $cacheMeta['headers'] ?? [];
					$etag = null;
					array_walk($headers, function ($values, $name) use (&$etag) {
						if (strpos($name, 'symfony') === 0) {
							return;
						}

						if (strtolower($name) == 'date') {
							return;
						}

						if (strtolower($name) == 'etag') {
							$etag = $values[0];
						}

						foreach ($values as $key => $value) {
							header("$name: $value", $key == 0);
						}
					});
					// Send real content length
					header("Content-Length: " . fstat($contentFile)['size']);

					// Set encoding and modify etag if compression is used
					if ($contentEncoding !== 'identity') {
						header("Content-Encoding: " . $contentEncoding);
						if ($etag !== null) {
							header("ETag: " . substr($etag, 0, -1).'-'.$contentEncoding.'"');
						}
					}

					if (is_callable($beforeSend)) {
						$beforeSend($domainMeta);
					}

					// Send cached response body
					fpassthru($contentFile);
					$usingCache = true;
				}
			}
			finally {
				flock($contentFile, LOCK_UN);
				fclose($contentFile);

				if ($usingCache) {
					exit;
				}
			}
		}
	}

	/**
	 * @param Domain $domain
	 * @param bool $keepMetadata
	 */
	public static function clear(Domain $domain, bool $keepMetadata=false): void
	{
		$user = $domain->getUser();
		assert($user instanceof User);

		// A single domain can be referenced in the cache by multiple IDs.
		// For example, if it was migrated from a JSON to a MongoDB store installation.
		$domainAliases = [
			$domain->getId(),
		];

		if (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB) {
			$domainAliases[] = Domain::convertObjectIdToJsonId($user->getId(), $domain->getId());
		}

		$domainAliases = array_unique($domainAliases);

		foreach ($domainAliases as $domainId) {
			$cacheDir = self::getDomainCacheDir($user->getApiKey(), $domainId);

			if (!is_dir($cacheDir)) {
				continue;
			}

			// Cache metadata
			if ($keepMetadata) {
				self::loadDomainMetadata($user->getApiKey(), $domainId);
			}

			rename($cacheDir, $cacheDir.'.tmp');
			Utils::removeDirectory($cacheDir.'.tmp');

			// Restore metadata (this preserves version identifiers)
			if ($keepMetadata) {
				self::storeDomainMetadata($domain, $domainId);
			}
			else {
				unset(self::$domainMetaCache["{$user->getApiKey()}/$domainId"]);
			}
		}
	}

	/**
	 * @param User $user
	 */
	public static function clearUser(User $user): void
	{
		$cacheDir = self::getUserCacheDir($user->getApiKey());

		if (is_dir($cacheDir)) {
			rename($cacheDir, $cacheDir.'.tmp');
			Utils::removeDirectory($cacheDir.'.tmp');
		}
	}

	public static function clearAll(): void
	{
		$cacheDir = Utils::getCacheDir() . DIRECTORY_SEPARATOR . 'frontend';

		if (is_dir($cacheDir)) {
			rename($cacheDir, $cacheDir.'.tmp');
			Utils::removeDirectory($cacheDir.'.tmp');
		}
	}
}
