<?php

namespace App\Component;

use App\Component\Curl;
use ErrorException;
use Exception;
use RuntimeException;

/**
 * @deprecated
 */
class Ccm19LicensingApiClient
{
	const API_URL = 'https://licence.ccm19.de/';

	/** @var bool $success */
	private $success;
	/** @var string $error */
	private $error;

	/** @var bool $connectionError */
	private $connectionError;
	/** @var bool $serverError */
	private $serverError;

	/** @var string $edition */
	private $edition;
	/** @var int $limit */
	private $limit;
	/** @var int $whitelabelMaxCount */
	private $whitelabelMaxCount;
	/** @var int $whitelabelCount */
	private $whitelabelCount;
	/** @var bool $whitelabel */
	private $whitelabel;

	/** @var int $maxcalls */
	private $maxcalls;

	/** @var string $tarif */
	private $tarif;

	public function __construct()
	{
		$this->reset();
	}

	private function reset(): void
	{
		$this->success = true;
		$this->error = '';

		$this->connectionError = false;
		$this->serverError = false;

		$this->edition = '';
		$this->limit = 0;
		$this->whitelabelMaxCount = 0;
		$this->whitelabelCount = 0;
		$this->whitelabel = false;
		$this->tarif ="free";
		$this->maxcalls =-1;
	}

	/**
	 * @param string $endpoint
	 * @param array $params
	 * @param array $headers
	 * @return array|null
	 * @throws ErrorException Wird geschmissen, wenn cURL nicht zur Verfügung steht.
	 * @throws RuntimeException Wenn ein Problem in der Kommunikation mit dem API-Server auftritt.
	 */
	private function get(string $endpoint, array $params = [], array $headers = []): ?array
	{
		$params = ['action' => $endpoint] + $params;

		$curl = new Curl();
		$curl->setOpt(CURLOPT_CONNECTTIMEOUT, 3);
		$curl->setOpt(CURLOPT_TIMEOUT, 7);

		foreach ($headers as $name => $value) {
			$curl->setHeader($name, $value);
		}

		$curl->get(self::API_URL, $params);

		if ($curl->curl_error) {
			$this->connectionError = true;
			throw new RuntimeException("An error has occurred in cURL module. (Code: {$curl->getErrorCode()}) ".$curl->getErrorMessage());
		}
		elseif ($curl->getResponseHeaders('Content-Type') != 'application/json') {
			$this->serverError = true;
			throw new RuntimeException('Unexpected response from license server.');
		}
		elseif ($curl->isServerError()) {
			$this->serverError = true;
		}

		$response = json_decode($curl->getResponse(), true);
		return is_array($response) ? $response : null;
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @return bool
	 */
	public function registerLicense(string $licenseKey, string $domain): bool
	{
		$this->reset();

		$response = null;
		try {
			$response = $this->get('registerLicense', [
				'licence_id' => $licenseKey,
				'dom' => $domain,
			]);

			$this->success = $response && ($response['ok'] ?? false);
		}
		catch (Exception $exception) {
			$this->setErrorMessage($exception->getMessage());
			$this->success = false;
			return false;
		}

		if ($this->isSuccess()) {
			$this
				->setEdition((string)($response['edition'] ?? ''))
				->setLimit(max(-1, (int)($response['limit'] ?? -1)))
				->setMaxCalls(max(-1, (int)($response['maxcalls'] ?? -1)))
				->setWhitelabel((bool)($response['whitelabel'] ?? false))
				->setTarif((string)($response['tarif'] ?? "free"))
			;
		}
		else {
			$this->setErrorMessage($response ? (string)($response['error'] ?? '') : 'License server unreachable');
		}

		return $this->isSuccess();
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @return bool
	 */
	public function unregisterLicense(string $licenseKey, string $domain): bool
	{
		$this->reset();

		$response = null;
		try {
			$response = $this->get('unregisterLicense', [
				'licenseKey' => $licenseKey,
				'domain' => $domain,
			]);

			$this->success = $response && ($response['success'] ?? false);
		}
		catch (Exception $exception) {
			$this->setErrorMessage($exception->getMessage());
			$this->success = false;
			return false;
		}

		if ($this->isSuccess()) {
			$this
				->setEdition((string)($response['edition'] ?? ''))
				->setLimit(max(-1, (int)($response['limit'] ?? -1)))
				->setMaxCalls(max(-1, (int)($response['maxcalls'] ?? -1)))
				->setWhitelabel((bool)($response['whitelabel'] ?? false))
			;
		}
		else {
			$this->setErrorMessage($response ? (string)($response['error'] ?? '') : 'Unknown error');
		}

		return $this->isSuccess();
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @return bool
	 */
	public function registerWhitelabel(string $licenseKey, string $domain): bool
	{
		$this->reset();

		$response = null;
		try {
			$response = $this->get('registerWhitelabel', [
				'licenseKey' => $licenseKey,
				'domain' => $domain,
			]);

			$this->success = $response && ($response['success'] ?? false);
		}
		catch (Exception $exception) {
			$this->setErrorMessage($exception->getMessage());
			$this->success = false;
			return false;
		}

		if ($this->isSuccess()) {
			$this
				->setEdition((string)($response['edition'] ?? ''))
				->setLimit(max(-1, (int)($response['limit'] ?? -1)))
				->setMaxCalls(max(-1, (int)($response['maxcalls'] ?? -1)))
				->setWhitelabelMaxCount((int)($response['whitelabelMaxCount'] ?? 0))
				->setWhitelabelCount((int)($response['whitelabelCount'] ?? 0))
				->setWhitelabel((bool)($response['whitelabel'] ?? false))
			;
		}
		else {
			$this->setErrorMessage($response ? (string)($response['error'] ?? '') : 'Unknown error');
		}

		return $this->isSuccess();
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @return bool
	 */
	public function unregisterWhitelabel(string $licenseKey, string $domain): bool
	{
		$this->reset();

		$response = null;
		try {
			$response = $this->get('unregisterWhitelabel', [
				'licenseKey' => $licenseKey,
				'domain' => $domain,
			]);

			$this->success = $response && ($response['success'] ?? false);
		}
		catch (Exception $exception) {
			$this->setErrorMessage($exception->getMessage());
			$this->success = false;
			return false;
		}

		if ($this->isSuccess()) {
			$this
				->setEdition((string)($response['edition'] ?? ''))
				->setLimit(max(-1, (int)($response['limit'] ?? -1)))
				->setMaxCalls(max(-1, (int)($response['maxcalls'] ?? -1)))
				->setWhitelabelMaxCount((int)($response['whitelabelMaxCount'] ?? 0))
				->setWhitelabelCount((int)($response['whitelabelCount'] ?? 0))
				->setWhitelabel((bool)($response['whitelabel'] ?? false))
			;
		}
		else {
			$this->setErrorMessage($response ? (string)($response['error'] ?? '') : 'Unknown error');
		}

		return $this->isSuccess();
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @param int $domainCount
	 * @param int $userCount
	 * @param string $sumCount
	 * @return bool
	 */
	public function reportLicensees(string $licenseKey, string $domain, int $domainCount, int $userCount, string $sumCount): bool
	{
		$this->reset();

		$response = null;
		try {
			$response = $this->get('reportLicensees', [
				'licenseKey' => $licenseKey,
				'domain' => $domain,
				'domainCount' => $domainCount,
				'userCount' => $userCount,
				'sumCount' => urlencode($sumCount),
			]);

			$this->success = $response && ($response['success'] ?? false);
		}
		catch (Exception $exception) {
			$this->setErrorMessage($exception->getMessage());
			$this->success = false;
			return false;
		}

		if ($this->isError()) {
			$this->setErrorMessage($response ? (string)($response['error'] ?? '') : 'Unknown error');
		}

		return $this->isSuccess();
	}

	/**
	 * @return bool
	 */
	public function isSuccess(): bool
	{
		return $this->success;
	}

	/**
	 * @return bool
	 */
	public function isError(): bool
	{
		return $this->isSuccess() == false;
	}

	/**
	 * @return bool
	 */
	public function isConnectionError(): bool
	{
		return $this->connectionError;
	}

	/**
	 * @return bool
	 */
	public function isServerError(): bool
	{
		return $this->serverError;
	}

	/**
	 * @param string $message
	 * @return $this
	 */
	private function setErrorMessage(string $message): self
	{
		$this->error = $message;
		return $this;
	}

	/**
	 * @return string
	 */
	public function getErrorMessage(): string
	{
		return $this->error;
	}

	/**
	 * @param string $edition
	 * @return $this
	 */
	private function setEdition(string $edition): self
	{
		$this->edition = $edition;
		return $this;
	}

	/**
	 * @return string
	 */
	public function getEdition(): string
	{
		return $this->edition;
	}

	/**
	 * @param int $limit
	 * @return $this
	 */
	private function setLimit(int $limit): self
	{
		$this->limit = max(-1, $limit);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getLimit(): int
	{
		return $this->limit;
	}

	/**
	 * @return int
	 */
	public function getmaxcalls(): int
	{
		return $this->maxcalls;
	}

	/**
	 * @param int $maxcalls
	 * @return $this
	 */
	private function setMaxCalls(int $maxcalls): self
	{
		$this->maxcalls = $maxcalls;
		return $this;
	}

	/**
	 * @param string $tarif
	 * @return $this
	 */
	private function setTarif(string $tarif): self
	{
		$this->tarif = $tarif;
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTarif(): string
	{
		return $this->tarif;
	}

	/**
	 * @param int $count
	 * @return $this
	 */
	private function setWhitelabelMaxCount(int $count): self
	{
		$this->whitelabelMaxCount = $count;
		return $this;
	}

	/**
	 * @return int
	 */
	public function getWhitelabelMaxCount(): int
	{
		return $this->whitelabelMaxCount;
	}

	/**
	 * @param int $count
	 * @return $this
	 */
	private function setWhitelabelCount(int $count): self
	{
		$this->whitelabelCount = $count;
		return $this;
	}

	/**
	 * @return int
	 */
	public function getWhitelabelCount(): int
	{
		return $this->whitelabelCount;
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	private function setWhitelabel(bool $state): self
	{
		$this->whitelabel = $state;
		return $this;
	}

	/**
	 * @return bool
	 */
	public function hasWhitelabel(): bool
	{
		return $this->whitelabel;
	}
}
