<?php

namespace App\Component;

use OutOfRangeException;
use UnexpectedValueException;
use LogicException;

class Color
{
	/** @var float $h */
	private $h = 0.0;
	/** @var float $s */
	private $s = 0.0;
	/** @var float $l */
	private $l = 0.0;
	/** @var float $a */
	private $a = 0.0;
	/** @var int $rgb */
	private $rgb = 0;

	/**
	 * @see Color::fromHSL()
	 * @see Color::fromRGB()
	 *
	 * @param array{h:float,s:float,l:float} $hsl ['h' => 0.f..360, 's' => 0..1, 'l' => 0..1]
	 * @param int $rgb 3-Byte RGB-Farbwert
	 * @param float $alpha Alpha-Wert (0..1)
	 * @throws OutOfRangeException bei aktivierten Assertions: wenn HSL-Werte out-of-range sind
	 */
	private function __construct(array $hsl, int $rgb, float $alpha)
	{
		$this->h = (float)$hsl['h'];
		$this->s = (float)$hsl['s'];
		$this->l = (float)$hsl['l'];
		$this->a = (float)$alpha;
		$this->rgb = $rgb;
	}

	/**
	 * Color-Objekt aus HSL-Farbwert erzeugen
	 *
	 * @param float[] $hsl ['h' => 0.f..360, 's' => 0..1, 'l' => 0..1]
	 * @param float $alpha
	 * @return Color
	 */
	public static function fromHSL(array $hsl, float $alpha=1.0): self
	{
		$rgb = Color::hsl2Rgb($hsl);
		return new self($hsl, $rgb, $alpha);
	}

	/**
	 * Color-Objekt aus RGB-Farbwert erzeugen
	 *
	 * @param int $rgb 3-Byte RGB-Farbwert
	 * @param float $alpha
	 * @return Color
	 */
	public static function fromRGB(int $rgb, float $alpha=1.0): self
	{
		$hsl = Color::rgb2Hsl($rgb);
		return new self($hsl, $rgb, $alpha);
	}

	/**
	 * Hilfsfunktion: `preg_match` auf mehreren Regexes, bis der erste passt
	 * @param string[] $patterns
	 * @param string $string
	 * @param string[] $matches
	 * @return bool
	 */
	private static function matchOne(array $patterns, string $string, &$matches): bool {
		foreach ($patterns as $pattern) {
			if (preg_match($pattern, $string, $matches)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * CSS-Farbstring validieren (derzeit nur RGB(A))
	 *
	 * @param string $colorString
	 * @return bool
	 */
	public static function validateCssString(string $colorString): bool {
		try {
			self::fromCssString($colorString);
			return true;
		} catch (UnexpectedValueException $ignore) {
			return false;
		}
	}

	/**
	 * Color-Objekt aus CSS-Farbstring erzeugen (derzeit nur RGB(A))
	 *
	 * @param string $colorString
	 * @return Color
	 * @throws UnexpectedValueException
	 */
	public static function fromCssString(string $colorString): self
	{
		if (preg_match('~^#([0-9a-f]{6})$~i', $colorString, $match)) {
			return Color::fromRGB(hexdec($match[1]));
		}
		elseif (preg_match('~^#([0-9a-f]{6})([0-9a-f]{2})$~i', $colorString, $match)) {
			return Color::fromRGB(hexdec($match[1]), hexdec($match[2])/255.0);
		}
		elseif (preg_match('~^#([0-9a-f])([0-9a-f])([0-9a-f])([0-9a-f])?$~i', $colorString, $match)) {
			return Color::fromRGB(
				(hexdec($match[1]) * 0x11 << 16) | (hexdec($match[2]) * 0x11 << 8) | (hexdec($match[3]) * 0x11),
				(isset($match[4])) ? hexdec($match[2]) * 0x11 / 255.0 : 1.0
			);
		}
		elseif (self::matchOne([
				'~^rgba?\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*(?:,\s*([0-9.]+%?)\s*)?\)$~i',
				'~^rgba?\(\s*([0-9]+)\s+([0-9]+)\s+([0-9]+)\s*(?:/\s*([0-9.]+%?)\s*)?\)$~i',
			], $colorString,
			$match)) {
			if (isset($match[4])) {
				$a = (strpos($match[4], '%') !== false) ? ((float)substr($match[4], 0, -1)) / 100 : (float)$match[4];
			} else {
				$a = 1.0;
			}
			$r = (int)min((int)$match[1], 255);
			$g = (int)min((int)$match[2], 255);
			$b = (int)min((int)$match[3], 255);
			return Color::fromRGB(
				($r << 16) | ($g << 8) | $b,
				$a
			);
		}
		elseif (self::matchOne([
				'~^rgba?\(\s*([0-9]+)%\s*,\s*([0-9]+)%\s*,\s*([0-9]+)%\s*(?:,\s*([0-9.]+%?)\s*)?\)$~i',
				'~^rgba?\(\s*([0-9]+)%\s+([0-9]+)%\s+([0-9]+)%\s*(?:/\s*([0-9.]+%?)\s*)?\)$~i',
			], $colorString,
			$match)) {
			if (isset($match[4])) {
				$a = (strpos($match[4], '%') !== false) ? ((float)substr($match[4], 0, -1)) / 100 : (float)$match[4];
			} else {
				$a = 1.0;
			}
			$r = (int)min((int)($match[1]) * 2.55, 255);
			$g = (int)min((int)($match[2]) * 2.55, 255);
			$b = (int)min((int)($match[3]) * 2.55, 255);
			return Color::fromRGB(
				($r << 16) | ($g << 8) | $b,
				$a
			);
		}
		else {
			throw new UnexpectedValueException('"'.addslashes($colorString).'" is not a supported color string');
		}

	}

	/**
	 * String-Darstellung der Farbe
	 *
	 * @note Eine eindeutige String-Darstellung ist wichtig,
	 *       damit array_unique() wie erwartet funktioniert.
	 * @return string HTML-Farbcode (#000000 - #FFFFFF), oder rgba(...) wenn Alpha != 1.0
	 */
	public function __toString(): string
	{
		if ($this->a < 1.0) {
			return sprintf('rgba(%d,%d,%d,%s)',
				($this->rgb >> 16) & 0xFF,
				($this->rgb >> 8) & 0xFF,
				$this->rgb & 0xFF,
				$this->a
			);
		} else {
			return sprintf('#%06x', $this->rgb);
		}
	}

	/**
	 * @return int RGB-Farbe als Integer
	 */
	public function getRgb(): int
	{
		return $this->rgb;
	}

	/**
	 * @return float Alpha
	 */
	public function getAlpha(): float
	{
		return $this->a;
	}

	/**
	 * @return bool
	 */
	public function isPseudoGrayscale(): bool
	{
		return $this->s < 0.05 || (($this->l < 0.075 || $this->l > 0.925) && ($this->s < 0.125 || $this->s == 1));
	}

	/**
	 * @return array{h:float,s:float,l:float} Alle Komponenten des HSL-Farbwerts: ['h' => 0..360, 's' => 0..1, 'l' => 0..1]
	 */
	public function getHsl(): array
	{
		return [
			'h' => $this->h,
			's' => $this->s,
			'l' => $this->l,
		];
	}

	/**
	 * @return float Farbton in ° (0..360)
	 */
	public function getHue(): float
	{
		return $this->h;
	}

	/**
	 * @return float Sättigung (0..1)
	 */
	public function getSaturation(): float
	{
		return $this->s;
	}

	/**
	 * @return float Helligkeit (0..1)
	 */
	public function getLightness(): float
	{
		return $this->l;
	}

	/**
	 * @param float $alpha
	 * @return $this
	 */
	public function setAlpha(float $alpha): self
	{
		$this->a = (float)min(max(0, $alpha), 1);
		return $this;
	}

	/**
	 * @param float $hue (0..360)
	 * @return $this
	 */
	private function setHue(float $hue): self
	{
		$hue = (float)min(max(0, $hue), 360);

		if ($hue < 0 || $hue > 360) {
			throw new OutOfRangeException('Hue must be in the range 0..360');
		}

		$this->h = $hue;
		$this->rgb = self::hsl2Rgb($this->getHsl());

		return $this;
	}

	/**
	 * @param float $saturation (0..1)
	 * @return $this
	 */
	private function setSaturation(float $saturation): self
	{
		$saturation = (float)min(max(0, $saturation), 1);

		if ($saturation < 0 || $saturation > 1) {
			throw new OutOfRangeException('Saturation value must be in the range 0..1');
		}

		$this->s = $saturation;
		$this->rgb = self::hsl2Rgb($this->getHsl());

		return $this;
	}

	/**
	 * @param float $lightness (0..1)
	 * @return $this
	 */
	private function setLightness(float $lightness): self
	{
		$lightness = (float)min(max(0, $lightness), 1);

		if ($lightness < 0 || $lightness > 1) {
			throw new OutOfRangeException('Lightness value must be in the range 0..1');
		}

		$this->l = $lightness;
		$this->rgb = self::hsl2Rgb($this->getHsl());

		return $this;
	}

	/**
	 * @param float $amount (0..1) Ein Wert zwischen 0, keine Änderung, und 1, voll abgedunkelt.
	 * @return $this
	 */
	public function darken(float $amount): self
	{
		$amount = (float)min(max(0, $amount), 1);

		$lightnessDecrease = $amount * $this->getLightness();
		$saturation = $this->isPseudoGrayscale()
			? ($this->getSaturation() == 1 ? 0 : min($this->getSaturation(), 0.05))
			: $this->getSaturation() + $amount * ((1.0 - $this->getSaturation()) / 2);

		$this->setLightness($this->getLightness() - $lightnessDecrease);
		$this->setSaturation($saturation);

		return $this;
	}

	/**
	 * @param float $amount (0..1) Ein Wert zwischen 0, keine Änderung, und 1, voll erhellt.
	 * @return $this
	 */
	public function lighten(float $amount): self
	{
		$amount = (float)min(max(0, $amount), 1);

		$lightnessIncrease = $amount * (1.0 - $this->getLightness());
		$saturation = $this->isPseudoGrayscale()
			? ($this->getSaturation() == 1 ? 0 : min($this->getSaturation(), 0.05))
			: $this->getSaturation() - $amount * ($this->getSaturation() / 2);

		$this->setLightness($this->getLightness() + $lightnessIncrease);
		$this->setSaturation($saturation);

		return $this;
	}

	/**
	 * @param self $tint
	 * @return $this
	 */
	public function tint(self $tint): self
	{
		$this->setHue($tint->getHue());
		$this->setSaturation($tint->getSaturation());

		return $this;
	}

	/**
	 * @return float The relative brightness of this color in the sRGB colorspace. (0..1)
	 * @see https://www.w3.org/TR/WCAG20/#relativeluminancedef
	 */
	public function relativeLuminance(): float
	{
		$rgb = array_map(function (int $component) {
			$component /= 255;
			return $component <= 0.03928 ? $component / 12.92 : (($component + 0.055) / 1.055) ** 2.4;
		}, array_combine(['r', 'g', 'b'], sscanf(sprintf('%06x', $this->rgb), '%2x%2x%2x')));

		return 0.2126 * $rgb['r'] + 0.7152 * $rgb['g'] + 0.0722 * $rgb['b'];
	}

	/**
	 * @param self $color
	 * @return float The contrast ratio between `$this` and `$color`. (1..21)
	 * @see https://www.w3.org/WAI/GL/wiki/Contrast_ratio
	 */
	public function contrastRatio(self $color): float
	{
		$l1 = max($this->relativeLuminance(), $color->relativeLuminance());
		$l2 = min($this->relativeLuminance(), $color->relativeLuminance());
		return ($l1 + 0.05) / ($l2 + 0.05);
	}

	/**
	 * Diese Methode wandelt einen RGB-Farbwert in entsprechende HSL-Komponenten um.
	 *
	 * @param int $rgb 3-Byte RGB-Farbwert
	 * @return array Alle Komponenten des HSL-Farbwerts: ['h' => 0..360, 's' => 0..1, 'l' => 0..1]
	 */
	public static function rgb2Hsl(int $rgb): array
	{
		$r = ($rgb >> 16 & 0xff) / 255;
		$g = ($rgb >> 8 & 0xff) / 255;
		$b = ($rgb & 0xff) / 255;

		$max = max($r, $g, $b);
		$min = min($r, $g, $b);

		if ($max == $min) {
			$h = 0;
		}
		else if ($max == $r) {
			$h = 60 * (0 + ($g - $b) / ($max - $min));
		}
		else if ($max == $g) {
			$h = 60 * (2 + ($b - $r) / ($max - $min));
		}
		else if ($max == $b) {
			$h = 60 * (4 + ($r - $g) / ($max - $min));
		}
		else {
			throw new LogicException('fatal error while converting color value');
		}

		if ($h < 0) {
			$h += 360;
		}

		$s = $max == 0 || $min == 1 ? 0 : ($max - $min) / (1 - abs($max + $min - 1));

		$l = ($max + $min) / 2;

		return [
			'h' => $h,
			's' => $s,
			'l' => $l,
		];
	}

	/**
	 * Diese Methode wandelt einen HSL-kodierten Farbwert in den entsprechenden RGB-Farbwert um.
	 *
	 * Implementiert nach: https://en.wikipedia.org/wiki/HSL_and_HSV#From_HSL
	 *
	 * @param array $hsl ['h' => 0.f..360, 's' => 0..1, 'l' => 0..1]
	 * @return int 3-Byte RGB-Farbwert
	 */
	private static function hsl2Rgb(array $hsl): int
	{
		$h = max(0, min($hsl['h'], 360));
		$s = max(0, min($hsl['s'], 1));
		$l = max(0, min($hsl['l'], 1));

		$c = (1 - abs(2 * $l - 1)) * $s;

		$h_ = $h / 60;

		// Absolut wichtig: nicht den normalen Integer-Modulo verwenden => 5.123 % 2 = 1; fmod(5.123, 2) = 1.123
		$x = $c * (1 - abs(fmod($h_, 2) - 1));

		if ($h_ >= 0 && $h_ <= 1) {
			list($r1, $g1, $b1) = [$c, $x, 0];
		}
		else if ($h_ > 1 && $h_ <= 2) {
			list($r1, $g1, $b1) = [$x, $c, 0];
		}
		else if ($h_ > 2 && $h_ <= 3) {
			list($r1, $g1, $b1) = [0, $c, $x];
		}
		else if ($h_ > 3 && $h_ <= 4) {
			list($r1, $g1, $b1) = [0, $x, $c];
		}
		else if ($h_ > 4 && $h_ <= 5) {
			list($r1, $g1, $b1) = [$x, 0, $c];
		}
		else if ($h_ > 5 && $h_ <= 6) {
			list($r1, $g1, $b1) = [$c, 0, $x];
		}
		else {
			list($r1, $g1, $b1) = [0, 0, 0];
		}

		$m = $l - $c / 2;

		list($r, $g, $b) = [
			(int)round(($r1 + $m) * 255),
			(int)round(($g1 + $m) * 255),
			(int)round(($b1 + $m) * 255),
		];

		return ($r << 16 & 0xff0000) + ($g << 8 & 0xff00) + ($b & 0xff);
	}
}
