<?php

namespace App\Component\Config;

use App\Exception\ConfigNotWritableException;
use App\Exception\ConfigNotReadableException;

interface RepositoryInterface
{
	/**
	 * @return array
	 */
	public function toArray(): array;

	/**
	 * @return static[]
	 * @param array $filter MongoDB-Query-Filter
	 */
	public function toArrayFiltered(array $filter): array;

	/**
	 * @return int
	 */
	public function count(): int;

	/**
	 * @return bool
	 */
	public function isEmpty(): bool;

	/**
	 * Daten aus Repository holen. Schlüssel werden im $identifier durch einen Punkt separiert.
	 * @param string $identifier
	 * @param mixed $fallback
	 * @return mixed
	 */
	public function get(string $identifier, $fallback = null);

	/**
	 * Daten in Repository speichern. Schlüssel werden im $identifier durch einen Punkt separiert.
	 * @param string $identifier
	 * @param mixed $data
	 * @return $this
	 */
	public function set(string $identifier, $data);

	/**
	 * Binäre Daten in Repository speichern. Schlüssel werden im $identifier durch einen Punkt separiert.
	 * @param string $identifier
	 * @param string $data
	 * @return $this
	 */
	public function setBinary(string $identifier, $data);

	/**
	 * Ein Zahlen-Datum erhöhen. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * Negative Werte können zum verringern verwendet werden. Wenn der Eintrag nicht existiert,
	 * wird er so behandelt als würde er 0 enthalten.
	 *
	 * @param string $identifier
	 * @param integer|float $amount
	 * @return $this
	 */
	public function increment(string $identifier, $amount=1);

	/**
	 * Ein Element an ein Array anhängen. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * Wenn der Eintrag nicht existiert, wird er so behandelt als wäre es ein leeres Array.
	 *
	 * @param string $identifier
	 * @param mixed $item
	 * @param bool $noDuplicate Duplikate vermeiden
	 * @return $this
	 */
	public function appendToArray(string $identifier, $item, bool $noDuplicate=true);

	/**
	 * Ein Element aus einem Array löschen, falls vorhanden. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * @param string $identifier
	 * @param mixed $item
	 * @return $this
	 */
	public function removeFromArray(string $identifier, $item);

	/**
	 * Daten in Repository speichern, falls ein Eintrag fehlt (oder NULL ist). Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * @param string $identifier
	 * @param mixed $value
	 * @return $this
	 */
	public function setIfMissing(string $identifier, $value);

	/**
	 * Diese Methode entfernt einen Datenstrang.
	 * @param string $identifier
	 * @return $this
	 * @see RepositoryInterface::flush()
	 */
	public function unset(string $identifier);

	/**
	 * Leeren des Repositorys.
	 * @see RepositoryInterface::flush()
	 */
	public function clear(): void;

	/**
	 * Committen aller Änderungen im Repository.
	 * @throws ConfigNotWritableException
	 */
	public function flush(): void;

	/**
	 * Erneutes Einlesen des Repository von der Festplatte. Sämtliche Änderungen gehen verloren.
	 * @param bool $forWrite
	 * @throws ConfigNotReadableException
	 */
	public function reload(bool $forWrite=false): void;

	/**
	 * Gibt an, ob eine Vorlagendatei beim Öffnen des Repository eingelesen wurde.
	 * @return bool
	 */
	public function isTemplateLoaded(): bool;

	/**
	 * @return bool
	 */
	public function isWriteLocked(): bool;

	/**
	 * @return bool
	 */
	public function isDirty(): bool;

	/**
	 * Sperren des Repositories zum Schreiben.
	 *
	 * Danach soll bis zu flush() oder reload() kein parallel laufender
	 * Prozess zugreifen können.
	 */
	public function beginWrite(): void;

	/**
	 * Diese Methode generiert eine ID, die zur Erzeugung eines neuen Objekts
	 * verwendet werden kann.
	 *
	 * Das Repository wird zum Schreiben gesperrt; flush() oder reload() zur
	 * darauffolgenden Freigabe verwenden.
	 *
	 * Ist das Repository `volatile`, so wird eine temporäre ID generiert.
	 * @return string
	 */
	public function generateUniqueObjectId(): string;

	/**
	 * Diese Methode reserviert eine ID, die zur Erzeugung eines neuen Objekts
	 * verwendet werden kann.
	 *
	 * Das Repository wird zum Schreiben gesperrt; flush() oder reload() zur
	 * darauffolgenden Freigabe verwenden.
	 *
	 * Ist das Repository `volatile`, wird es nicht gesperrt und die kann ID
	 * verwendet werden.
	 *
	 * @param string $id
	 * @return void
	 */
	public function reserveObjectId(string $id): void;

	/**
	 * @return bool
	 */
	public function isVolatile(): bool;

	/**
	 * Garantiere keinen weiteren Schreibzugriff auf das Repository.
	 *
	 * Es können weiterhin Modelinstanzen erzeugt und auch Setter verwendet werden,
	 * allerdings sind diese Änderungen nicht persistierbar.
	 *
	 * Verwende reload(), um diese Maßnahme aufzuheben.
	 */
	public function makeVolatile(): void;

	/**
	 * Diese Methode veranlasst das Repository, nur bestimmte Dokumentfelder zu laden.
	 *
	 * Schreibvorgänge sind nach dem Aufruf dieser Methode auf dem Repository nicht mehr
	 * zulässig, da ein Risiko auf Datenverlust bzw. unvollständige Dokumente besteht.
	 *
	 * Verwende reload(), um das Repository zurückzusetzen.
	 *
	 * @param array<string,bool> $projection
	 * @return void
	 */
	public function setProjection(array $projection): void;

	/**
	 * Größe des Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes (-1 wenn nicht vom Repository unterstützt)
	 */
	public function getPersistedSize(): int;

	/**
	 * @return int
	 */
	public function getLastModifiedTimestamp(): int;

	/**
	 * Escaping von Schlüsseln.
	 *
	 * @param string $key
	 * @return string
	 */
	public static function escapeKey($key): string;

	/**
	 * Unescaping von Schlüsseln.
	 *
	 * @param string $key
	 * @return string
	 * @see RepositoryInterface::escapeKey()
	 */
	public static function unescapeKey(string $key): string;
}
