<?php

namespace App\Component\Config\Store;

use App\Component\Config\RepositoryInterface;
use App\Component\LogStream\QueryFilterIterator;
use App\Exception\ConfigNotWritableException;
use App\Exception\ConfigNotReadableException;
use App\Exception\UniqueConstraintException;
use App\Utils;
use ArrayIterator;
use InvalidArgumentException;
use LogicException;

class JsonConfig implements RepositoryInterface
{
	/** @var string $filename */
	private $filename;

	/** @var array $data */
	private $data;

	/** @var array $cache */
	private $cache;

	/** @var bool $withLocking */
	private $withLocking;

	/** @var bool $dirty */
	private $dirty;

	/** @var string|null $templateDir */
	private $templateDir;

	/** @var bool $templateLoaded */
	private $templateLoaded;

	/** @var bool $volatile */
	private $volatile;

	/** @var bool $dirty */
	private $writeLocked;

	/** @var ?resource */
	private $handle;

	/** @var array */
	private static $heldLocks = [];

	private static $SAFE_LOCK_ORDER = ['cm-config', 'cm-users', 'cm-domains', 'cm-statistic', 'cm-embeddings', 'cm-embedding-assets'];

	/**
	 * Lädt bzw. erstellt eine Konfigurationsdatei relativ zum Projektverzeichnis <i>`var/`</i>.
	 * @param string $filename
	 * @param bool $withLocking Wenn true, wird Locking verwendet, damit keine parallelen Änderungen
	 *                          verloren gehen. Ansonsten wird atomar überschrieben.
	 * @param string $base Basis-Verzeichnis: 'var' oder 'share'
	 * @param string|null $templateDir Vorlagen-Verzeichnis relativ zum Base-Dir, oder NULL wenn keine Vorlage verwendet werden soll.
	 */
	public function __construct(string $filename, bool $withLocking=true, string $base='var', ?string $templateDir='share')
	{
		$this->filename = ($base == 'share' ? Utils::getShareDir()
			: ($base == 'var' ? Utils::getBundledVarDir()
			: (Utils::getBaseDir() . DIRECTORY_SEPARATOR . $base))
		) . DIRECTORY_SEPARATOR . $filename;
		$this->dirty = false;
		$this->templateDir = ($base !== $templateDir) ? $templateDir : null;
		$this->volatile = false;
		$this->writeLocked = false;
		$this->withLocking = $withLocking;

		$this->data = [];
		$this->cache = [];

		// Read template files from share/skeleton/clickskeks/ if available
		if ($templateDir === null && Utils::isClickskeksCloud()) {
			$clickskeksTemplatePath = preg_replace('~/share/skeleton/([^/]+)$~', '/share/skeleton/clickskeks/$1', $this->filename);
			if ($clickskeksTemplatePath != $this->filename && file_exists($clickskeksTemplatePath)) {
				$this->filename = $clickskeksTemplatePath;
			}
		}

		// Wenn die Konfigurationsdatei existiert, wird sie eingelesen, andernfalls wird versucht, sie zu erstellen.
		$this->reload();
	}

	/**
	 * Stellt sicher, dass die Datei entsperrt wird
	 *
	 * @phan-suppress PhanTypeMismatchProperty
	 */
	public function __destruct()
	{
		if (@$this->handle) {
			@$this->unlock();
			@fclose($this->handle);
			$this->handle = null;
		}
	}

	private function logPotentialDeadlock(): void {
		$msg = date('[Y-m-d H:i:s]').' Potential deadlock: ';
		$msg .= implode(', ', array_map(function ($filename, $type) {
			return "{$filename}[{$type}]";
		}, array_keys(self::$heldLocks), array_values(self::$heldLocks)));
		$msg .= "\n";

		$backtrace = array_slice(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 11), 2);
		foreach ($backtrace as $row) {
			$msg .= sprintf("%s%s%s at %s:%s\n", $row['class'] ?? '', $row['type'] ?? '', $row['function'] ?? '?', $row['file'] ?? '?', $row['line'] ?? '?');
		}
		$msg .= "\n";

		file_put_contents(Utils::getBundledVarDir().'/log/deadlock.log', $msg, FILE_APPEND);
	}

	private static function isLockOrderUnsafe(): bool
	{
		$lastIndex = -1;
		foreach (array_keys(self::$heldLocks) as $filepath) {
			$basename = basename($filepath, '.json');
			$index = array_search($basename, self::$SAFE_LOCK_ORDER);
			if ($index === false) {
				$index = 9999;
			}
			if ($index > $lastIndex) {
				$lastIndex = $index;
			} else {
				return true;
			}
		}
		return false;
	}

	private function lockWrite(): void
	{
		if (!empty($_ENV['DEBUG_LOCKS'])) {
			unset(self::$heldLocks[$this->filename]);
			self::$heldLocks[$this->filename] = 'w';
			if (count(self::$heldLocks) > 1 and self::isLockOrderUnsafe()) {
				$this->logPotentialDeadlock();
			}
		}
		if ($this->handle) {
			flock($this->handle, LOCK_EX);
		}
	}

	private function lockRead(): void
	{
		if (!empty($_ENV['DEBUG_LOCKS'])) {
			unset(self::$heldLocks[$this->filename]);
			self::$heldLocks[$this->filename] = 'r';
			if (count(self::$heldLocks) > 1 and self::isLockOrderUnsafe()) {
				$this->logPotentialDeadlock();
			}
		}
		if ($this->handle) {
			flock($this->handle, LOCK_SH);
		}
	}

	private function unlock(): void
	{
		if (!empty($_ENV['DEBUG_LOCKS'])) {
			unset(self::$heldLocks[$this->filename]);
		}
		if ($this->handle) {
			flock($this->handle, LOCK_UN);
		}
	}

	/**
	 * @return array
	 */
	public function toArray(): array
	{
		return $this->data;
	}

	/**
	 * @param array $filter MongoDB-Query-Filter
	 * @return array
	 * @throws InvalidArgumentException Wenn das Query nicht unterstützt wird
	 */
	public function toArrayFiltered(array $filter): array
	{
		return iterator_to_array(new QueryFilterIterator(new ArrayIterator($this->data), $filter));
	}

	/**
	 * Größe des Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes
	 */
	public function getPersistedSize(): int
	{
		return (int)@filesize($this->filename);
	}

	/**
	 * @return int
	 */
	public function getLastModifiedTimestamp(): int
	{
		return (int)@filemtime($this->filename);
	}

	/**
	 * @return int
	 */
	public function count(): int
	{
		return count($this->toArray());
	}

	/**
	 * @return bool
	 */
	public function isEmpty(): bool
	{
		return $this->count() === 0;
	}

	/**
	 * Diese Methode liest Daten aus der Konfiguration. Schlüssel werden im $identifier durch einen Punkt separiert.
	 * @param string $identifier
	 * @param mixed $fallback
	 * @return mixed
	 * @phan-side-effect-free
	 */
	public function get(string $identifier, $fallback = null)
	{
		if (isset($this->cache[$identifier])) {
			return $this->cache[$identifier];
		}

		$nodeNames = array_values(array_filter(explode('.', $identifier), function ($str) { return strlen($str) > 0; }));
		if (count($nodeNames) == 0) {
			return $fallback;
		}

		$value = $fallback;

		try {
			$value = array_reduce($nodeNames, function ($node, $key) use ($identifier) {
				if (is_array($node) && array_key_exists($key, $node)) {
					return $node[$key];
				}
				else {
					throw new InvalidArgumentException("Config `$identifier` not found.");
				}
			}, $this->data);
		} catch (InvalidArgumentException $exception) {}

		$this->cache[$identifier] = $value;

		return $value;
	}

	/**
	 * Diese Methode schreibt Daten in die Konfiguration. Schlüssel werden im $identifier durch einen Punkt separiert.
	 * @param string $identifier
	 * @param mixed $data
	 * @return $this
	 */
	public function set(string $identifier, $data): self
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}

		$nodeNames = array_values(array_filter(explode('.', $identifier), function ($str) { return strlen($str) > 0; }));
		if (count($nodeNames) == 0) {
			return $this;
		}

		$node = &$this->data;

		foreach ($nodeNames as $key => $childNode) {
			$isLast = $key == count($nodeNames) - 1;

			if (array_key_exists($childNode, $node) == false || ($isLast == false && is_array($node[$childNode]) == false)) {
				$node[$childNode] = [];
			}

			$tmpRef = &$node[$childNode];
			unset($node);

			$node = &$tmpRef;
			unset($tmpRef);
		}

		$hasChanged = is_array($node) || is_object($node) || $node === null || $node !== $data;

		if ($hasChanged) {
			$node = $data;
			$this->cache[$identifier] = $data;
			$this->dirty = true;

			// Obere Knotenpunkte im Cache invalidieren
			for ($i = count($nodeNames) - 1; $i > 0; $i--) {
				unset($this->cache[implode('.', array_slice($nodeNames, 0, $i))]);
			}

			// Untere Knotenpunkte im Cache invalidieren
			$prefix = $identifier.'.';
			$this->cache = array_filter($this->cache, function ($cacheIdentifier) use ($prefix) {
				return (strlen($cacheIdentifier) < strlen($prefix) or strpos($cacheIdentifier, $prefix) !== 0);
			}, ARRAY_FILTER_USE_KEY);
		}

		return $this;
	}

	/**
	 * @param string $identifier
	 * @param string $data
	 * @return $this
	 */
	public function setBinary(string $identifier, $data): self
	{
		return $this->set($identifier, $data);
	}

	/**
	 * Ein Zahlen-Datum erhöhen. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * Negative Werte können zum verringern verwendet werden. Wenn der Eintrag nicht existiert,
	 * wird er so behandelt als würde er 0 enthalten.
	 *
	 * @param string $identifier
	 * @param integer|float $amount
	 * @return $this
	 */
	public function increment(string $identifier, $amount=1)
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}
		$value = $this->get($identifier, 0);
		$this->set($identifier, $value + $amount);
		return $this;
	}

	/**
	 * Ein Element an ein Array anhängen. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * Wenn der Eintrag nicht existiert, wird er so behandelt als wäre es ein leeres Array.
	 *
	 * @param string $identifier
	 * @param mixed $item
	 * @param bool $noDuplicate Duplikate vermeiden
	 * @return $this
	 */
	public function appendToArray(string $identifier, $item, bool $noDuplicate=true)
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}
		$array = $this->get($identifier, []);
		if (!$noDuplicate or !in_array($item, $array, true)) {
			$array[] = $item;
			$this->set($identifier, $array);
		}
		return $this;
	}

	/**
	 * Ein Element aus einem Array löschen, falls vorhanden. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * @param string $identifier
	 * @param mixed $item
	 * @return $this
	 */
	public function removeFromArray(string $identifier, $item)
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}
		$array = $this->get($identifier, []);
		if ($array) {
			$count = count($array);
			$array = array_filter($array, function ($v) use ($item) {
				return $v !== $item;
			});
			if (count($array) !== $count) {
				$this->set($identifier, $array);
			}
		}
		return $this;
	}

	/**
	 * Daten in Repository speichern, falls ein Eintrag fehlt (oder NULL ist). Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * @param string $identifier
	 * @param mixed $value
	 * @return $this
	 */
	public function setIfMissing(string $identifier, $value)
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}
		$oldValue = $this->get($identifier, null);
		if ($oldValue === null) {
			$this->set($identifier, $value);
		}
		return $this;
	}

	/**
	 * Diese Methode entfernt einen Konfigurationsstrang. Die Änderung muss mit JsonConfig::flush() verewigt werden.
	 * @param string $identifier
	 * @return $this
	 * @see JsonConfig::flush()
	 */
	public function unset(string $identifier): self
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}

		$nodeNames = array_values(array_filter(explode('.', $identifier), function ($str) { return strlen($str) > 0; }));
		if (count($nodeNames) == 0) {
			return $this;
		}

		$node = &$this->data;
		foreach ($nodeNames as $key => $nodeName) {
			// Datenzweig entfernen
			if ($key + 1 == count($nodeNames)) {
				unset($node[$nodeName]);
				unset($this->cache[$identifier]);
				$this->dirty = true;
				break;
			}
			// Wenn die definierte Datenstruktur nicht existiert, muss nichts gemacht werden
			else if (array_key_exists($nodeName, $node) == false || is_array($node[$nodeName]) == false) {
				break;
			}

			$tmpRef = &$node[$nodeName];
			unset($node);

			$node = &$tmpRef;
			unset($tmpRef);
		}

		return $this;
	}

	/**
	 * Diese Methode leert die gesamte Konfiguration, muss jedoch noch mit JsonConfig::flush() geschrieben werden.
	 * @see JsonConfig::flush()
	 */
	public function clear(): void
	{
		$this->beginWrite();
		$this->data = [];
		$this->cache = [];
		$this->dirty = true;
	}

	/**
	 * Diese Methode schreibt alle an der Konfiguration vorgenommenen Änderungen auf die Platte.
	 * @throws ConfigNotWritableException
	 */
	public function flush(): void
	{
		$f = $this->handle;

		if ($this->isVolatile()) {
			throw new LogicException('Config is volatile, write access not allowed.');
		}
		elseif ($this->dirty == false) {
			$this->unlock();
			$this->writeLocked = false;
			return;
		}
		assert($f !== null);

		// Schreiben mit Locking:
		if ($this->withLocking) {
			if (!$f) {
				return;
			}

			// Sicher gehen, dass wir einen exklusiven Lock haben
			if (!$this->writeLocked) {
				$this->lockWrite();
			}

			$filename = $this->filename;
			$data     = json_encode($this->data);

			fseek($f, 0);
			if (@fwrite($f, $data) < strlen($data)) {
				throw new ConfigNotWritableException($filename);
			}
			ftruncate($f, ftell($f));

			// Datei wieder entsperren
			$this->unlock();
			$this->dirty = false;
			$this->writeLocked = false;
		}
		// Schreiben und atomar umbenennen:
		else {
			$filename    = $this->filename;
			$tmpfilename = $filename.'~'.bin2hex(random_bytes(8));
			$data        = json_encode($this->data);

			if (@file_put_contents($tmpfilename, $data) === false) {
				throw new ConfigNotWritableException($tmpfilename);
			}

			if (@rename($tmpfilename, $filename) === false) {
				throw new ConfigNotWritableException($filename);
			}

			// Falls Locks vorhanden, entsperren
			if ($f) {
				$this->unlock();
			}

			$this->handle = @fopen($filename, 'rb');
			$this->dirty = false;
			$this->writeLocked = false;
		}
	}

	/**
	 * Diese Methode liest die Konfiguration erneut von der Festplatte ein. Sämtliche Änderungen gehen verloren.
	 *
	 * @param bool $forWrite Ob die Datei zum Schreiben gesperrt werden soll
	 *
	 * @throws ConfigNotReadableException
	 */
	public function reload(bool $forWrite=false): void
	{
		// Dateihandle entsperren und schließen wenn offen
		$f = $this->handle;
		if ($f) {
			$this->unlock();
			fclose($f);
		}
		$this->writeLocked = false;

		// Eine Vorlagendatei wird nur bei Bedarf und Vorhandensein eingelesen
		$this->templateLoaded = false;

		// Datei öffnen
		$f = $this->handle = @fopen($this->filename, ($forWrite ? 'rb+' : 'rb'));
		if ($f) {
			if ($this->withLocking) {
				if ($forWrite) {
					$this->lockWrite();
				} else {
					$this->lockRead();
				}
			}
			$data = json_decode(@stream_get_contents($f), true);
			$this->dirty = false;
			$this->writeLocked = $forWrite;
			if (!$forWrite) {
				$this->unlock();
			}
		}
		else {
			$data = null;
			$error = error_get_last()['message'];
			if (is_file($this->filename)) {
				if ($forWrite) {
					throw new ConfigNotWritableException($this->filename);
				}
				else {
					throw new ConfigNotReadableException($this->filename);
				}
			}
			// Wenn keine Daten da sind, aber eine Vorlage existieren könnte
			if ($this->templateDir !== null) {

				// Bestimmen, ob eine Vorlagedatei in share/ liegt
				$pathName = $this->filename;
				$varPrefix = Utils::getBundledVarDir() . DIRECTORY_SEPARATOR;
				$sharePrefix = Utils::getBaseDir() . DIRECTORY_SEPARATOR . $this->templateDir . DIRECTORY_SEPARATOR;
				if (strpos($pathName, $varPrefix) === 0) {
					// User-/Domain-spezifisches liegt im `skeleton`-Unterverzeichnis
					$templatePath = $sharePrefix . preg_replace('%^user_data/[a-f0-9]{7}/(domains/[a-f0-9]{7}/)?%', 'skeleton/', substr($pathName, strlen($varPrefix)));

					// Read template files from share/skeleton/clickskeks/ if available
					if (Utils::isClickskeksCloud()) {
						$clickskeksTemplatePath = preg_replace('~/skeleton/([^/]+)$~', '/skeleton/clickskeks/$1', $templatePath);
						if ($clickskeksTemplatePath != $templatePath && file_exists($clickskeksTemplatePath)) {
							$templatePath = $clickskeksTemplatePath;
						}
					}
				}
				else {
					$templatePath = null;
				}

				// Datei erstellen, falls sie fehlt, ggf. mit Daten aus Vorlage füllen
				if ($this->withLocking) {
					$f = $this->handle = @fopen($this->filename, 'cb+');
					if (!$f) {
						throw new ConfigNotWritableException($this->filename);
					}
					$this->lockWrite();
					$data = json_decode(@stream_get_contents($f), true);

					if ($data === null) {
						$data = ($templatePath and file_exists($templatePath)) ? json_decode(file_get_contents($templatePath), true) : null;
					}

					$this->data = is_array($data) ? $data : [];
					$this->dirty = true;
					$this->writeLocked = true;
					if ($forWrite == false) {
						$this->flush();
					}
				}
				else {
					$data = ($templatePath and file_exists($templatePath)) ? json_decode(file_get_contents($templatePath), true) : null;
					$this->dirty = true;
				}

				// Festhalten, dass eine Vorlagendatei eingelesen wurde
				$this->templateLoaded = (bool)$data;
			}
			// Wenn es keine Vorlage gibt, aber geschrieben werden soll
			elseif ($forWrite) {
				// Datei erstellen, falls sie fehlt
				if ($this->withLocking) {
					$f = $this->handle = @fopen($this->filename, 'cb+');
					if (!$f) {
						throw new ConfigNotWritableException($this->filename);
					}
					$this->lockWrite();
					$data = json_decode(@stream_get_contents($f), true);
					$this->dirty = true;
					$this->writeLocked = true;
				}
				else {
					$data = null;
					$this->dirty = true;
				}
			}
		}
		$this->data = is_array($data) ? $data : [];
		$this->cache = [];
		$this->volatile = false;
	}

	public function isTemplateLoaded(): bool
	{
		return $this->templateLoaded;
	}

	public function isWriteLocked(): bool
	{
		return $this->writeLocked;
	}

	/**
	 * @return bool
	 */
	public function isDirty(): bool
	{
		return $this->dirty;
	}

	/**
	 * Sperrt die Konfiguration zum Schreiben, wenn Locking aktiviert.
	 *
	 * Danach kann bis zu flush() oder reload() kein parallel laufender
	 * Prozess auf die Datei zugreifen.
	 */
	public function beginWrite(): void
	{
		if ($this->isVolatile()) {
			throw new LogicException('Config is volatile, write access not allowed.');
		}
		if ($this->withLocking and $this->writeLocked == false) {
			$this->reload(true);
		}
	}

	/**
	 * @return string
	 */
	public function generateUniqueObjectId(): string
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}

		$id = null;
		do {
			$id = substr(hash('sha256', uniqid()), 0, 7);
		} while (isset($this->data[$id]));

		return $id;
	}

	/**
	 * @param string $id
	 * @return void
	 * @throws InvalidArgumentException if a parsing error occurs, e.g. invalid ID format.
	 * @throws UniqueConstraintException if the ID is already taken.
	 */
	public function reserveObjectId(string $id): void
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}

		$pattern = '/^[a-f0-9]{7}$/';
		if (preg_match($pattern, $id) == false) {
			throw new InvalidArgumentException('Invalid id, use the following format: '.$pattern);
		}
		elseif (isset($this->data[$id])) {
			throw new UniqueConstraintException("Object ID \"$id\" is already in use.");
		}
	}

	/**
	 * @return bool
	 */
	public function isVolatile(): bool
	{
		return $this->volatile;
	}

	/**
	 * Nachdem diese Methode aufgerufen wurde, kann das Repository nicht
	 * mehr gespeichert werden. Dieses Verhalten kann nur durch Neuladen
	 * des Repos umgangen werden.
	 */
	public function makeVolatile(): void
	{
		// Lock aufheben
		if ($this->handle) {
			$this->unlock();
			fclose($this->handle);
			$this->handle = null;
			$this->writeLocked = false;
		}
		$this->volatile = true;
	}

	/**
	 * @param array<string,bool> $projection
	 * @return void
	 */
	public function setProjection(array $projection): void
	{
		$this->makeVolatile();
		// Ignoring $projection for JSON store.
	}

	/**
	 * Escaping von Schlüsseln.
	 *
	 * Verhindert das . als Trenner für Unterelemente interpretiert wird.
	 *
	 * @param string $key
	 * @return string
	 */
	public static function escapeKey($key): string
	{
		$key = str_replace('%', '%25', (string)$key);
		$key = str_replace("\0", '%00', $key);
		$key = str_replace('.', '%2E', $key);
		return $key;
	}

	/**
	 * Unescaping von Schlüsseln.
	 *
	 * @param string $key
	 * @return string
	 * @see JsonConfig::escapeKey()
	 */
	public static function unescapeKey(string $key): string
	{
		$key = str_replace('%2E', '.', $key);
		$key = str_replace('%00', "\0", $key);
		$key = str_replace('%25', '%', $key);
		return $key;
	}
}
