<?php

namespace App\Component\Config\Store;

use App\Component\Config\RepositoryInterface;
use App\Component\MongoDbService;
use App\Component\LogStream\QueryFilterIterator;
use App\Exception\ConfigNotReadableException;
use App\Exception\ConfigNotWritableException;
use App\Exception\UniqueConstraintException;
use App\Model\Domain;
use App\Model\User;
use ArrayObject;
use MongoDB\BSON\Binary;
use MongoDB\BSON\UTCDateTime;
use MongoDB\BSON\ObjectId;
use MongoDB\Collection;
use MongoDB\Driver\Exception\BulkWriteException;
use MongoDB\Driver\Exception\CommandException;
use MongoDB\Driver\Session;
use MongoDB\Driver\Exception\InvalidArgumentException;
use MongoDB\Driver\ReadConcern;
use MongoDB\Driver\ReadPreference;
use MongoDB\Driver\WriteConcern;
use MongoDB\Driver\WriteConcernError;
use MongoDB\Driver\WriteError;
use MongoDB\Model\BSONDocument;
use ArrayIterator;

class MongoDb implements RepositoryInterface
{
	private const ERROR_DUPLICATE_KEY = 11000;

	/**
	 * Fehlercodes: https://github.com/mongodb/mongo/blob/master/src/mongo/base/error_codes.yml
	 */
	private const TRANSIENT_ERROR_CODES = [46, 64, 107, 109, 112, 117, 208, 216, 226];

	// Diese Collections werden nicht auf anderer Ebene (System/User/Domain) wiederverwendet.
	// Bei systemweiten Collections bedeutet dass, dass sie ohne
	// ['_domainId' => ['$exists' => false], '_userId' => ['$exists' => false]]
	// abgefragt werden können.
	private const EXCLUSIVE_COLLECTIONS = [ 'cm-users', 'cm-config', 'cm-versionlog' ];

	/** @var MongoDbService $service */
	private static $service;

	/** @var Session|null $session */
	private $session = null;

	/** @var Collection $collection */
	private $collection;

	/** @var User|null $user */
	private $user;
	/** @var Domain|null $domain */
	private $domain;

	/** @var array $data */
	private $data;
	/** @var array $cache */
	private $cache;

	/** @var bool $customIds */
	private $customIds;
	/** @var array $customIdsMap */
	private $customIdsMap;

	/** @var bool $dirty */
	private $dirty;

	/** @var array<string,array> $dirtyDocuments Liste der Änderungen pro Dokument */
	private $dirtyDocuments;

	/** @var string|null $templateDir */
	private $templateDir;

	/** @var bool $templateLoaded */
	private $templateLoaded;

	/** @var bool $volatile */
	private $volatile;

	/** @var array<string,int> $projection */
	private $projection;

	/** @var bool $writeLocked */
	private $writeLocked;

	/** @var bool $isExclusiveCollection Wenn TRUE, wird diese Collection nie für Daten auf einer anderen Ebene (System/User/Domain) wiederverwendet */
	private $isExclusiveCollection = false;

	/** @var array|null $currentDataFilter */
	private $currentDataFilter = null;

	/** @var bool $wantsMtimeIndex */
	private $wantsMtimeIndex;

	/**
	 * MongoDb constructor.
	 * @throws ConfigNotReadableException
	 * @param string $collectionName
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @param bool $customIds Ob nicht-zufällige IDs verwendet werden
	 * @param string|null $templateDir Vorlagen-Verzeichnis relativ zum Base-Dir, oder NULL wenn keine Vorlage verwendet werden soll.
	 * @param bool $wantsMtimeIndex Ob ein Index auf _mtime angelegt werden soll, um performant den Zeitstempel der letzten Änderung bestimmen zu können
	 */
	public function __construct(string $collectionName, ?User $user=null, ?Domain $domain=null, bool $customIds=false, ?string $templateDir='share', bool $wantsMtimeIndex=false)
	{
		try {
			$this->collection = self::getService()->getDatabase()->selectCollection($collectionName);
		} catch (\Exception $exception) {
			throw new ConfigNotReadableException($exception->getMessage());
		}

		$this->customIds = $customIds;
		$this->customIdsMap = [];
		$this->dirty = false;
		$this->dirtyDocuments = [];
		$this->templateDir = $templateDir;
		$this->volatile = false;
		$this->projection = [];
		$this->writeLocked = false;

		$this->data = [];
		$this->cache = [];

		$this->user = $user;
		$this->domain = $domain;
		$this->isExclusiveCollection = in_array($collectionName, self::EXCLUSIVE_COLLECTIONS);
		$this->wantsMtimeIndex = $wantsMtimeIndex;

		// Collection überprüfen und u. U. mit Daten aus share/skeleton/ vorbefüllen
		$this->reload();
	}

	public function __destruct()
	{
		// TODO: "Ungespeicherte" Dokumente löschen; möglicherweise obsolet, wenn Transaktionen (Commit,Rollback) implementiert wird
		if ($this->session and $this->session->isInTransaction()) {
			$this->session->abortTransaction();
		}
	}

	/**
	 * @return MongoDbService
	 * @throws ConfigNotReadableException
	 */
	private static function getService(): MongoDbService
	{
		if (!self::$service) {
			self::$service = MongoDbService::getInstance();
		}

		return self::$service;
	}

	/**
	 * @return Session
	 * @throws ConfigNotReadableException
	 */
	private function getSession(): Session
	{
		if (!$this->session) {
			try {
				$this->session = self::getService()->getClient()->startSession();
			} catch (\Exception $exception) {
				throw new ConfigNotReadableException($exception->getMessage());
			}
		}

		return $this->session;
	}

	/**
	 * Diese Methode entpackt ein Dokument das aus der MongoDB kommt. Dabei
	 * werden alle Objekte, BSONDocument, BSONArray etc., in PHP-native Arrays
	 * umgewandelt.
	 *
	 * Das Ergebnis ist zur weiteren Verwaltung in $this->data gedacht.
	 * @param ArrayObject<string,mixed> $document
	 * @return array
	 */
	private static function unpackDocument(ArrayObject $document): array
	{
		$data = [];

		foreach ((array)$document as $field => $value) {
			if (is_object($value)) {
				if ($value instanceof ArrayObject) {
					$value = self::unpackDocument($value);
				}
				elseif ($value instanceof ObjectId) {
					$value = (string)$value;
				}
			}

			$data[$field] = $value;
		}

		return $data;
	}

	/**
	 * Diese Methode fügt bei einem relevanten Repository die verknüpfte Domain
	 * oder den verknüpften Benutzer und einen Timestamp den Dokumentdaten hinzu.
	 *
	 * Jedes Dokument sollte hier durchlaufen, bevor es an MongoDB geht.
	 * @param array $document
	 * @return array
	 */
	private function prepareDocumentContext(array $document): array
	{
		unset($document['_domainId'], $document['_userId']);

		if ($this->domain) {
			$document = ['_domainId'=>new ObjectId($this->domain->getId())] + $document;
		}
		elseif ($this->user) {
			$document = ['_userId'=>new ObjectId($this->user->getId())] + $document;
		}

		$document = ['_mtime'=>new UTCDateTime()] + $document;
		return $document;
	}

	/**
	 * Diese Methode sorgt bei einem relevanten Repository dafür, dass nur
	 * Dokumente berücksichtigt werden, die mit einer bestimmten Domain oder
	 * einem bestimmten Benutzer verknüpft sind.
	 *
	 * Jeder Filter sollte hier durchlaufen, bevor er an MongoDB geht.
	 * @param array $filter
	 * @return array
	 */
	private function prepareFilterContext(array $filter): array
	{
		if ($this->domain) {
			$filter = ['_domainId'=>new ObjectId($this->domain->getId())] + $filter;
		}

		elseif ($this->user) {
			$filter = ['_userId'=>new ObjectId($this->user->getId())] + $filter;
		}

		elseif (! $this->isExclusiveCollection) {
			$filter = [
				'$and' => [
					['_domainId' => ['$exists' => false]],
					['_userId' => ['$exists' => false]],
					$filter,
				],
			];
			if (!$filter['$and'][2]) {
				unset($filter['$and'][2]);
			}
		}

		return $filter;
	}

	/**
	 * Diese Methode lädt alle Dokumente, die zum Filter passen, aus MongoDB in $this->data.
	 * @param array $filter
	 * @throws ConfigNotReadableException
	 */
	private function fetchDocuments(array $filter): void
	{
		$filter = $this->prepareFilterContext($filter);
		if ($this->currentDataFilter === $filter or $this->currentDataFilter === $this->prepareFilterContext([])) {
			// Wenn zuletzt schon genau das selbe oder alles abgefragt wurde, nicht neu laden.
			return;
		}
		$this->currentDataFilter = null;

		try {
			$cursor = $this->collection->find($filter, [
				'projection' => $this->projection,
				'session' => $this->getSession(),
			]);
		} catch (\Exception $exception) {
			throw new ConfigNotReadableException($exception->getMessage());
		}

		foreach ($cursor as $document) {
			/** @var BSONDocument $document */
			if ($this->customIds) {
				$this->customIdsMap[(string)$document['_customId']] = (string)$document['_id'];
				$documentId = (string)$document['_customId'];
			}
			else {
				$documentId = (string)$document['_id'];
			}

			if (isset($this->data[$documentId])) {
				continue;
			}

			// Dokument in PHP-native Datentypen umwandeln
			$data = self::unpackDocument($document);
			unset($data['_id'], $data['_userId'], $data['_domainId'], $data['_customId'], $data['_mtime']);

			$this->data[$documentId] = $data;
		}
		$this->currentDataFilter = $filter;
	}

	/**
	 * Diese Methode lädt ein bestimmtes Dokument in $this->data.
	 * @param string $id
	 * @throws \LogicException
	 * @throws InvalidArgumentException
	 */
	private function fetchDocument(string $id): void
	{
		if ($this->customIds && isset($this->customIdsMap[$id]) || isset($this->data[$id])) {
			return;
		}

		try {
			$filter = $this->customIds ? ['_customId' => $id] : ['_id' => new ObjectId($id)];
		} catch (InvalidArgumentException $exception) {
			return;
		}

		$this->fetchDocuments($filter);
	}

	/**
	 * @return bool
	 */
	public function isTemplateLoaded(): bool
	{
		return $this->templateLoaded;
	}

	/**
	 * @return bool
	 */
	public function isVolatile(): bool
	{
		return $this->volatile;
	}

	/**
	 * @return bool
	 */
	public function isWriteLocked(): bool
	{
		return $this->writeLocked;
	}

	/**
	 * @return bool
	 */
	public function isDirty(): bool
	{
		return $this->dirty;
	}

	public function beginWrite(): void
	{
		if ($this->isVolatile()) {
			throw new \LogicException('Config is volatile, write access not allowed.');
		}
		if ($this->writeLocked == false) {
			$this->reload(true);
		}
	}

	/**
	 * @return string
	 */
	public function generateUniqueObjectId(): string
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
			if (!$this->getSession()->isInTransaction()) {
				$this->getSession()->startTransaction(['readPreference' => new ReadPreference('primary')]);
			}
		}
		try {
			$this->currentDataFilter = null;
			return (string)$this->collection
				->insertOne($this->prepareDocumentContext([]), ['session'=>$this->getSession()])
				->getInsertedId();
		} catch (\Exception $exception) {
			throw new ConfigNotWritableException($exception->getMessage());
		}
	}

	/**
	 * @param string $id
	 * @return void
	 * @throws \InvalidArgumentException if a parsing error occurs, e.g. invalid object ID.
	 * @throws UniqueConstraintException if the object ID is already taken.
	 * @throws ConfigNotWritableException for any other error.
	 */
	public function reserveObjectId(string $id): void
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
			if (!$this->getSession()->isInTransaction()) {
				$this->getSession()->startTransaction(['readPreference' => new ReadPreference('primary')]);
			}
		}

		try {
			$this->currentDataFilter = null;
			$this->collection->insertOne(
				$this->prepareDocumentContext(['_id'=>new ObjectId($id)]),
				['session'=>$this->getSession()]
			);
		} catch (InvalidArgumentException $exception) {
			throw new \InvalidArgumentException($exception->getMessage());
		} catch (BulkWriteException $exception) {
			if ($exception->getCode() == self::ERROR_DUPLICATE_KEY) {
				throw new UniqueConstraintException("Object ID \"$id\" is already in use.");
			}
			else {
				throw new ConfigNotWritableException($exception->getMessage());
			}
		} catch (\Exception $exception) {
			throw new ConfigNotWritableException($exception->getMessage());
		}
	}

	public function clear(): void
	{
		if (!$this->getSession()->isInTransaction()) {
			$this->getSession()->startTransaction(['readPreference' => new ReadPreference('primary')]);
		}
		$this->currentDataFilter = null;
		$this->collection->deleteMany($this->prepareFilterContext([]), ['session'=>$this->getSession()]);

		// Aktion in flush() durchführen
		$this->dirty = true;
	}

	private function executeChanges(string $documentId, ?string $customId, array $changes)
	{
		$filter = $this->prepareFilterContext(['_id' => new ObjectId($documentId)]);
		if ($customId !== null) {
			$filter['_customId'] = $customId;
		}

		$operations = [];
		$currentOperation = [];
		$currentOperationIsRegularUpdate = false;

		foreach ($changes as $change) {
			$op = $change[0];
			$arg1 = $change[1];
			$arg2 = $change[2] ?? null;

			// Standard-Operationen die in einem Update zusammengefasst werden können
			if ($op[0] === '$') {
				$canReuseOperation = $currentOperationIsRegularUpdate;

				if ($currentOperationIsRegularUpdate) {
					// Das gleiche Feld kann nicht in zwei verschiedenen Operatoren auftauchen; dann splitten.
					// @phan-suppress-next-line PhanTypeInvalidDimOffset
					foreach ($currentOperation['updateOne'][1] as $oneOp => $oneArgs) {
						if ($oneOp !== $op) {
							foreach ($oneArgs as $oneKey => $oneValue) {
								if ($oneKey === $arg1 or strpos($oneKey, $arg1.'.') === 0 or strpos($arg1, $oneKey.'.') === 0) {
									$canReuseOperation = false;
									break;
								}
							}
						} else {
							foreach ($oneArgs as $oneKey => $oneValue) {
								if (strpos($oneKey, $arg1.'.') === 0 or strpos($arg1, $oneKey.'.') === 0) {
									$canReuseOperation = false;
									break;
								}
							}
						}
					}
				}

				if (!$canReuseOperation) {
					unset($currentOperation);

					// Sonderbehandlung Array-Operationen
					// Das macht das Mergen von Operationen einfacher
					if ($op === '$pull') {
						$op = '$pullAll';
						$arg2 = [$arg2];
					}
					elseif ($op === '$push' or $op === '$addToSet') {
						$arg2 = ['$each' => [$arg2]];
					}

					$currentOperation = ['updateOne' => [
						$filter, [
							'$max' => ['_mtime'=>new UTCDateTime()],
							$op => [$arg1 => $arg2],
						]
					]];

					$currentOperationIsRegularUpdate = true;
					$operations[] = &$currentOperation;
				} else {
					// Operation hinzufügen
					// @phan-suppress-next-line PhanTypeInvalidDimOffset
					$suboperations = &$currentOperation['updateOne'][1];

					if (!isset($suboperations[$op])) {
						if ($op === '$push' or $op === '$addToSet') {
							$suboperations[$op] = [$arg1 => ['$each' => [$arg2]]];
						}
						elseif ($op === '$pull') {
							$suboperations['$pullAll'] = [$arg1 => [$arg2]];
						}
						else {
							$suboperations[$op] = [$arg1 => $arg2];
						}
					}
					elseif ($op === '$set' or $op === '$unset') {
						$suboperations[$op][$arg1] = $arg2;
					}
					elseif ($op === '$inc') {
						$suboperations[$op][$arg1] = ($suboperations[$op][$arg1] ?? 0) + $arg2;
					}
					elseif ($op === '$push' or $op === '$addToSet') {
						if (isset($suboperations[$op][$arg1]['$each'])) {
							$suboperations[$op][$arg1]['$each'][] = $arg2;
						} else {
							$suboperations[$op][$arg1] = ['$each' => [$arg2]];
						}
					}
					elseif ($op === '$pull') {
						if (isset($suboperations['$pullAll'][$arg1])) {
							$suboperations['$pullAll'][$arg1][] = $arg2;
						} else {
							$suboperations[$op][$arg1] = [$arg2];
						}
					}
					else {
						throw new \LogicException("unknown operation: $op");
					}
				}
				unset($suboperations);
			}
			elseif ($op === 'replace') {
				if ($customId !== null) {
					$arg1['_customId'] = $customId;
				}
				$data = $this->prepareDocumentContext($arg1);
				$operations[] = ['replaceOne' => [$filter, $data]];
				$currentOperationIsRegularUpdate = false;
			}
			elseif ($op === 'setIfMissing') {
				$opFilter = array_merge([$arg1 => null], $filter);
				$operations[] = ['updateOne' => [ $opFilter, [ '$set' => [$arg1 => $arg2], '$max' => ['_mtime'=>new UTCDateTime()] ]]];
				$currentOperationIsRegularUpdate = false;
			}
			elseif ($op === 'insertIfMissing') {
				// Wird schon im Insert-Code behandelt.
				// *Sollte* also niemals erreicht werden -> no-op.
				$currentOperationIsRegularUpdate = false;
			}
			else {
				throw new \LogicException("unknown operation: $op");
			}
		}

		$session = $this->getSession();
		if (count($operations) === 0) {
			return;
		}
		elseif (count($operations) === 1 and isset($operations[0]['updateOne'])) {
			for ($i = 0; $i < 5; ++$i) {
				try {
					$this->collection->updateOne($operations[0]['updateOne'][0], $operations[0]['updateOne'][1], ['session' => $this->getSession()]);
					break;
				} catch (BulkWriteException $e) {
					$result = $e->getWriteResult();
					/** @var WriteError[]|WriteConcernError[] $errors */
					$errors = array_values(array_filter(array_merge($result->getWriteErrors(), [$result->getWriteConcernError()])));

					if (!$errors) {
						break;
					}
					// Neu versuchen, wenn es ein vorübergehender Fehler ist (WriteConflict etc.)
					// Fehlercodes: https://github.com/mongodb/mongo/blob/master/src/mongo/base/error_codes.yml
					elseif ($i < 4 and in_array($errors[0]->getCode(), self::TRANSIENT_ERROR_CODES)) {
						usleep(rand(100, 20000));
						continue;
					}
					else {
						throw $e;
					}
				}
			}
		}
		else {
			for ($i = 0; $i < 10; ++$i) {
				try {
					$this->collection->bulkWrite($operations, ['session' => $this->getSession()]);
					break;
				} catch (BulkWriteException $e) {
					$result = $e->getWriteResult();
					/** @var WriteError[]|WriteConcernError[] $errors */
					$errors = array_values(array_filter(array_merge($result->getWriteErrors(), [$result->getWriteConcernError()])));

					if (!$errors) {
						break;
					}

					// Die bereits erfolgreich abgeschlossenen Operationen aus der Liste entfernen
					$index = $errors[0]->getIndex();
					if ($index !== 0) {
						$operations = array_slice($operations, $errors[0]->getIndex());
					}

					// Neu versuchen, wenn es ein vorübergehender Fehler ist (WriteConflict etc.)
					// Fehlercodes: https://github.com/mongodb/mongo/blob/master/src/mongo/base/error_codes.yml
					elseif ($i < 4 and in_array($errors[0]->getCode(), self::TRANSIENT_ERROR_CODES)) {
						usleep(rand(100, 20000));
						continue;
					}
					else {
						throw $e;
					}
				}
			}
		}
	}

	public function flush(): void
	{
		if ($this->isVolatile()) {
			throw new \LogicException('Config is volatile, write access not allowed.');
		}
		elseif ($this->dirty == false) {
			$this->writeLocked = false;
			return;
		}

		$session = $this->getSession();
		$this->currentDataFilter = null;

		if ($this->getSession()->isInTransaction()) {
			try {
				$session->commitTransaction();
			}
			catch (CommandException $e) {
				if ($e->hasErrorLabel('UnknownTransactionCommitResult')) {
					// Retry commit
					usleep(rand(100, 20000));
					$session->commitTransaction();
				}
				else {
					throw $e;
				}
			}
		}

		$dirtyDocuments = array_keys($this->dirtyDocuments);

		if ($this->customIds) {
			// Zunächst werden alle neuen Datensätze in MongoDB eingespielt
			$newCustomIds = array_diff($dirtyDocuments, array_keys($this->customIdsMap));

			// Integer aus numerischem Index in String umwandeln
			$newCustomIds = array_map('strval', $newCustomIds);

			// Neue Dokumente ermitteln und benutzerdefinerte IDs den Daten beifügen
			$newDocuments = array_filter($this->data, function (string $customId) use ($newCustomIds) {
				return in_array($customId, $newCustomIds, true);
			}, ARRAY_FILTER_USE_KEY);
			$newDocuments = array_map(function ($data, $id) {
				return ['_customId'=>(string)$id] + $data;
			}, $newDocuments, array_keys($newDocuments));

			// Reihenfolge zwecks Zuordnung "CustomId -> ObjectId" übernehmen
			$newCustomIds = array_map(function (array $document) { return $document['_customId']; }, $newDocuments);

			// Neue Dokumente erzeugen
			if ($newDocuments) {
				$result = $this->collection->insertMany(array_map(function ($data) {
					return $this->prepareDocumentContext($data);
				}, array_values($newDocuments)), ['session'=>$session]);

				if ($result->isAcknowledged()) {
					if ($result->getInsertedCount() == count($newDocuments)) {
						// Neue ObjectIds festhalten
						array_map(function ($customId, $objectId) {
							/** @var ObjectId|string $objectId */
							$this->customIdsMap[$customId] = (string)$objectId;
						}, $newCustomIds, $result->getInsertedIds());
					}
					else {
						// TODO: Es wurden nicht alle Daten gespeichert, loggen?
					}
				}
				else {
					// TODO: Fehlgeschlagenen Speichervorgang loggen?
				}
			}

			// Zu aktualisierende Dokumente ermitteln
			$updateCustomIds = array_diff($dirtyDocuments, $newCustomIds);

			// Datensätze mit benutzerdefinierten IDs aktualisieren
			foreach ($updateCustomIds as $customId) {
				if (isset($this->customIdsMap[$customId]) == false) {
					// TODO: Sollte nicht vorkommen, trotzdem loggen?
					continue;
				}

				$this->executeChanges($this->customIdsMap[$customId], $customId, $this->dirtyDocuments[$customId]);
			}
		}
		else {
			// Datensätze aktualisieren
			foreach ($dirtyDocuments as $documentId) {
				$this->executeChanges($documentId, null, $this->dirtyDocuments[$documentId]);
			}
		}

		$this->dirty = false;
		$this->dirtyDocuments = [];
		$this->writeLocked = false;
	}

	public function count(): int
	{
		$filter = $this->prepareFilterContext([]);
		if ($this->currentDataFilter === $filter) {
			return count($this->data);
		} else {
			return $this->collection->countDocuments($filter, ['session'=>$this->getSession()]);
		}
	}

	public function get(string $identifier, $fallback = null)
	{
		if (isset($this->cache[$identifier])) {
			return $this->cache[$identifier];
		}

		$nodeNames = array_values(array_filter(explode('.', $identifier), function ($str) { return strlen($str) > 0; }));
		if (count($nodeNames) == 0) {
			return $fallback;
		}

		$documentId = reset($nodeNames);

		// Gesamtes Dokument aus MongoDB in den Zwischenspeicher laden
		$this->fetchDocument($documentId);

		$value = $fallback;

		// Identifier auflösen und angeforderte Daten heraussuchen
		try {
			$value = array_reduce($nodeNames, function ($node, $key) use ($identifier) {
				if (is_array($node) && array_key_exists($key, $node)) {
					return $node[$key];
				}
				else {
					throw new \InvalidArgumentException("Config `$identifier` not found.");
				}
			}, $this->data);
		} catch (\InvalidArgumentException $exception) {}

		if (count($nodeNames) === 1 and isset($value['_value'])) {
			$value = $value['_value'];
		}

		$this->cache[$identifier] = $value;

		return $value;
	}

	/**
	 * @param string $identifier
	 * @return array{string,string[]} [$documentId, $nodeNames]
	 */
	private function prepareModification(string $identifier): array
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}

		$nodeNames = array_values(array_filter(explode('.', $identifier), function ($str) { return strlen($str) > 0; }));
		if (count($nodeNames) == 0) {
			return ['', []];
		}

		$documentId = reset($nodeNames);

		// Sicherstellen, dass das ein bestehendes Dokument geladen wurde
		$this->fetchDocument($documentId);

		return [$documentId, $nodeNames];
	}

	/**
	 * Diese Methode gibt eine Referenz auf einen Knotenpunkt innerhalb von `$this->data` zurück.
	 * @param array $nodeNames
	 * @return mixed
	 */
	private function &getDocumentNode(array $nodeNames)
	{
		$node = &$this->data;
		foreach ($nodeNames as $key => $childNode) {
			$isLast = $key == count($nodeNames) - 1;

			if (array_key_exists($childNode, $node) == false || ($isLast == false && is_array($node[$childNode]) == false)) {
				$node[$childNode] = [];
			}

			$tmpRef = &$node[$childNode];
			unset($node);

			$node = &$tmpRef;
			unset($tmpRef);
		}

		return $node;
	}

	/**
	 * @param string[] $nodeNames
	 */
	private function invalidateCacheSupSub(string $identifier, array $nodeNames): void
	{
		$this->currentDataFilter = null;

		// Obere Knotenpunkte im Cache invalidieren
		for ($i = count($nodeNames) - 1; $i > 0; $i--) {
			unset($this->cache[implode('.', array_slice($nodeNames, 0, $i))]);
		}

		// Untere Knotenpunkte im Cache invalidieren
		$prefix = $identifier.'.';
		$this->cache = array_filter($this->cache, function ($cacheIdentifier) use ($prefix) {
			return (strlen($cacheIdentifier) < strlen($prefix) or strpos($cacheIdentifier, $prefix) !== 0);
		}, ARRAY_FILTER_USE_KEY);
	}

	/**
	 * @param string $identifier
	 * @param mixed $data
	 * @return $this
	 */
	public function set(string $identifier, $data): self
	{
		list($documentId, $nodeNames) = $this->prepareModification($identifier);
		if (count($nodeNames) == 0) {
			return $this;
		}

		$node = &$this->getDocumentNode($nodeNames);

		if (count($nodeNames) == 1 and !is_array($data)) {
			$dataForCache = $data;
			$data = ['_value' => $data];
			$subidentifier = '_value';
		} else {
			$dataForCache = $data;
			$subidentifier = substr($identifier, strlen($documentId)+1);
		}

		// Kompatibilität mit PHP < 8.0
		if ($subidentifier === false) {
			$subidentifier = '';
		}

		$hasChanged = is_array($node) || is_object($node) || $node === null || $node !== $data;

		if ($hasChanged) {
			$node = $data;
			$this->cache[$identifier] = $dataForCache;
			$this->dirty = true;
			$this->dirtyDocuments[$documentId] = array_merge(
				$this->dirtyDocuments[$documentId] ?? [],
				[($subidentifier === '') ? ['replace', $data] : ['$set', $subidentifier, $dataForCache]]
			);
			$this->invalidateCacheSupSub($identifier, $nodeNames);
		}

		return $this;
	}

	/**
	 * @param string $identifier
	 * @param string|Binary $data
	 * @return $this
	 */
	public function setBinary(string $identifier, $data): self
	{
		if (is_string($data)) {
			$data = new Binary($data, Binary::TYPE_GENERIC);
		}
		return $this->set($identifier, $data);
	}

	/**
	 * Ein Zahlen-Datum erhöhen. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * Negative Werte können zum verringern verwendet werden. Wenn der Eintrag nicht existiert,
	 * wird er so behandelt als würde er 0 enthalten.
	 *
	 * @param string $identifier
	 * @param integer|float $amount
	 * @return $this
	 */
	public function increment(string $identifier, $amount=1)
	{
		list($documentId, $nodeNames) = $this->prepareModification($identifier);
		if (count($nodeNames) == 0) {
			throw new \LogicException('called increment() with too short an identifier.');
		}

		$node = &$this->getDocumentNode($nodeNames);

		if (count($nodeNames) == 1) {
			$subidentifier = '_value';
			if ($node === null) {
				$node = [];
			}
			if (array_key_exists('_value', $node) == false) {
				$node['_value'] = null;
			}
			$tmpRef = &$node['_value'];
			unset($node);
			$node = &$tmpRef;
			unset($tmpRef);
		} else {
			$subidentifier = substr($identifier, strlen($documentId)+1);
		}

		$hasChanged = is_array($node) || is_object($node) || $node === null || $amount !== 0;

		if ($hasChanged) {
			if ($node === null || $node === []) {
				$node = $amount;
			} else {
				$node += $amount;
			}
			$this->cache[$identifier] = $node;
			$this->dirty = true;
			$this->dirtyDocuments[$documentId] = array_merge(
				$this->dirtyDocuments[$documentId] ?? [],
				[['$inc', $subidentifier, $amount]]
			);
			$this->invalidateCacheSupSub($identifier, $nodeNames);
		}

		return $this;
	}

	/**
	 * Ein Element an ein Array anhängen. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * Wenn der Eintrag nicht existiert, wird er so behandelt als wäre es ein leeres Array.
	 *
	 * @param string $identifier
	 * @param mixed $item
	 * @param bool $noDuplicate Duplikate vermeiden
	 * @return $this
	 */
	public function appendToArray(string $identifier, $item, bool $noDuplicate=true)
	{
		list($documentId, $nodeNames) = $this->prepareModification($identifier);
		if (count($nodeNames) < 2) {
			throw new \LogicException('called appendToArray() with too short an identifier.');
		}
		$subidentifier = substr($identifier, strlen($documentId)+1);

		$node = &$this->getDocumentNode($nodeNames);

		$hasChanged = !$noDuplicate || !in_array($item, $node ?? [], true);

		if ($hasChanged) {
			if ($node === null) {
				$node = [$item];
			} else {
				$node[] = $item;
			}
			$this->cache[$identifier] = $node;
			$this->dirty = true;
			$this->dirtyDocuments[$documentId] = array_merge(
				$this->dirtyDocuments[$documentId] ?? [],
				[[$noDuplicate ? '$addToSet' : '$push', $subidentifier, $item]]
			);
			$this->invalidateCacheSupSub($identifier, $nodeNames);
		}

		return $this;
	}

	/**
	 * Ein Element aus einem Array löschen, falls vorhanden. Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * @param string $identifier
	 * @param mixed $item
	 * @return $this
	 */
	public function removeFromArray(string $identifier, $item)
	{
		list($documentId, $nodeNames) = $this->prepareModification($identifier);
		if (count($nodeNames) < 2) {
			throw new \LogicException('called removeFromArray() with too short an identifier.');
		}
		$subidentifier = substr($identifier, strlen($documentId)+1);

		$node = &$this->getDocumentNode($nodeNames);

		$hasChanged = in_array($item, $node ?? [], true);

		if ($hasChanged) {
			if ($node === null) {
				$node = [];
			} else {
				$node = array_filter($node, function ($v) use ($item) {
					return $v !== $item;
				});
			}
			$this->cache[$identifier] = $node;
			$this->dirty = true;
			$this->dirtyDocuments[$documentId] = array_merge(
				$this->dirtyDocuments[$documentId] ?? [],
				[['$pull', $subidentifier, $item]]
			);
			$this->invalidateCacheSupSub($identifier, $nodeNames);
		}

		return $this;
	}

	/**
	 * Daten in Repository speichern, falls ein Eintrag fehlt (oder NULL ist). Schlüssel werden im $identifier durch einen Punkt separiert.
	 *
	 * @param string $identifier
	 * @param mixed $data
	 * @return $this
	 */
	public function setIfMissing(string $identifier, $data)
	{
		list($documentId, $nodeNames) = $this->prepareModification($identifier);
		if (count($nodeNames) == 0) {
			return $this;
		}

		$node = &$this->getDocumentNode($nodeNames);

		if (count($nodeNames) == 1 and !is_array($data)) {
			$dataForCache = $data;
			$data = ['_value' => $data];
			$subidentifier = '_value';
		} else {
			$dataForCache = $data;
			$subidentifier = substr($identifier, strlen($documentId)+1);

			// Kompatibilität mit PHP < 8.0
			if ($subidentifier === false) {
				$subidentifier = '';
			}
		}

		$hasChanged = ($node === null or $node === []);

		if ($hasChanged) {
			$node = $data;
			$this->cache[$identifier] = $dataForCache;
			$this->dirty = true;
			$this->dirtyDocuments[$documentId] = array_merge(
				$this->dirtyDocuments[$documentId] ?? [],
				[($subidentifier === '') ? ['insertIfMissing', $data] : ['setIfMissing', $subidentifier, $dataForCache]]
			);
			$this->invalidateCacheSupSub($identifier, $nodeNames);
		}

		return $this;
	}

	/**
	 * Diese Methode entfernt einen Konfigurationsstrang. Die Änderung muss mit MongoDb::flush() verewigt werden.
	 * Wird ein vollständiges Dokument gelöscht, indem lediglich die ID als $identifier übergeben wurde, passiert
	 * dies sofort in der Datenbank.
	 * @param string $identifier
	 * @return $this
	 * @see MongoDb::flush()
	 */
	public function unset(string $identifier): self
	{
		if ($this->isVolatile() == false) {
			$this->beginWrite();
		}

		$nodeNames = array_values(array_filter(explode('.', $identifier), function ($str) { return strlen($str) > 0; }));
		if (count($nodeNames) == 0) {
			return $this;
		}

		$documentId = reset($nodeNames);

		// Sicherstellen, dass das ein bestehendes Dokument geladen wurde
		$this->fetchDocument($documentId);

		// Komplettes Dokument direkt in MongoDB löschen
		if (count($nodeNames) == 1 && ($this->customIds == false || isset($this->customIdsMap[$documentId]))) {
			if (!$this->getSession()->isInTransaction()) {
				$this->getSession()->startTransaction(['readPreference' => new ReadPreference('primary')]);
			}

			// Dokument in MongoDB löschen und Referenzen aus dem Zwischenspeicher entfernen
			$this->collection->deleteOne(['_id' => new ObjectId($this->customIdsMap[$documentId] ?? $documentId)], ['session'=>$this->getSession()]);
			unset($this->cache[$documentId], $this->customIdsMap[$documentId], $this->data[$documentId], $this->dirtyDocuments[$documentId]);
			$this->currentDataFilter = null;

			// Löschung des Dokuments in flush() persistieren
			$this->dirty = true;

			// Alle Cache-Einträge für dieses Dokument entfernen
			$this->cache = array_filter($this->cache, function (string $identifier) use ($documentId) {
				return strpos($identifier, $documentId.'.') == false;
			}, ARRAY_FILTER_USE_KEY);

			return $this;
		}

		$subidentifier = substr($identifier, strlen($documentId)+1);
		// Kompatibilität mit PHP < 8.0
		if ($subidentifier === false) {
			$subidentifier = '';
		}

		$node = &$this->data;
		foreach ($nodeNames as $key => $nodeName) {
			// Datenzweig entfernen
			if ($key + 1 == count($nodeNames)) {
				unset($node[$nodeName]);
				unset($this->cache[$identifier]);
				$this->dirty = true;
				// FIXME: Übergeordnete Knotenpunkte werden nicht gelöscht
				// FIXME: Bsp.: Objekt-ID (1) { banane: [1, 2, 3] }
				// FIXME: unset('1.banane.1') entfernt die 2 aus dem Array
				// FIXME: unset('1.banane') entfernt allerdings nicht das Objekt "banane".
				$this->dirtyDocuments[$documentId] = array_merge(
					$this->dirtyDocuments[$documentId] ?? [],
					[['$unset', $subidentifier, null]]
				);
				break;
			}
			// Wenn die definierte Datenstruktur nicht existiert, muss nichts gemacht werden
			else if (array_key_exists($nodeName, $node) == false || is_array($node[$nodeName]) == false) {
				break;
			}

			$tmpRef = &$node[$nodeName];
			unset($node);

			$node = &$tmpRef;
			unset($tmpRef);
		}

		return $this;
	}

	public function getPersistedSize(): int
	{
		$db = self::getService()->getDatabase();

		try {
			$averageDocumentSize = 0;
			$collectionSize = 0;

			foreach ($db->command(['collStats' => $this->collection->getCollectionName()]) as $data) {
				$averageDocumentSize = $data['avgObjSize'] ?? 0;
				$collectionSize = $data['size'] ?? 0;
				break;
			}

			if (!$this->user and !$this->domain) {
				return $collectionSize;
			}
			else {
				// Use `avgObjSize` just like the `estimate` flag does for the `dataSize` command.
				// Use this approach over the `dataSize` command because the latter would always
				// return the full size of the collection, even with `keyPattern` defined.
				return $averageDocumentSize * $this->count();
			}
		} catch (CommandException $e) {
			// ggf. loggen
			return 0;
		}
		return 0;
	}

	public function getLastModifiedTimestamp(): int
	{
		$filter = $this->prepareFilterContext([]);
		$result = $this->collection->findOne($filter, [
			'projection' => ['_mtime'=>1, '_id'=>0],
			'sort' => ['_mtime'=>-1]
		]);
		return ($result and isset($result['_mtime'])) ? $result['_mtime']->toDateTime()->getTimestamp() : 0;
	}

	public function makeVolatile(): void
	{
		if (!$this->volatile) {
			// @phan-suppress-next-line PhanTypeMismatchArgumentInternal
			$this->getSession()->startTransaction(['readPreference' => new ReadPreference('primary'), 'readConcern' => new ReadConcern(ReadConcern::LOCAL), 'writeConcern' => new WriteConcern(1, 3000)]);
		}
		$this->writeLocked = false;
		$this->volatile = true;
	}

	/**
	 * @param array<string,bool> $projection
	 * @return void
	 */
	public function setProjection(array $projection): void
	{
		$mandatoryProjectionFields = [
			'_id',
			'_mtime',
			'_userId', '_domainId',
			'_customId',
		];

		if ($inclusionProjection = array_filter($projection)) {
			$projection = array_fill_keys($mandatoryProjectionFields, true) + $inclusionProjection;
		}
		elseif ($projection) {
			$projection = array_diff_key($projection, array_flip($mandatoryProjectionFields));
		}

		// $this->reload();
		$this->makeVolatile();
		$this->projection = array_map('intval', $projection);
	}

	private function inTransaction(array $options, callable $callback)
	{
		$session = $this->getSession();
		for ($i = 0; $i < 5; ++$i) {
			if (!$this->getSession()->isInTransaction()) {
				$this->getSession()->startTransaction(['readPreference' => new ReadPreference('primary')]);
			}
			try {
				$callback($session);
				$session->commitTransaction();
			}
			catch (CommandException $e) {
				if ($e->hasErrorLabel('UnknownTransactionCommitResult')) {
					// Retry commit
					usleep(rand(100, 20000));
					$session->commitTransaction();
				}
				elseif ($e->hasErrorLabel('TransientTransactionError')) {
					// Retry whole transaction if not last try
					usleep(rand(100, 20000));
					try {
						$session->abortTransaction();
					} catch (\Exception $ignore) { /* ignore */ }
					if ($i >= 4) {
						throw $e;
					}
					continue;
				}
				else {
					throw $e;
				}

			}
			catch (BulkWriteException $e) {
				$result = $e->getWriteResult();
				$errors = $result->getWriteErrors();
				$wcError = $result->getWriteConcernError();
				if ($wcError) {
					$errors[] = $wcError;
				}

				try {
					$session->abortTransaction();
				} catch (\Exception $ignore) { /* ignore */ }

				// Neu versuchen, wenn es ein vorübergehender Fehler ist (WriteConflict etc.)
				if ($errors and $i < 4 and in_array($errors[0]->getCode(), self::TRANSIENT_ERROR_CODES)) {
					usleep(rand(100, 20000));
					continue;
				}
				else {
					throw $e;
				}
			}
			catch (\Exception $e) {
				try {
					$session->abortTransaction();
				} catch (\Exception $ignore) { /* ignore */ }
				throw $e;
			}
			break;
		}
	}

	public function isEmpty(): bool
	{
		if ($this->data) {
			return false;
		}

		$filter = $this->prepareFilterContext([]);
		try {
			return ($this->collection->findOne($filter, [
				'projection' => [],
				'returnKey' => true,
				'session' => $this->getSession(),
			]) === null);
		} catch (\Exception $exception) {
			throw new ConfigNotReadableException($exception->getMessage());
		}
	}

	public function reload(bool $forWrite=false): void
	{
		if ($this->writeLocked or $this->volatile) {
			if ($this->getSession()->isInTransaction()) {
				$this->getSession()->abortTransaction();
			}
			$this->writeLocked = false;
			$this->volatile = false;
			$this->projection = [];
		}

		$data = [];
		$this->currentDataFilter = null;
		$this->dirty = false;
		$this->dirtyDocuments = [];
		$this->templateLoaded = false;

		// Skeleton importieren, wenn noch keine Dokumente vorhanden sind
		if ($this->isEmpty()) {

			// Für Support von MongoDB < 4.4 müssen Collections außerhalb einer Transaktion erzeugt werden.
			// Das Erstellen der Indices hier sichert das.
			if (count(iterator_to_array($this->collection->listIndexes())) == 0) {
				// @phan-suppress-next-line PhanTypeMismatchArgumentInternal
				$wc = new WriteConcern(1, 5000);

				if ($this->domain) {
					if ($this->customIds) {
						$this->collection->createIndex(['_domainId'=>1, '_customId'=>1], ['unique'=>true, 'writeConcern'=>$wc]);
					}
					if ($this->wantsMtimeIndex) {
						$this->collection->createIndex(['_domainId'=>1, '_mtime'=>-1], ['writeConcern'=>$wc]);
					} elseif (!$this->customIds) {
						$this->collection->createIndex(['_domainId'=>1], ['writeConcern'=>$wc]);
					}
				}
				elseif ($this->user) {
					if ($this->customIds) {
						$this->collection->createIndex(['_userId'=>1, '_customId'=>1], ['unique'=>true, 'writeConcern'=>$wc]);
					}
					if ($this->wantsMtimeIndex) {
						$this->collection->createIndex(['_userId'=>1, '_mtime'=>-1], ['writeConcern'=>$wc]);
					} elseif (!$this->customIds) {
						$this->collection->createIndex(['_userId'=>1], ['writeConcern'=>$wc]);
					}
				}
				else {
					if ($this->customIds) {
						$this->collection->createIndex(['_customId'=>'hashed'], ['writeConcern'=>$wc]);
					}
					if ($this->wantsMtimeIndex) {
						$this->collection->createIndex(['_mtime'=>-1], ['writeConcern'=>$wc]);
					}
				}
			}

			if ($this->templateDir !== null) {
				// User-/Domain-spezifizisches liegt in share/skeleton/ statt nur in share/
				$skeletonPrefix = ($this->user or $this->domain) ? 'skeleton'.DIRECTORY_SEPARATOR : '';
				$skeleton = new JsonConfig($skeletonPrefix.$this->collection->getCollectionName().'.json', false, $this->templateDir, null);

				if (!$skeleton->isEmpty()) {
					$data = array_map(function ($data, $id) {
						return $this->customIds ? ['_customId'=>$id] + $data : $data;
					}, $skeleton->toArray(), array_keys($skeleton->toArray()));

					$this->inTransaction([], function ($session) use (&$data, $skeleton) {
						if ($this->isEmpty()) {

							$result = $this->collection->insertMany(array_map(function ($data) {
								return $this->prepareDocumentContext($data);
							}, $data), ['session'=>$session]);

							if ($result->isAcknowledged()) {
								if ($result->getInsertedCount() == count($data)) {
									$data = array_combine(
										$this->customIds
											? array_keys($skeleton->toArray())
											: array_map(function (ObjectId $id) { return (string)$id; }, $result->getInsertedIds()),
										array_values($data)
									);
								}
								else {
									$this->collection->deleteMany([
										'_id' => ['$in' => $result->getInsertedIds()],
									], ['session'=>$session]);

									throw new \RuntimeException('Not all skeleton documents imported, action reverted');
								}
							}
							else {
								throw new \RuntimeException('Couldn\'t load skeleton into collection.');
							}

							$this->data = is_array($data) ? $data : [];

						}
					});

					if ($forWrite) {
						$this->dirty = true;
						$this->writeLocked = true;
					} else {
						$this->writeLocked = false;
					}
				}

				// Festhalten, dass eine Vorlagendatei eingelesen wurde
				$this->templateLoaded = (bool)$data;
			}
		}

		if ($forWrite and !$this->writeLocked) {
			$this->writeLocked = true;
		}

		$this->customIdsMap = [];
		$this->data = is_array($data) ? $data : [];
		$this->cache = [];
		$this->volatile = false;
		$this->projection = [];
	}

	/**
	 * @return array
	 */
	public function toArray(): array
	{
		$this->fetchDocuments([]);
		return $this->data;
	}

	/**
	 * @param array $filter MongoDB-Query-Filter
	 * @return array
	 */
	public function toArrayFiltered(array $filter): array
	{
		$this->fetchDocuments($filter);
		return iterator_to_array(new QueryFilterIterator(new ArrayIterator($this->data), $filter));
	}

	/**
	 * Bulk-Updates für MongoDB-spezifische Optimierungen
	 *
	 * Nach dem Vorgang wird das Repository neugeladen, nicht gespeicherte Änderungen gehen verloren.
	 * Beim Neuladen bleibt der Zustand `writeLocked` erhalten.
	 *
	 * @param array<string,mixed> $filter MongoDB-Filter
	 * @param array<string,mixed> $update Update-Dokument
	 * @param bool $performGlobally Wenn TRUE, werden Dokumente *aller* User und Domains bearbeitet.
	 *             Globale Updates werden außerhalb einer Transaktion durchgeführt und damit sofort wirksam.
	 * @return int Anzahl der veränderten Dokumente
	 */
	public function performCustomBulkUpdate(array $filter, array $update, bool $performGlobally = false): int
	{
		$options = [];
		if (!$performGlobally) {
			$options['session'] = $this->getSession();
			$filter = $this->prepareFilterContext($filter);
		}
		$result = $this->collection->updateMany($filter, $update, $options);

		// Repository neuladen, da abgerufene Dokumente möglicherweise überschrieben wurden
		if ($this->data || $this->dirty) {
			$this->reload($this->writeLocked);
		}

		return $result->getModifiedCount();
	}

	/**
	 * Escaping von Schlüsseln.
	 *
	 * Verhindert das . als Trenner für Unterelemente interpretiert wird.
	 *
	 * TODO: Convert RepositoryInterface to abstract class and implement this method there -- @see JsonConfig::escapeKey()
	 *
	 * @param string $key
	 * @return string
	 */
	public static function escapeKey($key): string
	{
		$key = str_replace('%', '%25', (string)$key);
		$key = str_replace("\0", '%00', $key);
		$key = str_replace('.', '%2E', $key);
		return $key;
	}

	/**
	 * Unescaping von Schlüsseln.
	 *
	 * TODO: Convert RepositoryInterface to abstract class and implement this method there -- @see JsonConfig::unescapeKey()
	 *
	 * @param string $key
	 * @return string
	 * @see MongoDb::escapeKey()
	 */
	public static function unescapeKey(string $key): string
	{
		$key = str_replace('%2E', '.', $key);
		$key = str_replace('%00', "\0", $key);
		$key = str_replace('%25', '%', $key);
		return $key;
	}
}
