<?php

namespace App\Component\Cron;
use App\Model\Domain;
use App\Model\Model;
use App\Model\User;
use App\Utils;
use App\Config;

class Cron
{
	/** @var JobBase[] $jobs */
	private $jobs = [];
	/** @var array[] $deferredJobs */
	private $deferredJobs = [];

	public function setJobList(iterable $jobs)
	{
		if (!is_array($jobs)) {
			$jobs = iterator_to_array($jobs);
		}
		$this->jobs = $jobs;
	}

	/**
	 * Setzt den Benutzer-/Domain-Kontext zum Durchführen eines oder mehrerer Cronjobs.
	 *
	 * Es wird in den User und die Domain gewechselt. Nicht gespeicherte Änderungen
	 * an allen Benutzer-/Domain-spezifischen Models gehen verloren.
	 *
	 * @return bool TRUE bei Erfolg, ansonsten bleiben User und Domain auf dem vorherigen Wert.
	 */
	public function setContext(?User $user, ?string $domainId): bool
	{
		$prevUser = User::loggedInUser();
		$prevDomain = Domain::activeDomain();

		// Alle relevanten offenen Models schließen
		Model::closeUserModels();

		// User wechseln
		if ($user !== null) {
			if (User::setApiUser($user->getApiKey())) {
				Domain::reloadRepository();
			} else {
				User::setApiUser($prevUser ? $prevUser->getApiKey() : '');
				return false;
			}
		}

		// Domain wechseln
		if ($domainId !== null and !Domain::setDomainForRequest($domainId)) {
			// Bei Fehler alles zurückstellen und FALSE zurückgeben
			if ($prevUser) {
				User::setApiUser($prevUser->getApiKey());
				Domain::switchRepositoryContext(null);
				Domain::setDomainForRequest(($prevDomain) ? $prevDomain->getId() : '');
			} else {
				User::setApiUser('');
				Domain::switchRepositoryContext(null);
			}
			return false;
		}

		foreach ($this->jobs as $job) {
			$job->reinitialize();
		}

		return true;
	}

	/**
	 * @return JobBase[]
	 */
	public function jobs(): array
	{
		return $this->jobs;
	}

	/**
	 * Alle fälligen Cronjobs
	 *
	 * Domainweite Jobs werden nur berücksichtigt, wenn eine Domain aktiv ist.
	 * Userweite Jobs werden nur berücksichtigt wenn ein User aktiv/eingeloggt ist.
	 * Systemweite Jobs werden immer berücksichtigt.
	 *
	 * @return JobBase[]
	 */
	public function dueJobs(): array
	{
		$hasActiveDomain = (bool)Domain::activeDomain();
		$hasActiveUser = (bool)User::loggedInUser();
		return array_filter($this->jobs, function ($job) use ($hasActiveDomain, $hasActiveUser) {
			/** @var JobBase $job */
			return ((($hasActiveDomain and $job instanceof DomainJob) or ($hasActiveUser and $job instanceof UserJob) or $job instanceof Job) and $job->isDue());
		});
	}

	/**
	 * Job-Objekt nach Namen suchen
	 *
	 * @param string $name
	 * @return JobBase|null
	 */
	public function jobByName(string $name): ?JobBase
	{
		return array_reduce($this->jobs, function ($previousJob, $currentJob) use ($name) {
			/**
			 * @var JobBase|null $previousJob
			 * @var JobBase $currentJob
			 */
			return $currentJob->getName() == $name ? $currentJob : $previousJob;
		}, null);
	}

	/**
	 * Plant die Ausführung eines Jobs im Anschluss an die aktuelle Verarbeitung ein.
	 * Diese Jobs werden (mehr oder weniger) asynchron über Curl durchgeführt.
	 *
	 * @param JobBase $job
	 */
	public function addDeferredJob(JobBase $job): void
	{
		$user = User::loggedInUser();
		$domain = Domain::activeDomain();
		$this->deferredJobs[] = [$job, $user->getApiKey(), $domain ? $domain->getId() : null];
	}

	/**
	 * Gibt die Liste der zurückgestellten Jobs aus. Die Liste wird nach der Rückgabe geleert.
	 *
	 * @return array[] [ [$job, $apiKey, ?$domainId] ]
	 */
	public function deferredJobs(): array
	{
		$result = $this->deferredJobs;
		$this->deferredJobs = [];
		return $result;
	}

	/**
	 * Führt die Funktion/Closure mit dem Cron-Lock aus
	 *
	 * @param callable() $callable
	 * @return bool TRUE bei Erfolg, FALSE wenn schon ein Cronjob läuft
	 */
	public function withCronLock(callable $callable): bool
	{
		$config = Config::getInstance();
		// Sicherstellen, dass immer nur ein Cronjob zeitgleich abgearbeitet wird
		$isCluster = (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB and !empty($_ENV['APP_CLUSTER_PEERS']));
		if ($isCluster) {
			// Bei MongoDB-Cluster: über DB synchronisieren
			// Dazu bis zu 10x versuchen den Lock zu bekommen (max 250ms lang), und ansonsten abbrechen
			$blocked = true;
			$lockf = null;
			for ($i = 0; ($i < 10); $i++) {
				$blocked = !$config->setCronjobRunning();
				if ($blocked) {
					usleep(rand(15000, 25000));
					$config->reload();
				} else {
					break;
				}
			}
		} else {
			// Ohne Cluster über Lock-Datei synchronisieren
			$lockf = @fopen(Utils::getBundledVarDir().'/.cron.lock', 'wb');
			// Dazu bis zu 30x versuchen den Lock zu bekommen (max 250ms lang), und ansonsten abbrechen
			$blocked = true;
			for ($i = 0; ($i < 25); $i++) {
				flock($lockf, LOCK_EX|LOCK_NB, $blocked);
				if ($blocked) {
					usleep(rand(750, 10000));
				} else {
					break;
				}
			}
		}
		if ($blocked) {
			return false;
		}

		try {
			$callable();
		}
		finally {
			// Nächsten Cronjob freigeben
			if ($isCluster) {
				$config->unsetCronjobRunning();
			} elseif ($lockf) {
				flock($lockf, LOCK_UN);
				fclose($lockf);
			}
		}
		return true;
	}
}
