<?php

namespace App\Component\Cron\Job;

use App\Component\Licensing\ApiClient as LicensingApiClient;
use App\Component\Cron\Job;
use App\Config;
use App\Exception\LicenseInvalidException;
use App\Model\Domain;
use App\Model\SystemLog;
use App\Model\User;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;

class LicenseCheck extends Job
{
	const COMPONENT = 'LicenseCheck';
	/** @var LicensingApiClient $api */
	private $api;

	public function __construct(LicensingApiClient $api)
	{
		parent::__construct(60*60*6);
		$this->api = $api;
	}

	public function run(Request $request): void
	{
		//Licence Check 4 x am Tag ausführen oder wenn hrCheck gesetzt ist und kein Call von außen kommt
		if(empty($request->query->get('key')) && $this->lastExecutionTimestamp()+$this->intervalSeconds > time())
		{
			return;
		}

		// Zeitstempel der letzten Ausführung aktualisieren
		$this->touchLastExecutionTimestamp();

		// Lizenzschlüssel validieren
		$this->checkLicense($request);

		// Wenn kein Lizenzschlüssel hinterlegt ist, Cronjob als erledigt betrachten
		if (!Utils::hasLicenseKey()) {
			return;
		}

		//Und nur durchführen wenn es sich um eine Agency-Edition handelt
		if (Utils::isExtendedEdition()) {
			$this->reportLicensees($request);
		}
	}

	/**
	 * Diese Methode validiert den im System hinterlegten Lizenzschlüssel.
	 */
	private function checkLicense(Request $request): void
	{
		$licenseKey = Utils::getLicenseKey();

		// Wenn kein Lizenzschlüssel hinterlegt ist, Vorgang überspringen
		if (empty($licenseKey)) {
			return;
		}

		$logger = SystemLog::current();

		try {
			// Erfolg indiziert die Validität der Lizenz
			$licenseInfo = $this->api->registerLicense($licenseKey, $request->getHttpHost());
		} catch (LicenseInvalidException $e) {
			// Lizenzinformationen nur in der Basisedition zurücksetzen; Konfiguration
			// einer Agenturversion unverändert lassen, um den Betrieb eines möglichen
			// Produktivsystems zu erhalten.
			/**
			 * TODO
			 * Hier müssen wir eigentlich mal ansetzen, das geht ja nicht dass die Versionen immer weiter laufen
			 * Die alten Base Editions laufen ja weiter, müssen die ja auch, nur die multi und agency Versionen müssen wir dann irgendwann mal abschalten.
			 * Oder zumindest Kenntniss darüber bekommen.
			 */
			/* Vorerst keine Lizenzinformationen zurücksetzen, um den Betrieb valider Installationen zu erhalten.
			if (Utils::isBaseEdition()) {
				Config::getInstance()
					->set('licensekey', '')
					->set('whitelabel', false)
					->flush()
				;
				Utils::setLocalEnvironment([
					'APP_EDITION' => '',
				]);
			}
			*/

			$translator = Utils::getTranslator();
			$logMessage = "License has expired.\nKey: {$licenseKey}\nReason: {$translator->trans($e->getMessage())}";
			$logger->logWarning(self::COMPONENT, $logMessage);
			return;
		} catch (\Exception $e) {

			$translator = Utils::getTranslator();
			$logMessage = 'License could not be verified.'.rtrim(' '.$translator->trans($e->getMessage()), ' ');
			$logger->logError(self::COMPONENT, $logMessage);
			return;
		}

		Config::getInstance()->set('whitelabel', $licenseInfo->whitelabel)->flush();
		Utils::setLocalEnvironment([
			'APP_EDITION' => $licenseInfo->edition,
		]);

		// Das Limit ist für Administratoren derzeit nur in der Multi-Domain-Edition relevant
		foreach (User::byRole(User::ROLE_ADMIN) as $user) {
			$user->setMaxDomainCount($licenseInfo->limit);
			$user->setUserTarifSettings($licenseInfo->tarif);

			//Multi Version - dann die Max Callcounts in den neue Versionen berücksichtigen.
			if($licenseInfo->edition === "multi")
			{
				$user->setMaxCallCount($licenseInfo->maxcalls);
			}
			//Sonder / Altfall eine Single Base Version oder Pro oder sonstige Lizenz - dann auf base stellen und Funktionen passend aktivieren.
			if($licenseInfo->edition === "base")
			{
				$user->setUserTarifSettings("base");
			}
		}
		//speichern
		User::flush();
	}

	/**
	 * Diese Methode erstattet Bericht, über die Anzahl aktiver Kunden, Domains etc., zwecks Fakturierung.
	 */
	private function reportLicensees(Request $request): void
	{
		$sumCount = [];
		$userCount = $domainCount = 0;
		try {
			$userCount = User::count()-1; //abzgl. root User
			// TODO: Optimieren: zwei Loops zu einem zusammenfassen
			$domainCount = array_reduce(User::getClients(), static function ($count, User $user) {
				Domain::switchToUser($user);
				return $count + Domain::countIncludingShareNames();
			}, 0);
			foreach (User::getClients() as $client)
			{
				Domain::switchToUser($client);
				$currentDomCount = Domain::countIncludingShareNames();
				if(empty($sumCount[$currentDomCount])) {
					$sumCount[$currentDomCount] = 1;
				}
				else{
					$sumCount[$currentDomCount]++;
				}
			}
		}
		finally {
			$user = User::loggedInUser();
			if ($user) {
				Domain::switchToUser($user);
			}
		}

		try {
			$this->api->reportLicensees(Utils::getLicenseKey(), $request->getHttpHost(), $domainCount, $userCount, serialize($sumCount));
		} catch (\Exception $e) {
			$translator = Utils::getTranslator();
			$logger = SystemLog::current();
			$logMessage = 'Failed to send licensee report: '.rtrim(' '.$translator->trans($e->getMessage()), ' ');
			$logger->logError(self::COMPONENT, $logMessage);
			return;
		}
	}
}
