<?php

namespace App\Component\Cron\Job;

use App\Component\Cron\DomainJob;
use App\Config;
use App\Controller\PageCheck;
use App\Component\Mail\MailMessage;
use App\Component\LoggerInterface;
use App\Model\Domain;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Mailer\MailerInterface;
use Symfony\Contracts\Translation\TranslatorInterface;
use Exception;

class NewCookieAndCoNotification extends DomainJob
{
	/** @var MailerInterface */
	private $mailer;
	/** @var LoggerInterface */
	private $logger;
	/** @var TranslatorInterface */
	private $translator;

	public function __construct(LoggerInterface $logger, TranslatorInterface $translator, MailerInterface $mailer)
	{
		//60 * 60 * 24
		parent::__construct(60 * 60 * 24);
		$this->logger = $logger;
		$this->mailer = $mailer;
		$this->translator = $translator;
	}

	public function run(Request $request): void
	{
		$startTime = time();
		//Timer neu setzen
		$this->touchLastExecutionTimestamp();

		//config einbinden
		$config = Config::getInstance();
		$user = Domain::activeDomain()->getUser();

		//Wenn die Versendung nicht aktiv ist, einfach wieder raus - timer wurde ja neu gesetzt...
		if ($config->get('sub_cookie_email_check') != "on")
		{
			return;
		}
		elseif (!$user->isMailingSubscribed('CookieNotification', true)) {
			return; // Nicht senden wenn für den User deaktiviert
		}
		elseif (!$user->isActive()) {
			return; // Nicht senden wenn User deaktiviert
		}


		//Erstmal schauen ob es neue Daten gibt...
		$found_cookies = new PageCheck();
		$foundData = $found_cookies->getCounter($request,true);

		//Wenn mindestens ein Eintrag vorliegt
		if(
			$foundData['cookiesCount'] > 0
			|| $foundData['LocalStoragesCount'] > 0
			|| $foundData['skriptCount'] > 0
			|| $foundData['contentCount'] > 0
		)
		{
			//ini
			$foundText = "";

			//Dann go... Text setzen
			if($foundData['cookiesCount'] > 0)
			{
				$foundText.= $foundData['cookiesCount'].$this->translator->trans(' cookies found.')."\n";
			}

			if($foundData['LocalStoragesCount'] > 0)
			{
				$foundText.= $foundData['LocalStoragesCount'].$this->translator->trans(' LocalStorage elements found.')."\n";
			}

			if($foundData['skriptCount'] > 0)
			{
				$foundText.= $foundData['skriptCount'].$this->translator->trans(' scripts found.')."\n";
			}

			if($foundData['contentCount'] > 0)
			{
				$foundText.= $foundData['contentCount'].$this->translator->trans(' other external elements found.')."\n";
			}

			$domain = Domain::activeDomain();
			$user = $domain->getUser();
			assert($user !== null);

			//Je nach Version die TO Adresse setzen
			if (Utils::isExtendedEdition()) {
				//E-Mail Adresse des Users der Domain...
				$mailto = $user->getEmailAddress();
				$domainName = $domain->getName();
			}
			elseif (Utils::hasMultiDomainSupport()) {
				$mailto = $config->get('email');
				$domainName = $domain->getName();
			}
			else {
				$mailto = $config->get('email');
				$domainName = "";
			}

			//Mail Vorlage...
			$subject = $config->get('sub_cookie_email_betreff');
			$body = $config->get('sub_cookie_email_content');
			$body = str_ireplace("#uncontrolledData#",$foundText,$body);
			$body = str_ireplace("{{ uncontrolledData }}",$foundText,$body);
			$body = str_ireplace("{{ domainvar }}",$domainName,$body);
			$subject = str_ireplace("{{ uncontrolledData }}",$foundText,$subject);
			$subject = str_ireplace("{{ domainvar }}",$domainName,$subject);

			//Mail versenden...
			$message = (new MailMessage($subject))
				->from($config->get('email')) // Von ist Admin E-Mail
				->to($mailto)   // to ist die MailAdresse des jeweiligen Kundens... oder Admin wenn Base...
				->listId('CookieNotification.'.$user->getId())
				->unsubscribe($user, 'CookieNotification')
				->autoSubmitted()
				->text($body);
			try {
				$this->mailer->send($message);
				$stopTime = time();
				$took = ($startTime - $stopTime);
				$took = " (".$took.") sec.";

				$this->logger->logInfo('CookieNotification', "Sent notification mail to ".$mailto." ".$took);
			} catch (Exception $e) {
				$stopTime = time();
				$took = ($startTime - $stopTime);
				$took = " (".$took.") sec.";
				$this->logger->logWarning('CookieNotification', "Could not send notification email to ".$mailto." ".$took);
			}
		}
		return;
	}
}
