<?php

namespace App\Component\Cron\Job;

use App\Component\Cron\DomainJob;
use App\Model\Domain;
use App\Model\Protocol;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use RuntimeException;

class ProtocolRotation extends DomainJob
{
	const MAX_FILESIZE_IN_MEGABYTES = 5;
	const MAX_ENTRIES_BEFORE_ROTATE = 10000;
	const MAX_DAYS = 365;
	const MAX_DAYS_REDIS = 7;

	public function __construct()
	{
		parent::__construct(30 * 60);
	}

	public function run(Request $request): void
	{
		$lastDateCheck = $this->readConfig('lastDateCheck');
		$this->touchLastExecutionTimestamp();
		$repositoryStore = Utils::getProtocolRepositoryStore();
		$protocol = Protocol::byDomain(Domain::activeDomain());
		$now = (int)time();

		// Auf jeden Fall rotieren, wenn der älteste Eintrag im Protokoll älter
		// als 365 Tage (bzw. 7 bei Redis) ist. (Das aber nur alle 2 Tage überprüfen.)
		if ($lastDateCheck === null or ($now - (int)$lastDateCheck) > 48*60*60) {
			$maxDays = $repositoryStore === Utils::REPOSITORY_STORE_REDIS ? self::MAX_DAYS_REDIS : self::MAX_DAYS;
			$this->writeConfig('lastDateCheck', $now);
			$this->state()->save();
			$firstEntry = $protocol->first();
			if ($firstEntry and !empty($firstEntry['timestamp']) and ($now - (int)$firstEntry['timestamp']) > $maxDays*24*60*60) {
				try {
					$protocol->rotateProtocol();
					return;
				}
				catch (RuntimeException $ignore) {}
			}
		}

		// Je nach Repository-Typ andere Größenbeschränkung verwenden
		switch ($repositoryStore) {
			case Utils::REPOSITORY_STORE_MONGODB:
			case Utils::REPOSITORY_STORE_REDIS:
				$this->runByCountLimit($protocol);
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$this->runBySizeLimit($protocol);
				break;
		}
	}

	private function runBySizeLimit(Protocol $protocol): void
	{
		// Sobald die Datei zu groß wird, komprimieren und leeren
		if ($protocol->persistedRepositorySize() > self::MAX_FILESIZE_IN_MEGABYTES * pow(1024, 2)) {
			$protocol->rotateProtocol();
		}
	}

	private function runByCountLimit(Protocol $protocol): void
	{
		if ($protocol->count() > self::MAX_ENTRIES_BEFORE_ROTATE) {
			try {
				$protocol->rotateProtocol();
			}
			catch (RuntimeException $ignore) {}
		}
	}
}
