<?php

namespace App\Component\Deployment;

use App\Model\Domain;
use App\Model\Model;
use App\Model\SystemTheme;
use App\Model\Theme;
use App\Model\User;
use Exception;

class SystemThemeDeployment
{
	/**
	 * This method distributes all `SystemTheme` objects to all domains of each client.
	 *
	 * This method closes all user and domain model repositories, locks the `Theme`
	 * repository and releases the lock after writing the changes to the database.
	 *
	 * @return void
	 * @throws Exception if an error occurs.
	 */
	public static function publishAllSystemThemes(): void
	{
		self::publishSystemThemes(SystemTheme::all());
	}

	/**
	 * This method distributes a `SystemTheme` to all domains of each client.
	 *
	 * This method closes all user and domain model repositories, locks the `Theme`
	 * repository and releases the lock after writing the changes to the database.
	 *
	 * @param SystemTheme $systemTheme
	 * @return void
	 * @throws Exception if an error occurs.
	 */
	public static function publishSystemTheme(SystemTheme $systemTheme): void
	{
		self::publishSystemThemes([$systemTheme->getId()=>$systemTheme]);
	}

	/**
	 * This method distributes multiple `SystemTheme` objects to all domains of each client.
	 *
	 * This method closes all user and domain model repositories, locks the `Theme`
	 * repository and releases the lock after writing the changes to the database.
	 *
	 * @param SystemTheme[] $systemThemes
	 * @return void
	 * @throws Exception if an error occurs.
	 */
	private static function publishSystemThemes(array $systemThemes): void
	{
		// Skip the entire process if there are no themes
		if (empty($systemThemes)) {
			return;
		}

		// Remove duplicates and index themes using their ID
		$systemThemes = array_combine(
			array_map(function (SystemTheme $theme) { return $theme->getId(); }, $systemThemes),
			$systemThemes
		);

		// Remember the current user context
		$requestUser = User::loggedInUser();
		// Remember the current domain context
		$requestDomainId = Domain::getDomainIdForRequest();

		try {
			foreach (User::getClients() as $client) {
				// Switch user context
				User::setApiUser($client->getApiKey());
				Model::closeUserModels();
				Domain::switchToUser($client);

				foreach (Domain::all() as $domain) {
					// Switch domain context
					Domain::setDomainForRequest($domain->getId());
					Model::closeDomainModels();

					// Upsert deployed themes
					foreach ($systemThemes as $systemTheme) {
						self::migrateSystemTheme($systemTheme);
					}

					// Delete obsolete themes
					foreach (Theme::all() as $theme) {
						if ($theme->isDangling() && $theme->isDefaultTheme() == false) {
							Theme::delete($theme->getId(), false);
						}
					}

					// Write changes to the database
					Theme::flush();
				}
			}
		} finally {
			// Switch back to the original user/domain context
			User::setApiUser($requestUser ? $requestUser->getApiKey() : '');
			Model::closeUserModels();
			if ($requestUser) {
				Domain::switchToUser($requestUser);
			}
			Domain::setDomainForRequest($requestDomainId);
		}

		// Update published values
		foreach ($systemThemes as $systemTheme) {
			$systemTheme
				->setPublishedSettingsIcon($systemTheme->getSettingsIcon())
				->setPublishedSettingsIconBackgroundColor($systemTheme->getSettingsIconBackgroundColor())
				->setPublishedSettingsIconForegroundColor($systemTheme->getSettingsIconForegroundColor())
			;
		}
		SystemTheme::flush();
	}

	/**
	 * @param SystemTheme $systemTheme
	 * @return void
	 */
	private static function migrateSystemTheme(SystemTheme $systemTheme): void
	{
		$theme = self::findOrCreateTheme($systemTheme);

		// Daten des System-Themes in das referenzierende Theme kopieren
		$theme->setSystemTheme($systemTheme);

		// Aktualisiere die Einstellungs-Icon-Farben des Themes, wenn sie noch mit denen des System-Themes übereinstimmen
		$previousValue = $systemTheme->getPublishedSettingsIconBackgroundColor();
		if ($theme->getSettingsIconBackgroundColor() === $previousValue || $previousValue === null) {
			$theme->setSettingsIconBackgroundColor($systemTheme->getSettingsIconBackgroundColor());
		}
		$previousValue = $systemTheme->getPublishedSettingsIconForegroundColor();
		if ($theme->getSettingsIconForegroundColor() === $previousValue || $previousValue === null) {
			$theme->setSettingsIconForegroundColor($systemTheme->getSettingsIconForegroundColor());
		}
	}

	/**
	 * @param SystemTheme $systemTheme
	 * @return Theme
	 */
	private static function findOrCreateTheme(SystemTheme $systemTheme): Theme
	{
		// Finde das Theme, das das aktuelle System-Theme referenziert; erstelle alternativ ein neues
		return Theme::bySystemTheme($systemTheme) ?? Theme::create()
			->setSystemTheme($systemTheme)
			->setSettingsIconBackgroundColor($systemTheme->getPublishedSettingsIconBackgroundColor() ?? $systemTheme->getSettingsIconBackgroundColor())
			->setSettingsIconForegroundColor($systemTheme->getPublishedSettingsIconForegroundColor() ?? $systemTheme->getSettingsIconForegroundColor())
		;
	}
}
