<?php

namespace App\Component\EmbeddingDb;

use App\Exception\ConfigNotWritableException;
use App\Exception\ConfigNotReadableException;
use App\Exception\DownloaderException;
use App\Exception\DownloaderHTTPException;
use App\Component\Curl;
use App\Utils;
use DateTime;
use DateTimeInterface;
use InvalidArgumentException;

class Repository
{
	private const APIKEY = 'jkn487gbf67igvfg6f2fhfbzg96ibjgbvzug';

	/** @var self[] $instances*/
	static private $instances = [];

	/** @var string $filename */
	protected $filename;

	/** @var array $data */
	private $data;

	/**
	 * Lädt eine EmbeddingDb-Datendatei.
	 * @param string $filename
	 * @return static
	 */
	static public function load(string $filename): self
	{
		if (isset(self::$instances[$filename])) {
			return self::$instances[$filename];
		}

		$instance = new static($filename);

		self::$instances[$filename] = $instance;
		return $instance;
	}

	/**
	 * Lädt eine EmbeddingDb-Datendatei.
	 * @param string $filename
	 */
	protected function __construct(string $filename)
	{
		$this->filename = $filename;

		$data = [];
		// Wenn die Datei existiert, wird sie eingelesen, andernfalls wird versucht, sie herunterzuladen.
		$this->reload();
	}

	/**
	 * Diese Methode gibt den Dateipfad der Konfigurationsdatei zurück.
	 * @return string
	 */
	protected function getPathname(): string
	{
		$filename = strtr($this->filename, ['/', '__']);
		@mkdir(Utils::getVarDir() . DIRECTORY_SEPARATOR . 'embeddingdb');
		return Utils::getVarDir() . DIRECTORY_SEPARATOR . 'embeddingdb' . DIRECTORY_SEPARATOR . $filename . '.json';
	}

	/**
	 * @return array
	 */
	public function toArray(): array
	{
		return $this->data;
	}

	/**
	 * Größe des Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes
	 */
	public function getPersistedSize(): int
	{
		return (int)@filesize($this->getPathname());
	}

	/**
	 * Diese Methode liest Daten aus der Konfiguration. Schlüssel werden im $identifier durch einen Punkt separiert.
	 * @param string $identifier
	 * @param mixed $fallback
	 * @return mixed
	 */
	public function get(string $identifier, $fallback = null)
	{
		$nodeNames = explode('.', $identifier);
		if (empty($identifier) || count($nodeNames) == 0) {
			return $fallback;
		}

		$value = $fallback;

		try {
			$value = array_reduce($nodeNames, function ($node, $key) use ($identifier) {
				if (is_array($node) && array_key_exists($key, $node)) {
					return $node[$key];
				}
				else {
					throw new InvalidArgumentException("Regulation data `$identifier` not found.");
				}
			}, $this->data);
		} catch (InvalidArgumentException $exception) {}

		return $value;
	}

	/**
	 * Angepeilter Download-Intervall in Sekunden
	 *
	 * @param int $fallback
	 * @return int
	 */
	private function getDownloadInterval(int $fallback=43200): int
	{
		$path = $this->getPathname().'.maxage';
		$value = (int)(@file_get_contents($path) ?: $fallback);

		return ($value > 86400) ? 86400 : $value;
	}

	/**
	 * @return DateTimeInterface
	 */
	public function getLastModified(): DateTimeInterface
	{
		$path = $this->getPathname();
		$mtime = @filemtime($path);
		if ($mtime) {
			return new DateTime('@'.$mtime);
		} else {
			return new DateTime('1970-01-01');
		}
	}

	/**
	 * Ob ein neuer Download durchgeführt werden sollte
	 *
	 * @return bool
	 */
	private function isStale(): bool
	{
		$path = $this->getPathname();
		return (!file_exists($path) or @filesize($path) == 0 or time() > (int)(@filemtime($path)) + $this->getDownloadInterval());
	}

	/**
	 * @return string[]
	 */
	protected function getDownloadUrls(): array
	{
		return [
			'https://cookiedb.ccm19.de/v2/'.$this->filename,
		];
	}

	/**
	 * Auf 10er-Stellen abrunden
	 */
	private function floorInt(int $value, int $precision): int {
		$factor = pow(10, $precision);
		return (int)($value / $factor) * $factor;
	}

	/**
	 * Führt den eigentlichen Download aus
	 *
	 * @return bool FALSE bei 403/404, TRUE bei Erfolg (200/304)
	 * @throws DownloaderException bei anderen Fehlern
	 * @throws ConfigNotWritableException Wenn das Umbenennen der temporären Datei fehlschlägt
	 */
	private function downloadInner(string $url, $f, string $tmpfile, string $destfile, $maxagef, $mtime, $maxage): bool
	{
		$curl = new Curl();

		// If-Not-Modified-Since setzen
		if ($mtime) {
			$curl->setOpt(CURLOPT_TIMEVALUE, $mtime);
			$curl->setOpt(CURLOPT_TIMECONDITION, CURL_TIMECOND_IFMODSINCE);
		}

		// Direkten Download in geöffnete Datei durchführen
		ftruncate($f, 0);
		fseek($f, 0, SEEK_SET);
		$curl->setOpt(CURLOPT_FILE, $f);
		$curl->get($url.'?key='.self::APIKEY);

		// Status prüfen
		$status = $curl->getHttpStatus();
		if ($status == 304) {
			// Wenn "Not Modified": mtime anpassen, sodass es in spätestens 12 Stunden noch mal versucht wird
			touch($destfile, $this->floorInt(time()-max($maxage-43200, 0), 2));
			clearstatcache(false, $destfile);
			// Falls die Datei während des Checks beschädigt wurde, löschen und nächsten Aufruf neu herunterladen lassen.
			if (@filesize($destfile) == 0) {
				@unlink($destfile);
			}
			@unlink($tmpfile);
		}
		elseif ($status == 200) {
			// Wenn "Ok": Intervall merken und umbenennen

			// Max-Age-Intervall aus Headern extrahieren und speichern
			$cacheinfo = $curl->getResponseHeaders('Cache-Control') ?: '';
			$maxage_match = [];
			preg_match('/(?:^|\s|,)max-age=([0-9]+)/i', $cacheinfo, $maxage_match);
			if ($maxage_match) {
				ftruncate($maxagef, 0);
				fwrite($maxagef, $maxage_match[1]);
			} else {
				// Default to 24 h
				ftruncate($maxagef, 0);
				fwrite($maxagef, '86400');
			}

			// Sicherstellen, dass die Datei fertig geschrieben wurde
			$curl->close();
			fflush($f);
			fflush($maxagef);

			// Prüfen, ob die Datei dekodierbar ist
			fseek($f, 0, SEEK_SET);
			if (json_decode(stream_get_contents($f)) === null) {
				throw new DownloaderHTTPException($url, 900);
			}

			// Umbenennen
			if (!rename($tmpfile, $destfile)) {
				throw new ConfigNotWritableException($destfile);
			}

			// Mtime setzen
			$mtime = @strtotime($curl->getResponseHeaders('Last-Modified'));
			if ($mtime) {
				touch($destfile, max($mtime, $this->floorInt(time() - 43200, 2)));
			}
		}
		elseif ($status == 403 or $status == 404) {
			return false;
		}
		else {
			// Exception bei anderen HTTP-Fehlern
			throw new DownloaderHTTPException($url, $curl->getHttpStatus());
		}
		return true;
	}

	/**
	 * Lädt eine neue Version einer Datei von der EmbeddingDb herunter.
	 *
	 * Wenn parallel schon ein anderer Prozess die Datei herunterlädt,
	 * wird, wenn es noch keine gecachte Datei gibt, der Prozess abgewartet,
	 * und ansonsten direkt FALSE zurück gegeben.
	 *
	 * @return bool FALSE, wenn ein paralleler Prozess schon ein Update durchführt,
	 *              das nicht abgewartet wurde.
	 * @throws DownloaderException
	 */
	protected function download(): bool
	{
		$urls = $this->getDownloadUrls();
		if (!$urls or empty($urls[0])) {
			throw new DownloaderException('no url found', 404);
		}
		$destfile = $this->getPathname();
		$maxagefile = $destfile.'.maxage';
		$tmpfile = $destfile.'.'.getmypid().'-'.rand().'.tmp';
		$mtime = (file_exists($destfile) and @filesize($destfile)) ? @filemtime($destfile) : null;
		$maxage = (int)(@file_get_contents($maxagefile) ?: 86400);

		$lockf = @fopen($maxagefile, 'cb');
		if (!$lockf) {
			throw new ConfigNotWritableException($maxagefile);
		}
		// Versuchen, die MaxAge-Datei exklusiv zu sperren
		if (flock($lockf, LOCK_EX | LOCK_NB)) {
			clearstatcache(false, $maxagefile);
			if (fstat($lockf)['ino'] !== stat($maxagefile)['ino']) {
				// Datei ersetzt während wir sperren wollten.. Race-Condition?
				return false;
			}

			$f = @fopen($tmpfile, 'wb+');
			if (!$f) {
				throw new ConfigNotWritableException($tmpfile);
			}

			try {
				foreach ($urls as $url) {
					$success = false;
					if (($success = $this->downloadInner($url, $f, $tmpfile, $destfile, $lockf, $mtime, $maxage))) {
						break;
					}
				}
				if (!$success) {
					throw new DownloaderException('no url for regulations data reachable', 404);
				}
			}
			catch (DownloaderHTTPException $e) {
				@unlink($tmpfile);
				// HTTP-Fehler ignorieren, solange noch eine alte Version der Datei existiert
				if (file_exists($destfile)) {
					return false;
				} else {
					throw $e;
				}
			}
			catch (\Throwable $e) {
				@unlink($tmpfile);
				throw $e;
			}
			finally {
				if ($f !== null) {
					flock($lockf, LOCK_UN);
					fclose($lockf);
				}
			}
			return true;
		}
		// Wenn noch keine Zieldatei vorhanden, aber ein anderer Update-Prozess läuft, diesen abwarten
		elseif (!file_exists($destfile)) {
			flock($lockf, LOCK_SH);
			flock($lockf, LOCK_UN);
			fclose($lockf);
			clearstatcache(false, $maxagefile);
			return file_exists($destfile);
		}
		// Ansonsten einfach abbrechen
		else {
			return false;
		}
	}

	/**
	 * Diese Methode liest die Datendatei erneut von der Festplatte oder aus dem Internet ein.
	 *
	 * @param bool $mayUpdate Ob ein Update aus dem Internet durchgeführt werden darf
	 *
	 * @throws ConfigNotReadableException
	 * @throws DownloaderHTTPException
	 */
	public function reload(bool $mayUpdate=true): void
	{
		$path = $this->getPathname();

		if ($mayUpdate && $this->isStale()) {
			$this->download();
		}

		// Datei öffnen
		try {
			$content = file_get_contents($path);
			// Leere Dateien löschen. Sollte im Normalbetrieb nicht auftreten.
			if ($content === '') {
				@unlink($path);
			}
		}
		catch (\ErrorException $e) {
			throw new ConfigNotReadableException($path, '', 0, $e);
		}
		$data = json_decode($content, true);
		if (!$data and $mayUpdate) {
			// Fehlerhafter Download? Noch mal versuchen
			@unlink($path);
			$this->download();
		}
		$this->data = $data ?? [];
	}

}
