<?php

namespace App\Component;

use App\Config;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\EmbeddingDb\EmbeddingDbLegacy;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingBase;
use App\Model\EmbeddingAssetBase;
use App\Model\Language;
use App\Model\Purpose;
use App\Model\SystemEmbedding;
use App\Model\SystemEmbeddingAsset;
use App\Model\User;
use App\Model\UserEmbedding;
use App\Model\UserEmbeddingAsset;
use App\Model\UserLocale;
use InvalidArgumentException;

final class EmbeddingUpdateService
{
	/** @var EmbeddingDb $db */
	private $db;
	/** @var EmbeddingDbLegacy $legacyDb */
	private $legacyDb;

	/** @var array<string,string> $GENERAL_MAPPING */
	private static $GENERAL_MAPPING = [
		'name'=>'Name',
		'code_snippet'=>'CodeSnippet',
		'embedding_vendor'=>'EmbeddingVendor',
		'script_loader_group' => 'ScriptLoaderGroup',
		'structured_legal_basis' => 'StructuredLegalBasis'
	];

	/** @var array<string,string> $TRANSLATION_MAPPING */
	private static $TRANSLATION_MAPPING = [
		'label'=>'Name',
		'description'=>'Description',
		'privacy_policy_url'=>'PrivacyPolicyUrl',
		'collected_data_info'=>'CollectedDataInfo',
		'purpose_of_data_collection'=>'PurposeOfDataCollection',
		'legal_basis'=>'LegalBasis',
		'legal_basis_supplement'=>'LegalBasisSupplement',
		'place_of_processing'=>'PlaceOfProcessing',
	];

	public function __construct(EmbeddingDb $db, EmbeddingDbLegacy $legacyDb) {
		$this->db = $db;
		$this->legacyDb = $legacyDb;
	}

	/**
	 * Versucht ein Embedding zu aktualisieren.
	 *
	 * EmbeddingAsset::flush()/SystemEmbeddingAsset::flush() und $embedding->save() werden *nicht*
	 * ausgeführt. Committen liegt in der Verantwortung des Aufrufenden.
	 *
	 * @param EmbeddingBase $embedding
	 * @param bool $force Wenn TRUE wird immer überschrieben, egal ob die Quelle neuer oder älter ist
	 * @return bool Ob ein Update durchgeführt wurde
	 * @throws InvalidArgumentException
	 */
	public function updateEmbedding(EmbeddingBase $embedding, bool $force=false): bool
	{
		if (!$force and !$embedding->getSourceTimestamp()) {
			return false;
		}

		switch ($embedding->getSource()) {
			case EmbeddingBase::SOURCE_GLOBAL_DB:
				return $this->sourceGlobalDatabase($embedding, $force);

			case EmbeddingBase::SOURCE_GLOBAL_DB_LEGACY:
				return $this->sourceGlobalDatabaseLegacy($embedding, $force);

			case EmbeddingBase::SOURCE_SYSTEM_TEMPLATE:
				return $this->sourceSystemTemplate($embedding, $force);

			case EmbeddingBase::SOURCE_USER_TEMPLATE:
				return $this->sourceUserTemplate($embedding, $force);

			default:
				return false;
		}
	}

	/**
	 * @param EmbeddingBase $embedding
	 * @param bool $force
	 * @return bool
	 */
	private function sourceGlobalDatabase(EmbeddingBase $embedding, bool $force=false): bool
	{
		$remoteEmbedding = $this->db->fetchEmbeddingIfModified((int)$embedding->getSourceId(), ($force) ? null : $embedding->getSourceTimestamp(), $this->listMissingEmbeddingTranslations($embedding));

		// Prüfen ob der Eintrag auf eine neue Quelle umgestellt werden soll - aber nur bei Domain-Embeddings
		if ($embedding instanceof Embedding) {
			$id = $embedding->getId();
			$upgradeSource = $this->db->getLastUpgradeSource();
			if ($upgradeSource and ($remoteEmbedding or $remoteEmbedding === null)) {
				list($source, $sourceId) = $upgradeSource;
				if ($embedding->getSource() !== $source or $embedding->getSourceId() !== $sourceId) {
					// Prüfen, ob eine Doppelung auftritt
					$others = array_filter(Embedding::all(), function ($embedding) use ($id, $source, $sourceId) {
						return $embedding->getId() !== $id
							and $embedding->getSource() === $source
							and $embedding->getSourceId() === $sourceId
							and $embedding->isActive()
							and $embedding->isAutoUpdated();
					});

					// Es gibt schon ein anderes Embedding, das genau diesen Eintrag aus der DB abbildet.
					// Dieses Embedding hier kann gelöscht werden.
					if ($others) {
						Embedding::delete($id, false);
						return true;
					}
					else {
						// Neue Quelle eintragen
						$embedding->setSource($source);
						$embedding->setSourceId($sourceId);

						if ($source === EmbeddingBase::SOURCE_GLOBAL_DB) {
							// Neue Daten abfragen
							$remoteEmbedding = $this->db->fetchEmbedding((int)$sourceId);
						}
					}

				}
			}
		}

		// Unverändert oder Fehler → überspringen
		if (!$remoteEmbedding) {
			return false;
		}

		$this->updateGeneralFromDb($embedding, $remoteEmbedding);
		$this->updateDefaultTranslationFromDb($embedding, $remoteEmbedding);
		$this->updateTranslationsFromDb($embedding, $remoteEmbedding);
		$this->updateAssetsFromDb($embedding, $remoteEmbedding);
		$embedding->setSourceTimestamp($remoteEmbedding['timestamp']);
		$embedding->updateLastModified();

		return true;
	}

	/**
	 * @param EmbeddingBase $embedding
	 * @param bool $force
	 * @return bool
	 */
	private function sourceGlobalDatabaseLegacy(EmbeddingBase $embedding, bool $force=false): bool
	{
		$remoteEmbedding = $this->legacyDb->fetchEmbeddingIfModified((int)$embedding->getSourceId(), ($force) ? null : $embedding->getSourceTimestamp(), $this->listMissingEmbeddingTranslations($embedding));

		// Prüfen ob der Eintrag auf eine neue Quelle umgestellt werden soll
		$upgradeSource = $this->legacyDb->getLastUpgradeSource();
		if ($upgradeSource and ($remoteEmbedding or $remoteEmbedding === null)) {
			list($source, $sourceId) = $upgradeSource;
			if ($embedding->getSource() !== $source or $embedding->getSourceId() !== $sourceId) {
				// Neue Quelle eintragen
				$embedding->setSource($source);
				$embedding->setSourceId($sourceId);

				if ($source === EmbeddingBase::SOURCE_GLOBAL_DB) {
					// Neue Daten abfragen
					$remoteEmbedding = $this->db->fetchEmbedding((int)$sourceId);
				}

			}
		}

		// Unverändert oder Fehler → überspringen
		if (!$remoteEmbedding) {
			return false;
		}

		$this->updateGeneralFromDb($embedding, $remoteEmbedding);
		$this->updateDefaultTranslationFromDb($embedding, $remoteEmbedding);
		$this->updateTranslationsFromDb($embedding, $remoteEmbedding);
		$this->updateAssetsFromDb($embedding, $remoteEmbedding);
		$embedding->setSourceTimestamp($remoteEmbedding['timestamp']);
		$embedding->updateLastModified();

		return true;
	}

	/**
	 * @param Embedding|UserEmbedding $embedding
	 * @param bool $force
	 * @return bool
	 */
	private function sourceSystemTemplate(EmbeddingBase $embedding, bool $force=false): bool
	{
		// Aktuell können nur Domain- und User-Embeddings von System-Embeddings abgeleitet sein
		if (!($embedding instanceof Embedding) and !($embedding instanceof UserEmbedding)) {
			return false;
		}
		$systemEmbedding = SystemEmbedding::find((string)$embedding->getSourceId());

		// Nicht gefunden → überspringen
		if (!$systemEmbedding) {
			return false;
		}

		$missingLanguages = $this->listMissingEmbeddingTranslations($embedding);
		$foundLanguages = $this->listExistingEmbeddingTranslations($systemEmbedding);
		$fetchableLanguages = array_intersect($missingLanguages, $foundLanguages);

		// Unverändert oder nicht gefunden → überspringen
		if ((!$force and $systemEmbedding->getLastModified() <= $embedding->getSourceTimestamp()) and !$fetchableLanguages) {
			return false;
		}

		$this->updateGeneralFromTemplate($embedding, $systemEmbedding);
		$this->updateDefaultTranslationFromTemplate($embedding, $systemEmbedding);
		$this->updateTranslationsFromTemplate($embedding, $systemEmbedding);
		$this->updateAssetsFromTemplate($embedding, $systemEmbedding);
		$embedding->setSourceTimestamp($systemEmbedding->getLastModified());
		$embedding->updateLastModified();

		return true;
	}

	/**
	 * @param Embedding $embedding
	 * @param bool $force
	 * @return bool
	 */
	private function sourceUserTemplate(EmbeddingBase $embedding, bool $force=false): bool
	{
		// Aktuell können nur Domain-Embeddings von User-Embeddings abgeleitet sein
		if (!($embedding instanceof Embedding)) {
			return false;
		}
		$userEmbedding = UserEmbedding::find((string)$embedding->getSourceId());

		// Nicht gefunden → überspringen
		if (!$userEmbedding) {
			return false;
		}

		$missingLanguages = $this->listMissingEmbeddingTranslations($embedding);
		$foundLanguages = $this->listExistingEmbeddingTranslations($userEmbedding);
		$fetchableLanguages = array_intersect($missingLanguages, $foundLanguages);

		// Unverändert oder nicht gefunden → überspringen
		if ((!$force and $userEmbedding->getLastModified() <= $embedding->getSourceTimestamp()) and !$fetchableLanguages) {
			return false;
		}

		$this->updateGeneralFromTemplate($embedding, $userEmbedding);
		$this->updateDefaultTranslationFromTemplate($embedding, $userEmbedding);
		$this->updateTranslationsFromTemplate($embedding, $userEmbedding);
		$this->updateAssetsFromTemplate($embedding, $userEmbedding);
		$embedding->setSourceTimestamp($userEmbedding->getLastModified());
		$embedding->updateLastModified();

		return true;
	}

	/**
	 * @return EmbeddingAsset|UserEmbeddingAsset|SystemEmbeddingAsset
	 * @phan-assert Embedding|UserEmbedding|SystemEmbedding $embedding
	 */
	private function newAsset(EmbeddingBase $embedding): EmbeddingAssetBase
	{
		if ($embedding instanceof Embedding) {
			return EmbeddingAsset::create();
		}
		elseif ($embedding instanceof UserEmbedding) {
			return UserEmbeddingAsset::create();
		}
		elseif ($embedding instanceof SystemEmbedding) {
			return SystemEmbeddingAsset::create();
		}
		else {
			throw new InvalidArgumentException('Embedding type not supported by EmbeddingUpdateService');
		}
	}

	/**
	 * Sprachunabhängige Daten von SystemEmbedding übernehmen
	 */
	private function updateGeneralFromTemplate(EmbeddingBase $embedding, EmbeddingBase $template): void
	{
		// Purpose übernehmen
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_PURPOSE)) {
			$purpose = $template->getPurpose();
			if ($purpose && (!$embedding->getPurpose() || !$embedding->getPurpose()->isMandatory())) {
				$embedding->setPurpose($purpose);
			}
		}

		// Generelle Daten übernehmen
		$generalMapping = array_filter(self::$GENERAL_MAPPING, static function (string $key) use ($embedding) {
			return $embedding->isUpdateParameterAllowed($key);
		}, ARRAY_FILTER_USE_KEY);

		foreach ($generalMapping as $property) {
			$value = call_user_func([$template, 'get'.$property]);
			if ($value != '') {
				call_user_func([$embedding, 'set'.$property], $value);
			}
		}

		// Strukturierte Rechtsgrundlage (de-)aktivieren
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_STRUCTURED_LEGAL_BASIS)) {
			$embedding->setLegalBasisStructured($template->isLegalBasisStructured());
		}

		// Code-Blocking übernehmen
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_CODE_BLOCK)) {
			$codeBlock = $template->getCodeBlock();
			if ($codeBlock) {
				$embedding->setCodeBlock($codeBlock);
			}
		}

		// URL-Excluding übernehmen
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_EXCLUDE_URLS)) {
			$excludeUrls = $template->getExcludeUrls();
			if ($excludeUrls) {
				$embedding->setExcludeUrls($excludeUrls);
			}
		}

		// Iframe-Blocking übernehmen
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_IFRAMES_BLOCK)) {
			$iframeBlock = $template->getIframeBlockingStrings();
			if ($iframeBlock) {
				$embedding->setIframeBlockingStrings($iframeBlock);
			}
		}

		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_IFRAMES_HIDE)) {
			$embedding->setIframesHiddenUntilConsent($template->isIframesHiddenUntilConsent());
		}

		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_IFRAMES_LOAD_EVENT)) {
			$embedding->setDispatchLoadEventForIframes($template->isDispatchLoadEventForIframes());
		}

		$embedding->setTcfVendor($template->getTcfVendor());
		$embedding->setTcfPurposes($template->getTcfPurposes());
		$embedding->setTcfSpecialFeatures($template->getTcfSpecialFeatures());
	}

	/**
	 * Standard-Übersetzung von SystemEmbedding übernehmen
	 *
	 * @param Embedding|UserEmbedding $embedding
	 * @param SystemEmbedding|UserEmbedding $template
	 */
	private function updateDefaultTranslationFromTemplate(EmbeddingBase $embedding, EmbeddingBase $template): void
	{
		// Standard-Übersetzung anhand des Standard-Locales übernehmen
		$locale = ($embedding instanceof UserEmbedding) ? LocaleService::getUserDefault() : LocaleService::getDefault();
		$language = ($template instanceof UserEmbedding) ? UserLocale::find($locale->getName()) : ($locale->getLanguage() ?? Language::find('en_US'));

		if ($language) {
			foreach (self::$TRANSLATION_MAPPING as $key => $property) {
				$value = call_user_func([$template, 'getTranslated'.$property], $language, true);
				if ($value != '' and $key !== 'label') {
					call_user_func([$embedding, 'set'.$property], $value);
				}
			}
		}
	}

	/**
	 * Übersetzungen von System-/UserEmbedding übernehmen
	 *
	 * Übersetzungen werden nur berücksichtigt, wenn sie nicht leer sind.
	 *
	 * @param Embedding|UserEmbedding $embedding
	 * @param SystemEmbedding|UserEmbedding $template
	 */
	private function updateTranslationsFromTemplate(EmbeddingBase $embedding, EmbeddingBase $template): void
	{
		// Standard-Übersetzung anhand des Standard-Locales übernehmen
		$locales = ($embedding instanceof Embedding) ? LocaleService::all() : UserLocale::all();

		foreach ($locales as $locale) {
			$language = ($template instanceof UserEmbedding) ? UserLocale::find($locale->getName()) : $locale->getLanguage();
			if ($language) {
				foreach (self::$TRANSLATION_MAPPING as $property) {
					$value = call_user_func([$template, 'getTranslated'.$property], $language, false);
					if ($value != '') {
						call_user_func([$embedding, 'setTranslated'.$property], $locale, $value);
					}
				}
			}
		}
	}

	/**
	 * Assets von System-/UserEmbedding übernehmen
	 *
	 * @param Embedding|UserEmbedding $embedding
	 * @param SystemEmbedding|UserEmbedding $template
	 */
	private function updateAssetsFromTemplate(EmbeddingBase $embedding, EmbeddingBase $template): void
	{
		$assets = $embedding->getAssets();

		if (!$assets) {
			return;
		}

		$assets = array_combine(array_map(function (EmbeddingAssetBase $e) { return $e->getName(); }, $assets), $assets);
		$templateAssets = $template->getAssets();
		$templateAssets = array_combine(array_map(function (EmbeddingAssetBase $a) { return $a->getName(); }, $templateAssets), $templateAssets);
		$deletedAssets = array_diff_key($assets, $templateAssets);

		$resultingAssets = [];

		foreach ($templateAssets as $assetName => $templateAsset) {
			$asset = $assets[$assetName] ?? $this->newAsset($embedding);
			assert($asset instanceof EmbeddingAsset or $asset instanceof UserEmbeddingAsset);
			assert($templateAsset instanceof SystemEmbeddingAsset or $templateAsset instanceof UserEmbeddingAsset);

			// Generelle Asset-Daten
			$asset->setName($templateAsset->getName());
			$asset->setDynamic($templateAsset->isDynamic());
			$asset->setStorageType($templateAsset->getStorageType());

			// Asset-Standardübersetzung
			// Standard-Übersetzung anhand des Standard-Locales übernehmen
			$locale = ($embedding instanceof Embedding) ? LocaleService::getDefault() : LocaleService::getUserDefault();
			$language = ($templateAsset instanceof UserEmbeddingAsset) ? UserLocale::find($locale->getName()) : $locale->getLanguage();
			if ($language) {
				$value = $templateAsset->getTranslatedLifetime($language, true);
				if ($value != '') {
					$asset->setLifetime($value);
				}
				$value = $templateAsset->getTranslatedDescription($language, true);
				if ($value != '') {
					$asset->setDescription($value);
				}
			}

			// Asset-Übersetzungen
			$locales = ($embedding instanceof Embedding) ? LocaleService::all() : UserLocale::all();
			foreach ($locales as $locale) {
				$language = ($templateAsset instanceof UserEmbeddingAsset) ? UserLocale::find($locale->getName()) : $locale->getLanguage();
				if (!$language) {
					continue;
				}
				$value = $templateAsset->getTranslatedLifetime($language, false);
				if ($value != '') {
					$asset->setTranslatedLifetime($locale, $value);
				}
				$value = $templateAsset->getTranslatedDescription($language, false);
				if ($value != '') {
					$asset->setTranslatedDescription($locale, $value);
				}
			}

			$resultingAssets[] = $asset;
		}

		$embedding->setAssets($resultingAssets);

		foreach ($deletedAssets as $asset) {
			$asset->delete($asset->getId());
		}
	}

	/**
	 * Sprachunabhängige Daten von EmbeddingDB-Daten übernehmen
	 */
	private function updateGeneralFromDb(EmbeddingBase $embedding, array $remoteEmbedding): void
	{
		// Purpose übernehmen
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_PURPOSE)) {
			$purpose = Purpose::find($remoteEmbedding['purpose']);
			if ($purpose && (!$embedding->getPurpose() || !$embedding->getPurpose()->isMandatory())) {
				$embedding->setPurpose($purpose);
			}
		}

		// Generelle Daten übernehmen
		$generalMapping = array_filter(self::$GENERAL_MAPPING, static function (string $key) use ($embedding) {
			return $embedding->isUpdateParameterAllowed($key);
		}, ARRAY_FILTER_USE_KEY);

		foreach ($generalMapping as $key => $setter) {
			if (($remoteEmbedding[$key] ?? '') != '') {
				call_user_func([$embedding, 'set'.$setter], $remoteEmbedding[$key]);
			}
		}

		// Strukturierte Rechtsgrundlage aktivieren, wenn entsprechende Daten vorhanden
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_STRUCTURED_LEGAL_BASIS)) {
			$embedding->setLegalBasisStructured(isset($remoteEmbedding['structured_legal_basis']));
		}

		// Code-Blocking übernehmen
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_CODE_BLOCK)) {
			$codeBlock = explode("\n", trim((string)($remoteEmbedding['code_block'] ?? '')));
			$codeBlock = array_filter($codeBlock, function ($line) {
				return trim($line) !== '';
			});
			$codeBlock = array_values(array_map(function ($line) {
				return trim($line);
			}, $codeBlock));
			if ($codeBlock) {
				$embedding->setCodeBlock($codeBlock);
			}
		}

		// Iframe-Blocking übernehmen
		if ($embedding->isUpdateParameterAllowed(EmbeddingBase::FIELD_IFRAMES_BLOCK)) {
			$iframeBlock = explode("\n", trim((string)($remoteEmbedding['iframe_blocking_strings'] ?? '')));
			$iframeBlock = array_filter($iframeBlock, function ($line) {
				return trim($line) !== '';
			});
			$iframeBlock = array_values(array_map(function ($line) {
				return trim($line);
			}, $iframeBlock));
			if ($iframeBlock) {
				$embedding->setIframeBlockingStrings($iframeBlock);
			}
		}
	}

	/**
	 * Standard-Übersetzung von EmbeddingDB-Daten übernehmen
	 *
	 * Bei Domain-Embeddings wird das Standard-Locale berücksichtigt. Ansonsten wird Englisch genommen.
	 */
	private function updateDefaultTranslationFromDb(EmbeddingBase $embedding, array $remoteEmbedding): void
	{
		if ($embedding instanceof Embedding) {
			// Standard-Übersetzung anhand des Standard-Locales übernehmen
			$locale = LocaleService::getDefault();
			$language = $locale->getLanguage();
			$languageCode = ($language) ? $language->getShortCode() : 'en';

			$defaultTranslation = $remoteEmbedding['translations'][$languageCode] ?? null;
		}
		elseif ($embedding instanceof UserEmbedding) {
			// Standard-Übersetzung anhand des Standard-Locales übernehmen
			$locale = LocaleService::getUserDefault();
			$language = $locale->getLanguage();
			$languageCode = ($language) ? $language->getShortCode() : 'en';

			$defaultTranslation = $remoteEmbedding['translations'][$languageCode] ?? null;
		}
		else {
			// Standard-Übersetzung EN übernehmen
			$defaultTranslation = $remoteEmbedding['translations']['en'] ?? null;
		}

		if ($defaultTranslation) {
			foreach (self::$TRANSLATION_MAPPING as $key=>$setter) {
				if (($defaultTranslation[$key] ?? '') != '' and $key !== 'label') {
					call_user_func([$embedding, 'set'.$setter], $defaultTranslation[$key]);
				}
			}
		}
	}

	/**
	 * @return Language[]
	 */
	private function listExistingEmbeddingTranslations(EmbeddingBase $embedding): array
	{
		$existingLanguages = [];
		if ($embedding instanceof Embedding) {
			foreach (LocaleService::all() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) {
					continue;
				}
				if (
					$embedding->getTranslatedDescription($locale, false) !== ''
					or $embedding->getTranslatedPrivacyPolicyUrl($locale, false) !== ''
					or $embedding->getTranslatedCollectedDataInfo($locale, false) !== ''
					or $embedding->getTranslatedPurposeOfDataCollection($locale, false) !== ''
					or $embedding->getTranslatedLegalBasis($locale, false) !== ''
					or $embedding->getTranslatedPlaceOfProcessing($locale, false) !== ''
				) {
					$existingLanguages[] = $language->getShortCode();
				}
			}
		}
		elseif ($embedding instanceof UserEmbedding) {
			$user = User::loggedInUser();
			foreach ($user->getUserEmbeddingLocales() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) {
					continue;
				}
				if (
					$embedding->getTranslatedDescription($locale, false) !== ''
					or $embedding->getTranslatedPrivacyPolicyUrl($locale, false) !== ''
					or $embedding->getTranslatedCollectedDataInfo($locale, false) !== ''
					or $embedding->getTranslatedPurposeOfDataCollection($locale, false) !== ''
					or $embedding->getTranslatedLegalBasis($locale, false) !== ''
					or $embedding->getTranslatedPlaceOfProcessing($locale, false) !== ''
				) {
					$existingLanguages[] = $language->getShortCode();
				}
			}
		}
		elseif ($embedding instanceof SystemEmbedding) {
			$config = Config::getInstance();
			foreach ($config->getSystemEmbeddingLanguages() as $language) {
				if (
					$embedding->getTranslatedDescription($language, false) !== ''
					or $embedding->getTranslatedPrivacyPolicyUrl($language, false) !== ''
					or $embedding->getTranslatedCollectedDataInfo($language, false) !== ''
					or $embedding->getTranslatedPurposeOfDataCollection($language, false) !== ''
					or $embedding->getTranslatedLegalBasis($language, false) !== ''
					or $embedding->getTranslatedPlaceOfProcessing($language, false) !== ''
				) {
					$existingLanguages[] = $language->getShortCode();
				}
			}
		}
		return $existingLanguages;
	}

	/**
	 * @return string[] Short-Code
	 */
	private function listMissingEmbeddingTranslations(EmbeddingBase $embedding): array
	{
		$missingLanguages = [];
		if ($embedding instanceof Embedding) {
			foreach (LocaleService::enabledLocales() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) {
					continue;
				}
				if (
					$embedding->getTranslatedDescription($locale, false) === ''
					and $embedding->getTranslatedPrivacyPolicyUrl($locale, false) === ''
					and $embedding->getTranslatedCollectedDataInfo($locale, false) === ''
					and $embedding->getTranslatedPurposeOfDataCollection($locale, false) === ''
					and $embedding->getTranslatedLegalBasis($locale, false) === ''
					and $embedding->getTranslatedPlaceOfProcessing($locale, false) === ''
				) {
					$missingLanguages[] = $language->getShortCode();
				}
			}
		}
		elseif ($embedding instanceof UserEmbedding) {
			$user = User::loggedInUser();
			foreach ($user->getUserEmbeddingLocales() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) {
					continue;
				}
				if (
					$embedding->getTranslatedDescription($locale, false) === ''
					and $embedding->getTranslatedPrivacyPolicyUrl($locale, false) === ''
					and $embedding->getTranslatedCollectedDataInfo($locale, false) !== ''
					and $embedding->getTranslatedPurposeOfDataCollection($locale, false) === ''
					and $embedding->getTranslatedLegalBasis($locale, false) === ''
					and $embedding->getTranslatedPlaceOfProcessing($locale, false) === ''
				) {
					$missingLanguages[] = $language->getShortCode();
				}
			}
		}
		elseif ($embedding instanceof SystemEmbedding) {
			$config = Config::getInstance();
			foreach ($config->getSystemEmbeddingLanguages() as $language) {
				if (
					$embedding->getTranslatedDescription($language, false) === ''
					and $embedding->getTranslatedPrivacyPolicyUrl($language, false) === ''
					and $embedding->getTranslatedCollectedDataInfo($language, false) !== ''
					and $embedding->getTranslatedPurposeOfDataCollection($language, false) === ''
					and $embedding->getTranslatedLegalBasis($language, false) === ''
					and $embedding->getTranslatedPlaceOfProcessing($language, false) === ''
				) {
					$missingLanguages[] = $language->getShortCode();
				}
			}
		}
		return $missingLanguages;
	}

	/**
	 * Übersetzungen von EmbeddingDB-Daten übernehmen
	 *
	 * Übersetzungen werden nur berücksichtigt, wenn sie nicht leer sind.
	 *
	 * @phan-assert Embedding|SystemEmbedding $embedding
	 */
	private function updateTranslationsFromDb(EmbeddingBase $embedding, array $remoteEmbedding): void
	{
		if ($embedding instanceof Embedding) {
			foreach (LocaleService::all() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) {
					continue;
				}
				$shortCode = $language->getShortCode();
				$translation = $remoteEmbedding['translations'][$shortCode] ?? null;
				if ($translation and !empty($translation['description'])) {
					foreach (self::$TRANSLATION_MAPPING as $key=>$setter) {
						if (($translation[$key] ?? '') != '') {
							call_user_func([$embedding, 'setTranslated'.$setter], $locale, $translation[$key]);
						}
					}
				}
			}
		}
		elseif ($embedding instanceof UserEmbedding) {
			foreach (UserLocale::all() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) {
					continue;
				}
				$shortCode = $language->getShortCode();
				$translation = $remoteEmbedding['translations'][$shortCode] ?? null;
				if ($translation and !empty($translation['description'])) {
					foreach (self::$TRANSLATION_MAPPING as $key=>$setter) {
						if (($translation[$key] ?? '') != '') {
							call_user_func([$embedding, 'setTranslated'.$setter], $locale, $translation[$key]);
						}
					}
				}
			}
		}
		elseif ($embedding instanceof SystemEmbedding) {
			foreach (Language::all() as $language) {
				$shortCode = $language->getShortCode();
				$translation = $remoteEmbedding['translations'][$shortCode] ?? null;
				if ($translation and !empty($translation['description'])) {
					foreach (self::$TRANSLATION_MAPPING as $key=>$setter) {
						if (($translation[$key] ?? '') != '') {
							call_user_func([$embedding, 'setTranslated'.$setter], $language, $translation[$key]);
						}
					}
				}
			}
		}
		else {
			throw new InvalidArgumentException('Embedding type not supported by EmbeddingUpdateService');
		}
	}

	/**
	 * Assets von EmbeddingDB-Daten übernehmen
	 */
	private function updateAssetsFromDb(EmbeddingBase $embedding, array $remoteEmbedding): void
	{
		if (empty($remoteEmbedding['assets'])) {
			return;
		}

		$assets = $embedding->getAssets();
		$assets = array_combine(array_map(function (EmbeddingAssetBase $e) { return $e->getName(); }, $assets), $assets);
		$remoteAssets = array_combine(array_map(function (array $a) { return $a['name']; }, $remoteEmbedding['assets']), $remoteEmbedding['assets']);
		$deletedAssets = array_diff_key($assets, $remoteAssets);

		$resultingAssets = [];

		foreach ($remoteAssets as $assetName => $remoteAsset) {
			$asset = $assets[$assetName] ?? $this->newAsset($embedding);

			// Generelle Asset-Daten
			$asset->setName($remoteAsset['name']);
			$asset->setDynamic($remoteAsset['dynamic'] ?? false);
			$asset->setStorageType($remoteAsset['storageType']);

			// Asset-Standardübersetzung
			if ($embedding instanceof Embedding) {
				// Standard-Übersetzung anhand des Standard-Locales übernehmen
				$locale = LocaleService::getDefault();
				$language = $locale->getLanguage();
				$languageCode = ($language) ? $language->getShortCode() : 'en';
				$defaultTranslation = $remoteAsset['translations'][$languageCode] ?? null;
			}
			elseif ($embedding instanceof UserEmbedding) {
				// Standard-Übersetzung anhand des Standard-Locales übernehmen
				$locale = LocaleService::getUserDefault();
				$language = $locale->getLanguage();
				$languageCode = ($language) ? $language->getShortCode() : 'en';
				$defaultTranslation = $remoteAsset['translations'][$languageCode] ?? null;
			}
			else {
				// Standard-Übersetzung EN übernehmen
				$defaultTranslation = $remoteAsset['translations']['en'] ?? null;
			}
			if ($defaultTranslation) {
				if (($defaultTranslation['lifetime'] ?? '') != '') {
					$asset->setLifetime($defaultTranslation['lifetime']);
				}
				if (($defaultTranslation['description'] ?? '') != '') {
					$asset->setDescription($defaultTranslation['description']);
				}
			}

			// Asset-Übersetzungen
			if ($embedding instanceof Embedding) {
				assert($asset instanceof EmbeddingAsset);
				foreach (LocaleService::all() as $locale) {
					$language = $locale->getLanguage();
					if (!$language) {
						continue;
					}
					$shortCode = $language->getShortCode();
					$translation = $remoteAsset['translations'][$shortCode] ?? null;
					if ($translation) {
						if (($translation['lifetime'] ?? '') != '') {
							$asset->setTranslatedLifetime($locale, $translation['lifetime']);
						}
						if (($translation['description'] ?? '') != '') {
							$asset->setTranslatedDescription($locale, $translation['description']);
						}
					}
				}
			}
			elseif ($embedding instanceof UserEmbedding) {
				assert($asset instanceof UserEmbeddingAsset);
				foreach (UserLocale::all() as $locale) {
					$language = $locale->getLanguage();
					if (!$language) {
						continue;
					}
					$shortCode = $language->getShortCode();
					$translation = $remoteAsset['translations'][$shortCode] ?? null;
					if ($translation) {
						if (($translation['lifetime'] ?? '') != '') {
							$asset->setTranslatedLifetime($locale, $translation['lifetime']);
						}
						if (($translation['description'] ?? '') != '') {
							$asset->setTranslatedDescription($locale, $translation['description']);
						}
					}
				}
			}
			elseif ($embedding instanceof SystemEmbedding) {
				assert($asset instanceof SystemEmbeddingAsset);
				foreach (Language::all() as $language) {
					$shortCode = $language->getShortCode();
					$translation = $remoteAsset['translations'][$shortCode] ?? null;
					if ($translation) {
						if (($translation['lifetime'] ?? '') != '') {
							$asset->setTranslatedLifetime($language, $translation['lifetime']);
						}
						if (($translation['description'] ?? '') != '') {
							$asset->setTranslatedDescription($language, $translation['description']);
						}
					}
				}
			}
			else {
				throw new InvalidArgumentException('Embedding type not supported by EmbeddingUpdateService');
			}

			$resultingAssets[] = $asset;
		}

		$embedding->setAssets($resultingAssets);

		foreach ($deletedAssets as $asset) {
			$asset->delete($asset->getId());
		}
	}
}
