<?php

namespace App\Component;

use Symfony\Component\HttpFoundation\Session\Session;

use Symfony\Component\HttpFoundation\Session\Flash\FlashBag;
use Symfony\Component\HttpFoundation\Session\Flash\FlashBagInterface;
use IteratorAggregate;

/**
 * @implements IteratorAggregate<string, mixed>
 */
class LazySession extends Session implements IteratorAggregate
{

	private function hasSessionCookie(): bool
	{
		return (isset($_COOKIE[$this->getName()]));
	}

	/**
	 * {@inheritdoc}
	 */
	public function invalidate($lifetime = null): bool
	{
		if ($lifetime !== null or $this->isStarted() or $this->hasSessionCookie()) {
			return parent::invalidate($lifetime);
		}
		return false;
	}

	/**
	 * {@inheritdoc}
	 */
	public function save(): void
	{
		if ($this->isStarted() or $this->hasSessionCookie()) {
			parent::save();
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function has($name): bool
	{
		if ($this->isStarted()) {
			return parent::has($name);
		} else {
			return false;
		}
	}

	/**
	 * {@inheritdoc}
	 * @return mixed
	 */
	#[\ReturnTypeWillChange]
	public function get($name, $default = null)
	{
		if ($this->isStarted() or $this->hasSessionCookie()) {
			return parent::get($name, $default);
		} else {
			return $default;
		}
	}

	/**
	 * {@inheritdoc}
	 * @return array<string,mixed>
	 */
	public function all(): array
	{
		if ($this->isStarted() or $this->hasSessionCookie()) {
			return parent::all();
		} else {
			return [];
		}
	}

	/**
	 * {@inheritdoc}
	 * @return mixed
	 */
	#[\ReturnTypeWillChange]
	public function remove($name)
	{
		if ($this->isStarted() or $this->hasSessionCookie()) {
			return parent::remove($name);
		} else {
			return null;
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function clear(): void
	{
		if ($this->isStarted() or $this->hasSessionCookie()) {
			parent::clear();
		}
	}

	/**
	 * Returns the number of attributes.
	 */
	public function count(): int
	{
		if (!$this->isStarted() and !$this->hasSessionCookie()) {
			return 0;
		}
		return parent::count();
	}

	/**
	 * Returns an iterator for attributes.
	 * @return \ArrayIterator<string, mixed>
	 */
	public function getIterator(): \Iterator
	{
		if (!$this->isStarted() and !$this->hasSessionCookie()) {
			return new \ArrayIterator([]);
		}
		return parent::getIterator();
	}

	/**
	 * {@inheritdoc}
	 *
	 * Returns an empty dummy bag if the session fails.
	 *
	 * @throws \Exception
	 */
	public function getFlashBag(): FlashBagInterface
	{
		try {
			return parent::getFlashBag();
		} catch (\Exception $e) {
			if (!$this->isStarted()) {
				return new FlashBag();
			} else {
				throw $e;
			}
		}
	}
}
