<?php
namespace App\Component\Licensing;

use App\Component\Curl;
use App\Component\CurlResponse;
use App\Exception\LicenseInvalidException;
use App\Exception\DownloaderException;
use App\Exception\DownloaderHTTPException;

/**
 * API-Client für Lizenz-API
 */
final class ApiClient {
	const API_URL = 'https://licence.ccm19.de/';

	/**
	 * @param string $endpoint
	 * @param array $params
	 * @param array $headers
	 * @return CurlResponse Response mit JSON-Array
	 * @throws \ErrorException Wird geschmissen, wenn cURL nicht zur Verfügung steht.
	 * @throws DownloaderException Wenn ein Problem in der Kommunikation mit dem API-Server auftritt.
	 */
	private function get(string $endpoint, array $params = [], array $headers = []): CurlResponse
	{
		$params = ['action' => $endpoint] + $params;

		$curl = new Curl();
		$curl->setOpt(CURLOPT_CONNECTTIMEOUT, 3);
		$curl->setOpt(CURLOPT_TIMEOUT, 7);

		foreach ($headers as $name => $value) {
			$curl->setHeader($name, $value);
		}

		$curl->get(self::API_URL, $params);

		if ($curl->curl_error) {
			$error = $curl->getErrorCode();
			throw new DownloaderException("An error has occurred in cURL module. (Code: $error) ".$curl->getErrorMessage(), $error);
		}

		$response = $curl->getResponseObject();

		if (!$response->isJson() or !is_array($response->getData())) {
			throw new DownloaderHTTPException('Unexpected response from license server.', $response->getStatusCode());
		}

		return $response;
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @return LicenseInfo Lizenzdaten bei Erfolg
	 * @throws DownloaderException Wenn ein Problem in der Kommunikation mit dem API-Server auftritt.
	 * @throws LicenseInvalidException Wenn ein Problem mit der Lizenz besteht
	 */
	public function registerLicense(string $licenseKey, string $domain): LicenseInfo
	{
		$response = $this->get('registerLicense', [
			'licence_id' => $licenseKey,
			'dom' => $domain,
		]);
		$data = $response->getData();
		assert(is_array($data));

		if ($data['ok'] ?? false) {
			return LicenseInfo::fromArray($data);
		} else {
			throw new LicenseInvalidException($data['error'] ?? '');
		}
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @throws DownloaderException Wenn ein Problem in der Kommunikation mit dem API-Server auftritt.
	 * @throws LicenseInvalidException Wenn ein Problem mit der Lizenz besteht
	 */
	public function unregisterLicense(string $licenseKey, string $domain): void
	{
		$response = $this->get('unregisterLicense', [
			'licenseKey' => $licenseKey,
			'domain' => $domain,
		]);
		$data = $response->getData();
		assert(is_array($data));

		if (!($data['success'] ?? false)) {
			throw new LicenseInvalidException($data['error'] ?? '');
		}
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @return WhitelabelInfo Lizenzdaten bei Erfolg
	 * @throws DownloaderException Wenn ein Problem in der Kommunikation mit dem API-Server auftritt.
	 * @throws LicenseInvalidException Wenn ein Problem mit der Lizenz besteht
	 */
	public function registerWhitelabel(string $licenseKey, string $domain): WhitelabelInfo
	{
		$response = $this->get('registerWhitelabel', [
			'licenseKey' => $licenseKey,
			'domain' => $domain,
		]);

		$data = $response->getData();
		assert(is_array($data));

		if ($data['success'] ?? false) {
			return WhitelabelInfo::fromArray($data);
		} else {
			throw new LicenseInvalidException($data['error'] ?? '');
		}
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @return WhitelabelInfo Lizenzdaten bei Erfolg
	 * @throws DownloaderException Wenn ein Problem in der Kommunikation mit dem API-Server auftritt.
	 * @throws LicenseInvalidException Wenn ein Problem mit der Lizenz besteht
	 */
	public function unregisterWhitelabel(string $licenseKey, string $domain): WhitelabelInfo
	{
		$response = $this->get('unregisterWhitelabel', [
			'licenseKey' => $licenseKey,
			'domain' => $domain,
		]);

		$data = $response->getData();
		assert(is_array($data));

		if ($data['success'] ?? false) {
			return WhitelabelInfo::fromArray($data);
		} else {
			throw new LicenseInvalidException($data['error'] ?? '');
		}
	}

	/**
	 * @param string $licenseKey
	 * @param string $domain
	 * @param int $domainCount
	 * @param int $userCount
	 * @param string $sumCount
	 * @throws DownloaderException Wenn ein Problem in der Kommunikation mit dem API-Server auftritt.
	 * @throws LicenseInvalidException Wenn ein Problem mit der Lizenz besteht
	 */
	public function reportLicensees(string $licenseKey, string $domain, int $domainCount, int $userCount, string $sumCount): void
	{
		$response = $this->get('reportLicensees', [
			'licenseKey' => $licenseKey,
			'domain' => $domain,
			'domainCount' => $domainCount,
			'userCount' => $userCount,
			'sumCount' => urlencode($sumCount),
		]);

		$data = $response->getData();
		assert(is_array($data));

		if (!($data['success'] ?? false)) {
			throw new LicenseInvalidException($data['error'] ?? '');
		}
	}
}
