<?php

namespace App\Component;

use App\Model\Purpose;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Trait for code sharing between Locale and UserLocale controllers
 */
trait LocaleManagerTrait
{
	/**
	 * @param string $fieldId
	 * @param string $type get/set
	 * @return array [methodname, ?extra]
	 * @phan-return array{0:string,1:?string}
	 */
	protected static function buildGetterSetterName(string $fieldId, string $type='get'): array
	{
		$method = $type.preg_replace_callback('/(?:^|_)([a-z])/',
			function ($match) { return strtoupper($match[1]); },
			$fieldId
		);
		$result = explode('-', $method, 2);
		if (count($result) == 1) {
			$result[] = null;
		}
		return $result;
	}

	/**
	 * @param string $fieldId
	 * @return array [field_name, ?extra]
	 * @phan-return array{0:string,1:?string}
	 */
	protected static function buildSkeletonName(string $fieldId): array
	{
		$result = explode('-', $fieldId, 2);
		if (count($result) == 1) {
			$result[] = null;
		}
		return $result;
	}

	/**
	 * @param string $fieldId
	 * @param array $languageData
	 * @return mixed|string|bool|null
	 */
	protected static function getSkeletonValue(string $fieldId, array $languageData)
	{
		[$fieldName, $purposeId] = self::buildSkeletonName($fieldId);
		if ($purposeId) {
			return $languageData[$fieldName][$purposeId] ?? null;
		} else {
			return $languageData[$fieldName] ?? null;
		}
	}

	/**
	 * @param TranslatorInterface $translator
	 * @return array
	 */
	protected static function getStandardFieldGroups(TranslatorInterface $translator): array
	{
		$result = [
			$translator->trans('Cookie widget texts') => [
				'widget_title' => [
					'field_description' => $translator->trans('Cookie layer headline'),
					'field_id' => 'widget_title',
					'type' => 'input_text',
					'required' => true
				],
				'widget_introduction' => [
					'field_description' => $translator->trans('Layer intro text'),
					'field_id' => 'widget_introduction',
					'field_note' => $translator->trans('Besides the user-defined placeholders you may use <code>:accept</code> for the "Accept all" button text, <code>:decline</code> for the "Decline" button text, <code>:settings</code> for the "Settings" button text and <code>:partnerCount</code> for the count of third-party-vendors.'),
					'type' => 'textarea',
					'required' => true
				],
				'widget_consent_button_text' => [
					'field_description' => $translator->trans('Button for "consent"'),
					'field_id' => 'widget_consent_button_text',
					'type' => 'input_text',
					'required' => true
				],
				'widget_decline_button_text' => [
					'field_description' => $translator->trans('Button: Decline cookies'),
					'field_id' => 'widget_decline_button_text',
					'type' => 'input_text',
					'required' => true
				],
				'widget_settings_button_text' => [
					'field_description' => $translator->trans('Text for Link to control panel'),
					'field_id' => 'widget_settings_button_text',
					'type' => 'input_text',
					'required' => true
				],
				'control_panel_title' => [
					'field_description' => $translator->trans('Control panel headline'),
					'field_id' => 'control_panel_title',
					'type' => 'input_text',
					'required' => true
				],
				'control_panel_cancel_button_text' => [
					'field_description' => $translator->trans('Control panel cancel button text'),
					'field_id' => 'control_panel_cancel_button_text',
					'type' => 'input_text',
					'required' => true
				],
				'control_panel_save_button_text' => [
					'field_description' => $translator->trans('Control panel save button text'),
					'field_id' => 'control_panel_save_button_text',
					'type' => 'input_text',
					'required' => true
				],
				'imprint_text' => [
					'field_description' => $translator->trans('Link text for “Imprint”'),
					'field_id' => 'imprint_text',
					'type' => 'input_text',
					'required' => true
				],
				'privacy_policy_text' => [
					'field_description' => $translator->trans('Link text for “Privacy notes”'),
					'field_id' => 'privacy_policy_text',
					'type' => 'input_text',
					'required' => true
				],
				'accessibility_text' => [
					'field_description' => $translator->trans('Link text for “Accessibility”'),
					'field_id' => 'accessibility_text',
					'type' => 'input_text',
					'required' => true,
				],
				'more_purpose_details_button_text' => [
					'field_description' => $translator->trans('Link text for more details on purposes'),
					'field_id' => 'more_purpose_details_button_text',
					'type' => 'input_text',
					'required' => true
				],
				'more_purpose_details_button_title' => [
					'field_description' => $translator->trans('Accessibility text for more details on purposes'),
					'field_id' => 'more_purpose_details_button_title',
					'field_note' => $translator->trans('Besides the user-defined placeholders you may use <code>:section</code> for the section name details are provided for.'),
					'type' => 'input_text',
					'required' => true
				],
				'consent_sharing_text' => [
					'field_description' => $translator->trans('Declaration for cross-domain consent sharing'),
					'field_id' => 'consent_sharing_text',
					'type' => 'input_text',
					'field_note' => $translator->trans('Besides the user-defined placeholders you may use <code>:group</code> for a link with the name of the group receiving consent.'),
					'required' => true
				],
				'language_chooser_usage' => [
					'field_description' => $translator->trans('Accessibility: Language chooser usage explanation'),
					'field_id' => 'language_chooser_usage',
					'type' => 'input_text',
					'required' => true
				],
			],
			$translator->trans('Widget category view texts') => [
				'categories_accept_all_button_text' => [
					'field_description' => $translator->trans('"Accept all" button'),
					'field_id' => 'categories_accept_all_button_text',
					'type' => 'input_text',
					'required' => true,
				],
				'withdraw_all_text' => [
					'field_description' => "{$translator->trans('Withdraw consent to all')}",
					'field_id' => 'withdraw_all_text',
					'type' => 'input_text',
					'required' => false,
				],
				'check_all_button_text' => [
					'field_description' => $translator->trans('"Check all" button'),
					'field_id' => 'check_all_button_text',
					'type' => 'input_text',
					'required' => false,
				],
				'uncheck_all_button_text' => [
					'field_description' => $translator->trans('"Uncheck all" button'),
					'field_id' => 'uncheck_all_button_text',
					'type' => 'input_text',
					'required' => false,
				],
				'status_active_text' => [
					'field_description' => $translator->trans('Status: "Active"'),
					'field_id' => 'status_active_text',
					'type' => 'input_text',
					'required' => false,
				],
				'status_inactive_text' => [
					'field_description' => $translator->trans('Status: "Inactive"'),
					'field_id' => 'status_inactive_text',
					'type' => 'input_text',
					'required' => false,
				],
				'status_partial_text' => [
					'field_description' => $translator->trans('Status: "Partial"'),
					'field_id' => 'status_partial_text',
					'type' => 'input_text',
					'required' => false,
				],
			],
			$translator->trans('Widget detail view texts') => [
				'details_title' => [
					'field_description' => $translator->trans('Detail view headline'),
					'field_id' => 'details_title',
					'type' => 'input_text',
					'required' => true
				],
				'details_close_button_text' => [
					'field_description' => $translator->trans('Detail view close button text'),
					'field_id' => 'details_close_button_text',
					'type' => 'input_text',
					'required' => true
				],
				'cookie_vendor_text' => [
					'field_description' => $translator->trans('Cookie ”Vendor” label'),
					'field_id' => 'cookie_vendor_text',
					'type' => 'input_text',
					'required' => true
				],
				'cookie_description_text' => [
					'field_description' => $translator->trans('Cookie “Description” label'),
					'field_id' => 'cookie_description_text',
					'type' => 'input_text',
					'required' => true
				],
				'cookie_privacy_policy_url_text' => [
					'field_description' => $translator->trans('Cookie “Privacy policy url” label'),
					'field_id' => 'cookie_privacy_policy_url_text',
					'type' => 'input_text',
					'required' => true
				],
				'cookie_lifetime_text' => [
					'field_description' => $translator->trans('Cookie “Lifetime” label'),
					'field_id' => 'cookie_lifetime_text',
					'type' => 'input_text',
					'required' => true
				],
				'cookie_collected_data_info_text' => [
					'field_description' => $translator->trans('Cookie „What data will be collected?“ label'),
					'field_id' => 'cookie_collected_data_info_text',
					'type' => 'input_text',
					'required' => true
				],
				'cookie_purpose_of_data_collection_text' => [
					'field_description' => $translator->trans('Cookie „Purpose of data collection“ label'),
					'field_id' => 'cookie_purpose_of_data_collection_text',
					'type' => 'input_text',
					'required' => true
				],
				'cookie_legal_basis_text' => [
					'field_description' => $translator->trans('Cookie „Legal basis“ label'),
					'field_id' => 'cookie_legal_basis_text',
					'type' => 'input_text',
					'required' => true
				],
				'cookie_place_of_processing_text' => [
					'field_description' => $translator->trans('Cookie „Place of processing“ label'),
					'field_id' => 'cookie_place_of_processing_text',
					'type' => 'input_text',
					'required' => true
				],
				'embedding_privacy_policy_desc' => [
					'field_description' => $translator->trans('Accessibility: Description for privacy policy link in embeddings'),
					'field_id' => 'embedding_privacy_policy_desc',
					'type' => 'input_text',
					'field_note' => $translator->trans('Use <code>:embedding</code> in the text to dynamically replace it with the embedding name.'),
					'required' => true
				],
			],
			$translator->trans('Embedded cookie declaration') => [
				'change_consent_label' => [
					'field_description' => $translator->trans('Change consent'),
					'field_id' => 'change_consent_label',
					'type' => 'input_text',
					'required' => false
				],
				'asset_name_label' => [
					'field_description' => $translator->trans('Column header “Name”'),
					'field_id' => 'asset_name_label',
					'type' => 'input_text',
					'required' => false
				],
				'asset_lifetime_label' => [
					'field_description' => $translator->trans('Column header “Lifetime”'),
					'field_id' => 'asset_lifetime_label',
					'type' => 'input_text',
					'required' => false
				],
				'asset_description_label' => [
					'field_description' => $translator->trans('Column header “Description”'),
					'field_id' => 'asset_description_label',
					'type' => 'input_text',
					'required' => false
				],
				'consent_state_label' => [
					'field_description' => $translator->trans('Label “Your current state:”'),
					'field_id' => 'consent_state_label',
					'type' => 'input_text',
					'required' => false
				],
				'consent_state_none' => [
					'field_description' => $translator->trans('State “no decision taken”'),
					'field_id' => 'consent_state_none',
					'type' => 'input_text',
					'required' => false
				],
				'consent_state_custom' => [
					'field_description' => $translator->trans('State “individually selected”'),
					'field_id' => 'consent_state_custom',
					'type' => 'input_text',
					'required' => false
				],
				'consent_state_accepted' => [
					'field_description' => $translator->trans('State “all accepted”'),
					'field_id' => 'consent_state_accepted',
					'type' => 'input_text',
					'required' => false
				],
				'consent_state_declined' => [
					'field_description' => $translator->trans('State “declined”'),
					'field_id' => 'consent_state_declined',
					'type' => 'input_text',
					'required' => false
				],
				'consent_id_label' => [
					'field_description' => $translator->trans('Label “Your consent ID:”'),
					'field_id' => 'consent_id_label',
					'type' => 'input_text',
					'required' => false
				],
				'consent_date_label' => [
					'field_description' => $translator->trans('Label “Consent date:”'),
					'field_id' => 'consent_date_label',
					'type' => 'input_text',
					'required' => false
				],
			],
			$translator->trans('TCFv2') => [
				'tcf_widget_introduction' => [
					'field_description' => $translator->trans('Layer intro text'),
					'field_id' => 'tcf_widget_introduction',
					'field_note' => $translator->trans('Besides the user-defined placeholders you may use <code>:accept</code> for the "Accept all" button text, <code>:decline</code> for the "Decline" button text, <code>:settings</code> for the "Settings" button text and <code>:partnerCount</code> for the count of third-party-vendors.'),
					'type' => 'textarea',
					'required' => false,
				],
				'consent_no_sharing_text' => [
					'field_description' => "\"{$translator->trans('You give your consent only for this website.')}\"",
					'field_id' => 'consent_no_sharing_text',
					'type' => 'input_text',
					'required' => true
				],
				'tcf_stacks_list_text' => [
					'field_description' => $translator->trans('Introduction text before permission stack list'),
					'field_id' => 'tcf_stacks_list_text',
					'field_note' => $translator->trans('Besides the user-defined placeholders you may use <code>:accept</code> for the "Accept all" button text, <code>:decline</code> for the "Decline" button text, <code>:settings</code> for the "Settings" button text and <code>:partnerCount</code> for the count of third-party-vendors.'),
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_legitimate_interest_text' => [
					'field_description' => $translator->trans('Disclosure text for using "Legitimate Interest"'),
					'field_id' => 'tcf_legitimate_interest_text',
					'field_note' => $translator->trans('Besides the user-defined placeholders you may use <code>:accept</code> for the "Accept all" button text, <code>:decline</code> for the "Decline" button text, <code>:settings</code> for the "Settings" button text and <code>:partnerCount</code> for the count of third-party-vendors.'),
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_vendor_granted_permissions_text' => [
					'field_description' => $translator->trans('Vendor permissions text'),
					'field_id' => 'tcf_vendor_granted_permissions_text',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_purposes_label' => [
					'field_description' => "\"{$translator->trans('Purposes')}\"",
					'field_id' => 'tcf_purposes_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_special_purposes_label' => [
					'field_description' => "\"{$translator->trans('Special purposes')}\"",
					'field_id' => 'tcf_special_purposes_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_features_label' => [
					'field_description' => "\"{$translator->trans('Features')}\"",
					'field_id' => 'tcf_features_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_special_features_label' => [
					'field_description' => "\"{$translator->trans('Special features')}\"",
					'field_id' => 'tcf_special_features_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_data_categories_label' => [
					'field_description' => "\"{$translator->trans('Data Categories')}\"",
					'field_id' => 'tcf_data_categories_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_opted_in_remark' => [
					'field_description' => "\"{$translator->trans('(if opted in)')}\"",
					'field_id' => 'tcf_opted_in_remark',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_third_parties_label' => [
					'field_description' => "\"{$translator->trans('Third parties')}\"",
					'field_id' => 'tcf_third_parties_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_consent_label' => [
					'field_description' => "\"{$translator->trans('Consent')}\"",
					'field_id' => 'tcf_consent_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_legitimate_interest_label' => [
					'field_description' => "\"{$translator->trans('Legitimate interest')}\"",
					'field_id' => 'tcf_legitimate_interest_label',
					'type' => 'input_text',
					'required' => false,
				],
				'non_tcf_purposes_label' => [
					'field_description' => "\"{$translator->trans('Purposes not defined by IAB')}\"",
					'field_id' => 'non_tcf_purposes_label',
					'type' => 'input_text',
					'required' => false,
				],
				'non_tcf_vendors_label' => [
					'field_description' => "\"{$translator->trans('Vendors not part of IAB')}\"",
					'field_id' => 'non_tcf_vendors_label',
					'type' => 'input_text',
					'required' => false,
				],
				'more_information_label' => [
					'field_description' => "\"{$translator->trans('More information')}\"",
					'field_id' => 'more_information_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_illustrations_label' => [
					'field_description' => "\"{$translator->trans('Illustrations:')}\"",
					'field_id' => 'tcf_illustrations_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_vendor_uses_non_cookie_storage_text' => [
					'field_description' => "\"{$translator->trans('This vendor may use browser storage other than cookies.')}\"",
					'field_id' => 'tcf_vendor_uses_non_cookie_storage_text',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_special_features_handled_globally_text' => [
					'field_description' => "\"{$translator->trans('Special features are handled globally instead of on per vendor basis. Please specify the permissions that you grant each IAB Framework vendor.')}\"",
					'field_id' => 'tcf_special_features_handled_globally_text',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_cookie_refresh_text' => [
					'field_description' => "\"{$translator->trans('The cookie lifetime may start running anew on each visit.')}\"",
					'field_id' => 'tcf_cookie_refresh_text',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_extended_disclosure_text' => [
					'field_description' => "{$translator->trans('Vendor cookie disclosure heading')}",
					'field_id' => 'tcf_extended_disclosure_text',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_asset_type_label' => [
					'field_description' => "{$translator->trans('Column header “Type”')}",
					'field_id' => 'tcf_asset_type_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_asset_purposes_label' => [
					'field_description' => "{$translator->trans('Column header “Purposes”')}",
					'field_id' => 'tcf_asset_purposes_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_data_retention_text' => [
					'field_description' => "{$translator->trans('Data retention duration information')}",
					'field_id' => 'tcf_data_retention_text',
					'field_note' => $translator->trans('You may use <code>:duration</code> as a placeholder for the duration in days/weeks/months/years.'),
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_data_retention_session_text' => [
					'field_description' => "{$translator->trans('Data retention duration is session')}",
					'field_id' => 'tcf_data_retention_session_text',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_purpose_vendor_list_label' => [
					'field_description' => "\"{$translator->trans('Vendors requesting this')}\"",
					'field_id' => 'tcf_purpose_vendor_list_label',
					'type' => 'input_text',
					'required' => false,
				],
				'tcf_consent_storage_text' => [
					'field_description' => $translator->trans('Disclosure of consent storage'),
					'field_id' => 'tcf_consent_storage_text',
					'field_note' => $translator->trans('You may use <code>:storage</code> as a placeholder for storage type and <code>:days</code> for the maximum duration in days.'),
					'type' => 'input_text',
					'required' => false,
				],
			],
			$translator->trans('Blocking dialog for external content') => [
				'blocked_content_title' => [
					'field_description' => $translator->trans('Dialog title'),
					'field_id' => 'blocked_content_title',
					'type' => 'input_text',
					'required' => true
				],
				'blocked_content_text' => [
					'field_description' => $translator->trans('Dialog text'),
					'field_id' => 'blocked_content_text',
					'type' => 'textarea',
					'required' => true,
					'field_note' => $translator->trans('<strong>Note:</strong> You can use <code>:domain</code> as a placeholder for the domain from which content was blocked.')
				],
				'blocked_content_button_text' => [
					'field_description' => $translator->trans('Confirm button'),
					'field_id' => 'blocked_content_button_text',
					'type' => 'input_text',
					'required' => true
				],
				'blocked_content_show_external_content' => [
					'field_description' => $translator->trans('Show external content'),
					'field_id' => 'blocked_content_show_external_content',
					'type' => 'input_text',
					'required' => true,
				],
				'blocked_content_allow_embedding' => [
					'field_description' => $translator->trans('Allow content for ":embedding"'),
					'field_note' => $translator->trans('Use <code>:embedding</code> in the text to dynamically replace it with the embedding name.'),
					'field_id' => 'blocked_content_allow_embedding',
					'type' => 'input_text',
					'required' => true,
				],
				'blocked_content_provider_details' => [
					'field_description' => $translator->trans('Provider details'),
					'field_id' => 'blocked_content_provider_details',
					'type' => 'input_text',
					'required' => true,
				],
			],
		];
		$purposes = array_reduce(Purpose::all(), function ($purposes, $purpose) use ($translator) {
			$purposes[] = [
				'field_description' => $purpose->getName(),
				'field_id' => 'purpose_name-'.$purpose->getId(),
				'type' => 'input_text',
				'required' => true
			];
			$purposes[] = [
				'field_description' => $translator->trans('Description'),
				'field_id' => 'purpose_description-'.$purpose->getId(),
				'type' => 'textarea',
				'required' => false,
			];
			return $purposes;
		}, []);
		$result[$translator->trans('Cookie purposes')] = array_combine(array_column($purposes, 'field_id'), $purposes);
		return $result;
	}

	/**
	 * @param TranslatorInterface $translator
	 * @return array
	 */
	protected final static function getStandardFields(TranslatorInterface $translator): array
	{
		return array_merge([], ...array_values(static::getStandardFieldGroups($translator)));
	}

	/**
	 * @param string $fieldId
	 * @param TranslatorInterface $translator
	 * @return array|null
	 */
	protected final static function getStandardField(string $fieldId, TranslatorInterface $translator): ?array
	{
		return self::getStandardFields($translator)[$fieldId] ?? null;
	}
}
