<?php

namespace App\Component;

use App\Model\Locale as DomainLocale;
use App\Model\LocaleBase;
use App\Model\LocaleOverride;
use App\Model\Purpose;
use App\Model\UserLocale;

class LocaleOverrideHelper
{
	use LocaleManagerTrait;

	/**
	 * Diese Methode aktualisiert die `$locales` entsprechend der `$overrides`.
	 *
	 * Es wird davon ausgegangen, dass die korrekten Repos geladen sind.
	 *
	 * Die Methode fordert ein Write-Lock auf dem Repository an, gibt es aber
	 * nicht selbständig frei. Nach dem Aufruf dieser Methode sollte `flush()`
	 * auf den entsprechenden `LocaleBase`-Klassen ausgeführt werden.
	 *
	 * @param LocaleOverride $override
	 * @param LocaleBase[] $locales
	 * @return void
	 * @see self::applyOverridesToDomainLocales
	 * @see self::applyOverridesToUserLocales
	 */
	private static function applyOverrideToLocales(LocaleOverride $override, array $locales): void
	{
		$locales = array_combine(array_map(function (LocaleBase $locale) {
			return $locale->getName();
		}, $locales), $locales);

		$translations = $override->getTranslations();
		foreach (array_intersect_key($locales, $translations) as $localeName => $locale) {
			if (in_array($override->getName(), $locale->getUpdateParameterBlacklist())) {
				continue;
			}

			[$setter, $purposeId] = self::buildGetterSetterName($override->getName(), 'set');
			if (!method_exists($locale, $setter)) {
				continue;
			}

			if ($purposeId and $purpose = Purpose::find((string)$purposeId)) {
				$locale->$setter($purpose, $translations[$localeName]);
			} else {
				$locale->$setter($translations[$localeName]);
			}
		}
	}

	/**
	 * Diese Methode aktualisiert die `Locale`-Objekte im aktiven `Domain`-Kontext entsprechend der `$overrides`.
	 *
	 * Es wird davon ausgegangen, dass die korrekten Repos geladen sind.
	 *
	 * @param LocaleOverride[] $overrides
	 * @return void
	 */
	public static function applyOverridesToDomainLocales(array $overrides): void
	{
		$locales = DomainLocale::all();
		foreach ($overrides as $override) {
			self::applyOverrideToLocales($override, $locales);
		}
		DomainLocale::flush();
	}

	/**
	 * Diese Methode aktualisiert die `UserLocale`-Objekte im aktiven `User`-Kontext entsprechend der `$overrides`.
	 *
	 * Es wird davon ausgegangen, dass die korrekten Repos geladen sind.
	 *
	 * @param LocaleOverride[] $overrides
	 * @return void
	 */
	public static function applyOverridesToUserLocales(array $overrides): void
	{
		$userLocales = UserLocale::all();
		foreach ($overrides as $override) {
			self::applyOverrideToLocales($override, $userLocales);
		}
		UserLocale::flush();
	}
}
