<?php

namespace App\Component;

use App\Model\User;
use App\Model\Domain;
use App\Model\Locale;
use App\Model\LocaleBase;
use App\Model\UserLocale;
use Symfony\Component\HttpFoundation\Request;

class LocaleService
{
	public const DEFAULT_PRIORITY = ['en_US', 'en_GB', 'en_DE', 'de_DE', 'de_AT', 'de_CH'];

	/**
	 * @return LocaleBase|null
	 */
	public static function find(string $localeString): ?LocaleBase
	{
		if (Domain::activeDomain()) {
			return Locale::find($localeString) ?? UserLocale::find($localeString);
		} else {
			return UserLocale::find($localeString);
		}
	}

	/**
	 * @param string $localeString
	 * @return LocaleBase|null
	 */
	public static function findFuzzy(string $localeString): ?LocaleBase
	{
		return self::fuzzySearch($localeString, self::all());
	}

	/**
	 * @return LocaleBase[]
	 */
	public static function all(): array
	{
		$locales = (Domain::activeDomain()) ? Locale::all() : [];
		$userLocales = array_filter(UserLocale::all(), function ($ul) use ($locales) {
			foreach ($locales as $l) {
				if ($l->getName() == $ul->getName()) {
					return false;
				}
			}
			return true;
		});
		return $locales + $userLocales;
	}

	/**
	 * @return LocaleBase[]
	 */
	public static function enabledLocales(): array
	{
		return array_filter(self::all(), function ($locale) {
			/** @var LocaleBase $locale */
			return $locale->isEnabled();
		});
	}

	/**
	 * @return LocaleBase
	 */
	public static function getDefault(): LocaleBase
	{
		$user = User::loggedInUser();
		return $user->getAccountOwner()->getDefaultLocale();
	}

	/**
	 * @return UserLocale
	 */
	public static function getUserDefault(): UserLocale
	{
		$result = self::getDefault();

		if (!($result instanceof UserLocale)) {
			$result = UserLocale::find($result->getName());
		}

		assert($result instanceof UserLocale);
		return $result;
	}

	public static function lastModifiedTimestamp(): int
	{
		return max(
			Locale::lastModifiedTimestamp(),
			UserLocale::lastModifiedTimestamp()
		);
	}

	/**
	 * Sucht das am besten passende aktive Locale zum angegebenen Locale-String.
	 *
	 * Hier wird die Variante mit _, nicht mit - erwartet!
	 *
	 * Existiert kein passendes Locale, wird NULL zurückgegeben.
	 *
	 * @param string $localeString
	 * @param bool $allowMoreSpecific Ob z.B. "de" auch "de_DE" matchen soll.
	 * @return LocaleBase|null
	 */
	public static function findEnabledFuzzy(string $localeString, bool $allowMoreSpecific = true): ?LocaleBase
	{
		$locale = self::find($localeString);
		if ($locale && $locale->isEnabled()) {
			return $locale;
		}

		$locales = self::enabledLocales();

		return self::fuzzySearch($localeString, $locales);
	}

	/**
	 * @param string $localeString
	 * @param array $locales
	 * @return LocaleBase|null
	 */
	private static function fuzzySearch(string $localeString, array $locales): ?LocaleBase
	{
		// Optionale Angabe des Schriftsystems rauslöschen (siehe BCP 47)
		// Sonst würde de-Latn-DE → de_LATN_DE nicht de_DE matchen.
		$localeString = preg_replace('/_[A-Z][a-zA-Z]{3}(_|$)/', '\1', $localeString);

		// Passendes Locale suchen
		foreach ($locales as $locale) {
			if ($locale->getName() === $localeString) {
				return $locale;
			}
		}
		foreach ($locales as $locale) {
			if (strpos($locale->getName(), $localeString.'_') === 0) {
				return $locale;
			}
		}
		foreach ($locales as $locale) {
			if (strpos($localeString, $locale->getName().'_') === 0) {
				return $locale;
			}
		}
		return null;
	}

	/**
	 * Sucht das passendste Locale anhand der vom Browser angegebenen
	 * Sprachen.
	 *
	 * Existiert kein passendes Locale, wird NULL zurückgegeben
	 *
	 * @param Request $request
	 * @return ?LocaleBase
	 */
	public static function guessFromBrowser(Request $request): ?LocaleBase
	{
		$locales = Locale::enabledLocales();
		foreach ($request->getLanguages() as $lang)
		{
			$locale = self::findEnabledFuzzy($lang, false);
			if ($locale) {
				return $locale;
			}
		}
		foreach ($request->getLanguages() as $lang)
		{
			$locale = self::findEnabledFuzzy($lang, true);
			if ($locale) {
				return $locale;
			}
		}
		return null;
	}

	/**
	 * @param string[] $localeNames Nach Priorität sortierte Locale-Namen
	 * @param LocaleBase[] $locales Bestimmte Auswahl an Locales; wenn leer, werden alle berücksichtigt.
	 * @return LocaleBase Die zuerst ermittelte Locale; entweder passend zu $localeNames, ansonsten die erstbeste.
	 */
	public static function anyLocale(array $localeNames=[], array $locales=[]): LocaleBase
	{
		if (empty($locales)) {
			$locales = self::all();
		}

		$match = array_reduce($localeNames, function ($match, $localeName) use ($locales) {
			return $match ?? array_reduce($locales, function ($match, $locale) use ($localeName) {
				/** @var LocaleBase $locale */
				return $match ?? (strtolower($localeName) == strtolower($locale->getName()) ? $locale : $match);
			});
		});

		return $match ?? reset($locales);
	}

	/**
	 * @param string[] $localeNames Nach Priorität sortierte Locale-Namen
	 * @param LocaleBase[] $locales Bestimmte Auswahl an Locales; wenn leer, werden alle berücksichtigt.
	 * @return LocaleBase Die zuerst ermittelte Locale; entweder passend zu $localeNames, ansonsten die erstbeste.
	 */
	public static function anyLocaleFuzzy(array $localeNames=[], array $locales=[]): LocaleBase
	{
		if (empty($locales)) {
			$locales = self::all();
		}

		foreach ($localeNames as $localeName) {
			if ($locale = self::fuzzySearch($localeName, $locales)) {
				return $locale;
			}
		}

		return reset($locales);
	}
}
