<?php

namespace App\Component\LogStream;

use Countable;
use OutOfBoundsException;
use Iterator;
use OuterIterator;
use SeekableIterator;

/**
 * Iterator, der das Dekodieren der Zeilen in der JSON-Datei möglicht lange herauszögert.
 *
 * Mit `seek()` kann der Startoffset festgelegt werden. Kompatibel mit `LimitIterator`.
 *
 * Damit ist Pagination einfach möglich.
 */
class JsonSeekableIterator implements OuterIterator, SeekableIterator, Countable
{
	/** @var callable $lineIteratorCallback */
	private $lineIteratorCallback;
	/** @var int $currentOffset */
	private $currentOffset = 0;
	/** @var array|null $currentValue */
	private $currentValue = null;
	/** @var Iterator|null $iterator */
	private $iterator;
	/** @var int|null $cachedCount */
	private $cachedCount = null;

	/**
	 * @param callable $lineIteratorCallback Funktion, die einen Iterator über die Zeilen liefert
	 */
	public function __construct(callable $lineIteratorCallback)
	{
		$this->lineIteratorCallback = $lineIteratorCallback;
		$this->iterator = $lineIteratorCallback();
	}

	/**
	 * {@inheritdoc}
	 * @return int
	 */
	#[\ReturnTypeWillChange]
	public function key()
	{
		return $this->currentOffset;
	}

	/**
	 * Prüft, ob die aktuelle Position zulässig ist
	 */
	public function valid(): bool
	{
		return ($this->iterator !== null and $this->iterator->valid());
	}

	/**
	 * {@inheritdoc}
	 */
	public function getInnerIterator(): Iterator
	{
		assert($this->iterator !== null);
		return $this->iterator;
	}

	/**
	 * {@inheritdoc}
	 * @return array|null
	 */
	#[\ReturnTypeWillChange]
	public function current()
	{
		if ($this->currentValue) {
			return $this->currentValue;
		}
		$data = null;
		while ($this->iterator->valid()) {
			$line = $this->iterator->current();
			$data = json_decode($line, true);
			if (is_array($data)) {
				break;
			}
			$this->iterator->next();
		}
		$this->currentValue = $data;
		return $data;
	}

	/**
	 * {@inheritdoc}
	 */
	public function next(): void
	{
		$this->current();
		$this->iterator->next();
		$this->currentValue = null;
		$this->currentOffset++;
	}

	/**
	 * {@inheritdoc}
	 */
	public function rewind(): void
	{
		$this->currentOffset = 0;
		$this->currentValue = null;
		$this->iterator->rewind();
		$this->iterator = null;
		$this->iterator = ($this->lineIteratorCallback)();
	}

	/**
	 * Springt zu einer Position
	 *
	 * @param int $position
	 * @return void
	 * @phan-suppress PhanParamSignatureRealMismatchHasNoParamTypeInternal
	 */
	public function seek($position): void
	{
		if ($this->currentOffset === $position) {
			return;
		}
		if ($position < 0) {
			throw new OutOfBoundsException('cannot seek to negative position');
		}

		// Am Anfang beginnen, wenn rückwärts gegangen werden soll
		if ($position < $this->currentOffset) {
			$this->rewind();
		}

		// Zum gewünschten Punkt wandern
		for ($i = $this->currentOffset; $i < $position; ++$i) {
			if ($this->iterator->valid()) {
				$this->iterator->next();
			}
		}
		$this->currentOffset = $position;
	}

	/**
	 * Gibt die Anzahl der Elemente zurück
	 *
	 * Die Anzahl wird nach dem ersten Aufruf gecached und muss bei
	 * korrumpierten Dateien nicht unbedingt exakt mit der Anzahl der
	 * iterierten Elemente übereinstimmen.
	 */
	public function count(): int
	{
		if ($this->cachedCount !== null) {
			return $this->cachedCount;
		}

		$position = $this->currentOffset;
		$count = $position-1;

		while ($this->iterator->valid()) {
			$count++;
			$this->iterator->next();
		}

		// An die Stelle zurückspringen, an der wir waren
		$this->rewind();
		$this->seek($position);

		// Ergebnis merken und zurückgeben
		$this->cachedCount = $count;
		return $count;
	}

}
