<?php

namespace App\Component\LogStream;

use Countable;
use EmptyIterator;
use OutOfBoundsException;
use Iterator;
use IteratorIterator;
use OuterIterator;
use SeekableIterator;
use MongoDB\Collection;
use MongoDB\Model\BSONDocument;

/**
 * Iterator, der das Absetzen eines `find`-Queries so lange wie möglich
 * herauszögert.
 *
 * Mit `seek()` kann der Startoffset festgelegt werden. Kompatibel mit `LimitIterator`.
 *
 * Damit ist Pagination einfach möglich.
 */
class MongoDbSeekableIterator implements OuterIterator, SeekableIterator, Countable
{
	/** @var Collection $collection */
	private $collection;
	/** @var array $filter */
	private $filter;
	/** @var array $options */
	private $options;
	/** @var int $startOffset */
	private $startOffset = 0;
	/** @var int $currentOffset */
	private $currentOffset = 0;
	/** @var callable $map */
	private $map;
	/** @var Iterator|null $iterator */
	private $iterator = null;

	/**
	 * @param Collection $collection
	 * @param array $filter MongoDB Query-Filter
	 * @param array $options MongoDB Query-Optionen
	 * @param callable(BSONDocument):mixed $mapDocument Mapping-Funktion für das Dokument
	 */
	public function __construct(Collection $collection, array $filter, array $options, callable $mapDocument)
	{
		$this->collection = $collection;
		$this->filter = $filter;
		$this->options = $options;
		$this->map = $mapDocument;
	}

	/**
	 * Startet die Datenbankabfrage, wenn noch nicht passiert.
	 *
	 * Setzt $this->iterator immer auf einen Iterator. Wenn die Datenbankabfrage
	 * fehlschlug ohne eine Exception auszulösen, wird ein EmptyIterator
	 * verwendet.
	 *
	 * @return Iterator $this->iterator
	 */
	private function ensureIterator(): Iterator
	{
		if ($this->iterator === null) {
			$options = $this->options;
			$options['skip'] = $this->startOffset;
			$cursor = $this->collection->find($this->filter, $options);
			$this->currentOffset = $this->startOffset;

			if ($cursor === null) {
				$this->iterator = new EmptyIterator();
			}
			elseif ($cursor instanceof Iterator) {
				$this->iterator = $cursor;
			}
			else {
				// Abwärtskompatibilität mit php-mongodb < 1.9
				$this->iterator = new IteratorIterator($cursor);
			}

			$this->iterator->rewind();
		}
		return $this->iterator;
	}

	/**
	 * {@inheritdoc}
	 */
	public function getInnerIterator(): Iterator
	{
		return $this->ensureIterator();
	}

	/**
	 * {@inheritdoc}
	 * @return int
	 */
	#[\ReturnTypeWillChange]
	public function key()
	{
		return $this->currentOffset;
	}

	/**
	 * {@inheritdoc}
	 */
	public function valid(): bool
	{
		$this->ensureIterator();
		return $this->iterator->valid();
	}

	/**
	 * {@inheritdoc}
	 */
	#[\ReturnTypeWillChange]
	public function current()
	{
		$this->ensureIterator();
		return ($this->map)($this->iterator->current());
	}

	/**
	 * {@inheritdoc}
	 */
	public function next(): void
	{
		$this->ensureIterator();
		$this->iterator->next();
		$this->currentOffset++;
	}

	/**
	 * {@inheritdoc}
	 */
	public function rewind(): void
	{
		$this->seek(0);
	}

	/**
	 * {@inheritdoc}
	 * @param int $position
	 * @return void
	 * @phan-suppress PhanParamSignatureRealMismatchHasNoParamTypeInternal
	 */
	public function seek($position): void
	{
		if ($this->currentOffset === $position) {
			$this->startOffset = $position;
			return;
		}
		if ($position < 0) {
			throw new OutOfBoundsException('cannot seek to negative position');
		}
		$this->startOffset = $position;

		if ($this->iterator !== null) {
			// Wenn der Iterator aktuell recht nah am gewünschten Punkt ist, da hin wandern
			$offset = $this->currentOffset;
			if ($offset < $position and $position-$offset < 50) {
				for ($i = $offset; $i < $position; ++$i) {
					if ($this->valid()) {
						$this->next();
					}
				}
			}
			// Ansonsten den aktuellen Iterator schließen und mit einem neuen DB-Query anfangen
			else {
				$this->iterator = null;
			}
		}
	}

	/**
	 * Gibt die Anzahl der Elemente zurück
	 *
	 * Dafür wird eine neue Datenbankabfrage gemacht.
	 */
	public function count(): int
	{
		return $this->collection->countDocuments($this->filter, $this->options);
	}

}
