<?php

namespace App\Component\LogStream;

use FilterIterator;
use InvalidArgumentException;
use Iterator;
use MongoDB\BSON\Regex;

/**
 * Filtert einen Iterator anhand eines MongoDB-Query-Filters.
 */
class QueryFilterIterator extends FilterIterator
{
	/** @var callable $filter */
	private $filter;

	/** @var callable|null $lastFilterArray */
	private static $lastFilterArray = null;
	/** @var callable|null $lastFilterCompiled */
	private static $lastFilterCompiled = null;

	/**
	 * @param Iterator $iterator
	 * @param array $filter MongoDB-Query-Filter
	 * @throws InvalidArgumentException Wenn das Query nicht unterstützt wird
	 */
	public function __construct(Iterator $iterator, array $filter)
	{
		if ($filter === self::$lastFilterArray) {
			$this->filter = self::$lastFilterCompiled;
		} else {
			$this->filter = $this->compileFilter('$implicit_and', $filter);
			self::$lastFilterArray = $filter;
			self::$lastFilterCompiled = $this->filter;
		}
		parent::__construct($iterator);
	}

	/**
	 * Kompiliert ein MongoDB-Filter-Prädikat
	 *
	 * @param string $op
	 * @param mixed $comp
	 * @param array $extra
	 * @return null|callable(array,string):bool
	 */
	private static function compileFilterPredicate(string $op, $comp, array $extra): ?callable
	{
		switch ($op) {
			case '$eq':
				return function ($item, $key) use ($comp) { return ($item[$key] ?? null) == $comp; };
			case '$ne':
				return function ($item, $key) use ($comp) { return ($item[$key] ?? null) != $comp; };
			case '$gt':
				return function ($item, $key) use ($comp) { $value = $item[$key] ?? null; return ($item[$key] !== null and $value > $comp); };
			case '$gte':
				return function ($item, $key) use ($comp) { $value = $item[$key] ?? null; return ($item[$key] !== null and $value >= $comp); };
			case '$lt':
				return function ($item, $key) use ($comp) { $value = $item[$key] ?? null; return ($item[$key] !== null and $value < $comp); };
			case '$lte':
				return function ($item, $key) use ($comp) { $value = $item[$key] ?? null; return ($item[$key] !== null and $value <= $comp); };
			case '$exists':
				return function ($item, $key) use ($comp) { return isset($item[$key]) == $comp; };
			case '$in':
				return function ($item, $key) use ($comp) { return in_array($item[$key] ?? null, $comp); };
			case '$nin':
				return function ($item, $key) use ($comp) { return !in_array($item[$key] ?? null, $comp); };
			case '$not':
				$subfilter = self::compileFilterPredicates($comp);
				return function ($item, $key) use ($subfilter) {
					foreach ($subfilter as $filter) {
						if (!$filter($item, $key)) {
							return false;
						}
						return true;
					}
				};
			case '$mod':
				if (count($comp) !== 2) {
					throw new InvalidArgumentException('invalid $mod argument count');
				}
				return function ($item, $key) use ($comp) { $value = $item[$key] ?? null; return ($item[$key] !== null and $value % $comp[0] === $comp[1]); };
			case '$regex':
				if ($comp instanceof Regex) {
					$pattern = $comp->getPattern();
					$flags = $comp->getFlags();
				} else {
					$pattern = (string)$comp;
					$flags = $extra['options'] ?? '';
				}

				// Optimierter Sonderfall: Einfacher Präfix
				if ($pattern !== '' and $pattern[0] == '^' and preg_match('~^\^[a-zA-Z0-9 #@,_-]*$~', $pattern)) {
					$prefix = substr($pattern, 1);
					if ($flags === '') {
						return function ($item, $key) use ($prefix) { $value = $item[$key] ?? null; return (is_string($value) and strpos($value, $prefix) === 0); };
					} elseif ($flags === 'i') {
						return function ($item, $key) use ($prefix) { $value = $item[$key] ?? null; return (is_string($value) and stripos($value, $prefix) === 0); };
					}
				}
				// Sonst als Preg-Regex verarbeiten
				$fullPattern = '~'.str_replace('~', '\~', $pattern).'~'.$flags;
				return function ($item, $key) use ($fullPattern) { $value = $item[$key] ?? null; return (is_string($value) and preg_match($fullPattern, $item[$key])); };
			case '$options':
				return null;
			default:
				throw new InvalidArgumentException('unsupported filter predicate "'.$op.'"');
		}
	}

	/**
	 * Kompiliert MongoDB-Filter-Prädikate in einem assoziativen Array
	 *
	 * @param mixed $filter
	 * @return array<callable(array,string):bool>
	 */
	private static function compileFilterPredicates($filter): array
	{
		if ($filter instanceof Regex) {
			$filter = ['$regex' => $filter->getPattern(), '$options' => $filter->getFlags()];
		}
		elseif (!is_array($filter)) {
			$filter = ['$eq' => $filter];
		}

		return array_filter(array_map(function ($op, $subfilter) use ($filter) {
			return self::compileFilterPredicate($op, $subfilter, $filter);
		}, array_keys($filter), $filter));
	}

	/**
	 * Kompiliert ein MongoDB-Filter-Query in eine Funktion, die für einen
	 * Eintrag true zurückgibt, wenn das Query matcht.
	 *
	 * @param string $key
	 * @param mixed $filter
	 * @return callable(array):bool
	 */
	public static function compileFilter(string $key, $filter): callable
	{
		if (in_array($key, ['$or', '$nor', '$and', '$implicit_and'])) {
			if ($key === '$implicit_and') {
				$subfilters = array_map(function ($key, $filter) {
					return self::compileFilter($key, $filter);
				}, array_keys($filter), $filter);
			} else {
				$subfilters = array_map(function ($filter) {
					return self::compileFilter('$implicit_and', $filter);
				}, $filter);
			}

			if (count($subfilters) == 1) {
				$subfilter = reset($subfilters);
				if ($key == '$nor') {
					return function (array $item) use ($subfilter) { return ! $subfilter($item); };
				} else {
					return $subfilter;
				}
			}

			switch ($key) {
				case '$implicit_and':
				case '$and':
					return function (array $item) use ($subfilters) {
						foreach ($subfilters as $filter) {
							if (!$filter($item)) {
								return false;
							}
						}
						return true;
					};
				case '$nor':
					return function (array $item) use ($subfilters) {
						foreach ($subfilters as $filter) {
							if ($filter($item)) {
								return false;
							}
						}
						return true;
					};
				case '$or':
				default:
					return function (array $item) use ($subfilters) {
						foreach ($subfilters as $filter) {
							if ($filter($item)) {
								return true;
							}
						}
						return false;
					};
			}
		}
		elseif ($key == '' or $key[0] == '$') {
			throw new InvalidArgumentException('unsupported filter "'.$key.'"');
		}

		$predicates = self::compileFilterPredicates($filter);

		return function (array $item) use ($key, $predicates) {
			foreach ($predicates as $predicate) {
				if (!$predicate($item, $key)) {
					return false;
				}
			}
			return true;
		};
	}

	/**
	 * @override
	 */
	public function accept(): bool
	{
		// @phan-suppress-next-line PhanUndeclaredMethod
		$item = $this->getInnerIterator()->current();
		if ($item === null) { return false; }
		return ($this->filter)($item);
	}
}
