<?php

namespace App\Component\LogStream;

use Countable;
use OutOfBoundsException;
use SeekableIterator;
use Predis\Client as Redis;

/**
 * Iterator, der das Absetzen eines `find`-Queries so lange wie möglich
 * herauszögert.
 *
 * Mit `seek()` kann der Startoffset festgelegt werden. Kompatibel mit `LimitIterator`.
 *
 * Damit ist Pagination einfach möglich.
 */
class RedisScoreRangeIterator implements SeekableIterator, Countable
{
	/** @var Redis $redis */
	private $redis;
	/** @var string $key */
	private $key;
	/** @var string|int|float $start */
	private $start;
	/** @var string|int|float $end */
	private $end;
	/** @var bool $reversed */
	private $reversed;
	/** @var int $startOffset */
	private $startOffset = 0;
	/** @var int $currentOffset */
	private $currentOffset = 0;
	/** @var callable $map */
	private $map;
	/** @var int $chunkSize */
	private $chunkSize;
	/** @var array|null $data */
	private $data = null;

	/**
	 * @param Redis $redis
	 * @param string $key Redis-Key zu einer sortierten Menge
	 * @param string|int|float $start Minimum-Score (z.B. '-inf', '(1', 2.0)
	 * @param string|int|float $end Maximum-Score (z.B. '+inf', '(100', 99.9)
	 * @param bool $reversed true = absteigend sortieren
	 * @param callable(string):mixed $mapDocument Mapping-Funktion für das Dokument
	 * @param int $chunkSize
	 */
	public function __construct(Redis $redis, string $key, $start, $end, bool $reversed, callable $mapDocument, int $chunkSize)
	{
		$this->redis = $redis;
		$this->key = $key;
		$this->start = $start;
		$this->end = $end;
		$this->reversed = $reversed;
		$this->map = $mapDocument;
		$this->chunkSize = $chunkSize;
	}

	/**
	 * Holt einen Chunk Datensätze, falls nötig.
	 *
	 * @return void
	 */
	private function ensureChunk()
	{
		if ($this->data === null or ($this->currentOffset-$this->startOffset) >= count($this->data)) {
			// Wenn die Anzahl der Einträge weit unter der Chunk-Größe liegt, sind wir am Ende angekommen
			// und müssen keine Abfrage mehr machen.
			if ($this->data !== null and count($this->data) < $this->chunkSize*0.60) {
				$this->data = [];
			}
			// Ansonsten neuen Chunk aus Redis holen
			elseif ($this->reversed) {
				$this->data = $this->redis->zRevRangeByScore($this->key, $this->end, $this->start, ['limit'=>[$this->currentOffset, $this->chunkSize]]);
			}
			else {
				$this->data = $this->redis->zRangeByScore($this->key, $this->start, $this->end, ['limit'=>[$this->currentOffset, $this->chunkSize]]);
			}
			$this->startOffset = $this->currentOffset;
		}
	}

	/**
	 * {@inheritdoc}
	 * @return int
	 */
	#[\ReturnTypeWillChange]
	public function key()
	{
		return $this->currentOffset;
	}

	/**
	 * {@inheritdoc}
	 */
	public function valid(): bool
	{
		$this->ensureChunk();
		return (bool)$this->data;
	}

	/**
	 * {@inheritdoc}
	 */
	#[\ReturnTypeWillChange]
	public function current()
	{
		$this->ensureChunk();
		assert(is_array($this->data));
		return ($this->map)($this->data[$this->currentOffset - $this->startOffset]);
	}

	/**
	 * {@inheritdoc}
	 */
	public function next(): void
	{
		$this->currentOffset++;
	}

	/**
	 * {@inheritdoc}
	 */
	public function rewind(): void
	{
		$this->seek(0);
	}

	/**
	 * {@inheritdoc}
	 * @param int $position
	 * @return void
	 * @phan-suppress PhanParamSignatureRealMismatchHasNoParamTypeInternal
	 */
	public function seek($position): void
	{
		if ($this->currentOffset === $position) {
			return;
		}
		if ($position < 0) {
			throw new OutOfBoundsException('cannot seek to negative position');
		}
		$this->currentOffset = $position;

		if ($this->data !== null) {
			// Wenn die Position nicht im aktuellen Chunk liegt, einen neuen Chunk holen
			$offset = $this->startOffset;
			$dataIdx = $position-$offset;
			if ($dataIdx < 0 or $dataIdx > count($this->data)) {
				$this->data = null;
			}
		}
	}

	/**
	 * Gibt die Anzahl der Elemente zurück
	 *
	 * Dafür wird eine neue Datenbankabfrage gemacht.
	 */
	public function count(): int
	{
		return (int)$this->redis->zCount($this->key, $this->start, $this->end);
	}

}
