<?php

namespace App\Component\LogStream\Store;

use App\Component\LogStream\RepositoryInterface;
use App\Component\LogStream\QueryFilterIterator;
use App\Component\LogStream\JsonSeekableIterator;
use App\Model\User;
use App\Model\Domain;
use App\Utils;
use ArrayIterator;
use BadMethodCallException;
use Generator;
use Iterator;
use NoRewindIterator;
use RuntimeException;
use App\Exception\LogNotReadableException;
use App\Exception\LogNotWritableException;
use App\Exception\LogNotDeletableException;

class GzJsonStore implements RepositoryInterface
{
	/** @var string $pathname */
	private $pathname;

	/** @var resource|null $fileHandle */
	private $fileHandle;

	/**
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @param string|null $base Basis-Verzeichnis (nur für Sonderfälle angeben, sonst automatisch)
	 */
	private static function getBasedir(?User $user, ?Domain $domain, ?string $base=null): string
	{
		if ($base) {
			$base = rtrim($base, '/\\');
			if ($domain) {
				$base .= DIRECTORY_SEPARATOR . $domain->getUser()->getId() . DIRECTORY_SEPARATOR . 'domains'. DIRECTORY_SEPARATOR . $domain->getId();
			}
			elseif ($user) {
				$base .= DIRECTORY_SEPARATOR . $user->getId();
			}
		} else {
			if ($domain) {
				$base = $domain->configDir(false);
			}
			elseif ($user) {
				$base = $user->configDir(false);
			}
			else {
				$base = Utils::getBundledVarDir();
			}
		}
		return $base;
	}

	/**
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @param string $name
	 * @param string|null $base Basis-Verzeichnis (nur für Sonderfälle angeben, sonst automatisch)
	 */
	public function __construct(?User $user, ?Domain $domain, string $name, ?string $base=null)
	{
		$this->fileHandle = null;

		$base = self::getBasedir($user, $domain, $base);
		$this->pathname = $base . DIRECTORY_SEPARATOR . $name . '.log.gz';
	}

	/**
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @param string $prefix
	 * @param string|null $base Basis-Verzeichnis (nur für Sonderfälle angeben, sonst automatisch)
	 */
	public static function listCollections(?User $user, ?Domain $domain, string $prefix, ?string $base=null): array
	{
		$base = self::getBasedir($user, $domain, $base);
		$fullPrefix = $base . DIRECTORY_SEPARATOR . $prefix;
		$files = glob(preg_replace('/(\*|\?|\[)/', '[$1]', $fullPrefix) . '~*.log.gz');
		return array_map(function ($path) { return substr(basename($path), 0, -strlen('.log.gz')); }, (array)$files);
	}

	public function __destruct()
	{
		$this->close();
	}

	/**
	 * Durchsuchen des Protokolls.
	 *
	 * @param array $query MongoDB-Query-Array mit einfachen Prädikaten
	 * @param bool $reversed
	 * @return Iterator<array>
	 * @see https://docs.mongodb.com/manual/reference/operator/query/
	 */
	public function filter(array $query, bool $reversed=false): Iterator
	{
		return new QueryFilterIterator($this->all($reversed), $query);
	}

	/**
	 * Iteriert über alle Einträge im Log
	 *
	 * @param bool $reversed
	 * @return \Iterator<array>
	 */
	public function all(bool $reversed=false): Iterator
	{
		if ($reversed) {
			// Geht leider nicht viel effizienter
			$result = [];
			foreach ($this->iterateLines() as $line) {
				if (is_array($data = json_decode($line, true))) {
					$result[] = $data;
				}
			}
			return new ArrayIterator(array_reverse($result));
		}
		else {
			return new JsonSeekableIterator(function () { return $this->iterateLines(); });
		}
	}

	public function count(): int
	{
		$this->openForReading();

		$counter = 0;
		while (($line = gzgets($this->fileHandle, 8192)) !== false) {
			if (empty($line) || substr($line, 0, 1) == '#') {
				continue;
			}

			$counter++;
		}

		$this->close();
		return $counter;
	}

	public function exists(): bool
	{
		return file_exists($this->pathname);
	}

	public function createOrOverwrite(): void
	{
		$this->openForWriting();
	}

	/**
	 * @throws RuntimeException
	 */
	public function append(array $data): void
	{
		if (!$this->fileHandle) {
			throw new RuntimeException('file must be open for writing');
		}
		$this->write(json_encode($data)."\n");
	}

	/**
	 * @throws RuntimeException
	 */
	public function appendMulti(iterable $datas): void
	{
		if (!$this->fileHandle) {
			throw new RuntimeException('file must be open for writing');
		}
		foreach ($datas as $data) {
			$this->write(json_encode($data)."\n");
		}
	}

	/**
	 * @throws BadMethodCallException
	 */
	public function rotateProtocol(?array $keepQueryHint = null): void
	{
		throw new BadMethodCallException('rotateProtocol() unsupported.');
	}

	/**
	 * @throws LogNotReadableException
	 * @throws LogNotWritableException
	 */
	public function trimBefore(array $query): int
	{
		$check = QueryFilterIterator::compileFilter('$implicit_and', $query);

		$tmpFile = $this->pathname.'.tmp';
		$lockHandle = fopen($tmpFile, 'cb+');
		@flock($lockHandle, LOCK_EX);

		// Abbruch bei parallelen Bearbeitung in anderem Thread
		if ($lockHandle and !file_exists($tmpFile)) {
			@flock($lockHandle, LOCK_UN);
			return 0;
		}
		$writeHandle = @gzopen($tmpFile, 'wb9');

		if (!$writeHandle or !$lockHandle) {
			@flock($lockHandle, LOCK_UN);
			throw new LogNotWritableException($tmpFile);
		}

		$this->openForReading();

		$countDeleted = 0;
		$fullDelete = false;
		try {
			$iterator = new NoRewindIterator($this->iterateLines());
			// Alles bis zur ersten nicht-matchenden Zeile überspringen
			foreach ($iterator as $line) {
				if (is_array($data = json_decode($line, true))) {
					if (!$check($data)) {
						break;
					}
					$countDeleted ++;
				}
			}

			// Nur wenn auch etwas gelöscht wird oder etwas erhalten werden soll, muss wirklich weitergemacht werden
			if ($countDeleted > 0 and $iterator->valid()) {
				// Alle zu erhaltenden Zeilen in neue Datei übernehmen
				foreach ($iterator as $line) {
					gzwrite($writeHandle, $line);
				}
			}
			elseif ($countDeleted > 0) {
				$fullDelete = true;
			}

			$iterator = null;
			$this->close();

			gzclose($writeHandle);

			// Wenn nichts gelöscht werden soll, muss auch die Datei nicht ersetzt werden.
			if ($countDeleted === 0) {
				// do nothing
			}
			// Wenn alles gelöscht werden soll, Datei selbst löschen
			elseif ($fullDelete) {
				unlink($this->pathname);
			}
			// Ansonsten Datei mit gekürzter Fassung ersetzen
			else {
				rename($tmpFile, $this->pathname);
			}
		}
		finally {
			if (file_exists($tmpFile)) {
				@unlink($tmpFile);
			}
			flock($lockHandle, LOCK_UN);
			fclose($lockHandle);
		}
		return $countDeleted;
	}

	/**
	 * Löscht ein komplettes .json.gz-Repository unwiderruflich.
	 *
	 * @throws LogNotDeletableException
	 */
	public function clear(): void
	{
		$this->close();
		$filename = $this->pathname;
		if (file_exists($filename) and @unlink($filename) === false) {
			throw new LogNotDeletableException($filename);
		}
	}

	/**
	 * Iteriert über alle Einträge im Log ohne sie zu dekodieren
	 *
	 * @return Generator|string[]
	 * @phan-return generator<string>
	 */
	private function iterateLines(): Generator
	{
		$this->openForReading();
		try {
			while (!gzeof($this->fileHandle) and ($line = gzgets($this->fileHandle, 8192)) !== false) {
				if (empty($line) || substr($line, 0, 1) == '#') {
					continue;
				}
				yield $line;
			}
		}
		finally {
			$this->close();
		}
	}

	/**
	 * Diese Methode gibt den Dateipfad der Konfigurationsdatei zurück.
	 * @return string
	 */
	public function getPathname(): string
	{
		return $this->pathname;
	}

	/**
	 * @param string $mode
	 * @throws LogNotReadableException
	 * @throws LogNotWritableException
	 */
	private function open(string $mode): void
	{
		if (is_resource($this->fileHandle)) {
			return;
		}

		$filename = $this->getPathname();

		$this->fileHandle = @gzopen($filename, $mode);

		if (is_resource($this->fileHandle) == false) {
			if ($mode == 'rb') {
				throw new LogNotReadableException($filename);
			} else {
				throw new LogNotWritableException($filename);
			}
		}
	}

	/**
	 * @throws RuntimeException
	 */
	private function openForReading(): void
	{
		$this->open('rb');
	}

	/**
	 * @throws RuntimeException
	 */
	private function openForWriting(): void
	{
		@mkdir(dirname($this->getPathname()), 0775, true);
		$this->open('wb9');
	}

	private function rewind(): void
	{
		if (is_resource($this->fileHandle) == false) {
			return;
		}

		gzrewind($this->fileHandle);
	}

	/**
	 * @param string $data
	 */
	private function write(string $data): void
	{
		gzwrite($this->fileHandle, $data);
	}

	public function close(): void
	{
		if (is_resource($this->fileHandle) == false) {
			return;
		}

		gzclose($this->fileHandle);

		$this->fileHandle = null;
	}

	/**
	 * Größe des JSON-Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes
	 */
	public function getPersistedSize(): int
	{
		return (int)@filesize($this->getPathname());
	}
}
