<?php

namespace App\Component\LogStream\Store;

use App\Component\LogStream\RepositoryInterface;
use App\Component\LogStream\QueryFilterIterator;
use App\Component\LogStream\JsonSeekableIterator;
use App\Model\User;
use App\Model\Domain;
use App\Utils;
use ArrayIterator;
use Generator;
use Iterator;
use RuntimeException;
use App\Event\ProtocolRotateEvent;
use App\Exception\LogNotReadableException;
use App\Exception\LogNotWritableException;
use App\Exception\LogNotCreatableException;
use App\Exception\LogNotDeletableException;

class JsonStore implements RepositoryInterface
{
	/** @var string $pathname */
	private $pathname;

	/** @var resource|null $fileHandle */
	private $fileHandle;

	/**
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @param string $name
	 * @throws LogNotCreatableException Wenn Logdatei nicht geschrieben werden kann.
	 */
	public function __construct(?User $user, ?Domain $domain, string $name)
	{
		$this->fileHandle = null;

		if ($domain) {
			$base = $domain->configDir(false);
		}
		elseif ($user) {
			$base = $user->configDir(false);
		}
		else {
			$base = Utils::getBundledVarDir();
		}
		$this->pathname = $base . DIRECTORY_SEPARATOR . $name . '.log';

		self::createFileIfNotExists();
	}

	public function __destruct()
	{
		$this->close();
	}

	/**
	 * Durchsuchen des Protokolls.
	 *
	 * @param array $query MongoDB-Query-Array mit einfachen Prädikaten
	 * @param bool $reversed
	 * @return Iterator<array>
	 * @see https://docs.mongodb.com/manual/reference/operator/query/
	 */
	public function filter(array $query, bool $reversed=false): Iterator
	{
		return new QueryFilterIterator($this->all($reversed), $query);
	}

	/**
	 * Iteriert über alle Einträge im Log
	 *
	 * @param bool $reversed
	 * @return \Iterator<array>
	 */
	public function all(bool $reversed=false): Iterator
	{
		if ($reversed) {
			// Geht leider nicht viel effizienter
			$result = [];
			foreach ($this->iterateLines() as $line) {
				if (is_array($data = json_decode($line, true))) {
					$result[] = $data;
				}
			}
			return new ArrayIterator(array_reverse($result));
		}
		else {
			return new JsonSeekableIterator(function () { return $this->iterateLines(); });
		}
	}

	public function count(): int
	{
		$this->openForReading();

		$counter = 0;
		while (($line = fgets($this->fileHandle)) !== false) {
			if (empty($line) || substr($line, 0, 1) == '#') {
				continue;
			}

			$counter++;
		}

		$this->close();
		return $counter;
	}

	/**
	 * @param array $data
	 */
	public function append(array $data): void
	{
		$this->openForWriting();
		$this->write(json_encode($data)."\n");
		$this->close();
	}

	/**
	 * @param iterable<array> $datas
	 */
	public function appendMulti(iterable $datas, bool $keepOpen = false): void
	{
		$this->openForWriting();

		foreach ($datas as $data) {
			$this->write(json_encode($data)."\n");
		}

		$this->close();
	}

	/**
	 * Protokoll rotieren/archivieren
	 *
	 * @param array|null $keepQueryHint (nicht unterstützt; ignoriert)
	 *
	 * Vor dem eigentlichen Prozess wird ein ProtocolRotateEvent ausgelöst.
	 */
	public function rotateProtocol(?array $keepQueryHint = null): void
	{
		if (!is_resource($this->fileHandle)) {
			$this->openForWriting();
		}
		flock($this->fileHandle, LOCK_SH);
		Utils::dispatchEvent(ProtocolRotateEvent::NAME, new ProtocolRotateEvent());
		flock($this->fileHandle, LOCK_EX);

		$pathinfo = pathinfo($this->pathname);
		$i = 0;
		do {
			// Pfad umbiegen, wenn über das Environment festgelegt
			$customArchiveRoot = Utils::getCustomArchiveRoot();
			if ($customArchiveRoot !== null) {
				$pathinfo['dirname'] = rtrim($customArchiveRoot, '/\\') . substr($pathinfo['dirname'], strlen(Utils::getBundledVarDir() . DIRECTORY_SEPARATOR . 'user_data'));
			}
			$destPathname = $pathinfo['dirname'].'/'.$pathinfo['filename'].date('~Ymd-His').(($i == 0) ? '' : '-'.$i).'.'.($pathinfo['extension'] ?? '');
			$i++;
		} while (file_exists($destPathname));

		$compressed = false;

		if (is_resource($logHandle = @fopen($this->pathname, 'rb'))) {
			if (is_resource($gzHandle = gzopen($destPathname.'.gz', 'wb9'))) {
				// Compress file in chunks of 64 KB in size
				while (feof($logHandle) == false && ($buffer = fread($logHandle, pow(2, 16))) !== false) {
					gzwrite($gzHandle, $buffer);
				}

				$compressed = true;
				gzclose($gzHandle);
			}

			fclose($logHandle);
		}
		elseif ($compressed == false) {
			// Datei kopieren statt umzubenennen, da ein Handle geöffnet ist
			copy($this->pathname, $destPathname);
		}

		ftruncate($this->fileHandle, 0);
	}

	/**
	 * {@inheritdoc}
	 */
	public function trimBefore(array $query): int
	{
		$check = QueryFilterIterator::compileFilter('$implicit_and', $query);

		$this->openForReadingAndWriting(false);
		flock($this->fileHandle, LOCK_EX);

		$position = 0;
		$skipHead = 0;
		$count = 0;
		try {
			// Datei-Position bestimmen, vor der gelöscht werden soll
			while (($line = fgets($this->fileHandle)) !== false) {
				if (empty($line) || substr($line, 0, 1) == '#') {
					if ($position == 0) {
						$skipHead = ftell($this->fileHandle);
					}
					continue;
				}

				$row = json_decode($line, true);
				if ($row and $check($row) == false) {
					break;
				}
				$position = ftell($this->fileHandle);
				$count++;
			}

			// Wenn alles aktuell, dann ist nichts zu tun
			if ($position-$skipHead <= 0) {
				return 0;
			}

			// Daten in Blöcken verschieben
			$blocksize = min($position-$skipHead, 524288);
			$readOffset = $position;
			$writeOffset = $skipHead;
			$retries = 5;
			while (!feof($this->fileHandle)) {
				// Block einlesen
				$this->seek($readOffset);
				$block = fread($this->fileHandle, $blocksize);
				if ($block === '') {
					// Ende der Datei erreicht
					break;
				}
				// Block schreiben
				$this->seek($writeOffset);
				$writtenCount = fwrite($this->fileHandle, $block);
				// Bei zu häufigen Schreibfehlern aufgeben
				if ($writtenCount == 0) {
					if ($retries == 0) {
						throw new RuntimeException('trimming logfile failed. Log might be in an inconsistent state!');
					}
					$retries--;
				}
				// Schreibgröße verkleinern, wenn nicht so viel auf einmal geschrieben werden kann
				if ($writtenCount > 0 and $writtenCount < $blocksize) {
					$blocksize = $writtenCount;
				}
				$readOffset += $writtenCount;
				$writeOffset += $writtenCount;
			}
			// Dateigröße anpassen (verkleinern)
			$this->rewind();
			ftruncate($this->fileHandle, $writeOffset);
		}
		finally {
			flock($this->fileHandle, LOCK_UN);
			$this->close();
		}
		return $count;
	}

	/**
	 * Leeren des Repositorys.
	 *
	 * Löscht die zugrundeliegende Datei.
	 *
	 * @throws LogNotDeletableException
	 */
	public function clear(): void
	{
		$this->close();
		$filename = $this->pathname;
		if (file_exists($filename) and @unlink($filename) === false) {
			throw new LogNotDeletableException($filename);
		}
	}

	/**
	 * Iteriert über alle Einträge im Log ohne sie zu dekodieren
	 *
	 * @return Generator|string[]
	 * @phan-return generator<string>
	 */
	private function iterateLines(): Generator
	{
		$this->openForReading();
		try {
			while (($line = fgets($this->fileHandle)) !== false) {
				if (empty($line) || substr($line, 0, 1) == '#') {
					continue;
				}
				yield $line;
			}
		}
		finally {
			$this->close();
		}
	}

	/**
	 * @throws LogNotCreatableException
	 */
	private function createFileIfNotExists(): void
	{
		$filename = $this->pathname;
		$firstLine = "# This file was generated automatically, please do not edit.\n";
		if (is_file($filename) == false && @file_put_contents($filename, $firstLine) === false) {
			throw new LogNotCreatableException($filename);
		}
	}

	/**
	 * Diese Methode gibt den Dateipfad der Konfigurationsdatei zurück.
	 * @return string
	 */
	public function getPathname(): string
	{
		return $this->pathname;
	}

	/**
	 * @param string $mode
	 * @param int $lock
	 * @throws LogNotReadableException
	 * @throws LogNotWritableException
	 */
	private function open(string $mode, int $lock): void
	{
		if (is_resource($this->fileHandle)) {
			return;
		}

		$filename = $this->getPathname();

		$this->fileHandle = @fopen($filename, $mode);

		if (is_resource($this->fileHandle) == false) {
			if ($mode == 'r') {
				throw new LogNotReadableException($filename);
			} else {
				throw new LogNotWritableException($filename);
			}
		}

		flock($this->fileHandle, $lock);
	}

	/**
	 * @throws RuntimeException
	 */
	private function openForReading(): void
	{
		$this->open('r', LOCK_SH);
	}

	/**
	 * @param bool $truncate
	 * @throws RuntimeException
	 */
	private function openForReadingAndWriting(bool $truncate = false): void
	{
		$this->open($truncate ? 'w+' : 'r+', LOCK_EX);
	}

	/**
	 * @param bool $truncate
	 * @throws RuntimeException
	 */
	private function openForWriting(bool $truncate = false): void
	{
		@mkdir(dirname($this->getPathname()));
		$this->open($truncate ? 'w' : 'a', LOCK_EX);
	}

	private function rewind(): void
	{
		if (is_resource($this->fileHandle) == false) {
			return;
		}

		rewind($this->fileHandle);
	}

	/**
	 * @see fseek()
	 * @throws RuntimeException
	 */
	private function seek($offset, $whence=SEEK_SET): void
	{
		if (fseek($this->fileHandle, $offset, $whence) == -1) {
			throw new RuntimeException('Could not seek in logfile');
		}
	}

	/**
	 * @param string $data
	 */
	private function write(string $data): void
	{
		fwrite($this->fileHandle, $data);
	}

	private function close(): void
	{
		if (is_resource($this->fileHandle) == false) {
			return;
		}

		fflush($this->fileHandle);
		flock($this->fileHandle, LOCK_UN);
		fclose($this->fileHandle);

		$this->fileHandle = null;
	}

	/**
	 * Größe des JSON-Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes
	 */
	public function getPersistedSize(): int
	{
		return (int)@filesize($this->getPathname());
	}
}
