<?php

namespace App\Component\LogStream\Store;

use App\Component\LogStream\RepositoryInterface;
use App\Component\LogStream\MongoDbSeekableIterator;
use App\Component\MongoDbService;
use App\Event\ProtocolRotateEvent;
use App\Exception\LogNotWritableException;
use App\Exception\LogNotDeletableException;
use App\Exception\LogNotCreatableException;
use App\Model\User;
use App\Model\Domain;
use App\Utils;
use ArrayObject;
use Iterator;
use Traversable;
use MongoDB\BSON\ObjectId;
use MongoDB\Collection;
use MongoDB\Driver\Exception\CommandException;
use MongoDB\Driver\WriteConcern;
use MongoDB\Model\BSONDocument;


class MongoDb implements RepositoryInterface
{
	const MAX_ENTRIES_BEFORE_ROTATE = 10000;

	/** @var Collection $collection */
	private $collection;
	/** @var string $name */
	private $name;
	/** @var User|null $user */
	private $user;
	/** @var Domain|null $domain */
	private $domain;

	/**
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @param string $name
	 * @param bool $createIndices Wenn true, wird beim ersten Zugriff ein Index auf die Collection angelegt
	 *             (für domain-spezifisches auf [_domainId, _id] und für User-spezifisches auf [_userId, _id])
	 * @throws LogNotCreatableException Wenn die Log-Collection nicht geschrieben werden kann.
	 */
	public function __construct(?User $user, ?Domain $domain, string $name, bool $createIndices=true)
	{
		$service = MongoDbService::getInstance();
		$this->collection = $collection = $service->getDatabase()->selectCollection($name);

		$this->name = $name;
		$this->domain = $domain;
		$this->user = $user;

		if ($createIndices) {
			// Erstellen der Indices beim ersten Zugriff
			try {
				if ($collection->estimatedDocumentCount() === 0) {

					if (iterator_count($this->collection->listIndexes()) == 0) {
						// @phan-suppress-next-line PhanTypeMismatchArgumentInternal
						$wc = new WriteConcern(1, 5000);
						if ($domain) {
							$this->collection->createIndex(['_domainId'=>1, '_id'=>-1], ['writeConcern'=>$wc]);
						}
						elseif ($user) {
							$this->collection->createIndex(['_userId'=>1, '_id'=>-1], ['writeConcern'=>$wc]);
						}
					}
				}
			}
			catch (\Exception $exception) {
				throw new LogNotCreatableException('mongodb:'.$name, $exception->getCode(), $exception);
			}
		}
	}

	/**
	 * Durchsuchen des Protokolls.
	 *
	 * @param array $query MongoDB-Query-Array mit einfachen Prädikaten
	 * @param bool $reversed
	 * @return Iterator<array>
	 * @see https://docs.mongodb.com/manual/reference/operator/query/
	 */
	public function filter(array $query, bool $reversed=false): Iterator
	{
		$sort = ['_id' => $reversed ? -1 : 1];

		return new MongoDbSeekableIterator(
			$this->collection,
			$this->prepareFilterContext($query),
			[
				'sort' => $sort,
			],
			function ($document) {
				// Dokument in PHP-native Datentypen umwandeln
				$data = self::unpackDocument($document);
				unset($data['_id'], $data['_userId'], $data['_domainId']);
				return $data;
			}
		);
	}

	/**
	 * Iteriert über alle Einträge im Log
	 *
	 * @param bool $reversed
	 * @return \Iterator<array>
	 */
	public function all(bool $reversed=false): Iterator
	{
		return $this->filter([], $reversed);
	}

	/**
	 * Diese Methode entpackt ein Dokument das aus der MongoDB kommt. Dabei
	 * werden alle Objekte, BSONDocument, BSONArray etc., in PHP-native Arrays
	 * umgewandelt.
	 *
	 * Das Ergebnis ist zur weiteren Verwaltung in $this->data gedacht.
	 * @param BSONDocument|ArrayObject $document
	 * @return array
	 */
	private static function unpackDocument(ArrayObject $document): array
	{
		return array_map(function ($value) {
			if (is_object($value)) {
				if ($value instanceof ArrayObject) {
					$value = self::unpackDocument($value);
				}
				elseif ($value instanceof ObjectId) {
					$value = (string)$value;
				}
			}
			return $value;
		}, (array)$document);
	}

	/**
	 * Diese Methode fügt bei einem relevanten Repository die verknüpfte Domain
	 * oder den verknüpften Benutzer hinzu.
	 *
	 * Jedes Dokument sollte hier durchlaufen, bevor es an MongoDB geht.
	 * @param array $document
	 * @return array
	 */
	private function prepareDocumentContext(array $document): array
	{
		unset($document['_domainId'], $document['_userId']);

		if ($this->domain) {
			$document = ['_domainId'=>new ObjectId($this->domain->getId())] + $document;
		}
		elseif ($this->user) {
			$document = ['_userId'=>new ObjectId($this->user->getId())] + $document;
		}

		return $document;
	}

	/**
	 * Diese Methode sorgt bei einem relevanten Repository dafür, dass nur
	 * Dokumente berücksichtigt werden, die mit einer bestimmten Domain oder
	 * einem bestimmten Benutzer verknüpft sind.
	 *
	 * Jeder Filter sollte hier durchlaufen, bevor er an MongoDB geht.
	 * @param array $filter
	 * @return array
	 */
	private function prepareFilterContext(array $filter): array
	{
		if ($this->domain) {
			$filter = ['_domainId'=>new ObjectId($this->domain->getId())] + $filter;
		}

		elseif ($this->user) {
			$filter = ['_userId'=>new ObjectId($this->user->getId())] + $filter;
		}

		else {
			$filter = [
				'$and' => [
					['_domainId' => ['$exists' => false]],
					['_userId' => ['$exists' => false]],
					$filter,
				],
			];
			if (!$filter['$and'][2]) {
				unset($filter['$and'][2]);
			}
		}

		return $filter;
	}

	public function count(): int
	{
		return $this->collection->countDocuments($this->prepareFilterContext([]), []);
	}

	/**
	 * @param array $data
	 */
	public function append(array $data): void
	{
		try {
			$this->collection->insertOne($this->prepareDocumentContext($data), []);
		} catch (\Exception $exception) {
			throw new LogNotWritableException('mongodb:'.$this->name, $exception->getCode(), $exception);
		}
	}

	/**
	 * @param iterable<array> $datas
	 */
	public function appendMulti(iterable $datas, bool $keepOpen = false): void
	{
		$documents = [];
		foreach ($datas as $data) {
			$documents[] = $this->prepareDocumentContext($data);
		}

		try {
			$this->collection->insertMany($documents, []);
		} catch (\Exception $exception) {
			throw new LogNotWritableException('mongodb:'.$this->name, $exception->getCode(), $exception);
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function rotateProtocol(?array $keepQueryHint = null): void
	{
		// Erstelle Archiv; wenn möglich GZip-komprimiert
		try {
			MongoDbService::getInstance()->getDatabase()->createCollection($this->name.'.archive', [
				'storageEngine' => [
					'wiredTiger' => [ 'configString' => "block_compressor=zlib" ],
				],
			]);
		}
		catch (CommandException $ignore) {
			// ignore
		}

		Utils::dispatchEvent(ProtocolRotateEvent::NAME, new ProtocolRotateEvent());

		$cutoffFilter = $keepQueryHint ? ['$nor' => [$keepQueryHint]] : [];

		// ID des neusten Eintrags bestimmen
		$cutoff = $this->collection->findOne($this->prepareFilterContext($cutoffFilter), [
			'sort' => ['_id' => -1],
			'projection'=>['_id'=>1, 'timestamp'=>1],
		]);
		// Abbrechen, wenn leer.
		if (!$cutoff) {
			return;
		}

		$filter = $this->prepareFilterContext(['_id' => ['$lte' => $cutoff['_id']]]);

		if (Utils::getArchiveRepositoryStore() === Utils::REPOSITORY_STORE_FILE) {
			// Wenn Archiv in Dateien erfolgen soll:
			$i = 0;
			do {
				$gzName = $this->name . date('~Ymd-His', $cutoff['timestamp']) . ($i == 0 ? '' : '-'.$i);
				$gzRepo = new GzJsonStore($this->user, $this->domain, $gzName, Utils::getCustomArchiveRoot());
				++$i;
			} while ($gzRepo->exists());

			$cursor = $this->collection->find($filter, ['sort'=>['_id'=>1]]);
			$gzRepo->createOrOverwrite();
			$gzRepo->appendMulti($cursor);
			$gzRepo->close();
			// Kopierte Einträge löschen
			$this->collection->deleteMany($filter);
		}
		else {
			// Ansonsten Archiv in anderer DB-Collection:
			try {
				// Einträge in Archiv kopieren
				try {
					$this->collection->aggregate([
						[ '$match' => $filter ],
						[ '$merge' => [ 'into' => $this->name.'.archive', ] ]
					]);
				}
				catch (CommandException $exception) {
					// Fallback für MongoDB < 4.2
					$archiveCollection = MongoDbService::getInstance()->getDatabase()->selectCollection($this->name.'.archive');
					$cursor = $this->collection->find($filter, ['sort'=>['_id'=>1]]);
					foreach (Utils::chunk(100, $cursor) as $chunk) {
						$archiveCollection->insertMany($chunk);
					}
				}
				// Kopierte Einträge löschen
				$this->collection->deleteMany($filter);
			} catch (\Exception $exception) {
				throw new LogNotWritableException('mongodb:'.$this->name.'.archive', $exception->getCode(), $exception);
			}
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function trimBefore(array $query): int
	{
		$cutoff = $this->collection->findOne($this->prepareFilterContext($query), [
			'sort' => ['_id' => 1],
			'projection'=>['_id'=>1],
		]);

		if (!$cutoff) {
			return 0;
		}

		try {
			$result = $this->collection->deleteMany($this->prepareFilterContext(['_id' => ['$lt' => $cutoff['_id']]]), []);
		} catch (\Exception $exception) {
			throw new LogNotDeletableException('mongodb:'.$this->name, $exception->getCode(), $exception);
		}

		$count = $result->getDeletedCount();
		return $count;
	}

	/**
	 * Löscht Einträge aus dem Log
	 *
	 * Nur in Ausnahmefällen verwenden, da nicht unbedingt von anderen Repository-Typen unterstützt.
	 *
	 * @param array $query
	 * @return int
	 */
	public function deleteByFilter(array $query): int
	{
		$filter = $this->prepareFilterContext($query);

		try {
			$result = $this->collection->deleteMany($filter, []);
		} catch (\Exception $exception) {
			throw new LogNotDeletableException('mongodb:'.$this->name, $exception->getCode(), $exception);
		}

		$count = $result->getDeletedCount();
		return $count;
	}

	/**
	 * Leeren des Repositorys.
	 */
	public function clear(): void
	{
		try {
			$this->collection->deleteMany($this->prepareFilterContext([]), []);
		} catch (\Exception $exception) {
			throw new LogNotDeletableException('mongodb:'.$this->name, $exception->getCode(), $exception);
		}
		try {
			MongoDbService::getInstance()->getDatabase()->command(['compact' => $this->name]);
		} catch (\Exception $exception) {
			// ignore
		}

	}

	/**
	 * Low-Level-Aggregation direkt auf der MongoDB-Datenbank durchführen.
	 *
	 * Nur in Ausnahmefällen verwenden, da nicht von anderen Repository-Typen unterstützt.
	 */
	public function aggregate(array $pipeline): Traversable
	{
		array_unshift($pipeline, ['$match'=>$this->prepareFilterContext([])]);
		return $this->collection->aggregate($pipeline);
	}

	/**
	 * N/A
	 *
	 * @return int Immer -1
	 */
	public function getPersistedSize(): int
	{
		return -1;
	}
}
