<?php

namespace App\Component\LogStream\Store;

use App\Component\LogStream\RepositoryInterface;
use App\Component\LogStream\RedisScoreRangeIterator;
use App\Component\LogStream\QueryFilterIterator;
use App\Component\MongoDbService;
use App\Event\ProtocolRotateEvent;
use App\Exception\LogNotWritableException;
use App\Exception\LogNotDeletableException;
use App\Model\User;
use App\Model\Domain;
use App\Utils;
use Iterator;
use MongoDB\BSON\ObjectId;
use MongoDB\Driver\Exception\CommandException;
use Predis\Client as Redis;

class RedisStore implements RepositoryInterface
{
	const MAX_ENTRIES_BEFORE_ROTATE = 10000;

	/** @var Redis $redis */
	private $redis;
	/** @var string $key */
	private $key;
	/** @var string $name */
	private $name;
	/** @var User|null $user */
	private $user;
	/** @var Domain|null $domain */
	private $domain;
	/** @var string $scoreKey */
	private $scoreKey;
	/** @var int|null $expire */
	private $expireCollection;

	/**
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @param string $name
	 * @param string $scoreKey Der Schlüssel in einem Eintrag mit dem Score (z.B. ein Unix-Timestamp)
	 * @param int|null $expireCollection Löscht die gesamte Collection (pro Domain) x Sekunden nach dem letzten Schreibzugriff
	 */
	public function __construct(?User $user, ?Domain $domain, string $name, string $scoreKey='timestamp', ?int $expireCollection=null)
	{
		$this->redis = $GLOBALS['kernel']->getContainer()->get('Redis');
		$key = $name;
		if ($user) {
			// API-Key statt User-ID für optimiertes Schreiben z.B. des Consent-Logs durch ein externes Programm
			$key .= ':' . $user->getApiKey();
		}
		if ($domain) {
			$key .= ':' . $domain->getId();
		}

		$this->name = $name;
		$this->key = $key;
		$this->user = $user;
		$this->domain = $domain;
		$this->scoreKey = $scoreKey;
		$this->expireCollection = $expireCollection;
	}

	/**
	 * Durchsuchen des Protokolls.
	 *
	 * @param array $query MongoDB-Query-Array mit einfachen Prädikaten
	 * @param bool $reversed
	 * @return Iterator<array>
	 * @see https://docs.mongodb.com/manual/reference/operator/query/
	 */
	public function filter(array $query, bool $reversed=false): Iterator
	{
		// Queries auf `timestamp` zu Score Ranges umbauen.
		// Maximal bis "Jetzt - 2 Sekunden", um Races mit neuen Einträgen zu reduzieren
		$maxtimestamp = time()-2;
		$start = '-inf';
		$end = $maxtimestamp;
		if (isset($query[$this->scoreKey]['$gte'])) {
			$start = $query[$this->scoreKey]['$gte'];
			unset($query[$this->scoreKey]['$gte']);
		}
		if (isset($query[$this->scoreKey]['$gt'])) {
			$start = '('.$query[$this->scoreKey]['$gt'];
			unset($query[$this->scoreKey]['$gt']);
		}
		if (isset($query[$this->scoreKey]['$lte'])) {
			if ($query[$this->scoreKey]['$lte'] < $maxtimestamp) {
				$end = $query[$this->scoreKey]['$lte'];
			}
			unset($query[$this->scoreKey]['$lte']);
		}
		if (isset($query[$this->scoreKey]['$lt'])) {
			if ($query[$this->scoreKey]['$lt'] <= $maxtimestamp) {
				$end = '('.$query[$this->scoreKey]['$lt'];
			}
			unset($query[$this->scoreKey]['$lt']);
		}
		if (isset($query[$this->scoreKey]['$eq'])) {
			$start = (int)$query[$this->scoreKey]['$lt'];
			$end = '('.($start + 1);
			unset($query[$this->scoreKey]['$eq']);
		}
		if (isset($query[$this->scoreKey]) and !$query[$this->scoreKey]) {
			unset($query[$this->scoreKey]);
		}

		// Iterator erstellen
		$iterator = $this->iterTimestampRange($start, $end, $reversed);

		// Falls noch ein Query übrig bleibt, in einen FilterIterator wrappen
		if ($query) {
			$iterator = new QueryFilterIterator($iterator, $query);
		}

		return $iterator;
	}

	/**
	 * @param string|int|float $start
	 * @param string|int|float $end
	 * @param bool $reversed
	 * @param int $chunkSize
	 * @return Iterator<array>
	 */
	private function iterTimestampRange($start='-inf', $end='+inf', bool $reversed=false, int $chunkSize=250): Iterator
	{
		return new RedisScoreRangeIterator(
			$this->redis,
			$this->key,
			$start,
			$end,
			$reversed,
			function ($document) {
				// Dokument in PHP-native Datentypen umwandeln
				$array = json_decode($document, true, JSON_INVALID_UTF8_SUBSTITUTE);
				if (!is_array($array)) {
					return [];
				}
				return $array;
			},
			$chunkSize
		);
	}

	/**
	 * Iteriert über alle Einträge im Log
	 *
	 * @param bool $reversed
	 * @return \Iterator<array>
	 */
	public function all(bool $reversed=false): Iterator
	{
		return $this->filter([], $reversed);
	}

	public function count(): int
	{
		return $this->redis->zCard($this->key);
	}

	/**
	 * @param array $data
	 */
	public function append(array $data): void
	{
		$score = (float)($data[$this->scoreKey] ?? microtime(true));
		try {
			// False Positive ignorieren
			// @phan-suppress-next-line PhanParamTooMany, PhanTypeMismatchArgumentProbablyReal
			$this->redis->zAdd($this->key, $score, (string)json_encode($data));
			if ($this->expireCollection !== null) {
				$this->redis->expire($this->key, $this->expireCollection);
			}
		} catch (\Exception $exception) {
			throw new LogNotWritableException('redis:'.$this->key, $exception->getCode(), $exception);
		}
	}

	/**
	 * @param iterable<array> $datas
	 */
	public function appendMulti(iterable $datas, bool $keepOpen = false): void
	{
		$timestamp = microtime(true);
		$arguments = [$this->key];
		foreach ($datas as $data) {
			$arguments[] = (float)($data[$this->scoreKey] ?? $timestamp);
			$arguments[] = (string)json_encode($data);
		}

		try {
			call_user_func_array([$this->redis, 'zAdd'], $arguments);
			if ($this->expireCollection !== null) {
				$this->redis->expire($this->key, $this->expireCollection);
			}
		} catch (\Exception $exception) {
			throw new LogNotWritableException('redis:'.$this->key, $exception->getCode(), $exception);
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function rotateProtocol(?array $keepQueryHint = null): void
	{
		Utils::dispatchEvent(ProtocolRotateEvent::NAME, new ProtocolRotateEvent());

		$maxtimestamp = time()-2;

		$customArchiveRoot = Utils::getCustomArchiveRoot();

		// Anzahl der Datensätze holen und bei 0 abbrechen
		$count = (int)$this->redis->zCount($this->key, '-inf', $maxtimestamp);
		if ($count === 0) {
			return;
		}
		// Kleinsten und größten Zeitstempel bestimmen
		$minScore = $this->redis->zRangeByScore($this->key, '-inf', $maxtimestamp, ['limit' => [0, 1], 'withscores' => TRUE]);
		$minScore = (int)reset($minScore);
		$maxScore = $this->redis->zRevRangeByScore($this->key, $maxtimestamp, '-inf', ['limit' => [0, 1], 'withscores' => TRUE]);
		$maxScore = (int)reset($maxScore);
		// Sehr große Logs in mehreren Schritten verarbeiten, unter der Annahme,
		// dass die Einträge über die Zeit ungefähr gleichverteilt sind.
		// Aufgeteilt wird in im Schnitt 50.000 Einträge, aber max. so viele, dass
		// mindestens 5 Minuten pro Chunk verarbeitet werden.
		$splitInto = max(1, min((int)ceil($count / 50000), (int)floor(($maxScore-$minScore) / 300)));
		$scoreChunkSize = (int)ceil(($maxScore-$minScore) / $splitInto);

		if (Utils::getArchiveRepositoryStore() === Utils::REPOSITORY_STORE_MONGODB) {
			// Archivieren in MongoDB:
			$db = MongoDbService::getInstance()->getDatabase();
			// Erstelle Archiv; wenn möglich GZip-komprimiert
			try {
				$db->createCollection($this->name.'.archive', [
					'storageEngine' => [
						'wiredTiger' => [ 'configString' => "block_compressor=zlib" ],
					],
				]);
			}
			catch (CommandException $ignore) {
				// ignore
			}

			try {
				$uniqId = bin2hex(random_bytes(5));
				$counter = rand(0, 16777215);
				$chunkMinScore = $minScore;
				while ($chunkMinScore <= $maxScore) {
					$chunkMaxScore = min($maxScore, $chunkMinScore + $scoreChunkSize);

					// Einträge in Archiv kopieren
					$archiveCollection = $db->selectCollection($this->name.'.archive');
					$cursor = $this->iterTimestampRange($chunkMinScore, $chunkMaxScore, false, 500);

					$userId = $this->user ? new ObjectId($this->user->getId()) : null;
					$domainId = $this->domain ? new ObjectId($this->domain->getId()) : null;

					foreach (Utils::chunk(100, $cursor) as $subchunk) {
						foreach ($subchunk as &$item) {
							$timestamp = (int)($item[$this->scoreKey] ?? time());
							$item['_id'] = new ObjectId(substr(str_pad(dechex($item[$this->scoreKey]), 8, '0', STR_PAD_LEFT), -8, 8).$uniqId.str_pad(dechex($counter), 6, '0', STR_PAD_LEFT));
							$counter = ($counter + 1) & 16777215;
							if ($domainId !== null) {
								$item['_domainId'] = $domainId;
							}
							elseif ($userId !== null) {
								$item['_userId'] = $userId;
							}
						}
						unset($item);
						$archiveCollection->insertMany($subchunk);
					}
					// Kopierte Einträge löschen
					$this->redis->zRemRangeByScore($this->key, $chunkMinScore, $chunkMaxScore);

					$chunkMinScore += $scoreChunkSize + 1;
				}
			} catch (\Exception $exception) {
				throw new LogNotWritableException('mongodb:'.$this->name.'.archive', $exception->getCode(), $exception);
			}


		}
		else {
			// Ansonsten Archiv in Dateien:
			$chunkMinScore = $minScore;
			while ($chunkMinScore <= $maxScore) {
				$chunkMaxScore = min($maxScore, $chunkMinScore + $scoreChunkSize);

				$i = 0;
				do {
					$gzName = $this->name . date('~Ymd-His', $chunkMaxScore) . ($i == 0 ? '' : '-'.$i);
					$gzRepo = new GzJsonStore($this->user, $this->domain, $gzName, Utils::getCustomArchiveRoot());
					++$i;
				} while ($gzRepo->exists());

				$cursor = $this->iterTimestampRange($chunkMinScore, $chunkMaxScore, false, 500);
				$gzRepo->createOrOverwrite();
				$gzRepo->appendMulti($cursor);
				$gzRepo->close();
				// Kopierte Einträge löschen
				$this->redis->zRemRangeByScore($this->key, $chunkMinScore, $chunkMaxScore);

				$chunkMinScore += $scoreChunkSize + 1;
			}
		}

		if ($this->expireCollection !== null) {
			$this->redis->expire($this->key, $this->expireCollection);
		}
	}

	/**
	 * {@inheritdoc}
	 *
	 * Nur $gt/$gte-Queries auf `timestamp` werden unterstützt.
	 */
	public function trimBefore(array $query): int
	{
		$start = '-inf';
		$end = time()-2;

		if (isset($query[$this->scoreKey]['$gte'])) {
			$end = '('.$query[$this->scoreKey]['$gte'];
			unset($query[$this->scoreKey]['$gte']);
		}
		if (isset($query[$this->scoreKey]['$gt'])) {
			$end = $query[$this->scoreKey]['$gt'];
			unset($query[$this->scoreKey]['$gt']);
		}
		if (isset($query[$this->scoreKey]) and !$query[$this->scoreKey]) {
			unset($query[$this->scoreKey]);
		}

		if ($query) {
			throw new \UnexpectedValueException("unsupported query for RedisStore::trimBefore()");
		}

		if ($this->expireCollection !== null) {
			$this->redis->expire($this->key, $this->expireCollection);
		}
		return (int)$this->redis->zRemRangeByScore($this->key, $start, $end);
	}

	/**
	 * Leeren des Repositorys.
	 */
	public function clear(): void
	{
		try {
			$this->redis->del($this->key);
		} catch (\Exception $exception) {
			throw new LogNotDeletableException('redis:'.$this->key, $exception->getCode(), $exception);
		}
	}

	/**
	 * Größe des Redis-Keys im Redis-Server (inklusive Overhead)
	 *
	 * @return int Größe in Bytes
	 */
	public function getPersistedSize(): int
	{
		try {
			return (int)$this->redis->rawCommand("MEMORY", "USAGE", $this->key);
		} catch (\Exception $ignore) {
			return -1;
		}
	}
}
