<?php

namespace App\Component\Mail;

use App\Component\MailSubscriptions;
use App\Config;
use App\Model\User;
use App\Utils;
use Symfony\Component\Mime\Email;
use Symfony\Component\Mime\Address;

/**
 * E-Mail-Nachricht
 */
class MailMessage extends Email
{
	/** @var string|null $unsubscribeLink */
	private $unsubscribeLink = null;

	/**
	 * @param ?string $subject Betreff
	 * @phan-suppress PhanTypeMismatchArgument
	 *
	 * Weitgehend wie `Email`, setzt aber automatisch den Absender.
	 */
	public function __construct($subject = null)
	{
		parent::__construct();
		if ($subject !== null) {
			$this->subject($subject);
		}
		$config = Config::getInstance();
		$sender = $config->get('senderEmail');
		$senderName = $config->get('senderName', '');
		$replyTo = $config->get('replyTo');

		if (!$sender) {
			$sender = $config->get('email');
		}
		if ($senderName != '') {
			$sender = new Address($sender, $senderName);
		}
		$this->from($sender);

		if ($replyTo and $replyTo !== $sender) {
			$this->replyTo($replyTo);
		}

		$headers = $this->getHeaders();
		$headers->addTextHeader('X-Face', 'C#x"pQhvrI4K*<4%v6[Lz[8[uDt0@Aju+SQcp|Htc`\%<kQgsm)p/78;pT*eQWjXV|$AU@]/Xe>P0|t|&@+hgqWhSpmR:+yG_uJUP(j/qTWfh*$aaGCxkA@MpZKp0M;9EyP?b`9u\'7"UW9#lMn)@#@ju#u');
		$headers->addTextHeader('X-Auto-Response-Suppress', 'All');
	}

	/**
	 * @param string $html HTML
	 * @return string Plaintext
	 */
	public static function html2text(string $html): string
	{
		// HTML to Text
		$result = strtr($html, array("\r\n"=>' ', "\n"=>' ', "\r"=>' '));
		$result = preg_replace('/.*<body/i', "<body", $result);
		$result = preg_replace('/<br ?\/?'.'>/i', "$0\n", $result);
		$result = preg_replace('/<li.*?>/i', "- $0", $result);
		$result = preg_replace('/<\/(?:p|pre|div|h[1-6]|aside|article|section|main|address|nav|ul|ol|li|blockquote|form|tr|dd)>/i', "$0\n\n", $result);
		$result = preg_replace('/<script.*?<\/script>/i', "\n", $result);
		$result = preg_replace('/<a[^>]*?href="([^"]*)"[^>]*?>\s*<img.*?<\/a>/i', '', $result);
		$result = preg_replace('/<a[^>]*?href="([^"]*)"[^>]*?>(.*?)<\/a>/i', '$0: $1', $result);
		$result = preg_replace('/<style.*?<\/style>/i', "\n", $result);
		$result = preg_replace('/<\/(?:td|th|dt)>/i', "$0 ", $result);
		$result = strip_tags($result);
		$result = html_entity_decode($result, ENT_HTML5|ENT_COMPAT, 'UTF-8');
		// Leading Spaces auf jeder Zeile entfernen
		$result = preg_replace('/^[ \t]+/m', "", $result);
		// Trailing Spaces auf jeder Zeile entfernen
		$result = preg_replace('/[ \t]+$/m', "", $result);
		// Newlines zusammenziehen
		$result = preg_replace('/\n\n+/is', "\n\n", $result);
		return trim($result)."\n";
	}

	/**
	 * @return $this
	 */
	public function voidReplyTo(): self
	{
		$headers = $this->getHeaders();
		$headers->remove('Reply-To');
		$headers->addTextHeader('Reply-To', '<>');
		return $this;
	}

	/**
	 * Mail als automatisiert markieren
	 *
	 * @param string $type no|auto-generated|auto-replied
	 * @return $this
	 */
	public function autoSubmitted(string $type='auto-generated'): self
	{
		$headers = $this->getHeaders();
		$headers->remove('Auto-Submitted');
		$headers->addTextHeader('Auto-Submitted', $type);
		return $this;
	}

	/**
	 * List-ID-Header hinzufügen
	 *
	 * Für Mailinglisten/Newsletter/automatische periodische Mails:
	 * ID die alle Mails einer Versandgruppe zusammenfasst, also z.B. `AutoUpdate-success.cloud.ccm19.de`.
	 * 'ccm19' und der Hostname des Servers wird automatisch angehängt.
	 *
	 * Erwünscht, aber nicht zwingend nötig.
	 *
	 * @param string $localPart Erste Hälfte der ID
	 * @return $this
	 */
	public function listId(string $localPart): self
	{
		$headers = $this->getHeaders();
		$headers->remove('List-Id');
		$headers->addTextHeader('List-Id', '<'.$localPart.'.ccm19.'.$_SERVER['HTTP_HOST'].'>');
		return $this;
	}

	/**
	 * Unsubscribe-Header und Link hinzufügen
	 *
	 * `text()` bzw. `html()` dürfen erst nach diesem Aufruf ausgeführt werden, damit der Link
	 * in den Text eingebaut werden kann.
	 *
	 * @param User $user Benutzer für den der Link generiert werden soll
	 * @param string $subscriptionKey ID der E-Mail-Benachrichtigung (siehe MailSubscriptions.php)
	 * @return $this
	 */
	public function unsubscribe(User $user, string $subscriptionKey): self
	{
		/** @var MailSubscriptions $subscriptions */
		$subscriptions = Utils::getService('mail.subscriptions');
		assert($subscriptions instanceof MailSubscriptions);

		$this->unsubscribeLink = $subscriptions->buildUnsubscribeLink($user, $subscriptionKey);
		$headers = $this->getHeaders();
		$headers->remove('List-Unsubscribe');
		$headers->remove('List-Unsubscribe-Post');
		$headers->addTextHeader('List-Unsubscribe', '<'.$this->unsubscribeLink.'>');
		$headers->addTextHeader('List-Unsubscribe-Post', 'List-Unsubscribe=One-Click');
		return $this;
	}

	/**
	 * Set the html body of this entity.
	 *
	 * Overridden to automatically add Unsubscribe link and generate a text body.
	 *
	 * @param resource|string|null  $body
	 * @param string $charset optional
	 *
	 * @return $this
	 * @override
	 */
	public function html($body, string $charset = 'utf-8'): self
	{
		if (!is_string($body)) {
			return parent::html($body, $charset);
		}

		if ($this->unsubscribeLink !== null)  {
			$translator = Utils::getTranslator();
			$textBody = static::html2text($body) . "\n".$translator->trans("Unsubscribe: :link", [':link' => $this->unsubscribeLink]);
			$body .= "\n<br /><br /><small><a href=\"".htmlspecialchars($this->unsubscribeLink)."\">".$translator->trans("Unsubscribe").'</a></small>';
			parent::text($textBody);
		} else {
			parent::text(static::html2text($body));
		}
		return parent::html($body);
	}

	/**
	 * Set the text body of this entity.
	 *
	 * Overridden to automatically add Unsubscribe link.
	 *
	 * @param resource|string|null  $body
	 * @param string $charset optional
	 *
	 * @return $this
	 * @override
	 */
	public function text($body, string $charset = 'utf-8'): self
	{
		if (!is_string($body)) {
			return parent::text($body, $charset);
		}

		if ($this->unsubscribeLink !== null)  {
			$translator = Utils::getTranslator();
			$body .= "\n".$translator->trans("Unsubscribe: :link", [':link' => $this->unsubscribeLink]);
		}
		return parent::text($body);
	}
}
