<?php

namespace App\Component\Mail;

use App\Component\LoggerInterface;
use Symfony\Component\Mailer\Envelope;
use Symfony\Component\Mailer\Exception\TransportExceptionInterface;
use Symfony\Component\Mailer\SentMessage;
use Symfony\Component\Mailer\Transport\TransportInterface;
use Symfony\Component\Mime\Address;
use Symfony\Component\Mime\RawMessage;

final class TransportLogWrapper implements TransportInterface
{
	/** @var TransportInterface $inner */
	private $inner;
	/** @var LoggerInterface $inner */
	private $logger;

	public function __construct(TransportInterface $inner, LoggerInterface $logger)
	{
		$this->inner = $inner;
		$this->logger = $logger;
	}

	public function send(RawMessage $message, Envelope $envelope = null): ?SentMessage
	{
		if ($envelope === null) {
			$envelope = Envelope::create($message);
		}

		$sender = $envelope->getSender()->getAddress();
		$recipients = array_map(function (Address $addr) {
			return '<'.$addr->getAddress().'>';
		}, $envelope->getRecipients());

		$subject = '<unknown>';
		if (method_exists($message, 'getHeaders')) {
			// @phan-suppress-next-line PhanUndeclaredMethod
			$subjectHeader = $message->getHeaders()->get('Subject');
			if ($subjectHeader) {
				$subject = (string)$subjectHeader->getBody();
			}
		}

		try {
			$result = $this->inner->send($message, $envelope);
			$lines = explode("\n", rtrim($result->getDebug()));
			$lastLine = trim(ltrim($lines[count($lines)-1], '<'));
			if (strlen($lastLine) > 100) {
				$lastLine = substr($lastLine, 0, 98).'…';
			}
			$this->logger->logInfo('Mail', sprintf(
				'Sent email to %s from <%s>; Subject: "%s", ID: <%s>, Result: "%s"',
				implode(', ', $recipients),
				$sender,
				addslashes($subject),
				$result->getMessageId(),
				$lastLine ?: '—'
			));
			return $result;
		} catch (TransportExceptionInterface $e) {
			$message = $e->getMessage() ?: 'Unknown error.';
			if (strlen($message) > 120) {
				$message = substr($message, 0, 118).'…';
			}
			$this->logger->logError('Mail', sprintf(
				'Failed to send email to %s from <%s>; Subject: "%s", Result: "%s"',
				implode(', ', $recipients),
				$sender,
				addslashes($subject),
				$message
			));
			throw $e;
		}
	}

	public function __toString(): string
	{
		return $this->inner->__toString();
	}

}
