<?php

namespace App\Component;

use App\Model\User;
use App\Utils;
use App\Config;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;
use Symfony\Component\Routing\RouterInterface;

final class MailSubscriptions
{
	private $router;

	public function __construct(RouterInterface $router)
	{
		$this->router = $router;
	}

	/**
	 * Listet alle für $user verfügbaren Mailings
	 *
	 * Wenn $user nicht angegeben wird, wird der aktuell angemeldete User genutzt
	 *
	 * Plugins können sich mit dem Filter-Event 'ccm19.mail.available-subscriptions'
	 * hinzufügen.
	 *
	 * @param User|null $user
	 * @return array<string, array>
	 */
	public function list(?User $user = null): array
	{
		if (!$user) {
			$user = User::loggedInUser();
		}
		if (!$user) {
			return [];
		}
		$subscriptions = [];
		if ($user->getRole() === User::ROLE_ADMIN) {
			$subscriptions['UpdateNotification'] = ['label'=>'An update is available'];
			$subscriptions['AutoUpdate-Success'] = ['label'=>'AutoUpdate succeeded'];
			$subscriptions['AutoUpdate-Failure'] = ['label'=>'AutoUpdate failed'];
			$subscriptions['AutoUpdate-Issues'] = ['label'=>'Issues prevent an AutoUpdate'];
		}
		if (!Utils::isExtendedEdition() or ($user->getRole() === User::ROLE_CLIENT and $user->isAccountOwner())) {
			$subscriptions['QuotaReport-SubMax'] = ['label'=>'Tariff call limit almost exceeded'];
			$subscriptions['QuotaReport-Exceeded'] = ['label'=>'Tariff call limit exceeded'];
			$subscriptions['CookieNotification'] = ['label'=>'New cookies found on a website'];
			$subscriptions['EmbeddingNotification'] = ['label'=>'New Embeddings found on a website'];
		}
		if (Utils::isExtendedEdition() and $user->getRole() === User::ROLE_CLIENT and $user->isAccountOwner()) {
			$subscriptions['ClientUpdateNotification'] = ['label'=>'Cookie management software was updated', 'default'=>false];
		}
		$subscriptions = Utils::filterWithEvent('ccm19.mail.available-subscriptions', $subscriptions);

		return $subscriptions;
	}

	/**
	 * @param User $user
	 * @param string $subscriptionKey
	 * @return string URL
	 */
	public function buildUnsubscribeLink(User $user, string $subscriptionKey): string  {
		$config = Config::getInstance();

		$scheme = $config->isForceHttpsConnection() ? 'https://' : $_SERVER['REQUEST_SCHEME'].'://';
		$baseUrl = $scheme.$_SERVER['HTTP_HOST'];
		return $baseUrl.$this->router->generate('app_mail_unsubscribe', [
			'key' => $subscriptionKey,
			'u' => $user->getApiKey(),
			't' => $this->buildValidationCode($user, $subscriptionKey),
		], UrlGeneratorInterface::ABSOLUTE_PATH);
	}

	/**
	 * @param User $user
	 * @param string $subscriptionKey
	 * @return string Schlüssel (Hex)
	 */
	public function buildValidationCode(User $user, string $subscriptionKey): string {
		return substr(hash_hmac('sha256', 'unsubscribe:'.strtolower($subscriptionKey), ($_ENV['APP_SECRET'] ?? '').Utils::getLicenseKey().$user->getUsername()), 0, 10);
	}
}
