<?php

namespace App\Component\Media;

use App\Model\User;
use App\Model\Domain;
use App\Utils;

class MediaFileService
{
	/** @var array<string, RepositoryInterface> $repositories */
	private $repositories = [];

	/**
	 * @return static
	 */
	public static function getInstance()
	{
		global $kernel;
		return $kernel->getContainer()->get('media.mediafileservice');
	}

	/**
	 * Repository öffnen
	 *
	 * Stattdessen sollte `getGlobalRepository()`, `getUserRepository()`
	 * oder `getDomainRepository()` verwendet werden.
	 *
	 * @param string $kind
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @return RepositoryInterface
	 */
	private function getRepository(string $kind, ?User $user, ?Domain $domain): RepositoryInterface
	{
		$storeType = Utils::getRepositoryStore();
		$cacheKey = sprintf('%s/%s/%s/%s', $storeType, ($user ? $user->getId() : ''), ($domain ? $domain->getId() : ''), $kind);
		if (isset($this->repositories[$cacheKey])) {
			return $this->repositories[$cacheKey];
		}

		if ($storeType == Utils::REPOSITORY_STORE_MONGODB) {
			$result = new Store\MongoDb('cm-media', $kind, $user, $domain);
		}
		else {
			$result = new Store\FileStore($kind, $user, $domain);
		}
		$this->repositories[$cacheKey] = $result;

		return $result;
	}

	/**
	 * Globales Repository öffnen
	 * @param string $kind
	 * @return RepositoryInterface
	 */
	public function getGlobalRepository(string $kind): RepositoryInterface
	{
		return $this->getRepository($kind, null, null);
	}

	/**
	 * (Haupt-)Benutzerspezifisches Repository öffnen
	 * @param string $kind
	 * @param User|null $user Abweichender Benutzer, ansonsten der aktuelle Hauptuser
	 * @return RepositoryInterface
	 */
	public function getUserRepository(string $kind, ?User $user=null): RepositoryInterface
	{
		if (!$user) {
			$user = User::loggedInUser();
		}
		return $this->getRepository($kind, $user->getAccountOwner(), null);
	}

	/**
	 * (Sub-)Benutzerspezifisches Repository öffnen
	 * @param string $kind
	 * @param User|null $user Abweichender Benutzer, ansonsten der aktuelle (Sub-)User
	 * @return RepositoryInterface
	 */
	public function getSubuserRepository(string $kind, ?User $user=null): RepositoryInterface
	{
		if (!$user) {
			$user = User::loggedInUser();
		}
		return $this->getRepository($kind, $user, null);
	}

	/**
	 * Domainspezifisches Repository öffnen
	 * @param string $kind
	 * @param Domain|null $domain Abweichende Domain, ansonsten die aktuelle
	 * @return RepositoryInterface
	 */
	public function getDomainRepository(string $kind, ?Domain $domain=null): RepositoryInterface
	{
		if (!$domain) {
			$domain = Domain::activeDomain();
		}

		return $this->getRepository($kind, $domain->getUser(), $domain);
	}
}
