<?php

namespace App\Component\Media\Store;

use App\Component\Media\RepositoryInterface;
use App\Exception\ConfigNotWritableException;
use App\Model\User;
use App\Model\Domain;
use App\Utils;
use InvalidArgumentException;
use RuntimeException;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\HttpFoundation\File\UploadedFile;

class FileStore implements RepositoryInterface
{
	/** @var string $path */
	private $path;

	/**
	 * Lädt ein Unterverzeichnis für Mediendateien relativ zum Projektverzeichnis <i>`var/`</i>.
	 * @param string $subdirName Name des Unterverzeichnisses
	 * @param User|null $user User, wenn benutzerspezifisch
	 * @param Domain|null $domain Domain, wenn domainspezifisch
	 */
	public function __construct(string $subdirName, ?User $user=null, ?Domain $domain=null)
	{
		if ($subdirName === '' or $subdirName[0] == '.' or strpos($subdirName, '/') !== false) {
			throw new InvalidArgumentException('subdirName invalid');
		}

		if ($domain) {
			$base = $domain->configDir(false);
		}
		elseif ($user) {
			$base = $user->configDir(false);
		}
		else {
			$base = Utils::getBundledVarDir();
		}
		$this->path = $base . DIRECTORY_SEPARATOR . $subdirName;
	}

	/**
	 * Leeren des Repositorys.
	 */
	public function clear(): void
	{
		Utils::removeDirectory($this->path);
	}

	/**
	 * {@inheritdoc}
	 */
	public function getFile(string $identifier): ?File
	{
		if ($identifier === '' or $identifier[0] == '.' or strpos($identifier, '/') !== false) {
			throw new InvalidArgumentException('file identifier invalid');
		}
		try {
			return new File($this->path . DIRECTORY_SEPARATOR . $identifier);
		} catch (\Exception $e) {
			return null;
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function addFile(File $file, ?string $filename=null, bool $overwrite=false): string
	{
		if ($filename === null) {
			$filename = $file->getFilename();
		} elseif ($filename != '') {
			// Erweiterung anhängen, wenn keine im Dateinamen angegeben
			if (strpos($filename, '.') === false) {
				$extension = ($file instanceof UploadedFile) ? $file->getClientOriginalExtension() : $file->getExtension();
				if ($extension) {
					$filename .= '.'.$extension;
				}
			}
		}
		$filename = ltrim($filename, '.');
		if ($filename === '' or strpos($filename, '/') !== false) {
			throw new InvalidArgumentException('file name invalid');
		}

		// Sicherstellen, dass das Zielverzeichnis existiert
		@mkdir($this->path);

		// Optimierung: Hochgeladene Dateien werden direkt verschoben statt kopiert
		if ($file instanceof UploadedFile) {
			$tempFile = $file;
		}
		else {
			$tempPath = @tempnam($this->path, $filename);
			if (copy((string)$file, $tempPath) === false) {
				throw new ConfigNotWritableException($tempPath);
			}
			$tempFile = new File($tempPath);
		}

		if (!$overwrite) {
			if (file_exists($this->path . DIRECTORY_SEPARATOR . $filename)) {
				$parts = explode('.', $filename);
				$basename = implode(array_slice($parts, 0, -1));
				$extension = (count($parts) > 1) ? '.'.$parts[count($parts)-1] : '';
				for ($i = 1; $i < 999; ++$i) {
					$new_filename = $basename . '_'.$i . $extension;
					if (file_exists($this->path . DIRECTORY_SEPARATOR . $new_filename) === false) {
						$filename = $new_filename;
						break;
					}
				}
				if (file_exists($this->path . DIRECTORY_SEPARATOR . $filename)) {
					throw new RuntimeException('failed to generate unique filename');
				}
			}
		}

		$tempFile->move($this->path, $filename);
		return $filename;
	}

	/**
	 * {@inheritdoc}
	 */
	public function replaceFile(?string $oldIdentifier, ?File $file, ?string $filename=null): string
	{
		if ($file !== null) {
			$identifier = $this->addFile($file, $filename, true);
		} else {
			$identifier = '';
		}
		if ($oldIdentifier != '' and $oldIdentifier !== $identifier) {
			$this->removeFile($oldIdentifier);
		}
		return $identifier;
	}

	/**
	 * {@inheritdoc}
	 */
	public function removeFile(string $identifier)
	{
		if ($identifier === '' or $identifier[0] == '.' or strpos($identifier, '/') !== false) {
			throw new InvalidArgumentException('file identifier invalid');
		}

		$filePath = $this->path . DIRECTORY_SEPARATOR . $identifier;

		@unlink($filePath);

		if (file_exists($filePath)) {
			throw new ConfigNotWritableException($filePath);
		}

		return $this;
	}

}
