<?php

namespace App\Component\Media\Store;

use App\Component\Media\CachedFile;
use App\Component\Media\RepositoryInterface;
use App\Component\MongoDbService;
use App\Exception\ConfigNotReadableException;
use App\Exception\ConfigNotWritableException;
use App\Model\User;
use App\Model\Domain;
use App\Utils;
use InvalidArgumentException;
use MongoDB\Driver\Exception\InvalidArgumentException as MongoDbInvalidArgumentException;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use MongoDB\BSON\ObjectId;
use MongoDB\BSON\UTCDateTime;
use MongoDB\BSON\Binary;
use MongoDB\Collection;
use MongoDB\Model\BSONDocument;

class MongoDb implements RepositoryInterface
{
	/** @var Collection $collection */
	private $collection;

	/** @var string $kind */
	private $kind;

	/** @var string|null $userId */
	private $userId;

	/** @var string|null $domainId */
	private $domainId;

	/** @var string $cacheDir */
	private $cacheDir;

	/**
	 * Lädt Mediendateien aus der MongoDb
	 * @param string $collection Name der Collection
	 * @param string $kind Unterart der Datei ('icon', usw.)
	 * @param User|null $user User, wenn benutzerspezifisch
	 * @param Domain|null $domain Domain, wenn domainspezifisch
	 */
	public function __construct(string $collection, string $kind, ?User $user=null, ?Domain $domain=null)
	{
		if ($collection === '' or $collection[0] == '.' or strpos($collection, '/') !== false) {
			throw new InvalidArgumentException('collection invalid');
		}

		$db = MongoDbService::getInstance()->getDatabase();
		try {
			$this->collection = $db->selectCollection($collection);
		} catch (\Exception $exception) {
			throw new ConfigNotReadableException($exception->getMessage());
		}

		if ($kind === '' or $kind[0] == '.' or strpos($kind, '/') !== false) {
			throw new InvalidArgumentException('kind invalid');
		}
		$this->kind = $kind;

		$this->domainId = $domainId = ($domain) ? $domain->getId() : null;
		$this->userId = $userId = ($user) ? $user->getId() : null;

		$this->cacheDir = Utils::getCacheDir() . DIRECTORY_SEPARATOR . 'media' . DIRECTORY_SEPARATOR . $collection.'-'.$kind;
		if ($userId) {
			$this->cacheDir .= DIRECTORY_SEPARATOR . $userId;

			if ($domainId) {
				$this->cacheDir .= DIRECTORY_SEPARATOR . $domainId;
			} else {
				$this->cacheDir .= DIRECTORY_SEPARATOR . 'global';
			}

		} else {
			$this->cacheDir .= DIRECTORY_SEPARATOR . 'global';
		}
	}

	/**
	 * Diese Methode sorgt bei einem relevanten Repository dafür, dass nur
	 * Dokumente berücksichtigt werden, die mit einer bestimmten Domain oder
	 * einem bestimmten Benutzer verknüpft sind.
	 *
	 * Jeder Filter sollte hier durchlaufen, bevor er an MongoDB geht.
	 * @param array $filter
	 * @return array
	 */
	private function prepareFilterContext(array $filter): array
	{
		$filter = ['kind' => $this->kind] + $filter;
		if ($this->domainId) {
			$filter = ['_domainId'=>new ObjectId($this->domainId)] + $filter;
		}

		elseif ($this->userId) {
			$filter = ['_userId'=>new ObjectId($this->userId)] + $filter;
		}

		else {
			$filter = [
				'$and' => [
					['_domainId' => ['$exists' => false]],
					['_userId' => ['$exists' => false]],
				],
			] + $filter;
		}

		return $filter;
	}

	private function getCachedFile($identifier): \SplFileInfo
	{
		if (!file_exists($this->cacheDir)) {
			@mkdir($this->cacheDir, 0777, true);
		}
		return new \SplFileInfo($this->cacheDir . DIRECTORY_SEPARATOR . $identifier);
	}

	/**
	 * Leeren des Repositorys.
	 */
	public function clear(): void
	{
		$this->collection->deleteMany($this->prepareFilterContext([]));
		if (file_exists($this->cacheDir)) {
			$tempPath = tempnam(dirname($this->cacheDir), basename($this->cacheDir));
			rename($this->cacheDir, $tempPath);
			Utils::removeDirectory($tempPath);
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function getFile(string $identifier): ?File
	{
		if ($identifier === '' or $identifier[0] == '.' or strpos($identifier, '/') !== false) {
			throw new InvalidArgumentException('file identifier invalid');
		}

		// Aktuelle Cache-Datei verwenden, falls vorhanden
		$cacheFile = $this->getCachedFile($identifier);
		if ($cacheFile->isReadable() and time() - $cacheFile->getCTime() < 15) {
			return new CachedFile((string)$cacheFile, file_get_contents(((string)$cacheFile).'.name'));
		}

		$document = $this->fetchDocument($identifier);
		$tempPath = tempnam($this->cacheDir, (string)$cacheFile);
		if ($document) {
			file_put_contents($tempPath, $document['content']->getData());
			file_put_contents($tempPath.'.name', $document['name']);
			@touch($tempPath, $document['mtime']->toDateTime()->getTimestamp());
			rename($tempPath.'.name', ((string)$cacheFile).'.name');
			rename($tempPath, (string)$cacheFile);
			return new CachedFile((string)$cacheFile, $document['name']);
		}
		else {
			return null;
		}

	}

	/**
	 * Diese Methode lädt ein bestimmtes Dokument
	 * @throws ConfigNotReadableException
	 */
	private function fetchDocument(string $identifier): ?BSONDocument
	{
		try {
			$filter = ['_id' => new ObjectId($identifier)];
		} catch (MongoDbInvalidArgumentException $exception) {
			return null;
		}

		try {
			return $this->collection->findOne($this->prepareFilterContext($filter));
		} catch (\Exception $exception) {
			throw new ConfigNotReadableException($exception->getMessage());
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function addFile(File $file, ?string $filename=null, bool $overwrite=false): string
	{
		if ($filename === null) {
			$filename = $file->getFilename();
		} elseif ($filename != '') {
			// Erweiterung anhängen, wenn keine im Dateinamen angegeben
			if (strpos($filename, '.') === false) {
				$extension = ($file instanceof UploadedFile) ? $file->getClientOriginalExtension() : $file->getExtension();
				if ($extension) {
					$filename .= '.'.$extension;
				}
			}
		}
		$filename = ltrim($filename, '.');
		if ($filename === '' or strpos($filename, '/') !== false) {
			throw new InvalidArgumentException('file name invalid');
		}

		$document = [
			'kind' => $this->kind,
			'name' => $filename,
			'mtime' => new UTCDateTime((int)($file->getMTime()*1000)),
			'type' => $file->getMimeType(),
			'content' => new Binary(file_get_contents((string)$file), Binary::TYPE_GENERIC),
		];

		if ($this->domainId) {
			$document = ['_domainId'=>new ObjectId($this->domainId)] + $document;
		}
		elseif ($this->userId) {
			$document = ['_userId'=>new ObjectId($this->userId)] + $document;
		}

		if ($overwrite) {
			$filter = $this->prepareFilterContext([ 'name' => $filename ]);
			$result = $this->collection->findOneAndReplace($filter, $document, [
				'upsert' => true,
				'returnDocument' => \MongoDB\Operation\FindOneAndReplace::RETURN_DOCUMENT_AFTER,
				'projection' => [ '_id'=>1 ]
			]);
			if (!$result) {
				throw new ConfigNotWritableException('could not write file to database');
			}
			$identifier = (string)($result['_id']);
		}
		else {
			$result = $this->collection->insertOne($document);
			if (!$result) {
				throw new ConfigNotWritableException('could not write file to database');
			}
			$identifier = (string)($result->getInsertedId());
		}
		@unlink($this->getCachedFile($identifier));
		return $identifier;
	}

	/**
	 * {@inheritdoc}
	 */
	public function replaceFile(?string $oldIdentifier, ?File $file, ?string $filename=null): string
	{
		if ($file !== null) {
			$identifier = $this->addFile($file, $filename, true);
		} else {
			$identifier = '';
		}
		if ($oldIdentifier != '' and $oldIdentifier !== $identifier) {
			$this->removeFile($oldIdentifier);
		}
		return $identifier;
	}

	/**
	 * {@inheritdoc}
	 */
	public function removeFile(string $identifier)
	{
		try {
			$filter = ['_id' => new ObjectId($identifier)];
		} catch (MongoDbInvalidArgumentException $exception) {
			return $this;
		}

		try {
			$this->collection->deleteOne($this->prepareFilterContext($filter));
		} catch (\Exception $exception) {
			throw new ConfigNotReadableException($exception->getMessage());
		}

		@unlink($this->getCachedFile($identifier));

		return $this;
	}

}
