<?php

namespace App\Component\Menu;

use Doctrine\Common\Annotations\Reader;
use Symfony\Component\Config\ConfigCacheFactoryInterface;
use Symfony\Component\Config\ConfigCacheInterface;
use Symfony\Component\Config\Resource\FileResource;
use Symfony\Contracts\Service\ServiceProviderInterface;
use App\Utils;
use App\Model\PluginState;
use App\Model\User;
use App\Component\Menu\Model\MenuItem;
use App\Component\Menu\Model\MenuGroup;
use App\Event\MenuGenerationEvent;

class Manager
{
	const DEFAULT_APP = 'CCM19';

	private static $ROUTE_ANNOTATION = 'Symfony\\Component\\Routing\\Annotation\\Route';
	private static $MENU_ANNOTATION = 'App\\Component\\Menu\\Annotation\\Menu';

	private $controllerLocator;
	private $projectDir;
	private $cacheDir;
	private $annotationReader;
	private $configCacheFactory;
	private $dumper;
	private $menuTree;
	private $accessedResources = [];
	private $menuCache = [];
	private $dirty = false;

	public function __construct(string $projectDir, ?string $cacheDir, ServiceProviderInterface $controllerLocator, Reader $annotationReader, ConfigCacheFactoryInterface $configCacheFactory, Dumper $dumper)
	{
		$this->controllerLocator = $controllerLocator;
		$this->annotationReader = $annotationReader;
		$this->projectDir = $projectDir;
		$this->cacheDir = $cacheDir;
		$this->configCacheFactory = $configCacheFactory;
		$this->dumper = $dumper;
	}

	/**
	 * Gibt die Menügruppen, die die Applikationen repräsentieren, zurück
	 *
	 * Die Daten sind nach den Benutzerberechtigungen gefiltert und werden während des Requests gecached.
	 *
	 * @param User|null $user Benutzer bzw. null für den aktuell angemeldeten
	 * @return array<string,MenuGroup>
	 */
	public function getApps(?User $user=null): array
	{
		$result = [];
		foreach ($this->menuTree as $appName => $app) {
			// Lade Menüs in den Cache
			$menu = $this->getMenu($user, $appName);
			$metamenu = $this->getMetaMenu($user, $appName);

			if ($menu or $metamenu) {
				$result[$appName] = $app;
			}
		}
		return $result;
	}

	/**
	 * Gibt die Menügruppen für das Haupt-/Domain-Menü zurück.
	 *
	 * Die Daten sind nach den Benutzerberechtigungen gefiltert und werden während des Requests gecached.
	 *
	 * @param User|null $user Benutzer bzw. null für den aktuell angemeldeten
	 * @param string $app
	 * @return MenuGroup[]
	 */
	public function getMenu(?User $user=null, $app='CCM19'): array
	{
		if ($user === null) {
			$user = User::loggedInUser();
		}
		if ($user === null) {
			return [];
		}
		$userId = $user->getId();
		if (isset($this->menuCache[$userId][$app]['default'])) {
			return $this->menuCache[$userId][$app]['default'];
		}

		$edition = Utils::getEdition() ?: 'base';
		$accessGroups = [];
		if ($user->getRole() == User::ROLE_ADMIN) {
			$accessGroups[] = MenuItem::ACCESS_ADMIN;
			if (!Utils::isExtendedEdition()) {
				$accessGroups[] = MenuItem::ACCESS_USER;
			}
		} elseif ($user->isManager()) {
			$accessGroups[] = MenuItem::ACCESS_MANAGER;
		} elseif ($user->isSubUser()) {
			$accessGroups[] = MenuItem::ACCESS_SUBUSER;
		} else {
			$accessGroups[] = MenuItem::ACCESS_USER;
		}

		$hasMultiDomainAccess = Utils::hasMultiDomainAccess($user);
		$result = isset($this->menuTree[$app]) ? $this->menuTree[$app]->filtered(
			function (MenuItem $item) use ($user, $accessGroups, $edition, $hasMultiDomainAccess) {
				return (
					($item->navigation == MenuItem::NAVIGATION_DOMAIN or (!$hasMultiDomainAccess and $item->navigation == MenuItem::NAVIGATION_MAIN))
					and in_array($edition, $item->editions)
					and Utils::any($accessGroups, function ($group) use ($item) { return in_array($group, $item->access); })
					and Utils::mayUserAccessRoute($user, $item->route, 'GET')
					and $item->isActiveForCurrentDomain()
				);
			})
		->items : [];

		if (!isset($this->menuCache[$userId])) {
			$this->menuCache[$userId] = [];
		}
		if (!isset($this->menuCache[$userId][$app])) {
			$this->menuCache[$userId][$app] = [];
		}
		$this->menuCache[$userId][$app]['default'] = $result;
		return $result;
	}

	/**
	 * Gibt die Menügruppen für das Meta-Menü zurück.
	 *
	 * Die Daten sind nach den Benutzerberechtigungen gefiltert und werden während des Requests gecached.
	 *
	 * @param User|null $user Benutzer bzw. null für den aktuell angemeldeten
	 * @param string $app
	 * @return MenuGroup[]
	 */
	public function getMetaMenu(?User $user=null, $app='CCM19'): array
	{
		if ($user === null) {
			$user = User::loggedInUser();
		}
		if ($user === null) {
			return [];
		}
		$userId = $user->getId();
		if (isset($this->menuCache[$userId][$app]['meta'])) {
			return $this->menuCache[$userId][$app]['meta'];
		}

		$edition = Utils::getEdition() ?: 'base';
		$accessGroups = [];
		if ($user->getRole() == User::ROLE_ADMIN) {
			$accessGroups[] = MenuItem::ACCESS_ADMIN;
			if (!Utils::isExtendedEdition()) {
				$accessGroups[] = MenuItem::ACCESS_USER;
			}
		} elseif ($user->isManager()) {
			$accessGroups[] = MenuItem::ACCESS_MANAGER;
		} elseif ($user->isSubUser()) {
			$accessGroups[] = MenuItem::ACCESS_SUBUSER;
		} else {
			$accessGroups[] = MenuItem::ACCESS_USER;
		}

		$hasMultiDomainAccess = Utils::hasMultiDomainAccess($user);

		$result = isset($this->menuTree[$app]) ? $this->menuTree[$app]->filtered(
			function (MenuItem $item) use ($user, $accessGroups, $edition, $hasMultiDomainAccess) {
				return (
					($item->navigation == MenuItem::NAVIGATION_META or ($hasMultiDomainAccess and $item->navigation == MenuItem::NAVIGATION_MAIN))
					and in_array($edition, $item->editions)
					and Utils::any($accessGroups, function ($group) use ($item) { return in_array($group, $item->access); })
					and Utils::mayUserAccessRoute($user, $item->route, 'GET')
					and $item->isActiveForCurrentUser()
				);
			})
		->items : [];

		if (!isset($this->menuCache[$userId])) {
			$this->menuCache[$userId] = [];
		}
		if (!isset($this->menuCache[$userId][$app])) {
			$this->menuCache[$userId][$app] = [];
		}
		$this->menuCache[$userId][$app]['meta'] = $result;
		return $result;
	}

	public function isRouteInMetaMenu($route, ?User $user=null): bool
	{
		$domainMenu = self::getMenu($user);
		$metaMenu = self::getMetaMenu($user);

		foreach ($domainMenu as $group) {
			foreach ($group->iterateItems() as $item) {
				if ($item->route == $route) {
					return false;
				}
			}
		}
		foreach ($metaMenu as $group) {
			foreach ($group->iterateItems() as $item) {
				if ($item->route == $route) {
					return true;
				}
			}
		}
		foreach ($domainMenu as $group) {
			foreach ($group->iterateItems() as $item) {
				if (fnmatch($item->routeGroup.'*', $route)) {
					return false;
				}
			}
		}
		foreach ($metaMenu as $group) {
			foreach ($group->iterateItems() as $item) {
				if (fnmatch($item->routeGroup.'*', $route)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * @required
	 */
	public function loadMenu()
	{
		// Wenn CacheDir gesetzt: Menüdaten cachen
		if ($this->cacheDir) {
			$cache = $this->configCacheFactory->cache($this->cacheDir.'/MenuData.php', function (ConfigCacheInterface $cache) {
				$this->makeBaseMenu();
				$cache->write($this->dumper->dump($this->menuTree), $this->accessedResources);
			});
			$this->menuTree = require $cache->getPath();
		}
		else {
			// Sonst direkt alle PHP-Dateien jedes mal auslesen
			$this->makeBaseMenu();
		}
		// Event für Plugins etc. auslösen, damit die noch Menüpunkte hinzufügen können
		Utils::dispatchEvent(MenuGenerationEvent::NAME, new MenuGenerationEvent($this));
		$this->finishMenuEdit();
	}

	/**
	 *
	 */
	private function makeBaseMenu()
	{
		$forceHidden = explode('|', $_ENV['APP_MENU_HIDE'] ?? '');
		$this->menuTree = [];
		$this->addApplication(self::DEFAULT_APP, -1, 'fa-dashboard');
		// Standard-Gruppe für Plugins hinzufügen
		$this->addMenuGroup('Plugins', 150);

		// Apps hinzufügen
		foreach (PluginState::all() as $plugin) {
			if ($plugin->getType() == 'ccm19-app') {
				$this->addApplication($plugin->getName(), null, $plugin->getIcon());
			}
		}

		// Menüpunkte aus Controller-Klassen sammeln
		foreach ($this->readControllerAnnotations() as $a) {
			if ($forceHidden and $a->getName() !== null and in_array($a->getName(), $forceHidden)) {
				continue;
			}
			if ($a->getSubmenu() === null and $a->getRoute() === null) {
				$this->addSubMenu($a->getGroup(), $a->getName(), $a->getOrder(), $a->getIcon(), $a->getApp() ?? self::DEFAULT_APP);
			} else {
				$this->addMenuItem($a->getGroup(), MenuItem::fromAnnotation($a), $a->getApp() ?? self::DEFAULT_APP, $a->getSubmenu());
			}
		}

		// Abschließendes Sortieren
		$this->finishMenuEdit();
	}

	/**
	 * @param string $group
	 * @param MenuItem $item
	 * @param string $app
	 * @param ?string $submenu
	 */
	public function addMenuItem(string $group, MenuItem $item, $app="CCM19", $submenu=null): void
	{
		$this->dirty = true;
		$this->menuCache = [];
		if (!isset($this->menuTree[$app]->items[$group])) {
			$this->addMenuGroup($group, null, $app);
		}
		$groupObj = $this->menuTree[$app]->items[$group];
		$items = &$groupObj->items;

		if ($submenu) {
			$submenuObj = array_reduce($items, function ($carry, $item) use ($submenu) {
				return $carry ? $carry: (($item->name == $submenu and $item instanceof MenuGroup) ? $item : $carry);
			}, null);
			if ($submenuObj) {
				unset($items);
				$items = &$submenuObj->items;
				$groupObj = $submenuObj;
			}
		}

		if ($item->order === null) {
			if ($items) {
				$item->order = array_reduce($items, function ($carry, $item) { return ($carry >= $item->order) ? $carry : $item->order; }, 0) + 10;
			} elseif ($groupObj->order !== null) {
				$item->order = $groupObj->order;
			} else {
				$item->order = 10;
			}
		}

		$items[] = $item;
	}

	/**
	 * @param string $group
	 * @param string $name
	 * @param int|null $order
	 * @param string|null $icon
	 * @param string $app
	 */
	public function addSubMenu(string $group, string $name, ?int $order=null, ?string $icon=null, $app="CCM19"): void
	{
		$this->dirty = true;
		$this->menuCache = [];
		if (!isset($this->menuTree[$app]->items[$group])) {
			$this->addMenuGroup($group, null, $app);
		}
		$items = &$this->menuTree[$app]->items[$group]->items;
		$submenu = array_reduce($items, function ($carry, $item) use ($name) {
			return $carry ? $carry: (($item->name == $name and $item instanceof MenuGroup) ? $item : $carry);
		}, null);
		if ($submenu) {
			if ($submenu->icon === null) {
				$submenu->icon = $icon;
			}
		} else {
			$items[] = $submenu = new MenuGroup($name, $order, $icon);
			if ($submenu->order === null) {
				if ($items) {
					$submenu->order = array_reduce($items, function ($carry, $item) { return ($carry >= $item->order) ? $carry : $item->order; }, 0) + 10;
				} elseif ($this->menuTree[$app]->items[$group]->order !== null) {
					$submenu->order = $this->menuTree[$app]->items[$group]->order;
				} else {
					$submenu->order = 10;
				}
			}
		}
	}

	/**
	 * @param string $app
	 * @param int|null $order
	 * @param string|null $icon
	 */
	public function addApplication(string $app, ?int $order=null, ?string $icon=null): void
	{
		if (!isset($this->menuTree[$app])) {
			if ($order === null) {
				$order = count($this->menuTree)*10;
			}
			$this->menuTree[$app] = new MenuGroup($app, $order, $icon);
		}
	}

	/**
	 * @param string $group
	 * @param int|null $order
	 * @param string $app
	 */
	public function addMenuGroup(string $group, ?int $order=null, $app="CCM19"): void
	{
		if (!isset($this->menuTree[$app])) {
			$this->menuTree[$app] = new MenuGroup($app);
		}

		if (isset($this->menuTree[$app]->items[$group])) {
			return;
		}
		$this->menuTree[$app]->items[$group] = new MenuGroup($group, $order);
	}

	private function finishRecursive(MenuGroup $group)
	{
		foreach ($group->items as $item) {
			if ($item instanceof MenuGroup) {
				$this->finishRecursive($item);
			}
		}
		uasort($group->items, function ($a, $b) { return $a->order <=> $b->order ?: $a->name <=> $b->name; });
		if ($group->order === null and $group->items) {
			$group->order = reset($group->items)->order;
		}
	}

	/**
	 *
	 */
	public function finishMenuEdit(): void
	{
		if (!$this->dirty) {
			return;
		}

		foreach ($this->menuTree as $appname => $app) {
			$this->finishRecursive($app);
		}

		uasort($this->menuTree, function ($a, $b) {
			return $a->order <=> $b->order ?: $a->name <=> $b->name;
		});
	}

	/**
	 * @return array
	 * @throws \ReflectionException
	 */
	private function readControllerAnnotations(): array
	{
		$this->accessedResources = [
			new FileResource(__FILE__),
			new FileResource($this->projectDir . '/.env'),
		];

		if (file_exists($this->projectDir . '/.env.local')) {
			$this->accessedResources[] = new FileResource($this->projectDir . '/.env.local');
		}

		$annotations = [];

		/** @var class-string[] $classes */
		$classes = $this->controllerLocator->getProvidedServices();

		// Alle Controller durchgehen
		foreach ($classes as $className) {
			$class = new \ReflectionClass($className);
			$path = $class->getFileName();

			// Plugin bestimmen
			$isPlugin = strpos($className, 'Plugins\\') === 0;
			$isApp = false;
			$appName = self::DEFAULT_APP;
			if ($isPlugin) {
				$pluginState = PluginState::findByClassName($className);
				if ($pluginState) {
					$pluginStateId = $pluginState->getId();
					$isApp = ($pluginState->getType() === 'ccm19-app');
					if ($isApp) {
						$appName = $pluginState->getName();
					}
				} else {
					// Defekte Plugins überspringen
					continue;
				}
			} else {
				$pluginStateId = null;
			}

			if ($path) {
				$this->accessedResources[] = new FileResource($path);
			}

			// Routing-Annotation der Klasse berücksichtigen
			$annotation = $this->annotationReader->getClassAnnotation($class, self::$ROUTE_ANNOTATION);
			$routeBase = ($annotation) ? $annotation->getName() : '';

			// Menü-Annotation der Klasse berücksichtigen
			$hasSubmenu = false;
			$classMenuAnnotation = $this->annotationReader->getClassAnnotation($class, self::$MENU_ANNOTATION);
			if ($classMenuAnnotation) {
				$classMenuAnnotation->setRoute(null);
				// App vervollständigen
				if ($classMenuAnnotation->getApp() === null) {
					$classMenuAnnotation->setApp($appName);
				}
				// Bei Plugins: Gruppe vervollständigen
				if ($pluginStateId !== null and $classMenuAnnotation->getGroup() === null) {
					$classMenuAnnotation->setGroup(($isApp) ? $appName : 'Plugins');
				}
				// Routen-Gruppe vervollständigen
				if ($classMenuAnnotation->getRouteGroup() === null) {
					$classMenuAnnotation->setRouteGroup($routeBase != '' ? $routeBase : null);
				}
				// Quell-Plugin setzen
				$classMenuAnnotation->setPluginStateId($pluginStateId);

				if ($classMenuAnnotation->getName() !== null) {
					$hasSubmenu = true;
					// Sub-Menü anlegen
					$annotations[] = $classMenuAnnotation;
				}
			}

			foreach ($class->getMethods() as $method) {
				$methodAnnotations = $this->annotationReader->getMethodAnnotations($method);
				$route = null;

				// Routing-Annotation der Methode zum Vervollständigen verwenden
				foreach ($methodAnnotations as $annotation) {
					if ($annotation instanceof self::$ROUTE_ANNOTATION and $annotation->getName() !== null) {
						$route = $routeBase . $annotation->getName();
						break;
					}
				}

				$classDomainDependant = $class->isSubclassOf('\\App\\Controller\\DomainDependantController');
				$classAdmin = $class->isSubclassOf('\\App\\Controller\\AdminController');
				$classHosting = $class->isSubclassOf('\\App\\Controller\\HostingController');
				$classUniversal = $class->isSubclassOf('\\App\\Controller\\UniversalController');
				$classUnrestricted = $class->isSubclassOf('\\App\\Controller\\UnrestrictedController');

				// Menü-Annotation suchen
				foreach ($methodAnnotations as $annotation) {
					if ($annotation instanceof self::$MENU_ANNOTATION) {
						// Einschränkung nach ENV berücksichtigen
						$env = $_SERVER['APP_ENV'] ?? 'prod';
						$allowedEnvs = $annotation->getEnvs();
						if ($allowedEnvs !== null and !in_array(trim(strtolower($env)), $allowedEnvs)) {
							continue;
						}

						if ($hasSubmenu and $classMenuAnnotation) {
							// Submenü-Einträge sind in der gleichen App und Gruppe wie ihr Elternelement
							$annotation->setApp($classMenuAnnotation->getApp());
							$annotation->setGroup($classMenuAnnotation->getGroup());
							$annotation->setSubmenu($classMenuAnnotation->getName());
						} else {
							// App vervollständigen
							if ($annotation->getApp() === null) {
								if ($classMenuAnnotation and $classMenuAnnotation->getApp()) {
									$annotation->setApp($classMenuAnnotation->getApp());
								} else {
									$annotation->setApp($appName);
								}
							}
							// Bei Plugins: Gruppe vervollständigen
							if ($pluginStateId !== null and $annotation->getGroup() === null) {
								if ($classMenuAnnotation and $classMenuAnnotation->getGroup()) {
									$annotation->setGroup($classMenuAnnotation->getGroup());
								} else {
									$annotation->setGroup(($isApp) ? $appName : 'Plugins');
								}
							}
						}
						// Route vervollständigen
						if ($annotation->getRoute() === null) {
							$annotation->setRoute($route);
						}
						// Routen-Gruppe vervollständigen
						if ($annotation->getRouteGroup() === null) {
							if ($classMenuAnnotation and $classMenuAnnotation->getRouteGroup()) {
								$annotation->setRouteGroup($hasSubmenu ? $route : $classMenuAnnotation->getRouteGroup());
							} else {
								$annotation->setRouteGroup($routeBase != '' ? $routeBase.'*' : $route);
							}
						}
						// Navigation vervollständigen
						if ($annotation->getNavigation() === null) {
							if ($classMenuAnnotation and $classMenuAnnotation->getNavigation()) {
								$annotation->setNavigation($classMenuAnnotation->getNavigation());
							} else {
								$annotation->setNavigation($classDomainDependant ? MenuItem::NAVIGATION_DOMAIN : MenuItem::NAVIGATION_MAIN);
							}
						}
						// Access vervollständigen
						if (!$annotation->getAccess()) {
							if ($classMenuAnnotation and $classMenuAnnotation->getAccess()) {
								$annotation->setAccess($classMenuAnnotation->getAccess());
							} else {
								$domainDependant = $annotation->getNavigation() == MenuItem::NAVIGATION_DOMAIN;
								if ($domainDependant) {
									$annotation->setAccess([MenuItem::ACCESS_USER, MenuItem::ACCESS_SUBUSER]);
								}
								elseif ($classAdmin or $classHosting) {
									$annotation->setAccess([MenuItem::ACCESS_ADMIN, MenuItem::ACCESS_MANAGER]);
								}
								elseif ($classUniversal or $classUnrestricted) {
									$annotation->setAccess([MenuItem::ACCESS_ADMIN, MenuItem::ACCESS_MANAGER, MenuItem::ACCESS_USER, MenuItem::ACCESS_SUBUSER]);
								}
								else {
									$annotation->setAccess([MenuItem::ACCESS_USER, MenuItem::ACCESS_SUBUSER]);
								}
							}
						}
						// Editions vervollständigen
						if ($annotation->getEditions() === null) {
							if ($classMenuAnnotation and $classMenuAnnotation->getEditions() !== null) {
								$annotation->setEditions($classMenuAnnotation->getEditions());
							} else {
								// Wenn nichts anderes angegeben, gibt es HostingController nur in der Extended-Edition.
								if ($classHosting) {
									$annotation->setEditions(['extended']);
								}
							}
						}
						// Envs vervollständigen
						if ($annotation->getEnvs() === null) {
							if ($classMenuAnnotation and $classMenuAnnotation->getEnvs() !== null) {
								$annotation->setEnvs($classMenuAnnotation->getEnvs());
							}
						}
						// Quell-Plugin setzen
						$annotation->setPluginStateId($pluginStateId);

						$annotations[] = $annotation;
					}
				}
			}

		}
		return $annotations;
	}

}
