<?php

namespace App\Component\Menu\Model;

use App\Component\Menu\Annotation\Menu;
use App\Model\PluginState;

final class MenuItem
{
	// Menüpunkte, die in der Metanavigation auftauchen sollen
	const NAVIGATION_META = 'meta';
	// Menüpunkte, die in der "Hauptnavigation" (Multidomain/Agency Kunde: Meta, sonst Domainspezifisch) auftauchen sollen
	const NAVIGATION_MAIN = 'main';
	// Menüpunkte, die nur in der domainspezifischen Navigation auftauchen sollen
	const NAVIGATION_DOMAIN = 'domain';

	// Zugriff für Administratoren
	const ACCESS_ADMIN = 'admin';
	// Zugriff für Manager
	const ACCESS_MANAGER = 'manager';
	// Zugriff für reguläre Benutzer
	const ACCESS_USER = 'user';
	// Zugriff für Unterbenutzer
	const ACCESS_SUBUSER = 'subuser';

	public $name;
	public $icon;
	public $access;
	public $navigation;
	public $route;
	public $routeGroup;
	public $order;
	public $editions;
	public $pluginStateId;

	/**
	 * @param string $name Name des Menüpunktes, wird übersetzt zum Anzeigetitel
	 * @param string|null $icon Icon-Name (z.B. fa-cog oder glyphicon-…)
	 * @param string[]|null $access Zugriffsrechte für diesen Menüpunkt (Array von `MenuItem::ACCESS_`-Konstanten)
	 * @param string $navigation Anzeige in welcher Navigation ('main', 'domain', 'meta')
	 * @param string $route Menüpunkt-Zielroute
	 * @param string|null $routeGroup Routen-Präfix für den der Menüpunkt als aktiv angesehen wird
	 * @param int|null $order Sortierungs-ID; je kleiner desto weiter oben
	 * @param string[]|null $editions Editionen in denen der Menüpunkt angezeigt werden soll
	 * @param string|null $_onlyIfPluginActive (intern) PluginState-ID, die im Domain-/User-Kontext aktiv sein muss, damit der Menüpunkt angezeigt wird.
	 */
	public function __construct(string $name, ?string $icon, ?array $access, string $navigation, string $route, ?string $routeGroup=null, ?int $order=null, ?array $editions=null, ?string $_onlyIfPluginActive=null)
	{
		$this->name = $name;
		$this->icon = $icon;
		$this->access = $access ?? [];
		$this->navigation = $navigation;
		$this->route = $route;
		$this->routeGroup = ($routeGroup === null) ? $route : $routeGroup;
		$this->order = $order;
		$this->editions = $editions ?? ['base', 'multi', 'extended'];
		$this->pluginStateId = $_onlyIfPluginActive;
	}

	public static function fromAnnotation(Menu $a): self
	{
		return new self(
			(string)$a->getName(),
			$a->getIcon(),
			$a->getAccess(),
			$a->getNavigation() ?? 'main',
			(string)$a->getRoute(),
			$a->getRouteGroup(),
			$a->getOrder(),
			$a->getEditions(),
			$a->getPluginStateId()
		);
	}

	/**
	 * Magic Method für Serialisierung/Caching
	 */
	public static function __set_state($arr)
	{
		return new self(
			$arr['name'],
			$arr['icon'],
			$arr['access'],
			$arr['navigation'],
			$arr['route'],
			$arr['routeGroup'],
			$arr['order'],
			$arr['editions'],
			$arr['pluginStateId']
		);
	}

	/**
	 * CSS-Klassen für das Icon
	 */
	public function getIconClasses(): string
	{
		if (!$this->icon) {
			return '';
		}
		elseif (strpos($this->icon, 'fa-') === 0) {
			return 'fa '.$this->icon;
		}
		elseif (strpos($this->icon, 'glyphicon-') === 0) {
			return 'glyphicon '.$this->icon;
		}
		else {
			return $this->icon;
		}
	}

	/**
	 * Ob ein Menüpunkte für den aktuellen User sichtbar sein soll
	 *
	 * Aktuell sind davon Menüpunkte von Plugins betroffen.
	 */
	public function isActiveForCurrentUser(): bool
	{
		if ($this->pluginStateId === null) {
			return true;
		}

		$pluginState = PluginState::find($this->pluginStateId);

		// Defekte Plugin-Menüpunkte ausblenden
		if ($pluginState === null) {
			return false;
		}

		return $pluginState->isAllowedForCurrentUser() || !$pluginState->isFullAccessForAllAccounts();
	}

	/**
	 * Ob ein Menüpunkte für die aktuelle Domain sichtbar sein soll
	 *
	 * Aktuell sind davon Domain-spezifische Menüpunkte von Plugins betroffen.
	 */
	public function isActiveForCurrentDomain(): bool
	{
		if ($this->pluginStateId === null) {
			return true;
		}

		$pluginState = PluginState::find($this->pluginStateId);

		// Defekte Plugin-Menüpunkte ausblenden
		if ($pluginState === null) {
			return false;
		}

		return $pluginState->isActiveForCurrentDomain();
	}

	/**
	 * Hilfsfunktion zur Vereinheitlichung mit `MenuGroup`
	 *
	 * Liefert sich selbst zurück.
	 *
	 * @return $this
	 */
	public function getFirstItem(): MenuItem
	{
		return $this;
	}

	public function activatedByRoute($route): bool
	{
		$route = (string)$route;
		return fnmatch(($this->routeGroup === null) ? $this->route : $this->routeGroup.'*', $route);
	}
}
