<?php

namespace App\Component\OnBoarding;

use Symfony\Component\HttpFoundation\ResponseHeaderBag;

/**
 * Eingebundene Ressource (Dokument, Script, Bild, XHR-Request, Stylesheet, usw.)
 */
class Resource
{
	/** @var array $data */
	private $data;
	/** @var array[]|null $urlMappings */
	private $urlMappings;

	public function __construct(array $data, ?array $urlMappings)
	{
		$this->data = $data;
		$this->urlMappings = $urlMappings;
	}

	/** Geladene URL */
	public function getUrl(): string
	{
		return $this->data['url'];
	}

	/** Ursprünglich aufgerufene URL (nur wenn eine Weiterleitung involviert war) */
	public function getOriginalUrl(): ?string
	{
		return $this->data['originalUrl'] ?? null;
	}

	/** Art der Resource ('document', 'stylesheet', 'image', 'script', 'xhr', 'other', …) */
	public function getType(): string
	{
		return $this->data['type'] ?? 'other';
	}

	/** Informationen, was den Request ausgelöst hat */
	public function getInitiator(): array
	{
		return $this->data['initiator'];
	}

	/** Status-Code */
	public function getStatus(): int
	{
		return (int)$this->data['status'];
	}

	/** Status-Code der Weiterleitung (falls eine Weiterleitung involviert war) */
	public function getRedirectStatus(): ?int
	{
		return $this->data['redirectStatus'];
	}

	/** Content-Type der Ressource (falls anwendbar) */
	public function getMimeType(): ?string
	{
		return $this->data['mimeType'] ?? null;
	}

	/** Response-Header */
	public function getResponseHeaders(): ResponseHeaderBag
	{
		return new ResponseHeaderBag($this->data['responseHeaders'] ?? []);
	}

	/** Ob der Request sicher (HTTPS) war */
	public function isSecure(): ?bool
	{
		return $this->data['secure'] ?? null;
	}

	/** Übertragungsprotokoll ('http/1.1', 'h2', …) */
	public function getProtocol(): ?string
	{
		return $this->data['protocol'] ?? null;
	}

	/**
	 * Cookies, die dynamisch vom Script gesetzt wurden (soweit erkennbar)
	 *
	 * @return string[]
	 */
	public function getScriptedCookies(): array
	{
		if (!$this->urlMappings) {
			return [];
		}
		$matching = array_filter(
			$this->urlMappings,
			function ($sourceUrls) {
				return (in_array($this->data['url'], $sourceUrls));
			}
		);
		return array_keys($matching);
	}
}
