<?php

namespace App\Component\OnBoarding;

use App\Component\Curl;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\ImpressumCrawler;
use App\Component\LocaleService;
use App\Component\LoggerInterface;
use App\Model\Cookie as CookieModel;
use App\Model\EmbeddingBase;
use App\Model\Purpose;
use App\Utils;
use RuntimeException;

class ScanService
{
	const API_URL = 'https://cookiecheck.ccm19.de/crawler.php';

	/**
	 * @param string $domain URL oder Domain
	 * @param LoggerInterface $logger
	 * @param string $domainBasicAuth
	 * @return ScanData
	 */
	public function scanDomain(string $domain, LoggerInterface $logger,string $domainBasicAuth=""): ScanData
	{
		// Ggf. Protokoll ergänzen
		if (stripos($domain, "http:") === false && stripos($domain, "https:") === false) {
			$domain = "http://" . $domain;
		}

		// URL bereinigen
		$domain_data = parse_url($domain);
		$host = $domain_data['host'];
		$domain = $domain_data['scheme'] . '://' . $host . ($domain_data['path'] ?? '/');

		// Auf lokale urls checken
		if (!$host || !(Utils::isPublicHost($host))) {
			return new ScanData([]);
		}

		// curl ini
		$curl = new Curl();

		// Daten holen
		$curl->get(self::API_URL, ["apikey" => "1234", "url" => $domain, "basicAuth"=>$domainBasicAuth]);

		if ($curl->curl_error) {
			throw new RuntimeException("An error has occurred in cURL module. (Code: {$curl->getErrorCode()}) " . $curl->getErrorMessage());
		}

		if ($curl->getResponseHeaders('Content-Type') !== 'application/json') {
			$logger->logError('ScanError', "This domain cannot be scanned at the moment because it produces an unknown error - Domain: " . $domain);
			throw new RuntimeException('This domain cannot be scanned at the moment because it produces an unknown error.');
		} elseif ($curl->isServerError()) {
		}

		$response = json_decode($curl->getResponse(), true);
		return new ScanData($response ?? array());
	}


	/**
	 * Holt Embedding-Informationen aus der DB für alle Cookies, die gefunden wurden.
	 *
	 * @param ScanData $scanData
	 * @param bool $includeCcm19Consent Ob das ccm_consent-Cookie immer mit eingeschlossen werden soll
	 * @return array[]
	 * @see EmbeddingDb::fetchEmbeddingsByCookieNames()
	 */
	public function fetchMatchingEmbeddings(ScanData $scanData, bool $includeCcm19Consent=true): array
	{
		$names = array_unique(array_merge(
			array_map(function ($item) { return $item->getName(); }, $scanData->getCookies()),
			array_map(function ($item) { return $item->getName(); }, $scanData->getLocalStorage()),
			array_map(function ($item) { return $item->getName(); }, $scanData->getSessionStorage()),
			($includeCcm19Consent) ? ['ccm_consent'] : []
		));

		if ($names) {
			return array_map(function ($embedding) {
				$embedding['source'] = EmbeddingBase::SOURCE_GLOBAL_DB;
				$embedding['source_id'] = $embedding['id'];
				$embedding['source_ts'] = $embedding['lastModified'];
				$embedding['autoUpdated'] = true;
				return $embedding;
			}, EmbeddingDb::fetchEmbeddingsByCookieNames($names));
		} else {
			return [];
		}
	}

	/**
	 * Holt Embedding-Informationen aus der DB für alle Embeddings die keine Cookies haben aber vorhanden sein könnten z.B. aus CM Banner Import
	 *
	 * @param string[] $names
	 * @param bool $includeCcm19Consent Ob das ccm_consent-Cookie immer mit eingeschlossen werden soll
	 * @return array[]
	 * @see EmbeddingDb::fetchEmbeddingsByCookieNames()
	 */
	public function fetchMatchingEmbeddingsByNames(array $names, bool $includeCcm19Consent=true): array
	{

		if ($names) {
			return array_map(function ($embedding) {
				$embedding['source'] = EmbeddingBase::SOURCE_GLOBAL_DB;
				$embedding['source_id'] = $embedding['id'];
				$embedding['source_ts'] = $embedding['lastModified'];
				$embedding['autoUpdated'] = true;
				return $embedding;
			}, EmbeddingDb::fetchEmbeddingsByNames($names));
		}

		return [];
	}

	/**
	 * Ergänzt die Embedding-Daten um Embedding(s) für Cookies zu denen kein Embedding gefunden wurde
	 *
	 * @param $embeddings &$embeddings
	 * @param ScanData $scanData
	 * @param bool $useLegacyCookies Ob das Cookie-Model genutzt werden soll, um alte Daten zu migrieren
	 * @return void
	 */
	public function supplementUnmatchedCookies(array &$embeddings, ScanData $scanData, bool $useLegacyCookies): void
	{
		$assets = [];
		$assetsNotwendig = array();
		$isTechOk = false;

		//print_r($embeddings);
		foreach ($scanData->getCookies() as $i => $cookie) {
			if ($this->cookieInEmbeddings($cookie->getName(), $embeddings)) {
				continue;
			}

			$asset = [
				'id' => -($i+1),
				'name' => $cookie->getName(),
				'storageType' => 'cookie',
				'dynamic' => false,
				'translations' => [
					'de' => [ 'lifetime' => $cookie->getLifetime(), 'description'=>'' ],
					'en' => [ 'lifetime' => $cookie->getLifetime(), 'description'=>'' ],
				],
			];

			$matchedLegacy = false;
			if ($useLegacyCookies) {
				foreach (CookieModel::all() as $legacyCookie) {
					if ($legacyCookie->matchName($cookie->getName())) {
						$embedding = [
							'id' => -($i+1),
							'name' => $legacyCookie->getName(),
							'purpose' => $legacyCookie->getPurpose()->getId(),
							'code_snippet' => $legacyCookie->getCodeSnippet(),
							'code_block' => implode("\n", $legacyCookie->getCodeBlock()),
							'embedding_vendor' => $legacyCookie->getCookieVendor(),
							'translations' => [],
							'assets' => [$asset],
						];

						$defaultLocale = LocaleService::getDefault();
						foreach (LocaleService::all() as $locale) {
							$language = $locale->getLanguage();
							if (!$language) {
								continue;
							}
							if ($legacyCookie->getTranslatedPrivacyPolicyUrl($locale, false) == '' and $legacyCookie->getTranslatedDescription($locale, false) == '') {
								continue;
							}
							$langShort = $language->getShortCode();
							$useFallback = ($locale->getName() == $defaultLocale->getName());
							$embedding['translations'][$langShort] = [
								'description' => $legacyCookie->getTranslatedDescription($locale, $useFallback),
								'privacy_policy_url' => $legacyCookie->getTranslatedPrivacyPolicyUrl($locale, $useFallback),
								'collected_data_info' => $legacyCookie->getTranslatedCollectedDataInfo($locale, $useFallback),
								'purpose_of_data_collection' => $legacyCookie->getTranslatedPurposeOfDataCollection($locale, $useFallback),
								'legal_basis' => $legacyCookie->getTranslatedLegalBasis($locale, $useFallback),
								'place_of_processing' => $legacyCookie->getTranslatedPlaceOfProcessing($locale, $useFallback),
							];
							$embedding['assets'][0]['translations'][$langShort] = [
								'lifetime' => $legacyCookie->getTranslatedLifetime($locale, true),
								'description' => '',
							];
						}

						$embeddings[] = $embedding;
						$matchedLegacy = true;
						break;
					}
				}
			}

			if (!$matchedLegacy) {
				if($cookie->isNecessary() || $cookie->getPurpose()==="c51")
				{
					$assetsNotwendig[] = $asset;
				}
				else{
					$assets[] = $asset;
				}
			}
		}

		//Noch Local Storage durchgehen
		foreach ($scanData->getLocalStorage() as $i => $cookie) {
			if ($this->cookieInEmbeddings($cookie->getName(), $embeddings)) {
				continue;
			}
			$asset = [
				'id' => -($i+1),
				'name' => $cookie->getName(),
				'storageType' => 'localStorage',
				'dynamic' => false,
				'translations' => [
					'de' => [ 'lifetime' => $cookie->getLifetime(), 'description'=>'' ],
					'en' => [ 'lifetime' => $cookie->getLifetime(), 'description'=>'' ],
				],
			];

			$matchedLegacy = false;
			if ($useLegacyCookies) {
				foreach (CookieModel::all() as $legacyCookie) {
					if ($legacyCookie->matchName($cookie->getName())) {
						$embedding = [
							'id' => -($i+1),
							'name' => $legacyCookie->getName(),
							'purpose' => $legacyCookie->getPurpose()->getId(),
							'code_snippet' => $legacyCookie->getCodeSnippet(),
							'code_block' => implode("\n", $legacyCookie->getCodeBlock()),
							'embedding_vendor' => $legacyCookie->getCookieVendor(),
							'translations' => [],
							'assets' => [$asset],
						];

						$defaultLocale = LocaleService::getDefault();
						foreach (LocaleService::all() as $locale) {
							$language = $locale->getLanguage();
							if (!$language) {
								continue;
							}
							if ($legacyCookie->getTranslatedPrivacyPolicyUrl($locale, false) == '' and $legacyCookie->getTranslatedDescription($locale, false) == '') {
								continue;
							}
							$langShort = $language->getShortCode();
							$useFallback = ($locale->getName() == $defaultLocale->getName());
							$embedding['translations'][$langShort] = [
								'description' => $legacyCookie->getTranslatedDescription($locale, $useFallback),
								'privacy_policy_url' => $legacyCookie->getTranslatedPrivacyPolicyUrl($locale, $useFallback),
								'collected_data_info' => $legacyCookie->getTranslatedCollectedDataInfo($locale, $useFallback),
								'purpose_of_data_collection' => $legacyCookie->getTranslatedPurposeOfDataCollection($locale, $useFallback),
								'legal_basis' => $legacyCookie->getTranslatedLegalBasis($locale, $useFallback),
								'place_of_processing' => $legacyCookie->getTranslatedPlaceOfProcessing($locale, $useFallback),
							];
							$embedding['assets'][0]['translations'][$langShort] = [
								'lifetime' => $legacyCookie->getTranslatedLifetime($locale, true),
								'description' => '',
							];
						}

						$embeddings[] = $embedding;
						$matchedLegacy = true;
						break;
					}
				}
			}

			if (!$matchedLegacy) {
				if($cookie->isNecessary() || $cookie->getPurpose()==="c51")
				{
					$assetsNotwendig[] = $asset;
				}
				else{
					$assets[] = $asset;
				}
			}
		}

		//Session Storage auch durchgehen
		foreach ($scanData->getSessionStorage() as $i => $cookie) {
			if ($this->cookieInEmbeddings($cookie->getName(), $embeddings)) {
				continue;
			}
			$asset = [
				'id' => -($i+1),
				'name' => $cookie->getName(),
				'storageType' => 'sessionStorage',
				'dynamic' => false,
				'translations' => [
					'de' => [ 'lifetime' => $cookie->getLifetime(), 'description'=>'' ],
					'en' => [ 'lifetime' => $cookie->getLifetime(), 'description'=>'' ],
				],
			];

			$matchedLegacy = false;
			if ($useLegacyCookies) {
				foreach (CookieModel::all() as $legacyCookie) {
					if ($legacyCookie->matchName($cookie->getName())) {
						$embedding = [
							'id' => -($i+1),
							'name' => $legacyCookie->getName(),
							'purpose' => $legacyCookie->getPurpose()->getId(),
							'code_snippet' => $legacyCookie->getCodeSnippet(),
							'code_block' => implode("\n", $legacyCookie->getCodeBlock()),
							'embedding_vendor' => $legacyCookie->getCookieVendor(),
							'translations' => [],
							'assets' => [$asset],
						];

						$defaultLocale = LocaleService::getDefault();
						foreach (LocaleService::all() as $locale) {
							$language = $locale->getLanguage();
							if (!$language) {
								continue;
							}
							if ($legacyCookie->getTranslatedPrivacyPolicyUrl($locale, false) == '' and $legacyCookie->getTranslatedDescription($locale, false) == '') {
								continue;
							}
							$langShort = $language->getShortCode();
							$useFallback = ($locale->getName() == $defaultLocale->getName());
							$embedding['translations'][$langShort] = [
								'description' => $legacyCookie->getTranslatedDescription($locale, $useFallback),
								'privacy_policy_url' => $legacyCookie->getTranslatedPrivacyPolicyUrl($locale, $useFallback),
								'collected_data_info' => $legacyCookie->getTranslatedCollectedDataInfo($locale, $useFallback),
								'purpose_of_data_collection' => $legacyCookie->getTranslatedPurposeOfDataCollection($locale, $useFallback),
								'legal_basis' => $legacyCookie->getTranslatedLegalBasis($locale, $useFallback),
								'place_of_processing' => $legacyCookie->getTranslatedPlaceOfProcessing($locale, $useFallback),
							];
							$embedding['assets'][0]['translations'][$langShort] = [
								'lifetime' => "Session",
								'description' => '',
							];
						}

						$embeddings[] = $embedding;
						$matchedLegacy = true;
						break;
					}
				}
			}

			if (!$matchedLegacy) {
				if($cookie->isNecessary() || $cookie->getPurpose()==="c51")
				{
					$assetsNotwendig[] = $asset;
				}
				else{
					$assets[] = $asset;
				}
			}
		}

		if (Utils::isClickskeksCloud()) {
			// Merge assets of the preset 13002, as it's just a bunch of cookies
			$preset = array_reduce($embeddings, function (?array $carry, array $preset) {
				return $preset['id'] == 13002 ? $preset : $carry;
			});

			if ($preset) {
				$assets = array_merge($preset['assets'], $assets);
				$embeddings = array_filter($embeddings, function(array $preset) { return $preset['id'] != 13002; });
			}
		}

		//Wenn es zusätzlich technisch notwendige Cookies gibt, die nicht zugeordnet werden konnten.
		if ($assetsNotwendig) {
			//Checken ob es den Betreiber der Webseite gibt als Embedding, falls nicht markieren
			if(!empty($embeddings))
			{
				foreach($embeddings as $embedKey => $oneEmbed)
				{
					if($oneEmbed['id']=="13002")
					{
						//$isTechOk = true;
						unset($embeddings[$embedKey]);
					}
				}
			}
			//wenn markiert dann Betreiber anlegen.
			if($isTechOk===false)
			{
				$techEmbed=EmbeddingDb::fetchEmbedding(13002);
				$techEmbed['assets']=array();
				$embeddings[] = $techEmbed;
			}

			//Und dem Betreiber dann die notwendigen Cookies zuweisen, sonst hängen die unter sonstiges und werden nicht zwangsläufig aktiviert
			foreach($embeddings as $embedKey => $oneEmbed)
			{
				if($oneEmbed['id']=="13002")
				{
					//Technisch notwendige hineinmergen.
					$embeddings[$embedKey]['assets']=array_merge($embeddings[$embedKey]['assets'],$assetsNotwendig);
				}
			}
		}

		//Sonstige Cookies an sonstiges Embedding hängen
		if ($assets) {
			$embeddings[] = [
				'id' => 0,
				'name' => 'Sonstige',
				'purpose' => '',
				'code_snippet' => '',
				'code_block' => '',
				'embedding_vendor' => 'N.A.',
				'translations' => [
					'de' => [
						'label' => '',
						'description' => '',
						'privacy_policy_url' => '',
						'collected_data_info' => '',
						'purpose_of_data_collection' => '',
						'legal_basis' => 'Einwilligung, Art. 6 Abs. 1 lit. a DSGVO',
						'place_of_processing' => '',
					],
					'en' => [
						'label' => 'Other',
						'description' => '',
						'privacy_policy_url' => '',
						'collected_data_info' => '',
						'purpose_of_data_collection' => '',
						'legal_basis' => 'Consent, Art. 6 para. 1 s. 1 lit. c GDPR',
						'place_of_processing' => '',
					]
				],
				'assets' => $assets,
			];
		}
	}

	/**
	 * Hilfsfunktion: Prüft ob ein Cookiename in der Embeddingliste vorhanden ist
	 *
	 * @param string $cookieName
	 * @param array[] $embeddings
	 * @return bool
	 */
	private function cookieInEmbeddings(string $cookieName, array $embeddings): bool
	{
		foreach ($embeddings as $embedding) {
			foreach ($embedding['assets'] ?? [] as $asset) {
				if (($asset['dynamic'] and fnmatch($asset['name'], $cookieName)) or $asset['name'] == $cookieName) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * @param array[] $embeddings
	 * @return array<string,array[]> [purposeId => embeddings]
	 */
	public function splitEmbeddingsByPurpose(array $embeddings): array
	{
		$purposes = Purpose::all();
		$fallbackPurpose = Purpose::findByName('Sonstiges') ?? $purposes[count($purposes)-1];

		// Datensatz vorbereiten
		$result = [];
		foreach ($purposes as $purpose) {
			$result[$purpose->getId()] = [];
		}

		// Einträge alphabetisch sortieren, mit "Sonstige" ganz unten
		usort($embeddings, function ($a, $b) {
			if ($a['id'] == 0 && $b['id'] != 0) { return 1; }
			if ($a['id'] != 0 && $b['id'] == 0) { return -1; }
			if ($a['ccm_special'] ?? false && $b['id'] != 0) { return 1; }
			if ($b['ccm_special'] ?? false && $a['id'] != 0) { return -1; }
			return $a['name'] <=> $b['name']; });

		// Aufsplitten nach Purposes
		foreach ($embeddings as $embedding) {
			$purpose = $purposes[$embedding['purpose']] ?? $fallbackPurpose;
			$embedding['purpose'] = $purpose->getId();
			$embedding['active'] = true;

            //Sonderfall CookieBot / Usercentrics - raus damit
            if(stripos($embedding['name'], 'cookiebot') !== false || stripos($embedding['name'], 'usercentrics') !== false) {
                $embedding['active'] = false;
            }
			$result[$purpose->getId()][] = $embedding;
		}
		return $result;
	}

	/**
	 * @param string $url
	 * @return array
	 */
	public function fetchImprintData(string $url): array
	{
		//Crawler starten und Links sowie Daten holen...
		$impcrawler = new ImpressumCrawler();
		$res = $impcrawler->extract_impressum_data($url) ?? [];
		//Name nochmal bereinigen um üblichen Kram...
		$res['name'] = $this->replaceDat($res['name'] ?? '');
		return $res;
	}

	/**
	 * @param string $string
	 * @return string
	 */
	private function replaceDat(string $string)
	{
		$repl = [
			'Medieninhaber: ',
			'Name und Sitz des Versicherers: ',
			'sind das Webangebot des',
			'für Texte und Inhalt: ',
			'(haftungsbeschränkt)',
			'Angaben gemäß § 5 Telemediengesetz (TMG):',
			'Ein Online-Service der',
			'für den Inhalt:',
			'Vorsitzender:',
			'Verlag:',
			'Redaktion und Inhalt:',
			'Betreiber der Internetseiten:',
			'gemäß §6 TDG §6 MDStV:',
			'(Einzelunternehmer)',
			'Betreiber und Verantwortlicher der Seite:',
			'Sitz:',
			'und verantwortlich',
			'ist die',
			'für den Inhalt der Angebote:',
			'Seitenbetreiber i.S.d. § 5 TMG',
			'Publisher:',
			'(UNTERNEHMENSSITZ)',
			'für den Inhalt dieser Website nach TDG:',
			'Unsere Bankverbindung lautet: Kontoinhaber:',
			'der Schulhomepage ist der Schulträger:',
			'Betreiber dieser Website:',
			'Seiteninhaber:',
			'ist:',
			'für diese Seiten:',
			'Betreiber dieser Webseite:',
			'Inhaltverantwortliches Unternehmen:',
			'Kontaktformular',
			'Content Verantwortliche:',
			'Gestaltung und Realisierung:',
			'Geschäftsführung:',
			'Inhaltlich Verantwortliche gemäß § 6 MDStV und § 55 Abs. 2 RStV',
			'Alle rechtlichen Informationen der',
			'Betriebsgesellschaft:',
			'Körperschaft des öffentlichen Rechts',
			'Unsere Praxen',
			'zuständige Kammer',
			'ein Unternehmen der ',
			'Vorsitzende des Vorstands',
			'Lohnsteuerhilfeverein',
			'Ein Projekt der',
			'für redaktionelle Inhalte:',
			'Konzernzentrale:',
			'Vertreten durch:',
			'Vorstand:',
			'im Sinne von § 5 Abs. 1 TMG :',
			'für den Inhalt laut § 5 TMG und § 55 RStV:',
			'Verantwortlicher:',
			'Standort:',
			'für diese Website:',
			'für den Inhalt nach § 55 Abs. 2 RStV',
			'Träger der Website',
			'Ein unverbindliches Angebot von: ',
			'sdfsdfsdfsdf',
		];

		foreach ($repl as $v)
		{
			$string = str_ireplace($v,"",$string);
		}
		return $string;
	}


}
