<?php

namespace App\Component;

use App\Utils;
use Countable;
use Iterator;
use LimitIterator;
use LogicException;

class Pagination
{
	/** @var iterable $collection */
	private $collection;
	/** @var int $limit */
	private $limit;
	/** @var int $padding */
	private $padding;

	/** @var bool $dirty */
	private $dirty;
	/** @var array $cachedPagination */
	private $cachedPagination;
	/** @var array $cachedPage */
	private $cachedPage = [];

	/** @var int|null $totalCount */
	private $totalCount = null;

	private $currentPage = 1;

	/**
	 * @param Iterator|array $items Array oder Iterator der mehrfach durchlaufen werden kann
	 * @param int $limit
	 * @param int $padding
	 */
	public function __construct(iterable $items, int $limit = 10, int $padding = 3)
	{
		$this->collection = $items;
		$this
			->setLimit($limit)
			->setPadding($padding)
		;

		$this->dirty = true;
		$this->cachedPagination = [];
	}

	/**
	 * Diese Methode ermittelt die Anzahl auf der aktuellen Seite anzuzeigender Einträge.
	 * @return int
	 */
	public function getItemCount(): int
	{
		return count($this->getItems());
	}

	/**
	 * Diese Methode gibt alle für die aktuelle Seite relevanten Einträge zurück.
	 * @return array
	 */
	public function getItems(): array
	{
		if ($this->totalCount === null) {
			$this->build();
		}

		$pageno = $this->getCurrentPage();
		if ($this->cachedPage['page'] ?? null === $pageno) {
			return $this->cachedPage['items'];
		}

		$offset = $this->limit * ($pageno - 1);

		if (is_array($this->collection)) {
			$items = array_slice($this->collection, $offset, $this->limit);
		} else {
			$items = iterator_to_array(new LimitIterator($this->collection, $offset, $this->limit), false);
		}
		$this->cachedPage = ['page'=>$pageno, 'items'=>$items];
		return $items;
	}

	/**
	 * @param int $limit
	 * @return $this
	 */
	public function setLimit(int $limit): self
	{
		if ($limit < 1) {
			throw new LogicException('Pagination row limit may not be lower than 1.');
		}

		$this->limit = $limit;
		$this->dirty = true;
		return $this;
	}

	/**
	 * @param int $padding
	 * @return $this
	 */
	public function setPadding(int $padding): self
	{
		if ($padding < 0) {
			throw new LogicException('Pagination pad value may not be lower than 0.');
		}

		$this->padding = $padding;
		$this->dirty = true;
		return $this;
	}

	public function setCurrentPage(int $currentPage): self
	{
		$this->currentPage = $currentPage;
		$this->dirty = true;
		return $this;
	}

	/**
	 * @return int
	 */
	public function getCurrentPage(): int
	{
		$request = Utils::getCurrentRequest();
		return ($request && $request->query->has('page')) ? max(1, $request->query->getInt('page')) : $this->currentPage;
	}

	public function getTotalCount(): int
	{
		if ($this->totalCount === null) {
			$this->build();
		}
		return $this->totalCount;
	}


	/**
	 * @return array
	 */
	public function build(): array
	{
		if ($this->dirty == false) {
			return $this->cachedPagination;
		}

		$pad = $this->padding;
		$limit = $this->limit;
		$currentPage = $this->getCurrentPage();

		if (is_array($this->collection) or $this->collection instanceof Countable) {
			$objectCount = count($this->collection);
		}
		else {
			$this->collection = iterator_to_array($this->collection, false);
			$objectCount = count($this->collection);
		}

		$this->totalCount = $objectCount;
		$this->cachedPage = [];

		$lastPage = (int)floor($objectCount / $limit) + ($objectCount == 0 || $objectCount % $limit > 0 ? 1 : 0);

		$pagination = [[
			'page' => 1,
			'dots' => $currentPage - $pad > 2,
		]];

		$lowerBounds = max(2, $currentPage - $pad - max(0, $pad - ($lastPage - $currentPage)));
		$upperBounds = min($currentPage + $pad + max(0, $pad - $currentPage + 1), $lastPage);

		for ($i = $lowerBounds; $i <= $upperBounds; $i++) {
			if ($i == $lastPage) {
				break;
			}

			$pagination[] = [
				'page' => $i,
				'dots' => $i == $upperBounds && $i + 1 < $lastPage,
			];
		}

		if ($lastPage > 1) {
			$pagination[] = [
				'page' => $lastPage,
				'dots' => false,
			];
		}

		$this->cachedPagination = $pagination;
		$this->dirty = false;
		return $pagination;
	}
}
