<?php

namespace App\Component\Plugin;

use App\Component\Curl;
use App\Downloader;
use App\Exception\TempFileNotWritableException;
use App\Exception\ZipArchiveException;
use App\Exception\ZipArchiveExtractException;
use App\Model\PluginState;
use App\Utils;
use FilesystemIterator;
use Symfony\Component\Filesystem\Filesystem;
use ZipArchive;

class Manager
{
	private $pluginListCache = null;

	/**
	 * Sammelt Metadaten über die im Dateisystem vorhandenen Plugins
	 */
	protected function discoverLocalPlugins(): array
	{
		$plugins = [];
		$folders = new FilesystemIterator(Utils::getBaseDir() . '/plugins/');
		foreach ($folders as $path => $folder) {
			if (!$folder->isDir()) {
				continue;
			}
			if (substr($folder->getFilename(), 0, 1) === '.') {
				continue;
			}
			$infoFile = $path . '/composer.json';
			if (!file_exists($infoFile)) {
				continue;
			}

			$data = json_decode(@file_get_contents($infoFile), true);
			if (!$data) {
				$data = [
					'broken' => true,
					'error'=>'composer.json damaged or unreadable'
				];
			}
			else {
				$data['broken'] = false;
			}

			if (!isset($data['name'])) {
				$data['name'] = $folder->getFilename();
			}

			if (!isset($data['extra']['preview'])) {
				if (!isset($data['extra'])) {
					$data['extra'] = [];
				}
				if (file_exists($path . '/preview.png')) {
					$data['extra']['preview'] = $path . '/preview.png';
				}
				elseif (file_exists($path . '/preview.jpeg')) {
					$data['extra']['preview'] = $path . '/preview.jpeg';
				}
				elseif (file_exists($path . '/preview.svg')) {
					$data['extra']['preview'] = $path . '/preview.svg';
				}
				elseif (file_exists($path . '/preview.webp')) {
					$data['extra']['preview'] = $path . '/preview.webp';
				}
				elseif (file_exists($path . '/preview.gif')) {
					$data['extra']['preview'] = $path . '/preview.gif';
				}
				else {
					$data['extra']['preview'] = null;
				}
			} else {
				$data['extra']['preview'] = $path . '/' . $data['extra']['preview'];
			}

			$plugins[$folder->getFilename()] = ['folder' => $folder, 'info' => $data];
		}
		return $plugins;
	}

	/**
	 * Listet alle lokal vorhandenen Plugins, egal ob installiert, unkonfiguriert oder beschädigt.
	 * @return PluginInfo[]
	 */
	public function installedPlugins(): array
	{
		if ($this->pluginListCache !== null) {
			return $this->pluginListCache;
		}

		// Plugin-Verzeichnis durchsuchen
		$localPlugins = $this->discoverLocalPlugins();

		// PluginState zu vorhandenen Plugins matchen, dabei ggf. Einträge aktualisieren
		$matchedState = [];
		$plugins = [];
		$configured = PluginState::all();
		foreach ($localPlugins as $plugin) {
			foreach ($configured as $state) {
				if ($state->getPluginUID() == $plugin['info']['name']) {
					$folder = $plugin['folder']->getFilename();
					$plugins[$plugin['info']['name']] = $matchedState[$state->getId()] = new PluginInfo($plugin['folder'], $plugin['info'], $state->getId());
					if ($state->getDirname() !== $folder) {
						$state->setDirname($folder);
						$state->save();
					}
					$activatable = (bool)($plugin['info']['activatePerDomain'] ?? false);
					if ($state->isActivatablePerDomain() !== $activatable) {
						$state->setActivatablePerDomain($activatable);
						$state->save();
					}
				}
			}
		}
		foreach ($localPlugins as $plugin) {
			if (!isset($plugins[$plugin['info']['name']]) and isset($plugin['info']['extra']['compatId'])) {
				foreach ($configured as $state) {
					if ($state->getPluginUID() == $plugin['info']['extra']['compatId']) {
						$plugins[$plugin['info']['name']] = $matchedState[$state->getId()] = new PluginInfo($plugin['folder'], $plugin['info'], $state->getId());
						$state->setDirname($plugin['folder']->getFilename());
						$state->setPluginUID($plugin['info']['name']);
						$state->cleanupCompat();
						$state->save();
					}
				}
			}
			if (!isset($plugins[$plugin['info']['name']])) {
				foreach ($configured as $state) {
					if (rtrim($state->getDirname(),'/') == $plugin['folder']->getFilename()) {
						$plugins[$plugin['info']['name']] = $matchedState[$state->getId()] = new PluginInfo($plugin['folder'], $plugin['info'], $state->getId());
					}
				}
			}
			if (!isset($plugins[$plugin['info']['name']])) {
				$plugins[$plugin['info']['name']] = new PluginInfo($plugin['folder'], $plugin['info'], null);
			}
		}

		foreach ($configured as $state) {
			if (!isset($matchedState[$state->getId()])) {
				$plugins[$state->getPluginUID()] = new PluginInfo(null, [
					'name' => $state->getName(),
					'description' => '',
					'version' => $state->getVersion(),
					'broken' => true,
					'error' => 'plugin directory is missing',
				], $state->getId());
			}
		}

		usort($plugins, function ($a, $b) {
			return -($a->isValidAndActive() <=> $b->isValidAndActive()) ?: -($a->isActive() <=> $b->isActive()) ?: ($a->getName() <=> $b->getName());
		});

		$this->pluginListCache = $plugins;

		return $plugins;
	}

	/**
	 * Listet alle installierten (unbeschädigt und konfiguriert) Plugins.
	 * @return PluginInfo[]
	 */
	public function configuredPlugins(): array
	{
		return array_filter(
			$this->installedPlugins(),
			function (PluginInfo $plugin) { return $plugin->isValidAndActive(); }
		);
	}

	/**
	 * Holt Plugin-Informationen anhand eine Plugin-ID
	 * @param string $uid UID (vendor/name) oder CompatID
	 * @return PluginInfo|null
	 */
	public function findInstalledPlugin(string $uid): ?PluginInfo
	{
		foreach ($this->installedPlugins() as $plugin) {
			if ($plugin->getName() == $uid) {
				return $plugin;
			}
			elseif ($plugin->getCompatId() == $uid) {
				return $plugin;
			}
		}
		return null;
	}

	/**
	 * @param string $search
	 * @return array
	 */
	public function getAvailablePlugins(string $search = ""): array
	{
		return []; // TODO
		//Plugins URL ... evtl. später hinter Passwort?
		$url = 'https://plugins.ccm19.de/plugins.php?search='.$search;

		//Liste der verfügbaren Plugins holen...
		$curl = new Curl;
		$curl->setOpt(CURLOPT_CONNECTTIMEOUT, 3);
		$curl->setOpt(CURLOPT_TIMEOUT, 7);
		$curl->get($url);
		$resultData = $curl->getResponse();
		//print_r($resultData);exit();
		$curl->close();
		$plugins = (json_decode($resultData,true));

		if (!is_array($plugins))
		{
			return array();
		}

		return $plugins;
	}

	/**
	 * @param string $uid
	 * @return false|string|null
	 */
	public function downloadAndInstallPlugin(string $uid)
	{
		$url = 'https://plugins.ccm19.de/plugins/'.$uid."/plugin_".$uid.".zip";
		$filename = Utils::getVarDir().'/plugin'.basename($uid).'.zip';
		//print_r($filename);
		//Liste der verfügbaren Plugins holen...
		$downloader = new Downloader($url);
		$downloader->downloadInto($filename, true);
		$path = Utils::getVarDir().'/'.$uid;;
		@mkdir($path, 0700);


		// ZIP-Datei öffnen
		$zip = new ZipArchive();
		$offset=0;
		$result = $zip->open($filename);
		$umask = umask();
		$timeLimit=-1;

		if ($result !== true) {
			throw new ZipArchiveException('zip file open failed', $result);
		}

		// Dateien extrahieren
		$startTS = time();
		$dirs = array();
		$countFiles = $zip->numFiles;
		for ($idx=$offset; $idx < $countFiles; $idx++) {

			$stat = $zip->statIndex($idx);
			$entry = $stat['name'];
			$dirs[] = $entry;

			// Zur Sicherheit alles mit ../ und / am Anfang ignorieren
			if ($entry === '' or $entry[0] === '/' or strpos($entry, '../') !== false) {
				continue;
			}

			// Zielpfad bestimmen
			$dest = $path . DIRECTORY_SEPARATOR . $entry;

			// Verzeichnisse: Existenz sicherstellen
			if (substr($entry, -1) === '/') {
				@mkdir($dest, 0777 & ~$umask);
			}
			// Dateien: extrahieren
			else {
				@mkdir(dirname($dest), 0777 & ~$umask, true);
				// Bei kleinen Dateien (<= 512k) gesamten Inhalt auf einmal kopieren
				if ($stat['size'] <= 512*1024) {
					$data = @$zip->getFromIndex($idx);
					if ($data === false) {
						throw new ZipArchiveExtractException($entry);
					}
					if (@file_put_contents($dest, $data) === false) {
						throw new TempFileNotWritableException($dest);
					}
					$data = null;
				}
				// Ansonsten als Stream in Chunks kopieren
				else {
					$sourceFd = @$zip->getStream($entry);
					if (!$sourceFd) {
						throw new ZipArchiveExtractException($entry);
					}

					$destFd = @fopen($dest, 'wb');
					if (!$destFd) {
						throw new TempFileNotWritableException($dest);
					}

					try {
						// Daten kopieren
						while (!feof($sourceFd)) {
							fwrite($destFd, fread($sourceFd, 512*1024));
						}
					}
					finally {
						fclose($sourceFd);
						fclose($destFd);
					}
				}

				if (strpos($entry, '/bin/') !== false) {
					@chmod($dest, 0777 & ~$umask);
				} elseif (strpos($entry, '/index.php') !== false) {
					@chmod($dest, 0777 & ~$umask);
				} else {
					@chmod($dest, 0666 & ~$umask);
				}

				// Stoppen, wenn Zeitlimit erreicht
				if ($timeLimit >= 0 and time() - $startTS >= $timeLimit) {
					break;
				}
			}
		}

		$zip->close();
		$pluginVerzeichnis = $dirs[0];
		$filesystem = new Filesystem();
		$filesystem->mirror($path."/".basename($pluginVerzeichnis),Utils::getVarDir()."/../plugins/".basename($pluginVerzeichnis)."/");

		//temp Datei löschen... $filename
		$filesystem->remove($filename);

		//temp Verzeichnis löschen...
		$filesystem->remove(Utils::getVarDir().'/'.basename($uid));

		return $pluginVerzeichnis;

	}

	/**
	 * @return bool Ob Plugin-State und/oder Dateien gelöscht wurden
	 */
	public function deletePlugin(PluginInfo $plugin): bool
	{
		$success = false;
		$state = $plugin->getState();
		if ($state) {
			PluginState::delete($state->getId());
			$success = true;
		}

		$dirname = $plugin->getDirname();
		if ($dirname) {
			$success = $success || $this->deleteFiles($dirname);
		}
		return false;
	}

	/**
	 * Löscht Dateien eines Plugins von der Platte
	 */
	private function deleteFiles(string $dirname): bool
	{
		$filesystem = new Filesystem();
		if (!empty($dirname))
		{
			$filesystem->remove(Utils::getVarDir()."/../plugins/".$dirname);
			return true;
		}
		return false;
	}
}
