<?php

namespace App\Component\Plugin;

use App\Model\PluginState;
use App\Utils;
use SplFileInfo;
use Symfony\Component\HttpFoundation\File\File;

class PluginInfo
{
	/** @var SplFileInfo|null $folder Plugin-Hauptverzeichnis (null, wenn Plugin defekt/nicht vorhanden) */
	private $folder;
	/** @var array<string, mixed> $data Plugin-Metadaten (größtenteils aus composer.json) */
	private $data;
	/** @var string|null $stateId ID des zugehörigen PluginStates (null wenn nicht installiert) */
	private $stateId;

	public function __construct(?SplFileInfo $folder = null, array $data = [], ?string $stateId=null)
	{
		$this->folder = $folder;
		$this->data = $data;
		$this->stateId = $stateId;
	}

	/** Technischer Name (= UID)  */
	public function getName(): string
	{
		return (string)$this->data['name'];
	}

	/** Typ (i.d.R. ccm19-plugin oder ccm19-app)  */
	public function getType(): string
	{
		return (string)($this->data['type'] ?? 'ccm19-plugin');
	}

	/** Icon (haptsächlich für Apps) */
	public function getIcon(): string
	{
		return (string)($this->data['extra']['icon'] ?? 'fa-plug');
	}

	public function getCssSafeName(): string {
		return preg_replace('/[^a-zA-Z0-9-]/', '_', strtr($this->data['name'], ['--'=>'_','/'=>'--']));
	}

	/** Schlüssel im Rechte-Array von Agency-Usern */
	public function getUserRightsKey(): string {
		return PluginState::buildUserRightsKey($this->getName());
	}

	/** Schlüssel im Rechte-Array von Subgruppen */
	public function getSubgroupRightsKey(): string {
		return PluginState::buildSubgroupRightsKey($this->getName());
	}

	public function getDirname(): string
	{
		return ($this->folder) ? $this->folder->getFilename() : '';
	}

	public function getDirPath(): ?string
	{
		return $this->folder;
	}

	private function getTranslatedExtra(string $key, string $language, $fallback=null)
	{
		$language = str_replace('_', '-', $language);

		$candidates = [$language, explode('-',$language)[0]];
		if (strpos($language, '-') === false) {
			foreach ($this->data['extra'][$key] ?? [] as $lang => $_) {
				if (strpos($lang, $language.'-') === 9) {
					$candidates[] = $language;
				}
			}
		}
		$candidates[] = 'en';
		$candidates[] = 'en_US';
		$candidates[] = 'en_GB';

		foreach ($candidates as $candidate) {
			if (isset($this->data['extra'][$key][$candidate])) {
				return $this->data['extra'][$key][$candidate];
			}
		}
		return $fallback;
	}

	public function getManufacturerName(): string
	{
		$authors = $this->data['authors'] ?? [];
		$list = array_filter($authors, function ($item) {
			return (is_array($item) and strtolower($item['role']) == 'manufacturer');
		});
		if ($list) {
			return reset($list)['name'] ?? '';
		} elseif ($authors) {
			return reset($authors)['name'] ?? '';
		} else {
			return '???';
		}
	}

	public function getLabel($language='en'): string
	{
		return (string)$this->getTranslatedExtra('label', $language, $this->data['name']);
	}

	public function getDescription($language='en'): string
	{
		return (string)$this->getTranslatedExtra('description', $language, $this->data['description'] ?? '');
	}

	public function getSupportLink($language='en'): ?string
	{
		return (string)$this->getTranslatedExtra('supportLink', $language, null);
	}

	public function getManufacturerLink($language='en'): ?string
	{
		return (string)$this->getTranslatedExtra('manufacturerLink', $language, null);
	}

	public function getCompatId(): ?string
	{
		return $this->data['extra']['compatId'] ?? null;
	}

	public function getVersion(): string
	{
		return $this->data['version'] ?? '?';
	}

    public function isAllowTrial(): bool
    {
        return (bool)($this->data['extra']['allowTrial'] ?? false);
    }

	public function getCopyright(): string
	{
		return $this->data['extra']['copyright'] ?? '';
	}

	public function isBroken(): bool
	{
		return (bool)$this->data['broken'];
	}

	public function isHidden(): bool
	{
		return (bool)($this->data['extra']['hidden'] ?? false);
	}

	public function isPreinstalled(): bool
	{
		return (bool)($this->data['extra']['preinstalled'] ?? false);
	}

	public function getError(): ?string
	{
		return $this->data['error'] ?? null;
	}

	/**
	 * Ob das Plugin für einzelne Domains aktiviert werden kann
	 *
	 * In Varianten außer der Agency-Edition immer `false`.
	 */
	public function isActivatablePerDomain(): bool
	{
		return (Utils::isExtendedEdition() and (bool)($this->data['extra']['activatePerDomain'] ?? false));
	}

	/**
	 * Ob es für den Admin Einstellungen gibt
	 *
	 * Aktuell immer `true`.
	 */
	public function hasSettings(): bool
	{
		return true;
	}

	public function getPreviewImage(): ?File
	{
		$path = $this->data['extra']['preview'] ?? null;
		if ($path) {
			return new File($path);
		} else {
			return null;
		}
	}

	/**
	 * @return string Screenshot als data-URI
	 */
	public function getPreviewAsDataUri(): string
	{
		$file = $this->getPreviewImage();
		if ($file) {
			$data = @base64_encode(file_get_contents($file->getPathname()));
		} else {
			$data = '';
		}
		// Die ersten 3 Base64-Zeichen reichen aus um die wichtigsten Bildformate zu unterscheiden
		$head = substr($data, 0, 3);
		switch ($head) {
			case '':
				// Leeres 1x1px GIF, wenn kein Screenshot vorhanden
				return 'data:image/gif;base64,R0lGODlhAQABAIAAAP///wAAACH5BAEAAAAALAAAAAABAAEAAAICRAEAOw==';
			case 'R0l':
				return 'data:image/gif;base64,'.$data;
			case 'iVB':
				return 'data:image/png;base64,'.$data;
			case 'Ukl':
				return 'data:image/webp;base64,'.$data;
			case 'PD9':
				return 'data:image/svg+xml;base64,'.$data;
			case 'AAA':
				return 'data:image/avif;base64,'.$data;
			case '/9j':
			default:
				return 'data:image/jpeg;base64,'.$data;
		}
	}

	public function getState(): ?PluginState
	{
		return ($this->stateId !== null) ? PluginState::find($this->stateId) : null;
	}

	/**
	 * Ob das Plugin im Dateisystem vorhanden ist
	 */
	public function exists(): bool
	{
		return $this->folder !== null;
	}

	/**
	 * Ob das Plugin aktiviert wurde
	 */
	public function isActive(): bool
	{
		return $this->getState() !== null;
	}

	/**
	 * Ob das Plugin lauffähig *und* aktiviert ist
	 */
	public function isValidAndActive(): bool
	{
		return ($this->exists() and !$this->isBroken() and $this->isActive());
	}
}
