<?php

namespace App\Component\Reflection;

use LogicException;
use ReflectionClass;
use ReflectionException;
use ReflectionMethod;
use ReflectionNamedType;
use ReflectionParameter;
use ReflectionType;

class PropertyProcessor
{
	/**
	 * This method looks for matching getter and setter methods between the two objects.
	 * If $className is omitted, then getter and setter methods are gathered from `$source`
	 * and `$target` respectively – otherwise methods are gathered from `$className`.
	 * If the found methods are abstract or if `$source` and `$target` are of different
	 * classes, properties with the type hint `array` will be ignored as the expected data
	 * type for the contained values can't be guaranteed. Such properties have to be merged
	 * manually.
	 * @param object $source The object that the data is read from.
	 * @param object $target The object that the data is written to.
	 * @param string|null $className Use only getter/setter methods defined specifically in $className, ignore inheritance.
	 * @param string[] $setterBlacklist An array of setter method names that should be ignored; dash-separated arguments, if any. `['setterA', 'setterB-firstArgObjectId', …]`
	 * @return void
	 * @throws LogicException if the class does not exist. This should only ever occur during development, if at all.
	 */
	public static function copyValues(object $source, object $target, ?string $className=null, array $setterBlacklist=[]): void
	{
		try {
			$sourceClass = new ReflectionClass($className ?? $source);
			$targetClass = new ReflectionClass($className ?? $target);
		} catch (ReflectionException $exception) {
			throw new LogicException($exception->getMessage(), $exception->getCode(), $exception);
		}

		$argumentCache = [];

		foreach ($targetClass->getMethods(ReflectionMethod::IS_PUBLIC) as $setter) {
			// Skip if the method isn't defined directly by the reflected class or its name doesn't match the setter pattern
			if ($className && $setter->class != $targetClass->name
				|| !preg_match('~^set(?<property>[A-Z].*)~', $setter->name, $match)
			) {
				continue;
			}

			// Skip if the setter method is blacklisted
			if (in_array($setter->name, $setterBlacklist)) {
				continue;
			}

			$property = $match['property'];

			// Find a getter method that matches the name of the setter
			$getter = array_reduce(['get', 'has', 'is'], function (?ReflectionMethod $method, string $prefix) use ($sourceClass, $property) {
				try {
					return $method ?? $sourceClass->getMethod($prefix.$property);
				} catch (ReflectionException $exception) {
					return $method;
				}
			});

			// Skip if the method isn't defined directly inside the reflected class or if it isn't publicly accessible
			if (!$getter || $className && $getter->class != $sourceClass->name || !$getter->isPublic()) {
				continue;
			}

			// Determine argument and return types for both methods as needed
			$getterArgumentTypes = array_map(function (ReflectionParameter $parameter) {
				return $parameter->getType();
			}, $getter->getParameters());
			$setterArgumentTypes = array_map(function (ReflectionParameter $parameter) {
				return $parameter->getType();
			}, $setter->getParameters());

			$getterReturnType = $getter->getReturnType();

			// Ignore methods that don't type hint their arguments or return type
			if ($getterReturnType === null
				|| in_array(null, $getterArgumentTypes, true)
				|| in_array(null, $setterArgumentTypes, true)
			) {
				continue;
			}

			/**
			 * @param ReflectionType $type
			 * @return string[]
			 */
			$extractReflectionType = function (ReflectionType $type): array {
				$types = $type->allowsNull() ? ['null'] : [];
				if ($type instanceof ReflectionNamedType) {
					$types[] = $type->getName();
				}
				// TODO Implement checks for ReflectionUnionType and ReflectionIntersectionType as necessary.
				else {
					$types = [];
				}

				return $types;
			};

			/**
			 * @param ReflectionType|string $subject
			 * @param ReflectionType|string $test
			 * @return bool Returns `true` if $test accepts all data types that $subject can provide, `false` otherwise.
			 */
			$testReflectionType = function ($subject, $test) use ($extractReflectionType): bool {
				$subjectTypes = is_string($subject) ? [$subject] : $extractReflectionType($subject);
				$testTypes = is_string($test) ? [$test] : $extractReflectionType($test);

				if (empty($subjectTypes) || empty($testTypes)) {
					return false;
				}

				return empty(array_diff($subjectTypes, $testTypes));
			};

			// Skip array properties if the data type of their elements could vary from one object to the other
			$skipArrayTypes = $className && ($getter->isAbstract() || $setter->isAbstract())
				|| !$className && $getter->class != $setter->class;

			// Copy value from $source to $target
			try {
				// getValue(): T -> setValue(T)
				if (count($getterArgumentTypes) == 0 && count($setterArgumentTypes) == 1) {
					if ($testReflectionType($getterReturnType, $setterArgumentTypes[0])
						&& !($skipArrayTypes && $testReflectionType('array', $getterReturnType))
					) {
						$setter->invoke($target, $getter->invoke($source));
					}
				}
				elseif (count($getterArgumentTypes) == 1 && count($setterArgumentTypes) == 2) {
					// getValue(T2): T -> setValue(T2, T)
					if ($testReflectionType($getterReturnType, $setterArgumentTypes[1])
						&& !($skipArrayTypes && $testReflectionType('array', $getterReturnType))
						&& $getterArgumentTypes[0] instanceof ReflectionNamedType
						&& $setterArgumentTypes[0] instanceof ReflectionNamedType
						&& $getterArgumentTypes[0]->getName() == $setterArgumentTypes[0]->getName()
						&& class_exists($firstArgumentClassName = $getterArgumentTypes[0]->getName())
					) {
						// Add support for getter/setter methods that access values based on, for example, an App\Model\Purpose
						if (is_a($firstArgumentClassName, 'App\Model\ImmutableModel', true)) {
							$firstArgumentValues = $argumentCache[$firstArgumentClassName] ?? ($argumentCache[$firstArgumentClassName] = $firstArgumentClassName::all());
							foreach ($firstArgumentValues as $firstArgumentId => $firstArgumentValue) {
								// Skip if the setter method is blacklisted for the this specific argument
								if (in_array($setter->name . '-' . $firstArgumentId, $setterBlacklist)) {
									continue;
								}
								$setter->invoke($target, $firstArgumentValue, $getter->invoke($source, $firstArgumentValue));
							}
						}
					}
				}
			} catch (ReflectionException $exception) {}
		}
	}
}
