<?php

namespace App\Component\Tcf\v2;

use App\Exception\ConfigNotWritableException;
use App\Exception\DownloaderHTTPException;
use App\Component\Curl;

final class GoogleRepository extends Repository
{
	/**
	 * @override
	 * @return string[]
	 */
	protected function getDownloadUrls(): array
	{
		$filename = str_replace('.json', '.csv', $this->filename);
		$filename = preg_replace('/^google-/', '', $filename);
		return [$_ENV['TCF_GOOGLE_BASE_URL'].'/'.$filename];
	}

	/**
	 * Lädt eine neue Version einer Datei von storage.googleapis.com herunter.
	 *
	 * Wenn parallel schon ein anderer Prozess die Datei herunterlädt,
	 * wird, wenn es noch keine gecachte Datei gibt, der Prozess abgewartet,
	 * und ansonsten direkt FALSE zurück gegeben.
	 *
	 * @return bool FALSE, wenn ein paralleler Prozess schon ein Update durchführt,
	 *              das nicht abgewartet wurde.
	 * @throws DownloaderHTTPException
	 * @override
	 */
	protected function download(): bool
	{
		$url = $this->getDownloadUrls()[0];
		$destfile = $this->getPathname();
		$maxagefile = $destfile.'.maxage';
		$tmpfile = $destfile.'.csv.tmp';
		$tmpfile2 = $destfile.'.tmp';
		$mtime = @filemtime($destfile);
		$maxage = (int)(@file_get_contents($maxagefile) ?: 86400);

		$f = @fopen($tmpfile, 'cb+');
		if (!$f) {
			throw new ConfigNotWritableException($tmpfile);
		}
		// Versuchen, die Datei exklusiv zu sperren
		if (flock($f, LOCK_EX | LOCK_NB)) {
			try {
				$curl = new Curl();

				// If-Not-Modified-Since setzen
				if ($mtime) {
					$curl->setOpt(CURLOPT_TIMEVALUE, $mtime);
					$curl->setOpt(CURLOPT_TIMECONDITION, CURL_TIMECOND_IFMODSINCE);
				}

				// Direkten Download in geöffnete Datei durchführen
				ftruncate($f, 0);
				$curl->setOpt(CURLOPT_FILE, $f);
				$curl->get($url);

				// Status prüfen
				$status = $curl->getHttpStatus();
				if ($status == 304) {
					// Wenn "Not Modified": mtime anpassen, sodass es in spätestens 12 Stunden noch mal versucht wird
					touch($destfile, time()-max($maxage-43200, 0));
					@unlink($tmpfile);
					clearstatcache(false, $destfile);
					// Falls die Datei während des Checks beschädigt wurde, löschen und nächsten Aufruf neu herunterladen lassen.
					if (@filesize($destfile) == 0) {
						@unlink($destfile);
					}
					return true;
				}
				elseif ($status == 200) {
					// Wenn "Ok": Intervall merken und umbenennen

					// Max-Age-Intervall aus Headern extrahieren und speichern
					$cacheinfo = $curl->getResponseHeaders('Cache-Control') ?: '';
					$maxage_match = [];
					preg_match('/(?:^|\s|,)max-age=([0-9]+)/i', $cacheinfo, $maxage_match);
					if ($maxage_match) {
						file_put_contents($maxagefile, $maxage_match[1]);
					} else {
						// Default to 24 h
						file_put_contents($maxagefile, '86400');
					}

					// CSV zu JSON umbauen
					$f2 = @fopen($tmpfile2, 'wb');
					if (!$f2) {
						throw new ConfigNotWritableException($tmpfile2);
					}
					fseek($f, 0);
					$header = fgetcsv($f);
					if (!$header or $header[0] === null) {
						// Abbrechen, wenn Datei nicht verarbeitbar
						return false;
					}
					fwrite($f2, '{"data":{');
					$first = true;
					while (($line = fgetcsv($f))) {
						if (count($line) == count($header) and $line[0] !== null) {
							if ($header[3] == 'domains') {
								$domains = trim((string)$line[3]);
								if ($domains) {
									$line[3] = explode(' ', $domains);
								} else {
									$line[3] = [];
								}
							}
							fwrite($f2, ($first ? '' : ",").json_encode((string)$line[0]).':'.json_encode(array_combine($header, $line)));
							$first = false;
						}
					}
					fwrite($f2, '}}');
					fclose($f2);

					// Mtime setzen
					$mtime = @strtotime($curl->getResponseHeaders('Last-Modified'));
					if ($mtime) {
						touch($tmpfile2, $mtime);
					}

					// Umbenennen
					if (!rename($tmpfile2, $destfile)) {
						throw new ConfigNotWritableException($destfile);
					}
					@unlink($tmpfile);
				}
				else {
					// Exception bei anderen HTTP-Fehlern
					throw new DownloaderHTTPException($url, $curl->getHttpStatus());
				}
			}
			catch (DownloaderHTTPException $e) {
				@unlink($tmpfile);
				// HTTP-Fehler ignorieren, solange noch eine alte Version der Datei existiert
				if (file_exists($destfile)) {
					return false;
				} else {
					throw $e;
				}
			}
			catch (\Throwable $e) {
				@unlink($tmpfile);
				throw $e;
			}
			finally {
				flock($f, LOCK_UN);
				fclose($f);
			}
			return true;
		}
		// Wenn noch keine Zieldatei vorhanden, aber ein anderer Update-Prozess läuft, diesen abwarten
		elseif (!file_exists($destfile)) {
			flock($f, LOCK_SH);
			flock($f, LOCK_UN);
			fclose($f);
			clearstatcache(false, $destfile);
			return true;
		}
		// Ansonsten einfach abbrechen
		else {
			return false;
		}
	}

}
