<?php

namespace App\Component\Tcf\v2\Model;

use App\Component\Tcf\v2\Repository;
use App\Component\Tcf\v2\Model\Translation\Stack as Translated;
use App\Exception\ConfigAccessException;
use App\Utils;

class Stack extends Model
{
	protected static $repositoryFilename = 'vendor-list.json';
	protected static $repositoryKey = 'stacks';

	/** @var Repository|null $repository */
	protected static $repository = null;

	public function getName(): string
	{
		return $this->get('name');
	}

	public function getDescription(): string
	{
		return $this->get('description');
	}

	/**
	 * Alle Verwendungszwecke, die von diesem Stack abgedeckt werden
	 * @return Purpose[]
	 */
	public function getPurposes(): array
	{
		return array_filter(array_map(function ($id) {
			return Purpose::find($id);
		}, $this->get('purposes')));
	}

	/**
	 * Alle besonderen Funktionen, die von diesem Stack abgedeckt werden
	 * @return SpecialFeature[]
	 */
	public function getSpecialFeatures(): array
	{
		return array_filter(array_map(function ($id) {
			return SpecialFeature::find($id);
		}, $this->get('specialFeatures')));
	}

	/**
	 * Liefert die offizielle Übersetzungen für Name und Beschreibung, falls eine existiert
	 * @param string $langCode Sprachcode
	 * @param bool $fallbackUntranslated Wenn true, wird das unübersetzte Model zurückgegeben,
	 *        wenn die Sprache nicht unterstützt wird. Ansonsten gibt es eine Exception.
	 * @return static|Translated
	 */
	public function getTranslation(string $langCode, bool $fallbackUntranslated=true)
	{
		try {
			$result = Translated::find($langCode, $this->getId());
			return ($result) ? $result : $this;
		}
		catch (ConfigAccessException $e) {
			if ($fallbackUntranslated) {
				return $this;
			} else {
				throw $e;
			}
		}
	}

	public function __toString(): string
	{
		return sprintf('[TCFv2 Stack %s: "%s"]', $this->getId(), $this->getName());
	}

	/**
	 * Prüft, ob dieser Stack eine Übermenge von $stack ist
	 *
	 * $this muss alle Purposes und SpecialFeatures enthalten, die
	 * auch $stack enthält, aber darf nicht der selbe Stack sein.
	 */
	public function supersetOf(Stack $stack): bool
	{
		$selfPurposeIds = $this->get('purposes');
		$selfSpecialFeatureIds = $this->get('specialFeatures');
		$otherPurposeIds = $stack->get('purposes');
		$otherSpecialFeatureIds = $stack->get('specialFeatures');

		return (
			$stack->getId() !== $this->getId()
			and
			array_intersect($otherPurposeIds, $selfPurposeIds) == $otherPurposeIds
			and
			array_intersect($otherSpecialFeatureIds, $selfSpecialFeatureIds) == $otherSpecialFeatureIds
		);
	}

	/**
	 * Prüft, ob dieser Stack eine bestimmte Purpose oder ein SpecialFeature enthält
	 * @param Purpose|SpecialFeature $purposeOrSpecialFeature
	 * @return bool
	 */
	public function contains($purposeOrSpecialFeature): bool
	{
		$id = $purposeOrSpecialFeature->getId();
		$ids = $this->get(($purposeOrSpecialFeature instanceof SpecialFeature) ? 'specialFeatures' : 'purposes');
		return in_array($id, $ids);
	}

	/**
	 * Produziert eine möglichst kurze Liste von Stacks, die möglichst alle übergebenen Purposes und SpecialFeatures abdeckt
	 *
	 * Wenn $appendRemains TRUE ist, werden zusätzlich alle Purposes und SpecialFeatures zurückgeliefert, die nicht
	 * von einem Stack abgedeckt werden konnten.
	 *
	 * @param Purpose[] $purposes
	 * @param SpecialFeature[] $specialFeatures
	 * @param bool $appendRemains Append remaining stack-less purposes and special features to the result array
	 * @return (Stack|Purpose|SpecialFeature)[]
	 */
	public static function findMinimalSet(array $purposes, array $specialFeatures, bool $appendRemains=true): array
	{
		$purposeIds = array_map(function ($p) { return $p->getId(); }, $purposes);
		$purposes = array_combine($purposeIds, $purposes);
		$specialFeatureIds = array_map(function ($p) { return $p->getId(); }, $specialFeatures);
		$specialFeatures = array_combine($specialFeatureIds, $specialFeatures);

		// Stacks bestimmen, die die Purposes abdecken
		$result = [];
		foreach (self::all() as $stack) {
			$stackPurposeIds = $stack->get('purposes');
			$stackSpecialFeatureIds = $stack->get('specialFeatures');
			$intersectingPurposes = array_intersect($stackPurposeIds, $purposeIds);
			$intersectingSpecialFeatures = array_intersect($stackSpecialFeatureIds, $specialFeatureIds);

			if (
				($stackPurposeIds and $intersectingPurposes == $stackPurposeIds)
				or
				($stackSpecialFeatureIds and $intersectingSpecialFeatures == $stackSpecialFeatureIds)
			) {
				$result[$stack->getId()] = $stack;
			}
		}

		// Stacks rausfiltern, die von einem anderen Stack auch abgedeckt werden
		$result = array_filter($result, function ($item) use ($result) {
			return ! Utils::any($result, function ($stack) use ($item) { return $stack->supersetOf($item); });
		});

		// Übrigbleibende Purposes sammeln
		$purposes = array_filter($purposes, function ($p) use ($result) {
			return ! Utils::any($result, function ($stack) use ($p) {
				return $stack->contains($p);
			});
		});

		// Übrigbleibende SpecialFeatures sammeln
		$specialFeatures = array_filter($specialFeatures, function ($sf) use ($result) {
			return ! Utils::any($result, function ($stack) use ($sf) {
				return $stack->contains($sf);
			});
		});

		return ($appendRemains)
			? array_merge(array_values($result), array_values($purposes), array_values($specialFeatures))
			: array_values($result);
	}
}
