<?php

namespace App\Component\Tcf\v2\Model\Translation;
use App\Exception\ConfigAccessException;
use App\Component\Tcf\v2\Repository;
use LogicException;

/**
 * Basisklasse für TCFv2 Übersetzungs-Models
 */
abstract class Model
{
	/** @var array $dataset */
	private $dataset;

	/** @var string $language */
	private $language;

	/** @var string|null $repositoryFilenameTemplate Dateiname mit "%" als Platzhalter für den Sprachcode */
	protected static $repositoryFilenameTemplate = 'purposes-%.json';

	/** @var array<string,string> $supportedLanguages */
	protected static $supportedLanguages = [
		'bg' => 'bg', 'ca' => 'ca', 'cs' => 'cs', 'da' => 'da', 'de' => 'de', 'el' => 'el',
		'es' => 'es', 'et' => 'et', 'fi' => 'fi', 'fr' => 'fr', 'hr' => 'hr', 'hu' => 'hu',
		'it' => 'it', 'ja' => 'ja', 'lt' => 'lt', 'lv' => 'lv', 'mt' => 'mt', 'nl' => 'nl',
		'no' => 'no', 'pl' => 'pl', 'pt' => 'pt', 'ro' => 'ro', 'ru' => 'ru', 'sk' => 'sk',
		'sl' => 'sl', 'sr' => 'sr-Latn', 'sv' => 'sv', 'tr' => 'tr', 'zh' => 'zh'
	];

	/** @var array<string,Repository> $repositoryKey */
	protected static $repositories = [];

	/** @var string|null $repositoryKey */
	protected static $repositoryKey = null;


	/**
	 * @param string $langCode
	 * @param array $dataset
	 */
	protected function __construct(string $langCode, array $dataset)
	{
		$this->language = $langCode;
		$this->dataset = $dataset;
	}

	/**
	 * Diese Methode stellt die Konfiguration des aktuellen Models bereit.
	 * @param string $langCode
	 * @return Repository
	 * @phan-suppress PhanUndeclaredStaticProperty
	 * @phan-suppress PhanUndeclaredStaticMethod
	 */
	protected static function getRepository(string $langCode): Repository
	{
		if (empty(static::$repositories[$langCode])) {
			static::$repositories[$langCode] = static::openRepository($langCode);
		}
		return static::$repositories[$langCode];
	}

	/**
	 * Diese Methode stellt die Konfiguration des aktuellen Models bereit und lädt dieses ggf. neu.
	 *
	 * @param string $langCode
	 * @return Repository
	 */
	protected static function openRepository(string $langCode): Repository
	{
		$repoFilename = static::$repositoryFilenameTemplate;

		// Gewährleisten, dass dem Model ein Dateiname als Repository beiliegt
		if (empty($repoFilename)) {
			throw new LogicException('No repository defined for '.static::class);
		}
		if (empty(static::$repositoryKey)) {
			throw new LogicException('No repository key defined for '.static::class);
		}

		$langCode = explode('_', $langCode, 2)[0];
		$langCode = explode('-', $langCode, 2)[0];
		$langCode = strtolower($langCode);

		if (!isset(static::$supportedLanguages[$langCode])) {
			throw new ConfigAccessException('Language '.$langCode.' is not supported for TCF translations');
		}

		$langCode = static::$supportedLanguages[$langCode];

		return Repository::load(str_replace('%', $langCode, $repoFilename ?? ''));
	}

	/**
	 * @param string $langCode
	 * @param string $id
	 * @return bool
	 */
	public static function exists(string $langCode, string $id): bool
	{
		return $id ? is_array(self::getRepository($langCode)->get(static::$repositoryKey.'.'.$id)) : false;
	}

	/**
	 * @param string $langCode
	 * @param string $id
	 * @return static|null
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function find(string $langCode, string $id): ?self
	{
		return self::exists($langCode, $id) ? new static($langCode, self::getRepository($langCode)->get(static::$repositoryKey.'.'.$id)) : null;
	}

	/**
	 * @param string $langCode
	 * @return static[]
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function all(string $langCode): array
	{
		/** @var static[] $entities */
		$entities = [];

		foreach (self::getRepository($langCode)->get(static::$repositoryKey) as $id => $data) {
			$entities[$id] = new static($langCode, $data);
		}

		return $entities;
	}

	/**
	 * @param string $langCode
	 * @return array
	 */
	public static function rawAll(string $langCode): array
	{
		return self::getRepository($langCode)->get(static::$repositoryKey, []);
	}

	/**
	 * Zählt die vorhandenen Instanzen dieses Models.
	 * @param string $langCode
	 * @return int
	 */
	public static function count(string $langCode): int
	{
		return count(self::getRepository($langCode)->get(static::$repositoryKey));
	}

	/**
	 * Größe des Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @param string $langCode
	 * @return int Größe in Bytes
	 */
	public static function persistedRepositorySize(string $langCode): int
	{
		return static::getRepository($langCode)->getPersistedSize();
	}

	/**
	 * @param string $property
	 * @param mixed|null $fallback
	 * @return mixed
	 */
	protected final function get(string $property, $fallback = null)
	{
		return isset($this->dataset[$property]) ? $this->dataset[$property] : $fallback;
	}

	/**
	 * Liest alle Einträge neu von der Festplatte oder ggf. dem Internet ein.
	 * @param string $langCode
	 * @throws ConfigAccessException
	 */
	public static function reload(string $langCode): void
	{
		self::getRepository($langCode)->reload();
	}

	/**
	 * @return string
	 */
	public final function getId(): string
	{
		return (string)$this->get('id');
	}

	/**
	 * @return string Sprach-Code
	 */
	public final function getLanguage(): string
	{
		return (string)$this->language;
	}
}

