<?php

namespace App\Component\Tcf\v2\Model;

use App\Component\Tcf\v2\Repository;
use App\Component\Tcf\v2\Model\Translation\SpecialPurpose as SpecialPurposeTranslated;
use \DateTimeInterface;
use \DateTime;

class Vendor extends Model
{
	protected static $repositoryFilename = 'vendor-list.json';
	protected static $repositoryKey = 'vendors';

	/** @var Repository|null $repository */
	protected static $repository = null;

	public function getName(): string
	{
		return (string)$this->get('name', '');
	}

	public function getPolicyUrl(string $langCode = 'en'): string
	{
		if (strpos($langCode, '_') !== false) {
			$langCode = explode('_', $langCode)[0];
		} elseif (strpos($langCode, '-') !== false) {
			$langCode = explode('-', $langCode)[0];
		}
		$urls = $this->get('urls', []);
		// Passende URL in TCFv2.2-Daten suchen
		foreach ($urls as $urlItem) {
			if (($urlItem['langId'] ?? '') == $langCode and !empty($urlItem['privacy'])) {
				return $urlItem['privacy'];
			}
		}
		// Fallback zu erster deklarierter URL
		foreach ($urls as $urlItem) {
			if (!empty($urlItem['privacy'])) {
				return $urlItem['privacy'];
			}
		}
		// Fallback für TCFv2
		return (string)$this->get('policyUrl', '');
	}

	/**
	 * @return Purpose[]
	 */
	public function getPurposes(): array
	{
		return array_filter(array_map(function ($id) {
			return Purpose::find($id);
		}, $this->get('purposes', [])));
	}

	/**
	 * @return Purpose[]
	 */
	public function getFlexiblePurposes(): array
	{
		return array_filter(array_map(function ($id) {
			return Purpose::find($id);
		}, $this->get('flexiblePurposes', [])));
	}


	/**
	 * @return int[]
	 */
	public function getFlexiblePurposeIds(): array
	{
		return (array)$this->get('flexiblePurposes', []);
	}

	/**
	 * @return Purpose[]
	 */
	public function getLegitimateInterestPurposes(): array
	{
		return array_filter(array_map(function ($id) {
			return Purpose::find($id);
		}, $this->get('legIntPurposes', [])));
	}

	/**
	 * @return Purpose[]
	 */
	public function getAllPurposes(): array
	{
		return array_merge(
			$this->getPurposes(),
			$this->getLegitimateInterestPurposes()
		);
	}

	/**
	 * @return SpecialPurpose[]
	 */
	public function getSpecialPurposes(): array
	{
		return array_filter(array_map(function ($id) {
			return SpecialPurpose::find($id);
		}, $this->get('specialPurposes', [])));
	}

	/**
	 * @return Feature[]
	 */
	public function getFeatures(): array
	{
		return array_filter(array_map(function ($id) {
			return Feature::find($id);
		}, $this->get('features', [])));
	}

	/**
	 * @return SpecialFeature[]
	 */
	public function getSpecialFeatures(): array
	{
		return array_filter(array_map(function ($id) {
			return SpecialFeature::find($id);
		}, $this->get('specialFeatures', [])));
	}

	/**
	 * @return DataCategory[]
	 */
	public function getDataCategories(): array
	{
		return array_filter(array_map(function ($id) {
			return DataCategory::find($id);
		}, $this->get('dataDeclaration', [])));
	}

	/**
	 * @return DateTimeInterface|null
	 */
	public function getDeletedDate(): ?DateTimeInterface
	{
		$data = $this->get('deletedDate');
		return ($data ? new DateTime($data) : null);
	}

	public function isDeleted(): bool
	{
		$date = $this->getDeletedDate();
		if ($date) {
			return $date <= (new DateTime('now'));
		} else {
			return false;
		}
	}

	public function isUsingCookies(): bool
	{
		return (bool)($this->get('usesCookies') ?? true);
	}

	public function getCookieMaxAgeSeconds(): ?int
	{
		$value = $this->get('cookieMaxAgeSeconds');
		return $value === null ? $value : (int)$value;
	}

	public function hasCookieRefresh(): ?bool
	{
		return $this->get('cookieRefresh');
	}

	public function isUsingNonCookieAccess(): ?bool
	{
		return $this->get('usesNonCookieAccess');
	}

	public function getHttpGetLimit(): ?int
	{
		$data = $this->get('overflow', []);
		return $data['httpGetLimit'] ?? null;
	}

	public function getDeviceStorageDisclosureUrl(): ?string
	{
		return $this->get('deviceStorageDisclosureUrl') ?? null;
	}

	/**
	 * @return DeviceStorageDisclosure[]|null
	 */
	public function getDeviceStorageDisclosures(): ?array
	{
		if ($this->getDeviceStorageDisclosureUrl()) {
			DeviceStorageDisclosure::selectVendor($this);
			return DeviceStorageDisclosure::all();
		} else {
			return null;
		}
	}

	/**
	 * @param Purpose|SpecialPurpose $purpose
	 * @return int|null
	 */
	public function getDataRetentionDays($purpose): ?int
	{
		$array = $this->get('dataRetention', []);
		$default = $array['stdRetention'] ?? null;
		if ($purpose instanceof SpecialPurpose or $purpose instanceof SpecialPurposeTranslated) {
			$array = $array['specialPurposes'] ?? [];
		} else {
			$array = $array['purposes'] ?? [];
		}
		return $array[$purpose->getId()] ?? $default;
	}

	public static function getGvlSpecificationVersion(): int
	{
		return self::getRepository()->get('gvlSpecificationVersion', 2);
	}

	public static function getVendorListVersion(): int
	{
		return self::getRepository()->get('vendorListVersion', 0);
	}

	public static function getTcfPolicyVersion(): int
	{
		return self::getRepository()->get('tcfPolicyVersion', 0);
	}

	public function __toString(): string
	{
		return sprintf('[TCFv2 Vendor %s: "%s"]', $this->getId(), $this->getName());
	}
}
