<?php

namespace App\Component\Validator;

use LogicException;

class DataTypeValidator
{
	/**
	 * @param $value
	 * @param string $type
	 * @return bool
	 * @throws LogicException
	 */
	private static function checkPrimitive($value, string $type): bool
	{
		switch ($type) {
			case 'bool':
				$valid = is_bool($value);
				break;
			case 'string':
				$valid = is_string($value);
				break;
			case 'int':
				$valid = is_int($value);
				break;
			case 'float':
				$valid = is_float($value);
				break;
			case 'date':
				$valid = is_string($value) && preg_match('~^\d{4}-\d\d-\d\d$~', $value) && strtotime($value) !== false;
				break;
			case 'array':
				$valid = is_array($value);
				break;
			case 'null':
				$valid = $value === null;
				break;
			default:
				throw new LogicException("Can't evaluate payload, unknown data type \"$type\".");
		}

		return $valid;
	}

	/**
	 * This method checks whether a string is a valid representation of an unsigned integer.
	 * @param string $value
	 * @return bool
	 */
	public static function checkStringForUInt(string $value): bool
	{
		/** @var int|false $number */
		$number = filter_var($value, FILTER_VALIDATE_INT);
		return $number !== false && $number >= 0;
	}

	/**
	 * @param mixed $value
	 * @param string $type e.g. "bool", "int[]", "?string", "float|int" or "array"
	 * @return bool
	 * @throws LogicException if a check for a specific data type needs to be implemented. (only during development)
	 */
	public static function verify($value, string $type): bool
	{
		return array_reduce(explode('|', $type), function (bool $valid, string $type) use ($value) {
			if ($valid) {
				return true;
			}

			if (preg_match('~^(?<nullable>\?)?(?<type>[a-z]+)(?<collection>\[])?$~', $type, $typeInfo) == false) {
				throw new LogicException("Can't evaluate payload, unknown data type \"$type\".");
			}

			$isNullable = (bool)($typeInfo['nullable'] ?? false);
			$isCollection = (bool)($typeInfo['collection'] ?? false);

			if ($isNullable && $isCollection) {
				throw new LogicException('Mixing nullable and collection modifier is ambiguous, no behavior implemented.');
			}

			if ($isCollection) {
				if (is_array($value)) {
					$valid = array_reduce($value, function (bool $valid, $value) use ($typeInfo) {
						return $valid && self::checkPrimitive($value, $typeInfo['type']);
					}, true);
				} else {
					$valid = false;
				}
			}
			elseif ($isNullable && $value === null) {
				$valid = true;
			}
			else {
				$valid = self::checkPrimitive($value, $typeInfo['type']);
			}

			return $valid;
		}, false);
	}
}
