<?php

namespace App;

use App\Component\Config\RepositoryInterface;
use App\Component\Config\Store\JsonConfig;
use App\Component\Config\Store\MongoDb;
use App\Component\Media\MediaFileService;
use App\Exception\ConfigNotWritableException;
use App\Model\Language;
use App\Utils;
use MongoDB\Driver\Exception\WriteException;
use Symfony\Component\HttpFoundation\File\File;

class Config
{
	const CONFIG_COLLECTION = 'cm-config';

	/** @var self|null $instance */
	private static $instance = null;

	/** @var RepositoryInterface $config */
	private $config;

	private function __construct()
	{
		$this->config = self::openRepository();
	}

	private static function openRepository(): RepositoryInterface
	{
		$repositoryStore = Utils::getRepositoryStore();
		switch ($repositoryStore) {
			case Utils::REPOSITORY_STORE_MONGODB:
				$repository = new MongoDb(self::CONFIG_COLLECTION, null, null, true, null);
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$repository = new JsonConfig(self::CONFIG_COLLECTION.'.json');
				break;
		}
		return $repository;
	}

	/**
	 * Diese Methode gibt die Instanz der Klasse Config zurück, die nach dem Singleton-Pattern erzeugt wird.
	 * @return self
	 */
	public static function getInstance(): self
	{
		if (self::$instance === null) {
			self::$instance = new self;
		}

		return self::$instance;
	}

	/**
	 * Diese Methode liest Daten aus der Konfiguration. Schlüssel werden im $identifier durch einen Punkt separiert.
	 * @param string $identifier
	 * @param mixed $fallback
	 * @return mixed
	 */
	public function get(string $identifier, $fallback = null)
	{
		return $this->config->get($identifier, $fallback);
	}

	/**
	 * Diese Methode schreibt Daten in die Konfiguration. Schlüssel werden im $identifier durch einen Punkt separiert.
	 * @param string $identifier
	 * @param mixed $data
	 * @return $this
	 */
	public function set(string $identifier, $data): self
	{
		$this->config->set($identifier, $data);
		return $this;
	}

	/**
	 * @param string $identifier
	 */
	public function unset(string $identifier): void
	{
		$this->config->unset($identifier);
	}

	/**
	 * Diese Methode schreibt alle an der Konfiguration vorgenommenen Änderungen auf die Platte.
	 * @throws ConfigNotWritableException
	 */
	public function flush(): void
	{
		$this->config->flush();
	}

	/**
	 * Diese Methode liest die Konfiguration erneut von der Festplatte ein. Sämtliche Änderungen gehen verloren.
	 */
	public function reload(): void
	{
		$this->config->reload();
		$this->config = self::openRepository();
	}

	/**
	 * Sperrt die Konfiguration zum Schreiben.
	 *
	 * Danach kann bis zu flush() oder reload() kein parallel laufender
	 * Prozess auf die Datei zugreifen.
	 */
	public function beginWrite(): void
	{
		$this->config->beginWrite();
	}

	/**
	 * @return array
	 */
	public function getListOfInstalledPlugins(): array
	{
		return (array)$this->get('ListOfInstalledPlugins', array());
	}

	/**
	 * @param array $plugins
	 * @return $this
	 */
	public function setListOfInstalledPlugins(array $plugins): self
	{
		$this->set('ListOfInstalledPlugins', $plugins);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isForceHttpsConnection(): bool
	{
		return (bool)$this->get('forceHttpsConnection', false);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setForceHttpsConnection(bool $state): self
	{
		$this->set('forceHttpsConnection', $state);
		return $this;
	}


	/**
	 * Maximale Generation, auf die bei Updates derzeit gesprungen werden kann.
	 */
	public function getMaximumGeneration(): int
	{
		return (int)$this->get('maxGeneration', 1);
	}

	/**
	 * @param int $generationNumber
	 * @return $this
	 */
	public function setMaximumGeneration(int $generationNumber): self
	{
		$this->set('maxGeneration', $generationNumber);
		return $this;
	}

	/**
	 * @param int $generationNumber
	 * @return $this
	 */
	public function raiseMaximumGeneration(int $generationNumber): self
	{
		$this->beginWrite();
		$generationNumber = max($this->getMaximumGeneration(), Utils::getVersionGeneration(), $generationNumber);
		$this->set('maxGeneration', $generationNumber);
		return $this;
	}

	/**
	 * Liste der Sprachen für System-Embeddings
	 *
	 * @return Language[]
	 */
	public function getSystemEmbeddingLanguages(): array
	{
		$result = $this->get('embeddingLanguages', ['de_DE', 'en_US']);

		return array_filter(array_combine($result, array_map(function ($code) {
			return Language::find($code);
		}, $result)));
	}

	/**
	 * @param Language[] $languages
	 * @return $this
	 */
	public function setSystemEmbeddingLanguages(array $languages): self
	{
		$this->set('embeddingLanguages', array_map(function (Language $l) {
			return $l->getId();
		}, $languages));
		return $this;
	}

	/**
	 * Ob die globale Cookie-DB für Endnutzer verwendet wird
	 */
	public function isGlobalCookieDbUsed(): bool
	{
		return $this->get('useGlobalCookieDb', true);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setGlobalCookieDbUsed(bool $value): self
	{
		$this->set('useGlobalCookieDb', $value);
		return $this;
	}

	/**
	 * IP-Adressen die nicht gezählt werden sollen
	 *
	 * @return string[]
	 */
	public function getExemptAddresses(): array
	{
		return (array)$this->get('exemptAddresses', []);
	}

	/**
	 * @param string[] $value
	 * @return $this
	 */
	public function setExemptAddresses(array $value): self
	{
		$this->set('exemptAddresses', $value);
		return $this;
	}

	/**
	 * User-Agents die nicht gezählt werden sollen
	 *
	 * @return string[]
	 */
	public function getExemptUserAgents(): array
	{
		return (array)$this->get('exemptUserAgents', []);
	}

	/**
	 * @param string[] $value
	 * @return $this
	 */
	public function setExemptUserAgents(array $value): self
	{
		$this->set('exemptUserAgents', $value);
		return $this;
	}

	public function getCustomBackendLogo(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getGlobalRepository('backendLogos');
		$identifier = $this->get('backendLogo', '');
		if ($identifier != '') {
			$result = $repo->getFile($identifier);
		} else {
			// Abwärtskompatibilität
			$identifier = 'logo.png';
			$result = null;
		}
		if (!$result and strpos($identifier, '/') === false and file_exists(Utils::getVarDir() . '/' . $identifier)) {
			return new File(Utils::getVarDir() . '/' . $identifier);
		} else {
			return $result;
		}
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setCustomBackendLogo(?File $file): self
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getGlobalRepository('backendLogos');
		$prevIdentifier = $this->get('backendLogo', '');

		$identifier = $repo->replaceFile($prevIdentifier, $file, 'logo');
		$this->set('backendLogo', $identifier);

		// Abwärtskompatibilität: Logo an altem Speicherort löschen
		if (strpos($prevIdentifier, '/') === false and file_exists(Utils::getVarDir() . '/' . $prevIdentifier)) {
			@unlink(Utils::getVarDir() . '/' . $prevIdentifier ?: 'logo.png');
		}

		return $this;
	}

	public function getCustomLoginLogo(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getGlobalRepository('loginLogos');
		$identifier = $this->get('loginLogo', '');
		if ($identifier != '') {
			$result = $repo->getFile($identifier);
		} else {
			// Abwärtskompatibilität
			$identifier = 'loginlogo.png';
			$result = null;
		}
		if (!$result and strpos($identifier, '/') === false and file_exists(Utils::getVarDir() . '/' . $identifier)) {
			return new File(Utils::getVarDir() . '/' . $identifier);
		} else {
			return $result;
		}
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setCustomLoginLogo(?File $file): self
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getGlobalRepository('loginLogos');
		$prevIdentifier = $this->get('loginLogo', '');

		$identifier = $repo->replaceFile($prevIdentifier, $file, 'loginlogo');
		$this->set('loginLogo', $identifier);

		// Abwärtskompatibilität: Logo an altem Speicherort löschen
		if (strpos($prevIdentifier, '/') === false and file_exists(Utils::getVarDir() . '/' . $prevIdentifier)) {
			@unlink(Utils::getVarDir() . '/' . $prevIdentifier ?: 'loginlogo.png');
		}

		return $this;
	}

	/**
	 * @return File|null
	 */
	public function getCustomFrontendImage(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getGlobalRepository('imageFrontend');
		$identifier = $this->get('imageFrontend', '');
		//$identifier='';
		if ($identifier != '') {
			$result = $repo->getFile($identifier);
		} else {
			// Fallback Standard.
			$imageData = Utils::getBackgroundImage();
			$identifier = $imageData['src'];
			$result = null;
		}
		if (!$result  and file_exists(Utils::getBaseDir() . '/public/' . $identifier)) {
			return new File(Utils::getBaseDir() . '/public/' . $identifier);
		} else {
			return $result;
		}
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setCustomFrontendImage(?File $file): self
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getGlobalRepository('imageFrontend');
		$prevIdentifier = $this->get('imageFrontend', '');

		$identifier = $repo->replaceFile($prevIdentifier, $file, 'imageFrontend');
		$this->set('imageFrontend', $identifier);

		return $this;
	}

	/**
	 * @return File|null
	 */
	public function getCustomFavicon(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getGlobalRepository('branding');
		$identifier = (string)$this->get('favicon');
		return $identifier == '' ? null : $repo->getFile($identifier);
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setCustomFavicon(?File $file): self
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getGlobalRepository('branding');
		$prevIdentifier = (string)$this->get('favicon');

		$identifier = $repo->replaceFile($prevIdentifier, $file, 'favicon');
		$this->set('favicon', $identifier);

		return $this;
	}

	/**
	 * Cluster-System: Ob ein Cronjob noch läuft
	 */
	public function isCronjobRunning(): bool
	{
		$value = $this->get('cronjobRunningTS');
		if (!$value) {
			return false;
		}
		if ((time() - (int)$value) > 30) {
			return false;
		}
		return true;
	}

	/**
	 * Cluster-System: Cronjob als laufend markieren (für max. 2 Minuten)
	 *
	 * NOTE: Führt immer ein flush() und bei Fehlern reload() aus.
	 *
	 * @return bool Erfolg; FALSE wenn bereits ein Cronjob läuft
	 */
	public function setCronjobRunning(): bool
	{
		$this->beginWrite();
		$value = $this->get('cronjobRunningTS');
		if ($value and (time() - (int)$value) <= 120) {
			return false;
		}
		try {
			$this->set('cronjobRunningTS', (int)time());
			$this->flush();
		} catch (WriteException $ignore) {
			return false;
		}
		return true;
	}

	/**
	 * Cluster-System: Cronjob als nicht mehr laufend markieren
	 *
	 * NOTE: Führt immer ein flush() aus
	 *
	 * @return $this
	 */
	public function unsetCronjobRunning(): self
	{
		for ($i = 0; $i < 10; ++$i) {
			try {
				$this->set('cronjobRunningTS', null);
				$this->flush();
				break;
			} catch (WriteException $ignore) {
				$this->reload();
				usleep(rand(10000, 30000));
			}
		}
		return $this;
	}

	/**
	 * Basis-URL, die verwendet wird, wenn HTTP_HOST nicht verfügbar ist
	 *
	 * @return string
	 */
	public function getBaseUrl(): string
	{
		return (string)$this->get('baseUrl', '');
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setBaseUrl(?string $value): self
	{
		$this->set('baseUrl', $value ?? '');
		return $this;
	}

	/**
	 * Whitelabel-Agency: Eigenes Backend-CSS
	 *
	 * @return bool
	 * @phan-side-effect-free
	 */
	public function isWhiteLabel(): bool
	{
		return (bool)$this->get('whitelabel', false);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setWhiteLabel(bool $value): self
	{
		$this->set('whitelabel', $value);
		return $this;
	}

	/**
	 * Whitelabel-Agency: Eigenes Backend-CSS
	 *
	 * @return string|null
	 * @phan-side-effect-free
	 */
	public function getBackendCustomCss(): ?string
	{
		$value = $this->get('themeBack_customCss', null);
		return ($value === null) ? null : (string)$value;
	}

	/**
	 * @param string|null $customCss
	 * @return $this
	 */
	public function setBackendCustomCss(?string $customCss): self
	{
		$origins = Utils::extractCSPOrigins($customCss ?? '');
		$this->set('themeBack_customCssOrigins', $origins);

		$this->set('themeBack_customCss', $customCss);
		return $this;
	}

	/**
	 * Externe Origins die im Backend-CSS gefunden wurden.
	 *
	 * Dienen dem Einfügen in eine Content-Security-Policy
	 *
	 * @return string[]
	 */
	public function getBackendCustomCssOrigins(): array
	{
		return (array)$this->get('themeBack_customCssOrigins');
	}

	/**
	 * @return bool
	 */
	public function isShowFrontendWidgetOnlyIfEmbeddingsExist(): bool
	{
		return (bool)$this->get('showFrontendWidgetOnlyIfEmbeddingsExist');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowFrontendWidgetOnlyIfEmbeddingsExist(bool $state): self
	{
		$this->set('showFrontendWidgetOnlyIfEmbeddingsExist', $state);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getImprintURL(): string
	{
		return (string)($this->get('imprintURL') ?? '');
	}

	/**
	 * @return string
	 */
	public function getDataPrivacyURL(): string
	{
		return (string)($this->get('dataPrivacyURL') ?? '');
	}

	/**
	 * (Kurzer) Name dieser Anwendung
	 * @return string
	 */
	public function getAppName(): string
	{
		$result = (string)($this->get('appName'));
		if ($result === '') {
			$result = Utils::isClickskeksCloud() ? 'Clickskeks' : 'CCM19';
		}
		return $result;
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setAppName(string $name): self
	{
		$this->set('appName', $name);
		return $this;
	}
}
