<?php

namespace App\Controller;

use App\Component\MailSubscriptions;
use App\Component\Menu\Annotation\Menu;
use App\Event\AccountUpdateEvent;
use App\Exception\CsrfTokenException;
use App\Model\User;
use App\Model\Country;
use App\Model\CsrfToken;
use App\Config;
use App\Utils;
use DateTimeZone;
use RuntimeException;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class Account extends UniversalController
{
	/**
	 * @Menu("Account settings", group="Account Settings", icon="fa-cog",
	 *     navigation="meta", order=12)
	 * @Route("/account", name="app_account", methods={"HEAD", "GET"})
	 * @param MailSubscriptions $mailSubscriptions
	 * @return Response
	 */
	public function account(MailSubscriptions $mailSubscriptions): Response
	{
		$user = User::loggedInUser();

		return $this->render('account/index.html.twig', [
			'countries' => Country::all(),
			'csrfToken' => CsrfToken::get(),
			'timestamp' => time(),
			'timeZones' => DateTimeZone::listIdentifiers() ?: ['Europe/Berlin', 'UTC'],
		]);
	}

	/**
	 * @Route("/account/image", name="app_account_userdata_image",
	 *     methods={"HEAD", "GET"})
	 * @param Request $request
	 * @return Response
	 */
	public function userImage(Request $request): Response
	{

		$user = User::loggedInUser();
		$avatar = $user->getUserImage();

		if (!$avatar) {
			return new Response('Not found.', 404);
		}
		if ($avatar->isReadable() == false) {
			return new Response('Reading file not permitted.', 403);
		}

		$response = new BinaryFileResponse($avatar, 200, [], false);
		$response->headers->set('X-Content-Type-Options', '');

		// Caching
		Utils::resetCacheControl($response);
		if ($request->query->has('nocache')) {
			$response->setCache(['max_age' => 0, 'public' => false]);
			$response->headers->addCacheControlDirective('no-cache');
			$response->headers->addCacheControlDirective('must-revalidate');
		} else {
			$response->setCache([
				'max_age' => 450,
				'public' => false,
			]);
			$response->isNotModified($request);
		}

		return $response;
	}

	/**
	 * @Route("/account/password", name="app_account_password",
	 *     methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function changePassword(Request $request, TranslatorInterface $translator): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('app_account');
		}

		$password = trim($request->request->get('current_password'));
		$newPassword = trim($request->request->get('new_password'));
		$newPasswordRetyped = trim($request->request->get('new_password_retyped'));

		$user = User::loggedInUser();

		if ($user->isPasswordLocked()) {
			$this->addFlash('danger', $translator->trans('Password changes are disabled for your account.'));
			return $this->redirectToRoute('app_account');
		}

		if (($user->hasPassword() or $password !== '') and $user->checkPassword($password) == false) {
			$this->addFlash('danger', $translator->trans('The password you\'ve entered is incorrect.'));
		} else if (empty($newPassword)) {
			$this->addFlash('danger', $translator->trans('New password may not be empty.'));
		} else if ($newPassword != $newPasswordRetyped) {
			$this->addFlash('danger', $translator->trans('Retyped password doesn\'t match new password.'));
		} else {
			$event = new AccountUpdateEvent($user, AccountUpdateEvent::CAUSE_SELFSERVICE);
			$event->setPasswordChanged();
			$user->setPassword($newPassword);
			Utils::dispatchEvent(AccountUpdateEvent::NAME_PRE_SAVE, $event);
			$user->save();
			$this->addFlash('success', $translator->trans('Your password has been updated!'));
			Utils::dispatchEvent(AccountUpdateEvent::NAME_POST_SAVE, $event);
		}

		return $this->redirectToRoute('app_account');
	}

	/**
	 * @Route("/account", name="app_account_userdata", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param MailSubscriptions $mailSubscriptions
	 * @return Response
	 */
	public function changeUserdata(Request $request, TranslatorInterface $translator, MailSubscriptions $mailSubscriptions): Response
	{
		$action = $request->request->getAlnum('action');
		if ($action != "saveColorScheme") {
			try {
				CsrfToken::get()->checkRequest($request);
			} catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('app_account');
			}
		}

		$user = User::loggedInUser();
		assert($user instanceof User);

		switch ($action) {
			case "save":
				$email = trim($request->request->get('email'));
				$firstName = trim($request->request->get('firstName'));
				$lastName = trim($request->request->get('lastName'));
				$company = trim($request->request->get('company'));
				$country = trim($request->request->get('country'));

				$user->setEmailAddress($email)
					->setFirstName($firstName)
					->setLastName($lastName)
					->setCompany($company)
					->setCountry(Country::find($country))
					->setPlz((string) $request->request->get("plz",""))
					->setCity((string)$request->request->get("city",""))
					->setStrNr((string)$request->request->get("streetNr","") )
					->setVatId((string)$request->request->get("vatId",""));

				if ($request->request->has('invoiceEmailAddress')) {
					$invoiceEmail = trim((string)$request->request->get('invoiceEmailAddress'));
					$user->setInvoiceEmailAddress($invoiceEmail);
				}
				if ($request->request->has('dunningEmailAddress')) {
					$dunningEmail = trim((string)$request->request->get('dunningEmailAddress'));
					$user->setDunningEmailAddress($dunningEmail);
				}

				$event = new AccountUpdateEvent($user, AccountUpdateEvent::CAUSE_SELFSERVICE);
				$event->setAccountInfoChanged();
				Utils::dispatchEvent(AccountUpdateEvent::NAME_PRE_SAVE, $event);

				$user->save();

				Utils::dispatchEvent(AccountUpdateEvent::NAME_POST_SAVE, $event);

				if ($user->getRole() === User::ROLE_ADMIN && !empty($email)) {
					$config = Config::getInstance();
					$config->set('email', $email);
				}
				$this->addFlash('success', $translator->trans('Your user information has been updated!'));
				break;

			case "uploadAvatar":
				$image = $request->files->get('image');
				if ($image) {
					$this->handleUploadErrors($image, $translator);

					$user->setUserImage($image);
					$user->save();

					$this->addFlash('success', $translator->trans('Your avatar has been updated!'));
				}
				break;

			case "saveColorScheme":
				$scheme = $request->request->getAlnum("preferredColorScheme");

				// Wenn der Admin in einen Kundenaccount wechselt, Farbschema für den Admin speichern, nicht für den Kunden
				$user = (User::originalLoggedInUser() ?: User::loggedInUser());
				$user->setPreferredColorScheme($scheme ?: null);
				$user->save();

				$response = new JsonResponse([ "success"=>true, "preferredColorScheme" => $user->getPreferredColorScheme() ]);

				//For Dev JSON_PRETTY_PRINT
				$response->setEncodingOptions($response->getEncodingOptions() | JSON_PRETTY_PRINT);

				//und raus...
				return $response;
			default:
		}

		return $this->redirectToRoute('app_account');
	}

	/**
	 * @param UploadedFile $image
	 * @param TranslatorInterface $translator
	 * @return void
	 */
	private function handleUploadErrors(UploadedFile $image, TranslatorInterface $translator): void
	{
		if ($image->getError() != UPLOAD_ERR_OK) {
			$this->addFlash(
				'danger',
				$translator->trans('The image could not be uploaded! (Error code %errorcode% %errormessage%)', [
					'%errorcode%' => $image->getError(),
					'%errormessage%' => $image->getErrorMessage()
				])
			);
			throw new RuntimeException();
		}

		if (preg_match('/^image\/.*$/', $image->getMimeType()) == false) {
			$this->addFlash('danger', $translator->trans('The uploaded file is not an image!'));
			throw new RuntimeException();
		}
	}
}
