<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\LegalText;
use App\Component\LocaleService;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class Barrierefreiheit extends DomainDependantController
{
	/**
	 * @Menu("Accessibility", group="Content & Texts", icon="fa-heart", order=100)
	 * @Route("/domains/{_domainId}/barrierefreiheit", name="app_barrierefreiheit")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function accessibility(Request $request, TranslatorInterface $translator): Response
	{
		$locales = LocaleService::enabledLocales();
		$imprintL10n = (array)$request->request->get('accessibility_l10n');
		$imprintUrl = (array)$request->request->get('accessibilityUrl');

		if ($request->getRealMethod() === 'POST') {
			foreach (LegalText::all() as $legalText) {
				$legalText
					->setAccessibilityScheme($request->request->getAlpha('accessibilityScheme'))
					->setShowAccessibilityInFrontendWidget($request->request->getBoolean('showAccessibilityInFrontendWidget'))
				;
			}

			foreach ($locales as $locale) {
				/** @var LegalText|null $legalText */
				$legalText = LegalText::byLocale($locale);

				if ($legalText === null) {
					$legalText = LegalText::create()->setLocale($locale);
				}

				$legalText
					->setAccessibility($imprintL10n[$locale->getName()] ?? '')
					->setAccessibilityUrl($imprintUrl[$locale->getName()] ?? '')
				;
			}

			try {
				CsrfToken::get()->checkRequest($request);
				LegalText::flush();

				$this->addFlash('success', $translator->trans('Localizations have been written successfully.'));
				return $this->redirectToRoute('app_barrierefreiheit');
			}
			catch (ConfigNotWritableException $exception) {
				$this->addFlash('danger', $translator->trans('Localizations could not be written to disk.'));
			}
			catch (CsrfTokenException $exception) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			}
		}

		// Erstbeste Instanz einlesen, um sprachübergreifende Attribute auszulesen
		$firstLocale = reset($locales) ?: LocaleService::find('de_DE') ?? LocaleService::find('en_US');
		$legalText = LegalText::byLocale($firstLocale) ?? LegalText::create()->setLocale($firstLocale);

		return $this->render('barrierefreiheit/index.html.twig', [
			'locales' => $locales,
			'legalTexts' => LegalText::havingLocaleEnabled(),
			'csrfToken' => CsrfToken::get(),
			'accessibilityScheme' => $legalText->getAccessibilityScheme(),
			'showAccessibilityInFrontendWidget' => $legalText->isShowAccessibilityInFrontendWidget(),
		]);
	}
}
