<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Component\LocaleService;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Purpose;
use App\Model\User;
use App\Model\Cookie as CookieModel;
use App\Model\Locale as Locale;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Controller zur Cookie-Verwaltung
 *
 * @Route("/domains/{_domainId}/cookies", name="app_cookies_")
 */
class Cookie extends DomainDependantController
{
	/**
	 * @Menu("Cookies", group="Cookies & Settings", icon="fa-certificate", order=29)
	 * @Route("", name="list")
	 *
	 * @throws \Exception Wenn kein Zugriff auf das Cookie-File erfolgen kann
	 *
	 * @return Response
	 */
	public function listCookies(): Response
	{
		$cookies = CookieModel::all();
		$purposes = Purpose::all();

		return $this->render('cookie/list.html.twig', [
			'cookies' => $cookies,
			'purposes' => $purposes,
			'defaultLocale' => LocaleService::getDefault(),
		]);

	}

	/**
	 * Daten aus dem Array $cookieData in $cookie übertragen
	 */
	private function applyChanges(CookieModel $cookie, array $cookieData, string $defaultLocaleId): CookieModel
	{
		// Script-Blocking
		$codeBlock = explode("\n", trim((string)($cookieData['code_block'] ?? '')));
		$codeBlock = array_filter($codeBlock, function ($line) {
			return trim($line) !== '';
		});
		$codeBlock = array_values(array_map(function ($line) {
			return trim($line);
		}, $codeBlock));
		// Zweck
		$purpose = Purpose::find($cookieData['purpose']) ?? Purpose::find('15c61c3');
		assert($purpose !== null);

		// Allgemeines speichern
		$cookie
			->setName($cookieData['name'] ?? '')
			->setDescription($cookieData['description'][$defaultLocaleId] ?? '')
			->setCodeSnippet($cookieData['code_snippet'] ?? '')
			->setCodeBlock($codeBlock)
			->setCookieVendor($cookieData['cookie_vendor'] ?? '')
			->setPrivacyPolicyUrl($cookieData['privacy_policy_url'][$defaultLocaleId] ?? '')
			->setHttpCookieName($cookieData['http_cookie_name'] ?? '')
			->setLifetime($cookieData['lifetime'][$defaultLocaleId] ?? '')
			->setCollectedDataInfo($cookieData['collected_data_info'][$defaultLocaleId] ?? '')
			->setPurposeOfDataCollection($cookieData['purpose_of_data_collection'][$defaultLocaleId] ?? '')
			->setLegalBasis($cookieData['legal_basis'][$defaultLocaleId] ?? '')
			->setPlaceOfProcessing($cookieData['place_of_processing'][$defaultLocaleId] ?? '')
			->setActive($cookieData['active'] ?? false)
			->setPurpose($purpose)
		;

		// Übersetzungen speichern
		foreach (Locale::enabledLocales() as $locale) {
			$localeId = $locale->getId();
			if (!isset($cookieData['description'][$localeId])) {
				continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
			}
			$cookie
				->setTranslatedDescription($locale, $cookieData['description'][$localeId] ?? '')
				->setTranslatedPrivacyPolicyUrl($locale, $cookieData['privacy_policy_url'][$localeId] ?? '')
				->setTranslatedLifetime($locale, $cookieData['lifetime'][$localeId] ?? '')
				->setTranslatedCollectedDataInfo($locale, $cookieData['collected_data_info'][$localeId] ?? '')
				->setTranslatedPurposeOfDataCollection($locale, $cookieData['purpose_of_data_collection'][$localeId] ?? '')
				->setTranslatedLegalBasis($locale, $cookieData['legal_basis'][$localeId] ?? '')
				->setTranslatedPlaceOfProcessing($locale, $cookieData['place_of_processing'][$localeId] ?? '')
			;
		}

		return $cookie;
	}

	/**
	 * @Route("/new", name="new")
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function createCookie(Request $request, TranslatorInterface $translator): Response
	{
		$csrfToken = CsrfToken::get();
		$cookieData = [];
		$purposes = Purpose::all();
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleId = $defaultLocale->getId();

		if ($request->getMethod() === 'GET') {
			//sanitized Name holen
			$cookieData['name']=CookieModel::getNameFromGet($_GET['cookiename'] ?? '');
		}

		if ($request->getMethod() === 'POST') {
			try {
				$csrfToken->checkRequest($request);
				// Cookie-Daten holen und aufbereiten
				$cookieData = $request->request->all();
				unset($cookieData['action']);

				$formData = $request->request->all();

				// ohne Namen nicht speichern
				if (trim($cookieData['name']) === '') {
					$this->addFlash('warning', 'Ein Cookie kann nicht ohne Namen gespeichert werden');
					return $this->render('cookie/new.html.twig', [
						'formData' => ['name' => '',] + $formData,
						'purposes' => $purposes,
						'locales' => Locale::enabledLocales(),
						'defaultLocale' => $defaultLocale,
						'csrfToken' => $csrfToken,
					]);
				}

				// ohne Verwendungszweck nicht speichern
				if (Purpose::exists($cookieData['purpose']) == false) {
					$this->addFlash('warning', 'Ein Cookie kann nicht ohne Verwendungszweck gespeichert werden');
					return $this->render('cookie/new.html.twig', [
						'formData' => ['purpose' => '',] + $formData,
						'purposes' => $purposes,
						'locales' => Locale::enabledLocales(),
						'defaultLocale' => $defaultLocale,
						'csrfToken' => $csrfToken,
					]);
				}

				// Speichern
				$cookie = $this->applyChanges(CookieModel::create(), $cookieData, $defaultLocaleId);
				$cookie->save();

				$this->addFlash('success', 'Das Cookie "' . $cookieData['name'] . '" wurde erfolgreich erstellt!');

				return $this->redirectToRoute('app_cookies_list');
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			}
		}

		return $this->render('cookie/new.html.twig', [
			'purposes' => $purposes,
			'locales' => Locale::enabledLocales(),
			'defaultLocale' => $defaultLocale,
			'csrfToken' => $csrfToken,
			'formData' => $cookieData ?? [],
		]);
	}

	/**
	 * @Route("/{id}/edit", methods={"HEAD", "GET"}, name="edit")
	 *
	 * @param string $id
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editCookie(string $id, Request $request): Response
	{
		if (CookieModel::exists($id) == false) {
			return $this->redirectToRoute('app_cookies_list');
		}

		$cookie = CookieModel::find($id);

		$purposes = Purpose::all();

		return $this->render('cookie/edit.html.twig', [
			'cookie' => $cookie,
			'purposes' => $purposes,
			'locales' => Locale::enabledLocales(),
			'defaultLocale' => LocaleService::getDefault(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}/edit", methods={"POST"}, name="edit_post")
	 *
	 * @param string $id
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editCookiePost(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleId = $defaultLocale->getId();
		$cookie = CookieModel::find($id);

		try {
			CsrfToken::get()->checkRequest($request);

			// Existiert nicht → zurück zur Übersicht
			if ($cookie === null) {
				return $this->redirectToRoute('app_cookies_list');
			}

			// Cookie-Daten holen und aufbereiten
			$cookieData = $request->request->all();

			CookieModel::beginWrite();
			if (CookieModel::exists($id) == false) {
				return $this->redirectToRoute('app_cookies_list');
			}

			// ohne Verwendungszweck nicht speichern
			if (Purpose::exists($cookieData['purpose']) == false) {
				$this->addFlash('warning', 'Ein Cookie kann nicht ohne Verwendungszweck gespeichert werden');
				return $this->redirectToRoute('app_cookies_edit', ['id' => $cookie->getId()]);
			}


			// Script-Blocking
			$codeBlock = explode("\n", trim((string)($cookieData['code_block'] ?? '')));
			$codeBlock = array_filter($codeBlock, function ($line) {
				return trim($line) !== '';
			});
			$codeBlock = array_values(array_map(function ($line) {
				return trim($line);
			}, $codeBlock));

			// ohne Namen nicht speichern
			if (trim($cookieData['name']) === '') {
				$this->addFlash('warning', 'Ein Cookie kann nicht ohne Namen gespeichert werden');
				return $this->redirectToRoute('app_cookies_edit', ['id' => $cookie->getId()]);
			}

			// Speichern
			$cookie = $this->applyChanges($cookie, $cookieData, $defaultLocaleId);
			$cookie->save();

			$this->addFlash('success', 'Das Cookie "' . $cookieData['name'] . '" wurde erfolgreich bearbeitet!');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		return $this->redirectToRoute('app_cookies_list');
	}

	/**
	 * @Route("/{id}/delete", name="delete")
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function deleteCookie(string $id, Request $request, TranslatorInterface $translator): Response
	{
		CookieModel::beginWrite();
		if (CookieModel::exists($id) == false) {
			return $this->redirectToRoute('app_cookies_list');
		}

		$cookie = CookieModel::find($id);
		$cookieName = $cookie->getName();

		if ($request->getRealMethod() === 'POST') {
			try {
				CsrfToken::get()->checkRequest($request);
				CookieModel::delete($id);
				$this->addFlash('success', 'Das Cookie "' . $cookieName . '" wurde erfolgreich gelöscht!');
				return $this->redirectToRoute('app_cookies_list');
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			}
		}

		return $this->render('cookie/delete.html.twig', [
			'user' => User::loggedInUser(),
			'cookie' => $cookie,
			'csrfToken' => CsrfToken::get(),
		]);
	}
}
