<?php

namespace App\Controller;

use App\Component\Cron\Cron;
use App\Component\Cron\DomainJob;
use App\Component\Cron\Job;
use App\Component\Cron\JobBase;
use App\Config;
use App\Model\UserCronState;
use App\Model\CronState;
use App\Model\Domain;
use App\Model\User;
use App\Utils;
use App\Component\Curl;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;

class CronWorker extends UnrestrictedController
{
	/** @var Cron $cron */
	private $cron;

	/**
	 * @required
	 */
	public function init(Cron $cron)
	{
		$this->cron = $cron;
		// Einige Klassen laden, damit der Autoloader während eines Updates nicht anspringen muss.
		class_exists('\\App\\Component\\Curl', true);
		class_exists('\\Curl\\Curl', true);
		class_exists('\\Symfony\\Component\\Mailer\\EventListener\\MessageListener', true);
		class_exists('\\Symfony\\Component\\Mailer\\EventListener\\MessageLoggerListener', true);
		class_exists('\\Symfony\\Component\\Mailer\\EventListener\\EnvelopeListener', true);
		class_exists('\\Symfony\\Component\\HttpKernel\\EventListener\\ExceptionListener', true);
		class_exists('\\Symfony\\Component\\HttpKernel\\Log\\Logger', true);
	}

	/**
	 * Prüft den API-Key und die Domain-ID bzw. den CronKey im Request
	 *
	 * @param Request $request
	 * @param bool $randomOnCronKey Ob bei fehlendem API-Key, aber gültigem
	 *             Cron-Key eine zufällige Domain gewählt werden soll.
	 * @return bool True, wenn gültige Daten gefunden wurden.
	 */
	private function selectUserAndDomain(Request $request, bool $randomOnCronKey=true): bool
	{
		$apiKey = (string)$request->query->get('apiKey');
		$domainId = (string)$request->query->get('domain');
		// Zu User wechseln, wenn API-Key im Request
		if ($apiKey !== '') {
			$user = User::byApiKey($apiKey);
			if (!$user) {
				return false;
			}
			return $this->cron->setContext($user, $domainId);
		}

		$key = (string)$request->query->get('key');
		if ($randomOnCronKey && hash_equals(Utils::getCronKey(), $key)) {
			$user = User::randomClient();
			User::setApiUser($user->getApiKey());
			Domain::reloadRepository();
			$domain = Domain::random();
			return $this->cron->setContext($user, $domain ? $domain->getId() : null);
		}

		return false;
	}

	/**
	 * @Route("/cron", name="app_cron", methods={"OPTIONS", "GET", "POST"})
	 * @param Request $request
	 * @return Response
	 */
	public function cron(Request $request): Response
	{
		ignore_user_abort(true);
		set_time_limit(300);

		// Zugriffsschlüssel prüfen und User/Domain setzen
		if (!$this->selectUserAndDomain($request, true)) {
			return $this->json([
				'error' => 'Unauthorized: no valid key or domain'
			], 403);
		}

		$success = $this->cron->withCronLock(function () use ($request) {
			// API-Key -> Anfrage für eine Domain
			if ($request->query->has('apiKey')) {
				$this->doCronForActiveDomain($request);
			}
			// Cron-Key -> Alle Kundendomains soweit möglich abarbeiten
			else {
				$start = microtime(true);

				// Zuerst System-Jobs, User-Jobs und Jobs einer zufälligen Domain abarbeiten
				$this->doCronForActiveDomain($request, 1.5, false);

				// Dann alle Kunden und Domains durchloopen, solange es geht
				$users = (Utils::isBaseEdition()) ? [current(User::byRole(User::ROLE_ADMIN))] : User::getClients();
				shuffle($users);
				foreach ($users as $user) {
					if (!$this->cron->setContext($user, null)) {
						continue;
					}
					$domains = Domain::all();
					shuffle($domains);
					foreach ($domains as $domain) {
						if (!$this->cron->setContext($user, $domain->getId())) {
							continue;
						}
						$this->doCronForActiveDomain($request, 0.4, true);

						// Wenn schon 10 Sekunden lang gearbeitet wurde, keinen neuen Job mehr
						// beginnen.
						if (microtime(true) - $start > 9.99) {
							break 2;
						}
					}
				}
			}
		});

		if (!$success) {
			if ($request->getMethod() === 'GET') {
				$result = $this->json([
					'result' => 'wouldblock',
				], 200);
				$result->headers->set('Cache-Control', 'no-cache, must-revalidate');
			} else {
				$result = new Response('', 204);
				$result->headers->set('X-Status', 'wouldblock');
			}
			return $result;
		}

		$this->runDeferredJobs($request);

		if ($request->getMethod() === 'GET') {
			$result = $this->json([
				'result' => 'success'
			], 200);
			$result->headers->set('Cache-Control', 'no-cache, must-revalidate');
		} else {
			$result = new Response('', 204);
			$result->headers->set('X-Status', 'done');
		}
		return $result;
	}

	/**
	 * @param Request $request
	 * @param float $timelimit
	 * @param bool $onlyDomainJobs
	 * @return void
	 */
	private function doCronForActiveDomain(Request $request, float $timelimit=0.2, bool $onlyDomainJobs=false)
	{
		$config = Config::getInstance();
		$activeDomain = Domain::activeDomain();

		// Immer auf frischen Daten arbeiten; Jobs in Warteschlange berücksichtigen
		$config->reload();
		UserCronState::reload();
		if ($activeDomain) {
			CronState::reload();
		}
		$start = microtime(true);

		/** @var JobBase[] $jobs */
		$jobs = $this->cron->dueJobs();

		if ($onlyDomainJobs) {
			$jobs = array_filter($jobs, static function ($job) { return $job instanceof DomainJob; });
		}

		// Aufgaben in zufälliger Reihenfolge beginnen
		shuffle($jobs);

		// Fällige Jobs abarbeiten
		foreach ($jobs as $job) {
			try {
				$job->run($request);
			} finally {
				// Konfiguration schreiben, damit sie im nächsten Cronjob direkt verfügbar ist
				$config->flush();
				UserCronState::flush();
				if ($activeDomain) {
					CronState::flush();
				}
			}

			// Wenn schon 200ms lang gearbeitet wurde, keinen neuen Job mehr
			// beginnen.
			if (microtime(true) - $start > $timelimit) {
				break;
			}
		}
	}

	/**
	 * @Route("/cron/{jobName}", name="app_cron_job", methods={"OPTIONS", "GET", "POST"})
	 * @param Request $request
	 * @param string $jobName
	 * @return Response
	 */
	public function cronOne(Request $request, string $jobName): Response
	{
		ignore_user_abort(true);
		set_time_limit(300);
		$config = Config::getInstance();

		// Sicherstellen, dass immer nur ein Cronjob zeitgleich abgearbeitet wird
		$isCluster = (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB and !empty($_ENV['APP_CLUSTER_PEERS']));
		if ($isCluster) {
			// Bei MongoDB-Cluster: über DB synchronisieren
			// Dazu bis zu 50x versuchen den Lock zu bekommen (max 2s lang), und ansonsten abbrechen
			$blocked = true;
			$lockf = null;
			for ($i = 0; ($i < 50); $i++) {
				$blocked = !$config->setCronjobRunning();
				if ($blocked) {
					usleep(random_int(30000, 40000));
					$config->reload();
				} else {
					break;
				}
			}
		} else {
			$lockf = @fopen(Utils::getBundledVarDir().'/.cron.lock', 'wb');
			// Dazu bis zu 1000x versuchen den Lock zu bekommen (max 2s lang), und ansonsten abbrechen
			$blocked = true;
			for ($i = 0; ($i < 1000); $i++) {
				flock($lockf, LOCK_EX|LOCK_NB, $blocked);
				if ($blocked) {
					usleep(random_int(400, 2000));
				} else {
					break;
				}
			}
		}
		if ($blocked) {
			if ($request->getMethod() === 'GET') {
				$result = $this->json([
					'result' => 'wouldblock',
				], 200);
				$result->headers->set('Cache-Control', 'no-cache, must-revalidate');
			} else {
				$result = new Response('', 204);
				$result->headers->set('X-Status', 'wouldblock');
			}
			if ($request->headers->get('Origin') === 'https://update.ccm19.de') {
				$result->headers->set('Access-Control-Allow-Origin', $request->headers->get('Origin'));
			}
			return $result;
		}

		// Immer auf frischen Daten arbeiten; Jobs in Warteschlange berücksichtigen
		$config = Config::getInstance();
		$config->reload();

		try {
			/** @var JobBase[] $jobs */
			$job = $this->cron->jobByName($jobName);

			// Zugriffsschlüssel prüfen und User/Domain setzen
			if (!$this->selectUserAndDomain($request, ($job instanceof Job))) {
				return $this->json([
					'error' => 'Unauthorized: no valid key or domain or improper job for authentication type',
				], 403);
			}

			if ($job) {
				$job->run($request);
			} else {
				return $this->json(['error'=>'404 Not found', 'description'=>'No cronjob with this name'], 404);
			}
		}
		finally {
			// Konfiguration schreiben, damit sie im nächsten Cronjob direkt verfügbar ist
			$config->flush();

			// Nächsten Cronjob freigeben
			if ($isCluster) {
				$config->unsetCronjobRunning();
			} elseif ($lockf) {
				flock($lockf, LOCK_UN);
				fclose($lockf);
			}
		}

		$this->runDeferredJobs($request);

		if ($request->getMethod() === 'GET') {
			$result = $this->json([
				'result' => 'success',
			], 200);
			$result->headers->set('Cache-Control', 'no-cache, must-revalidate');
		} else {
			$result = new Response('', 204);
			$result->headers->set('X-Status', 'done');
		}
		if ($request->headers->get('Origin') === 'https://update.ccm19.de') {
			$result->headers->set('Access-Control-Allow-Origin', $request->headers->get('Origin'));
		}
		return $result;
	}

	/**
	 * @param Request $request
	 */
	private function runDeferredJobs(Request $request): void
	{
		foreach ($this->cron->deferredJobs() as $jobData) {
			list($job, $apiKey, $domainId) = $jobData;
			$url = $this->generateUrl('app_cron_job', array_filter([
				'jobName' => $job->getName(),
				'apiKey' => $apiKey,
				'domain' => $domainId,
				'key' => $request->query->get('key')
			], static function ($v) { return $v !== null; }), UrlGeneratorInterface::ABSOLUTE_URL);

			$f = @fopen('/dev/null', 'ab');
			if (!$f) {
				$f = @fopen('php://temp', 'wb');
			}
			$curl = new Curl();
			$curl->setOpt(CURLOPT_CONNECTTIMEOUT, 4);
			$curl->setOpt(CURLOPT_TIMEOUT, 4);
			$curl->setOpt(CURLOPT_RETURNTRANSFER, false);
			$curl->setOpt(CURLOPT_FILE, $f);
			$curl->post($url, []);
			fclose($f);
		}
	}
}
