<?php

namespace App\Controller;


use App\Component\LocaleService;
use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Component\ThumbnailExtractor\ThumbnailExtractorService;
use App\Component\Update;
use App\Config;
use App\Exception\CsrfTokenException;
use App\Model\Country;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\Embedding as EmbeddingModel;
use App\Model\LegalText;
use App\Model\LocaleBase;
use App\Model\Theme;
use App\Model\User;
use App\Model\Cookie;
use App\Model\Purpose;
use App\Component\HttpFoundation\JavascriptResponse;
use App\Utils;
use App\Model\StatisticsData;
use DateTimeZone;
use RuntimeException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Validator\Constraints\Length;
use Symfony\Component\Validator\Constraints\Url;
use Symfony\Component\Validator\ConstraintViolation;
use Symfony\Component\Validator\Validator\ValidatorInterface;
use Symfony\Contracts\Translation\TranslatorInterface;
use Throwable;

/**
 * @Route("/domains/{_domainId}/dashboard", name="app_dashboard")
 */
class Dashboard extends DomainDependantController
{
    /**
     * @Menu("Dashboard & Usage", group="Cookies & Settings", icon="glyphicon-home", order=0)
     * @Route("", name="")
     * @param Request $request
     * @param TranslatorInterface $translator
     * @param ValidatorInterface $validator
     * @param LoggerInterface $logger
     * @return Response
     */
    public function index(Request $request, TranslatorInterface $translator, ValidatorInterface $validator, LoggerInterface $logger): Response
    {
        // Verwende Daten des Hauptbenutzers (z. B. API-Schlüssel für CCM19-Einbindung)
        $user = User::loggedInUser()->getAccountOwner();
        $domain = Domain::activeDomain();
        $session = $request->getSession();
        $locales = LocaleService::enabledLocales();

        //Sprachdaten für die Auswahl des Schnipsels
        $localeNames = [];
        foreach ($locales as $locale) {
            /** @var LocaleBase $locale */
            $localeName = $locale->getName();
            $language = $locale->getLanguage();
            $localeNames[$localeName] = ($language ? $language->getDisplayName($request->getLocale()) : $localeName);
        }

        $cookieCounts = [];
        $cookieCountTotal = 0;
        foreach (Purpose::all() as $purpose) {
            $count = count(Cookie::byPurpose($purpose));
            $cookieCounts[$purpose->getName()] = $count;
            $cookieCountTotal += $count;
        }

        // Auf Update prüfen
        $this->checkForUpdateGlobal();

        //nur dann, wenn mehrere Domains verwaltet werden... Einzelplatz Version macht keinen Sinn da kein Domname vorhanden ist
        if (!empty($domain->getName())) {
            $firstLogin = $domain->isOnBoarding();
        } else {
            $firstLogin = true;
        }


        if (Utils::isBaseEdition() && empty($domain->getName()) && empty(Embedding::all()) && $domain->isOnBoarding() == false) {
            return $this->redirectToRoute('app_onboarding', [], 303);
        }

        $embedCheck = $domain->get("embeddingCheck");
        $isBothChecked = false;
        if (!empty($embedCheck)) {
            if ($embedCheck['position'] == 1 && $embedCheck['loaded'] == 1) {
                $isBothChecked = true;
            }
        }

        $activeWidgetsCount = Domain::frontendWidgetCountForAllDomains();
        $domainMaxCount = $user->getMaxDomainCount();

        $allowFrontendWidget = !Utils::isExtendedEdition() || $domainMaxCount < 0 || $activeWidgetsCount < $domainMaxCount || $domain->isFrontendWidgetEnabled();

        if (!$allowFrontendWidget) {
            $this->addFlash('warning', $translator->trans('You cannot activate the frontend widget at the moment due to insufficient domains in your plan.'));
        }

        if ($domain->isFrontendWidgetEnabled()) {
            $domain
                ->setFrontendLegalEnabled(true)
                ->flush();
        }

        $firstLocale = reset($locales) ?: LocaleService::find('de_DE') ?? LocaleService::find('en_US');
        $legalText = LegalText::byLocale($firstLocale) ?? LegalText::create()->setLocale($firstLocale);

        $whitelabelAvailableCount = DomainManager::whitelabelAvailableCountForUser($user);
        $whitelabelAvailable = $whitelabelAvailableCount != 0 || $domain->hasWhitelabel();

        $formData = [
            'country' => '',
            'domainName' => '',
            'whitelabel' => '',
            'domainHtPassWd' => '',
            'whitelabel_verfuegbar' => $whitelabelAvailable,
            'whitelabel_verfuegbar_count' => $whitelabelAvailableCount < 0 ? $translator->trans('unlimited') : $whitelabelAvailableCount,
        ];

        $range = 30;
        if(Utils::isCCMCloud() && User::loggedInUser()->getMaxCallCount() <= 5000 && User::loggedInUser()->getMaxCallCount() !== -1) {
            $range = 8;
        }

        //First Login
        if ($firstLogin == false) {
            return $this->redirectToRoute('app_onboarding', [], 303);
        } else {
            return $this->render('dashboard/index.html.twig', [
                'embedCodeSnippetBaseParams' => ['apiKey' => $user->getApiKey(), 'domain' => $domain->getId()],
                'themes' => Theme::allOrderedByTypeAndName(),
                'localeNames' => $localeNames,
                'cookieCounts' => $cookieCounts,
                'cookieCountTotal' => $cookieCountTotal,
                'update_notifier' => $session->get('update_notifier', null),
                'embeddingCheck' => $domain->get("embeddingCheck"),
                'domName' => $domain->getName(),
                'csrfToken' => CsrfToken::get(),
                'isBothChecked' => $isBothChecked ?? false,
                'allowFrontendWidget' => $allowFrontendWidget,
                'frontendLegalEnabled' => $domain->isFrontendLegalEnabled(),
                'purposes' => Purpose::all(),
                'embeddings' => EmbeddingModel::all(),
                'legaltext' => $legalText,
                'domain' => $domain,
                'countries' => Country::all(),
                'timeZones' => DateTimeZone::listIdentifiers() ?: ['Europe/Berlin', 'UTC'],
                'formData' => $formData,
                'range' => $range,
            ]);
        }
    }

    /**
     *
     * @Route("/check_embedding", name="_check_embedding", methods={"POST"} )
     *
     */
    public function checkembeddingCode(Request $request, TranslatorInterface $translator, ValidatorInterface $validator, LoggerInterface $logger)
    {
        $user = User::loggedInUser()->getAccountOwner();
        $domain = Domain::activeDomain();
        //Extended...
        if (!empty($request->query->get("checkembedding")) && !empty($domain->getName())) {
            $domain->set("embeddingCheck", $domain->checkEmbedding());
            $domain->save();
        }


        if ($request->getMethod() === 'POST' && empty($request->query->get("checkembedding"))) {
            try {
                CsrfToken::get()->checkRequest($request);

                $domName = $request->request->get("domName");
                $constraints = [
                    new Url(['relativeProtocol' => true]), // Allow relative protocol
                    new Length(['min' => 12, 'allowEmptyString' => false]), // Min len
                ];
                $violations = $validator->validate($domName, $constraints);

                if ($violations->count()) {
                    foreach ($violations as $violation) {
                        if ($violation instanceof ConstraintViolation) {
                            // Könnte man später mal nutzen: $message = $violation->getMessage();
                            $logger->logError('Dom Setting Error', "Domain: '" . $domName . "' -> " . $violation->getMessage());
                            $this->addFlash('danger', $translator->trans('Please enter a valid URL or domain.'));
                            return $this->redirectToRoute('app_dashboard');
                        }
                    }
                } else {
                    $domain->setName($domName);
                    $domain->save();
                    $this->addFlash('success', $translator->trans('Domain name was successfully saved.'));
                    return $this->redirectToRoute('app_dashboard');
                }

            } catch (CsrfTokenException $exception) {
                $this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
            }
            return $this->redirectToRoute('app_dashboard');
        }
    }

    /**
     * @Route("/activate", methods={"POST"}, name="_activation")
     * @param Request $request
     * @param TranslatorInterface $translator
     * @param ThumbnailExtractorService $thumbnails
     * @return Response
     */
    public function saveFrontendSettings(Request $request, TranslatorInterface $translator, ThumbnailExtractorService $thumbnails): Response
    {

        /** @var User $user */
        $user = User::loggedInUser()->getAccountOwner();
        /** @var Domain $domain */
        $domain = Domain::activeDomain();
        Domain::beginWrite();

        $activeWidgetsCount = Domain::frontendWidgetCountForAllDomains();
        $domainMaxCount = $user->getMaxDomainCount();
        $allowFrontendWidget = !Utils::isExtendedEdition() || $domainMaxCount < 0 || $activeWidgetsCount < $domainMaxCount || $domain->isFrontendWidgetEnabled();

        $allowFrontendLegal = $domain->isFrontendLegalEnabled();
        if (!$allowFrontendLegal) {
            $allowFrontendLegal = (bool)$request->request->get('enableFrontendLegal');
        }

        $domain
            ->setFrontendWidgetEnabled($allowFrontendWidget && (bool)$request->request->get('enableFrontendWidget'))
            ->setFrontendLegalEnabled($allowFrontendLegal)
            ->save();

        $this->addFlash('success', $translator->trans('Settings have been saved.'));
        return $this->redirectToRoute('app_dashboard', [], 303);
    }

    /**
     * @Route("/checkembedding", name="_checkembedding")
     * @return JsonResponse
     */
    public function checkEmbedding(): JsonResponse
    {
        return $this->json([
            'checked' => true
        ]);
    }

    /**
     * @return false
     */
    private function checkForUpdateGlobal()
    {
        $config = Config::getInstance();
        $session = Utils::getSession();

        //AutoUpdate - dann nix checken
        if ($config->get("enableAutoUpdate")) {
            return false;
        }

        $updater = new Update();
        $config = Config::getInstance();
        try {
            $updateInfo = $updater->checkForUpdate();
            $session->set('update_notifier', ($updateInfo and Utils::getVersionId() != $updateInfo['versionId']));
        } catch (Throwable $e) {
            $session->set('update_notifier', false);
        }
    }

    /**
     * @Route("/chart.js", name="_chart_js")
     * @param Request $request
     * @param TranslatorInterface $translator
     * @return Response
     * @phan-suppress PhanTypeMismatchReturnSuperType
     * @throws \Exception
     */
    public function chartJs(Request $request, TranslatorInterface $translator): Response
    {
        $response = new JavascriptResponse();
        Utils::resetCacheControl($response);
        $response->setCache([
            'max_age' => 90,
            'public' => false
        ]);

        //Bei jedem Aufruf die aktuellen Daten aus der Standard Statistik übernehmen... damit es immer aktuell ist...
        StatisticsData::createFullStats(false);

        $dateForm = "%Y-%m-%d";
        $statDat = array_filter($request->query->all(), static function ($key) {
            return $key && $key[0] !== '_' && $key !== 'chart';
        }, ARRAY_FILTER_USE_KEY);

        // Statistik kann von Plugins überarbeitet werden
        $parameters = Utils::filterWithEvent(
            'ccm19.dashboard.chart.values', [
                "dateForm" => $dateForm,
                "statdat" => $statDat,
            ]
        );

        // Session schließen
        $request->getSession()->save();

        return $this->render('dashboard/chart.js.twig', $parameters, $response);
    }

    /**
     * @Route("/basic_data.json", name="_basic_data_json", methods={"HEAD", "GET"})
     * @param Request $request
     * @param TranslatorInterface $translator
     * @return JSONResponse
     */
    public function showJsonData(Request $request, TranslatorInterface $translator): JSONResponse
    {
        $daterange = (string)$request->query->get('daterange');

        $range = 30;
		$maxCallCount = User::loggedInUser()->getMaxCallCount();

        if(Utils::isCCMCloud() && $maxCallCount != -1 && $maxCallCount <= 5000) {
            $range = 8;
        }

        //Datum immer setzen
        $selDate = [
            'from' => time() - (60 * 60 * 24 *  $range),
            'to' => time(),
        ];
        if ($daterange) {
            $daDates = explode(" - ", $daterange);
            if ($daDates) {
                $selDate['from'] = (int)strtotime($daDates[0]);
                $selDate['to'] = (int)(strtotime($daDates[1]));
            }
        }

        //nur die Daten ab Mitte Februar - vorher nur crap...
        if ($selDate['from'] < 1613857795) {
            $selDate['from'] = 1613857795;
        }

        $statistics = array();
        $count = 0;

        $domain = Domain::activeDomain();
        assert($domain !== null);

        // Einträge für Datums-Range ziehen
        $fromDay = strftime('%Y-%m-%d', $selDate['from']);
        $toDay = strftime('%Y-%m-%d', $selDate['to']);
        $allData = array_filter(StatisticsData::all(), static function ($dateKey) use ($fromDay, $toDay) {
            return ($dateKey >= $fromDay and $dateKey <= $toDay);
        }, ARRAY_FILTER_USE_KEY);
        // Nach Datum sortieren
        ksort($allData);

        if(empty($allData))
        {
            //NUllwert setzen - falls null ist
            $dateKey = strftime('%Y-%m-%d', $selDate['from']);
            $statistics["opened"][$dateKey] = 0;
            $statistics["acceptAll"][$dateKey] = 0;
            $statistics["leave"][$dateKey] = 0;
            $statistics["decline"][$dateKey] = 0;
            $statistics["Calls"][$dateKey] = 0;
        }


        try {
            foreach ($allData as $date => $stat) {
                $statistics["x"][$date] = $date;
            }
        } catch (RuntimeException $exception) {
        }

        try {
            foreach ($allData as $dateKey => $stat) {
                //print_r($stat);
                $statistics["opened"][$dateKey] = 0;
                $statistics["acceptAll"][$dateKey] = 0;
                $statistics["leave"][$dateKey] = 0;
                $statistics["decline"][$dateKey] = 0;
                $statistics["Calls"][$dateKey] = 0;

                $count += $stat->getViewCount();
                $details = $stat->getDetails();

                if (!empty($details)) {
                    foreach ($details as $themeID => $themeData) {
                        $statistics[$themeID][$dateKey] = $themeData['all'] ?? 0;
                        //print_r($statistics);
                        //exit();
                    }
                    if (!empty($details['opened']['all']) && !empty($details['acceptAll']['all'])) {
                        if (empty($details['decline']['all'])) {
                            $details['decline']['all'] = 0;
                            $statistics["decline"][$dateKey] = 0;
                        }
                        $decline = $details['opened']['all'] - $details['decline']['all'] - $details['acceptAll']['all'];
                        //print_r($decline."\n");
                        //oha - da wurde vermutlich durch Bots Murks produziert... korrigieren.
                        if ($decline < 0) {
                            $decline = 0;
                            //$statistics['opened'][$dateKey] =  $statistics['opened'][$dateKey] + $decline;

                        }
                        $statistics["leave"][$dateKey] = $decline;
                    } else {
                        $statistics["acceptAll"][$dateKey] = 0;
                        $statistics["leave"][$dateKey] = 0;
                        $statistics["decline"][$dateKey] = 0;
                        //$statistics["opened"][] = 0;
                    }
                }
                //print_r($statistics);
                $statistics["Calls"][$dateKey] = $stat->getCallCount();
            }
        } catch (RuntimeException $exception) {
        }

        // Datums-Keys entfernen
        $statistics = array_map(function ($subarray) {
            if (is_array($subarray)) {
                return array_values($subarray);
            } else {
                return $subarray;
            }
        }, $statistics);

        $statistics = [
            'x' => $statistics['x'] ?? [],
            // Die Reihenfolge wird in der Diagramm-Legende beibehalten
            $translator->trans('Calls') => $statistics['Calls'] ?? [],
            $translator->trans('Banner views') => $statistics['opened'] ?? [],
            $translator->trans('acceptAll') => $statistics['acceptAll'] ?? [],
            $translator->trans('decline') => $statistics['decline'] ?? [],
            $translator->trans('leave') => $statistics['leave'] ?? [],
        ];

        // Session schließen
        $request->getSession()->save();

        $response = new JsonResponse($statistics);
        //For Dev JSON_PRETTY_PRINT
        $response->setEncodingOptions($response->getEncodingOptions() | JSON_PRETTY_PRINT);

        //und raus...
        return $response;
    }
}
