<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\LegalText;
use App\Component\LocaleService;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class Datenschutz extends DomainDependantController
{
	/**
	 * @Menu("Privacy protection", group="Content & Texts", icon="fa-eye-slash", order=90)
	 * @Route("/domains/{_domainId}/datenschutz", name="app_datenschutz")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function privacyPolicy(Request $request, TranslatorInterface $translator): Response
	{
		$locales = LocaleService::enabledLocales();
		$privacyPolicyL10n = (array)$request->request->get('privacy_policy_l10n');
		$privacyUrl = (array)$request->request->get('privacyUrl');

		if ($request->getRealMethod() === 'POST') {
			foreach (LegalText::all() as $legalText) {
				$legalText
					->setPrivacyScheme($request->request->getAlpha('privacyScheme'))
					->setShowWidgetOnPrivacyPage($request->request->getBoolean('showWidgetOnPrivacyPage'))
					->settargetBlankPrivacy($request->request->getBoolean('targetBlankPrivacy'))
				;
			}

			foreach ($locales as $locale) {
				/** @var LegalText|null $legalText */
				$legalText = LegalText::byLocale($locale);

				if ($legalText === null) {
					$legalText = LegalText::create()->setLocale($locale);
				}

				$legalText
					->setPrivacyPolicy($privacyPolicyL10n[$locale->getName()] ?? '')
					->setPrivacyUrl($privacyUrl[$locale->getName()] ?? '')
				;
			}

			try {
				CsrfToken::get()->checkRequest($request);
				LegalText::flush();

				$this->addFlash('success', $translator->trans('Localizations have been written successfully.'));
				return $this->redirectToRoute('app_datenschutz');
			}
			catch (ConfigNotWritableException $exception) {
				$this->addFlash('danger', $translator->trans('Localizations could not be written to disk.'));
			}
			catch (CsrfTokenException $exception) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			}
		}

		// Erstbeste Instanz einlesen, um sprachübergreifende Attribute auszulesen
		$firstLocale = reset($locales) ?: LocaleService::find('de_DE') ?? LocaleService::find('en_US');
		$legalText = LegalText::byLocale($firstLocale) ?? LegalText::create()->setLocale($firstLocale);

		return $this->render('datenschutz/index.html.twig', [
			'locales' => $locales,
			'legalTexts' => LegalText::havingLocaleEnabled(),
			'csrfToken' => CsrfToken::get(),
			'privacyScheme' =>$legalText->getPrivacyScheme(),
			'targetBlankPrivacy' => $legalText->istargetBlankPrivacy() ?? false,
		]);
	}
}
