<?php

namespace App\Controller;

use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Config;
use App\Event\DomainInitializedEvent;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Country;
use App\Model\Domain;
use App\Model\Model;
use App\Model\Statistic;
use App\Model\Theme as ThemeModel;
use App\Model\User;
use App\Model\FoundCookie;
use App\Model\FoundScript;
use App\Model\FoundStorage;
use App\Model\FoundObject;
use App\Model\Index\DomainIndex;
use App\Model\Regulations\Regulation;
use App\Utils;
use DateTimeZone;
use Exception;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use DateTime;

class DomainManager extends AbstractController
{
	/**
	 * @Menu("Manage domains", group="Settings", icon="glyphicon-list", routeGroup="app_domains_", order=10)
	 * @Route("/domains", name="app_domains_index", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function domainIndex(Request $request, TranslatorInterface $translator,LoggerInterface $logger): Response
	{
		/** @var Domain[] $domains */
		$domains = Domain::all();
		$config = Config::getInstance();
		$user = User::loggedInUser();
		assert($user instanceof User);

		//Falls noch keine Domains da sind... und der User nicht abgelehnt hat...
		if (empty($domains) && $user->isOnBoarding() == true && Utils::mayUserAccessRoute($user, 'app_onboarding_extended', 'GET')) {
			return $this->redirectToRoute('app_onboarding_extended');
		}

		// Sortiere Domains dem Namen nach
		usort($domains, function (Domain $a, Domain $b) {
			return strnatcasecmp($a->getName(), $b->getName());
		});

		$pagination = new Pagination($domains);
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		// Remember the subuser to switch back to it after updating the domain info
		$subuser = $user->isSubUser() ? $user : null;
		$user = $user->getAccountOwner();

		// Switch to account owner context
		if ($subuser) {
			User::setApiUser($user->getApiKey());
			Model::closeUserModels();
			Domain::switchToUser($user);
		}

		$user->updateDomainInfo();
		$user->save();

		// Switch back to subuser context
		if ($subuser) {
			User::setApiUser($subuser->getApiKey());
			Model::closeUserModels();
			Domain::switchToUser($subuser);
		}

		$domainLimitString = $user->getMaxDomainCount() < 0
			? $translator->trans('unlimited')
			: (string)$user->getMaxDomainCount();

		$maxWLCount = $user->getMaxWLCount();
		$maxCallCount = $user->getMaxCallCount();
		$actualCallCount = $user->getActualCallCount();

		$maxWLCount = $maxWLCount < 0 ? $translator->trans('unlimited') : $maxWLCount;
		$maxCallCount = $maxCallCount < 0 ? $translator->trans('unlimited') : $maxCallCount;

		//Speziell Clickskeks - check ob sich was getan hat in der Abrechnungsliste
		if(Utils::isClickskeksCloud())
		{
			$ckDoms = $user->getDomCKList();
			$updateCKList = false;
			if(!empty($ckDoms))
			{
				foreach($ckDoms as $domKey => $domTarif)
				{
					//Die Domain gibts nicht mehr
					if(Domain::find($domKey) === null)
					{
						//unset($ckDoms[$domKey]);
						$updateCKList = true;
						$logger->logError("Domain","Domain nicht mehr vorhanden - wurde anscheinend gelöscht - DomainId: ".$domKey." User: ".$user->getId());
					}
				}
			}
			//Neue Liste speichern, wenn sich was getan hat
			if($updateCKList)
			{
				$user->setDomCKList($ckDoms);
				$user->save();
			}
		}

		if ($user->isAutoAddDomains()) {
			$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
			$baseUrl = $scheme.$request->getHttpHost();
			$embedCodeSnippetAutoDomain = '<script src="'.htmlspecialchars($baseUrl.$this->generateUrl(
				'app_external_main_javascript_file',
				[
					'apiKey' => $user->getApiKey(),
				]
			)).'" referrerpolicy="origin"></script>';
		}
		else {
			$embedCodeSnippetAutoDomain = '';
		}

		//Check für PermScan Anzeige, damit der Menüpunkt korrekt an oder aus ist
		foreach ($domains as $domain) {
			//Den Permscan Menüpunkt deaktivieren, wenn Crawl aktiv ist
			if($domain->getDoCrawl() && $domain->isActivatePermaScan())
			{
				$domain->setActivatePermaScan(false);
				$domain->save();

			}
			if(!$domain->getDoCrawl() && !$domain->isActivatePermaScan())
			{
				$domain->setActivatePermaScan(true);
				$domain->save();
			}
		}

		return $this->render('domain/index.html.twig', [
			'pagination' => $pagination,
			'domainCount' => $user->getActualDomainCount(),
			'domainLimit' => $domainLimitString,
			'whitelabelUsedCount' => $user->getActualWLCount(),
			'maxWLCount' => $maxWLCount,
			'maxCallCount' => $maxCallCount,
			'crawlBudget' =>  $user->getAccountFeatureMaxCrawlBudget(),
			'crawlBudgetUsed' =>  $user->getAccountFeatureMaxCrawlBudget() - Domain::getRestCrawlBudget(),
			'slidetariff' => $user->get("SlidingTarif"),
			'actualCallCount' => $actualCallCount,
			'embedCodeSnippetAutoDomain' => $embedCodeSnippetAutoDomain,
			'regulations' => Regulation::all(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/domains", name="app_domains_bulk_action", methods={"POST"})
	 */
	public function domainBulkAction(Request $request, TranslatorInterface $translator,LoggerInterface $logger): RedirectResponse
	{
		$user = User::loggedInUser()->getAccountOwner();

		$activeDomain = Domain::activeDomain();
		$action = $request->request->get('action');
		$targetIds = $request->request->get('targets', '');
		if (!is_array($targetIds)) {
			if ($targetIds !== '*') {
				$targetIds = explode(',', $targetIds);
			}
		}
		/** @var Domain[] $targets */
		$targets = ($targetIds === '*') ? Domain::all() : array_filter(array_map(function (string $id) {
			return Domain::find(trim($id));
		}, $targetIds));

		try {
			CsrfToken::get()->checkRequest($request);
		}
		catch (CsrfTokenException $e) {
			$error = $translator->trans('CSRF token is missing or invalid. Please try again.');
			$this->addFlash('danger', $error);
			return $this->redirectToRoute('app_page_check', [], 303);
		}

		switch ($action) {
			case 'clear-page-check':
				foreach ($targets as $domain) {
					Domain::select($domain->getId());
					FoundCookie::reloadRepository();
					FoundScript::reloadRepository();
					FoundStorage::reloadRepository();
					FoundObject::reloadRepository();
					FoundCookie::clear();
					FoundScript::clear();
					FoundStorage::clear();
					FoundObject::clear();
					gc_collect_cycles();
				}

				$this->addFlash('success', $translator->trans('All page check entries were successfully cleared.'));
				break;

			case 'activate':
				//Alle durchlaufen und aktivieren.
				foreach ($targets as $domain) {
					$domain->setFrontendWidgetEnabled(true);
					$domain->setFrontendLegalEnabled(true);
					$domain->save();
				}

				$this->addFlash('success', $translator->trans('All domains are now activated.'));
				break;

			case 'delete':
				$count = 0;
				foreach ($targets as $domain) {
					$logger->logWarning('Domain ', "DomainDeleted Bulk: '" . $domain->getName() . "' User -> " . $user->getId());
					Domain::delete($domain->getId(), false);
					$count++;
				}
				Domain::flush();

				// Benutzer aktualisieren
				$user->updateDomainInfo();
				$user->save();

				if ($count > 0) {
					$this->addFlash('success', $translator->trans('The selected :n domain(s) were successfully deleted.', [':n' => $count]));
				}
				break;
			default:
				// unbekannte Aktion... ignorieren
		}

		// aktive Domain zurücksetzen
		if ($activeDomain) {
			Domain::select($activeDomain->getId());
		}
		// Index neu generieren
		DomainIndex::rebuild();

		return $this->redirectToRoute('app_domains_index', [], 303);
	}

	/**
	 * @Route("/domains/bulk_delete", name="app_domains_bulk_delete", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @return Response
	 */
	public function domainBulkDeleteForm(Request $request): Response
	{
		if(Utils::isClickskeksCloud())
		{
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		/** @var Domain[] $domains */
		$domains = Domain::all();

		// Sortiere Domains nach Aktiv und dann dem Namen nach
		usort($domains, function (Domain $a, Domain $b) {
			return ($a->isFrontendWidgetEnabled() <=> $b->isFrontendWidgetEnabled()) ?: strnatcasecmp($a->getName(), $b->getName());
		});

		return $this->render('domain/bulk_delete.html.twig', [
			'domains' => $domains,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/domains/bulk_widget", name="app_domains_bulk_widget", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @return Response
	 */
	public function domainBulkWidgetActivateForm(Request $request): Response
	{
		/** @var Domain[] $domains */
		$domains = Domain::all();

		// Sortiere Domains nach Aktiv und dann dem Namen nach
		usort($domains, function (Domain $a, Domain $b) {
			return ($a->isFrontendWidgetEnabled() <=> $b->isFrontendWidgetEnabled()) ?: strnatcasecmp($a->getName(), $b->getName());
		});

		return $this->render('domain/activate_widget.html.twig', [
			'domains' => $domains,
			'csrfToken' => CsrfToken::get(),
		]);
	}


	/**
	 * @param User $user
	 * @return int
	 */
	public static function whitelabelAvailableCountForUser(User $user): int
	{
		$maxWhitelabelCount = $user->getMaxWLCount();
		return $maxWhitelabelCount >= 0
			? max(0, $maxWhitelabelCount - Domain::whitelabelCountForAllDomains())
			: -1;
	}

	/**
	 * @Route("/domains/list.json", name="app_domains_list_ajax")
	 * @param Request $request
	 *
	 * @return JsonResponse
	 * @throws \Exception
	 */
	public function ajaxDomainSearch(Request $request): JsonResponse
	{
		$activeDomain = Domain::activeDomain();

		$domains = array_values(array_filter(array_map(function ($domain) use ($activeDomain) {
			$user = User::loggedInUser();
			$activeDoms = $user->getDomActive();
			$totalCallCount = 0;

			// Unterbenutzer haben möglicherweise keinen Zugriff, dann raus
			if (!Domain::setDomainForRequest($domain->getId())) {
				return null;
			}

			Statistic::reloadRepository();
			ThemeModel::switchRepository($user,$domain);

			// Aktives Theme bestimmen
			$theme = ThemeModel::activeTheme();
			$themeName = $theme ? $theme->getName() : "default";

			// Statistik-Daten holen
			foreach (Statistic::inDateRange(new DateTime('first day of this month'), new DateTime('today')) as $stat) {
				$totalCallCount += $stat->getCallCount();
			}

			if ($user->isAccountOwner() || isset($activeDoms[$domain->getId()])) {
				return [
					'id' => $domain->getId(),
					'name' => $domain->getName(),
					'urlPath' => $domain->getUrlPath(),
					'url' => $domain->getUrl(),
					'widgetEnabled' => $domain->isFrontendWidgetEnabled(),
					'whitelabel' => $domain->hasWhitelabel(),
					'managementStructure' => $domain->get('managementStructure') ?? null,
					'themeName' =>$themeName,
					'crawlBudget' =>$domain->getDomainMaxCrawlBudget(),
					'crawlFrequency' =>$domain->getDomainExternalCrawlerFrequency(),
					'active' => $activeDomain && $domain->getId() == $activeDomain->getId(),
					'shareDomains' => $domain->getShareDomainNames(),
					'statistics' => $totalCallCount,
					'urls' => [
						'delete' => $this->generateUrl('app_domains_delete_form', ['id'=>$domain->getId()]),
						'edit' => $this->generateUrl('app_domains_edit_form', ['id'=>$domain->getId()]),
						'copy' => $this->generateUrl('app_domains_copy', ['id'=>$domain->getId()]),
						'select' => $this->generateUrl('app_dashboard', ['_domainId'=>$domain->getId()]),
					],
				];
			}
			else {
				return null;
			}
		}, Domain::all())));

		return $this->json([
			'data' => $domains,
		]);

	}

	/**
	 * @Route("/domains/create", name="app_domains_create")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function domainCreate(Request $request, TranslatorInterface $translator): Response
	{
		$id = 0;

		/** @var User $user */
		$user = User::loggedInUser()->getAccountOwner();

		//verfügbare WL Lizenzen
		$verfuegbarWhitelabelCount = self::whitelabelAvailableCountForUser($user);
		$verfuegbarWhitelabel = $verfuegbarWhitelabelCount != 0;

		if ($user->isDomainLimitReached()) {
			$this->addFlash('danger', $translator->trans('You\'ve reached the domain limit. Please contact your provider to raise that limit.'));
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		$formData = [
			'country' => '',
			'domainName' => '',
			'timeZone' => '',
			'whitelabel' => '',
			'domainHtPassWd' => '',
			'domainExternalCrawlerFrequency' => '8',
			'legalBasesDefault' => true,
			'whitelabel_verfuegbar' => $verfuegbarWhitelabel,
			'whitelabel_verfuegbar_count' => $verfuegbarWhitelabelCount < 0 ? $translator->trans('unlimited') : $verfuegbarWhitelabelCount,
		];

		if ($request->getRealMethod() == 'POST') {
			$action = $request->request->getAlnum("action");

			foreach (array_keys($formData) as $fieldName) {
				$formData[$fieldName] = trim((string)$request->request->get($fieldName));
			}
			$formData['legalBases'] = array_filter(array_map(
				function ($item) { return (string)$item; },
				$request->request->all("legalBases")
			));
			$formData['legalBasesDefault'] = (bool)$formData['legalBasesDefault'];

			$domainInfo = parse_url(filter_var(
				preg_match('~^https?://~', $formData['domainName']) ? $formData['domainName'] : 'https://'.$formData['domainName'],
				FILTER_SANITIZE_URL
			));
			if (isset($domainInfo['host'])) {
				$urlPath = '/'.ltrim($domainInfo['path'] ?? '/', '/');
				$urlScheme = $domainInfo['scheme'] ?? 'https';
			}
			else {
				$urlPath = '/';
				$urlScheme = 'https';
			}

			self::sanitizeFormData($formData);
			///print_r($formData);exit();

			try {
				$timeZone = new DateTimeZone($formData['timeZone']);
			} catch (Exception $exception) {
				$timeZone = null;
			}

			if ($formData['domainName']) {
				$domain = Domain::createForUser($user);
				$domain
					->setName($formData['domainName'])
					->setUrlPath($urlPath)
					->setUrlScheme($urlScheme)
					->setFrontendWidgetEnabled(false)
					->setWhitelabel((bool)$formData['whitelabel'])
					->setCountry($formData['country'] ? Country::find($formData['country']) : null)
					->setTimeZone($timeZone)
					->setLegalBases($formData['legalBasesDefault'] ? null: $formData['legalBases'])
					->setDomainHtPassWd($formData['domainHtPassWd'])
					->setInlineScriptsBlocked(false)
					->setSameDomainScriptsBlocked(false)
					->setManagementStructurePreLegalConsent(true)
					->setManagementStructure(Domain::MANAGEMENT_STRUCTURE_EMBEDDING)
					->setOnlyActiveOnMatchingDomain($user->isNewDomainOnlyActiveOnMatchingDomain())
					->save();

				// Grant access to subuser
				if (User::loggedInUser()->isSubUser()) {
					User::loggedInUser()
						->grantDomainAccess($domain)
						->save();
				}

				if($user->hasAccountFeature('canUseExternalCrawler')) {
					//Check, ob das Crawlerbudget nicht überschritten wird, der Wert der eigenen Domain muss hier wieder drauf, weil er bei der Anzeige der RestCrawlBudget abgezogen wird
					if ($request->request->getInt('domainMaxCrawlBudget') > (Domain::getRestCrawlBudget() + $domain->getDomainActualMaxCrawlBudget())) {
						$request->request->set('domainMaxCrawlBudget', Domain::getRestCrawlBudget() + $domain->getDomainActualMaxCrawlBudget());
						$this->addFlash('warning', $translator->trans('The CrawlerBudget was reset to the maximum allowed value: ') . (Domain::getRestCrawlBudget() + $domain->getDomainActualMaxCrawlBudget()));
					}

					//Noch die Crawler Filter formatieren um , und Leerzeichen usw. zu entfernen
					$filterData = trim($request->request->get('domainFilterForCrawler'));
					$filterDataArray = preg_split("/[\s,;]+/", $filterData);
					$filterData = trim(implode("\n", $filterDataArray));

					//Noch die Crawler Include Filter formatieren um , und Leerzeichen usw. zu entfernen
					$includeFilterData = trim($request->request->get('domainIncludeFilterForCrawler'));
					$includeFilterDataArray = preg_split("/[\s,;]+/", $includeFilterData);
					$includeFilterData = trim(implode("\n", $includeFilterDataArray));

					//Generelle User Settings überschreiben diese Einstellung hier
					if($user->isAllowAutoActivationOfNewEmbeddings()===false){
						$request->request->set('activateEmbeddingsFoundByCrawler', false);
						$this->addFlash('danger', 'You are not allowed to deactivate the crawl for this domain because of general Account Settings.');
					}
					if($user->isCrawlAllDomains()===false){
						$request->request->set('doCrawl', false);
						$this->addFlash('danger', 'You are not allowed to crawl this domain because of general Account Settings.');
					}

					//activatePermaScan deaktivieren wenn der Crawler aktiviert wird.
					$activatePermaScan = true;
					if($request->request->getBoolean('doCrawl'))
					{
						$activatePermaScan = false;
					}

					//CrawlerBudget setzen
					$domain
						->setDomainMaxCrawlBudget($request->request->getInt('domainMaxCrawlBudget'))
						->setDomainExternalCrawlerFrequency($request->request->getInt('domainExternalCrawlerFrequency'))
						->setDomainFilterForCrawler($filterData)
						->setDomainIncludeFilterForCrawler($includeFilterData)
						->setActivateEmbeddingsFoundByCrawler($request->request->getBoolean('activateEmbeddingsFoundByCrawler'))
						->setDoCrawl($request->request->getBoolean('doCrawl'))
						->setActivatePermaScan($activatePermaScan)
						->save();
				}

				$this->addFlash('success', $translator->trans('Domain :name has been created!', [
					':name' => $formData['domainName'],
				]));

				// Index neu generieren
				DomainIndex::rebuild();

				// Give plugins a chance to initialize their data for the new domain
				Utils::dispatchEvent(DomainInitializedEvent::NAME, new DomainInitializedEvent($domain));

				switch ($action) {
					case 'addDomainClose':
						return $this->redirectToRoute('app_domains_index', [], 303);
					case 'addDomain':
					default:
						return $this->redirectToRoute('app_domains_edit_form', ['id' => $domain->getId()], 303);
				}
			}
			else {
				$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
			}
		} else {
			$formData['legalBases'] = $user->getDefaultLegalBases();
		}



		return $this->render('domain/create.html.twig', [
			'formData' => $formData,
			'countries' => Country::all(),
			'regulations' => Regulation::all(),
			'csrfToken' => CsrfToken::get(),
			'restCrawlBudget' => Domain::getRestCrawlBudget() ?? 0,
			'restCrawlBudgetProDomain' => Domain::getVerfuegbaresCrawlBudgetProDomain() ?? 0,
			'possibleCrawls' => $user->getAccountFeatureExternalCrawlerFrequency(),
			'timeZones' => DateTimeZone::listIdentifiers() ?: ['Europe/Berlin', 'UTC'],
		]);
	}

	/**
	 * @Route("/domains/{id}", methods={"HEAD", "GET"}, name="app_domains_edit_form")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function domainEditForm(string $id, Request $request, TranslatorInterface $translator): Response
	{
		/** @var User $user */
		$user = User::loggedInUser()->getAccountOwner();

		/** @var Domain|null $domain */
		$domain = Domain::find($id);

		if ($domain == null) {
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		$whitelabelAvailableCount = self::whitelabelAvailableCountForUser($user);
		$whitelabelAvailable = $whitelabelAvailableCount != 0 || $domain->hasWhitelabel();

		$formData = [
			'country' => '',
			'domainName' => '',
			'whitelabel' => '',
			'domainHtPassWd' => '',
			'whitelabel_verfuegbar' => $whitelabelAvailable,
			'whitelabel_verfuegbar_count' => $whitelabelAvailableCount < 0 ? $translator->trans('unlimited') : $whitelabelAvailableCount,
		];

		return $this->render('domain/edit.html.twig', [
			'domain' => $domain,
			'csrfToken' => CsrfToken::get(),
			'countries' => Country::all(),
			'regulations' => Regulation::all(),
			'restCrawlBudget' => Domain::getRestCrawlBudget() ?? 0,
			'possibleCrawls' => $user->getAccountFeatureExternalCrawlerFrequency(),
			'timeZones' => DateTimeZone::listIdentifiers() ?: ['Europe/Berlin', 'UTC'],
			'formData' => $formData
		]);
	}

	/**
	 * @Route("/domains/{id}/copy", methods={"POST"}, name="app_domains_copy")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function copyEditForm(string $id, Request $request, TranslatorInterface $translator): Response
	{

		/** @var User $user */
		$user = User::loggedInUser()->getAccountOwner();
		/** @var Domain|null $domain */
		$domain = Domain::find($id);

		/**
		 * ok - neue Domain erzeugen mit dem aktuellen Namen und Copy hinten dran...
		 *
		 */

		//OK? - sonst raus...
		if ($user->isDomainLimitReached()) {
			$this->addFlash('danger', $translator->trans('You\'ve reached the domain limit. Please contact your provider to raise that limit.'));
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		$newDomain = clone $domain;

		// Index neu generieren
		DomainIndex::rebuild();

		$this->addFlash('success', $translator->trans('Domain :name has been copyied.', [
			':name' => $domain->getName(),
		]));

		return $this->redirectToRoute('app_domains_index', [], 303);
	}

	/**
	 * @Route("/domains/{id}", methods={"PATCH"}, name="app_domains_edit")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function domainEdit(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$action = $request->request->getAlnum('action');
		Domain::beginWrite();
		/** @var User $user */
		$user = User::loggedInUser()->getAccountOwner();
		/** @var Domain|null $domain */
		$domain = Domain::find($id);
		$filterData = "";
		$includeFilterData = "";

		if ($domain == null) {
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		$whitelabelAvailableCount = self::whitelabelAvailableCountForUser($user);
		$whitelabelAvailable = $whitelabelAvailableCount != 0 || $domain->hasWhitelabel();

		$formData = [
			'country' => '',
			'domainName' => '',
			'timeZone' => '',
			'whitelabel' => '',
			'domainHtPassWd' => '',
			'whitelabel_verfuegbar' => $whitelabelAvailable,
			'whitelabel_verfuegbar_count' => $whitelabelAvailableCount < 0 ? $translator->trans('unlimited') : $whitelabelAvailableCount,
		];

		foreach (array_keys($formData) as $fieldName) {
			$formData[$fieldName] = trim((string)$request->request->get($fieldName));
		}

		$domainInfo = parse_url(filter_var(
			preg_match('~^https?://~', $formData['domainName']) ? $formData['domainName'] : 'https://'.$formData['domainName'],
			FILTER_SANITIZE_URL
		));
		if (isset($domainInfo['host'])) {
			$urlPath = '/'.ltrim($domainInfo['path'] ?? '/', '/');
			$urlScheme = $domainInfo['scheme'] ?? 'https';
		}
		else {
			$urlPath = '/';
			$urlScheme = 'https';
		}

		self::sanitizeFormData($formData);

		try {
			$timeZone = new DateTimeZone($formData['timeZone']);
		} catch (Exception $exception) {
			$timeZone = null;
		}

		$legalBases = $request->request->getBoolean("legalBasesDefault") ? null : array_filter(array_map(
			static function ($item) { return (string)$item; },
			$request->request->all("legalBases")
		));

		//Darf der User den Crawler nutzen?
		if($user->hasAccountFeature('canUseExternalCrawler'))
		{
			//-1 abchecken, darf nicht gesetzt sein, wenn es nicht vom Tarif her erlaubt ist
			if(Domain::getRestCrawlBudget() !== -1 && $request->request->getInt('domainMaxCrawlBudget') === -1)
			{
				$request->request->set('domainMaxCrawlBudget', Domain::getRestCrawlBudget()+$domain->getDomainActualMaxCrawlBudget());
			}

			//Check, ob das Crawlerbudget nicht überschritten wird, der Wert der eigenen Domain muss hier wieder drauf, weil er bei der Anzeige der RestCrawlBudget abgezogen wird
			if(Domain::getRestCrawlBudget() !== -1 &&  $request->request->getInt('domainMaxCrawlBudget') > (Domain::getRestCrawlBudget()+$domain->getDomainActualMaxCrawlBudget()))
			{
				if($request->request->getInt('domainMaxCrawlBudget')!=0)
				{
					$request->request->set('domainMaxCrawlBudget', Domain::getRestCrawlBudget()+$domain->getDomainActualMaxCrawlBudget());
				}
				$this->addFlash('warning', $translator->trans('The CrawlerBudget was reset to the maximum allowed value: ').(Domain::getRestCrawlBudget()+$domain->getDomainActualMaxCrawlBudget()));
			}

			//CrawlerFrequency auf den passenden Wert setzen, wenn notwendig
			if($request->request->getInt('domainExternalCrawlerFrequency') !== 0 && $request->request->getInt('domainExternalCrawlerFrequency') < ($user->getAccountFeatureExternalCrawlerFrequency())) {
				$request->request->set('domainExternalCrawlerFrequency', $user->getAccountFeatureExternalCrawlerFrequency());
				$this->addFlash('warning', $translator->trans('The Crawler Frequency was reset to the maximum allowed value: ').($user->getAccountFeatureExternalCrawlerFrequency()));
			}
			//Nur festgelegte Werte erlauben
			$crawlerFrequenz = array(8,4,1,0);
			if(!in_array($request->request->getInt('domainExternalCrawlerFrequency'),$crawlerFrequenz)) {
				$request->request->set('domainExternalCrawlerFrequency', 8);
			}

			//Noch die Crawler Filter formatieren um , und Leerzeichen usw. zu entfernen
			$filterData = trim($request->request->get('domainFilterForCrawler'));
			$filterDataArray = preg_split("/[\s,;]+/", $filterData);
			$filterData = trim(implode("\n", $filterDataArray));

			//Noch die Crawler Include Filter formatieren um , und Leerzeichen usw. zu entfernen
			$includeFilterData = trim($request->request->get('domainIncludeFilterForCrawler'));
			$includeFilterDataArray = preg_split("/[\s,;]+/", $includeFilterData);
			$includeFilterData = trim(implode("\n", $includeFilterDataArray));

			//Generelle User Settings überschreiben diese Einstellung hier
			if($user->isAllowAutoActivationOfNewEmbeddings()===false){
				$request->request->set('activateEmbeddingsFoundByCrawler', false);
				$this->addFlash('danger', 'You are not allowed to deactivate the crawl for this domain because of general Account Settings.');
			}
			if($user->isCrawlAllDomains()===false){
				$request->request->set('doCrawl', false);
				$this->addFlash('danger', 'You are not allowed to crawl this domain because of general Account Settings.');
			}
		}

		//Darf nicht crawlen, dann auf 0 setzen
		else{
			$request->request->set('domainExternalCrawlerFrequency', 0);
			$request->request->set('domainMaxCrawlBudget', 0);
		}

		//Sonderfall - Account hat 0 Budget, dann auch die Domain auf 0 setzen
		if( $user->getAccountFeatureMaxCrawlBudget()===0 && $domain->getDomainActualMaxCrawlBudget() > 0)
		{
			$request->request->set('domainMaxCrawlBudget', 0);
			$this->addFlash('warning', $translator->trans('The CrawlerBudget was reset to the maximum allowed value: 0'));
		}

		//activatePermaScan deaktivieren wenn der Crawler aktiviert wird.
		$activatePermaScan = true;
		if($request->request->getBoolean('doCrawl'))
		{
			$activatePermaScan = false;
		}

		// Daten temporär zuweisen, damit im Fehlerfall die Benutzereingaben im Formular ausgegeben werden whitelabel
		$domain
			->setName($formData['domainName'])
			->setUrlPath($urlPath)
			->setUrlScheme($urlScheme)
			->setDomainHtPassWd($formData['domainHtPassWd'])
			->setWhitelabel((bool)$formData['whitelabel'])
			->setCountry($formData['country'] ? Country::find($formData['country']) : null)
			->setLegalBases($legalBases)
			->setTimeZone($timeZone)
			->setDomainMaxCrawlBudget($request->request->getInt('domainMaxCrawlBudget'))
			->setDomainFilterForCrawler($filterData)
			->setDomainIncludeFilterForCrawler($includeFilterData)
			->setDomainExternalCrawlerFrequency($request->request->getInt('domainExternalCrawlerFrequency'))
			->setActivateEmbeddingsFoundByCrawler($request->request->getBoolean('activateEmbeddingsFoundByCrawler'))
			->setDoCrawl($request->request->getBoolean('doCrawl'))
			->setActivatePermaScan($activatePermaScan)
		;

		$ok = true;
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$error = $translator->trans('CSRF token is missing or invalid. Please try again.');
			$this->addFlash('danger', $error);
			$ok = false;
		}

		if (!$formData['domainName']) {
			$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
			$ok = false;
		}

		if (!empty($formData['domainHtPassWd']) && strpos($formData['domainHtPassWd'], ":") === false) {
			$this->addFlash('danger', $translator->trans('Please enter : between user and password.'));
			$ok = false;
		}

		if ($ok) {
			$domain->save();

			$this->addFlash('success', $translator->trans('Domain :name has been updated!', [
				':name' => $formData['domainName'],
			]));

			// Bei CK Cloud noch den Tarif setzen wenn die ID vorhanden ist
			$tarifId = $request->request->get("tarifId");
			if (Utils::isClickskeksCloud() && is_string($tarifId) && is_numeric($tarifId) && User::originalLoggedInUser()) {
				$domTarifList = $user->getDomCKList();
				if ($tarifId != 999) {
					$domTarifList[$domain->getId()] = $tarifId;
				}
				elseif ($tarifId == 999) {
					unset($domTarifList[$domain->getId()]);
					$freeData = $user->getDomCK30DayFreeList();
					$freeData[$domain->getId()] = array("id"=>$domain->getId(),"start"=>time(),"name"=>$domain->getName());
					$user->setDomCK30DayFreeList($freeData);
				}

				$user->setDomCKList($domTarifList);
				$user->save();
			}

			// Index neu generieren
			DomainIndex::rebuild();

			switch ($action) {
				case 'saveClose':
					return $this->redirectToRoute('app_domains_index', [], 303);
				case 'save':
				default:
					return $this->redirectToRoute('app_domains_edit_form', ['id' => $domain->getId()], 303);
			}
		}

		return $this->render('domain/edit.html.twig', [
			'domain' => $domain,
			'csrfToken' => CsrfToken::get(),
			'countries' => Country::all(),
			'regulations' => Regulation::all(),
			'timeZones' => DateTimeZone::listIdentifiers() ?: ['Europe/Berlin', 'UTC'],
			'formData' => $formData
		]);
	}

	/**
	 * @Route("/domains/{id}/delete", methods={"HEAD", "GET"}, name="app_domains_delete_form")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function domainDeleteForm(string $id, Request $request, TranslatorInterface $translator): Response
	{
		/** @var Domain|null $domain */
		$domain = Domain::find($id);

		if ($domain == null) {
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		return $this->render('domain/delete.html.twig', [
			'domain' => $domain,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/domains/{id}", methods={"DELETE"}, name="app_domains_delete")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function domainDelete(string $id, Request $request, TranslatorInterface $translator,LoggerInterface $logger): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$error = $translator->trans('CSRF token is missing or invalid. Please try again.');
			$this->addFlash('danger', $error);
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		Domain::beginWrite();
		/** @var Domain|null $domain */
		$domain = Domain::find($id);

		if ($domain == null) {
			return $this->redirectToRoute('app_domains_index', [], 303);
		}

		$domainName = $domain->getName();
		Domain::delete($domain->getId());
		$user = User::loggedInUser();

		$logger->logWarning('Domain ', "DomainDeleted: '" . $domainName . "' User -> " . $user->getId());
		$this->addFlash('success', $translator->trans('Domain :name has been deleted.', [
			':name' => $domainName,
		]));

		// Index neu generieren
		DomainIndex::rebuild();

		return $this->redirectToRoute('app_domains_index', [], 303);
	}

	/**
	 * @Route("/domains/{id}/select", name="app_domains_select")
	 * @deprecated
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function domainSelect(string $id, Request $request, TranslatorInterface $translator): Response
	{
		if (Domain::select($id)) {
			/** @var Domain|null $domain */
			$domain = Domain::find($id);

			$this->addFlash('success', $translator->trans('Domain :name has been selected.', [
				':name' => $domain->getName(),
			]));
			return $this->redirectToRoute('app_dashboard', ['_domainId'=>$domain->getId()], 303);
		}
		else {
			$this->addFlash('danger', $translator->trans('Invalid domain selected, please try again.'));
			return $this->redirectToRoute('app_domains_index', [], 303);
		}
	}

	/**
	 * @param array $data
	 */
	private static function sanitizeFormData(array &$data): void
	{
		// Leerzeichen entfernen und Domain auf den Namen trimmen
		$data['domainName'] = trim(preg_replace('~\s~', '',
			// Alles bis zu einem möglichen Protokoll und hinter einem abschließenden Slash löschen
			// Eingabe: e r23rhttps://http:////www.beispiel.xyz/weorihwe heorihw   howeihroih /er wo=-w er=rj
			// Ausgabe: www.beispiel.xyz
			preg_replace('~^(.*https?://)?(?(1)|)/*([^/]+).*~', '$2', $data['domainName'])
		));
	}
}
