<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Model\Domain;
use App\Model\Embedding;
use App\Component\Tcf\v2\Model\Google\AdTechProvider;
use App\Component\Tcf\v2\Model\SpecialFeature;
use App\Component\Tcf\v2\Model\Vendor;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

/**
 * Controller zur Verwaltung domainspezifischer TCFv2-Daten
 * @Route("/domains/{_domainId}/tcf/v2/", name="app_tcf_")
 */
class DomainTcf extends DomainDependantController
{
	/**
	 * @Route("vendor/{vendor}/form/{embedding}", name="xhr_embedding_form")
	 *
	 * @return Response
	 */
	public function form(Vendor $vendor, ?Embedding $embedding): Response
	{
		// Embedding-Daten nur berücksichtigen, wenn die Vendor-ID stimmt
		if ($embedding and $embedding->getTcfVendor() != $vendor) {
			$embedding = null;
		}

		return $this->render('tcf/embedding_form.html.twig', [
			'vendor' => $vendor,
			'flexiblePurposeIds' => $vendor->getFlexiblePurposeIds(),
			'embeddingPurposes' => ($embedding) ? $embedding->getTcfPurposes() : null,
			'embeddingSpecialFeatures' => ($embedding) ? $embedding->getTcfSpecialFeatures() : null,
		]);
	}

	/**
	 * @Menu("IAB Framework", group="Cookies & Settings", icon="glyphicon-list", routeGroup="app_tcf_", order=35)
	 * @Route("global-vendor-list", name="global_vendor_list", methods={"GET", "HEAD"})
	 * @return Response
	 */
	public function indexGlobalVendorList(Request $request): Response
	{
		$domain = Domain::activeDomain();
		$blockedVendorIds = array_map(function (Vendor $vendor) {
			return $vendor->getId();
		}, $domain->getTcfVendorBlacklist());

		$selectedGoogleAdTechProviderIds = array_map(function (AdTechProvider $provider) {
			return $provider->getId();
		}, $domain->getGoogleAdditionalConsentWhitelist());

		// Hole aktive Embeddings, um referenzierte Anbieter zu ermitteln
		$activeEmbeddings = array_filter(Embedding::all(), function (Embedding $embedding) {
			return $embedding->isActive();
		});

		// Ermittle Special Features, die in Embeddings referenziert werden
		/** @var string[] $referencedSpecialFeatureIds */
		$referencedSpecialFeatureIds = array_reduce($activeEmbeddings, function (array $ids, Embedding $embedding) {
			foreach ($embedding->getTcfSpecialFeatures() as $feature) {
				$ids[$feature->getId()] = $feature->getId();
			}
			return $ids;
		}, []);

		$enabledSpecialFeatureIds = array_reduce($domain->getExplicitTcfSpecialFeatures(), function (array $ids, SpecialFeature $feature) {
			$ids[$feature->getId()] = $feature->getId();
			return $ids;
		}, []);

		$specialFeatures = array_map(function (SpecialFeature $feature) use ($referencedSpecialFeatureIds, $enabledSpecialFeatureIds) {
			$isEnabled = isset($enabledSpecialFeatureIds[$feature->getId()]);
			$isReferenced = isset($referencedSpecialFeatureIds[$feature->getId()]);

			return [
				'id' => $feature->getId(),
				'feature' => $feature,
				'checked' => $isEnabled || $isReferenced,
				'referenced' => $isReferenced,
			];
		}, SpecialFeature::all());

		return $this->render('tcf/global_vendor_list/index.html.twig', [
			'blockedVendorIds' => $blockedVendorIds,
			'specialFeatures' => $specialFeatures,
			'selectedGoogleAdTechProviderIds' => $selectedGoogleAdTechProviderIds,
		]);
	}

	/**
	 * @Route("global-vendor-list.json", name="xhr_global_vendor_list_ajax")
	 * @return JsonResponse
	 */
	public function ajaxGlobalVendorList(Request $request): JsonResponse
	{
		$vendors = Vendor::all();

		// Sortiere Anbieter nach ihren Namen
		usort($vendors, function (Vendor $a, Vendor $b) {
			return strnatcasecmp($a->getName(), $b->getName());
		});

		// Hole aktive Embeddings, um referenzierte Anbieter zu ermitteln
		$activeEmbeddings = array_filter(Embedding::all(), function (Embedding $embedding) {
			return $embedding->isActive();
		});

		// Ermittle Anbieter, die in Embeddings referenziert werden
		/** @var string[] $referencedVendorIds */
		$referencedVendorIds = array_reduce($activeEmbeddings, function (array $ids, Embedding $embedding) {
			if ($vendor = $embedding->getTcfVendor()) {
				$ids[$vendor->getId()] = $vendor->getId();
			}
			return $ids;
		}, []);

		$blockedVendorIds = array_map(function (Vendor $vendor) {
			return $vendor->getId();
		}, Domain::activeDomain()->getTcfVendorBlacklist());

		// Session schließen
		$request->getSession()->save();

		$data = array_map(function (Vendor $vendor) use ($blockedVendorIds, $referencedVendorIds) {
			$isBlockedVendor = in_array($vendor->getId(), $blockedVendorIds);
			$isReferencedVendor = in_array($vendor->getId(), $referencedVendorIds);

			return [
				'id' => $vendor->getId(),
				'name' => $vendor->getName(),
				'checked' => !$isBlockedVendor || $isReferencedVendor,
				'referenced' => $isReferencedVendor,
			];
		}, $vendors);

		return $this->json([
			'data' => $data,
		]);
	}

	/**
	 * @Route("google-additional-consent-list.json", name="xhr_google_additional_consent_list_ajax")
	 * @return JsonResponse
	 */
	public function ajaxGoogleAdditionalConsentList(Request $request): JsonResponse
	{
		$providers = AdTechProvider::all();

		// Sortiere Anbieter nach ihren Namen
		usort($providers, function (AdTechProvider $a, AdTechProvider $b) {
			return strnatcasecmp($a->getName(), $b->getName());
		});

		$selectedProviderIds = array_map(function (AdTechProvider $provider) {
			return $provider->getId();
		}, Domain::activeDomain()->getGoogleAdditionalConsentWhitelist());

		// Session schließen
		$request->getSession()->save();

		$data = array_map(function (AdTechProvider $provider) use ($selectedProviderIds) {
			$isSelectedProvider = in_array($provider->getId(), $selectedProviderIds);

			return [
				'id' => $provider->getId(),
				'name' => $provider->getName(),
				'checked' => $isSelectedProvider,
			];
		}, $providers);

		return $this->json([
			'data' => $data,
		]);
	}

	/**
	 * @Route("global-vendor-list", name="global_vendor_list_update", methods={"POST"})
	 * @param Request $request
	 * @return Response
	 */
	public function updateGlobalVendorList(Request $request): Response
	{
		$domain = Domain::activeDomain();
		$vendors = Vendor::all();
		$googleAdditionalConsentMode = $request->request->getBoolean('googleAdditionalConsentMode');
		$selectedGoogleAdTechProviderIds = explode(',', $request->request->get('selectedGoogleAdTechProviders', ''));
		$tcfCompatVersion = ((string)$request->request->get('tcfCompatVersion', '2.2'));

		switch ($request->request->get('action')) {
			case 'allow-all':
				$blockedVendorIds = [];
				$specialFeatures = SpecialFeature::all();
				$selectedGoogleAdTechProviderIds = $googleAdditionalConsentMode ?
					array_map(function (AdTechProvider $provider) { return $provider->getId(); }, AdTechProvider::all()) :
					$selectedGoogleAdTechProviderIds;
				break;
			case 'block-all':
				$blockedVendorIds = array_map(function (Vendor $vendor) {
					return $vendor->getId();
				}, $vendors);
				$specialFeatures = [];
				$selectedGoogleAdTechProviderIds = $googleAdditionalConsentMode ? [] : $selectedGoogleAdTechProviderIds;
				break;
			case 'save':
			default:
				$blockedVendorIds = explode(',', $request->request->get('blockedVendors'));
				$domain->setTcfVendorsEnabled($request->request->getBoolean('tcfVendorsEnabled'));
				$domain->setNewTcfVendorsActive($request->request->getBoolean('newTcfVendorsActive'));
				$domain->setTcfAdTagsEnabled($request->request->getBoolean('tcfAdTagsEnabled'));
				$domain->setTcfAdTagsAutoAppend($request->request->getBoolean('tcfAdTagsAutoAppend'));
				$domain->setTcfLegitimateInterestEnabled($request->request->getBoolean('tcfLegitimateInterestEnabled'));
				$specialFeatures = array_filter(array_map(function ($id) {
					return SpecialFeature::find((string)$id);
				}, (array)$request->request->get('specialFeatures')));
				break;
		}

		// Hole aktive Embeddings, um referenzierte Anbieter zu ermitteln
		$activeEmbeddings = array_filter(Embedding::all(), function (Embedding $embedding) {
			return $embedding->isActive();
		});

		// Ermittle Anbieter, die in Embeddings referenziert werden
		/** @var string[] $referencedVendorIds */
		$referencedVendorIds = array_reduce($activeEmbeddings, function (array $ids, Embedding $embedding) {
			if ($vendor = $embedding->getTcfVendor()) {
				$ids[$vendor->getId()] = $vendor->getId();
			}
			return $ids;
		}, []);

		// Sicherstellen, dass nur Anbieter zu blockieren sind, die nicht referenziert werden
		$blockedVendorIds = array_diff($blockedVendorIds, $referencedVendorIds);

		// Sicherstellen, dass Anbieter existieren
		/** @var Vendor[] $blockedVendors */
		$blockedVendors = array_filter(array_map(function ($id) {
			return Vendor::find((string)$id);
		}, $blockedVendorIds));

		// Neusten Anbieter festhalten, um Blacklist nach Update der GVL zu erweitern
		/** @var Vendor $mostRecentVendor */
		$mostRecentVendor = array_reduce($vendors, function (?Vendor $carry, Vendor $item) {
			return $carry ? ($item->getId() > $carry->getId() ? $item : $carry) : $item;
		});

		// Lücken innerhalb der GVL festhalten, um Blacklist nach GVL-Update zu erweitern
		$unusedVendorIds = array_values(array_filter(range(1, (int)$mostRecentVendor->getId()), function (int $id) {
			return Vendor::exists((string)$id) == false;
		}));

		// Googles AdditionalConsent-Liste bauen
		/** @var AdTechProvider[] $selectedGoogleAdTechProviders */
		$selectedGoogleAdTechProviders = array_filter(array_map(function ($id) {
			return AdTechProvider::find((string)$id);
		}, $selectedGoogleAdTechProviderIds));

		$domain
			->setTcfVendorBlacklist($blockedVendors)
			->setExplicitTcfSpecialFeatures($specialFeatures)
			->setMostRecentTcfVendor($mostRecentVendor)
			->setUnusedTcfVendorIds($unusedVendorIds)
			->setGoogleAdditionalConsentMode($googleAdditionalConsentMode)
			->setGoogleAdditionalConsentWhitelist($selectedGoogleAdTechProviders)
			->setTcfCompatVersion($tcfCompatVersion)
			->updateTcfStacks()
			->touchFrontendWidgetCacheTimestamp()
			->save();

		return $this->redirectToRoute('app_tcf_global_vendor_list');
	}
}

