<?php

namespace App\Controller;

use App\Component\EmbeddingUpdateService;
use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingBase;
use App\Model\SystemEmbedding;
use App\Model\Purpose;
use App\Model\User;
use App\Model\UserEmbedding;
use App\Model\UserLocale;
use App\Model\Embedding as EmbeddingModel;
use App\Model\Language;
use App\Model\LocaleBase;
use App\Model\Regulations\Article;
use App\Model\Regulations\Regulation;
use App\Model\Regulations\Group as RegulationGroup;
use App\Config;
use App\Component\LocaleService;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\EmbeddingDb\EmbeddingDbLegacy;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Utils;
use Exception;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

use DateTimeImmutable;
use DateTimeInterface;

/**
 * Controller zur Embedding-Verwaltung
 * @Route("/domains/{_domainId}/embeddings", name="app_embeddings_")
 */
class Embeddings extends DomainDependantController
{
	/**
	 * @Menu("Embeddings", group="Cookies & Settings", icon="fa-certificate", order=30)
	 * @Route("", name="list")
	 *
	 * @param Request $request
	 * @throws Exception Wenn kein Zugriff auf das Embedding-File erfolgen kann
	 *
	 * @return Response
	 */
	public function listEmbeddings(Request $request): Response
	{
		$embeddings = EmbeddingModel::sortAllByPurpose();
		$purposes = Purpose::all();
		$purposeCount = array();
		$onlyTechNotwendig = 0;

		//Alles durchlaufen und checken ob es nur 1 Purpose gibt
		foreach ($embeddings as $embk=>$embedding)
		{
			$ppID = $embedding->getPurpose()->getId();
			$purposeCount[$ppID]=1;
		}
		if(count($purposeCount)==1 && !empty($purposeCount['41ba25c']) )
		{
			$onlyTechNotwendig = 1;
		}

		return $this->render('embedding/list.html.twig', [
			'embeddings' => $embeddings,
			'purposes' => $purposes,
			'onlyTechNotwendig' =>$onlyTechNotwendig,
		]);
	}

	/**
	 * @Route("/list.json", name="list_ajax")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return JsonResponse
	 */
	public function ajaxEmbeddingSearch(Request $request, TranslatorInterface $translator): JsonResponse
	{
		$currentLanguageCode = explode('_', $request->getLocale(), 2)[0];
		$defaultLocale = LocaleService::getDefault();
		$defaultLanguage = $defaultLocale->getLanguage() ?? Language::find('de_DE');
		$backendLocale = LocaleService::findFuzzy($request->getLocale());
		$backendLanguage = $backendLocale ? $backendLocale->getLanguage() : null;
		$language = $backendLanguage ?? Language::find($currentLanguageCode) ?? $defaultLanguage;
		$locale = ($backendLocale ? UserLocale::find($backendLocale->getName()) : null) ?? LocaleService::getUserDefault();

		$search = trim((string)($request->query->all('search')['value'] ?? ''));

		// Lade alle Benutzervorlagen
		$userTemplateNames = [];
		$userTemplates = UserEmbedding::all();
		foreach ($userTemplates as $userTemplate) {
			$userTemplateNames[$userTemplate->getName()] = $userTemplate->getName();
		}

		// Mappe Daten der Benutzervorlagen
		$userTemplates = array_values(array_filter(array_map(function ($embedding) use ($locale, $translator) {
			$assets = [];
			foreach ($embedding->getAssets() as $asset) {
				$assets[] = $asset->getName();
			}
			return [
				'_score' => null,
				'id' => $embedding->getId(),
				'name' => $embedding->getTranslatedName($locale),
				'source' => EmbeddingBase::SOURCE_USER_TEMPLATE,
				'embedding_vendor' => $embedding->getEmbeddingVendor(),
				'description' => $embedding->getTranslatedDescription($locale),
				'purpose' => $embedding->getPurpose()->getName(),
				'assets' => $assets,
				'category' => $translator->trans('User embedding templates')
			];
		}, $userTemplates)));

		// Lade alle Systemweite Einbindungsvorlagen
		$templates = array_filter(SystemEmbedding::all(), function (SystemEmbedding $e) use ($userTemplateNames) {
			return !isset($userTemplateNames[$e->getName()]);
		});

		// Mappe Daten der Systemweiten Embeddings
		$templates = array_values(array_filter(array_map(function ($embedding) use ($language, $translator) {
			$assets = [];
			foreach ($embedding->getAssets() as $asset) {
				$assets[] = $asset->getName();
			}
			return [
				'_score' => null,
				'id' => $embedding->getId(),
				'name' => $embedding->getTranslatedName($language),
				'source' => EmbeddingBase::SOURCE_SYSTEM_TEMPLATE,
				'embedding_vendor' => $embedding->getEmbeddingVendor(),
				'description' => $embedding->getTranslatedDescription($language),
				'purpose' => $embedding->getPurpose()->getName(),
				'assets' => $assets,
				'category' => $translator->trans('System-wide embedding templates')
			];
		}, $templates)));

		// Globale DB abfragen, wenn aktiviert (Einbindungs-Daten aus externer Datenbank)
		if (Config::getInstance()->isGlobalCookieDbUsed()) {
			// FIXME Total record count is unknown when searching
			$embeddingsDB = $this->searchGlobalCookieDb(strlen($search) ? $search : null, $language->getShortCode());
		} else {
			$embeddingsDB = [];
		}

		// Mappe Daten der Embeddings aus externer DB
		$embeddingsDB = array_values(array_filter(array_map(function ($embedding) use ($translator) {
			$assets = [];
			foreach ($embedding['assets'] as $asset) {
				$assets[] = $asset['name'];
			}
			return [
				'_score' => $embedding['@score'] ?? null,
				'id' => $embedding['id'],
				'name' => $embedding['translations']['current']['label'],
				'source' => EmbeddingBase::SOURCE_GLOBAL_DB,
				'embedding_vendor' => $embedding['embedding_vendor'],
				'description' => $embedding['translations']['current']['description'],
				'purpose' => Purpose::find($embedding['purpose'])->getName(),
				'assets' => $assets,
				'category' => $translator->trans('Embedding Data from remote Database')
			];
		}, $embeddingsDB)));

		// Sortierung der Embeddings
		$order = $request->query->all('order')[0];
		$sortColumn = $order['column'] ?? '';
		$sortDirection = $order['dir'] ?? 'asc';

		$userTemplates = Utils::sortNatNum($userTemplates, $sortColumn, $sortDirection);
		$templates = Utils::sortNatNum($templates, $sortColumn, $sortDirection);
		$embeddingsDB = Utils::sortNatNum($embeddingsDB, $sortColumn, $sortDirection);

		$visibleEmbeddings = array_merge($templates, $userTemplates, $embeddingsDB);

		// Filtere bereits bestehende Embeddings aus angezeigten Embeddings heraus
		$activeEmbeddings = EmbeddingModel::all();
		foreach ($activeEmbeddings as $activeEmbedding) {
			foreach ($visibleEmbeddings as $key => $embeddingTemplate) {
				if ($embeddingTemplate['id'] == $activeEmbedding->getSourceId()) {
					unset($visibleEmbeddings[$key]);
				}
			}
		}

		// Filterung nach Suchbegriff
		$embeddingsFiltered = strlen($search) > 0 ? self::searchInArray($search, $visibleEmbeddings) : $visibleEmbeddings;

		// Paginierung der Einträge
		$start = $request->query->getInt('start');
		$length = $request->query->getInt('length') ?: 10;

		$pagination = new Pagination($embeddingsFiltered, $length);
		$currentPage = ($start / $length) + 1;
		$pagination->setCurrentPage($currentPage);

		$embeddingsPaginated = $pagination->getItems();

		return $this->json([
			'data' => array_values($embeddingsPaginated),
			'draw' => $request->query->get('draw'),
			'recordsTotal' => count($visibleEmbeddings),
			'recordsFiltered' => count($embeddingsFiltered),
		]);
	}

	/**
	 * @Route("/add-{kind<cookie|localStorage|sessionStorage|indexedDb|webSql|other>}/{name}", name="add_asset", methods={"HEAD", "GET"})
	 *
	 * @param Request $request
	 * @param string $kind
	 * @param string $name
	 * @throws Exception Wenn kein Zugriff auf das Embedding-File erfolgen kann
	 * @return Response
	 */
	public function addAssetPage(Request $request, string $kind, string $name): Response
	{
		$name = urldecode($name);
		$currentLanguageCode = explode('_', $request->getLocale(), 2)[0];
		$embeddings = EmbeddingModel::all();
		$purposes = Purpose::all();

		$usedSources = array_filter(array_map(function (EmbeddingModel $embedding) {
			$source = $embedding->getSource();
			$sourceId = $embedding->getSourceId();
			if ($sourceId and $source) {
				return [$source, $sourceId];
			} else {
				return null;
			}
		}, $embeddings));

		// Passende User-Embeddings suchen
		$userEmbeddings = array_filter(UserEmbedding::all(), function (UserEmbedding $userEmbedding) use ($kind, $name, $usedSources) {
			foreach ($userEmbedding->getAssets() as $asset) {
				if ($asset->getStorageType() === $kind and $asset->matchName($name)) {
					// Nur anzeigen, wenn nicht schon ein Embedding mit dieser Quelle existiert
					if (!in_array([EmbeddingModel::SOURCE_USER_TEMPLATE, $userEmbedding->getId()], $usedSources, true)) {
						return true;
					}
				}
			}
			return false;
		});
		if (!$userEmbeddings) {
			$userEmbeddings = array_filter(UserEmbedding::all(), function (UserEmbedding $userEmbedding) use ($name, $usedSources) {
				foreach ($userEmbedding->getAssets() as $asset) {
					if ($asset->matchName($name)) {
						// Nur anzeigen, wenn nicht schon ein Embedding mit dieser Quelle existiert
						if (!in_array([EmbeddingModel::SOURCE_USER_TEMPLATE, $userEmbedding->getId()], $usedSources, true)) {
							return true;
						}
					}
				}
				return false;
			});
		}
		$userEmbeddingNames = [];
		foreach ($userEmbeddings as $embedding) {
			$userEmbeddingNames[$embedding->getName()] = $embedding->getName();
		}
		// Passende System-Embeddings suchen
		$systemEmbeddings = array_filter(SystemEmbedding::all(), function (SystemEmbedding $systemEmbedding) use ($kind, $name, $usedSources) {
			foreach ($systemEmbedding->getAssets() as $asset) {
				if ($asset->getStorageType() === $kind and $asset->matchName($name)) {
					// Nur anzeigen, wenn nicht schon ein Embedding mit dieser Quelle existiert
					if (!in_array([EmbeddingModel::SOURCE_SYSTEM_TEMPLATE, $systemEmbedding->getId()], $usedSources, true)) {
						return true;
					}
				}
			}
			return false;
		});
		if (!$systemEmbeddings) {
			$systemEmbeddings = array_filter(SystemEmbedding::all(), function (SystemEmbedding $systemEmbedding) use ($name, $usedSources) {
				foreach ($systemEmbedding->getAssets() as $asset) {
					if ($asset->matchName($name)) {
						// Nur anzeigen, wenn nicht schon ein Embedding mit dieser Quelle existiert
						if (!in_array([EmbeddingModel::SOURCE_SYSTEM_TEMPLATE, $systemEmbedding->getId()], $usedSources, true)) {
							return true;
						}
					}
				}
				return false;
			});
		}
		if ($userEmbeddings) {
			$systemEmbeddings = array_filter($systemEmbeddings, function (SystemEmbedding $e) use ($userEmbeddingNames) {
				return !isset($userEmbeddingNames[$e->getName()]);
			});
		}

		// Globale DB abfragen, wenn aktiviert
		if (Config::getInstance()->isGlobalCookieDbUsed()) {
			$embeddingsDB = $this->searchGlobalCookieDb($name, $currentLanguageCode);
			$embeddingsDB = array_filter($embeddingsDB, function (array $embedding) use ($name, $usedSources) {
				foreach ($embedding['assets'] ?? [] as $asset) {
					if (EmbeddingAsset::matchAssetName($asset['name'], $name, $asset['dynamic'])) {
						// Nur anzeigen, wenn nicht schon ein Embedding mit dieser Quelle existiert
						if (!in_array([EmbeddingModel::SOURCE_GLOBAL_DB, (string)$embedding['id']], $usedSources, true)) {
							return true;
						}
					}
				}
				return false;
			});
		} else {
			$embeddingsDB = [];
		}

		return $this->render('embedding/add_asset.html.twig', [
			'assetKind' => $kind,
			'assetName' => $name,
			'embeddings' => $embeddings,
			'purposes' => $purposes,
			'userEmbeddingTemplates' => $userEmbeddings,
			'embeddingTemplates' => $systemEmbeddings,
			'dbEmbeddings' => $embeddingsDB,
			'defaultLocale' => LocaleService::getDefault(),
			'defaultUserLocale' => LocaleService::getUserDefault(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/add-{kind<cookie|localStorage|sessionStorage|indexedDb|webSql|other>}/{name}", name="add_asset_post", methods={"POST"})
	 *
	 * @param Request $request
	 * @param string $kind
	 * @param string $name
	 * @param TranslatorInterface $translator
	 * @throws Exception Wenn kein Zugriff auf das Embedding-File erfolgen kann
	 * @return Response
	 */
	public function addAssetAction(Request $request, string $kind, string $name, TranslatorInterface $translator): Response
	{
		$name = urldecode($name);
		$embeddingId = $request->request->getAlnum("add-to-embedding");
		$ok = true;

		try {
			CsrfToken::get()->checkRequest($request);
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
			$ok = false;
			return $this->redirectToRoute('app_embeddings_add_asset', ['kind' => $kind, 'name' => urlencode($name)], 303);
		}

		$embedding = EmbeddingModel::find($embeddingId);
		if (!$embedding) {
			$this->addFlash('danger', 'Unknown embedding. Please try again.');
			$ok = false;
			return $this->redirectToRoute('app_embeddings_add_asset', ['kind' => $kind, 'name' => urlencode($name)], 303);
		}

		foreach ($embedding->getAssets() as $asset) {
			// Embedding enthält dieses Asset schon.
			if ($asset->getName() === $name and $asset->getStorageType() == $kind) {
				$this->addFlash('warning', 'The embedding already contains this cookie.');
				return $this->redirectToRoute('app_embeddings_edit', ['id'=>$embeddingId, '_fragment'=>'tab-assets'], 303);
			}
		}

		$asset = EmbeddingAsset::create()
			->setName($name)
			->setStorageType($kind)
		;
		$embedding->addAsset($asset);
		$embedding->save();
		EmbeddingAsset::flush();

		$this->addFlash('success', $translator->trans('“%name%” was successfully added to this embedding.', ['%name%'=>htmlspecialchars($name)]));
		return $this->redirectToRoute('app_embeddings_edit', ['id'=>$embeddingId, '_fragment'=>'tab-assets'], 303);
	}

	private function searchGlobalCookieDb(?string $search, string $languageCode, bool $filterSystemEmbeddings=true): array
	{
		$defaultLocale = LocaleService::getDefault();
		$defaultLanguage = $defaultLocale->getLanguage() ?? Language::find('de_DE');
		$defaultLanguageCode = $defaultLanguage->getShortCode();

		$embeddingsDB = EmbeddingDb::fetchEmbeddingList($search, Language::find($languageCode) ?? $defaultLanguage);

		if ($filterSystemEmbeddings) {
			$cookieMap = [];
			$templateNames = [];

			$templates = ($search == null) ? SystemEmbedding::all() : SystemEmbedding::search($search);
			foreach ($templates as $template) {
				$templateNames[$template->getName()] = $template->getName();
				foreach ($template->getAssets() as $asset) {
					$cookieMap[$asset->getName()] = $template;
				}
			}

			// User-Templates auch berücksichtigen
			$templates = ($search == null) ? UserEmbedding::all() : UserEmbedding::search($search);
			foreach ($templates as $template) {
				$templateNames[$template->getName()] = $template->getName();
				foreach ($template->getAssets() as $asset) {
					$cookieMap[$asset->getName()] = $template;
				}
			}

			$sysEmbeddingId = Utils::getWhiteLabelEmbedding();

			// Einträge ausfiltern, die schon einen Eintrag in den System-/UserEmbeddings haben
			$embeddingsDB = array_filter($embeddingsDB ?? [], static function ($item) use ($cookieMap, $templateNames,$sysEmbeddingId) {
				
				//Existiert schon ein Embedding mit diesem Namen?
				if (isset($templateNames[$item['name']])) {
					return false;
				}

				//Es handelt sich um das CCM19 Standard Embedding
				if(!empty($sysEmbeddingId) && $item['id']===13308){
					return false;
				}

				foreach ($item['assets'] as $asset) {
					if (!isset($asset['name'])) {
						var_dump($asset);exit();
					}
					if (isset($cookieMap[$asset['name']])) {
						return false;
					}
				}
				return true;
			});
		}

		// Anzeige-Übersetzung auswählen nach aktueller Backend-Sprache, Standard-Locale und Fallback DE.
		foreach ($embeddingsDB as &$embedding) {
			$embedding['translations']['current'] = ($embedding['translations'][$languageCode] ?? [])
				+ ($embedding['translations'][$defaultLanguageCode] ?? [])
				+ $embedding['translations']['de'];
		}
		unset($embedding);

		return $embeddingsDB;
	}

	/**
	 * Embeddings aus der Embedding-Vorlagen und/oder der CCM19.de-Datenbank updaten
	 * @Route("/update", name="update_all", methods={"POST"})
	 * @param EmbeddingUpdateService $updater
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws Exception
	 */
	public function updateAllEmbeddings(EmbeddingUpdateService $updater, TranslatorInterface $translator): Response
	{
		foreach (Embedding::all() as $embedding) {
			// Skip if auto update is disabled
			if (!$embedding->isAutoUpdated()) {
				continue;
			}
			$updater->updateEmbedding($embedding, true);
		}

		EmbeddingAsset::flush();
		Embedding::flush();

		$this->addFlash('success', $translator->trans('All embeddings have been updated.'));
		return $this->redirectToRoute('app_embeddings_list', [], 303);
	}

	/**
	 * Embeddings aus der Embedding-Vorlagen und/oder der CCM19.de-Datenbank updaten
	 * @Route("/update/{id}", name="update", methods={"POST"})
	 * @param string $id
	 * @param EmbeddingUpdateService $updater
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws Exception
	 */
	public function updateEmbedding(string $id, EmbeddingUpdateService $updater, TranslatorInterface $translator): Response
	{
		$embedding = EmbeddingModel::find($id);
		if (!$embedding) {
			return $this->redirectToRoute('app_embeddings_list', [], 303);
		}
		// Abort if auto update is disabled
		elseif (!$embedding->isAutoUpdated()) {
			return $this->redirectToRoute('app_embeddings_list', [], 303);
		}

		if ($updater->updateEmbedding($embedding, true)) {
			EmbeddingAsset::flush();
			Embedding::flush();
			$this->addFlash('success', $translator->trans('The embedding :name has been updated.', [
				':name' => $embedding->getName(),
			]));
		}
		else {
			EmbeddingAsset::reload();
			Embedding::reload();
			$this->addFlash('danger', $translator->trans('The embedding could not be updated.'));
		}

		return $this->redirectToRoute('app_embeddings_list', [], 303);
	}

	/**
	 * Embeddings aus den Embedding-Vorlagen und/oder der CCM19.de-Datenbank auflisten
	 * @Route("/db", name="from_db" )
	 * @param Request $request
	 * @return Response
	 */
	public function listEmbeddingsFromDatabase(Request $request): Response
	{
		return $this->render('embedding/list_db.html.twig', [
		]);
	}

	/**
	 * Diese Methode filtert die Embeddings nach dem Suchbegriff.
	 * @param string $search Suchbegriff
	 * @param array $embeddings Embeddings, die gefiltert werden sollen
	 * @return array
	 */
	public static function searchInArray(string $search, array $embeddings): array
	{
		return array_filter($embeddings, function (array $embedding) use ($search) {
			return stristr($embedding['name'], $search)
				|| 	stristr($embedding['embedding_vendor'] ,$search)
				|| 	stristr($embedding['description'] ,$search)
				|| 	stristr($embedding['purpose'] ,$search)
				|| array_reduce($embedding['assets'], function (bool $match, string $name) use ($search) {
					return $match || stripos($name, $search) !== false;
				}, false)
			;
		});
	}

	/**
	 * Entpackt mit `data-pack="asset,add_asset"` als JSON gepackte POST-Daten
	 */
	private function unpackPostData(array $array): array
	{
		foreach (['asset', 'add_asset'] as $key) {
			if (isset($array[$key]) and is_string($array[$key])) {
				$array[$key] = json_decode($array[$key], true, 64, JSON_INVALID_UTF8_IGNORE);
			}
		}
		return $array;
	}

	/**
	 *  Daten aus dem Array $embeddingData in $embedding übertragen
	 *
	 * @param EmbeddingModel $embedding
	 * @param array $embeddingData
	 * @param string $defaultLocaleId
	 *
	 * @return EmbeddingModel
	 * @throws Exception
	 */
	public function applyChanges(EmbeddingModel $embedding, array $embeddingData, string $defaultLocaleId): EmbeddingModel
	{
		$request = Utils::getCurrentRequest();
		$enabledLocales = LocaleService::enabledLocales();

		// Script-Blocking
		$codeBlock = explode("\n", trim((string)($embeddingData['code_block'] ?? '')));
		$codeBlock = array_filter($codeBlock, function ($line) {
			return trim($line) !== '';
		});
		$codeBlock = array_values(array_map(function ($line) {
			return trim($line);
		}, $codeBlock));

		// Iframes blockieren
		$iframeBlockingStrings = array_filter(array_map(function ($line) {
			return trim($line);
		}, preg_split('~\r\n|\n|\r~', (string)($embeddingData['iframe_blocking_strings'] ?? ''))));

		// Exclude Urls
		$excludeUrls = explode("\n", trim((string)($embeddingData['excludeUrls'] ?? '')));
		$excludeUrls = array_filter($excludeUrls, function ($line) {
			return trim($line) !== '';
		});
		$excludeUrls = array_values(array_map(function ($line) {
			return trim($line);
		}, $excludeUrls));
		// Zweck
		$purpose = Purpose::find($embeddingData['purpose'] ?? '');

		if ($request) {
			$dispatchLoadEventForIframes = $request->isMethodCacheable()
				// GET: Wenn der Parameter nicht vorhanden ist, dann soll der Standardwert verwendet werden
				? (bool)($embeddingData['dispatchLoadEventForIframes'] ?? $embedding->isDispatchLoadEventForIframes())
				// POST/PUT: Wenn der Parameter nicht vorhanden ist, wurde der Schalter im Formular deaktiviert
				: (bool)($embeddingData['dispatchLoadEventForIframes'] ?? false);
		}
		else {
			// Sollte nicht auftreten; Standardwert verwenden
			$dispatchLoadEventForIframes = $embedding->isDispatchLoadEventForIframes();
		}

		// Allgemeines speichern
		$embedding
			->setName($embeddingData['name'] ?? '')
			->setDescription($embeddingData['description'][$defaultLocaleId] ?? '')
			->setCodeBlock($codeBlock)
			->setScriptLoaderGroup($embeddingData['scriptLoaderGroup'] ?? '')
			->setIframeBlockingStrings($iframeBlockingStrings)
			->setIframesHiddenUntilConsent((bool)($embeddingData['hideIframesUntilConsent'] ?? false))
			->setDispatchLoadEventForIframes($dispatchLoadEventForIframes)
			->setExcludeUrls($excludeUrls)
			->setExcludeUrlsMode($embeddingData['excludeUrlsMode'] ?? $embedding->getExcludeUrlsMode())
			->setEmbeddingVendor($embeddingData['embedding_vendor'] ?? '')
			->setPrivacyPolicyUrl($embeddingData['privacy_policy_url'][$defaultLocaleId] ?? '')
			->setHttpEmbeddingName($embeddingData['http_embedding_name'] ?? '')
			->setLifetime($embeddingData['lifetime'][$defaultLocaleId] ?? '')
			->setCollectedDataInfo($embeddingData['collected_data_info'][$defaultLocaleId] ?? '')
			->setPurposeOfDataCollection($embeddingData['purpose_of_data_collection'][$defaultLocaleId] ?? '')
			->setLegalBasis($embeddingData['legal_basis'][$defaultLocaleId] ?? '')
			->setLegalBasisSupplement($embeddingData['legal_basis_supplement'][$defaultLocaleId] ?? '')
			->setLegalBasisStructured((bool)($embeddingData['legal_basis_structured'] ?? false))
			->setStructuredLegalBasis((array)($embeddingData['structured_legal_basis'] ?? []))
			->setPlaceOfProcessing($embeddingData['place_of_processing'][$defaultLocaleId] ?? '')
			->setActive($embeddingData['active'] ?? false)
			->setAutoUpdated($embeddingData['autoUpdated'] ?? false)
			->setUpdateParameterBlacklist((array)($embeddingData['updateParameterBlacklist'] ?? []))
			->updateLastModified()
		;
		if ($purpose) {
			$embedding->setPurpose($purpose);
		}

		$user = User::loggedInUser();

		if (!$user || Utils::mayUserAccessRoute($user, 'meta__app_embeddings__source_code', 'POST')) {
			$embedding->setCodeSnippet($embeddingData['code_snippet'] ?? '');
		}

		if (isset($embeddingData['source'])) {
			$source_ts = ($embeddingData['source_ts'] ?? null) ?: null;
			if ($source_ts and !($source_ts instanceof DateTimeInterface)) {
				$source_ts = new DateTimeImmutable((string)$source_ts);
			}
			if (!$source_ts) {
				$source_ts = null;
			}
			$embedding->setSource($embeddingData['source'] ?? EmbeddingModel::SOURCE_UNKNOWN)
				->setSourceId(!empty($embeddingData['source_id']) ? $embeddingData['source_id'] : null)
				->setSourceTimestamp($source_ts);
		}

		// TCF-Daten speichern
		$embedding->setTcfVendor(
			(is_string($embeddingData['tcf_vendor_id'] ?? null) or $embedding->getTcfVendor())
			? TcfVendor::find($embeddingData['tcf_vendor_id'] ?? $embedding->getTcfVendor()->getId())
			: null);
		$embedding->setTcfPurposes(array_filter(array_map(
			function ($id) {
				return TcfPurpose::find($id);
			},
			(array)($embeddingData['tcf'][$embeddingData['tcf_vendor_id'] ?? '']['p'] ?? [])
		)));
		$embedding->setTcfSpecialFeatures(array_filter(array_map(
			function ($id) {
				return TcfSpecialFeature::find($id);
			},
			(array)($embeddingData['tcf'][$embeddingData['tcf_vendor_id'] ?? '']['sf'] ?? [])
		)));

		// Google Consent Mode speichern
		if (isset($embeddingData['gcm']) and is_array($embeddingData['gcm']) and $embedding->hasVendorGoogle()) {
			$embedding->setGoogleConsentMode($embeddingData['gcm']);
		} else {
			$embedding->setGoogleConsentMode([]);
		}

		// Assets bearbeiten
		foreach ($embeddingData['asset'] ?? [] as $assetId => $data) {
			/** @var EmbeddingAsset $asset */
			$asset = $embedding->getAsset($assetId);
			$assetName = (string)($data['name'] ?? '');

			if ($asset === null || empty($assetName)) {
				continue;
			}

			$asset
				->setName($assetName)
				->setDynamic((bool)($data['dynamic'] ?? false))
				->setStorageType((string)($data['storageType'] ?? ''))
				->setLifetime((string)($data['lifetime'][$defaultLocaleId] ?? ''))
				->setDescription((string)($data['description'][$defaultLocaleId] ?? ''))
			;

			//assets lang Data
			foreach ($enabledLocales as $locale) {
				$localeId = $locale->getName();
				if (!isset($data['description'][$localeId])) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				// @phan-suppress-next-line PhanUndeclaredMethod
				$asset
					->setTranslatedDescription($locale, $data['description'][$localeId] ?? '')
					->setTranslatedLifetime($locale, $data['lifetime'][$localeId] ?? '')
				;
			}
		}
		if(!empty($embeddingData['add_asset'][$defaultLocaleId])) {
			$addAssets = $embeddingData['add_asset'][$defaultLocaleId];
		}
		elseif (is_array($embeddingData['add_asset'] ?? null)) {
			$addAssets = $embeddingData['add_asset'];
		}

		// Assets hinzufügen
		foreach ($addAssets ?? [] as $data) {
			$assetName = (string)($data['name'] ?? '');
			if (empty($assetName)) {
				continue;
			}
			$asset = EmbeddingAsset::create()
				->setName($assetName)
				->setDynamic((bool)($data['dynamic'] ?? false))
				->setStorageType((string)($data['storageType'] ?? ''))
				->setLifetime((string)($data['lifetime'][$defaultLocaleId] ?? ''))
				->setDescription((string)($data['description'][$defaultLocaleId] ?? ''))
			;
			//assets lang Data
			foreach (LocaleService::enabledLocales() as $locale) {
				$localeId = $locale->getName();
				if (!isset($data['description'][$localeId])) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				$asset
					->setTranslatedDescription($locale, $data['description'][$localeId] ?? '')
					->setTranslatedLifetime($locale, $data['lifetime'][$localeId] ?? '');
			}
			//embed now...
			$embedding->addAsset($asset);
		}

		// Assets löschen
		foreach ($embeddingData['delete_asset'] ?? [] as $assetId) {
			$embedding->removeAsset($assetId);
		}

		// Übersetzungen speichern
		foreach ($enabledLocales as $locale) {
			$localeId = $locale->getName();
			if (!isset($embeddingData['description'][$localeId])) {
				continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
			}
			$embedding
				->setTranslatedName($locale, $embeddingData['label'][$localeId] ?? '')
				->setTranslatedDescription($locale, $embeddingData['description'][$localeId] ?? '')
				->setTranslatedPrivacyPolicyUrl($locale, $embeddingData['privacy_policy_url'][$localeId] ?? '')
				->setTranslatedLifetime($locale, $embeddingData['lifetime'][$localeId] ?? '')
				->setTranslatedCollectedDataInfo($locale, $embeddingData['collected_data_info'][$localeId] ?? '')
				->setTranslatedPurposeOfDataCollection($locale, $embeddingData['purpose_of_data_collection'][$localeId] ?? '')
				->setTranslatedLegalBasis($locale, $embeddingData['legal_basis'][$localeId] ?? '')
				->setTranslatedLegalBasisSupplement($locale, $embeddingData['legal_basis_supplement'][$localeId] ?? '')
				->setTranslatedPlaceOfProcessing($locale, $embeddingData['place_of_processing'][$localeId] ?? '')
				->setTranslatedBlockedContentTitle($locale, $embeddingData['blocked_content_title'][$localeId] ?? '')
				->setTranslatedBlockedContentText($locale, $embeddingData['blocked_content_text'][$localeId] ?? '')
				->setTranslatedBlockedContentButtonText($locale, $embeddingData['blocked_content_button_text'][$localeId] ?? '')
			;
		}

		return $embedding;
	}

	/**
	 * @param Request $request
	 * @param LocaleBase $defaultLocale
	 * @return array|string[]
	 */
	private function prepareNewEmbeddingData(Request $request, LocaleBase $defaultLocale): array
	{
		$defaultLocaleName = $defaultLocale->getName();
		$templateId = $request->query->getAlnum('template');
		$userTemplateId = $request->query->getAlnum('user_template');

		// User-Vorlagen
		if ($userTemplateId != '') {
			return $this->createUserEmbedding($userTemplateId, $defaultLocaleName);
		}
		// Embedding-Vorlagen
		elseif ($templateId != '') {
			return $this->createSystemEmbedding($templateId, $defaultLocaleName);
		}
		// Import aus DB nach ID
		elseif (($remoteId = $request->query->getInt('remoteId'))) {
			$remoteDb = (string)$request->query->get('remoteDb', 'v2.2');
			if ($remoteDb === 'v2') {
				$remoteEmbedding = EmbeddingDbLegacy::fetchEmbedding($remoteId);
			} else {
				$remoteEmbedding = EmbeddingDb::fetchEmbedding($remoteId);
			}

			return $this->createRemoteEmbeddingFromDbEntry($remoteEmbedding, $defaultLocale, $remoteDb);
		}
		// Neu erstellen nur mit Name und Asset-Typ
		elseif (($name = (string)$request->query->get('name', '')) !== '') {
			$kind = (string)$request->query->get('kind', 'cookie');

			return $embeddingData = [
				'name' => $name,
				'source' => EmbeddingModel::SOURCE_LOCAL,
				'add_asset' => [[
					'name' => $name,
					'storageType' => $kind,
					'dynamic' => false,
					'lifetime' => [],
					'description' => [],
				]],
			];
		}

		return ['source'=>EmbeddingModel::SOURCE_LOCAL];
	}

	private function createUserEmbedding(string $userTemplateId, string $defaultLocaleName): array
	{
		$template = UserEmbedding::find($userTemplateId);
		if ($template) {
			$tcfVendor = $template->getTcfVendor();
			if ($tcfVendor) {
				$tcf = [
					$tcfVendor->getId() => [
						'p' => array_map(function (TcfPurpose $p) { return $p->getId(); }, $template->getTcfPurposes()),
						'sf' => array_map(function (TcfSpecialFeature $sf) { return $sf->getId(); }, $template->getTcfSpecialFeatures()),
					]
				];
			} else {
				$tcf = [];
			}

			$assets = array_values(array_map(function ($asset) use ($defaultLocaleName) {
				$result = [
					'name' => $asset->getName(),
					'dynamic' => $asset->isDynamic(),
					'storageType' => $asset->getStorageType(),
					'lifetime' => [$defaultLocaleName => $asset->getLifetime()],
					'description' => [$defaultLocaleName => $asset->getDescription()],
				];
				foreach (UserLocale::all() as $locale) {
					$result['lifetime'][$locale->getName()] = $asset->getTranslatedLifetime($locale, false);
					$result['description'][$locale->getName()] = $asset->getTranslatedDescription($locale, false);
				}
				return $result;
			}, $template->getAssets()));

			$result = [
				"name" => $template->getName(),
				'hinweis' => "from_template",
				"code_snippet" => $template->getCodeSnippet(),
				"code_block" => implode("\n", $template->getCodeBlock()),
				"excludeUrls" => implode("\n", $template->getExcludeUrls()),
				'excludeUrlsMode' => $template->getExcludeUrlsMode(),
				"scriptLoaderGroup" => $template->getScriptLoaderGroup(),
				"iframe_blocking_strings" => implode("\n", $template->getIframeBlockingStrings()),
				"hideIframesUntilConsent" => $template->isIframesHiddenUntilConsent(),
				"dispatchLoadEventForIframes" => $template->isDispatchLoadEventForIframes(),
				"description" => [$defaultLocaleName => $template->getDescription()],
				"purpose" => $template->getPurpose()->getId(),
				"privacy_policy_url" => [$defaultLocaleName => $template->getPrivacyPolicyUrl()],
				"embedding_vendor" => $template->getEmbeddingVendor(),
				"tcf_vendor_id" => $tcfVendor ? $tcfVendor->getId() : null,
				"tcf" => $tcf,
				"lifetime" => [$defaultLocaleName => $template->getLifetime()],
				"active" => true,
				"updateParameterBlacklist" => $template->getUpdateParameterBlacklist(),
				"collected_data_info" => [$defaultLocaleName => $template->getCollectedDataInfo()],
				"purpose_of_data_collection" => [$defaultLocaleName => $template->getPurposeOfDataCollection()],
				"legal_basis" => [$defaultLocaleName => $template->getLegalBasis()],
				"structured_legal_basis" => array_map(static function (array $articles) {
					return array_map(static function (Article $article) { return (int)$article->getId(); }, $articles);
				}, $template->getStructuredLegalBasis()),
				"legal_basis_structured" => $template->isLegalBasisStructured(),
				"legal_basis_supplement" => [$defaultLocaleName => $template->getLegalBasisSupplement()],
				"place_of_processing" => [$defaultLocaleName => $template->getPlaceOfProcessing()],
				'add_asset' => $assets,
				'gcm' => $template->getGoogleConsentMode(),
				"source" => EmbeddingModel::SOURCE_USER_TEMPLATE,
				"source_id" => $template->getId(),
				"source_ts" => $template->getLastModified(),
			];

			foreach (UserLocale::all() as $locale) {
				$result['label'][$locale->getName()] = $template->getTranslatedName($locale, false);
				$result['description'][$locale->getName()] = $template->getTranslatedDescription($locale, false);
				$result['privacy_policy_url'][$locale->getName()] = $template->getTranslatedPrivacyPolicyUrl($locale, false);
				$result['lifetime'][$locale->getName()] = $template->getTranslatedLifetime($locale, false);
				$result['collected_data_info'][$locale->getName()] = $template->getTranslatedCollectedDataInfo($locale, false);
				$result['purpose_of_data_collection'][$locale->getName()] = $template->getTranslatedPurposeOfDataCollection($locale, false);
				$result['legal_basis'][$locale->getName()] = $template->getTranslatedLegalBasis($locale, false);
				$result['legal_basis_supplement'][$locale->getName()] = $template->getTranslatedLegalBasisSupplement($locale, false);
				$result['place_of_processing'][$locale->getName()] = $template->getTranslatedPlaceOfProcessing($locale, false);
				$result['blocked_content_title'][$locale->getName()] = $template->getTranslatedBlockedContentTitle($locale, false);
				$result['blocked_content_text'][$locale->getName()] = $template->getTranslatedBlockedContentText($locale, false);
				$result['blocked_content_button_text'][$locale->getName()] = $template->getTranslatedBlockedContentButtonText($locale, false);
			}

			return $result;
		}
		return [];
	}

	private function createSystemEmbedding(string $templateId, string $defaultLocaleName): array
	{
		$template = SystemEmbedding::find($templateId);
		if ($template) {
			$tcfVendor = $template->getTcfVendor();
			if ($tcfVendor) {
				$tcf = [
					$tcfVendor->getId() => [
						'p' => array_map(function (TcfPurpose $p) { return $p->getId(); }, $template->getTcfPurposes()),
						'sf' => array_map(function (TcfSpecialFeature $sf) { return $sf->getId(); }, $template->getTcfSpecialFeatures()),
					]
				];
			} else {
				$tcf = [];
			}

			$assets = array_values(array_map(function ($asset) use ($defaultLocaleName) {
				$result = [
					'name' => $asset->getName(),
					'dynamic' => $asset->isDynamic(),
					'storageType' => $asset->getStorageType(),
					'lifetime' => [$defaultLocaleName => $asset->getLifetime()],
					'description' => [$defaultLocaleName => $asset->getDescription()],
				];
				foreach (LocaleService::all() as $locale) {
					$language = $locale->getLanguage();
					if (!$language) { continue; }
					$result['lifetime'][$locale->getName()] = $asset->getTranslatedLifetime($language, false);
					$result['description'][$locale->getName()] = $asset->getTranslatedDescription($language, false);
				}
				return $result;
			}, $template->getAssets()));

			$result = [
				"name" => $template->getName(),
				'hinweis' => "from_template",
				"code_snippet" => $template->getCodeSnippet(),
				"code_block" => implode("\n", $template->getCodeBlock()),
				"excludeUrls" => implode("\n", $template->getExcludeUrls()),
				'excludeUrlsMode' => $template->getExcludeUrlsMode(),
				"scriptLoaderGroup" => $template->getScriptLoaderGroup(),
				"iframe_blocking_strings" => implode("\n", $template->getIframeBlockingStrings()),
				"hideIframesUntilConsent" => $template->isIframesHiddenUntilConsent(),
				"dispatchLoadEventForIframes" => $template->isDispatchLoadEventForIframes(),
				"description" => [$defaultLocaleName => $template->getDescription()],
				"purpose" => $template->getPurpose()->getId(),
				"privacy_policy_url" => [$defaultLocaleName => $template->getPrivacyPolicyUrl()],
				"embedding_vendor" => $template->getEmbeddingVendor(),
				"tcf_vendor_id" => $tcfVendor ? $tcfVendor->getId() : null,
				"tcf" => $tcf,
				"lifetime" => [$defaultLocaleName => $template->getLifetime()],
				"active" => true,
				"updateParameterBlacklist" => $template->getUpdateParameterBlacklist(),
				"collected_data_info" => [$defaultLocaleName => $template->getCollectedDataInfo()],
				"purpose_of_data_collection" => [$defaultLocaleName => $template->getPurposeOfDataCollection()],
				"legal_basis" => [$defaultLocaleName => $template->getLegalBasis()],
				"structured_legal_basis" => array_map(static function (array $articles) {
					return array_map(static function (Article $article) { return (int)$article->getId(); }, $articles);
				}, $template->getStructuredLegalBasis()),
				"legal_basis_structured" => $template->isLegalBasisStructured(),
				"legal_basis_supplement" => [$defaultLocaleName => $template->getLegalBasisSupplement()],
				"place_of_processing" => [$defaultLocaleName => $template->getPlaceOfProcessing()],
				'add_asset' => $assets,
				'gcm' => $template->getGoogleConsentMode(),
				"source" => EmbeddingModel::SOURCE_SYSTEM_TEMPLATE,
				"source_id" => $template->getId(),
				"source_ts" => $template->getLastModified(),
			];

			foreach (LocaleService::all() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) { continue; }
				$result['label'][$locale->getName()] = $template->getTranslatedName($language, false);
				$result['description'][$locale->getName()] = $template->getTranslatedDescription($language, false);
				$result['privacy_policy_url'][$locale->getName()] = $template->getTranslatedPrivacyPolicyUrl($language, false);
				$result['lifetime'][$locale->getName()] = $template->getTranslatedLifetime($language, false);
				$result['collected_data_info'][$locale->getName()] = $template->getTranslatedCollectedDataInfo($language, false);
				$result['purpose_of_data_collection'][$locale->getName()] = $template->getTranslatedPurposeOfDataCollection($language, false);
				$result['legal_basis'][$locale->getName()] = $template->getTranslatedLegalBasis($language, false);
				$result['legal_basis_supplement'][$locale->getName()] = $template->getTranslatedLegalBasisSupplement($language, false);
				$result['place_of_processing'][$locale->getName()] = $template->getTranslatedPlaceOfProcessing($language, false);
				$result['blocked_content_title'][$locale->getName()] = $template->getTranslatedBlockedContentTitle($language, false);
				$result['blocked_content_text'][$locale->getName()] = $template->getTranslatedBlockedContentText($language, false);
				$result['blocked_content_button_text'][$locale->getName()] = $template->getTranslatedBlockedContentButtonText($language, false);
			}

			return $result;
		}
		return [];
	}

	public function createRemoteEmbeddingFromDbEntry($remoteEmbedding, $defaultLocale, string $remoteDb = 'v2.2'): array
	{
		if (is_array($remoteEmbedding) and $remoteEmbedding) {
			$embeddingData=['hinweis'=>"hinweis_remote"];

			$embeddingData['source'] = ($remoteDb === 'v2') ? EmbeddingModel::SOURCE_GLOBAL_DB_LEGACY : EmbeddingModel::SOURCE_GLOBAL_DB;
			$embeddingData['source_id'] = $remoteEmbedding['id'];
			$embeddingData['source_ts'] = $remoteEmbedding['timestamp'];

			// Upgrade zu neuer DB
			if (isset($remoteEmbedding['@upgrade'])) {
				$upgradeSource = $remoteEmbedding['@upgrade']['source'];
				$upgradeId = $remoteEmbedding['@upgrade']['sourceId'];
				if ($upgradeSource !== $embeddingData['source'] or (string)$upgradeId !== (string)$remoteEmbedding['id']) {
					$embeddingData['source'] = $upgradeSource;
					$embeddingData['source_id'] = $upgradeId;
					if ($upgradeSource === EmbeddingModel::SOURCE_GLOBAL_DB) {
						$remoteEmbedding = EmbeddingDb::fetchEmbedding($upgradeId);
						$embeddingData['source_ts'] = $remoteEmbedding['timestamp'];
					}
				}
			}

			//Standardmäßig auf aktiv - will man in der Regel ja auch...
			$embeddingData["active"] = true;

			// Basisdaten übertragen
			foreach (['name', 'purpose', 'code_snippet', 'code_block', 'embedding_vendor', 'iframe_blocking_strings'] as $key) {
				$embeddingData[$key] = $remoteEmbedding[$key];
			}
			foreach (['label', 'description', 'privacy_policy_url', 'collected_data_info', 'purpose_of_data_collection', 'legal_basis', "legal_basis_supplement", 'place_of_processing'] as $key) {
				$embeddingData[$key] = [];
			}
			if (isset($remoteEmbedding['structured_legal_basis'])) {
				$embeddingData['structured_legal_basis'] = $remoteEmbedding['structured_legal_basis'];
				$embeddingData['legal_basis_structured'] = true;

				// Falls für DSGVO, DSG oder CCPA keine Daten existieren, wird ein leeres Array übergeben
				$regulationIds = array_keys(Regulation::rawAll());
				$embeddingData['structured_legal_basis'] += array_fill_keys($regulationIds, []);
			} else {
				$embeddingData['legal_basis_structured'] = false;
			}
			$embeddingData["autoUpdated"] = true;

			// Namen aus Standardsprache nehmen
			$defaultLanguage = $defaultLocale->getLanguage();
			$defaultLangShort = ($defaultLanguage) ? $defaultLanguage->getShortCode() : 'en';
			if ($defaultLanguage and isset($remoteEmbedding['translations'][$defaultLangShort])) {
				$embeddingData['name'] = $remoteEmbedding['translations'][$defaultLangShort]['label'];
			}

			// Übersetzungen übertragen
			foreach (Language::all() as $languageId => $language) {
				$shortCode = $language->getShortCode();
				if (!empty($remoteEmbedding['translations'][$shortCode]) and !empty($remoteEmbedding['translations'][$shortCode]['description'])) {
					$translation = $remoteEmbedding['translations'][$shortCode];
					foreach ($translation as $key => $value) {
						$embeddingData[$key][$languageId] = $value;
					}
				}
			}

			// Assets übertragen
			$embeddingData['add_asset'] = array_map(function ($asset) {
				$result = [
					'name' => $asset['name'],
					'storageType' => $asset['storageType'],
					'dynamic' => $asset['dynamic'],
					'lifetime' => [],
					'description' => [],
				];

				// Übersetzungen übertragen
				foreach (Language::all() as $languageId => $language) {
					$shortCode = $language->getShortCode();
					if (!empty($asset['translations'][$shortCode]) and !empty($asset['translations'][$shortCode]['lifetime'])) {
						$translation = $asset['translations'][$shortCode];
						foreach ($translation as $key => $value) {
							$result[$key][$languageId] = $value;
						}
					}
				}

				return $result;
			}, $remoteEmbedding['assets']);
			return $embeddingData;
		}
		return [];
	}

	/**
	 * @Route("/new", name="new", methods={"HEAD","GET"})
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function createEmbeddingForm(Request $request, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$csrfToken = CsrfToken::get();
		$purposes = Purpose::all();
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleId = $defaultLocale->getName();

		$embeddingData = $this->prepareNewEmbeddingData($request, $defaultLocale);

		$addVar="add_";

		return $this->render('embedding/new.html.twig', [
			'purposes' => $purposes,
			'locales' => LocaleService::enabledLocales(),
			'csrfToken' => $csrfToken,
			'embedding' => $this->applyChanges(EmbeddingModel::createNonPersistent(), $embeddingData, $defaultLocaleId),
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'canUseIAB' => $user->hasAccountFeature('canUseIAB'),
			'add_var' => $addVar,
		]);
	}

	/**
	 * @Route("/new", name="new_post", methods={"POST"})
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws Exception
	 */
	public function createEmbedding(Request $request, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$action = $request->request->getAlnum('action');
		$csrfToken = CsrfToken::get();
		$domain = Domain::activeDomain();
		$embeddingData = [];
		$purposes = Purpose::all();
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleId = $defaultLocale->getName();

		$embedding = $this->applyChanges(EmbeddingModel::create(), $this->unpackPostData($request->request->all()), $defaultLocaleId);

		try {
			$csrfToken->checkRequest($request);

			// ohne Namen nicht speichern
			if (trim($embedding->getName()) === '') {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Namen gespeichert werden');
				return $this->render('embedding/new.html.twig', [
					'embedding' => $embedding,
					'purposes' => $purposes,
					'locales' => LocaleService::enabledLocales(),
					'csrfToken' => $csrfToken,
				]);
			}

			// ohne Verwendungszweck nicht speichern
			if (!$embedding->getPurpose()) {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Verwendungszweck gespeichert werden');
				return $this->render('embedding/new.html.twig', [
					'embedding' => $embedding,
					'purposes' => $purposes,
					'locales' => LocaleService::enabledLocales(),
					'regulations' => Regulation::all(),
					'regulationGroups' => RegulationGroup::all(),
					'csrfToken' => $csrfToken,
					'canUseIAB' => $user->hasAccountFeature('canUseIAB'),
				]);
			}

			// Speichern
			$embedding->save();
			EmbeddingAsset::flush();

			// Entferne referenzierten Anbieter aus Blacklist
			if ($embedding->isActive() && ($vendor = $embedding->getTcfVendor())) {
				$tcfVendorBlacklist = array_filter($domain->getTcfVendorBlacklist(), function (TcfVendor $item) use ($vendor) {
					return $item->getId() != $vendor->getId();
				});
				$domain
					->setTcfVendorBlacklist($tcfVendorBlacklist)
					->updateTcfStacks()
					->save();
			}

			$this->addFlash('success', 'Das Embedding "' . $embedding->getName() . '" wurde erfolgreich erstellt!');

			switch ($action) {
				case 'saveClose':
					return $this->redirectToRoute('app_embeddings_list', [], 303);
				case 'save':
				default:
					return $this->redirectToRoute('app_embeddings_edit', ['id' => $embedding->getId()], 303);
			}
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		return $this->render('embedding/new.html.twig', [
			'purposes' => $purposes,
			'locales' => LocaleService::enabledLocales(),
			'csrfToken' => $csrfToken,
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'formData' => $embeddingData ?? [],
			'canUseIAB' => $user->hasAccountFeature('canUseIAB'),
			'add_var' => "",
		]);
	}

	/**
	 * @Route("/new_bulk", name="new_bulk", methods={"POST"})
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws Exception
	 */
	public function createEmbeddingBulk(Request $request, TranslatorInterface $translator): Response
	{
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleName = $defaultLocale->getName();

		// Füge ausgewählte User Embeddings hinzu
		$embeddingsUserTemplate = explode(',', $request->request->get('selectedUserTemplates'));
		foreach($embeddingsUserTemplate as $embeddingsUserTemplateId) {
			$userEmbeddingData = $this->createUserEmbedding($embeddingsUserTemplateId, $defaultLocaleName);
			if (!empty($userEmbeddingData)) {
				$userEmbedding = $this->applyChanges(EmbeddingModel::create(), $userEmbeddingData, $defaultLocaleName);
				$userEmbedding->save();
			}
		}

		// Füge ausgewählte System Embeddings hinzu
		$embeddingsSystemTemplate = explode(',', $request->request->get('selectedTemplates'));
		foreach($embeddingsSystemTemplate as $embeddingSystemTemplateId) {
			$systemEmbeddingData = $this->createSystemEmbedding($embeddingSystemTemplateId, $defaultLocaleName);
			if (!empty($systemEmbeddingData)) {
				$systemEmbedding = $this->applyChanges(EmbeddingModel::create(), $systemEmbeddingData, $defaultLocaleName);
				$systemEmbedding->save();
			}
		}

		// Füge ausgewähltes Embedding aus DB hinzu
		$embeddingsExternalDb = $request->request->get('selectedRemoteIds') ? explode(',', $request->request->get('selectedRemoteIds')) : [];
		foreach($embeddingsExternalDb as $remoteId) {
			$remoteDb = (string)$request->query->get('remoteDb', 'v2.2');
			if ($remoteDb === 'v2') {
				$remoteEmbedding = EmbeddingDbLegacy::fetchEmbedding((int)$remoteId);
			} else {
				$remoteEmbedding = EmbeddingDb::fetchEmbedding((int)$remoteId);
			}

			$remoteEmbeddingData = $this->createRemoteEmbeddingFromDbEntry($remoteEmbedding, $defaultLocale, (string)$remoteDb);
			if (!empty($remoteEmbeddingData)) {
				$remoteEmbedding = $this->applyChanges(EmbeddingModel::create(), $remoteEmbeddingData, $defaultLocaleName);
				$remoteEmbedding->save();
			}
		}

		EmbeddingAsset::flush();
		$this->addFlash('success', 'Die ausgewählten Embeddings wurden erfolgreich erstellt!');
		return $this->redirectToRoute('app_embeddings_list', [], 303);
	}

	/**
	 * @Route("/{id}/edit", methods={"HEAD", "GET"}, name="edit")
	 *
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function editEmbeddingForm(string $id, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$embedding = EmbeddingModel::find($id);

		if ($embedding === null) {
			return $this->redirectToRoute('app_embeddings_list');
		}

		if ($embedding->getExcludeUrlsMode() == EmbeddingBase::EXCLUDE_URLS_MODE_WHITELIST && empty($embedding->getExcludeUrls())) {
			$this->addFlash('warning', $translator->trans(
				'The embedding may have been accidentally blocked for all pages. It uses the ":mode" mode in conjunction with an empty filter list for the ":excludeUrls" option.',
				[':mode'=>$translator->trans('Whitelist'), ':excludeUrls'=>$translator->trans('Exclude URLs')]
			));
		}

		$purposes = Purpose::all();

		return $this->render('embedding/edit.html.twig', [
			'embedding' => $embedding,
			'purposes' => $purposes,
			'locales' => LocaleService::enabledLocales(),
			'csrfToken' => CsrfToken::get(),
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'canUseIAB' => $user->hasAccountFeature('canUseIAB'),
			'add_var' => "",
		]);
	}

	/**
	 * Führt Daten aus zwei Embedding-Arrays zusammen.
	 *
	 * Die Einträge des zweiten überschreiben dabei die des ersten.
	 * NULL und leere Strings werden nicht berücksichtigt.
	 *
	 * Sonderbehandlung für 'add_asset' -> 'asset'.
	 *
	 * @param array $data1
	 * @param array $data2
	 * @return array
	 */
	private static function mergeEmbeddingData(array $data1, array $data2): array
	{
		foreach ($data2 as $key => $value)
		{
			if (is_array($value)) {
				if ($key === 'add_asset' and isset($data1['asset'])) {
					foreach ($value as $asset) {
						$name = $asset['name'] ?? '';
						if ($name == '') {
							continue;
						}
						$assetId = null;
						foreach ($data1['asset'] as $assetKey => $assetData) {
							if (($assetData['name'] ?? '') === $name) {
								$assetId = $assetKey;
								break;
							}
						}
						if ($assetId !== null) {
							$data1['asset'][$assetId] = self::mergeEmbeddingData($data1['asset'][$assetId], $asset);
						} else {
							$data1['add_asset'][] = $asset;
						}
					}
				} else {
					$data1[$key] = self::mergeEmbeddingData($data1[$key] ?? [], $data2[$key]);
				}
			} elseif ($value !== null and $value !== '') {
				$data1[$key] = $value;
			}
		}
		return $data1;
	}

	/**
	 * @Route("/{id}/edit", methods={"PUT"}, name="edit_post")
	 *
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws Exception
	 */
	public function editEmbedding(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$action = $request->request->getAlnum('action');
		$domain = Domain::activeDomain();
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleId = $defaultLocale->getName();
		$embedding = EmbeddingModel::find($id);
		$hadTcfVendor = ($embedding->getTcfVendor() !== null);

		if ($embedding === null) {
			return $this->redirectToRoute('app_embeddings_list');
		}

		try {
			CsrfToken::get()->checkRequest($request);

			// Embedding-Daten holen und aufbereiten
			$embeddingData = $this->unpackPostData($request->request->all());

			$purpose = Purpose::find($embeddingData['purpose']);

			// Abbrechen wenn zu editierendes Embedding nicht existiert
			EmbeddingModel::beginWrite();
			if (EmbeddingModel::exists($id) == false) {
				return $this->redirectToRoute('app_embeddings_list');
			}

			// ohne Verwendungszweck nicht speichern
			if (Purpose::exists($embeddingData['purpose']) == false) {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Verwendungszweck gespeichert werden');
				return $this->redirectToRoute('app_embeddings_edit', ['id' => $embedding->getId()]);
			}

			// Script-Blocking
			$codeBlock = explode("\n", trim((string)($embeddingData['code_block'] ?? '')));
			$codeBlock = array_filter($codeBlock, function ($line) {
				return trim($line) !== '';
			});
			$codeBlock = array_values(array_map(function ($line) {
				return trim($line);
			}, $codeBlock));

			// ohne Namen nicht speichern
			if (trim($embeddingData['name']) === '') {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Namen gespeichert werden');
				return $this->redirectToRoute('app_embeddings_edit', ['id' => $embedding->getId()]);
			}

			// Wenn AutoUpdate aktiviert wurde, direkt Daten übernehmen
			if (!$embedding->isAutoUpdated() and ($embeddingData['autoUpdated'] ?? false)) {
				switch ($embeddingData['source'] ?? '') {
					case EmbeddingModel::SOURCE_GLOBAL_DB:
						$request->query->set('remoteId', $embeddingData['source_id'] ?? '');
						$request->query->set('remoteDb', 'v2.2');
						break;
					case EmbeddingModel::SOURCE_GLOBAL_DB_LEGACY:
						$request->query->set('remoteId', $embeddingData['source_id'] ?? '');
						$request->query->set('remoteDb', 'v2');
						break;
					case EmbeddingModel::SOURCE_SYSTEM_TEMPLATE:
						$request->query->set('template', $embeddingData['source_id'] ?? '');
						break;
					case EmbeddingModel::SOURCE_USER_TEMPLATE:
						$request->query->set('user_template', $embeddingData['source_id'] ?? '');
						break;
					default:
						break;
				}
				$remoteData = $this->prepareNewEmbeddingData($request, $defaultLocale);
				$currentSourceTs = new DateTimeImmutable($embeddingData['source_ts']);
				if (!empty($remoteData['source_ts']) and $remoteData['source_ts'] > $currentSourceTs) {
					$embeddingData = $this->mergeEmbeddingData($embeddingData, $remoteData);
				}
			}

			// Speichern
			$embedding = $this->applyChanges($embedding, $embeddingData, $defaultLocaleId);
			$embedding->save();
			EmbeddingAsset::flush();

			// Entferne referenzierten Anbieter aus Blacklist
			if ($embedding->isActive() && ($vendor = $embedding->getTcfVendor())) {
				$tcfVendorBlacklist = array_filter($domain->getTcfVendorBlacklist(), function (TcfVendor $item) use ($vendor) {
					return $item->getId() != $vendor->getId();
				});
				$domain
					->setTcfVendorBlacklist($tcfVendorBlacklist)
					->updateTcfStacks()
					->save();
			}
			elseif ($hadTcfVendor) {
				$domain
					->updateTcfStacks()
					->save();
			}

			$this->addFlash('success', 'Das Embedding "' . $embeddingData['name'] . '" wurde erfolgreich bearbeitet!');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		switch ($action) {
			case 'saveClose':
				return $this->redirectToRoute('app_embeddings_list', [], 303);
			case 'save':
			default:
				return $this->redirectToRoute('app_embeddings_edit', ['id' => $embedding->getId()], 303);
		}
	}

	/**
	 * @Route("/{id}/delete", name="delete")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deleteEmbedding(string $id, Request $request, TranslatorInterface $translator): Response
	{
		EmbeddingModel::beginWrite();

		$embedding = EmbeddingModel::find($id);

		if ($embedding === null) {
			return $this->redirectToRoute('app_embeddings_list');
		}

		$embeddingName = $embedding->getName();

		if ($request->getRealMethod() === 'POST') {
			try {
				CsrfToken::get()->checkRequest($request);

				foreach ($embedding->getAssets() as $asset) {
					EmbeddingAsset::delete($asset->getId(), false);
				}
				EmbeddingAsset::flush();

				EmbeddingModel::delete($id);
				$this->addFlash('success', 'Das Embedding "' . $embeddingName . '" wurde erfolgreich gelöscht!');
				Domain::activeDomain()->updateTcfStacks()->save();
				return $this->redirectToRoute('app_embeddings_list');
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			}
		}

		return $this->render('embedding/delete.html.twig', [
			'user' => User::loggedInUser(),
			'embedding' => $embedding,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{embedding}/gcm", name="gcm_form")
	 * @Route("/gcm", name="gcm_form_new")
	 *
	 * @param EmbeddingModel|null $embedding
	 * @return Response
	 */
	public function gcmForm(?EmbeddingModel $embedding): Response
	{
		return $this->render('embedding/gcm_form.html.twig', [
			'embedding' => $embedding,
		]);
	}

	/**
	 * @Route("/{embedding}/importSource", name="import_source")
	 *
	 * @param EmbeddingModel $embedding
	 * @return JsonResponse
	 */
	public function sourceInfo(EmbeddingModel $embedding): JsonResponse
	{
		$defaultLocale = LocaleService::getDefault();
		$request = new Request();

		switch ($embedding->getSource()) {
			case EmbeddingModel::SOURCE_GLOBAL_DB:
				$request->query->set('remoteId', $embedding->getSourceId());
				$request->query->set('remoteDb', 'v2.2');
				break;
			case EmbeddingModel::SOURCE_GLOBAL_DB_LEGACY:
				$request->query->set('remoteId', $embedding->getSourceId());
				$request->query->set('remoteDb', 'v2');
				break;
			case EmbeddingModel::SOURCE_SYSTEM_TEMPLATE:
				$request->query->set('template', $embedding->getSourceId());
				break;
			case EmbeddingModel::SOURCE_USER_TEMPLATE:
				$request->query->set('user_template', $embedding->getSourceId());
				break;
			default:
				break;
		}

		return $this->json($this->prepareNewEmbeddingData($request, $defaultLocale));
	}
}
