<?php

namespace App\Controller\External;

use App\Component\HttpFoundation\CssResponse;
use App\Component\ThumbnailExtractor\ThumbnailExtractorService;
use App\Component\ThumbnailExtractor\AbstractThumbnailExtractor;
use App\Config;
use App\Controller\ExternalController;
use App\Model\Embedding;
use App\Model\Theme;
use App\Model\User;
use App\Component\LocaleService;
use App\Utils;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\File\Exception\FileNotFoundException;
use Symfony\Component\Routing\Annotation\Route;

class ContentBlocker extends ExternalController
{
	/**
	 * @Route("/x-content-blocked.html", name="app_external_x_content_blocked", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @return Response
	 */
	public function contentBlocked(Request $request, ThumbnailExtractorService $thumbnailer): Response
	{
		$iframeUrl = $request->query->get('url', '');
		$apiKey = $request->query->get('apiKey', '');
		$domainId = $request->query->get('domain', '');

		$config = Config::getInstance();

		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
		$baseUrl = $scheme.$request->getHttpHost();

		$apiParams = [
			'apiKey' => $apiKey,
			'domain' => $domainId,
		];

		/** @var Theme $theme */
		$theme = Theme::deriveFromQuery($request->query);
		if ($theme == null) {
			return new JsonResponse([
				'error' => 'No theme set as active.',
			], 500);
		}

		// Verwende angeforderte Sprache, sonst Standard
		$localeName = (string)$request->query->get('lang', '');
		$locale = LocaleService::find($localeName);
		if (!$locale) {
			$locale = User::loggedInUser()->getDefaultLocale();
		}

		// Thumbnails von IFrame holen, falls möglich
		$embedInfo = $thumbnailer->getEmbedInformation($iframeUrl);
		if ($embedInfo and count($embedInfo->getUsableThumbnails()) == 0) {
			$embedInfo = null;
		}

		$parameters = [
			'theme' => $theme,
			'locale' => $locale,
			'embedInfo' => $embedInfo,
			'embedding' => Embedding::find((string)$request->query->get('embedding')),
			'apiKey' => $apiKey,
			'breakpoints' => [
				'desktop' => '64.0625em',
			],
			'customCssUrl' => $baseUrl.$this->generateUrl('app_external_x_content_blocked_css', $apiParams + [
					'theme' => $theme->getId(),
					'v' => $theme->getLastModifiedTimestamp(),
				]),
		];

		$response = $this->render('external/content-blocked.html.twig', $parameters);
		Utils::resetCacheControl($response);
		$response->setCache([
			'max_age'=>300,
			'public'=>true,
		]);

		$origins = $theme->getCustomCssForIframeBlockerOrigins();
		$originsText = implode(' ', $origins);

		if (!in_array('data:', $origins)) {
			$origins[] = 'data:';
		}
		$imgOriginsText = implode(' ', $origins);

		$response->headers->set('Content-Security-Policy',
			"default-src 'self' $originsText; ".
			"script-src 'self' 'unsafe-inline'; " .
			"style-src 'self' 'unsafe-inline' $originsText; " .
			"img-src 'self' $imgOriginsText; " .
			"frame-src 'self'"
		);
		$response->headers->set('X-Robots-Tag', 'noindex');

		return $response;
	}

	/**
	 * @Route("/x-content-blocked.css", name="app_external_x_content_blocked_css", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @param Request $request
	 * @return CssResponse
	 */
	public function customCss(Request $request): CssResponse
	{
		$response = new CssResponse();

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		// Lebensdauer der CSS-Datei auf ein Jahr setzen, da ein Zeitstempel an die CSS-Route angehängt wird
		$response->headers->remove('Cache-Control');
		$response->setCache([
			'max_age' => 31536000,
			'immutable' => true,
			'public' => true,
		]);

		/** @var Theme $theme */
		$theme = Theme::find((string)$request->query->get('theme'));
		if ($theme == null) {
			return new CssResponse('/* No theme set as active. */', 500);
		}

		$origins = $theme->getCustomCssForIframeBlockerOrigins();
		$originsText = implode(' ', $origins);

		$response->headers->set('Content-Security-Policy',
			"default-src 'self' $originsText; ".
			"script-src 'none'; ".
			"connect-src 'none'; ".
			"form-action 'none'"
		);

		$response->setContent($theme->getCustomCssForIframeBlocker());

		return $response;
	}

	/**
	 * @Route("/iframe-thumbnail/{urlHash}-{width}.jpeg", name="app_external_iframe_thumbnail", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @return Response
	 */
	public function backgroundImage(string $urlHash, int $width): Response
	{
		$path = AbstractThumbnailExtractor::getThumbnailCachePath($urlHash, $width);
		try {
			$response = new BinaryFileResponse($path);
		}
		catch (FileNotFoundException $e) {
			$response = new Response(base64_decode('iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR4nGNiYAAAAAkAAxkR2eQAAAAASUVORK5CYII='), 404, ['content-type'=>'image/png']);
		}

		Utils::resetCacheControl($response);
		$response->setCache([
			'max_age'=>300,
			'public'=>true,
		]);
		$response->headers->set('X-Robots-Tag', 'noindex');
		return $response;
	}
}
