<?php

namespace App\Controller\External;

use App\Component\LoggerInterface;
use App\Component\Tcf\v2\TcfHelper;
use App\Controller\ExternalController;
use App\Model\Cookie;
use App\Model\Domain;
use App\Model\Embedding;
use App\Component\LocaleService;
use App\Model\LocaleBase;
use App\Model\Purpose;
use App\Model\Regulations\Regulation;
use App\Model\Theme;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;

class CookieDeclaration extends ExternalController
{
	/** @var LoggerInterface $logger */
	private $logger;

	/**
	 * @param LoggerInterface $logger
	 */
	public function __construct(LoggerInterface $logger)
	{
		$this->logger = $logger;
	}

	/**
	 * @Route("/cookie-declaration", name="app_external_cookie_declaration", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @param Request $request
	 * @return Response
	 */
	public function cookieDeclaration(Request $request): Response
	{
		/** @var Domain $domain */
		$domain = Domain::activeDomain();
		if ($domain == null) {
			return new JsonResponse(['error' => 'Active domain could not be determined.'], 500);
		}

		// Je nach Einstellung entweder auf Embedding-Implementierung wechseln oder mit Cookies fortfahren
		if ($domain->getManagementStructure() == Domain::MANAGEMENT_STRUCTURE_EMBEDDING) {
			return $this->cookieDeclaration_managementStructureEmbedding($request);
		}

		/** @var Theme $theme */
		$theme = Theme::deriveFromQuery($request->query);
		if ($theme == null) {
			return new JsonResponse(['error' => 'No theme set as active.'], 500);
		}

		$localeName = (string)$request->query->get('lang', '');
		if ($localeName == '') {
			return new JsonResponse(['error' => '`lang` parameter required.'], 422);
		}

		/** @var LocaleBase $locale */
		$locale = LocaleService::findEnabledFuzzy((string)$localeName, true);
		if ($locale == null) {
			return new JsonResponse(['error' => 'Locale not found.'], 500);
		}

		// Nur aktive Cookies berücksichtigen
		$cookies = array_filter(Cookie::all(), function ($cookie) {
			/** @var Cookie $cookie */
			return $cookie->isActive();
		});

		// Purposes anhand selektierter Cookies ermitteln
		$purposes = array_reduce(
			$cookies,
			function ($purposes, $cookie) {
				/** @var Purpose $purpose */
				$purpose = $cookie->getPurpose();

				if ($purpose) {
					$purposes[$purpose->getId()] = $purpose;
				}

				return $purposes;
			},
			[]
		);

		// ORDER BY mandatory, name
		usort($purposes, function ($a, $b) {
			/**
			 * @var Purpose $a
			 * @var Purpose $b
			 */
			if ($a->isMandatory() xor $b->isMandatory()) {
				return $a->isMandatory() ? -1 : 1;
			}
			else {
				return strnatcasecmp($a->getName(), $b->getName());
			}
		});

		// Daten für das Template zusammenstellen
		$parameters = [
			'domain' => $domain,
			'theme' => $theme,
			'locale' => $locale,
			'locales' => LocaleService::enabledLocales(),
			'cookies' => $cookies,
			'purposes' => $purposes,
			'whitelabel' => Utils::hasWhitelabelLicense() && $theme->getWhitelabel(),
			'poweredByText' => Utils::getPoweredByText(),
			'poweredByTooltip' => Utils::getPoweredByTooltip(),
			'poweredByUrl' => Utils::getPoweredByUrl(),
			'h' => $domain->getConsentTableHeadingStartLevel(),
		];

		$response = $this->render('external/cookiedeclaration/index.html.twig', $parameters);

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		// Hier nicht text/html verwenden. Einige Proxies (z.B. in Mobilnetzen)
		// können sonst das unvollständige HTML "reparieren" und Werbung etc.
		// einbauen
		$response->headers->set('Content-Type', 'text/x-html-fragment; charset=utf-8');

		// Robots sollen das nicht crawlen
		$response->headers->set('X-Robots-Tag', 'noindex');

		// Letzten Zeitstempel einer möglichen CSS-Anpassung ermitteln
		$cssLastModifiedTimestamp = (int)max(
			filemtime(realpath(Utils::getBaseDir().'/templates/external/cookie-management.css.twig')),
			$theme->getLastModifiedTimestamp()
		);
		$response->headers->set('Link', '<'.$this->generateUrl('app_external_main_css_file', [
			'apiKey' => $request->query->get('apiKey'),
			'domain' => $request->query->get('domain'),
			'theme' => $theme->getId(),
			'v' => $cssLastModifiedTimestamp
		]).'>;rel="stylesheet"');


		$language = $locale->getLanguage();
		if ($language) {
			$response->headers->set('Content-Language', $language->getIsoCode());
		}

		return $response->setCache([
			'max_age'=>120,
			'public'=>true,
		]);
	}

	/**
	 * @param Request $request
	 * @return Response
	 */
	public function cookieDeclaration_managementStructureEmbedding(Request $request): Response
	{
		/** @var Domain $domain */
		$domain = Domain::activeDomain();
		if ($domain == null) {
			return new JsonResponse(['error' => 'Active domain could not be determined.'], 500);
		}

		/** @var Theme $theme */
		$theme = Theme::find((string)$request->query->get('theme'));
		if ($theme == null) {
			return new JsonResponse(['error' => 'No theme set as active.'], 500);
		}

		$localeName = (string)$request->query->get('lang', '');
		$localeSpecified = strlen($localeName) > 0;

		if (!$localeName) {
			// Locale raten
			$locale = LocaleService::guessFromBrowser($request);
			$localeName = ($locale) ? $locale->getName() : '';
		}

		/** @var LocaleBase $locale */
		$locale = LocaleService::findEnabledFuzzy((string)$localeName, true);
		if ($locale == null) {
			return new JsonResponse(['error' => 'Locale not found.'], 500);
		}

		// Nur aktive Embeddings berücksichtigen
		$embeddings = array_filter(Embedding::all(), function (Embedding $embedding) {
			return $embedding->isActive();
		});

		// Purposes anhand selektierter Cookies ermitteln
		$purposes = array_reduce(
			$embeddings,
			function (array $purposes, Embedding $embedding) {
				/** @var Purpose $purpose */
				$purpose = $embedding->getPurpose();

				if ($purpose) {
					$purposes[$purpose->getId()] = $purpose;
				}

				return $purposes;
			},
			[]
		);

		// ORDER BY mandatory, name
		usort($purposes, function ($a, $b) {
			/**
			 * @var Purpose $a
			 * @var Purpose $b
			 */
			if ($a->isMandatory() xor $b->isMandatory()) {
				return $a->isMandatory() ? -1 : 1;
			}
			else {
				return strnatcasecmp($a->getName(), $b->getName());
			}
		});

		$tcfEnabled = $domain->isTcfVendorsEnabled();
		$extraTcfVendors = $tcfEnabled ? TcfHelper::getExtraTcfVendors($domain, $embeddings) : [];

		// Anzahl der Partner bestimmen
		$embeddingVendorCount = count(Embedding::listActiveThirdPartyVendors());
		$partnerCount = $embeddingVendorCount + count($extraTcfVendors);

		// Daten für das Template zusammenstellen
		$parameters = [
			'domain' => $domain,
			'theme' => $theme,
			'locale' => $locale,
			'locales' => LocaleService::enabledLocales(),
			'embeddings' => $embeddings,
			'purposes' => $purposes,
			'legalBases' => $domain->getEffectiveLegalBases(),
			'regulations' => Regulation::all(),
			'whitelabel' => Utils::hasWhitelabelLicense() && $theme->getWhitelabel(),
			'poweredByText' => Utils::getPoweredByText(),
			'poweredByTooltip' => Utils::getPoweredByTooltip(),
			'poweredByUrl' => Utils::getPoweredByUrl(),
			'partnerCount' => $partnerCount,
			'h' => $domain->getConsentTableHeadingStartLevel(),
		];

		$response = $this->render('external/cookiedeclaration/index_using-embeddings.html.twig', $parameters);

		// Vary-Header für Caching bearbeiten
		if (!$localeSpecified) {
			$vary = $response->getVary();
			$vary[] = 'Accept-Language';
			$response->setVary($vary);
		}

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		// Hier nicht text/html verwenden. Einige Proxies (z.B. in Mobilnetzen)
		// können sonst das unvollständige HTML "reparieren" und Werbung etc.
		// einbauen
		$response->headers->set('Content-Type', 'text/x-html-fragment; charset=utf-8');

		// Robots sollen das nicht crawlen
		$response->headers->set('X-Robots-Tag', 'noindex');

		// Letzten Zeitstempel einer möglichen CSS-Anpassung ermitteln
		$cssLastModifiedTimestamp = (int)max(
			filemtime(realpath(Utils::getBaseDir().'/templates/external/cookie-management.css.twig')),
			$theme->getLastModifiedTimestamp()
		);
		$response->headers->set('Link', '<'.$this->generateUrl('app_external_main_css_file', [
			'apiKey' => $request->query->get('apiKey'),
			'domain' => $request->query->get('domain'),
			'theme' => $theme->getId(),
			'v' => $cssLastModifiedTimestamp
		]).'>;rel="stylesheet"');

		$language = $locale->getLanguage();
		if ($language) {
			$response->headers->set('Content-Language', $language->getIsoCode());
		}

		return $response->setCache([
			'max_age'=>120,
			'public'=>true,
		]);

	}

}
