<?php

namespace App\Controller\External;

use App\Controller\UnrestrictedController;
use App\Component\Tcf\v2\Model\Vendor;
use App\Component\Tcf\v2\Model\Purpose;
use App\Component\Tcf\v2\Model\SpecialPurpose;
use App\Component\Tcf\v2\Model\Feature;
use App\Component\Tcf\v2\Model\SpecialFeature;
use App\Component\Tcf\v2\Model\Translation\Purpose as TranslatedPurpose;
use App\Component\Tcf\v2\Model\Translation\SpecialPurpose as TranslatedSpecialPurpose;
use App\Component\Tcf\v2\Model\Translation\Feature as TranslatedFeature;
use App\Component\Tcf\v2\Model\Translation\SpecialFeature as TranslatedSpecialFeature;
use App\Exception\ConfigAccessException;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;

class Tcf extends UnrestrictedController
{
	/**
	 * @Route("/tcf/vendors.json", name="app_external_tcf_vendors", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @param Request $request
	 * @return JsonResponse
	 */
	public function vendorList(Request $request): JsonResponse
	{
		$response = new JsonResponse(Vendor::rawAll());

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		$response->headers->set('X-Robots-Tag', 'noindex, nofollow');
		$response->setLastModified(Vendor::lastModified());
		$response->setEtag(Vendor::getTcfPolicyVersion().'-'.Vendor::getGvlSpecificationVersion().'-'.Vendor::getVendorListVersion(), true);
		$response->setCache([
			'max_age'=>86400,
			'public'=>true,
		]);
		return $response;
	}

	/**
	 * Übersetzbarer Teil der Vendor/Purpose-List (purposes, special purposes, features, …)
	 *
	 * @Route("/tcf/purposes_{lang<[a-z]{2,3}(_[A-Z]{2})?>}.json", name="app_external_tcf_purposes", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @param Request $request
	 * @return JsonResponse
	 */
	public function purposeList(Request $request, string $lang): JsonResponse
	{
		try {
			$response = new JsonResponse([
				'purposes' => TranslatedPurpose::rawAll($lang),
				'specialPurposes' => TranslatedSpecialPurpose::rawAll($lang),
				'features' => TranslatedFeature::rawAll($lang),
				'specialFeatures' => TranslatedSpecialFeature::rawAll($lang),
			]);
			$response->setLastModified(Purpose::lastModified());
		}
		catch (ConfigAccessException $e) {
			// Fallback zu englischen Originaltexten
			$response = new JsonResponse([
				'purposes' => Purpose::rawAll(),
				'specialPurposes' => SpecialPurpose::rawAll(),
				'features' => Feature::rawAll(),
				'specialFeatures' => SpecialFeature::rawAll(),
			]);
			$response->setLastModified(Purpose::lastModified());
			if ($lang != 'en' and strpos($lang, 'en_') !== 0) {
				$response->headers->set('Link', '<purposes-en.json>;rel="canonical"', false);
				$response->headers->set('Content-Location', 'purposes-en.json');
			}
		}

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		$response->headers->set('X-Robots-Tag', 'noindex, nofollow');
		$response->setCache([
			'max_age'=>86400,
			'public'=>true,
		]);
		return $response;
	}
}
