<?php

namespace App\Controller\External;

use App\Component\Cache\Cache;
use App\Component\HttpFoundation\CssResponse;
use App\Component\HttpFoundation\JavascriptResponse;
use App\Component\LoggerInterface;
use App\Config;
use App\Controller\ExternalController;
use App\Component\LocaleService;
use App\Component\Tcf\v2\TcfHelper;
use App\Component\Tcf\v2\Model\Google\AdTechProvider;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialPurpose as TcfSpecialPurpose;
use App\Component\Tcf\v2\Model\Feature as TcfFeature;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Component\Tcf\v2\Model\DataCategory as TcfDataCategory;
use App\Component\Tcf\v2\Model\Translation\SpecialFeature as TranslatedTcfSpecialFeature;
use App\Component\Tcf\v2\Model\Stack as TcfStack;
use App\Event\FilterEvent;
use App\Event\StatisticsCronJobEvent;
use App\Event\ThemeDisplayEvent;
use App\Event\WidgetPluginsInitEvent;
use App\Component\Mail\MailMessage;
use App\Model\AccessCountJournal;
use App\Model\Cookie;
use App\Model\Domain;
use App\Model\Country;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingBase;
use App\Model\FoundCookie;
use App\Model\FoundScript;
use App\Model\FoundObject;
use App\Model\FoundStorage;
use App\Model\Statistic;
use App\Model\GeoIp;
use App\Model\LegalText;
use App\Model\LocaleBase;
use App\Model\Placeholder;
use App\Model\Purpose;
use App\Model\Regulations\Regulation;
use App\Model\Script;
use App\Model\Theme;
use App\Model\User;
use App\Utils;
use Exception;
use Symfony\Component\Asset\Packages;
use Symfony\Component\HttpFoundation\IpUtils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Mailer\MailerInterface;
use Symfony\Component\Routing\Annotation\Route;
use DateTime;
use Twig\Environment;
use stdClass;

class WidgetEmbedding extends ExternalController
{
	const EVENT_FILTER_APP_JS_PARAMETERS = 'ccm19.app_js.parameters';

	const EVENT_FILTER_WIDGET_PARAMETERS = 'ccm19.widget.parameters';

	const SCRIPT_COUNT_LIMIT = 100;
	const FOUND_REPO_BYTES_LIMIT = 524288;
	const TOUCH_INTERVAL = 21600;

	private $EXCLUDE_LOCALSTORAGE = [
		'_piggy.*',
		'panel-positions_*',
		'jiraBaseUrl',
		't3-*',
		'ec_*_ec_details',
	];
	private $MERGE_LOCALSTORAGE = [
		'optimizely_data$$',
		'st_shares_',
		'at-lojson-cache-',
		'_at.hist.',
		'widget-settings-shopware-',
	];

	/** @var LoggerInterface $logger */
	private $logger;
	/** @var MailerInterface $mailer */
	private $mailer;

	/**
	 * @param LoggerInterface $logger
	 * @param MailerInterface $mailer
	 */
	public function __construct(LoggerInterface $logger, MailerInterface $mailer)
	{
		$this->logger = $logger;
		$this->mailer = $mailer;
	}

	/**
	 * @param Domain $domain
	 * @param EmbeddingBase[] $embeddings
	 * @param LocaleBase|null $locale
	 * @return TcfSpecialFeature[]|TranslatedTcfSpecialFeature[]
	 */
	private static function getActiveTcfSpecialFeatures(Domain $domain, array $embeddings, ?LocaleBase $locale): array
	{
		return array_values(array_reduce(
			$embeddings,
			function (array $features, EmbeddingBase $embedding) use ($locale) {
				// @phan-suppress-next-line PhanUndeclaredMethod
				$currentFeatures = ($locale !== null) ? $embedding->getTranslatedTcfSpecialFeatures($locale) : $embedding->getTcfSpecialFeatures();
				foreach ($currentFeatures as $feature) {
					$features[$feature->getId()] = $features[$feature->getId()] ?? $feature;
				}
				return $features;
			},
			array_reduce(
				$domain->getExplicitTcfSpecialFeatures(),
				function (array $features, TcfSpecialFeature $feature) use ($locale) {
					if ($locale !== null) {
						$feature = $feature->getTranslation($locale->getName());
					}
					$features[$feature->getId()] = $features[$feature->getId()] ?? $feature;
					return $features;
				},
				[]
			)
		));
	}

	/**
	 * @Route("/app.js", name="app_external_main_javascript_file", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @Route("/ccm19.js", name="app_external_main_javascript_file_alias", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @param Request $request
	 * @param Packages $assetManager
	 * @return JavascriptResponse
	 */
	public function mainJavascriptFile(Request $request, Packages $assetManager): JavascriptResponse
	{
		$domain = Domain::activeDomain();
		assert($domain instanceof Domain);

		// Je nach Einstellung entweder auf Embedding-Implementierung wechseln oder mit Cookies fortfahren
		if ($domain->getManagementStructure() == Domain::MANAGEMENT_STRUCTURE_EMBEDDING) {
			return $this->mainJavascriptFile_managementStructureEmbedding($request, $assetManager);
		}

		$config = Config::getInstance();

		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
		$baseUrl = $scheme.$request->getHttpHost();
		$webPath = $request->getBasePath();

		$response = new JavascriptResponse();
		$response->headers->set('X-Robots-Tag', 'noindex, nofollow');

		// Verwende angeforderte Sprache, sofern vorhanden, andernfalls rate das Locale
		$localeName = (string)$request->query->get('lang', '');
		$localeSpecified = strlen($localeName) > 0;

		$locales = LocaleService::enabledLocales();
		$fallbackLocale = User::loggedInUser()->getDefaultLocale();
		$locale = LocaleService::find($localeName);

		// Wenn das Fallback-Locale nicht für diese Domain aktiv ist, ein anderes wählen,
		// mit Priorisierung von Englisch und Deutsch.
		if (!$fallbackLocale->isEnabled() && $locales) {
			$fallbackLocale = LocaleService::anyLocale(LocaleService::DEFAULT_PRIORITY, $locales);
		}

		// Im weiteren Verlauf unbenötigte Variablen aus dem Scope entfernen
		unset($localeName);

		// Wenn ein nicht-existentes Locale angegeben wird, Fallback zum Standard
		if ($localeSpecified and !$locale) {
			$locale = $fallbackLocale;
		}

		if ($localeSpecified and ($locale == null || $locale->isEnabled() == false)) {
			$message = $locale ? "Translation '{$locale->getName()}' disabled" : 'Translation not found';
			return new JavascriptResponse('console.error(\'[CCM19] '.addcslashes($message, '\'').'\');');
		}

		// Benutzer anhand des verwendeten API-Schlüssels ermitteln - siehe AuthenticationHandler
		$apiUser = User::loggedInUser();

		$apiParams = [
			'apiKey' => $apiUser->getApiKey(),
			'domain' => $domain->getId(),
			'gen' => 1,
		];

		if (strlen(Utils::getLicenseKey()) < 10) {
			//aussteigen
			return new JavascriptResponse('console.error("[CCM19] Not licensed / Nicht lizenziert");');
		};

		// Prüfen ob Cronjobs über das Frontend ausgelöst werden sollen
		if (Utils::getCronMode() !== Utils::CRON_MODE_FRONTEND) {
			$cronProbability = 0.0;
		} else {
			// Aus Statistik von gestern schließen, wie oft pro Minute aufgerufen wird, und entsprechend oft
			// den Cronjob laufen lassen, maximal jedes 3. mal und minimal jedes 1000. mal.
			$stat = Statistic::find(date("Y-m-d", strtotime("yesterday noon")));
			if (!$stat) {
				$cronProbability = 0.34;
			} else {
				$cronProbability = max(min(1440 / ($stat->getCallCount()+1), 0.34), 0.001);
			}
		}

		// Anzahl der Aufrufe nach Tarif limitieren
		$userAgent = $request->headers->get('User-Agent', '');
		if (!IpUtils::checkIp($request->getClientIp(), $config->getExemptAddresses()) and Utils::all($config->getExemptUserAgents(), function ($token) use ($userAgent) {
			return strpos($userAgent, $token) === false;
		})) {
			AccessCountJournal::byDomainAndKind($domain, 'call')->increment();
		}
		if ((Utils::isExtendedEdition() && $apiUser->isClient()) || (Utils::isMultiDomainEdition() && $apiUser->isAdministrator())) {
			$this->sendQuotaReport($apiUser);

			if ($apiUser->callCountReached()) {
				if (rand(1, 4) == 1) {
					Utils::dispatchEvent(StatisticsCronJobEvent::NAME, new StatisticsCronJobEvent(false));
				}
				return new JavascriptResponse('console.warn(\'[CCM19] Quota exceeded\');');
			}
		}

		/** @var Theme $theme */
		$theme = Theme::deriveFromQuery($request->query, $apiParams);

		if ($theme == null) {
			return new JavascriptResponse('console.error(\'[CCM19] No theme set as active.\');', 200);
		}

		// Nur aktive Cookies berücksichtigen
		$cookies = array_filter(Cookie::all(), function ($cookie) {
			/** @var Cookie $cookie */
			return $cookie->isActive();
		});

		$activeCookiesNames = array_values(array_map(function ($cookie) {
			/** @var Cookie $cookie */
			return $cookie->getHttpCookieName();
		}, $cookies));

		// Zeitstempel anhand jüngster Aktualisierung diverser Komponenten ermitteln
		$widgetLastModifiedTimestamp = (int)max(
			filemtime(realpath(Utils::getBaseDir().'/templates/external/widget/index.html.twig')),
			Cookie::lastModifiedTimestamp(),
			LegalText::lastModifiedTimestamp(),
			LocaleService::lastModifiedTimestamp(),
			Placeholder::lastModifiedTimestamp(),
			$domain->getFrontendWidgetCacheTimestamp(),
			$theme->getLastModifiedTimestamp()
		);

		$widgetUrls = array_reduce($locales, function ($widgetUrls, $locale) use ($baseUrl, $apiParams, $widgetLastModifiedTimestamp) {
			$widgetUrls[$locale->getName()] = $baseUrl.$this->generateUrl('app_external_widget', $apiParams + [
				'lang' => $locale->getName(),
				'v' => $widgetLastModifiedTimestamp,
			]);
			return $widgetUrls;
		}, []);

		if ($domain->isNewScriptsBlocked()) {
			$allowedScripts = array_values(array_map(function ($script) {
				return $script->getId();
			}, Script::byEnabled(true)));

			$deniedScripts = [];
		}
		else {
			$deniedScripts = array_values(array_map(function ($script) {
				return $script->getId();
			}, Script::byEnabled(false)));

			$allowedScripts = [];
		}

		// Letzten Zeitstempel einer möglichen CSS-Anpassung ermitteln
		$cssLastModifiedTimestamp = (int)max(
			filemtime(realpath(Utils::getBaseDir().'/templates/external/cookie-management.css.twig')),
			$theme->getLastModifiedTimestamp()
		);

		// Zu sperrende Scripte pro Zweck holen
		$blockPurposeScriptMarkers = [];
		foreach ($cookies as $cookie) {
			/** @var Cookie $cookie */
			$purpose = $cookie->getPurpose() ? $cookie->getPurpose()->getId() : null;
			$mandatory = $cookie->getPurpose() ? $cookie->getPurpose()->isMandatory() : true;

			if (!$mandatory) {
				if (!isset($blockPurposeScriptMarkers[$purpose])) {
					$blockPurposeScriptMarkers[$purpose] = $cookie->getCodeBlock();
				}
				else {
					$blockPurposeScriptMarkers[$purpose] = array_merge($blockPurposeScriptMarkers[$purpose], $cookie->getCodeBlock());
				}
			}
		}

		// GeoIP-Überprüfung, wenn aktiviert
		$onlyInEu = $theme->getOnlyInEu();
		$geoDependsOnClientIP = false;
		if ($onlyInEu) {
			$countryCode = $request->headers->get("X-Country", $request->headers->get("CF-IPCountry", ''));
			if ($countryCode) {
				$country = Country::find($countryCode);
				// Vary-Header für Caching bearbeiten
				$vary = $response->getVary();
				$vary[] = 'X-Country';
				if ($request->headers->has('CF-IPCountry')) {
					$vary[] = 'CF-IPCountry';
				}
				$response->setVary($vary);
			} else {
				$geoip = GeoIp::find($request->getClientIp());
				$country = $geoip ? $geoip->getCountry() : null;
				$geoDependsOnClientIP = true;
			}
			$noConsentRequired = $country ? !$country->hasGdprOrSimilarLaw() : false;
		}
		else {
			$noConsentRequired = false;
		}
		$behavior = [
			'respectDoNotTrack' => $theme->isComplyingDoNotTrack(),
			'noConsentRequired' => $noConsentRequired
		];

		$tcfEnabled = $domain->isTcfVendorsEnabled();

		// URLs für Javascript-Dateien, die im public-Verzeichnis zur Einbindung im Frontend liegen, vorbereiten
		$scriptAssets = array_map(
			function ($asset) use ($assetManager, $baseUrl) {
				$timestamp = (int)filemtime(realpath(Utils::getBaseDir().'/public/'.$asset));
				return $baseUrl.$assetManager->getUrl($asset).'?'.http_build_query(['v' => $timestamp]);
			},
			// NOTE: Derzeit ist TCF die einzige Komponente, deshalb kann erstmal alles ausgeschaltet werden.
			($tcfEnabled) ? [
				'js/frontend/ccm19.components.min.js',
			] : []
		);

		$exemptUrls = array_values(array_filter(array_reduce(
			LegalText::havingLocaleEnabled(),
			function ($urls, $legalText) {
				if ($legalText->getImprintScheme() == 'link') {
					$urls[$legalText->getImprintUrl()] = $legalText->getImprintUrl();
				}

				if ($legalText->getPrivacyScheme() == 'link') {
					$urls[$legalText->getPrivacyUrl()] = $legalText->getPrivacyUrl();
				}

				if ($legalText->getAccessibilityScheme() == 'link') {
					$urls[$legalText->getAccessibilityUrl()] = $legalText->getAccessibilityUrl();
				}

				return $urls;
			},
			[]
		)));

		$consentCookieReset = (int)strtotime($domain->getConsentCookieReset());
		if ($domain->hasConsentResetOnRelevantChange()) {
			$consentCookieReset = array_reduce($cookies, function (int $carry, Cookie $item) {
				$dt = $item->getLegalLastModified();
				return max($carry, $dt ? $dt->getTimestamp() : 0);
			}, $consentCookieReset);
		}

		$parameters = [
			'domain' => Domain::activeDomain(),
			'uniqueCookieId' => Utils::generateUniqueCookieId(),
			'theme' => $theme,
			'scripts' => $scriptAssets,
			'styles' => [
				$baseUrl.$this->generateUrl('app_external_main_css_file', $apiParams + [
					'v' => $cssLastModifiedTimestamp,
				]),
			],
			'widgetUrls' => $widgetUrls,
			'statisticsUrl' => $baseUrl.$this->generateUrl('app_external_statistics_consent', $apiParams),
			'consentValidateUrl' => $baseUrl.$this->generateUrl('app_external_statistics_consent_validate', $apiParams),
			'pageCheckUrl' => $baseUrl.$this->generateUrl('app_external_page_check_report', $apiParams),
			'cookieDeclarationUrl' => $baseUrl.$this->generateUrl('app_external_cookie_declaration', $apiParams),
			'contentBlockedUrl' =>
				$baseUrl.$this->generateUrl('app_external_x_content_blocked', $apiParams + [
					'lang' => ($locale ?? $fallbackLocale)->getName(),
				]),
			'cronUrl' => $baseUrl.$this->generateUrl('app_cron', $apiParams),
			'locale' => $locale,
			'fallbackLocale' => $fallbackLocale,
			'ccmOrigin' => $baseUrl,
			'cookieDataActive' => $activeCookiesNames,
			'allowedScripts' => $allowedScripts,
			'deniedScripts' => $deniedScripts,
			'blockPurposeScriptMarkers' => $blockPurposeScriptMarkers,
			'behavior' => $behavior,
			'isWidgetPreview' => $request->query->get('preview') == 'true',
			'consentCookieReset' => $consentCookieReset,
			'exemptUrls' => $exemptUrls,
			'cronProbability' => $cronProbability,
			'consoleLogsEnabled' => $domain->hasCodeMinification() == false,
		];

		$response = $this->render('external/cookie-management.js.twig', $parameters, $response);

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);
		$response->headers->set('Cache-Control', 'max-age=0, public, stale-while-revalidate=3600, stale-if-error=3600');

		if ($onlyInEu) {
			// Für externes Caching: Header, ob Version für die EU oder nicht, wenn Banner nur in der EU angezeigt werden soll
			$response->headers->set('X-Target-Region', $noConsentRequired ? '!EU' : 'EU');
			// Caching auf Privat setzen, wenn GeoIP von der IP bestimmt wurde
			if ($geoDependsOnClientIP) {
				$response->headers->set('Cache-Control', 'max-age=0, s-maxage=0, private');
			}
		}
		foreach ($parameters['styles'] as $styleUrl) {
			$response->headers->set('Link', sprintf('<%s>; rel="preload";as="style";nopush', $styleUrl), false);
		}
		if (count($widgetUrls) === 1) {
			$response->headers->set('Link', sprintf('<%s>;rel="preload";as="fetch";crossorigin="anonymous";nopush', reset($widgetUrls)), false);
		}
		foreach ($scriptAssets as $scriptUrl) {
			$response->headers->set('Link', sprintf('<%s>; rel="preload";as="script";nopush', $scriptUrl), false);
		}
		// @phan-suppress-next-line PhanTypeMismatchReturnSuperType
		return $response;
	}

	/**
	 * @param Request $request
	 * @param Packages $assetManager
	 * @return JavascriptResponse
	 */
	public function mainJavascriptFile_managementStructureEmbedding(Request $request, Packages $assetManager): JavascriptResponse
	{
		$config = Config::getInstance();

		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
		$baseUrl = $scheme.$request->getHttpHost();
		$webPath = $request->getBasePath();

		$response = new JavascriptResponse();
		$response->headers->set('X-Robots-Tag', 'noindex, nofollow');
		$response->headers->set('Content-Type', 'application/javascript;charset=utf-8');

		// Verwende angeforderte Sprache, sofern vorhanden
		if ($localeSpecified = $request->query->has('lang')) {
			$localeName = (string)$request->query->get('lang', '');
		}
		else {
			$localeName = null;
		}

		$fallbackLocale = User::loggedInUser()->getDefaultLocale();
		$locales = LocaleService::enabledLocales();
		$locale = ($localeName !== null) ? LocaleService::find($localeName) : null;

		// Wenn das Fallback-Locale nicht für diese Domain aktiv ist, ein anderes wählen,
		// mit Priorisierung von Englisch und Deutsch.
		if (!$fallbackLocale->isEnabled() && $locales) {
			$fallbackLocale = LocaleService::anyLocale(LocaleService::DEFAULT_PRIORITY, $locales);
		}

		// Im weiteren Verlauf unbenötigte Variablen aus dem Scope entfernen
		unset($localeName);

		// Wenn ein nicht-existentes Locale angegeben wird, Fallback zum Standard
		if ($localeSpecified and !$locale) {
			$locale = $fallbackLocale;
		}

		// Fehlermeldung, wenn bis hier hin ein deaktiviertes Locale gewählt wurde
		if ($localeSpecified and ($locale == null or $locale->isEnabled() == false)) {
			$message = $locale ? "Translation '{$locale->getName()}' disabled" : 'Translation not found';
			return new JavascriptResponse('console.error(\'[CCM19] '.addcslashes($message, '\'').'\');');
		}

		// Benutzer anhand des verwendeten API-Schlüssels ermitteln - siehe AuthenticationHandler
		$apiUser = User::loggedInUser();
		if ($apiUser->isSubUser())
		{
			$apiUser = $apiUser->getAccountOwner();
		}
		$domain = Domain::activeDomain();
		assert($domain !== null);

		$apiParams = [
			'apiKey' => $apiUser->getApiKey(),
			'domain' => $domain->getId(),
			'gen' => 2,
		];

		if (strlen(Utils::getLicenseKey()) < 10) {
			//aussteigen
			return new JavascriptResponse('console.error("[CCM19] Not licensed / Nicht lizenziert");');
		};

		// Rate zur Durchführung von Page-Checks berechnen
		$statFetched = false;
		$stat = null;
		if ($domain->isActivatePermaScan()) {
			$stat = Statistic::find(date("Y-m-d", strtotime("yesterday noon")));
			$statFetched = true;
			if (!$stat) {
				$reportProbability = 0.25;
			} else {
				// Aus Statistik von gestern schließen, wie oft pro Minute aufgerufen wird, und entsprechend oft
				// den Page-Check laufen lassen, maximal jedes 4. mal und minimal jedes 1000. mal.
				$reportProbability = max(min(1440 / ($stat->getCallCount()+1), 0.25), 0.001);
			}
		} else {
			$reportProbability = 0;
		}

		// Prüfen ob Cronjobs über das Frontend ausgelöst werden sollen
		if (Utils::getCronMode() !== Utils::CRON_MODE_FRONTEND) {
			$cronProbability = 0.0;
		} else {
			if (!$statFetched) {
				$stat = Statistic::find(date("Y-m-d", strtotime("yesterday noon")));
				$statFetched = true;
			}
			// Aus Statistik von gestern schließen, wie oft pro Minute aufgerufen wird, und entsprechend oft
			// den Cronjob laufen lassen, maximal jedes 3. mal und minimal jedes 1000. mal.
			if (!$stat) {
				$cronProbability = 0.34;
			} else {
				$cronProbability = max(min(1440 / ($stat->getCallCount()+1), 0.34), 0.001);
			}
		}

		// Anzahl der Aufrufe nach Tarif limitieren
		$userAgent = $request->headers->get('User-Agent', '');
		if (!IpUtils::checkIp($request->getClientIp(), $config->getExemptAddresses()) and Utils::all($config->getExemptUserAgents(), function ($token) use ($userAgent) {
			return strpos($userAgent, $token) === false;
		})) {
			AccessCountJournal::byDomainAndKind($domain, 'call')->increment();
		}
		if ((Utils::isExtendedEdition() && $apiUser->isClient()) || (Utils::isMultiDomainEdition() && $apiUser->isAdministrator()) ) {
			$this->sendQuotaReport($apiUser);

			if ($apiUser->callCountReached()) {
				if (rand(1, 4) == 1) {
					Utils::dispatchEvent(StatisticsCronJobEvent::NAME, new StatisticsCronJobEvent(false));
				}
				return new JavascriptResponse('console.warn(\'[CCM19] Quota exceeded\');');
			}
		}

		$theme = Theme::deriveFromQuery($request->query, $apiParams);

		if ($theme == null) {
			return new JavascriptResponse('console.error(\'[CCM19] No theme set as active.\');', 200);
		}

		// Mit Event durch Plugins verarbeitbar machen
		$event = new ThemeDisplayEvent($theme, $request->query->getAlnum('theme') ?: null, $request->query->getBoolean('preview'));
		Utils::dispatchEvent(ThemeDisplayEvent::NAME, $event);

		// Endgültiges Theme zurückgeben
		$theme = $event->getTheme();
		$apiParams['theme'] = $theme->getId();

		// Nur aktive Embeddings berücksichtigen
		$embeddings = array_filter(Embedding::all(), function (Embedding $embedding) {
			return $embedding->isActive();
		});

		$tcfEnabled = $domain->isTcfVendorsEnabled();
		$tcfLegIntEnabled = $domain->isTcfLegitimateInterestEnabled();

		$blacklistedTcfVendorIds = array_map(function (TcfVendor $vendor) {
			return $vendor->getId();
		}, $domain->getTcfVendorBlacklist());

		$relevantTcfVendors = $tcfEnabled ?
			array_filter(TcfVendor::all(), function ($v) use ($blacklistedTcfVendorIds) {
				return ! in_array($v->getId(), $blacklistedTcfVendorIds);
			})
			: [];

		// Dieses Array wird JSON-kodiert im Template ausgegeben, um den Zugriff über Javascript zu erleichtern
		$embeddingRepository = array_map(static function (Embedding $embedding) use (
			$domain, $fallbackLocale, $locales, &$relevantTcfVendors, $tcfEnabled
		) {
			$purpose = $embedding->getPurpose();
			$purposeId = $purpose ? $purpose->getId() : null;
			$mandatory = $purpose ? $purpose->isMandatory() : true;
			$tcfVendor = $embedding->getTcfVendor();

			if ($tcfEnabled && $tcfVendor) {
				$relevantTcfVendors[$tcfVendor->getId()] = $tcfVendor;
				$tcfData = [
					$tcfVendor->getId() => [
						'p' => array_map(function (TcfPurpose $p): int {
							return (int)$p->getId();
						}, $embedding->getTcfPurposes()),
						'sf' => array_values(array_map(function (TcfSpecialFeature $sf) {
							return (int)$sf->getId();
						}, $embedding->getTcfSpecialFeatures())),
					]
				];
			}
			else {
				$tcfData = null;
			}

			$result = [
				'id' => $embedding->getId(),
				'name' => $embedding->getName(),
				'vendor' => Placeholder::applyAll($embedding->getEmbeddingVendor(), false),
				'code' => $embedding->getCodeSnippet(),
				'iframeBlockingStrings' => $embedding->getIframeBlockingStrings(),
				'hideIframesUntilConsent' => $embedding->isIframesHiddenUntilConsent(),
				'dispatchLoadEventForIframes' => $embedding->isDispatchLoadEventForIframes(),
				'scriptLoaderGroup' => $embedding->getScriptLoaderGroup(),
				'excludeUrls' => $embedding->getexcludeUrls(),
				'excludeUrlsMode' => $embedding->getExcludeUrlsMode(),
				'purpose' => $purposeId,
				'mandatory' => $mandatory,
				'assets' => array_map(static function (EmbeddingAsset $asset) use ($domain, $locales) {
					$result = [
						'id' => $asset->getId(),
						'name' => $asset->getName(),
						'type' => $asset->getStorageType(),
						'glob' => $asset->isDynamic(),
						'lifetime' => $asset->getLifetime(),
						'hours' => $asset->getLifetimeHours(),
					];

					if ($domain->isPartOfAdAndTrackingFreeNetwork()) {
						// The l10n object is passed through in the JavaScript API.
						// Do not change the structure, because ad and tracking free networks rely on it.
						$result['l10n'] = array_combine(
							array_map(static function (LocaleBase $locale) { return $locale->getName(); }, $locales),
							array_map(static function (LocaleBase $locale) use ($asset) {
								return [
									'lifetime' => $asset->getTranslatedLifetime($locale, false),
									'description' => $asset->getTranslatedDescription($locale, false),
								];
							}, $locales)
						);
						uksort($result['l10n'], 'strnatcasecmp');
					}

					return $result;
				}, $embedding->getAssets()),
				'tcf' => $tcfData,
			];

			$gcm = $embedding->getGoogleConsentMode();
			if ($gcm) {
				$result['gcm'] = $gcm;
			}

			if ($domain->isPartOfAdAndTrackingFreeNetwork()) {
				// The l10n object is passed through in the JavaScript API.
				// Do not change the structure, because ad and tracking free networks rely on it.
				$result['l10n'] = array_combine(
					array_map(static function (LocaleBase $locale) { return $locale->getName(); }, $locales),
					array_map(static function (LocaleBase $locale) use ($embedding, $purpose) {
						return [
							'name' => $embedding->getTranslatedName($locale, false),
							'description' => $embedding->getTranslatedDescription($locale, false),
							'purpose' => $purpose ? $locale->getPurposeName($purpose) : '',
							'privacyPolicyUrl' => $embedding->getTranslatedPrivacyPolicyUrl($locale, false),
							'collectedDataInfo' => $embedding->getTranslatedCollectedDataInfo($locale, false),
							'purposeOfDataCollection' => $embedding->getTranslatedPurposeOfDataCollection($locale, false),
							'legalBasis' => $embedding->getTranslatedLegalBasis($locale, false),
							'placeOfProcessing' => $embedding->getTranslatedPlaceOfProcessing($locale, false),
						];
					}, $locales)
				);
				uksort($result['l10n'], 'strnatcasecmp');
			}

			return $result;
		}, $embeddings);

		// How many embeddings are there that aren't the CCM19 embedding?
		$nonCmpEmbeddingCount = array_reduce($embeddings, function (int $count, Embedding $embedding) {
			$isCmpEmbedding = array_reduce($embedding->getAssets(), function (bool $carry, EmbeddingAsset $asset) {
				return $carry || $asset->getName() == 'ccm_consent';
			}, false);

			return $isCmpEmbedding ? $count : $count + 1;
		}, 0);

		$tcfVendorData = array_map(function (TcfVendor $v) {
			$mapPurposeId = function (TcfPurpose $p): int {
				return (int)$p->getId();
			};
			$flexTcfPurposeIds = $v->getFlexiblePurposeIds();
			$liTcfPurposeIds = array_map($mapPurposeId, $v->getLegitimateInterestPurposes());
			$consentTcfPurposeIds = array_map($mapPurposeId, $v->getPurposes());
			$specialPurposeIds = array_map(function (TcfSpecialPurpose $p): int { return (int)$p->getId(); }, $v->getSpecialPurposes());

			return [
				'n' => $v->getName(),
				'p' => array_values($consentTcfPurposeIds),
				'pF' => array_values($flexTcfPurposeIds),
				'pL' => array_values($liTcfPurposeIds),
				'sp' => array_values($specialPurposeIds),
			];
		}, $relevantTcfVendors);

		$tcfCompatVersion = $domain->getTcfCompatVersion();

		// Zeitstempel anhand jüngster Aktualisierung diverser Komponenten ermitteln
		$widgetLastModifiedTimestamp = (int)max(
			filemtime(realpath(Utils::getBaseDir().'/templates/external/widget/index_using-embeddings.html.twig')),
			Embedding::lastModifiedTimestamp(),
			LegalText::lastModifiedTimestamp(),
			LocaleService::lastModifiedTimestamp(),
			Placeholder::lastModifiedTimestamp(),
			$domain->getFrontendWidgetCacheTimestamp(),
			$theme->getLastModifiedTimestamp()
		);

		$widgetUrls = array_reduce($locales, function ($widgetUrls, $locale) use ($baseUrl, $apiParams, $widgetLastModifiedTimestamp) {
			$widgetUrls[$locale->getName()] = $baseUrl.$this->generateUrl('app_external_widget', $apiParams + [
				'lang' => $locale->getName(),
				'v' => $widgetLastModifiedTimestamp,
			]);
			return $widgetUrls;
		}, []);

		$widgetDetailsUrls = array_reduce($locales, function ($urls, LocaleBase $locale) use ($baseUrl, $apiParams, $widgetLastModifiedTimestamp) {
			$urls[$locale->getName()] = $baseUrl.$this->generateUrl('app_external_widget_details', $apiParams + [
				'lang' => $locale->getName(),
				'v' => $widgetLastModifiedTimestamp,
			]);
			return $urls;
		}, []);

		if ($domain->isNewScriptsBlocked()) {
			$allowedScripts = array_values(array_map(function ($script) {
				return $script->getId();
			}, Script::byEnabled(true)));

			$deniedScripts = [];
		}
		else {
			$deniedScripts = array_values(array_map(function ($script) {
				return $script->getId();
			}, Script::byEnabled(false)));

			$allowedScripts = [];
		}

		// Letzten Zeitstempel einer möglichen CSS-Anpassung ermitteln
		$cssLastModifiedTimestamp = (int)max(
			filemtime(realpath(Utils::getBaseDir().'/templates/external/cookie-management.css.twig')),
			$domain->getFrontendWidgetCacheTimestamp(),
			$theme->getLastModifiedTimestamp()
		);

		// Sammle zu sperrende Skripte pro Embedding
		$blockEmbeddingScriptMarkers = [];
		foreach ($embeddings as $embedding) {
			/** @var Embedding $embedding */
			$mandatory = $embedding->getPurpose() ? $embedding->getPurpose()->isMandatory() : true;

			if (!$mandatory) {
				$blockEmbeddingScriptMarkers[$embedding->getId()] = array_merge(
					$blockEmbeddingScriptMarkers[$embedding->getId()] ?? [],
					$embedding->getCodeBlock()
				);
			}
		}

		// GeoIP-Überprüfung, wenn aktiviert
		$onlyInEu = $theme->getOnlyInEu();
		$geoDependsOnClientIP = false;
		if ($onlyInEu) {
			$countryCode = $request->headers->get("X-Country", $request->headers->get("CF-IPCountry", ''));
			if ($countryCode) {
				$country = Country::find($countryCode);
				// Vary-Header für Caching bearbeiten
				$vary = $response->getVary();
				$vary[] = 'X-Country';
				if ($request->headers->has('CF-IPCountry')) {
					$vary[] = 'CF-IPCountry';
				}
				$response->setVary($vary);
			} else {
				$geoip = GeoIp::find($request->getClientIp());
				$country = $geoip ? $geoip->getCountry() : null;
				$geoDependsOnClientIP = true;
			}
			$noConsentRequired = $country ? !$country->hasGdprOrSimilarLaw() : false;
		}
		else {
			$noConsentRequired = false;
		}
		$behavior = [
			'respectDoNotTrack' => $theme->isComplyingDoNotTrack(),
			'noConsentRequired' => $noConsentRequired
		];

		// URLs für Javascript-Dateien, die im public-Verzeichnis zur Einbindung im Frontend liegen, vorbereiten
		$scriptAssets = array_map(
			function ($asset) use ($assetManager, $baseUrl) {
				$timestamp = (int)filemtime(realpath(Utils::getBaseDir().'/public/'.$asset));
				return $baseUrl.$assetManager->getUrl($asset).'?'.http_build_query(['v' => $timestamp]);
			},
			// NOTE: Derzeit ist TCF die einzige Komponente, deshalb kann erstmal alles ausgeschaltet werden.
			($tcfEnabled) ? [
				'js/frontend/ccm19.components.min.js',
			] : []
		);

		// URLs für CSS-Dateien, die das Banner zusätzlich einbinden soll (auch app.css)
		$styleAssets = [];

		$hasBaseCss = !$theme->getBaseLayout()->isEmpty();
		$hasCustomCss = strlen(trim($theme->getCustomCss())) > 0;

		if ($hasBaseCss || $hasCustomCss) {
			$styleAssets[] = $baseUrl . $this->generateUrl('app_external_main_css_file', $apiParams + [
				'v' => $cssLastModifiedTimestamp,
			]);
		}

		$exemptUrls = array_values(array_filter(array_reduce(
			LegalText::havingLocaleEnabled(),
			function ($urls, $legalText) {
				if ($legalText->getImprintScheme() == 'link' && $legalText->isShowWidgetOnImprintPage() == false) {
					$urls[$legalText->getImprintUrl()] = $legalText->getImprintUrl();
				}

				if ($legalText->getPrivacyScheme() == 'link' && $legalText->isShowWidgetOnPrivacyPage() == false) {
					$urls[$legalText->getPrivacyUrl()] = $legalText->getPrivacyUrl();
				}

				if ($legalText->getAccessibilityScheme() == 'link') {
					$urls[$legalText->getAccessibilityUrl()] = $legalText->getAccessibilityUrl();
				}

				return $urls;
			},
			$domain->getWidgetExclusionUrls()
		)));

		// Consent-Stichtag bestimmen
		$consentCookieReset = (int)strtotime($domain->getConsentCookieReset());
		if ($domain->hasConsentResetOnRelevantChange()) {
			$consentCookieReset = array_reduce($embeddings, function (int $carry, Embedding $item) {
				$dt = $item->getLegalLastModified();
				return max($carry, $dt ? $dt->getTimestamp() : 0);
			}, $consentCookieReset);
		}

		// Collect plugin info
		$widgetPluginsInitEvent = new WidgetPluginsInitEvent($domain, $theme, $locale, $apiParams, $baseUrl, $widgetLastModifiedTimestamp);
		Utils::dispatchEvent(WidgetPluginsInitEvent::NAME, $widgetPluginsInitEvent);

		if ($domain->hasCrossDomainSharing()) {
			$consentSharingDomains = array_merge([$domain->getName()], $domain->getShareDomainNames());
			$consentSharingDomains = array_unique($consentSharingDomains);
			natcasesort($consentSharingDomains);
			$consentSharingDomains = array_values($consentSharingDomains);
		}
		else {
			$consentSharingDomains = array_values($domain->getShareDomainNames());
		}

		if ($apiUser->isSendFullConsentAndRejectionAcrossDomains()) {
			$accountWideConsentSharingDomains = call_user_func_array('array_merge', array_values(array_map(function (Domain $domain) {
				return array_merge([$domain->getName()], $domain->getShareDomainNames());
			}, Domain::all())));
			$accountWideConsentSharingDomains = array_unique($accountWideConsentSharingDomains);
			natcasesort($accountWideConsentSharingDomains);
			$accountWideConsentSharingDomains = array_values($accountWideConsentSharingDomains);
		}
		else {
			$accountWideConsentSharingDomains = [];
		}

		$countryCode = $domain->getCountryCode();
		if (!$countryCode) {
			$countryCode = $apiUser->getDefaultServiceCountry()->getCode();
		}

		$activeTcfPurposes = [];
		$activeTcfLegIntPurposes = [];
		if ($tcfEnabled) {
			array_walk($embeddings, function (Embedding $embedding) use (&$activeTcfPurposes, &$activeTcfLegIntPurposes) {
				$tcfVendor = $embedding->getTcfVendor();
				if ($tcfVendor) {
					$legIntPurposeIds = array_map(static function (TcfPurpose $purpose) {
						return $purpose->getId();
					}, $tcfVendor->getLegitimateInterestPurposes());
					foreach ($embedding->getTcfPurposes() as $purpose) {
						$purposeId = $purpose->getId();
						$activeTcfPurposes[$purposeId] = $purpose;
						if (in_array($purposeId, $legIntPurposeIds)) {
							$activeTcfLegIntPurposes[$purposeId] = $purpose;
						}
					}
				}
			});
			array_walk($relevantTcfVendors, function (TcfVendor $vendor) use (&$activeTcfPurposes, &$activeTcfLegIntPurposes, $tcfLegIntEnabled) {
				$legIntPurposeIds = array_map(static function (TcfPurpose $purpose) {
					return $purpose->getId();
				}, $vendor->getLegitimateInterestPurposes());
				$purposes = array_merge(
					$vendor->getPurposes(),
					$tcfLegIntEnabled ? $vendor->getLegitimateInterestPurposes() : $vendor->getFlexiblePurposes()
				);
				foreach ($purposes as $purpose) {
					$purposeId = $purpose->getId();
					$activeTcfPurposes[$purposeId] = $purpose;
					if (in_array($purposeId, $legIntPurposeIds)) {
						$activeTcfLegIntPurposes[$purposeId] = $purpose;
					}
				}
			});
			ksort($activeTcfPurposes);
			ksort($activeTcfLegIntPurposes);
		}

		$parameters = [
			'versionId' => Utils::getVersionId(),
			'versionName' => Utils::getVersionName(),
			'domainId' => $domain->getId(),
			'widgetUrls' => $widgetUrls,
			'detailsUrls' => $widgetDetailsUrls,
			'locale' => $locale ? $locale->getName() : null,
			'fallbackLocale' => $fallbackLocale->getName(),
			'prioritizeLangAttribute' => !$domain->isPrioritizeLanguageHeaderOverLangAttribute(),
			'l10n' => array_combine(
				array_map(function (LocaleBase $locale) {
					return $locale->getName();
				}, $locales),
				array_map(function (LocaleBase $locale) {
					return [
						'blockedContentAllowEmbedding' => $locale->getBlockedContentAllowEmbedding(),
						'blockedContentProviderDetails' => $locale->getBlockedContentProviderDetails(),
						'blockedContentShowExternalContent' => $locale->getBlockedContentShowExternalContent(),
					];
				}, $locales)
			),
			'statisticsUrl' => $baseUrl.$this->generateUrl('app_external_statistics_consent', $apiParams),
			'pageCheckUrl' => $baseUrl.$this->generateUrl('app_external_page_check_report', $apiParams),
			'cronUrl' => $cronProbability > 0.0 ? $baseUrl.$this->generateUrl('app_cron', $apiParams) : '',
			'origin' => $baseUrl,
			'jsFiles' => $scriptAssets,
			'cssFiles' => $styleAssets,
			'plugins' => array_values($widgetPluginsInitEvent->getPluginInfo()),
			'isActive' => (
				(
					$domain->isFrontendWidgetEnabled()
					and ($config->isShowFrontendWidgetOnlyIfEmbeddingsExist() == false or $nonCmpEmbeddingCount > 0)
				)
				or $request->query->get('preview') == 'true'
			),
			'isPreview' => $request->query->get('preview') == 'true',
			'triggeringExtGtmEvents' => $domain->isTriggeringExtGtmEvents(),
			'widgetDelayInMilliseconds' => $domain->getWidgetDelayInMilliseconds(),
			'deleteAllCookiesConfig' => $domain->getDeleteUnknownCookies(),
			'deleteAllCookiesReload' => $domain->getDeleteUnknownCookiesForceReload(),
			'runMandatoryEmbeddingsWithoutConsent' => $domain->isLoadTechnicallyNecessaryEmbeddingsWithoutConsent(),
			'cookieLifetime' => $domain->getConsentCookieLifetime(),
			'cookieUseSecure' => $domain->hasSecureConsentCookie(),
			'consentStorageMethod' => $domain->getConsentStorage(),
			'consentShareDomains' => $consentSharingDomains,
			'consentCrossDomain' => $domain->hasCrossDomainSharing(),
			'accountWideConsentSharingDomains' => $accountWideConsentSharingDomains,
			'accountWideConsentSharing' => $apiUser->isSendFullConsentAndRejectionAcrossDomains(),
			'cookieDeclarationUrl' => $baseUrl.$this->generateUrl('app_external_cookie_declaration', $apiParams),
			'storageNamespace' => $domain->getUrlPath(),
			'storageNamespaceUrlParameter' => $domain->getConsentStorageNamespaceUrlParameter(),
			'settingsIconEnabled' => $theme->isSettingsIconEnabled(),
			'settingsIconTarget' => $theme->getSettingsIconTarget(),
			'repository' => $embeddingRepository ?: new stdClass(),
			'tcfData' => [
				'enabled' => $tcfEnabled,
				'legIntEnabled' => $tcfLegIntEnabled,
				'compatVersion' => $tcfCompatVersion,
				'purposes' => $activeTcfPurposes,
				'legIntPurposes' => $tcfLegIntEnabled ? $activeTcfLegIntPurposes : [],
				'specialFeatures' => array_reduce(
					self::getActiveTcfSpecialFeatures($domain, $embeddings, $locale),
					function (array $features, $feature) {
						$features[$feature->getId()] = $feature;
						return $features;
					},
					[]
				),
				'stacks' => array_values(array_map(function (TcfStack $stack) {
					return (int)$stack->getId();
				}, array_filter($domain->getTcfStacks(), function ($x) { return $x instanceof TcfStack; }))),
				'vendor' => $tcfVendorData,
				'gvlVersion' => TcfVendor::getGvlSpecificationVersion(),
				'vlVersion'  => TcfVendor::getVendorListVersion(),
				'tcfVersion' => TcfVendor::getTcfPolicyVersion(),
				'googleACMode' => $domain->hasGoogleAdditionalConsentMode(),
				'googleATPs' => array_reduce($domain->getGoogleAdditionalConsentWhitelist(), function (array $providers, AdTechProvider $provider) {
					$providers[$provider->getId()] = $provider;
					return $providers;
				}, []),
				'disclosureUrlTemplate' => $baseUrl.$this->generateUrl('app_external_tcf_disclosures', $apiParams + [
					'vendor' => ':VENDOR',
					'lang' => ':LANG'
				]),
				'adTags' => $domain->isTcfAdTagsEnabled(),
				'adTagsAutoAppend' => $domain->hasTcfAdTagsAutoAppend(),
			],
			'behavior' => $behavior,
			'partOfAdAndTrackingFreeNetwork' => $domain->isPartOfAdAndTrackingFreeNetwork(),
			'jsApiFeatures' => [
				'embeddingConsentState' => $domain->isAllowEmbeddingConsentViaJsApi(),
			],
			'evalInlineScripts' => $domain->isEvalInlineScripts(),
			'blockIframes' => $domain->isIframesBlocked() && $apiUser->hasAccountFeature('canUseIframes'),
			'rememberIframeConsent' => $domain->isRememberIframeConsentPerDomain(),
			'appendConsentSwitchToIframes' => $domain->isAppendConsentSwitchToIframes(),
			'iframeBlockMode' => $domain->getIframeBlockMode(),
			'blockNewScripts' => $domain->isNewScriptsBlocked(),
			'blockSameDomainScripts' => $domain->isSameDomainScriptsBlocked(),
			'blockInlineScripts' => $domain->isInlineScriptsBlocked(),
			'blockEmbeddingScriptMarkers' => $blockEmbeddingScriptMarkers ?: new stdClass(),
			'deniedScripts' => $deniedScripts,
			'allowedScripts' => $allowedScripts,
			'allowedScriptMarkers' => $domain->getAllowedScriptMarkers(),
			'deniedScriptMarkers' => $domain->getDeniedScriptMarkers(),
			'iframeMarkers' => $domain->getIframeMarkers(),
			'country' => $countryCode,
			'exemptUrls' => $exemptUrls,
			'scripTracker' => $domain->isScriptsTrack(),
			'mustMatchDomains' => $domain->isOnlyActiveOnMatchingDomain(),
			'mainMatchDomain' => $domain->getName(),
			'detectCloakedScripts' => $domain->isDetectCloakedScripts(),
			'selfUrlPattern' => explode('?', $request->getUri(), 2)[0] . '*',
			'consts' => [
				'MODE_WHITELIST' => EmbeddingBase::EXCLUDE_URLS_MODE_WHITELIST,
				'TYPE_COOKIE' => EmbeddingAsset::STORAGE_TYPE_COOKIE,
				'TYPE_LOCAL_STORAGE' => EmbeddingAsset::STORAGE_TYPE_LOCAL_STORAGE,
				'TYPE_SESSION_STORAGE' => EmbeddingAsset::STORAGE_TYPE_SESSION_STORAGE,
			],
			'contentBlockedUrl' =>
				$baseUrl.$this->generateUrl('app_external_x_content_blocked', $apiParams + [
					'lang' => ($locale ?? $fallbackLocale)->getName(),
					'v' => $widgetLastModifiedTimestamp,
				]),
			'cronProbability' => $cronProbability,
			'reportProbability' => $reportProbability,
			'manipulationPrevention' => $theme->getManipulationPrevention(),
			'showFrontendWidgetOnlyIfEmbeddingsExist' => $config->isShowFrontendWidgetOnlyIfEmbeddingsExist(),
			'nonCmpEmbeddingCount' => $nonCmpEmbeddingCount,
			'consentCookieReset' => $consentCookieReset,
			'consentLifetime' => $domain->getConsentLifetime(),
			'consentLifetimeForPartialConsent' => $domain->getConsentLifetimeForPartialConsent(),
		];

		// Sonderfall Clickskeks
		if (Utils::isClickskeksCloud()) {
			$parameters['clickskeksLegacy'] = [
				'allowedConfigChangeCallbacks' => [],
			];
		}

		$apiUrl ="";
		foreach ($apiParams as $k=>$v)
		{
			$apiUrl.=$k."=".$v."&";
		}
		//Theme und Version anhängen
		$apiUrl.="theme=".$theme->getId()."&v=".$cssLastModifiedTimestamp;

		//Parameter an ExtraCSS Links hängen
		$extraCss = $domain->getExtraCss();
		if (is_array($extraCss))
		{
			foreach ($extraCss as $k=>$v)
			{
				$extraCss[$k]=$v.="?".$apiUrl;
			}
		}
		//Parameter an ExtraJS Links hängen
		$extraJs = $domain->getExtraJS();
		if (is_array($extraJs))
		{
			foreach ($extraJs as $k=>$v)
			{
				$extraJs[$k]=$v.="?".$apiUrl;
			}
		}

		// Zusätzliches CSS von Plugins laden
		$parameters['cssFiles'] = array_merge($parameters['cssFiles'],array_values($extraCss));

		// Zusätzliches JS von Plugins laden
		$parameters['jsFiles'] = array_merge($parameters['jsFiles'],array_values($extraJs));

		// Parameter durch Core-Komponenten und Plugins jagen
		$parametersFilterEvent = new FilterEvent($parameters);
		Utils::dispatchEvent(self::EVENT_FILTER_APP_JS_PARAMETERS, $parametersFilterEvent);
		$parameters = $parametersFilterEvent->getReturn();

		// Parameter kodieren und Basis-Datei wählen
		$injections = [
			'_CCM19_CONFIG_' => json_encode($parameters, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE),
		];
		if ($domain->hasCodeMinification()) {
			$baseFile = Utils::getBaseDir() . '/public/js/frontend/ccm19.core.min.inc.js';
		} else {
			$baseFile = Utils::getBaseDir() . '/public/js/frontend/ccm19.core.inc.js';
			$injections['"use strict";'] = '"use strict"; var CCM_DEBUG=true;';
		}

		// Response-Inhalt bauen
		$response->setContent(strtr(
			file_get_contents($baseFile),
			$injections
		));

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);
		// `must-revalidate` ist nicht nötig. `max-age=0` reicht aus, um zuverlässig Calls zu zählen,
		// aber hilft bei Verbindungsproblemen den Browser-Cache weiterzuverwenden.
		// Die `stale-*`-Erweiterungen aus RFC-5861 erlauben einem Cache noch eine Stunde nach dem
		// Abruf eine Hintergrund-Validierung durchzuführen, aber die gecachte Version währenddessen
		// und bei 5xx-Fehlern zu verwenden.
		$response->headers->set('Cache-Control', 'max-age=0, public, stale-while-revalidate=3600, stale-if-error=3600');

		if ($onlyInEu) {
			// Für externes Caching: Header, ob Version für die EU oder nicht, wenn Banner nur in der EU angezeigt werden soll
			$response->headers->set('X-Target-Region', $noConsentRequired ? '!EU' : 'EU');
			// Caching auf Privat setzen, wenn GeoIP von der IP bestimmt wurde
			if ($geoDependsOnClientIP) {
				$response->headers->set('Cache-Control', 'max-age=0, s-maxage=0, private');
			}
		}
		foreach ($parameters['cssFiles'] as $styleUrl) {
			$response->headers->set('Link', sprintf('<%s>;rel="preload";as="style";nopush', $styleUrl), false);
		}
		if (count($widgetUrls) === 1) {
			$response->headers->set('Link', sprintf('<%s>;rel="preload";as="fetch";crossorigin="anonymous";nopush', reset($widgetUrls)), false);
		}
		foreach ($scriptAssets as $scriptUrl) {
			$response->headers->set('Link', sprintf('<%s>;rel="preload";as="script";fetchpriority="low";nopush', $scriptUrl), false);
		}

		// Only cache requests that have valid parameters
		$cacheable = $request->query->get('preview') != 'true'
			&& ($request->query->has('lang') == false || ($locale && (string)$request->query->get('lang') == $locale->getName()))
			&& (
				!$request->query->has('systemTheme') && !$request->query->has('userTheme') && !$request->query->has('theme')
				|| $theme->getSystemTheme() && (string)$request->query->get('systemTheme') == $theme->getSystemTheme()->getId()
				|| $theme->getUserTheme() && (string)$request->query->get('userTheme') == $theme->getUserTheme()->getId()
				|| (string)$request->query->get('theme') == $theme->getId()
			)
		;

		if ($domain->isWidgetCachingEnabled() && $cacheable) {
			// Allow requests that are using these timestamps to be cached
			Cache::registerVersionIdentifier($domain, $widgetLastModifiedTimestamp);
			Cache::registerVersionIdentifier($domain, $cssLastModifiedTimestamp);

			Cache::storeResponse($request, $response, $domain, [
				'lang' => $locale ? $locale->getName() : null, // Explicitly specify the language
				'noConsentRequired' => $noConsentRequired,
			], 43200); // Set cache lifetime to enforce the call limit etc. at a frequent interval
		}

		// @phan-suppress-next-line PhanTypeMismatchReturnSuperType
		return $response;
	}

	/**
	 * @Route("/app.css", name="app_external_main_css_file", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @Route("/ccm19.css", name="app_external_main_css_file_alias", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @param Request $request
	 * @return CssResponse
	 */
	public function mainCssFile(Request $request): CssResponse
	{
		$response = new CssResponse();
		$domain = Domain::activeDomain();
		if ($domain === null) {
			return new CssResponse('/* Active domain could not be determined. */', 500);
		}

		/** @var Theme $theme */
		$theme = Theme::deriveFromQuery($request->query);

		if ($theme === null) {
			return new CssResponse('/* No theme set as active. */', 500);
		}

		// Letzten Zeitstempel einer möglichen CSS-Anpassung ermitteln
		$cssLastModifiedTimestamp = (int)max(
			filemtime(realpath(Utils::getBaseDir().'/templates/external/cookie-management.css.twig')),
			$domain->getFrontendWidgetCacheTimestamp(),
			$theme->getLastModifiedTimestamp()
		);

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		// Lebensdauer der CSS-Datei auf ein Jahr setzen, da ein Zeitstempel an die CSS-Route angehängt wird
		$response->headers->remove('Cache-Control');
		$response->setLastModified(new DateTime('@'.$cssLastModifiedTimestamp));
		$response->setCache([
			'max_age' => 31536000,
			'immutable' => true,
			'public' => true,
		]);

		$origins = $theme->getCustomCssOrigins();
		$originsText = implode(' ', $origins);
		$response->headers->set('Content-Security-Policy',
			"default-src 'self' $originsText; ".
			"script-src 'none'; ".
			"connect-src 'none'; ".
			"form-action 'none'"
		);

		$parameters = [
			'theme' => $theme,
			'breakpoints' => [
				'mobileMax' => '40em',
				'tabletMax' => '64em',
				'desktop' => '64.0625em',
			],
		];

		/** @var CssResponse $response */
		$response = $this->render('external/cookie-management.css.twig', $parameters, $response);

		if ($domain->isWidgetCachingEnabled() && $request->query->get('preview') != 'true') {
			// Store response in CacheHandler to include CORS headers
			Cache::storeResponseAddToQueue($request, $response, $domain);
		}

		// @phan-suppress-next-line PhanTypeMismatchReturnSuperType
		return $response;
	}

	/**
	 * @Route("/widget", name="app_external_widget", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @param Request $request
	 * @return Response
	 */
	public function widget(Request $request): Response
	{
		/** @var Domain $domain */
		$domain = Domain::activeDomain();
		if ($domain == null) {
			return new JsonResponse(['error' => 'Active domain could not be determined.'], 500);
		}

		// Je nach Einstellung entweder auf Embedding-Implementierung wechseln oder mit Cookies fortfahren
		if ($domain->getManagementStructure() == Domain::MANAGEMENT_STRUCTURE_EMBEDDING) {
			return $this->widget_managementStructureEmbedding($request);
		}

		/** @var Theme $theme */
		$theme = Theme::deriveFromQuery($request->query);

		if ($theme === null) {
			return new JsonResponse(['error' => 'No theme set as active.'], 500);
		}

		/** @var LocaleBase $locale */
		$locale = LocaleService::find((string)$request->query->get('lang'));
		if ($locale === null) {
			return new JsonResponse(['error' => 'Locale not found.'], 500);
		}

		/** @var LegalText $legalText */
		$legalText = LegalText::byLocale($locale);
		if ($legalText === null) {
			return new JsonResponse(['error' => 'Legal texts aren\'t translated.'], 500);
		}

		if ($request->headers->get('X-CCM19State') == 'new-view') {
			AccessCountJournal::byDomainAndKind($domain, 'view')->increment();
		}

		// Nur aktive Cookies berücksichtigen
		$cookies = array_filter(Cookie::all(), function ($cookie) {
			/** @var Cookie $cookie */
			return $cookie->isActive();
		});

		// Purposes anhand selektierter Cookies ermitteln
		$purposes = array_reduce(
			$cookies,
			function ($purposes, $cookie) {
				/** @var Purpose $purpose */
				$purpose = $cookie->getPurpose();

				if ($purpose) {
					$purposes[$purpose->getId()] = $purpose;
				}

				return $purposes;
			},
			[]
		);

		// ORDER BY mandatory, name
		usort($purposes, function ($a, $b) {
			/**
			 * @var Purpose $a
			 * @var Purpose $b
			 */
			if ($a->isMandatory() xor $b->isMandatory()) {
				return $a->isMandatory() ? -1 : 1;
			}
			else {
				return strnatcasecmp($a->getName(), $b->getName());
			}
		});

		// Dieses Array wird JSON-kodiert im Template ausgegeben, um den Zugriff über Javascript zu erleichtern
		$scriptRepository = array_map(function ($cookie) {
			/** @var Cookie $cookie */
			$purpose = $cookie->getPurpose() ? $cookie->getPurpose()->getId() : null;
			$mandatory = $cookie->getPurpose() ? $cookie->getPurpose()->getId() : null;

			return [
				'name' => $cookie->getHttpCookieName(),
				'code' => $cookie->getCodeSnippet(),
				'purpose' => $cookie->getPurpose()->getId(),
				'mandatory' => $cookie->getPurpose()->isMandatory(),
			];
		}, $cookies);

		// GeoIP-Überprüfung, wenn benötigt
		$geoDependsOnClientIP = false;
		$geoDependsOnXCountry = false;
		if ($theme->hasDeclineButtonRestrictions()) {
			$regionRestrictions = $theme->getDeclineButtonRestrictions();
			$countryCode = $request->headers->get("X-Country", $request->headers->get("CF-IPCountry", ''));
			if ($countryCode) {
				$country = Country::find($countryCode);
				$geoDependsOnXCountry = true;
			} else {
				$geoip = GeoIp::find($request->getClientIp());
				$country = $geoip ? $geoip->getCountry() : null;
				$countryCode = $country ? $country->getCode() : '';
				$geoDependsOnClientIP = true;
			}
			if (!$country) {
				$targetRegion = 'EU';
				$showDeclineButton = true;
			} elseif ($country->hasGdpr()) {
				$targetRegion = 'EU';
				$showDeclineButton = in_array('EU', $regionRestrictions);
			} elseif (in_array($countryCode, Theme::RESTRICTABLE_COUNTRIES)) {
				$targetRegion = $countryCode;
				$showDeclineButton = in_array($countryCode, $regionRestrictions);
			} else {
				$targetRegion = implode(',', array_merge(
					["!EU"],
					array_map(function ($code) { return "!".$code; }, Theme::RESTRICTABLE_COUNTRIES)
				));
				$showDeclineButton = in_array('world', $regionRestrictions);
			}
		}
		else {
			$targetRegion = null;
			$showDeclineButton = true;
		}

		$config = Config::getInstance();
		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
		$baseUrl = $scheme.$request->getHttpHost();

		// Daten für das Template zusammenstellen
		$parameters = [
			'baseUrl' => $baseUrl,
			'domain' => $domain,
			'theme' => $theme,
			'locale' => $locale,
			'locales' => LocaleService::enabledLocales(),
			'legalText' => $legalText,
			'cookies' => $cookies,
			'purposes' => $purposes,
			'scriptRepository' => $scriptRepository,
			'whitelabel' => Utils::hasWhitelabelLicense() && $theme->getWhitelabel(),
			'poweredByText' => Utils::getPoweredByText(),
			'poweredByTooltip' => Utils::getPoweredByTooltip(),
			'poweredByUrl' => Utils::getPoweredByUrl(),
			'showDeclineButton' => $showDeclineButton,
		];

		$response = $this->render('external/widget/index.html.twig', $parameters);

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		if ($targetRegion !== null) {
			$response->headers->set("X-Target-Region", $targetRegion);
		}

		if ($geoDependsOnClientIP) {
			// Lebensdauer kurz und private, wenn es von GeoIP abhängt
			$response->setCache([
				'max_age' => 30,
				'immutable' => true,
				'private' => true,
			]);
		} else {
			if ($geoDependsOnXCountry) {
				// Vary-Header für Caching bearbeiten
				$vary = $response->getVary();
				$vary[] = 'X-Country';
				if ($request->headers->has('CF-IPCountry')) {
					$vary[] = 'CF-IPCountry';
				}
				$response->setVary($vary);
			}
			// Lebensdauer auf ca. einen Monat setzen, da ein Zeitstempel an die Widget-Route angehängt wird
			$response->setCache([
				'max_age' => 2592000,
				'immutable' => true,
				'public' => true,
			]);
		}

		// Hier nicht text/html verwenden. Einige Proxies (z.B. in Mobilnetzen)
		// können sonst das unvollständige HTML "reparieren" und Werbung etc.
		// einbauen
		$response->headers->set('Content-Type', 'text/x-html-fragment; charset=utf-8');

		// Robots sollen das Widget nicht crawlen
		$response->headers->set('X-Robots-Tag', 'noindex');

		$language = $locale->getLanguage();
		if ($language) {
			$response->headers->set('Content-Language', $language->getIsoCode());
		}

		return $response;
	}

	/**
	 * @param Request $request
	 * @return Response
	 */
	public function widget_managementStructureEmbedding(Request $request): Response
	{
		/** @var Domain $domain */
		$domain = Domain::activeDomain();
		if ($domain == null) {
			return new JsonResponse(['error' => 'Active domain could not be determined.'], 500);
		}

		$apiParams = [];
		$theme = Theme::deriveFromQuery($request->query, $apiParams);

		if ($theme == null) {
			return new JsonResponse(['error' => 'No theme set as active.'], 500);
		}

		/** @var LocaleBase $locale */
		$locale = LocaleService::find((string)$request->query->get('lang'));
		if ($locale == null) {
			return new JsonResponse(['error' => 'Locale not found.'], 500);
		}

		/** @var LegalText $legalText */
		$legalText = LegalText::byLocale($locale);
		if ($legalText == null) {
			return new JsonResponse(['error' => 'Legal texts aren\'t translated.'], 500);
		}

		$user = User::loggedInUser();
		assert($user instanceof User);

		$owner = $user->getAccountOwner();

		// Nur aktive Embeddings berücksichtigen
		$embeddings = array_filter(Embedding::all(), function (Embedding $embedding) {
			return $embedding->isActive();
		});

		// Purposes anhand selektierter Cookies ermitteln
		$purposeUsageCounts = array_fill_keys(array_keys(Purpose::all()), 0);
		$purposes = array_reduce(
			$embeddings,
			function (array $purposes, Embedding $embedding) use (&$purposeUsageCounts) {
				/** @var Purpose $purpose */
				$purpose = $embedding->getPurpose();

				if ($purpose) {
					$purposeId = $purpose->getId();
					$purposes[$purposeId] = $purpose;
					$purposeUsageCounts[$purposeId]++;
				}

				return $purposes;
			},
			[]
		);

		$tcfEnabled = $domain->isTcfVendorsEnabled();
		$tcfLegIntEnabled = $domain->isTcfLegitimateInterestEnabled();
		$activeTcfSpecialFeatures = self::getActiveTcfSpecialFeatures($domain, $embeddings, $locale);
		$extraTcfVendors = $tcfEnabled ? TcfHelper::getExtraTcfVendors($domain, $embeddings) : [];
		$activeTcfPurposes = [];
		$tcfUsageCounts = [
			'purposes' => [
				'all' => [],
				'legitimate_interest' => [],
			],
			'special_features' => [],
			'vendors' => [
				'all' => 0, // Includes embedding vendors if legitimate interest is enabled
				'legitimate_interest' => 0,
			],
		];

		/**
		 * @param TcfVendor $vendor
		 * @param TcfPurpose[] $purposes
		 * @param TcfSpecialFeature[] $features
		 * @param bool $isEmbeddingVendor
		 * @return void
		 */
		$addTcfVendorStats = static function (TcfVendor $vendor, array $purposes, array $features, bool $isEmbeddingVendor=false)
		use (&$tcfUsageCounts, &$activeTcfPurposes, $tcfLegIntEnabled) {
			$legIntPurposeIds = array_map(static function (TcfPurpose $purpose) {
				return $purpose->getId();
			}, $vendor->getLegitimateInterestPurposes());

			foreach ($purposes as $purpose) {
				$purposeId = $purpose->getId();
				if (!isset($activeTcfPurposes[$purposeId])) {
					$activeTcfPurposes[$purposeId] = $purpose;
					$tcfUsageCounts['purposes']['all'][$purposeId] = 1;
					$tcfUsageCounts['purposes']['legitimate_interest'][$purposeId] = 0;
				} else {
					$tcfUsageCounts['purposes']['all'][$purposeId]++;
				}

				if (in_array($purposeId, $legIntPurposeIds)) {
					$tcfUsageCounts['purposes']['legitimate_interest'][$purposeId]++;
				}
			}

			foreach ($features as $feature) {
				$featureId = $feature->getId();
				if (!isset($tcfUsageCounts['special_features'][$featureId])) {
					$tcfUsageCounts['special_features'][$featureId] = 1;
				} else {
					$tcfUsageCounts['special_features'][$featureId]++;
				}
			}

			if ($tcfLegIntEnabled || !$isEmbeddingVendor) {
				$tcfUsageCounts['vendors']['all']++;
			}
			if ($legIntPurposeIds) {
				$tcfUsageCounts['vendors']['legitimate_interest']++;
			}
		};

		if ($tcfEnabled) {
			array_walk($embeddings, function (Embedding $embedding) use ($addTcfVendorStats) {
				$tcfVendor = $embedding->getTcfVendor();
				if ($tcfVendor) {
					$purposes = $embedding->getTcfPurposes();
					$features = $embedding->getTcfSpecialFeatures();
					$addTcfVendorStats($tcfVendor, $purposes, $features, true);
				}
			});
			array_walk($extraTcfVendors, function (TcfVendor $vendor) use ($addTcfVendorStats, $tcfLegIntEnabled) {
				$purposes = array_merge(
					$vendor->getPurposes(),
					$tcfLegIntEnabled ? $vendor->getLegitimateInterestPurposes() : $vendor->getFlexiblePurposes()
				);
				$features = $vendor->getSpecialFeatures();
				$addTcfVendorStats($vendor, $purposes, $features);
			});
			ksort($activeTcfPurposes);

			$adTechProviders = $domain->hasGoogleAdditionalConsentMode() ? $domain->getGoogleAdditionalConsentWhitelist() : [];
			if ($adTechProviders) {
				$tcfUsageCounts['vendors']['all'] += count($adTechProviders);
			}
		}

		// Anzahl der Partner bestimmen
		$embeddingVendorCount = count(Embedding::listActiveThirdPartyVendors());
		$partnerCount = $embeddingVendorCount + count($extraTcfVendors);

		// ORDER BY mandatory, name
		usort($purposes, function ($a, $b) {
			/**
			 * @var Purpose $a
			 * @var Purpose $b
			 */
			if ($a->isMandatory() xor $b->isMandatory()) {
				return $a->isMandatory() ? -1 : 1;
			}
			else {
				return strnatcasecmp($a->getName(), $b->getName());
			}
		});

		$settingsIconArgs = [
			'user' => $owner->getId(),
			'domain' => $domain->getId(),
		] + array_intersect_key($apiParams, array_flip(['systemTheme', 'userTheme', 'theme']));

		if (isset($settingsIconArgs['systemTheme'])) {
			unset($settingsIconArgs['user'], $settingsIconArgs['domain']);
		}
		elseif (isset($settingsIconArgs['userTheme'])) {
			unset($settingsIconArgs['domain']);
		}
		elseif (isset($settingsIconArgs['theme']) == false) {
			$settingsIconArgs['theme'] = $theme->getId();
		}

		// GeoIP-Überprüfung, wenn benötigt
		$geoDependsOnClientIP = false;
		$geoDependsOnXCountry = false;
		if ($theme->hasDeclineButtonRestrictions()) {
			$regionRestrictions = $theme->getDeclineButtonRestrictions();
			$countryCode = $request->headers->get("X-Country", $request->headers->get("CF-IPCountry", ''));
			if ($countryCode) {
				$country = Country::find($countryCode);
				$geoDependsOnXCountry = true;
			} else {
				$geoip = GeoIp::find($request->getClientIp());
				$country = $geoip ? $geoip->getCountry() : null;
				$countryCode = $country ? $country->getCode() : '';
				$geoDependsOnClientIP = true;
			}
			if (!$country) {
				$targetRegion = 'EU';
				$showDeclineButton = true;
			} elseif ($country->hasGdpr()) {
				$targetRegion = 'EU';
				$showDeclineButton = in_array('EU', $regionRestrictions);
			} elseif (in_array($countryCode, Theme::RESTRICTABLE_COUNTRIES)) {
				$targetRegion = $countryCode;
				$showDeclineButton = in_array($countryCode, $regionRestrictions);
			} else {
				$targetRegion = implode(',', array_merge(
					["!EU"],
					array_map(function ($code) { return "!".$code; }, Theme::RESTRICTABLE_COUNTRIES)
				));
				$showDeclineButton = in_array('world', $regionRestrictions);
			}
		}
		else {
			$targetRegion = null;
			$showDeclineButton = true;
		}

		if ($domain->hasCrossDomainSharing()) {
			$shareGroupName = $domain->getShareGroupName();
			if ($shareGroupName === null) {
				$firmaPlaceholder = Placeholder::find("Firma") ?? Placeholder::find("FIRMA");
				$shareGroupName = $firmaPlaceholder ? $firmaPlaceholder->getValue() : $domain->getUrlHost();
			}
			$shareGroupUrl = $domain->getShareGroupUrl();
			if ($shareGroupUrl === null) {
				$datenschutzPlaceholder = Placeholder::find("Firma_Datenschutz");
				$shareGroupUrl = $datenschutzPlaceholder ? $datenschutzPlaceholder->getValue() : '';
			}
		} else {
			$shareGroupName = '';
			$shareGroupUrl = null;
		}

		$config = Config::getInstance();
		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
		$baseUrl = $scheme.$request->getHttpHost();
		$locales = LocaleService::enabledLocales();
		// Sortieren der Locales nach Sprachcode
		usort($locales, function (LocaleBase $a, LocaleBase $b) {
			return $a->getName() <=> $b->getName();
		});

		// Daten für das Template zusammenstellen
		$parameters = [
			'baseUrl' => $baseUrl,
			'domain' => $domain,
			'theme' => $theme,
			'locale' => $locale,
			'locales' => $locales,
			'legalText' => $legalText,
			'embeddings' => $embeddings,
			'purposes' => $purposes,
			'purposeUsageCounts' => $purposeUsageCounts,
			'whitelabel' => Utils::hasWhitelabelLicense() && $theme->getWhitelabel(),
			'poweredByText' => Utils::getPoweredByText(),
			'poweredByTooltip' => Utils::getPoweredByTooltip(),
			'poweredByUrl' => Utils::getPoweredByUrl(),
			'partnerCount' => $partnerCount,
			'settingsIconArgs' => $settingsIconArgs,
			'showDeclineButton' => $showDeclineButton,
			'tcf' => [
				'specialFeatures' => $activeTcfSpecialFeatures,
				'extraVendors' => $extraTcfVendors,
				'usageCounts' => $tcfUsageCounts,
				'descriptions' => [
					'purposes' => array_map(function (TcfPurpose $item) use ($locale) {
						$translated = $item->getTranslation($locale->getName());
						return [
							'name' => $translated->getName(),
							'desc' => $translated->getDescription(),
							'legal' => $translated->getLegalDescription(),
							'illustrations' => $translated->getIllustrations(),
						];
					}, $activeTcfPurposes),
					'special_purposes' => array_map(function (TcfSpecialPurpose $item) use ($locale) {
						$translated = $item->getTranslation($locale->getName());
						return [
							'desc' => $translated->getDescription(),
							'legal' => $translated->getLegalDescription(),
							'illustrations' => $translated->getIllustrations(),
						];
					}, TcfSpecialPurpose::all()),
					'features' => array_map(function (TcfFeature $item) use ($locale) {
						$translated = $item->getTranslation($locale->getName());
						return [
							'desc' => $translated->getDescription(),
							'legal' => $translated->getLegalDescription(),
							'illustrations' => $translated->getIllustrations(),
						];
					}, TcfFeature::all()),
					'special_features' => array_map(function ($translated) {
						return [
							'desc' => $translated->getDescription(),
							'legal' => $translated->getLegalDescription(),
							'illustrations' => $translated->getIllustrations(),
						];
					}, $activeTcfSpecialFeatures),
				],
			],
		   'shareGroupName' => $shareGroupName,
		   'shareGroupUrl' => $shareGroupUrl,
		];


		$parametersFilterEvent = new FilterEvent($parameters);
		Utils::dispatchEvent(self::EVENT_FILTER_WIDGET_PARAMETERS, $parametersFilterEvent);
		$parameters = $parametersFilterEvent->getReturn();


		$response = $this->render('external/widget/index_using-embeddings.html.twig', $parameters);

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);
		$response->headers->remove('Cache-Control');

		if ($targetRegion !== null) {
			$response->headers->set("X-Target-Region", $targetRegion);
		}

		if ($geoDependsOnClientIP) {
			// Lebensdauer kurz und private, wenn es von GeoIP abhängt
			$response->setCache([
				'max_age' => 30,
				'immutable' => true,
				'private' => true,
			]);
		} else {
			if ($geoDependsOnXCountry) {
				// Vary-Header für Caching bearbeiten
				$vary = $response->getVary();
				$vary[] = 'X-Country';
				if ($request->headers->has('CF-IPCountry')) {
					$vary[] = 'CF-IPCountry';
				}
				$response->setVary($vary);
			}
			// Lebensdauer auf ca. einen Monat setzen, da ein Zeitstempel an die Widget-Route angehängt wird
			$response->setCache([
				'max_age' => 2592000,
				'immutable' => true,
				'public' => true,
			]);
		}

		// Hier nicht text/html verwenden. Einige Proxies (z.B. in Mobilnetzen)
		// können sonst das unvollständige HTML "reparieren" und Werbung etc.
		// einbauen
		$response->headers->set('Content-Type', 'text/x-html-fragment; charset=utf-8');

		// Robots sollen das Widget nicht crawlen
		$response->headers->set('X-Robots-Tag', 'noindex');

		$language = $locale->getLanguage();
		if ($language) {
			$response->headers->set('Content-Language', $language->getIsoCode());
		}

		if ($domain->isWidgetCachingEnabled() && $request->query->get('preview') != 'true') {
			// Store response in CacheHandler to include CORS headers
			Cache::storeResponseAddToQueue($request, $response, $domain, [ 'targetRegion' => $targetRegion ]);
		}

		return $response;
	}

	/**
	 * @Route("/widget/details", name="app_external_widget_details", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 * @param Request $request
	 * @return Response
	 */
	public function widgetDetails(Request $request): Response
	{
		/** @var Domain $domain */
		$domain = Domain::activeDomain();
		if ($domain == null) {
			return new JsonResponse(['error' => 'Active domain could not be determined.'], 500);
		}

		$theme = Theme::deriveFromQuery($request->query);

		if ($theme == null) {
			return new JsonResponse(['error' => 'No theme set as active.'], 500);
		}

		/** @var LocaleBase $locale */
		$locale = LocaleService::find((string)$request->query->get('lang'));
		if ($locale == null) {
			return new JsonResponse(['error' => 'Locale not found.'], 500);
		}

		/** @var LegalText $legalText */
		$legalText = LegalText::byLocale($locale);
		if ($legalText == null) {
			return new JsonResponse(['error' => 'Legal texts aren\'t translated.'], 500);
		}

		// Nur aktive Embeddings berücksichtigen
		$embeddings = array_filter(Embedding::all(), function (Embedding $embedding) {
			return $embedding->isActive();
		});

		// Purposes anhand selektierter Cookies ermitteln
		$purposes = array_reduce(
			$embeddings,
			function (array $purposes, Embedding $embedding) {
				/** @var Purpose $purpose */
				$purpose = $embedding->getPurpose();

				if ($purpose) {
					$purposes[$purpose->getId()] = $purpose;
				}

				return $purposes;
			},
			[]
		);

		$tcfEnabled = $domain->isTcfVendorsEnabled();
		$tcfLegIntEnabled = $domain->isTcfLegitimateInterestEnabled();
		$activeTcfSpecialFeatures = self::getActiveTcfSpecialFeatures($domain, $embeddings, $locale);
		$extraTcfVendors = $tcfEnabled ? TcfHelper::getExtraTcfVendors($domain, $embeddings) : [];
		$activeTcfPurposes = [];
		$tcfUsageCounts = [
			'purposes' => [
				'all' => [],
				'legitimate_interest' => [],
			],
			'special_features' => [],
			'vendors' => [
				'all' => 0, // Includes embedding vendors if legitimate interest is enabled
				'legitimate_interest' => 0,
			],
		];

		/**
		 * @param TcfVendor $vendor
		 * @param TcfPurpose[] $purposes
		 * @param TcfSpecialFeature[] $features
		 * @param bool $isEmbeddingVendor
		 * @return void
		 */
		$addTcfVendorStats = static function (TcfVendor $vendor, array $purposes, array $features, bool $isEmbeddingVendor=false)
		use (&$tcfUsageCounts, &$activeTcfPurposes, $tcfLegIntEnabled) {
			$legIntPurposeIds = array_map(static function (TcfPurpose $purpose) {
				return $purpose->getId();
			}, $vendor->getLegitimateInterestPurposes());

			foreach ($purposes as $purpose) {
				$purposeId = $purpose->getId();
				if (!isset($activeTcfPurposes[$purposeId])) {
					$activeTcfPurposes[$purposeId] = $purpose;
					$tcfUsageCounts['purposes']['all'][$purposeId] = 1;
					$tcfUsageCounts['purposes']['legitimate_interest'][$purposeId] = 0;
				} else {
					$tcfUsageCounts['purposes']['all'][$purposeId]++;
				}

				if (in_array($purposeId, $legIntPurposeIds)) {
					$tcfUsageCounts['purposes']['legitimate_interest'][$purposeId]++;
				}
			}

			foreach ($features as $feature) {
				$featureId = $feature->getId();
				if (!isset($tcfUsageCounts['special_features'][$featureId])) {
					$tcfUsageCounts['special_features'][$featureId] = 1;
				} else {
					$tcfUsageCounts['special_features'][$featureId]++;
				}
			}

			if ($tcfLegIntEnabled || !$isEmbeddingVendor) {
				$tcfUsageCounts['vendors']['all']++;
			}
			if ($legIntPurposeIds) {
				$tcfUsageCounts['vendors']['legitimate_interest']++;
			}
		};

		if ($tcfEnabled) {
			array_walk($embeddings, function (Embedding $embedding) use ($addTcfVendorStats) {
				$tcfVendor = $embedding->getTcfVendor();
				if ($tcfVendor) {
					$purposes = $embedding->getTcfPurposes();
					$features = $embedding->getTcfSpecialFeatures();
					$addTcfVendorStats($tcfVendor, $purposes, $features, true);
				}
			});
			array_walk($extraTcfVendors, function (TcfVendor $vendor) use ($addTcfVendorStats, $tcfLegIntEnabled) {
				$purposes = array_merge(
					$vendor->getPurposes(),
					$tcfLegIntEnabled ? $vendor->getLegitimateInterestPurposes() : $vendor->getFlexiblePurposes()
				);
				$features = $vendor->getSpecialFeatures();
				$addTcfVendorStats($vendor, $purposes, $features);
			});
			ksort($activeTcfPurposes);

			$adTechProviders = $domain->hasGoogleAdditionalConsentMode() ? $domain->getGoogleAdditionalConsentWhitelist() : [];
			if ($adTechProviders) {
				$tcfUsageCounts['vendors']['all'] += count($adTechProviders);
			}
		}

		// ORDER BY mandatory, name
		usort($purposes, function ($a, $b) {
			/**
			 * @var Purpose $a
			 * @var Purpose $b
			 */
			if ($a->isMandatory() xor $b->isMandatory()) {
				return $a->isMandatory() ? -1 : 1;
			}
			else {
				return strnatcasecmp($a->getName(), $b->getName());
			}
		});

		$config = Config::getInstance();
		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
		$baseUrl = $scheme.$request->getHttpHost();

		// Daten für das Template zusammenstellen
		$parameters = [
			'baseUrl' => $baseUrl,
			'domain' => $domain,
			'theme' => $theme,
			'locale' => $locale,
			'locales' => LocaleService::enabledLocales(),
			'legalText' => $legalText,
			'regulations' => Regulation::all(),
			'embeddings' => $embeddings,
			'purposes' => $purposes,
			'whitelabel' => Utils::hasWhitelabelLicense() && $theme->getWhitelabel(),
			'poweredByText' => Utils::getPoweredByText(),
			'poweredByTooltip' => Utils::getPoweredByTooltip(),
			'poweredByUrl' => Utils::getPoweredByUrl(),
			'tcf' => [
				'usageCounts' => $tcfUsageCounts,
				'specialFeatures' => $activeTcfSpecialFeatures,
				'extraVendors' => $extraTcfVendors,
				'descriptions' => [
					'purposes' => array_map(function (TcfPurpose $item) use ($locale) {
						$translated = $item->getTranslation($locale->getName());
						return [
							'name' => $translated->getName(),
							'desc' => $translated->getDescription(),
							'legal' => $translated->getLegalDescription(),
							'illustrations' => $translated->getIllustrations(),
						];
					}, $activeTcfPurposes),
					'special_purposes' => array_map(function (TcfSpecialPurpose $item) use ($locale) {
						$translated = $item->getTranslation($locale->getName());
						return [
							'desc' => $translated->getDescription(),
							'legal' => $translated->getLegalDescription(),
							'illustrations' => $translated->getIllustrations(),
						];
					}, TcfSpecialPurpose::all()),
					'features' => array_map(function (TcfFeature $item) use ($locale) {
						$translated = $item->getTranslation($locale->getName());
						return [
							'desc' => $translated->getDescription(),
							'legal' => $translated->getLegalDescription(),
							'illustrations' => $translated->getIllustrations(),
						];
					}, TcfFeature::all()),
					'special_features' => array_map(function (TcfSpecialFeature $item) use ($locale) {
						$translated = $item->getTranslation($locale->getName());
						return [
							'desc' => $translated->getDescription(),
							'legal' => $translated->getLegalDescription(),
							'illustrations' => $translated->getIllustrations(),
						];
					}, TcfSpecialFeature::all()),
					'data_categories' => array_map(function (TcfDataCategory $item) use ($locale) {
						$translated = $item->getTranslation($locale->getName());
						return [
							'desc' => $translated->getDescription(),
						];
					}, TcfDataCategory::all()),
				],
			],
		];

		$response = $this->render('external/widget/details.html.twig', $parameters);

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		// Lebensdauer auf ein Jahr setzen, da ein Zeitstempel an die Widget-Route angehängt wird
		$response->headers->remove('Cache-Control');
		$response->setCache([
			'max_age' => 31536000,
			'immutable' => true,
			'public' => true,
		]);

		// Hier nicht text/html verwenden. Einige Proxies (z.B. in Mobilnetzen)
		// können sonst das unvollständige HTML "reparieren" und Werbung etc.
		// einbauen
		$response->headers->set('Content-Type', 'text/x-html-fragment; charset=utf-8');

		// Robots sollen das Widget nicht crawlen
		$response->headers->set('X-Robots-Tag', 'noindex');

		$language = $locale->getLanguage();
		if ($language) {
			$response->headers->set('Content-Language', $language->getIsoCode());
		}

		if ($domain->isWidgetCachingEnabled() && $request->query->get('preview') != 'true') {
			// Store response in CacheHandler to include CORS headers
			Cache::storeResponseAddToQueue($request, $response, $domain);
		}

		return $response;
	}

	private function escapeNulBytes(string $string): string
	{
		return str_replace("\0", '\0', $string);
	}

	/**
	 * @Route("/page_check/report", name="app_external_page_check_report", methods={"OPTIONS", "POST"}, stateless=true)
	 * @param Request $request
	 * @return JsonResponse
	 * @phan-suppress PhanUndeclaredProperty
	 */
	public function pageCheckReport(Request $request): JsonResponse
	{
		// Kein Reporting im Vorschaumodus
		if ($request->query->getBoolean('preview')) {
			return new JsonResponse(['success'=>false, 'message'=>'No reports in preview mode']);
		}

		$now = new DateTime('now');
		$domain = Domain::activeDomain();

		if (Utils::decodeJsonRequest($request, $data, false, 5) == false) {
			/** @var JsonResponse $data ($data ist im Fehlerfall eine Fehlermeldung) */
			assert($data instanceof JsonResponse);
			return $data;
		}

		try {
			if (isset($data->url) and is_string($data->url)) {
				$pageUrl = preg_replace('/#[^#]*$/', '', $data->url);
			} else {
				return $this->json(['success'=>false, 'message'=>'invalid data: url missing'], 400);
			}

			// Cookies verarbeiten
			if (isset($data->c) and is_array($data->c)) {
				foreach ($data->c as $cookie) {
					if (is_string($cookie)) {
						$cookie = $this->escapeNulBytes($cookie);

						if (FoundCookie::exists($cookie)) {
							$item = FoundCookie::find($cookie);
						}
						// Nur neue Einträge speichern, wenn Dateigrößenlimit noch nicht erreicht
						elseif (FoundCookie::persistedRepositorySize() <= self::FOUND_REPO_BYTES_LIMIT) {
							$item = FoundCookie::create($cookie, $now);
							//Primär auf ignore setzen, erst extern bestätigen lassen durch Scanner
							$item->setIgnore(true);
						}
						else {
							$item = null;
						}

						if ($item) {
							if (($now->getTimestamp() - $item->getLastSeenOn()->getTimestamp()) > self::TOUCH_INTERVAL) {
								$item->setLastSeenOn($now);
							}
							if (count($item->getFoundOnPages()) < 10) {
								$item->addFoundOnPage($pageUrl);
							}
						}

					}
				}
				FoundCookie::flush();
			}

			// Skripte verarbeiten
			if (isset($data->s) and is_array($data->s)) {
				foreach ($data->s as $script) {
					if (is_string($script)) {
						// Parameter und Fragment aus URL entfernen
						$script = preg_replace('/#[^#]*$/', '', $script);
						$script = preg_replace('/\?[^?]*$/', '?', $script);
						// Weitere Pfad-Parameter entfernen
						$script = preg_replace('%/(intl/|releases/|viewthroughconversion/|www-widgetapi-)[^/]+/%', '/\1…/', $script);
						// Ergebnis speichern
						if (preg_match('%^(?:https:|http:|ftp:)?//%', $script)) {

							if (FoundScript::exists($script)) {
								$item = FoundScript::find($script);
							}
							// Nur neue Einträge speichern, wenn Dateigrößenlimit noch nicht erreicht
							elseif (FoundScript::persistedRepositorySize() <= self::FOUND_REPO_BYTES_LIMIT) {
								$item = FoundScript::create($script, $now);
								//Primär auf ignore setzen, erst extern bestätigen lassen durch Scanner
								$item->setIgnore(true);
							}
							else {
								$item = null;
							}

							if ($item) {
								if (($now->getTimestamp() - $item->getLastSeenOn()->getTimestamp()) > self::TOUCH_INTERVAL) {
									$item->setLastSeenOn($now);
								}
								if (count($item->getFoundOnPages()) < 10) {
									$item->addFoundOnPage($pageUrl);
								}
							}

						}
					}
				}
				FoundScript::flush();
			}

			// Objekte/IFrames verarbeiten
			if (isset($data->o) and is_array($data->o)) {
				foreach ($data->o as $object) {
					if (is_string($object)) {
						$object = $this->escapeNulBytes($object);

						// Parameter und Fragment aus URL entfernen
						$object = preg_replace('/#[^#]*$/', '', $object);
						$object = preg_replace('/\?[^?]*$/', '?', $object);
						// URL weiter bereinigen (Sonderfall Google SafeFrame)
						$object = preg_replace('/\/\/[0-9a-f]{16,}\.safeframe\./', '//….safeframe.', $object);

						if (preg_match('%^(?:https:|http:|ftp:)?//%', $object)) {

							if (FoundObject::exists($object)) {
								$item = FoundObject::find($object);
							}
							// Nur neue Einträge speichern, wenn Dateigrößenlimit noch nicht erreicht
							elseif (FoundObject::persistedRepositorySize() <= self::FOUND_REPO_BYTES_LIMIT) {
								$item = FoundObject::create($object, $now);
								//Primär auf ignore setzen, erst extern bestätigen lassen durch Scanner
								$item->setIgnore(true);
							}
							else {
								$item = null;
							}

							if ($item) {
								if (($now->getTimestamp() - $item->getLastSeenOn()->getTimestamp()) > self::TOUCH_INTERVAL) {
									$item->setLastSeenOn($now);
								}
								if (count($item->getFoundOnPages()) < 10) {
									$item->addFoundOnPage($pageUrl);
								}
							}

						}
					}
				}
				FoundObject::flush();
			}

			// LocalStorage verarbeiten
			if (isset($data->ls) and is_array($data->ls)) {
				foreach ($data->ls as $object) {
					if (is_string($object)) {
						$object = $this->escapeNulBytes($object);
						$drop = false;
						foreach ($this->EXCLUDE_LOCALSTORAGE as $pattern) {
							if (fnmatch($pattern, substr($object, 0, 4095))) {
								$drop = true;
								break;
							}
						}
						if ($drop) {
							continue;
						}
						foreach ($this->MERGE_LOCALSTORAGE as $prefix) {
							if (strpos($object, $prefix) === 0) {
								$object = $prefix.'…';
							}
						}

						if (FoundStorage::exists($object)) {
							$item = FoundStorage::find($object);
						}
						// Nur neue Einträge speichern, wenn Dateigrößenlimit noch nicht erreicht
						elseif (FoundStorage::persistedRepositorySize() <= self::FOUND_REPO_BYTES_LIMIT) {
							$item = FoundStorage::create($object, $now);
							//Primär auf ignore setzen, erst extern bestätigen lassen durch Scanner
							$item->setIgnore(true);
							$item->setIsSessionStorage(false);
						}
						else {
							$item = null;
						}

						if ($item) {
							if (($now->getTimestamp() - $item->getLastSeenOn()->getTimestamp()) > self::TOUCH_INTERVAL) {
								$item->setLastSeenOn($now);
							}
							if (count($item->getFoundOnPages()) < 10) {
								$item->addFoundOnPage($pageUrl);
							}
						}
					}
				}
				FoundStorage::flush();
			}

			// SessionStorage verarbeiten
			if (isset($data->ss) and is_array($data->ss)) {
				foreach ($data->ss as $object) {
					if (is_string($object)) {
						$object = $this->escapeNulBytes($object);
						$drop = false;
						foreach ($this->EXCLUDE_LOCALSTORAGE as $pattern) {
							if (fnmatch($pattern, substr($object, 0, 4095))) {
								$drop = true;
								break;
							}
						}
						if ($drop) {
							continue;
						}
						foreach ($this->MERGE_LOCALSTORAGE as $prefix) {
							if (strpos($object, $prefix) === 0) {
								$object = $prefix.'…';
							}
						}

						if (FoundStorage::exists($object)) {
							$item = FoundStorage::find($object);
						}
						// Nur neue Einträge speichern, wenn Dateigrößenlimit noch nicht erreicht
						elseif (FoundStorage::persistedRepositorySize() <= self::FOUND_REPO_BYTES_LIMIT) {
							$item = FoundStorage::create($object, $now);
							//Primär auf ignore setzen, erst extern bestätigen lassen durch Scanner
							$item->setIgnore(true);
							$item->setIsSessionStorage(true);
						}
						else {
							$item = null;
						}

						if ($item) {
							if (($now->getTimestamp() - $item->getLastSeenOn()->getTimestamp()) > self::TOUCH_INTERVAL) {
								$item->setLastSeenOn($now);
							}
							if (count($item->getFoundOnPages()) < 10) {
								$item->addFoundOnPage($pageUrl);
							}
						}
					}
				}
				FoundStorage::flush();
			}

			// Blockierbare Scripte verarbeiten
			if (isset($data->as) and is_object($data->as)) {
				$added = false;
				$enable = !($domain->isNewScriptsBlocked());
				foreach ((array)$data->as as $key => $script) {
					$code = null;
					if (is_string($script)) {
						$code = $script;
						$cookies = [];
					}
					elseif (is_object($script)) {
						$code = $script->s;
						// Base64 dekodieren
						if (is_string($code) && $code !== '' && $code[0] != '<') {
							$code = base64_decode($code, false);
							if (!is_string($code)) {
								$code = '';
							}
						}
						$cookies = array_map(function ($cookie) { return $this->escapeNulBytes((string)$cookie); }, (array)$script->c);
					}
					else {
						continue;
					}
					if ($code && Script::count() < self::SCRIPT_COUNT_LIMIT) {
						if (preg_match('%^<script\b%i', $code)) {
							$script = Script::find($key);
							if (!$script) {
								$added = true;
								$script = Script::create($key)->setContent($code)->addCookies($cookies)->setEnableScript($enable)->addFoundOnPage($pageUrl);
							}
							elseif ($cookies) {
								$new_cookies = array_diff($cookies, $script->getCookies());
								if ($new_cookies) {
									$script->addCookies($new_cookies);
									$added = true;
								}
							}
						}
					}
				}
				if ($added) {
					Script::flush();
				}
			}

		}
		catch (Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		$response = $this->json(['success'=>true]);

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		$response->setCache([
			'max_age'=>30,
			'private'=>true,
		]);

		return $response;
	}

	/**
	 * Diese Methode benachrichtigt einen Benutzer über den Stand des Aufruflimits, sofern eine bestimmte
	 * Grenze überschritten wurde.
	 * @param User $user
	 */
	private function sendQuotaReport(User $user): void
	{
		if($user->isSubUser())
		{
			$user = $user->getAccountOwner();
		}

		$callCount 		= $user->getActualCallCount();
		$callMaxCount 	= $user->getMaxCallCount();
		$config 		= Config::getInstance();
		$domainName		= Domain::activeDomain()->getName();

        // Wenn die Anzahl der Widgetausgabe nicht limitiert ist, sofort raus
		if ($callMaxCount == -1) {
			return;
		}

		//Settings Call Limits...
		$clArray = explode("\n",trim((string)$config->get('callLimitNotificationSteps')));

		if (empty($clArray))
		{
			$clArray = array(90,95);
		}
		// Bei 100% = überschritten auch immer benachrichtigen!
		$clArray[] = 100;

		//Match Call Limits from Settings
		$quota = $callMaxCount > 0
			? min((int)floor($callCount / $callMaxCount * 100), 100)
			: 100;

		$matchedQuota = array_reduce($clArray, function ($match, $item) use ($quota) {
			return $item <= $quota && $item > $match ? $item : $match;
		});

		// Wenn eine bestimmte Grenze noch nicht erreicht ist, keine Benachritigung verschicken
		if ($matchedQuota <= $user->getLastCallReportQuota()) {
			return;
		}

		/** @var Environment $twig */
		$twig = $this->container->get('twig');

		if ($matchedQuota < 100)
		{
			$subject  	= $config->get('sub_max_traffic_email_betreff');
			$body  		= $config->get('sub_max_traffic_email_content');
			$mailingKey = 'QuotaReport-SubMax';
		}
		else
		{
			$subject  	= $config->get('max_traffic_email_betreff');
			$body  		= $config->get('max_traffic_email_content');
			$mailingKey = 'QuotaReport-Exceeded';
		}

		if (!empty($body) && !empty($subject))
		{
			//quota
			$body = str_ireplace('{{ quota }}', (string)$matchedQuota, $body);
			$subject = str_ireplace('{{ quota }}', (string)$matchedQuota, $subject);
			//$callMaxCount
			$body = str_ireplace('{{ callMaxCount }}', (string)$callMaxCount, $body);
			$subject = str_ireplace('{{ callMaxCount }}', (string)$callMaxCount, $subject);
			//DomName
			$body = str_ireplace('{{ domainvar }}', (string)$domainName, $body);
			$subject = str_ireplace('{{ domainvar }}', (string)$domainName, $subject);
		}
		else{
			$subject = $matchedQuota.'% of your quota exceeded';
			$body = $twig->render('emails/quota_exceeded.txt.twig', [
				'quota' => $matchedQuota,
			]);
		}

		$addr = $user->getEmailAddress();
		if (!$addr) {
			$addr = $user->getUsername();
		}

		if ($user->isActive() and $user->isMailingSubscribed($mailingKey, true)) {
			$mail = new MailMessage($subject);

			try {
				$mail->to($addr);
				$mail->autoSubmitted();
				$mail->listId($mailingKey.'.'.$user->getId());
				$mail->unsubscribe($user, $mailingKey);
				$mail->text($body);
				$sender = $mail->getFrom()[0];

				// BCC setzen wenn Kopie-Funktion aktiviert
				if ($config->get('sendQuotaCopies')) {
					$mail->bcc($config->get('email'));
				}

				$this->mailer->send($mail);

				// Loggen
				$this->logger->logInfo("QuotaReport", "QuotaReport an ".$addr." - ".$subject." - Von: ".$sender->getAddress());
			} catch (Exception $exception) {
				$this->logger->logWarning('QuotaReport', 'Could not send email. '.$exception->getMessage());
			}
		}

		$user->setLastCallReportQuota($matchedQuota)->save();
	}

	/**
	 * @Route("/page_check/report", name="app_external_page_check_report_get", methods={"HEAD", "GET"}, stateless=true)
	 * @param Request $request
	 * @return JsonResponse
	 *
	 * Dummy-Methode: Vermeidet Exceptions bei GET auf die Report-Route
	 */
	public function pageCheckReportGet(Request $request): JsonResponse
	{
		return $this->json(['error'=>'GET not allowed for this action']);
	}

	/**
	 * @Route("tcf-disclosure/{vendor}", name="app_external_tcf_disclosures", methods={"HEAD", "GET", "OPTIONS"}, stateless=true)
	 *
	 * @param Request $request
	 * @param TcfVendor $vendor
	 * @return Response HTML-Fragment
	 */
	public function disclosures(Request $request, TcfVendor $vendor): Response
	{
		// Verwende angeforderte Sprache
		$localeName = (string)$request->query->get('lang', '');
		$locale = LocaleService::find($localeName);
		// Wenn ein nicht-existentes Locale angegeben wird, Fallback zum Standard
		if (!$locale) {
			$locale = User::loggedInUser()->getDefaultLocale();
		}

		$disclosures = $vendor->getDeviceStorageDisclosures();

		$response = $this->render('external/widget/tcf_extended_disclosure.html.twig', ['locale'=>$locale, 'disclosures'=>$disclosures ?? []]);
		if ($disclosures === null) {
			$response->setStatusCode(404);
		}
		//$response->headers->set('Content-Type', 'text/x-html-fragment; charset=utf-8');
		$response->headers->set('Content-Type', 'text/html; charset=utf-8');

		// Workaround: HTTP-Header Cache-Control würde sonst auf private, must-revalidate etc. gesetzt
		Utils::resetCacheControl($response);

		$response->headers->set('X-Robots-Tag', 'noindex, nofollow');
		$response->setCache([
			'max_age'=>3600,
			'public'=>true,
		]);
		return $response;
	}
}
