<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\Exception\BadRequestException;
use App\Controller\HostingApiController;
use App\Exception\LocaleAlreadyExistsException;
use App\Exception\UserAlreadyExistsException;
use App\Model\Domain;
use App\Model\Language;
use App\Model\LegalText;
use App\Model\Locale;
use App\Model\LocaleBase;
use App\Model\Purpose;
use App\Model\SystemTheme;
use App\Model\Theme;
use App\Model\User;
use App\Model\UserLocale;
use App\Model\UserTheme;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\HttpKernelInterface;
use Symfony\Component\HttpKernel\KernelInterface;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/hosting/api/aio", name="app_hosting_api_aio_")
 */
class AioApi extends HostingApiController
{
	private const WEBSITE_AIO_SCHEMA = [
		'username' => 'string',
		'websiteUrl' => 'string',

		'domainMaxCount' => 'int',
		'whitelabelMaxCount' => 'int',
		'widgetDisplayMaxCount' => 'int',

		'locales' => 'string[]',

		'systemTheme' => 'string',
		'userTheme' => 'string',
	];

	/**
	 * @Route("/website", name="website", methods={"POST"})
	 * @param Request $request
	 * @param KernelInterface $kernel
	 * @return JsonResponse
	 */
	public function processWebsite(Request $request, KernelInterface $kernel): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}


		// Verify data types and check for mandatory properties
		try {
			self::checkPayloadDataTypes($payload, self::WEBSITE_AIO_SCHEMA);

			foreach (['username', 'websiteUrl'] as $property) {
				if (
					array_key_exists($property, $payload) == false
					|| is_string($payload[$property]) && strlen(trim($payload[$property])) == 0
					|| is_array($payload[$property]) && count($payload[$property]) == 0
				) {
					throw new InvalidArgumentException("Empty or missing $property.", 1);
				}
			}

			if ($invalidLocaleNames = array_diff($payload['locales'] ?? [], array_keys(Language::all()))) {
				throw new InvalidArgumentException('Invalid locale: '.implode(', ', $invalidLocaleNames), 2);
			}

			if (array_key_exists('systemTheme', $payload) && array_key_exists('userTheme', $payload)) {
				throw new InvalidArgumentException('Define either systemTheme or userTheme.', 3);
			}
			elseif (array_key_exists('systemTheme', $payload) && !SystemTheme::exists($payload['systemTheme'])) {
				throw new InvalidArgumentException('Invalid systemTheme.', 3);
			}
		}
		catch (InvalidArgumentException $exception) {
			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}


		// Load or create the specified user and update properties
		try {
			$username = trim($payload['username']);
			$user = User::find($username) ?? User::create($username)->setRole(User::ROLE_CLIENT);

			// Set proper user context
			self::assertClientExists($user->getId());

			// Update user properties
			if (array_key_exists('domainMaxCount', $payload)) {
				$user->setMaxDomainCount(max(-1, $payload['domainMaxCount']));
			}
			if (array_key_exists('whitelabelMaxCount', $payload)) {
				$user->setMaxWLCount(max(-1, $payload['whitelabelMaxCount']));
			}
			if (array_key_exists('widgetDisplayMaxCount', $payload)) {
				$user->setMaxCallCount(max(-1, $payload['widgetDisplayMaxCount']));
			}

			$user->save();
		}
		catch (InvalidArgumentException|LogicException|UserAlreadyExistsException $exception) {
			// Revert changes if an error has occurred (rollback)
			User::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (BadRequestException $exception) {
			// Revert changes if an error has occurred (rollback)
			User::reload();

			return $exception->getResponse();
		}


		// Check user dependant payload info
		if (array_key_exists('userTheme', $payload) && !UserTheme::exists($payload['userTheme'])) {
			return $this->json([
				'error' => 'Invalid userTheme.',
			], 400);
		}


		// Create the domain in a sub request via the domain API
		$attributes = [
			'_controller' => DomainApi::class.'::createDomain',
			'userId' => $user->getId(),
		];
		$body = json_encode([
			'name' => trim($payload['websiteUrl']),
			'frontendWidgetEnabled' => true,
			'frontendWidgetMinifyCode' => true,
		]);

		$createDomainRequest = new Request([], [], $attributes, [], [], [], $body);
		$createDomainRequest->setMethod('POST');
		$createDomainRequest->headers->set('Content-Type', 'application/json');

		$createDomainResponse = $kernel->handle($createDomainRequest, HttpKernelInterface::SUB_REQUEST);
		$domainResult = json_decode((string)$createDomainResponse->getContent(), true);

		if ($createDomainResponse->getStatusCode() != 201) {
			return $this->json([
				'error' => $domainResult['error'] ?? 'Unexpected error while creating the domain',
			], $createDomainResponse->getStatusCode() ?: 500);
		}

		$domain = Domain::find($domainResult['id'] ?? '');
		if (!$domain) {
			return $this->json([
				'error' => 'An unexpected error occurred while creating the domain.',
			], 500);
		}

		// Set proper domain context
		self::assertDomainExists($user->getId(), $domain->getId());


		// Scan the domain in a sub request via the onboarding API
		$attributes = [
			'_controller' => OnboardingApi::class.'::scan',
			'userId' => $user->getId(),
			'domainId' => $domain->getId(),
		];
		$query = [
			'ignoreCrawlerError' => $request->query->getBoolean('ignoreCrawlerError'),
		];

		$onboardingScanRequest = new Request($query, [], $attributes);
		$onboardingScanResponse = $kernel->handle($onboardingScanRequest, HttpKernelInterface::SUB_REQUEST);

		$scanResult = json_decode((string)$onboardingScanResponse->getContent(), true);
		if (!$scanResult) {
			Domain::delete($domain->getId());
			$user->updateDomainInfo()->save();
			return $this->json([
				'error' => 'Sub request OnboardingApi::scan exited with an unexpected response.',
			], 500);
		}
		elseif ($onboardingScanResponse->getStatusCode() != 200) {
			Domain::delete($domain->getId());
			$user->updateDomainInfo()->save();
			return $this->json([
				'error' => $scanResult['error'] ?? 'Unexpected error while scanning website',
			], $onboardingScanResponse->getStatusCode() ?: 500);
		}


		// Import embeddings in a sub request via the onboarding API
		if ($presetIds = array_filter(array_map(function ($preset) { return (int)$preset['id']; }, $scanResult['availablePresets']))) {
			$attributes = [
				'_controller' => OnboardingApi::class.'::bulkImportPresets',
				'userId' => $user->getId(),
				'domainId' => $domain->getId(),
			];
			$body = json_encode(array_values(array_map(function (int $presetId) {
				return ['id'=>$presetId, 'active'=>true];
			}, $presetIds)));

			$onboardingPresetsRequest = new Request([], [], $attributes, [], [], [], $body);
			$onboardingPresetsRequest->setMethod('POST');
			$onboardingPresetsRequest->headers->set('Content-Type', 'application/json');

			$kernel->handle($onboardingPresetsRequest, HttpKernelInterface::SUB_REQUEST);
		}


		// Create an embedding for unmatched cookies via the embedding API
		$orphansPresetMandatory = array_reduce($scanResult['availablePresets'], function ($carry, $item) {
			return $item['id'] == 13002 ? $item : $carry;
		});
		$orphansPresetOptional = array_reduce($scanResult['availablePresets'], function ($carry, $item) {
			return $item['id'] ? $carry : $item;
		});

		// Collect all presets that have to be manually created next
		$orphansPresets = array_filter([
			$orphansPresetMandatory,
			$orphansPresetOptional,
		]);

		foreach ($orphansPresets as $orphansPreset) {
			$attributes = [
				'_controller' => EmbeddingApi::class.'::createEmbedding',
				'userId' => $user->getId(),
				'domainId' => $domain->getId(),
			];
			$body = json_encode([
				'name' => $orphansPreset['name'],
				'active' => true,
				'purpose' => $orphansPreset['purpose'],
				'vendor' => $orphansPreset['vendor'],
				'translations' => [
					'de_DE' => [
						'name' => $orphansPreset['translations']['de']['name'],
						'description' => $orphansPreset['translations']['de']['description'],
						'legalBasis' => $orphansPreset['translations']['de']['legalBasis'],
					],
					'en_US' => [
						'name' => $orphansPreset['translations']['en']['name'],
						'description' => $orphansPreset['translations']['en']['description'],
						'legalBasis' => $orphansPreset['translations']['en']['legalBasis'],
					],
				],
				'assets' => array_map(function ($asset) {
					return [
						'name' => $asset['name'],
						'dynamic' => (bool)$asset['dynamic'],
						'storageType' => $asset['storageType'],
						'translations' => [
							'de_DE' => [
								'lifetime' => $asset['translations']['de']['lifetime'],
								'description' => $asset['translations']['de']['description'],
							],
							'en_US' => [
								'lifetime' => $asset['translations']['en']['lifetime'],
								'description' => $asset['translations']['en']['description'],
							],
						],
					];
				}, $orphansPreset['assets'] ?? []),
			]);

			$subRequest = new Request([], [], $attributes, [], [], [], $body);
			$subRequest->setMethod('POST');
			$subRequest->headers->set('Content-Type', 'application/json');

			$kernel->handle($subRequest, HttpKernelInterface::SUB_REQUEST);
		}


		// Create placeholders in a sub request via the placeholder API
		$attributes = [
			'_controller' => PlaceholderApi::class.'::bulkCreatePlaceholders',
			'userId' => $user->getId(),
			'domainId' => $domain->getId(),
		];
		$body = json_encode([
			['name'=>'FIRMA', 'value'=>$scanResult['address']['company'] ?? ''],
			['name'=>'Firma_Datenschutz', 'value'=>$scanResult['privacyPolicyUrl'] ?? ''],
			['name'=>'Firma_Ort', 'value'=>trim("{$scanResult['address']['zipCode']} {$scanResult['address']['city']}")],
		]);

		$placeholderRequest = new Request([], [], $attributes, [], [], [], $body);
		$placeholderRequest->setMethod('PUT');
		$placeholderRequest->headers->set('Content-Type', 'application/json');

		$kernel->handle($placeholderRequest, HttpKernelInterface::SUB_REQUEST);


		// Enable locales
		$localeNames = array_unique(($payload['locales'] ?? []) ?: ['de_DE']);
		self::loadLocalesAndCreateIfMissing(array_merge($localeNames, array_keys(UserLocale::enabledLocales())));

		foreach (Locale::all() as $locale) {
			$locale->setEnabled(in_array($locale->getName(), $localeNames));

			LegalText::byLocale($locale)
				->setImprintScheme(LegalText::DISPLAY_MODE_LINK)
				->setImprintUrl($scanResult['imprintUrl'] ?? '')
				->setPrivacyScheme(LegalText::DISPLAY_MODE_LINK)
				->setPrivacyUrl($scanResult['privacyPolicyUrl'] ?? '')
			;
		}

		LegalText::flush();
		Locale::flush();


		// Enable the requested theme
		if (
			array_key_exists('systemTheme', $payload) and $systemTheme = SystemTheme::find($payload['systemTheme']) and $theme = Theme::bySystemTheme($systemTheme)
			or array_key_exists('userTheme', $payload) and $userTheme = UserTheme::find($payload['userTheme']) and $theme = Theme::byUserTheme($userTheme)
		) {
			$theme->makeDefault();
		}


		// Get embed code snippets in a sub request via the onboarding API
		$attributes = [
			'_controller' => OnboardingApi::class.'::getCodeSnippets',
			'userId' => $user->getId(),
			'domainId' => $domain->getId(),
		];

		$onboardingEmbedRequest = new Request([], [], $attributes);
		$onboardingEmbedResponse = $kernel->handle($onboardingEmbedRequest, HttpKernelInterface::SUB_REQUEST);

		$embedResult = json_decode((string)$onboardingEmbedResponse->getContent(), true);
		if (!$embedResult) {
			Domain::delete($domain->getId());
			$user->updateDomainInfo()->save();
			return $this->json([
				'error' => 'Sub request OnboardingApi::getCodeSnippets exited with an unexpected response.',
			], 500);
		}
		elseif ($onboardingEmbedResponse->getStatusCode() != 200) {
			Domain::delete($domain->getId());
			$user->updateDomainInfo()->save();
			return $this->json([
				'error' => $scanResult['error'] ?? 'Unexpected error while fetching embed code snippets',
			], $onboardingEmbedResponse->getStatusCode() ?: 500);
		}


		return $this->json([
			'clientId' => $user->getId(),
			'domainId' => $domain->getId(),
			'embed' => $embedResult,
		], 201);
	}

	/**
	 * @param string[] $localeNames
	 * @return Locale[]
	 */
	private static function loadLocalesAndCreateIfMissing(array $localeNames): array
	{
		$locales = Locale::all();
		$locales = array_combine(
			array_map(function (Locale $locale) { return $locale->getName(); }, $locales),
			$locales
		);

		$locales = array_intersect_key($locales, array_flip($localeNames));

		// Create new locales
		if ($pendingLocaleNames = array_diff($localeNames, array_keys($locales))) {
			$userLocales = UserLocale::all();
			$userLocales = array_combine(
				array_map(function (UserLocale $locale) { return $locale->getName(); }, $userLocales),
				$userLocales
			);

			foreach ($pendingLocaleNames as $localeName) {
				try {
					$locale = Locale::create($localeName);
					$locales[$locale->getName()] = $locale;
				} catch (LocaleAlreadyExistsException $e) {
					continue; // Shouldn't happen
				}

				// If a corresponding UserLocale exists, copy its contents
				if (isset($userLocales[$localeName])) {
					self::fillLocale($locale, $userLocales[$localeName]);
				}
			}

			LegalText::flush();
			Locale::flush();
		}

		return $locales;
	}

	/**
	 * This method locks collections Locale and LegalText in this order; flush should occur in reverse.
	 * @param Locale $target
	 * @param LocaleBase $source
	 * @param bool $updateLegalText
	 * @return void
	 */
	private static function fillLocale(Locale $target, LocaleBase $source, bool $updateLegalText=false): void
	{
		$target
			->setEnabled($source->isEnabled())

			->setWidgetTitle($source->getWidgetTitle())
			->setWidgetIntroduction($source->getWidgetIntroduction())
			->setWidgetDeclineButtonText($source->getWidgetDeclineButtonText())
			->setWidgetConsentButtonText($source->getWidgetConsentButtonText())
			->setWidgetSettingsButtonText($source->getWidgetSettingsButtonText())

			->setCheckAllButtonText($source->getCheckAllButtonText())
			->setUncheckAllButtonText($source->getUncheckAllButtonText())

			->setImprintText($source->getImprintText())

			->setPrivacyPolicyText($source->getPrivacyPolicyText())

			->setAccessibilityText($source->getAccessibilityText())

			->setControlPanelTitle($source->getControlPanelTitle())
			->setControlPanelCancelButtonText($source->getControlPanelCancelButtonText())
			->setCategoriesAcceptAllButtonText($source->getCategoriesAcceptAllButtonText())
			->setControlPanelSaveButtonText($source->getControlPanelSaveButtonText())

			->setDetailsTitle($source->getDetailsTitle())
			->setDetailsCloseButtonText($source->getDetailsCloseButtonText())

			->setCookieVendorText($source->getCookieVendorText())
			->setCookieDescriptionText($source->getCookieDescriptionText())
			->setCookiePrivacyPolicyUrlText($source->getCookiePrivacyPolicyUrlText())
			->setCookieLifetimeText($source->getCookieLifetimeText())
			->setCookieCollectedDataInfoText($source->getCookieCollectedDataInfoText())
			->setCookiePurposeOfDataCollectionText($source->getCookiePurposeOfDataCollectionText())
			->setCookieLegalBasisText($source->getCookieLegalBasisText())
			->setCookiePlaceOfProcessingText($source->getCookiePlaceOfProcessingText())

			->setAssetNameLabel($source->getAssetNameLabel())
			->setAssetLifetimeLabel($source->getAssetLifetimeLabel())
			->setAssetDescriptionLabel($source->getAssetDescriptionLabel())

			->setBlockedContentTitle($source->getBlockedContentTitle())
			->setBlockedContentText($source->getBlockedContentText())
			->setBlockedContentButtonText($source->getBlockedContentButtonText())
			->setBlockedContentShowExternalContent($source->getBlockedContentShowExternalContent())
			->setBlockedContentAllowEmbedding($source->getBlockedContentAllowEmbedding())
			->setBlockedContentProviderDetails($source->getBlockedContentProviderDetails())

			->setChangeConsentLabel($source->getChangeConsentLabel())
			->setMoreInformationLabel($source->getMoreInformationLabel())
			->setMorePurposeDetailsButtonText($source->getMorePurposeDetailsButtonText())

			->setTcfVendorGrantedPermissionsText($source->getTcfVendorGrantedPermissionsText())
			->setTcfPurposesLabel($source->getTcfPurposesLabel())
			->setTcfSpecialPurposesLabel($source->getTcfSpecialPurposesLabel())
			->setTcfFeaturesLabel($source->getTcfFeaturesLabel())
			->setTcfSpecialFeaturesLabel($source->getTcfSpecialFeaturesLabel())
			->setTcfOptedInRemark($source->getTcfOptedInRemark())
			->setTcfThirdPartiesLabel($source->getTcfThirdPartiesLabel())
			->setTcfConsentLabel($source->getTcfConsentLabel())
			->setTcfLegitimateInterestLabel($source->getTcfLegitimateInterestLabel())
			->setNonTcfPurposesLabel($source->getNonTcfPurposesLabel())
			->setNonTcfVendorsLabel($source->getNonTcfVendorsLabel())
			->setTcfVendorUsesNonCookieStorageText($source->getTcfVendorUsesNonCookieStorageText())
			->setTcfSpecialFeaturesHandledGloballyText($source->getTcfSpecialFeaturesHandledGloballyText())
			->setTcfWidgetIntroduction($source->getTcfWidgetIntroduction())
			->setTcfStacksListText($source->getTcfStacksListText())
			->setTcfCookieRefreshText($source->getTcfCookieRefreshText())
			->setTcfExtendedDisclosureText($source->getTcfExtendedDisclosureText())
			->setTcfAssetTypeLabel($source->getTcfAssetTypeLabel())
			->setTcfAssetPurposesLabel($source->getTcfAssetPurposesLabel())
		;

		foreach (Purpose::all() as $purpose) {
			$target
				->setPurposeName($purpose, $source->getPurposeName($purpose))
				->setPurposeDescription($purpose, $source->getPurposeDescription($purpose))
			;
		}

		if ($updateLegalText) {
			$legalTextSource = LegalText::byLocale($source);
			LegalText::byLocale($target)
				->setImprint($legalTextSource->getImprint())
				->setImprintScheme($legalTextSource->getImprintScheme())
				->setImprintUrl($legalTextSource->getImprintUrl())
				->setPrivacyPolicy($legalTextSource->getPrivacyPolicy())
				->setPrivacyScheme($legalTextSource->getPrivacyScheme())
				->setPrivacyUrl($legalTextSource->getPrivacyUrl())
				->setAccessibility($legalTextSource->getAccessibility())
				->setAccessibilityScheme($legalTextSource->getAccessibilityScheme())
				->setAccessibilityUrl($legalTextSource->getAccessibilityUrl())
				->setShowAccessibilityInFrontendWidget($legalTextSource->isShowAccessibilityInFrontendWidget())
				->setShowWidgetOnImprintPage($legalTextSource->isShowWidgetOnImprintPage())
				->setShowWidgetOnPrivacyPage($legalTextSource->isShowWidgetOnPrivacyPage())
				->settargetBlank($legalTextSource->istargetBlank())
				->settargetBlankPrivacy($legalTextSource->istargetBlankPrivacy())
			;
		}
	}
}
