<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\ClusterApi;
use App\Component\Api\Exception\ApiRestrictionException;
use App\Component\Api\Exception\BadRequestException;
use App\Component\Cache\Cache;
use App\Component\LoggerInterface;
use App\Component\ThumbnailExtractor\ThumbnailExtractorService;
use App\Controller\HostingApiController;
use App\Event\AccountUpdateEvent;
use App\Exception\UserAlreadyExistsException;
use App\Model\Country;
use App\Model\Domain;
use App\Model\Model;
use App\Model\User;
use App\Utils;
use DateTimeZone;
use Exception;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use App\Model\Statistic;

/**
 * @Route("/hosting/api/client", name="app_hosting_api_client_")
 */
class ClientApi extends HostingApiController
{
	private const CLIENT_SCHEMA = [
		'username' => 'string',
		'password' => 'string',
		'passwordHash' => 'string',
		'passwordLocked' => 'bool',
		'active' => 'bool',

		'firstName' => 'string',
		'lastName' => 'string',
		'emailAddress' => 'string',
		'company' => 'string',
		'street' => 'string',
		'zipCode' => 'string',
		'city' => 'string',
		'country' => 'string',
		'vatId' => 'string',
		'invoiceEmailAddress' => '?string',
		'dunningEmailAddress' => '?string',

		'domainMaxCount' => 'int',
		'widgetDisplayCount' => 'int',
		'whitelabelMaxCount' => 'int',
		'widgetDisplayMaxCount' => 'int',

		'crawlAllDomains' => 'bool',
		'allowAutoActivationOfNewEmbeddings' => 'bool',

		'autoCreateDomains' => 'bool',

		'defaultTimeZone' => 'string',

		'agencyId' => 'string',
		'domcklist' => 'int[]',
		'status' => 'int',
		'token' => 'string',

		'features' => 'array',
		'settings' => 'array',  // legacy

		'externalMetadata' => 'array',
	];

	/**
	 * Maps readable properties of the component to document field names that are used inside the model setter.
	 * @var array<string,string|null> $projectionMap
	 */
	protected static $projectionMap = [
		'id' => null,
		'apiKey' => 'apiKey',
	];

	/**
	 * @Route("", name="create", methods={"POST"})
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function createClient(Request $request, LoggerInterface $logger): Response
	{
		try {
			$payload = self::decodeRequestPayload($request);
			$this->parseRequestParameters($request);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		try {
			// Create the new user
			$user = User::create((string)($payload['username'] ?? ''));
		}
		catch (LogicException $exception) {
			return $this->json([
				'code' => 1,
				'error' => 'The client\'s username may not be empty.',
			], 400);
		}
		catch (UserAlreadyExistsException $exception) {
			// Release the exclusive lock
			User::reload();
			return $this->json([
				'code' => 2,
				'error' => 'The client\'s username is already in use.'
			], 400);
		}

		try {
			// Try to process the payload and update relevant properties
			self::migrateClientPayload($user, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Delete the user if any error occurred (rollback)
			User::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Delete the user if any error occurred (rollback)
			User::reload();

			// Dev time, let it rip
			throw $exception;
		}

		// Predefine properties that aren't accessible via the API
		$user
			->setRole(User::ROLE_CLIENT)
		;

		$user->save();

		$logger->logInfo('Client-API',"User created: {$user->getUsername()} [{$user->getId()}]");

		// Set up the proper repository contexts
		User::setApiUser($user->getApiKey());
		Domain::switchToUser($user);

		return $this->json($this->clientToSchemaArray($user), 201);
	}

	/**
	 * @Route("", name="read_all", methods={"GET"})
	 * @Route("_all", name="read_all_alias", methods={"GET"})
	 * @param Request $request
	 * @return Response
	 */
	public function readAllClients(Request $request): Response
	{
		try {
			$this->parseRequestParameters($request);

			// Apply projection on the database level to reduce the memory usage
			User::setProjection($this->buildModelProjection());
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		$users = User::getClients();

		try {
			// Apply the _limit and _skip query parameters
			$users = $this->applyPagination($users);
		}
		catch (ApiRestrictionException $exception) {
			return $exception->getResponse();
		}

		$dataForJson = array();
		$selDate = [
			'from' => time()-(60*60*24*(int)date("d")),
			'to' => time(),
		];

		foreach ($users as $k => $userVal)
		{
			if ($userVal && $userVal->getRole() == User::ROLE_CLIENT) {
				$domForJsonData = array();
				$domains = $userVal->getDomList();
				$user = User::byId($userVal->getId());
				if ($user === null) { continue; }

				foreach ($domains as $domKey => $domName) {
					//Keine Domain - dann auf 0
					if(stristr($domName,"none"))
					{
						$domForJsonData[$domName] = 0;
						continue;
					}
					//DOmain setzen
					Domain::switchToUser($user);
					$domData = Domain::byName($domName);

					//echt jetzt? Keine Daten? Dann auch auf 0
					if(empty($domData))
					{
						$domForJsonData[$domName] = 0;
						continue;
					}
					Domain::setDomainForRequest((string)$domData->getId());
					Statistic::reloadRepository();

					$statistics = 0;

					//Daten durchloopen und calls rausholen
					if (!empty($domData)) {
						foreach (Statistic::all() as $date => $stat) {
							if (strtotime($stat->getId()) >= $selDate['from'] && strtotime($stat->getId()) <= $selDate['to']) {
								$statistics += $stat->getCallCount();
							}
						}
					}
					//Für jede Domain explizit Crawlen j/n
                    $domForJsonData[$domData->getId()]['doCrawl'] = $domData->getDoCrawl();
					//Für jede Domain explizit Embeddings aktivieren j/n
					$domForJsonData[$domData->getId()]['activateEmbeddingsFoundByCrawler'] = $domData->isActivateEmbeddingsFoundByCrawler();
					//Anzahl der zu crawlenden Seiten für diese Domain
					$domForJsonData[$domData->getId()]['crawlPagesNumber'] = $domData->getDomainActualMaxCrawlBudget();
					//Die Crawl Frequenz die genutzt werden soll
					$domForJsonData[$domData->getId()]['crawlFrequency'] = $domData->getDomainExternalCrawlerFrequency();
					//Und die Filter die genutzt werden sollen
					$domForJsonData[$domData->getId()]['crawlUrlFilter'] = $domData->getDomainFilterForCrawler();
					//Und die Positiv Liste die genutzt werden sollen
					$domForJsonData[$domData->getId()]['crawlIncludeUrlFilter'] = $domData->getDomainIncludeFilterForCrawler();
					//Htpasswd Daten um dahinter crawlen zu können domainHtPassWd
					$domForJsonData[$domData->getId()]['domainHtPassWd'] = $domData->getDomainHtPassWd();

					//Name der Domain
					$domForJsonData[$domData->getId()]['domName'] = $domName;
					//Die Calls bisher
					$domForJsonData[$domData->getId()]['calls'] = $statistics;
				}

				//Daten zuweisen für Ausgabe
				$dataForJson[$k]['userId'] = $user->getId();
				$dataForJson[$k]['actualCallCount'] = $user->getActualCallCount();
				$dataForJson[$k]['domCKList'] = $user->getDomCKList();
				$dataForJson[$k]['dom30DayFreeList'] = $user->getDomCK30DayFreeList();
				$dataForJson[$k]['domainCountData'] = $domForJsonData;
				$dataForJson[$k]['externalMetadata'] = $user->getExternalMetadata();
				$dataForJson[$k]['accountCreatedTimestamp'] = $user->getCreateTimeStamp();
				$dataForJson[$k]['lastLoginTimestamp'] = $user->getLastLogin();
				$dataForJson[$k]['agencyId'] = $user->getAgencyId();
				$dataForJson[$k]['affiliateId'] = $user->getAgencyAffiliateId();

			}
		}

		// Fallback for the old implementation that provided a semantically wrong status code
		$successCode = $request->attributes->get('_route') == 'app_hosting_api_client_read_all_alias' ? 201 : 200;

		return $this->json(array_values(array_map(function (User $user, array $additionalData) {
			// Set up the proper repository contexts
			User::setApiUser($user->getApiKey());
			Domain::switchToUser($user);

			return $this->applyProjection(self::clientToSchemaArray($user) + $additionalData);
		}, $users, $dataForJson)), $successCode);
	}

	/**
	 * @Route("/{userId}", name="read", methods={"GET"})
	 * @param string $userId
	 * @param Request $request
	 * @return Response
	 */
	public function readClient(string $userId, Request $request): Response
	{
		try {
			$this->parseRequestParameters($request);
			self::assertClientExists($userId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var User $user */
		$user = User::byId($userId);
		assert($user instanceof User);

		return $this->json($this->clientToSchemaArray($user));
	}

	/**
	 * @Route("/by-apiKey/{apiKey}", name="by_apikey", methods={"HEAD", "GET"})
	 * @param string $apiKey
	 * @param Request $request
	 * @return Response
	 */
	public function clientByApiKey(string $apiKey, Request $request): Response
	{
		/** @var User|null $user */
		$user = User::byApiKey($apiKey);
		if ($user && $user->isClient()) {
			try {
				$userId = $user->getId();
				$this->parseRequestParameters($request);
				self::assertClientExists($userId);
			}
			catch (BadRequestException $exception) {
				return $exception->getResponse();
			}
		}
		else {
			return $this->json([
			'error' => 'Client not found.',
			], 404);
		}

		assert($user && $user instanceof User);

		$url = $this->generateUrl('app_hosting_api_client_read', ['userId'=>$user->getId()]);

		$response = $this->json($this->clientToSchemaArray($user));
		$response->headers->set("Content-Location", $url);
		return $response;
	}

	/**
	 * @Route("/by-username/{username}", name="by_username", methods={"HEAD", "GET"})
	 * @param string $username
	 * @param Request $request
	 * @return Response
	 */
	public function clientByUserName(string $username, Request $request): Response
	{
		/** @var User|null $user */
		$user = User::find($username);
		if ($user && $user->isClient()) {
			try {
				$userId = $user->getId();
				$this->parseRequestParameters($request);
				self::assertClientExists($userId);
			}
			catch (BadRequestException $exception) {
				return $exception->getResponse();
			}
		}
		else {
			return $this->json([
			'error' => 'Client not found.',
			], 404);
		}

		assert($user && $user instanceof User);

		$url = $this->generateUrl('app_hosting_api_client_read', ['userId'=>$user->getId()]);

		$response = $this->json($this->clientToSchemaArray($user));
		$response->headers->set("Content-Location", $url);
		return $response;
	}

	/**
	 * @Route("getid/{userName}", name="get_id", methods={"GET"})
	 * @param string $userName
	 * @param Request $request
	 * @return Response
	 */
	public function clientByNameOrMail(string $userName, Request $request): Response
	{
		//Checken ob es einen User gibt wo der String in den Daten zu finden ist
		if ($users = User::search($userName, User::ROLE_CLIENT)) {
			try {
				$userId = reset($users)->getId();
				$this->parseRequestParameters($request);
				self::assertClientExists($userId);
			}
			catch (BadRequestException $exception) {
				return $exception->getResponse();
			}
		}
		else {
			return $this->json([
			'error' => 'Client not found.',
			], 404);
		}

		/** @var User $user */
		$user = User::byId($userId);
		assert($user instanceof User);

		//checken ob es Username oder E-Mail mit den Daten gibt - dann ID ausgeben
		if($user->getUsername()==$userName || $user->getEmailAddress()==$userName)
		{
			$url = $this->generateUrl('app_hosting_api_client_read', ['userId'=>$user->getId()]);
			$response = $this->json($this->clientToSchemaArray($user));
			$response->headers->set("Content-Location", $url);
			return $response;
		}
		else{
			return $this->json([
				'error' => 'Client not found.',
			], 404);
		}
	}

	/**
	 * @Route("/{userId}", name="update", methods={"PUT"})
	 * @param string $userId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function updateClient(string $userId, Request $request, LoggerInterface $logger): Response
	{
		try {
			$payload = self::decodeRequestPayload($request);
			$this->parseRequestParameters($request);
			self::assertClientExists($userId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var User $user */
		$user = User::byId($userId);
		assert($user instanceof User);

		try {
			// Try to process the payload and update relevant properties
			self::migrateClientPayload($user, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Delete the user if any error occurred (rollback)
			User::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Delete the user if any error occurred (rollback)
			User::reload();

			// Dev time, let it rip
			throw $exception;
		}

		$event = new AccountUpdateEvent($user, AccountUpdateEvent::CAUSE_API);
		if ($payload and (count($payload) > 1 or !array_key_exists('password', $payload) and !array_key_exists('passwordHash', $payload))) {
			$event->setAccountInfoChanged();
		}
		if (array_key_exists('password', $payload) or array_key_exists('passwordHash', $payload)) {
			$event->setPasswordChanged();
		}
		Utils::dispatchEvent(AccountUpdateEvent::NAME_PRE_SAVE, $event);

		$user->save();

		$logger->logInfo('Client-API',"User updated: {$user->getUsername()} [$userId]");
		Utils::dispatchEvent(AccountUpdateEvent::NAME_POST_SAVE, $event);

		return $this->json($this->clientToSchemaArray($user));
	}

	/**
	 * @Route("/{userId}", name="delete", methods={"DELETE"})
	 * @param string $userId
	 * @param LoggerInterface $logger
	 * @param ThumbnailExtractorService $thumbnails
	 * @return Response
	 */
	public function deleteClient(string $userId, LoggerInterface $logger, ThumbnailExtractorService $thumbnails): Response
	{
		try {
			self::assertClientExists($userId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var User $user */
		$user = User::byId($userId);
		assert($user instanceof User);

		$thumbnails->clearCache($user);

		foreach ($user->getSubusers() as $subuser) {
			User::delete($subuser->getId(), false);
		}
		//Cache löschen
		Cache::clearUser($user);
		// Broadcast clear cache request to other peers of this cluster
		ClusterApi::getInstance()->callRemoteApiBroadcast('app_cluster_api_user_cache_clear', [
			'userId' => $user->getId(),
			'apiKey' => $user->getApiKey(),
		], 'DELETE');

		$logger->logInfo('Client-API',"User deleted: {$user->getUsername()} [$userId]");
		User::delete($userId);
		// Repository schreiben
		User::flush();

		// Liste der veränderten Models leeren, damit der CacheHandler nicht annehmen muss,
		// dass sich Daten anderer User geändert haben
		Model::clearWrittenModelsClassNames();

		return new Response('', 204);
	}

	/**
	 * @param User $user
	 * @param array $payload
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migrateClientPayload(User $user, array $payload): void
	{
		self::checkPayloadDataTypes($payload, self::CLIENT_SCHEMA);

		if (array_key_exists('username', $payload) or $user->getUsername() === '') {
			if (($payload['username'] ?? '') === '') {
				throw new InvalidArgumentException('Empty or missing username.', 1);
			}

			if ($user->getUsername() !== $payload['username']) {
				$otherUser = User::find($payload['username']);
				if ($otherUser !== null and $otherUser->getId() !== $user->getId()) {
					throw new InvalidArgumentException('The client\'s username is already in use.', 2);
				}
				else {
					$user->setUsername($payload['username']);
				}
			}
		}

		if (array_key_exists('password', $payload) || array_key_exists('passwordHash', $payload)) {
			if (array_key_exists('password', $payload) && strlen($payload['password']) > 0) {
				$user->setPassword($payload['password']);
			}
			elseif (array_key_exists('passwordHash', $payload) && strlen($payload['passwordHash']) > 0) {
				$user->setPasswordHash($payload['passwordHash']);
			}
			else {
				throw new InvalidArgumentException('Invalid password or passwordHash, it mustn\'t be empty.', 1);
			}
		}
		elseif (empty($user->getPasswordHash())) {
			throw new InvalidArgumentException('Provide either password or passwordHash.', 1);
		}

		if (array_key_exists('active', $payload)) {
			$user->setActive($payload['active']);
		}

		if (array_key_exists('passwordLocked', $payload)) {
			$user->setPasswordLocked($payload['passwordLocked']);
		}

		if (array_key_exists('firstName', $payload)) {
			$user->setFirstName($payload['firstName']);
		}
		if (array_key_exists('lastName', $payload)) {
			$user->setLastName($payload['lastName']);
		}
		if (array_key_exists('emailAddress', $payload)) {
			if ($payload['emailAddress'] || Utils::validateEmailAddress($payload['emailAddress'])) {
				$user->setEmailAddress($payload['emailAddress']);
			}
			else {
				throw new InvalidArgumentException('The client\'s email address is either empty or invalid.', 3);
			}
		}
		if (array_key_exists('company', $payload)) {
			$user->setCompany($payload['company']);
		}
		if (array_key_exists('street', $payload)) {
			$user->setStrNr($payload['street']);
		}
		if (array_key_exists('zipCode', $payload)) {
			$user->setPlz($payload['zipCode']);
		}
		if (array_key_exists('city', $payload)) {
			$user->setCity($payload['city']);
		}
		if (array_key_exists('country', $payload)) {
			$user->setCountry(Country::find($payload['country']));
		}
		if (array_key_exists('vatId', $payload)) {
			$user->setVatId($payload['vatId']);
		}
		if (array_key_exists('invoiceEmailAddress', $payload)) {
			if ($payload['invoiceEmailAddress'] === null or $payload['invoiceEmailAddress'] === '') {
				$user->setInvoiceEmailAddress(null);
			}
			elseif (Utils::validateEmailAddress($payload['invoiceEmailAddress'])) {
				$user->setInvoiceEmailAddress($payload['invoiceEmailAddress']);
			}
			else {
				throw new InvalidArgumentException('The client\'s invoice email address is invalid.', 3);
			}
		}
		if (array_key_exists('dunningEmailAddress', $payload)) {
			if ($payload['dunningEmailAddress'] === null or $payload['dunningEmailAddress'] === '') {
				$user->setDunningEmailAddress(null);
			}
			elseif (Utils::validateEmailAddress($payload['dunningEmailAddress'])) {
				$user->setDunningEmailAddress($payload['dunningEmailAddress']);
			}
			else {
				throw new InvalidArgumentException('The client\'s dunning email address is invalid.', 3);
			}
		}
		if (array_key_exists('domainMaxCount', $payload)) {
			$user->setMaxDomainCount(max(-1, $payload['domainMaxCount']));
		}

		if (array_key_exists('widgetDisplayCount', $payload)) {
			$user->setCallCount(max(0, $payload['widgetDisplayCount']));
		}
		if (array_key_exists('whitelabelMaxCount', $payload)) {
			$user->setMaxWLCount(max(-1, $payload['whitelabelMaxCount']));
		}
		if (array_key_exists('widgetDisplayMaxCount', $payload)) {
			$user->setMaxCallCount(max(-1, $payload['widgetDisplayMaxCount']));
		}

		if (array_key_exists('crawlAllDomains', $payload)) {
			$user->setCrawlAllDomains(max(-1, $payload['crawlAllDomains']));
		}
		if (array_key_exists('allowAutoActivationOfNewEmbeddings', $payload)) {
			$user->setAllowAutoActivationOfNewEmbeddings(max(-1, $payload['allowAutoActivationOfNewEmbeddings']));
		}

		//Agency ID - um die auch von außen setzen zu können - nötig für CK
		if (array_key_exists('agencyId', $payload)) {
			$user->setAgencyId($payload['agencyId']);
		}

		if (array_key_exists('autoCreateDomains', $payload)) {
			$user->setAutoAddDomains($payload['autoCreateDomains']);
		}

		if (array_key_exists('defaultTimeZone', $payload)) {
			try {
				if ($timeZone = new DateTimeZone((string)($payload['defaultTimeZone'] ?? ''))) {
					$user->setDefaultTimeZone($timeZone);
				}
			} catch (Exception $exception) {}
		}

		// DomCKList (Clickskeks-spezifisch)
		if (array_key_exists('domcklist', $payload)) {
			$user->setDomCKList($payload['domcklist']);
		}

		// Metadaten für Drittsysteme
		if (array_key_exists('externalMetadata', $payload)) {
			$user->setExternalMetadata($payload['externalMetadata']);
		}

		// agencyId
		if (array_key_exists('agencyId', $payload)) {
			$user->setAgencyId($payload['agencyId']);
		}

		if (array_key_exists('status', $payload)) {
			$user->setStatus($payload['status']);
		}

		if (array_key_exists('token', $payload)) {
			$user->setToken($payload['token']);
		}

		if (array_key_exists('features', $payload)) {
			$user->setSettings($payload['features'] + $user->getSettings() + User::buildDefaultSettings(false));
		}
		elseif (array_key_exists('settings', $payload)) { // legacy
			$user->setSettings($payload['settings'] + $user->getSettings() + User::buildDefaultSettings(false));
		}
		elseif (empty($user->getSettings())) {
			$user->setSettings(User::buildDefaultSettings(true));
		}

		// Undocumented property -- TODO: Move to separate route; e.g. "/hosting/api/client/{clientId}/resetCallCount"
		if (array_key_exists('isUpgrade', $payload)) {
			$user->resetCallCountForNextMonth();
		}
	}

	/**
	 * @param User $user
	 * @return array
	 */
	private function clientToSchemaArray(User $user): array
	{
		$additionalFields = [];

		// Add individual fields for the Clickskeks environment
		if (Utils::isClickskeksCloud()) {
			$additionalFields += [
				'domcklist' => $user->getDomCKList(),
			];
		}

		return $this->applyProjection([
			'id' => $user->getId(),
			'username' => $user->getUsername(),
			'passwordHash' => $user->getPasswordHash(),
			'userPasswordHash' => $user->getPasswordHash(), // legacy
			'active' => $user->isActive(),
			'firstName' => $user->getFirstName(),
			'lastName' => $user->getLastName(),
			'emailAddress' => $user->getEmailAddress(),
			'company' => $user->getCompany(),
			'street' => (string)$user->getStrNr(),
			'zipCode' => (string)$user->getPlz(),
			'city' => (string)$user->getCity(),
			'country' => $user->getCountryCode(),
			'vatId' => (string)$user->getVatId(),
			'invoiceEmailAddress' => $user->getInvoiceEmailAddress(),
			'dunningEmailAddress' => $user->getDunningEmailAddress(),
			'domainCount' => Domain::count(),
			'domainMaxCount' => $user->getMaxDomainCount(),
			'whitelabelCount' => Domain::whitelabelCountForAllDomains(),
			'whitelabelMaxCount' => $user->getMaxWLCount(),
			'widgetDisplayCount' => $user->getActualCallCount(),
			'widgetDisplayMaxCount' => $user->getMaxCallCount(),
			'crawlAllDomains' => $user->isCrawlAllDomains(),
			'allowAutoActivationOfNewEmbeddings' => $user->isAllowAutoActivationOfNewEmbeddings(),
			'autoCreateDomains' => $user->isAutoAddDomains(),
			'defaultTimeZone' => $user->getDefaultTimeZone(),
			'apiKey' => $user->getApiKey(),
			'features' => $user->getSettings(),
			'settings' => $user->getSettings(), // legacy
			'externalMetadata' => $user->getExternalMetadata(),
		] + $additionalFields);
	}
}
