<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\Exception\BadRequestException;
use App\Component\LoggerInterface;
use App\Component\Mail\MailMessage;
use App\Component\Tcf\v2\Model\Google\AdTechProvider;
use App\Component\Tcf\v2\Model\SpecialFeature;
use App\Component\Tcf\v2\Model\Vendor;
use App\Config;
use App\Controller\HostingApiController;
use App\Model\Country;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\Index\DomainIndex;
use App\Model\LegalText;
use App\Model\User;
use App\Utils;
use DateTime;
use DateTimeZone;
use Exception;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Mailer\MailerInterface;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/hosting/api/client/{userId}/domain", name="app_hosting_api_domain_")
 */
class DomainApi extends HostingApiController
{
	private const DOMAIN_SCHEMA = [
		'name' => 'string',
		'country' => '?string',
		'timeZone' => '?string',
		'whitelabel' => 'bool',

		'consentStorage' => 'string',
		'consentStorageNamespaceUrlParameter' => 'string',
		'consentCookieLifetime' => 'int',
		'consentCookieSecure' => 'bool',
		'consentLifetime' => 'int',
		'consentLifetimeForPartialConsent' => '?int',
		'consentResetBeforeDate' => 'date',
		'consentResetOnLegalChange' => 'bool',
		'consentSharingDomains' => 'string[]',
		'consentSharingCrossDomain' => 'bool',

		'frontendWidgetEnabled' => 'bool',
		'frontendWidgetCache' => 'bool',
		'frontendWidgetMinifyCode' => 'bool',
		'partOfAdAndTrackingFreeNetwork' => 'bool',

		'managementStructure' => 'string',

		'deleteUnknownCookies' => 'bool',
		'deleteUnknownCookiesForceReload' => 'bool',
		'requireConsentForMandatoryEmbeddings' => 'bool',
		'declineButtonEnforced' => 'bool',
		'forcedCssImportant' => 'bool',
		'lcpOptimization' => 'bool',
		'onlyActiveOnMatchingDomain' => 'bool',
		'embeddingConsentViaJsApi' => 'bool',
		'dispatchAdditionalGtmEvents' => 'bool',
		'detectCloakedScripts' => 'bool',
		'prioritizeLanguageHeaderOverLangAttribute' => 'bool',
		'widgetDelayInMilliseconds' => 'int',
		'widgetExclusionUrls' => 'string[]',

		'allowedScriptMarkers' => 'string[]',
		'blockedScriptMarkers' => 'string[]',

		'evalInlineScripts' => 'bool',

		'recordScriptsOnWebsite' => 'bool',
		'blockNewScripts' => 'bool',
		'blockFirstPartyScripts' => 'bool',
		'blockInlineScripts' => 'bool',
		'permanentScanEnabled' => 'bool',

		'tcfEnabled' => 'bool',
		'tcfEnableNewVendorsAfterGvlUpdate' => 'bool',
		'tcfVendorBlacklist' => 'int[]',
		'tcfSpecialFeatures' => 'int[]',
		'tcfGoogleAdditionalConsentMode' => 'bool',
		'tcfGoogleAdditionalConsentAdProviders' => 'int[]',
		'tcfUrlMacroAutoAppend' => 'bool',
		'tcfUrlMacroHandling' => 'bool',

        'blockIframes' => 'bool',
		'doCrawl' => 'bool',
		'activateEmbeddingsFoundByCrawler' => 'bool',
		'rememberIframeConsentPerDomain' => 'bool',
		'appendConsentSwitchToIframes' => 'bool',
		'iframeBlockMode' => 'string',
		'iframeMarkers' => 'string[]',

		'cookieDeclarationTableHeadingStartLevel' => 'int',
		'cookieDeclarationTablePrependConsentChangeLink' => 'bool',
		'cookieDeclarationTablePrependIntroText' => 'bool',

		'imprintDisplayMode' => 'string', // LegalText
		'imprintOpenInNewTab' => 'bool', // LegalText
		'imprintShowWidget' => 'bool', // LegalText

		'privacyPolicyDisplayMode' => 'string', // LegalText
		'privacyPolicyOpenInNewTab' => 'bool', // LegalText
		'privacyPolicyShowWidget' => 'bool', // LegalText

		'accessibilityDisplayMode' => 'string', // LegalText
		'accessibilityLinkEnabled' => 'bool', // LegalText

		'realCrawlCount' => 'int', // Count tatsächlich gecrawlte Urls
		'lastScanDate' => 'string', // DateTime | timestamp
		'domainMaxCrawlBudget' => 'int', // Maximaler Crawl-Budget
		'domainExternalCrawlerFrequency' => 'int', // Crawl Frequency

		//Hier die Daten die für den Versand der Crawler Mails genutzt werden
		'uncontrolledData' => 'string', // Daten für den Platzhalter in Crawler Mails

	];


	/**
	 * @Route("", name="create", methods={"POST"})
	 * @param string $userId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function createDomain(string $userId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertClientExists($userId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		$user = User::byId($userId);
		assert($user instanceof User);

		if ($user->isDomainLimitReached()) {
			return $this->json([
				'code' => 20,
				'error' => 'Can\'t create domain, limit reached.',
			], 403);
		}

		// Check the domain limit with regard to consentSharingDomains
		if (isset($payload['consentSharingDomains']) && count($payload['consentSharingDomains']) > 0) {
			// Include the domain itself (+1) in the count
			$requestedSlots = count(array_diff($payload['consentSharingDomains'], [$payload['name'] ?? ''])) + 1;

			// Abort if there aren't enough free domain slots
			if (!$user->hasFreeDomainSlots($requestedSlots)) {
				return $this->json([
					'code' => 20,
					'error' => 'Can\'t create domain, limit reached due to consentSharingDomains.',
				], 403);
			}
		}

		if ($payload['whitelabel'] ?? false) {
			$whitelabelCount = Domain::whitelabelCountForAllDomains();
			$whitelabelMaxCount = $user->getMaxWLCount();

			if ($whitelabelMaxCount >= 0 && $whitelabelCount >= $whitelabelMaxCount) {
				return $this->json([
					'code' => 21,
					'error' => 'Can\'t use whitelabel, limit reached.',
				], 403);
			}
		}

		// Create the new domain
		$domain = Domain::createForUser($user);
		Domain::setDomainForRequest($domain->getId());

		// Enforce embedding structure as the default setting
		$domain
			->setManagementStructurePreLegalConsent(true)
			->setManagementStructure(Domain::MANAGEMENT_STRUCTURE_EMBEDDING)
		;

		try {
			// Try to process the payload and update relevant properties
			self::migrateDomainPayload($domain, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Delete the domain if any error occurred (rollback)
			Domain::delete($domain->getId());
			Domain::setDomainForRequest('');

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Delete the domain if any error occurred (rollback)
			Domain::delete($domain->getId());
			Domain::setDomainForRequest('');

			// Dev time, let it rip
			throw $exception;
		}

		// Determine the latest vendor by ID to complete the blacklist after a GVL update
		/** @var Vendor $mostRecentVendor */
		$mostRecentVendor = array_reduce(Vendor::all(), function (?Vendor $carry, Vendor $item) {
			return $carry ? ($item->getId() > $carry->getId() ? $item : $carry) : $item;
		});

		// Store ID gaps within the GVL to properly complete the blacklist after a GVL update
		$unusedVendorIds = array_values(array_filter(range(1, (int)$mostRecentVendor->getId()), function (int $id) {
			return Vendor::exists((string)$id) == false;
		}));

		// Predefine properties that aren't accessible via the API
		$domain
			->setMostRecentTcfVendor($mostRecentVendor)
			->setOnBoarding(true)
			->setUnusedTcfVendorIds($unusedVendorIds)
		;

		LegalText::flush();
		$domain->save();

		$logger->logInfo("Domain-API", "Domain created: {$domain->getName()} [{$domain->getUser()->getId()}::{$domain->getId()}]");

		// Index neu generieren
		DomainIndex::rebuild();

		//Bei der CK Cloud noch bei jeder Erzeugung via API noch die 30 Tage Ablauf setzen
		if(Utils::isClickskeksCloud())
		{
			$freeData = $user->getDomCK30DayFreeList();
			$freeData[$domain->getId()] = array("id"=>$domain->getId(),"start"=>time(),"name"=>$domain->getName());
			$user->setDomCK30DayFreeList($freeData);
			$user->save();
		}

		// Benutzer aktualisieren
		$user
			->updateDomainInfo()
			->save();

		return $this->json(self::domainToSchemaArray($domain), 201);
	}

	/**
	 * @Route("", name="read_all", methods={"GET"})
	 * @param string $userId
	 * @return JsonResponse
	 */
	public function readAllDomains(string $userId): JsonResponse
	{
		try {
			self::assertClientExists($userId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		return $this->json(array_values(array_map(function (Domain $domain) {
			return self::domainToSchemaArray($domain);
		}, Domain::all())), 200);
	}

	/**
	 * @Route("/{domainId}", name="read", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @return JsonResponse
	 */
	public function readDomain(string $userId, string $domainId): JsonResponse
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Domain $domain */
		$domain = Domain::find($domainId);
		assert($domain instanceof Domain);

		return $this->json(self::domainToSchemaArray($domain), 200);
	}

	/**
	 * @Route("/{domainId}", name="update", methods={"PUT"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function updateDomain(string $userId, string $domainId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		$user = User::byId($userId);
		assert($user instanceof User);
		/** @var Domain $domain */
		$domain = Domain::find($domainId);
		assert($domain instanceof Domain);

		// Check the domain limit with regard to consentSharingDomains
		if (isset($payload['consentSharingDomains']) && count($payload['consentSharingDomains']) > 0) {
			// Determine the number of domains that require a free slot
			$requestedSlots = count(array_diff(
				$payload['consentSharingDomains'],
				$domain->getShareDomainNames(), // Exclude domains that are already in the list
				[$payload['name'] ?? $domain->getUrlHost()] // Exclude the domain itself
			));

			// Abort if there aren't enough free domain slots
			if (!$user->hasFreeDomainSlots($requestedSlots)) {
				return $this->json([
					'code' => 20,
					'error' => 'Can\'t update domain, limit reached due to consentSharingDomains.',
				], 403);
			}
		}

		// Check limit only if whitelabel will change to enabled
		if ($payload['whitelabel'] ?? false and $domain->hasWhitelabel() == false) {
			$whitelabelCount = Domain::whitelabelCountForAllDomains();
			$whitelabelMaxCount = $user->getMaxWLCount();

			if ($whitelabelMaxCount >= 0 && $whitelabelCount >= $whitelabelMaxCount) {
				return $this->json([
					'code' => 21,
					'error' => 'Can\'t use whitelabel, limit reached.',
				], 403);
			}
		}

		try {
			// Try to process the payload and update relevant properties
			self::migrateDomainPayload($domain, $payload);
		}
		catch (InvalidArgumentException $exception) {
			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		// Determine the latest vendor by ID to complete the blacklist after a GVL update
		/** @var Vendor $mostRecentVendor */
		$mostRecentVendor = array_reduce(Vendor::all(), function (?Vendor $carry, Vendor $item) {
			return $carry ? ($item->getId() > $carry->getId() ? $item : $carry) : $item;
		});

		// Store ID gaps within the GVL to properly complete the blacklist after a GVL update
		$unusedVendorIds = array_values(array_filter(range(1, (int)$mostRecentVendor->getId()), function (int $id) {
			return Vendor::exists((string)$id) == false;
		}));

		// Update properties that aren't accessible via the API
		$domain
			->setMostRecentTcfVendor($mostRecentVendor)
			->setOnBoarding(true)
			->setUnusedTcfVendorIds($unusedVendorIds)
		;

		LegalText::flush();
		$domain->save();

		// Index neu generieren
		DomainIndex::rebuild();

		// Benutzer aktualisieren
		$user
			->updateDomainInfo()
			->save();

		$logger->logInfo("Domain-API", "Domain updated: {$domain->getName()} [{$domain->getUser()->getId()}::{$domain->getId()}]");

		return $this->json(self::domainToSchemaArray($domain), 200);
	}

	/**
	 * @Route("/{domainId}", name="delete", methods={"DELETE"})
	 * @param string $userId
	 * @param string $domainId
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteDomain(string $userId, string $domainId, LoggerInterface $logger): Response
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		$user = User::byId($userId);
		assert($user instanceof User);
		/** @var Domain $domain */
		$domain = Domain::find($domainId);
		assert($domain instanceof Domain);

		$logger->logInfo("Domain-API", "Domain deleted: {$domain->getName()} [{$domain->getUser()->getId()}::{$domain->getId()}]");
		Domain::delete($domainId);

		// Index neu generieren
		DomainIndex::rebuild();

		// Benutzer aktualisieren
		$user
			->updateDomainInfo()
			->save();

		return new Response('', 204);
	}

	/**
	 * @Route("/{domainId}/crawlermail", name="crawler_mail", methods={"POST"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @param \Symfony\Component\Mailer\MailerInterface $mailer
	 *
	 * @return JsonResponse
	 * @throws \Symfony\Component\Mailer\Exception\TransportExceptionInterface
	 * @internal - nur zum Gebrauch durch die CCM19 Crawler API
	 */
	public function crawlerMailDomain(string $userId, string $domainId, Request $request, LoggerInterface $logger, MailerInterface $mailer): JsonResponse
	{

		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}
		//Fürs Logging der Mailversendung
		$startTime = time();

		$user = User::byId($userId);
		assert($user instanceof User);

		//Checken ob der User überhaupt Mails bekommen will
		$userMailPrefs = $user->getMailingPreferences();
		if($userMailPrefs['EmbeddingNotification']===false)
		{
			$logger->logWarning("Domain-API", "CrawlerMail Aufruf aber Mail soll nicht versendet werden für = : ".$user->getUsername());
			return new JsonResponse(array("error"=>"User does not want E-Mails"), 405);
		}

		/** @var Domain $domain */
		$domain = Domain::find($domainId);
		assert($domain instanceof Domain);

		$payload = self::decodeRequestPayload($request);

		$templateId = 1;
		if($domain->isActivateEmbeddingsFoundByCrawler())
		{
			$templateId = 2;
		}

		//Wenn die Daten die per POST kommen nicht leer sind
		// 'uncontrolledData' => 'string', Daten für den Platzhalter in Crawler Mails
		//'crawlerMailTemplateId' => 'int', Welches Template genutzt werden soll, um die Mails zu versenden
		if(!empty($payload['uncontrolledData']))
		{
			$config = Config::getInstance();
			$subject = "";
			$body = "";

			//Je nach Version die TO Adresse setzen
			if(Utils::isExtendedEdition()) {
				//E-Mail Adresse des Users der Domain...
				$mailto = $user->getEmailAddress();
				$domainName = $domain->getName();
			}
			elseif (Utils::hasMultiDomainSupport()) {
				//Betreiber Mail
				$mailto = $config->get('email');
				$domainName = $domain->getName();
			}
			else {
				//Single Instanz - aber ob das tatsächlich gebraucht wird...
				$mailto = $config->get('email');
				$domainName = "";
			}
			//Id 1 = nicht aktiviert - 2 ist dann aktiviert
			if($templateId===1)
			{
				//Check ob Mail überhaupt gesendet werden soll - falls nicht, dann mit Error raus
				if(!$config->get('sub_crawler_email_check'))
				{
					$logger->logWarning("Domain-API", "CrawlerMail Aufruf aber Mail Versand ist nicht aktiviert für Found by Crawler ");
					return new JsonResponse(array("error"=>"Mailsend type 1 not activated"), 405);
				}

				$subject = $config->get('sub_crawler_email_betreff');
				$body = $config->get('sub_crawler_email_content');
			}
			//ID 2 ist dann aktiviert
			if($templateId===2){

				//Check ob Mail überhaupt gesendet werden soll - falls nicht, dann mit Error raus
				if(!$config->get('sub_crawler_email_check'))
				{
					$logger->logWarning("Domain-API", "CrawlerMail Aufruf aber Mail Versand ist nicht aktiviert für Found by Crawler AND activate ");
					return new JsonResponse(array("error"=>"Mailsend type 2 not activated"), 405);
				}

				$subject = $config->get('sub_crawler_activated_email_betreff');
				$body = $config->get('sub_crawler_activated_email_content');
			}

			if(empty($subject)) {
				$logger->logError("Domain-API", "CrawlerMail Aufruf aber Mail Content nicht definiert für TemplateId = : ".(int) $templateId);
				return new JsonResponse(array("error"=>"TemplateID wrong"), 405);
			}

			//Mail zusammenstellen
			$subject = str_ireplace("{{ domainvar }}",$domainName,$subject);

			$body = str_ireplace("#uncontrolledData#",$payload['uncontrolledData'],$body);
			$body = str_ireplace("{{ uncontrolledData }}",$payload['uncontrolledData'],$body);
			$body = str_ireplace("{{ domainvar }}",$domainName,$body);

			//Noch den Link zu den Embeddings bauen - wir gehen hier davon aus, dass es kein Unterverzeichnis gibt
			$link = "https://".$request->getHttpHost()."/domains/".$domain->getId()."/embeddings"; //
			$body = str_ireplace("{{ embeddingLink }}",$link,$body);

			//Mail versenden...
			$message = (new MailMessage($subject))
				->from($config->get('email')) // Von ist Admin E-Mail
				->to($mailto)   // to ist die MailAdresse des jeweiligen Kundens... oder Admin wenn Base...
				->listId('CookieNotification.'.$user->getId())
				->unsubscribe($user, 'CookieNotification')
				->autoSubmitted()
				->text($body);
			try {
				$mailer->send($message);
				$stopTime = time();
				$took = ($startTime - $stopTime);
				$took = " (".$took.") sec.";
				$logger->logInfo('CrawlerNotification', "Domain-API - Sent notification type ".(int) $templateId." mail to ".$mailto." ".$took);
			} catch (Exception $e) {
				$stopTime = time();
				$took = ($startTime - $stopTime);
				$took = " (".$took.") sec.";
				$logger->logWarning('CrawlerNotification', "Domain-API - Could not send notification  type ".(int) $templateId."  email to ".$mailto." ".$took);
			}
		}
		else{
			$logger->logError("Domain-API", "CrawlerMail Aufruf mit falschen uncrontolledData - ".(int) $templateId);
			return new JsonResponse(array("error"=>"TemplateID && uncrontolledData wrong"), 405);
		}

		$logger->logInfo("Domain-API", "CrawlerMail Aufruf mit Daten: {$domain->getName()} [{$domain->getUser()->getId()}::{$domain->getId()}]");
		return new JsonResponse(array("success"=>"Mail send"), 200);
	}

	/**
	 * @param Domain $domain
	 * @param array $payload
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migrateDomainPayload(Domain $domain, array $payload): void
	{
		self::checkPayloadDataTypes($payload, self::DOMAIN_SCHEMA);

		if (array_key_exists('name', $payload) || empty($domain->getName())) {
			if ($payload['name'] ?? '') {
				$domainInfo = parse_url(filter_var($payload['name'], FILTER_SANITIZE_URL));

				if (isset($domainInfo['host'])) {
					$urlHost = $domainInfo['host'];
					$urlPath = '/'.ltrim($domainInfo['path'] ?? '/', '/');
					$urlScheme = $domainInfo['scheme'] ?? 'https';
				}
				else {
					$parts = explode('/', trim($payload['name']), 2);
					$urlHost = $parts[0];
					$urlPath = '/'.ltrim($parts[1] ?? '', '/');
					$urlScheme = 'https';
				}

				$domain
					->setName($urlHost)
					->setUrlPath($urlPath)
					->setUrlScheme($urlScheme)
				;
			}
			else {
				throw new InvalidArgumentException('Empty or missing domain name.', 1);
			}
		}

		if (array_key_exists('country', $payload)) {
			if ($payload['country']) {
				$country = Country::find($payload['country']);
				if (!$country) {
					throw new InvalidArgumentException('Invalid country code.', 2);
				}
			}
			else {
				$country = null;
			}

			$domain->setCountry($country);
		}

		if (array_key_exists('timeZone', $payload)) {
			try {
				$timeZone = new DateTimeZone((string)($payload['timeZone'] ?? ''));
			} catch (Exception $exception) {
				$timeZone = null;
			}
			$domain->setTimeZone($timeZone);
		}

		if (array_key_exists('whitelabel', $payload)) {
			$domain->setWhitelabel($payload['whitelabel'] ?? false);
		}

		if (array_key_exists('consentStorage', $payload)) {
			if (in_array($payload['consentStorage'], Domain::ALLOWED_CONSENT_STORAGE_VALUES)) {
				$domain->setConsentStorage($payload['consentStorage']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid value for consentStorage, allowed values: '.implode(', ', Domain::ALLOWED_CONSENT_STORAGE_VALUES),
					3
				);
			}
		}
		if (array_key_exists('consentStorageNamespaceUrlParameter', $payload)) {
			$domain->setConsentStorageNamespaceUrlParameter($payload['consentStorageNamespaceUrlParameter']);
		}
		if (array_key_exists('consentCookieLifetime', $payload)) {
			$domain->setConsentCookieLifetime(max(0, (int)$payload['consentCookieLifetime']));
		}
		if (array_key_exists('consentCookieSecure', $payload)) {
			$domain->setSecureConsentCookie((bool)$payload['consentCookieSecure']);
		}
		if (array_key_exists('consentLifetime', $payload)) {
			$domain->setConsentLifetime($payload['consentLifetime']);
		}
		if (array_key_exists('consentLifetimeForPartialConsent', $payload)) {
			$domain->setConsentLifetimeForPartialConsent($payload['consentLifetimeForPartialConsent']);
		}
		if (array_key_exists('consentResetBeforeDate', $payload)) {
			$domain->setconsentCookieReset((string)$payload['consentResetBeforeDate']);
		}
		if (array_key_exists('consentResetOnLegalChange', $payload)) {
			$domain->setConsentResetOnRelevantChange((bool)$payload['consentResetOnLegalChange']);
		}
		if (array_key_exists('consentSharingDomains', $payload)) {
			$domain->setShareDomainNames(array_values(array_filter(array_map('trim', (array)$payload['consentSharingDomains']))));
		}
		if (array_key_exists('consentSharingCrossDomain', $payload)) {
			$domain->setCrossDomainSharing((bool)$payload['consentSharingCrossDomain']);
		}

		if (array_key_exists('frontendWidgetEnabled', $payload)) {
			if ($payload['frontendWidgetEnabled']) {
				$domain->setfrontendLegalEnabled(true);
			}
			$domain->setFrontendWidgetEnabled((bool)$payload['frontendWidgetEnabled']);
		}
		if (array_key_exists('frontendWidgetCache', $payload)) {
			$domain->setWidgetCachingEnabled((bool)$payload['frontendWidgetCache']);
		}
		if (array_key_exists('frontendWidgetMinifyCode', $payload)) {
			$domain->setCodeMinification((bool)$payload['frontendWidgetMinifyCode']);
		}
		if (array_key_exists('partOfAdAndTrackingFreeNetwork', $payload)) {
			$domain->setPartOfAdAndTrackingFreeNetwork($payload['partOfAdAndTrackingFreeNetwork']);
		}

		if (array_key_exists('managementStructure', $payload)) {
			try {
				if ($payload['managementStructure'] == Domain::MANAGEMENT_STRUCTURE_EMBEDDING) {
					$domain->setManagementStructurePreLegalConsent(true);
				}
				$domain->setManagementStructure($payload['managementStructure'] ?? '');
			}
			catch (LogicException $exception) {
				throw new InvalidArgumentException(
					'Invalid value for managementStructure, allowed values: '.implode(', ', Domain::ALLOWED_MANAGEMENT_STRUCTURE_VALUES),
					4
				);
			}
		}

		if (array_key_exists('deleteUnknownCookies', $payload)) {
			$domain->setDeleteUnknownCookies((bool)$payload['deleteUnknownCookies']);
		}
		if (array_key_exists('deleteUnknownCookiesForceReload', $payload)) {
			$domain->setDeleteUnknownCookiesForceReload((bool)$payload['deleteUnknownCookiesForceReload']);
		}
		if (array_key_exists('requireConsentForMandatoryEmbeddings', $payload)) {
			$domain->setLoadTechnicallyNecessaryEmbeddingsWithoutConsent($payload['requireConsentForMandatoryEmbeddings'] == false);
		}
		if (array_key_exists('declineButtonEnforced', $payload)) {
			$domain->setDeclineButtonEnforced($payload['declineButtonEnforced']);
		}
		if (array_key_exists('forcedCssImportant', $payload)) {
			$domain->setForcedCssImportant($payload['forcedCssImportant']);
		}
		if (array_key_exists('lcpOptimization', $payload)) {
			$domain->setLcpOptimizationEnabled($payload['lcpOptimization']);
		}
		if (array_key_exists('onlyActiveOnMatchingDomain', $payload)) {
			$domain->setOnlyActiveOnMatchingDomain($payload['onlyActiveOnMatchingDomain']);
		}
		if (array_key_exists('embeddingConsentViaJsApi', $payload)) {
			$domain->setAllowEmbeddingConsentViaJsApi($payload['embeddingConsentViaJsApi']);
		}
		if (array_key_exists('dispatchAdditionalGtmEvents', $payload)) {
			$domain->setTriggeringExtGtmEvents($payload['dispatchAdditionalGtmEvents']);
		}
		if (array_key_exists('detectCloakedScripts', $payload)) {
			$domain->setDetectCloakedScripts($payload['detectCloakedScripts']);
		}
		if (array_key_exists('prioritizeLanguageHeaderOverLangAttribute', $payload)) {
			$domain->setPrioritizeLanguageHeaderOverLangAttribute($payload['prioritizeLanguageHeaderOverLangAttribute']);
		}
		if (array_key_exists('widgetDelayInMilliseconds', $payload)) {
			$domain->setWidgetDelayInMilliseconds($payload['widgetDelayInMilliseconds']);
		}
		if (array_key_exists('widgetExclusionUrls', $payload)) {
			$domain->setWidgetExclusionUrls(array_values(array_filter(
				array_map('trim', $payload['widgetExclusionUrls']),
				function (string $line) { return $line != ''; }
			)));
		}

		if (array_key_exists('allowedScriptMarkers', $payload)) {
			$domain->setAllowedScriptMarkers(array_values(array_filter(array_map('trim', (array)$payload['allowedScriptMarkers']))));
		}
		if (array_key_exists('blockedScriptMarkers', $payload)) {
			$domain->setDeniedScriptMarkers(array_values(array_filter(array_map('trim', (array)$payload['blockedScriptMarkers']))));
		}

		if (array_key_exists('evalInlineScripts', $payload)) {
			$domain->setEvalInlineScripts((bool)$payload['evalInlineScripts']);
		}

		if (array_key_exists('recordScriptsOnWebsite', $payload)) {
			$domain->setScriptsTrack((bool)$payload['recordScriptsOnWebsite']);
		}
		if (array_key_exists('blockNewScripts', $payload)) {
			$domain->setNewScriptsBlocked((bool)$payload['blockNewScripts']);
		}
		if (array_key_exists('blockFirstPartyScripts', $payload)) {
			$domain->setSameDomainScriptsBlocked((bool)$payload['blockFirstPartyScripts']);
		}
		if (array_key_exists('blockInlineScripts', $payload)) {
			$domain->setInlineScriptsBlocked((bool)$payload['blockInlineScripts']);
		}
		if (array_key_exists('permanentScanEnabled', $payload)) {
			$domain->setActivatePermaScan($payload['permanentScanEnabled']);
		}

		if (array_key_exists('tcfEnabled', $payload)) {
			$domain->setTcfVendorsEnabled((bool)$payload['tcfEnabled']);
		}
		if (array_key_exists('tcfEnableNewVendorsAfterGvlUpdate', $payload)) {
			$domain->setNewTcfVendorsActive((bool)$payload['tcfEnableNewVendorsAfterGvlUpdate']);
		}
		if (array_key_exists('tcfVendorBlacklist', $payload)) {
			/** @var int[] $referencedVendorIds */
			$referencedVendorIds = array_reduce(Embedding::all(), function (array $ids, Embedding $embedding) {
				if ($embedding->isActive() && $vendor = $embedding->getTcfVendor()) {
					$ids[$vendor->getId()] = (int)$vendor->getId();
				}
				return $ids;
			}, []);

			/** @var Vendor[] $blacklistedVendors */
			$blacklistedVendors = array_filter(array_map(function ($id) {
				return Vendor::find((string)$id);
			}, array_diff($payload['tcfVendorBlacklist'], $referencedVendorIds)));

			$domain->setTcfVendorBlacklist($blacklistedVendors);
		}
		if (array_key_exists('tcfSpecialFeatures', $payload)) {
			/** @var SpecialFeature[] $specialFeatures */
			$specialFeatures = array_filter(array_map(function ($id) {
				return SpecialFeature::find((string)$id);
			}, $payload['tcfSpecialFeatures']));

			$domain->setExplicitTcfSpecialFeatures($specialFeatures);
		}
		if (array_key_exists('tcfGoogleAdditionalConsentMode', $payload)) {
			$domain->setGoogleAdditionalConsentMode((bool)$payload['tcfGoogleAdditionalConsentMode']);
		}
		if (array_key_exists('tcfGoogleAdditionalConsentAdProviders', $payload)) {
			/** @var AdTechProvider[] $selectedGoogleAdTechProviders */
			$adTechProviders = array_filter(array_map(function ($id) {
				return AdTechProvider::find((string)$id);
			}, $payload['tcfGoogleAdditionalConsentAdProviders']));

			$domain->setGoogleAdditionalConsentWhitelist($adTechProviders);
		}
		if (array_key_exists('tcfUrlMacroAutoAppend', $payload)) {
			$domain->setTcfAdTagsAutoAppend((bool)$payload['tcfUrlMacroAutoAppend']);
		}
		if (array_key_exists('tcfUrlMacroHandling', $payload)) {
			$domain->setTcfAdTagsEnabled((bool)$payload['tcfUrlMacroHandling']);
		}

		if (array_key_exists('blockIframes', $payload)) {
			$domain->setIframesBlocked((bool)$payload['blockIframes']);
		}
		if (array_key_exists('rememberIframeConsentPerDomain', $payload)) {
			$domain->setRememberIframeConsentPerDomain((bool)$payload['rememberIframeConsentPerDomain']);
		}
		if (array_key_exists('appendConsentSwitchToIframes', $payload)) {
			$domain->setAppendConsentSwitchToIframes((bool)$payload['appendConsentSwitchToIframes']);
		}
		if (array_key_exists('iframeBlockMode', $payload)) {
			if (in_array($payload['iframeBlockMode'], Domain::ALLOWED_IFRAME_BLOCK_MODE_VALUES)) {
				$domain->setIframeBlockMode((string)$payload['iframeBlockMode']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid iframeBlockMode, allowed values: '.implode(', ', Domain::ALLOWED_IFRAME_BLOCK_MODE_VALUES),
					6
				);
			}
		}
		if (array_key_exists('iframeMarkers', $payload)) {
			$domain->setIframeMarkers(array_values(array_filter(array_map('trim', (array)$payload['iframeMarkers']))));
		}

		if (array_key_exists('cookieDeclarationTableHeadingStartLevel', $payload)) {
			if (in_array($payload['cookieDeclarationTableHeadingStartLevel'], [2, 3, 4, 5])) {
				$domain->setConsentTableHeadingStartLevel((int)$payload['cookieDeclarationTableHeadingStartLevel']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid value for cookieDeclarationTableHeadingStartLevel, allowed values: 2, 3, 4 and 5',
					5
				);
			}
		}
		if (array_key_exists('cookieDeclarationTablePrependConsentChangeLink', $payload)) {
			$domain->setConsentTablePrependConsentChangeLink((bool)$payload['cookieDeclarationTablePrependConsentChangeLink']);
		}
		if (array_key_exists('cookieDeclarationTablePrependIntroText', $payload)) {
			$domain->setConsentTablePrependIntroText((bool)$payload['cookieDeclarationTablePrependIntroText']);
		}


		//Immer einen LastScanDate erstellen - damit es da keinen kaputten Wert in der Maske gibt
		if (array_key_exists('lastScanDate', $payload)) {
			$value = new DateTime((string)$payload['lastScanDate']);
		}
		else{
			$value = new DateTime((string)"now");
		}
		$domain->setLastScanTimestamp($value);

		//CrawlCount
		if (array_key_exists('realCrawlCount', $payload)) {
			$domain->setRealCrawlCount((int)$payload['realCrawlCount']);
		}
		//activateEmbeddingsFoundByCrawler
		if (array_key_exists('activateEmbeddingsFoundByCrawler', $payload)) {
			$domain->setActivateEmbeddingsFoundByCrawler((bool)$payload['activateEmbeddingsFoundByCrawler']);
		}
		//domainExternalCrawlerFrequency
		if (array_key_exists('domainExternalCrawlerFrequency', $payload)) {
			$domain->setDomainExternalCrawlerFrequency((int)$payload['domainExternalCrawlerFrequency']);
		}
		//domainMaxCrawlBudget
		if (array_key_exists('domainMaxCrawlBudget', $payload)) {
			$domain->setDomainMaxCrawlBudget((int)$payload['domainMaxCrawlBudget']);
		}
		//doCrawl
		if (array_key_exists('doCrawl', $payload)) {
			$domain->setDoCrawl((bool)$payload['doCrawl']);
		}

		foreach (LegalText::all() as $legalText) {
			if (array_key_exists('imprintDisplayMode', $payload)) {
				if (in_array($payload['imprintDisplayMode'], LegalText::ALLOWED_DISPLAY_MODE_VALUES)) {
					$legalText->setImprintScheme($payload['imprintDisplayMode']);
				}
				else {
					throw new InvalidArgumentException(
						'Invalid imprintDisplayMode, allowed values: '.implode(', ', LegalText::ALLOWED_DISPLAY_MODE_VALUES),
						6
					);
				}
			}
			if (array_key_exists('imprintOpenInNewTab', $payload)) {
				$legalText->settargetBlank($payload['imprintOpenInNewTab']);
			}
			if (array_key_exists('imprintShowWidget', $payload)) {
				$legalText->setShowWidgetOnImprintPage($payload['imprintShowWidget']);
			}

			if (array_key_exists('privacyPolicyDisplayMode', $payload)) {
				if (in_array($payload['privacyPolicyDisplayMode'], LegalText::ALLOWED_DISPLAY_MODE_VALUES)) {
					$legalText->setPrivacyScheme($payload['privacyPolicyDisplayMode']);
				}
				else {
					throw new InvalidArgumentException(
						'Invalid privacyPolicyDisplayMode, allowed values: '.implode(', ', LegalText::ALLOWED_DISPLAY_MODE_VALUES),
						6
					);
				}
			}
			if (array_key_exists('privacyPolicyOpenInNewTab', $payload)) {
				$legalText->settargetBlankPrivacy($payload['privacyPolicyOpenInNewTab']);
			}
			if (array_key_exists('privacyPolicyShowWidget', $payload)) {
				$legalText->setShowWidgetOnPrivacyPage($payload['privacyPolicyShowWidget']);
			}

			if (array_key_exists('accessibilityDisplayMode', $payload)) {
				if (in_array($payload['accessibilityDisplayMode'], LegalText::ALLOWED_DISPLAY_MODE_VALUES)) {
					$legalText->setAccessibilityScheme($payload['accessibilityDisplayMode']);
				}
				else {
					throw new InvalidArgumentException(
						'Invalid accessibilityDisplayMode, allowed values: '.implode(', ', LegalText::ALLOWED_DISPLAY_MODE_VALUES),
						6
					);
				}
			}
			if (array_key_exists('accessibilityLinkEnabled', $payload)) {
				$legalText->setShowAccessibilityInFrontendWidget($payload['accessibilityLinkEnabled']);
			}

		}

		$domain->updateTcfStacks();
	}

	/**
	 * @param Domain $domain
	 * @return array
	 */
	private static function domainToSchemaArray(Domain $domain): array
	{
		// Set up proper repository context (required for bulk operations)
		Domain::setDomainForRequest($domain->getId());

		/** @var LegalText|null $legalText */
		$legalText = current(LegalText::all()) ?: null;

		return [
			'id' => $domain->getId(),
			'name' => $domain->getName(),
			'path' => $domain->getUrlPath(),
			'country' => $domain->getCountryCode(),
			'timeZone' => $domain->getTimeZone() ? $domain->getTimeZone()->getName() : null,
			'whitelabel' => $domain->hasWhitelabel(),

			'consentStorage' => $domain->getConsentStorage(),
			'consentStorageNamespaceUrlParameter' => $domain->getConsentStorageNamespaceUrlParameter(),
			'consentCookieLifetime' => $domain->getConsentCookieLifetime(),
			'consentCookieSecure' => $domain->hasSecureConsentCookie(),
			'consentLifetime' => $domain->getConsentLifetime(),
			'consentLifetimeForPartialConsent' => $domain->getConsentLifetimeForPartialConsent(),
			'consentResetBeforeDate' => $domain->getconsentCookieReset(),
			'consentResetOnLegalChange' => $domain->hasConsentResetOnRelevantChange(),
			'consentSharingDomains' => $domain->getShareDomainNames(),
			'consentSharingCrossDomain' => $domain->hasCrossDomainSharing(),

			'frontendWidgetEnabled' => $domain->isFrontendWidgetEnabled(),
			'frontendWidgetCache' => $domain->isWidgetCachingEnabled(),
			'frontendWidgetMinifyCode' => $domain->hasCodeMinification(),
			'partOfAdAndTrackingFreeNetwork' => $domain->isPartOfAdAndTrackingFreeNetwork(),

			'managementStructure' => $domain->getManagementStructure(),

			'deleteUnknownCookies' => $domain->getDeleteUnknownCookies(),
			/* Feature disabled indefinitely
			'deleteUnknownCookiesForceReload' => $domain->getDeleteUnknownCookiesForceReload(),
			*/
			'requireConsentForMandatoryEmbeddings' => $domain->isLoadTechnicallyNecessaryEmbeddingsWithoutConsent() == false,
			'declineButtonEnforced' => $domain->isDeclineButtonEnforced(),
			'forcedCssImportant' => $domain->hasForcedCssImportant(),
			'lcpOptimization' => $domain->isLcpOptimizationEnabled(),
			'onlyActiveOnMatchingDomain' => $domain->isOnlyActiveOnMatchingDomain(),
			'embeddingConsentViaJsApi' => $domain->isAllowEmbeddingConsentViaJsApi(),
			'dispatchAdditionalGtmEvents' => $domain->isTriggeringExtGtmEvents(),
			'detectCloakedScripts' => $domain->isDetectCloakedScripts(),
			'prioritizeLanguageHeaderOverLangAttribute' => $domain->isPrioritizeLanguageHeaderOverLangAttribute(),
			'widgetDelayInMilliseconds' => $domain->getWidgetDelayInMilliseconds(),
			'widgetExclusionUrls' => $domain->getWidgetExclusionUrls(),

			'allowedScriptMarkers' => $domain->getAllowedScriptMarkers(),
			'blockedScriptMarkers' => $domain->getDeniedScriptMarkers(),

			'evalInlineScripts' => $domain->isEvalInlineScripts(),

			'recordScriptsOnWebsite' => $domain->isScriptsTrack(),
			'blockNewScripts' => $domain->isNewScriptsBlocked(),
			'blockFirstPartyScripts' => $domain->isSameDomainScriptsBlocked(),
			'blockInlineScripts' => $domain->isInlineScriptsBlocked(),
			'permanentScanEnabled' => $domain->isActivatePermaScan(),

			'tcfEnabled' => $domain->isTcfVendorsEnabled(),
			'tcfEnableNewVendorsAfterGvlUpdate' => $domain->isNewTcfVendorsActive(),
			'tcfVendorBlacklist' => array_map(function (Vendor $vendor) {
				return (int)$vendor->getId();
			}, $domain->getTcfVendorBlacklist()),
			'tcfSpecialFeatures' => array_map(function (SpecialFeature $feature) {
				return (int)$feature->getId();
			}, $domain->getExplicitTcfSpecialFeatures()),
			'tcfGoogleAdditionalConsentMode' => $domain->hasGoogleAdditionalConsentMode(),
			'tcfGoogleAdditionalConsentAdProviders' => array_map(function (AdTechProvider $provider) {
				return (int)$provider->getId();
			}, $domain->getGoogleAdditionalConsentWhitelist()),
			'tcfUrlMacroAutoAppend' => $domain->hasTcfAdTagsAutoAppend(),
			'tcfUrlMacroHandling' => $domain->isTcfAdTagsEnabled(),

			'blockIframes' => $domain->isIframesBlocked(),
			'rememberIframeConsentPerDomain' => $domain->isRememberIframeConsentPerDomain(),
			'appendConsentSwitchToIframes' => $domain->isAppendConsentSwitchToIframes(),
			'iframeBlockMode' => $domain->getIframeBlockMode(),
            'iframeMarkers' => $domain->getIframeMarkers(),
			'doCrawl' => $domain->getDoCrawl(),
			'activateEmbeddingsFoundByCrawler' => $domain->isActivateEmbeddingsFoundByCrawler(),

			'cookieDeclarationTableHeadingStartLevel' => $domain->getConsentTableHeadingStartLevel(),
			'cookieDeclarationTablePrependConsentChangeLink' => $domain->getConsentTablePrependConsentChangeLink(),
			'cookieDeclarationTablePrependIntroText' => $domain->getConsentTablePrependIntroText(),

			'imprintDisplayMode' => $legalText ? $legalText->getImprintScheme() : null,
			'imprintOpenInNewTab' => $legalText ? $legalText->istargetBlank() : null,
			'imprintShowWidget' => $legalText ? $legalText->isShowWidgetOnImprintPage() : null,

			'privacyPolicyDisplayMode' => $legalText ? $legalText->getPrivacyScheme() : null,
			'privacyPolicyOpenInNewTab' => $legalText ? $legalText->istargetBlankPrivacy() : null,
			'privacyPolicyShowWidget' => $legalText ? $legalText->isShowWidgetOnPrivacyPage() : null,

			'accessibilityDisplayMode' => $legalText ? $legalText->getAccessibilityScheme() : null,
            'accessibilityLinkEnabled' => $legalText ? $legalText->isShowAccessibilityInFrontendWidget() : null,
		];
	}
}
