<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\Exception\BadRequestException;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\LocaleService;
use App\Component\LoggerInterface;
use App\Component\Reflection\PropertyProcessor;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Config;
use App\Controller\Embeddings;
use App\Controller\HostingApiController;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingAssetBase;
use App\Model\EmbeddingBase;
use App\Model\LocaleBase;
use App\Model\Purpose;
use App\Model\SystemEmbedding;
use App\Model\SystemEmbeddingAsset;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/hosting/api/client/{userId}/domain/{domainId}/embedding", name="app_hosting_api_embedding_")
 */
class EmbeddingApi extends HostingApiController
{
	private const EMBEDDING_ASSET_SCHEMA = [
		'name' => 'string',
		'dynamic' => 'bool',
		'storageType' => 'string',
		'lifetime' => 'string',
		'description' => 'string',

		'translations' => [
			'@collection' => true,
			'@schema' => [
				'lifetime' => 'string',
				'description' => 'string',
			],
		],
	];

	private const EMBEDDING_SCHEMA = [
		'name' => 'string',
		'active' => 'bool',
		'purpose' => 'string',
		'vendor' => 'string',

		'systemEmbeddingId' => 'string',
		'autoUpdate' => 'bool',

		'scriptHtmlCode' => 'string',
		'scriptLoaderGroup' => 'string',
		'scriptMarkers' => 'string[]',

		'iframeMarkers' => 'string[]',
		'hideIframesUntilConsent' => 'bool',
		'dispatchLoadEventForIframes' => 'bool',

		'excludeUrlsMode' => 'string',
		'excludeFromUrls' => 'string[]',

		'description' => 'string',
		'privacyPolicyUrl' => 'string',
		'whatDataIsCollected' => 'string',
		'purposeOfDataCollection' => 'string',
		'legalBasis' => 'string',
		'placeOfProcessing' => 'string',

		'translations' => [
			'@collection' => true,
			'@schema' => [
				'name' => 'string',
				'description' => 'string',
				'privacyPolicyUrl' => 'string',
				'whatDataIsCollected' => 'string',
				'purposeOfDataCollection' => 'string',
				'legalBasis' => 'string',
				'placeOfProcessing' => 'string',
				'blockedContentTitle' => 'string',
				'blockedContentText' => 'string',
				'blockedContentButtonText' => 'string',
			],
		],

		'tcfVendor' => '?int',
		'tcfPurposes' => 'int[]',
		'tcfSpecialFeatures' => 'int[]',

		'googleConsentMode' => 'string[]',

		'assets' => [
			'@collection' => true,
			'@schema' => self::EMBEDDING_ASSET_SCHEMA,
		],
	];

	private const BULK_PATCH_SCHEMA = [
		'removeAll' => 'bool',
		'remove' => 'string[]',
		'update' => [
			'@collection' => true,
			'@schema' => self::EMBEDDING_SCHEMA,
		],
		'add' => [
			'@collection' => true,
			'@schema' => self::EMBEDDING_SCHEMA,
		],
	];

	private const FOR_DB_IDS = [
		'dbIds' => 'array',
	];

	/**
	 * @Route("", name="create", methods={"POST"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function createEmbedding(string $userId, string $domainId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request, 6);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Create the new embedding
		$embedding = Embedding::create();

		try {
			// Try to process the payload and update relevant properties
			self::migrateEmbeddingPayload($embedding, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Delete new objects if an error has occurred (rollback)
			EmbeddingAsset::reload();
			Embedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Delete new objects if an error has occurred (rollback)
			EmbeddingAsset::reload();
			Embedding::reload();

			// Dev time, let it rip
			throw $exception;
		}

		// Predefine properties that aren't accessible via the API
		if ($embedding->isImported() == false) {
			$embedding
				->setSource(EmbeddingBase::SOURCE_LOCAL)
				->setAutoUpdated(false)
			;
		}

		EmbeddingAsset::flush();
		$embedding->save();

		$logger->logInfo('Embedding-API', "Embedding created: {$embedding->getName()} [$userId::$domainId]");

		return $this->json(self::embeddingToSchemaArray($embedding), 201);
	}

	/**
	 * Route um aus eine übertragenen DB Ids aus unserer CookieDB ein neues Embedding zu erstellen für die jeweilige Domain
	 *
	 * @internal Diese Route ist nur für den internen Gebrauch und sollte nicht öffentlich zugänglich sein, nur Cloud oder EP Versionen
	 * @Route("/bulk_from_db_ids", name="create_embedding_from_dbId", methods={"POST"})")
	 * @param string $userId
	 * @param string $domainId
	 * @param \Symfony\Component\HttpFoundation\Request $request
	 * @param \App\Component\LoggerInterface $logger
	 *
	 * @return \Symfony\Component\HttpFoundation\JsonResponse
	 * @throws \Exception
	 */
	public function bulkCreateEmbeddingsFromDbIDs(string $userId, string $domainId, Request $request, LoggerInterface $logger): JsonResponse
	{
		//Genereller Check ob Daten ok sind
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
			self::checkPayloadDataTypes($payload, self::FOR_DB_IDS);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}
		catch (InvalidArgumentException $exception) {
			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		//Check, ob leer - falls leer - Fehlermeldung ausgeben
		if(empty($payload) || empty($payload['dbIds']))
		{
			return $this->json([
				'code' => "false",
				'error' => "Payload is empty",
			], 400);
		}

		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleName = $defaultLocale->getName();
		$vorhandeneEmbeddings = Embedding::all();

		//Aktiv j/n bestimmen, kommt aus den Domain Settings
		$domain = Domain::activeDomain();
		$sollAktivieren = true;
		if($domain!==null)
		{
			$sollAktivieren = $domain->isActivateEmbeddingsFoundByCrawler();
		}

		//Arrays vorbelegen
		$returnEmbeddingData = array();
		$remoteIds = array();

		//Vorhandene durchgehen und die Remote DB Ids rausholen zum Abgleich
		foreach($vorhandeneEmbeddings as $vorhandenesEmbedding) {
			$rId = $vorhandenesEmbedding->getSourceId();
			$remoteIds[$rId]=$rId;
		}

		//Checken welche überhaupt valide sind
		$embeddingDataSetFromDB = EmbeddingDb::fetchEmbeddings(array_keys($payload['dbIds']));

		//Die IDs durchgehen die geliefert wurden
		foreach ($payload['dbIds'] as $embeddingDbId => $foundUrl) {

			//Checken, ob das Embedding schon existiert
			if(!empty($remoteIds[$embeddingDbId]))
			{
				$returnEmbeddingData[$embeddingDbId] = "exists";
				continue;
			}

			//Checken, ob das Embedding bekannt ist aus der Cookie DB Remote
			if(empty($embeddingDataSetFromDB[$embeddingDbId]))
			{
				$returnEmbeddingData[$embeddingDbId] = "not Found";
				continue;
			}

			//Embedding erstellen aus der DB ID
			$remoteEmbedding = EmbeddingDb::fetchEmbedding($embeddingDbId);

			//Embedding Klasse ini
			$embeddingsClass = new Embeddings();
			$remoteEmbeddingData = $embeddingsClass->createRemoteEmbeddingFromDbEntry($remoteEmbedding, $defaultLocale);
			if (!empty($remoteEmbeddingData)) {
				$theRemoteEmbedding = $embeddingsClass->applyChanges(Embedding::create(), $remoteEmbeddingData, $defaultLocaleName);
				if($sollAktivieren===false)
				{
					//Nur auf inaktiv setzen, wenn es so gewünscht ist, Standard ist eh aktiv
					$theRemoteEmbedding->setActive(false);
				}
				//Wo wurde das Embedding zuerst gefunden
				$theRemoteEmbedding->setFoundUrl(trim($foundUrl));
				//Wann das Embedding gefunden wurde
				$theRemoteEmbedding->setFoundTime(time());
				//Neues Embedding speichern
				$theRemoteEmbedding->save();

				//Oha - es soll ein Iframe geblockt werden, UND das Embedding soll aktiviert werden
				if($sollAktivieren===true && count($theRemoteEmbedding->getIframeBlockingStrings()) > 0)
				{
					//Dann besser auch das Iframe Blocking aktivieren, Modus steht auch Blocklist, sollte daher kein Problem machen
					$domain->setIframesBlocked(true);
					$domain->save();
					$logger->logInfo('Embedding-API From DB', "IFRAME Blocking aktiviert: {$theRemoteEmbedding->getName()} [$userId::$domainId]");
				}

				//noch kurz loggen
				$logger->logInfo('Embedding-API From DB', "Embedding created: {$theRemoteEmbedding->getName()} [$userId::$domainId]");

				//Name für die Rückgabe holen
				$returnEmbeddingData[$embeddingDbId] = $theRemoteEmbedding->getName()." created!";
			}
		}
		//Sicherheitshalber auch die Assest flashen
		EmbeddingAsset::flush();

		return $this->json(array("result"=>$returnEmbeddingData));
	}

	/**
	 * @Route("", name="read_all", methods={"GET", "HEAD"})
	 * @param string $userId
	 * @param string $domainId
	 * @return JsonResponse
	 */
	public function readAllEmbeddings(string $userId, string $domainId): JsonResponse
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		return $this->json(array_values(array_map(function (Embedding $embedding) {
			return self::embeddingToSchemaArray($embedding);
		}, Embedding::all())), 200);
	}

	/**
	 * @Route("/{embeddingId}", name="read", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @return JsonResponse
	 */
	public function readEmbedding(string $userId, string $domainId, string $embeddingId): JsonResponse
	{
		try {
			self::assertEmbeddingExists($userId, $domainId, $embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Embedding $embedding */
		$embedding = Embedding::find($embeddingId);
		assert($embedding instanceof Embedding);

		return $this->json(self::embeddingToSchemaArray($embedding), 200);
	}

	/**
	 * @Route("/{embeddingId}", name="update", methods={"PUT"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function updateEmbedding(string $userId, string $domainId, string $embeddingId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertEmbeddingExists($userId, $domainId, $embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Embedding $embedding */
		$embedding = Embedding::find($embeddingId);
		assert($embedding instanceof Embedding);

		try {
			// Try to process the payload and update relevant properties
			self::migrateEmbeddingPayload($embedding, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			EmbeddingAsset::reload();
			Embedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		EmbeddingAsset::flush();
		$embedding->save();

		$logger->logInfo('Embedding-API', "Embedding updated: {$embedding->getName()} [$userId::$domainId]");

		return $this->json(self::embeddingToSchemaArray($embedding), 200);
	}

	/**
	 * @Route("/{embeddingId}", name="delete", methods={"DELETE"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteEmbedding(string $userId, string $domainId, string $embeddingId, LoggerInterface $logger): Response
	{
		try {
			self::assertEmbeddingExists($userId, $domainId, $embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Embedding $embedding */
		$embedding = Embedding::find($embeddingId);
		assert($embedding instanceof Embedding);

		try {
			foreach ($embedding->getAssets() as $asset) {
				EmbeddingAsset::delete($asset->getId(), false);
			}
		}
		finally {
			EmbeddingAsset::flush();
		}

		$logger->logInfo('Embedding-API', "Embedding deleted: {$embedding->getName()} [$userId::$domainId]");
		Embedding::delete($embedding->getId());

		return new Response('', 204);
	}

	/**
	 * @Route("", name="patch_bulk", methods={"PATCH"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function patchEmbeddings(string $userId, string $domainId, Request $request, LoggerInterface $logger): Response
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
			self::checkPayloadDataTypes($payload, self::BULK_PATCH_SCHEMA);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}
		catch (InvalidArgumentException $exception) {
			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		Embedding::beginWrite();
		EmbeddingAsset::beginWrite();

		$removeAll = $payload['removeAll'] ?? false;
		$toRemove = $payload['remove'] ?? [];
		$toUpdate = $payload['update'] ?? [];
		$toAdd = $payload['add'] ?? [];
		$removed = $updated = $added = [];
		if (!is_array($payload) or array_diff(array_keys($payload), ['remove', 'update', 'add', 'removeAll'])) {
			return $this->json(['error' => 'invalid payload structure'], 400);
		}
		if ($toRemove and $removeAll) {
			return $this->json(['error' => '"remove" and "removeAll" are mutually exclusive'], 400);
		}

		try {
			if ($removeAll) {
				$removed = array_values(array_map(function (Embedding $item) { return $item->getId(); }, Embedding::all()));
				Embedding::clear(false);
				EmbeddingAsset::clear(false);
				$logger->logInfo('Embedding-API', "All embeddings deleted in bulk operation [$userId::$domainId]");
			}

			foreach ($toRemove as $embeddingId) {
				/** @var Embedding $embedding */
				$embedding = Embedding::find($embeddingId);
				if (!$embedding) { continue; }

				foreach ($embedding->getAssets() as $asset) {
					try {
						EmbeddingAsset::delete($asset->getId(), false);
					}
					catch (\Throwable $ignore) {
						// ignore
					}
				}
				Embedding::delete($embedding->getId(), false);
				$removed[] = $embedding->getId();
				$logger->logInfo('Embedding-API', "Embedding deleted in bulk operation: {$embedding->getName()} [$userId::$domainId]");
			}

			foreach ($toUpdate as $embeddingId => $embeddingData) {
				self::assertEmbeddingExists($userId, $domainId, $embeddingId);

				/** @var Embedding $embedding */
				$embedding = Embedding::find($embeddingId);
				assert($embedding instanceof Embedding);
				$preUpdate = self::embeddingToSchemaArray($embedding);

				// Try to process the payload and update relevant properties
				self::migrateEmbeddingPayload($embedding, $embeddingData);
				$postUpdate = self::embeddingToSchemaArray($embedding);

				$logger->logInfo('Embedding-API', "Embedding updated in bulk operation: {$embedding->getName()} [$userId::$domainId]");
				if ($preUpdate !== $postUpdate) {
					$updated[$embedding->getId()] = self::embeddingToSchemaArray($embedding);
				}
			}

			foreach ($toAdd as $embeddingData) {
				// Create the new embedding
				$embedding = Embedding::create();

				// Try to process the payload and update relevant properties
				self::migrateEmbeddingPayload($embedding, $embeddingData);

				// Predefine properties that aren't accessible via the API
				if ($embedding->isImported() == false) {
					$embedding
						->setSource(EmbeddingBase::SOURCE_LOCAL)
						->setAutoUpdated(false)
					;
				}

				$added[] = self::embeddingToSchemaArray($embedding);
				$logger->logInfo('Embedding-API', "Embedding created in bulk operation: {$embedding->getName()} [$userId::$domainId]");
			}
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}
		catch (InvalidArgumentException $exception) {
			// Restore previous state if an error has occurred (rollback)
			EmbeddingAsset::reload();
			Embedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Restore previous state if an error has occurred (rollback)
			EmbeddingAsset::reload();
			Embedding::reload();

			// Dev time, let it rip
			throw $exception;
		}

		EmbeddingAsset::flush();
		Embedding::flush();

		$logger->logInfo('Embedding-API', "Embedding bulk operation finished [$userId::$domainId]");

		return new JsonResponse([
			'removed' => $removed,
			'updated' => $updated,
			'added' => $added,
		], 200);
	}

	/**
	 * @Route("/{embeddingId}/asset", name="asset_create", methods={"POST"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function createEmbeddingAsset(string $userId, string $domainId, string $embeddingId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertEmbeddingExists($userId, $domainId, $embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Lock embedding repository first, then create the new embedding asset
		Embedding::beginWrite();
		/** @var Embedding $embedding */
		$embedding = Embedding::find($embeddingId);
		assert($embedding instanceof Embedding);
		$asset = EmbeddingAsset::create();

		try {
			// Try to process the payload and update relevant properties
			self::migrateEmbeddingAssetPayload($asset, $payload);
			$embedding->addAsset($asset);
			$embedding->updateLastModified();
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			EmbeddingAsset::reload();
			Embedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Revert changes if an error has occurred (rollback)
			EmbeddingAsset::reload();
			Embedding::reload();

			// Dev time, let it rip
			throw $exception;
		}

		$asset->save();
		$embedding->save();

		$logger->logInfo('Embedding-API', "Asset \"{$asset->getName()}\" created for \"{$embedding->getName()}\". [$userId::$domainId::$embeddingId]");

		return $this->json(self::embeddingAssetToSchemaArray($asset), 201);
	}

	/**
	 * @Route("/{embeddingId}/asset", name="asset_read_all", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @return JsonResponse
	 */
	public function readAllEmbeddingAssets(string $userId, string $domainId, string $embeddingId): JsonResponse
	{
		try {
			self::assertEmbeddingExists($userId, $domainId, $embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Embedding $embedding */
		$embedding = Embedding::find($embeddingId);
		assert($embedding instanceof Embedding);

		return $this->json(array_values(array_map(function (EmbeddingAsset $asset) {
			return self::embeddingAssetToSchemaArray($asset);
		}, $embedding->getAssets())), 200);
	}

	/**
	 * @Route("/{embeddingId}/asset/{assetId}", name="asset_read", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @param string $assetId
	 * @return JsonResponse
	 */
	public function readEmbeddingAsset(string $userId, string $domainId, string $embeddingId, string $assetId): JsonResponse
	{
		try {
			self::assertEmbeddingAssetExists($userId, $domainId, $embeddingId, $assetId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Directly query the asset repository, as it's read only.
		/** @var EmbeddingAsset $asset */
		$asset = EmbeddingAsset::find($assetId);
		assert($asset instanceof EmbeddingAsset);

		return $this->json(self::embeddingAssetToSchemaArray($asset), 200);
	}

	/**
	 * @Route("/{embeddingId}/asset/{assetId}", name="asset_update", methods={"PUT"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @param string $assetId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function updateEmbeddingAsset(string $userId, string $domainId, string $embeddingId, string $assetId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertEmbeddingAssetExists($userId, $domainId, $embeddingId, $assetId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Embedding $embedding */
		$embedding = Embedding::find($embeddingId);
		assert($embedding instanceof Embedding);
		/** @var EmbeddingAsset $asset */
		$asset = $embedding->getAsset($assetId);
		assert($asset instanceof EmbeddingAsset);

		try {
			// Lock embedding repository first
			Embedding::beginWrite();
			// Try to process the payload and update relevant properties
			self::migrateEmbeddingAssetPayload($asset, $payload);
			$embedding->updateLastModified();
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			EmbeddingAsset::reload();
			Embedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		$asset->save();
		$embedding->save();

		$logger->logInfo('Embedding-API', "Asset \"{$asset->getName()}\" updated for \"{$embedding->getName()}\". [$userId::$domainId::$embeddingId]");

		return $this->json(self::embeddingAssetToSchemaArray($asset), 200);
	}

	/**
	 * @Route("/{embeddingId}/asset/{assetId}", name="asset_delete", methods={"DELETE"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @param string $assetId
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteEmbeddingAsset(string $userId, string $domainId, string $embeddingId, string $assetId, LoggerInterface $logger): Response
	{
		try {
			self::assertEmbeddingAssetExists($userId, $domainId, $embeddingId, $assetId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Embedding $embedding */
		$embedding = Embedding::find($embeddingId);
		assert($embedding instanceof Embedding);
		/** @var EmbeddingAsset $asset */
		$asset = $embedding->getAsset($assetId);
		assert($asset instanceof EmbeddingAsset);

		$logger->logInfo('Embedding-API', "Asset \"{$asset->getName()}\" deleted from \"{$embedding->getName()}\". [$userId::$domainId::$embeddingId]");

		$embedding->removeAsset($assetId);
		$embedding->updateLastModified();
		EmbeddingAsset::flush();
		$embedding->save();

		return new Response('', 204);
	}

	/**
	 * @param Embedding $embedding
	 * @param SystemEmbedding $systemEmbedding
	 * @return void
	 */
	private static function baseOnSystemEmbedding(Embedding $embedding, SystemEmbedding $systemEmbedding): void
	{
		// Copy trivial properties
		PropertyProcessor::copyValues($systemEmbedding, $embedding, EmbeddingBase::class);

		$config = Config::getInstance();
		$languages = $config->getSystemEmbeddingLanguages();

		// Manually copy properties that, for example, use abstract methods or mix nullable and non-nullable getter/setter
		foreach ($embedding->getAssets() as $asset) {
			EmbeddingAsset::delete($asset->getId(), false);
		}

		$assets = array_map(function (SystemEmbeddingAsset $systemEmbeddingAsset) use ($languages) {
			$embeddingAsset = EmbeddingAsset::create();
			PropertyProcessor::copyValues($systemEmbeddingAsset, $embeddingAsset, EmbeddingAssetBase::class);

			// Copy language-specific properties
			foreach ($languages as $language) {
				if ($locale = LocaleService::find($language->getId())) {
					$embeddingAsset
						->setTranslatedLifetime($locale, $systemEmbeddingAsset->getTranslatedLifetime($language, false))
						->setTranslatedDescription($locale, $systemEmbeddingAsset->getTranslatedDescription($language, false))
					;
				}
			}

			return $embeddingAsset;
		}, $systemEmbedding->getAssets());

		$purpose = $systemEmbedding->getPurpose() ?? Purpose::find('15c61c3');
		assert($purpose instanceof Purpose);

		$embedding
			->setAssets($assets)
			->setPurpose($purpose)
			->setSource(EmbeddingBase::SOURCE_SYSTEM_TEMPLATE)
			->setSourceId($systemEmbedding->getId())
			->setSourceTimestamp($systemEmbedding->getLastModified())
		;

		// Copy language-specific properties
		foreach ($languages as $language) {
			if ($locale = LocaleService::find($language->getId())) {
				$embedding
					->setTranslatedName($locale, $systemEmbedding->getTranslatedName($language, false))
					->setTranslatedDescription($locale, $systemEmbedding->getTranslatedDescription($language, false))
					->setTranslatedPrivacyPolicyUrl($locale, $systemEmbedding->getTranslatedPrivacyPolicyUrl($language, false))
					->setTranslatedCollectedDataInfo($locale, $systemEmbedding->getTranslatedCollectedDataInfo($language, false))
					->setTranslatedPurposeOfDataCollection($locale, $systemEmbedding->getTranslatedPurposeOfDataCollection($language, false))
					->setTranslatedLegalBasis($locale, $systemEmbedding->getTranslatedLegalBasis($language, false))
					->setTranslatedPlaceOfProcessing($locale, $systemEmbedding->getTranslatedPlaceOfProcessing($language, false))
					->setTranslatedBlockedContentTitle($locale, $systemEmbedding->getTranslatedBlockedContentTitle($language, false))
					->setTranslatedBlockedContentText($locale, $systemEmbedding->getTranslatedBlockedContentText($language, false))
					->setTranslatedBlockedContentButtonText($locale, $systemEmbedding->getTranslatedBlockedContentButtonText($language, false))
				;
			}
		}
	}

	/**
	 * @param Embedding $embedding
	 * @param array $payload
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migrateEmbeddingPayload(Embedding $embedding, array $payload): void
	{
		self::checkPayloadDataTypes($payload, self::EMBEDDING_SCHEMA);

		if (array_key_exists('systemEmbeddingId', $payload)) {
			// When creating
			if ($embedding->getName() == '') {
				if ($systemEmbedding = SystemEmbedding::find($payload['systemEmbeddingId'])) {
					self::baseOnSystemEmbedding($embedding, $systemEmbedding);
				}
				else {
					throw new InvalidArgumentException('Unknown systemEmbeddingId, object not found.', 6);
				}
			}
			else {
				throw new InvalidArgumentException('Currently, systemEmbeddingId is only allowed when creating a new object.', 7);
			}
		}

		if (array_key_exists('name', $payload) || empty($embedding->getName())) {
			if ($payload['name'] ?? '') {
				$embedding->setName($payload['name']);
			}
			else {
				throw new InvalidArgumentException('Empty or missing name.', 1);
			}
		}
		if (array_key_exists('active', $payload)) {
			$embedding->setActive($payload['active']);
		}
		if (array_key_exists('purpose', $payload) || $embedding->getPurpose() == null) {
			if ($purpose = Purpose::find($payload['purpose'] ?? '')) {
				$embedding->setPurpose($purpose);
			}
			else {
				throw new InvalidArgumentException('Empty, invalid or missing purpose, allowed values: '.implode(', ', array_keys(Purpose::all())), 2);
			}
		}
		if (array_key_exists('vendor', $payload)) {
			$embedding->setEmbeddingVendor($payload['vendor']);

			$tcfVendor = array_reduce(TcfVendor::all(), function (?TcfVendor $carry, TcfVendor $vendor) use ($payload) {
				return $carry ?? ($vendor->getName() == $payload['vendor'] ? $vendor : $carry);
			}, null);
			$embedding->setTcfVendor($tcfVendor);
		}

		if (array_key_exists('autoUpdate', $payload) && $embedding->isImported()) {
			$embedding->setAutoUpdated($payload['autoUpdate']);
		}

		if (array_key_exists('scriptHtmlCode', $payload)) {
			$embedding->setCodeSnippet($payload['scriptHtmlCode']);
		}
		if (array_key_exists('scriptLoaderGroup', $payload)) {
			$embedding->setScriptLoaderGroup($payload['scriptLoaderGroup']);
		}
		if (array_key_exists('scriptMarkers', $payload)) {
			$embedding->setCodeBlock(array_values(array_filter(array_map('trim', (array)$payload['scriptMarkers']))));
		}

		if (array_key_exists('iframeMarkers', $payload)) {
			$embedding->setIframeBlockingStrings(array_values(array_filter(array_map('trim', (array)$payload['iframeMarkers']))));
		}
		if (array_key_exists('hideIframesUntilConsent', $payload)) {
			$embedding->setIframesHiddenUntilConsent($payload['hideIframesUntilConsent']);
		}
		if (array_key_exists('dispatchLoadEventForIframes', $payload)) {
			$embedding->setDispatchLoadEventForIframes($payload['dispatchLoadEventForIframes']);
		}

		if (array_key_exists('excludeUrlsMode', $payload)) {
			if (in_array($payload['excludeUrlsMode'], EmbeddingBase::ALLOWED_EXCLUDE_URLS_MODE_VALUES)) {
				$embedding->setExcludeUrlsMode((string)$payload['excludeUrlsMode']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid excludeUrlsMode, allowed values: '.implode(', ', EmbeddingBase::ALLOWED_EXCLUDE_URLS_MODE_VALUES),
					3
				);
			}
		}
		if (array_key_exists('excludeFromUrls', $payload)) {
			$embedding->setExcludeUrls(array_values(array_filter(array_map('trim', (array)$payload['excludeFromUrls']))));
		}

		if (array_key_exists('description', $payload)) {
			$embedding->setDescription($payload['description']);
		}
		if (array_key_exists('privacyPolicyUrl', $payload)) {
			$embedding->setPrivacyPolicyUrl($payload['privacyPolicyUrl']);
		}
		if (array_key_exists('whatDataIsCollected', $payload)) {
			$embedding->setCollectedDataInfo($payload['whatDataIsCollected']);
		}
		if (array_key_exists('purposeOfDataCollection', $payload)) {
			$embedding->setPurposeOfDataCollection($payload['purposeOfDataCollection']);
		}
		if (array_key_exists('legalBasis', $payload)) {
			$embedding->setLegalBasis($payload['legalBasis']);
		}
		if (array_key_exists('placeOfProcessing', $payload)) {
			$embedding->setPlaceOfProcessing($payload['placeOfProcessing']);
		}

		if (array_key_exists('translations', $payload)) {
			foreach ($payload['translations'] as $localeName => $translation) {
				if ($locale = LocaleService::find((string)$localeName)) {
					if (array_key_exists('name', $translation)) {
						$embedding->setTranslatedName($locale, $translation['name']);
					}
					if (array_key_exists('description', $translation)) {
						$embedding->setTranslatedDescription($locale, $translation['description']);
					}
					if (array_key_exists('privacyPolicyUrl', $translation)) {
						$embedding->setTranslatedPrivacyPolicyUrl($locale, $translation['privacyPolicyUrl']);
					}
					if (array_key_exists('whatDataIsCollected', $translation)) {
						$embedding->setTranslatedCollectedDataInfo($locale, $translation['whatDataIsCollected']);
					}
					if (array_key_exists('purposeOfDataCollection', $translation)) {
						$embedding->setTranslatedPurposeOfDataCollection($locale, $translation['purposeOfDataCollection']);
					}
					if (array_key_exists('legalBasis', $translation)) {
						$embedding->setTranslatedLegalBasis($locale, $translation['legalBasis']);
					}
					if (array_key_exists('placeOfProcessing', $translation)) {
						$embedding->setTranslatedPlaceOfProcessing($locale, $translation['placeOfProcessing']);
					}
					if (array_key_exists('blockedContentTitle', $translation)) {
						$embedding->setTranslatedBlockedContentTitle($locale, $translation['blockedContentTitle']);
					}
					if (array_key_exists('blockedContentText', $translation)) {
						$embedding->setTranslatedBlockedContentText($locale, $translation['blockedContentText']);
					}
					if (array_key_exists('blockedContentButtonText', $translation)) {
						$embedding->setTranslatedBlockedContentButtonText($locale, $translation['blockedContentButtonText']);
					}
				}
				else {
					throw new InvalidArgumentException("Invalid locale name \"$localeName\" for translation.", 3);
				}
			}
		}

		if (array_key_exists('tcfVendor', $payload)) {
			if ($payload['tcfVendor']) {
				if ($tcfVendor = TcfVendor::find((string)$payload['tcfVendor'])) {
					$embedding->setTcfVendor($tcfVendor);
					$embedding->setEmbeddingVendor($tcfVendor->getName());
				}
				else {
					throw new InvalidArgumentException('Invalid TCF vendor id.', 4);
				}
			}
			else {
				$embedding->setTcfVendor(null);
				if (array_key_exists('vendor', $payload) == false) {
					$embedding->setEmbeddingVendor('');
				}
			}
		}
		if (array_key_exists('tcfPurposes', $payload)) {
			/** @var TcfPurpose[] $purposes  */
			$purposes = array_filter(array_map(function ($id) {
				return TcfPurpose::find((string)$id);
			}, $payload['tcfPurposes']));

			$embedding->setTcfPurposes($purposes);
		}
		if (array_key_exists('tcfSpecialFeatures', $payload)) {
			/** @var SpecialFeature[] $specialFeatures */
			$specialFeatures = array_filter(array_map(function ($id) {
				return SpecialFeature::find((string)$id);
			}, $payload['tcfSpecialFeatures']));

			$embedding->setTcfSpecialFeatures($specialFeatures);
		}

		if (array_key_exists('googleConsentMode', $payload)) {
			$valid = array_reduce($payload['googleConsentMode'], function (bool $valid, string $mode) {
				return $valid && in_array($mode, EmbeddingBase::ALLOWED_GCM_VALUES);
			}, true);

			if ($valid) {
				$embedding->setGoogleConsentMode($payload['googleConsentMode']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid Google consent mode, allowed values: '.implode(', ', EmbeddingBase::ALLOWED_GCM_VALUES),
					5
				);
			}
		}

		if (array_key_exists('assets', $payload)) {
			foreach ($embedding->getAssets() as $asset) {
				EmbeddingAsset::delete($asset->getId(), false);
			}

			$assets = array_map(function (array $payload) {
				$asset = EmbeddingAsset::create();
				self::migrateEmbeddingAssetPayload($asset, $payload);
				return $asset;
			}, $payload['assets']);

			$embedding->setAssets($assets);
		}

		$embedding->updateLastModified();
	}

	/**
	 * @param EmbeddingAsset $asset
	 * @param array $payload
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migrateEmbeddingAssetPayload(EmbeddingAsset $asset, array $payload): void
	{
		self::checkPayloadDataTypes($payload, self::EMBEDDING_ASSET_SCHEMA);

		if (array_key_exists('name', $payload) || empty($asset->getName())) {
			if ($payload['name'] ?? '') {
				$asset->setName($payload['name']);
			}
			else {
				throw new InvalidArgumentException('Empty or missing name.', 1);
			}
		}
		if (array_key_exists('dynamic', $payload)) {
			$asset->setDynamic($payload['dynamic']);
		}
		if (array_key_exists('storageType', $payload)) {
			if (in_array($payload['storageType'], EmbeddingAssetBase::ALLOWED_STORAGE_TYPE_VALUES)) {
				$asset->setStorageType($payload['storageType']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid storage type, allowed values: '.implode(', ', EmbeddingAssetBase::ALLOWED_STORAGE_TYPE_VALUES),
					6
				);
			}
		}
		if (array_key_exists('lifetime', $payload)) {
			$asset->setLifetime($payload['lifetime']);
		}
		if (array_key_exists('description', $payload)) {
			$asset->setDescription($payload['description']);
		}

		if (array_key_exists('translations', $payload)) {
			foreach ($payload['translations'] as $localeName => $translation) {
				if ($locale = LocaleService::find((string)$localeName)) {
					if (array_key_exists('lifetime', $translation)) {
						$asset->setTranslatedLifetime($locale, $translation['lifetime']);
					}
					if (array_key_exists('description', $translation)) {
						$asset->setTranslatedDescription($locale, $translation['description']);
					}
				}
				else {
					throw new InvalidArgumentException("Invalid locale name \"$localeName\" for translation.", 3);
				}
			}
		}
	}

	/**
	 * @param Embedding $embedding
	 * @return array
	 */
	private static function embeddingToSchemaArray(Embedding $embedding): array
	{
		$locales = LocaleService::enabledLocales();
		$locales = array_combine(
			array_map(function (LocaleBase $locale) { return $locale->getName(); }, $locales),
			$locales
		);

		return [
			'id' => $embedding->getId(),
			'name' => $embedding->getName(),
			'active' => $embedding->isActive(),
			'purpose' => $embedding->getPurpose()->getId(),
			'vendor' => $embedding->getEmbeddingVendor(),

			'systemEmbeddingId' => $embedding->getSource() == EmbeddingBase::SOURCE_SYSTEM_TEMPLATE ? $embedding->getSourceId() : null,
			'autoUpdate' => $embedding->isAutoUpdated(),

			'scriptHtmlCode' => $embedding->getCodeSnippet(),
			'scriptLoaderGroup' => $embedding->getScriptLoaderGroup(),
			'scriptMarkers' => $embedding->getCodeBlock(),

			'iframeMarkers' => $embedding->getIframeBlockingStrings(),
			'hideIframesUntilConsent' => $embedding->isIframesHiddenUntilConsent(),
			'dispatchLoadEventForIframes' => $embedding->isDispatchLoadEventForIframes(),

			'excludeUrlsMode' => $embedding->getExcludeUrlsMode(),
			'excludeFromUrls' => $embedding->getExcludeUrls(),

			'description' => $embedding->getDescription(),
			'privacyPolicyUrl' => $embedding->getPrivacyPolicyUrl(),
			'whatDataIsCollected' => $embedding->getCollectedDataInfo(),
			'purposeOfDataCollection' => $embedding->getPurposeOfDataCollection(),
			'legalBasis' => $embedding->getLegalBasis(),
			'placeOfProcessing' => $embedding->getPlaceOfProcessing(),
			'sourceId' => $embedding->getSourceId(),

			'translations' => array_map(function (LocaleBase $locale) use ($embedding) {
				return [
					'name' => $embedding->getTranslatedName($locale, false),
					'description' => $embedding->getTranslatedDescription($locale, false),
					'privacyPolicyUrl' => $embedding->getTranslatedPrivacyPolicyUrl($locale, false),
					'whatDataIsCollected' => $embedding->getTranslatedCollectedDataInfo($locale, false),
					'purposeOfDataCollection' => $embedding->getTranslatedPurposeOfDataCollection($locale, false),
					'legalBasis' => $embedding->getTranslatedLegalBasis($locale, false),
					'placeOfProcessing' => $embedding->getTranslatedPlaceOfProcessing($locale, false),
					'blockedContentTitle' => $embedding->getTranslatedBlockedContentTitle($locale, false),
					'blockedContentText' => $embedding->getTranslatedBlockedContentText($locale, false),
					'blockedContentButtonText' => $embedding->getTranslatedBlockedContentButtonText($locale, false),
				];
			}, $locales),

			'tcfVendor' => $embedding->getTcfVendor() ? (int)$embedding->getTcfVendor()->getId() : null,
			'tcfPurposes' => array_values(array_map(function (TcfPurpose $purpose) {
				return (int)$purpose->getId();
			}, $embedding->getTcfPurposes())),
			'tcfSpecialFeatures' => array_values(array_map(function (SpecialFeature $feature) {
				return (int)$feature->getId();
			}, $embedding->getTcfSpecialFeatures())),

			'googleConsentMode' => $embedding->getGoogleConsentMode(),

			'assets' => array_values(array_map(function (EmbeddingAsset $asset) {
				return self::embeddingAssetToSchemaArray($asset);
			}, $embedding->getAssets())),
		];
	}

	/**
	 * @param EmbeddingAsset $asset
	 * @return array
	 */
	private static function embeddingAssetToSchemaArray(EmbeddingAsset $asset): array
	{
		$locales = LocaleService::enabledLocales();
		$locales = array_combine(
			array_map(function (LocaleBase $locale) { return $locale->getName(); }, $locales),
			$locales
		);

		return [
			'id' => $asset->getId(),
			'name' => $asset->getName(),
			'dynamic' => $asset->isDynamic(),
			'storageType' => $asset->getStorageType(),
			'lifetime' => $asset->getLifetime(),
			'description' => $asset->getDescription(),

			'translations' => array_map(function (LocaleBase $locale) use ($asset) {
				return [
					'lifetime' => $asset->getTranslatedLifetime($locale, false),
					'description' => $asset->getTranslatedDescription($locale, false),
				];
			}, $locales),
		];
	}
}
