<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\Exception\BadRequestException;
use App\Component\LocaleService;
use App\Component\LoggerInterface;
use App\Component\Reflection\PropertyProcessor;
use App\Controller\HostingApiController;
use App\Exception\LocaleAlreadyExistsException;
use App\Model\LegalText;
use App\Model\Locale;
use App\Model\LocaleBase;
use App\Model\Purpose;
use App\Model\UserLocale;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/hosting/api/client/{userId}/domain/{domainId}/locale", name="app_hosting_api_locale_")
 */
class LocaleApi extends HostingApiController
{
	private const LOCALE_SCHEMA = [
		'name' => 'string',
		'active' => 'bool',

		'widgetTitle' => 'string',
		'widgetIntroText' => 'string',
		'widgetDeclineButtonText' => 'string',
		'widgetFullConsentButtonText' => 'string',
		'widgetSettingsButtonText' => 'string',

		'checkAllButtonText' => 'string',
		'uncheckAllButtonText' => 'string',
		'withdrawAllButtonText' => 'string',

		'consentNoSharingText' => 'string',
		'consentSharingText' => 'string',

		'consentStateOffText' => 'string',
		'consentStateOnText' => 'string',
		'consentStatePartialText' => 'string',

		'imprint' => 'string',
		'imprintHtmlCode' => 'string', // LegalText
		'imprintUrl' => 'string', // LegalText

		'privacyPolicy' => 'string',
		'privacyPolicyHtmlCode' => 'string', // LegalText
		'privacyPolicyUrl' => 'string', // LegalText

		'accessibility' => 'string',
		'accessibilityHtmlCode' => 'string', // LegalText
		'accessibilityUrl' => 'string', // LegalText

		'categoriesTitle' => 'string',
		'categoriesCancelButtonText' => 'string',
		'categoriesFullConsentButtonText' => 'string',
		'categoriesSaveButtonText' => 'string',

		'detailsTitle' => 'string',
		'detailsCloseButtonText' => 'string',

		'embeddingVendorLabel' => 'string',
		'embeddingDescriptionLabel' => 'string',
		'embeddingPrivacyPolicyUrlLabel' => 'string',
		'embeddingLifetimeLabel' => 'string',
		'embeddingWhichDataIsCollectedLabel' => 'string',
		'embeddingPurposeOfDataCollectionLabel' => 'string',
		'embeddingLegalBasisLabel' => 'string',
		'embeddingPlaceOfProcessingLabel' => 'string',

		'embeddingAssetNameLabel' => 'string',
		'embeddingAssetLifetimeLabel' => 'string',
		'embeddingAssetDescriptionLabel' => 'string',

		'blockedContentTitle' => 'string',
		'blockedContentDisclaimer' => 'string',
		'blockedContentConsentButtonText' => 'string',
		'blockedContentConsentSwitchShowExternalContent' => 'string',
		'blockedContentConsentSwitchAllowEmbedding' => 'string',
		'blockedContentConsentSwitchProviderDetails' => 'string',

		'purposeNames' => 'string[]',
		'purposeDescriptions' => 'string[]',

		'changeConsentText' => 'string',
		'moreInformationLabel' => 'string',
		'questionMarkButtonText' => 'string',

		'tcfVendorIntroText' => 'string',
		'tcfPurposeLabel' => 'string',
		'tcfSpecialPurposeLabel' => 'string',
		'tcfFeaturesLabel' => 'string',
		'tcfSpecialFeaturesLabel' => 'string',
		'tcfOptedInRemark' => 'string',
		'tcfThirdPartiesLabel' => 'string',
		'tcfConsentLabel' => 'string',
		'tcfLegitimateInterestLabel' => 'string',
		'nonTcfPurposesLabel' => 'string',
		'nonTcfVendorsLabel' => 'string',
		'tcfVendorUsesNonCookieStorageText' => 'string',
		'tcfSpecialFeaturesHandledGloballyText' => 'string',
		'tcfWidgetIntroText' => 'string',
		'tcfStacksListText' => 'string',
		'tcfCookieRefreshText' => 'string',
		'tcfExtendedDisclosureText' => 'string',
		'tcfAssetTypeLabel' => 'string',
		'tcfAssetPurposesLabel' => 'string',
		'tcfDataCategoriesLabel' => 'string',
		'tcfDataRetentionText' => 'string',
		'tcfDataRetentionSessionText' => 'string',
		'tcfIllustrationsLabel' => 'string',
		'tcfPurposeVendorListLabel' => 'string',
	];

	/**
	 * @Route("", name="create", methods={"POST"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function createLocale(string $userId, string $domainId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		$localeName = (string)($payload['name'] ?? '');

		try {
			// Create the new locale
			$locale = Locale::create($localeName);
		}
		catch (LocaleAlreadyExistsException $exception) {
			return $this->json([
				'code' => 2,
				'error' => "The locale $localeName already exists."
			], 400);
		}

		$legalText = LegalText::byLocale($locale);
		assert($legalText instanceof LegalText);

		try {
			// Fill in omitted properties with values from the user locale
			if ($userLocale = UserLocale::find($localeName)) {
				PropertyProcessor::copyValues($userLocale, $locale, LocaleBase::class);
			}

			// Try to process the payload and update relevant properties
			self::migrateLocalePayload($locale, $legalText, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Delete new objects if an error has occurred (rollback)
			LegalText::delete($legalText->getId());
			Locale::delete($locale->getId());

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Delete new objects if an error has occurred (rollback)
			LegalText::delete($legalText->getId());
			Locale::delete($locale->getId());

			// Dev time, let it rip
			throw $exception;
		}

		$legalText->save();
		$locale->save();

		$logger->logInfo('Locale-API', "Locale created: {$locale->getName()} [$userId::$domainId]");

		return $this->json(self::localeToSchemaArray($locale, $legalText), 201);
	}

	/**
	 * @Route("", name="read_all", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @return JsonResponse
	 */
	public function readAllLocales(string $userId, string $domainId): JsonResponse
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		return $this->json(array_values(array_map(function (Locale $locale) {
			return self::localeToSchemaArray($locale, LegalText::byLocale($locale));
		}, Locale::all())), 200);
	}

	/**
	 * @Route("/{localeName}", name="read", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $localeName
	 * @return JsonResponse
	 */
	public function readLocale(string $userId, string $domainId, string $localeName): JsonResponse
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		if ($locale = Locale::find($localeName)) {
			return $this->json(self::localeToSchemaArray($locale, LegalText::byLocale($locale)), 200);
		}
		else {
			return $this->json([
				'error' => 'Locale not found.',
			], 404);
		}
	}

	/**
	 * @Route("/{localeName}", name="update", methods={"PUT"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $localeName
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function updateLocale(string $userId, string $domainId, string $localeName, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		$locale = Locale::find($localeName);

		if (!$locale) {
			return $this->json([
				'error' => 'Locale not found.',
			], 404);
		}

		try {
			$legalText = LegalText::byLocale($locale);
			assert($legalText instanceof LegalText);

			// Try to process the payload and update relevant properties
			self::migrateLocalePayload($locale, $legalText, $payload, true);
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			LegalText::reload();
			Locale::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		$legalText->save();
		$locale->save();

		$logger->logInfo('Locale-API', "Locale updated: {$locale->getName()} [$userId::$domainId]");

		return $this->json(self::localeToSchemaArray($locale, $legalText), 200);
	}

	/**
	 * @Route("/{localeName}", name="delete", methods={"DELETE"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $localeName
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteLocale(string $userId, string $domainId, string $localeName, LoggerInterface $logger): Response
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		if ($locale = Locale::find($localeName)) {
			$logger->logInfo('Locale-API', "Locale deleted: {$locale->getName()} [$userId::$domainId]");

			if ($legalText = LegalText::byLocale($locale)) {
				LegalText::delete($legalText->getId());
			}

			Locale::delete($locale->getId());
		}
		else {
			return $this->json([
				'error' => 'Locale not found.',
			], 404);
		}

		return new Response('', 204);
	}

	/**
	 * @param Locale $locale
	 * @param LegalText $legalText
	 * @param array $payload
	 * @param bool $update
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migrateLocalePayload(Locale $locale, LegalText $legalText, array $payload, bool $update=false): void
	{
		self::checkPayloadDataTypes($payload, self::LOCALE_SCHEMA);

		if (array_key_exists('name', $payload) || empty($locale->getName())) {
			if ($payload['name'] ?? '') {
				if ($update) {
					throw new InvalidArgumentException('Locale name is read-only.', 3);
				}
				else {
					$locale->setName($payload['name']);
				}
			}
			else {
				throw new InvalidArgumentException('Empty or missing locale name.', 1);
			}
		}

		if (array_key_exists('active', $payload)) {
			$locale->setEnabled($payload['active']);

			if (count(LocaleService::enabledLocales()) == 0) {
				throw new InvalidArgumentException('At least one locale has to be active.', 5);
			}
		}

		if (array_key_exists('widgetTitle', $payload)) {
			$locale->setWidgetTitle($payload['widgetTitle']);
		}
		if (array_key_exists('widgetIntroText', $payload)) {
			$locale->setWidgetIntroduction($payload['widgetIntroText']);
		}
		if (array_key_exists('widgetDeclineButtonText', $payload)) {
			$locale->setWidgetDeclineButtonText($payload['widgetDeclineButtonText']);
		}
		if (array_key_exists('widgetFullConsentButtonText', $payload)) {
			$locale->setWidgetConsentButtonText($payload['widgetFullConsentButtonText']);
		}
		if (array_key_exists('widgetSettingsButtonText', $payload)) {
			$locale->setWidgetSettingsButtonText($payload['widgetSettingsButtonText']);
		}

		if (array_key_exists('checkAllButtonText', $payload)) {
			$locale->setCheckAllButtonText($payload['checkAllButtonText']);
		}
		if (array_key_exists('uncheckAllButtonText', $payload)) {
			$locale->setUncheckAllButtonText($payload['uncheckAllButtonText']);
		}
		if (array_key_exists('withdrawAllButtonText', $payload)) {
			$locale->setWithdrawAllText($payload['withdrawAllButtonText']);
		}

		if (array_key_exists('consentNoSharingText', $payload)) {
			$locale->setConsentNoSharingText($payload['consentNoSharingText']);
		}
		if (array_key_exists('consentSharingText', $payload)) {
			$locale->setConsentSharingText($payload['consentSharingText']);
		}

		if (array_key_exists('consentStateOffText', $payload)) {
			$locale->setStatusInactiveText($payload['consentStateOffText']);
		}
		if (array_key_exists('consentStateOnText', $payload)) {
			$locale->setStatusActiveText($payload['consentStateOnText']);
		}
		if (array_key_exists('consentStatePartialText', $payload)) {
			$locale->setStatusPartialText($payload['consentStatePartialText']);
		}

		if (array_key_exists('imprint', $payload)) {
			$locale->setImprintText($payload['imprint']);
		}
		if (array_key_exists('imprintHtmlCode', $payload)) {
			$legalText->setImprint($payload['imprintHtmlCode']);
		}
		if (array_key_exists('imprintUrl', $payload)) {
			$legalText->setImprintUrl($payload['imprintUrl']);
		}

		if (array_key_exists('privacyPolicy', $payload)) {
			$locale->setPrivacyPolicyText($payload['privacyPolicy']);
		}
		if (array_key_exists('privacyPolicyHtmlCode', $payload)) {
			$legalText->setPrivacyPolicy($payload['privacyPolicyHtmlCode']);
		}
		if (array_key_exists('privacyPolicyUrl', $payload)) {
			$legalText->setPrivacyUrl($payload['privacyPolicyUrl']);
		}

		if (array_key_exists('accessibility', $payload)) {
			$locale->setAccessibilityText($payload['accessibility']);
		}
		if (array_key_exists('accessibilityHtmlCode', $payload)) {
			$legalText->setAccessibility($payload['accessibilityHtmlCode']);
		}
		if (array_key_exists('accessibilityUrl', $payload)) {
			$legalText->setAccessibilityUrl($payload['accessibilityUrl']);
		}

		if (array_key_exists('categoriesTitle', $payload)) {
			$locale->setControlPanelTitle($payload['categoriesTitle']);
		}
		if (array_key_exists('categoriesCancelButtonText', $payload)) {
			$locale->setControlPanelCancelButtonText($payload['categoriesCancelButtonText']);
		}
		if (array_key_exists('categoriesFullConsentButtonText', $payload)) {
			$locale->setCategoriesAcceptAllButtonText($payload['categoriesFullConsentButtonText']);
		}
		if (array_key_exists('categoriesSaveButtonText', $payload)) {
			$locale->setControlPanelSaveButtonText($payload['categoriesSaveButtonText']);
		}

		if (array_key_exists('detailsTitle', $payload)) {
			$locale->setDetailsTitle($payload['detailsTitle']);
		}
		if (array_key_exists('detailsCloseButtonText', $payload)) {
			$locale->setDetailsCloseButtonText($payload['detailsCloseButtonText']);
		}

		if (array_key_exists('embeddingVendorLabel', $payload)) {
			$locale->setCookieVendorText($payload['embeddingVendorLabel']);
		}
		if (array_key_exists('embeddingDescriptionLabel', $payload)) {
			$locale->setCookieDescriptionText($payload['embeddingDescriptionLabel']);
		}
		if (array_key_exists('embeddingPrivacyPolicyUrlLabel', $payload)) {
			$locale->setCookiePrivacyPolicyUrlText($payload['embeddingPrivacyPolicyUrlLabel']);
		}
		if (array_key_exists('embeddingLifetimeLabel', $payload)) {
			$locale->setCookieLifetimeText($payload['embeddingLifetimeLabel']);
		}
		if (array_key_exists('embeddingWhichDataIsCollectedLabel', $payload)) {
			$locale->setCookieCollectedDataInfoText($payload['embeddingWhichDataIsCollectedLabel']);
		}
		if (array_key_exists('embeddingPurposeOfDataCollectionLabel', $payload)) {
			$locale->setCookiePurposeOfDataCollectionText($payload['embeddingPurposeOfDataCollectionLabel']);
		}
		if (array_key_exists('embeddingLegalBasisLabel', $payload)) {
			$locale->setCookieLegalBasisText($payload['embeddingLegalBasisLabel']);
		}
		if (array_key_exists('embeddingPlaceOfProcessingLabel', $payload)) {
			$locale->setCookiePlaceOfProcessingText($payload['embeddingPlaceOfProcessingLabel']);
		}

		if (array_key_exists('embeddingAssetNameLabel', $payload)) {
			$locale->setAssetNameLabel($payload['embeddingAssetNameLabel']);
		}
		if (array_key_exists('embeddingAssetLifetimeLabel', $payload)) {
			$locale->setAssetLifetimeLabel($payload['embeddingAssetLifetimeLabel']);
		}
		if (array_key_exists('embeddingAssetDescriptionLabel', $payload)) {
			$locale->setAssetDescriptionLabel($payload['embeddingAssetDescriptionLabel']);
		}

		if (array_key_exists('blockedContentTitle', $payload)) {
			$locale->setBlockedContentTitle($payload['blockedContentTitle']);
		}
		if (array_key_exists('blockedContentDisclaimer', $payload)) {
			$locale->setBlockedContentText($payload['blockedContentDisclaimer']);
		}
		if (array_key_exists('blockedContentConsentButtonText', $payload)) {
			$locale->setBlockedContentButtonText($payload['blockedContentConsentButtonText']);
		}
		if (array_key_exists('blockedContentConsentSwitchShowExternalContent', $payload)) {
			$locale->setBlockedContentShowExternalContent($payload['blockedContentConsentSwitchShowExternalContent']);
		}
		if (array_key_exists('blockedContentConsentSwitchAllowEmbedding', $payload)) {
			$locale->setBlockedContentAllowEmbedding($payload['blockedContentConsentSwitchAllowEmbedding']);
		}
		if (array_key_exists('blockedContentConsentSwitchProviderDetails', $payload)) {
			$locale->setBlockedContentProviderDetails($payload['blockedContentConsentSwitchProviderDetails']);
		}

		$purposes = Purpose::all();
		if (array_key_exists('purposeNames', $payload)) {
			array_walk($payload['purposeNames'], function ($translation, $purposeId) use ($locale, $purposes) {
				if ($purpose = $purposes[$purposeId] ?? null) {
					$locale->setPurposeName($purpose, $translation);
				}
				else {
					throw new InvalidArgumentException("Purpose $purposeId doesn't exist, valid keys: ".implode(', ', array_keys($purposes)), 4);
				}
			});
		}
		if (array_key_exists('purposeDescriptions', $payload)) {
			array_walk($payload['purposeDescriptions'], function ($translation, $purposeId) use ($locale, $purposes) {
				if ($purpose = $purposes[$purposeId] ?? null) {
					$locale->setPurposeDescription($purpose, $translation);
				}
				else {
					throw new InvalidArgumentException("Purpose $purposeId doesn't exist, valid keys: ".implode(', ', array_keys($purposes)), 4);
				}
			});
		}

		if (array_key_exists('changeConsentText', $payload)) {
			$locale->setChangeConsentLabel($payload['changeConsentText']);
		}
		if (array_key_exists('moreInformationLabel', $payload)) {
			$locale->setMoreInformationLabel($payload['moreInformationLabel']);
		}
		if (array_key_exists('questionMarkButtonText', $payload)) {
			$locale->setMorePurposeDetailsButtonText($payload['questionMarkButtonText']);
		}

		if (array_key_exists('tcfVendorIntroText', $payload)) {
			$locale->setTcfVendorGrantedPermissionsText($payload['tcfVendorIntroText']);
		}
		if (array_key_exists('tcfPurposeLabel', $payload)) {
			$locale->setTcfPurposesLabel($payload['tcfPurposeLabel']);
		}
		if (array_key_exists('tcfSpecialPurposeLabel', $payload)) {
			$locale->setTcfSpecialPurposesLabel($payload['tcfSpecialPurposeLabel']);
		}
		if (array_key_exists('tcfFeaturesLabel', $payload)) {
			$locale->setTcfFeaturesLabel($payload['tcfFeaturesLabel']);
		}
		if (array_key_exists('tcfSpecialFeaturesLabel', $payload)) {
			$locale->setTcfSpecialFeaturesLabel($payload['tcfSpecialFeaturesLabel']);
		}
		if (array_key_exists('tcfOptedInRemark', $payload)) {
			$locale->setTcfOptedInRemark($payload['tcfOptedInRemark']);
		}
		if (array_key_exists('tcfThirdPartiesLabel', $payload)) {
			$locale->setTcfThirdPartiesLabel($payload['tcfThirdPartiesLabel']);
		}
		if (array_key_exists('tcfConsentLabel', $payload)) {
			$locale->setTcfConsentLabel($payload['tcfConsentLabel']);
		}
		if (array_key_exists('tcfLegitimateInterestLabel', $payload)) {
			$locale->setTcfLegitimateInterestLabel($payload['tcfLegitimateInterestLabel']);
		}
		if (array_key_exists('nonTcfPurposesLabel', $payload)) {
			$locale->setNonTcfPurposesLabel($payload['nonTcfPurposesLabel']);
		}
		if (array_key_exists('nonTcfVendorsLabel', $payload)) {
			$locale->setNonTcfVendorsLabel($payload['nonTcfVendorsLabel']);
		}
		if (array_key_exists('tcfVendorUsesNonCookieStorageText', $payload)) {
			$locale->setTcfVendorUsesNonCookieStorageText($payload['tcfVendorUsesNonCookieStorageText']);
		}
		if (array_key_exists('tcfSpecialFeaturesHandledGloballyText', $payload)) {
			$locale->setTcfSpecialFeaturesHandledGloballyText($payload['tcfSpecialFeaturesHandledGloballyText']);
		}
		if (array_key_exists('tcfWidgetIntroText', $payload)) {
			$locale->setTcfWidgetIntroduction($payload['tcfWidgetIntroText']);
		}
		if (array_key_exists('tcfStacksListText', $payload)) {
			$locale->setTcfStacksListText($payload['tcfStacksListText']);
		}
		if (array_key_exists('tcfCookieRefreshText', $payload)) {
			$locale->setTcfCookieRefreshText($payload['tcfCookieRefreshText']);
		}
		if (array_key_exists('tcfExtendedDisclosureText', $payload)) {
			$locale->setTcfExtendedDisclosureText($payload['tcfExtendedDisclosureText']);
		}
		if (array_key_exists('tcfAssetTypeLabel', $payload)) {
			$locale->setTcfAssetTypeLabel($payload['tcfAssetTypeLabel']);
		}
		if (array_key_exists('tcfAssetPurposesLabel', $payload)) {
			$locale->setTcfAssetPurposesLabel($payload['tcfAssetPurposesLabel']);
		}
		if (array_key_exists('tcfDataCategoriesLabel', $payload)) {
			$locale->setTcfDataCategoriesLabel($payload['tcfDataCategoriesLabel']);
		}
		if (array_key_exists('tcfDataRetentionText', $payload)) {
			$locale->setTcfDataRetentionText($payload['tcfDataRetentionText']);
		}
		if (array_key_exists('tcfDataRetentionSessionText', $payload)) {
			$locale->setTcfDataRetentionSessionText($payload['tcfDataRetentionSessionText']);
		}
		if (array_key_exists('tcfIllustrationsLabel', $payload)) {
			$locale->setTcfIllustrationsLabel($payload['tcfIllustrationsLabel']);
		}
		if (array_key_exists('tcfPurposeVendorListLabel', $payload)) {
			$locale->setTcfPurposeVendorListLabel($payload['tcfPurposeVendorListLabel']);
		}
	}

	/**
	 * @param Locale $locale
	 * @param LegalText|null $legalText
	 * @return array
	 */
	private static function localeToSchemaArray(Locale $locale, ?LegalText $legalText): array
	{
		$purposes = Purpose::all();

		return [
			'name' => $locale->getName(),
			'active' => $locale->isEnabled(),

			'widgetTitle' => $locale->getWidgetTitle(),
			'widgetIntroText' => $locale->getWidgetIntroduction(),
			'widgetDeclineButtonText' => $locale->getWidgetDeclineButtonText(),
			'widgetFullConsentButtonText' => $locale->getWidgetConsentButtonText(),
			'widgetSettingsButtonText' => $locale->getWidgetSettingsButtonText(),

			'checkAllButtonText' => $locale->getCheckAllButtonText(),
			'uncheckAllButtonText' => $locale->getUncheckAllButtonText(),
			'withdrawAllButtonText' => $locale->getWithdrawAllText(),

			'consentNoSharingText' => $locale->getConsentNoSharingText(),
			'consentSharingText' => $locale->getConsentSharingText(),

			'consentStateOffText' => $locale->getStatusInactiveText(),
			'consentStateOnText' => $locale->getStatusActiveText(),
			'consentStatePartialText' => $locale->getStatusPartialText(),

			'imprint' => $locale->getImprintText(),
			'imprintHtmlCode' => $legalText ? $legalText->getImprint() : '',
			'imprintUrl' => $legalText ? $legalText->getImprintUrl() : '',

			'privacyPolicy' => $locale->getPrivacyPolicyText(),
			'privacyPolicyHtmlCode' => $legalText ? $legalText->getPrivacyPolicy() : '',
			'privacyPolicyUrl' => $legalText ? $legalText->getPrivacyUrl() : '',

			'accessibility' => $locale->getAccessibilityText(),
			'accessibilityHtmlCode' => $legalText ? $legalText->getAccessibility() : '',
			'accessibilityUrl' => $legalText ? $legalText->getAccessibilityUrl() : '',

			'categoriesTitle' => $locale->getControlPanelTitle(),
			'categoriesCancelButtonText' => $locale->getControlPanelCancelButtonText(),
			'categoriesFullConsentButtonText' => $locale->getCategoriesAcceptAllButtonText(),
			'categoriesSaveButtonText' => $locale->getControlPanelSaveButtonText(),

			'detailsTitle' => $locale->getDetailsTitle(),
			'detailsCloseButtonText' => $locale->getDetailsCloseButtonText(),

			'embeddingVendorLabel' => $locale->getCookieVendorText(),
			'embeddingDescriptionLabel' => $locale->getCookieDescriptionText(),
			'embeddingPrivacyPolicyUrlLabel' => $locale->getCookiePrivacyPolicyUrlText(),
			'embeddingLifetimeLabel' => $locale->getCookieLifetimeText(),
			'embeddingWhichDataIsCollectedLabel' => $locale->getCookieCollectedDataInfoText(),
			'embeddingPurposeOfDataCollectionLabel' => $locale->getCookiePurposeOfDataCollectionText(),
			'embeddingLegalBasisLabel' => $locale->getCookieLegalBasisText(),
			'embeddingPlaceOfProcessingLabel' => $locale->getCookiePlaceOfProcessingText(),

			'embeddingAssetNameLabel' => $locale->getAssetNameLabel(),
			'embeddingAssetLifetimeLabel' => $locale->getAssetLifetimeLabel(),
			'embeddingAssetDescriptionLabel' => $locale->getAssetDescriptionLabel(),

			'blockedContentTitle' => $locale->getBlockedContentTitle(),
			'blockedContentDisclaimer' => $locale->getBlockedContentText(),
			'blockedContentConsentButtonText' => $locale->getBlockedContentButtonText(),
			'blockedContentConsentSwitchShowExternalContent' => $locale->getBlockedContentShowExternalContent(),
			'blockedContentConsentSwitchAllowEmbedding' => $locale->getBlockedContentAllowEmbedding(),
			'blockedContentConsentSwitchProviderDetails' => $locale->getBlockedContentProviderDetails(),

			'purposeNames' => array_map(function (Purpose $purpose) use ($locale) {
				return $locale->getPurposeName($purpose);
			}, $purposes),
			'purposeDescriptions' => array_map(function (Purpose $purpose) use ($locale) {
				return $locale->getPurposeDescription($purpose);
			}, $purposes),

			'changeConsentText' => $locale->getChangeConsentLabel(),
			'moreInformationLabel' => $locale->getMoreInformationLabel(),
			'questionMarkButtonText' => $locale->getMorePurposeDetailsButtonText(),

			'tcfVendorIntroText' => $locale->getTcfVendorGrantedPermissionsText(),
			'tcfPurposeLabel' => $locale->getTcfPurposesLabel(),
			'tcfSpecialPurposeLabel' => $locale->getTcfSpecialPurposesLabel(),
			'tcfFeaturesLabel' => $locale->getTcfFeaturesLabel(),
			'tcfSpecialFeaturesLabel' => $locale->getTcfSpecialFeaturesLabel(),
			'tcfOptedInRemark' => $locale->getTcfOptedInRemark(),
			'tcfThirdPartiesLabel' => $locale->getTcfThirdPartiesLabel(),
			'tcfConsentLabel' => $locale->getTcfConsentLabel(),
			'tcfLegitimateInterestLabel' => $locale->getTcfLegitimateInterestLabel(),
			'nonTcfPurposesLabel' => $locale->getNonTcfPurposesLabel(),
			'nonTcfVendorsLabel' => $locale->getNonTcfVendorsLabel(),
			'tcfVendorUsesNonCookieStorageText' => $locale->getTcfVendorUsesNonCookieStorageText(),
			'tcfSpecialFeaturesHandledGloballyText' => $locale->getTcfSpecialFeaturesHandledGloballyText(),
			'tcfWidgetIntroText' => $locale->getTcfWidgetIntroduction(),
			'tcfStacksListText' => $locale->getTcfStacksListText(),
			'tcfCookieRefreshText' => $locale->getTcfCookieRefreshText(),
			'tcfExtendedDisclosureText' => $locale->getTcfExtendedDisclosureText(),
			'tcfAssetTypeLabel' => $locale->getTcfAssetTypeLabel(),
			'tcfAssetPurposesLabel' => $locale->getTcfAssetPurposesLabel(),
			'tcfDataCategoriesLabel' => $locale->getTcfDataCategoriesLabel(),
			'tcfDataRetentionText' => $locale->getTcfDataRetentionText(),
			'tcfDataRetentionSessionText' => $locale->getTcfDataRetentionSessionText(),
			'tcfIllustrationsLabel' => $locale->getTcfIllustrationsLabel(),
			'tcfPurposeVendorListLabel' => $locale->getTcfPurposeVendorListLabel(),
		];
	}
}
