<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\Exception\BadRequestException;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\LocaleService;
use App\Component\LoggerInterface;
use App\Component\OnBoarding\Cookie;
use App\Component\OnBoarding\ScanData;
use App\Component\OnBoarding\ScanService;
use App\Controller\HostingApiController;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingAssetBase;
use App\Model\EmbeddingBase;
use App\Model\Purpose;
use App\Model\User;
use DateTimeImmutable;
use Exception;
use InvalidArgumentException;
use RuntimeException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\HttpKernelInterface;
use Symfony\Component\HttpKernel\KernelInterface;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;

/**
 * @Route("/hosting/api/client/{userId}/domain/{domainId}/onboarding", name="app_hosting_api_onboarding_")
 */
class OnboardingApi extends HostingApiController
{
	private const PRESET_SCHEMA = [
		'id' => 'int',
		'active' => 'bool',
	];

	private const COOKIE_LIST_SCHEMA = [
		'name' => 'string',
		'storageType' => 'string',
	];

	private const STORAGE_TYPE_COOKIE = 'cookie';
	private const STORAGE_TYPE_LOCAL_STORAGE = 'localStorage';
	private const STORAGE_TYPE_SESSION_STORAGE = 'sessionStorage';

	public const ALLOWED_STORAGE_TYPES = [
		self::STORAGE_TYPE_COOKIE,
		self::STORAGE_TYPE_LOCAL_STORAGE,
		self::STORAGE_TYPE_SESSION_STORAGE,
	];

	/**
	 * @Route("/scan", name="scan", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param ScanService $scanService
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function scan(string $userId, string $domainId, Request $request, ScanService $scanService, LoggerInterface $logger): JsonResponse
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Domain $domain */
		$domain = Domain::find($domainId);
		assert($domain instanceof Domain);

		$siteInfo = $scanService->fetchImprintData($domain->getUrl());

		$contentLength = (int)($siteInfo['index_length'] ?? 0);
		$httpStatusCode = (int)($siteInfo['httpcode'] ?? 0);
		if ($httpStatusCode < 200 || $httpStatusCode > 299 || $contentLength == 0) {
			return $this->json([
				'code' => 1,
				'error' => "The website didn't send a success response code. ({$domain->getUrl()})",
			], 400);
		}

		try {
			$scanResult = $scanService->scanDomain($domain->getUrl(), $logger);
		}
		catch (RuntimeException $exception) {
			if ($request->query->getBoolean('ignoreCrawlerError')) {
				$scanResult = new ScanData([]);
			}
			else {
				return $this->json(['error'=>$exception->getMessage()], 500);
			}
		}

		$embeddings = $scanService->fetchMatchingEmbeddings($scanResult);
		$scanService->supplementUnmatchedCookies($embeddings, $scanResult, false);
		$embeddingsGroupedByPurpose = $scanService->splitEmbeddingsByPurpose($embeddings);
		$embeddings = array_reduce($embeddingsGroupedByPurpose, function (array $embeddings, array $group) {
			return array_merge($embeddings, $group);
		}, []);

		return $this->json([
			'address' => [
				'company' => trim($siteInfo['name'] ?? '') ?: null,
				'street' => trim($siteInfo['addr'] ?? '') ?: null,
				'zipCode' => trim($siteInfo['plz'] ?? '') ?: null,
				'city' => trim($siteInfo['stadt'] ?? '') ?: null,
			],
			'imprintUrl' => trim($siteInfo['impressum'] ?? '') ?: null,
			'privacyPolicyUrl' => trim($siteInfo['datenschutz_url'] ?? '') ?: null,
			'availablePresets' => array_values(array_map(function (array $preset) {
				$varyingProperties = [];

				// Add unmappable assets to their respective fallback embedding
				if (in_array($preset['id'], [
					0, // optional assets | "Sonstige"
					13002, // mandatory assets | "Betreiber der Webseite"
				])) {
					$varyingProperties['assets'] = array_map(function (array $asset) {
						return [
							'name' => $asset['name'],
							'dynamic' => $asset['dynamic'],
							'storageType' => $asset['storageType'],
							'translations' => array_map(function (array $translation) {
								return [
									'lifetime' => $translation['lifetime'] ?? '',
									'description' => $translation['description'] ?? '',
								];
							}, $asset['translations'] ?? []),
						];
					}, $preset['assets'] ?? []);
				}

				return [
					'id' => (int)$preset['id'],
					'name' => $preset['name'],
					'purpose' => $preset['purpose'],
					'vendor' => $preset['embedding_vendor'],
					'translations' => array_map(function (array $translation) {
						return [
							'name' => $translation['label'] ?? '',
							'description' => $translation['description'] ?? '',
							/* Hide this information to encourage the use of the preset import endpoint(s)
							'privacyPolicyUrl' => $translation['privacy_policy_url'] ?? '',
							'whatDataIsCollected' => $translation['collected_data_info'] ?? '',
							'purposeOfDataCollection' => $translation['purpose_of_data_collection'] ?? '',
							*/
							'legalBasis' => $translation['legal_basis'] ?? '',
							/*
							'placeOfProcessing' => $translation['place_of_processing'] ?? '',
							*/
						];
					}, $preset['translations'] ?? []),
				] + $varyingProperties;
			}, $embeddings)),
			'storage' => [
				'cookies' => array_values(array_map(function (Cookie $cookie) {
					return [
						'name' => $cookie->getName(),
						'domain' => (string)$cookie->getDomain(),
						'path' => (string)$cookie->getPath(),
					];
				}, $scanResult->getCookies())),
				'localStorage' => array_values(array_map(function (Cookie $cookie) {
					return [
						'name' => $cookie->getName(),
					];
				}, $scanResult->getLocalStorage())),
				'sessionStorage' => array_values(array_map(function (Cookie $cookie) {
					return [
						'name' => $cookie->getName(),
					];
				}, $scanResult->getSessionStorage())),
			],
		]);
	}

	/**
	 * @Route("/presets", name="preset_import_bulk", methods={"POST"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function bulkImportPresets(string $userId, string $domainId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Validate the payload and parse preset IDs
		try {
			array_walk($payload, function ($item) {
				if (is_array($item) == false) {
					throw new InvalidArgumentException('Invalid payload, collection of presets expected.');
				}
				self::checkPayloadDataTypes($item, self::PRESET_SCHEMA);
			});

			/** @var int[] $presetIds */
			$presetIds = array_filter(array_map(function (array $preset) {
				if (array_key_exists('id', $preset) == false) {
					throw new InvalidArgumentException('Preset ID is required.');
				}
				return $preset['id'];
			}, $payload));

			// Silently ignore preset 13002
			// TODO Maybe later perform another domain scan to find missing assets and add this embedding instead.
			// TODO If you do, remove custom handling of 13002 in AioApi::processWebsite.
			$presetIds = array_filter($presetIds, function (int $id) {
				return $id != 13002;
			});
		}
		catch (InvalidArgumentException $exception) {
			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		if (empty($presetIds)) {
			return $this->json([
				'code' => 1,
				'error' => 'No preset IDs defined.',
			], 400);
		}

		$alreadyImportedPresetIds = array_filter(array_map(function (Embedding $embedding) use ($presetIds) {
			return $embedding->getSource() == EmbeddingBase::SOURCE_GLOBAL_DB && in_array($embedding->getSourceId(), $presetIds)
				? (int)$embedding->getSourceId()
				: null;
		}, Embedding::all()));
		sort($alreadyImportedPresetIds);

		if ($alreadyImportedPresetIds && !$request->query->getBoolean('ignoreExistingPresets')) {
			return $this->json([
				'code' => 2,
				'error' => 'The following presets are already active: '.implode(',', $alreadyImportedPresetIds),
			], 400);
		}

		// Subtract already imported presets
		$presetIds = array_diff($presetIds, $alreadyImportedPresetIds);

		if (empty($presetIds)) {
			return $this->json([]);
		}

		$presets = EmbeddingDb::fetchEmbeddings($presetIds);

		if ($unknownPresetIds = array_diff($presetIds, array_keys($presets))) {
			return $this->json([
				'code' => 3,
				'error' => 'The following presets don\'t exist: '.implode(',', $unknownPresetIds),
			], 400);
		}

		// Determine the active state by considering the payload for each preset
		array_walk($presets, function (&$preset) use ($payload) {
			$preset['active'] = array_reduce($payload, function (bool $active, array $item) use ($preset) {
				return $item['id'] == $preset['id'] ? ($item['active'] ?? false) : $active;
			}, false);
		});

		Embedding::beginWrite();
		EmbeddingAsset::beginWrite();

		$embeddings = array_map(function (array $preset) {
			return self::createEmbeddingFromPreset($preset);
		}, $presets);

		EmbeddingAsset::flush();
		Embedding::flush();

		// Determine if embeddings have been created that use the iframe blocker
		$forceIframeBlockerEnabled = array_reduce($embeddings, function (bool $useIframeBlocker, Embedding $embedding) {
			return $useIframeBlocker || count($embedding->getIframeBlockingStrings()) > 0;
		}, false);

		// Update domain settings
		$domain = Domain::activeDomain();

		if ($forceIframeBlockerEnabled) {
			$domain->setIframesBlocked(true);
		}

		$domain->save();

		$logger->logInfo("Onboarding-API", "Presets imported. (".implode(';', $presetIds).") [$userId::$domainId]");

		return $this->json(array_values(array_map(function (Embedding $embedding) {
			return [
				'presetId' => (int)$embedding->getSourceId(),
				'embeddingId' => $embedding->getId(),
			];
		}, $embeddings)), 201);
	}

	/**
	 * @Route("/embeddingsFromCookies", name="embeddings_from_cookies", methods={"POST"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param ScanService $scanService
	 * @param KernelInterface $kernel
	 * @return Response
	 */
	public function bulkCreateEmbeddingsFromCookieNames(string $userId, string $domainId, Request $request, ScanService $scanService, KernelInterface $kernel): Response
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Validate the payload
		try {
			array_walk($payload, function ($item) {
				if (is_array($item) == false) {
					throw new InvalidArgumentException('Invalid payload, OnboardingCookie collection expected.');
				}
				self::checkPayloadDataTypes($item, self::COOKIE_LIST_SCHEMA);

				if (strlen(trim($item['name'] ?? '')) == 0) {
					throw new InvalidArgumentException('Empty or missing name.');
				}
				elseif (in_array($item['storageType'] ?? '', self::ALLOWED_STORAGE_TYPES) == false) {
					throw new InvalidArgumentException('Invalid or missing storageType, allowed values: '.implode(', ', self::ALLOWED_STORAGE_TYPES));
				}
			});

			if (empty($payload)) {
				throw new InvalidArgumentException('Empty payload, at least a single item is necessary.');
			}
		}
		catch (InvalidArgumentException $exception) {
			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		// Compile embedding presets
		$scanResult = new ScanData([
			'cookies' => array_map(function ($item) { return [
				'name' => $item['name'],
			]; }, array_filter($payload, function ($item) { return $item['storageType'] == self::STORAGE_TYPE_COOKIE; })),
			'localStorage' => array_map(function ($item) { return [
				'name' => $item['name'],
			]; }, array_filter($payload, function ($item) { return $item['storageType'] == self::STORAGE_TYPE_LOCAL_STORAGE; })),
			'sessionStorage' => array_map(function ($item) { return [
				'name' => $item['name'],
			]; }, array_filter($payload, function ($item) { return $item['storageType'] == self::STORAGE_TYPE_SESSION_STORAGE; })),
		]);

		$presets = $scanService->fetchMatchingEmbeddings($scanResult);
		$scanService->supplementUnmatchedCookies($presets, $scanResult, false);

		$presetsGroupedByPurpose = $scanService->splitEmbeddingsByPurpose($presets);
		$presets = array_reduce($presetsGroupedByPurpose, function (array $presets, array $group) {
			return array_merge($presets, $group);
		}, []);

		// Import embeddings in a sub request via the onboarding API
		if ($presetIds = array_filter(array_map(function ($preset) { return (int)$preset['id']; }, $presets))) {
			$attributes = [
				'_controller' => self::class.'::bulkImportPresets',
				'userId' => $userId,
				'domainId' => $domainId,
			];
			$body = json_encode(array_values(array_map(function (int $presetId) {
				return ['id'=>$presetId, 'active'=>true];
			}, $presetIds)));

			$subRequest = new Request(['ignoreExistingPresets'=>true], [], $attributes, [], [], [], $body);
			$subRequest->setMethod('POST');
			$subRequest->headers->set('Content-Type', 'application/json');

			$response = $kernel->handle($subRequest, HttpKernelInterface::SUB_REQUEST);
			if (!$response->isSuccessful()) {
				return $response;
			}
		}

		// Create an embedding for unmatched cookies via the embedding API
		$orphansPreset = array_reduce($presets, function ($carry, $item) {
			return $item['id'] ? $carry : $item;
		});
		// Check if corresponding EmbeddingAssets exist to prevent duplicates
		if ($orphansPreset) {
			$availableAssets = array_reduce(Embedding::all(), function (array $assets, Embedding $embedding) {
				return array_merge_recursive($assets, array_reduce(
					$embedding->getAssets(),
					function (array $assets, EmbeddingAsset $asset) {
						$assets[$asset->getStorageType()][] = $asset->getName();
						return $assets;
					},
					array_fill_keys(EmbeddingAssetBase::ALLOWED_STORAGE_TYPE_VALUES, [])
				));
			}, array_fill_keys(EmbeddingAssetBase::ALLOWED_STORAGE_TYPE_VALUES, []));

			$orphansPreset['assets'] = array_filter($orphansPreset['assets'], function (array $asset) use ($availableAssets) {
				return in_array($asset['name'], $availableAssets[$asset['storageType']] ?? []) == false;
			});
		}
		if ($orphansPreset && $orphansPreset['assets']) {
			$attributes = [
				'_controller' => EmbeddingApi::class.'::createEmbedding',
				'userId' => $userId,
				'domainId' => $domainId,
			];
			$body = json_encode([
				'name' => $orphansPreset['name'],
				'active' => true,
				'purpose' => $orphansPreset['purpose'],
				'vendor' => $orphansPreset['embedding_vendor'],
				'translations' => [
					'de_DE' => [
						'name' => $orphansPreset['translations']['de']['label'],
						'legalBasis' => $orphansPreset['translations']['de']['legal_basis'],
					],
					'en_US' => [
						'name' => $orphansPreset['translations']['en']['label'],
						'legalBasis' => $orphansPreset['translations']['en']['legal_basis'],
					],
				],
				'assets' => array_map(function ($asset) {
					return [
						'name' => $asset['name'],
						'dynamic' => (bool)$asset['dynamic'],
						'storageType' => $asset['storageType'],
						'translations' => [
							'de_DE' => [
								'lifetime' => $asset['translations']['de']['lifetime'],
								'description' => $asset['translations']['de']['description'],
							],
							'en_US' => [
								'lifetime' => $asset['translations']['en']['lifetime'],
								'description' => $asset['translations']['en']['description'],
							],
						],
					];
				}, $orphansPreset['assets']),
			]);

			$subRequest = new Request([], [], $attributes, [], [], [], $body);
			$subRequest->setMethod('POST');
			$subRequest->headers->set('Content-Type', 'application/json');

			$kernel->handle($subRequest, HttpKernelInterface::SUB_REQUEST);
		}

		return new Response('', 201);
	}

	/**
	 * @Route("/embed", name="embed", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @return JsonResponse
	 */
	public function getCodeSnippets(string $userId, string $domainId): JsonResponse
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var User $user */
		$user = User::byId($userId);
		assert($user instanceof User);

		$params = [
			'apiKey' => $user->getApiKey(),
			'domain' => $domainId,
		];

		$urls = [
			'default' => $this->generateUrl('app_external_main_javascript_file', $params, UrlGeneratorInterface::ABSOLUTE_URL),
		];

		foreach (LocaleService::enabledLocales() as $locale) {
			$urls[$locale->getName()] = $this->generateUrl('app_external_main_javascript_file', $params + [
				'lang' => $locale->getName(),
			], UrlGeneratorInterface::ABSOLUTE_URL);
		}

		$snippets = array_map(function (string $url) {
			return '<script src="'.htmlspecialchars($url).'" referrerpolicy="origin"></script>';
		}, $urls);

		return $this->json($snippets);
	}

	/**
	 * @param array $preset
	 * @return Embedding
	 */
	private static function createEmbeddingFromPreset(array $preset): Embedding
	{
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleName = strtolower(explode('_', $defaultLocale->getName())[0]);
		$enabledLocales = LocaleService::enabledLocales();

		/** @var DateTimeImmutable|null $lastModified */
		if ($lastModified = (string)($preset['lastModified'] ?? '') ?: null) {
			try {
				$lastModified = new DateTimeImmutable($lastModified);
			}
			catch (Exception $exception) {
				$lastModified = null;
			}
		}

		$purpose = Purpose::find($preset['purpose'] ?? '') ?? Purpose::find('15c61c3');
		assert($purpose !== null);
		$scriptMarkers = array_filter(array_map('trim', preg_split('~\r\n|\n|\r~', (string)($preset['code_block'] ?? ''))));
		$iframeMarkers = array_filter(array_map('trim', preg_split('~\r\n|\n|\r~', (string)($preset['iframe_blocking_strings'] ?? ''))));

		/** @var Embedding $embedding */
		$embedding = Embedding::create();

		$embedding
			->setName((string)($preset['name'] ?? ''))
			->setActive((bool)($preset['active'] ?? false))
			->setCodeSnippet((string)($preset['code_snippet'] ?? ''))
			->setCodeBlock($scriptMarkers)
			->setIframeBlockingStrings($iframeMarkers)
			->setPurpose($purpose)
			->setEmbeddingVendor((string)($preset['embedding_vendor'] ?? '') ?? '')
			->setDescription((string)($preset['translations'][$defaultLocaleName]['description'] ?? ''))
			->setPrivacyPolicyUrl((string)($preset['translations'][$defaultLocaleName]['privacy_policy_url'] ?? ''))
			->setCollectedDataInfo((string)($preset['translations'][$defaultLocaleName]['collected_data_info'] ?? ''))
			->setPurposeOfDataCollection((string)($preset['translations'][$defaultLocaleName]['purpose_of_data_collection'] ?? ''))
			->setLegalBasis((string)($preset['translations'][$defaultLocaleName]['legal_basis'] ?? ''))
			->setPlaceOfProcessing((string)($preset['translations'][$defaultLocaleName]['place_of_processing'] ?? ''))
			->setSource(EmbeddingBase::SOURCE_GLOBAL_DB)
			->setSourceId((int)($preset['id'] ?? 0) ?: null)
			->setSourceTimestamp($lastModified)
			->setAutoUpdated(true)
		;

		foreach ($enabledLocales as $locale) {
			$localeName = strtolower(explode('_', $locale->getName())[0]);
			if (isset($preset['translations'][$localeName]['description']) == false) {
				continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
			}
			$embedding
				->setTranslatedDescription($locale, (string)($preset['translations'][$localeName]['description'] ?? ''))
				->setTranslatedPrivacyPolicyUrl($locale, (string)($preset['translations'][$localeName]['privacy_policy_url'] ?? ''))
				->setTranslatedCollectedDataInfo($locale, (string)($preset['translations'][$localeName]['collected_data_info'] ?? ''))
				->setTranslatedPurposeOfDataCollection($locale, (string)($preset['translations'][$localeName]['purpose_of_data_collection'] ?? ''))
				->setTranslatedLegalBasis($locale, (string)($preset['translations'][$localeName]['legal_basis'] ?? ''))
				->setTranslatedPlaceOfProcessing($locale, (string)($preset['translations'][$localeName]['place_of_processing'] ?? ''))
			;
		}

		/** @var EmbeddingAsset[] $assets */
		$assets = array_filter(array_map(function (array $data) use ($defaultLocaleName, $enabledLocales) {
			if (($assetName = trim((string)($data['name'] ?? ''))) == '') {
				return null;
			}

			/** @var EmbeddingAsset $asset */
			$asset = EmbeddingAsset::create();

			$asset
				->setName($assetName)
				->setDynamic((bool)($data['dynamic'] ?? false))
				->setStorageType((string)($data['storageType'] ?? ''))
				->setLifetime((string)($data['translations'][$defaultLocaleName]['lifetime'] ?? ''))
				->setDescription((string)($data['translations'][$defaultLocaleName]['description'] ?? ''))
			;

			foreach ($enabledLocales as $locale) {
				$localeName = strtolower(explode('_', $locale->getName())[0]);
				if (isset($data['translations'][$localeName]['description']) == false) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				$asset
					->setTranslatedLifetime($locale, (string)($data['translations'][$localeName]['lifetime'] ?? ''))
					->setTranslatedDescription($locale, (string)($data['translations'][$localeName]['description'] ?? ''))
				;
			}

			return $asset;
		}, $preset['assets'] ?? []));

		$embedding->setAssets($assets);

		return $embedding;
	}
}
