<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\Exception\BadRequestException;
use App\Component\LoggerInterface;
use App\Controller\HostingApiController;
use App\Model\Placeholder;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/hosting/api/client/{userId}/domain/{domainId}/placeholder", name="app_hosting_api_placeholder_")
 */
class PlaceholderApi extends HostingApiController
{
	private const PLACEHOLDER_SCHEMA = [
		'name' => 'string',
		'value' => 'string',
	];

	/**
	 * @Route("", name="create", methods={"POST"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function createPlaceholder(string $userId, string $domainId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		try {
			$placeholder = self::createPlaceholderFromPayload($payload);
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if any error occurred (rollback)
			Placeholder::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Revert changes if any error occurred (rollback)
			Placeholder::reload();

			// Dev time, let it rip
			throw $exception;
		}

		$placeholder->save();

		$logger->logInfo('Hosting-API', "Placeholder created: \"{$placeholder->getId()}\" [$userId::$domainId]");

		return $this->json(self::placeholderToSchemaArray($placeholder), 201);
	}

	/**
	 * @Route("", name="create_bulk", methods={"PUT"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function bulkCreatePlaceholders(string $userId, string $domainId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		if (empty($payload)) {
			return $this->json([
				'code' => 4,
				'error' => 'Empty payload, nothing to do.'
			], 400);
		}

		// Get exclusive lock here to prevent the skeleton from being applied when the first object is created
		Placeholder::beginWrite();
		// Delete every placeholder when creating in bulk
		Placeholder::clear(false);

		$placeholders = [];
		foreach ($payload as $item) {
			if (is_array($item)) {
				try {
					$placeholders[] = self::createPlaceholderFromPayload($item);
				}
				catch (InvalidArgumentException $exception) {
					// Revert changes if any error occurred (rollback)
					Placeholder::reload();

					return $this->json([
						'code' => $exception->getCode(),
						'error' => $exception->getMessage(),
					], 400);
				}
				catch (LogicException $exception) {
					// Revert changes if any error occurred (rollback)
					Placeholder::reload();

					// Dev time, let it rip
					throw $exception;
				}
			}
			else {
				return $this->json([
					'code' => 3,
					'error' => 'Invalid payload, collection of placeholders expected.'
				], 400);
			}
		}

		Placeholder::flush();

		$logger->logInfo('Hosting-API', "Placeholder(s) created. (bulk) [$userId::$domainId]");

		return $this->json(array_map(function (Placeholder $placeholder) {
			return self::placeholderToSchemaArray($placeholder);
		}, $placeholders), 201);
	}

	/**
	 * @Route("", name="read_all", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @return JsonResponse
	 */
	public function readAllPlaceholders(string $userId, string $domainId): JsonResponse
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		return $this->json(array_values(array_map(function (Placeholder $placeholder) {
			return self::placeholderToSchemaArray($placeholder);
		}, Placeholder::all())), 200);
	}

	/**
	 * @Route("/{placeholderId}", name="read", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $placeholderId
	 * @return JsonResponse
	 */
	public function readPlaceholder(string $userId, string $domainId, string $placeholderId): JsonResponse
	{
		try {
			self::assertPlaceholderExists($userId, $domainId, $placeholderId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Placeholder $placeholder */
		$placeholder = Placeholder::find($placeholderId);
		assert($placeholder instanceof Placeholder);

		return $this->json(self::placeholderToSchemaArray($placeholder), 200);
	}

	/**
	 * @Route("/{placeholderId}", name="update", methods={"PUT"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $placeholderId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function updatePlaceholder(string $userId, string $domainId, string $placeholderId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertPlaceholderExists($userId, $domainId, $placeholderId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Placeholder $placeholder */
		$placeholder = Placeholder::find($placeholderId);
		assert($placeholder instanceof Placeholder);

		try {
			// Try to process the payload and update relevant properties
			self::migratePlaceholderPayload($placeholder, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			Placeholder::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		$placeholder->save();

		$logger->logInfo('Hosting-API', "Placeholder updated: \"{$placeholder->getId()}\" [$userId::$domainId]");

		return $this->json(self::placeholderToSchemaArray($placeholder), 200);
	}

	/**
	 * @Route("/{placeholderId}", name="delete", methods={"DELETE"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $placeholderId
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deletePlaceholder(string $userId, string $domainId, string $placeholderId, LoggerInterface $logger): Response
	{
		try {
			self::assertPlaceholderExists($userId, $domainId, $placeholderId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Placeholder $placeholder */
		$placeholder = Placeholder::find($placeholderId);
		assert($placeholder instanceof Placeholder);

		$logger->logInfo('Hosting-API', "Placeholder deleted: \"{$placeholder->getId()}\" [$userId::$domainId]");
		Placeholder::delete($placeholder->getId());

		return new Response('', 204);
	}

	/**
	 * @param Placeholder $placeholder
	 * @param array $payload
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migratePlaceholderPayload(Placeholder $placeholder, array $payload): void
	{
		self::checkPayloadDataTypes($payload, self::PLACEHOLDER_SCHEMA);

		if (array_key_exists('value', $payload)) {
			$placeholder->setValue($payload['value']);
		}
	}

	/**
	 * @param Placeholder $placeholder
	 * @return array
	 */
	private static function placeholderToSchemaArray(Placeholder $placeholder): array
	{
		return [
			'name' => $placeholder->getName(),
			'value' => $placeholder->getValue(),
		];
	}

	/**
	 * @param array $payload
	 * @return Placeholder
	 */
	private static function createPlaceholderFromPayload(array $payload): Placeholder
	{
		$placeholderName = trim((string)($payload['name'] ?? ''));

		if (strlen($placeholderName) == 0) {
			throw new InvalidArgumentException('Empty or missing placeholder name.', 1);
		}
		elseif (Placeholder::exists($placeholderName)) {
			throw new InvalidArgumentException("The placeholder $placeholderName already exists.", 2);
		}

		$placeholder = Placeholder::create($placeholderName);

		// Try to process the payload and update relevant properties
		self::migratePlaceholderPayload($placeholder, $payload);

		return $placeholder;
	}
}
