<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\Exception\BadRequestException;
use App\Component\LoggerInterface;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Config;
use App\Controller\HostingApiController;
use App\Model\EmbeddingAssetBase;
use App\Model\EmbeddingBase;
use App\Model\Language;
use App\Model\Purpose;
use App\Model\SystemEmbedding;
use App\Model\SystemEmbeddingAsset;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/hosting/api/embedding", name="app_hosting_api_system_embedding_")
 */
class SystemEmbeddingApi extends HostingApiController
{
	private const SYSTEM_EMBEDDING_ASSET_SCHEMA = [
		'name' => 'string',
		'dynamic' => 'bool',
		'storageType' => 'string',
		'lifetime' => 'string',
		'description' => 'string',

		'translations' => [
			'@collection' => true,
			'@schema' => [
				'lifetime' => 'string',
				'description' => 'string',
			],
		],
	];

	private const SYSTEM_EMBEDDING_SCHEMA = [
		'name' => 'string',
		'active' => 'bool',
		'purpose' => 'string',
		'vendor' => 'string',

		'scriptHtmlCode' => 'string',
		'scriptLoaderGroup' => 'string',
		'scriptMarkers' => 'string[]',

		'iframeMarkers' => 'string[]',
		'hideIframesUntilConsent' => 'bool',
		'dispatchLoadEventForIframes' => 'bool',

		'excludeUrlsMode' => 'string',
		'excludeFromUrls' => 'string[]',

		'description' => 'string',
		'privacyPolicyUrl' => 'string',
		'whatDataIsCollected' => 'string',
		'purposeOfDataCollection' => 'string',
		'legalBasis' => 'string',
		'placeOfProcessing' => 'string',

		'translations' => [
			'@collection' => true,
			'@schema' => [
				'name' => 'string',
				'description' => 'string',
				'privacyPolicyUrl' => 'string',
				'whatDataIsCollected' => 'string',
				'purposeOfDataCollection' => 'string',
				'legalBasis' => 'string',
				'placeOfProcessing' => 'string',
				'blockedContentTitle' => 'string',
				'blockedContentText' => 'string',
				'blockedContentButtonText' => 'string',
			],
		],

		'tcfVendor' => '?int',
		'tcfPurposes' => 'int[]',
		'tcfSpecialFeatures' => 'int[]',

		'googleConsentMode' => 'string[]',

		'assets' => [
			'@collection' => true,
			'@schema' => self::SYSTEM_EMBEDDING_ASSET_SCHEMA,
		],
	];

	private const BULK_PATCH_SCHEMA = [
		'removeAll' => 'bool',
		'remove' => 'string[]',
		'update' => [
			'@collection' => true,
			'@schema' => self::SYSTEM_EMBEDDING_SCHEMA,
		],
		'add' => [
			'@collection' => true,
			'@schema' => self::SYSTEM_EMBEDDING_SCHEMA,
		],
	];

	/**
	 * @Route("", name="create", methods={"POST"})
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function createEmbedding(Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request, 6);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Create the new embedding
		$embedding = SystemEmbedding::create();

		try {
			// Try to process the payload and update relevant properties
			self::migrateEmbeddingPayload($embedding, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Delete new objects if an error has occurred (rollback)
			SystemEmbeddingAsset::reload();
			SystemEmbedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Delete new objects if an error has occurred (rollback)
			SystemEmbeddingAsset::reload();
			SystemEmbedding::reload();

			// Dev time, let it rip
			throw $exception;
		}

		// Predefine properties that aren't accessible via the API
		$embedding
			->setSource(EmbeddingBase::SOURCE_LOCAL)
			->setAutoUpdated(false)
		;

		SystemEmbeddingAsset::flush();
		$embedding->save();

		$logger->logInfo('Hosting-API', "SystemEmbedding created: {$embedding->getName()} [{$embedding->getId()}]");

		return $this->json(self::embeddingToSchemaArray($embedding), 201);
	}

	/**
	 * @Route("", name="read_all", methods={"GET"})
	 * @return JsonResponse
	 */
	public function readAllEmbeddings(): JsonResponse
	{
		return $this->json(array_values(array_map(function (SystemEmbedding $embedding) {
			return self::embeddingToSchemaArray($embedding);
		}, SystemEmbedding::all())));
	}

	/**
	 * @Route("/{embeddingId}", name="read", methods={"GET"})
	 * @param string $embeddingId
	 * @return JsonResponse
	 */
	public function readEmbedding(string $embeddingId): JsonResponse
	{
		try {
			self::assertSystemEmbeddingExists($embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var SystemEmbedding $embedding */
		$embedding = SystemEmbedding::find($embeddingId);
		assert($embedding instanceof SystemEmbedding);

		return $this->json(self::embeddingToSchemaArray($embedding));
	}

	/**
	 * @Route("/{embeddingId}", name="update", methods={"PUT"})
	 * @param string $embeddingId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function updateEmbedding(string $embeddingId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertSystemEmbeddingExists($embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var SystemEmbedding $embedding */
		$embedding = SystemEmbedding::find($embeddingId);
		assert($embedding instanceof SystemEmbedding);

		try {
			// Try to process the payload and update relevant properties
			self::migrateEmbeddingPayload($embedding, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			SystemEmbeddingAsset::reload();
			SystemEmbedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		SystemEmbeddingAsset::flush();
		$embedding->save();

		$logger->logInfo('Hosting-API', "SystemEmbedding updated: {$embedding->getName()} [$embeddingId]");

		return $this->json(self::embeddingToSchemaArray($embedding));
	}

	/**
	 * @Route("/{embeddingId}", name="delete", methods={"DELETE"})
	 * @param string $embeddingId
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteEmbedding(string $embeddingId, LoggerInterface $logger): Response
	{
		try {
			self::assertSystemEmbeddingExists($embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var SystemEmbedding $embedding */
		$embedding = SystemEmbedding::find($embeddingId);
		assert($embedding instanceof SystemEmbedding);

		try {
			foreach ($embedding->getAssets() as $asset) {
				SystemEmbeddingAsset::delete($asset->getId(), false);
			}
		}
		finally {
			SystemEmbeddingAsset::flush();
		}

		$logger->logInfo('Hosting-API', "SystemEmbedding deleted: {$embedding->getName()} [{$embedding->getId()}]");
		SystemEmbedding::delete($embedding->getId());

		return new Response('', 204);
	}

	/**
	 * @Route("", name="patch_bulk", methods={"PATCH"})
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function patchEmbeddings(Request $request, LoggerInterface $logger): Response
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::checkPayloadDataTypes($payload, self::BULK_PATCH_SCHEMA);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}
		catch (InvalidArgumentException $exception) {
			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		SystemEmbedding::beginWrite();
		SystemEmbeddingAsset::beginWrite();

		$removeAll = $payload['removeAll'] ?? false;
		$toRemove = $payload['remove'] ?? [];
		$toUpdate = $payload['update'] ?? [];
		$toAdd = $payload['add'] ?? [];
		$removed = $updated = $added = [];
		if (!is_array($payload) or array_diff(array_keys($payload), ['remove', 'update', 'add', 'removeAll'])) {
			return $this->json(['error' => 'invalid payload structure'], 400);
		}
		if ($toRemove and $removeAll) {
			return $this->json(['error' => '"remove" and "removeAll" are mutually exclusive'], 400);
		}

		try {
			if ($removeAll) {
				$removed = array_values(array_map(function (SystemEmbedding $item) { return $item->getId(); }, SystemEmbedding::all()));
				SystemEmbedding::clear(false);
				SystemEmbeddingAsset::clear(false);
				$logger->logInfo('Hosting-API', "All system embeddings deleted in bulk operation");
			}

			foreach ($toRemove as $embeddingId) {
				/** @var SystemEmbedding $embedding */
				$embedding = SystemEmbedding::find($embeddingId);
				if (!$embedding) { continue; }

				foreach ($embedding->getAssets() as $asset) {
					try {
						SystemEmbeddingAsset::delete($asset->getId(), false);
					}
					catch (\Throwable $ignore) {
						// ignore
					}
				}
				SystemEmbedding::delete($embedding->getId(), false);
				$removed[] = $embedding->getId();
				$logger->logInfo('Hosting-API', "SystemEmbedding deleted in bulk operation: {$embedding->getName()}");
			}

			foreach ($toUpdate as $embeddingId => $embeddingData) {
				self::assertSystemEmbeddingExists($embeddingId);

				/** @var SystemEmbedding $embedding */
				$embedding = SystemEmbedding::find($embeddingId);
				assert($embedding instanceof SystemEmbedding);
				$preUpdate = self::embeddingToSchemaArray($embedding);

				// Try to process the payload and update relevant properties
				self::migrateEmbeddingPayload($embedding, $embeddingData);
				$postUpdate = self::embeddingToSchemaArray($embedding);

				$logger->logInfo('Hosting-API', "SystemEmbedding updated in bulk operation: {$embedding->getName()}");
				if ($preUpdate !== $postUpdate) {
					$updated[$embedding->getId()] = self::embeddingToSchemaArray($embedding);
				}
			}

			foreach ($toAdd as $embeddingData) {
				// Create the new embedding
				$embedding = SystemEmbedding::create();

				// Try to process the payload and update relevant properties
				self::migrateEmbeddingPayload($embedding, $embeddingData);

				// Predefine properties that aren't accessible via the API
				if ($embedding->isImported() == false) {
					$embedding
						->setSource(EmbeddingBase::SOURCE_LOCAL)
						->setAutoUpdated(false)
					;
				}

				$added[] = self::embeddingToSchemaArray($embedding);
				$logger->logInfo('Hosting-API', "SystemEmbedding created in bulk operation: {$embedding->getName()}");
			}
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}
		catch (InvalidArgumentException $exception) {
			// Restore previous state if an error has occurred (rollback)
			SystemEmbeddingAsset::reload();
			SystemEmbedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Restore previous state if an error has occurred (rollback)
			SystemEmbeddingAsset::reload();
			SystemEmbedding::reload();

			// Dev time, let it rip
			throw $exception;
		}

		SystemEmbeddingAsset::flush();
		SystemEmbedding::flush();

		$logger->logInfo('Hosting-API', "SystemEmbedding bulk operation finished");

		return new JsonResponse([
			'removed' => $removed,
			'updated' => $updated,
			'added' => $added,
		], 200);
	}

	/**
	 * @Route("/{embeddingId}/asset", name="asset_create", methods={"POST"})
	 * @param string $embeddingId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function createEmbeddingAsset(string $embeddingId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertSystemEmbeddingExists($embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Lock embedding repository first, then create the new embedding asset
		SystemEmbedding::beginWrite();
		/** @var SystemEmbedding $embedding */
		$embedding = SystemEmbedding::find($embeddingId);
		assert($embedding instanceof SystemEmbedding);
		$asset = SystemEmbeddingAsset::create();

		try {
			// Try to process the payload and update relevant properties
			self::migrateEmbeddingAssetPayload($asset, $payload);
			$embedding->addAsset($asset);
			$embedding->updateLastModified();
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			SystemEmbeddingAsset::reload();
			SystemEmbedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (LogicException $exception) {
			// Revert changes if an error has occurred (rollback)
			SystemEmbeddingAsset::reload();
			SystemEmbedding::reload();

			// Dev time, let it rip
			throw $exception;
		}

		$asset->save();
		$embedding->save();

		$logger->logInfo('Hosting-API', "SystemEmbeddingAsset created: \"{$asset->getName()}\" in \"{$embedding->getName()}\"  [$embeddingId::{$asset->getId()}]");

		return $this->json(self::embeddingAssetToSchemaArray($asset), 201);
	}

	/**
	 * @Route("/{embeddingId}/asset", name="asset_read_all", methods={"GET"})
	 * @param string $embeddingId
	 * @return JsonResponse
	 */
	public function readAllEmbeddingAssets(string $embeddingId): JsonResponse
	{
		try {
			self::assertSystemEmbeddingExists($embeddingId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var SystemEmbedding $embedding */
		$embedding = SystemEmbedding::find($embeddingId);
		assert($embedding instanceof SystemEmbedding);

		return $this->json(array_values(array_map(function (SystemEmbeddingAsset $asset) {
			return self::embeddingAssetToSchemaArray($asset);
		}, $embedding->getAssets())));
	}

	/**
	 * @Route("/{embeddingId}/asset/{assetId}", name="asset_read", methods={"GET"})
	 * @param string $embeddingId
	 * @param string $assetId
	 * @return JsonResponse
	 */
	public function readEmbeddingAsset(string $embeddingId, string $assetId): JsonResponse
	{
		try {
			self::assertSystemEmbeddingAssetExists($embeddingId, $assetId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Directly query the asset repository, as it's read only.
		/** @var SystemEmbeddingAsset $asset */
		$asset = SystemEmbeddingAsset::find($assetId);
		assert($asset instanceof SystemEmbeddingAsset);

		return $this->json(self::embeddingAssetToSchemaArray($asset));
	}

	/**
	 * @Route("/{embeddingId}/asset/{assetId}", name="asset_update", methods={"PUT"})
	 * @param string $embeddingId
	 * @param string $assetId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function updateEmbeddingAsset(string $embeddingId, string $assetId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertSystemEmbeddingAssetExists($embeddingId, $assetId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var SystemEmbedding $embedding */
		$embedding = SystemEmbedding::find($embeddingId);
		assert($embedding instanceof SystemEmbedding);
		/** @var SystemEmbeddingAsset $asset */
		$asset = $embedding->getAsset($assetId);
		assert($asset instanceof SystemEmbeddingAsset);

		try {
			// Lock embedding repository first
			SystemEmbedding::beginWrite();
			// Try to process the payload and update relevant properties
			self::migrateEmbeddingAssetPayload($asset, $payload);
			$embedding->updateLastModified();
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			SystemEmbeddingAsset::reload();
			SystemEmbedding::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		$asset->save();
		$embedding->save();

		$logger->logInfo('Hosting-API', "SystemEmbeddingAsset updated: \"{$asset->getName()}\" in \"{$embedding->getName()}\"  [$embeddingId::$assetId]");

		return $this->json(self::embeddingAssetToSchemaArray($asset));
	}

	/**
	 * @Route("/{embeddingId}/asset/{assetId}", name="asset_delete", methods={"DELETE"})
	 * @param string $embeddingId
	 * @param string $assetId
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteEmbeddingAsset(string $embeddingId, string $assetId, LoggerInterface $logger): Response
	{
		try {
			self::assertSystemEmbeddingAssetExists($embeddingId, $assetId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var SystemEmbedding $embedding */
		$embedding = SystemEmbedding::find($embeddingId);
		assert($embedding instanceof SystemEmbedding);
		/** @var SystemEmbeddingAsset $asset */
		$asset = $embedding->getAsset($assetId);
		assert($asset instanceof SystemEmbeddingAsset);

		$logger->logInfo('Hosting-API', "SystemEmbeddingAsset deleted: \"{$asset->getName()}\" from \"{$embedding->getName()}\"  [$embeddingId::$assetId]");

		$embedding->removeAsset($assetId);
		$embedding->updateLastModified();
		SystemEmbeddingAsset::flush();
		$embedding->save();

		return new Response('', 204);
	}

	/**
	 * @param SystemEmbedding $embedding
	 * @param array $payload
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migrateEmbeddingPayload(SystemEmbedding $embedding, array $payload): void
	{
		self::checkPayloadDataTypes($payload, self::SYSTEM_EMBEDDING_SCHEMA);

		$config = Config::getInstance();
		$languages = $config->getSystemEmbeddingLanguages();

		if (array_key_exists('name', $payload) || empty($embedding->getName())) {
			if ($payload['name'] ?? '') {
				$embedding->setName($payload['name']);
			}
			else {
				throw new InvalidArgumentException('Empty or missing name.', 1);
			}
		}
		if (array_key_exists('active', $payload)) {
			$embedding->setActive($payload['active']);
		}
		if (array_key_exists('purpose', $payload) || $embedding->getPurpose() == null) {
			if ($purpose = Purpose::find($payload['purpose'] ?? '')) {
				$embedding->setPurpose($purpose);
			}
			else {
				throw new InvalidArgumentException('Empty, invalid or missing purpose, allowed values: '.implode(', ', array_keys(Purpose::all())), 2);
			}
		}
		if (array_key_exists('vendor', $payload)) {
			$embedding->setEmbeddingVendor($payload['vendor']);

			$tcfVendor = array_reduce(TcfVendor::all(), function (?TcfVendor $carry, TcfVendor $vendor) use ($payload) {
				return $carry ?? ($vendor->getName() == $payload['vendor'] ? $vendor : $carry);
			});
			$embedding->setTcfVendor($tcfVendor);
		}

		if (array_key_exists('scriptHtmlCode', $payload)) {
			$embedding->setCodeSnippet($payload['scriptHtmlCode']);
		}
		if (array_key_exists('scriptLoaderGroup', $payload)) {
			$embedding->setScriptLoaderGroup($payload['scriptLoaderGroup']);
		}
		if (array_key_exists('scriptMarkers', $payload)) {
			$embedding->setCodeBlock(array_values(array_filter(array_map('trim', (array)$payload['scriptMarkers']))));
		}

		if (array_key_exists('iframeMarkers', $payload)) {
			$embedding->setIframeBlockingStrings(array_values(array_filter(array_map('trim', (array)$payload['iframeMarkers']))));
		}
		if (array_key_exists('hideIframesUntilConsent', $payload)) {
			$embedding->setIframesHiddenUntilConsent($payload['hideIframesUntilConsent']);
		}
		if (array_key_exists('dispatchLoadEventForIframes', $payload)) {
			$embedding->setDispatchLoadEventForIframes($payload['dispatchLoadEventForIframes']);
		}

		if (array_key_exists('excludeUrlsMode', $payload)) {
			if (in_array($payload['excludeUrlsMode'], EmbeddingBase::ALLOWED_EXCLUDE_URLS_MODE_VALUES)) {
				$embedding->setExcludeUrlsMode((string)$payload['excludeUrlsMode']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid excludeUrlsMode, allowed values: '.implode(', ', EmbeddingBase::ALLOWED_EXCLUDE_URLS_MODE_VALUES),
					3
				);
			}
		}
		if (array_key_exists('excludeFromUrls', $payload)) {
			$embedding->setExcludeUrls(array_values(array_filter(array_map('trim', (array)$payload['excludeFromUrls']))));
		}

		if (array_key_exists('description', $payload)) {
			$embedding->setDescription($payload['description']);
		}
		if (array_key_exists('privacyPolicyUrl', $payload)) {
			$embedding->setPrivacyPolicyUrl($payload['privacyPolicyUrl']);
		}
		if (array_key_exists('whatDataIsCollected', $payload)) {
			$embedding->setCollectedDataInfo($payload['whatDataIsCollected']);
		}
		if (array_key_exists('purposeOfDataCollection', $payload)) {
			$embedding->setPurposeOfDataCollection($payload['purposeOfDataCollection']);
		}
		if (array_key_exists('legalBasis', $payload)) {
			$embedding->setLegalBasis($payload['legalBasis']);
		}
		if (array_key_exists('placeOfProcessing', $payload)) {
			$embedding->setPlaceOfProcessing($payload['placeOfProcessing']);
		}

		if (array_key_exists('translations', $payload)) {
			foreach ($payload['translations'] as $localeName => $translation) {
				if ($language = $languages[$localeName] ?? null) {
					if (array_key_exists('name', $translation)) {
						$embedding->setTranslatedName($language, $translation['name']);
					}
					if (array_key_exists('description', $translation)) {
						$embedding->setTranslatedDescription($language, $translation['description']);
					}
					if (array_key_exists('privacyPolicyUrl', $translation)) {
						$embedding->setTranslatedPrivacyPolicyUrl($language, $translation['privacyPolicyUrl']);
					}
					if (array_key_exists('whatDataIsCollected', $translation)) {
						$embedding->setTranslatedCollectedDataInfo($language, $translation['whatDataIsCollected']);
					}
					if (array_key_exists('purposeOfDataCollection', $translation)) {
						$embedding->setTranslatedPurposeOfDataCollection($language, $translation['purposeOfDataCollection']);
					}
					if (array_key_exists('legalBasis', $translation)) {
						$embedding->setTranslatedLegalBasis($language, $translation['legalBasis']);
					}
					if (array_key_exists('placeOfProcessing', $translation)) {
						$embedding->setTranslatedPlaceOfProcessing($language, $translation['placeOfProcessing']);
					}
					if (array_key_exists('blockedContentTitle', $translation)) {
						$embedding->setTranslatedBlockedContentTitle($language, $translation['blockedContentTitle']);
					}
					if (array_key_exists('blockedContentText', $translation)) {
						$embedding->setTranslatedBlockedContentText($language, $translation['blockedContentText']);
					}
					if (array_key_exists('blockedContentButtonText', $translation)) {
						$embedding->setTranslatedBlockedContentButtonText($language, $translation['blockedContentButtonText']);
					}
				}
				elseif (Language::find($localeName)) {
					throw new InvalidArgumentException("Language \"$localeName\" is disabled.", 4);
				}
				else {
					throw new InvalidArgumentException("Invalid locale name \"$localeName\" for translation.", 3);
				}
			}
		}

		if (array_key_exists('tcfVendor', $payload)) {
			if ($payload['tcfVendor']) {
				if ($tcfVendor = TcfVendor::find((string)$payload['tcfVendor'])) {
					$embedding->setTcfVendor($tcfVendor);
					$embedding->setEmbeddingVendor($tcfVendor->getName());
				}
				else {
					throw new InvalidArgumentException('Invalid TCF vendor id.', 4);
				}
			}
			else {
				$embedding->setTcfVendor(null);
				if (array_key_exists('vendor', $payload) == false) {
					$embedding->setEmbeddingVendor('');
				}
			}
		}
		if (array_key_exists('tcfPurposes', $payload)) {
			/** @var TcfPurpose[] $purposes  */
			$purposes = array_filter(array_map(function ($id) {
				return TcfPurpose::find((string)$id);
			}, $payload['tcfPurposes']));

			$embedding->setTcfPurposes($purposes);
		}
		if (array_key_exists('tcfSpecialFeatures', $payload)) {
			/** @var SpecialFeature[] $specialFeatures */
			$specialFeatures = array_filter(array_map(function ($id) {
				return SpecialFeature::find((string)$id);
			}, $payload['tcfSpecialFeatures']));

			$embedding->setTcfSpecialFeatures($specialFeatures);
		}

		if (array_key_exists('googleConsentMode', $payload)) {
			$valid = array_reduce($payload['googleConsentMode'], function (bool $valid, string $mode) {
				return $valid && in_array($mode, EmbeddingBase::ALLOWED_GCM_VALUES);
			}, true);

			if ($valid) {
				$embedding->setGoogleConsentMode($payload['googleConsentMode']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid Google consent mode, allowed values: '.implode(', ', EmbeddingBase::ALLOWED_GCM_VALUES),
					5
				);
			}
		}

		if (array_key_exists('assets', $payload)) {
			foreach ($embedding->getAssets() as $asset) {
				SystemEmbeddingAsset::delete($asset->getId(), false);
			}

			$assets = array_map(function (array $payload) {
				$asset = SystemEmbeddingAsset::create();
				self::migrateEmbeddingAssetPayload($asset, $payload);
				return $asset;
			}, $payload['assets']);

			$embedding->setAssets($assets);
		}

		$embedding->updateLastModified();
	}

	/**
	 * @param SystemEmbeddingAsset $asset
	 * @param array $payload
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migrateEmbeddingAssetPayload(SystemEmbeddingAsset $asset, array $payload): void
	{
		self::checkPayloadDataTypes($payload, self::SYSTEM_EMBEDDING_ASSET_SCHEMA);

		$config = Config::getInstance();
		$languages = $config->getSystemEmbeddingLanguages();

		if (array_key_exists('name', $payload) || empty($asset->getName())) {
			if ($payload['name'] ?? '') {
				$asset->setName($payload['name']);
			}
			else {
				throw new InvalidArgumentException('Empty or missing name.', 1);
			}
		}
		if (array_key_exists('dynamic', $payload)) {
			$asset->setDynamic($payload['dynamic']);
		}
		if (array_key_exists('storageType', $payload)) {
			if (in_array($payload['storageType'], EmbeddingAssetBase::ALLOWED_STORAGE_TYPE_VALUES)) {
				$asset->setStorageType($payload['storageType']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid storage type, allowed values: '.implode(', ', EmbeddingAssetBase::ALLOWED_STORAGE_TYPE_VALUES),
					6
				);
			}
		}
		if (array_key_exists('lifetime', $payload)) {
			$asset->setLifetime($payload['lifetime']);
		}
		if (array_key_exists('description', $payload)) {
			$asset->setDescription($payload['description']);
		}

		if (array_key_exists('translations', $payload)) {
			foreach ($payload['translations'] as $localeName => $translation) {
				if ($language = $languages[$localeName] ?? null) {
					if (array_key_exists('lifetime', $translation)) {
						$asset->setTranslatedLifetime($language, $translation['lifetime']);
					}
					if (array_key_exists('description', $translation)) {
						$asset->setTranslatedDescription($language, $translation['description']);
					}
				}
				elseif (Language::find($localeName)) {
					throw new InvalidArgumentException("Language \"$localeName\" is disabled.", 4);
				}
				else {
					throw new InvalidArgumentException("Invalid locale name \"$localeName\" for translation.", 3);
				}
			}
		}
	}

	/**
	 * @param SystemEmbedding $embedding
	 * @return array
	 */
	private static function embeddingToSchemaArray(SystemEmbedding $embedding): array
	{
		$config = Config::getInstance();
		$languages = $config->getSystemEmbeddingLanguages();

		return [
			'id' => $embedding->getId(),
			'name' => $embedding->getName(),
			'active' => $embedding->isActive(),
			'purpose' => $embedding->getPurpose()->getId(),
			'vendor' => $embedding->getEmbeddingVendor(),

			'scriptHtmlCode' => $embedding->getCodeSnippet(),
			'scriptLoaderGroup' => $embedding->getScriptLoaderGroup(),
			'scriptMarkers' => $embedding->getCodeBlock(),

			'iframeMarkers' => $embedding->getIframeBlockingStrings(),
			'hideIframesUntilConsent' => $embedding->isIframesHiddenUntilConsent(),
			'dispatchLoadEventForIframes' => $embedding->isDispatchLoadEventForIframes(),

			'excludeUrlsMode' => $embedding->getExcludeUrlsMode(),
			'excludeFromUrls' => $embedding->getExcludeUrls(),

			'description' => $embedding->getDescription(),
			'privacyPolicyUrl' => $embedding->getPrivacyPolicyUrl(),
			'whatDataIsCollected' => $embedding->getCollectedDataInfo(),
			'purposeOfDataCollection' => $embedding->getPurposeOfDataCollection(),
			'legalBasis' => $embedding->getLegalBasis(),
			'placeOfProcessing' => $embedding->getPlaceOfProcessing(),

			'translations' => array_map(function (Language $language) use ($embedding) {
				return [
					'name' => $embedding->getTranslatedName($language, false),
					'description' => $embedding->getTranslatedDescription($language, false),
					'privacyPolicyUrl' => $embedding->getTranslatedPrivacyPolicyUrl($language, false),
					'whatDataIsCollected' => $embedding->getTranslatedCollectedDataInfo($language, false),
					'purposeOfDataCollection' => $embedding->getTranslatedPurposeOfDataCollection($language, false),
					'legalBasis' => $embedding->getTranslatedLegalBasis($language, false),
					'placeOfProcessing' => $embedding->getTranslatedPlaceOfProcessing($language, false),
					'blockedContentTitle' => $embedding->getTranslatedBlockedContentTitle($language, false),
					'blockedContentText' => $embedding->getTranslatedBlockedContentText($language, false),
					'blockedContentButtonText' => $embedding->getTranslatedBlockedContentButtonText($language, false),
				];
			}, $languages),

			'tcfVendor' => $embedding->getTcfVendor() ? (int)$embedding->getTcfVendor()->getId() : null,
			'tcfPurposes' => array_values(array_map(function (TcfPurpose $purpose) {
				return (int)$purpose->getId();
			}, $embedding->getTcfPurposes())),
			'tcfSpecialFeatures' => array_values(array_map(function (SpecialFeature $feature) {
				return (int)$feature->getId();
			}, $embedding->getTcfSpecialFeatures())),

			'googleConsentMode' => $embedding->getGoogleConsentMode(),

			'assets' => array_values(array_map(function (SystemEmbeddingAsset $asset) {
				return self::embeddingAssetToSchemaArray($asset);
			}, $embedding->getAssets())),
		];
	}

	/**
	 * @param SystemEmbeddingAsset $asset
	 * @return array
	 */
	private static function embeddingAssetToSchemaArray(SystemEmbeddingAsset $asset): array
	{
		$config = Config::getInstance();
		$languages = $config->getSystemEmbeddingLanguages();

		return [
			'id' => $asset->getId(),
			'name' => $asset->getName(),
			'dynamic' => $asset->isDynamic(),
			'storageType' => $asset->getStorageType(),
			'lifetime' => $asset->getLifetime(),
			'description' => $asset->getDescription(),

			'translations' => array_map(function (Language $language) use ($asset) {
				return [
					'lifetime' => $asset->getTranslatedLifetime($language, false),
					'description' => $asset->getTranslatedDescription($language, false),
				];
			}, $languages),
		];
	}
}
