<?php

namespace App\Controller\Hosting\Api;

use App\Component\Api\Exception\BadRequestException;
use App\Component\Color;
use App\Component\LoggerInterface;
use App\Controller\HostingApiController;
use App\Exception\FilesizeLimitExceededException;
use App\Exception\MimeTypeNotSupportedException;
use App\Exception\TempFileNotWritableException;
use App\Model\BaseLayout;
use App\Model\Theme;
use App\Model\ThemeBase;
use App\Utils;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use UnexpectedValueException;

/**
 * @Route("/hosting/api/client/{userId}/domain/{domainId}/theme", name="app_hosting_api_theme_")
 */
class ThemeApi extends HostingApiController
{
	private const THEME_SCHEMA = [
		'name' => 'string',
		'baseLayout' => 'string',
		'default' => 'bool',

		'widgetBlocking' => 'bool',
		'widgetButtons' => 'string[]',
		'widgetCloseButton' => 'bool',
		'widgetPosition' => 'string',
		'widgetHeightMobile' => 'int',
		'widgetSwitchType' => 'string',
		'showPurposesInMainWindow' => 'bool',

		'settingsShowEmbeddings' => 'bool',
		'settingsAcceptAllButton' => 'bool',
		'settingsWithdrawConsentButton' => 'bool',
		'settingsCheckAndUncheckAllButtons' => 'bool',
		'settingsFooter' => 'bool',

		'declineButtonRestrictions' => 'string[]',
		'detailsSaveButton' => 'bool',

		'complyWithDoNotTrack' => 'bool',
		'enableOnlyInEu' => 'bool',
		'manipulationPrevention' => 'bool',

		'logoEnabled' => 'bool',
		'logoImage' => '?string',

		'settingsIconEnabled' => 'bool',
		'settingsIconImage' => '?string',
		'settingsIconTarget' => 'string',

		'settingsIconBackgroundColor' => 'string',
		'settingsIconForegroundColor' => 'string',

		'whitelabel' => 'bool',

		'primaryButtonBackgroundColor' => 'string',
		'primaryButtonForegroundColor' => 'string',
		'primaryButtonBorderColor' => 'string',
		'declineButtonBackgroundColor' => 'string',
		'declineButtonForegroundColor' => 'string',
		'declineButtonBorderColor' => 'string',
		'buttonBackgroundColor' => 'string',
		'buttonForegroundColor' => 'string',
		'buttonBorderColor' => 'string',

		'windowBackgroundColor' => 'string',
		'windowForegroundColor' => 'string',
		'windowLinkColor' => 'string',
		'windowBorderColor' => 'string',

		'iframeBlockerBackgroundColor' => 'string',
		'iframeBlockerForegroundColor' => 'string',
		'iframeBlockerButtonBackgroundColor' => 'string',
		'iframeBlockerButtonForegroundColor' => 'string',

		'customCss' => 'string',
		'customCssTemplated' => 'bool',
		'customCssForIframeBlocker' => 'string',
		'customCssForIframeBlockerTemplated' => 'bool',
	];

	/**
	 * @Route("", name="create", methods={"POST"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function createTheme(string $userId, string $domainId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		// Create the new theme
		$theme = Theme::create();

		try {
			// Try to process the payload and update relevant properties
			self::migrateThemePayload($theme, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Delete the theme if any error occurred (rollback)
			Theme::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}
		catch (TempFileNotWritableException $exception) {
			// Delete the theme if any error occurred (rollback)
			Theme::reload();

			return $this->json([
				'error' => $exception->getMessage(),
			], 500);
		}
		catch (LogicException $exception) {
			// Delete the theme if any error occurred (rollback)
			Theme::reload();

			// Dev time, let it rip
			throw $exception;
		}

		$theme->save();

		$logger->logInfo('Theme-API', "Theme created: {$theme->getName()} ({$theme->getId()}) [$userId::$domainId]");

		return $this->json(self::themeToSchemaArray($theme), 201);
	}

	/**
	 * @Route("", name="read_all", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @return JsonResponse
	 */
	public function readAllThemes(string $userId, string $domainId): JsonResponse
	{
		try {
			self::assertDomainExists($userId, $domainId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		return $this->json(array_values(array_map(function (Theme $theme) {
			return self::themeToSchemaArray($theme);
		}, Theme::all())), 200);
	}

	/**
	 * @Route("/{themeId}", name="read", methods={"GET"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $themeId
	 * @return JsonResponse
	 */
	public function readTheme(string $userId, string $domainId, string $themeId): JsonResponse
	{
		try {
			self::assertThemeExists($userId, $domainId, $themeId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Theme $theme */
		$theme = Theme::find($themeId);
		assert($theme instanceof Theme);

		return $this->json(self::themeToSchemaArray($theme), 200);
	}

	/**
	 * @Route("/{themeId}", name="update", methods={"PUT"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $themeId
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return JsonResponse
	 */
	public function updateTheme(string $userId, string $domainId, string $themeId, Request $request, LoggerInterface $logger): JsonResponse
	{
		try {
			$payload = self::decodeRequestPayload($request);
			self::assertThemeExists($userId, $domainId, $themeId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Theme $theme */
		$theme = Theme::find($themeId);
		assert($theme instanceof Theme);

		try {
			// Try to process the payload and update relevant properties
			self::migrateThemePayload($theme, $payload);
		}
		catch (InvalidArgumentException $exception) {
			// Revert changes if an error has occurred (rollback)
			Theme::reload();

			return $this->json([
				'code' => $exception->getCode(),
				'error' => $exception->getMessage(),
			], 400);
		}

		$theme->save();

		$logger->logInfo('Theme-API', "Theme updated: {$theme->getName()} ({$theme->getId()}) [$userId::$domainId]");

		return $this->json(self::themeToSchemaArray($theme), 200);
	}

	/**
	 * @Route("/{themeId}", name="delete", methods={"DELETE"})
	 * @param string $userId
	 * @param string $domainId
	 * @param string $themeId
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteTheme(string $userId, string $domainId, string $themeId, LoggerInterface $logger): Response
	{
		try {
			self::assertThemeExists($userId, $domainId, $themeId);
		}
		catch (BadRequestException $exception) {
			return $exception->getResponse();
		}

		/** @var Theme $theme */
		$theme = Theme::find($themeId);
		assert($theme instanceof Theme);

		$themes = Theme::all();
		$domainSpecificThemeCount = array_reduce($themes, function ($count, Theme $theme) {
			return ($theme->getSystemTheme() || $theme->getUserTheme()) ? $count : $count + 1;
		}, 0);

		if ($theme->getSystemTheme() || $theme->getUserTheme()) {
			return $this->json([
				'code' => 1,
				'error' => 'System and account themes cannot be deleted.',
			], 400);
		}
		elseif ($domainSpecificThemeCount == 1) {
			return $this->json([
				'code' => 2,
				'error' => 'You may not delete the last theme.',
			], 400);
		}
		elseif ($theme->isDefaultTheme()) {
			$defaultThemeId = array_values(array_diff(array_keys($themes), [$themeId]))[0];
			$themes[$defaultThemeId]
				->touchLastModifiedTimestamp()
				->makeDefault();
		}

		$theme->setLogo(null);
		$theme->setSettingsIcon(null);

		$logger->logInfo('Theme-API', "Theme deleted: {$theme->getName()} ({$theme->getId()}) [$userId::$domainId]");
		Theme::delete($theme->getId());

		return new Response('', 204);
	}

	/**
	 * @param Theme $theme
	 * @param array $payload
	 * @return void
	 * @throws InvalidArgumentException if the payload is invalid in any way.
	 */
	private static function migrateThemePayload(Theme $theme, array $payload): void
	{
		self::checkPayloadDataTypes($payload, self::THEME_SCHEMA);

		$theme->touchLastModifiedTimestamp();

		if (array_key_exists('default', $payload)) {
			if ($payload['default']) {
				$theme->makeDefault();
			}
			else {
				throw new InvalidArgumentException('Make another theme the default by setting its default property to `true`.', 2);
			}
		}

		if (array_key_exists('logoEnabled', $payload)) {
			$theme->setLogoDisplayed($payload['logoEnabled']);
		}
		if (array_key_exists('logoImage', $payload)) {
			if ($payload['logoImage'] === null) {
				$theme->setLogo(null);
			}
			else {
				$mimeTypes = [
					'image/jpeg',
					'image/png',
				];

				try {
					$file = Utils::decodeBase64Image($payload['logoImage'], 524288, $mimeTypes);
				} catch (FilesizeLimitExceededException $exception) {
					throw new InvalidArgumentException('Invalid logoImage, maximum image upload size is 512 KiB.', 5);
				} catch (MimeTypeNotSupportedException $exception) {
					throw new InvalidArgumentException('Invalid logoImage, valid image formats: '.implode(', ', $mimeTypes), 6);
				} catch (InvalidArgumentException $exception) {
					throw new InvalidArgumentException('Invalid logoImage, expecting base64 encoded string.', 7);
				}

				$theme->setLogo($file);

				$tempFilename = $file->getPathname();
				unset($file);
				unlink($tempFilename);
			}
		}

		if (array_key_exists('whitelabel', $payload)) {
			if (Utils::hasWhitelabelLicense()) {
				$theme->setWhitelabel($payload['whitelabel']);
			}
			else {
				throw new InvalidArgumentException('Whitelabel is inactive for this domain.', 9);
			}
		}

		// Restrict access to further properties of system and account themes
		if ($theme->getSystemTheme() || $theme->getUserTheme()) {
			return;
		}

		if (array_key_exists('name', $payload) || empty($theme->getName())) {
			if ($payload['name'] ?? '') {
				$theme->setName($payload['name']);
			}
			else {
				throw new InvalidArgumentException('Empty or missing name.', 1);
			}
		}
		if (array_key_exists('baseLayout', $payload)) {
			$theme->setBaseLayout(BaseLayout::find($payload['baseLayout']) ?? BaseLayout::default());
		}

		if (array_key_exists('widgetBlocking', $payload)) {
			$theme->setBlocking($payload['widgetBlocking']);
		}
		if (array_key_exists('widgetButtons', $payload)) {
			array_walk($payload['widgetButtons'], function ($button) {
				if (in_array($button, ThemeBase::ALLOWED_WIDGET_BUTTON_VALUES) == false) {
					throw new InvalidArgumentException(
						'Invalid widgetButtons, allowed values: '.implode(', ', ThemeBase::ALLOWED_WIDGET_BUTTON_VALUES),
						3
					);
				}
			});

			$theme->setWidgetButtons($payload['widgetButtons']);
		}
		if (array_key_exists('widgetCloseButton', $payload)) {
			$theme->setCloseButton($payload['widgetCloseButton']);
		}
		if (array_key_exists('widgetPosition', $payload)) {
			if (in_array($payload['widgetPosition'] ?? '', ThemeBase::ALLOWED_WIDGET_POSITION_VALUES)) {
				$theme->setPosition($payload['widgetPosition']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid or missing widgetPosition, allowed values: '.implode(', ', ThemeBase::ALLOWED_WIDGET_POSITION_VALUES),
					4
				);
			}
		}
		elseif (empty($theme->getPosition())) {
			$theme->setPosition(ThemeBase::WIDGET_POSITION_CENTER);
		}

		if (array_key_exists('widgetHeightMobile', $payload)) {
			$theme->setBannerHeightMobile(($payload['widgetHeightMobile']));
		}

		if (array_key_exists('widgetSwitchType', $payload)) {
			$theme->setSelectType($payload['widgetSwitchType']);
		}
		if (array_key_exists('showPurposesInMainWindow', $payload)) {
			$theme->setShowPurposesInMainWindow($payload['showPurposesInMainWindow']);
		}

		if (array_key_exists('settingsShowEmbeddings', $payload)) {
			$theme->setShowEmbeddingsInControlPanel($payload['settingsShowEmbeddings']);
		}
		if (array_key_exists('settingsAcceptAllButton', $payload)) {
			$theme->setShowAcceptAllButtonInControlPanel($payload['settingsAcceptAllButton']);
		}
		if (array_key_exists('settingsWithdrawConsentButton', $payload)) {
			$theme->setShowWithdrawAllButtonInControlPanel($payload['settingsWithdrawConsentButton']);
		}
		if (array_key_exists('settingsCheckAndUncheckAllButtons', $payload)) {
			$theme->setShowCheckAndUncheckAllButtonInControlPanel($payload['settingsCheckAndUncheckAllButtons']);
		}
		if (array_key_exists('settingsFooter', $payload)) {
			$theme->setShowExtraImprintFooter($payload['settingsFooter']);
		}

		if (array_key_exists('declineButtonRestrictions', $payload)) {
			$theme->setDeclineButtonRestrictions($payload['declineButtonRestrictions']);
		}
		if (array_key_exists('detailsSaveButton', $payload)) {
			$theme->setSaveButtonInDetailsWindow($payload['detailsSaveButton']);
		}

		if (array_key_exists('complyWithDoNotTrack', $payload)) {
			$theme->setComplyingDoNotTrack($payload['complyWithDoNotTrack']);
		}
		if (array_key_exists('enableOnlyInEu', $payload)) {
			$theme->setOnlyInEu($payload['enableOnlyInEu']);
		}
		if (array_key_exists('manipulationPrevention', $payload)) {
			$theme->setManipulationPrevention($payload['manipulationPrevention']);
		}

		if (array_key_exists('settingsIconEnabled', $payload)) {
			$theme->setSettingsIconEnabled($payload['settingsIconEnabled']);
		}
		if (array_key_exists('settingsIconImage', $payload)) {
			if ($payload['settingsIconImage'] === null) {
				$theme->setSettingsIcon(null);
			}
			else {
				$mimeTypes = [
					'image/jpeg',
					'image/png',
				];

				try {
					$file = Utils::decodeBase64Image($payload['settingsIconImage'], 65536, $mimeTypes);
				} catch (FilesizeLimitExceededException $exception) {
					throw new InvalidArgumentException('Invalid settingsIconImage, maximum image upload size is 64 KiB.', 5);
				} catch (MimeTypeNotSupportedException $exception) {
					throw new InvalidArgumentException('Invalid settingsIconImage, valid image formats: '.implode(', ', $mimeTypes), 6);
				} catch (InvalidArgumentException $exception) {
					throw new InvalidArgumentException('Invalid settingsIconImage, expecting base64 encoded string.', 7);
				}

				$theme->setSettingsIcon($file);

				$tempFilename = $file->getPathname();
				unset($file);
				unlink($tempFilename);
			}
		}
		if (array_key_exists('settingsIconTarget', $payload)) {
			if (in_array($payload['settingsIconTarget'], ThemeBase::ALLOWED_SETTINGS_ICON_TARGET_VALUES)) {
				$theme->setSettingsIconTarget($payload['settingsIconTarget']);
			}
			else {
				throw new InvalidArgumentException(
					'Invalid settingsIconTarget, allowed values: '.implode(', ', ThemeBase::ALLOWED_SETTINGS_ICON_TARGET_VALUES),
					8
				);
			}
		}

		$colorProperties = [
			'settingsIconBackgroundColor' => [$theme, 'setSettingsIconBackgroundColor'],
			'settingsIconForegroundColor' => [$theme, 'setSettingsIconForegroundColor'],
			'primaryButtonBackgroundColor' => [$theme, 'setOkButtonBackgroundColor'],
			'primaryButtonForegroundColor' => [$theme, 'setOkButtonColor'],
			'primaryButtonBorderColor' => [$theme, 'setOkButtonBorderColor'],
			'declineButtonBackgroundColor' => [$theme, 'setDeclineButtonBackgroundColor'],
			'declineButtonForegroundColor' => [$theme, 'setDeclineButtonTextColor'],
			'declineButtonBorderColor' => [$theme, 'setDeclineButtonBorderColor'],
			'buttonBackgroundColor' => [$theme, 'setMoreInfoButtonBackgroundColor'],
			'buttonForegroundColor' => [$theme, 'setMoreInfoButtonColor'],
			'buttonBorderColor' => [$theme, 'setMoreInfoButtonBorderColor'],
			'windowBackgroundColor' => [$theme, 'setWindowBackgroundColor'],
			'windowForegroundColor' => [$theme, 'setInfotextColor'],
			'windowLinkColor' => [$theme, 'setLinkInInfotextColor'],
			'windowBorderColor' => [$theme, 'setWindowBorderColor'],
			'iframeBlockerBackgroundColor' => [$theme, 'setIframeBlockerBackgroundColor'],
			'iframeBlockerForegroundColor' => [$theme, 'setIframeBlockerForegroundColor'],
			'iframeBlockerButtonBackgroundColor' => [$theme, 'setIframeBlockerButtonBackgroundColor'],
			'iframeBlockerButtonForegroundColor' => [$theme, 'setIframeBlockerButtonForegroundColor'],
		];
		foreach ($colorProperties as $property => $setter) {
			$getter = [$theme, substr_replace($setter[1], 'g', 0, 1)];
			if (array_key_exists($property, $payload) || (is_callable($getter) && empty($getter()))) {
				try {
					$setter((string)Color::fromCssString($payload[$property] ?? '')->setAlpha(1));
				} catch (UnexpectedValueException $exception) {
					throw new InvalidArgumentException("Invalid or missing $property, use format \"#rrggbb\".", 10);
				}
			}
		}

		if (array_key_exists('customCss', $payload)) {
			$theme->setCustomCss($payload['customCss']);
		}
		if (array_key_exists('customCssTemplated', $payload)) {
			$theme->setCustomCssTemplated($payload['customCssTemplated']);
		}
		if (array_key_exists('customCssForIframeBlocker', $payload)) {
			$theme->setCustomCssForIframeBlocker($payload['customCssForIframeBlocker']);
		}
		if (array_key_exists('customCssForIframeBlockerTemplated', $payload)) {
			$theme->setCustomCssForIframeBlockerTemplated($payload['customCssForIframeBlockerTemplated']);
		}
	}

	/**
	 * @param Theme $theme
	 * @return array
	 */
	private static function themeToSchemaArray(Theme $theme): array
	{
		if ($theme->getSystemTheme() || $theme->getUserTheme()) {
			$result = [
				'id' => $theme->getId(),
				'systemTheme' => $theme->getSystemTheme() ? $theme->getSystemTheme()->getId() : null,
				'userTheme' => $theme->getUserTheme() ? $theme->getUserTheme()->getId() : null,
				'name' => $theme->getName(),
				'default' => $theme->isDefaultTheme(),

				'logoEnabled' => $theme->isLogoDisplayed(),
				'logoImage' => $theme->getLogo() ? '[binary]' : null,

				'whitelabel' => $theme->getWhitelabel(),
			];

			if ($result['systemTheme']) {
				unset($result['userTheme']);
			}
			else {
				unset($result['systemTheme']);
			}

			return $result;
		}

		// Send empty array if all regions are present
		$declineButtonRestrictions = $theme->getDeclineButtonRestrictions();
		if (empty(array_diff(ThemeBase::allRestrictRegions(), $declineButtonRestrictions))) {
			$declineButtonRestrictions = [];
		}

		return [
			'id' => $theme->getId(),
			'name' => $theme->getName(),
			'baseLayout' => $theme->getBaseLayout()->getId(),
			'default' => $theme->isDefaultTheme(),

			'widgetBlocking' => $theme->isBlocking(),
			'widgetButtons' => $theme->getWidgetButtons(),
			'widgetCloseButton' => $theme->hasCloseButton(),
			'widgetPosition' => $theme->getPosition(),
			'widgetHeightMobile' => $theme->getBannerHeightMobile(),
			'widgetSwitchType' => $theme->getSelectType(),
			'showPurposesInMainWindow' => $theme->isShowPurposesInMainWindow(),

			'settingsShowEmbeddings' => $theme->isShowEmbeddingsInControlPanel(),
			'settingsAcceptAllButton' => $theme->isShowAcceptAllButtonInControlPanel(),
			'settingsWithdrawConsentButton' => $theme->isShowWithdrawAllButtonInControlPanel(),
			'settingsCheckAndUncheckAllButtons' => $theme->isShowCheckAndUncheckAllButtonInControlPanel(),
			'settingsFooter' => $theme->isShowExtraImprintFooter(),

			'declineButtonRestrictions' => $declineButtonRestrictions,
			'detailsSaveButton' => $theme->hasSaveButtonInDetailsWindow(),

			'complyWithDoNotTrack' => $theme->isComplyingDoNotTrack(),
			'enableOnlyInEu' => $theme->getOnlyInEu(),
			'manipulationPrevention' => $theme->getManipulationPrevention(),

			'logoEnabled' => $theme->isLogoDisplayed(),
			'logoImage' => $theme->getLogo() ? '[binary]' : null,

			'settingsIconEnabled' => $theme->isSettingsIconEnabled(),
			'settingsIconImage' => $theme->getSettingsIcon() ? '[binary]' : null,
			'settingsIconTarget' => $theme->getSettingsIconTarget(),

			'settingsIconBackgroundColor' => $theme->getSettingsIconBackgroundColor(),
			'settingsIconForegroundColor' => $theme->getSettingsIconForegroundColor(),

			'whitelabel' => $theme->getWhitelabel(),

			'primaryButtonBackgroundColor' => $theme->getOkButtonBackgroundColor(),
			'primaryButtonForegroundColor' => $theme->getOkButtonColor(),
			'primaryButtonBorderColor' => $theme->getOkButtonBorderColor(),
			'declineButtonBackgroundColor' => $theme->getDeclineButtonBackgroundColor(),
			'declineButtonForegroundColor' => $theme->getDeclineButtonTextColor(),
			'declineButtonBorderColor' => $theme->getDeclineButtonBorderColor(),
			'buttonBackgroundColor' => $theme->getMoreInfoButtonBackgroundColor(),
			'buttonForegroundColor' => $theme->getMoreInfoButtonColor(),
			'buttonBorderColor' => $theme->getMoreInfoButtonBorderColor(),

			'windowBackgroundColor' => $theme->getWindowBackgroundColor(),
			'windowForegroundColor' => $theme->getInfotextColor(),
			'windowLinkColor' => $theme->getLinkInInfotextColor(),
			'windowBorderColor' => $theme->getWindowBorderColor(),

			'iframeBlockerBackgroundColor' => $theme->getIframeBlockerBackgroundColor(),
			'iframeBlockerForegroundColor' => $theme->getIframeBlockerForegroundColor(),
			'iframeBlockerButtonBackgroundColor' => $theme->getIframeBlockerButtonBackgroundColor(),
			'iframeBlockerButtonForegroundColor' => $theme->getIframeBlockerButtonForegroundColor(),

			'customCss' => $theme->getCustomCss(),
			'customCssTemplated' => $theme->isCustomCssTemplated(),
			'customCssForIframeBlocker' => $theme->getCustomCssForIframeBlocker(),
			'customCssForIframeBlockerTemplated' => $theme->isCustomCssForIframeBlockerTemplated(),
		];
	}
}
