<?php

namespace App\Controller\Hosting;

use App\Component\Api\HostingApi;
use App\Component\LocaleService;
use App\Component\Menu\Annotation\Menu;
use App\Component\Validator\DataTypeValidator;
use App\Controller\HostingController;
use App\Model\BaseLayout;
use App\Model\Purpose;
use App\Model\ThemeBase;
use App\Model\User;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/hosting/api/", name="app_hosting_api_")
 */
class ApiDocs extends HostingController
{
	/**
	 * @Menu("Settings", group="Hosting API", routeGroup="app_hosting_api_setup", icon="fa-cog", order=20)
	 * @Route("setup", name="setup")
	 * @param Request $request
	 * @return Response
	 */
	public function setup(Request $request): Response
	{
		return $this->render('hosting/api/setup.html.twig', [
			'api' => HostingApi::getInstance(),
		]);
	}

	/**
	 * @Menu("Endpoints", group="Hosting API", routeGroup="app_hosting_api_endpoints", icon="fa-list-ul", order=21)
	 * @Route("endpoints", name="endpoints")
	 * @return Response
	 */
	public function endpoints(): Response
	{
		return $this->render('hosting/api/endpoints.html.twig');
	}

	/**
	 * @Menu("Components", group="Hosting API", routeGroup="app_hosting_api_components", icon="fa-sitemap", order=22)
	 * @Route("components", name="components")
	 * @param Request $request
	 * @return Response
	 */
	public function components(Request $request): Response
	{
		$permissions = User::buildDefaultSettings();
		uksort($permissions, 'strnatcasecmp');
		$permissions = array_keys($permissions);

		return $this->render('hosting/api/components.html.twig', [
			'permissions' => $permissions,
			'purposes' => Purpose::all(),
			'locale' => LocaleService::findFuzzy($request->getLocale()) ?? LocaleService::find('en_US'),
			'userLocales' => LocaleService::all(),
			'baseLayouts' => BaseLayout::all(),
			'defaultBaseLayout' => BaseLayout::default(),
			'restrictRegions' => ThemeBase::allRestrictRegions(),
		]);
	}

	/**
	 * @Route("update_key", name="update_key", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function updateApiKey(Request $request, TranslatorInterface $translator): Response
	{
		$apiKey = trim((string)$request->request->get('apiKey'));

		if (strlen($apiKey) >= HostingApi::API_KEY_LENGTH) {
			HostingApi::getInstance()->setApiKey($apiKey);
		}
		else {
			$this->addFlash('danger', $translator->trans('The API key must be at least :x characters long.', [
				':x' => HostingApi::API_KEY_LENGTH,
			]));
		}

		return $this->redirectToRoute('app_hosting_api_setup', [], 303);
	}

	/**
	 * @Route("update_hawk", name="update_hawk", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function updateHawkHttpAuthentication(Request $request, TranslatorInterface $translator): Response
	{
		$useHawk = (bool)$request->request->get('useHawk');
		$sharedSecret = trim((string)$request->request->get('sharedSecret'));

		$api = HostingApi::getInstance();

		$api->setUseHawk($useHawk);

		if (strlen($sharedSecret) >= HostingApi::SHARED_SECRET_LENGTH) {
			$api->setSharedSecret($sharedSecret);
		}
		else {
			$this->addFlash('danger', $translator->trans('The shared secret should be at least :x characters long.', [
				':x' => HostingApi::SHARED_SECRET_LENGTH,
			]));
		}

		return $this->redirectToRoute('app_hosting_api_setup', [], 303);
	}

	/**
	 * @Route("update_restrictions", name="update_restrictions", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function updateRestrictions(Request $request, TranslatorInterface $translator): Response
	{
		$api = HostingApi::getInstance();

		// Update item limit
		$itemLimit = (string)$request->request->get('itemLimit');
		if (strlen($itemLimit) > 0) {
			if (DataTypeValidator::checkStringForUInt($itemLimit)) {
				$api->setItemLimitRestriction($request->request->getInt('itemLimit'));
			} else {
				$this->addFlash('danger', $translator->trans('The item limit must be a positive number, leave blank to disable the limit.'));
			}
		} else {
			$api->setItemLimitRestriction(null);
		}

		return $this->redirectToRoute('app_hosting_api_setup', [], 303);
	}
}
