<?php

namespace App\Controller\Hosting;

use App\Component\Api\ClusterApi;
use App\Component\Cache\Cache;
use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Component\Plugin\Manager as PluginManager;
use App\Component\ThumbnailExtractor\ThumbnailExtractorService;
use App\Controller\HostingController;
use App\Event\AccountUpdateEvent;
use App\Exception\CsrfTokenException;
use App\Exception\UserAlreadyExistsException;
use App\Model\Country;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Index\DomainIndex;
use App\Model\Model;
use App\Model\User;
use App\Utils;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Menu("Clients", group="Management", icon="fa-users", route_group="app_hosting_client_", order=10)
 * @Route("/hosting/client", name="app_hosting_client_main_account_")
 */
class Client extends HostingController
{
	/**
	 * @Menu("Main accounts", route_group="app_hosting_client_main_account_", order=0)
	 * @Route("", methods={"HEAD", "GET"}, name="index")
	 * @param Request $request
	 * @return Response
	 */
	public function clientIndex(Request $request): Response
	{
		$clients = User::getClients();
		return $this->render('hosting/client/index.html.twig', [
			'clients' => $clients,
			'countries' => Country::all(),
			'searchQuery' => (string)$request->query->get('search'),
		]);
	}

	/**
	 * @Route("list.json", name="list_ajax")
	 * @param Request $request
	 * @return JsonResponse
	 */
	public function ajaxClientSearch(Request $request): JsonResponse
	{
		$clients = User::getClients();

		// Filterung der Kunden nach Suchstring (falls vorhanden)
		$search = (string)($request->query->all('search')['value'] ?? '');
		$clientsFiltered = strlen($search) > 0 ? User::searchInArray($search, $clients) : $clients;

		// Mapping der Kundendaten auf Tabellenspalten
		$clientsFiltered = array_values(array_filter(array_map(function ($client) {
			return [
				'id' => $client->getId(),
				'username' => $client->getUsername(),
				'active' => $client->isActive(),
				'agencyId' => $client->getAgencyId(),
				'admininfo' => $client->getAdminInfo(),
				'name' => $client->getFirstName() . ' ' . $client->getLastName(),
				'emailAddress' => $client->getEmailAddress(),
				'company' => $client->getCompany(),'firstName' => $client->getFirstName(),
				'lastName' => $client->getLastName(),
				'maxWLCount' => $client->getMaxWLCount(),
				'actualWLCount' => $client->getActualWLCount(),
				'maxDomainCount' => $client->getMaxDomainCount(),
				'actualDomainCount' => $client->getActualDomainCount(),
				'domlist' => $client->getDomList(),
				'maxCallCount' => $client->getMaxCallCount(),
				'actualCallCount' => $client->getActualCallCount(),
				'accountCreatedTimestamp' => $client->getCreateTimeStamp(),
				'lastLoginTimestamp' => $client->getLastLogin(),
			];
		}, $clientsFiltered)));

		// Sortierung der Kunden
		$order = $request->query->all('order')[0];
		$sortColumn = $order['column'] ?? '';
		$sortDirection = $order['dir'] ?? 'asc';

		$clientsFiltered = Utils::sortNatNum($clientsFiltered, $sortColumn, $sortDirection);

		// Paginierung der Einträge
		$start = $request->query->getInt('start');
		$length = $request->query->getInt('length') ?: 10;

		$pagination = new Pagination($clientsFiltered, $length);
		$currentPage = ($start / $length) + 1;
		$pagination->setCurrentPage($currentPage);

		$clientsPaginated = $pagination->getItems();

		return $this->json([
			'data' => $clientsPaginated,
			'draw' => $request->query->get('draw'),
			'recordsTotal' => count($clients),
			'recordsFiltered' => count($clientsFiltered),
		]);
	}

	/**
	 * @Route("_delete_list", methods={"POST"}, name="delete_list_post")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param ThumbnailExtractorService $thumbnails
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function clientDeleteListAll(Request $request, TranslatorInterface $translator, ThumbnailExtractorService $thumbnails, LoggerInterface $logger): Response
	{
		$delete_clients = $request->request->all('delete_clients');

		if (count($delete_clients) > 0) {
			// Repository zum Schreiben sperren
			User::beginWrite();

			$user = User::loggedInUser();
			assert($user instanceof User);

			foreach ($delete_clients as $clientId) {
				/** @var User|null $client */
				$client = User::byId($clientId);

				if ($client == null || $client->getRole() != User::ROLE_CLIENT || $client->isAccountOwner() == false) {
					continue;
				}
				elseif (!$user->mayAccessClient($client)) {
					continue;
				}

				// Subuser vorher löschen...
				foreach ($client->getSubusers() as $subuser) {
					// Repo nicht zwischenspeichern, um keine Performance zu verlieren
					User::delete($subuser->getId(), false);
				}

				// Caches des Users leeren
				$thumbnails->clearCache($client);
				Cache::clearUser($client);
				// Broadcast clear cache request to other peers of this cluster
				ClusterApi::getInstance()->callRemoteApiBroadcast('app_cluster_api_user_cache_clear', [
					'userId' => $client->getId(),
					'apiKey' => $client->getApiKey(),
				], 'DELETE');

				// Benutzer löschen, Repository noch nicht schreiben
				User::delete($client->getId(), false);
			}

			// Repository schreiben
			User::flush();

			// Liste der veränderten Models leeren, damit der CacheHandler nicht annehmen muss,
			// dass sich Daten anderer User geändert haben
			Model::clearWrittenModelsClassNames();

			$logger->logSuccess('ClientManagement', $translator->trans(':x users deleted.', [
				':x' => count($delete_clients),
			]));
			$this->addFlash('success', $translator->trans('Accounts have been terminated.'));
		}
		else {
			$this->addFlash('info', $translator->trans('Nothing to do.'));
		}

		return $this->redirectToRoute('app_hosting_client_main_account_delete_list', [], 303);
		//exit();
	}

	/**
	 * @Route("_delete_list", methods={"HEAD", "GET"}, name="delete_list")
	 * @param Request $request
	 * @return Response
	 */
	public function clientDeleteList(Request $request): Response
	{
		$clients = User::byRole(User::ROLE_CLIENT);

		//Search
		$usersearch = trim((string)$request->query->get("search", ''));
		if ($usersearch) {
			$clients = User::search($usersearch, User::ROLE_CLIENT);
		}

		//Subuser hier ausschließen... evtl. aktivieren wir das später mal wieder... aber das ist der Trigger...
		$clients = array_filter($clients, function (User $client) {
			//alle die keinen Callcount haben
			if($client->getActualCallCount()<1 && ($client->getLastLogin() < (time()-(60*60*24*60))) && $client->getMaxCallCount()<5001)
			{
				if($client->isAccountOwner())
				{
					return true;
				}
			}

			//Alle die keine Domain haben
			if($client->getDomList()){
				foreach($client->getDomList() as $dom)
				{
					if(stristr($dom,"none") && ($client->getLastLogin() < (time()-(60*60*24*60)))) //60 Tage
					{
						if($client->isAccountOwner())
						{
							return true;
						}
					}
				}
			}

			return false;
		});

		$pagination = new Pagination($clients,1000);
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_hosting_client_main_account_delete_list');
		}

		return $this->render('hosting/client/delete_list.html.twig', [
			'pagination' => $pagination,
			'usersearch' => $usersearch,
		]);
	}

	/**
	 * @Route("/new", name="new", methods={"GET","HEAD"})
	 * @param Request $request
	 * @param PluginManager $pluginManager
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function clientCreateForm(Request $request, PluginManager $pluginManager, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$formData = [
			'username' => '',
			'password' => '',
			'emailAddress' => '',
			'firstName' => '',
			'lastName' => '',
			'company' => '',
			'country' => 'DE',
			'useractive' =>'1',
			'maxDomainCount' => '10',
			'maxWLCount' => '0',
			'maxCallCount' => '-1',
			'autoAddDomains' => false,
			'clientGroups' => [],
		];

		$newUserSettings = User::buildDefaultSettings(true);

		$formData['settings'] = $newUserSettings;

		return $this->render('hosting/client/new.html.twig', [
			'formData' => $formData,
			'clientGroups' => $user->getAccessibleClientGroups(),
			'countries' => Country::all(),
			'plugins' => $pluginManager->configuredPlugins(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/new", name="new_post", methods={"POST"})
	 * @param Request $request
	 * @param PluginManager $pluginManager
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 * @throws UserAlreadyExistsException
	 */
	public function clientCreate(Request $request, PluginManager $pluginManager, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$formData = [
			'username' => '',
			'password' => '',
			'emailAddress' => '',
			'firstName' => '',
			'lastName' => '',
			'company' => '',
			'useractive' =>'1',
			'maxDomainCount' => '10',
			'maxWLCount' => '0',
			'maxCallCount' => '1000',
			'autoAddDomains' => false,
			'clientGroups' => [],
			'settings' => [],
			'passwordChangeable' => true,
			'noPassword' => false,
			'invoiceEmailAddress' => null,
			'dunningEmailAddress' => null,
		];

		// Alle Formulardaten übernehmen, damit sie bei fehlerhaften Eingaben im Template bereitstehen
		$formData = array_map(function ($value) {
			return is_array($value) ? $value : trim((string)$value);
		}, $request->request->all()) + $formData;

		$newUserSettings = [];

		foreach ((array)$request->request->get("userSettings") as $fieldName => $value) {
			$newUserSettings[$fieldName] = !empty($value);
			if(($fieldName === "maxCrawlBudget" || $fieldName === "externalCrawlerFrequency") && is_numeric($value))
			{
				$newUserSettings[$fieldName] = (int)$value;
			}
		}
		// Versteckte Plugins (payment) immer auf aktiv setzen
		foreach ($pluginManager->configuredPlugins() as $plugin) {
			if ($plugin->isHidden()) {
				$newUserSettings[$plugin->getUserRightsKey()] = true;
			}
		}

		$formData['settings'] = $newUserSettings;

		try {
			CsrfToken::get()->checkRequest($request);

			User::beginWrite();

			if (User::exists($formData['username'] ?? '')) {
				$this->addFlash('danger', $translator->trans('Username %username% already exists.', [
					'%username%' => $formData['username'],
				]));
			}
			elseif (Utils::validateEmailAddress($formData['emailAddress'] ?? '') == false) {
				$this->addFlash('danger', $translator->trans('Invalid email address, check format.'));
			}
			elseif (empty($formData['username']) or (empty($formData['password']) and !$formData['noPassword'])) {
				$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
			}
			else {
				$client = User::create($formData['username'])
					->setEmailAddress($formData['emailAddress'])
					->setFirstName($formData['firstName'])
					->setLastName($formData['lastName'])
					->setCompany($formData['company'])
					->setMaxDomainCount((int)$formData['maxDomainCount'])
					->setRole(User::ROLE_CLIENT)
					->setMaxWLCount((int)$formData['maxWLCount'])
					->setMaxCallCount((int)$formData['maxCallCount'])
					->setAutoAddDomains((bool)$formData['autoAddDomains'])
					->setActive((bool)$formData['useractive'])
					->setPasswordLocked(!(bool)$formData['passwordChangeable'])
					->setPlz((string)($request->request->get("plz") ?? ""))
					->setCity((string)($request->request->get("city") ?? ""))
					->setStrNr((string)($request->request->get("streetNr") ?? ""))
					->setCountry(Country::find($request->request->getAlpha('country')))
					->setVatId((string)($request->request->get("vatId") ?? ""))
					->setAgencyId((string)($request->request->get("agencyId") ?? ""))
					->setAdminInfo((string)$request->request->get('admininfo'))
					->setSettings((array)$newUserSettings);

				if ($formData['invoiceEmailAddress']) {
					$client->setInvoiceEmailAddress(trim((string)$formData['invoiceEmailAddress']));
				}
				if ($formData['dunningEmailAddress']) {
					$client->setDunningEmailAddress(trim((string)$formData['dunningEmailAddress']));
				}

				if ($formData['noPassword']) {
					$client->unsetPassword();
				} else {
					$client->setPassword($formData['password']);
				}

				$client->setClientGroups(array_intersect_key(
					$user->getAccessibleClientGroups(),
					array_flip($request->request->all('clientGroups'))
				));

				$client->save();

				$this->addFlash('success', $translator->trans('Client %username% has been created!', [
					'%username%' => $formData['username'],
				]));

				switch ($request->request->getAlpha('action')) {
					case 'saveClose':
						return $this->redirectToRoute('app_hosting_client_main_account_index');
					case 'saveLogin':
						return $this->clientLogin($client->getId(), $translator, $logger);
					case 'addClient':
					default:
						return $this->redirectToRoute('app_hosting_client_main_account_edit_form', ['id'=>$client->getId()], 303);
				}
			}
		} catch (CsrfTokenException $e) {
			$error = $translator->trans('CSRF token is missing or invalid. Please try again.');
			$this->addFlash('danger', $error);
		}

		return $this->render('hosting/client/new.html.twig', [
			'formData' => $formData,
			'clientGroups' => $user->getAccessibleClientGroups(),
			'countries' => Country::all(),
			'plugins' => $pluginManager->configuredPlugins(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/from_csv", name="from_csv", methods={"GET","HEAD"})
	 * @param Request $request
	 * @param PluginManager $pluginManager
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function clientQuickCreateFromCsvForm(Request $request, PluginManager $pluginManager, TranslatorInterface $translator): Response
	{
		return $this->render('hosting/client/from_csv.html.twig', [
			'countries' => Country::all(),
		]);
	}

	/**
	 * @Route("/from_csv", name="from_csv_post", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function clientQuickCreateFromCsv(Request $request, TranslatorInterface $translator): Response
	{
		/** @var UploadedFile|null $file */
		$file = $request->files->get('csv_file');

		if (!$file) {
			$this->addFlash('danger', $translator->trans('No file uploaded.'));
			return $this->redirectToRoute('app_hosting_client_main_account_from_csv', [], 303);
		}
		elseif ($file->getError() != UPLOAD_ERR_OK) {
			$this->addFlash('danger', $translator->trans('An upload error occurred: (:code) :error', [
				':code' => $file->getError(),
				':error' => $file->getErrorMessage(),
			]));
			return $this->redirectToRoute('app_hosting_client_main_account_from_csv', [], 303);
		}

		$file = $file->openFile();

		// Read column names from uploaded file
		$columnNames = $file->fgetcsv(',', '"', '\\') ?: [];
		$mandatoryColumns = [
			'username',
		];

		// Abort if the file contains no columns or if mandatory columns have been omitted
		if (empty($columnNames) || array_diff($mandatoryColumns, $columnNames)) {
			$this->addFlash('danger', $translator->trans('Please provide all necessary columns: :cols', [
				':cols' => implode(', ', $mandatoryColumns),
			]));
			return $this->redirectToRoute('app_hosting_client_main_account_from_csv', [], 303);
		}

		// Read remaining rows from uploaded file
		$rows = [];
		while (is_array($row = $file->fgetcsv(',', '"', '\\'))) {
			// Silently ignore rows whose number of columns differs
			if (count($row) != count($columnNames)) {
				continue;
			}

			$row = array_map('trim', $row);
			$rows[] = array_combine($columnNames, $row);
		}

		// Close handler to uploaded file
		unset($file);

		// Validate input data
		if ($error = $this->validateCsvForClients($rows, $translator)) {
			return $error;
		}

		// Remember current user
		$user = User::loggedInUser();
		assert($user instanceof User);

		// Create or update clients
		$this->createOrUpdateClientsFromCsv($rows, $translator);

		// Switch back to the original user context
		Domain::setDomainForRequest('');
		User::setApiUser($user->getApiKey());
		Model::closeUserModels();

		return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
	}

	/**
	 * @Route("/example.csv", name="csv_example_file", methods={"GET","HEAD"})
	 * @return Response
	 */
	public function clientQuickCreateCsvExampleFile(): Response
	{
		$data = [[
			'username' => 'john_doe',
			'password' => '',
			'firstName' => 'John',
			'lastName' => 'Doe',
			'emailAddress' => 'john.doe@example.domain',
			'street' => 'Hauptstraße 1',
			'zipCode' => '12345',
			'city' => 'Netzstadt',
			'country' => 'DE',
			'company' => 'ACME Corp.',
			'vatId' => 'DE999999999',
			'maxCallCount' => 10000,
			'maxDomainCount' => 5,
			'maxWhitelabelCount' => 0,
			'domainUrls' => implode('|', [
				'https://a.domain/',
				'https://b.domain/path/to/website/2/',
				'https://b.domain/path/to/website/3/',
			]),
		]];

		$handle = fopen('php://temp', 'w');

		// Write column names
		fputcsv($handle, array_keys($data[0]), ',', '"', '\\');

		foreach ($data as $row) {
			// Skip rows that have a different number of columns
			if (count($row) != count($data[0])) {
				continue;
			}

			// Write row to the dataset
			fputcsv($handle, $row, ',', '"', '\\');
		}

		rewind($handle);
		$csv = stream_get_contents($handle);
		fclose($handle);

		return new Response($csv, 200, [
			'Content-Type' => 'text/csv',
			'Content-Disposition' => 'attachment; filename="example.csv"',
		]);
	}

	/**
	 * @Route("/{id}", methods={"HEAD", "GET"}, name="edit_form")
	 * @param string $id
	 * @param Request $request
	 * @param PluginManager $pluginManager
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function clientEditForm(string $id, Request $request, PluginManager $pluginManager, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		/** @var User|null $client */
		$client = User::byId($id);
        //print_r($client);

        if ($client == null || $client->getRole() != User::ROLE_CLIENT) {
            return $this->redirectToRoute('app_hosting_client_main_account_index');
        }
        if (!$user->mayAccessClient($client)) {
            return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
        }
        $trialPeriods = $client->getTrialPeriods();
        //print_r($user);
        $hasAccess = array();
        $hasAccessTest = array();
        foreach($pluginManager->configuredPlugins() as $plugin)
        {
            $state = $plugin->getState();
            if($state=== null){
                continue;
            }
            //Besteht Zugriff auf das Plugin durch eine TarifVoragbe seitens des Plugins aus dem Plugin Manager?
            if($state->checkFullAccessPerCount($client)){
                $hasAccess[$state->getPluginUID()] = true;
                continue;
            }

            //Besteht ein Zugriff durch einen Testzeitraum? Und wie lange ist der noch in Tagen?
            if($state->checkTestAccessPerCount($client)){
				//Falls noch leer, dann auf Testzeitraum setzen
				if(empty($trialPeriods[$state->getPluginUID()]))
				{
					$trialPeriods[$state->getPluginUID()] = ($state->getTestAccessTime()*86400)+time();
				}
                $hasAccessTest[$state->getPluginUID()]['access'] = true;
                $hasAccessTest[$state->getPluginUID()]['days'] = round(((max(0, $trialPeriods[$state->getPluginUID()] - time()))/86400));
                $hasAccessTest[$state->getPluginUID()]['testtime'] = $state->getTestAccessTime();
            }
        }

		return $this->render('hosting/client/edit.html.twig', [
			'user' => User::loggedInUser(),
			'plugins' => $pluginManager->configuredPlugins(),
			'client' => $client,
			'clientGroups' => $user->getAccessibleClientGroups(),
			'countries' => Country::all(),
			'csrfToken' => CsrfToken::get(),
            'hasAccess' => $hasAccess,
            'hasAccessTest' => $hasAccessTest,
		]);
	}

	/**
	 * @Route("/{id}", methods={"PATCH"}, name="edit")
	 * @param string $id
	 * @param Request $request
	 * @param PluginManager $pluginManager
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function clientEdit(string $id, Request $request, PluginManager $pluginManager, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		User::beginWrite();
		/** @var User|null $client */
		$client = User::byId($id);

		if ($client == null || $client->getRole() != User::ROLE_CLIENT) {
			return $this->redirectToRoute('app_hosting_client_main_account_index');
		}
		elseif (!$user->mayAccessClient($client)) {
			return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
		}

		$formData = [
			'username' => $client->getUsername(),
			'password' => '',
			'emailAddress' => '',
			'firstName' => '',
			'lastName' => '',
			'company' => '',
			'plz' => '',
			'city' => '',
			'streetNr' => '',
			'vatId' => '',
			'agencyId' => '',
			'admininfo' => '',
			'useractive' => false,
			'maxDomainCount' => '10',
			'maxWLCount' => '0',
			'maxCallCount' => '1000',
			'autoAddDomains' => false,
			'passwordChangeable' => false,
			'noPassword' => false,
			'invoiceEmailAddress' => null,
			'dunningEmailAddress' => null,
		];

		$newUserSettings = array(
			'moreThan2Lang' => false,
			'canUseAnalytics' => false,
			'canUseABTests' => false,
			'canUseScripts' => false,
			'canUseIframes' => false,
			'canUseConsentSharing' => false,
			'canUseOwnLogo' => false,
			'canUseIndCSS' => false,
			'canUseIframeCSS' => false,
			'canUseMultiDom' => false,
			'canUseIAB' => false,
			'canUseUserGroupRights' => false,
			'canUseDomAPI' => false,
			'canUseExternalCrawler' => false,
			'externalCrawlerFrequency' => 0,
			'maxCrawlBudget' => 1,
		);

		foreach (array_keys($formData) as $fieldName) {
			$formData[$fieldName] = trim((string)$request->request->get($fieldName));
		}

		$canUseExternalCrawler = false;
		foreach ((array)$request->request->get("userSettings") as $fieldName => $value) {
			$newUserSettings[$fieldName] = !empty($value);
			if(($fieldName === "maxCrawlBudget" || $fieldName === "externalCrawlerFrequency") && is_numeric($value))
			{
				$newUserSettings[$fieldName] = (int)$value;
			}
			if($fieldName === "canUseExternalCrawler" && $value == true)
			{
				$canUseExternalCrawler = true;
			}
		}
		// Versteckte Plugins (payment) immer auf aktiv setzen
		foreach ($pluginManager->configuredPlugins() as $plugin) {
			if ($plugin->isHidden()) {
				$newUserSettings[$plugin->getUserRightsKey()] = true;
			}
		}

		$ok = true;

		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$error = $translator->trans('CSRF token is missing or invalid. Please try again.');
			$this->addFlash('danger', $error);
			$ok = false;
		}

		$usernameChanged = $client->getUsername() !== $formData['username'];

		// Überprüfen, ob der Username bereits verwendet wird
		if ($usernameChanged and User::exists($formData['username'])) {
			$error = $translator->trans('This username is already in use. Please choose another name.');
			$this->addFlash('danger', $error);
			$ok = false;
		}

		// Daten temporär zuweisen, damit im Fehlerfall die Benutzereingaben im Formular ausgegeben werden
		$client
			->setUsername($formData['username'])
			->setEmailAddress($formData['emailAddress'])
			->setFirstName($formData['firstName'])
			->setLastName($formData['lastName'])
			->setCompany($formData['company'])
			->setMaxWLCount((int)$formData['maxWLCount'])
			->setMaxDomainCount((int)$formData['maxDomainCount'])
			->setMaxCallCount((int)$formData['maxCallCount'])
			->setAutoAddDomains((bool)$formData['autoAddDomains'])
			->setPasswordLocked(!(bool)$formData['passwordChangeable'])
			->setSettings((array)$newUserSettings)
			->setPlz((string)$formData['plz'])
			->setCity((string)$formData['city'])
			->setStrNr((string)$formData['streetNr'])
			->setCountry(Country::find($request->request->getAlpha('country')))
			->setVatId((string)$formData['vatId'])
			->setAgencyId((string)$formData['agencyId'])
			->setAdminInfo((string)$request->request->get('admininfo'))
			->setActive((bool)$formData['useractive'])
			->setInvoiceEmailAddress(trim((string)$formData['invoiceEmailAddress']))
			->setDunningEmailAddress(trim((string)$formData['dunningEmailAddress']))
		;


		//Wenn der User den Crawler aktivieren können soll, dann muss das natürlich auch direkt aktiv sein.
		if($canUseExternalCrawler)
		{
			$client
				->setCrawlAllDomains(true)
				->setAllowAutoActivationOfNewEmbeddings(true)
			;
		}


        //Trial Zeiträume der Plugins neue setzen
        if ($pluginTrialResets = $request->request->all('userSettings_resetTest')) {
            $trialPeriods = $client->getTrialPeriods();
            foreach ($pluginTrialResets as $pluginKey => $days) {
                $trialPeriods[$pluginKey] = $days * 86400 + time();
            }
            $client->setTrialPeriods($trialPeriods);
        }

		if ($usernameChanged and strlen(trim($formData['username'])) < 2) {
			$error = $translator->trans('The username is too short. Please choose another name.');
			$this->addFlash('danger', $error);
			$ok = false;
		}

		if (! Utils::validateEmailAddress($formData['emailAddress'])) {
			$this->addFlash('danger', $translator->trans('Invalid email address, check format.'));
			$ok = false;
		}

		$accessibleClientGroups = $user->getAccessibleClientGroups();
		$selectedClientGroupIds = $request->request->all('clientGroups');

		if ($user->hasAccessToAllClients()) {
			$client->setClientGroups(array_intersect_key($accessibleClientGroups, array_flip($selectedClientGroupIds)));
		}
		else {
			// Preserve assigned client groups that the current user is not allowed to manage
			$client->setClientGroups(array_merge(
				array_diff_key($client->getClientGroups(), $accessibleClientGroups), // preserve
				array_intersect_key($accessibleClientGroups, array_flip($selectedClientGroupIds)) // add
			));
		}

		if ($ok) {
			$event = new AccountUpdateEvent($client, AccountUpdateEvent::CAUSE_ADMIN);
			$event->setAccountInfoChanged();

			if (strlen($formData['password']) > 0) {
				$client->setPassword($formData['password']);
				$event->setPasswordChanged();
			}
			elseif ($formData['noPassword'] and $client->hasPassword()) {
				$client->unsetPassword();
			}
			Utils::dispatchEvent(AccountUpdateEvent::NAME_PRE_SAVE, $event);

			$client->save();

			$this->addFlash('success', $translator->trans('Client %username% has been edited!', [
				'%username%' => $client->getUsername(),
			]));

			Utils::dispatchEvent(AccountUpdateEvent::NAME_POST_SAVE, $event);

			switch ($request->request->getAlpha('action')) {
				case 'saveClose':
					return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
				case 'saveLogin':
					return $this->clientLogin($id, $translator, $logger);
				case 'save':
				default:
					return $this->redirectToRoute('app_hosting_client_main_account_edit_form', ['id'=>$id], 303);
			}
		}

		return $this->render('hosting/client/edit.html.twig', [
			'user' => User::loggedInUser(),
			'plugins' => $pluginManager->configuredPlugins(),
			'client' => $client,
			'clientGroups' => $user->getAccessibleClientGroups(),
			'countries' => Country::all(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}", methods={"POST"}, name="action")
	 * @param string $id
	 * @param Request $request
	 * @return JsonResponse|RedirectResponse
	 */
	public function clientAction(string $id, Request $request): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		User::beginWrite();
		$action = (string)$request->request->get('action');
		$isJson = $request->request->getAlnum('format', 'html') === 'json';
		/** @var User|null $client */
		$client = User::byId($id);

		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			if ($isJson) {
				return $this->json([ 'success'=>false, 'error'=>'CSRF token is missing or invalid.' ], 403);
			} else {
				$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
				return $this->redirectToRoute('app_hosting_client_main_account_edit_form', ['id'=>$id], 303);
			}
		}

		if (!$client or $client->getRole() != User::ROLE_CLIENT) {
			if ($isJson) {
				return $this->json([ 'success'=>false, 'error'=>'client not found' ], 404);
			} else {
				$this->addFlash('danger', 'Client not found.');
				return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
			}
		}
		elseif (!$user->mayAccessClient($client)) {
			return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
		}

		switch ($action) {
			case 'resetCallCount':
				$client->setCallCount(0);
				$client->save();
				if ($isJson) {
					return $this->json([ 'success'=>true, 'callCount'=>$client->getActualCallCount() ]);
				} else {
					$this->addFlash('success', 'Counter was successfully reset.');
					return $this->redirectToRoute('app_hosting_client_main_account_edit_form', ['id'=>$id], 303);
				}
			case 'clear2FA':
				$client->clearAllTwoFactorAuthParameters();
				$client->save();
				if ($isJson) {
					return $this->json([ 'success'=>true ]);
				} else {
					$this->addFlash('success', 'Two-factor authentication was successfully reset.');
					return $this->redirectToRoute('app_hosting_client_main_account_edit_form', ['id'=>$id], 303);
				}
			default:
				if ($isJson) {
					return $this->json([ 'success'=>false, 'error'=>'unknown action' ]);
				} else {
					return $this->redirectToRoute('app_hosting_client_main_account_edit_form', ['id'=>$id], 303);
				}
		}
	}

	/**
	 * @Route("/{id}/login", methods={"POST"}, name="login")
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function clientLogin(string $id, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		/** @var User|null $client */
		$client = User::byId($id);
		$currentUser = User::loggedInUser();

		if ($client == null || $client->getRole() != User::ROLE_CLIENT) {
			return $this->redirectToRoute('app_hosting_client_main_account_index');
		}
		elseif (!$currentUser->mayAccessClient($client)) {
			return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
		}

		$client->switchTo(true);

		$logger->logInfo("Login", "User ".$currentUser->getUsername()." logged into account ".$client->getUsername());

		$this->addFlash('success', $translator->trans('You are now in the account “%USER%”. You may switch back to the admin account using the account menu in the top right corner.', ['%USER%' => $client->getUsername()]));

		$response = $this->redirectToRoute('app_root', [], 303);
		$response->headers->set('Clear-Site-Data', '"cache"');
		return $response;
	}

	/**
	 * @Route("/{id}/delete", methods={"HEAD", "GET"}, name="delete_form")
	 * @param string $id
	 * @return Response
	 */
	public function clientDeleteForm(string $id): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		/** @var User|null $client */
		$client = User::byId($id);

		if ($client == null || $client->getRole() != User::ROLE_CLIENT) {
			return $this->redirectToRoute('app_hosting_client_main_account_index');
		}
		elseif (!$user->mayAccessClient($client)) {
			return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
		}

		return $this->render('hosting/client/delete.html.twig', [
			'user' => User::loggedInUser(),
			'client' => $client,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}", methods={"DELETE"}, name="delete")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param ThumbnailExtractorService $thumbnails
	 * @return Response
	 */
	public function clientDelete(string $id, Request $request, TranslatorInterface $translator, ThumbnailExtractorService $thumbnails): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$error = $translator->trans('CSRF token is missing or invalid. Please try again.');
			$this->addFlash('danger', $error);
			return $this->redirectToRoute('app_hosting_client_main_account_delete_form', ['id'=>$id], 303);
		}

		User::beginWrite();
		/** @var User|null $client */
		$client = User::byId($id);

		if ($client == null || $client->getRole() != User::ROLE_CLIENT || $client->isAccountOwner() == false) {
			return $this->redirectToRoute('app_hosting_client_main_account_index');
		}
		elseif (!$user->mayAccessClient($client)) {
			return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
		}

		$username = $client->getUsername();

		//Subuser vorher löschen...
		foreach ($client->getSubusers() as $subuser) {
			// Repo nicht zwischenspeichern, um keine Performance zu verlieren
			User::delete($subuser->getId(), false);
		}

		// Caches des Users leeren
		$thumbnails->clearCache($client);
		Cache::clearUser($client);
		// Broadcast clear cache request to other peers of this cluster
		ClusterApi::getInstance()->callRemoteApiBroadcast('app_cluster_api_user_cache_clear', [
			'userId' => $client->getId(),
			'apiKey' => $client->getApiKey(),
		], 'DELETE');

		// Benutzer und alle Benutzerdaten entfernen
		User::delete($client->getId());

		// Liste der veränderten Models leeren, damit der CacheHandler nicht annehmen muss,
		// dass sich Daten anderer User geändert haben
		Model::clearWrittenModelsClassNames();

		$this->addFlash('success', $translator->trans('Account %username% has been terminated.', [
			'%username%' => $username,
		]));

		return $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);

	}

	/**
	 * @param array $rows
	 * @param TranslatorInterface $translator
	 * @return Response|null
	 */
	private function validateCsvForClients(array $rows, TranslatorInterface $translator): ?Response
	{
		$usernameCounts = array_count_values(array_map('mb_strtolower', array_column($rows, 'username')));
		$hasDuplicateUsername = (bool)array_diff_key(array_flip($usernameCounts), [1=>null]);
		$hasEmptyUsername = isset($usernameCounts['']);

		if ($hasEmptyUsername) {
			$this->addFlash('danger', $translator->trans('Some rows have an empty username, please provide a non-empty value.'));
			return $this->redirectToRoute('app_hosting_client_main_account_from_csv', [], 303);
		}
		elseif ($hasDuplicateUsername) {
			$this->addFlash('danger', $translator->trans('Some rows define the same username, please provide only one row per client.'));
			return $this->redirectToRoute('app_hosting_client_main_account_from_csv', [], 303);
		}

		$passwordCounts = array_count_values(array_map('mb_strtolower', array_column($rows, 'password')));
		unset($passwordCounts['']); // Disregard empty password
		$hasDuplicatePassword = (bool)array_diff_key(array_flip($passwordCounts), [1=>null]);

		if ($hasDuplicatePassword) {
			$this->addFlash('danger', $translator->trans('Some rows define the same password, please provide a different password for each client.'));
			return $this->redirectToRoute('app_hosting_client_main_account_from_csv', [], 303);
		}

		$invalidCountryCodes = array_unique(array_filter(array_column($rows, 'country'), function (string $country) {
			return strlen($country) > 0 && Country::exists($country) == false;
		}));

		if ($invalidCountryCodes) {
			$this->addFlash('danger', $translator->trans('Some rows provide an invalid country code: :codes', [
				':codes' => implode(', ', $invalidCountryCodes),
			]));
			return $this->redirectToRoute('app_hosting_client_main_account_from_csv', [], 303);
		}

		$invalidEmailAddresses = array_unique(array_filter(array_column($rows, 'emailAddress'), function (string $emailAddress) {
			return strlen($emailAddress) > 0 && Utils::validateEmailAddress($emailAddress) == false;
		}));

		if ($invalidEmailAddresses) {
			$this->addFlash('danger', $translator->trans('Some rows provide an invalid email address: :issues', [
				':issues' => implode(', ', $invalidEmailAddresses),
			]));
			return $this->redirectToRoute('app_hosting_client_main_account_from_csv', [], 303);
		}

		return null;
	}

	/**
	 * @param array $rows
	 * @param TranslatorInterface $translator
	 * @return void
	 */
	private function createOrUpdateClientsFromCsv(array $rows, TranslatorInterface $translator): void
	{
		$statistics = [
			'clientsCreated' => 0,
			'clientsUpdated' => 0,
			'domainsCreated' => 0,
			'domainsSkipped' => 0,
		];

		foreach ($rows as $row) {
			User::beginWrite();
			$client = User::find($row['username']);

			if ($client) {
				// Silently ignore users other than clients' main accounts
				if ($client->isClient() == false || $client->isSubUser()) {
					continue;
				}
				$statistics['clientsUpdated']++;
			}
			else {
				try {
					$client = User::create($row['username'])
						->setRole(User::ROLE_CLIENT)
						->setActive(true)
					;
					$statistics['clientsCreated']++;
				} catch (UserAlreadyExistsException $exception) {
					continue; // Ignore silently
				}

				// Disable password login for new clients if the password was omitted
				if (strlen($row['password'] ?? '') == 0) {
					$client
						->setSettings(array_fill_keys([
								'canUsePlugin/ccm19/singlesignon',
							], true)
							+ $client->getSettings() + User::buildDefaultSettings())
						->unsetPassword()
					;
				}
			}

			if (isset($row['password']) && strlen($row['password']) > 0) {
				$client->setPassword($row['password']);
			}
			if (isset($row['firstName']) && strlen($row['firstName']) > 0) {
				$client->setFirstName($row['firstName']);
			}
			if (isset($row['lastName']) && strlen($row['lastName']) > 0) {
				$client->setLastName($row['lastName']);
			}
			if (isset($row['emailAddress']) && strlen($row['emailAddress']) > 0) {
				$client->setEmailAddress($row['emailAddress']);
			}
			if (isset($row['street']) && strlen($row['street']) > 0) {
				$client->setStrNr($row['street']);
			}
			if (isset($row['zipCode']) && strlen($row['zipCode']) > 0) {
				$client->setPlz($row['zipCode']);
			}
			if (isset($row['city']) && strlen($row['city']) > 0) {
				$client->setCity($row['city']);
			}
			if (isset($row['country']) && strlen($row['country']) > 0) {
				$client->setCountry(Country::find($row['country']));
			}
			if (isset($row['company']) && strlen($row['company']) > 0) {
				$client->setCompany($row['company']);
			}
			if (isset($row['vatId']) && strlen($row['vatId']) > 0) {
				$client->setVatId($row['vatId']);
			}

			// Write client to database
			$client->save();

			// Switch the user context
			User::setApiUser($client->getApiKey());
			Model::closeUserModels();
			Domain::switchToUser($client);

			// Process domains
			$domainUrls = array_filter(
				array_map('mb_strtolower', array_map('trim', explode('|', $row['domainUrls'] ?? ''))),
				function (string $url) { return preg_match('~^https?://.+~', $url); }
			);
			$domains = Domain::all();

			foreach ($domainUrls as $url) {
				// Look for a domain with the same URL
				$domain = array_reduce($domains, function (?Domain $match, Domain $item) use ($url) {
					return $match ?? (
					mb_strtolower(rtrim($item->getUrl(), '/')) == mb_strtolower(rtrim($url, '/'))
						? $item
						: $match
					);
				}, null);

				if ($domain) {
					$statistics['domainsSkipped']++;
					continue;
				}
				else {
					$domain = Domain::createForUser($client);
					$statistics['domainsCreated']++;
				}

				Domain::setDomainForRequest($domain->getId());
				Model::closeDomainModels();

				$urlInfo = parse_url(filter_var($url, FILTER_SANITIZE_URL));

				if (isset($urlInfo['host'])) {
					$urlHost = $urlInfo['host'];
					$urlPath = '/'.ltrim($urlInfo['path'] ?? '/', '/');
					$urlScheme = $urlInfo['scheme'] ?? 'https';
				}
				else {
					$parts = explode('/', trim($url), 2);
					$urlHost = $parts[0];
					$urlPath = '/'.ltrim($parts[1] ?? '', '/');
					$urlScheme = 'https';
				}

				// Write domain to database
				$domain
					->setName($urlHost)
					->setUrlPath($urlPath)
					->setUrlScheme($urlScheme)
					->save();
			}

			if ($statistics['domainsCreated'] > 0) {
				// Index neu generieren
				DomainIndex::rebuild();
			}

			// Update client info to set account limits next
			$client->updateDomainInfo();

			// Apply account limits
			if (isset($row['maxCallCount']) && is_numeric($row['maxCallCount'])) {
				$client->setMaxCallCount(max(-1, (int)$row['maxCallCount']));
			}
			if (isset($row['maxDomainCount']) && is_numeric($row['maxDomainCount'])) {
				$client->setMaxDomainCount(max(-1, max($client->getActualDomainCount(), (int)$row['maxDomainCount'])));
			}
			if (isset($row['maxWhitelabelCount']) && is_numeric($row['maxWhitelabelCount'])) {
				$client->setMaxWLCount(max(-1, max($client->getActualWLCount(), (int)$row['maxWhitelabelCount'])));
			}

			// Write changes to database
			$client->save();
		}

		$this->addFlash('success', $translator->trans('File processed! :cc clients created; :cu clients updated; :dc domains created; :ds domains skipped.', [
			':cc' => $statistics['clientsCreated'],
			':cu' => $statistics['clientsUpdated'],
			':dc' => $statistics['domainsCreated'],
			':ds' => $statistics['domainsSkipped'],
		]));
	}
}
