<?php

namespace App\Controller\Hosting;

use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Controller\HostingController;
use App\Exception\CsrfTokenException;
use App\Exception\UniqueConstraintException;
use App\Model\ClientGroup;
use App\Model\CsrfToken;
use App\Model\User;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Menu("Clients", group="Management", icon="fa-users", route_group="app_hosting_client_", order=10)
 * @Route("/hosting/client_group", name="app_hosting_client_client_group_")
 */
class ClientGroupController extends HostingController
{
	/**
	 * @Menu("Client groups", route_group="app_hosting_client_client_group_", order=5)
	 * @Route("", methods={"HEAD", "GET"}, name="index")
	 * @return Response
	 */
	public function listClientGroups(): Response
	{
		return $this->render('hosting/client_group/index.html.twig');
	}

	/**
	 * @Route("/index.json", methods={"HEAD", "GET"}, name="index_json")
	 * @return JsonResponse
	 */
	public function ajaxClients(): JsonResponse
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$clientGroups = $user->getAccessibleClientGroups();
		$userCount = array_reduce(User::getClients(), function (array $map, User $client) {
			foreach ($client->getClientGroups() as $group) {
				if (isset($map[$group->getId()])) {
					$map[$group->getId()]++;
				}
			}
			return $map;
		}, array_fill_keys(array_keys($clientGroups), 0));

		$tableData = array_map(function (ClientGroup $group) use ($user, $userCount) {
			return [
				'id' => $group->getId(),
				'name' => $group->getName(),
				'userCount' => $userCount[$group->getId()],
				'mayDelete' => $userCount[$group->getId()] == 0 || $user->isAdministrator(),
			];
		}, $clientGroups);

		return $this->json([
			'data' => array_values($tableData),
		]);
	}

	/**
	 * @Route("/new", methods={"GET","HEAD"}, name="new_form")
	 * @return Response
	 */
	public function createClientGroupForm(): Response
	{
		return $this->render('hosting/client_group/new.html.twig', [
			'clientGroup' => ClientGroup::createNonPersistent(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/new", methods={"POST"}, name="new")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function createClientGroup(Request $request, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$name = trim((string)$request->request->get('name'));

		$group = null;
		$success = false;

		if ($name == '') {
			$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
		}
		else {
			try {
				$group = ClientGroup::create($name);
				$success = true;
			} catch (UniqueConstraintException $e) {
				ClientGroup::reload();
				$this->addFlash('danger', $translator->trans('The name is already in use.'));
			}
		}

		// Use a temporary client group if there was an error
		$group = $group ?? ClientGroup::createNonPersistent();

		$group
			->setName($name)
		;

		try {
			CsrfToken::get()->checkRequest($request);

			if ($success) {
				$group->save();
				$logger->logSuccess('Hosting::ClientGroups',
					"User \"{$user->getUsername()}\" ({$user->getId()}) has created \"{$group->getName()}\" ({$group->getId()})"
				);

				// Grant the user access to the created client group
				if (!$user->isAdministrator()) {
					$user->assignClientGroup($group)->save();
				}

				$this->addFlash('success', $translator->trans('The client group ":name" has been created.', [
					':name' => $group->getName(),
				]));
				return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
			}
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		return $this->render('hosting/client_group/new.html.twig', [
			'clientGroup' => $group,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}", methods={"GET","HEAD"}, name="edit_form")
	 * @param string $id
	 * @return Response
	 */
	public function updateClientGroupForm(string $id): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$group = ClientGroup::find($id);
		if (!$group || !$user->mayAccessClientGroup($group)) {
			return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
		}

		return $this->render('hosting/client_group/edit.html.twig', [
			'clientGroup' => $group,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}", methods={"PUT"}, name="edit")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function updateClientGroup(string $id, Request $request, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$group = ClientGroup::find($id);
		if (!$group || !$user->mayAccessClientGroup($group)) {
			return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
		}

		$name = trim((string)$request->request->get('name'));

		$success = false;

		if ($name == '') {
			$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
		}
		elseif ($other = ClientGroup::byName($name) and $other->getId() != $group->getId()) {
			$this->addFlash('danger', $translator->trans('The name is already in use.'));
		}
		else {
			$success = true;
		}

		$oldName = $group->getName();

		$group
			->setName($name)
		;

		try {
			CsrfToken::get()->checkRequest($request);

			if ($success) {
				$group->save();
				$logger->logInfo('Hosting::ClientGroups',
					$name == $oldName
						? "User \"{$user->getUsername()}\" ({$user->getId()}) has updated \"{$group->getName()}\" ({$group->getId()})"
						: "User \"{$user->getUsername()}\" ({$user->getId()}) has renamed \"$oldName\" ({$group->getId()}) to {$group->getName()}"
				);

				$this->addFlash('success', $translator->trans('The client group ":name" has been updated.', [
					':name' => $group->getName(),
				]));

				switch ($request->request->getAlpha('action')) {
					case 'saveClose':
						return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
					case 'save':
					default:
						return $this->redirectToRoute('app_hosting_client_client_group_edit_form', ['id' => $group->getId()], 303);
				}
			}
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		return $this->render('hosting/client_group/edit.html.twig', [
			'clientGroup' => $group,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}/delete", methods={"GET","HEAD"}, name="delete_confirm")
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deleteClientGroupConfirmDialog(string $id, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$group = ClientGroup::find($id);
		if (!$group || !$user->mayAccessClientGroup($group)) {
			return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
		}

		// Check if the client group is assigned to any client
		$isAssignedToClients = array_reduce(User::getClients(), function (bool $carry, User $client) use ($group) {
			return $carry || $client->hasClientGroup($group);
		}, false);

		// Users other than admins aren't allowed to delete the client group
		// if it is still assigned to clients
		if ($isAssignedToClients && !$user->isAdministrator()) {
			$this->addFlash('danger', $translator->trans('You cannot delete this client group, because it is still assigned to clients.'));
			return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
		}

		return $this->render('hosting/client_group/delete.html.twig', [
			'clientGroup' => $group,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}/delete", methods={"DELETE"}, name="delete")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteClientGroup(string $id, Request $request, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$group = ClientGroup::find($id);
		if (!$group || !$user->mayAccessClientGroup($group)) {
			return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
		}

		// Check if the client group is assigned to any client
		$isAssignedToClients = array_reduce(User::getClients(), function (bool $carry, User $client) use ($group) {
			return $carry || $client->hasClientGroup($group);
		}, false);

		// Users other than admins aren't allowed to delete the client group
		// if it is still assigned to clients
		if ($isAssignedToClients && !$user->isAdministrator()) {
			$this->addFlash('danger', $translator->trans('You cannot delete this client group, because it is still assigned to clients.'));
			return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
		}

		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
		}

		$name = $group->getName();

		ClientGroup::delete($group->getId());

		$logger->logSuccess('Hosting::ClientGroups',
			"User \"{$user->getUsername()}\" ({$user->getId()}) has deleted \"$name\" ($id)"
		);

		$this->addFlash('success', $translator->trans('The client group ":name" has been deleted.', [
			':name' => $name,
		]));

		return $this->redirectToRoute('app_hosting_client_client_group_index', [], 303);
	}
}
