<?php

namespace App\Controller\Hosting;

use App\Component\HttpFoundation\JavascriptResponse;
use App\Component\Menu\Annotation\Menu;
use App\Controller\HostingController;
use App\Model\User;
use App\Model\Domain;
use App\Model\Purpose;
use App\Model\Statistic;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use DateTimeImmutable;
use RuntimeException;


class Dashboard extends HostingController
{
	/**
	 * @Menu("Dashboard", group="Management", icon="glyphicon-home", order=0)
	 * @Route("/hosting/dashboard", name="app_hosting_dashboard")
	 * @return Response
	 */
	public function dashboard(): Response
	{
		$clientUsers = User::byRole(User::ROLE_CLIENT);

		$clients = array_filter($clientUsers, static function (User $user) { return $user->isAccountOwner(); });
		$subusers = array_diff_key($clientUsers, $clients);

		$activeClientsCount = count(array_filter($clients, static function (User $user) { return $user->isActive(); }));

		$domainCount = array_reduce($clients, static function (int $count, User $client) {
			return $count + $client->getActualDomainCount();
		}, 0);

		switch ($_ENV['APP_SESSION_HANDLER'] ?? 'php') {
			case 'redis':
				if (!class_exists('Redis')) {
					$this->addFlash('danger', 'Redis session handler is configured, but Redis PHP module is missing. Please install the Redis extension.');
				}
				break;
			default:
				$this->addFlash('danger', 'Unknown session handler configured. Please check your .env.local.');
				break;
			case 'php':
				// ignore
		}

		return $this->render('hosting/dashboard/index.html.twig', [
			'clientCount' => count($clients),
			'clientCountInActive' => count($clients) - $activeClientsCount,
			'domainCountTotal' => $domainCount,
			'subuserCount' => count($subusers),
		]);
	}

	/**
	 * @Route("/hosting/chart.js", methods={"HEAD","GET"}, name="app_hosting_dashboard_chart_js")
	 * @param Request $request
	 * @return JavascriptResponse
	 */
	public function chartJs(Request $request): JavascriptResponse
	{
		$response = new JavascriptResponse();
		Utils::resetCacheControl($response);
		$response->setMaxAge(120);
		$response->setCache([
			'max_age'=>120,
			'public'=>false
		]);
		$loggedInUser = User::loggedInUser();

		$purposes = Purpose::all();
		$statistics = [];
		foreach ($purposes as $purpose) {
			$statistics[$purpose->getId()] = 0;
		}

		$startDate = new DateTimeImmutable('midnight first day of this month');

		$count = 0;
		$countCall = 0;
		// Statistik über alle User erstellen
		try {
			foreach (User::getClients() as $client) {
				Domain::switchToUser($client);
				foreach (Domain::all() as $domain) {
					Domain::setDomainForRequest($domain->getId());
					Statistic::reloadRepository();
					$statList = Statistic::inDateRange($startDate, null);
					foreach ($statList as $date => $stat) {
						$countCall += $stat->getCallCount();
						foreach ($purposes as $purpose) {
							$statistics[$purpose->getId()] += $stat->getPurposeAcceptCount($purpose->getId());
						}
						$count += $stat->getViewCount();
					}
				}
			}
		}
		catch (RuntimeException $exception) {}
		finally {
			Domain::switchToUser($loggedInUser);
			Domain::setDomainForRequest('');
		}

		// Session schließen.
		$request->getSession()->save();

		$parameters = [
			'purposes' => $purposes,
			'statistics' => $statistics,
			'viewCountTotal' => $count,
			'callCountTotal' => $countCall,
		];

		// @phan-suppress-next-line PhanTypeMismatchReturnSuperType
		return $this->render('dashboard/chart.js.twig', $parameters, $response);
	}
}
