<?php

namespace App\Controller\Hosting;

use App\Component\LocaleService;
use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingAssetBase;
use App\Model\EmbeddingBase;
use App\Model\Language;
use App\Model\LocaleBase;
use App\Model\Model;
use App\Model\Purpose;
use App\Model\Regulations\Article;
use App\Model\Regulations\Regulation;
use App\Model\Regulations\Group as RegulationGroup;
use App\Model\SystemEmbedding;
use App\Model\SystemEmbeddingAsset;
use App\Config;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\EmbeddingDb\EmbeddingDbLegacy;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Controller\HostingController;
use App\Model\User;
use App\Model\UserEmbedding;
use App\Model\UserEmbeddingAsset;
use App\Model\UserLocale;
use App\Utils;
use Exception;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use DateTimeImmutable;
use DateTimeInterface;

/**
 * Controller zur Admin-Embedding-Verwaltung
 * @Route("/hosting/embeddings", name="app_hosting_embeddings_")
 */
class Embeddings extends HostingController
{
	/**
	 * @Menu("Embedding templates", group="Management", icon="fa-certificate", order=30)
	 * @Route("", name="list", methods={"GET", "HEAD"})
	 *
	 * @param Request $request
	 * @throws \Exception Wenn kein Zugriff auf das Embedding-File erfolgen kann
	 *
	 * @return Response
	 */
	public function listEmbeddings(Request $request): Response
	{
		$config = Config::getInstance();
		$embeddings = SystemEmbedding::all();
		$purposes = Purpose::all();

		return $this->render('hosting/embedding/list.html.twig', [
			'embeddings' => $embeddings,
			'purposes' => $purposes,
			'selectedLanguages' => $config->getSystemEmbeddingLanguages(),
			'allLanguages' => Language::all(),
			'useGlobalCookieDb' => $config->isGlobalCookieDbUsed(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/list.json", name="list_ajax")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return JsonResponse
	 */
	public function ajaxEmbeddingSearch(Request $request, TranslatorInterface $translator): JsonResponse
	{
		$currentLanguageCode = explode('_', $request->getLocale(), 2)[0];
		$defaultLocale = LocaleService::getDefault();
		$defaultLanguage = $defaultLocale->getLanguage() ?? Language::find('de_DE');

		$search = trim((string)($request->query->all('search')['value'] ?? ''));
		$search = $search == '' ? null : $search;

		// Lade Einbindungsvorlagen
		$templates = SystemEmbedding::all();
		// FIXME Total record count is unknown when searching
		$embeddingsDB = EmbeddingDb::fetchEmbeddingList($search, Language::find($currentLanguageCode) ?? $defaultLanguage);

		// Mappe Daten der Embeddings aus externer DB
		$embeddingsDB = array_values(array_filter(array_map(function ($embedding) use ($translator, $currentLanguageCode) {
			return [
				'_score' => $embedding['@score'] ?? null,
				'id' => $embedding['id'],
				'name' => $embedding['translations'][$currentLanguageCode]['label'],
				'source' => EmbeddingBase::SOURCE_GLOBAL_DB,
				'embedding_vendor' => $embedding['embedding_vendor'],
				'description' => $embedding['translations'][$currentLanguageCode]['description'],
				'purpose' => Purpose::find($embedding['purpose'])->getName(),
				'assets' => $embedding['assets'],
			];
		}, $embeddingsDB)));

		// Sortierung der Embeddings
		$order = $request->query->all('order')[0];
		$sortColumn = $order['column'] ?? '';
		$sortDirection = $order['dir'] ?? 'asc';

		$embeddingsDB = Utils::sortNatNum($embeddingsDB, $sortColumn, $sortDirection);

		$visibleEmbeddings = $embeddingsDB;

		// Filtere bereits bestehende Embeddings aus angezeigten Embeddings heraus
		foreach ($templates as $activeEmbedding) {
			foreach ($visibleEmbeddings as $key => $embeddingTemplate) {
				if ($embeddingTemplate['id'] == $activeEmbedding->getSourceId()) {
					unset($visibleEmbeddings[$key]);
				}
			}
		}

		// Filterung nach Suchbegriff
		// Code aus User- und Domain-Embedding-Controller übernommen, daher nur die einfache Zuweisung
		$embeddingsFiltered = $visibleEmbeddings;

		// Paginierung der Einträge
		$start = $request->query->getInt('start');
		$length = $request->query->getInt('length') ?: 10;

		$pagination = new Pagination($embeddingsFiltered, $length);
		$currentPage = ($start / $length) + 1;
		$pagination->setCurrentPage($currentPage);

		$embeddingsPaginated = $pagination->getItems();

		return $this->json([
			'data' => array_values($embeddingsPaginated),
			'draw' => $request->query->getInt('draw'),
			'recordsTotal' => count($visibleEmbeddings),
			'recordsFiltered' => count($embeddingsFiltered),
		]);
	}

	/**
	 * @Route("", name="save_settings", methods={"POST"})
	 *
	 * @param Request $request
	 * @throws \Exception Wenn kein Zugriff auf das Config-File erfolgen kann
	 *
	 * @return Response
	 */
	public function saveSettings(Request $request): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
			$config = Config::getInstance();

			$config->setGlobalCookieDbUsed($request->request->getBoolean('useGlobalCookieDb'));
			$config->setSystemEmbeddingLanguages(array_filter(array_map(function ($code) {
				return Language::find($code);
			}, (array)$request->request->get('selectedLanguages'))));
			$config->flush();

			$this->addFlash('success', 'Settings have been saved.');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		return $this->redirectToRoute('app_hosting_embeddings_list', [], 303);
	}

	/**
	 * Embeddings aus der CCM19.de Datenbank anzeigen zur Auswahl...
	 * @Route("/db", name="from_db", methods={"HEAD", "GET"} )
	 * @return Response
	 *
	 * @todo überarbeiten
	 */
	public function listEmbeddingsFromDatabase(Request $request): Response
	{
		return $this->render('hosting/embedding/list_db.html.twig', [
		]);
	}

	/**
	 * Einbindungsvorlagen von Kunden oder Einbindungen von Domains als
	 * systemweite Einbindungsvorlage übernehmen.
	 *
	 * @Route("/client", name="from_client", methods={"HEAD", "GET"})
	 *
	 * @return Response
	 */
	public function listEmbeddingsFromClient(): Response
	{
		return $this->render('hosting/embedding/list_client.html.twig');
	}

	/**
	 * Kunden anzeigen, die mindestens eine accountweite Einbindung haben.
	 *
	 * @Route("/client/clients.json", name="from_client_ajax_clients", methods={"HEAD", "GET"})
	 *
	 * @return JsonResponse
	 */
	public function ajaxClients(): JsonResponse
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$tableData = [];

		foreach (User::getClients() as $client) {
			// Switch the user context
			User::setApiUser($client->getApiKey());
			Model::closeUserModels();

			if (UserEmbedding::isEmpty()) {
				continue;
			}

			$tableData[$client->getId()] = [
				'client' => [
					'id' => $client->getId(),
					'username' => $client->getUsername(),
				],
				'embeddings' => array_values(array_map(function (UserEmbedding $embedding) {
					return [
						'id' => $embedding->getId(),
						'name' => $embedding->getName(),
					];
				}, UserEmbedding::all())),
			];
		}

		// Switch back to the original user context
		User::setApiUser($user->getApiKey());
		Model::closeUserModels();

		return $this->json([
			'data' => array_values($tableData),
		]);
	}

	/**
	 * Kunden anzeigen, die mindestens eine Domain mit einer Einbindung haben.
	 *
	 * @Route("/client/domains.json", name="from_client_ajax_domains", methods={"HEAD", "GET"})
	 *
	 * @return JsonResponse
	 */
	public function ajaxDomains(): JsonResponse
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$tableData = [];

		foreach (User::getClients() as $client) {
			// Switch the user context
			User::setApiUser($client->getApiKey());
			Model::closeUserModels();
			Domain::switchToUser($client);

			foreach (Domain::all() as $domain) {
				Domain::setDomainForRequest($domain->getId());
				Model::closeDomainModels();

				$tableData["{$client->getId()}-{$domain->getId()}"] = [
					'client' => [
						'id' => $client->getId(),
						'username' => $client->getUsername(),
					],
					'domain' => [
						'id' => $domain->getId(),
						'host' => (string)parse_url($domain->getUrl(), PHP_URL_HOST),
						'path' => $domain->getUrlPath(),
					],
					'embeddings' => array_values(array_map(function (Embedding $embedding) {
						return [
							'id' => $embedding->getId(),
							'name' => $embedding->getName(),
						];
					}, Embedding::all())),
				];
			}
		}

		// Switch back to the original user context
		Domain::setDomainForRequest('');
		User::setApiUser($user->getApiKey());
		Model::closeUserModels();

		// Select only domains that have embeddings
		$tableData = array_filter($tableData, function (array $row) {
			return count($row['embeddings']) > 0;
		});

		return $this->json([
			'data' => array_values($tableData),
		]);
	}

	/**
	 * Entpackt mit `data-pack="asset,add_asset"` als JSON gepackte POST-Daten
	 */
	private function unpackPostData(array $array): array
	{
		foreach (['asset', 'add_asset'] as $key) {
			if (isset($array[$key]) and is_string($array[$key])) {
				$array[$key] = json_decode($array[$key], true, 64, JSON_INVALID_UTF8_IGNORE);
			}
		}
		return $array;
	}

	/**
	 * Daten aus dem Array $embeddingData in $embedding übertragen
	 */
	private function applyChanges(SystemEmbedding $embedding, array $embeddingData, string $defaultLanguageId): SystemEmbedding
	{
		$request = Utils::getCurrentRequest();
		$config = Config::getInstance();

		// Script-Blocking
		$codeBlock = explode("\n", trim((string)($embeddingData['code_block'] ?? '')));
		$codeBlock = array_filter($codeBlock, function ($line) {
			return trim($line) !== '';
		});
		$codeBlock = array_values(array_map(function ($line) {
			return trim($line);
		}, $codeBlock));

		// Iframes blockieren
		$iframeBlockingStrings = array_filter(array_map(function ($line) {
			return trim($line);
		}, preg_split('~\r\n|\n|\r~', (string)($embeddingData['iframe_blocking_strings'] ?? ''))));

		// Exclude Urls
		$excludeUrls = explode("\n", trim((string)($embeddingData['excludeUrls'] ?? '')));
		$excludeUrls = array_filter($excludeUrls, function ($line) {
			return trim($line) !== '';
		});
		$excludeUrls = array_values(array_map(function ($line) {
			return trim($line);
		}, $excludeUrls));
		// Zweck
		$purpose = Purpose::find($embeddingData['purpose'] ?? '');

		if ($request) {
			$dispatchLoadEventForIframes = $request->isMethodCacheable()
				// GET: Wenn der Parameter nicht vorhanden ist, dann soll der Standardwert verwendet werden
				? (bool)($embeddingData['dispatchLoadEventForIframes'] ?? $embedding->isDispatchLoadEventForIframes())
				// POST/PUT: Wenn der Parameter nicht vorhanden ist, wurde der Schalter im Formular deaktiviert
				: (bool)($embeddingData['dispatchLoadEventForIframes'] ?? false);
		}
		else {
			// Sollte nicht auftreten; Standardwert verwenden
			$dispatchLoadEventForIframes = $embedding->isDispatchLoadEventForIframes();
		}

		// Allgemeines speichern
		$embedding
			->setName($embeddingData['name'] ?? '')
			->setDescription($embeddingData['description'][$defaultLanguageId] ?? '')
			->setCodeSnippet($embeddingData['code_snippet'] ?? '')
			->setCodeBlock($codeBlock)
			->setScriptLoaderGroup($embeddingData['scriptLoaderGroup'] ?? '')
			->setIframeBlockingStrings($iframeBlockingStrings)
			->setIframesHiddenUntilConsent((bool)($embeddingData['hideIframesUntilConsent'] ?? false))
			->setDispatchLoadEventForIframes($dispatchLoadEventForIframes)
			->setExcludeUrls($excludeUrls)
			->setExcludeUrlsMode($embeddingData['excludeUrlsMode'] ?? $embedding->getExcludeUrlsMode())
			->setEmbeddingVendor($embeddingData['embedding_vendor'] ?? '')
			->setPrivacyPolicyUrl($embeddingData['privacy_policy_url'][$defaultLanguageId] ?? '')
			->setHttpEmbeddingName($embeddingData['http_embedding_name'] ?? '')
			->setLifetime($embeddingData['lifetime'][$defaultLanguageId] ?? '')
			->setCollectedDataInfo($embeddingData['collected_data_info'][$defaultLanguageId] ?? '')
			->setPurposeOfDataCollection($embeddingData['purpose_of_data_collection'][$defaultLanguageId] ?? '')
			->setLegalBasis($embeddingData['legal_basis'][$defaultLanguageId] ?? '')
			->setLegalBasisStructured((bool)($embeddingData['legal_basis_structured'] ?? false))
			->setStructuredLegalBasis((array)($embeddingData['structured_legal_basis'] ?? []))
			->setLegalBasisSupplement($embeddingData['legal_basis_supplement'][$defaultLanguageId] ?? '')
			->setPlaceOfProcessing($embeddingData['place_of_processing'][$defaultLanguageId] ?? '')
			->setAutoUpdated($embeddingData['autoUpdated'] ?? false)
			->setUpdateParameterBlacklist((array)($embeddingData['updateParameterBlacklist'] ?? []))
			->updateLastModified()
		;
		if ($purpose) {
			$embedding->setPurpose($purpose);
		}

		if (isset($embeddingData['source'])) {
			$source_ts = ($embeddingData['source_ts'] ?? null) ?: null;
			if ($source_ts and !($source_ts instanceof DateTimeInterface)) {
				$source_ts = new DateTimeImmutable((string)$source_ts);
			}
			if (!$source_ts) {
				$source_ts = null;
			}
			$embedding->setSource($embeddingData['source'] ?? EmbeddingBase::SOURCE_UNKNOWN)
				->setSourceId(!empty($embeddingData['source_id']) ? $embeddingData['source_id'] : null)
				->setSourceTimestamp($source_ts);
		}

		// Google Consent Mode speichern
		if (isset($embeddingData['gcm']) and is_array($embeddingData['gcm']) and $embedding->hasVendorGoogle()) {
			$embedding->setGoogleConsentMode($embeddingData['gcm']);
		} else {
			$embedding->setGoogleConsentMode([]);
		}

		// TCF-Daten speichern
		$embedding->setTcfVendor(
			(is_string($embeddingData['tcf_vendor_id'] ?? null) or $embedding->getTcfVendor())
			? TcfVendor::find($embeddingData['tcf_vendor_id'] ?? $embedding->getTcfVendor()->getId())
			: null);
		$embedding->setTcfPurposes(array_filter(array_map(
			function ($id) {
				return TcfPurpose::find($id);
			},
			(array)($embeddingData['tcf'][$embeddingData['tcf_vendor_id'] ?? '']['p'] ?? [])
		)));
		$embedding->setTcfSpecialFeatures(array_filter(array_map(
			function ($id) {
				return TcfSpecialFeature::find($id);
			},
			(array)($embeddingData['tcf'][$embeddingData['tcf_vendor_id'] ?? '']['sf'] ?? [])
		)));

		// Assets bearbeiten
		foreach ($embeddingData['asset'] ?? [] as $assetId => $data) {
			/** @var SystemEmbeddingAsset $asset */
			$asset = $embedding->getAsset($assetId);
			$assetName = (string)($data['name'] ?? '');

			if ($asset === null || empty($assetName)) {
				continue;
			}

			$asset
				->setName($assetName)
				->setDynamic((bool)($data['dynamic'] ?? false))
				->setStorageType((string)($data['storageType'] ?? ''))
				->setLifetime((string)($data['lifetime'][$defaultLanguageId] ?? ''))
				->setDescription((string)($data['description'][$defaultLanguageId] ?? ''))
			;

			// Asset-Übersetzungen
			foreach ($config->getSystemEmbeddingLanguages() as $languageId => $language) {
				if (!isset($data['description'][$languageId])) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				// @phan-suppress-next-line PhanUndeclaredMethod
				$asset
					->setTranslatedDescription($language, $data['description'][$languageId] ?? '')
					->setTranslatedLifetime($language, $data['lifetime'][$languageId] ?? '')
				;
			}
		}
		if(!empty($embeddingData['add_asset'][$defaultLanguageId]))
		{
			$addAssets = $embeddingData['add_asset'][$defaultLanguageId];
		}
		elseif (is_array($embeddingData['add_asset'] ?? null)) {
			$addAssets = $embeddingData['add_asset'];
		}

		// Assets hinzufügen
		foreach ($addAssets ?? [] as $data) {
			$assetName = (string)($data['name'] ?? '');
			if (empty($assetName)) {
				continue;
			}
			$asset = SystemEmbeddingAsset::create()
				->setName($assetName)
				->setDynamic((bool)($data['dynamic'] ?? false))
				->setStorageType((string)($data['storageType'] ?? ''))
				//->setLifetime((string)($data['lifetime'] ?? ''))
				//->setDescription((string)($data['description'] ?? ''))
			;

			//assets lang Data
			foreach ($config->getSystemEmbeddingLanguages() as $languageId => $language) {
				if (!isset($data['description'][$languageId])) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				// @phan-suppress-next-line PhanUndeclaredMethod
				$asset
					->setTranslatedDescription($language, $data['description'][$languageId] ?? '')
					->setTranslatedLifetime($language, $data['lifetime'][$languageId] ?? '');
			}
			//embed now...
			$embedding->addAsset($asset);
		}

		// Assets löschen
		foreach ($embeddingData['delete_asset'] ?? [] as $assetId) {
			$embedding->removeAsset($assetId);
		}

		// Übersetzungen speichern
		foreach ($config->getSystemEmbeddingLanguages() as $languageId => $language) {
			if (!isset($embeddingData['description'][$languageId])) {
				continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
			}
			$embedding
				->setTranslatedName($language, $embeddingData['label'][$languageId] ?? '')
				->setTranslatedDescription($language, $embeddingData['description'][$languageId] ?? '')
				->setTranslatedPrivacyPolicyUrl($language, $embeddingData['privacy_policy_url'][$languageId] ?? '')
				->setTranslatedLifetime($language, $embeddingData['lifetime'][$languageId] ?? '')
				->setTranslatedCollectedDataInfo($language, $embeddingData['collected_data_info'][$languageId] ?? '')
				->setTranslatedPurposeOfDataCollection($language, $embeddingData['purpose_of_data_collection'][$languageId] ?? '')
				->setTranslatedLegalBasis($language, $embeddingData['legal_basis'][$languageId] ?? '')
				->setTranslatedLegalBasisSupplement($language, $embeddingData['legal_basis_supplement'][$languageId] ?? '')
				->setTranslatedPlaceOfProcessing($language, $embeddingData['place_of_processing'][$languageId] ?? '')
				->setTranslatedBlockedContentTitle($language, $embeddingData['blocked_content_title'][$languageId] ?? '')
				->setTranslatedBlockedContentText($language, $embeddingData['blocked_content_text'][$languageId] ?? '')
				->setTranslatedBlockedContentButtonText($language, $embeddingData['blocked_content_button_text'][$languageId] ?? '')
			;
		}

		return $embedding;
	}

	/**
	 * @param Request $request
	 * @param string $defaultLanguageId
	 * @return array
	 */
	private static function prepareNewEmbeddingData(Request $request, string $defaultLanguageId): array
	{
		$embeddingData = [];
		// Import aus DB nach ID
		if (($remoteId = $request->query->getInt('remoteId'))) {
			$remoteDb = $request->query->get('remoteDb', 'v2.2');
			if ($remoteDb === 'v2') {
				$remoteEmbedding = EmbeddingDbLegacy::fetchEmbedding($remoteId);
			} else {
				$remoteEmbedding = EmbeddingDb::fetchEmbedding($remoteId);
			}

			$embeddingData = self::createRemoteEmbeddingFromDbEntry($remoteEmbedding, (string)$remoteDb);
		}
		elseif (
			$request->query->has('client') and $request->query->has('embedding')
			and $client = User::byId($request->query->getAlnum('client'))
			and $client->isClient() and $client->isAccountOwner()
		) {
			$embeddingData = ['source'=>EmbeddingBase::SOURCE_LOCAL];

			$user = User::loggedInUser();
			assert($user instanceof User);

			// Switch the user context
			User::setApiUser($client->getApiKey());
			Model::closeUserModels();

			try {
				if ($request->query->has('domain')) {
					Domain::switchToUser($client);
					if (Domain::setDomainForRequest($request->query->getAlnum('domain'))
						and $embedding = Embedding::find($request->query->getAlnum('embedding'))
					) {
						$embeddingData += self::extractDataFromDomainEmbedding($embedding, $defaultLanguageId);
					}
				}
				elseif ($userEmbedding = UserEmbedding::find($request->query->getAlnum('embedding'))) {
					$embeddingData += self::extractDataFromUserEmbedding($userEmbedding, $defaultLanguageId);
				}
			}
			finally {
				// Switch back to the original user context
				Domain::setDomainForRequest('');
				User::setApiUser($user->getApiKey());
				Model::closeUserModels();
			}
		}
		else {
			$embeddingData = ['source'=>EmbeddingBase::SOURCE_LOCAL];
		}
		return $embeddingData;
	}

	private static function createRemoteEmbeddingFromDbEntry($remoteEmbedding, string $remoteDb = 'v2.2'): array
	{
		if (is_array($remoteEmbedding) and $remoteEmbedding) {
			$embeddingData = [];
			$embeddingData['hinweis'] = "hinweis_remote";

			$embeddingData['source'] = ($remoteDb === 'v2') ? EmbeddingBase::SOURCE_GLOBAL_DB_LEGACY : EmbeddingBase::SOURCE_GLOBAL_DB;
			$embeddingData['source_id'] = $remoteEmbedding['id'];
			$embeddingData['source_ts'] = $remoteEmbedding['timestamp'];

			// Upgrade zu neuer DB
			if (isset($remoteEmbedding['@upgrade'])) {
				$upgradeSource = $remoteEmbedding['@upgrade']['source'];
				$upgradeId = $remoteEmbedding['@upgrade']['sourceId'];
				if ($upgradeSource !== $embeddingData['source'] or (string)$upgradeId !== (string)$remoteEmbedding['id']) {
					$embeddingData['source'] = $upgradeSource;
					$embeddingData['source_id'] = $upgradeId;
					if ($upgradeSource === EmbeddingBase::SOURCE_GLOBAL_DB) {
						$remoteEmbedding = EmbeddingDb::fetchEmbedding($upgradeId);
						$embeddingData['source_ts'] = $remoteEmbedding['timestamp'];
					}
				}
			}

			// Basisdaten übertragen
			foreach (['name', 'purpose', 'code_snippet', 'code_block', 'embedding_vendor', 'iframe_blocking_strings'] as $key) {
				$embeddingData[$key] = $remoteEmbedding[$key];
			}
			foreach (['label', 'description', 'privacy_policy_url', 'collected_data_info', 'purpose_of_data_collection', 'legal_basis', 'legal_basis_supplement', 'place_of_processing'] as $key) {
				$embeddingData[$key] = [];
			}
			if (isset($remoteEmbedding['structured_legal_basis'])) {
				$embeddingData['structured_legal_basis'] = $remoteEmbedding['structured_legal_basis'];
				$embeddingData['legal_basis_structured'] = true;

				// Falls für DSGVO, DSG oder CCPA keine Daten existieren, wird ein leeres Array übergeben
				$regulationIds = array_keys(Regulation::rawAll());
				$embeddingData['structured_legal_basis'] += array_fill_keys($regulationIds, []);
			} else {
				$embeddingData['legal_basis_structured'] = false;
			}

			// Übersetzungen übertragen
			foreach (Language::all() as $languageId => $language) {
				$shortCode = $language->getShortCode();
				if (!empty($remoteEmbedding['translations'][$shortCode]) and !empty($remoteEmbedding['translations'][$shortCode]['description'])) {
					$translation = $remoteEmbedding['translations'][$shortCode];
					foreach ($translation as $key => $value) {
						$embeddingData[$key][$languageId] = $value;
					}
				}
			}

			// Assets übertragen
			$embeddingData['add_asset'] = array_map(function ($asset) {
				$result = [
					'name' => $asset['name'],
					'storageType' => $asset['storageType'],
					'dynamic' => $asset['dynamic'],
					'lifetime' => [],
					'description' => [],
				];

				// Übersetzungen übertragen
				foreach (Language::all() as $languageId => $language) {
					$shortCode = $language->getShortCode();
					if (!empty($asset['translations'][$shortCode]) and !empty($asset['translations'][$shortCode]['lifetime'])) {
						$translation = $asset['translations'][$shortCode];
						foreach ($translation as $key => $value) {
							$result[$key][$languageId] = $value;
						}
					}
				}

				return $result;
			}, $remoteEmbedding['assets']);

		} else {
			$embeddingData = ['source' => EmbeddingBase::SOURCE_LOCAL];
		}
		return $embeddingData;
	}

	/**
	 * @param EmbeddingBase $embedding
	 * @param string $defaultLanguageId
	 * @param callable $assetL10nFn
	 * @return array
	 */
	private static function extractDataFromEmbeddingBase(EmbeddingBase $embedding, string $defaultLanguageId, callable $assetL10nFn): array
	{
		// Basisdaten extrahieren
		$data = [
			'name' => $embedding->getName(),
			'purpose' => $embedding->getPurpose() ? $embedding->getPurpose()->getId() : '',
			'code_snippet' => $embedding->getCodeSnippet(),
			'code_block' => implode("\n", $embedding->getCodeBlock()),
			'embedding_vendor' => $embedding->getEmbeddingVendor(),
			'iframe_blocking_strings' => implode("\n", $embedding->getIframeBlockingStrings()),
			'hideIframesUntilConsent' => $embedding->isIframesHiddenUntilConsent(),
			'dispatchLoadEventForIframes' => $embedding->isDispatchLoadEventForIframes(),
			'excludeUrls' => implode("\n", $embedding->getExcludeUrls()),
			'excludeUrlsMode' => $embedding->getExcludeUrlsMode(),
			'scriptLoaderGroup' => $embedding->getScriptLoaderGroup(),
			'structured_legal_basis' => array_map(static function (array $articles) {
				return array_map(static function (Article $article) { return (int)$article->getId(); }, $articles);
			}, $embedding->getStructuredLegalBasis()),
			'legal_basis_structured' => $embedding->isLegalBasisStructured(),
			'autoUpdated' => false,
			'updateParameterBlacklist' => $embedding->getUpdateParameterBlacklist(),
		];

		// Übersetzungen extrahieren
		$data += [
			'blocked_content_button_text' => [],
			'blocked_content_text' => [],
			'blocked_content_title' => [],
			'collected_data_info' => [$defaultLanguageId=>$embedding->getCollectedDataInfo()],
			'description' => [$defaultLanguageId=>$embedding->getDescription()],
			'legal_basis' => [$defaultLanguageId=>$embedding->getLegalBasis()],
			'legal_basis_supplement' => [$defaultLanguageId=>$embedding->getLegalBasisSupplement()],
			'place_of_processing' => [$defaultLanguageId=>$embedding->getPlaceOfProcessing()],
			'privacy_policy_url' => [$defaultLanguageId=>$embedding->getPrivacyPolicyUrl()],
			'purpose_of_data_collection' => [$defaultLanguageId=>$embedding->getPurposeOfDataCollection()],
		];

		// Assets extrahieren
		$data['add_asset'] = array_map(function (EmbeddingAssetBase $asset) use ($defaultLanguageId, $assetL10nFn) {
			$result = [
				'name' => $asset->getName(),
				'storageType' => $asset->getStorageType(),
				'dynamic' => $asset->isDynamic(),
			];

			// Übersetzungen extrahieren
			$result += [
				'description' => [$defaultLanguageId=>$asset->getDescription()],
				'lifetime' => [$defaultLanguageId=>$asset->getLifetime()],
			];

			$assetL10nFn($asset, $result);

			return $result;
		}, $embedding->getAssets());

		return $data;
	}

	/**
	 * @param Embedding $embedding
	 * @param string $defaultLanguageId
	 * @return array
	 */
	private static function extractDataFromDomainEmbedding(Embedding $embedding, string $defaultLanguageId): array
	{
		$locales = LocaleService::enabledLocales();
		$locales = array_combine(
			array_map(function (LocaleBase $locale) {
				return $locale->getLanguage() ? $locale->getLanguage()->getCode() : $locale->getName();
			}, $locales),
			$locales
		);

		$data = self::extractDataFromEmbeddingBase(
			$embedding,
			$defaultLanguageId,
			// Übersetzung für jedes EmbeddingAsset extrahieren
			function (EmbeddingAsset $asset, array &$data) use ($locales) {
				foreach ($locales as $languageId => $locale) {
					$data['description'][$languageId] = $asset->getTranslatedDescription($locale, false);
					$data['lifetime'][$languageId] = $asset->getTranslatedLifetime($locale, false);
				}
			}
		);

		// Übersetzungen extrahieren
		foreach ($locales as $languageId => $locale) {
			$data['blocked_content_button_text'][$languageId] = $embedding->getTranslatedBlockedContentButtonText($locale, false);
			$data['blocked_content_text'][$languageId] = $embedding->getTranslatedBlockedContentText($locale, false);
			$data['blocked_content_title'][$languageId] = $embedding->getTranslatedBlockedContentTitle($locale, false);
			$data['collected_data_info'][$languageId] = $embedding->getTranslatedCollectedDataInfo($locale, false);
			$data['description'][$languageId] = $embedding->getTranslatedDescription($locale, false);
			$data['legal_basis'][$languageId] = $embedding->getTranslatedLegalBasis($locale, false);
			$data['legal_basis_supplement'][$languageId] = $embedding->getTranslatedLegalBasisSupplement($locale, false);
			$data['place_of_processing'][$languageId] = $embedding->getTranslatedPlaceOfProcessing($locale, false);
			$data['privacy_policy_url'][$languageId] = $embedding->getTranslatedPrivacyPolicyUrl($locale, false);
			$data['purpose_of_data_collection'][$languageId] = $embedding->getTranslatedPurposeOfDataCollection($locale, false);
		}

		return $data;
	}

	/**
	 * @param UserEmbedding $embedding
	 * @param string $defaultLanguageId
	 * @return array
	 */
	private static function extractDataFromUserEmbedding(UserEmbedding $embedding, string $defaultLanguageId): array
	{
		/** @var UserLocale[] $locales */
		$locales = array_filter(array_map(function (Language $language) {
			return UserLocale::findByLanguage($language);
		}, Language::all()));

		$data = self::extractDataFromEmbeddingBase(
			$embedding,
			$defaultLanguageId,
			// Übersetzung für jedes UserEmbeddingAsset extrahieren
			function (UserEmbeddingAsset $asset, array &$data) use ($locales) {
				foreach ($locales as $languageId => $locale) {
					$data['description'][$languageId] = $asset->getTranslatedDescription($locale, false);
					$data['lifetime'][$languageId] = $asset->getTranslatedLifetime($locale, false);
				}
			}
		);

		// Übersetzungen extrahieren
		foreach ($locales as $languageId => $locale) {
			$data['blocked_content_button_text'][$languageId] = $embedding->getTranslatedBlockedContentButtonText($locale, false);
			$data['blocked_content_text'][$languageId] = $embedding->getTranslatedBlockedContentText($locale, false);
			$data['blocked_content_title'][$languageId] = $embedding->getTranslatedBlockedContentTitle($locale, false);
			$data['collected_data_info'][$languageId] = $embedding->getTranslatedCollectedDataInfo($locale, false);
			$data['description'][$languageId] = $embedding->getTranslatedDescription($locale, false);
			$data['legal_basis'][$languageId] = $embedding->getTranslatedLegalBasis($locale, false);
			$data['legal_basis_supplement'][$languageId] = $embedding->getTranslatedLegalBasisSupplement($locale, false);
			$data['place_of_processing'][$languageId] = $embedding->getTranslatedPlaceOfProcessing($locale, false);
			$data['privacy_policy_url'][$languageId] = $embedding->getTranslatedPrivacyPolicyUrl($locale, false);
			$data['purpose_of_data_collection'][$languageId] = $embedding->getTranslatedPurposeOfDataCollection($locale, false);
		}

		return $data;
	}

	/**
	 * @Route("/new", name="new", methods={"HEAD","GET"})
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function createEmbeddingForm(Request $request, TranslatorInterface $translator): Response
	{
		$csrfToken = CsrfToken::get();
		$embeddingData = [];
		$purposes = Purpose::all();
		$config = Config::getInstance();
		$enabledLanguages = $config->getSystemEmbeddingLanguages();
		/** @var Language|null $defaultLanguage */
		$defaultLanguage = array_reduce($enabledLanguages, function (?Language $carry, Language $language) {
			return $language->getCode() == 'de_DE' ? $language : ($carry ?? $language);
		});
		$defaultLanguageId = $defaultLanguage ? $defaultLanguage->getCode() : 'de_DE';

		$embeddingData = self::prepareNewEmbeddingData($request, $defaultLanguageId);

		$addVar="add_";

		return $this->render('hosting/embedding/new.html.twig', [
			'purposes' => $purposes,
			'languages' => $enabledLanguages,
			'csrfToken' => $csrfToken,
			'embedding' => $this->applyChanges(SystemEmbedding::createNonPersistent(), $embeddingData, $defaultLanguageId),
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'tcfVendors' => TcfVendor::all(),
			'add_var' => $addVar,
		]);
	}

	/**
	 * @Route("/new", name="new_post", methods={"POST"})
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function createEmbedding(Request $request, TranslatorInterface $translator): Response
	{
		$config = Config::getInstance();
		$action = $request->request->getAlnum('action');
		$csrfToken = CsrfToken::get();
		$embeddingData = [];
		$purposes = Purpose::all();
		$enabledLanguages = $config->getSystemEmbeddingLanguages();
		/** @var Language|null $defaultLanguage */
		$defaultLanguage = array_reduce($enabledLanguages, function (?Language $carry, Language $language) {
			return $language->getCode() == 'de_DE' ? $language : ($carry ?? $language);
		});
		$defaultLanguageId = $defaultLanguage ? $defaultLanguage->getCode() : 'de_DE';

		$embedding = $this->applyChanges(SystemEmbedding::create(), $this->unpackPostData($request->request->all()), $defaultLanguageId);

		try {
			$csrfToken->checkRequest($request);

			// ohne Namen nicht speichern
			if (trim($embedding->getName()) === '') {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Namen gespeichert werden');
				return $this->render('hosting/embedding/new.html.twig', [
					'embedding' => $embedding,
					'purposes' => $purposes,
					'regulations' => Regulation::all(),
					'regulationGroups' => RegulationGroup::all(),
					'languages' => $enabledLanguages,
					'csrfToken' => $csrfToken,
				]);
			}

			// ohne Verwendungszweck nicht speichern
			if (!$embedding->getPurpose()) {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Verwendungszweck gespeichert werden');
				return $this->render('hosting/embedding/new.html.twig', [
					'embedding' => $embedding,
					'purposes' => $purposes,
					'regulations' => Regulation::all(),
					'regulationGroups' => RegulationGroup::all(),
					'languages' => $enabledLanguages,
					'csrfToken' => $csrfToken,
					'tcfVendors' => TcfVendor::all(),
				]);
			}

			// Speichern
			$embedding->save();
			SystemEmbeddingAsset::flush();

			$this->addFlash('success', 'Das Embedding "' . $embedding->getName() . '" wurde erfolgreich erstellt!');

			switch ($action) {
				case 'saveClose':
					return $this->redirectToRoute('app_hosting_embeddings_list', [], 303);
				case 'save':
				default:
					return $this->redirectToRoute('app_hosting_embeddings_edit', ['id' => $embedding->getId()], 303);
			}
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		return $this->render('hosting/embedding/new.html.twig', [
			'purposes' => $purposes,
			'languages' => $enabledLanguages,
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'csrfToken' => $csrfToken,
			'formData' => $embeddingData ?? [],
			'tcfVendors' => TcfVendor::all(),
			'add_var' => "",
		]);
	}

	/**
	 * @Route("/new_bulk", name="new_bulk", methods={"POST"})
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws Exception
	 */
	public function createEmbeddingBulk(Request $request, TranslatorInterface $translator): Response
	{
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleName = $defaultLocale->getName();

		// Füge ausgewähltes Embedding aus DB hinzu
		$embeddingsExternalDb = $request->request->get('selectedRemoteIds') ? explode(',', $request->request->get('selectedRemoteIds')) : [];
		foreach($embeddingsExternalDb as $remoteId) {
			$remoteDb = (string)$request->query->get('remoteDb', 'v2.2');
			if ($remoteDb === 'v2') {
				$remoteEmbedding = EmbeddingDbLegacy::fetchEmbedding((int)$remoteId);
			} else {
				$remoteEmbedding = EmbeddingDb::fetchEmbedding((int)$remoteId);
			}

			$remoteEmbeddingData =  self::createRemoteEmbeddingFromDbEntry($remoteEmbedding, (string)$remoteDb);
			if (!empty($remoteEmbeddingData)) {
				$remoteEmbedding = $this->applyChanges(SystemEmbedding::create(), $remoteEmbeddingData, $defaultLocaleName);
				$remoteEmbedding->save();
			}
		}

		SystemEmbeddingAsset::flush();
		$this->addFlash('success', 'Die ausgewählten Embeddings wurden erfolgreich erstellt!');
		return $this->redirectToRoute('app_hosting_embeddings_list', [], 303);
	}

	/**
	 * @Route("/{embedding}/gcm", name="gcm_form")
	 * @Route("/gcm", name="gcm_form_new")
	 *
	 * @param SystemEmbedding|null $embedding
	 * @return Response
	 */
	public function gcmForm(?SystemEmbedding $embedding): Response
	{
		return $this->render('embedding/gcm_form.html.twig', [
			'embedding' => $embedding,
		]);
	}

	/**
	 * @Route("/{id}", methods={"HEAD", "GET"}, name="edit")
	 *
	 * @param string $id
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editEmbeddingForm(string $id, Request $request): Response
	{
		$embedding = SystemEmbedding::find($id);

		if ($embedding === null) {
			return $this->redirectToRoute('app_hosting_embeddings_list');
		}

		$purposes = Purpose::all();

		return $this->render('hosting/embedding/edit.html.twig', [
			'embedding' => $embedding,
			'purposes' => $purposes,
			'languages' => Config::getInstance()->getSystemEmbeddingLanguages(),
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'csrfToken' => CsrfToken::get(),
			'tcfVendors' => TcfVendor::all(),
			'add_var' => "",
		]);
	}

	/**
	 * Führt Daten aus zwei Embedding-Arrays zusammen.
	 *
	 * Die Einträge des zweiten überschreiben dabei die des ersten.
	 * NULL und leere Strings werden nicht berücksichtigt.
	 *
	 * Sonderbehandlung für 'add_asset' -> 'asset'.
	 */
	private static function mergeEmbeddingData(array $data1, array $data2): array
	{
		foreach ($data2 as $key => $value)
		{
			if (is_array($value)) {
				if ($key === 'add_asset' and isset($data1['asset'])) {
					foreach ($value as $asset) {
						$name = $asset['name'] ?? '';
						if ($name == '') {
							continue;
						}
						$assetId = null;
						foreach ($data1['asset'] as $assetKey => $assetData) {
							if (($assetData['name'] ?? '') === $name) {
								$assetId = $assetKey;
								break;
							}
						}
						if ($assetId !== null) {
							$data1['asset'][$assetId] = self::mergeEmbeddingData($data1['asset'][$assetId], $asset);
						} else {
							$data1['add_asset'][] = $asset;
						}
					}
				} else {
					$data1[$key] = self::mergeEmbeddingData($data1[$key] ?? [], $data2[$key]);
				}
			} elseif ($value !== null and $value !== '') {
				$data1[$key] = $value;
			}
		}
		return $data1;
	}

	/**
	 * @Route("/{id}", methods={"PUT"}, name="edit_post")
	 *
	 * @param string $id
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editEmbedding(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$config = Config::getInstance();
		$action = $request->request->getAlnum('action');
		$enabledLanguages = $config->getSystemEmbeddingLanguages();
		/** @var Language|null $defaultLanguage */
		$defaultLanguage = array_reduce($enabledLanguages, function (?Language $carry, Language $language) {
			return $language->getCode() == 'de_DE' ? $language : ($carry ?? $language);
		});
		$defaultLanguageId = $defaultLanguage ? $defaultLanguage->getCode() : 'de_DE';
		$embedding = SystemEmbedding::find($id);
		$hadTcfVendor = ($embedding->getTcfVendor() !== null);

		try {
			CsrfToken::get()->checkRequest($request);

			if ($embedding === null) {
				return $this->redirectToRoute('app_hosting_embeddings_list', [], 303);
			}

			// Embedding-Daten holen und aufbereiten
			$embeddingData = $this->unpackPostData($request->request->all());

			$purpose = Purpose::find($embeddingData['purpose']);

			// Abbrechen wenn zu editierendes Embedding nicht existiert
			SystemEmbedding::beginWrite();
			if (SystemEmbedding::exists($id) == false) {
				return $this->redirectToRoute('app_hosting_embeddings_list');
			}

			// ohne Verwendungszweck nicht speichern
			if (Purpose::exists($embeddingData['purpose']) == false) {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Verwendungszweck gespeichert werden');
				return $this->redirectToRoute('app_hosting_embeddings_edit', ['id' => $embedding->getId()]);
			}

			// Script-Blocking
			$codeBlock = explode("\n", trim((string)($embeddingData['code_block'] ?? '')));
			$codeBlock = array_filter($codeBlock, function ($line) {
				return trim($line) !== '';
			});
			$codeBlock = array_values(array_map(function ($line) {
				return trim($line);
			}, $codeBlock));

			// ohne Namen nicht speichern
			if (trim($embeddingData['name']) === '') {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Namen gespeichert werden');
				return $this->redirectToRoute('app_hosting_embeddings_edit', ['id' => $embedding->getId()]);
			}

			// Wenn AutoUpdate aktiviert wurde, direkt Daten übernehmen
			if (!$embedding->isAutoUpdated() and ($embeddingData['autoUpdated'] ?? false)) {
				switch ($embeddingData['source'] ?? '') {
					case EmbeddingBase::SOURCE_GLOBAL_DB:
						$request->query->set('remoteId', $embeddingData['source_id'] ?? '');
						$request->query->set('remoteDb', 'v2.2');
						break;
					case EmbeddingBase::SOURCE_GLOBAL_DB_LEGACY:
						$request->query->set('remoteId', $embeddingData['source_id'] ?? '');
						$request->query->set('remoteDb', 'v2');
						break;
					default:
						break;
				}
				$remoteData = self::prepareNewEmbeddingData($request, $defaultLanguageId);
				$currentSourceTs = new DateTimeImmutable($embeddingData['source_ts']);
				if (!empty($remoteData['source_ts']) and $remoteData['source_ts'] > $currentSourceTs) {
					$embeddingData = $this->mergeEmbeddingData($embeddingData, $remoteData);
				}
			}

			// Speichern
			$embedding = $this->applyChanges($embedding, $embeddingData, $defaultLanguageId);
			$embedding->save();
			SystemEmbeddingAsset::flush();

			$this->addFlash('success', 'Das Embedding "' . $embeddingData['name'] . '" wurde erfolgreich bearbeitet!');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		switch ($action) {
			case 'saveClose':
				return $this->redirectToRoute('app_hosting_embeddings_list', [], 303);
			case 'save':
			default:
				return $this->redirectToRoute('app_hosting_embeddings_edit', ['id' => $embedding->getId()], 303);
		}
	}

	/**
	 * @Route("/{id}/delete", name="delete_form")
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function deleteEmbeddingForm(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$embedding = SystemEmbedding::find($id);
		if ($embedding === null) {
			return $this->redirectToRoute('app_hosting_embeddings_list');
		}
		$embeddingName = $embedding->getName();

		return $this->render('hosting/embedding/delete.html.twig', [
			'embedding' => $embedding,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}", name="delete", methods={"DELETE"})
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function deleteEmbedding(string $id, Request $request, TranslatorInterface $translator): Response
	{
		SystemEmbedding::beginWrite();
		if (SystemEmbedding::exists($id) == false) {
			return $this->redirectToRoute('app_hosting_embeddings_list');
		}

		$embedding = SystemEmbedding::find($id);
		$embeddingName = $embedding->getName();

		try {
			CsrfToken::get()->checkRequest($request);

			foreach ($embedding->getAssets() as $asset) {
				SystemEmbeddingAsset::delete($asset->getId(), false);
			}
			SystemEmbeddingAsset::flush();

			SystemEmbedding::delete($id);
			$this->addFlash('success', 'Das Embedding "' . $embeddingName . '" wurde erfolgreich gelöscht!');
			return $this->redirectToRoute('app_hosting_embeddings_list');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		return $this->redirectToRoute('app_hosting_embeddings_delete_form', ['id'=>$embedding->getId()]);
	}

	/**
	 * @Route("/{embedding}/importSource", name="import_source")
	 *
	 * @param SystemEmbedding $embedding
	 * @return JsonResponse
	 */
	public function sourceInfo(SystemEmbedding $embedding): JsonResponse
	{
		$request = new Request();

		switch ($embedding->getSource()) {
			case EmbeddingBase::SOURCE_GLOBAL_DB:
				$request->query->set('remoteId', $embedding->getSourceId());
				$request->query->set('remoteDb', 'v2.2');
				break;
			case EmbeddingBase::SOURCE_GLOBAL_DB_LEGACY:
				$request->query->set('remoteId', $embedding->getSourceId());
				$request->query->set('remoteDb', 'v2');
				break;
			default:
				break;
		}

		$config = Config::getInstance();
		$enabledLanguages = $config->getSystemEmbeddingLanguages();
		/** @var Language|null $defaultLanguage */
		$defaultLanguage = array_reduce($enabledLanguages, function (?Language $carry, Language $language) {
			return $language->getCode() == 'de_DE' ? $language : ($carry ?? $language);
		});
		$defaultLanguageId = $defaultLanguage ? $defaultLanguage->getCode() : 'de_DE';

		return $this->json(self::prepareNewEmbeddingData($request, $defaultLanguageId));
	}
}
