<?php

namespace App\Controller\Hosting;

use App\Component\LocaleManagerTrait;
use App\Component\LocaleOverrideHelper;
use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Controller\HostingController;
use App\Model\Domain;
use App\Model\Language;
use App\Model\LocaleOverride;
use App\Model\Model;
use App\Model\User;
use App\Model\UserLocale;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/hosting/locale_override", name="app_hosting_locale_override_")
 */
class LocaleOverrideController extends HostingController
{
	use LocaleManagerTrait;

	/**
	 * @Menu("Override default texts", icon="fa-file-text-o", group="Management")
	 * @Route("", name="index", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function index(Request $request, TranslatorInterface $translator): Response
	{
		$overrides = LocaleOverride::all();
		$fieldGroups = self::buildFieldGroupsForOverrides($overrides, $translator);

		// Access overrides by their proper name
		$overrides = array_combine(
			array_map(function (LocaleOverride $override) { return $override->getName(); }, $overrides),
			$overrides
		);

		if (count($overrides) > 0) {
			$language = current(array_filter(array_merge(array_fill_keys([
				$request->getLocale(),
				explode('_', $request->getLocale(), 2)[0],
				'en_US',
				'de_DE',
			], null), reset($overrides)->getLanguages())));
			assert($language instanceof Language);
		}
		else {
			$language = null;
		}

		return $this->render('hosting/locale_override/index.html.twig', [
			'fieldGroups' => $fieldGroups,
			'overrides' => $overrides,
			'language' => $language,
		]);
	}

	/**
	 * @Route("/override/{fieldName}", name="edit", methods={"HEAD", "GET"})
	 * @param string $fieldName
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function edit(string $fieldName, TranslatorInterface $translator): Response
	{
		$field = self::getStandardField($fieldName, $translator);

		// Bail if the requested field doesn't exist
		if (!$field) {
			return $this->redirectToRoute('app_hosting_locale_override_index', [], 303);
		}

		// Use the existing override or create a temporary instance
		$override = LocaleOverride::find($fieldName) ?? LocaleOverride::createNonPersistent($fieldName);

		// Get languages that are present in the skeleton data
		$translations = UserLocale::readSkeletonData();
		$languages = array_intersect_key(Language::all(), $translations);

		// Get translations for the requested field
		$translations = array_map(function (array $translation) use ($field) {
			return self::getSkeletonValue($field['field_id'], $translation);
		}, $translations);

		return $this->render('hosting/locale_override/edit.html.twig', [
			'field' => $field,
			'languages' => $languages,
			'override' => $override,
			'translations' => $translations,
		]);
	}

	/**
	 * @Route("/override/{fieldName}", name="save", methods={"POST"})
	 * @param string $fieldName
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function save(string $fieldName, Request $request, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		if (self::getStandardField($fieldName, $translator) === null) {
			return $this->redirectToRoute('app_hosting_locale_override_index', [], 303);
		}

		$languages = Language::all();

		// Use the provided translations for known languages
		$translations = array_intersect_key($request->request->all('translations'), $languages);

		if (empty($translations)) {
			return $this->redirectToRoute('app_hosting_locale_override_index', [], 303);
		}

		// Update the existing override or create a new instance
		$override = LocaleOverride::find($fieldName) ?? LocaleOverride::create($fieldName);

		foreach ($translations as $languageCode => $translation) {
			$override->setTranslation($languages[$languageCode], (string)$translation);
		}

		// Write changes to database
		$override->save();

		$user = User::loggedInUser();
		assert($user instanceof User);

		$logger->logInfo('LocaleOverride', "Item updated: \"{$override->getName()}\" by user {$user->getId()}");

		$this->addFlash('success', $translator->trans('The settings have been saved.'));

		switch ($request->request->getAlpha('action')) {
			case 'saveClose':
				return $this->redirectToRoute('app_hosting_locale_override_index', [], 303);
			case 'save':
			default:
				return $this->redirectToRoute('app_hosting_locale_override_edit', ['fieldName' => $override->getName()], 303);
		}
	}

	/**
	 * @Route("/override/{fieldName}/delete", name="delete_confirm", methods={"HEAD", "GET"})
	 * @param string $fieldName
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deleteConfirmDialog(string $fieldName, TranslatorInterface $translator): Response
	{
		$override = LocaleOverride::find($fieldName);
		if (!$override) {
			return $this->redirectToRoute('app_hosting_locale_override_index', [], 303);
		}

		return $this->render('hosting/locale_override/delete.html.twig', [
			'field' => self::getStandardField($fieldName, $translator),
			'override' => $override,
		]);
	}

	/**
	 * @Route("/override/{fieldName}", name="delete", methods={"DELETE"})
	 * @param string $fieldName
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function delete(string $fieldName, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$override = LocaleOverride::find($fieldName);
		if (!$override) {
			return $this->redirectToRoute('app_hosting_locale_override_index', [], 303);
		}

		$user = User::loggedInUser();
		assert($user instanceof User);

		$logger->logInfo('LocaleOverride', "Item deleted: \"{$override->getName()}\" by user {$user->getId()}");
		LocaleOverride::delete($override->getName());

		$this->addFlash('success', $translator->trans('Item has been deleted.'));
		return $this->redirectToRoute('app_hosting_locale_override_index', [], 303);
	}

	/**
	 * @Route("/select_field", name="select_field", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function selectField(Request $request, TranslatorInterface $translator): Response
	{
		// Provide remaining fields that haven't been overridden yet
		$fieldGroups = self::buildFieldGroupsAndSubtractOverrides(LocaleOverride::all(), $translator);

		// Get localization data
		$backendLanguage = $request->getLocale();
		$translations = UserLocale::readSkeletonData();

		/** @var array[]|array $translation */
		$translation = $translations[$backendLanguage]
			?? $translations[explode('_', $backendLanguage, 2)[0]]
			?? $translations['en_US']
			?? $translations['de_DE']
			?? reset($translations);
		assert(is_array($translation));

		$translation = array_map(function (array $field) use ($translation) {
			return self::getSkeletonValue($field['field_id'], $translation);
		}, array_merge([], ...array_values($fieldGroups)));

		return $this->render('hosting/locale_override/select_field.html.twig', [
			'fieldGroups' => $fieldGroups,
			'translation' => $translation,
		]);
	}

	/**
	 * @Route("/deploy", name="deploy_confirm", methods={"HEAD", "GET"})
	 * @return Response
	 */
	public function deployConfirmDialog(): Response
	{
		return $this->render('hosting/locale_override/deploy.html.twig');
	}

	/**
	 * @Route("/deploy", name="deploy", methods={"POST"})
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deploy(TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$overrides = LocaleOverride::all();

		$user = User::loggedInUser();
		assert($user instanceof User);

		foreach (User::getClients() as $client) {
			// Switch the user context
			User::setApiUser($client->getApiKey());
			Model::closeUserModels();
			Domain::switchToUser($client);

			LocaleOverrideHelper::applyOverridesToUserLocales($overrides);

			foreach (Domain::all() as $domain) {
				Domain::setDomainForRequest($domain->getId());
				Model::closeDomainModels();

				LocaleOverrideHelper::applyOverridesToDomainLocales($overrides);
			}
		}

		// Switch back to the original user context
		Domain::setDomainForRequest('');
		User::setApiUser($user->getApiKey());
		Model::closeUserModels();

		$logger->logInfo('LocaleOverride', "Translations deployed by user {$user->getId()}");

		$this->addFlash('success', $translator->trans('The translations have been distributed to all users and domains.'));
		return $this->redirectToRoute('app_hosting_locale_override_index', [], 303);
	}

	/**
	 * @param LocaleOverride[] $overrides
	 * @param TranslatorInterface $translator
	 * @return array
	 */
	private static function buildFieldGroupsAndSubtractOverrides(array $overrides, TranslatorInterface $translator): array
	{
		$fieldGroups = self::getStandardFieldGroups($translator);
		$overrideNames = array_map(function (LocaleOverride $override) {
			return $override->getName();
		}, $overrides);

		// Filter out locale fields that have already been overridden
		array_walk($fieldGroups, function (array &$group) use ($overrideNames) {
			$group = array_diff_key($group, array_flip($overrideNames));
		});

		// Filter out empty field groups
		return array_filter($fieldGroups);
	}

	/**
	 * @param LocaleOverride[] $overrides
	 * @param TranslatorInterface $translator
	 * @return array
	 */
	private static function buildFieldGroupsForOverrides(array $overrides, TranslatorInterface $translator): array
	{
		$fieldGroups = self::getStandardFieldGroups($translator);
		$overrideNames = array_map(function (LocaleOverride $override) {
			return $override->getName();
		}, $overrides);

		// Filter out locale fields that don't match an override
		array_walk($fieldGroups, function (array &$group) use ($overrideNames) {
			$group = array_intersect_key($group, array_flip($overrideNames));
		});

		// Filter out empty field groups
		return array_filter($fieldGroups);
	}
}
