<?php

namespace App\Controller\Hosting;

use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Component\Menu\Manager as MenuManager;
use App\Component\Menu\Model\MenuGroup;
use App\Component\Menu\Model\MenuItem;
use App\Component\Pagination;
use App\Controller\HostingController;
use App\Event\AccountUpdateEvent;
use App\Exception\CsrfTokenException;
use App\Exception\UserAlreadyExistsException;
use App\Model\Country;
use App\Model\CsrfToken;
use App\Model\User;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/hosting/manager", name="app_hosting_manager_")
 */
class Manager extends HostingController
{
	public const EVENT_VALIDATE_INPUT = 'ccm19.hosting.manager.validateInput';
	public const EVENT_UPDATE = 'ccm19.hosting.manager.update';
	public const EVENT_DELETE = 'ccm19.hosting.manager.delete';

	private static $mandatoryPrivileges = [
		'app_account',
		'app_hosting_dashboard',
	];

	/**
	 * @Menu("Managers", group="System", icon="fa-users", order=210)
	 * @Route("", methods={"HEAD", "GET"}, name="index")
	 * @param Request $request
	 * @return Response
	 */
	public function listManagers(Request $request): Response
	{
		if ($searchTerm = trim((string)$request->query->get('search'))) {
			$users = User::search($searchTerm, User::ROLE_MANAGER);
			$managersExist = count(User::byRole(User::ROLE_MANAGER)) > 0;
		} else {
			$users = User::byRole(User::ROLE_MANAGER);
			$managersExist = count($users) > 0;
		}

		// Filtering by role should be obsolete
		$users = array_filter($users, function (User $user) {
			return $user->isManager();
		});

		// Sort by username
		usort($users, function (User $a, User $b) {
			return strnatcasecmp($a->getUsername(), $b->getUsername());
		});

		$pagination = new Pagination($users);
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_hosting_manager_index', [], 303);
		}

		return $this->render('hosting/manager/index.html.twig', [
			'managersExist' => $managersExist,
			'pagination' => $pagination,
			'searchTerm' => $searchTerm,
		]);
	}

	/**
	 * @Route("/new", methods={"GET","HEAD"}, name="new_form")
	 * @param MenuManager $menuManager
	 * @return Response
	 */
	public function createManagerForm(MenuManager $menuManager): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		// Create temporary manager to prefill the form
		$manager = User::createNonPersistent()
			->setRole(User::ROLE_MANAGER)
		;

		return $this->render('hosting/manager/new.html.twig', [
			'clientGroups' => $user->getAccessibleClientGroups(),
			'countries' => Country::all(),
			'csrfToken' => CsrfToken::get(),
			'manager' => $manager,
			'mandatoryPrivileges' => self::$mandatoryPrivileges,
			'menuGroups' => self::getMenuGroupsForCurrentUser($menuManager),
		]);
	}

	/**
	 * @Route("/new", methods={"POST"}, name="new")
	 * @param Request $request
	 * @param MenuManager $menuManager
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function createManager(Request $request, MenuManager $menuManager, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$username = trim((string)$request->request->get('username'));
		$password = trim((string)$request->request->get('password'));
		$usePassword = !$request->request->getBoolean('noPassword');
		$emailAddress = trim((string)$request->request->get('emailAddress'));

		$manager = null;
		$success = false;

		if ($username == '' || ($usePassword && $password == '') || $emailAddress == '') {
			$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
		}
		elseif (Utils::validateEmailAddress($emailAddress) == false) {
			$this->addFlash('danger', $translator->trans('Invalid email address, check format.'));
		}
		// Give other components a chance to validate custom properties
		elseif ($errors = Utils::filterWithEvent(self::EVENT_VALIDATE_INPUT, [])) {
			foreach ($errors as $error) {
				$this->addFlash('danger', $error);
			}
		}
		else {
			try {
				$manager = User::create($username);
				$success = true;
			} catch (UserAlreadyExistsException $e) {
				User::reload();
				$this->addFlash('danger', $translator->trans('Username %username% already exists.', [
					'%username%' => $username,
				]));
			}
		}

		// Use a temporary user if there was an error
		$manager = $manager ?? User::createNonPersistent();

		$manager
			->setRole(User::ROLE_MANAGER)
			->setUsername($username)
			->setEmailAddress($emailAddress)
			->setFirstName(trim((string)$request->request->get('firstName')))
			->setLastName(trim((string)$request->request->get('lastName')))
			->setStrNr(trim((string)$request->request->get('streetNr')))
			->setPlz(trim((string)$request->request->get('plz')))
			->setCity(trim((string)$request->request->get('city')))
			->setCountry(Country::find($request->request->getAlpha('country')))
			->setCompany(trim((string)$request->request->get('company')))
			->setVatId(trim((string)$request->request->get('vatId')))
			->setActive($request->request->getBoolean('active'))
		;

		if ($usePassword) {
			$manager->setPassword($password);
		} else {
			$manager->unsetPassword();
		}

		// Update full access to all clients in accordance with the current user's privileges
		if ($user->hasAccessToAllClients()) {
			$manager->setAccessToAllClients($request->request->getBoolean('accessToAllClients'));
		}

		// Update access to individual clients if the manager doesn't already have full access
		if ($manager->hasAccessToAllClients() == false) {
			$manager->setClientGroups(array_intersect_key(
				$user->getAccessibleClientGroups(),
				array_flip($request->request->all('clientGroups'))
			));
		}

		// Update all privileges in accordance with the current user's privileges
		if ($user->hasAllPrivileges()) {
			$manager->setAllPrivileges($request->request->getBoolean('allPrivileges'));
		}

		// Update individual privileges if the manager doesn't already have all privileges
		if ($manager->hasAllPrivileges() == false) {
			$manager->setPrivileges(array_merge(
				self::$mandatoryPrivileges,
				array_intersect(self::getPrivilegesForCurrentUser($menuManager), $request->request->all('privileges'))
			));
		}

		try {
			CsrfToken::get()->checkRequest($request);

			if ($success) {
				// Give other components a chance to update custom properties. However, they must not save the manager.
				Utils::notifyWithEvent(self::EVENT_UPDATE, $manager, ['flush'=>true]);

				$manager->save();
				$logger->logSuccess('Hosting::Managers',
					"User \"{$user->getUsername()}\" ({$user->getId()}) has created \"{$manager->getUsername()}\" ({$manager->getId()})"
				);

				$this->addFlash('success', $translator->trans('The user account :username has been created.', [
					':username' => $manager->getUsername(),
				]));
				return $this->redirectToRoute('app_hosting_manager_index', [], 303);
			}
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		// Give other components a chance to update custom properties. However, changes must not be flushed.
		Utils::notifyWithEvent(self::EVENT_UPDATE, $manager, ['flush'=>false]);

		return $this->render('hosting/manager/new.html.twig', [
			'clientGroups' => $user->getAccessibleClientGroups(),
			'countries' => Country::all(),
			'csrfToken' => CsrfToken::get(),
			'manager' => $manager,
			'mandatoryPrivileges' => self::$mandatoryPrivileges,
			'menuGroups' => self::getMenuGroupsForCurrentUser($menuManager),
		]);
	}

	/**
	 * @Route("/{id}", methods={"GET","HEAD"}, name="edit_form")
	 * @param string $id
	 * @param MenuManager $menuManager
	 * @return Response
	 */
	public function updateManagerForm(string $id, MenuManager $menuManager): Response
	{
		$manager = User::byId($id);
		if (!$manager || !$manager->isManager()) {
			return $this->redirectToRoute('app_hosting_manager_index', [], 303);
		}

		$user = User::loggedInUser();
		assert($user instanceof User);

		// Redirect the manager to their personal account settings
		if ($manager->getId() == $user->getId()) {
			return $this->redirectToRoute('app_account', [], 303);
		}

		return $this->render('hosting/manager/edit.html.twig', [
			'clientGroups' => $user->getAccessibleClientGroups(),
			'countries' => Country::all(),
			'csrfToken' => CsrfToken::get(),
			'manager' => $manager,
			'mandatoryPrivileges' => self::$mandatoryPrivileges,
			'menuGroups' => self::getMenuGroupsForCurrentUser($menuManager),
		]);
	}

	/**
	 * @Route("/{id}", methods={"PUT"}, name="edit")
	 * @param string $id
	 * @param Request $request
	 * @param MenuManager $menuManager
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function updateManager(string $id, Request $request, MenuManager $menuManager, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$manager = User::byId($id);
		if (!$manager || !$manager->isManager()) {
			return $this->redirectToRoute('app_hosting_manager_index', [], 303);
		}

		$user = User::loggedInUser();
		assert($user instanceof User);

		// Redirect the manager to their personal account settings
		if ($manager->getId() == $user->getId()) {
			return $this->redirectToRoute('app_account', [], 303);
		}

		$username = trim((string)$request->request->get('username'));
		$password = trim((string)$request->request->get('password'));
		$usePassword = !$request->request->getBoolean('noPassword');
		$emailAddress = trim((string)$request->request->get('emailAddress'));

		$success = false;

		if ($username == '' || $emailAddress == '') {
			$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
		}
		elseif (mb_strtolower($manager->getUsername()) != mb_strtolower($username) && User::exists($username)) {
			$this->addFlash('danger', $translator->trans('Username %username% already exists.', [
				'%username%' => $username,
			]));
		}
		elseif (Utils::validateEmailAddress($emailAddress) == false) {
			$this->addFlash('danger', $translator->trans('Invalid email address, check format.'));
		}
		// Give other components a chance to validate custom properties
		elseif ($errors = Utils::filterWithEvent(self::EVENT_VALIDATE_INPUT, [])) {
			foreach ($errors as $error) {
				$this->addFlash('danger', $error);
			}
		}
		else {
			$success = true;
		}

		$manager
			->setUsername($username)
			->setEmailAddress($emailAddress)
			->setFirstName(trim((string)$request->request->get('firstName')))
			->setLastName(trim((string)$request->request->get('lastName')))
			->setStrNr(trim((string)$request->request->get('streetNr')))
			->setPlz(trim((string)$request->request->get('plz')))
			->setCity(trim((string)$request->request->get('city')))
			->setCountry(Country::find($request->request->getAlpha('country')))
			->setCompany(trim((string)$request->request->get('company')))
			->setVatId(trim((string)$request->request->get('vatId')))
			->setActive($request->request->getBoolean('active'))
		;

		// Event auslösen
		$event = new AccountUpdateEvent($manager, AccountUpdateEvent::CAUSE_ADMIN);
		$event->setAccountInfoChanged();
		if ($usePassword) {
			// Keep the current password if the new password is omitted
			if ($password != '') {
				$manager->setPassword($password);
				$event->setPasswordChanged();
			}
		} else {
			$manager->unsetPassword();
		}

		// Update full access to all clients in accordance with the current user's privileges
		if ($user->hasAccessToAllClients()) {
			$manager->setAccessToAllClients($request->request->getBoolean('accessToAllClients'));
		}

		// Update access to individual clients if the manager doesn't already have full access
		if ($manager->hasAccessToAllClients() == false) {
			$accessibleClientGroups = $user->getAccessibleClientGroups();
			$manager->setClientGroups(array_merge(
				array_diff_key($manager->getClientGroups(), $accessibleClientGroups), // preserve
				array_intersect_key($accessibleClientGroups, array_flip($request->request->all('clientGroups'))) // add
			));
		}

		// Update all privileges in accordance with the current user's privileges
		if ($user->hasAllPrivileges()) {
			$manager->setAllPrivileges($request->request->getBoolean('allPrivileges'));
		}

		// Update individual privileges if the manager doesn't already have all privileges
		if ($manager->hasAllPrivileges() == false) {
			$grantablePrivileges = self::getPrivilegesForCurrentUser($menuManager);
			$preservedPrivileges = array_diff($manager->getPrivileges(), $grantablePrivileges);
			$manager->setPrivileges(array_merge(
				self::$mandatoryPrivileges,
				$preservedPrivileges,
				array_intersect($grantablePrivileges, $request->request->all('privileges'))
			));
		}

		try {
			CsrfToken::get()->checkRequest($request);

			if ($success) {
				// Give other components a chance to update custom properties. However, they must not save the manager.
				Utils::notifyWithEvent(self::EVENT_UPDATE, $manager, ['flush'=>true]);
				Utils::dispatchEvent(AccountUpdateEvent::NAME_PRE_SAVE, $event);

				$manager->save();
				$logger->logInfo('Hosting::Managers',
					"User \"{$user->getUsername()}\" ({$user->getId()}) has updated \"{$manager->getUsername()}\" ({$manager->getId()})"
				);
				Utils::dispatchEvent(AccountUpdateEvent::NAME_POST_SAVE, $event);

				$this->addFlash('success', $translator->trans('The user account :username has been updated.', [
					':username' => $manager->getUsername(),
				]));

				switch ($request->request->getAlpha('action')) {
					case 'saveClose':
						return $this->redirectToRoute('app_hosting_manager_index', [], 303);
					case 'save':
					default:
						return $this->redirectToRoute('app_hosting_manager_edit_form', ['id' => $manager->getId()], 303);
				}
			}
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		// Give other components a chance to update custom properties. However, changes must not be flushed.
		Utils::notifyWithEvent(self::EVENT_UPDATE, $manager, ['flush'=>false]);

		return $this->render('hosting/manager/edit.html.twig', [
			'clientGroups' => $user->getAccessibleClientGroups(),
			'countries' => Country::all(),
			'csrfToken' => CsrfToken::get(),
			'manager' => $manager,
			'mandatoryPrivileges' => self::$mandatoryPrivileges,
			'menuGroups' => self::getMenuGroupsForCurrentUser($menuManager),
		]);
	}

	/**
	 * @Route("/{id}/delete", methods={"GET","HEAD"}, name="delete_confirm")
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deleteManagerConfirmDialog(string $id, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$manager = User::byId($id);
		if (!$manager || !$manager->isManager()) {
			return $this->redirectToRoute('app_hosting_manager_index', [], 303);
		}
		elseif ($manager->getId() == $user->getId()) {
			$this->addFlash('danger', $translator->trans('You can\'t delete yourself, it\'s simply impossible.'));
			return $this->redirectToRoute('app_hosting_manager_index', [], 303);
		}

		return $this->render('hosting/manager/delete.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'manager' => $manager,
		]);
	}

	/**
	 * @Route("/{id}", methods={"DELETE"}, name="delete")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deleteManager(string $id, Request $request, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$user = User::loggedInUser();
		assert($user instanceof User);

		$manager = User::byId($id);
		if (!$manager || !$manager->isManager()) {
			return $this->redirectToRoute('app_hosting_manager_index', [], 303);
		}
		elseif ($manager->getId() == $user->getId()) {
			$this->addFlash('danger', $translator->trans('You can\'t delete yourself, it\'s simply impossible.'));
			return $this->redirectToRoute('app_hosting_manager_index', [], 303);
		}

		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('app_hosting_manager_index', [], 303);
		}

		$username = $manager->getUsername();

		// Give other components a chance to delete their custom data
		Utils::notifyWithEvent(self::EVENT_DELETE, $manager);

		User::delete($manager->getId());

		$logger->logSuccess('Hosting::Managers',
			"User \"{$user->getUsername()}\" ({$user->getId()}) has deleted \"$username\" ($id)"
		);

		$this->addFlash('success', $translator->trans('Account %username% has been terminated.', [
			'%username%' => $username,
		]));

		return $this->redirectToRoute('app_hosting_manager_index', [], 303);
	}

	/**
	 * Returns all menu items, which the user can access, in their respective menu groups.
	 * @param MenuManager $menuManager
	 * @return MenuGroup[]
	 */
	private static function getMenuGroupsForCurrentUser(MenuManager $menuManager): array
	{
		return array_merge_recursive(
			$menuManager->getMenu(),
			$menuManager->getMetaMenu()
		);
	}

	/**
	 * Lists all privileges that the current user can grant to other managers.
	 * @param MenuManager $menuManager
	 * @return string[]
	 */
	private static function getPrivilegesForCurrentUser(MenuManager $menuManager): array
	{
		$privileges = array_map(function (MenuItem $item) {
			return rtrim($item->routeGroup ?: $item->route, '*');
		}, call_user_func_array('array_merge', array_values(array_map(function (MenuGroup $group) {
			return iterator_to_array($group->iterateItems(), false);
		}, self::getMenuGroupsForCurrentUser($menuManager)))));

		natcasesort($privileges);
		return array_values($privileges);
	}
}
