<?php

namespace App\Controller\Hosting\Settings;

use App\Component\Menu\Annotation\Menu;
use App\Component\Update;
use App\Controller\HostingController;
use App\Config;
use App\Model\SystemEmbedding;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Translation\Loader\YamlFileLoader;
use Symfony\Component\Translation\Loader\LoaderInterface;
use Symfony\Component\Translation\Dumper\DumperInterface;
use Symfony\Component\Validator\Constraints\Length;
use Symfony\Component\Validator\Constraints\Url;
use Symfony\Component\Validator\ConstraintViolation;
use Symfony\Component\Validator\Validator\ValidatorInterface;
use Symfony\Contracts\Translation\TranslatorInterface;
use Symfony\Component\Translation\MessageCatalogue;
use RuntimeException;

/**
 * @Route("/hosting/settings/branding", name="app_hosting_settings_branding_")
 */
class Branding extends HostingController
{
	private $translationLoader;
	private $yamlLoader;

	/**
	 * @required
	 */
	public function setTranslationLoader(LoaderInterface $loader): void
	{
		$this->translationLoader = $loader;
	}

	/**
	 * @required
	 */
	public function setYamlTranslationLoader(YamlFileLoader $loader): void
	{
		$this->yamlLoader = $loader;
	}

	/**
	 * @Menu("Branding", group="Whitelabel", routeGroup="app_hosting_settings_branding_form", icon="fa-building", order=225)
	 * @Route("", methods={"HEAD","GET"}, name="form")
	 * @param Request $request
	 * @return Response
	 */
	public function settingsForm(Request $request): Response
	{
		$config 		= 	Config::getInstance();

		$formData = [
			'appName' => $config->getAppName(),
			'poweredByText' => Utils::getPoweredByText(),
			'poweredByTooltip' => Utils::getPoweredByTooltip(),
			'poweredByUrl' => Utils::getPoweredByUrl(),
			'themeBack_customCss' => $config->getBackendCustomCss(),
			'Search' => "",
			'Replace' => "",
			'docsurl' => Utils::getDocsUrl(),
			"enablehelp" => Utils::isHelpEnabled(),
			"unsplashHref" =>  Utils::getUnsplashHref(),
			"unsplashName" => Utils::getUnsplashName(),
			"headlineColor" => Utils::getHeadlineColor(),
			"unsplashLinkColor" => Utils::getUnsplashLinkColor(),
			"unsplashLinkBackgroundColor" => Utils::getUnsplashLinkBackgroundColor(),
			"unsplashLinkBackgroundColorOpacity" => Utils::getUnsplashLinkBackgroundColorOpacity(),
			"portalName" => Utils::getPortalName(),
			"portalURL" => Utils::getPortalURL(),
			"imprintURL" => $config->getImprintURL(),
			"dataPrivacyURL" => $config->getDataPrivacyURL(),
			"isWhiteLabel" => $config->isWhiteLabel(),
			"wlEmbedding" => Utils::getWhiteLabelEmbedding(),
		];

		return $this->render('hosting/settings/branding/index.html.twig', [
			'formData' => $formData,
			'langSelect' => (string)$request->request->get('langSelect'),
			"isWhiteLabel" => $config->isWhiteLabel(),
			"embeddings" => SystemEmbedding::all(),
		]);
	}

	/**
	 * @Menu("Layout & CSS", group="Whitelabel", icon="fa-desktop", routeGroup="app_hosting_settings_branding_layout", order=235)
	 * @Route("/layout", methods={"HEAD","GET"}, name="layout")
	 * @param Request $request
	 * @return Response
	 */
	public function settingsLayout(Request $request): Response
	{
		$config 		= 	Config::getInstance();

		$formData = [
			'themeBack_customCss' => $config->getBackendCustomCss() ?? '.specialCss {} ',
		];

		return $this->render('hosting/settings/branding/index_layout.html.twig', [
			'formData' => $formData,
			'langSelect' => (string)$request->request->get('langSelect'),
			"isWhiteLabel" => $config->isWhiteLabel(),
		]);
	}

	/**
	 * @Menu("Content / Text", group="Whitelabel", icon="fa-file-text", routeGroup="app_hosting_settings_branding_content",  order=245)
	 * @Route("/l10n", methods={"HEAD","GET"}, name="content")
	 * @param Request $request
	 * @return Response
	 */
	public function settingsContent(Request $request): Response
	{
		$config 		= 	Config::getInstance();

		//powerdByUrl
		$formData = [
			'Search' => "",
			'Replace' => "",
		];

		return $this->render('hosting/settings/branding/index_content.html.twig', [
			'formData' => $formData,
			'langSelect' => (string)$request->request->get('langSelect'),
			"isWhiteLabel" => $config->isWhiteLabel(),
		]);
	}

	/**
	 * @Route("/content_post", methods={"POST"}, name="content_post")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param DumperInterface $dumper
	 * @param Update $updater
	 * @return Response
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function settingsPost(Request $request, TranslatorInterface $translator, DumperInterface $dumper, Update $updater): Response
	{
		$config 		= 	Config::getInstance();
		if (!$config->isWhiteLabel()) {
			$this->addFlash('error', $translator->trans(
				'You are not allowed to save these settings!'));
			return $this->redirectToRoute('app_hosting_settings_branding_form', [], 303);
		}

		if ((string)$request->request->get('action') === "edit")
		{
			// Übersetzungen
			$lang = (string)$request->request->get('langSelect');
			$translations = $this->getTranslationData($lang);
			$translationResult = [];
			$translationChanges = (array)$request->request->get('yaml');
			$catalogue = new MessageCatalogue($lang);
			// Bestimmen, was sich im Vergleich zum Original geändert hat
			foreach ($translationChanges as $key => $value) {
				$key = urldecode($key);
				$value = str_replace("\r\n", "\n", $value);
				$stockValue = str_replace("\r\n", "\n", $translations[$key]['stock'] ?? '');
				if ($value !== '' and $value !== $stockValue) {
					$catalogue->set($key, $value);
				}
			}

			$dumper->dump($catalogue, ['path'=>Utils::getVarDir() . '/translations']);

			// Cache komplett leeren... nicht immer nötig, aber funktioniert
			$updater->deleteFullCacheOnExit();
		}

		$this->addFlash('success', $translator->trans(
			'The settings were saved successfully!'));
		return $this->redirectToRoute('app_hosting_settings_branding_form', [], 303);
	}



	/**
	 * @Route("/layout_post", methods={"POST"}, name="layout_post")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param Update $updater
	 * @return Response
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function layoutPost(Request $request, TranslatorInterface $translator, Update $updater): Response
	{
		$config 		= 	Config::getInstance();

		if (!$config->isWhiteLabel()) {
			$this->addFlash('error', $translator->trans(
				'You are not allowed to save these settings!'));
			return $this->redirectToRoute('app_hosting_settings_branding_layout', [], 303);
		}

		//neuer Kram

		$config->setBackendCustomCss((string)$request->request->get('themeBack_customCss'));
		$config->flush();

		$this->addFlash('success', $translator->trans(
			'The settings were saved successfully!'));
		return $this->redirectToRoute('app_hosting_settings_branding_layout', [], 303);
	}

	/**
	 * @Route("/whitelabel_reset", methods={"GET"}, name="whitelabel_reset")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param DumperInterface $dumper
	 * @param Update $updater
	 * @return Response
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function whitelabelReset(Request $request, TranslatorInterface $translator, DumperInterface $dumper, Update $updater): Response
	{
		$config 		= 	Config::getInstance();
		$config->unset('unsplashHref');
		$config->unset('unsplashName');
		$config->unset('headlineColor');
		$config->unset('unsplashLinkColor');
		$config->unset('unsplashLinkBackgroundColor');
		$config->unset('unsplashLinkBackgroundColorOpacity');
		$config->unset('portalName');
		$config->unset('portalURL');
		$config->unset('imprintURL');
		$config->unset('dataPrivacyURL');
		$config->unset('docsurl');
		$config->unset('enablehelp');
		$config->unset('poweredByText');
		$config->unset('poweredByTooltip');
		$config->unset('poweredByUrl');
		$config->unset('loginLogo');
		$config->unset('imageFrontend');
		$config->unset('backendLogo');
		$config->unset('wlEmbedding');
		$config->flush();

		$this->addFlash('info', $translator->trans(
			'The settings were resetted successfully!'));
		return $this->redirectToRoute('app_hosting_settings_branding_form', [], 303);
	}

	/**
	 * @Route("/settings_data_post", methods={"POST"}, name="settings_data_post")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param DumperInterface $dumper
	 * @param Update $updater
	 * @param ValidatorInterface $validator
	 * @return Response
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function settingsDataPost(Request $request, TranslatorInterface $translator, DumperInterface $dumper, Update $updater,ValidatorInterface $validator): Response
	{
		$config 		= 	Config::getInstance();

		if($config->isWhiteLabel()) {
			/** @var UploadedFile[] $images */
			$images = array_filter(array_intersect_key($request->files->all(), array_flip([
				'favicon',
				'imageFrontend',
				'imageLogin',
				'image',
			])));

			// Auf Probleme beim Upload prüfen
			foreach ($images as $file) {
				try {
					$this->handleUploadErrors($file, $translator);
				} catch (RuntimeException $e) {
					return $this->settingsForm($request);
				}
			}

			// Bilder speichern
			foreach ($images as $name => $file) {
				switch ($name) {
					case 'favicon':
						$config->setCustomFavicon($file);
						break;
					case 'imageFrontend':
						$config->setCustomFrontendImage($file);
						break;
					case 'imageLogin':
						$config->setCustomLoginLogo($file);
						break;
					case 'image':
						$config->setCustomBackendLogo($file);
						break;
					default:
						break;
				}
			}
		}

		$domName = (string)$request->request->get('docsurl');
		$constraints = [
			new Url(['relativeProtocol' => true]), // Allow relative protocol
			new Length(['min' => 12, 'allowEmptyString' => false]), // Min len
		];
		$violations = $validator->validate($domName, $constraints);

		if ($violations->count()) {
			foreach ($violations as $violation) {
				if ($violation instanceof ConstraintViolation) {
					// Könnte man später mal nutzen: $message = $violation->getMessage();
					$this->addFlash('danger', $translator->trans('Please enter a valid URL or domain.'));
					return $this->redirectToRoute('app_hosting_settings_branding_form', ['switch'=>$request->query->get('switch')]);
				}
			}
		} else {

			if($config->isWhiteLabel()) {
				$config->setAppName((string)$request->request->get('appName'));
				$config->set('unsplashHref', (string)$request->request->get('unsplashHref'));
				$config->set('unsplashName', (string)$request->request->get('unsplashName'));
				$config->set('headlineColor', (string)$request->request->get('headlineColor'));
				$config->set('unsplashLinkColor', (string)$request->request->get('unsplashLinkColor'));
				$config->set('unsplashLinkBackgroundColor', (string)$request->request->get('unsplashLinkBackgroundColor'));
				$config->set('unsplashLinkBackgroundColorOpacity', (string)$request->request->get('unsplashLinkBackgroundColorOpacity'));
				$config->set('portalName', (string)$request->request->get('portalName'));
				$config->set('portalURL', (string)$request->request->get('portalURL'));

				$imprintUrl = (string)$request->request->get('imprintURL');
				$dataPrivacyURL = (string)$request->request->get('dataPrivacyURL');
				$urlPattern = '#^https?://.+#';
				if (strlen($imprintUrl) > 0 && !preg_match($urlPattern, $imprintUrl)) {
					$this->addFlash('danger', $translator->trans('The link entered does not match the required format.'));
				} else {
					$config->set('imprintURL', $imprintUrl);
				}
				if (strlen($dataPrivacyURL) > 0 && !preg_match($urlPattern, $dataPrivacyURL)) {
					$this->addFlash('danger', $translator->trans('The link entered does not match the required format.'));
				} else {
					$config->set('dataPrivacyURL', $dataPrivacyURL);
				}

				$config->set('wlEmbedding', (string)$request->request->get('wlEmbedding'));

				$config->set('docsurl', (string)$request->request->get('docsurl'));
				$config->set('enablehelp', (string)$request->request->get('enablehelp'));
			}
			$config->set('poweredByText', (string)$request->request->get('poweredByText'));
			$config->set('poweredByTooltip', (string)$request->request->get('poweredByTooltip'));
			$config->set('poweredByUrl', (string)$request->request->get('poweredByUrl'));

			$this->addFlash('success', $translator->trans(
				'The settings were saved successfully!'));
			return $this->redirectToRoute('app_hosting_settings_branding_form', [], 303);
		}

		return $this->redirectToRoute('app_hosting_settings_branding_form', [], 303);
	}

	/**
	 * Übersetzungen-Cache leeren.
	 *
	 * Reicht aus, solange alle Dateien bereits existierten. Veränderungen werden
	 * dann erkannt, aber neu hinzugekommene Dateien von Symfony nicht
	 * berücksichtigt. Dafür muss der komplette Cache geleert werden.
	 */
	private function clearTranslationsCache(): void
	{
		$baseCacheDir = Utils::getCacheDir();
		$cacheDir = $baseCacheDir . DIRECTORY_SEPARATOR . 'translations';
		$rnd = bin2hex(random_bytes(3));
		$suffix = '~'.date('Ymd').'-'.$rnd;

		if (@rename($cacheDir, $cacheDir.$suffix)) {
			Utils::removeDirectory($cacheDir.$suffix);
		}
	}

	/**
	 * @param string $lang
	 * @return array
	 */
	private function getTranslationData(string $lang): array
	{
		$customPath = Utils::getVarDir() . '/translations/messages.'.$lang.'.yaml';
		$originalPath = Utils::getBaseDir() . '/translations/messages.'.$lang.'.yaml';
		$originalData = $this->yamlLoader->load($originalPath, $lang)->all('messages');

		// Sonderfall Englisch: zusätzlich Schlüssel aus Deutsch holen
		if ($lang === 'en') {
			foreach ($this->yamlLoader->load(Utils::getBaseDir() . '/translations/messages.de.yaml', $lang)->all('messages') as $key => $text) {
				if (!isset($originalData[$key])) {
					$originalData[$key] = null;
				}
			}
		}

		// der YAML-Loader verwendet den Dateinamen; der MongoDB-Loader nicht.
		// Deswegen kann hier über die Dependency Injection zwischen den beiden Klassen gewechselt werden.
		$customData = $this->translationLoader->load($customPath, $lang)->all('messages');
		$data = [];
		foreach ($originalData as $key => $text) {
			$data[$key] = [
				'stock'=>$text,
				'custom'=>$customData[$key] ?? null
			];
		}
		foreach ($customData as $key => $text) {
			if (!isset($data[$key])) {
				$data[$key] = [
					'stock'=>null,
					'custom'=>$text
				];
			}
		}
		ksort($data);

		return $data;
	}

	/**
	 * @Route("/translations/{lang<[a-z]{2,4}>}.json", methods={"HEAD","GET"}, name="ajax_translation")
	 * @param string $lang
	 * @param Request $request
	 * @return JsonResponse
	 */
	public function ajaxTranslation(string $lang, Request $request): JsonResponse
	{
		$customPath = Utils::getVarDir() . '/translations/messages.'.$lang.'.yaml';
		$originalPath = Utils::getBaseDir() . '/translations/messages.'.$lang.'.yaml';

		if (!file_exists($originalPath)) {
			return $this->json([], 404);
		}

		$data = $this->getTranslationData($lang);

		return $this->json($data);
	}

	/**
	 * @param UploadedFile $image
	 * @param TranslatorInterface $translator
	 * @return void
	 */
	private function handleUploadErrors(UploadedFile $image, TranslatorInterface $translator): void
	{
		if ($image->getError() != UPLOAD_ERR_OK) {
			$this->addFlash(
				'danger',
				$translator->trans('The image could not be uploaded! (Error code %errorcode% %errormessage%)', [
					'%errorcode%' => $image->getError(),
					'%errormessage%' => $image->getErrorMessage()
				])
			);
			throw new RuntimeException('upload failed');
		}

		if (preg_match('/^image\/.*$/', $image->getMimeType()) == false) {
			$this->addFlash('danger', $translator->trans('The uploaded file is not an image!'));
			throw new RuntimeException('not an image');
		}
	}
}
