<?php

namespace App\Controller\Hosting;

use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Controller\HostingController;
use App\Model\User;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Menu("Clients", group="Management", icon="fa-users", route_group="app_hosting_client_", order=10)
 * @Route("/hosting/subuser", name="app_hosting_client_subuser_")
 */
class Subuser extends HostingController
{
	/**
	 * @Menu("Subusers", route_group="app_hosting_client_subuser_", order=10)
	 * @Route("", methods={"HEAD", "GET"}, name="index")
	 * @param Request $request
	 * @return Response
	 */
	public function listSubusers(Request $request): Response
	{
		if ($searchTerm = trim((string)$request->query->get('search'))) {
			$clients = User::search($searchTerm, User::ROLE_CLIENT);
		} else {
			$clients = User::byRole(User::ROLE_CLIENT);
		}

		$clients = array_reverse($clients);

		$clients = array_filter($clients, function (User $client) {
			return $client->isSubUser();
		});

		$pagination = new Pagination($clients);
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_hosting_client_subuser_index');
		}

		return $this->render('hosting/subuser/index.html.twig', [
			'pagination' => $pagination,
			'searchTerm' => $searchTerm,
		]);
	}

	/**
	 * @Route("/{id}/login", methods={"POST"}, name="login")
	 * @param string $id
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function logInAsSubuser(string $id, TranslatorInterface $translator, LoggerInterface $logger): Response
	{
		$subuser = User::byId($id);
		$user = User::loggedInUser();
		assert($user instanceof User);

		if (!$subuser || !($subuser->isClient() && $subuser->isSubUser())) {
			return $this->redirectToRoute('app_hosting_client_subuser_index');
		}

		$subuser->switchTo(true);

		$logger->logInfo("Login", "User ".$user->getUsername()." logged into account ".$subuser->getUsername());

		$this->addFlash('success', $translator->trans('You are now in the account “%USER%”. You may switch back to the admin account using the account menu in the top right corner.', [
			'%USER%' => $subuser->getUsername(),
		]));

		$response = $this->redirectToRoute('app_root', [], 303);
		$response->headers->set('Clear-Site-Data', '"cache"');
		return $response;
	}
}
