<?php

namespace App\Controller\Hosting;

use App\Component\Color;
use App\Component\Deployment\SystemThemeDeployment;
use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Config;
use App\Controller\HostingController;
use App\Model\BaseLayout;
use App\Model\Domain;
use App\Model\SystemTheme;
use App\Model\ThemeBase;
use App\Model\User;
use App\Utils;
use Exception;
use RuntimeException;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class SystemThemeController extends HostingController
{
	/**
	 * @Menu("Themes", group="Management", icon="fa-picture-o", route_group="app_hosting_system_theme_", order=40)
	 * @Route("/hosting/system_theme", name="app_hosting_system_theme_index")
	 * @param Request $request
	 * @return Response
	 */
	public function themeIndex(Request $request): Response
	{
		return $this->render('hosting/system_theme/index.html.twig', [
			'themes' => SystemTheme::all(),
		]);
	}

	/**
	 * @Route("/hosting/system_theme/{id}/default", name="app_hosting_system_theme_default", methods={"POST"})
	 * @param string $id
	 * @return Response
	 */
	public function setDefaultTheme(string $id): Response
	{
		$theme = SystemTheme::find($id);

		if ($theme) {
			$theme->makeDefault();
		}

		return $this->redirectToRoute('app_hosting_system_theme_index', [], 303);
	}

	/**
	 * @Route("/hosting/system_theme/{id}/preview", name="app_hosting_system_theme_preview")
	 * @param string $id
	 * @param Request $request
	 * @return Response
	 */
	public function previewTheme(string $id, Request $request): Response
	{
		$config = Config::getInstance();
		$user = User::loggedInUser();
		// Hole leere Domain, die beim Setup erstellt worden sein sollte
		$domain = array_reduce(Domain::all(), function (?Domain $carry, Domain $item) { return $carry ?? $item; });
		// Hole das angeforderte Theme
		$theme = SystemTheme::find($id);

		if (!$theme) {
			return $this->redirectToRoute('app_hosting_system_theme_index');
		}
		elseif (!$domain) {
			// This should not happen as a domain is created in \App\Model\Domain::onRepositoryLoaded.
			$this->addFlash('warning', 'No domain available to simulate theme preview.');
			return $this->redirectToRoute('app_hosting_system_theme_index');
		}

		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme() . '://';
		$baseUrl = $scheme . $request->getHttpHost();

		$embedScriptUrl = $baseUrl . $this->generateUrl(
			'app_external_main_javascript_file',
			[
				'apiKey' => $user->getApiKey(),
				'domain' => $domain->getId(),
				'systemTheme' => $theme->getId(),
				'preview' => 'true',
			]
		);

		$embedCodeSnippet = '<script src="' . htmlspecialchars($embedScriptUrl) . '" referrerpolicy="origin"></script>';

		$response = $this->render('theme/preview.html.twig', [
			'embedScript' => $embedCodeSnippet,
		]);

		$origins = $theme->getCustomCssOrigins();
		$originsText = implode(' ', $origins);
		$response->headers->set('Content-Security-Policy',
			"default-src 'self' 'unsafe-inline' $originsText; ".
			"img-src 'self' data:; ".
			"script-src 'self' 'unsafe-inline'; ".
			"connect-src 'self'; ".
			"form-action 'none'"
		);

		return $response;
	}

	/**
	 * @Route("/hosting/system_theme/deploy", name="app_hosting_system_theme_deploy", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function deployThemes(Request $request, TranslatorInterface $translator,  LoggerInterface $logger): Response
	{
		try {
			SystemThemeDeployment::publishAllSystemThemes();

			$this->addFlash('success', $translator->trans('System-wide themes have been deployed into each client account.'));
			$logger->logInfo('ThemeDeployment', 'System themes have been deployed successfully.');
		} catch (Exception $exception) {
			$this->addFlash('danger', $translator->trans('Oops! An error has occurred. Please refer to the system log for further information.'));
			$logger->logException('ThemeDeployment', $exception, 'An error occurred while deploying system-wide themes.');
		}

		return $this->redirectToRoute('app_hosting_system_theme_index', [], 303);
	}

	/**
	 * @Route("/hosting/system_theme/new", name="app_hosting_system_theme_new")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function themeCreate(Request $request, TranslatorInterface $translator): Response
	{
		$formData = $this->formDataTemplate();
		$mayGeoIP = (phpversion('bcmath') !== false or extension_loaded('bcmath') !== false or phpversion('gmp') !== false or extension_loaded('gmp') !== false);

		if ($request->getRealMethod() == 'POST') {
			$action = $request->request->getAlnum('action');

			// Benutzereingaben einlesen und validieren; $formData und $formErrors per Referenz
			$this->validateFormInput($request, $translator, $formData, $formErrors);

			if ($formErrors) {
				// Im Fehlerfall entsprechende Meldungen ausgeben
				array_walk($formErrors, function ($error) {
					$this->addFlash('danger', $error);
				});
			}
			else {
				// Neues Theme anlegen
				$theme = SystemTheme::create();
				$this->injectFormData($theme, $formData);

				try {
					// Bilduploads verarbeiten
					$this->migrateFileUploads($theme, $request, $translator);
				} catch (RuntimeException $e) {
					return $this->render('hosting/system_theme/new.html.twig', [
						'formData' => $formData,
						'mayGeoIP' => $mayGeoIP,
						'baseLayouts' => BaseLayout::all(),
						'allRestrictRegions' => ThemeBase::allRestrictRegions(),
					]);
				}

				// Das erste Theme wird automatisch zum Standard
				if (SystemTheme::count() == 1) {
					$theme->makeDefault();
				}

				$theme->save();

				$this->addFlash('success', $translator->trans(
					'The theme "%themename%" has been created successfully!', [
						'%themename%' => $theme->getName(),
					]
				));

				if ($action === 'saveClose') {
					return $this->redirectToRoute('app_hosting_system_theme_index', [], 303);
				} else {
					return $this->redirectToRoute('app_hosting_system_theme_edit', ['id'=>$theme->getId()], 303);
				}
			}
		}

		return $this->render('hosting/system_theme/new.html.twig', [
			'formData' => $formData,
			'mayGeoIP' => $mayGeoIP,
			'baseLayouts' => BaseLayout::all(),
			'allRestrictRegions' => ThemeBase::allRestrictRegions(),
		]);
	}

	/**
	 * @Route("/hosting/system_theme/{id}", name="app_hosting_system_theme_edit")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function themeEdit(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$theme = SystemTheme::find($id);

		if ($theme === null) {
			return $this->redirectToRoute('app_hosting_system_theme_index', [], 303);
		}

		$formData = $this->formDataTemplate($theme);
		$mayGeoIP = (phpversion('bcmath') !== false or extension_loaded('bcmath') !== false or phpversion('gmp') !== false or extension_loaded('gmp') !== false);

		if ($request->getRealMethod() == 'POST') {
			$action = $request->request->getAlnum('action');

			// Benutzereingaben einlesen und validieren; $formData und $formErrors per Referenz
			$this->validateFormInput($request, $translator, $formData, $formErrors);

			if ($formErrors) {
				// Im Fehlerfall entsprechende Meldungen ausgeben
				array_walk($formErrors, function ($error) {
					$this->addFlash('danger', $error);
				});
			}
			else {
				// Theme aktualisieren
				$this->injectFormData($theme, $formData);

				try {
					// Bilduploads verarbeiten
					$this->migrateFileUploads($theme, $request, $translator);
				} catch (RuntimeException $e) {
					return $this->render('hosting/system_theme/edit.html.twig', [
						'theme' => $theme,
						'formData' => $formData,
						'mayGeoIP' => $mayGeoIP,
						'baseLayouts' => BaseLayout::all(),
						'allRestrictRegions' => ThemeBase::allRestrictRegions(),
					]);
				}

				$theme->save();

				$this->addFlash('success', $translator->trans(
					'The theme "%themename%" has been edited successfully!', [
						'%themename%' => $theme->getName(),
					]
				));

				if ($action === 'saveClose') {
					return $this->redirectToRoute('app_hosting_system_theme_index', [], 303);
				} else {
					return $this->redirectToRoute('app_hosting_system_theme_edit', ['id'=>$theme->getId()], 303);
				}
			}
		}

		return $this->render('hosting/system_theme/edit.html.twig', [
			'theme' => $theme,
			'formData' => $formData,
			'mayGeoIP' => $mayGeoIP,
			'baseLayouts' => BaseLayout::all(),
			'allRestrictRegions' => ThemeBase::allRestrictRegions(),
		]);
	}

	/**
	 * @Route("/hosting/system_theme/{id}/copy", name="app_hosting_system_theme_copy", methods={"POST"})
	 * @param string $id
	 * @return Response
	 */
	public function themeCopy(string $id): Response
	{
		if ($theme = SystemTheme::find($id)) {
			SystemTheme::create()
				->importTheme($theme)
				->setName("COPY: {$theme->getName()}")
				->save();
		}

		return $this->redirectToRoute('app_hosting_system_theme_index', [], 303);
	}

	/**
	 * @Route("/hosting/system_theme/{id}/delete", name="app_hosting_system_theme_delete")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function themeDelete(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$theme = SystemTheme::find($id);

		if ($theme === null) {
			return $this->redirectToRoute('app_hosting_system_theme_index', [], 303);
		}

		if ($request->getRealMethod() == 'POST') {
			// Sofern das Standardtheme gelöscht wird, das erstbeste als neuen Standard festlegen
			if ($theme->isDefaultTheme()) {
				foreach (SystemTheme::all() as $currentTheme) {
					if ($currentTheme->getId() != $theme->getId()) {
						$currentTheme->makeDefault();
						$this->addFlash(
							'info',
							$translator->trans('The theme "%themename%" has now been set as default.', [
								'%themename%' => $currentTheme->getName(),
							])
						);
						break;
					}
				}
			}

			// Ressourcen bereinigen
			$theme
				->setPublishedSettingsIcon(null)
				->setSettingsIcon(null)
			;

			// Theme entfernen
			$themeName = $theme->getName();
			SystemTheme::delete($theme->getId());

			$this->addFlash('success', $translator->trans('The theme "%themename%" has been deleted successfully!', [
				'%themename%' => $themeName,
			]));

			return $this->redirectToRoute('app_hosting_system_theme_index', [], 303);
		}

		return $this->render('hosting/system_theme/delete.html.twig', [
			'theme' => $theme,
		]);
	}

	/**
	 * Diese Methode erstellt ein assoziatives Array, das sämtliche Parameter enthält, die für die Theme-Konfiguration
	 * relevant sind. Wenn ein Objekt übergeben wird, werden anstelle von Standardwerten die des Objekts hinterlegt.
	 * @param SystemTheme|null $theme
	 * @return array
	 */
	private function formDataTemplate(?SystemTheme $theme = null): array
	{
		return [
			'name' => $theme ? $theme->getName() : '',
			'baseLayout' => $theme ? $theme->getBaseLayout()->getId() : BaseLayout::default()->getId(),
			'position' => $theme ? $theme->getPosition() : '',
			'selectType' => $theme ? $theme->getSelectType() : '',
			'showPurposesInMainWindow' => $theme ? $theme->isShowPurposesInMainWindow() : false,
			'blocking' => $theme ? $theme->isBlocking() : false,
			'showEmbeddingsInControlPanel' => $theme ? $theme->isShowEmbeddingsInControlPanel() : false,
			'showAcceptAllButtonInControlPanel' => $theme ? $theme->isShowAcceptAllButtonInControlPanel() : false,
			'showWithdrawAllButtonInControlPanel' => $theme ? $theme->isShowWithdrawAllButtonInControlPanel() : false,
			'showCheckAndUncheckAllButtonInControlPanel' => $theme ? $theme->isShowCheckAndUncheckAllButtonInControlPanel() : false,
			'showExtraImprintFooter' => $theme ? $theme->isShowExtraImprintFooter() : false,
			'saveButtonInDetailsWindow' => $theme ? $theme->hasSaveButtonInDetailsWindow() : false,
			'settingsIconEnabled' => $theme ? $theme->isSettingsIconEnabled() : true,
			'settingsIconTarget' => $theme ? $theme->getSettingsIconTarget() : '',
			'settingsIconBackgroundColor' => $theme ? $theme->getSettingsIconBackgroundColor() : (Utils::isClickskeksCloud() ? '#4cb474' : '#003366'),
			'settingsIconForegroundColor' => $theme ? $theme->getSettingsIconForegroundColor() : '#ffffff',
			'widgetButtons' => $theme ? $theme->getWidgetButtons() : ['accept', 'cpanel', 'decline'],
			'complyingDoNotTrack' => $theme ? $theme->isComplyingDoNotTrack() : false,
			'onlyInEu' => $theme ? $theme->getOnlyInEu() : false,
			'okButtonBackgroundColor' => $theme ? $theme->getOkButtonBackgroundColor() : '#000000',
			'okButtonColor' => $theme ? $theme->getOkButtonColor() : '#FFFFFF',
			'okButtonBorderColor' => $theme ? $theme->getOkButtonBorderColor() : '#000000',
			'declineButtonBackgroundColor' => $theme ? $theme->getDeclineButtonBackgroundColor() : '#000000',
			'declineButtonTextColor' => $theme ? $theme->getDeclineButtonTextColor() : '#FFFFFF',
			'declineButtonBorderColor' => $theme ? $theme->getDeclineButtonBorderColor() : '#000000',
			'moreInfoButtonBackgroundColor' => $theme ? $theme->getMoreInfoButtonBackgroundColor() : '#000000',
			'moreInfoButtonColor' => $theme ? $theme->getMoreInfoButtonColor() : '#FFFFFF',
			'moreInfoButtonBorderColor' => $theme ? $theme->getMoreInfoButtonBorderColor() : '#000000',
			'windowBackgroundColor' => $theme ? $theme->getWindowBackgroundColor() : '#000000',
			'infotextColor' => $theme ? $theme->getInfotextColor() : '#FFFFFF',
			'windowBorderColor' => $theme ? $theme->getWindowBorderColor() : '#000000',
			'linkInInfotextColor' => $theme ? $theme->getLinkInInfotextColor() : '#FFFFFF',
			'iframeBlockerBackgroundColor' => $theme ? $theme->getIframeBlockerBackgroundColor() : '#000000',
			'iframeBlockerForegroundColor' => $theme ? $theme->getIframeBlockerForegroundColor() : '#FFFFFF',
			'iframeBlockerButtonBackgroundColor' => $theme ? $theme->getIframeBlockerButtonBackgroundColor() : '#000000',
			'iframeBlockerButtonForegroundColor' => $theme ? $theme->getIframeBlockerButtonForegroundColor() : '#FFFFFF',
			'customCss' => $theme ? $theme->getCustomCss() : '',
			'customCssTemplated' => $theme ? $theme->isCustomCssTemplated() : true,
			'customCssForIframeBlocker' => $theme ? $theme->getCustomCssForIframeBlocker() : '',
			'customCssForIframeBlockerTemplated' => $theme ? $theme->isCustomCssForIframeBlockerTemplated() : true,
			'manipulationPrevention' => $theme ? $theme->getManipulationPrevention() : true,
			'showCloseButton' => $theme ? $theme->hasCloseButton() : true,
			'declineButtonRestrictions' => $theme ? $theme->getDeclineButtonRestrictions() : ThemeBase::allRestrictRegions(),
			'bannerHeightMobile' => $theme ? $theme->getBannerHeightMobile() : 100,
		];
	}

	/**
	 * Diese Methoden validiert die Benutzereingaben und generiert u. U. Fehlermeldungen.
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param array $formData [Referenz]
	 * @param null|array $formErrors [Referenz] @phan-output-reference
	 */
	private function validateFormInput(Request $request, TranslatorInterface $translator, array &$formData, ?array &$formErrors): void
	{
		$formErrors = [];

		array_walk($formData, function (&$value, $fieldName) use ($request, $translator, &$formErrors) {
			$value = $request->request->get($fieldName);

			if (is_string($value)) {
				$value = trim($value);
			}

			if ($fieldName == 'name' && empty($value)) {
				$formErrors[] = $translator->trans('Theme name may not be empty.');
			}
			elseif (preg_match('~Color$~', $fieldName) && !Color::validateCssString((string)$value)) {
				$formErrors[] = $translator->trans('Invalid color values! Please use the color picker and/or provide hex color codes.');
			}
		});

		$formErrors = array_unique($formErrors);
	}

	/**
	 * Diese Methode überträgt die Benutzereingaben in das übergebene Theme-Objekt, ohne das Repository zu speichern.
	 * @param ThemeBase $theme
	 * @param array $formData
	 */
	private function injectFormData(ThemeBase $theme, array $formData): void
	{
		$theme
			->touchLastModifiedTimestamp()
			->setName($formData['name'])
			->setBaseLayout(BaseLayout::find((string)$formData['baseLayout']) ?? BaseLayout::default())
			->setPosition($formData['position'])
			->setSelectType($formData['selectType'])
			->setShowPurposesInMainWindow((bool)($formData['showPurposesInMainWindow'] ?? false))
			->setBlocking((bool)($formData['blocking'] ?? false))
			->setShowEmbeddingsInControlPanel((bool)($formData['showEmbeddingsInControlPanel'] ?? false))
			->setShowAcceptAllButtonInControlPanel((bool)($formData['showAcceptAllButtonInControlPanel'] ?? false))
			->setShowWithdrawAllButtonInControlPanel((bool)($formData['showWithdrawAllButtonInControlPanel'] ?? false))
			->setShowCheckAndUncheckAllButtonInControlPanel((bool)($formData['showCheckAndUncheckAllButtonInControlPanel'] ?? false))
			->setShowExtraImprintFooter((bool)($formData['showExtraImprintFooter'] ?? false))
			->setSaveButtonInDetailsWindow((bool)($formData['saveButtonInDetailsWindow'] ?? false))
			->setSettingsIconEnabled((bool)($formData['settingsIconEnabled'] ?? false ))
			->setSettingsIconTarget($formData['settingsIconTarget'])
			->setSettingsIconBackgroundColor($formData['settingsIconBackgroundColor'])
			->setSettingsIconForegroundColor($formData['settingsIconForegroundColor'])
			->setWidgetButtons((array)$formData['widgetButtons'])
			->setComplyingDoNotTrack((bool)($formData['complyingDoNotTrack'] ?? false))
			->setOnlyInEu((bool)($formData['onlyInEu'] ?? false))
			->setOkButtonBackgroundColor($formData['okButtonBackgroundColor'])
			->setOkButtonColor($formData['okButtonColor'])
			->setOkButtonBorderColor($formData['okButtonBorderColor'])
			->setDeclineButtonBackgroundColor($formData['declineButtonBackgroundColor'])
			->setDeclineButtonTextColor($formData['declineButtonTextColor'])
			->setDeclineButtonBorderColor($formData['declineButtonBorderColor'])
			->setMoreInfoButtonBackgroundColor($formData['moreInfoButtonBackgroundColor'])
			->setMoreInfoButtonColor($formData['moreInfoButtonColor'])
			->setMoreInfoButtonBorderColor($formData['moreInfoButtonBorderColor'])
			->setWindowBackgroundColor($formData['windowBackgroundColor'])
			->setInfotextColor($formData['infotextColor'])
			->setWindowBorderColor($formData['windowBorderColor'])
			->setLinkInInfotextColor($formData['linkInInfotextColor'])
			->setIframeBlockerBackgroundColor($formData['iframeBlockerBackgroundColor'])
			->setIframeBlockerForegroundColor($formData['iframeBlockerForegroundColor'])
			->setIframeBlockerButtonBackgroundColor($formData['iframeBlockerButtonBackgroundColor'])
			->setIframeBlockerButtonForegroundColor($formData['iframeBlockerButtonForegroundColor'])
			->setCustomCss($formData['customCss'])
			->setCustomCssTemplated((bool)($formData['customCssTemplated'] ?? false))
			->setCustomCssForIframeBlocker($formData['customCssForIframeBlocker'])
			->setCustomCssForIframeBlockerTemplated((bool)($formData['customCssForIframeBlockerTemplated'] ?? false))
			->setManipulationPrevention((bool)($formData['manipulationPrevention'] ?? false))
			->setCloseButton((bool)($formData['showCloseButton'] ?? false))
			->setDeclineButtonRestrictions($formData['declineButtonRestrictions'] ?? [])
			->setBannerHeightMobile((int)($formData['bannerHeightMobile'] ?? 100))
		;
	}

	/**
	 * @param SystemTheme $theme
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 */
	private function migrateFileUploads(SystemTheme $theme, Request $request, TranslatorInterface $translator): void
	{
		$fileUploads = array_filter([
			[
				'fieldName' => 'settings_icon',
				'setter' => [$theme, 'setSettingsIcon'],
			],
		]);

		foreach ($fileUploads as $uploadInfo) {
			/** @var UploadedFile $file */
			$file = $request->files->get($uploadInfo['fieldName']);
			$shouldRemove = $request->request->getBoolean($uploadInfo['fieldName'].'__delete');

			if ($file) {
				$this->handleUploadErrors($file, $translator);
				$uploadInfo['setter']($file);
			}
			elseif ($shouldRemove) {
				$uploadInfo['setter'](null);
			}
		}
	}

	/**
	 * @param UploadedFile $image
	 * @param TranslatorInterface $translator
	 */
	private function handleUploadErrors(UploadedFile $image, TranslatorInterface $translator): void
	{
		if ($image->getError() != UPLOAD_ERR_OK) {
			$this->addFlash(
				'danger',
				$translator->trans('The image could not be uploaded! (Error code %errorcode% %errormessage%)', [
					'%errorcode%' => $image->getError(),
					'%errormessage%' => $image->getErrorMessage()
				])
			);
			throw new RuntimeException();
		}

		if (strpos($image->getMimeType(), 'image/') !== 0) {
			$this->addFlash('danger', $translator->trans('The uploaded file is not an image!'));
			throw new RuntimeException();
		}
	}
}
