<?php

namespace App\Controller;

use App\Component\Api\Exception\ApiRestrictionException;
use App\Component\Api\Exception\BadRequestException;
use App\Component\Api\HostingApi;
use App\Component\Validator\DataTypeValidator;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\Model;
use App\Model\Placeholder;
use App\Model\SystemEmbedding;
use App\Model\SystemEmbeddingAsset;
use App\Model\SystemTheme;
use App\Model\Theme;
use App\Model\User;
use App\Utils;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

abstract class HostingApiController extends AbstractController
{
	/**
	 * Maps readable properties of the component to document field names that are used inside the model setter.
	 * @var array<string,string> $projectionMap
	 */
	protected static $projectionMap = [];

	/** @var array<string,bool>|null $projection */
	private $projection;

	/** @var int|null $limit */
	private $limit;

	/** @var int $skip */
	private $skip = 0;

	/** @var bool $requestParametersParsed */
	private $requestParametersParsed = false;

	/**
	 * @param array $payload
	 * @param array<string,mixed> $schema [
	 * 		propertyName => dataType,
	 * 		propertyName => [ propertyName => dataType ],
	 * 		propertyName => [ "@collection" => true|false, "@schema" => schemaDefinition ],
	 * 		…
	 * ]
	 * @return void
	 * @throws InvalidArgumentException if the payload contains invalid data types.
	 */
	protected static function checkPayloadDataTypes(array $payload, array $schema): void
	{
		foreach ($schema as $property => $type) {
			if (array_key_exists($property, $payload) == false) {
				continue;
			}
			elseif (is_array($type)) {
				if (is_array($payload[$property])) {
					if (is_array($nestedSchema = $type['@schema'] ?? null)) {
						$isCollection = (bool)($type['@collection'] ?? false);
						if ($isCollection) {
							foreach ($payload[$property] as $item) {
								if (is_array($item)) {
									// Treat the property as a collection of objects using the nested schema definition
									self::checkPayloadDataTypes($item, $nestedSchema);
								}
								else {
									throw new InvalidArgumentException("Invalid data type for \"$property\", collection of objects expected.", 30);
								}
							}
						}
						else {
							// Treat the property as single object using the nested schema definition
							self::checkPayloadDataTypes($payload[$property], $type);
						}
					}
					else {
						// Treat the property as normal object
						self::checkPayloadDataTypes($payload[$property], $type);
					}
				}
				else {
					throw new InvalidArgumentException("Invalid data type for \"$property\", object expected.", 30);
				}
			}
			elseif (DataTypeValidator::verify($payload[$property], $type) == false) {
				throw new InvalidArgumentException("Invalid data type for \"$property\", $type expected.", 30);
			}
		}
	}

	/**
	 * @param Request $request
	 * @param int $depth
	 * @return array
	 * @throws BadRequestException
	 */
	protected static function decodeRequestPayload(Request $request, int $depth=6): array
	{
		if (Utils::decodeJsonRequest($request, $payload, true, $depth) == false) {
			/** @var JsonResponse $payload */
			assert($payload instanceof JsonResponse);
			throw new BadRequestException($payload);
		}
		elseif (!is_array($payload)) {
			throw new BadRequestException(new JsonResponse([
				'success' => false,
				'message' => 'Invalid payload. Make sure to send a JSON encoded object.',
			], 400));
		}

		return $payload;
	}

	/**
	 * @param Request $request
	 * @return array<string,bool>
	 * @throws BadRequestException
	 */
	private static function parseProjectionParameters(Request $request): array
	{
		$projection = [];

		if ($request->query->has('_source_includes')) {
			$projection = array_fill_keys(array_filter(explode(',', $request->query->get('_source_includes'))), true);
		}

		$projection = array_fill_keys(array_filter(explode(',', $request->query->get('_source_excludes'))), false) + $projection;

		if ($unsupportedProjectionKeys = array_keys(array_diff_key($projection, static::$projectionMap))) {
			throw new BadRequestException(new JsonResponse([
				'code' => 900,
				'error' => 'Unsupported projection keys: '.implode(',', $unsupportedProjectionKeys),
			], 400));
		}

		return $projection;
	}

	/**
	 * @param Request $request
	 * @return void
	 * @throws BadRequestException
	 */
	protected final function parseRequestParameters(Request $request): void
	{
		// Parse projection parameters
		$this->projection = self::parseProjectionParameters($request);

		// Parse _limit parameter
		if ($request->query->has('_limit')) {
			if (DataTypeValidator::checkStringForUInt((string)$request->query->get('_limit'))) {
				$this->limit = $request->query->getInt('_limit');
			} else {
				throw new BadRequestException(new JsonResponse([
					'code' => 901,
					'error' => 'Invalid _limit value, must be a positive integer.',
				], 400));
			}
		} else {
			$this->limit = null;
		}

		// Parse _skip parameter
		if ($request->query->has('_skip')) {
			if (DataTypeValidator::checkStringForUInt((string)$request->query->get('_skip'))) {
				$this->skip = $request->query->getInt('_skip');
			} else {
				throw new BadRequestException(new JsonResponse([
					'code' => 902,
					'error' => 'Invalid _skip value, must be a positive integer.',
				], 400));
			}
		} else {
			$this->skip = 0;
		}

		// Mark request parameters as parsed
		$this->requestParametersParsed = true;
	}

	/**
	 * This method maps the API component projection `$this->projection` to `static::projectionMap()`.
	 * The resulting array may be supplied to `Model::setProjection()`.
	 * Call `$this->parseRequestParameters()` before calling this method.
	 * @return array
	 * @throws LogicException if projection hasn't been initialized. This should only ever happen during development.
	 * @see self::parseRequestParameters()
	 * @see Model::setProjection()
	 */
	protected final function buildModelProjection(): array
	{
		if ($this->projection === null) {
			throw new LogicException('Projection has not been initialized, call method parseRequestParameters first.');
		}
		elseif (empty($this->projection)) {
			return [];
		}

		// Map API component properties to their respective models' counterparts.
		// Filter out entries that don't point to a database field, i.e. they are null.
		return array_filter(array_combine(
			array_map(function (string $key) {
				return static::$projectionMap[$key];
			}, array_keys($this->projection)),
			$this->projection
		) ?: [], 'strlen', ARRAY_FILTER_USE_KEY);
	}

	/**
	 * @param array<string,mixed> $data
	 * @return array<string,mixed>
	 * @throws LogicException if projection hasn't been initialized. This should only ever happen during development.
	 */
	protected final function applyProjection(array $data): array
	{
		if ($this->projection === null) {
			throw new LogicException('Projection has not been initialized, call method parseRequestParameters first.');
		}
		elseif (empty($this->projection)) {
			return $data;
		}

		$include = array_filter($this->projection);
		$exclude = array_diff_key($this->projection, $include);

		if ($include) {
			$data = array_intersect_key($data, $include);
		}

		return array_diff_key($data, $exclude);
	}

	/**
	 * @return void
	 * @throws LogicException if query parameters haven't been parsed. This should only ever happen during development.
	 */
	private function assertQueryParametersParsed(): void
	{
		if ($this->requestParametersParsed == false) {
			throw new LogicException('Query parameters have not been parsed, call method parseRequestParameters first.');
		}
	}

	/**
	 * @return int|null How many documents to return at most. If null, no limit is to be applied.
	 */
	protected final function getQueryParameterLimit(): ?int
	{
		$this->assertQueryParametersParsed();
		return $this->limit;
	}

	/**
	 * @return int How many documents to skip.
	 */
	protected final function getQueryParameterSkip(): int
	{
		$this->assertQueryParametersParsed();
		return $this->skip;
	}

	/**
	 * This method applies the query parameters `_limit` and `_skip` to `$data`.
	 * If the limit is not set, the method will apply the respective API restriction.
	 * @template T
	 * @param array<T> $data
	 * @return array<T>
	 */
	protected final function applyPagination(array $data): array
	{
		// Apply item limit restriction if no limit was defined in the query parameters
		if ($this->getQueryParameterLimit() === null) {
			$apiLimit = HostingApi::getInstance()->getItemLimitRestriction();
			$itemCount = max(0, count($data) - $this->getQueryParameterSkip());
			if ($apiLimit !== null && $itemCount > $apiLimit) {
				throw new ApiRestrictionException(new JsonResponse([
					'error' => "API restriction: The result set exceeds the limit of $apiLimit items, use _limit parameter.",
				], 416));
			}
		}

		return ($this->getQueryParameterSkip() > 0 || $this->getQueryParameterLimit() !== null)
			? array_slice($data, $this->getQueryParameterSkip(), $this->getQueryParameterLimit(), true)
			: $data;
	}

	/**
	 * This method checks whether the client exists. It also sets up the proper context
	 * for user and domain specific repositories.
	 * Also, subusers aren't considered clients.
	 * @param string $userId
	 * @return void
	 * @throws BadRequestException if client doesn't exist or subuser was selected.
	 */
	protected static function assertClientExists(string $userId): void
	{
		$user = User::byId($userId);

		if (!$user || $user->getRole() != User::ROLE_CLIENT) {
			throw new BadRequestException(new JsonResponse(['error' => 'Client not found.'], 404));
		}
		elseif ($user->isSubUser()) {
			throw new BadRequestException(new JsonResponse([
				'code' => 10,
				'error' => 'Can\'t perform this action for a subuser.',
			], 400));
		}

		User::setApiUser($user->getApiKey());
		Model::closeUserModels();
		Domain::switchToUser($user);
	}

	/**
	 * This method checks whether the user and domain exist. It also sets up the proper context
	 * for user and domain specific repositories.
	 * @param string $userId
	 * @param string $domainId
	 * @return void
	 * @throws BadRequestException if client-domain combination doesn't exist or subuser was selected.
	 */
	protected static function assertDomainExists(string $userId, string $domainId): void
	{
		self::assertClientExists($userId);

		if (Domain::setDomainForRequest($domainId) == false) {
			throw new BadRequestException(new JsonResponse(['error' => 'Domain not found.'], 404));
		}
	}

	/**
	 * This method checks whether the user, domain and embedding exist. It also sets up the proper context
	 * for user and domain specific repositories.
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @return void
	 * @throws BadRequestException
	 */
	protected static function assertEmbeddingExists(string $userId, string $domainId, string $embeddingId): void
	{
		self::assertDomainExists($userId, $domainId);

		if (Embedding::exists($embeddingId) == false) {
			throw new BadRequestException(new JsonResponse(['error' => 'Embedding not found.'], 404));
		}
	}

	/**
	 * This method checks whether the user, domain, embedding and asset exist. It also sets up the proper context
	 * for user and domain specific repositories.
	 * @param string $userId
	 * @param string $domainId
	 * @param string $embeddingId
	 * @param string $assetId
	 * @return void
	 * @throws BadRequestException
	 */
	protected static function assertEmbeddingAssetExists(string $userId, string $domainId, string $embeddingId, string $assetId): void
	{
		self::assertEmbeddingExists($userId, $domainId, $embeddingId);

		/** @var Embedding $embedding */
		$embedding = Embedding::find($embeddingId);
		assert($embedding instanceof Embedding);
		/** @var EmbeddingAsset|null $asset */
		$asset = $embedding->getAsset($assetId);

		if (!$asset) {
			throw new BadRequestException(new JsonResponse(['error' => 'Asset not found.'], 404));
		}
	}

	/**
	 * This method checks whether the user, domain and theme exist. It also sets up the proper context
	 * for user and domain specific repositories.
	 * @param string $userId
	 * @param string $domainId
	 * @param string $themeId
	 * @return void
	 * @throws BadRequestException
	 */
	protected static function assertThemeExists(string $userId, string $domainId, string $themeId): void
	{
		self::assertDomainExists($userId, $domainId);

		if (Theme::exists($themeId) == false) {
			throw new BadRequestException(new JsonResponse(['error' => 'Theme not found.'], 404));
		}
	}

	/**
	 * This method checks whether the user, domain and placeholder exist. It also sets up the proper context
	 * for user and domain specific repositories.
	 * @param string $userId
	 * @param string $domainId
	 * @param string $placeholderId
	 * @return void
	 * @throws BadRequestException
	 */
	protected static function assertPlaceholderExists(string $userId, string $domainId, string $placeholderId): void
	{
		self::assertDomainExists($userId, $domainId);

		if (Placeholder::exists($placeholderId) == false) {
			throw new BadRequestException(new JsonResponse(['error' => 'Placeholder not found.'], 404));
		}
	}

	/**
	 * This method checks whether the SystemEmbedding exists.
	 * @param string $embeddingId
	 * @return void
	 * @throws BadRequestException
	 */
	protected static function assertSystemEmbeddingExists(string $embeddingId): void
	{
		if (SystemEmbedding::exists($embeddingId) == false) {
			throw new BadRequestException(new JsonResponse(['error' => 'Embedding not found.'], 404));
		}
	}

	/**
	 * This method checks whether the system-wide embedding and asset exist.
	 * @param string $embeddingId
	 * @param string $assetId
	 * @return void
	 * @throws BadRequestException
	 */
	protected static function assertSystemEmbeddingAssetExists(string $embeddingId, string $assetId): void
	{
		self::assertSystemEmbeddingExists($embeddingId);

		/** @var SystemEmbedding $embedding */
		$embedding = SystemEmbedding::find($embeddingId);
		assert($embedding instanceof SystemEmbedding);
		/** @var SystemEmbeddingAsset|null $asset */
		$asset = $embedding->getAsset($assetId);

		if (!$asset) {
			throw new BadRequestException(new JsonResponse(['error' => 'Asset not found.'], 404));
		}
	}

	/**
	 * This method checks whether the system-wide theme exists.
	 * @param string $themeId
	 * @return void
	 * @throws BadRequestException
	 */
	protected static function assertSystemThemeExists(string $themeId): void
	{
		if (SystemTheme::exists($themeId) == false) {
			throw new BadRequestException(new JsonResponse(['error' => 'Theme not found.'], 404));
		}
	}
}
