<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\LegalText;
use App\Component\LocaleService;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class Impressum extends DomainDependantController
{

	/**
	 * @Menu("Imprint", group="Content & Texts", icon="fa-italic", order=80)
	 * @Route("/domains/{_domainId}/impressum", name="app_impressum")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function imprint(Request $request, TranslatorInterface $translator): Response
	{
		$locales = LocaleService::enabledLocales();
		$imprintL10n = (array)$request->request->get('imprint_l10n');
		$imprintUrl = (array)$request->request->get('imprintUrl');

		if ($request->getRealMethod() === 'POST') {
			foreach (LegalText::all() as $legalText) {
				$legalText
					->setImprintScheme($request->request->getAlpha('imprintScheme'))
					->setShowWidgetOnImprintPage($request->request->getBoolean('showWidgetOnImprintPage'))
					->settargetBlank($request->request->getBoolean('targetBlank'))
				;
			}

			foreach ($locales as $locale) {
				/** @var LegalText|null $legalText */
				$legalText = LegalText::byLocale($locale);

				if ($legalText === null) {
					$legalText = LegalText::create()->setLocale($locale);
				}

				$legalText
					->setImprint($imprintL10n[$locale->getName()] ?? '')
					->setImprintUrl($imprintUrl[$locale->getName()] ?? '')
				;
			}

			try {
				CsrfToken::get()->checkRequest($request);
				LegalText::flush();

				$this->addFlash('success', $translator->trans('Localizations have been written successfully.'));
				return $this->redirectToRoute('app_impressum');
			}
			catch (ConfigNotWritableException $exception) {
				$this->addFlash('danger', $translator->trans('Localizations could not be written to disk.'));
			}
			catch (CsrfTokenException $exception) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			}
		}

		// Erstbeste Instanz einlesen, um sprachübergreifende Attribute auszulesen
		$firstLocale = reset($locales) ?: LocaleService::find('de_DE') ?? LocaleService::find('en_US');
		$legalText = LegalText::byLocale($firstLocale) ?? LegalText::create()->setLocale($firstLocale);

		return $this->render('impressum/index.html.twig', [
			'locales' => $locales,
			'legalTexts' => LegalText::havingLocaleEnabled(),
			'csrfToken' => CsrfToken::get(),
			'imprintScheme' =>$legalText->getImprintScheme(),
			'targetBlank' => $legalText->istargetBlank() ?? false,
		]);
	}
}
