<?php

namespace App\Controller;

use App\Component\Licensing\ApiClient;
use App\Component\Menu\Annotation\Menu;
use App\Config;
use App\Model\Domain;
use App\Model\User;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/license_manager", name="app_license_manager_")
 */
class LicenseManager extends AdminController
{
	/**
	 * @Menu("License management", group="Settings", icon="fa-key", navigation="meta", order=70)
	 * @Route("", name="index")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function index(Request $request, TranslatorInterface $translator, ApiClient $api): Response
	{
		if ($request->getMethod() === 'POST') {
			$this->updateLicense($request, $translator, $api);

			return $this->redirectToRoute('app_license_manager_index', [], 303);
		}

		return $this->render('license_manager/index.html.twig', [
			'licenseKey' => Utils::getLicenseKey(),
			'whitelabel' => Utils::hasWhitelabelLicense(),
		]);
	}

	/**
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 */
	private function updateLicense(Request $request, TranslatorInterface $translator, ApiClient $api): void
	{
		$oldLicenseKey = Utils::getLicenseKey();
		$oldWhitelabelState = Utils::hasWhitelabelLicense();

		$newLicenseKey = trim($request->request->get('licenseKey'));
		$newWhitelabelState = (bool)$request->request->get('whitelabel');

		$licensed = strlen($oldLicenseKey) > 0;
		$edition = $_ENV['APP_EDITION'] ?? '';

		//erstmal auskommentiert - damit man durch Speichern des Keys die Daten zuverlässig neu abfragen kann
		//if ($oldLicenseKey != $newLicenseKey)
		//{
			// Registrierung der u. U. aktiven Lizenz inkl. Whitelabel zunächst aufheben
			if ($licensed) {
				try {
					$api->unregisterLicense($oldLicenseKey, $request->getHttpHost());
				} catch (\Exception $e) {
					$this->addFlash('danger', $translator->trans($e->getMessage()));
				}
				$licensed = false;
				$edition = '';
			}

			// Versuche neuen Lizenzschlüssel zu registrieren
			if ($newLicenseKey !== '') {
				$licenseInfo = null;
				try {
					$licenseInfo = $api->registerLicense($newLicenseKey, $request->getHttpHost());
					$licensed = true;
				}
				catch (\Exception $e) {
					$this->addFlash('danger', $translator->trans($e->getMessage()));
				}

				if ($licenseInfo !== null) {
					$edition = $licenseInfo->edition;

					Config::getInstance()
						->set('licensekey', $newLicenseKey)
						->set('whitelabel', $licenseInfo->whitelabel)
					;

					Utils::setLocalEnvironment([
						'APP_EDITION' => $edition,
					]);

					// Das Limit ist für Administratoren derzeit nur in der Multi-Domain-Edition relevant
					foreach (User::byRole(User::ROLE_ADMIN) as $user) {
						$user->setMaxDomainCount($licenseInfo->limit);
						$user->setUserTarifSettings($licenseInfo->tarif);

						//nur nicht bei extended setzen
						if ($licenseInfo->edition === "multi") {
							//Max Call Counts setzen - unbegrenzt oder laut Tarif
							$user->setMaxCallCount($licenseInfo->maxcalls);
						}
						//Sonder / Altfall eine Single Base Version oder Pro oder sonstige Lizenz - dann auf base stellen und Funktionen passend aktivieren.
						if ($licenseInfo->edition === "base") {
							$user->setUserTarifSettings("base");
						}

					}
					User::flush();

					$this->addFlash('success', $translator->trans('License key has been activated successfully.'));
				}
			}

			// Wenn nicht mehr lizenziert, sämtliche Lizenzdaten zurücksetzen und raus
			if (!$licensed) {
				Config::getInstance()
					->set('licensekey', '')
					->set('whitelabel', false)
				;
				Utils::setLocalEnvironment([
					'APP_EDITION' => '',
				]);

				if (empty($newLicenseKey)) {
					$this->addFlash('success', $translator->trans('License data has been reset.'));
				}

				return;
			}
		//}

		if ($licensed && $edition !== 'extended') {
			$whitelabelInfo = null;
			try {
				if ($newWhitelabelState) {
					$whitelabelInfo = $api->registerWhitelabel($newLicenseKey, $request->getHttpHost());
				}
				else {
					$whitelabelInfo = $api->unregisterWhitelabel($newLicenseKey, $request->getHttpHost());
				}
			}
			catch (\Exception $e) {
				$this->addFlash('danger', $translator->trans($e->getMessage()));
			}

			if ($whitelabelInfo !== null) {
				Config::getInstance()
					->set('whitelabel', $whitelabelInfo->whitelabel)
				;
				foreach (User::byRole(User::ROLE_ADMIN) as $user) {
					$user->setMaxWLCount($whitelabelInfo->whitelabelMaxCount);
				}
				User::flush();

				$keyChanged = $oldLicenseKey != $newLicenseKey;
				$wlChanged = $oldWhitelabelState != $newWhitelabelState;

				if (($keyChanged == false && $wlChanged) || ($keyChanged && $newWhitelabelState)) {
					$this->addFlash('success', $translator->trans(
						$newWhitelabelState
							? 'Whitelabel has been registered successfully.'
							: 'Whitelabel has been unregistered successfully.'
					));
				}
			}
		}
		/**
		 * Extended Edition HR Daten übergeben
		 */
		if ($edition === "extended") {
			$sumCount = [];
			$userCount = $domainCount = 0;
			try {
				$userCount = User::count()-1; //abzgl. root User
				// TODO: Optimieren: zwei Loops zu einem zusammenfassen
				$domainCount = array_reduce(User::getClients(), static function ($count, User $user) {
					Domain::switchToUser($user);
					return $count + Domain::countIncludingShareNames();
				}, 0);
				foreach (User::getClients() as $client)
				{
					Domain::switchToUser($client);
					$currentDomCount = Domain::countIncludingShareNames();
					if(empty($sumCount[$currentDomCount])) {
						$sumCount[$currentDomCount] = 1;
					}
					else {
						$sumCount[$currentDomCount]++;
					}
				}
			}
			finally {
				$user = User::loggedInUser();
				if ($user) {
					Domain::switchToUser($user);
				}
			}
			try {
				$api->reportLicensees(Utils::getLicenseKey(), $request->getHttpHost(), $domainCount, $userCount, serialize($sumCount));
			} catch (\Exception $e) {
				// ignore for now
			}
		}
	}
}
