<?php

namespace App\Controller;

use App\Component\EmbeddingUpdateService;
use App\Component\LocaleService;
use App\Component\Menu\Annotation\Menu;
use App\Component\LocaleManagerTrait;
use App\Controller\DomainDependantController;
use App\Exception\LocaleAlreadyExistsException;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\Locale as LocaleModel;
use App\Model\LocaleBase;
use App\Model\UserLocale;
use App\Model\Language as Language;
use App\Model\User;
use App\Model\Purpose;
use Exception;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/domains/{_domainId}/locale", name="app_setting_locale_")
 */
class Locale extends DomainDependantController
{
	use LocaleManagerTrait;

	/**
	 * @Menu("Texts", group="Content & Texts", icon="fa-file-text-o", order=70)
	 * @Route("", name="list")
	 * @return Response
	 */
	public function listLocales(): Response
	{
		//Evel. noch nicht genutzten Skeletons importieren.
		$this->importSkeletons();

		$locales = LocaleModel::sortByEnabledAndName(LocaleModel::all());
		$userLocales = UserLocale::sortByEnabledAndName(UserLocale::all());

		// Locale-Namen als Index verwenden, wie es auch UserLocales tun
		$locales = array_combine(
			array_map(function (LocaleModel $locale) { return $locale->getName(); }, $locales),
			$locales
		);

		$overriddenUserLocales = array_filter($userLocales, function (UserLocale $ul) use ($locales) {
			foreach ($locales as $l) {
				if ($l->getName() == $ul->getName()) {
					return true;
				}
			}
			return false;
		});

		return $this->render('locale/list.html.twig', [
			'user' => User::loggedInUser(),
			'locales' => $locales,
			'userLocales' => $userLocales,
			'effectiveLocales' => LocaleBase::sortByEnabledAndName($locales + $userLocales),
			'overriddenUserLocales' => $overriddenUserLocales,
		]);
	}

	/**
	 * Importiert neue Layouts aus dem Skeleton Themes - damit auch die alten Accounts aktuelle Themes nutzen können
	 * die als Vorlagen neu reinkommen
	 */
	public function importSkeletons()
	{
		//Alle Themes
		$userLocales = UserLocale::all();

		//Alle Skeletons
		$skeletons = UserLocale::readSkeletonData();

		//loopen
		foreach($skeletons as $key=>$skeletonItem)
		{
			$exists = false;
			foreach($userLocales as $userLocale)
			{
				//Nach Key checken
				if($key==$userLocale->getId())
				{
					$exists = true;
				}
			}

			//gibts noch nicht, dann ein neues anlegen...
			if($exists==false)
			{
				$lang = Language::find($key);
				if (!empty($lang) && strlen($lang->getId())>2) {
					$newULoc = UserLocale::create($lang->getCode());
					foreach($skeletonItem as $newUserLocale => $newItem)
					{
						$newULoc->set($newUserLocale,$newItem);
					}
					//und auch den Skeletonkey setzen, für den Fall dass ein Theme umbennant wird
					$newULoc->set("skeletonKey",$key);
					$newULoc->set("enabled",0);
					$newULoc->save();
				}
			}
		}
	}

	/**
	 * @Route("/new", name="new_form", methods={"GET","HEAD"})
	 * @Route("/new/{localeName}", name="new_from_template", methods={"GET","HEAD"})
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function createLocaleForm(Request $request, TranslatorInterface $translator, ?string $localeName = null): Response
	{
		$fields = self::getStandardFieldGroups($translator);
		$localeTemplate = ($localeName === null) ? null : UserLocale::find($localeName);

		if ($localeTemplate) {
			$fields = array_map(function ($fieldset) use ($localeTemplate) {
				return array_map(function ($field) use ($localeTemplate) {
						list($getter, $purposeId) = $this->buildGetterSetterName($field['field_id'], 'get');
						if ($getter == 'getPurposeName' or $getter == 'getPurposeDescription') {
							$purpose = Purpose::find((string)$purposeId);
							if ($purpose) {
								$field['default'] = $field['user_input'] = $localeTemplate->$getter($purpose);
							}
						} else {
							$field['default'] = $field['user_input'] = $localeTemplate->$getter();
						}
						return $field;
					}, $fieldset);
			}, $fields);

			$localeEnabled = $localeTemplate->isEnabled();
		} else {
			$localeEnabled = false;
		}

		$locales = LocaleService::all();

		// Consider user locales a "used name" because the index lists them in addition to domain locales.
		// To create the respective domain locale, one should use the "edit" button.
		$used_names = array_map(function (LocaleBase $item) { return $item->getName(); }, $locales);

		return $this->render('locale/new.html.twig', [
			'user' => User::loggedInUser(),
			'fields' => $fields,
			'used_names' => $used_names,
			'languages' => Language::all(),
			'locale_name' => $localeName,
			'localeEnabled' => $localeEnabled,
			'localeTemplates' => $localeTemplate ? [] : $locales,
		]);
	}

	/**
	 * @Route("/new", name="new", methods={"POST"})
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param EmbeddingUpdateService $updater
	 * @return Response
	 */
	public function createLocale(Request $request, TranslatorInterface $translator, EmbeddingUpdateService $updater): Response
	{
		$fields = self::getStandardFieldGroups($translator);
		$localeName = $request->request->get('locale_name', '');

		try {
			$locale = LocaleModel::create($localeName);

			$locale
				->setEnabled($request->request->getBoolean('localeEnabled'))
			;

			// Alle Daten in den Feldern speichern
			foreach ($fields as $fieldset) {
				foreach ($fieldset as $field) {
					list($setter, $purposeId) = $this->buildGetterSetterName($field['field_id'], 'set');
					$value = trim($request->request->get($field['field_id'], ''));
					if ($setter == 'setPurposeName' or $setter == 'setPurposeDescription') {
						$purpose = Purpose::find((string)$purposeId);
						if ($purpose) {
							$locale->$setter($purpose, $value);
						}
					} else {
						$locale->$setter($value);
					}
				}
			}
			$locale->save();

			// Update the translations of embeddings if the locale is enabled
			if ($locale->isEnabled()) {
				self::updateEmbeddings($updater);
			}

			$this->addFlash('success',
				$translator->trans('Locale %locale% has been updated.', ['%locale%' => $localeName])
			);

			if ($request->request->getAlnum('action') === 'saveClose') {
				return $this->redirectToRoute('app_setting_locale_list', [], 303);
			} else {
				return $this->redirectToRoute('app_setting_locale_edit', ["id"=>$locale->getId()], 303);
			}
		}
		catch (LocaleAlreadyExistsException $e) {
			$this->addFlash('danger',
				$translator->trans('Locale %locale% already exists.', ['%locale%' => $localeName])
			);
		}
		catch (Exception $e) {
			$this->addFlash('danger',
				$translator->trans('An unknown error has occured.')
			);
		}

		$fields = self::getStandardFieldGroups($translator);
		$fields = array_map(function ($fieldset) use ($request) {
			return array_map(function ($field) use ($request) {
					$field['user_input'] = $request->request->get($field['field_id'], '');
					return $field;
				}, $fieldset);
		}, $fields);

		$locales = LocaleService::all();

		// Consider user locales a "used name" because the index lists them in addition to domain locales.
		// To create the respective domain locale, one should use the "edit" button.
		$used_names = array_map(function (LocaleBase $item) { return $item->getName(); }, $locales);

		return $this->render('locale/new.html.twig', [
			'user' => User::loggedInUser(),
			'fields' => $fields,
			'used_names' => $used_names,
			'languages' => Language::all(),
			'locale_name' => $localeName,
			'localeEnabled' => (bool)$request->request->get('localeEnabled'),
		]);
	}

	/**
	 * @Route("/{localeName}.json", name="ajax", methods={"HEAD", "GET"})
	 * @param string $localeName
	 * @param TranslatorInterface $translator
	 * @return JsonResponse
	 */
	public function ajaxLocale(string $localeName, TranslatorInterface $translator): JsonResponse
	{
		$locale = LocaleService::find($localeName);
		if (!$locale) {
			return $this->json(['success'=>false, 'error'=>'locale not found'], 404);
		}

		$fieldGroups = self::getStandardFieldGroups($translator);
		$translation = array_map(static function (array $field) use ($locale) {
			[$getter, $purposeId] = self::buildGetterSetterName($field['field_id']);
			if (!method_exists($locale, $getter)) {
				return null;
			}
			elseif ($purposeId and $purpose = Purpose::find((string)$purposeId)) {
				return $locale->$getter($purpose);
			}
			else {
				return $locale->$getter();
			}
		}, array_merge([], ...array_values($fieldGroups)));

		// Discard empty values
		$translation = array_filter($translation, static function ($value) { return is_string($value) && $value != ''; });

		return $this->json($translation);
	}

	/**
	 * @Route("/{id}", name="edit")
	 *
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param EmbeddingUpdateService $updater
	 * @return Response
	 */
	public function editLocale(string $id, Request $request, TranslatorInterface $translator, EmbeddingUpdateService $updater): Response
	{
		$locale = LocaleModel::byId($id);
		if ($locale === null) {
			return $this->redirectToRoute('app_setting_locale_list');
		}

		if ($request->getRealMethod() === 'POST') {
			$fields = self::getStandardFieldGroups($translator);
			try {
				$locale
					->setEnabled((bool)$request->request->get('localeEnabled'))
					->setUpdateParameterBlacklist($request->request->all('updateParameterBlacklist'))
				;

				// Alle Daten in den Feldern speichern
				foreach ($fields as $fieldset) {
					foreach ($fieldset as $field) {
						list($setter, $purposeId) = $this->buildGetterSetterName($field['field_id'], 'set');
						$value = trim($request->request->get($field['field_id'], ''));
						if ($setter == 'setPurposeName' or $setter == 'setPurposeDescription') {
							if (($purpose = Purpose::find((string)$purposeId))) {
								$locale->$setter($purpose, $value);
							}
						} else {
							$locale->$setter($value);
						}
					}
				}
				$locale->save();

				// Update the translations of embeddings if the locale is enabled
				if ($locale->isEnabled()) {
					self::updateEmbeddings($updater);
				}

				$this->addFlash('success',
					$translator->trans('Locale %locale% has been updated.', ['%locale%' => $locale->getName()])
				);

				if ($request->request->getAlnum('action') === 'saveClose') {
					return $this->redirectToRoute('app_setting_locale_list', [], 303);
				} else {
					return $this->redirectToRoute('app_setting_locale_edit', ["id"=>$id], 303);
				}
			}
			catch (Exception $e) {
				$this->addFlash('danger', $translator->trans('An unknown error has occured.'));
			}
		}

		// Gespeicherte Werte setzen
		$fields = self::getStandardFieldGroups($translator);
		$fields = array_map(function ($fieldset) use ($locale) {
			return array_map(function ($field) use ($locale) {
				$getter = 'get'.preg_replace_callback('/(?:^|_)([a-z])/',
					function ($match) { return strtoupper($match[1]); },
					$field['field_id']
				);
				if (strpos($getter, 'getPurposeName-') === 0) {
					if (($purpose = Purpose::find(substr($getter, 15)))) {
						$getter = substr($getter, 0, 14);
						$field['user_input'] = $locale->$getter($purpose) ?? '';
					}
				} elseif (preg_match('~^getPurposeDescription-(?<purposeId>[0-9a-f]+)$~', $getter, $match)) {
					if (($purpose = Purpose::find($match['purposeId']))) {
						$field['user_input'] = $locale->getPurposeDescription($purpose);
					}
				} else {
					$field['user_input'] = $locale->$getter() ?? '';
				}
				return $field;
			}, $fieldset);
		}, $fields);

		// Standardwerte aus User-Theme setzen
		$localeTemplate = UserLocale::find($locale->getName());
		if ($localeTemplate) {
			$fields = array_map(function ($fieldset) use ($localeTemplate) {
				return array_map(function ($field) use ($localeTemplate) {
						list($getter, $purposeId) = $this->buildGetterSetterName($field['field_id'], 'get');
						if ($getter == 'getPurposeName' or $getter == 'getPurposeDescription') {
							if (($purpose = Purpose::find((string)$purposeId))) {
								$field['default'] = $localeTemplate->$getter($purpose);
							}
						} else {
							$field['default'] = $localeTemplate->$getter();
						}
						return $field;
					}, $fieldset);
			}, $fields);
		}

		return $this->render('locale/edit.html.twig', [
			'user' => User::loggedInUser(),
			'locale' => $locale,
			'fields' => $fields,
		]);
	}

	/**
	 * @Route("/{id}/delete", name="delete")
	 *
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws \App\Exception\ConfigNotWritableException
	 */
	public function deleteLocale(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$locale = LocaleModel::byId($id);
		if ($locale === null) {
			return $this->redirectToRoute('app_setting_locale_list');
		}

		$localeName = $locale->getName();
		$localeBlacklist = [
			'de_DE',
			'en_US',
		];

		// Bestimmte Übersetzungen dürfen nicht gelöscht werden
		if (in_array($localeName, $localeBlacklist) and !UserLocale::exists($localeName)) {
			$this->addFlash('info',
				$translator->trans('Locale %locale% may not be deleted.', ['%locale%' => $localeName])
			);
			return $this->redirectToRoute('app_setting_locale_list');
		}

		if ($request->getRealMethod() === 'POST') {
			LocaleModel::delete($locale->getId());

			$this->addFlash('success',
				$translator->trans('Locale %locale% has been deleted.', ['%locale%' => $localeName])
			);

			return $this->redirectToRoute('app_setting_locale_list');
		}

		return $this->render('locale/delete.html.twig', [
			'user' => User::loggedInUser(),
			'locale' => $locale,
		]);
	}

	/**
	 * @param EmbeddingUpdateService $updater
	 * @return void
	 */
	private static function updateEmbeddings(EmbeddingUpdateService $updater): void
	{
		$user = User::loggedInUser()->getAccountOwner();

		if ($user->isAllowAutoUpdateOfEmbeddings() == false) {
			return;
		}

		foreach (Embedding::all() as $embedding) {
			if ($embedding->isAutoUpdated()) {
				$updater->updateEmbedding($embedding);
			}
		}
		EmbeddingAsset::flush();
		Embedding::flush();
	}
}
