<?php

namespace App\Controller;

use App\Component\Api\HostingApi;
use App\Component\Licensing\ApiClient as LicensingApiClient;
use App\Component\LoggerInterface;
use App\Config;
use App\Event\AccountUpdateEvent;
use App\Exception\CsrfTokenException;
use App\Exception\SecondFactorRequiredException;
use App\Exception\UserIsDeactivatedException;
use App\Component\Mail\MailMessage;
use App\Model\CsrfToken;
use App\Model\BackendLanguage;
use App\Model\LoginAttemptsLog;
use App\Model\User;
use App\Utils;
use DateTime;
use DateTimeImmutable;
use DateTimeZone;
use Exception;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use Symfony\Component\Mailer\MailerInterface;

class Login extends AbstractController
{
	public const HOSTING_LOGIN_PBKDF2_ALGORITHM = 'sha256';
	public const HOSTING_LOGIN_PBKDF2_ITERATIONS = 500;
	public const HOSTING_LOGIN_PBKDF2_KEY_LENGTH = 24;

	/**
	 * Diese Methode bildet den Einstiegspunkt der Loginroutine.
	 *
	 * @Route("/login", methods={"HEAD","GET"}, name="app_login")
	 */
	public function loginPage(Request $request): Response
	{
		/** @var Config $config */
		$config = Config::getInstance();
		$session = $request->getSession();
		$loggedInUser = User::loggedInUser();
		$loggingInUser = User::loggingInUser();
		$username = $request->query->get('username', null);

		if (User::count() == 0) {
			return $this->redirectToRoute('app_setup', [], 307);
		}
		else if ($loggedInUser and ($username == null or $username === $loggedInUser->getUsername())) {
			return $this->redirectAfterLogin($request);
		}
		else if ($loggingInUser and ($username == null or $username === $loggingInUser->getUsername())) {
			return $this->redirectToRoute('app_login_2fa', ['backTo' => $request->query->get('backTo', null)], 303);
		}

		list($route, $parameters) = $this->parseReturnTo($request->query->get('backTo', ''));

		$response = $this->render('login/'.$this->getCkDir().'index.html.twig', [
			'error' => null,
			'username' => $request->query->get('username', null),
			'loginToken' => $request->query->get('loginToken', null),
			'backTo' => $request->query->get('backTo', null),
			'backToRoute' => $route,
			'backToArgs' => json_encode($parameters),
			'lang' => $request->query->get('lang'),
			'isck' => Utils::isClickskeksCloud(),
		]);

		Utils::resetCacheControl($response);
		if ($username !== null or $session->has('locale')) {
			$response->setCache(['max_age'=>0, 'public'=>false]);
		} else {
			$response->setCache(['max_age'=>180, 'public'=>true]);
		}

		return $response;
	}

	/**
	 * Parsen eines eventuellen 'returnTo'-Parameters
	 *
	 * @param string|null $string
	 * @return array [$route, $parameters]
	 */
	private function parseReturnTo($string): array
	{
		$route = 'app_root';
		$parameters = null;

		// Sicherheitscheck
		if (!$string or !is_string($string)) {
			return [$route, $parameters];
		}

		// String dekodieren und bei falschem Format abbrechen
		$routeinfo = @json_decode(base64_decode(strtr($string, ['-'=>'+', '_'=>'/'])), true, 5);
		if (!is_array($routeinfo) or count($routeinfo) !== 2) {
			return [$route, $parameters];
		}

		// Route und Parameter bestimmen
		list($route, $parameters) = $routeinfo;

		// Weiterer Sicherheitscheck
		if (!$route or !is_string($route) or !is_array($parameters)) {
			return ['app_root', []];
		}

		return [$route, $parameters];
	}

	/**
	 * Weiterleiten zur in GET/POST gespeicherten Route, ansonsten zum Dashboard
	 *
	 * @param Request $request
	 * @return RedirectResponse
	 */
	public function redirectAfterLogin(Request $request): RedirectResponse
	{
		//self::$IS_LOGIN = true;

		// Routen-Parameter bestimmen
		if ($request->request->has('backToRoute')) {
			$route = $request->request->get('backToRoute', 'app_root');
			$parameters = json_decode($request->request->get('backToArgs', '[]'), true, 5);
		} else {
			list($route, $parameters) = $this->parseReturnTo($request->query->get('backTo', ''));
		}
		if (!$route) {
			$route = 'app_root';
			$parameters = [];
		}
		if (!is_array($parameters)) {
			$route = 'app_root';
			$parameters = [];
		}

		// Redirect zurück
		try {
			return $this->redirectToRoute(
				$route,
				$parameters,
				303
			);
		} catch (\Throwable $e) {
			return $this->redirectToRoute('app_root', [], 303);
		}
	}

	/**
	 * Aktualisiert den Lizenzstatus.
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param LicensingApiClient $api
	 * @return void
	 */
	private function licenseCheck(Request $request, LoggerInterface $logger, TranslatorInterface $translator, LicensingApiClient $api): void
	{
		try {
			$licenseData = $api->registerLicense(Utils::getLicenseKey(), $request->getHttpHost());
		}
		catch (Exception $e) {
			$logger->logWarning('LicenseCheck', 'Error while checking license: '.$e->getMessage());
			$this->addFlash('warning', $translator->trans($e->getMessage()));
			return;
		}

		$config = Config::getInstance();
		$mayWhitelabel = $licenseData->whitelabel;
		$currentWhitelabel = $config->get('whitelabel');
		if ($currentWhitelabel != $mayWhitelabel) {
			Config::getInstance()->set('whitelabel', $mayWhitelabel);
			$logger->logInfo('LicenseCheck', ($mayWhitelabel) ? 'Whitelabel license unlocked' : 'Whitelabel license revoked');
		}

		// Das Limit ist für Administratoren derzeit nur in der Multi-Domain-Edition relevant
		foreach (User::byRole(User::ROLE_ADMIN) as $user) {
			$user->setMaxDomainCount($licenseData->limit);
		}
		User::flush();
	}

	private function setSessionLanguage(User $user, Request $request)
	{
		// Sprache setzen
		$selectedLanguage = BackendLanguage::byFuzzyLangCode((string)$request->request->get('locale_name', ''));
		$languageCode = null;

		if ($selectedLanguage) {
			$languageCode = $selectedLanguage->getLongCode();
		}
		else {
			$languageCode = $user->getDefaultLocale()->getName();
		}

		if ($languageCode) {
			$session = $request->getSession();
			$request->setLocale( $languageCode);
			$session->set('locale', $languageCode);
			$session->save();
		}
	}

	/**
	 * @param User $user
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @param TranslatorInterface $translator
	 * @param LicensingApiClient $licenseApi
	 * @return Response
	 */
	public function afterSuccessfulLogin(User $user, Request $request, LoggerInterface $logger, TranslatorInterface $translator, LicensingApiClient $licenseApi): Response
	{
		$username = $user->getUsername();
		$originalUsername = null;
		$targetUser = $user;

		if ($user->getRole() == User::ROLE_ADMIN) {
			$this->licenseCheck($request, $logger, $translator, $licenseApi);

			$switchTo = $request->request->get('switchTo');
			if ($switchTo and $switchTo !== $username) {
				$targetUser = User::find($switchTo);
				if ($targetUser and $targetUser->getRole() === User::ROLE_CLIENT) {
					$targetUser->switchTo(true);
					$originalUsername = $username;
					$username = $switchTo;
				}
			}
		}

		$this->setSessionLanguage($user, $request);

		if ($originalUsername !== null) {
			$logger->logInfo("Login","Successful login into '$username' by user: ".$originalUsername);
		} else {
			$logger->logInfo("Login","Successful login by user: ".$username);
		}
		switch ($request->request->get('format', 'html')) {
			case 'json':
				return $this->json([
					'userid' => $targetUser->getId(),
					'username' => $username,
					'originalUsername' => $originalUsername,
					'csrfToken' => (string)CsrfToken::get(),
				], 200, ['Clear-Site-Data' => '"cache"']);
			default:
				return $this->redirectAfterLogin($request);
		}
	}

	/**
	 * Check des Login-Tokens für passwortlose Anmeldung mit der Hosting-API
	 */
	private function checkHostingLoginToken(string $username, string $token): bool
	{
		$timestamp = new DateTimeImmutable('now', new DateTimeZone('UTC'));
		$api = HostingApi::getInstance();
		$key = hash_pbkdf2(
			self::HOSTING_LOGIN_PBKDF2_ALGORITHM,
			$api->getApiKey(),
			'login:'.$username,
			self::HOSTING_LOGIN_PBKDF2_ITERATIONS,
			self::HOSTING_LOGIN_PBKDF2_KEY_LENGTH,
			true
		);
		// Check ± 4 Minuten, damit Logins bei nicht perfekt synchronen Uhren
		// trotzdem funktionieren
		$expectedTokens = [
			hash_hmac('sha256', $timestamp->modify('- 4 minutes')->format('Y-m-d H:i'), $key, false),
			hash_hmac('sha256', $timestamp->modify('- 3 minutes')->format('Y-m-d H:i'), $key, false),
			hash_hmac('sha256', $timestamp->modify('- 2 minutes')->format('Y-m-d H:i'), $key, false),
			hash_hmac('sha256', $timestamp->modify('- 1 minutes')->format('Y-m-d H:i'), $key, false),
			hash_hmac('sha256', $timestamp->format('Y-m-d H:i'), $key, false),
			hash_hmac('sha256', $timestamp->modify('+ 1 minutes')->format('Y-m-d H:i'), $key, false),
			hash_hmac('sha256', $timestamp->modify('+ 2 minutes')->format('Y-m-d H:i'), $key, false),
			hash_hmac('sha256', $timestamp->modify('+ 3 minutes')->format('Y-m-d H:i'), $key, false),
			hash_hmac('sha256', $timestamp->modify('+ 4 minutes')->format('Y-m-d H:i'), $key, false),
		];
		foreach ($expectedTokens as $expectedToken) {
			if (hash_equals($expectedToken, $token)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Legacy Login-Einstiegspunkt
	 * @deprecated
	 *
	 * @Route("/login_now", methods={"POST","GET"}, name="app_login_now_legacy")
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @param TranslatorInterface $translator
	 * @param LicensingApiClient $licenseApi
	 */
	public function loginActionLegacy(Request $request, LoggerInterface $logger, TranslatorInterface $translator, LicensingApiClient $licenseApi): Response
	{
		// Legacy: Login über den Shop per GET mit Timekey
		$timekey = sha1(date("d.G"));
		if (!$request->request->has('username') && (string)$request->query->get('timekey', $request->request->get('timekey')) === $timekey) {
			$request->request->set('username', trim((string)$request->query->get('username', $request->request->get('username'))));
			$request->request->set('password', trim((string)$request->query->get('password', $request->request->get('password'))));
			$session = Utils::getSession();
			$session->set('secChecked', true);
		}

		return $this->loginAction($request, $logger, $translator, $licenseApi);
	}

	/**
	 * Diese Methode bildet den Einstiegspunkt der Loginroutine.
	 *
	 * @Route("/login", methods={"POST"}, name="app_login_post")
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @param TranslatorInterface $translator
	 * @param LicensingApiClient $licenseApi
	 * @return Response
	 */
	public function loginAction(Request $request, LoggerInterface $logger, TranslatorInterface $translator, LicensingApiClient $licenseApi): Response
	{
		$response = new Response();
		$error = null;
		$passwordless = false;

		// Wenn das Formular abgeschickt wurde, Daten prüfen und ggf. Benutzer einloggen
		$username = trim((string)$request->request->get('username'));
		$password = trim((string)$request->request->get('password'));

		// IP-Sperre
		$attempts = LoginAttemptsLog::forKind('login');
		if ($attempts->countAttempts($request->getClientIp()) > 5) {
			$response->setStatusCode(401);
			$error = $translator->trans('Too many attempts. Your device is blocked for up to %n% minutes.', [
				'%n%' => (int)ceil(LoginAttemptsLog::TIMEWINDOW_MINUTES),
			]);
		}

		// Hosting-Token-Login
		if ($error === null and $request->request->has('loginToken')) {
			$token = trim((string)$request->request->get('loginToken'));
			$tokenValid = $this->checkHostingLoginToken($username, $token);
			if ($tokenValid) {
				$passwordless = true;
				// für Onboarding von Checkout aus:
				$session = Utils::getSession();
				$session->set('secChecked', true);
			} else {
				$attempts->append($request->getClientIp());
				$response->setStatusCode(401);
				$error = $translator->trans("invalid or expired login token");
			}
		}

		// Login durchführen, wenn bis hierhin noch kein Fehler
		if ($error === null) {
			try {
				if ($passwordless) {
					$user = User::loginPasswordless($username);
				} else {
					$user = User::login($username, $password);
				}
				if ($user) {
					return $this->afterSuccessfulLogin($user, $request, $logger, $translator, $licenseApi);
				}
				else {
					$attempts->append($request->getClientIp());
					$response->setStatusCode(401);
					$error = $translator->trans('Please check your credentials.');
					$logger->logError("Login", 'Failed login due to wrong credentials for user '.$username);
				}
			}
			catch (UserIsDeactivatedException $exception) {
				$reason = $exception->getExtendedReason();
				if ($reason) {
					$error = $translator->trans($reason);
					$logger->logError("Login", 'Failed login due to a plugin rejecting the login for user '.$username.': '.$reason);
					$response->setStatusCode(401);
				} else {
					$error = $translator->trans('User is deactivated from logging into the system.');
					$logger->logError("Login", 'Failed login due to deactivation for user '.$username);
					$response->setStatusCode(401);
				}
			}
			catch (SecondFactorRequiredException $exception) {
				$user = $exception->getUser();
				$this->setSessionLanguage($user, $request);

				// Session-Re-Login: auf 2FA verzichten
				if ((string)$request->request->get('resumeToken') === $user->getSessionResumeToken()) {
					$user->switchTo();
					return $this->afterSuccessfulLogin($user, $request, $logger, $translator, $licenseApi);
				}

				$error = $translator->trans('Two Factor Authentication required');
				if ($request->request->has('backToRoute')) {
					$params = [
						'backTo' => Utils::encodeReturnTo(
							(string)$request->request->get('backToRoute', 'app_root'),
							(array)json_decode($request->request->get('backToArgs', '{}'), true)
						)
					];
				} else {
					$params = [];
				}

				$methods = $exception->getMethods();
				// Wenn nur eine Methode, dann direkt zu dieser springen
				if (count($methods) === 1) {
					$method = reset($methods);
					$response = $this->redirectToRoute($method['route'], array_merge($method['routeParams'] ?? [], $params), 303);
				} else {
					$response = $this->redirectToRoute('app_login_2fa', $params, 303);
				}
			}

			$attempts->cleanup();
		}

		switch ($request->request->get('format', 'html')) {
			case 'json':
				$json = [
					'error' => $error,
					'username' => $username,
					'backToRoute' => $request->request->get('backToRoute', 'app_root'),
					'backToArgs' => $request->request->get('backToArgs', ''),
				];
				$statusCode = $response->getStatusCode();
				$location = $response->headers->get('Location');
				if ($location && $statusCode == 303) {
					$json['nextStep'] = $location;
					$statusCode = 401;
				}
				return $this->json($json, $statusCode);
			default:

				return $this->render('login/'.$this->getCkDir().'index.html.twig', [
					'error' => $error,
					'username' => $username,
					'backTo' => $request->query->get('backTo', null),
					'backToRoute' => $request->request->get('backToRoute', 'app_root'),
					'backToArgs' => $request->request->get('backToArgs', ''),
					'lang' => $request->query->get('lang'),
					'isck' => Utils::isClickskeksCloud(),
				], $response);

		}
	}

	/**
	 * @return string
	 */
	private function getCkDir()
	{
		if (Utils::isClickskeksCloud()) {
			return "ck/";
		}
		return "";
	}


	/**
	 * Über diesen Endpunkt kann nach einem Benutzerwechsel zum ursprünglichen Benutzer
	 * zurückgesprungen werden.
	 *
	 * @Route("/login/switchBack", methods={"POST"}, name="app_login_switch_back")
	 * @param Request $request
	 * @return Response
	 */
	public function switchBackAction(Request $request): Response
	{
		$user = User::originalLoggedInUser();

		if ($user) {
			$user->switchTo();
			$response = $this->redirectToRoute('app_hosting_client_main_account_index', [], 303);
		}
		else {
			$response = $this->redirectToRoute('app_login', [], 303);
		}
		$response->headers->set('Clear-Site-Data', '"cache"');
		return $response;
	}

	/**
	 * Diese Methode loggt den Benutzer aus und leitet zur Loginmaske weiter.
	 *
	 * @Route("/logout", methods={"HEAD","GET"}, name="app_logout")
	 */
	public function logoutPage(): Response
	{
		return $this->render('login/logout.html.twig', []);
	}

	/**
	 * Diese Methode loggt den Benutzer aus und leitet zur Loginmaske weiter.
	 *
	 * @Route("/logout", methods={"POST"}, name="app_logout_post")
	 */
	public function logoutAction(Request $request): RedirectResponse
	{
		User::logout();
		if ($request->request->has('backTo')) {
			return $this->redirectToRoute('app_login', [ 'backTo'=>$request->request->get('backTo', '') ], 303);
		} else {
			return $this->redirectToRoute('app_login', [], 303);
		}
	}

	/**
	 * Diese Methode gibt zurück, ob und als wer der User eingeloggt ist
	 *
	 * @Route("/login/check", methods={"HEAD","GET"}, name="app_login_check")
	 */
	public function loginCheckAction(): JsonResponse
	{
		$user = User::loggedInUser();
		$origUser = User::originalLoggedInUser();

		$session = Utils::getSession();
		$session->save();

		$data = [
			'auth' => ($user !== null),
		];

		if ($user) {
			$data['userid'] = $user->getId();
			$data['user'] = $user->getUsername();
		}
		if ($origUser) {
			$data['origUser'] = $origUser->getUsername();
		}

		return $this->json($data, 200)->setEtag(sha1(($data['userid'] ?? '').'~'.($data['user'] ?? '').'~'.($data['origUser'] ?? '')), true);
	}

	/**
	 * Passwort-Recovery: 1. Schritt - Formular
	 *
	 * @Route("/login/forgot_password",  methods={"HEAD","GET"}, name="app_login_forgot_password")
	 */
	public function forgotPasswordPage(Request $request, ?TranslatorInterface $translator=null): Response
	{
			return $this->render('login/'.$this->getCkDir().'forgot_password.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'username' => $request->query->get('username', null),
			'backTo' => $request->query->get('backTo', null),
			'lang' => $request->query->get('lang'),
			'isck' => Utils::isClickskeksCloud(),
		]);
	}

	/**
	 * Passwort-Recovery: 1. Schritt - Formular absenden, E-Mail verschicken
	 *
	 * @Route("/login/forgot_password",  methods={"POST"}, name="app_login_forgot_password_post")
	 */
	public function forgotPasswordAction(Request $request, MailerInterface $mailer, ?TranslatorInterface $translator=null): Response
	{
		usleep(800000);
		$response = new Response();
		$username = trim($request->request->get('username', ''));
		$error = null;
		try {
			CsrfToken::get()->checkRequest($request);
		}
		catch (CsrfTokenException $e) {
			$error = $translator->trans('CSRF token is missing or invalid. Please try again.');
			return $this->render('login/forgot_password.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'error' => $error,
				'username' => $username,
				'lang' => $request->query->get('lang'),
			], $response);
		}

		$attempts = LoginAttemptsLog::forKind('recovery');
		if ($attempts->countAttempts($request->getClientIp()) > 4) {
			$error = $translator->trans('Too many attempts. Your device is blocked for up to %n% minutes.', [
				'%n%' => (int)ceil(LoginAttemptsLog::TIMEWINDOW_MINUTES),
			]);
			return $this->render('login/forgot_password.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'error' => $error,
				'username' => $username,
				'lang' => $request->query->get('lang'),
			], $response);
		}

		$user = User::find($username);

		if (!$user or !$user->isActive())
		{
			$attempts->cleanup();
			$attempts->append($request->getClientIp());
			$error = $translator->trans('Please check your input.');
			return $this->render('login/forgot_password.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'error' => $error,
				'username' => $username,
				'lang' => $request->query->get('lang'),
			], $response);
		}

		$timediff = (new Datetime('now'))->getTimestamp() - $user->getPasswordRecoveryTime()->getTimestamp();
		if ($timediff < 300) {
			$attempts->cleanup();
			$attempts->append($request->getClientIp());
			$error = $translator->trans('You already requested a password recovery in the last five minutes.');
			return $this->render('login/forgot_password.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'error' => $error,
				'username' => $username,
				'lang' => $request->query->get('lang'),
			], $response);
		}

		$attempts->cleanup();
		$token = bin2hex(random_bytes(20));
		$user->setPasswordRecoveryKey($token);

		// Bei gesperrtem Passwort ist kein Passwort-Recovery möglich
		if ($user->isPasswordLocked()) {
			$attempts->append($request->getClientIp());
			$error = $translator->trans('Password recovery not available. Please contact your administrator.');
			return $this->render('login/forgot_password.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'error' => $error,
				'username' => $username,
				'lang' => $request->query->get('lang'),
			], $response);
		}

		$message = (new MailMessage($translator->trans('CCM19 Account recovery')))
		->to($user->getEmailAddress())
		->text($this->renderView(
				'emails/password_recovery.txt.twig',
				['user' => $user, 'token' => $token]
			), 'text/plain');

		try {
			$mailer->send($message);
		} catch (Exception $e) {
			return $this->render('login/forgot_password.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'error' => $translator->trans('Email delivery failed. Please contact an administrator.'),
				'username' => $username,
				'lang' => $request->query->get('lang'),
				'email' => $user->getEmailAddress(),
			], $response);
		}

		$user->save();

		return $this->render('login/forgot_password.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'success' => $translator->trans('In a few minutes you should receive an email with further instructions to reset your account.'),
			'username' => $username,
			'lang' => $request->query->get('lang'),
		], $response);
	}

	/**
	 * Passwort-Recovery: 2. Schritt - Passwort ändern
	 *
	 * @Route("/login/recover",  methods={"HEAD","GET", "POST"}, name="app_login_recover")
	 */
	public function recoverPasswordAction(Request $request, LoggerInterface $logger, TranslatorInterface $translator, LicensingApiClient $licenseApi): Response
	{
		$response = new Response();
		$now = new DateTime('now');
		usleep(500000);
		User::beginWrite();
		$user = User::find($request->query->get('u', ''));
		if ($user) {
			$token = $user->getPasswordRecoveryKey();
			$time = $user->getPasswordRecoveryTime();
		}
		else {
			$token = null;
			$time = null;
		}

		// Token und Zeitraum prüfen
		if ($token === '' or $token === null or $now->getTimestamp() - $time->getTimestamp() > 86400 or !hash_equals($token, $request->query->get('t'))) {
			return $this->redirectToRoute('app_login', [], 303);
		}
		assert($user !== null);

		if ($request->getRealMethod() !== 'POST') {
			// Bei GET: Formular anzeigen
			return $this->render('login/recover.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'username' => $user->getUsername(),
				'token' => $request->query->get('t'),
			]);
		}

		$password = $request->request->get('password');
		$password2 = $request->request->get('password2');

		if (!hash_equals($password, $password2)) {
			return $this->render('login/recover.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'error' => $translator->trans('The passwords did not match!'),
				'username' => $user->getUsername(),
				'token' => $request->query->get('t'),
				'lang' => $request->query->get('lang'),
			]);
		}

		// Bei gesperrtem Passwort ist kein Passwort-Recovery möglich
		if ($user->isPasswordLocked()) {
			return $this->render('login/recover.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'error' => $translator->trans('Password recovery not available. Please contact your administrator.'),
				'username' => $user->getUsername(),
				'token' => $request->query->get('t'),
				'lang' => $request->query->get('lang'),
			]);
		}

		$user->setPassword($password);
		$user->setPasswordRecoveryKey(null);
		$user->generateSessionResumeToken();
		// 2FA deaktivieren
		$user->clearAllTwoFactorAuthParameters();

		$event = new AccountUpdateEvent($user, AccountUpdateEvent::CAUSE_PWRECOVERY);
		$event->setPasswordChanged();
		Utils::dispatchEvent(AccountUpdateEvent::NAME_PRE_SAVE, $event);

		$user->save();

		Utils::dispatchEvent(AccountUpdateEvent::NAME_POST_SAVE, $event);

		try {
			$user = User::login($user->getUsername(), $password);
			if ($user) {
				if ($user->getRole() == User::ROLE_ADMIN) {
					$this->licenseCheck($request, $logger, $translator, $licenseApi);
				}
				return $this->redirectToRoute('app_root', [], 303);
			}
			else {
				$error = $translator->trans('Please check your credentials.');
				$response->setStatusCode(401);
			}
		}

		catch (UserIsDeactivatedException $exception) {
			$error = $translator->trans('User is deactivated from logging into the system.');
			$response->setStatusCode(401);
		}

		return $this->render('login/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'error' => $error,
			'username' => ($user ? $user->getUsername() : null),
			'lang' => $request->query->get('lang'),
		], $response);
	}

	/**
	 * Login-Popup bei abgelaufener Sitzung. Route wird intern verwendet.
	 *
	 * @Route("/login/relogin_popup",  methods={"HEAD","GET"}, name="app_login_relogin_popup")
	 */
	public function reloginPopup(Request $request): Response
	{
		$response = $this->render('_components/relogin_popup.html.twig', []);
		Utils::resetCacheControl($response);
		$response->setCache(['max_age'=>30, 'public'=>false]);
		$vary = $response->getVary();
		$vary[] = 'Set-Cookie';
		$response->setVary($vary);
		return $response;
	}

	/**
	 * Einsprungpunkt 2-Faktor-Authentifizierung
	 *
	 * @Route("/login/2fa", methods={"HEAD","GET"}, name="app_login_2fa")
	 */
	public function secondFactorPage(Request $request): Response
	{
		$user = User::loggingInUser();
		if (!$user) {
			return $this->redirectToRoute('app_login');
		}

		$backTo = $request->query->get('backTo', null);

		$methods = User::currentTwoFactorAuths();
		if ($backTo) {
			$methods = array_map(function (array $method) use ($backTo) {
				$method['routeParams']['backTo'] = $backTo;
				return $method;
			}, $methods);
		}

		$response = $this->render('login/2fa.html.twig', [
			'error' => null,
			'csrfToken' => CsrfToken::get(),
			'backTo' => $backTo,
			'username' => $user->getUsername(),
			'methods' => $methods,
		]);

		Utils::resetCacheControl($response);
		$response->setCache(['max_age'=>30, 'public'=>false]);

		return $response;
	}
}
