<?php

namespace App\Controller;

use App\Component\MailSubscriptions;
use App\Model\LoginAttemptsLog;
use App\Model\User;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class MailUnsubscribe extends UnrestrictedController
{
	private function checkCredentials(Request $request, ?User $user, $key, $suppliedToken, MailSubscriptions $subscriptions, TranslatorInterface $translator): ?Response
	{
		// Verhindern, dass zu oft durchprobiert werden kann
		$attempts = LoginAttemptsLog::forKind('unsubscribe');
		if ($attempts->countAttempts($request->getClientIp()) > 30) {
			$error = $translator->trans('Too many attempts. Your device is blocked for up to %n% minutes.', [
				'%n%' => (int)ceil(LoginAttemptsLog::TIMEWINDOW_MINUTES),
			]);
			$response = $this->render('mail_unsubscribe/index.html.twig', ['error'=>$error, 'user'=>null, 'subscriptionKey'=>$key]);
			$response->setStatusCode(429);
			return $response;
		}
		$attempts->cleanup();

		// API-Key prüfen
		if (!$user) {
			$error = $translator->trans('Invalid unsubscribe link. Please check if the link was wrapped or click the link in a more recent mail.');
			$response = $this->render('mail_unsubscribe/index.html.twig', ['error'=>$error, 'user'=>null, 'subscriptionKey'=>$key]);
			$response->setStatusCode(403);
			return $response;
		}

		// Sicherheits-Token prüfen
		$correctToken = strtolower($subscriptions->buildValidationCode($user, $key));
		if (!hash_equals($correctToken, $suppliedToken)) {
			$attempts->append($request->getClientIp());

			$error = $translator->trans('Invalid unsubscribe link. Please check if the link was wrapped or click the link in a more recent mail.');
			$response = $this->render('mail_unsubscribe/index.html.twig', ['error'=>$error, 'user'=>null, 'subscriptionKey'=>$key]);
			$response->setStatusCode(403);
			return $response;
		}
		return null;
	}

	/**
	 * @Route("/unsubscribe/{key}", name="app_mail_unsubscribe", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @return Response
	 */
	public function unsubscribeForm(Request $request, string $key, MailSubscriptions $subscriptions, TranslatorInterface $translator): Response
	{
		$apiKey = (string)$request->query->getAlnum('u');
		$user = User::byApiKey($apiKey);
		$suppliedToken = str_replace(' ', '', strtolower($request->query->getAlnum('t')));

		$error = $this->checkCredentials($request, $user, $key, $suppliedToken, $subscriptions, $translator);
		if ($error) {
			return $error;
		}

		$possibleSubscriptions = $subscriptions->list($user);
		if (!isset($possibleSubscriptions[$key])) {
			// Redirect bei Groß-/Kleinschreibungs-Abweichung
			foreach ($possibleSubscriptions as $k => $v) {
				if (strtolower($k) == strtolower($key)) {
					return $this->redirectToRoute('app_external_unsubscribe', ['key'=>$key, 'apiKey'=>$apiKey, 'token'=>$suppliedToken], 301);
				}
			}
		}

		$info = $possibleSubscriptions[$key] ?? ['label'=>$key];
		$response = $this->render('mail_unsubscribe/index.html.twig', [
			'user'=>$user, 'subscriptionKey'=>$key, 'subscriptionInfo'=>$info, 'token'=>$suppliedToken
		]);
		return $response;
	}

	/**
	 * @Route("/unsubscribe/{key}", name="app_mail_unsubscribe_post", methods={"POST"})
	 * @param Request $request
	 * @return Response
	 */
	public function unsubscribe(Request $request, string $key, MailSubscriptions $subscriptions, TranslatorInterface $translator): Response
	{
		$apiKey = (string)$request->query->getAlnum('u');
		$user = User::byApiKey($apiKey);
		$suppliedToken = str_replace(' ', '', strtolower($request->query->getAlnum('t')));

		$error = $this->checkCredentials($request, $user, $key, $suppliedToken, $subscriptions, $translator);
		if ($error) {
			return $error;
		}

		$user->unsubscribeMailing($key)->save();

		$possibleSubscriptions = $subscriptions->list($user);
		$info = $possibleSubscriptions[$key] ?? ['label'=>$key];
		$response = $this->render('mail_unsubscribe/index.html.twig', [
			'user'=>$user, 'success'=>true, 'subscriptionKey'=>$key, 'subscriptionInfo'=>$info, 'token'=>$suppliedToken
		]);
		return $response;
	}
}
