<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Component\Plugin\Manager as PluginManager;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\MenuData;
use App\Model\SubGroups;
use App\Model\User;
use App\Model\CsrfToken;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class ManageSubGroups extends AbstractController
{
	/**
	 * @Menu("Group management", group="Users and Groups", icon="fa-group", routeGroup="app_sub_group_", order=80)
	 * @Route("/manage_groups", name="app_sub_group_index", methods={"HEAD", "GET", "POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function index(Request $request, TranslatorInterface $translator, PluginManager $pluginManager): Response
	{
		$user = User::loggedInUser();
		$owner = $user->getAccountOwner();

		if ($request->getRealMethod() === 'POST') {

			try {
				CsrfToken::get()->checkRequest($request);
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('app_user_user_index');
			}

			//Daten rausholen
			$submit = $request->request->all();

			if(!empty($submit['groupright']))
			{
				foreach ($submit['groupright'] as $groupid=>$rights)
				{
					$group =  SubGroups::find($groupid);
					if ($group) {
						$group->setGroupRights((array)$rights);
						$group->save();
					}
					else{
						$this->addFlash('danger', $translator->trans('Group rights couldn\'t be saved! Group id not valid in this context!'));
						return $this->redirectToRoute('app_sub_group_index');
					}
				}
			}
			else{
				$this->addFlash('info', $translator->trans('No rights are set!'));
				return $this->redirectToRoute('app_sub_group_index');
			}

			$this->addFlash('success', $translator->trans('Group rights were saved'));
			return $this->redirectToRoute('app_sub_group_index');
		}

		//Im Zweifel erstmal die SubAdmin Gruppe anlegen
		if (SubGroups::createSubAdminStarterGroup()) {
			return $this->redirectToRoute('app_sub_group_index');
		};

		$menu = MenuData::all();
		$groups = SubGroups::all();
		$plugins = array_filter(
			$pluginManager->configuredPlugins(),
			function ($plugin) use ($owner) {
				return $plugin->getState()->isAllowedForUser($owner) and !$plugin->isHidden();
			}
		);

		//print_r($menu);
		$pagination = new Pagination($menu, max(1, count($menu)));
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_sub_group_index');
		}

		//Ausgabe der Daten
		return $this->render('subgroups/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'grouplist' => $groups,
			'pagination' => $pagination,
			'menus' => $menu,
			'plugins' => $plugins,
			'usersearch' =>"",
		]);
	}




	/**
	 * @Route("/new_group", name="app_sub_group_new")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function createGroup(Request $request, TranslatorInterface $translator): Response
	{
		$mainUser = User::loggedInUser()->getAccountOwner();

		if ($request->getRealMethod() === 'POST') {

			try {
				CsrfToken::get()->checkRequest($request);
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('app_user_user_index');
			}

			//Daten rausholen
			$submit = $request->request->all();
			$data = [
				'groupname' => $submit['sub_groupname'] ?? '',
			];

			if (empty($data['groupname'])) {
				$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
			}
			else {

				//neuen Datensatz erstellen
				$newgroup = SubGroups::create();
				$newgroup->setArray($data);
				$newgroup->save();

				//yeah!!!
				$this->addFlash('success', $translator->trans('Group %groupname% has been created!', [
					'%groupname%' => $data['groupname'],
				]));

				//In Edit Maske springen...
				return $this->redirectToRoute('app_sub_group_edit', ['id' => $newgroup->getId()]);
			}
		}

		//Ausgabe der Daten
		return $this->render('subgroups/new_group.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'subgroup' => array(),
		]);
	}

	/**
	 * @Route("/edit_group", name="app_sub_group_edit")
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editGroup(Request $request, TranslatorInterface $translator): Response
	{
		$mainUser = User::loggedInUser()->getAccountOwner();

		if ($request->getRealMethod() === 'POST') {

			try {
				CsrfToken::get()->checkRequest($request);
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('app_user_user_index');
			}

			//Daten rausholen
			$submit = $request->request->all();
			$data = [
				'groupname' => $submit['sub_groupname'] ?? '',
			];

			if (empty($data['groupname'])) {
				$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
			}
			else {

				$subgroup = SubGroups::find($submit['id']);

				//Keine Rechte an der Gruppe...
				if (!$subgroup) {
					$this->addFlash('danger', $translator->trans('No valid group id selected.'));

					//In Edit Maske springen...
					return $this->redirectToRoute('app_sub_group_edit', ['id' => $submit['id']]);
				}

				//Datensatz bearbeiten
				$subgroup->setArray($data);
				$subgroup->save();

				//yeah!!!
				$this->addFlash('success', $translator->trans('Group %groupname% has been edited!', [
					'%groupname%' => $data['groupname'],
				]));

				//In Edit Maske springen...
				return $this->redirectToRoute('app_sub_group_edit', ['id' => $submit['id']]);
			}
		}

		if(empty($request->query->get("id")))
		{
			$this->addFlash('danger', $translator->trans('No valid group id selected.'));
			$subgroup = array();
		}
		else{
			$subgroup = SubGroups::find($request->query->get("id"));
		}

		//Ausgabe der Daten
		return $this->render('subgroups/edit_group.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'subgroup' => $subgroup,
		]);
	}

	/**
	 * @Route("/delete_group", name="app_sub_group_delete")
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function deleteGroup(Request $request, TranslatorInterface $translator): Response
	{
		$mainUser = User::loggedInUser()->getAccountOwner();

		if ($request->getRealMethod() === 'POST') {

			try {
				CsrfToken::get()->checkRequest($request);
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('app_sub_group_index');
			}
			$subgroup = SubGroups::find($request->request->get("id"));
			if ($subgroup) {
				SubGroups::delete($request->request->get("id"));
				$this->addFlash('success', $translator->trans('The group was deleted!'));
				return $this->redirectToRoute('app_sub_group_index');
			}
			else{
				$this->addFlash('danger', $translator->trans('No valid group id selected.'));
				return $this->redirectToRoute('app_sub_group_index');
			}
		}
		else{
			if(empty($request->query->get("id")))
			{
				$this->addFlash('danger', $translator->trans('No valid group id selected.'));
				$editGroup = array();
			}
			else{
				$editGroup = SubGroups::find($request->query->get("id"));
			}

			//Falls es ein anderer User ist - dann sofort raus...fünf scheidet völlig aus
			if (empty($editGroup))
			{
				$this->addFlash('danger', $translator->trans('No valid group id selected.'));
				return $this->redirectToRoute('app_sub_group_index');
			}

			//Ausgabe der Daten
			return $this->render('subgroups/delete_group.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'subgroup' => $editGroup,
			]);
		}


	}
}
