<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Event\AccountUpdateEvent;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\Domain;
use App\Model\SubGroups;
use App\Model\User;
use App\Model\CsrfToken;
use App\Utils;
use RuntimeException;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/manage_users", name="app_user_user")
 */
class ManageSubUser extends AbstractController
{
	/**
	 * @Menu("User management", group="Users and Groups", icon="fa-user", routeGroup="app_user_user_", order=70)
	 * @Route("/", name="_index", methods={"HEAD", "GET", "POST"})
	 * @param Request $request
	 * @return Response
	 */
	public function index(Request $request): Response
	{
		$mainUser = User::loggedInUser()->getAccountOwner();

		//Im Zweifel erstmal die SudAdmin Gruppe anlegen
		if (SubGroups::createSubAdminStarterGroup()) {
			return $this->redirectToRoute('app_user_user_index');
		};

		$groups = SubGroups::all();
		$domains = Domain::all();

		$allUSer = $mainUser->getSubusers();

		$usersearch = "";

		$pagination = new Pagination($allUSer);
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_user_user_index');
		}

		//Ausgabe der Daten
		return $this->render('subuser/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'userlist' => $allUSer,
			'pagination' => $pagination,
			'usersearch' => $usersearch,
			'groupList' => $groups,
			'domainList' => $domains,
			'timestamp' => time()
		]);
	}

	/**
	 * @param array $activeGroups
	 * @return string
	 */
	private function getUserGroups(array $activeGroups)
	{
		$groups = SubGroups::all();
		$userGroup = "";
		foreach ($activeGroups as $k=>$v)
		{
			if(!empty($groups[$k]))
			{
				$userGroup.= $groups[$k]->get("groupname")."<br />";
			}
		}
		return $userGroup;
	}

	/**
	 * @param array $activeDoms
	 * @return string
	 */
	private function getUserDoms(array $activeDoms)
	{
		$domains = Domain::all();
		$userDoms = "";
		foreach ($activeDoms as $k=>$v)
		{
			if(!empty($domains[$k]))
			{
				$userDoms.= $domains[$k]->get("name")."<br />";
			}
		}
		return $userDoms;
	}

	//

	/**
	 * @Route("/list.json", name="_json", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function listJson(Request $request,TranslatorInterface $translator): Response
	{
		$mainUser = User::loggedInUser()->getAccountOwner();

		//Datumsformat...
		$dtForm = $translator->trans('Y-m-d H:i:s');

		$list = array_map(function ($user) use ($dtForm) {
			return [
				'id' => $user->getId(),
				'userImageFilename' => $this->generateUrl('app_user_user_image', ['imageName'=>$user->getId()])."?".time(),
				'username' => $user->get("username"),
				'firstName' =>  $user->get("firstName")." ".$user->get("lastName"),
				'emailAddress' =>  $user->get("emailAddress"),
				'company' =>  $user->get("company"),
				'groupList' =>  $this->getUserGroups($user->getGroupActive()),
				'domainList' =>  $this->getUserDoms($user->getDomActive()),
				'createtimestamp' =>  date($dtForm,$user->get("createtimestamp")),
				'lastlogin' =>  date($dtForm,$user->get("lastlogin")),
				'active' =>  $user->get("active"),
				'urls' => [
					'edit' => $this->generateUrl('app_user_user_edit', ['id'=>$user->getId()]),
					'delete' => $this->generateUrl('app_user_user_delete', ['id'=>$user->getId()]),
				],
			];
		}, array_values($mainUser->getSubusers()));

		$response = new JsonResponse([
			'data' => $list
		]);
		// Mit JSON_PRETTY_PRINT
		$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );
		return $response;
	}


	/**
	 * @Route("/edit", name="_edit")
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editUser(Request $request, TranslatorInterface $translator): Response
	{
		$mainUser = User::loggedInUser()->getAccountOwner();
		$groups = SubGroups::all();

		if ($request->getRealMethod() === 'POST' && !empty($request->request->get("sub_username")) ) {
			$action = $request->request->getAlnum('action');
			$newUser = User::byId($request->request->get("id"));

			//Falls es ein anderer User ist - dann sofort raus...
			if (!$newUser || $newUser->getSubUserOf() != $mainUser->getId()) {
				$this->addFlash('danger', $translator->trans('No valid user id selected.'));
				return $this->redirectToRoute('app_user_user_index');
			}

			try {
				CsrfToken::get()->checkRequest($request);
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('app_user_user_index');
			}
			//Daten Array
			$data = array();
			$fields = array("sub_username",
				"sub_emailAddress",
				"sub_firstName",
				"sub_lastName",
				"sub_company",
				"sub_new_password",
				"sub_passwordChangeable",
				"sub_active",
				"dom_active",
				"group_active",
				"noPassword",
			);
			$submit = $request->request->all();

			//Daten übergeben
			foreach ($fields as $v)
			{
				$realName = str_ireplace("sub_new_","",$v);
				$realName = str_ireplace("sub_","",$realName);
				if(!empty($submit[$v]))
				{
					$data[$realName] = $submit[$v];
				}
				else{
					$data[$realName] = "";
				}
			}

			//Die ID des Users der bearbeitet wird...
			$data['id'] = $request->request->get("id");

			$existingUser = User::find($data['username']);
			if ($existingUser and $existingUser->getId() != $data['id']) {
				$this->addFlash('danger', $translator->trans('Username %username% already exists.', [
					'%username%' => $data['username'],
				]));
			}
			elseif (Utils::validateEmailAddress($data['emailAddress']) == false) {
				$this->addFlash('danger', $translator->trans('Invalid email address, check format.'));
			}
			elseif (empty($data['username'])) {
				$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
			}
			elseif ($data['password'] && $data['password'] != $submit['new_password_retyped']) {
				$this->addFlash('danger', $translator->trans('Passwords do not match – please try again.'));
			}
			else {
				$event = new AccountUpdateEvent($newUser, AccountUpdateEvent::CAUSE_ADMIN);
				$event->setAccountInfoChanged();

				$newUser->setUsername($data['username']);
				$newUser->setEmailAddress($data['emailAddress']);
				$newUser->setFirstName($data['firstName']);
				$newUser->setLastName($data['lastName']);
				$newUser->setCompany($data['company']);
				$newUser->setDomActive($data['dom_active']);
				$newUser->setGroupActive($data['group_active']);
				$newUser->setActive((bool)$data['active']);
				$newUser->setPasswordLocked(!(bool)$data['passwordChangeable']);

				if (!empty($data['password'])) {
					$newUser->setPassword($data['password']);
					$event->setPasswordChanged();
				}
				elseif (!empty($data['noPassword']) and $newUser->hasPassword()) {
					$newUser->unsetPassword();
				}

				/** @var UploadedFile $image */
				$image = $request->files->get('image');
				if ($image) {
					$this->handleUploadErrors($image, $translator);
					$logoFilename = $newUser->getId() . '.' . $image->getClientOriginalExtension();
					$newUser->setUserImage($image);
				}

				// Extra-Aktionen
				switch ($action) {
					case 'clear2FA':
						$newUser->clearAllTwoFactorAuthParameters();
						break;
					default:
				}

				Utils::dispatchEvent(AccountUpdateEvent::NAME_PRE_SAVE, $event);
				$newUser->save();
				Utils::dispatchEvent(AccountUpdateEvent::NAME_POST_SAVE, $event);

				if (empty($data['group_active'])) {
					$this->addFlash('danger', $translator->trans('Please add group rights – otherwise user will not be able to log in!'));
				}

				$this->addFlash('success', $translator->trans('Client %username% has been edited!', [
					'%username%' => $data['username'],
				]));

				switch ($action) {
					case 'saveClose':
						return $this->redirectToRoute('app_user_user_index', [], 303);
					case 'save':
					default:
						//In Edit Maske springen...
						return $this->redirectToRoute('app_user_user_edit', ['id' => $newUser->getId()], 303);
				}

			}

			//Ausgabe der Daten
			return $this->render('subuser/edit_user.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'subuser' => $data,
				'domainList' => Domain::all(),
				'groupList' => $groups,
				'timestamp' => time()
			]);
		}
		else{
			$editUser = User::byId($request->query->get('id', ''));

			//Falls es ein anderer User ist - dann sofort raus...fünf scheidet völlig aus
			if (!$editUser || $editUser->getSubUserOf() != $mainUser->getId()) {
				$this->addFlash('danger', $translator->trans('No valid user id selected.'));
				return $this->redirectToRoute('app_user_user_index');
			}

			//Ausgabe der Daten
			return $this->render('subuser/edit_user.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'subuser' => $editUser,
				'domainList' => Domain::all(),
				'groupList' => $groups,
				'timestamp' => time()
			]);
		}
	}

	/**
	 * @Route("/delete", name="_delete")
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function deleteUser(Request $request, TranslatorInterface $translator): Response
	{
		$mainUser = User::loggedInUser()->getAccountOwner();

		if ($request->getRealMethod() === 'POST') {

			$newUser = User::byId($request->request->get("id"));

			//Falls es ein anderer User ist - dann sofort raus...
			if (!$newUser || $newUser->getSubUserOf() != $mainUser->getId()) {
				$this->addFlash('danger', $translator->trans('No valid user id selected.'));
				return $this->redirectToRoute('app_user_user_index');
			}

			try {
				CsrfToken::get()->checkRequest($request);
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('app_user_user_index');
			}

			User::delete($newUser->getId());
			$this->addFlash('success', 'The User was deleted!');
			return $this->redirectToRoute('app_user_user_index');

		}
		else{
			$editUser = User::byId($request->query->get('id', ''));

			//Falls es ein anderer User ist - dann sofort raus...fünf scheidet völlig aus
			if(!$editUser || $editUser->getSubUserOf() != $mainUser->getId()) {
				$this->addFlash('danger', $translator->trans('No valid user id selected.'));
				return $this->redirectToRoute('app_user_user_index');
			}

			//Ausgabe der Daten
			return $this->render('subuser/delete_user.html.twig', [
				'csrfToken' => CsrfToken::get(),
				'subuser' => $editUser,
				'timestamp' => time()
			]);
		}


	}

	/**
	 * @Route("/new", name="_new")
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function createUser(Request $request, TranslatorInterface $translator): Response
	{
		$mainUser = User::loggedInUser()->getAccountOwner();
		//Den User auf aktiv setzen
		if($mainUser->getRole()==User::ROLE_ADMIN && $mainUser->isActive()==false)
		{
			$mainUser->setActive(true)
				->save();
		}

		if ($request->getRealMethod() === 'POST') {
			$action = $request->request->getAlnum('action');
			try {
				CsrfToken::get()->checkRequest($request);
			}
			catch (CsrfTokenException $e) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
				return $this->redirectToRoute('app_user_user_index');
			}
			//Daten Array
			$data = array();
			$fields = array("sub_username",
							"sub_emailAddress",
							"sub_firstName",
							"sub_lastName",
							"sub_company",
							"sub_new_password",
							"sub_passwordChangeable",
							"sub_active",
							"dom_active",
							"group_active",
							"noPassword",
				);
			$submit = $request->request->all();

			//Daten übergeben
			foreach ($fields as $v)
			{
				$realName = str_ireplace("sub_new_","",$v);
				$realName = str_ireplace("sub_","",$realName);
				if(!empty($submit[$v]))
				{
					$data[$realName] = $submit[$v];
				}
				else{
					$data[$realName] = "";
				}
			}

			User::beginWrite();

			if (User::exists($data['username'])) {
				$this->addFlash('danger', $translator->trans('Username %username% already exists.', [
					'%username%' => $data['username'],
				]));
			}
			elseif (Utils::validateEmailAddress($data['emailAddress']) == false) {
				$this->addFlash('danger', $translator->trans('Invalid email address, check format.'));
			}
			elseif (empty($data['username']) or (empty($data['password']) and empty($data['noPassword']))) {
				$this->addFlash('danger', $translator->trans('Please fill out all fields marked with an asterisk.'));
			}
			elseif (($data['password'] ?? '') != ($submit['new_password_retyped'] ?? '')) {
				$this->addFlash('danger', $translator->trans('Passwords do not match – please try again.'));
			}
			else {
				$newUser = User::create($data['username']);
				if (!empty($data['noPassword'])) {
					$newUser->unsetPassword();
				} else {
					$newUser->setPassword($data['password']);
				}
				$newUser->setEmailAddress($data['emailAddress']);
				$newUser->setFirstName($data['firstName']);
				$newUser->setLastName($data['lastName']);
				$newUser->setCompany($data['company']);
				$newUser->setMaxDomainCount(1);
				$newUser->setRole(User::ROLE_CLIENT);
				$newUser->setMaxWLCount(0);
				$newUser->setMaxCallCount(1000); //Damit man intern das Layout checken kann...
				$newUser->setAutoAddDomains(false);
				$newUser->setDomActive($data['dom_active']);
				$newUser->setGroupActive($data['group_active']);
				$newUser->setActive((bool) $data['active']);
				$newUser->setPasswordLocked(!(bool) $data['passwordChangeable']);
				$newUser->setSubUserOf($mainUser->getId());
				$newUser->setSettings($mainUser->getSettings());

				/** @var UploadedFile $image */
				$image = $request->files->get('image');
				if ($image) {
					$this->handleUploadErrors($image, $translator);
					$logoFilename = $newUser->getId() . '.' . $image->getClientOriginalExtension();
					$newUser->setUserImage($image);
				}

				$newUser->save();

				if (empty($data['group_active'])) {
					$this->addFlash('danger', $translator->trans('Please add group rights – otherwise user will not be able to log in!'));
				}

				$this->addFlash('success', $translator->trans('Client %username% has been created!', [
					'%username%' => $data['username'],
				]));

				switch ($action) {
					case 'saveClose':
						return $this->redirectToRoute('app_user_user_index', [], 303);
					case 'save':
					default:
						//In Edit Maske springen...
						return $this->redirectToRoute('app_user_user_edit', ['id' => $newUser->getId()], 303);
				}
			}

			// Exklusiven Lock freigeben
			User::reload();
		}

		//gruppen
		$groups = SubGroups::all();

		//Ausgabe der Daten
		return $this->render('subuser/new_user.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'subuser' => $data ?? [],
			'groupList' => $groups,
			'domainList' => Domain::all(),
			'timestamp' => time()
		]);
	}

	/**
	 * @Route("/image/{imageName}", name="_image", methods={"HEAD", "GET"})
	 * @param string $imageName
	 * @return Response
	 */
	public function userImage(Request $request, string $imageName=""): Response
	{
		$imageName = basename($imageName);
		$imageBasename = explode('.',$imageName);
		$user = User::byId(trim($imageBasename[0]));
		//wenn das quatisch ist - kein User - dann nix...
		$image = null;
		if ($user) {
			$image = $user->getUserImage();
		}
		if (!$image) {
			$image = new File(Utils::getBaseDir() . DIRECTORY_SEPARATOR ."public/img/schmetterfink.jpg");
		}

		if (!$image->isFile()) {
			return new Response('File not found.', 404);
		}
		elseif (!$image->isReadable()) {
			return new Response('Reading file not permitted.', 403);
		}

		$response = new BinaryFileResponse($image, 200, [], true, null, true, true);
		$response->headers->set('X-Content-Type-Options', '');

		// Caching
		if ($request->query->has('nocache')) {
			Utils::resetCacheControl($response);
			$response->setCache(['max_age'=>0, 'public'=>true]);
			$response->headers->addCacheControlDirective('no-cache');
			$response->headers->addCacheControlDirective('must-revalidate');
		} else {
			Utils::resetCacheControl($response);
			$response->setCache([
				'max_age'=>450,
				'public'=>true,
			]);
		}

		return $response;
	}

	/**
	 * @param UploadedFile $image
	 * @param TranslatorInterface $translator
	 * @return void
	 */
	private function handleUploadErrors(UploadedFile $image, TranslatorInterface $translator): void
	{
		if ($image->getError() != UPLOAD_ERR_OK) {
			$this->addFlash(
				'danger',
				$translator->trans('The image could not be uploaded! (Error code %errorcode% %errormessage%)', [
					'%errorcode%' => $image->getError(),
					'%errormessage%' => $image->getErrorMessage()
				])
			);
			throw new RuntimeException();
		}

		if (preg_match('/^image\/.*$/', $image->getMimeType()) == false) {
			$this->addFlash('danger', $translator->trans('The uploaded file is not an image!'));
			throw new RuntimeException();
		}
	}
}
