<?php

namespace App\Controller;

use App\Component\Curl;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\LocaleService;
use App\Component\LoggerInterface;
use App\Component\OnBoarding\ScanData;
use App\Component\OnBoarding\ScanService;
use App\Config;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Exception\FilesizeLimitExceededException;
use App\Exception\MimeTypeNotSupportedException;
use App\Model\Cookie as CookieModel;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Embedding as EmbeddingModel;
use App\Model\EmbeddingAsset;
use App\Model\LegalText;
use App\Model\LocaleBase;
use App\Model\Placeholder as PlaceholderModel;
use App\Model\Purpose;
use App\Model\SystemEmbedding;
use App\Model\Theme as ThemeModel;
use App\Model\User;
use App\Model\Index\DomainIndex;
use App\Utils;
use InvalidArgumentException;
use RuntimeException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Validator\Constraints\Length;
use Symfony\Component\Validator\Constraints\Url;
use Symfony\Component\Validator\ConstraintViolation;
use Symfony\Component\Validator\Validator\ValidatorInterface;
use Symfony\Contracts\Translation\TranslatorInterface;
use DateTime;

/**
 * @Route("/domains/{_domainId}/onboarding", name="app_onboarding")
 */
class OnBoardingEmbeddings extends DomainDependantController
{
	/**
	 * @Route("/", name="", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @return Response
	 */
	public function landingPage(Request $request): Response
	{
		//Die Domain Daten holen
		$domain = Domain::activeDomain();

		return $this->render('onboarding_embeddings/install.html.twig', [
			'actualDomain' => $domain->getName(),
			'switch' => $request->query->get('switch'),
			'enableFrontendLegal' => $domain->isFrontendLegalEnabled()
		]);
	}

	/**
	 * @Route("/", name="_post", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param ValidatorInterface $validator
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function landingPageAction(Request $request, TranslatorInterface $translator, ValidatorInterface $validator, LoggerInterface $logger): Response
	{
		//Die Domain Daten holen
		$domain = Domain::activeDomain();

		$action = (string)($request->request->getAlnum('action'));
		switch ($action) {
			case 'skip':
				//auf set setzen...
				$domain->setOnBoarding(true);
				$domain->flush();

				// Index neu generieren
				DomainIndex::rebuild();

				return $this->redirectToRoute('app_dashboard');
			case "start":
				//Die Domain Daten holen
				$domain = Domain::activeDomain();

				$domName = $request->request->get('domain');

				if (!$request->request->getBoolean('enableFrontendLegal')) {
					$this->addFlash('danger', $translator->trans('Please acknowledge the security advice.'));
					return $this->redirectToRoute('app_onboarding', ['switch' => $request->query->get('switch')]);
				}

				if (Utils::isBaseEdition()) {
					$constraints = [
						new Url(['relativeProtocol' => true]), // Allow relative protocol
						new Length(['min' => 12, 'allowEmptyString' => false]), // Min len
					];
					$violations = $validator->validate($domName, $constraints);

					if ($violations->count()) {
						foreach ($violations as $violation) {
							if ($violation instanceof ConstraintViolation) {
								// Könnte man später mal nutzen: $message = $violation->getMessage();
								$logger->logError('Dom Setting Error', "Domain: '" . $domName . "' -> " . $violation->getMessage());
								$this->addFlash('danger', $translator->trans('Please enter a valid URL or domain.'));
								return $this->redirectToRoute('app_onboarding', ['switch' => $request->query->get('switch')]);
							}
						}
					} else {
						$domain->setName($domName);
						$domain->save();
						$this->addFlash('success', $translator->trans('Domain name was successfully saved.'));
					}
				}

				return $this->redirectToRoute('app_onboarding_scan', ['switch' => $request->query->get('switch')], 303);
			default:
				$this->addFlash('danger', $translator->trans('Internal error. Please try again.'));
				return $this->redirectToRoute('app_onboarding', ['switch' => $request->query->get('switch')], 303);
		}
	}

	/**
	 * @Route("/scan", name="_scan", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @param ScanService $service
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function OnBoardingScan(Request $request, ScanService $service, LoggerInterface $logger): Response
	{
		$inMigration = $request->query->get('switch') == "embedding";
		// Die Domain-Daten holen
		$domain = Domain::activeDomain();
		$user = User::loggedInUser();

		$domainUrl = $domain->getUrl();
		$domainBasicAuth = $domain->getDomainHtPassWd();

		// Grundinformationen holen
		$purposes = Purpose::all();
		$locales = LocaleService::enabledLocales();

		// Daten holen
		try {
			$scanData = $service->scanDomain($domainUrl, $logger, $domainBasicAuth);
		} catch (RuntimeException $e) {
			$scanData = new ScanData([]);
		}

		//CB Daten holen wenn vorhanden
		$cbDaten = $scanData->getCbCookiesUndColor();
		$countPages = array();
		//Google Page Count holen für Crawler Settings
		$countPages['googlePageCount'] = $scanData->getCloudDataPageCount();

		//Es liegen CB Cookie vor die genutzt werden sollen, dann die manuelle Skript Erkennung von CB aktivieren
		if (!empty($cbDaten['cookies'])) {
			$domain
				->setDetectCloakedScripts(true)
				->flush();
		}

		//Wenn die Google Page Count Daten vorliegen, diese setzen
		if(!empty($countPages['googlePageCount']))
		{
			$domain
				->setCountPages($countPages)
				->setCMS($scanData->getCloudDataCMS())
				->flush();
		}

		//nur wenn auch mindestens Standard Tarif ausgewählt wurde
		if($user->getMaxDomainCount()>=2)
		{
			//Crawling Daten setzen mit Standard werten
			//Frequenz alle 8 Tage
			//Budget Anzahl der Seiten die gefunden wurden $countPages['googlePageCount'] - aber nur wenn die Anzahl kleiner ist als möglich
			$maxCrawlBudget = $countPages['googlePageCount'];
			if($maxCrawlBudget>(Domain::getRestCrawlBudget()+$domain->getDomainActualMaxCrawlBudget()))
			{
				$maxCrawlBudget = Domain::getRestCrawlBudget()+$domain->getDomainActualMaxCrawlBudget();
			}
			$domain
				->setDomainExternalCrawlerFrequency(8)
				->setDomainMaxCrawlBudget($maxCrawlBudget)
				->setActivatePermaScan(false)
				->flush();
		}


		//Oha - es gibt sogar sinnvolle Button / Textfarben, diese dann vorhalten damit Sie eingestellt werden können, wenn das Theme initialisiert wird
		if (!empty($cbDaten['cbColorButton']) && ($theme = ThemeModel::activeTheme()) && $theme->isDefaultTheme()) {

			if(!empty($cbDaten['cbColorButton']))
			{
				$cbDaten['cbColorButton'] = str_ireplace("##","#",$cbDaten['cbColorButton']);
				$theme->setDeclineButtonBackgroundColor((string)$cbDaten['cbColorButton']);
				$theme->setMoreInfoButtonBackgroundColor((string)$cbDaten['cbColorButton']);
				$theme->setOkButtonBackgroundColor((string)$cbDaten['cbColorButton']);
				$theme->setSettingsIconBackgroundColor((string)$cbDaten['cbColorButton']);
			}

			if(!empty($cbDaten['cbColorBorder']))
			{
				$cbDaten['cbColorBorder'] = str_ireplace("##","#",$cbDaten['cbColorBorder']);
				$theme->setDeclineButtonBorderColor((string)$cbDaten['cbColorBorder']);
				$theme->setMoreInfoButtonBorderColor((string)$cbDaten['cbColorBorder']);
				$theme->setOkButtonBorderColor((string)$cbDaten['cbColorBorder']);
			}

			if(!empty($cbDaten['cbColorText']))
			{
				$cbDaten['cbColorText'] = str_ireplace("##","#",$cbDaten['cbColorText']);
				$theme->setDeclineButtonTextColor((string)$cbDaten['cbColorText']);
				$theme->setMoreInfoButtonColor((string)$cbDaten['cbColorText']);
				$theme->setOkButtonColor((string)$cbDaten['cbColorText']);
				$theme->setSettingsIconForegroundColor((string)$cbDaten['cbColorText']);
			}



			//Fensterfarben setzen, wenn sie aus dem Scanner kommen
			if(!empty($cbDaten['cbColorBackWindow']) && !empty($cbDaten['cbColorBackWindowText']))
			{
				$cbDaten['cbColorBackWindow'] = str_ireplace("##","#",$cbDaten['cbColorBackWindow']);
				$cbDaten['cbColorBackWindowText'] = str_ireplace("##","#",$cbDaten['cbColorBackWindowText']);
				$theme->setWindowBackgroundColor((string)$cbDaten['cbColorBackWindow']);
				$theme->setWindowBorderColor((string)$cbDaten['cbColorBackWindow']);
				$theme->setInfotextColor((string)$cbDaten['cbColorBackWindowText']);
				$theme->setLinkInInfotextColor((string)$cbDaten['cbColorBackWindowText']);
			}

			if(!empty($cbDaten['bannerProvider']))
			{
				//Es ist ein UC Banner, dann noch extra CSS einfügen
				if((string)$cbDaten['bannerProvider']==="uc")
				{
					$customCSS = $theme->getCustomCSS().$this->getCustomUCCSS();
					$theme->setCustomCss($customCSS);

					//Position überprüfen
					if((string)$cbDaten['bannerType']==="BANNER")
					{
						$theme->setPosition("bottom");
						//Wenn Banner unten, optisch etwas anpassen
						$theme->setCustomCss($customCSS.$this->getCustomUCExtraCSS());
					}
				}
			}

			if(!empty($cbDaten['logoUrl']))
			{
				$curl = new Curl();
				$curl->setOpt(CURLOPT_CONNECTTIMEOUT, 10);
				$curl->setOpt(CURLOPT_TIMEOUT, 10);
				$curl->setOpt(CURLOPT_RETURNTRANSFER, 1);
				$curl->setOpt(CURLOPT_HEADER, 0);
				$curl->setOpt(CURLOPT_FOLLOWLOCATION, 1);
				$curl->get((string)$cbDaten['logoUrl']);
				$imageData = base64_encode($curl->getResponse());
				$curl->close();

				if(!$curl->error && !empty($imageData)) {

					$mimeTypes = [
						'image/jpeg',
						'image/png',
						'image/svg+xml',
					];
					$exception = null;
					$file = null;
					try {
						$file = Utils::decodeBase64Image($imageData, 524288, $mimeTypes);
					} catch (FilesizeLimitExceededException $exception) {
						$logger->logError('Logo Upload Error', "Domain: '" . $domain->getName() . "' -> " . 'Invalid logoImage, maximum image upload size is 512 KiB.');
					} catch (MimeTypeNotSupportedException $exception) {
						$logger->logError('Logo Upload Error', "Domain: '" . $domain->getName() . "' -> " . 'Invalid logoImage.');
					} catch (InvalidArgumentException $exception) {
						$logger->logError('Logo Upload Error', "Domain: '" . $domain->getName() . "' -> " . 'Invalid logoImage, expecting base64 encoded string');
					}
					if($exception === null && $file !== null)
					{
						//Logo setzen und aktivieren
						$theme->setLogo($file);
						$theme->setLogoDisplayed(true);

						//temp Daten löschen
						$tempFilename = $file->getPathname();
						unset($file);
						unlink($tempFilename);
					}
				}
			}

			if($cbDaten['showCategories'] === 1)
			{
				$theme->setShowPurposesInMainWindow(true);
			}

			$theme->save();
		}

		// upgrade path to embeddings... - die alten cookies mit übernehmen...
		if ($inMigration) {
			// alte Cookies auch rausholen und mit übernehmen...
			foreach (CookieModel::all() as $cookie) {
				if ($cookie->getHttpCookieName() !== "ccm_consent") {
					$scanData->addLegacyCookie($cookie);
				}
			}
		}

		//Setting falls Agency WL Settings für Embeddings vorliegen
		$includeCCM19ConsentData = true;
		$localCCM19ConsentData = array();

		//System Embedding holen wenn es eine WL Einstellung gibt
		if($sysEmbeddingId = Utils::getWhiteLabelEmbedding())
		{
			// Passende System Embeddings aus Datenbank holen
			$includeCCM19ConsentData = false;
			$localCCM19ConsentData[$sysEmbeddingId] = $this->getFromSystemEmbedding($sysEmbeddingId);
		}

		//Embeddings holen
		$embeddings = $service->fetchMatchingEmbeddings($scanData,$includeCCM19ConsentData);

		//Wenn z.B. aus CM Banner oder CB Banner Embeddings ohne Cookies vorliegen, dann versuchen diese zu finden.
		if(!empty($cbDaten['vendorsNoCookies']))
		{
			$embeddingsSpecial = $service->fetchMatchingEmbeddingsByNames($cbDaten['vendorsNoCookies']);
			if(!empty($embeddingsSpecial))
			{
				$embeddings = array_merge($embeddings, $embeddingsSpecial);
			}
		}

		// Embedding(s) für Cookies ohne Embedding in der DB ergänzen
		$service->supplementUnmatchedCookies($embeddings, $scanData, $inMigration);

		//Embeddings zusammenführen, falls es ein durch Agency Version erstelltes Embedding gibt
		$embeddings = array_merge($embeddings, $localCCM19ConsentData);

		//Embeddings nach Zweck sortieren
		$purposeEmbeddings = $service->splitEmbeddingsByPurpose($embeddings);
		$cookiesCount = count($scanData->getCookies()) + count($scanData->getLocalStorage()) + count($scanData->getSessionStorage());

		if (count($embeddings) == 1 and $cookiesCount == 0) {
			//Für die Domain auf geonboarded setzen
			$domain->setOnBoarding(true);
			$domain->flush();
		}


		return $this->render('onboarding_embeddings/install_scan.html.twig', [
			'hasScreenshot' => $scanData->hasScreenshot(),
			'screenshotUri' => $scanData->getScreenshotAsUri(),
			'cookiesCount' => $cookiesCount,
			'embeddingsCount' => count($embeddings),
			'purposeEmbeddings' => $purposeEmbeddings,
			'purposes' => $purposes,
			'locales' => $locales,
			'actualDomain' => $domain->getName(),
		]);
	}

	/**
	 * Erstellt aus einem System Embedding ein Array für die Ausgabe im Scanning UI
	 *
	 * @param string $templateId
	 * @return array|void
	 */
	public function getFromSystemEmbedding(string $templateId)
	{
		$template = SystemEmbedding::find($templateId);
		$result = array();

		//Asset Daten
		if ($template) {
			$assets = array_values(array_map(static function ($asset)  {
				$result = [
					'id' => $asset->getId(),
					'name' => $asset->getName(),
					'dynamic' => $asset->isDynamic(),
					'storageType' => $asset->getStorageType(),
				];
				foreach (LocaleService::all() as $locale) {
					$language = $locale->getLanguage();
					if (!$language) { continue; }
					$shortLocale = substr($locale->getId(),0,2);
					$result['translations'][$shortLocale]['lifetime'] = $asset->getTranslatedLifetime($language, false);
					$result['translations'][$shortLocale]['description'] = $asset->getTranslatedDescription($language, false);
				}
				return $result;
			}, $template->getAssets()));

			//Basis Daten des Embeddings
			$result = [
				'id' => $template->getId(),
				"name" => $template->getName(),
				"embedding_vendor" => $template->getEmbeddingVendor(),
				"purpose" => $template->getPurpose()->getId(),
				"code_snippet" => $template->getCodeSnippet(),
				"code_block" => implode("\n", $template->getCodeBlock()),
				"iframe_blocking_strings" => implode("\n", $template->getIframeBlockingStrings()),
				"ccm_special" => true,
			];

			foreach (LocaleService::all() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) { continue; }
				$shortLocale = substr($locale->getId(),0,2);
				$result['translations'][$shortLocale]['label'] = $template->getTranslatedName($language, false);
				$result['translations'][$shortLocale]['description'] = $template->getTranslatedDescription($language, false);
				$result['translations'][$shortLocale]['privacy_policy_url'] = $template->getTranslatedPrivacyPolicyUrl($language, false);
				$result['translations'][$shortLocale]['lifetime'] = $template->getTranslatedLifetime($language, false);
				$result['translations'][$shortLocale]['collected_data_info'] = $template->getTranslatedCollectedDataInfo($language, false);
				$result['translations'][$shortLocale]['purpose_of_data_collection'] = $template->getTranslatedPurposeOfDataCollection($language, false);
				$result['translations'][$shortLocale]['legal_basis'] = $template->getTranslatedLegalBasis($language, false);
				$result['translations'][$shortLocale]['place_of_processing'] = $template->getTranslatedPlaceOfProcessing($language, false);
				$result['translations'][$shortLocale]['blocked_content_title'] = $template->getTranslatedBlockedContentTitle($language, false);
				$result['translations'][$shortLocale]['blocked_content_text'] = $template->getTranslatedBlockedContentText($language, false);
				$result['translations'][$shortLocale]['blocked_content_button_text'] = $template->getTranslatedBlockedContentButtonText($language, false);
			}
			$result['assets'] = $assets;
		}

		return $result;
	}

	/**
	 * @Route("/scan", name="_scan_post", methods={"POST"})
	 * @param Request $request
	 * @return Response
	 */
	public function OnBoardingScanSave(Request $request): Response
	{
		//Daten speichern und zum nächsten Schritt
		$this->saveEmbeddingData($request);

		//Nächste Seite...
		return $this->redirectToRoute('app_onboarding_text');
	}

	/**
	 * @Route("/text", name="_text")
	 * @param Request $request
	 * @param ScanService $service
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function startOnBoardingText(Request $request, ScanService $service, TranslatorInterface $translator): Response
	{
		$domain = Domain::activeDomain();
		$domName = $domain->getName();
		$session = $request->getSession();
		$resData = $service->fetchImprintData($domName);
		$session->set('resDataPage', $resData);

		//Daten speichern und zum nächsten Schritt
		if (!empty($request->request->get("action"))) {
			//Daten speichern
			$this->saveTextData($request, $translator);

			//Nächste Seite...
			return $this->redirectToRoute('app_onboarding_settings');
		}

		$locales = LocaleService::enabledLocales();
		//$imprintL10n = (array)$request->request->get('imprint_l10n');

		return $this->render('onboarding_embeddings/install_text.html.twig', [
			'locales' => $locales,
			'legalTexts' => LegalText::all(),
			'csrfToken' => CsrfToken::get(),
			'resDataPage' => $resData,
		]);
	}

	/**
	 * @Route("/settings", name="_settings")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function startOnBoardingSettings(Request $request, TranslatorInterface $translator): Response
	{
		// Daten speichern und zum nächsten Schritt
		if (!empty($request->request->get("action"))) {
			//Daten speichern
			$this->saveSettingsData($request, $translator);

			// Nächste Seite...
			return $this->redirectToRoute('app_onboarding_finish');
		}

		$session = $request->getSession();

		// Platzhalter raussuchen
		$placeholders = PlaceholderModel::all();

		// Platzhalter ersetzen wenn möglich...
		if (array_key_exists("FIRMA", $placeholders)) {
			if (!empty($session->get('resDataPage', [])['name'])) {
				$placeholders['FIRMA']->setValue($session->get('resDataPage', [])['name']);
			}
		}
		if (array_key_exists("Firma_Datenschutz", $placeholders)) {
			if (!empty($session->get('resDataPage', [])['datenschutz_url'])) {
				$placeholders['Firma_Datenschutz']->setValue($session->get('resDataPage', [])['datenschutz_url']);
			}
		}
		if (array_key_exists("Firma_Ort", $placeholders)) {
			if (!empty($session->get('resDataPage', [])['stadt'])) {
				$placeholders['Firma_Ort']->setValue($session->get('resDataPage', [])['plz'] . " " . $session->get('resDataPage', [])['stadt']);
			}
		}

		PlaceholderModel::flush();

		return $this->render('onboarding_embeddings/install_settings.html.twig', [
			'placeholders' => $placeholders,
			'csrfToken' => CsrfToken::get(),
			'resDataPage' => $session->get('resDataPage', []),
		]);
	}


	/**
	 * @Route("/finish", name="_finish")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function startOnBoardingFinish(Request $request, TranslatorInterface $translator): Response
	{
		/** @var User $user */
		$user = User::loggedInUser()->getAccountOwner();
		/** @var Domain $domain */
		$domain = Domain::activeDomain();
		$config = Config::getInstance();

		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme() . '://';
		$baseUrl = $scheme . $request->getHttpHost();

		$activeWidgetsCount = Domain::frontendWidgetCountForAllDomains();
		$domainMaxCount = $user->getMaxDomainCount();

		$allowFrontendWidget = Utils::isExtendedEdition()
			? $domainMaxCount < 0 || $activeWidgetsCount < $domainMaxCount || $domain->isFrontendWidgetEnabled()
			: true;

		if ($allowFrontendWidget == false) {
			$this->addFlash('warning', $translator->trans('You cannot activate the frontend widget at the moment due to insufficient domains in your plan.'));
		} else {
			$domain
				->setFrontendWidgetEnabled(true)
				->setManagementStructurePreLegalConsent(true)
				->setManagementStructure('embedding')
				->setOnBoarding(true)
				->save();
		}

		$locales = LocaleService::enabledLocales();
		$localeNames = [];
		$embedCodeSnippets = [];
		foreach ($locales as $locale) {
			/** @var LocaleBase $locale */
			$localeName = $locale->getName();

			$embedScriptUrl = $baseUrl . $this->generateUrl(
					'app_external_main_javascript_file',
					[
						'apiKey' => $user->getApiKey(),
						'domain' => $domain->getId(),
						'lang' => $localeName,
					]
				);
			$embedCodeSnippets[$localeName] = '<script src="' . htmlspecialchars($embedScriptUrl) . '" referrerpolicy="origin"></script>';
			$language = $locale->getLanguage();
			$localeNames[$localeName] = ($language ? $language->getDisplayName($request->getLocale()) : $localeName);
		}

		$embedCodeSnippetAuto = '<script src="' . htmlspecialchars($baseUrl . $this->generateUrl(
					'app_external_main_javascript_file',
					[
						'apiKey' => $user->getApiKey(),
						'domain' => $domain->getId(),
					]
				)) . '" referrerpolicy="origin"></script>';

		// Index neu generieren
		DomainIndex::rebuild();

		return $this->render('onboarding_embeddings/install_finish.html.twig', [
			'embedCodeSnippets' => $embedCodeSnippets,
			'embedCodeSnippetAuto' => $embedCodeSnippetAuto,
			'localeNames' => $localeNames,
			'domain' => $domain,
		]);

	}

	/**
	 * @Route("/save", name="_save")
	 * @param Request $request
	 * @return Response
	 */
	public function startOnBoardingSave(Request $request): Response
	{
		$domain = Domain::activeDomain();

		//Only set to embedding if checked
		$saveEmbedding = $request->request->get("managementStructurePreLegalConsent");

		if ($saveEmbedding == "on") {
			//Für die Domain auf geonboarded setzen
			$domain->setOnBoarding(true);
			$domain->setManagementStructure("embedding");
			$domain->flush();
		}

		return $this->redirectToRoute('app_dashboard');

	}

	/**
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 */
	public function saveSettingsData(Request $request, TranslatorInterface $translator)
	{
		$placeholders = PlaceholderModel::all();

		// Löschen
		$toDelete = $request->request->all('delete_placeholder');
		foreach ($toDelete as $key) {
			PlaceholderModel::delete($key);
		}

		// Bestehende aktualisieren
		foreach ($request->request->all('set_placeholder') as $key => $value) {
			$key = urldecode($key);
			PlaceholderModel::create($key)->setValue($value);
		}

		// Neu hinzufügen
		$newItems = array_combine(
			$request->request->all('add_placeholder_name'),
			$request->request->all('add_placeholder_value')
		);
		foreach ($newItems as $key => $value) {
			if ($key !== '' and $key !== null) {
				PlaceholderModel::create($key)->setValue($value);
			}
		}
		PlaceholderModel::flush();
	}


	/**
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return \Symfony\Component\HttpFoundation\RedirectResponse
	 */
	public function saveTextData(Request $request, TranslatorInterface $translator)
	{
		/** @var LocaleBase[] $locales */
		$locales = LocaleService::enabledLocales();
		$imprintL10n = (array)$request->request->get('imprintUrl');
		$privacyPolicyL10n = (array)$request->request->get('privacyUrl');

		if ($request->getRealMethod() === 'POST') {
			foreach ($locales as $locale) {
				/** @var LegalText|null $legalText */
				$legalText = LegalText::byLocale($locale);

				if ($legalText === null) {
					$legalText = LegalText::create()->setLocale($locale);
				}

				$legalText
					->setImprintScheme('link')
					->setPrivacyScheme('link')
					->setImprintUrl($imprintL10n[$locale->getId()] ?? '')
					->setPrivacyUrl($privacyPolicyL10n[$locale->getId()] ?? '');
			}

			try {
				CsrfToken::get()->checkRequest($request);
				LegalText::flush();

				//$this->addFlash('success', $translator->trans('Localizations have been written successfully.'));
				return $this->redirectToRoute('app_onboarding_settings');
			} catch (ConfigNotWritableException $exception) {
				$this->addFlash('danger', $translator->trans('Localizations could not be written to disk.'));
			} catch (CsrfTokenException $exception) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			}
		}
	}

	/**
	 * @param Request $request
	 */
	public function saveEmbeddingData(Request $request): void
	{
		//Alles zurücksetzen
		EmbeddingModel::clear();
		EmbeddingAsset::clear();

		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleName = $defaultLocale->getName();

		if (strpos($defaultLocaleName, "_") !== false) {
			$df = explode("_", $defaultLocaleName);
			$defaultLocaleName = trim($df[0]);
		}

		// Festhalten, ob Einbindungen angelegt werden, die den Iframe-Blocker verwenden
		$forceIframeBlockerEnabled = false;

		//Daten aus dem Formular
		$embeddings = (array)$request->request->get('embeddings');
		//print_r($embeddings);exit();

		foreach ($embeddings as $embeddingKey => $embeddingData) {
			$embedding = EmbeddingModel::create();

			// Script-Blocking
			$codeBlock = explode("\n", trim((string)($embeddingData['code_block'] ?? '')));
			$codeBlock = array_filter($codeBlock, static function ($line) {
				return trim($line) !== '';
			});
			$codeBlock = array_values(array_map(static function ($line) {
				return trim($line);
			}, $codeBlock));

			// Iframes blockieren
			$iframeBlockingStrings = array_values(array_filter(
				array_map('trim', preg_split('~\r\n|\n|\r~', $embeddingData['iframe_blocking_strings'] ?? '')),
				static function (string $line) {
					return $line != '';
				}
			));

			$forceIframeBlockerEnabled = $forceIframeBlockerEnabled || count($iframeBlockingStrings) > 0;

			// Zweck
			$purpose = Purpose::find($embeddingData['purpose'] ?? '') ?? Purpose::find('15c61c3');
			assert($purpose !== null);

			// Allgemeines speichern
			$embedding
				->setName($embeddingData['name'] ?? '')
				->setActive((bool)($embeddingData['active'] ?? false))
				->setCodeSnippet($embeddingData['code_snippet'] ?? '')
				->setCodeBlock($codeBlock)
				->setIframeBlockingStrings($iframeBlockingStrings)
				->setPurpose($purpose)
				->setEmbeddingVendor($embeddingData['embedding_vendor'] ?? '')
				->setDescription($embeddingData['translations'][$defaultLocaleName]['description'] ?? '')
				->setPrivacyPolicyUrl($embeddingData['translations'][$defaultLocaleName]['privacy_policy_url'] ?? '')
				->setCollectedDataInfo($embeddingData['translations'][$defaultLocaleName]['collected_data_info'] ?? '')
				->setPurposeOfDataCollection($embeddingData['translations'][$defaultLocaleName]['purpose_of_data_collection'] ?? '')
				->setLegalBasis($embeddingData['translations'][$defaultLocaleName]['legal_basis'] ?? '')
				->setPlaceOfProcessing($embeddingData['translations'][$defaultLocaleName]['place_of_processing'] ?? '')
				->setAutoUpdated((bool)($embeddingData['autoUpdated'] ?? false))
				->setLegalBasisStructured((false));

			//Structured Legal Basis Daten ergänzen
			if(!empty($embeddingData['source_id']))
			{
				$dbEmbeddingData = EmbeddingDb::fetchEmbedding($embeddingData['source_id']);
				if(isset($dbEmbeddingData['structured_legal_basis'])) {
					$embedding
						->setLegalBasisStructured((true))
						->setStructuredLegalBasis((array)($dbEmbeddingData['structured_legal_basis']));
				}
			}


			// Zuordnung für Auto-Updates speichern
			if (isset($embeddingData['source'])) {
				$embedding->setSource($embeddingData['source'] ?? '');
				$embedding->setSourceId($embeddingData['source_id'] ?? '');
				if (isset($embeddingData['source_ts'])) {
					$embedding->setSourceTimestamp(new DateTime($embeddingData['source_ts']));
				}
			}

			// Übersetzungen speichern
			foreach (LocaleService::enabledLocales() as $locale) {
				$localeName = $locale->getName();
				if (!isset($embeddingData['translations'][$localeName]['description'])) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				$embedding
					->setTranslatedDescription($locale, $embeddingData['translations'][$localeName]['description'] ?? '')
					->setTranslatedPrivacyPolicyUrl($locale, $embeddingData['translations'][$localeName]['privacy_policy_url'] ?? '')
					->setTranslatedCollectedDataInfo($locale, $embeddingData['translations'][$localeName]['collected_data_info'] ?? '')
					->setTranslatedPurposeOfDataCollection($locale, $embeddingData['translations'][$localeName]['purpose_of_data_collection'] ?? '')
					->setTranslatedLegalBasis($locale, $embeddingData['translations'][$localeName]['legal_basis'] ?? '')
					->setTranslatedPlaceOfProcessing($locale, $embeddingData['translations'][$localeName]['place_of_processing'] ?? '');
			}

			if(!empty($embeddingData['assets']))
			{
				// Assets hinzufügen
				foreach ($embeddingData['assets'] as $assetKey => $assetData) {
					$assetName = (string)($assetData['name'] ?? '');

					if ($assetName == '') {
						continue;
					}

					$asset = EmbeddingAsset::create()
						->setName($assetName)
						->setDynamic((bool)($assetData['dynamic'] ?? false))
						->setStorageType((string)($assetData['storageType'] ?? ''))
						->setLifetime((string)($assetData['translations'][$defaultLocaleName]['lifetime'] ?? ''))
						->setDescription((string)($assetData['translations'][$defaultLocaleName]['description'] ?? ''));

					// Übersetzungen speichern
					foreach (LocaleService::enabledLocales() as $locale) {
						$localeName = $locale->getName();
						if (!isset($assetData['translations'][$localeName]['description'])) {
							continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
						}
						$asset
							->setTranslatedLifetime($locale, $assetData['translations'][$localeName]['lifetime'] ?? '')
							->setTranslatedDescription($locale, $assetData['translations'][$localeName]['description'] ?? '');
					}

					$embedding->addAsset($asset);
				}
				EmbeddingAsset::flush();
			}

			$embedding->save();
		}

		// Domain-Einstellungen aktualisieren
		$domain = Domain::activeDomain();

		if ($forceIframeBlockerEnabled) {
			$domain->setIframesBlocked(true);
		}

		$domain->save();
	}

	/**
	 * Extra CSS für UC Banner
	 * @return string
	 */
	private function getCustomUCCSS(): string
	{
		return "\n 
/** Custom CSS UC Banner Layout */
.ccm-root {
    font-weight: 400;
}

.ccm-root .ccm-widget--buttons .button, .ccm-root .ccm-widget--buttons button {
    text-transform: uppercase;
    font-weight: 600;
}
.ccm-widget--logo img {
    height: 50px;
}
.ccm-widget--logo {
    justify-content: left;
    padding-top: 20px;
}

";
	}

	/**
	 * Extra CSS für den Banner Bottom
	 * @return string
	 */
	private function getCustomUCExtraCSS(): string
	{
		return "\n
/** Custom CSS UC Banner Bottom Layout */
.ccm--p {
    margin-bottom: 3px !important;
}
.ccm-widget--buttons {
    margin-left: 40%;
}
.ccm-root .ccm-widget--buttons .button, .ccm-root .ccm-widget--buttons button {
    min-width: 30%;
    max-width: 600px;
}
.ccm-modal--body {
    padding-bottom: 5px;
}
.ccm-widget--text .ccm-inner-overflow {
    padding-bottom: 10px;
}

";
	}
}
