<?php

namespace App\Controller;

use App\Component\LoggerInterface;
use App\Event\DomainInitializedEvent;
use App\Model\Domain;
use App\Model\User;
use App\Model\Index\DomainIndex;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use Symfony\Component\Validator\Constraints\Length;
use Symfony\Component\Validator\Constraints\Url;
use Symfony\Component\Validator\ConstraintViolation;
use Symfony\Component\Validator\Validator\ValidatorInterface;

/**
 * @Route("/onboarding", name="app_onboarding_extended")
 */
class OnBoardingExtendedVersion extends AbstractController
{

	/**
	 * @Route("/start", name="", methods={"HEAD", "GET", "POST"})
     * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param ValidatorInterface $validator
	 * @param LoggerInterface $logger
     * @return Response
     */
    public function startExtendedOnBoarding(Request $request, TranslatorInterface $translator, ValidatorInterface $validator, LoggerInterface $logger): Response
    {
    	$user = User::loggedInUser();
		$session = $request->getSession();

		if ($session) {
			//print_r($request);
			// Locale-Wechsel bei POST mit locale
			$wantedLocale = $request->query->get('switchToLocale');
			if ($wantedLocale) {
				$session->set('locale', $wantedLocale);
				$request->setLocale($wantedLocale);
				return $this->redirectToRoute('app_onboarding_extended');
			}
		}

    	if (!empty($request->request->get("action"))) {

			if(!empty($request->request->get('domain')))
			{
				if(!stristr($request->request->get('domain'),"http"))
				{
					$domName = "https://".$request->request->get('domain');
				}
				else{
					$domName = (string)$request->request->get('domain');
				}

			}
			else{
				$domName = "";
			}

			//check urls...
			$constraints = [
				new Url(['relativeProtocol' => true]), // Allow relative protocol
				new Length(['min' => 12, 'allowEmptyString' => false]), // Min len
			];
			$violations = $validator->validate($domName, $constraints);

			if(empty($request->request->get('enableFrontendLegal')))
			{
				$logger->logError('Onboarding Error', "Haken Legal Aktivierung -> wurde nicht aktiviert... "." User: ".$user->getEmailAddress());
				$this->addFlash('danger', $translator->trans('Please acknowledge the security advice.'));
				return $this->render('onboarding_embeddings/index_extended.html.twig', [
					'actualDomain' =>$domName,
					'enableFrontendLegal' => $request->request->get('enableFrontendLegal'),
					'showLegal' => $session->get('secChecked', false),
				]);
			}

			if ($violations->count()) {
				foreach ($violations as $violation) {
					if ($violation instanceof ConstraintViolation) {
						// Könnte man später mal nutzen: $message = $violation->getMessage();
						$logger->logError('Onboarding Error', "Domain: '".$request->request->get('domain')."' -> ".$violation->getMessage()." User: ".$user->getEmailAddress());
						$this->addFlash('danger', $translator->trans('Please enter a valid URL or domain.'));
						return $this->render('onboarding_embeddings/index_extended.html.twig', [
							'actualDomain' => $domName,
							'enableFrontendLegal' => $request->request->get('enableFrontendLegal'),
							'showLegal' => $session->get('secChecked', false),
						]);
					}
				}
			} else {
				if ($this->saveDomain($domName, $translator, (string)$request->request->get("action"))) {
					$logger->logSuccess('Onboarding Ready', "Domain: '".$request->request->get('domain')."' -> Hat geklappt - Domain wurde eingetragen!"." User: ".$user->getEmailAddress());
					$session->set('form_switch', 'embedding');
					$session->set('form_domain', $domName);
					return $this->redirectToRoute('app_onboarding_scan', ["_domainId"=>Domain::activeDomain()->getId()], 303);
				};

			}
		}
		else
		{
			$logger->logInfo('Onboarding gestartet', "Von User: ".$user->getEmailAddress());
			return $this->render('onboarding_embeddings/index_extended.html.twig', [
				'actualDomain' => "",
				'showLegal' => $session->get('secChecked', false),
			]);
		}

		$logger->logInfo('Onboarding gestartet', "Von User: ".$user->getEmailAddress());

		return $this->render('onboarding_embeddings/index_extended.html.twig', [
			'actualDomain' => "",
			'showLegal' => $session->get('secChecked', false),
		]);
    }

	/**
	 * @Route("/stop", name="_manual")
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @return Response
	 */
	public function stopOnBoarding(Request $request, LoggerInterface $logger): Response
	{
		//Die Domain Daten holen
		$user = User::loggedInUser();

		//auf set setzen...
		$user->setOnBoarding(false);
		$user->flush();

		$logger->logWarning('Onboarding break', "Manuell abgebrochen von User: ".$user->getEmailAddress());

		return $this->redirectToRoute('app_domains_index');
	}


	/**
	 * @param string $newDomain
	 * @param TranslatorInterface $translator
	 * @param string $action
	 * @return bool
	 */
	private function saveDomain(string $newDomain, TranslatorInterface $translator, string $action): bool
	{
		/** @var User $user */
		$user = User::loggedInUser()->getAccountOwner();
		if ($user->isDomainLimitReached()) {
			$this->addFlash('danger', $translator->trans('You\'ve reached the domain limit. Please contact your provider to raise that limit.'));
			return false;
		}

		$domainInfo = parse_url(filter_var(
			preg_match('~^https?://~', $newDomain) ? $newDomain : 'https://'.$newDomain,
			FILTER_SANITIZE_URL
		));
		if (isset($domainInfo['host'])) {
			$newDomain = $domainInfo['host'];
			$urlPath = '/'.ltrim($domainInfo['path'] ?? '/', '/');
			$urlScheme = $domainInfo['scheme'] ?? 'https';
			$url = $urlScheme."://".$newDomain.$urlPath;
		}
		else {
			$urlPath = '/';
			$urlScheme = 'https';
			$url = $newDomain;
		}

		if ($newDomain) {
			$domain = Domain::createForUser($user);
			$domain
				->setName($newDomain)
				->setUrlPath($urlPath)
				->setUrlScheme($urlScheme)
				->setFrontendWidgetEnabled(false)
				->setWhitelabel(false)
				->setCountry(null)
				->setInlineScriptsBlocked(false)
				->setSameDomainScriptsBlocked(false)
				->setManagementStructurePreLegalConsent(true)
				->setManagementStructure(Domain::MANAGEMENT_STRUCTURE_EMBEDDING)
				->save();

			// Grant access to subuser
			if (User::loggedInUser()->isSubUser()) {
				User::loggedInUser()
					->grantDomainAccess($domain)
					->save();
			}

			//Userdatensatz aktualisieren, Anzahl der Domains UND die Domainliste
			$user
				->updateDomainInfo()
				->save();

			// Index neu generieren
			DomainIndex::rebuild();

			// Give plugins a chance to initialize their data for the new domain
			Utils::dispatchEvent(DomainInitializedEvent::NAME, new DomainInitializedEvent($domain));

			Utils::getSession()->set('form_domain', $url);
			Domain::select($domain->getId());
			return true;
		}
		else {
			return false;
		}
	}
}
