<?php

namespace App\Controller;

use App\Component\Curl;
use App\Component\Menu\Annotation\Menu;
use App\Exception\CsrfTokenException;
use App\Model\User;
use App\Model\Cookie;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\FoundCookie;
use App\Model\FoundScript;
use App\Model\FoundObject;
use App\Model\FoundStorage;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use DateTime;

/**
 *
 * @Route("/domains/{_domainId}/page_check", name="app_page_check")
 *
 */
class PageCheck extends DomainDependantController
{

	private $externCounter = 0;

	/**
	 * @Menu("Permanent Crawler", group="Cookies & Settings", icon="fa-eye", order=50)
	 * @Route("", name="")
	 *
	 * @param TranslatorInterface $translator
	 * @param Request $request
	 * @param string $ignore
	 * @return Response
	 */
	public function doPagesCheck(TranslatorInterface $translator, Request $request, string $ignore=""): Response
	{
		//Spezialfall page_check / Permanent Crawler inaktiv
		$domain = Domain::activeDomain();
		if($domain!==null && $domain->isActivatePermaScan()===false)
		{
			return $this->redirectToRoute('app_domains_index');
		}

		//Skripte, und andere Daten bereinigen durch Zusammenfassung von Ergebnissen
		$this->cleanupFoundScripts();
		$this->cleanupFoundObjects();
		$this->cleanupFoundStorage();

		//Skripte, Objekte, Cookies und Local Storage löschen und resetten wenn > 14 Tage nicht gesehen, dann vermutlich Artefakte
		$this->cleanupFoundScriptsAndReset();
		$this->cleanupFoundObjectsAndReset();
		$this->cleanupFoundStorageAndReset();
		$this->cleanupFoundCookiesAndReset();


		$showIgnoredEntries = $request->query->get('filter') == 'ignore';


		//$this->addRequestCookiesToFound($request);
		$uncontrolledCookies = $this->listUncontrolledCookies($request, $showIgnoredEntries);
		$uncontrolledStorage = $this->listUncontrolledStorage($request, $showIgnoredEntries,true);
		$uncontrolledScripts = $this->listUncontrolledScripts($request, $showIgnoredEntries);
		$uncontrolledContent = $this->listUncontrolledContent($request, $showIgnoredEntries);

		return $this->render('pagecheck/index.html.twig', [
			'user' => User::loggedInUser(),
			'uncontrolledCookies' => $this->sortedById($uncontrolledCookies),
			'uncontrolledStorage' => $this->sortedById($uncontrolledStorage),
			'uncontrolledScripts' => $this->sortedById($uncontrolledScripts),
			'uncontrolledObjects' => $this->sortedById($uncontrolledContent),
			'showIgnoredEntries' => $showIgnoredEntries,
			'dateTimeFormat' => $translator->trans('Y-m-d H:i:s'),
			'csrfToken' => CsrfToken::get()
		]);

	}

	/**
	 * @param Request $request
	 * @param bool $docurl
	 * @return array
	 */
	public function getCounter(Request $request,$docurl=true): array
	{
		$uncontrolledData = array();
		$uncontrolledData['cookiesCount'] 		= count($this->listUncontrolledCookies($request,false,$docurl));
		$uncontrolledData['LocalStoragesCount'] = count($this->listUncontrolledStorage($request,false,$docurl));
		$uncontrolledData['skriptCount'] 		= count($this->listUncontrolledScripts($request,false,$docurl));
		$uncontrolledData['contentCount'] 		= count($this->listUncontrolledContent($request,false,$docurl));
		return $uncontrolledData;
	}

	/**
	 * @param $iterable array|\Traversable
	 * @return array
	 */
	private function sortedById($iterable): array
	{
		if (is_array($iterable)) {
			$array = $iterable;
		}
		else {
			$array = [];
			foreach ($iterable as $key=>$value) {
				$array[$key] = $value;
			}
		}
		usort($array, function ($a,$b) {
			$aid = $a->getId();
			$bid = $b->getId();
			if ($aid === $bid) {
				return 0;
			}
			return ($aid < $bid) ? -1 : 1;
		});
		return $array;
	}

	/**
	 * Fasst Scripte, die bis auf den Query-Teil gleich sind, zusammen.
	 * Das bereinigt Einträge aus älteren Versionen des Page Reports
	 */
	private function cleanupFoundScripts(): void
	{
		$startts = time();
		$flush = false;
		$scripts = FoundScript::all();
		$foundNames = [];
		foreach ($scripts as $item) {
			$name = $item->getName();
			$cleanName = preg_replace('/\?[^?]*$/', '?', $name);

			if (isset($foundNames[$cleanName])) {
				$canonicalItem = FoundScript::find($foundNames[$cleanName]);
				$canonicalItem->setLastSeenOn(max($canonicalItem->getLastSeenOn(), $item->getLastSeenOn()));
				foreach ($item->getFoundOnPages() as $page) {
					$canonicalItem->addFoundOnPage($page);
				}
				FoundScript::delete($name, false);
				$flush = true;
			}
			else {
				$foundNames[$cleanName] = $name;
			}
			if ((time() - $startts) > 5) {
				break;
			}
		}
		if ($flush) {
			FoundScript::flush();
		}
	}

	/**
	 * Fasst externe URLs, die bis auf den Query-Teil gleich sind, zusammen
	 *
	 * Das bereinigt Einträge aus älteren Versionen des Page Reports
	 */
	private function cleanupFoundObjects(): void
	{
		$startts = time();
		$flush = false;
		$objects = FoundObject::all();
		$foundNames = [];
		foreach ($objects as $item) {
			$name = $item->getName();
			$cleanName = preg_replace('/\?[^?]*$/', '?', $name);

			if (isset($foundNames[$cleanName])) {
				$canonicalItem = FoundObject::find($foundNames[$cleanName]);
				$canonicalItem->setLastSeenOn(max($canonicalItem->getLastSeenOn(), $item->getLastSeenOn()));
				foreach ($item->getFoundOnPages() as $page) {
					$canonicalItem->addFoundOnPage($page);
				}
				FoundObject::delete($name, false);
				$flush = true;
			}
			else {
				$foundNames[$cleanName] = $name;
			}
			if ((time() - $startts) > 5) {
				break;
			}
		}
		if ($flush) {
			FoundObject::flush();
		}
	}

	/**
	 * Fasst LocalStorage-Elemente, die einen bekannten Präfix haben, zusammen
	 *
	 * Das bereinigt Einträge aus älteren Versionen des Page Reports
	 */
	private function cleanupFoundStorage(): void
	{
		$prefixes = [
			't3-tabs-DTM-',
			'optimizely_data$$',
			'st_shares_',
			'_piggy.',
			'at-lojson-cache-',
			't3-icon_',
			'_at.hist.',
		];
		$startts = time();
		$flush = false;
		$items = FoundStorage::all();
		$foundNames = [];
		foreach ($items as $item) {
			$name = $item->getName();
			$cleanName = null;
			foreach ($prefixes as $prefix) {
				if (strpos($name, $prefix) === 0) {
					$cleanName = $prefix.'…';
					break;
				}
			}
			if ($cleanName === null) {
				continue;
			}

			if (isset($foundNames[$cleanName])) {
				$canonicalItem = FoundStorage::find($foundNames[$cleanName]);
				$canonicalItem->setLastSeenOn(max($canonicalItem->getLastSeenOn(), $item->getLastSeenOn()));
				foreach ($item->getFoundOnPages() as $page) {
					$canonicalItem->addFoundOnPage($page);
				}
				FoundStorage::delete($name, false);
				$flush = true;
			}
			else {
				if ($cleanName === $name) {
					$canonicalItem = $item;
				} else {
					$canonicalItem = FoundStorage::create($cleanName)
						->setLastSeenOn($item->getLastSeenOn())
						->setFoundOnPages($item->getFoundOnPages())
						->setIsSessionStorage($item->getIsSessionStorage());
					FoundStorage::delete($name, false);
				}
				$foundNames[$cleanName] = $cleanName;
			}
			if ((time() - $startts) > 7) {
				break;
			}
		}
		if ($flush) {
			FoundStorage::flush();
		}
	}


	/**
	 * @Route("/all", methods={"DELETE"}, name="_delete_all")
	 *
	 * @return JsonResponse|RedirectResponse
	 */
	public function doDeleteAll(Request $request, TranslatorInterface $translator): Response
	{
		//Spezialfall page_check / Permanent Crawler inaktiv
		$domain = Domain::activeDomain();
		if($domain!==null && $domain->isActivatePermaScan()===false)
		{
			return $this->redirectToRoute('app_domains_index');
		}

		try {
			CsrfToken::get()->checkRequest($request);
			FoundCookie::clear();
			FoundScript::clear();
			FoundStorage::clear();
			FoundObject::clear();
		}
		catch (CsrfTokenException $e) {
			$error = $translator->trans('CSRF token is missing or invalid. Please try again.');
			if ($request->getRealMethod() === 'POST') {
				$this->addFlash('danger', $error);
				return $this->redirectToRoute('app_page_check', [], 303);
			}
			else {
				return $this->json(['success'=>false, 'message'=>$error]);
			}
		}
		catch (\Exception $e) {
			if ($request->getRealMethod() === 'POST') {
				$this->addFlash('danger', $e->getMessage());
				return $this->redirectToRoute('app_page_check', [], 303);
			}
			else {
				return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
			}
		}

		// Wenn über POST-Formular ausgelöst, umleiten, ansonsten JSON zurückgeben
		if ($request->getRealMethod() === 'POST') {
			$this->addFlash('success', $translator->trans('All page check entries were successfully cleared.'));
			return $this->redirectToRoute('app_page_check', [], 303);
		}
		else {
			return $this->json(['success'=>true]);
		}
	}

	/**
	 * @Route("/cookie_ignore/{cookie}", methods={"DELETE"}, name="_ignore_cookie", requirements={"cookie"=".+"})
	 *
	 * @param Request $request
	 * @param string $cookie
	 * @return JsonResponse
	 */
	public function doIgnoreCookie(Request $request, string $cookie): JsonResponse
	{
		$cookie = base64_decode(strtr($cookie, '-_', '+/'));
		//print_r($cookie);

		try {
			$object = FoundCookie::find($cookie);
			$object->setIgnore(true);
			$object->save();
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		return $this->json(['success'=>true,""],200);
	}

	/**
	 * @Route("/cookie/{cookie}", methods={"DELETE"}, name="_delete_cookie", requirements={"cookie"=".+"})
	 *
	 * @param Request $request
	 * @param string $cookie
	 * @return JsonResponse
	 */
	public function doDeleteCookie(Request $request, string $cookie): JsonResponse
	{
		$cookie = base64_decode(strtr($cookie, '-_', '+/'));

		try {
			FoundCookie::delete($cookie);
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		return $this->json(['success'=>true]);
	}

	/**
	 * @Route("/cookie/{cookie}", methods={"GET","HEAD"}, name="_get_cookie", requirements={"cookie"=".+"})
	 *
	 * @param Request $request
	 * @param string $cookie
	 * @return JsonResponse
	 */
	public function doGetCookie(Request $request, string $cookie): JsonResponse
	{
		$cookie = base64_decode(strtr($cookie, '-_', '+/'));

		try {
			$object = FoundCookie::find($cookie);
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		if ($object) {
			return $this->json(['name'=>$object->getName(), 'last_seen_on'=>$object->getLastSeenOn()->format('Y-m-d H:i:s')]);
		} else {
			return $this->json(null, 404);
		}
	}

	/**
	 * @Route("/script/{script}", methods={"DELETE"}, name="_delete_script", requirements={"script"=".+"})
	 *
	 * @param Request $request
	 * @param string $script
	 * @return JsonResponse
	 */
	public function doDeleteScript(Request $request, string $script): JsonResponse
	{
		$script = base64_decode(strtr($script, '-_', '+/'));

		try {
			FoundScript::delete($script);
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		return $this->json(['success'=>true]);
	}


	/**
	 * @Route("/script_ignore/{script}", methods={"DELETE"}, name="_ignore_script", requirements={"script"=".+"})
	 *
	 * @param Request $request
	 * @param string $script
	 * @return JsonResponse
	 */
	public function doIgnoreScript(Request $request, string $script): JsonResponse
	{
		$script = base64_decode(strtr($script, '-_', '+/'));

		try {
			$object = FoundScript::find($script);
			$object->setIgnore(true);
			$object->save();
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		return $this->json(['success'=>true]);
	}

	/**
	 * @Route("/script/{script}", methods={"GET","HEAD"}, name="_get_script", requirements={"script"=".+"})
	 *
	 * @param Request $request
	 * @param string $script
	 * @return JsonResponse
	 */
	public function doGetScript(Request $request, string $script): JsonResponse
	{
		$script = base64_decode(strtr($script, '-_', '+/'));

		try {
			$object = FoundScript::find($script);
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		if ($object) {
			return $this->json(['name'=>$object->getName(), 'last_seen_on'=>$object->getLastSeenOn()->format('Y-m-d H:i:s')]);
		} else {
			return $this->json(null, 404);
		}
	}

	/**
	 * @Route("/object_ignore/{object}", methods={"DELETE"}, name="_ignore_object", requirements={"object"=".+"})
	 *
	 * @param Request $request
	 * @param string $object
	 * @return JsonResponse
	 */
	public function doIgnoreObject(Request $request, string $object): JsonResponse
	{
			$object = base64_decode(strtr($object, '-_', '+/'));

		try {
			$object = FoundObject::find($object);
			$object->setIgnore(true);
			$object->save();
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		return $this->json(['success'=>true]);
	}

	/**
	 * @Route("/object/{object}", methods={"DELETE"}, name="_delete_object", requirements={"object"=".+"})
	 *
	 * @param Request $request
	 * @param string $object
	 * @return JsonResponse
	 */
	public function doDeleteObject(Request $request, string $object): JsonResponse
	{
		$object = base64_decode(strtr($object, '-_', '+/'));

		try {
			FoundObject::delete($object);
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		return $this->json(['success'=>true]);
	}

	/**
	 * @Route("/object/{object}", methods={"GET","HEAD"}, name="_get_object", requirements={"object"=".+"})
	 *
	 * @param Request $request
	 * @param string $object
	 * @return JsonResponse
	 */
	public function doGetObject(Request $request, string $object): JsonResponse
	{
		$object = base64_decode(strtr($object, '-_', '+/'));

		try {
			$theObject = FoundScript::find($object);
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		if ($theObject) {
			return $this->json(['name'=>$theObject->getName(), 'last_seen_on'=>$theObject->getLastSeenOn()->format('Y-m-d H:i:s')]);
		} else {
			return $this->json(null, 404);
		}
	}

	/**
	 * @Route("/storage_ignore/{storage}", name="_ignore_storage", requirements={"storage"=".+"})
	 *
	 * @param Request $request
	 * @param string $storage
	 * @return JsonResponse
	 */
	public function doIgnoreStorage(Request $request, string $storage): JsonResponse
	{
		$storage = base64_decode(strtr($storage, '-_', '+/'));

		try {
			$object = FoundStorage::find($storage);
			$object->setIgnore(true);
			$object->save();
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		return $this->json(['success'=>true]);
	}

	/**
	 * @Route("/storage/{storage}", methods={"DELETE"}, name="_delete_storage", requirements={"storage"=".+"})
	 *
	 * @param Request $request
	 * @param string $storage
	 * @return JsonResponse
	 */
	public function doDeleteStorage(Request $request, string $storage): JsonResponse
	{
		$storage = base64_decode(strtr($storage, '-_', '+/'));

		try {
			FoundStorage::delete($storage);
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		return $this->json(['success'=>true]);
	}

	/**
	 * @Route("/storage/{storage}", methods={"GET","HEAD"}, name="_get_storage", requirements={"storage"=".+"})
	 *
	 * @param Request $request
	 * @param string $storage
	 * @return JsonResponse
	 */
	public function doGetStorage(Request $request, string $storage): JsonResponse
	{
		$storage = base64_decode(strtr($storage, '-_', '+/'));

		try {
			$theObject = FoundStorage::find($storage);
		}
		catch (\Exception $e) {
			return $this->json(['success'=>false, 'message'=>$e->getMessage()], 500);
		}

		if ($theObject) {
			return $this->json(['name'=>$theObject->getName(), 'last_seen_on'=>$theObject->getLastSeenOn()->format('Y-m-d H:i:s')]);
		} else {
			return $this->json(null, 404);
		}
	}

	/**
	 * Fügt alle Cookies, die aktuell im Request sind, der Liste gefundener
	 * Cookies hinzu.
	 *
	 * @deprecated
	 * @param Request $request
	 * @return void
	 */
	private function addRequestCookiesToFound(Request $request): void
	{
		$now = new DateTime('now');
		$foundCookies = [];
		foreach ($request->cookies as $cookieName => $cookieContent) {
			FoundCookie::create($cookieName)->setLastSeenOn($now)->save();
		}
	}

	/**
	 * @param FoundCookie|FoundStorage|FoundScript|FoundObject $object
	 * @param string $type
	 * @return array
	 */
	private function checkExternScanner($object, string $type="cookies")
	{
		$this->externCounter++;

		//Damit es nicht zu timeouts kommt... mehr als 2 auf einmal nicht checken... das sind schon 6 AUfrufe a 10 Sekunden
		if($this->externCounter > 5)
		{
			return array();
		}

		//INI
		$name = $object->getId();
		$curl = new Curl();
		$i=0;
		$exists = 0;
		$found = array();
		$start = time();

		foreach ($object->getFoundOnPages() as $url)
		{
			//Sicherheitsfeature - Abbruche nach max. 5 Sekunden...
			if ($start+5 < time())
			{
				continue;
			}
			if($i>=10 || $exists > 0)
			{
				//continue;
			}
			$params = ["apikey" => "1234", "url" => $url];

			//Timeout setzen, damit es nicht bis unendlich läuft...
			$curl->setOpt(CURLOPT_TIMEOUT,2);

			//Daten holen
			//$curl->get('https://cookiecheck.ccm19.de/crawler.php', $params);
			$curl->get('https://www.ccm19.de/crawler/newcrawler/crawler.php', $params);

			if ($curl->curl_error) {
				continue;
			} elseif ($curl->getResponseHeaders('Content-Type') != 'application/json') {
				continue;
			} elseif ($curl->isServerError()) {
			}

			//Antwort vom Scanner Server...
			$response = ($curl->getResponse());
			$responseData = unserialize($response);

			if(($responseData!="null"))
			{
				$found[$url] = 0;
				foreach($responseData as $rpd)
				{
					if($type=="cookies")
					{
						foreach($rpd['cookies'] as $rpdLS)
						{
							if($rpdLS['name']==$name)
							{
								$found[$url] = 1;
							}
						}
					}

					if($type=="localStorage")
					{
						foreach($rpd['localStorage'] as $rpdLS)
						{
							if($rpdLS['name']==$name)
							{
								$found[$url] = 1;
							}
						}
					}

					if($type=="resources")
					{
						foreach($rpd['resources'] as $rpdLS)
						{
							if($rpdLS['url']==$name)
							{
								$found[$url] = 1;
							}
						}
					}
				}
			}
			else
			{
				$found[$url] = "nd";
			}
			$i++;
		}

		return $found;
	}


	/**
	 * @param Request $request
	 * @param bool $showIgnoredEntries
	 * @param bool $docurl
	 * @return FoundCookie[]
	 */
	private function listUncontrolledCookies(Request $request, bool $showIgnoredEntries=false,$docurl=true): array
	{

		// Uns bekannte Cookies sammeln
		$domain = Domain::activeDomain();
		assert($domain !== null);
		$knownCookies = [];
		// Embedding-Verwaltung
		if ($domain->getManagementStructure() == Domain::MANAGEMENT_STRUCTURE_EMBEDDING) {
			foreach (Embedding::all() as $embedding) {
				foreach ($embedding->getAssetsWithType('cookie') as $asset) {
					$knownCookies[$asset->getName()] = $asset;
				}
			}
		}
		// Legacy-Verwaltung
		else {
			// Uns bekannte Cookies sammeln
			foreach (Cookie::all() as $cookie) {
				$cookieName = $cookie->getHttpCookieName();
				$knownCookies[$cookieName] = $cookie;
			}
		}

		// Vorhandene Cookies sammeln
		$foundCookies = [];
		$allFoundCookies = FoundCookie::all();
		foreach ($allFoundCookies as $cookie) {
			// Liste entweder ignorierte oder aber die restlichen Einträge
			if ($cookie->getIgnore() != $showIgnoredEntries) {
				continue;
			}

			//Nur wenn 10 Einträge und Zeit weniger als 14 Tage her - dann isses vermutlich auch ein richtiges Cookie
			if (true or count($cookie->getFoundOnPages()) == 10 && $cookie->getLastSeenOn()->getTimestamp() > time() - 1209600) {
				//print_r($cookie);//exit();
				// && $cookie->get("scanned")==false
				if($showIgnoredEntries!=1)
				{
					$object = FoundCookie::find($cookie->getName());
					if (!$object) { continue; }
					$isFound = false;
					$retry = false;

					if($docurl)
					{
						$responseData = $this->checkExternScanner($object,"cookies");
						if(!empty($responseData))
						{
							foreach($responseData as $rpdat)
							{
								if($rpdat==1)
								{
									$isFound = true;
								}
								if($rpdat === "nd")
								{
									$retry = "retry";
								}
							}
						}
						else{
							$retry = "retry";
						}
						if($isFound==false)
						{
							$object->setIgnore(true);
						};
					}
					if($retry!="retry") {
						$object->set("scanned", true);
						$object->save();
					}
				}
				$foundCookies[$cookie->getName()] = $cookie;
			}
		}

		// Unsere eigene Session sollte nicht auftauchen
		$session = $request->getSession();
		if ($session) {
			unset($foundCookies[$session->getName()]);
		}
		// Unser eigenes Consent-Cookie sollte nicht auftauchen
		unset($foundCookies['ccm_consent']);

		// Unkontrollierte Cookies bestimmen
		$uncontrolledCookies = array_filter($foundCookies, function (string $key) use ($knownCookies) {
			foreach ($knownCookies as $cookie) {
				if ($cookie->matchName($key)) {
					return false;
				}
			}
			return true;
		}, ARRAY_FILTER_USE_KEY);

		return $uncontrolledCookies;
	}

	/**
	 * @param Request $request
	 * @param bool $showIgnoredEntries
	 * @param bool $docurl
	 * @return FoundStorage[]
	 */
	private function listUncontrolledStorage(Request $request, bool $showIgnoredEntries=false,$docurl=true): array
	{
		// Uns bekannte Storage-Keys sammeln
		$domain = Domain::activeDomain();
		assert($domain !== null);
		/** @var Cookie[]|EmbeddingAsset[] $knownLocalStorage */
		$knownLocalStorage = [];
		/** @var Cookie[]|EmbeddingAsset[] $knownSessionStorage */
		$knownSessionStorage = [];
		// Embedding-Verwaltung
		if ($domain->getManagementStructure() == Domain::MANAGEMENT_STRUCTURE_EMBEDDING) {
			foreach (Embedding::all() as $embedding) {
				foreach ($embedding->getAssetsWithType('localStorage') as $asset) {
					$knownLocalStorage[$asset->getName()] = $asset;
				}
				foreach ($embedding->getAssetsWithType('sessionStorage') as $asset) {
					$knownSessionStorage[$asset->getName()] = $asset;
				}
			}
		}
		// Legacy-Verwaltung
		else {
			// Uns bekannte Cookies sammeln
			foreach (Cookie::all() as $cookie) {
				$cookieName = $cookie->getHttpCookieName();
				$knownLocalStorage[$cookieName] = $cookie;
			}
			$knownSessionStorage = $knownLocalStorage;
		}

		// Vorhandene Storage-Keys sammeln
		$foundStorage = [];
		foreach (FoundStorage::all() as $storage) {
			// Liste entweder ignorierte oder aber die restlichen Einträge
			if ($storage->getIgnore() != $showIgnoredEntries) {
				continue;
			}

			//Nur wenn 10 Einträge und Zeit weniger als 14 Tage her - dann isses vermutlich auch ein richtiges Storage Element
			if (count($storage->getFoundOnPages()) == 10 && $storage->getLastSeenOn()->getTimestamp() > time() - 1209600) {
				if($showIgnoredEntries!=1  && $storage->get("scanned")==true)
				{
					$object = FoundStorage::find($storage->getName());
					if (!$object) { continue; }
					$isFound = false;
					$retry = false;

					if($docurl)
					{
						$responseData = $this->checkExternScanner($object,"localStorage");
						if(!empty($responseData))
						{
							foreach($responseData as $rpdat)
							{
								if($rpdat==1)
								{
									$isFound = true;
								}
								if($rpdat === "nd")
								{
									$retry = "retry";
								}
							}
						}
						else{
							$retry = "retry";
						}
						if($isFound==false)
						{
							$object->setIgnore(true);
						};
					}
					if($retry!="retry") {
						$object->set("scanned", true);
						$object->save();
					}
				}
				$foundStorage[$storage->getName()] = $storage;
			}
		}

		unset($foundStorage['ccm_consent']);

		// Unkontrollierte Storage bestimmen
		$uncontrolledStorage = array_filter($foundStorage, function (FoundStorage $value, string $key) use ($knownLocalStorage, $knownSessionStorage) {
			$knownStorage = ($value->getIsSessionStorage()) ? $knownSessionStorage : $knownLocalStorage;
			foreach ($knownStorage as $storage) {
				if ($storage->matchName($key)) {
					return false;
				}
			}
			return true;
		}, ARRAY_FILTER_USE_BOTH);

		return $uncontrolledStorage;
	}

	/**
	 * @param Request $request
	 * @param bool $showIgnoredEntries
	 * @param bool $docurl
	 * @return FoundScript[]
	 */
	private function listUncontrolledScripts(Request $request, bool $showIgnoredEntries=false,$docurl=true): array
	{
		// Uns bekannte Storage-Keys sammeln
		$domain = Domain::activeDomain();
		assert($domain !== null);

		// Vorhandene Storage-Keys sammeln
		$foundScripts = [];
		foreach (FoundScript::all() as $scripts) {
			// Liste entweder ignorierte oder aber die restlichen Einträge
			if ($scripts->getIgnore() != $showIgnoredEntries) {
				continue;
			}

			//Nur wenn 10 Einträge und Zeit weniger als 14 Tage her - dann isses vermutlich auch ein richtiges Storage Element
			if (count($scripts->getFoundOnPages()) == 10 && $scripts->getLastSeenOn()->getTimestamp() > time() - 1209600)
			{
				if($showIgnoredEntries!=1  && $scripts->get("scanned")==false)
				{
					$object = FoundScript::find($scripts->getName());
					if (!$object) { continue; }
					$isFound = false;
					$retry = false;

					if($docurl)
					{
						$responseData = $this->checkExternScanner($object,"resources");
						if(!empty($responseData))
						{
							foreach($responseData as $rpdat)
							{
								if($rpdat==1)
								{
									$isFound = true;
								}
								if($rpdat === "nd")
								{
									$retry = "retry";
								}
							}
						}
						else{
							$retry = "retry";
						}
						if($isFound==false)
						{
							$object->setIgnore(true);
						};
					}
					if($retry!="retry") {
						$object->set("scanned", true);
						$object->save();
					}
				}
				$foundScripts[$scripts->getName()] = $scripts;
			}
		}

		return $foundScripts;
	}

	/**
	 * @param Request $request
	 * @param bool $showIgnoredEntries
	 * @param bool $docurl
	 * @return FoundObject[]
	 */
	private function listUncontrolledContent(Request $request, bool $showIgnoredEntries=false,$docurl=true): array
	{
		// Uns bekannte Storage-Keys sammeln
		$domain = Domain::activeDomain();
		assert($domain !== null);

		// Vorhandene Storage-Keys sammeln
		$foundContent = array();
		foreach (FoundObject::all() as $content) {
			// Liste entweder ignorierte oder aber die restlichen Einträge
			if ($content->getIgnore() != $showIgnoredEntries) {
				continue;
			}

			//Nur wenn 10 Einträge und Zeit weniger als 14 Tage her - dann isses vermutlich auch ein richtiges Storage Element
			if (count($content->getFoundOnPages()) == 10 && $content->getLastSeenOn()->getTimestamp() > time() - 1209600) {
				if($showIgnoredEntries!=1  && $content->get("scanned")==false)
				{
					$object = FoundObject::find($content->getName());
					if (!$object) { continue; }
					$isFound = false;
					$retry = false;

					if($docurl)
					{
						$responseData = $this->checkExternScanner($object,"resources");
						if(!empty($responseData))
						{
							foreach($responseData as $rpdat)
							{
								if($rpdat==1)
								{
									$isFound = true;
								}
								if($rpdat === "nd")
								{
									$retry = "retry";
								}
							}
						}
						else{
							$retry = "retry";
						}
						if($isFound==false)
						{
							$object->setIgnore(true);
						};
					}
					if($retry!="retry") {
						$object->set("scanned", true);
						$object->save();
					}
				}
				$foundContent[$content->getName()] = $content;
			}
		}

		return $foundContent;
	}


	/**
	 * Gefundene Skripte bereinigen und zurücksetzen.
	 * Löscht gefundene Skripte, die in den letzten 14 Tagen nicht gesehen wurden, und leert den gefundenen Speicher.
	 *
	 * @return void
	 */
	private function cleanupFoundScriptsAndReset(): void
	{
		$items = FoundScript::all();
		foreach ($items as $item) {
			if(strtotime($item->getLastSeenOn()->format('Y-m-d H:i:s')) < strtotime("-14 days"))
			{
				FoundScript::delete($item->getName(), false);
			}
		}
		FoundScript::flush();
	}

	/**
	 * Gefundene Cookies bereinigen und zurücksetzen.
	 * Löscht gefundene Cookies, die in den letzten 14 Tagen nicht gesehen wurden, und leert den gefundenen Speicher.
	 *
	 * @return void
	 */
	private function cleanupFoundCookiesAndReset(): void
	{
		$items = FoundCookie::all();
		foreach ($items as $item) {
			if(strtotime($item->getLastSeenOn()->format('Y-m-d H:i:s')) < strtotime("-14 days"))
			{
				FoundCookie::delete($item->getName(), false);
			}
		}
		FoundCookie::flush();
	}

	/**
	 * Gefundene Storage Elemente bereinigen und zurücksetzen.
	 * Löscht gefundene Storage Elemente, die in den letzten 14 Tagen nicht gesehen wurden, und leert den gefundenen
	 * Speicher.
	 *
	 * @return void
	 */
	private function cleanupFoundStorageAndReset(): void
	{
		$items = FoundStorage::all();
		foreach ($items as $item) {
			if(strtotime($item->getLastSeenOn()->format('Y-m-d H:i:s')) < strtotime("-14 days"))
			{
				FoundStorage::delete($item->getName(), false);
			}
		}
		FoundStorage::flush();
	}

	/**
	 * Gefundene Objekt Elemente bereinigen und zurücksetzen.
	 * Löscht gefundene Objekt Elemente, die in den letzten 14 Tagen nicht gesehen wurden, und leert den gefundenen
	 * Speicher.
	 *
	 * @return void
	 */
	private function cleanupFoundObjectsAndReset(): void
	{
		$items = FoundObject::all();
		foreach ($items as $item) {
			if(strtotime($item->getLastSeenOn()->format('Y-m-d H:i:s')) < strtotime("-14 days"))
			{
				FoundObject::delete($item->getName(), false);
			}
		}
		FoundObject::flush();
	}
}
