<?php

namespace App\Controller;

use App\Utils;
use App\Model\PluginState;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\HttpFoundation\File\Exception\FileNotFoundException;

class PluginAssets extends UnrestrictedController
{
	const BASE64_EMPTY_GIF = 'R0lGODlhAQABAIAAAP///wAAACH5BAEAAAAALAAAAAABAAEAAAICRAEAOw==';

	/**
	 * @Route("/plugin_assets/{pluginuid<[^/]+/[^/]+>}/{path<.+>}", name="app_plugin_assets", stateless=true)
	 * @return Response
	 */
	public function asset(Request $request, string $pluginuid, string $path): Response
	{
		$plugin = PluginState::findByUid($pluginuid);

		if (!$plugin) {
			return $this->notFound($request);
		}

		// Pfad bereinigen
		$canonPath = $this->canonicalizePath($path);
		if ($path !== $canonPath) {
			$query = array_filter($request->query->all(), function ($key) { return substr($key, 0, 1) !== '_'; }, ARRAY_FILTER_USE_KEY);
			return $this->redirectToRoute('app_plugin_assets', array_merge($query, ['pluginuid'=>$pluginuid, 'path'=>$canonPath]), 301);
		}

		// Versteckte Dateien, PHP-Dateien und Path-Traversing-Versuche werden nicht ausgeliefert
		if (substr($path, 0, 1) === '.' or strpos($path, '/.') !== false or strpos($path, '.php') !== false) {
			return $this->notFound($request);
		}

		// Content-Type von Extension bestimmen
		$headers = [];
		$extIndex = strrpos($canonPath, '.');
		if ($extIndex !== false) {
			$extension = strtolower(substr($canonPath, strrpos($canonPath, '.')));
			switch ($extension) {
				case '.js':
					$headers['Content-Type'] = 'application/javascript';
					break;
				case '.json':
					$headers['Content-Type'] = 'application/json';
					break;
				case '.css':
					$headers['Content-Type'] = 'text/css';
					break;
				case '.gif':
					$headers['Content-Type'] = 'image/gif';
					break;
				case '.png':
					$headers['Content-Type'] = 'image/png';
					break;
				case '.jpg':
				case '.jpeg':
					$headers['Content-Type'] = 'image/jpeg';
					break;
				case '.webp':
					$headers['Content-Type'] = 'image/webp';
					break;
				default:
			}
		}

		$path = Utils::getBaseDir().'/plugins/'.$plugin->getDirname().'/assets/'.$canonPath;
		try {
			$response = new BinaryFileResponse($path, 200, $headers, true, null, true, true);
		}
		catch (FileNotFoundException $ignore) {
			$response = $this->notFound($request);
		}

		Utils::resetCacheControl($response);
		$response->setCache(['max_age'=>300, 'public'=>true]);
		return $response;
	}

	private function canonicalizePath(string $path): string
	{
		$path = str_replace('\\', '/', ltrim($path, '/\\'));
		$parts = array_filter(explode('/', $path), 'strlen');
		$absolutes = [];
		foreach ($parts as $part) {
			switch ($part) {
				case '.':
					break;
				case '..':
					array_pop($absolutes);
					break;
				default:
					$absolutes[] = $part;
			}
		}
		return implode('/', $absolutes);
	}

	private function notFound(Request $request): Response
	{
		$accept = $request->getAcceptableContentTypes();
		foreach ($accept as $item) {
			switch ($item) {
				case 'text/html':
				case 'text/*':
					return $this->render('bundles/TwigBundle/Exception/error404.html.twig', ['exception'=>['message'=>'File not found']], new Response('', 404));
				case 'text/plain':
					return new Response('Not found', 404, ['Content-Type'=>'text/plain']);
				case 'image/gif':
				case 'image/*':
					return new Response(base64_decode(self::BASE64_EMPTY_GIF), 404, ['Content-Type'=>'image/gif']);
				case 'text/javascript':
				case 'application/javascript':
					return new Response('console.error("not found");', 404, ['Content-Type'=>'application/javascript;charset=ascii']);
				case 'text/json':
				case 'application/json':
					return new Response('{"error":"not found"}', 404, ['Content-Type'=>'application/json;charset=ascii']);
				default:
			}
		}
		return $this->render('bundles/TwigBundle/Exception/error404.html.twig', ['exception'=>['message'=>'File not found']], new Response('', 404));
	}
}
