<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Exception\CsrfTokenException;
use App\Model\Cookie;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Component\LocaleService;
use App\Model\Protocol;
use App\Model\ProtocolArchive;
use App\Model\Purpose;
use App\Model\Statistic;
use App\Model\User;
use ArrayIterator;
use EmptyIterator;
use Exception;
use LimitIterator;
use RuntimeException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use DateTime;

class Protokoll extends DomainDependantController
{
	const ASSUMED_SESSION_COOKIE_LIFETIME = 14*24;
	const ASSUMED_UNKNOWN_COOKIE_LIFETIME = 365*24;

	const ARCHIVE_READ_LIMIT = 10000;

	/**
	 * @param iterable<array<string,mixed>> $protocolEntries
	 * @return int
	 */
	private function calculateMaxCookieLifetimeHours(iterable $protocolEntries): int
	{
		// Lebenszeit aller definierten Cookies bestimmen
		$cookieLifetimes = array_map(function ($cookie) {
			return $cookie->getLifetimeHours(self::ASSUMED_SESSION_COOKIE_LIFETIME, self::ASSUMED_UNKNOWN_COOKIE_LIFETIME);
		}, Cookie::all());

		$maxCookieLifetime = 0;

		// Lebenszeit der im Protokoll vorhandenen Cookies berücksichtigen
		foreach ($protocolEntries as $row) {
			try {
				$maxCookieLifetime = max($maxCookieLifetime, max(array_map(function ($id) use ($cookieLifetimes) { return $cookieLifetimes[$id] ?? self::ASSUMED_UNKNOWN_COOKIE_LIFETIME; }, $row['cookies'] ?? [])));
			} catch (\Throwable $e) {
				$maxCookieLifetime = self::ASSUMED_UNKNOWN_COOKIE_LIFETIME;
			}
		}

		return $maxCookieLifetime;
	}

	/**
	 * @param iterable<array<string,mixed>> $protocolEntries
	 * @return int
	 */
	private function calculateMaxCookieLifetimeHours_managementStructureEmbedding(iterable $protocolEntries): int
	{
		// Maximale Lebenszeit aller definierten Cookies pro Embedding bestimmen
		$cookieLifetimes = array_map('max', array_filter(
			array_map(
				function (Embedding $embedding) {
					return array_map(
						function (EmbeddingAsset $asset) {
							return $asset->getLifetimeHours(self::ASSUMED_SESSION_COOKIE_LIFETIME, self::ASSUMED_UNKNOWN_COOKIE_LIFETIME);
						},
						array_filter(
							$embedding->getAssets(),
							function (EmbeddingAsset $asset) {
								return $asset->getStorageType() == 'cookie';
							}
						)
					);
				},
				Embedding::all()
			)
		));

		$maxCookieLifetime = 0;

		// Lebenszeit der im Protokoll vorhandenen Cookies berücksichtigen
		foreach ($protocolEntries as $row) {
			// Verhindere leeres Array in max-Funktion
			if (count($row['embeddings'] ?? []) == 0) {
				continue;
			}

			try {
				$maxCookieLifetime = max($maxCookieLifetime, max(array_map(function ($id) use ($cookieLifetimes) {
					return $cookieLifetimes[$id] ?? self::ASSUMED_UNKNOWN_COOKIE_LIFETIME;
				}, $row['embeddings'] ?? [])));
			} catch (\Throwable $e) {
				$maxCookieLifetime = self::ASSUMED_UNKNOWN_COOKIE_LIFETIME;
			}
		}

		return $maxCookieLifetime;
	}

	/**
	 * @Menu("Protocol", group="Journals", icon="fa-edit", order=120)
	 * @Route("/domains/{_domainId}/protokoll", methods={"HEAD", "GET"}, name="app_protokoll")
	 * @Route("/domains/{_domainId}/protokoll/export_{format<json|csv>}", methods={"HEAD", "GET"}, name="app_protokoll_export")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param string $format
	 * @return Response
	 * @throws Exception
	 */
	public function protocol(Request $request, TranslatorInterface $translator, string $format = 'html'): Response
	{
		$user = User::loggedInUser();
		$domain = Domain::activeDomain();
		assert($domain !== null);
		$protocolEntries = [];

		// Je nach Einstellung entweder auf Embedding-Implementierung wechseln oder mit Cookies fortfahren
		if ($domain->getManagementStructure() === Domain::MANAGEMENT_STRUCTURE_EMBEDDING) {
			return $this->protocol_managementStructureEmbedding($request, $translator, $format);
		}

		// Suche
		$usersearch = trim((string)$request->query->get("search", ''));
		if ($usersearch !== '') {
			$protocol = Protocol::byDomain($domain);
			$protocolEntries = array_reverse($protocol->filterByConsentId($usersearch));
		}

		try {
			if (empty($protocolEntries) && empty($usersearch))
			{
				$protocol = Protocol::byDomain($domain);
				// TODO Logdatei nicht komplett in den Speicher laden, ggf. SQLite oder ähnliches verwenden
				$protocolEntries = $protocol->all(true);
			}

		}
		catch (RuntimeException $exception) {
			$this->addFlash('danger', $translator->trans('Logfile %file% could not be read, check permissions.', [
				'%file%' => Protocol::staticPathname($domain),
			]));
		}

		//Export als JSON der aktuellen Daten...
		if ($format === 'json') {
			if (!is_array($protocolEntries)) {
				$protocolEntries = iterator_to_array($protocolEntries, false);
			}
			$response = new JsonResponse($protocolEntries);
			$response->headers->set('Content-Type', 'application/json');
			$response->headers->set('Content-Disposition','attachment; filename="jsonLog_'.time().'.json"');
			// Use the JSON_PRETTY_PRINT
			//$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );
			return $response;
		}
		elseif ($format === 'csv') {
			// TODO
		}

		$pagination = new Pagination($protocolEntries, 20);
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_protokoll');
		}

		$maxCookieLifetimeHours = $this->calculateMaxCookieLifetimeHours($protocolEntries);
		$earliestDeleteHours = $maxCookieLifetimeHours + 3*24;
		$earliestDeleteDate = new DateTime('now - '.(int)$earliestDeleteHours.' hours');

		$locale = LocaleService::anyLocale([
			// Versuche, Locale anhand aktiver Backend-Sprache zu ermitteln
			$request->getLocale() == 'de' ? 'de_DE' : $request->getLocale(),
			'en_US',
			'de_DE',
		]);

		return $this->render('protokoll/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'locale' => $locale,
			'purposes' => Purpose::all(),
			'earliestDeleteDate' => $earliestDeleteDate,
			'pagination' => $pagination,
			'usersearch' => $usersearch,
			'totalCount' => count($protocolEntries),
		]);
	}

	/**
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param string $format
	 * @return Response
	 * @throws Exception
	 */
	public function protocol_managementStructureEmbedding(Request $request, TranslatorInterface $translator, string $format = 'html'): Response
	{
		$user = User::loggedInUser();
		$domain = Domain::activeDomain();
		assert($domain !== null);
		$protocolEntries = [];
		try {
			$protocol = Protocol::byDomain($domain);
		}
		catch (RuntimeException $exception) {
			$this->addFlash('danger', $translator->trans('Logfile %file% could not be read, check permissions.', [
				'%file%' => Protocol::staticPathname($domain),
			]));
			$protocol = null;
		}

		// Suche
		$ucid = trim((string)$request->query->get("search", ''));
		$startDate = trim((string)$request->query->get("startDate", ''));
		$endDate = trim((string)$request->query->get("endDate", ''));
		$isSearch = ($startDate !== '');
		try {
			if ($startDate !== '') {
				$endDateTime = new DateTime($endDate);
				$endDateTime->modify('+1 day midnight');

				$protocolArchive = ProtocolArchive::byDomain($domain);
				$protocolEntries = $protocolArchive->filter(new DateTime($startDate), $endDateTime, $ucid, true, true, true);
			}
			else {
				$protocolEntries = null;
			}
		}
		catch (RuntimeException $exception) {
			$protocolEntries = [];
			$this->addFlash('danger', $translator->trans('Logfile %file% could not be read, check permissions.', [
				'%file%' => Protocol::staticPathname($domain),
			]));
		}

		// Keine Suche -> aktuelles Protokoll
		if ($protocolEntries === null and $protocol) {
			try {
				$protocolEntries = $protocol->allConsented(true);
			}
			catch (RuntimeException $exception) {
				$this->addFlash('danger', $translator->trans('Logfile %file% could not be read, check permissions.', [
					'%file%' => Protocol::staticPathname($domain),
				]));
				$protocolEntries = [];
			}
		}

		//Export als JSON der aktuellen Daten...
		if ($format === 'json') {
			// Nicht ideal, wenn es sehr viele Einträge sind…
			if (!is_array($protocolEntries) and $protocolEntries !== null) {
				$protocolEntries = iterator_to_array($protocolEntries, false);
			}
			$response = new JsonResponse($protocolEntries);
			$response->headers->set('Content-Type', 'application/json');
			$response->headers->set('Content-Disposition','attachment; filename="jsonLog_'.time().'.json"');
			// Use the JSON_PRETTY_PRINT
			//$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );
			return $response;
		}

		if ($format === 'csv') {
			if ($protocolEntries === null) {
				$protocolEntries = [];
			}
			elseif (!is_array($protocolEntries)) {
				$protocolEntries = iterator_to_array($protocolEntries, false);
			}

			$fp = fopen('php://temp', 'w');
			$keys= array_keys($protocolEntries[0]);
			fputcsv($fp, $keys,"\t");

			foreach ($protocolEntries as $fields) {
				$result = [];
				array_walk($fields, static function ($item, $key) use (&$result) {
					if (!$item && $item !== "0") {
						$result[$key] = "0";
					}
					elseif (is_array($item)){
						$result[$key] = implode(";", $item);
					}
					else {
						$result[$key] = $item;
					}
                });
				fputcsv($fp, $result, "\t");
			}


			rewind($fp);
			$response = new Response(stream_get_contents($fp));
			fclose($fp);

			$response->headers->set('Content-Type', 'text/csv');
			$response->headers->set('Content-Disposition', 'attachment; filename="csvLog_'.time().'.csv"');

			return $response;
		}

		// Paginieren
		if ($isSearch) {
			if (is_array($protocolEntries)) {
				$protocolEntries = new ArrayIterator($protocolEntries);
			} elseif ($protocolEntries === null) {
				$protocolEntries = new EmptyIterator;
			}
			$pagination = new Pagination(new LimitIterator($protocolEntries, 0, self::ARCHIVE_READ_LIMIT), 20);
		} else {
			if ($protocolEntries === null) {
				$protocolEntries = [];
			}
			$pagination = new Pagination($protocolEntries, 20);
		}
		$totalCount = $pagination->getTotalCount();
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_protokoll', [
				'search' => $request->query->get("search"),
				'startDate' => $request->query->get("startDate"),
				'endDate' => $request->query->get("endDate"),
			]);
		}

		$embeddings = Embedding::all();
		usort($embeddings, function (Embedding $a, Embedding $b) {
			return strnatcasecmp($a->getName(), $b->getName());
		});

		$locale = LocaleService::anyLocale([
			// Versuche, Locale anhand aktiver Backend-Sprache zu ermitteln
			$request->getLocale() == 'de' ? 'de_DE' : $request->getLocale(),
			'en_US',
			'de_DE',
		]);

		if (!$startDate) {
			$firstItem = $protocol ? $protocol->first() : null;
			if (!$firstItem) {
				$startDate = (new DateTime('first day of this month midnight'))->format('Y-m-d');
			} else {
				$startDate = date('Y-m-d', $firstItem['timestamp']);
			}
		}

		return $this->render('protokoll/index_using-embeddings.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'locale' => $locale,
			'purposes' => Purpose::all(),
			'embeddings' => $embeddings,
			'tcf' => [
				'purposes' => array_map(function (TcfPurpose $item) use ($locale) {
					return $item->getTranslation($locale->getName());
				}, TcfPurpose::all()),
				'specialFeatures' => array_map(function (TcfSpecialFeature $item) use ($locale) {
					return $item->getTranslation($locale->getName());
				}, TcfSpecialFeature::all()),
				'vendors' => TcfVendor::all(),
			],
			'pagination' => $pagination,
			'ucid' => $ucid,
			'startDate' => $startDate,
			'endDate' => $endDate ?: (new DateTime('today'))->format('Y-m-d'),
			'totalCount' => $totalCount,
			'isSearch' => $isSearch,
			'wasCutoff' => $isSearch and $totalCount >= self::ARCHIVE_READ_LIMIT,
		]);
	}



	/**
	 * @Route("/domains/{_domainId}/protokoll", methods={"POST"}, name="app_protokoll_action")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function protocolAction(Request $request, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$domain = Domain::activeDomain();
		$action = $request->request->get("action");
		assert($domain !== null);

		try {
			CsrfToken::get()->checkRequest($request);
			switch ($action) {
				case "archive":
					$clearStatistics = $request->request->getBoolean("clearStatistics");
					try {
						$protocol = Protocol::byDomain($domain);
						$protocol->rotateProtocol();
						$this->addFlash('success', $translator->trans('Protocol was successfully archived.'));
					}
					catch (RuntimeException $e) {
						$this->addFlash('danger', $translator->trans('Archiving the protocol failed, please check file system permissions.'));
						return $this->redirectToRoute('app_protokoll', [], 303);
					}

					try {
						if ($clearStatistics) {
							Statistic::clear();
							$this->addFlash('success', $translator->trans('Domain statistics were successfully cleared.'));
						}
					}
					catch (RuntimeException $e) {
						$this->addFlash('danger', $translator->trans('Clearing domain statistics failed, please check file system permissions.'));
						return $this->redirectToRoute('app_protokoll', [], 303);
					}
					break;

				default:
					break;
			}
		} catch (CsrfTokenException $exception) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}


		$response = $this->redirectToRoute('app_protokoll', [], 303);
		$response->headers->set('Clear-Site-Data', '"cache"');
		return $response;
	}

}
