<?php

namespace App\Controller;

use App\Component\Cache\Cache;
use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\User;
use App\Model\Script;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/domains/{_domainId}/settings/scripts", name="app_scripts_")
 */
class Scripts extends DomainDependantController
{
	/**
	 * @Menu("Developer Settings", group="System", icon="fa-code", order=280)
	 * @Route("", methods={"GET", "HEAD"}, name="list")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function listScripts(Request $request, TranslatorInterface $translator): Response
	{
			// Scripte holen
		$scripts = Script::all();

		return $this->render('scripts/list.html.twig', [
			'scripts' => $scripts,
		]);
	}

	/**
	 * @Route("", methods={"POST"}, name="list_post")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function updateScriptsStatus(Request $request, TranslatorInterface $translator, LoggerInterface $logger): Response
	{

		/** @var Domain $domain */
		$domain = Domain::activeDomain();
		Domain::beginWrite();
        $user = User::loggedInUser();

		if ($domain->isManagementStructurePreLegalConsent()) {
			$domain->setManagementStructure($request->request->get('managementStructure', Domain::MANAGEMENT_STRUCTURE_COOKIE));
		}

		$declineButtonEnforced = $request->request->getBoolean('declineButtonEnforced');
		if ($declineButtonEnforced != $domain->isDeclineButtonEnforced()) {
			if ($declineButtonEnforced) {
				$logger->logInfo('DeveloperSettings', sprintf('User "%s" reactivated Decline button enforcement on domain "%s" [%s]',
					$user->getUsername(),
					$domain->getName(),
					$domain->getId()
				));
			} else {
				$logger->logWarning('DeveloperSettings', sprintf('User "%s" deactivated Decline button enforcement on domain "%s" [%s]',
					$user->getUsername(),
					$domain->getName(),
					$domain->getId()
				));
			}
		}

		$widgetExclusionUrls = array_values(array_filter(
			array_map('trim', preg_split('~\r\n|\n|\r~', (string)$request->request->get('widgetExclusionUrls'))),
			function ($line) { return $line != ''; }
		));

		$domain
			->setManagementStructurePreLegalConsent($domain->isManagementStructurePreLegalConsent() || (bool)$request->request->get('managementStructurePreLegalConsent'))
			->setDeleteUnknownCookies((bool)$request->request->get('deleteUnkownCookies'))
			->setDeleteUnknownCookiesForceReload((bool)$request->request->get('deleteUnkownCookiesForceReload'))
			->setLoadTechnicallyNecessaryEmbeddingsWithoutConsent($request->request->getBoolean('loadTechnicallyNecessaryEmbeddingsWithoutConsent'))
			->setEvalInlineScripts($request->request->getBoolean('evalInlineScripts'))
			->setNewScriptsBlocked((bool)$request->request->get('blockScripts'))
			->setScriptsTrack((bool)$request->request->get('scriptsTrack'))
			->setSameDomainScriptsBlocked((bool)$request->request->get('blockSameDomainScripts'))
			->setInlineScriptsBlocked((bool)$request->request->get('blockInlineScripts'))
			->setActivatePermaScan((bool)$request->request->get('activatePermaScan'))
			->setDeclineButtonEnforced($declineButtonEnforced)
			->setForcedCssImportant($request->request->getBoolean('forcedCssImportant'))
			->setLcpOptimizationEnabled($request->request->getBoolean('lcpOptimizationEnabled'))
			->setOnlyActiveOnMatchingDomain($request->request->getBoolean('onlyActiveOnMatchingDomain'))
			->setAllowEmbeddingConsentViaJsApi($request->request->getBoolean('allowEmbeddingConsentViaJsApi'))
			->setTriggeringExtGtmEvents($request->request->getBoolean('triggeringExtGtmEvents'))
			->setDetectCloakedScripts($request->request->getBoolean('detectCloakedScripts'))
			->setPrioritizeLanguageHeaderOverLangAttribute($request->request->getBoolean('prioritizeLanguageHeaderOverLangAttribute'))
            ->setWidgetDelayInMilliseconds($request->request->getInt('widgetDelayInMilliseconds'))
			->setWidgetExclusionUrls($widgetExclusionUrls)
			->setConsentStorageNamespaceUrlParameter((string)$request->request->get('consentStorageNamespaceUrlParameter'))
            ->setWidgetCachingEnabled($request->request->getBoolean('widgetCachingEnabled'))
            ->setCodeMinification((bool)$request->request->get('codeMinification'))
            ->setPartOfAdAndTrackingFreeNetwork($request->request->getBoolean('partOfAdAndTrackingFreeNetwork'))
			->save();

		// Clear the domain cache
		Cache::clear($domain, true);

		//Now the Script Part...
		Script::beginWrite();
		$scripts = Script::all();

		//Die Daten aus dem Form
		$enableScripts = (array)$request->request->get('enableScript');

		//Einträge durchloopen
		foreach ($scripts as $id=>$script)
		{
			//key gibts - dann aktiv - ansonsten inaktiv
			$data = $script->setEnableScript(array_key_exists($id, $enableScripts));
		}
		// Und Änderungen speichern
		Script::flush();

		// Inline-Script-Ausnahmen speichern
		$markers = explode("\n", trim((string)$request->request->get('allowedScriptMarkers', '')));
		$markers = array_filter($markers, function ($line) {
			return trim($line) !== '';
		});
		$markers = array_values(array_map(function ($line) {
			return trim($line);
		}, $markers));
		$domain->setAllowedScriptMarkers($markers)->save();
		unset($markers);

		//Blocklisting Skripte
		$markers = explode("\n", trim((string)$request->request->get('deniedScriptMarkers', '')));
		$markers = array_filter($markers, function ($line) {
			return trim($line) !== '';
		});
		$markers = array_values(array_map(function ($line) {
			return trim($line);
		}, $markers));
		$domain->setDeniedScriptMarkers($markers)->save();

		$this->addFlash('success', $translator->trans('The settings have been saved.'));
		return $this->redirectToRoute('app_scripts_list');
	}

	/**
	 * @Route("/{id}/delete", methods={"GET", "HEAD"}, name="delete")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deleteScriptPage(string $id, Request $request, TranslatorInterface $translator): Response
	{

		$script = Script::find($id);

		if (!$script) {
			return $this->redirectToRoute('app_scripts_list');
		}

		return $this->render('scripts/delete.html.twig', [
			'user' => User::loggedInUser(),
			'script' => $script,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}/delete", methods={"POST"}, name="delete_post")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deleteScript(string $id, Request $request, TranslatorInterface $translator): Response
	{

		Script::beginWrite();
		$script = Script::find($id);

		if (!$script) {
			return $this->redirectToRoute('app_scripts_list');
		}

		$scriptHash = $script->getId();

		try {
			CsrfToken::get()->checkRequest($request);
			Script::delete($id);
			$this->addFlash('success', 'Das Skript mit dem Hash "' . $scriptHash . '" wurde erfolgreich gelöscht!');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}
		return $this->redirectToRoute('app_scripts_list');
	}


}
