<?php

namespace App\Controller\Settings;

use App\Component\Api\ClusterApi;
use App\Component\Cache\Cache as FrontendCache;
use App\Component\Menu\Annotation\Menu;
use App\Component\Update;
use App\Component\ThumbnailExtractor\ThumbnailExtractorService;
use App\Config;
use App\Controller\AdminController;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use FilesystemIterator;
use DirectoryIterator;

class Cache extends AdminController
{
	/**
	 * @Menu("Caching", group="System", icon="fa-database", order=230)
	 * @Route("/settings/cache", methods={"HEAD", "GET"}, name="app_settings_cache")
	 * @param Request $request
	 * @return Response
	 */
	public function cacheView(Request $request): Response
	{
		$config = Config::getInstance();

		$cacheDir = Utils::getCacheDir();

		$cacheTotal = $this->calculateDirSize($cacheDir);
		$frontendSize = $this->calculateDirSize($cacheDir.'/frontend');
		$tmplSize = $this->calculateDirSize($cacheDir.'/twig');
		$i18nSize = $this->calculateDirSize($cacheDir.'/translations');
		$tcfSize = $this->calculateDirSize(Utils::getVarDir().'/tcfv2');
		$thumbSize = $this->calculateDirSize(ThumbnailExtractorService::getCacheDir());
		$symfonySize = $cacheTotal - $tmplSize - $i18nSize - $frontendSize;
		$cacheTotal += $thumbSize + $tcfSize;

		$opCacheSize = $this->getOpcacheSize();
		$cacheTotal += $opCacheSize;

		return $this->render('settings/cache/index.html.twig', [
			'tmplSize' => $tmplSize,
			'frontendSize' => $frontendSize,
			'i18nSize' => $i18nSize,
			'opCacheSize' => $opCacheSize,
			'symfonySize' => $symfonySize,
			'thumbSize' => $thumbSize,
			'tcfSize' => $tcfSize,
			'cacheTotal' => $cacheTotal,
			'peers' => array_map(function ($url) {
					$array = parse_url($url) ?: [];
					$array['url'] = $url;
					$array['id'] = 'peer--'.preg_replace('/[^a-zA-Z0-9:._-]/', '-', str_replace('://', ':', $url));
					return $array;
				},
				ClusterApi::listPeers()
			),
			'hostname' => explode('.', gethostname() ?: 'localhost')[0],
		]);
	}

	private function getOpcacheSize(): int
	{
		if (!function_exists('opcache_get_status')) {
			return 0;
		}
		$baseDir = Utils::getBaseDir();
		$size = 0;

		$status = opcache_get_status();
		foreach (($status['scripts'] ?? []) as $script) {
			if (strpos(($script['full_path'] ?? ''), $baseDir) === 0) {
				$size += ($script['memory_consumption'] ?? 0);
			}
		}
		return $size;
	}

	private function calculateDirSize(string $path): int
	{
		if (file_exists($path)) {
			return $this->_calculateDirSizeRecurse(new FilesystemIterator($path));
		} else {
			return 0;
		}
	}

	private function _calculateDirSizeRecurse(DirectoryIterator $it): int
	{
		$size = 0;

		foreach ($it as $fileinfo) {
			if ($fileinfo->isLink()) {
				continue;
			}
			elseif ($fileinfo->isDir()) {
				$size += $this->_calculateDirSizeRecurse(new FilesystemIterator($fileinfo->getPathname()));
			}
			elseif ($fileinfo->isFile()) {
				$size += $fileinfo->getSize();
			}
		}

		return $size;
	}

	/**
	 * @Route("/settings/cache", methods={"POST"}, name="app_settings_cache_save")
	 * @param Request $request
	 * @param ?TranslatorInterface $translator
	 * @param ?ThumbnailExtractorService $thumbnails
	 * @return Response
	 */
	public function cacheSave(Request $request, ?TranslatorInterface $translator, ?ThumbnailExtractorService $thumbnails): Response
	{
		$clusterApi = ClusterApi::getInstance();

		$trans = ($translator) ? [$translator, 'trans'] : function ($x) { return $x; };

		// Translation-System warmlaufen lassen
		$trans(' ');

		$clear = $request->request->get('clear');

		if (ClusterApi::listPeers()) {
			$peers = array_unique($request->request->all('peers'));
		} else {
			$peers = ['localhost'];
		}

		// Abbrechen, wenn kein Host ausgewählt
		if (!$peers) {
			return $this->redirectToRoute('app_settings_cache', [], 303);
		}

		if ($clear == 'all') {
			foreach ($peers as $peer) {
				if ($peer == 'localhost') {
					$updater = new Update;
					$updater->deleteFullCacheOnExit();

					if (function_exists('opcache_reset')) {
						opcache_reset();
					}

					$thumbnails->clearCache();
					Utils::clearDirectory(Utils::getVarDir() . '/tcfv2');
				} else {
					$clusterApi->callRemoteApi($peer, 'app_cluster_api_cache_clear', [], 'DELETE');
				}
			}
			$this->addFlash('success', $trans('All CCM19 caches were cleared.'));
			$response = $this->redirectToRoute('app_settings_cache', [], 303);
			$response->headers->set('Clear-Site-Data', '"cache"');
			return $response;
		}

		if ($clear == 'opcache') {
			$ok = true;
			foreach ($peers as $peer) {
				if ($peer == 'localhost') {
					if (function_exists('opcache_reset')) {
						opcache_reset();
					} else {
						$ok = false;
					}
				} else {
					$clusterApi->callRemoteApi($peer, 'app_cluster_api_opcache_clear', [], 'DELETE');
				}
			}

			if ($ok) {
				$this->addFlash('success', $trans('Zend OPcache was reset.'));
			} else {
				$this->addFlash('danger', $trans('Zend OPcache could not be cleared.'));
			}
			return $this->redirectToRoute('app_settings_cache', [], 303);
		}

		if ($clear == 'thumbnails') {
			$ok = true;
			foreach ($peers as $peer) {
				if ($peer == 'localhost') {
					if ($thumbnails) {
						$thumbnails->clearCache();
					} else {
						$ok = false;
					}
				} else {
					$clusterApi->callRemoteApi($peer, 'app_cluster_api_thumbnail_cache_clear', [], 'DELETE');
				}
			}
			if ($ok) {
				$this->addFlash('success', $trans('The thumbnail caches of all users were cleared.'));
			} else {
				$this->addFlash('danger', $trans('An error occured while clearing the thumbnail caches. Please try to clear all caches.'));
			}
			return $this->redirectToRoute('app_settings_cache', [], 303);
		}

		if ($clear == 'frontend') {
			foreach ($peers as $peer) {
				if ($peer == 'localhost') {
					FrontendCache::clearAll();
				} else {
					$clusterApi->callRemoteApi($peer, 'app_cluster_api_frontend_cache_clear', [], 'DELETE');
				}
			}

			$this->addFlash('success', $trans('The frontend caches of all users were cleared.'));
			$response = $this->redirectToRoute('app_settings_cache', [], 303);
			$response->headers->set('Clear-Site-Data', '"cache"');
			return $response;
		}

		if ($clear == 'tcf') {
			foreach ($peers as $peer) {
				if ($peer == 'localhost') {
					Utils::clearDirectory(Utils::getVarDir() . '/tcfv2');
				} else {
					$clusterApi->callRemoteApi($peer, 'app_cluster_api_tcf_cache_clear', [], 'DELETE');
				}
			}

			$this->addFlash('success', $trans('The TCFv2 cache was cleared.'));
			$response = $this->redirectToRoute('app_settings_cache', [], 303);
			return $response;
		}

		return $this->redirectToRoute('app_settings_cache', [], 303);
	}
}
