<?php

namespace App\Controller\Settings;

use App\Component\Menu\Annotation\Menu;
use App\Component\ThumbnailExtractor\ThumbnailExtractorService;
use App\Controller\DomainDependantController;
use App\Exception\CsrfTokenException;
use App\Model\Domain;
use App\Model\User;
use App\Model\CsrfToken;
use App\Model\Index\DomainIndex;
use App\Model\Placeholder;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/domains/{_domainId}/settings/consentstorage")
 */
class ConsentStorage extends DomainDependantController
{
	/**
	 * @Menu("Consent Storage", group="System", icon="fa-hdd-o", order=210)
	 * @Route("", methods={"GET", "HEAD"}, name="app_settings_consentstorage")
	 * @param Request $request
	 * @return Response
	 */
	public function showSettings(Request $request): Response
	{
		/** @var User $user */
		$user = User::loggedInUser();
		$accountOwner = $user->getAccountOwner();
		/** @var Domain $domain */
		$domain = Domain::activeDomain();

		$exampleDomain = preg_replace('/^www[0-9]?\./', '', explode(' ', strtolower(trim($domain->getName())), 2)[0]) ?: 'example.com';

		$crossDomainSharingAllowed = (Utils::isExtendedEdition() or Utils::isMultiDomainEdition() or $_ENV['APP_ALLOW_CROSS_DOMAIN_SHARING'] ?? false);
		$availableDomains = self::getAvailableConsentSharingDomainSlots($user, $domain);

		$firmaPlaceholder = Placeholder::find("Firma") ?? Placeholder::find("FIRMA");
		$datenschutzPlaceholder = Placeholder::find("Firma_Datenschutz");
		$shareGroupNameFallback = $firmaPlaceholder ? $firmaPlaceholder->getValue() : $domain->getUrlHost();
		$shareGroupUrlFallback = $datenschutzPlaceholder ? $datenschutzPlaceholder->getValue() : '';

		return $this->render('settings/consentstorage/index.html.twig', [
			'domain' => $domain,
			'exampleDomain' => $exampleDomain,
			'csrfToken' => CsrfToken::get(),
			'crossDomainSharingAllowed' => $crossDomainSharingAllowed,
			'availableDomains' => $availableDomains,
			'shareGroupNameFallback' => $shareGroupNameFallback,
			'shareGroupUrlFallback' => $shareGroupUrlFallback,
		]);
	}

	/**
	 * @Route("", methods={"POST"}, name="app_settings_consentstorage_post")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function saveSettings(Request $request, TranslatorInterface $translator, ThumbnailExtractorService $thumbnails): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
			/** @var User $user */
			$user = User::loggedInUser();
			$accountOwner = $user->getAccountOwner();
			/** @var Domain $domain */
			$domain = Domain::activeDomain();
			Domain::beginWrite();

			$availableDomains = self::getAvailableConsentSharingDomainSlots($user, $domain);
			$shareDomainNames = $request->request->all('shareDomain');

			// If the list contains the active domain, increment the available domain slot count
			if (in_array($domain->getUrlHost(), $shareDomainNames)) {
				$availableDomains++;
			}

			$shareDomainNames = Utils::reduceDomainNamesAndIpAddresses($shareDomainNames, $availableDomains);

			// Trigger cache refresh for the whole account if account-wide consent sharing is active
			// and consent sharing domains are updated
			if ($accountOwner->isSendFullConsentAndRejectionAcrossDomains() && $shareDomainNames != $domain->getShareDomainNames()) {
				$request->attributes->set('clearCacheScope', 'user');
			}

			if ($availableDomains < 0) {
				$this->addFlash('warning', $translator->trans('Some domain names could not be saved as you don\'t have enough domains available in your plan.'));
			}

			$consentLifetimeForPartialConsent = $request->request->get('consentLifetimeForPartialConsent');
			$consentLifetimeForPartialConsent = is_numeric($consentLifetimeForPartialConsent)
				? max(0, (int)$consentLifetimeForPartialConsent)
				: null;

			$domain
				->setShareDomainNames($shareDomainNames)
				->setConsentStorage((string)$request->request->get('storageMedium', 'localStorage'))
				->setConsentCookieLifetime($request->request->getInt('cookieLifetime', 0))
				->setSecureConsentCookie($request->request->getBoolean('cookieSecure'))
				->setCrossDomainSharing($request->request->getBoolean('cookieLinkSharing'))
				->setConsentCookieReset($request->request->get('consentCookieReset') ?: '')
				->setConsentResetOnRelevantChange($request->request->getBoolean('consentResetOnRelevantChange'))
				->setConsentLifetime($request->request->getInt('consentLifetime', 0))
				->setConsentLifetimeForPartialConsent($consentLifetimeForPartialConsent)
				->setShareGroupName((string)$request->request->get('shareGroupName', ''))
				->setShareGroupUrl((string)$request->request->get('shareGroupUrl', ''))
				->save();

			$user->setActualDomainCount(Domain::countIncludingShareNames())->save();

			// Index neu generieren
			DomainIndex::rebuild();

			$this->addFlash('success', $translator->trans('Settings have been saved.'));

		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		return $this->redirectToRoute('app_settings_consentstorage', [], 303);
	}

	/**
	 * This method determines the number of available consent sharing domain slots.
	 * @param User $user
	 * @param Domain $domain
	 * @return int
	 */
	private static function getAvailableConsentSharingDomainSlots(User $user, Domain $domain): int
	{
		$owner = $user->getAccountOwner();
		$maxDomainCount = (Utils::isBaseEdition() and $_ENV['APP_ALLOW_CROSS_DOMAIN_SHARING'] ?? false) ? 2147483647 : ($owner->getMaxDomainCount() ?: 1);
		// Count the sharing domains that are not the same as the domain itself
		$distinctDomainCount = count(array_diff($domain->getShareDomainNames(), [$domain->getUrlHost()]));
		// Determine the number of free sharing domain slots
		$freeDomainSlots = $maxDomainCount - $owner->getActualDomainCount() + $distinctDomainCount;
		return ($maxDomainCount < 0) ? 2147483647 : max(0, $freeDomainSlots);
	}
}
