<?php

namespace App\Controller\Settings;

use App\Component\Menu\Annotation\Menu;
use App\Controller\AdminController;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\BackendLanguage;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class LanguageSettings extends AdminController
{
	/**
	 * @Menu("Language Settings", group="System", icon="fa-beer", order=240)
	 * @Route("/settings/app_language", name="app_language", methods={"HEAD", "GET"})
	 * @param Request $request
	 * @return Response
	 */
	public function index(Request $request): Response
	{
		//Sprachliste aktualisieren
		BackendLanguage::updateAvailableLanguages();

		//Ausgabe der Daten
		return $this->render('settings/language/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'allLangs' => BackendLanguage::all(),
		]);
	}

	/**
	 * @Route("/settings/app_language", name="app_language_clear", methods={"DELETE"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function resetTranslations(Request $request, TranslatorInterface $translator): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);

			BackendLanguage::clear(true);
			//Sprachliste aktualisieren
			BackendLanguage::updateAvailableLanguages();

		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		return $this->redirectToRoute('app_language', [], 303);
	}

	/**
	 * @Route("/settings/app_language", name="app_language_save", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function saveNow(Request $request, TranslatorInterface $translator): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('app_language');
		}

		$langId = $request->request->get("setDefaultLang");
		if ($langId) {
			$newLanguage = BackendLanguage::byLongCode($langId);
			if ($newLanguage) {
				// Alten Standard löschen
				$oldDefault = BackendLanguage::findDefault();
				if ($oldDefault) {
					$oldDefault->setDefault(false);
				}

				//Locale Daten holen
				$newLanguage->setDefault(true);

				BackendLanguage::flush();

				$this->addFlash('success', $translator->trans('New default backend language successfully set.'));
			}
			return $this->redirectToRoute('app_language');
		}

		$setLangs = $request->request->all("active");
		$editedLangs = BackendLanguage::all();

		foreach ($editedLangs as $langVal) {
			$langVal->setActive(false);
			if (!empty($setLangs[$langVal->getId()])) {
				$langVal->setActive(true);
			}
		}
		BackendLanguage::flush();

		$this->addFlash('success', $translator->trans('The language selection is now set.'));
		return $this->redirectToRoute('app_language', [], 303);
	}


}
