<?php

namespace App\Controller\Settings;

use App\Config;
use App\Component\Menu\Annotation\Menu;
use App\Controller\AdminController;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class Network extends AdminController
{
	/**
	 * @Menu("Network", group="System", icon="fa-globe", order=240)
	 * @Route("/settings/network", methods={"HEAD", "GET"}, name="app_settings_network")
	 * @param Request $request
	 * @return Response
	 */
	public function networkSettings(Request $request): Response
	{
		$config = Config::getInstance();
		$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
		$currentHostUrl = $scheme.$request->getHttpHost();

		return $this->render('settings/network/index.html.twig', [
			'forceHttpsConnection' => $config->isForceHttpsConnection(),
			'baseUrl' => $config->getBaseUrl(),
			'currentUrl' => $currentHostUrl.$this->generateUrl('app_root'),
			'proxyUrl' => $_ENV['APP_PROXY'] ?? '',
			'maxSessionIdle' => (int)$_ENV['APP_MAX_SESSION_IDLE'] ?? 0,
			'exemptAddresses' => $config->getExemptAddresses(),
			'exemptUserAgents' => $config->getExemptUserAgents(),
			'reverseProxies' => explode(',', $_ENV['TRUSTED_PROXIES']) ?? [],
		]);
	}

	/**
	 * @Route("/settings/network", methods={"POST"}, name="app_settings_network_save")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function networkSettingsSave(Request $request, TranslatorInterface $translator): Response
	{
		$config = Config::getInstance();

		$trustedProxies = $request->request->get('reverseProxy') ?: null;
		if ($trustedProxies !== null) {
			$trustedProxies = implode(',', array_filter(array_map(function ($subnet) {
				$subnet = strtr(trim($subnet), [' '=>'', "\t"=>'',','=>'', '['=>'', ']'=>'', '"'=>'', "'"=>'', '$'=>'', '\\'=>'']);
				if (strtoupper($subnet) === 'REMOTE_ADDR') {
					return 'REMOTE_ADDR';
				}
				$parts = explode('/', $subnet, 2);
				$addr = filter_var($parts[0], FILTER_VALIDATE_IP);
				if ($addr === false) {
					return null;
				}
				if (isset($parts[1]) and !ctype_digit($parts[1])) {
					return null;
				}

				$maxCidr = (strpos($addr, ':') === false) ? 32 : 128;
				$cidr = (int)($parts[1] ?? $maxCidr);
				$cidr = max(0, min($cidr, $maxCidr));
				return $addr.'/'.$cidr;
			}, (array)$trustedProxies)));
		}

		// Call-Count-Ausschlüsse sammeln
		$exemptAddresses = array_unique(array_filter((array)$request->request->get('exemptAddress')));
		$exemptUserAgents = array_unique(array_filter((array)$request->request->get('exemptUserAgent')));

		$exemptAddresses = Utils::sanitizeIpAddresses($exemptAddresses);

		// Force-HTTPS etc. setzen
		$config
			->setForceHttpsConnection((bool)$request->request->get('forceHttpsConnection'))
			->setBaseUrl(trim((string)$request->request->get('baseUrl', '')))
			->setExemptAddresses($exemptAddresses)
			->setExemptUserAgents($exemptUserAgents)
			->flush();

		// Proxy und Session-Lifetime setzen
		Utils::setLocalEnvironment([
			'TRUSTED_PROXIES' => $trustedProxies,
			'APP_PROXY' => $request->request->get('proxyUrl'),
			'APP_MAX_SESSION_IDLE' => ((int)$request->request->get('maxSessionIdle')) ? ceil($request->request->get('maxSessionIdle')*60) : null,
		]);

		$this->addFlash('success', $translator->trans('Settings have been saved.'));

		return $this->redirectToRoute('app_settings_network', [], 303);
	}
}
