<?php

namespace App\Controller\Settings;

use App\Component\Api\ClusterApi;
use App\Component\Menu\Annotation\Menu;
use App\Controller\AdminController;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\PluginState;
use App\Model\User;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use App\Component\Plugin\Manager;
use App\Component\Plugin\PluginInfo;
use App\Component\Update;

/**
 * @Route("/settings/plugin_manager", name="app_plugins")
 */
class PluginManager extends AdminController
{
	/**
	 * @Menu("Plugin Manager", group="System", icon="fa-sitemap", order=250)
	 * @Route("", name="", methods={"HEAD","GET"})
	 * @param Request $request
	 * @param Manager $manager
	 * @param TranslatorInterface $translator
	 *
	 * @return Response
	 */
	public function index(Request $request, Manager $manager, TranslatorInterface $translator): Response
	{
		$search = $request->query->get("search") ?: '';

		$localPlugins = array_values(array_filter($manager->installedPlugins(), static function (PluginInfo $info) {
			return $info->isHidden() == false;
		}));

		$installedPlugins = array_values(array_filter($localPlugins, static function (PluginInfo $info) {
			return $info->isActive();
		}));
		$installablePlugins = array_values(array_filter($localPlugins, static function (PluginInfo $info) {
			return $info->isActive() == false;
		}));

		$availablePlugins = []; // TODO $manager->getAvailablePlugins($search);

		//Template ausgeben
		return $this->render('settings/plugin_manager/index.html.twig', [
			'installedPlugins' => $installedPlugins,
			'installablePlugins' => $installablePlugins,
			'availablePlugins' => $availablePlugins,
			'csrfToken' => CsrfToken::get(),
			'search' => $search,
		]);
	}

	/**
	 * @param array $users
	 * @param \App\Model\PluginState|null $pluginState
	 *
	 * @return int
	 */
	private function countUsersWithAllowedPlugin(array $users, ?PluginState $pluginState): int
	{
		if ($pluginState === null) {
			return 0;
		}
		return array_reduce($users, static function (int $carry, User $user) use ($pluginState) {
			return $carry + (int)$pluginState->isAllowedForUser($user);
		}, 0);
	}

	/**
	 * @Route("/{plugin<.+>}", name="_settings", methods={"HEAD","GET"})
	 * @param PluginInfo|null $plugin
	 * @param Manager $manager
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 *
	 * @return Response
	 */
	public function showSettings(?PluginInfo $plugin, Manager $manager, Request $request, TranslatorInterface $translator): Response
	{
		// Wenn es keine Einstellungen gibt, zurück.
		if (!$plugin or !$plugin->hasSettings()) {
			return $this->redirectToRoute('app_plugins');
		}
		if (!$plugin or !$plugin->isValidAndActive()) {
			$this->addFlash('danger', $translator->trans('Plugin not found or not yet activated.'));
			return $this->redirect('app_plugins');
		} else {
			$state = $plugin->getState();
			$state->setVersion($plugin->getVersion());
			$state->save();

			$clients = User::getClients();
			$pluginActiveClientCount = $this->countUsersWithAllowedPlugin($clients, $state);

			return $this->render('settings/plugin_manager/settings.html.twig', [
				'plugin' => $plugin,
				'clientCount' => count($clients),
				'pluginActiveClientCount' => $pluginActiveClientCount,
				'csrfToken' => CsrfToken::get(),
			]);
		}
	}

	/**
	 * @Route("/{plugin<.+>}", name="_settings_post", methods={"POST"})
	 * @param PluginInfo|null $plugin
	 * @param Manager $manager
	 * @param Update $updater
	 * @param TranslatorInterface $translator
	 * @param Request $request
	 *
	 * @return Response
	 */
	public function configurePlugin(?PluginInfo $plugin, Manager $manager, Update $updater, TranslatorInterface $translator, Request $request): Response
	{
		$action = $request->request->getAlnum('action', 'edit');
		$format = $request->request->getAlnum('format', 'html');
		if (!$plugin) {
			if ($format === 'json') {
				return $this->json(['success' => false, 'error' => "no such plugin"], 404);
			}

			return $this->redirectToRoute('app_plugins', [], 303);
		}

		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			if ($format === 'json') {
				return $this->json(['success' => false, 'error' => "CSRF token is missing or invalid"], 403);
			}

			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('app_plugins', [], 303);
		}

		switch ($action) {
			case "deactivate":
				if (($state = $plugin->getState())) {
					PluginState::delete($state->getId());
					PluginState::flush();
					$success = true;
					$updater->deleteFullCacheOnExit();
					$this->clusterApiBroadcast($plugin, null);
				} else {
					$success = false;
				}
				if ($format === 'json') {
					return $this->json(['success' => $success]);
				} else {
					if ($success) {
						$this->addFlash('success', 'The plugin was successfully deactivated.');
					}
					return $this->redirectToRoute('app_plugins', [], 303);
				}
			case "activate":
				if (!$plugin->getState()) {
					$state = PluginState::create();
					$state->setName($plugin->getLabel());
					$state->setType($plugin->getType());
					$state->setVersion($plugin->getVersion());
					$state->setDirname($plugin->getDirname());
					$state->setPluginUID($plugin->getName());
					$state->setIcon($plugin->getIcon());
					$state->setAllowTrial($plugin->isAllowTrial());
					if (!$plugin->isActivatablePerDomain()) {
						$state->setActivatedForAllAccounts(true);
					}
					$state->save();
					$success = true;

					$updater->deleteFullCacheOnExit();
					$this->clusterApiBroadcast($plugin, $state);
				} else {
					$success = false;
				}

				if ($format === 'json') {
					return $this->json(['success' => $success]);
				}

				if ($success) {
					$this->addFlash('success', 'The plugin was successfully activated.');
				}
				if ($plugin->hasSettings()) {
					return $this->redirectToRoute('app_plugins_settings', ['plugin' => $plugin->getName()], 303);
				}

				return $this->redirectToRoute('app_plugins', [], 303);
				break;
			case "activateForAllClients":
				$clients = User::getClients();
				if (($state = $plugin->getState())) {
					foreach ($clients as $client) {
						$settings = $client->getSettings();
						$settings[$state->getUserRightsKey()] = true;
						$client->setSettings($settings);
					}
					User::flush();
					$success = true;
				} else {
					$success = false;
				}
				if ($format === 'json') {
					return $this->json([
						'success' => $success,
						'clientCount' => count($clients),
						'pluginActiveClientCount' => $this->countUsersWithAllowedPlugin($clients, $plugin->getState()),
					]);
				}

				if ($success) {
					$this->addFlash('success', 'Plugin access rights were successfully granted to all clients.');
				}
				return $this->redirectToRoute('app_plugins_settings', ['plugin' => $plugin->getName()], 303);
			case "deactivateForAllClients":
				$clients = User::getClients();
				if (($state = $plugin->getState())) {
					foreach ($clients as $client) {
						$settings = $client->getSettings();
						$settings[$state->getUserRightsKey()] = false;
						$client->setSettings($settings);
					}
					User::flush();
					$success = true;
				} else {
					$success = false;
				}
				if ($format === 'json') {
					return $this->json([
						'success' => $success,
						'clientCount' => count($clients),
						'pluginActiveClientCount' => $this->countUsersWithAllowedPlugin($clients, $plugin->getState()),
					]);
				}

				if ($success) {
					$this->addFlash('success', 'Plugin access rights were successfully removed from all clients.');
				}
				return $this->redirectToRoute('app_plugins_settings', ['plugin' => $plugin->getName()], 303);
			case "edit":
			case "save":
				if (!$plugin->getState()) {
					return $this->redirectToRoute('app_plugins');
				}
				$pluginState = $plugin->getState();
				$fullAccess = $request->request->getBoolean("fullAccessForAllAccounts");
				$pluginState->setActivatedForAllAccounts((!$plugin->isActivatablePerDomain()) or $request->request->getBoolean("activateForAllAccounts"));
				$pluginState->setActivatedForAllAccountsOnTheirOwn($request->request->getBoolean("activateForAllAccountsOnTheirOwn"));
				$pluginState->setFullAccessForAllAccounts($fullAccess);

				$ok = true;

				if (!$fullAccess and $pluginState->isAllowTrial()) {
					//Call Counts Check
					if ($request->request->getInt("fullAccessCallCount") <= $request->request->getInt("testAccessCallCount")) {
						$this->addFlash('danger', 'Full Access Count hast to be higher than test count.');
						$ok = false;
					}

					//Dom Count Check
					if ($ok and $request->request->getInt("fullAccessDomCount") <= $request->request->getInt("testAccessDomCount")) {
						$this->addFlash('danger', 'Full Access Count hast to be higher than test count.');
						$ok = false;
					}
				}

				//Full AccessData
				$pluginState->setFullAccessCallCount($request->request->getInt("fullAccessCallCount"));
				$pluginState->setFullAccessDomCount($request->request->getInt("fullAccessDomCount"));

				//test AccessData
				$pluginState->setTestAccessCallCount($request->request->getInt("testAccessCallCount"));
				$pluginState->setTestAccessDomCount($request->request->getInt("testAccessDomCount"));

				//Testzeitraum
				$pluginState->setTestAccessTime($request->request->getInt("testAccessTime"));

				//Als gespeichert setzen
				$pluginState->setSavedAccess(true);

				//Es stimmt alles, also speichern
				if ($ok) {
					$pluginState->save();
					$this->clusterApiBroadcast($plugin, $pluginState);
					$this->addFlash('success', 'The plugin was successfully configured.');
				}
				//Upsi - ein Fehler, also Formular mit den Daten nochmal anbieten inkl. Fehlermeldung
				else {
					$clients = User::getClients();
					$pluginActiveClientCount = $this->countUsersWithAllowedPlugin($clients, $pluginState);

					return $this->render('settings/plugin_manager/settings.html.twig', [
						'plugin' => $plugin,
						'clientCount' => count($clients),
						'pluginActiveClientCount' => $pluginActiveClientCount,
						'csrfToken' => CsrfToken::get(),
					]);
				}
		}

		return $this->redirectToRoute('app_plugins_settings', ['plugin' => $plugin->getName()], 303);
	}

	/**
	 * @Route("", name="_install", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param Manager $manager
	 *
	 * @return Response
	 */
	public function installPlugin(Request $request, TranslatorInterface $translator, Manager $manager): Response
	{
		//Daten des Plugins holen anhand der ID
		$pluginUid = $request->request->get("pluginUID");
		$pluginData = $manager->getAvailablePlugins();
		$postPluginData = $pluginData[$pluginUid];

		//Liste installierter Plugins...
		$installed = $manager->installedPlugins();
		foreach ($installed as $k => $v) {
			if ($v->getName() == $pluginUid) {
				//Rücksprung und als installiert anzeigen
				$this->addFlash('danger', 'Das Plugin wurde schon installiert');
				return $this->redirectToRoute('app_plugins');
			}
		}

		//Zip Datei runterladen und kopieren und installieren im /plugins Verzeichnis unter dem Namen des Plugins
		$postPluginData['dirname'] = $manager->downloadAndInstallPlugin($pluginUid);

		//bei Base Version direkt aktivieren damit es sichtbar ist... bei extended brauchen wir aber mehr Kontrolle - daher unterschiedlich...
		/* TODO
		if (Utils::isBaseEdition()) {
			$postPluginData['activateForAllAccounts'] = "on";
		}
		*/

		//Rücksprung und als installiert anzeigen
		$this->addFlash('success', 'Das Plugin wurde installiert');
		return $this->redirectToRoute('app_plugins');
	}

	/**
	 * @Route("/{plugin<.+>}", name="_delete", methods={"DELETE"})
	 * @param \App\Component\Plugin\PluginInfo|null $plugin
	 * @param \Symfony\Component\HttpFoundation\Request $request
	 * @param \App\Component\Plugin\Manager $manager
	 * @param \Symfony\Contracts\Translation\TranslatorInterface $translator
	 * @param \App\Component\Update $updater
	 *
	 * @return Response
	 */
	public function deletePlugin(?PluginInfo $plugin, Request $request, Manager $manager, TranslatorInterface $translator, Update $updater): Response
	{
		if (!$plugin) {
			return $this->redirectToRoute('app_plugins', [], 303);
		}

		try {
			CsrfToken::get()->checkRequest($request);

			$manager->deletePlugin($plugin);
			$this->addFlash('success', $translator->trans('The Plugin was deleted!'));
			$updater->deleteFullCacheOnExit();
			return $this->redirectToRoute('app_plugins');
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('app_plugins');
		}

		return $this->redirectToRoute('app_plugins', [], 303);
	}

	/***
	 * Senden des aktuellen Plugin-Zustands und der -Konfiguration an alle Cluster-Mitglieder.
	 */
	public function clusterApiBroadcast(PluginInfo $plugin, ?PluginState $state)
	{
		if (!$state) {
			ClusterApi::getInstance()->callRemoteApiBroadcast('app_cluster_api_plugin_disable', ['plugin' => $plugin->getName()], 'DELETE');
		}
		else {
			$config = [
				'stateId' => $state->getId(),
				'activateForAllAccounts' => $state->isActivatedForAllAccounts(),
				'activateForAllAccountsOnTheirOwn' => $state->isActivatedForAllAccountsOnTheirOwn(),
				'fullAccessForAllAccounts' => $state->isFullAccessForAllAccounts(),
				'fullAccessCallCount' => $state->getFullAccessCallCount(),
				'fullAccessDomCount' => $state->getFullAccessDomCount(),
				'testAccessCallCount' => $state->getTestAccessCallCount(),
				'testAccessDomCount' => $state->getTestAccessDomCount(),
				'testAccessTime' => $state->getTestAccessTime(),
				'savedAccess' => $state->getSavedAccess(),
			];
			ClusterApi::getInstance()->callRemoteApiBroadcast('app_cluster_api_plugin_config', ['plugin' => $plugin->getName()], 'PUT', $config);
		}
	}
}
