<?php

namespace App\Controller;

use App\Component\Licensing\ApiClient as LicensingApiClient;
use App\Component\LoggerInterface;
use App\Component\MailerConfigurator;
use App\Component\Update;
use App\Config;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\User;
use App\Model\VersionLog;
use App\Utils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;
use Symfony\Contracts\Translation\TranslatorInterface;
use MongoDB;

class Setup extends AbstractController
{
	/**
	 * Diese Methode bildet den Einstiegspunkt der Setuproutine.
	 * @Route("/setup", name="app_setup")
	 */
	public function setup(TranslatorInterface $translator, LoggerInterface $logger, MailerConfigurator $mailconfig, Update $updater, LicensingApiClient $licenseApi, Request $request): Response
	{
		/** @var Config $config */
		$config = Config::getInstance();
		$session = $request->getSession();

		if ($session) {
			// Locale-Wechsel bei POST mit locale
			$wantedLocale = $request->query->get('switchToLocale');
			if ($wantedLocale) {
				$session->set('locale', $wantedLocale);
				$request->setLocale($wantedLocale);
				return $this->redirectToRoute('app_setup');
			}
		}

		// Wenn es bereits Benutzer gibt, gilt das Setup als durchgeführt
		if (User::count() > 0) {
			return $this->render('setup/already_done.html.twig', [])->setStatusCode(403);
		}

		$username = 'admin';
		$password = Utils::randomString(16);
		$licensekey = '';
		$email = '';
		$repositoryStorage = Utils::REPOSITORY_STORE_FILE;
		$consentArchiveStorage = Utils::REPOSITORY_STORE_FILE;
		$consentArchiveRoot = '';
		$mongoDbConnectionString = '';
		$mongoDbDatabase = '';
		$errors = [];
		$proxyUrl = $_ENV['APP_PROXY'] ?? $_ENV['HTTP_PROXY'] ?? '';

		// Wenn das Formular abgeschickt wurde, Daten prüfen und ggf. Einrichtung des Systems vornehmen
		if ($request->getRealMethod() === 'POST') {
			$username = trim($request->request->get('username'));
			$password = trim($request->request->get('password'));
			$licensekey = trim($request->request->get('licensekey'));
			$email = trim($request->request->get('email'));
			$proxyUrl = trim($request->request->get('proxyUrl'));
			$repositoryStorage = trim($request->request->get('repositoryStorage'));
			$consentArchiveStorage = ($repositoryStorage == Utils::REPOSITORY_STORE_FILE) ? Utils::REPOSITORY_STORE_FILE : trim($request->request->get('consentArchiveStorage'));
			$mongoDbConnectionString = trim($request->request->get('mongoDbConnectionString'));
			$mongoDbDatabase = trim($request->request->get('mongoDbDatabase'));
			$consentArchiveRoot = trim($request->request->get('consentArchiveRoot'));

			// MongoDB-Connection überprüfen und ggf. Fehlermeldung anzeigen
			if ($repositoryStorage == Utils::REPOSITORY_STORE_MONGODB) {
				$mongodbCheckResult = $this->checkMongoDbAccess($mongoDbConnectionString, $mongoDbDatabase);
				if ($mongodbCheckResult['uri']['success'] === false) {
					switch ($mongodbCheckResult['uri']['reason'] ?? '') {
						case 'version':
						case 'transactions':
						case 'php-extension':
							$errors[] = $translator->trans($mongodbCheckResult['uri']['message'], array_combine(
								array_map(function ($key) { return ':'.$key; }, array_keys($mongodbCheckResult['uri'])),
								array_values($mongodbCheckResult['uri'])
							));
							break;
						default:
							$errors[] = $translator->trans('MongoDB Connection URI invalid: :message', [':message'=>$mongodbCheckResult['uri']['message']]);
					}
				}
				if ($mongodbCheckResult['database']['success'] === false) {
					$errors[] = $translator->trans('MongoDB Database Name invalid: :message', [':message'=>$mongodbCheckResult['database']['message']]);
				}
			}

			// Proxy und Datenbank direkt verwenden, wenn eingerichtet
			$_ENV['APP_PROXY'] = $proxyUrl;
			if (!$errors) {
				$_ENV['APP_REPOSITORY_STORE'] = $repositoryStorage;
				$_ENV['APP_MONGODB_CONNECTION_STRING'] = $mongoDbConnectionString;
				$_ENV['APP_MONGODB_DATABASE'] = $mongoDbDatabase;
			}

			// Lizenzschlüssel prüfen
			$licenseData = null;
			try {
				$licenseData = $licenseApi->registerLicense($licensekey, $request->getHttpHost());
			} catch (\Exception $e) {
				$logger->logError('Setup', "Installation failed: ".$e->getMessage());
				$errors[] = $translator->trans($e->getMessage());
			}

			// Lizenz-Einschränkungen
			if ($licenseData and $licenseData->edition === 'base' and $repositoryStorage == Utils::REPOSITORY_STORE_MONGODB) {
				$errors[] = $translator->trans("Your license doesn't include MongoDB support. Please select Files storage to continue.");
			}

			// Archiv-Pfad überprüfen
			if (($consentArchiveRoot) !== '' and $consentArchiveStorage === Utils::REPOSITORY_STORE_FILE) {
				if (!file_exists($consentArchiveRoot) or !is_dir($consentArchiveRoot)) {
					$errors[] = $translator->trans("The given consent archive directory doesn't exist. Please choose another directory or leave the field empty.");
				} elseif (!is_writable($consentArchiveRoot)) {
					$errors[] = $translator->trans("The given consent archive directory is not writable. Please choose another directory or leave the field empty.");
				}
			}

			if ($username && $password && $email && $licenseData && !$errors) {
				$environ = [
					'APP_SECRET' => bin2hex(random_bytes(16)),
					'APP_EDITION' => $licenseData->edition,
					'APP_REPOSITORY_STORE' => $repositoryStorage,
					'APP_PROTOCOL_ARCHIVE_STORE' => $consentArchiveStorage,
				];

				if ($repositoryStorage == Utils::REPOSITORY_STORE_MONGODB) {
					$environ['APP_MONGODB_CONNECTION_STRING'] = $mongoDbConnectionString;
					$environ['APP_MONGODB_DATABASE'] = $mongoDbDatabase;
				}

				if ($consentArchiveStorage == Utils::REPOSITORY_STORE_FILE and $consentArchiveRoot !== '') {
					$environ['APP_PROTOCOL_ARCHIVE_ROOT'] = $consentArchiveRoot;
				}

				Utils::setLocalEnvironment($environ);

				User::reloadRepository();

				// Cache komplett leeren
				$updater->deleteFullCacheOnExit();

				// Wenn nach Zugriff auf die Datenbank schon eine Installation gefunden wird:
				if (User::count() > 0) {
					// Already-Done-Nachricht anzeigen
					return $this->redirectToRoute('app_setup', [], 303);
				}

				User::create($username)
					->setPassword($password)
					->setRole(User::ROLE_ADMIN)
					->setActive(true)
					->setMaxDomainCount($licenseData->limit)
					->save()
				;
				$user = User::login($username, $password);
				assert($user instanceof User);

				$config->reload();
				$config->set('multiDomain', true); // Zum Unterscheiden alter und neuer Versionen (Updateroutinen)
				$config->set('licensekey', $licensekey);
				$config->set('email', $email);
				$config->set('whitelabel', $licenseData->whitelabel);
				$config->setBaseUrl($this->generateUrl('app_root', [], UrlGeneratorInterface::ABSOLUTE_URL));

				//Standard Warn Mail vorbelegen sub_cookie_email_check
				$config->set('sub_cookie_email_check', "on");
				$config->set('sub_cookie_email_betreff', "Es wurden neue unkontrollierte Inhalte auf Ihrer Webseite {{ domainvar }} entdeckt.");
				$config->set('sub_cookie_email_content', "Sehr geehrte Damen und Herren,

CCM19 hat folgende neue unkontrollierte Inhalte auf Ihrer Webseite {{ domainvar }} entdeckt:

#uncontrolledData#

Bitte überprüfen Sie die Daten in Ihrem CCM19 Account unter dem Menüpunkt Pagecheck unter Cookies & Einstellung.

Mit freundlichen Grüßen
CCM19");
				//to much Traffic mail
				$config->set('max_traffic_email_betreff', "ALARM - Ihr  CCM19 Cookie Consent Manager wird für den Rest des Monats nicht mehr angezeigt.");
				$config->set('max_traffic_email_content', "Sehr geehrte Damen und Herren,

Sie haben die maximale Anzahl von Aufrufen in Ihrem Tarif für {{ domainvar }} erreicht. Die maximale Anzahl beträgt {{ callMaxCount }}, Ihr Kontingent für diesen Monat ist zu {{ quota }} % erschöpft.
Details dazu finden Sie hier: https://ServiceUrl.de/app/public/login

Da Sie die maximale Anzahl erreicht haben, wird das Cookie WIdget auf allen Ihren Seiten nicht mehr ausgespiel und die Cookies werden ungefragt gesetzt werden können.

Damit das CCM19 Cookie Consent Widget auch weiterhin auf Ihrer Seite ausgespielt werden kann, ordern Sie bitte einen höheren Tarif als derzeit. 

Eine Übersicht der Tarife mit Buchungsmöglichkeit finden Sie hier: 
https://www.....

Loggen Sie sich dort mit Ihren Benutzerdaten ein und suchen Sie sich hier den passenden Tarif für Ihren Traffic aus.

Um ein Tarifupgrade durchzuführen, klicken Sie auf den Button \"Buchen\" des passenden Tarifs, danach werden Sie in den Warenkorb weitergeleitet.

Auf diesem Wege ist sichergestellt, dass der gebuchte Tarif auch auf Ihren Account aufgebucht wird.

Falls etwas mit dem Login nicht klappt, antworten Sie am besten auf diese Mail und teilen Sie uns das Problem mit damit wir uns für Sie darum kümmern können.

Mit herzlichen Grüßen,
Ihr Team von ....");

				//Max traffic erreicht mail
				$config->set('sub_max_traffic_email_betreff', "Achtung - Ihr CCM19 Cookie Consent Manager wird bald nicht mehr angezeigt.");
				$config->set('sub_max_traffic_email_content', "Sehr geehrte Damen und Herren,

Sie haben die maximale Anzahl von Aufrufen in Ihrem Tarif für {{ domainvar }} bald erreicht. Die maximale Anzahl beträgt {{ callMaxCount }}, Ihr Kontingent für diesen Monat ist zu {{ quota }} % erschöpft.
Details dazu finden Sie hier: https://ServiceUrl.de/app/public/login

Sobald Sie die maximale Anzahl erreicht haben, wird das Cookie Widget auf allen Ihren Seiten in diesem Tarif nicht mehr ausgespielt und die Cookies werden ungefragt gesetzt werden können.

Damit das CCM19 Cookie Consent Widget auch weiterhin auf Ihrer Seite ausgespielt werden kann, ordern Sie bitte einen höheren Tarif als derzeit. 

Eine Übersicht der Tarife mit Buchungsmöglichkeit finden Sie hier: 
https://www.....

Loggen Sie sich dort mit Ihren Benutzerdaten ein und suchen Sie sich hier den passenden Tarif für Ihren Traffic aus.

Um ein Tarifupgrade durchzuführen, klicken Sie auf den Button \"Buchen\" des passenden Tarifs, danach werden Sie in den Warenkorb weitergeleitet.

Auf diesem Wege ist sichergestellt, dass der gebuchte Tarif auch auf Ihren Account aufgebucht wird.

Falls etwas mit dem Login nicht klappt, antworten Sie am besten auf diese Mail und teilen Sie uns das Problem mit damit wir uns für Sie darum kümmern können.

Sehr gerne stehen wir Ihnen für weitere Fragen zur Verfügung.

Mit herzlichen Grüßen,
Ihr Team von ....");

				//Crawler Found Mail nicht aktiviert
				$config->set('sub_crawler_email_betreff', "Achtung - auf {{ domainvar }} hat CCM19 neue Einbindungen und Cookies gefunden.");
				$config->set('sub_crawler_email_content', "Sehr geehrte Damen und Herren,

der CCM19 Crawler hat auf Ihrer Domain: {{ domainvar }} neue Cookies oder Skripte gefunden. 
{{ uncontrolledData }}
Bitte überprüfen Sie die Daten in Ihrem CCM19 Account unter dem Menüpunkt Einbindungen.
Gemäß Ihrer Vorgaben wurden die Einbindungen nicht aktiviert - bitte überprüfen Sie die Einbindungen und aktivieren Sie diese, wenn Sie diese benötigen.
Falls Sie diese nicht aktivieren, werden die Cookies und Skripte ohne Consent für Ihre Besucher ausgespielt, was rechtlich problematisch sein könnte.

Bei Fragen wenden Sie sich bitte an unser Support Team!

Mit herzlichen Grüßen,
Ihr Team von ....");

				//Crawler Found Mail wurde aktiviert
				$config->set('sub_crawler_activated_email_betreff', "Achtung - auf {{ domainvar }} hat CCM19 neue Einbindungen und Cookies gefunden.");
				$config->set('sub_crawler_activated_email_content', "Sehr geehrte Damen und Herren,

der CCM19 Crawler hat auf Ihrer Domain: {{ domainvar }} neue Cookies oder Skripte gefunden. 
{{ uncontrolledData }}
Bitte überprüfen Sie die Daten in Ihrem CCM19 Account unter dem Menüpunkt Einbindungen.
Gemäß Ihrer Vorgaben wurden die Einbindungen direkt aktiviert - bitte überprüfen Sie die Einbindungen und deaktivieren Sie diese, falls Sie diese nicht benötigen.

Bei Fragen wenden Sie sich bitte an unser Support Team!

Mit herzlichen Grüßen,
Ihr Team von ....");

				// Sendmail-Pfad setzen
				if ($_ENV['MAILER_URL'] === 'sendmail:') {
					$mailconfig->setMailer('sendmail');
				}

				Utils::setLocalEnvironment([
					'APP_PROXY' => $proxyUrl
				]);

				// Versions-Installation merken
				VersionLog::logCurrentVersion();

				$logger->logInfo('Setup', "Installation successful");

				return $this->redirectToRoute('app_root', [], 303);
			}
		}

		return $this->render('setup/index.html.twig', [
			'username' => $username,
			'password' => $password,
			'licensekey' => $licensekey,
			'email' => $email,
			'proxyUrl' => $proxyUrl,
			'repositoryStorage' => $repositoryStorage,
			'mongoDbConnectionString' => $mongoDbConnectionString,
			'mongoDbDatabase' => $mongoDbDatabase,
			'consentArchiveStorage' => $consentArchiveStorage,
			'consentArchiveRoot' => $consentArchiveRoot,
			'errors' => $errors,
			'mongoDbSupported' => Utils::mongoDbSupportAvailable(),
			'csrfToken' => CsrfToken::get(),
			'userDataDir' => Utils::getBundledVarDir() . DIRECTORY_SEPARATOR . 'user_data',
		]);
	}

	private function checkMongoDbAccess(string $uri, string $db): array
	{
		if (!Utils::mongoDbSupportAvailable()) {
			return [
				'uri' => [
					'success' => false,
					'reason' => 'php-extension',
					'message' => 'MongoDB PHP extension is missing',
				],
				'database' => [
					'success' => null,
					'message' => '',
				],
				'databases' => [],
			];
		}
		try {
			$client = new MongoDB\Client($uri);
			$session = $client->startSession();
		} catch (\Exception $exception) {
			return [
				'uri' => [
					'success' => false,
					'message' => $exception->getMessage(),
				],
				'database' => [
					'success' => null,
					'message' => '',
				],
				'databases' => [],
			];
		}

		$databases = [];
		try {
			foreach ($client->listDatabases() as $foundDb) {
				$databases[] = $foundDb->getName();
			}
		} catch (\Exception $exception) {
			// ignore
		}

		try {
			$database = $client->selectDatabase($db);
			$database->listCollections();
		} catch (\Exception $exception) {
			return [
				'uri' => [
					'success' => true,
					'message' => '',
				],
				'database' => [
					'success' => false,
					'message' => $exception->getMessage(),
				],
				'databases' => $databases,
			];
		}

		// MongoDb-Version prüfen
		$mongodbInfo = $database->command(['serverStatus'=>true]);
		$mongodbVersion = $mongodbInfo->toArray()[0]['version'];
		if (explode('.', $mongodbVersion, 2)[0] < 4) {
			return [
				'uri' => [
					'success' => false,
					'reason' => 'version',
					'version' => $mongodbVersion,
					'message' => 'MongoDB server version :version too old. At least MongoDB 4.0.0 is required.',
				],
				'database' => [
					'success' => true,
					'message' => '',
				],
				'databases' => $databases,
			];
		}

		// MongoDb-Transaktionssupport
		try {
			$session = $client->startSession();
			$session->startTransaction([]);
			$database->selectCollection('cm-config')->findOne([], ['projection'=>[], 'session'=>$session]);
			$session->commitTransaction();
		} catch (\Exception $exception) {
			return [
				'uri' => [
					'success' => false,
					'reason' => 'transactions',
					'message' => 'MongoDB is configured without transaction support. Please activate replica sets (4.0+) or sharded clustering (4.2+) in your MongoDB server configuration. (:submessage)',
					'submessage' => $exception->getMessage(),
				],
				'database' => [
					'success' => true,
					'message' => '',
				],
				'databases' => $databases,
			];
		}

		return [
			'uri' => [
				'success' => true,
				'message' => '',
			],
			'database' => [
				'success' => true,
				'message' => '',
			],
			'databases' => $databases,
		];
	}

	/**
	 * AJAX-Check der MongoDB-Verbindung
	 * @Route("/setup/mongodb_check", name="app_setup_mongodb_check", methods={"POST"})
	 */
	public function mongoDbAjaxCheck(TranslatorInterface $translator, Request $request): JsonResponse
	{
		// Wenn es bereits Benutzer gibt, gilt das Setup als durchgeführt
		if (User::count() > 0) {
			return $this->json(['error'=>'not available'], 404);
		}

		try {
			CsrfToken::get()->checkRequest($request);
		}
		catch (CsrfTokenException $e) {
			return $this->json(
				['error'=>'csrftoken invalid'], 400
			);
		}
		if ($request->request->get('uri', '') == '') {
			return $this->json([
				'uri' => [
					'success' => false,
					'message' => '',
				],
				'database' => [
					'success' => null,
					'message' => '',
				],
				'databases' => [],
			]);
		}

		$data = $this->checkMongoDbAccess($request->request->get('uri', ''), $request->request->get('db', ''));

		if (!empty($data['uri']['reason'])) {
			$data['uri']['message'] = $translator->trans($data['uri']['message'], array_combine(
				array_map(function ($key) { return ':'.$key; }, array_keys($data['uri'])),
				array_values($data['uri'])
			));
		}

		return $this->json(
			$data
		);
	}
}
