<?php

namespace App\Controller;

use App\Component\LoggerInterface;
use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Exception\CsrfTokenException;
use App\Exception\LogAccessException;
use App\Model\CsrfToken;
use App\Model\SystemLog as SystemLogModel;
use App\Model\User;
use DateTime;
use RuntimeException;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\StreamedResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class SystemLog extends AdminController
{
	/**
	 * @Menu("System log", group="Journals", icon="fa-book", order=120)
	 * @Route("/systemlog", methods={"HEAD", "GET"}, name="app_systemlog")
	 * @Route("/systemlog/{year<\d{4,5}>}/{month<\d{1,2}>}", methods={"HEAD", "GET"}, name="app_systemlog_month")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param int|null $year
	 * @param int|null $month
	 * @return Response
	 */
	public function logPage(Request $request, TranslatorInterface $translator, ?int $year=null, ?int $month=null): Response
	{
		$date = ($year === null or $month === null) ? new DateTime('now') : new DateTime("$year-$month-01 12:00:00 Z");
		$thisMonth = new DateTime('first day of this month midnight');

		$dates = SystemLogModel::listDateRanges();

		$logEntries = [];
		try {
			$protocol = SystemLogModel::byDate($date);
			$search = $request->query->get('search', '');
			if ($search == '') {
				$logEntries = $protocol->all();
			} else {
				$logEntries = $protocol->search($search);
			}
		}
		catch (LogAccessException $exception) {
			$this->addFlash('danger', $translator->trans('Logfile %file% could not be read, check permissions.', [
				'%file%' => $exception->getFilename(),
			]));
		}
		catch (RuntimeException $exception) {
			$this->addFlash('danger', $translator->trans('An unexpected error occured while reading the logfile: %msg%.', [
				'%msg%' => $exception->getMessage(),
			]));
		}

		$pagination = new Pagination($logEntries, 20);
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			return $this->redirectToRoute('app_systemlog');
		}

		return $this->render('systemlog/index.html.twig', [
			'pagination' => $pagination,
			'date' => $date,
			'dates' => $dates,
			'showDelete' => $date < $thisMonth,
			'csrfToken' => (string)CsrfToken::get(),
			'usersearch' => $search ?? null,
		]);
	}



	/**
	 * @Route("/systemlog/{year<\d{4,5}>}/{month<\d{1,2}>}.log", methods={"HEAD", "GET"}, name="app_systemlog_month_export")
	 *
	 * @param Request $request
	 * @param int $year
	 * @param int $month
	 * @return Response
	 */
	public function logExport(Request $request, int $year, int $month): Response
	{
		$date = new DateTime("$year-$month-01 12:00:00 Z");
		$thisMonth = new DateTime('first day of this month midnight');
		$filename = ($date < $thisMonth) ? sprintf('ccm19-system-%04d-%02d.log', $year, $month) : 'ccm19-system-'.date('Y-m-d_Hi').'.log';

		$protocol = SystemLogModel::byDate($date);
		return new StreamedResponse(function () use ($protocol) {
			foreach ($protocol->all() as $entry) {
				echo json_encode($entry,  JSON_UNESCAPED_UNICODE)."\n";
			}
		}, 200, [
			'Content-Type' => 'text/plain; charset=utf-8',
			'Cache-Control' => ($date < $thisMonth) ? 'private, max-age=600' : 'private, max-age=0',
			'Content-Disposition' => 'attachment; filename="'.$filename.'"'
		]);
	}

	/**
	 * @Route("/systemlog/{year<\d{4,5}>}/{month<\d{1,2}>}", methods={"DELETE"}, name="app_systemlog_month_delete")
	 *
	 * @param Request $request
	 * @param LoggerInterface $logger
	 * @param TranslatorInterface $translator
	 * @param int $year
	 * @param int $month
	 * @return Response
	 */
	public function deleteLog(Request $request, LoggerInterface $logger, TranslatorInterface $translator, int $year, int $month): Response
	{
		$date = new DateTime("$year-$month-01 12:00:00 Z");
		$thisMonth = new DateTime('first day of this month midnight');
		$user = User::loggedInUser();

		if ($date < $thisMonth) {

			try {
				CsrfToken::get()->checkRequest($request);
				SystemLogModel::deleteDateRange($date);
				$logger->logWarning('SystemLog', 'Logs for '.$date->format('M Y').' deleted by user '.$user->getUsername());

				$this->addFlash('success', $translator->trans('The logs for :month were deleted successfully.', [
					':month' => $date->format('M Y'),
				]));
				return $this->redirectToRoute('app_systemlog');
			}
			catch (LogAccessException $exception) {
				$this->addFlash('danger', $translator->trans('Logfile %file% could not be deleted, check permissions.', [
					'%file%' => $exception->getFilename(),
				]));
			}
			catch (CsrfTokenException $exception) {
				$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			}
			catch (RuntimeException $exception) {
				$this->addFlash('danger', $translator->trans('An unexpected error occured while deleting the logfile: %msg%.', [
					'%msg%' => $exception->getMessage(),
				]));
			}
		} else {
			$this->addFlash('danger', $translator->trans('Only logs in the past can be deleted.'));
		}

		return $this->redirectToRoute('app_systemlog', ['year' => $year, 'month' => $month]);
	}
}
