<?php
/** @noinspection PhpUnused */
/** @noinspection PhpRouteMissingInspection */
/** @noinspection PhpTemplateMissingInspection */

namespace App\Controller;

use App\Component\Color;
use App\Component\LocaleService;
use App\Component\Menu\Annotation\Menu;
use App\Exception\ConfigNotWritableException;
use App\Model\BaseLayout;
use App\Model\Domain;
use App\Model\User;
use App\Model\Theme as ThemeModel;
use App\Model\SystemTheme;
use App\Model\UserTheme;
use App\Utils;
use App\Config;
use RuntimeException;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use Twig\Error\SyntaxError;

/**
 * Theme-Verwaltung
 * @package App\Controller
 *
 * @Route("/domains/{_domainId}/themes", name="app_theme_")
 */
class Theme extends DomainDependantController
{
	/**
	 * @Menu("Themes", group="Cookies & Settings", icon="fa-picture-o", order=10)
	 * @Route("", name="list")
	 *
	 * @return Response
	 */
	public function listThemes(): Response
	{
		//Evel. noch nicht genutzten Skeletons importieren.
		$this->importSkeletons();

		$themes = ThemeModel::allOrderedByTypeAndName();

		$hasSystemThemes = array_reduce($themes, function ($carry, ThemeModel $theme) {
			return $carry || (bool)$theme->getSystemTheme();
		}, false);
		$hasUserThemes = array_reduce($themes, function ($carry, ThemeModel $theme) {
			return $carry || (bool)$theme->getUserTheme();
		}, false);

		return $this->render('theme/list.html.twig', [
			'user' => User::loggedInUser(),
			'themes' => $themes,
			'hasSystemThemes' => $hasSystemThemes,
			'hasUserThemes' => $hasUserThemes,
		]);
	}

	/**
	 * Importiert neue Layouts aus dem Skeleton Themes - damit auch die alten Accounts aktuelle Themes nutzen können
	 * die als Vorlagen neu reinkommen
	 */
	public function importSkeletons(): void
	{
		//Alle Themes
		$themes = ThemeModel::all();

		//Alle Skeletons
		$skeletons = ThemeModel::readSkeletonData();

		//loopen
		foreach($skeletons as $key=>$skeletonItem)
		{
			$exists = false;
			foreach($themes as $theme)
			{
				//Nach Name und Skeleton Key checken - ja - initial kann es sein dass es zu Doppelungen kommt falls jmd. die Org Themes umbenannt hat... läßt sich nicht vermeiden ist auch nicht schlimm...
				if($theme->getName()==$skeletonItem['name'] || $theme->get('skeletonKey')==$key)
				{
					$exists = true;
				}
			}

			//gibts noch nicht, dann ein neues anlegen...
			if($exists==false)
			{
				$newTheme = ThemeModel::create();
				foreach($skeletonItem as $newThemeKey => $newTthemeItem)
				{
					// Standard-Flag für neue Themes nicht übernehmen
					if ($newThemeKey == 'isDefault') {
						continue;
					}

					$newTheme->set($newThemeKey,$newTthemeItem);
				}
				//und auch den Skeletonkey setzen, für den Fall dass ein Theme umbennant wird
				$newTheme->set("skeletonKey",$key);

                //Changelog auf 0 setzen, das gibts ja nix drin da der User nix geändert hat
                $newTheme->set("changeLog",array());
				$newTheme->save();
			}
		}
	}

	/**
	 * Erzeugt ein Bild - entweder Screenshot der Domain oder Standard Hintergrund.
	 *
	 * @Route("/{id}/preview_img", name="preview_img" )
	 * @param string $id
	 * @param Request $request
	 * @return BinaryFileResponse
	 */
	public function previewImage(string $id, Request $request): BinaryFileResponse
	{
		$domain = Domain::activeDomain();

		//Refreshtime setzen für den Refresh des Screeshots der Webseite
		$refresh = false;
		if($domain->get("ScreenShotRefreshTime") < (time()-(30*60)))
		{
			$domain->set("ScreenShotRefreshTime",time());
			$domain->save();
			$refresh = true;
		}

		//wenn eine Domain vorgegeben ist... dann Screenshot holen
		if(!empty($domain->getName()))
		{
			//refresh mitgeben true/false
			$pageScreenshot = $domain->getScreenShot($domain->getName(), $refresh);
		} else {
			$pageScreenshot = null;
		}

		//Keine Domain oder Screenshot nicht verfügbar - dann Standard Background Image aus public
		if (!$pageScreenshot) {
			$bgimg = Utils::getBackgroundImage();
			$pageScreenshot = Utils::getBaseDir()."/public/".$bgimg['src'] ;
		}
		//Bild erzeugen
		return new BinaryFileResponse($pageScreenshot, 200, [], true, null, true, true);
	}

	/**
	 * @Route("/{id}/preview", name="preview" )
	 *
	 * @param string $id
	 * @param Request $request
	 * @return Response
	 */
	public function previewTheme(string $id, Request $request): Response
	{
		if (ThemeModel::exists($id)) {
			$theme = ThemeModel::find($id);
			$domain = Domain::activeDomain();

			$config = Config::getInstance();
			$user = User::loggedInUser();

			$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme() . '://';
			$baseUrl = $scheme . $request->getHttpHost();

			$embedScriptUrl = $baseUrl . $this->generateUrl(
				'app_external_main_javascript_file',
				[
					'apiKey' => $user->getApiKey(),
					'domain' => $domain->getId(),
					'theme' => $theme->getId(),
					'preview' => 'true',
				]
			);

			$embedCodeSnippet = '<script src="' . htmlspecialchars($embedScriptUrl) . '" referrerpolicy="origin"></script>';

			$response = $this->render('theme/preview.html.twig', [
				'user' => User::loggedInUser(),
				'embedScript' => $embedCodeSnippet,
				'theme' => $theme,
				'domName' => $domain->getName(),
			]);

			$origins = $theme->getCustomCssOrigins();
			$originsText = implode(' ', $origins);
			$response->headers->set('Content-Security-Policy',
				"default-src 'self' 'unsafe-inline' $originsText; ".
				"img-src 'self' data:; ".
				"script-src 'self' 'unsafe-inline'; ".
				"connect-src 'self'; ".
				"form-action 'none'"
			);

			return $response;
		} else {
			return $this->redirectToRoute('app_theme_list');
		}
	}

	/**
	 * @param ThemeModel $theme
	 * @return void
	 * @throws RuntimeException if selected features are not available with the current plan.
	 */
	private function checkPlanConstraints(ThemeModel $theme): void
	{
		if (Utils::isFreePlan()) {
			$baseLayout = $theme->getBaseLayout();
			if ($baseLayout->isEmpty()) {
				$message = Utils::getTranslator()->trans('Base layout :name is not available with your current plan.', [
					':name' => '<q>' . htmlspecialchars($baseLayout->getName()) . '</q>',
				]);
				$this->addFlash('danger|raw', $message);
				throw new RuntimeException($message);
			}
		}
	}

	/**
	 * Prüft, ob das CSS für das Template so gerendert werden kann.
	 *
	 * @throws SyntaxError
	 */
	private function checkThemeCssTemplate(ThemeModel $theme): void
	{
		$this->render('external/cookie-management.css.twig', [
			'theme' => $theme,
			'breakpoints' => [
				'mobileMax' => '40em',
				'tabletMax' => '64em',
				'desktop' => '64.0625em',
			],
		]);
	}

	/**
	 * @Route("/new", name="new", methods={"GET","HEAD","POST"})
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function createTheme(Request $request, TranslatorInterface $translator): Response
	{
		$action = $request->request->getAlnum('action');
		$mayWhitelabel = Utils::hasWhitelabelLicense();

		$mayGeoIP = (phpversion('bcmath') !== false or extension_loaded('bcmath') !== false or phpversion('gmp') !== false or extension_loaded('gmp') !== false);

		if ($action === 'save' or $action === 'saveClose') {
			$themeData = $request->request->all();

			// System- und User-Theme-Models hier schon laden, um Deadlocks zu vermeiden
			SystemTheme::all();
			UserTheme::all();

			$domain = Domain::activeDomain();
			assert($domain !== null);
			$theme = ThemeModel::create();
			$this->setThemeObjectAttributes($theme, $themeData, null, false);

			try {
				$this->handleNameError($themeData['name'], $translator);
				$this->handleColorErrors($themeData, $translator);
				$this->handlePositionError($themeData['position'], $translator);

				$this->checkPlanConstraints($theme);
				$this->checkThemeCssTemplate($theme);

				// Bilduploads verarbeiten
				$this->migrateFileUploads($domain, $theme, $request, $translator);
			} catch (SyntaxError $e) {
				$message = preg_replace('/"__string_template_[^"]*"/', '‹Custom CSS›', $e->getMessage());

				$this->addFlash(
					'danger',
					$translator->trans(
						'Variable syntax error in CSS: %msg%', [
							'%msg%' => $message,
						]
					)
				);
				return $this->render('theme/new.html.twig', [
					'theme' => $theme,
					'whitelabel' => $mayWhitelabel,
					'mayGeoIP' => $mayGeoIP,
					'baseLayouts' => BaseLayout::all(),
					'allRestrictRegions' => ThemeModel::allRestrictRegions(),
					'poweredByText' => Utils::getPoweredByText(),
				]);
			} catch (RuntimeException $e) {
				return $this->render('theme/new.html.twig', [
					'theme' => $theme,
					'whitelabel' => $mayWhitelabel,
					'mayGeoIP' => $mayGeoIP,
					'baseLayouts' => BaseLayout::all(),
					'allRestrictRegions' => ThemeModel::allRestrictRegions(),
					'poweredByText' => Utils::getPoweredByText(),
				]);
			}
			$theme->save();

			$this->addFlash('success', $translator->trans(
				'The theme "%themename%" has been created successfully!', [
					'%themename%' => $themeData['name']
				]
			));

			if ($action === 'saveClose') {
				return $this->redirectToRoute('app_theme_list', [], 303);
			} else {
				return $this->redirectToRoute('app_theme_edit', ['id' => $theme->getId()], 303);
			}
		}

		return $this->render('theme/new.html.twig', [
			'whitelabel' => $mayWhitelabel,
			'mayGeoIP' => $mayGeoIP,
			'baseLayouts' => BaseLayout::all(),
			'allRestrictRegions' => ThemeModel::allRestrictRegions(),
			'poweredByText' => Utils::getPoweredByText(),
		]);
	}

	/**
	 * @Route("/{id}", name="edit", methods={"GET","HEAD"})
	 *
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editTheme(string $id, Request $request, TranslatorInterface $translator): Response
	{
		if (ThemeModel::exists($id) == false) {
			return $this->redirectToRoute('app_theme_list');
		}

		$theme = ThemeModel::find($id);
		$domain = Domain::activeDomain();
		$user = User::loggedInUser();
		assert($domain !== null);

		// System- und User-Theme-Models hier schon laden, um Deadlocks zu vermeiden
		$theme->getSystemTheme();
		$theme->getUserTheme();

		$mayWhitelabel = Utils::hasWhitelabelLicense();

		$mayGeoIP = (phpversion('bcmath') !== false or extension_loaded('bcmath') !== false or phpversion('gmp') !== false or extension_loaded('gmp') !== false);

		$logoUrl = $this->generateUrl('app_logo_frontend', [
			'userId' => $user->getId(),
			'domainId' => $domain->getId(),
			'themeId' => $theme->getId(),
		]);

		if ($theme->getSystemTheme() || $theme->getUserTheme()) {
			$this->addFlash('info',
				$translator->trans('Please note that e. g. the positioning and coloring of global themes may not be edited. Duplicate the theme to change those parameters.')
			);
		}

		if ($domain->isPartOfAdAndTrackingFreeNetwork()) {
			$this->addFlash('info', $translator->trans('Changes to most settings will have no effect, as the <q>%setting%</q> setting is active and the banner is not displayed.', [
				'%setting%' => $translator->trans('My website is part of an ad and tracking-free network'),
			]));
		}

		return $this->render('theme/edit.html.twig', [
			'user' => User::loggedInUser(),
			'locale' => LocaleService::findEnabledFuzzy($request->getLocale()),
			'theme' => $theme,
			'logoUrl' => $logoUrl,
			'whitelabel' => $mayWhitelabel,
			'mayGeoIP' => $mayGeoIP,
			'themeId' => $theme->getId(),
			'baseLayouts' => BaseLayout::all(),
			'allRestrictRegions' => ThemeModel::allRestrictRegions(),
			'poweredByText' => Utils::getPoweredByText(),
		]);
	}

	/**
	 * @Route("/{id}", name="action", methods={"POST"})
	 *
	 * @param string $id
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function themeAction(string $id, Request $request, TranslatorInterface $translator)
	{
		if (ThemeModel::exists($id) == false) {
			return $this->redirectToRoute('app_theme_list');
		}

		$theme = ThemeModel::find($id);
		$domain = Domain::activeDomain();
		$user = User::loggedInUser();
		assert($domain !== null);

		// Wenn das Theme nicht existiert, abbrechen
		if ($theme === null) {
			return $this->redirectToRoute('app_theme_list', [], 303);
		}

		// System- und User-Theme-Models hier schon laden, um Deadlocks zu vermeiden
		$theme->getSystemTheme();
		$theme->getUserTheme();

		$mayWhitelabel = Utils::hasWhitelabelLicense();

		$mayGeoIP = (phpversion('bcmath') !== false or extension_loaded('bcmath') !== false or phpversion('gmp') !== false or extension_loaded('gmp') !== false);

		$logoUrl = $this->generateUrl('app_logo_frontend', [
			'userId' => $user->getId(),
			'domainId' => $domain->getId(),
			'themeId' => $theme->getId(),
		]);

		$action = $request->request->getAlnum('action');

		switch ($action) {
			case 'save':
			case 'saveClose':
				$themeData = $request->request->all();

				if ($mayWhitelabel == false) {
					$themeData['whitelabel'] = false;
				}

				$logo = $theme->getLogo();

				$this->setThemeObjectAttributes($theme, $themeData, $logo);

				try {
					// Überprüfe die Felder nur, wenn das Theme kein System-Theme oder User-Theme referenziert
					if (!$theme->getSystemTheme() && !$theme->getUserTheme()) {
						$this->handleNameError($themeData['name'], $translator);
						$this->handleColorErrors($themeData, $translator);
						$this->handlePositionError($themeData['position'], $translator);

						$this->checkPlanConstraints($theme);
					}
					else {
						$this->handleColorErrors(array_intersect_key($themeData, array_flip([
							'settingsIconBackgroundColor',
							'settingsIconForegroundColor',
						])), $translator);
					}

					// Prüfe, ob das resultierende CSS gültig ist
					$this->checkThemeCssTemplate($theme);

					// Bilduploads verarbeiten
					$this->migrateFileUploads($domain, $theme, $request, $translator);
				} catch (SyntaxError $e) {
					$message = preg_replace('/"__string_template_[^"]*"/', '‹Custom CSS›', $e->getMessage());

					$this->addFlash(
						'danger',
						$translator->trans(
							'Variable syntax error in CSS: %msg%', [
								'%msg%' => $message,
							]
						)
					);
					return $this->render('theme/edit.html.twig', [
						'baseLayouts' => BaseLayout::all(),
						'user' => User::loggedInUser(),
						'locale' => LocaleService::findEnabledFuzzy($request->getLocale()),
						'theme' => $theme,
						'logoPath' => $logoUrl,
						'whitelabel' => $mayWhitelabel,
						'mayGeoIP' => $mayGeoIP,
						'allRestrictRegions' => ThemeModel::allRestrictRegions(),
						'poweredByText' => Utils::getPoweredByText(),
					]);
				} catch (RuntimeException $e) {
					return $this->render('theme/edit.html.twig', [
						'baseLayouts' => BaseLayout::all(),
						'user' => User::loggedInUser(),
						'locale' => LocaleService::findEnabledFuzzy($request->getLocale()),
						'theme' => $theme,
						'logoPath' => $logoUrl,
						'whitelabel' => $mayWhitelabel,
						'mayGeoIP' => $mayGeoIP,
						'allRestrictRegions' => ThemeModel::allRestrictRegions(),
						'poweredByText' => Utils::getPoweredByText(),
					]);
				}

				$theme->save();

				$this->addFlash(
					'success',
					$translator->trans(
						'The theme "%themename%" has been edited successfully!', [
							'%themename%' => $theme->getName(),
						]
					)
				);

				if ($action === 'saveClose') {
					return $this->redirectToRoute('app_theme_list', [], 303);
				} else {
					return $this->redirectToRoute('app_theme_edit', ['id' => $theme->getId()], 303);
				}

			case 'copy':
				/** @var ThemeModel $themeCopy */
				$themeCopy = ThemeModel::create()
					->importTheme($theme)
					->setSystemTheme(null)
					->setUserTheme(null)
					->setName("COPY: {$theme->getName()}")
					->setLogo($theme->getLogo())
					->setLogoDisplayed($theme->isLogoDisplayed())
					->setSettingsIcon($theme->getSettingsIcon())
					->setWhitelabel($theme->getWhitelabel())
				;

				$themeCopy->save();

				return $this->redirectToRoute('app_theme_list', [], 303);

			case 'setDefault':
				$theme->makeDefault();
				Domain::activeDomain()
					->touchFrontendWidgetCacheTimestamp()
					->save();

				return $this->redirectToRoute('app_theme_list', [], 303);

			default:
				return $this->redirectToRoute('app_theme_list');
		}
	}

	/**
	 * @Route("/{id}/delete", name="delete_form")
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deleteThemeForm(string $id, Request $request, TranslatorInterface $translator): Response
	{
		if (ThemeModel::exists($id) == false) {
			return $this->redirectToRoute('app_theme_list');
		}

		$theme = ThemeModel::find($id);
		$allThemes = ThemeModel::all();

		if ($theme->getSystemTheme()) {
			$this->addFlash('danger', $translator->trans('System themes cannot be deleted.'));
			return $this->redirectToRoute('app_theme_list', [], 303);
		} elseif ($theme->getUserTheme()) {
			$this->addFlash('danger', $translator->trans('Manage your global themes <a href=":url">here</a>.', [
				':url' => $this->generateUrl('app_user_theme_index'),
			]));
			return $this->redirectToRoute('app_theme_list', [], 303);
		}

		$domainSpecificThemeCount = array_reduce($allThemes, function ($count, ThemeModel $theme) {
			return $theme->getUserTheme() || $theme->getSystemTheme() ? $count : $count + 1;
		}, 0);

		if ($domainSpecificThemeCount == 1) {
			$this->addFlash('danger', $translator->trans('You cannot delete the last theme!'));
			return $this->redirectToRoute('app_theme_list');
		}

		return $this->render('theme/delete.html.twig', [
			'user' => User::loggedInUser(),
			'theme' => $theme
		]);
	}

	/**
	 * @Route("/{id}", name="delete", methods={"DELETE"})
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function deleteTheme(string $id, Request $request, TranslatorInterface $translator): Response
	{
		if (ThemeModel::exists($id) == false) {
			return $this->redirectToRoute('app_theme_list');
		}

		$theme = ThemeModel::find($id);
		$allThemes = ThemeModel::all();
		$themeName = $theme->getName();

		if ($theme->getSystemTheme() or $theme->getUserTheme()) {
			return $this->redirectToRoute('app_theme_list', [], 303);
		}
		$domainSpecificThemeCount = array_reduce($allThemes, function ($count, ThemeModel $theme) {
			return $theme->getUserTheme() || $theme->getSystemTheme() ? $count : $count + 1;
		}, 0);

		if ($domainSpecificThemeCount == 1) {
			$this->addFlash('danger', $translator->trans('You cannot delete the last theme!'));
			return $this->redirectToRoute('app_theme_list');
		}

		// wenn das Theme als Standard registriert ist, dann das erstbeste Theme als Standard setzen
		if ($theme->isDefaultTheme()) {
			/** @var ThemeModel[] $allThemes */
			foreach ($allThemes as $themeId => $themeObject) {
				if ($id !== $themeId) {
					$themeObject
						->touchLastModifiedTimestamp()
						->makeDefault();
					$this->addFlash(
						'info',
						$translator->trans('The theme "%themename%" has now been set as default.', [
							'%themename%' => $themeObject->getName()
						])
					);
					break;
				}
			}
		}

		$theme->setLogo(null);
		$theme->setSettingsIcon(null);
		$theme->setIsDeleted(true);
		$theme->flush();

		$this->addFlash('success', $translator->trans(
			'The theme "%themename%" has been deleted successfully!', [
				'%themename%' => $themeName
			]
		));
		return $this->redirectToRoute('app_theme_list');
	}

	/**
	 * @param ThemeModel $theme
	 * @param array<string,mixed> $themeData
	 * @param File|null $newLogo
	 * @param bool $addToChangelog
	 * @return void
	 */
	private function setThemeObjectAttributes(ThemeModel $theme, array $themeData, ?File $newLogo, bool $addToChangelog=true): void
	{
		// Bool Defaults
		foreach ([
				'logoDisplayed', 'whitelabel', 'showPurposesInMainWindow', 'blocking', 'showAcceptAllButtonInControlPanel',
				'showEmbeddingsInControlPanel',
				'showWithdrawAllButtonInControlPanel',
				'showCheckAndUncheckAllButtonInControlPanel', 'settingsIconEnabled', 'complyingDoNotTrack', 'onlyInEu',
				'manipulationPrevention', 'showExtraImprintFooter', 'saveButtonInDetailsWindow', 'showCloseButton',
				'customCssTemplated', 'customCssForIframeBlockerTemplated',
				] as $key) {
			$themeData[$key] = (bool)($themeData[$key] ?? false);
		}
		// String Defaults
		foreach ([
				'name', 'position', 'settingsIconTarget', 'selectType', 'okButtonBackgroundColor',
				'moreInfoButtonBackgroundColor', 'windowBackgroundColor', 'windowBorderColor',
				'linkInInfotextColor', 'okButtonColor', 'moreInfoButtonColor', 'okButtonBorderColor',
				'moreInfoButtonBorderColor', 'declineButtonBackgroundColor', 'declineButtonBorderColor',
				'declineButtonTextColor', 'infotextColor', 'infotextFontFamily', 'infotextFontSize',
				'iframeBlockerBackgroundColor', 'iframeBlockerForegroundColor', 'iframeBlockerButtonBackgroundColor',
				'iframeBlockerButtonForegroundColor', 'customCss', 'customCssForIframeBlocker',
				'settingsIconBackgroundColor', 'settingsIconForegroundColor',
				'baseLayout',
				] as $key) {
			$themeData[$key] = trim((string)($themeData[$key] ?? ''));
		}
		$themeData['bannerHeightMobile'] = (int)($themeData['bannerHeightMobile'] ?? 100);

		$changes = [];
		$theme->touchLastModifiedTimestamp();

		if ($theme->isLogoDisplayed() !== $themeData['logoDisplayed']) {
			$changes['logoDisplayed'] = ['old' => $theme->isLogoDisplayed(), 'new' => $themeData['logoDisplayed']];
			$theme->setLogoDisplayed((bool)($themeData['logoDisplayed'] ?? false));
		}
		if (strlen($themeData['settingsIconBackgroundColor']) > 0 && $theme->getSettingsIconBackgroundColor() !== $themeData['settingsIconBackgroundColor']) {
			$changes['settingsIconBackgroundColor'] = ['old' => $theme->getSettingsIconBackgroundColor(), 'new' => $themeData['settingsIconBackgroundColor']];
			$theme->setSettingsIconBackgroundColor($themeData['settingsIconBackgroundColor']);
		}
		if (strlen($themeData['settingsIconForegroundColor']) > 0 && $theme->getSettingsIconForegroundColor() !== $themeData['settingsIconForegroundColor']) {
			$changes['settingsIconForegroundColor'] = ['old' => $theme->getSettingsIconForegroundColor(), 'new' => $themeData['settingsIconForegroundColor']];
			$theme->setSettingsIconForegroundColor($themeData['settingsIconForegroundColor']);
		}
		if ($theme->getWhitelabel() !== $themeData['whitelabel']) {
			$changes['whitelabel'] = ['old' => $theme->getWhitelabel(), 'new' => $themeData['whitelabel']];
			$theme->setWhitelabel((bool)($themeData['whitelabel'] ?? false));
		}

		if ($newLogo) {
			$oldLogoIdentifier = $theme->getLogoIdentifier();
			$theme->setLogo($newLogo);
			$changes['logo'] = ['old' => $oldLogoIdentifier, 'new' => $theme->getLogoIdentifier()];
		}

		// Parameter nur aktualisieren, wenn kein System- oder User-Theme referenziert wird.
		if (!$theme->getSystemTheme() && !$theme->getUserTheme()) {
			if ($theme->getName() !== $themeData['name']) {
				$changes['name'] = ['old' => $theme->getName(), 'new' => $themeData['name']];
				$theme->setName($themeData['name']);
			}
			if (
				$baseLayout = BaseLayout::find($themeData['baseLayout'])
				and $theme->getBaseLayout()->getId() != $baseLayout->getId()
			) {
				$changes['baseLayout'] = ['old' => $theme->getBaseLayout()->getId(), 'new' => $baseLayout->getId()];
				$theme->setBaseLayout($baseLayout);
			}
			if ($theme->getPosition() !== $themeData['position']) {
				$changes['position'] = ['old' => $theme->getPosition(), 'new' => $themeData['position']];
				$theme->setPosition($themeData['position']);
			}
			if ($theme->isShowPurposesInMainWindow() !== $themeData['showPurposesInMainWindow']) {
				$changes['showPurposesInMainWindow'] = ['old' => $theme->isShowPurposesInMainWindow(), 'new' => $themeData['showPurposesInMainWindow']];
				$theme->setShowPurposesInMainWindow($themeData['showPurposesInMainWindow']);
			}
			if ($theme->isBlocking() !== $themeData['blocking']) {
				$changes['blocking'] = ['old' => $theme->isBlocking(), 'new' => $themeData['blocking']];
				$theme->setBlocking($themeData['blocking']);
			}
			if ($theme->isShowEmbeddingsInControlPanel() !== $themeData['showEmbeddingsInControlPanel']) {
				$changes['showEmbeddingsInControlPanel'] = ['old' => $theme->isShowEmbeddingsInControlPanel(), 'new' => $themeData['showEmbeddingsInControlPanel']];
				$theme->setShowEmbeddingsInControlPanel($themeData['showEmbeddingsInControlPanel']);
			}
			if ($theme->isShowAcceptAllButtonInControlPanel() !== $themeData['showAcceptAllButtonInControlPanel']) {
				$changes['showAcceptAllButtonInControlPanel'] = ['old' => $theme->isShowAcceptAllButtonInControlPanel(), 'new' => $themeData['showAcceptAllButtonInControlPanel']];
				$theme->setShowAcceptAllButtonInControlPanel($themeData['showAcceptAllButtonInControlPanel']);
			}
			if ($theme->isShowWithdrawAllButtonInControlPanel() !== $themeData['showWithdrawAllButtonInControlPanel']) {
				$changes['showWithdrawAllButtonInControlPanel'] = ['old' => $theme->isShowWithdrawAllButtonInControlPanel(), 'new' => $themeData['showWithdrawAllButtonInControlPanel']];
				$theme->setShowWithdrawAllButtonInControlPanel($themeData['showWithdrawAllButtonInControlPanel']);
			}
			if ($theme->isShowCheckAndUncheckAllButtonInControlPanel() !== $themeData['showCheckAndUncheckAllButtonInControlPanel']) {
				$changes['showCheckAndUncheckAllButtonInControlPanel'] = ['old' => $theme->isShowCheckAndUncheckAllButtonInControlPanel(), 'new' => $themeData['showCheckAndUncheckAllButtonInControlPanel']];
				$theme->setShowCheckAndUncheckAllButtonInControlPanel($themeData['showCheckAndUncheckAllButtonInControlPanel']);
			}
			if ($theme->isSettingsIconEnabled() !== $themeData['settingsIconEnabled']) {
				$changes['settingsIconEnabled'] = ['old' => $theme->isSettingsIconEnabled(), 'new' => $themeData['settingsIconEnabled']];
				$theme->setSettingsIconEnabled($themeData['settingsIconEnabled']);
			}
			if ($theme->getSettingsIconTarget() !== $themeData['settingsIconTarget']) {
				$changes['settingsIconTarget'] = ['old' => $theme->getSettingsIconTarget(), 'new' => $themeData['settingsIconTarget']];
				$theme->setSettingsIconTarget($themeData['settingsIconTarget']);
			}
			if ($theme->getWidgetButtons() !== $themeData['widgetButtons']) {
				$changes['widgetButtons'] = ['old' => $theme->getWidgetButtons(), 'new' => $themeData['widgetButtons']];
				$theme->setWidgetButtons($themeData['widgetButtons']);
			}
			if ($theme->getSelectType() !== $themeData['selectType']) {
				$changes['selectType'] = ['old' => $theme->getSelectType(), 'new' => $themeData['selectType']];
				$theme->setSelectType($themeData['selectType']);
			}
			if ($theme->isComplyingDoNotTrack() !== $themeData['complyingDoNotTrack']) {
				$changes['complyingDoNotTrack'] = ['old' => $theme->isComplyingDoNotTrack(), 'new' => $themeData['complyingDoNotTrack']];
				$theme->setComplyingDoNotTrack($themeData['complyingDoNotTrack']);
			}
			if ($theme->getOkButtonBackgroundColor() !== $themeData['okButtonBackgroundColor']) {
				$changes['okButtonBackgroundColor'] = ['old' => $theme->getOkButtonBackgroundColor(), 'new' => $themeData['okButtonBackgroundColor']];
				$theme->setOkButtonBackgroundColor($themeData['okButtonBackgroundColor']);
			}
			if ($theme->getMoreInfoButtonBackgroundColor() !== $themeData['moreInfoButtonBackgroundColor']) {
				$changes['moreInfoButtonBackgroundColor'] = ['old' => $theme->getMoreInfoButtonBackgroundColor(), 'new' => $themeData['moreInfoButtonBackgroundColor']];
				$theme->setMoreInfoButtonBackgroundColor($themeData['moreInfoButtonBackgroundColor']);
			}
			if ($theme->getWindowBackgroundColor() !== $themeData['windowBackgroundColor']) {
				$changes['windowBackgroundColor'] = ['old' => $theme->getWindowBackgroundColor(), 'new' => $themeData['windowBackgroundColor']];
				$theme->setWindowBackgroundColor($themeData['windowBackgroundColor']);
			}
			if ($theme->getWindowBorderColor() !== $themeData['windowBorderColor']) {
				$changes['windowBorderColor'] = ['old' => $theme->getWindowBorderColor(), 'new' => $themeData['windowBorderColor']];
				$theme->setWindowBorderColor($themeData['windowBorderColor']);
			}
			if ($theme->getLinkInInfotextColor() !== $themeData['linkInInfotextColor']) {
				$changes['linkInInfotextColor'] = ['old' => $theme->getLinkInInfotextColor(), 'new' => $themeData['linkInInfotextColor']];
				$theme->setLinkInInfotextColor($themeData['linkInInfotextColor']);
			}
			if ($theme->getOkButtonColor() !== $themeData['okButtonColor']) {
				$changes['okButtonColor'] = ['old' => $theme->getOkButtonColor(), 'new' => $themeData['okButtonColor']];
				$theme->setOkButtonColor($themeData['okButtonColor']);
			}
			if ($theme->getMoreInfoButtonColor() !== $themeData['moreInfoButtonColor']) {
				$changes['moreInfoButtonColor'] = ['old' => $theme->getMoreInfoButtonColor(), 'new' => $themeData['moreInfoButtonColor']];
				$theme->setMoreInfoButtonColor($themeData['moreInfoButtonColor']);
			}
			if ($theme->getOkButtonBorderColor() !== $themeData['okButtonBorderColor']) {
				$changes['okButtonBorderColor'] = ['old' => $theme->getOkButtonBorderColor(), 'new' => $themeData['okButtonBorderColor']];
				$theme->setOkButtonBorderColor($themeData['okButtonBorderColor']);
			}
			if ($theme->getMoreInfoButtonBorderColor() !== $themeData['moreInfoButtonBorderColor']) {
				$changes['moreInfoButtonBorderColor'] = ['old' => $theme->getMoreInfoButtonBorderColor(), 'new' => $themeData['moreInfoButtonBorderColor']];
				$theme->setMoreInfoButtonBorderColor($themeData['moreInfoButtonBorderColor']);
			}
			if ($theme->getDeclineButtonBackgroundColor() !== $themeData['declineButtonBackgroundColor']) {
				$changes['declineButtonBackgroundColor'] = ['old' => $theme->getDeclineButtonBackgroundColor(), 'new' => $themeData['declineButtonBackgroundColor']];
				$theme->setDeclineButtonBackgroundColor($themeData['declineButtonBackgroundColor']);
			}
			if ($theme->getDeclineButtonBorderColor() !== $themeData['declineButtonBorderColor']) {
				$changes['declineButtonBorderColor'] = ['old' => $theme->getDeclineButtonBorderColor(), 'new' => $themeData['declineButtonBorderColor']];
				$theme->setDeclineButtonBorderColor($themeData['declineButtonBorderColor']);
			}
			if ($theme->getDeclineButtonTextColor() !== $themeData['declineButtonTextColor']) {
				$changes['declineButtonTextColor'] = ['old' => $theme->getDeclineButtonTextColor(), 'new' => $themeData['declineButtonTextColor']];
				$theme->setDeclineButtonTextColor($themeData['declineButtonTextColor']);
			}
			if ($theme->getInfotextColor() !== $themeData['infotextColor']) {
				$changes['infotextColor'] = ['old' => $theme->getInfotextColor(), 'new' => $themeData['infotextColor']];
				$theme->setInfotextColor($themeData['infotextColor']);
			}
			if ($theme->getInfotextFontFamily() !== $themeData['infotextFontFamily']) {
				$changes['infotextFontFamily'] = ['old' => $theme->getInfotextFontFamily(), 'new' => $themeData['infotextFontFamily']];
				$theme->setInfotextFontFamily($themeData['infotextFontFamily']);
			}
			if ($theme->getInfotextFontSize() !== $themeData['infotextFontSize']) {
				$changes['infotextFontSize'] = ['old' => $theme->getInfotextFontSize(), 'new' => $themeData['infotextFontSize']];
				$theme->setInfotextFontSize($themeData['infotextFontSize']);
			}
			if ($theme->getIframeBlockerBackgroundColor() !== $themeData['iframeBlockerBackgroundColor']) {
				$changes['iframeBlockerBackgroundColor'] = ['old' => $theme->getIframeBlockerBackgroundColor(), 'new' => $themeData['iframeBlockerBackgroundColor']];
				$theme->setIframeBlockerBackgroundColor($themeData['iframeBlockerBackgroundColor']);
			}
			if ($theme->getIframeBlockerForegroundColor() !== $themeData['iframeBlockerForegroundColor']) {
				$changes['iframeBlockerForegroundColor'] = ['old' => $theme->getIframeBlockerForegroundColor(), 'new' => $themeData['iframeBlockerForegroundColor']];
				$theme->setIframeBlockerForegroundColor($themeData['iframeBlockerForegroundColor']);
			}
			if ($theme->getIframeBlockerButtonBackgroundColor() !== $themeData['iframeBlockerButtonBackgroundColor']) {
				$changes['iframeBlockerButtonBackgroundColor'] = ['old' => $theme->getIframeBlockerButtonBackgroundColor(), 'new' => $themeData['iframeBlockerButtonBackgroundColor']];
				$theme->setIframeBlockerButtonBackgroundColor($themeData['iframeBlockerButtonBackgroundColor']);
			}
			if ($theme->getIframeBlockerButtonForegroundColor() !== $themeData['iframeBlockerButtonForegroundColor']) {
				$changes['iframeBlockerButtonForegroundColor'] = ['old' => $theme->getIframeBlockerButtonForegroundColor(), 'new' => $themeData['iframeBlockerButtonForegroundColor']];
				$theme->setIframeBlockerButtonForegroundColor($themeData['iframeBlockerButtonForegroundColor']);
			}
			if ($theme->getCustomCss() !== $themeData['customCss']) {
				$changes['customCss'] = ['old' => $theme->getCustomCss(), 'new' => $themeData['customCss']];
				$theme->setCustomCss($themeData['customCss']);
			}
			if ($theme->isCustomCssTemplated() !== $themeData['customCssTemplated']) {
				$changes['customCssTemplated'] = ['old' => $theme->isCustomCssTemplated(), 'new' => $themeData['customCssTemplated']];
				$theme->setCustomCssTemplated($themeData['customCssTemplated']);
			}
			if ($theme->getCustomCssForIframeBlocker() !== $themeData['customCssForIframeBlocker']) {
				$changes['customCssForIframeBlocker'] = ['old' => $theme->getCustomCssForIframeBlocker(), 'new' => $themeData['customCssForIframeBlocker']];
				$theme->setCustomCssForIframeBlocker($themeData['customCssForIframeBlocker']);
			}
			if ($theme->isCustomCssForIframeBlockerTemplated() !== $themeData['customCssForIframeBlockerTemplated']) {
				$changes['customCssForIframeBlockerTemplated'] = ['old' => $theme->isCustomCssForIframeBlockerTemplated(), 'new' => $themeData['customCssForIframeBlockerTemplated']];
				$theme->setCustomCssForIframeBlockerTemplated($themeData['customCssForIframeBlockerTemplated']);
			}
			if ($theme->getOnlyInEu() !== $themeData['onlyInEu']) {
				$changes['onlyInEu'] = ['old' => $theme->getOnlyInEu(), 'new' => $themeData['onlyInEu']];
				$theme->setOnlyInEu($themeData['onlyInEu']);
			}
			if ($theme->getManipulationPrevention() !== $themeData['manipulationPrevention']) {
				$changes['manipulationPrevention'] = ['old' => $theme->getManipulationPrevention(), 'new' => $themeData['manipulationPrevention']];
				$theme->setManipulationPrevention($themeData['manipulationPrevention']);
			}
			if ($theme->isShowExtraImprintFooter() !== $themeData['showExtraImprintFooter']) {
				$changes['showExtraImprintFooter'] = ['old' => $theme->isShowExtraImprintFooter(), 'new' => $themeData['showExtraImprintFooter']];
				$theme->setShowExtraImprintFooter($themeData['showExtraImprintFooter']);
			}
			if ($theme->hasSaveButtonInDetailsWindow() !== $themeData['saveButtonInDetailsWindow']) {
				$changes['saveButtonInDetailsWindow'] = ['old' => $theme->hasSaveButtonInDetailsWindow(), 'new' => $themeData['saveButtonInDetailsWindow']];
				$theme->setSaveButtonInDetailsWindow($themeData['saveButtonInDetailsWindow']);
			}
			if ($theme->hasCloseButton() !== $themeData['showCloseButton']) {
				$changes['showCloseButton'] = ['old' => $theme->hasCloseButton(), 'new' => $themeData['showCloseButton']];
				$theme->setCloseButton($themeData['showCloseButton']);
			}
			if (!in_array("EU", $themeData['declineButtonRestrictions'])) {
				array_unshift($themeData['declineButtonRestrictions'], 'EU');
			}
			if ($theme->getDeclineButtonRestrictions() !== $themeData['declineButtonRestrictions']) {
				$changes['declineButtonRestrictions'] = ['old' => $theme->getDeclineButtonRestrictions(), 'new' => $themeData['declineButtonRestrictions']];
				$theme->setDeclineButtonRestrictions($themeData['declineButtonRestrictions'] ?? []);
			}

			if ($theme->getBannerHeightMobile() !== $themeData['bannerHeightMobile']) {
				$changes['bannerHeightMobile'] = ['old' => $theme->getBannerHeightMobile(), 'new' => $themeData['bannerHeightMobile']];
				$theme->setBannerHeightMobile($themeData['bannerHeightMobile']);
			}
		}

		if ($addToChangelog and $changes) {
			// Update MongoDB property type from array to object
			if (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB) {
				$changelog = $theme->get('changeLog', []);
				$isArray = array_reduce(array_keys($changelog), function (bool $isArray, $key) {
					return $isArray && is_int($key);
				}, true);

				if ($isArray) {
					$changelog['ts'.time()] = $changes;
					$theme->set('changeLog', $changelog);
					return;
				}
			}

			$theme->set('changeLog.ts'.(string)((int)time()), $changes);
		}
	}

	/**
	 * @param Domain $domain
	 * @param ThemeModel $theme
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 */
	private function migrateFileUploads(Domain $domain, ThemeModel $theme, Request $request, TranslatorInterface $translator): void
	{
		$fileUploads = [
			[
				'fieldName' => 'image',
				'setter' => [$theme, 'setLogo'],
			], [
				'fieldName' => 'settings_icon',
				'setter' => [$theme, 'setSettingsIcon'],
			],
		];

		foreach ($fileUploads as $uploadInfo) {
			/** @var UploadedFile $file */
			$file = $request->files->get($uploadInfo['fieldName']);
			$shouldRemove = $request->request->getBoolean($uploadInfo['fieldName'].'__delete');

			if (!$file) {
				if ($shouldRemove) {
					$uploadInfo['setter'](null);
				}
				continue;
			}

			$this->handleUploadErrors($file, $translator);

			$uploadInfo['setter']($file);
		}
	}

	/**
	 * @param UploadedFile $image
	 * @param TranslatorInterface $translator
	 * @return void
	 */
	private function handleUploadErrors(UploadedFile $image, TranslatorInterface $translator): void
	{
		if ($image->getError() != UPLOAD_ERR_OK) {
			$this->addFlash(
				'danger',
				$translator->trans('The image could not be uploaded! (Error code %errorcode% %errormessage%)', [
					'%errorcode%' => $image->getError(),
					'%errormessage%' => $image->getErrorMessage()
				])
			);
			throw new RuntimeException();
		}

		if (preg_match('/^image\/.*$/', $image->getMimeType()) == false) {
			$this->addFlash('danger', $translator->trans('The uploaded file is not an image!'));
			throw new RuntimeException();
		}
	}

	/**
	 * @param array<string,mixed> $themeData
	 * @param TranslatorInterface $translator
	 * @return void
	 */
	private function handleColorErrors(array $themeData, TranslatorInterface $translator): void
	{
		foreach ($themeData as $name => $value) {
			if (preg_match('~Color$~', $name) == false || Color::validateCssString($value)) {
				continue;
			}

			$this->addFlash('danger', $translator->trans('Invalid color values! Please use the color picker and/or provide hex color codes.'));
			throw new RuntimeException();
		}
	}

	/**
	 * @param string $name
	 * @param TranslatorInterface $translator
	 * @return void
	 * @throws RuntimeException
	 */
	private function handleNameError($name, TranslatorInterface $translator): void
	{
		if (trim($name) === '') {
			$this->addFlash('warning', $translator->trans('A theme cannot be saved without a name!'));
			throw new RuntimeException();
		}
	}

	/**
	 * @param string $position
	 * @param TranslatorInterface $translator
	 * @return void
	 * @throws RuntimeException
	 */
	private function handlePositionError($position, TranslatorInterface $translator): void
	{
		if (in_array($position, ['top', 'center', 'bottom','bottomleft','bottomright',]) == false) {
			$this->addFlash('danger', $translator->trans('The value for field "%field%" is invalid!', ['%field%' => 'Position']));
			throw new RuntimeException();
		}
	}
}
