<?php

namespace App\Controller;

use App\Component\Menu\Annotation\Menu;
use App\Component\Pagination;
use App\Exception\ConfigNotWritableException;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\EmbeddingBase;
use App\Model\Language;
use App\Model\Model;
use App\Model\Purpose;
use App\Model\Regulations\Article;
use App\Model\Regulations\Regulation;
use App\Model\Regulations\Group as RegulationGroup;
use App\Model\SystemEmbedding;
use App\Model\UserEmbedding;
use App\Model\UserEmbeddingAsset;
use App\Model\UserLocale;
use App\Model\User;
use App\Config;
use App\Component\EmbeddingDb\EmbeddingDb;
use App\Component\EmbeddingDb\EmbeddingDbLegacy;
use App\Component\LocaleService;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Utils;
use Exception;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;
use DateTimeImmutable;
use DateTimeInterface;

/**
 * Controller zur Admin-Embedding-Verwaltung
 * @Route("/user_embeddings", name="app_user_embeddings_")
 */
class UserEmbeddings extends AbstractController
{
	/**
	 * @Menu("Embedding templates", group="Settings", icon="fa-certificate", order=50)
	 * @Route("", name="list", methods={"GET", "HEAD"})
	 *
	 * @param Request $request
	 * @throws \Exception Wenn kein Zugriff auf das Embedding-File erfolgen kann
	 *
	 * @return Response
	 */
	public function listEmbeddings(Request $request): Response
	{
		$user = User::loggedInUser();
		$locales = $user->getUserEmbeddingLocales();
		$embeddings = UserEmbedding::all();
		$purposes = Purpose::all();

		return $this->render('user_embedding/list.html.twig', [
			'embeddings' => $embeddings,
			'purposes' => $purposes,
			'selectedLocales' => $locales,
			'allLocales' => UserLocale::all(),
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/list.json", name="list_ajax")
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return JsonResponse
	 */
	public function ajaxEmbeddingSearch(Request $request, TranslatorInterface $translator): JsonResponse
	{
		$currentLanguageCode = explode('_', $request->getLocale(), 2)[0];
		$defaultLocale = LocaleService::getDefault();
		$defaultLanguage = $defaultLocale->getLanguage() ?? Language::find('de_DE');
		$backendLocale = LocaleService::findFuzzy($request->getLocale());
		$backendLanguage = $backendLocale ? $backendLocale->getLanguage() : null;
		$language = $backendLanguage ?? Language::find($currentLanguageCode) ?? $defaultLanguage;

		$search = trim((string)($request->query->all('search')['value'] ?? ''));

		// Lade alle Benutzervorlagen
		$userTemplateNames = [];
		$userTemplates = UserEmbedding::all();
		foreach ($userTemplates as $userTemplate) {
			$userTemplateNames[$userTemplate->getName()] = $userTemplate->getName();
		}

		// Lade alle Systemweite Einbindungsvorlagen
		$templates = array_filter(SystemEmbedding::all(), function (SystemEmbedding $e) use ($userTemplateNames) {
			return !isset($userTemplateNames[$e->getName()]);
		});
		// Mappe Daten der Systemweiten Embeddings
		$templates = array_values(array_filter(array_map(function ($embedding) use ($language, $translator) {
			$assets = [];
			foreach ($embedding->getAssets() as $asset) {
				$assets[] = $asset->getName();
			}
			return [
				'_score' => null,
				'id' => $embedding->getId(),
				'name' => $embedding->getTranslatedName($language),
				'source' => EmbeddingBase::SOURCE_SYSTEM_TEMPLATE,
				'embedding_vendor' => $embedding->getEmbeddingVendor(),
				'description' => $embedding->getTranslatedDescription($language),
				'purpose' => $embedding->getPurpose()->getName(),
				'assets' => $assets,
				'category' => $translator->trans('System-wide embedding templates')
			];
		}, $templates)));

		// Globale DB abfragen, wenn aktiviert (Einbindungs-Daten aus externer Datenbank)
		if (Config::getInstance()->isGlobalCookieDbUsed()) {
			// FIXME Total record count is unknown when searching
			$embeddingsDB = $this->searchGlobalCookieDb(strlen($search) ? $search : null, $language->getShortCode());
		} else {
			$embeddingsDB = [];
		}
		// Mappe Daten der Embeddings aus externer DB
		$embeddingsDB = array_values(array_filter(array_map(function ($embedding) use ($translator) {
			$assets = [];
			foreach ($embedding['assets'] as $asset) {
				$assets[] = $asset['name'];
			}
			return [
				'_score' => $embedding['@score'] ?? null,
				'id' => $embedding['id'],
				'name' => $embedding['translations']['current']['label'],
				'source' => EmbeddingBase::SOURCE_GLOBAL_DB,
				'embedding_vendor' => $embedding['embedding_vendor'],
				'description' => $embedding['translations']['current']['description'],
				'purpose' => Purpose::find($embedding['purpose'])->getName(),
				'assets' => $assets,
				'category' => $translator->trans('Embedding Data from remote Database')
			];
		}, $embeddingsDB)));

		// Sortierung der Embeddings
		$order = $request->query->all('order')[0] ?? null;
		$sortColumn = $order['column'] ?? '';
		$sortDirection = $order['dir'] ?? 'asc';

		$templates = Utils::sortNatNum($templates, $sortColumn, $sortDirection);
		$embeddingsDB = Utils::sortNatNum($embeddingsDB, $sortColumn, $sortDirection);

		$visibleEmbeddings = array_merge($templates, $embeddingsDB);

		// Filtere bereits bestehende Embeddings aus angezeigten Embeddings heraus
		foreach ($userTemplates as $activeEmbedding) {
			foreach ($visibleEmbeddings as $key => $embeddingTemplate) {
				if ($embeddingTemplate['id'] == $activeEmbedding->getSourceId()) {
					unset($visibleEmbeddings[$key]);
				}
			}
		}

		// Filterung nach Suchbegriff
		$embeddingsFiltered = strlen($search) > 0 ? self::searchInArray($search, $visibleEmbeddings) : $visibleEmbeddings;

		// Paginierung der Einträge
		$start = $request->query->getInt('start');
		$length = $request->query->getInt('length') ?: 10;

		$pagination = new Pagination($embeddingsFiltered, $length);
		$currentPage = ($start / $length) + 1;
		$pagination->setCurrentPage($currentPage);

		$embeddingsPaginated = $pagination->getItems();

		return $this->json([
			'data' => array_values($embeddingsPaginated),
			'draw' => $request->query->get('draw'),
			'recordsTotal' => count($visibleEmbeddings),
			'recordsFiltered' => count($embeddingsFiltered),
		]);
	}

	/**
	 * @Route("", name="save_settings", methods={"POST"})
	 *
	 * @param Request $request
	 * @throws \Exception Wenn kein Zugriff auf das Config-File erfolgen kann
	 *
	 * @return Response
	 */
	public function saveSettings(Request $request): Response
	{
		try {
			CsrfToken::get()->checkRequest($request);
			$user = User::loggedInUser();

			$user->setUserEmbeddingLocales(array_filter(array_map(function ($code) {
				return UserLocale::find($code);
			}, (array)$request->request->get('selectedLanguages'))));
			$user->save();

			$this->addFlash('success', 'Settings have been saved.');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		return $this->redirectToRoute('app_user_embeddings_list', [], 303);
	}

	/**
	 * Embeddings aus der CCM19.de Datenbank anzeigen zur Auswahl...
	 * @Route("/db", name="from_db", methods={"HEAD", "GET"} )
	 * @return Response
	 *
	 * @todo überarbeiten
	 */
	public function listEmbeddingsFromDatabase(Request $request): Response
	{
		return $this->render('user_embedding/list_db.html.twig', [
		]);
	}

	/**
	 * Einbindungsvorlagen von Kunden oder Einbindungen von Domains als
	 * systemweite Einbindungsvorlage übernehmen.
	 *
	 * @Route("/domain", name="from_domain", methods={"HEAD", "GET"})
	 *
	 * @return Response
	 */
	public function listEmbeddingsFromDomain(): Response
	{
		return $this->render('user_embedding/list_domain.html.twig');
	}

	/**
	 * Domains anzeigen, die mindestens eine Einbindung haben.
	 *
	 * @Route("/client/domains.json", name="from_domain_ajax", methods={"HEAD", "GET"})
	 *
	 * @return JsonResponse
	 */
	public function ajaxDomains(): JsonResponse
	{
		$tableData = [];

		foreach (Domain::all() as $domain) {
			Domain::setDomainForRequest($domain->getId());
			Model::closeDomainModels();

			$tableData[$domain->getId()] = [
				'domain' => [
					'id' => $domain->getId(),
					'host' => (string)parse_url($domain->getUrl(), PHP_URL_HOST),
					'path' => $domain->getUrlPath(),
				],
				'embeddings' => array_values(array_map(function (Embedding $embedding) {
					return [
						'id' => $embedding->getId(),
						'name' => $embedding->getName(),
					];
				}, Embedding::all())),
			];
		}

		// Reset context
		Domain::setDomainForRequest('');
		Model::closeDomainModels();

		// Select only domains that have embeddings
		$tableData = array_filter($tableData, function (array $row) {
			return count($row['embeddings']) > 0;
		});

		return $this->json([
			'data' => array_values($tableData),
		]);
	}

	/**
	 * Diese Methode filtert die Embeddings nach dem Suchbegriff.
	 * @param string $search Suchbegriff
	 * @param array $embeddings Embeddings, die gefiltert werden sollen
	 * @return array
	 */
	public static function searchInArray(string $search, array $embeddings): array
	{
		return array_filter($embeddings, function (array $embedding) use ($search) {
			return stristr($embedding['name'], $search)
				|| 	stristr($embedding['embedding_vendor'] ,$search)
				|| 	stristr($embedding['description'] ,$search)
				|| 	stristr($embedding['purpose'] ,$search)
				|| array_reduce($embedding['assets'], function (bool $match, string $name) use ($search) {
					return $match || stripos($name, $search) !== false;
				}, false)
				;
		});
	}

	private function searchGlobalCookieDb(?string $search, string $languageCode, bool $filterSystemEmbeddings=true): array
	{
		$defaultLocale = LocaleService::getDefault();
		$defaultLanguage = $defaultLocale->getLanguage() ?? Language::find('de_DE');
		$defaultLanguageCode = $defaultLanguage->getShortCode();

		$embeddingsDB = EmbeddingDb::fetchEmbeddingList($search, Language::find($languageCode) ?? $defaultLanguage);

		if ($filterSystemEmbeddings) {
			$cookieMap = [];
			$templateNames = [];

			$templates = ($search == null) ? SystemEmbedding::all() : SystemEmbedding::search($search);
			foreach ($templates as $template) {
				$templateNames[$template->getName()] = $template->getName();
				foreach ($template->getAssets() as $asset) {
					$cookieMap[$asset->getName()] = $template;
				}
			}

			// Einträge ausfiltern, die schon einen Eintrag in den System-/UserEmbeddings haben
			$embeddingsDB = array_filter($embeddingsDB ?? [], function ($item) use ($cookieMap, $templateNames) {
				if (isset($templateNames[$item['name']])) {
					return false;
				}

				foreach ($item['assets'] as $asset) {
					if (!isset($asset['name'])) {
						var_dump($asset);exit();
					}
					if (isset($cookieMap[$asset['name']])) {
						return false;
					}
				}
				return true;
			});
		}

		// Anzeige-Übersetzung auswählen nach aktueller Backend-Sprache, Standard-Locale und Fallback DE.
		foreach ($embeddingsDB as &$embedding) {
			$embedding['translations']['current'] = ($embedding['translations'][$languageCode] ?? [])
				+ ($embedding['translations'][$defaultLanguageCode] ?? [])
				+ $embedding['translations']['de'];
		}
		unset($embedding);

		return $embeddingsDB;
	}

	/**
	 * Entpackt mit `data-pack="asset,add_asset"` als JSON gepackte POST-Daten
	 */
	private function unpackPostData(array $array): array
	{
		foreach (['asset', 'add_asset'] as $key) {
			if (isset($array[$key]) and is_string($array[$key])) {
				$array[$key] = json_decode($array[$key], true, 64, JSON_INVALID_UTF8_IGNORE);
			}
		}
		return $array;
	}

	/**
	 * Daten aus dem Array $embeddingData in $embedding übertragen
	 */
	private function applyChanges(UserEmbedding $embedding, array $embeddingData, string $defaultLanguageId): UserEmbedding
	{
		$request = Utils::getCurrentRequest();
		$user = User::loggedInUser();

		// Script-Blocking
		$codeBlock = explode("\n", trim((string)($embeddingData['code_block'] ?? '')));
		$codeBlock = array_filter($codeBlock, function ($line) {
			return trim($line) !== '';
		});
		$codeBlock = array_values(array_map(function ($line) {
			return trim($line);
		}, $codeBlock));

		// Iframes blockieren
		$iframeBlockingStrings = array_filter(array_map(function ($line) {
			return trim($line);
		}, preg_split('~\r\n|\n|\r~', (string)($embeddingData['iframe_blocking_strings'] ?? ''))));

		// Exclude Urls
		$excludeUrls = explode("\n", trim((string)($embeddingData['excludeUrls'] ?? '')));
		$excludeUrls = array_filter($excludeUrls, function ($line) {
			return trim($line) !== '';
		});
		$excludeUrls = array_values(array_map(function ($line) {
			return trim($line);
		}, $excludeUrls));
		// Zweck
		$purpose = Purpose::find($embeddingData['purpose'] ?? '');

		if ($request) {
			$dispatchLoadEventForIframes = $request->isMethodCacheable()
				// GET: Wenn der Parameter nicht vorhanden ist, dann soll der Standardwert verwendet werden
				? (bool)($embeddingData['dispatchLoadEventForIframes'] ?? $embedding->isDispatchLoadEventForIframes())
				// POST/PUT: Wenn der Parameter nicht vorhanden ist, wurde der Schalter im Formular deaktiviert
				: (bool)($embeddingData['dispatchLoadEventForIframes'] ?? false);
		}
		else {
			// Sollte nicht auftreten; Standardwert verwenden
			$dispatchLoadEventForIframes = $embedding->isDispatchLoadEventForIframes();
		}

		// Allgemeines speichern
		$embedding
			->setName($embeddingData['name'] ?? '')
			->setDescription($embeddingData['description'][$defaultLanguageId] ?? '')
			->setCodeBlock($codeBlock)
			->setScriptLoaderGroup($embeddingData['scriptLoaderGroup'] ?? '')
			->setIframeBlockingStrings($iframeBlockingStrings)
			->setIframesHiddenUntilConsent((bool)($embeddingData['hideIframesUntilConsent'] ?? false))
			->setDispatchLoadEventForIframes($dispatchLoadEventForIframes)
			->setExcludeUrls($excludeUrls)
			->setExcludeUrlsMode($embeddingData['excludeUrlsMode'] ?? $embedding->getExcludeUrlsMode())
			->setEmbeddingVendor($embeddingData['embedding_vendor'] ?? '')
			->setPrivacyPolicyUrl($embeddingData['privacy_policy_url'][$defaultLanguageId] ?? '')
			->setHttpEmbeddingName($embeddingData['http_embedding_name'] ?? '')
			->setLifetime($embeddingData['lifetime'][$defaultLanguageId] ?? '')
			->setCollectedDataInfo($embeddingData['collected_data_info'][$defaultLanguageId] ?? '')
			->setPurposeOfDataCollection($embeddingData['purpose_of_data_collection'][$defaultLanguageId] ?? '')
			->setLegalBasis($embeddingData['legal_basis'][$defaultLanguageId] ?? '')
			->setLegalBasisSupplement($embeddingData['legal_basis_supplement'][$defaultLanguageId] ?? '')
			->setLegalBasisStructured((bool)($embeddingData['legal_basis_structured'] ?? false))
			->setStructuredLegalBasis((array)($embeddingData['structured_legal_basis'] ?? []))
			->setPlaceOfProcessing($embeddingData['place_of_processing'][$defaultLanguageId] ?? '')
			->setAutoUpdated($embeddingData['autoUpdated'] ?? false)
			->setUpdateParameterBlacklist((array)($embeddingData['updateParameterBlacklist'] ?? []))
			->updateLastModified()
		;
		if ($purpose) {
			$embedding->setPurpose($purpose);
		}

		if (!$user || Utils::mayUserAccessRoute($user, 'meta__app_user_embeddings__source_code', 'POST')) {
			$embedding->setCodeSnippet($embeddingData['code_snippet'] ?? '');
		}

		if (isset($embeddingData['source'])) {
			$source_ts = ($embeddingData['source_ts'] ?? null) ?: null;
			if ($source_ts and !($source_ts instanceof DateTimeInterface)) {
				$source_ts = new DateTimeImmutable((string)$source_ts);
			}
			if (!$source_ts) {
				$source_ts = null;
			}
			$embedding->setSource($embeddingData['source'] ?? EmbeddingBase::SOURCE_UNKNOWN)
				->setSourceId(!empty($embeddingData['source_id']) ? $embeddingData['source_id'] : null)
				->setSourceTimestamp($source_ts);
		}

		// Google Consent Mode speichern
		if (isset($embeddingData['gcm']) and is_array($embeddingData['gcm']) and $embedding->hasVendorGoogle()) {
			$embedding->setGoogleConsentMode($embeddingData['gcm']);
		} else {
			$embedding->setGoogleConsentMode([]);
		}

		// TCF-Daten speichern
		$embedding->setTcfVendor(
			(is_string($embeddingData['tcf_vendor_id'] ?? null) or $embedding->getTcfVendor())
			? TcfVendor::find($embeddingData['tcf_vendor_id'] ?? $embedding->getTcfVendor()->getId())
			: null);
		$embedding->setTcfPurposes(array_filter(array_map(
			function ($id) {
				return TcfPurpose::find($id);
			},
			(array)($embeddingData['tcf'][$embeddingData['tcf_vendor_id'] ?? '']['p'] ?? [])
		)));
		$embedding->setTcfSpecialFeatures(array_filter(array_map(
			function ($id) {
				return TcfSpecialFeature::find($id);
			},
			(array)($embeddingData['tcf'][$embeddingData['tcf_vendor_id'] ?? '']['sf'] ?? [])
		)));

		// Assets bearbeiten
		foreach ($embeddingData['asset'] ?? [] as $assetId => $data) {
			/** @var UserEmbeddingAsset $asset */
			$asset = $embedding->getAsset($assetId);
			$assetName = (string)($data['name'] ?? '');

			if ($asset === null || empty($assetName)) {
				continue;
			}

			$asset
				->setName($assetName)
				->setDynamic((bool)($data['dynamic'] ?? false))
				->setStorageType((string)($data['storageType'] ?? ''))
				->setLifetime((string)($data['lifetime'][$defaultLanguageId] ?? ''))
				->setDescription((string)($data['description'][$defaultLanguageId] ?? ''))
			;

			// Asset-Übersetzungen
			foreach ($user->getUserEmbeddingLocales() as $locale) {
				$languageId = $locale->getName();
				if (!isset($data['description'][$languageId])) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				// @phan-suppress-next-line PhanUndeclaredMethod
				$asset
					->setTranslatedDescription($locale, $data['description'][$languageId] ?? '')
					->setTranslatedLifetime($locale, $data['lifetime'][$languageId] ?? '')
				;
			}
		}
		if(!empty($embeddingData['add_asset'][$defaultLanguageId]))
		{
			$addAssets = $embeddingData['add_asset'][$defaultLanguageId];
		}
		elseif (is_array($embeddingData['add_asset'] ?? null)) {
			$addAssets = $embeddingData['add_asset'];
		}

		// Assets hinzufügen
		foreach ($addAssets ?? [] as $data) {
			$assetName = (string)($data['name'] ?? '');
			if (empty($assetName)) {
				continue;
			}
			$asset = UserEmbeddingAsset::create()
				->setName($assetName)
				->setDynamic((bool)($data['dynamic'] ?? false))
				->setStorageType((string)($data['storageType'] ?? ''))
			;

			// Asset-Übersetzungen
			foreach ($user->getUserEmbeddingLocales() as $locale) {
				$languageId = $locale->getName();
				if (!isset($data['description'][$languageId])) {
					continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
				}
				// @phan-suppress-next-line PhanUndeclaredMethod
				$asset
					->setTranslatedDescription($locale, $data['description'][$languageId] ?? '')
					->setTranslatedLifetime($locale, $data['lifetime'][$languageId] ?? '');
			}
			//embed now...
			$embedding->addAsset($asset);
		}

		// Assets löschen
		foreach ($embeddingData['delete_asset'] ?? [] as $assetId) {
			$embedding->removeAsset($assetId);
		}

		// Übersetzungen speichern
		foreach ($user->getUserEmbeddingLocales() as $locale) {
			$languageId = $locale->getName();
			if (!isset($embeddingData['description'][$languageId])) {
				continue; // Wenn description in der Sprache komplett fehlt, ist die Sprache wohl nicht dabei
			}
			$embedding
				->setTranslatedName($locale, $embeddingData['label'][$languageId] ?? '')
				->setTranslatedDescription($locale, $embeddingData['description'][$languageId] ?? '')
				->setTranslatedPrivacyPolicyUrl($locale, $embeddingData['privacy_policy_url'][$languageId] ?? '')
				->setTranslatedLifetime($locale, $embeddingData['lifetime'][$languageId] ?? '')
				->setTranslatedCollectedDataInfo($locale, $embeddingData['collected_data_info'][$languageId] ?? '')
				->setTranslatedPurposeOfDataCollection($locale, $embeddingData['purpose_of_data_collection'][$languageId] ?? '')
				->setTranslatedLegalBasis($locale, $embeddingData['legal_basis'][$languageId] ?? '')
				->setTranslatedLegalBasisSupplement($locale, $embeddingData['legal_basis_supplement'][$languageId] ?? '')
				->setTranslatedPlaceOfProcessing($locale, $embeddingData['place_of_processing'][$languageId] ?? '')
				->setTranslatedBlockedContentTitle($locale, $embeddingData['blocked_content_title'][$languageId] ?? '')
				->setTranslatedBlockedContentText($locale, $embeddingData['blocked_content_text'][$languageId] ?? '')
				->setTranslatedBlockedContentButtonText($locale, $embeddingData['blocked_content_button_text'][$languageId] ?? '')
			;
		}

		return $embedding;
	}

	private static function prepareNewEmbeddingData(Request $request, UserLocale $defaultLocale): array
	{
		$defaultLocaleName = $defaultLocale->getName();
		$embeddingData = [];
		// Embedding-Vorlagen
		$templateId = $request->query->getAlnum('template');
		if ($templateId != '') {
			return self::createSystemEmbedding($templateId, $defaultLocaleName);
		}
		// Import aus DB nach ID
		elseif (($remoteId = $request->query->getInt('remoteId'))) {
			$remoteDb = $request->query->get('remoteDb', 'v2.2');
			if ($remoteDb === 'v2') {
				$remoteEmbedding = EmbeddingDbLegacy::fetchEmbedding($remoteId);
			} else {
				$remoteEmbedding = EmbeddingDb::fetchEmbedding($remoteId);
			}

			$embeddingData = self::createRemoteEmbeddingFromDbEntry($remoteEmbedding, (string)$remoteDb);
		}
		elseif ($request->query->has('domain') && $request->query->has('embedding')) {
			$embeddingData = ['source'=>EmbeddingBase::SOURCE_LOCAL];

			try {
				if (Domain::setDomainForRequest($request->query->getAlnum('domain'))
					and $embedding = Embedding::find($request->query->getAlnum('embedding'))
				) {
					$embeddingData += self::extractDataFromDomainEmbedding($embedding, $defaultLocaleName);
				}
			}
			finally {
				// Reset context
				Domain::setDomainForRequest('');
				Model::closeDomainModels();
			}
		}
		else {
			$embeddingData = ['source'=>EmbeddingBase::SOURCE_LOCAL];
		}
		return $embeddingData;
	}

	private static function createSystemEmbedding(string $templateId, string $defaultLocaleName): array
	{
		$template = SystemEmbedding::find($templateId);
		if ($template) {
			$tcfVendor = $template->getTcfVendor();
			if ($tcfVendor) {
				$tcf = [
					$tcfVendor->getId() => [
						'p' => array_map(function (TcfPurpose $p) { return $p->getId(); }, $template->getTcfPurposes()),
						'sf' => array_map(function (TcfSpecialFeature $sf) { return $sf->getId(); }, $template->getTcfSpecialFeatures()),
					]
				];
			} else {
				$tcf = [];
			}

			$assets = array_values(array_map(function ($asset) use ($defaultLocaleName) {
				$result = [
					'name' => $asset->getName(),
					'dynamic' => $asset->isDynamic(),
					'storageType' => $asset->getStorageType(),
					'lifetime' => [$defaultLocaleName => $asset->getLifetime()],
					'description' => [$defaultLocaleName => $asset->getDescription()],
				];
				foreach (LocaleService::all() as $locale) {
					$language = $locale->getLanguage();
					if (!$language) { continue; }
					$result['lifetime'][$locale->getName()] = $asset->getTranslatedLifetime($language, false);
					$result['description'][$locale->getName()] = $asset->getTranslatedDescription($language, false);
				}
				return $result;
			}, $template->getAssets()));

			$result = [
				"name" => $template->getName(),
				'hinweis' => "from_template",
				"code_snippet" => $template->getCodeSnippet(),
				"code_block" => implode("\n", $template->getCodeBlock()),
				"excludeUrls" => implode("\n", $template->getExcludeUrls()),
				'excludeUrlsMode' => $template->getExcludeUrlsMode(),
				"scriptLoaderGroup" => $template->getScriptLoaderGroup(),
				"iframe_blocking_strings" => implode("\n", $template->getIframeBlockingStrings()),
				"hideIframesUntilConsent" => $template->isIframesHiddenUntilConsent(),
				"dispatchLoadEventForIframes" => $template->isDispatchLoadEventForIframes(),
				"description" => [$defaultLocaleName => $template->getDescription()],
				"purpose" => $template->getPurpose()->getId(),
				"privacy_policy_url" => [$defaultLocaleName => $template->getPrivacyPolicyUrl()],
				"embedding_vendor" => $template->getEmbeddingVendor(),
				"tcf_vendor_id" => $tcfVendor ? $tcfVendor->getId() : null,
				"tcf" => $tcf,
				"lifetime" => [$defaultLocaleName => $template->getLifetime()],
				"active" => true,
				"updateParameterBlacklist" => $template->getUpdateParameterBlacklist(),
				"collected_data_info" => [$defaultLocaleName => $template->getCollectedDataInfo()],
				"purpose_of_data_collection" => [$defaultLocaleName => $template->getPurposeOfDataCollection()],
				"legal_basis" => [$defaultLocaleName => $template->getLegalBasis()],
				"structured_legal_basis" => array_map(static function (array $articles) {
					return array_map(static function (Article $article) { return (int)$article->getId(); }, $articles);
				}, $template->getStructuredLegalBasis()),
				"legal_basis_structured" => $template->isLegalBasisStructured(),
				"legal_basis_supplement" => [$defaultLocaleName => $template->getLegalBasisSupplement()],
				"place_of_processing" => [$defaultLocaleName => $template->getPlaceOfProcessing()],
				'add_asset' => $assets,
				'gcm' => $template->getGoogleConsentMode(),
				"source" => EmbeddingBase::SOURCE_SYSTEM_TEMPLATE,
				"source_id" => $template->getId(),
				"source_ts" => $template->getLastModified(),
			];

			foreach (UserLocale::all() as $locale) {
				$language = $locale->getLanguage();
				if (!$language) { continue; }
				$result['label'][$locale->getName()] = $template->getTranslatedName($language, false);
				$result['description'][$locale->getName()] = $template->getTranslatedDescription($language, false);
				$result['privacy_policy_url'][$locale->getName()] = $template->getTranslatedPrivacyPolicyUrl($language, false);
				$result['lifetime'][$locale->getName()] = $template->getTranslatedLifetime($language, false);
				$result['collected_data_info'][$locale->getName()] = $template->getTranslatedCollectedDataInfo($language, false);
				$result['purpose_of_data_collection'][$locale->getName()] = $template->getTranslatedPurposeOfDataCollection($language, false);
				$result['legal_basis'][$locale->getName()] = $template->getTranslatedLegalBasis($language, false);
				$result['legal_basis_supplement'][$locale->getName()] = $template->getTranslatedLegalBasisSupplement($language, false);
				$result['place_of_processing'][$locale->getName()] = $template->getTranslatedPlaceOfProcessing($language, false);
				$result['blocked_content_title'][$locale->getName()] = $template->getTranslatedBlockedContentTitle($language, false);
				$result['blocked_content_text'][$locale->getName()] = $template->getTranslatedBlockedContentText($language, false);
				$result['blocked_content_button_text'][$locale->getName()] = $template->getTranslatedBlockedContentButtonText($language, false);
			}

			return $result;
		}
		return [];
	}

	private static function createRemoteEmbeddingFromDbEntry($remoteEmbedding, string $remoteDb = 'v2.2'): array
	{
		if (is_array($remoteEmbedding) and $remoteEmbedding) {
			$embeddingData = [];
			$embeddingData['hinweis'] = "hinweis_remote";

			$embeddingData['source'] = ($remoteDb === 'v2') ? EmbeddingBase::SOURCE_GLOBAL_DB_LEGACY : EmbeddingBase::SOURCE_GLOBAL_DB;
			$embeddingData['source_id'] = $remoteEmbedding['id'];
			$embeddingData['source_ts'] = $remoteEmbedding['timestamp'];

			// Upgrade zu neuer DB
			if (isset($remoteEmbedding['@upgrade'])) {
				$upgradeSource = $remoteEmbedding['@upgrade']['source'];
				$upgradeId = $remoteEmbedding['@upgrade']['sourceId'];
				if ($upgradeSource !== $embeddingData['source'] or (string)$upgradeId !== (string)$remoteEmbedding['id']) {
					$embeddingData['source'] = $upgradeSource;
					$embeddingData['source_id'] = $upgradeId;
					if ($upgradeSource === EmbeddingBase::SOURCE_GLOBAL_DB) {
						$remoteEmbedding = EmbeddingDb::fetchEmbedding($upgradeId);
						$embeddingData['source_ts'] = $remoteEmbedding['timestamp'];
					}
				}
			}

			// Basisdaten übertragen
			foreach (['name', 'purpose', 'code_snippet', 'code_block', 'embedding_vendor', 'iframe_blocking_strings'] as $key) {
				$embeddingData[$key] = $remoteEmbedding[$key];
			}
			foreach (['label', 'description', 'privacy_policy_url', 'collected_data_info', 'purpose_of_data_collection', 'legal_basis', 'legal_basis_supplement', 'place_of_processing'] as $key) {
				$embeddingData[$key] = [];
			}
			if (isset($remoteEmbedding['structured_legal_basis'])) {
				$embeddingData['structured_legal_basis'] = $remoteEmbedding['structured_legal_basis'];
				$embeddingData['legal_basis_structured'] = true;

				// Falls für DSGVO, DSG oder CCPA keine Daten existieren, wird ein leeres Array übergeben
				$regulationIds = array_keys(Regulation::rawAll());
				$embeddingData['structured_legal_basis'] += array_fill_keys($regulationIds, []);
			} else {
				$embeddingData['legal_basis_structured'] = false;
			}

			// Übersetzungen übertragen
			foreach (Language::all() as $languageId => $language) {
				$shortCode = $language->getShortCode();
				if (!empty($remoteEmbedding['translations'][$shortCode]) and !empty($remoteEmbedding['translations'][$shortCode]['description'])) {
					$translation = $remoteEmbedding['translations'][$shortCode];
					foreach ($translation as $key => $value) {
						$embeddingData[$key][$languageId] = $value;
					}
				}
			}

			// Assets übertragen
			$embeddingData['add_asset'] = array_map(function ($asset) {
				$result = [
					'name' => $asset['name'],
					'storageType' => $asset['storageType'],
					'dynamic' => $asset['dynamic'],
					'lifetime' => [],
					'description' => [],
				];

				// Übersetzungen übertragen
				foreach (Language::all() as $languageId => $language) {
					$shortCode = $language->getShortCode();
					if (!empty($asset['translations'][$shortCode]) and !empty($asset['translations'][$shortCode]['lifetime'])) {
						$translation = $asset['translations'][$shortCode];
						foreach ($translation as $key => $value) {
							$result[$key][$languageId] = $value;
						}
					}
				}

				return $result;
			}, $remoteEmbedding['assets']);

		} else {
			$embeddingData = ['source' => EmbeddingBase::SOURCE_LOCAL];
		}
		return $embeddingData;
	}

	/**
	 * @param Embedding $embedding
	 * @param string $defaultLocaleName
	 * @return array
	 */
	private static function extractDataFromDomainEmbedding(Embedding $embedding, string $defaultLocaleName): array
	{
		$locales = UserLocale::all();

		// Basisdaten extrahieren
		$data = [
			'name' => $embedding->getName(),
			'purpose' => $embedding->getPurpose() ? $embedding->getPurpose()->getId() : '',
			'code_snippet' => $embedding->getCodeSnippet(),
			'code_block' => implode("\n", $embedding->getCodeBlock()),
			'embedding_vendor' => $embedding->getEmbeddingVendor(),
			'iframe_blocking_strings' => implode("\n", $embedding->getIframeBlockingStrings()),
			'hideIframesUntilConsent' => $embedding->isIframesHiddenUntilConsent(),
			'dispatchLoadEventForIframes' => $embedding->isDispatchLoadEventForIframes(),
			'excludeUrls' => implode("\n", $embedding->getExcludeUrls()),
			'excludeUrlsMode' => $embedding->getExcludeUrlsMode(),
			'scriptLoaderGroup' => $embedding->getScriptLoaderGroup(),
			'structured_legal_basis' => array_map(static function (array $articles) {
				return array_map(static function (Article $article) { return (int)$article->getId(); }, $articles);
			}, $embedding->getStructuredLegalBasis()),
			'legal_basis_structured' => $embedding->isLegalBasisStructured(),
			'autoUpdated' => false,
			'updateParameterBlacklist' => $embedding->getUpdateParameterBlacklist(),
		];

		// Übersetzungen extrahieren
		$data += [
			'blocked_content_button_text' => [],
			'blocked_content_text' => [],
			'blocked_content_title' => [],
			'collected_data_info' => [$defaultLocaleName=>$embedding->getCollectedDataInfo()],
			'description' => [$defaultLocaleName=>$embedding->getDescription()],
			'legal_basis' => [$defaultLocaleName=>$embedding->getLegalBasis()],
			'legal_basis_supplement' => [$defaultLocaleName=>$embedding->getLegalBasisSupplement()],
			'place_of_processing' => [$defaultLocaleName=>$embedding->getPlaceOfProcessing()],
			'privacy_policy_url' => [$defaultLocaleName=>$embedding->getPrivacyPolicyUrl()],
			'purpose_of_data_collection' => [$defaultLocaleName=>$embedding->getPurposeOfDataCollection()],
		];

		foreach ($locales as $localeName => $locale) {
			$data['blocked_content_button_text'][$localeName] = $embedding->getTranslatedBlockedContentButtonText($locale, false);
			$data['blocked_content_text'][$localeName] = $embedding->getTranslatedBlockedContentText($locale, false);
			$data['blocked_content_title'][$localeName] = $embedding->getTranslatedBlockedContentTitle($locale, false);
			$data['collected_data_info'][$localeName] = $embedding->getTranslatedCollectedDataInfo($locale, false);
			$data['description'][$localeName] = $embedding->getTranslatedDescription($locale, false);
			$data['legal_basis'][$localeName] = $embedding->getTranslatedLegalBasis($locale, false);
			$data['legal_basis_supplement'][$localeName] = $embedding->getTranslatedLegalBasisSupplement($locale, false);
			$data['place_of_processing'][$localeName] = $embedding->getTranslatedPlaceOfProcessing($locale, false);
			$data['privacy_policy_url'][$localeName] = $embedding->getTranslatedPrivacyPolicyUrl($locale, false);
			$data['purpose_of_data_collection'][$localeName] = $embedding->getTranslatedPurposeOfDataCollection($locale, false);
		}

		// Assets extrahieren
		$data['add_asset'] = array_map(function (EmbeddingAsset $asset) use ($defaultLocaleName, $locales) {
			$result = [
				'name' => $asset->getName(),
				'storageType' => $asset->getStorageType(),
				'dynamic' => $asset->isDynamic(),
			];

			// Übersetzungen extrahieren
			$result += [
				'description' => [$defaultLocaleName=>$asset->getDescription()],
				'lifetime' => [$defaultLocaleName=>$asset->getLifetime()],
			];

			foreach ($locales as $localeName => $locale) {
				$result['description'][$localeName] = $asset->getTranslatedDescription($locale, false);
				$result['lifetime'][$localeName] = $asset->getTranslatedLifetime($locale, false);
			}

			return $result;
		}, $embedding->getAssets());

		return $data;
	}

	/**
	 * @Route("/new", name="new", methods={"HEAD","GET"})
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function createEmbeddingForm(Request $request, TranslatorInterface $translator): Response
	{
		$csrfToken = CsrfToken::get();
		$purposes = Purpose::all();
		$user = User::loggedInUser();
		$defaultLocale = LocaleService::getUserDefault();
		$defaultLanguage = $defaultLocale->getLanguage();
		$defaultLanguageId = $defaultLanguage ? $defaultLanguage->getCode() : 'de_DE';

		$embeddingData = self::prepareNewEmbeddingData($request, $defaultLocale);

		$addVar="add_";

		return $this->render('user_embedding/new.html.twig', [
			'purposes' => $purposes,
			'locales' => $user->getUserEmbeddingLocales(),
			'csrfToken' => $csrfToken,
			'embedding' => $this->applyChanges(UserEmbedding::createNonPersistent(), $embeddingData, $defaultLanguageId),
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'tcfVendors' => TcfVendor::all(),
			'add_var' => $addVar,
		]);
	}

	/**
	 * @Route("/new", name="new_post", methods={"POST"})
	 *
	 * @param Request $request
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function createEmbedding(Request $request, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$action = $request->request->getAlnum('action');
		$csrfToken = CsrfToken::get();
		$embeddingData = [];
		$purposes = Purpose::all();
		$enabledLocales = $user->getUserEmbeddingLocales();
		$defaultLocale = LocaleService::getUserDefault();
		$defaultLanguage = $defaultLocale->getLanguage();
		$defaultLanguageId = $defaultLanguage ? $defaultLanguage->getCode() : 'de_DE';

		$embedding = $this->applyChanges(UserEmbedding::create(), $this->unpackPostData($request->request->all()), $defaultLanguageId);

		try {
			$csrfToken->checkRequest($request);

			// ohne Namen nicht speichern
			if (trim($embedding->getName()) === '') {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Namen gespeichert werden');
				return $this->render('user_embedding/new.html.twig', [
					'embedding' => $embedding,
					'regulations' => Regulation::all(),
					'regulationGroups' => RegulationGroup::all(),
					'purposes' => $purposes,
					'locales' => $enabledLocales,
					'csrfToken' => $csrfToken,
				]);
			}

			// ohne Verwendungszweck nicht speichern
			if (!$embedding->getPurpose()) {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Verwendungszweck gespeichert werden');
				return $this->render('user_embedding/new.html.twig', [
					'embedding' => $embedding,
					'purposes' => $purposes,
					'locales' => $enabledLocales,
					'regulations' => Regulation::all(),
					'regulationGroups' => RegulationGroup::all(),
					'csrfToken' => $csrfToken,
					'tcfVendors' => TcfVendor::all(),
				]);
			}

			// Speichern
			$embedding->save();
			UserEmbeddingAsset::flush();

			$this->addFlash('success', 'Das Embedding "' . $embedding->getName() . '" wurde erfolgreich erstellt!');

			switch ($action) {
				case 'saveClose':
					return $this->redirectToRoute('app_user_embeddings_list', [], 303);
				case 'save':
				default:
					return $this->redirectToRoute('app_user_embeddings_edit', ['id' => $embedding->getId()], 303);
			}
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
		}

		return $this->render('user_embedding/new.html.twig', [
			'purposes' => $purposes,
			'locales' => $enabledLocales,
			'csrfToken' => $csrfToken,
			'formData' => $embeddingData ?? [],
			'tcfVendors' => TcfVendor::all(),
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'add_var' => "",
		]);
	}

	/**
	 * @Route("/new_bulk", name="new_bulk", methods={"POST"})
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws Exception
	 */
	public function createEmbeddingBulk(Request $request, TranslatorInterface $translator): Response
	{
		$defaultLocale = LocaleService::getDefault();
		$defaultLocaleName = $defaultLocale->getName();

		// Füge ausgewählte System Embeddings hinzu
		$embeddingsSystemTemplate = explode(',', $request->request->get('selectedTemplates'));
		foreach($embeddingsSystemTemplate as $embeddingSystemTemplateId) {
			$systemEmbeddingData = self::createSystemEmbedding($embeddingSystemTemplateId, $defaultLocaleName);
			if (!empty($systemEmbeddingData)) {
				$systemEmbedding = $this->applyChanges(UserEmbedding::create(), $systemEmbeddingData, $defaultLocaleName);
				$systemEmbedding->save();
			}
		}

		// Füge ausgewähltes Embedding aus DB hinzu
		$embeddingsExternalDb = $request->request->get('selectedRemoteIds') ? explode(',', $request->request->get('selectedRemoteIds')) : [];
		foreach($embeddingsExternalDb as $remoteId) {
			$remoteDb = (string)$request->query->get('remoteDb', 'v2.2');
			if ($remoteDb === 'v2') {
				$remoteEmbedding = EmbeddingDbLegacy::fetchEmbedding((int)$remoteId);
			} else {
				$remoteEmbedding = EmbeddingDb::fetchEmbedding((int)$remoteId);
			}

			$remoteEmbeddingData =  self::createRemoteEmbeddingFromDbEntry($remoteEmbedding, $remoteDb);
			if (!empty($remoteEmbeddingData)) {
				$remoteEmbedding = $this->applyChanges(UserEmbedding::create(), $remoteEmbeddingData, $defaultLocaleName);
				$remoteEmbedding->save();
			}
		}

		UserEmbeddingAsset::flush();
		$this->addFlash('success', 'Die ausgewählten Embeddings wurden erfolgreich erstellt!');
		return $this->redirectToRoute('app_user_embeddings_list', [], 303);
	}

	/**
	 * @Route("/{embedding}/gcm", name="gcm_form")
	 * @Route("/gcm", name="gcm_form_new")
	 *
	 * @param UserEmbedding|null $embedding
	 * @return Response
	 */
	public function gcmForm(?UserEmbedding $embedding): Response
	{
		return $this->render('embedding/gcm_form.html.twig', [
			'embedding' => $embedding,
		]);
	}

	/**
	 * @Route("/{id}", methods={"HEAD", "GET"}, name="edit")
	 *
	 * @param string $id
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editEmbeddingForm(string $id, Request $request): Response
	{
		$embedding = UserEmbedding::find($id);
		$user = User::loggedInUser();

		if ($embedding === null) {
			return $this->redirectToRoute('app_user_embeddings_list');
		}

		$purposes = Purpose::all();

		return $this->render('user_embedding/edit.html.twig', [
			'embedding' => $embedding,
			'purposes' => $purposes,
			'locales' => $user->getUserEmbeddingLocales(),
			'csrfToken' => CsrfToken::get(),
			'tcfVendors' => TcfVendor::all(),
			'regulations' => Regulation::all(),
			'regulationGroups' => RegulationGroup::all(),
			'add_var' => "",
		]);
	}

	/**
	 * Führt Daten aus zwei Embedding-Arrays zusammen.
	 *
	 * Die Einträge des zweiten überschreiben dabei die des ersten.
	 * NULL und leere Strings werden nicht berücksichtigt.
	 *
	 * Sonderbehandlung für 'add_asset' -> 'asset'.
	 */
	private static function mergeEmbeddingData(array $data1, array $data2): array
	{
		foreach ($data2 as $key => $value)
		{
			if (is_array($value)) {
				if ($key === 'add_asset' and isset($data1['asset'])) {
					foreach ($value as $asset) {
						$name = $asset['name'] ?? '';
						if ($name == '') {
							continue;
						}
						$assetId = null;
						foreach ($data1['asset'] as $assetKey => $assetData) {
							if (($assetData['name'] ?? '') === $name) {
								$assetId = $assetKey;
								break;
							}
						}
						if ($assetId !== null) {
							$data1['asset'][$assetId] = self::mergeEmbeddingData($data1['asset'][$assetId], $asset);
						} else {
							$data1['add_asset'][] = $asset;
						}
					}
				} else {
					$data1[$key] = self::mergeEmbeddingData($data1[$key] ?? [], $data2[$key]);
				}
			} elseif ($value !== null and $value !== '') {
				$data1[$key] = $value;
			}
		}
		return $data1;
	}

	/**
	 * @Route("/{id}", methods={"PUT"}, name="edit_post")
	 *
	 * @param string $id
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function editEmbedding(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$user = User::loggedInUser();
		$action = $request->request->getAlnum('action');
		$enabledLocales = $user->getUserEmbeddingLocales();
		$defaultLocale = LocaleService::getUserDefault();
		$defaultLanguage = $defaultLocale->getLanguage();
		$defaultLanguageId = $defaultLanguage ? $defaultLanguage->getCode() : 'de_DE';
		$embedding = UserEmbedding::find($id);
		$hadTcfVendor = ($embedding->getTcfVendor() !== null);

		try {
			CsrfToken::get()->checkRequest($request);

			if ($embedding === null) {
				return $this->redirectToRoute('app_user_embeddings_list', [], 303);
			}

			// Embedding-Daten holen und aufbereiten
			$embeddingData = $this->unpackPostData($request->request->all());

			$purpose = Purpose::find($embeddingData['purpose']);

			// Abbrechen wenn zu editierendes Embedding nicht existiert
			UserEmbedding::beginWrite();
			if (UserEmbedding::exists($id) == false) {
				return $this->redirectToRoute('app_user_embeddings_list');
			}

			// ohne Verwendungszweck nicht speichern
			if (Purpose::exists($embeddingData['purpose']) == false) {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Verwendungszweck gespeichert werden');
				return $this->redirectToRoute('app_user_embeddings_edit', ['id' => $embedding->getId()]);
			}

			// Script-Blocking
			$codeBlock = explode("\n", trim((string)($embeddingData['code_block'] ?? '')));
			$codeBlock = array_filter($codeBlock, function ($line) {
				return trim($line) !== '';
			});
			$codeBlock = array_values(array_map(function ($line) {
				return trim($line);
			}, $codeBlock));

			// ohne Namen nicht speichern
			if (trim($embeddingData['name']) === '') {
				$this->addFlash('warning', 'Ein Embedding kann nicht ohne Namen gespeichert werden');
				return $this->redirectToRoute('app_user_embeddings_edit', ['id' => $embedding->getId()]);
			}

			// Wenn AutoUpdate aktiviert wurde, direkt Daten übernehmen
			if (!$embedding->isAutoUpdated() and ($embeddingData['autoUpdated'] ?? false)) {
				switch ($embeddingData['source'] ?? '') {
					case EmbeddingBase::SOURCE_GLOBAL_DB:
						$request->query->set('remoteId', $embeddingData['source_id'] ?? '');
						$request->query->set('remoteDb', 'v2.2');
						break;
					case EmbeddingBase::SOURCE_GLOBAL_DB_LEGACY:
						$request->query->set('remoteId', $embeddingData['source_id'] ?? '');
						$request->query->set('remoteDb', 'v2');
						break;
					case EmbeddingBase::SOURCE_SYSTEM_TEMPLATE:
						$request->query->set('template', $embeddingData['source_id'] ?? '');
						break;
					default:
						break;
				}
				$remoteData = self::prepareNewEmbeddingData($request, $defaultLocale);
				$currentSourceTs = new DateTimeImmutable($embeddingData['source_ts']);
				if (!empty($remoteData['source_ts']) and $remoteData['source_ts'] > $currentSourceTs) {
					$embeddingData = $this->mergeEmbeddingData($embeddingData, $remoteData);
				}
			}

			// Speichern
			$embedding = $this->applyChanges($embedding, $embeddingData, $defaultLanguageId);
			$embedding->save();
			UserEmbeddingAsset::flush();

			$this->addFlash('success', 'Das Embedding "' . $embeddingData['name'] . '" wurde erfolgreich bearbeitet!');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		switch ($action) {
			case 'saveClose':
				return $this->redirectToRoute('app_user_embeddings_list', [], 303);
			case 'save':
			default:
				return $this->redirectToRoute('app_user_embeddings_edit', ['id' => $embedding->getId()], 303);
		}
	}

	/**
	 * @Route("/{id}/delete", name="delete_form")
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function deleteEmbeddingForm(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$embedding = UserEmbedding::find($id);
		if ($embedding === null) {
			return $this->redirectToRoute('app_user_embeddings_list');
		}
		$embeddingName = $embedding->getName();

		return $this->render('user_embedding/delete.html.twig', [
			'embedding' => $embedding,
			'csrfToken' => CsrfToken::get(),
		]);
	}

	/**
	 * @Route("/{id}", name="delete", methods={"DELETE"})
	 * @return Response
	 * @throws ConfigNotWritableException
	 */
	public function deleteEmbedding(string $id, Request $request, TranslatorInterface $translator): Response
	{
		UserEmbedding::beginWrite();
		if (UserEmbedding::exists($id) == false) {
			return $this->redirectToRoute('app_user_embeddings_list');
		}

		$embedding = UserEmbedding::find($id);
		$embeddingName = $embedding->getName();

		try {
			CsrfToken::get()->checkRequest($request);

			foreach ($embedding->getAssets() as $asset) {
				UserEmbeddingAsset::delete($asset->getId(), false);
			}
			UserEmbeddingAsset::flush();

			UserEmbedding::delete($id);
			$this->addFlash('success', 'Das Embedding "' . $embeddingName . '" wurde erfolgreich gelöscht!');
			return $this->redirectToRoute('app_user_embeddings_list');
		}
		catch (CsrfTokenException $e) {
			$this->addFlash('danger', 'CSRF token is missing or invalid. Please try again.');
		}

		return $this->redirectToRoute('app_user_embeddings_delete_form', ['id'=>$embedding->getId()]);
	}

	/**
	 * @Route("/{embedding}/importSource", name="import_source")
	 *
	 * @param UserEmbedding $embedding
	 * @return JsonResponse
	 */
	public function sourceInfo(UserEmbedding $embedding): JsonResponse
	{
		$defaultLocale = LocaleService::getUserDefault();
		$request = new Request();

		switch ($embedding->getSource()) {
			case EmbeddingBase::SOURCE_GLOBAL_DB:
				$request->query->set('remoteId', $embedding->getSourceId());
				$request->query->set('remoteDb', 'v2.2');
				break;
			case EmbeddingBase::SOURCE_GLOBAL_DB_LEGACY:
				$request->query->set('remoteId', $embedding->getSourceId());
				$request->query->set('remoteDb', 'v2');
				break;
			case EmbeddingBase::SOURCE_SYSTEM_TEMPLATE:
				$request->query->set('template', $embedding->getSourceId());
				break;
			default:
				break;
		}

		return $this->json(self::prepareNewEmbeddingData($request, $defaultLocale));
	}
}
