<?php

namespace App\Controller;

use App\Component\LocaleManagerTrait;
use App\Component\Menu\Annotation\Menu;
use App\Component\Reflection\PropertyProcessor;
use App\Exception\LocaleAlreadyExistsException;
use App\Model\Domain;
use App\Model\LocaleOverride;
use App\Model\UserLocale as LocaleModel;
use App\Model\Language as Language;
use App\Model\Locale;
use App\Model\LocaleBase;
use App\Model\User;
use App\Model\Purpose;
use Exception;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/user_locale", name="app_user_locale_")
 */
class UserLocale extends AbstractController
{
	use LocaleManagerTrait;

	/**
	 * @Menu("Manage default texts", group="Settings", icon="fa-file-text-o", order=40)
	 * @Route("", name="list", methods={"GET", "HEAD"})
	 * @return Response
	 */
	public function listLocales(): Response
	{
		$locales = LocaleModel::sortByEnabledAndName(LocaleModel::all());
		$user = User::loggedInUser();
		$fallbackLocale = $user->getAccountOwner()->getDefaultLocale();

		return $this->render('user_locale/list.html.twig', [
			'user' => $user,
			'locales' => $locales,
			'fallback' => $fallbackLocale->getName(),
		]);
	}

	/**
	 * @Route("", name="bulk_action", methods={"POST"})
	 * @param Request $request
	 * @return Response
	 */
	public function bulkAction(Request $request, TranslatorInterface $translator): Response
	{
		$action = $request->request->get('action');

		switch ($action) {
			case 'deploy':
				return $this->deployLocales($request, $translator);
			default:
				$this->addFlash('danger', 'Invalid action');
				return $this->redirectToRoute('app_user_locale_list', [], 303);
		}
	}

	/**
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	private function deployLocales(Request $request, TranslatorInterface $translator): Response
	{
		$userLocales = LocaleModel::all();
		$domains = Domain::all();

		foreach ($domains as $dom)
		{
			Locale::switchRepositoryContext($dom->getUser(), $dom);
			foreach (Locale::all() as $locale) {
				if ($userLocale = $userLocales[$locale->getName()] ?? null) {
					// Remember properties that are not supposed to be overwritten
					$enabled = $locale->isEnabled();
					$updateParameterBlacklist = $locale->getUpdateParameterBlacklist();

					// Build setter blacklist for usage in PropertyProcessor (dash-separated arguments)
					$setterBlacklist = array_map(static function (string $fieldId) {
						[$setter, $purposeId] = self::buildGetterSetterName($fieldId, 'set');
						return rtrim("$setter-$purposeId", '-');
					}, $updateParameterBlacklist);

					// Overwrite locale
					PropertyProcessor::copyValues($userLocale, $locale, LocaleBase::class, $setterBlacklist);

					// Restore properties
					$locale->setEnabled($enabled);
					$locale->setUpdateParameterBlacklist($updateParameterBlacklist);
				}
			}
			Locale::flush();
		}
		$this->addFlash('info',
			$translator->trans('Locales have been successfully deployed.')
		);
		return $this->redirectToRoute('app_user_locale_list');
	}

	/**
	 * @Route("/new", name="new")
	 *
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function createLocale(Request $request, TranslatorInterface $translator): Response
	{
		if ($request->getMethod() === 'POST') {
			$localeName = $request->request->get('locale_name', '');
			$fields = self::getStandardFieldGroups($translator);

			try {
				$action = $request->request->getAlnum('action');
				$locale = LocaleModel::create($localeName);

				$locale
					->setEnabled($request->request->getBoolean('localeEnabled'))
				;

				// Alle Daten in den Feldern speichern
				foreach ($fields as $fieldset) {
					foreach ($fieldset as $field) {
						$setter = 'set'.preg_replace_callback('/(?:^|_)([a-z])/',
							function ($match) { return strtoupper($match[1]); },
							$field['field_id']
						);
						$value = trim($request->request->get($field['field_id'], ''));
						if (strpos($setter, 'setPurposeName-') === 0) {
							$purpose = Purpose::find(substr($setter, 15));
							$setter = substr($setter, 0, 14);
							$locale->$setter($purpose, $value);
						} elseif (preg_match('~^setPurposeDescription-(?<purposeId>[0-9a-f]+)$~', $setter, $match)) {
							if ($purpose = Purpose::find($match['purposeId'])) {
								$locale->setPurposeDescription($purpose, $value);
							}
						} else {
							$locale->$setter($value);
						}
					}
				}
				$locale->save();

				$this->addFlash('success',
					$translator->trans('Locale %locale% has been updated.', ['%locale%' => $localeName])
				);

				switch ($action) {
					case 'createClose':
						return $this->redirectToRoute('app_user_locale_list', [], 303);
					case 'create':
					default:
						return $this->redirectToRoute('app_user_locale_edit', ['id' => $locale->getId()], 303);
				}
			}
			catch (LocaleAlreadyExistsException $e) {
				$this->addFlash('danger',
					$translator->trans('Locale %locale% already exists.', ['%locale%' => $localeName])
				);
			}
			catch (Exception $e) {
				$this->addFlash('danger',
					$translator->trans('An unknown error has occured.')
				);
			}
		}

		$fields = self::getStandardFieldGroups($translator);
		$fields = array_map(function ($fieldset) use ($request) {
			return array_map(function ($field) use ($request) {
					$field['user_input'] = $request->request->get($field['field_id'], '');
					return $field;
				}, $fieldset);
		}, $fields);

		$locales = LocaleModel::all();

		$used_names = array_map(function ($item) { return $item->getName(); }, $locales);

		return $this->render('user_locale/new.html.twig', [
			'user' => User::loggedInUser(),
			'fields' => $fields,
			'used_names' => $used_names,
			'languages' => Language::all(),
			'locale_name' => $request->request->get('locale_name', ''),
			'localeEnabled' => $request->request->getBoolean('localeEnabled'),
			'localeTemplates' => $request->getMethod() == 'GET' ? $locales : [],
		]);
	}

	/**
	 * @Route("/{localeName}.json", name="ajax", methods={"HEAD", "GET"})
	 * @param string $localeName
	 * @param TranslatorInterface $translator
	 * @return JsonResponse
	 */
	public function ajaxLocale(string $localeName, TranslatorInterface $translator): JsonResponse
	{
		$locale = LocaleModel::find($localeName);
		if (!$locale) {
			return $this->json(['success'=>false, 'error'=>'locale not found'], 404);
		}

		$fieldGroups = self::getStandardFieldGroups($translator);
		$translation = array_map(static function (array $field) use ($locale) {
			[$getter, $purposeId] = self::buildGetterSetterName($field['field_id']);
			if (!method_exists($locale, $getter)) {
				return null;
			}
			elseif ($purposeId and $purpose = Purpose::find((string)$purposeId)) {
				return $locale->$getter($purpose);
			}
			else {
				return $locale->$getter();
			}
		}, array_merge([], ...array_values($fieldGroups)));

		// Discard empty values
		$translation = array_filter($translation, static function ($value) { return is_string($value) && $value != ''; });

		return $this->json($translation);
	}

	/**
	 * @Route("/{id}", name="edit")
	 *
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function editLocale(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$locale = LocaleModel::find($id);
		if ($locale === null) {
			return $this->redirectToRoute('app_user_locale_list');
		}

		if ($request->getRealMethod() === 'POST') {
			$action = $request->request->getAlnum('action');
			$fields = self::getStandardFieldGroups($translator);
			try {
				$locale
					->setEnabled((bool)$request->request->get('localeEnabled'))
					->setUpdateParameterBlacklist($request->request->all('updateParameterBlacklist'))
				;

				// Alle Daten in den Feldern speichern
				foreach ($fields as $fieldset) {
					foreach ($fieldset as $field) {
						$setter = 'set'.preg_replace_callback('/(?:^|_)([a-z])/',
							function ($match) { return strtoupper($match[1]); },
							$field['field_id']
						);
						$value = trim($request->request->get($field['field_id'], ''));
						if (strpos($setter, 'setPurposeName-') === 0) {
							$purpose = Purpose::find(substr($setter, 15));
							$setter = substr($setter, 0, 14);
							$locale->$setter($purpose, $value);
						} elseif (preg_match('~^setPurposeDescription-(?<purposeId>[0-9a-f]+)$~', $setter, $match)) {
							if ($purpose = Purpose::find($match['purposeId'])) {
								$locale->setPurposeDescription($purpose, $value);
							}
						} else {
							$locale->$setter($value);
						}
					}
				}
				$locale->save();

				$this->addFlash('success',
					$translator->trans('Locale %locale% has been updated.', ['%locale%' => $locale->getName()])
				);

				switch ($action) {
					case 'saveClose':
						return $this->redirectToRoute('app_user_locale_list', [], 303);
					case 'save':
					default:
						return $this->redirectToRoute('app_user_locale_edit', ['id' => $locale->getId()], 303);
				}
			}
			catch (Exception $e) {
				$this->addFlash('danger', $translator->trans('An unknown error has occured.'));
			}
		}

		// Gespeicherte Werte setzen
		$fields = self::getStandardFieldGroups($translator);
		$fields = array_map(function ($fieldset) use ($locale) {
			return array_map(function ($field) use ($locale) {
				$getter = 'get'.preg_replace_callback('/(?:^|_)([a-z])/',
					function ($match) { return strtoupper($match[1]); },
					$field['field_id']
				);
				if (strpos($getter, 'getPurposeName-') === 0) {
					$purpose = Purpose::find(substr($getter, 15));
					$getter = substr($getter, 0, 14);
					$field['user_input'] = $locale->$getter($purpose) ?? '';
				} elseif (preg_match('~^getPurposeDescription-(?<purposeId>[0-9a-f]+)$~', $getter, $match)) {
					if ($purpose = Purpose::find($match['purposeId'])) {
						$field['user_input'] = $locale->getPurposeDescription($purpose);
					}
				} else {
					$field['user_input'] = $locale->$getter() ?? '';
				}
				return $field;
			}, $fieldset);
		}, $fields);

		// Standardwerte aus dem UserLocale-Skeleton übernehmen
		if ($localeDefaults = LocaleModel::readSkeletonData()[$locale->getName()] ?? null) {
			$fields = array_map(function ($fieldset) use ($localeDefaults) {
				return array_map(function ($field) use ($localeDefaults) {
					list($fieldName, $purposeId) = self::buildSkeletonName($field['field_id']);
					if ($purposeId) {
						$field['default'] = $localeDefaults[$fieldName][$purposeId] ?? null;
					} else {
						$field['default'] = $localeDefaults[$fieldName] ?? null;
					}
					return $field;
				}, $fieldset);
			}, $fields);
		}

		// Systemweite LocaleOverride-Übersetzungen als Standardwerte vorziehen
		array_walk($fields, function (&$group) use ($locale) {
			array_walk($group, function (&$field) use ($locale) {;
				if ($override = LocaleOverride::find($field['field_id'])) {
					$field['default'] = $override->getTranslations()[$locale->getName()] ?? $field['default'];
				}
			});
		});

		return $this->render('user_locale/edit.html.twig', [
			'user' => User::loggedInUser(),
			'locale' => $locale,
			'fields' => $fields,
		]);
	}

	/**
	 * @Route("/{id}/delete", name="delete")
	 *
	 * @param string $id
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 * @throws \App\Exception\ConfigNotWritableException
	 */
	public function deleteLocale(string $id, Request $request, TranslatorInterface $translator): Response
	{
		$locale = LocaleModel::find($id);
		if ($locale === null) {
			return $this->redirectToRoute('app_user_locale_list');
		}

		$localeName = $locale->getName();
		$localeBlacklist = [
			'de_DE',
			'en_US',
		];


		// Bestimmte Übersetzungen dürfen nicht gelöscht werden
		if (in_array($localeName, $localeBlacklist)) {
			$this->addFlash('info',
				$translator->trans('Locale %locale% may not be deleted.', ['%locale%' => $localeName])
			);
			return $this->redirectToRoute('app_user_locale_list');
		}

		if ($request->getRealMethod() === 'POST') {
			LocaleModel::delete($locale->getId());

			$this->addFlash('success',
				$translator->trans('Locale %locale% has been deleted.', ['%locale%' => $localeName])
			);

			return $this->redirectToRoute('app_user_locale_list');
		}

		return $this->render('user_locale/delete.html.twig', [
			'user' => User::loggedInUser(),
			'locale' => $locale,
		]);
	}
}
