<?php

namespace App\Controller\UserSettings;

use App\Component\MailSubscriptions;
use App\Component\Menu\Annotation\Menu;
use App\Controller\AbstractController;
use App\Exception\CsrfTokenException;
use App\Model\CsrfToken;
use App\Model\Country;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\LocaleBase;
use App\Model\User;
use App\Model\UserLocale;
use App\Model\Regulations\Regulation;
use DateTimeZone;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * @Route("/manage_usettings", name="app_usettings")
 */
class ManageSettings extends AbstractController
{
	/**
	 * @Menu("Global Account Settings", group="Account Settings", icon="fa-server", routeGroup="app_usettings_", order=30)
	 * @Route("/", name="_index", methods={"HEAD", "GET" })
	 * @param Request $request
	 * @param MailSubscriptions $mailSubscriptions
	 * @return Response
	 */
	public function index(Request $request,MailSubscriptions $mailSubscriptions): Response
	{
		$user = User::loggedInUser();
		$subscriptions = $mailSubscriptions->list($user);

		$owner = User::loggedInUser()->getAccountOwner();
		$locales = UserLocale::sortByEnabledAndName(UserLocale::all());
		$fallbackLocale = $owner->getDefaultLocale();


		//Ausgabe der Daten
		return $this->render('user_settings/account_settings/index.html.twig', [
			'csrfToken' => CsrfToken::get(),
			'timestamp' => time(),
			'timeZones' => DateTimeZone::listIdentifiers() ?: ['Europe/Berlin', 'UTC'],
			'countries' => Country::all(),
			'subscriptions' => $subscriptions,
			'regulations' => Regulation::all(),
			'locales' => $locales,
			'fallback' => $fallbackLocale->getName(),
		]);
	}

	/**
	 * @Route("/", name="_save_settings", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @param MailSubscriptions $mailSubscriptions
	 *
	 * @return Response
	 * @throws \Exception
	 */
	public function changeUserdata(Request $request, TranslatorInterface $translator, MailSubscriptions $mailSubscriptions): Response
	{
		$action = $request->request->getAlnum('action');

		//Csrf checken
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('app_account');
		}

		$user = User::loggedInUser();

		//Kein User - dann zurück auf Los
		if($user===null)
		{
			return $this->redirectToRoute('app_usettings_index');
		}

		switch ($action) {

			case "saveMenuPresets":
				$user
					->setActivateCrossDomainThemes($request->request->getBoolean('activateCrossDomainThemes'))
					->setActivateCrossDomainLocales($request->request->getBoolean('activateCrossDomainLocales'))
					->setActivateCrossDomainGeneralAPI($request->request->getBoolean('activateCrossDomainGeneralAPI'))
					->setActivateCrossDomainUserGroup($request->request->getBoolean('activateCrossDomainUserGroup'))
					->setActivateCrossDomainIab($request->request->getBoolean('activateCrossDomainIab'))
					->setActivateCrossDomainAccessibility($request->request->getBoolean('activateCrossDomainAccessibility'))
					->save();

				$this->addFlash('success', $translator->trans('Your preferences have been saved!'));
				break;

			case "saveSubscriptions":
				$availableSubscriptions = $mailSubscriptions->list();
				$wantedSubscriptions = $request->request->all('mailSubscriptions');
				$subscriptions = array_map(function ($key) use ($wantedSubscriptions) {
					return filter_var($wantedSubscriptions[$key] ?? false, FILTER_VALIDATE_BOOLEAN, FILTER_NULL_ON_FAILURE) ?? true;
				}, array_combine(array_keys($availableSubscriptions), array_keys($availableSubscriptions)));

				$user->setMailingPreferences($subscriptions)->save();

				$this->addFlash('success', $translator->trans('Your preferences have been saved!'));
				break;

			case 'saveDomainPresets':
				$countryCode = (string)$request->request->get('defaultServiceCountry');
				$country = $countryCode ? Country::find($countryCode) : null;

				if ($user->isSendFullConsentAndRejectionAcrossDomains() != $request->request->getBoolean('sendFullConsentAndRejectionAcrossDomains')) {
					$request->attributes->set('clearCacheScope', 'user');
				}
				$locale = UserLocale::find($request->request->get('fallbackLocale', '')) ?? UserLocale::find('en_US') ?? UserLocale::find('de_DE');
				assert($locale !== null);

				$user
					->getAccountOwner()
					->setNewDomainOnlyActiveOnMatchingDomain($request->request->getBoolean('newDomainOnlyActiveOnMatchingDomain'))
					->setDefaultTimeZone(new DateTimeZone($request->request->get('defaultTimeZone')))
					->setDefaultServiceCountry($country)
					->setDefaultLegalBases(array_filter(array_map(function ($item) { return (string)$item; }, $request->request->all("defaultLegalBases"))))
					->setDefaultLocale($locale)
					->setSendFullConsentAndRejectionAcrossDomains($request->request->getBoolean('sendFullConsentAndRejectionAcrossDomains'))
					->save();

				$this->updateEmbeddingFallbackTexts($user->getAccountOwner(), $locale);

				$this->addFlash('success', $translator->trans('Your preferences have been saved!'));
				break;

			case 'saveCrawlingPresets':

				$setActivate = false;

				//Nur wenn Crawling aktiv ist, dann kann man auch Embeddings automatisch aktivieren
				if($user->isCrawlAllDomains() || $request->request->getBoolean('crawlAllDomains')) {
					if($request->request->getBoolean('allowAutoActivationOfNewEmbeddings'))
					{
						$setActivate = true;
					}
				}
				else
				{
					if($request->request->getBoolean('allowAutoActivationOfNewEmbeddings'))
					{
						$this->addFlash('danger', $translator->trans('You can not activate auto activation of new embeddings when you are not crawling all domains!'));
					}
				}
				//Aktivierung der neuen Embeddings deaktivieren wenn Crawling deaktiviert ist, für den Fall dass der Crawl noch läuft, und dann hier deaktiviert wird.
				//Der Crawlermanager checkt direkt vor dem Update ob das gesetzt ist und aktiviert dann die Embeddings
				if(empty($request->request->getBoolean('crawlAllDomains')) && !empty($user->isAllowAutoActivationOfNewEmbeddings()))
				{
					$this->addFlash('danger', $translator->trans('You can not activate auto activation of new embeddings when you are not crawling all domains!'));
					$setActivate = false;
				}

				/**
				 * Das bleibt erstmal auskommentiert, will man wirklich alles überschreiben ohne zu prüfen ob der User das wirklich will?
				 * Besser nicht.
				 *
				 * User::loggedInUser();
				* $domains = Domain::all();
				* foreach ($domains as $domain) {
					* //Crawling und Embeddings aktivieren wie vom User angegeben
					* //noch die Crawlfrequenz einbauen und das Crawlbudget festlegen
					* $domain
						* ->setDoCrawl($request->request->getBoolean('crawlAllDomains'))
						* ->setActivateEmbeddingsFoundByCrawler($setActivate)
						* ->save();
				* }
				 * **/

				//Daten updaten
				$user
					->setCrawlAllDomains($request->request->getBoolean('crawlAllDomains'))
					->setAllowAutoActivationOfNewEmbeddings($setActivate)
					->setAllowAutoUpdateOfEmbeddings($request->request->getBoolean('allowAutoUpdateOfEmbeddings'))
					->save();

				$this->addFlash('success', $translator->trans('Your preferences have been saved!'));
				break;

			default:
		}

		return $this->redirectToRoute('app_usettings_index');
	}

	/**
	 * @param User $owner Hauptuser
	 * @param LocaleBase $locale Fallback-Sprache
	 * @return void
	 */
	private function updateEmbeddingFallbackTexts(User $owner, LocaleBase $locale): void
	{
		// Alle Embedding-Fallback-Texte ändern
		$domains = Domain::all();
		foreach ($domains as $domain) {
			Embedding::switchRepositoryContext($owner, $domain);
			EmbeddingAsset::switchRepositoryContext($owner, $domain);

			foreach (Embedding::all() as $embedding) {
				foreach (['Description', 'PrivacyPolicyUrl', 'Lifetime', 'CollectedDataInfo', 'PurposeOfDataCollection', 'LegalBasis', 'PlaceOfProcessing'] as $key) {
					$value = call_user_func([$embedding, 'getTranslated'.$key], $locale, false);
					if ($value !== '') {
						call_user_func([$embedding, 'set'.$key], $value);
					}
				}
			}

			Embedding::flush();
			foreach (EmbeddingAsset::all() as $asset) {
				foreach (['Description', 'Lifetime'] as $key) {
					$value = call_user_func([$asset, 'getTranslated'.$key], $locale, false);
					if ($value !== '') {
						call_user_func([$asset, 'set'.$key], $value);
					}
				}
			}
			EmbeddingAsset::flush();
		}
	}

	/**
	 * @Route("/reset_crawler", name="_reset_crawler", methods={"POST"})
	 * @param Request $request
	 * @param TranslatorInterface $translator
	 * @return Response
	 */
	public function resetCrawler(Request $request, TranslatorInterface $translator): Response
	{
		$action = $request->request->getAlnum('action');

		//Csrf checken
		try {
			CsrfToken::get()->checkRequest($request);
		} catch (CsrfTokenException $e) {
			$this->addFlash('danger', $translator->trans('CSRF token is missing or invalid. Please try again.'));
			return $this->redirectToRoute('app_account');
		}

		switch ($action) {
			case 'resetEnableCrawling':
				$account = User::loggedInUser()->getAccountOwner();
				$accountCrawlBudget = (int)($account->getSettings()['maxCrawlBudget'] ?? false);
				if($accountCrawlBudget !== -1)
				{
					$accountCrawlBudget = round($accountCrawlBudget / $account->getActualDomainCount());
				}

				foreach (Domain::all() as $domain) {
					$domain
						->setDoCrawl(true)
						->setActivateEmbeddingsFoundByCrawler(true)
						->setDomainExternalCrawlerFrequency(8)
						->setDomainMaxCrawlBudget($accountCrawlBudget)
						->setActivatePermaScan(false);
				}
				Domain::flush();
				$this->addFlash('success', $translator->trans('All Crawlsettings of all your Domains have been set!'));
				break;

			case 'disableCrawling':
				foreach (Domain::all() as $domain) {
					//NICHT Crawlen und NICHT Embeddings aktivieren
					$domain
						->setDoCrawl(false)
						->setActivateEmbeddingsFoundByCrawler(false);
				}
				Domain::flush();
				$this->addFlash('success', $translator->trans('All Crawlsettings of all your Domains have been set!'));
				break;

			default:
		}

		return $this->redirectToRoute('app_usettings_index');
	}

}
