<?php

namespace App\Controller\api;

use App\Component\Api\Exception\BadRequestException;
use App\Component\Cache\Cache;
use App\Component\ThumbnailExtractor\ThumbnailExtractorService;
use App\Component\Update;
use App\Component\Plugin\PluginInfo;
use App\Model\PluginState;
use App\Config;
use App\Controller\AbstractController;
use App\Event\StatisticsCronJobEvent;
use App\Model\AccessCountJournal;
use App\Model\Country;
use App\Model\Domain;
use App\Model\GeoIp;
use App\Model\Language;
use App\Model\Model;
use App\Model\User;
use App\Utils;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Routing\Annotation\Route;

/**
 * Interne API zum Austausch zwischen Servern des gleichen Clusters
 *
 * @Route("/api/_cluster/", name="app_cluster_api_")
 */
class ClusterAPI extends AbstractController
{
	/**
	 * @Route("test", name="test", methods={"GET", "HEAD"})
	 * @return JsonResponse
	 */
	public function index(): JsonResponse
	{
		//Test Response erstellen
		$response = new JsonResponse([
			'status' => 'ok',
		]);

		// Use the JSON_PRETTY_PRINT
		$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );

		//und raus...
		return $response;
	}

	/**
	 * Liefert allgemeine Statistik-Daten
	 *
	 * Aktuell hauptsächlich die Anzahl der Benutzer, der bekannten Länder und Sprachen.
	 *
	 * @Route("stats", name="stats", methods={"GET", "HEAD"})
	 * @return JsonResponse
	 */
	public function stats(): JsonResponse
	{
		$response = new JsonResponse([
			'status' => 'ok',
			'countries' => Country::count(),
			'languages' => Language::count(),
			'users' => User::count(),
		]);

		//und raus...
		return $response;
	}

	/**
	 * Liefert Daten über die Netzwerk-Konfiguration
	 *
	 * Aktuell hauptsächlich die erlaubten Reverse-Proxies und die Aufrufzähler-Ausnahmen.
	 *
	 * @Route("network", name="network", methods={"GET", "HEAD"})
	 * @return JsonResponse
	 */
	public function network(): JsonResponse
	{
		$config = Config::getInstance();
		$trustedProxies = empty($_ENV['TRUSTED_PROXIES']) ? [] : array_filter(explode(',', $_ENV['TRUSTED_PROXIES']));

		$response = new JsonResponse([
			'status' => 'ok',
			'trusted_proxies' => $trustedProxies,
			'exempt_addresses' => $config->getExemptAddresses(),
			'exempt_user_agents' => $config->getExemptUserAgents(),
			'force_https' => $config->isForceHttpsConnection(),
		]);

		//und raus...
		return $response;
	}

	/**
	 * Liefert die GeoIP-Country-Datenbank
	 *
	 * Dient dazu, die GeoIP-Datenbank mit Frontend-Caches abzugleichen
	 *
	 * @Route("geoip.mmdb", name="geoip_mmdb", methods={"GET", "HEAD"})
	 * @return BinaryFileResponse
	 */
	public function geoIpMmdb(): BinaryFileResponse
	{
		$file = GeoIp::getRepositoryFile();
		return new BinaryFileResponse($file, 200, ['Content-Security-Policy' => "default-src: 'none'; img-src 'self'; form-action 'none'"]);
	}

	/**
	 * Erhöht mehrere Call-Counter auf einmal
	 *
	 * Der Payload muss ein JSON-Objekt in dieser Form sein:
	 * ```
	 * {
	 *   apiKey: {
	 *     domainId: counter,
	 *     ...
	 *   },
	 *   ...
	 * }
	 * ```
	 *
	 * Gedacht für Frontend-Caches, die selbst einen Call-Counter führen und nicht jeden
	 * Request durchleiten.
	 *
	 * @Route("call_counter/bulk", name="update_call_counters", methods={"POST"})
	 * @return JsonResponse
	 */
	public function updateCallCounters(Request $request): JsonResponse
	{
		$payload = null;
		if (Utils::decodeJsonRequest($request, $payload, true, 4) == false) {
			/** @var JsonResponse $payload */
			assert($payload instanceof JsonResponse);
			throw new BadRequestException($payload);
		}
		elseif (!is_array($payload)) {
			throw new BadRequestException(new JsonResponse([
				'success' => false,
				'message' => 'Invalid payload. Make sure to send a JSON encoded object.',
			], 400));
		}

		$statusCode = 202; // 202 Accepted
		$sum = 0;
		foreach ($payload as $apiKey => $domains) {
			$user = User::byApiKey($apiKey);
			if (!$user) {
				continue;
			}
			foreach ($domains as $domainId => $counter) {
				$counter = max(0, (int)$counter);
				$sum += $counter;
				AccessCountJournal::byUserIdDomainIdAndKind($user->getId(), $domainId, 'call')->increment($counter);
			}
		}

		// Wenn `?sync=1`: Zähler am Ende in Datenbank übertragen
		if ($request->query->getBoolean("sync")) {
			$statusCode = 200; // 200 OK
			foreach ($payload as $apiKey => $domains) {
				$user = User::byApiKey($apiKey);
				if (!$user) {
					continue;
				}
				User::setApiUser($user->getApiKey());
				Model::closeUserModels();
				foreach ($domains as $domainId => $ignore) {
					if (!Domain::setDomainForRequest($domainId)) {
						continue;
					}
					Utils::dispatchEvent(StatisticsCronJobEvent::NAME, new StatisticsCronJobEvent(false));
					Model::closeDomainModels();
				}
			}
		}

		$response = new JsonResponse([
			'status' => 'ok',
			'sum' => $sum,
		], $statusCode);

		return $response;
	}

	// Leeren der Caches

	/**
	 * @Route("cache", name="cache_clear", methods={"DELETE"})
	 * @param Update $updater
	 * @return JsonResponse
	 */
	public function fullCacheClear(Update $updater, ThumbnailExtractorService $thumbnails): JsonResponse
	{
		$updater->deleteFullCacheOnExit();
		$thumbnails->clearCache();
		Utils::clearDirectory(Utils::getVarDir() . '/tcfv2');

		if (function_exists('opcache_reset')) {
			opcache_reset();
		}

		return $this->json([
			'status' => 'ok',
		], 202);
	}

	/**
	 * @Route("cache/thumbnails", name="thumbnail_cache_clear", methods={"DELETE"})
	 * @return JsonResponse
	 */
	public function thumbnailCacheClear(ThumbnailExtractorService $thumbnails): JsonResponse
	{
		$thumbnails->clearCache();

		return $this->json([
			'status' => 'ok',
		]);
	}

	/**
	 * @Route("cache/opcache", name="opcache_clear", methods={"DELETE"})
	 * @return JsonResponse
	 */
	public function opcacheClear(): JsonResponse
	{
		if (function_exists('opcache_reset')) {
			opcache_reset();
			return $this->json([
				'status' => 'ok',
			]);
		} else {
			return $this->json([
				'status' => 'no action required',
			]);
		}
	}

	/**
	 * @Route("cache/frontend", name="frontend_cache_clear", methods={"DELETE"})
	 * @return JsonResponse
	 */
	public function frontendCacheClear(): JsonResponse
	{
		Cache::clearAll();

		return $this->json([
			'status' => 'ok',
		]);
	}

	/**
	 * @Route("cache/frontend/{userId}", name="user_cache_clear", methods={"DELETE"})
	 * @param string $userId
	 * @return JsonResponse
	 */
	public function userFrontendCacheClear(string $userId): JsonResponse
	{
		$user = User::byId($userId);
		if (!$user) {
			return $this->json(['status'=>'invalid user'], 404);
		}
		elseif ($user->isSubUser()) {
			return $this->json(['status'=>'subuser not allowed'], 400);
		}

		// Setup proper repository context
		User::setApiUser($user->getApiKey());

		Cache::clearUser($user);

		return $this->json([
			'status' => 'ok',
		]);
	}

	/**
	 * @Route("cache/frontend/{userId}/{domainId}", name="domain_cache_clear", methods={"DELETE"})
	 * @param string $userId
	 * @param string $domainId
	 * @param Request $request
	 * @return JsonResponse
	 */
	public function domainCacheClear(string $userId, string $domainId, Request $request): JsonResponse
	{
		$user = User::byId($userId);
		if (!$user) {
			return $this->json(['status'=>'invalid user'], 404);
		}
		elseif ($user->isSubUser()) {
			return $this->json(['status'=>'subuser not allowed'], 400);
		}

		// Setup proper repository context
		User::setApiUser($user->getApiKey());
		Domain::switchToUser($user);

		if (Domain::setDomainForRequest($domainId) == false) {
			return $this->json(['status'=>'invalid domain'], 404);
		}

		/** @var Domain $domain */
		$domain = Domain::find($domainId);
		assert($domain instanceof Domain);

		Cache::clear($domain, $request->query->getBoolean('keepMetadata'));

		return $this->json([
			'status' => 'ok',
		]);
	}

	/**
	 * @Route("cache/tcf", name="tcf_cache_clear", methods={"DELETE"})
	 * @return JsonResponse
	 */
	public function tcfCacheClear(): JsonResponse
	{
		Utils::clearDirectory(Utils::getVarDir() . '/tcfv2');

		return $this->json([
			'status' => 'ok',
		]);
	}

	/**
	 * @Route("plugins/config/{plugin<.+>}", name="plugin_config", methods={"PUT"})
	 * @return JsonResponse
	 */
	public function configurePlugin(?PluginInfo $plugin, Request $request, Update $updater): JsonResponse
	{
		if (!$plugin) {
			return $this->json(['success' => false, 'error' => "no such plugin"], 404);
		}

		$payload = null;
		if (Utils::decodeJsonRequest($request, $payload, true, 4) == false) {
			/** @var JsonResponse $payload */
			assert($payload instanceof JsonResponse);
			throw new BadRequestException($payload);
		}
		elseif (!is_array($payload)) {
			throw new BadRequestException(new JsonResponse([
				'success' => false,
				'message' => 'Invalid payload. Make sure to send a JSON encoded object.',
			], 400));
		}

		// Aktivieren / Konfigurieren
		$state = $plugin->getState();
		if (!$state) {
			$state = PluginState::create($payload['stateId'] ?? null);
			$state->setName($plugin->getLabel());
			$state->setType($plugin->getType());
			$state->setDirname($plugin->getDirname());
			$state->setPluginUID($plugin->getName());
			$state->setIcon($plugin->getIcon());
			$updater->deleteFullCacheOnExit();
		}
		$state->setVersion($plugin->getVersion());
		$state->setAllowTrial($plugin->isAllowTrial());
		if (!$plugin->isActivatablePerDomain()) {
			$state->setActivatedForAllAccounts(true);
		}
		$state->setActivatedForAllAccounts($payload['activateForAllAccounts'] ?? false);
		$state->setActivatedForAllAccountsOnTheirOwn($payload['activateForAllAccountsOnTheirOwn'] ?? false);
		$state->setFullAccessForAllAccounts($payload['fullAccessForAllAccounts'] ?? false);
		$state->setFullAccessCallCount($payload['fullAccessCallCount']);
		$state->setFullAccessDomCount($payload['fullAccessDomCount']);
		$state->setTestAccessCallCount($payload['testAccessCallCount']);
		$state->setTestAccessDomCount($payload['testAccessDomCount']);
		$state->setTestAccessTime($payload['testAccessTime']);
		$state->setSavedAccess($payload['savedAccess'] ?? false);
		$state->save();
		return $this->json(['success' => true]);
	}

	/**
	 * @Route("plugins/config/{plugin<.+>}", name="plugin_disable", methods={"DELETE"})
	 * @return JsonResponse
	 */
	public function disablePlugin(string $plugin, Update $updater): JsonResponse
	{
		// Deaktivieren
		$changed = false;
		foreach (PluginState::all() as $state) {
			if ($state->getName() === $plugin) {
				PluginState::delete($state->getId(), false);
				$changed = true;
			}
		}
		PluginState::flush();
		if ($changed) {
			$updater->deleteFullCacheOnExit();
		}
		return $this->json(['success' => $changed]);
	}
}
