<?php

namespace App\Controller\api;

use App\Component\Pagination;
use App\Controller\GeneralApiController;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\Protocol;
use App\Model\Purpose;
use App\Model\User;
use App\Utils;
use Dflydev\Hawk\Server\UnauthorizedException;
use RuntimeException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

class GeneralAPI extends GeneralApiController
{
	/**
	 * @Route("/api/test", name="app_user_api_test", methods={"GET"})
	 * @return Response
	 */
	public function index(): Response
	{
		$user = User::loggedInUser();

		//Test Response erstellen
		$response = new JsonResponse([
			'username' => $user->getUsername(),
			'apiKey' => $user->getGeneralApiKey(),
		]);

		// Use the JSON_PRETTY_PRINT
		$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );

		//und raus...
		return $response;
	}

	/**
	 * @Route("/api/domains", name="app_user_api_domain_list", methods={"GET"})
	 * @return JsonResponse
	 */
	public function getDomains(): JsonResponse
	{
		//Alle Domains ziehen
		$domains = array_values(Domain::all());

		//Die Domainliste erstellen
		$payload = array_map(function (Domain $domain) {
			return [
				'id' => $domain->getId(),
				'name' => $domain->getName(),
				'whiteLabel' => $domain->hasWhitelabel(),
				'viewCount' => $domain->getViewCount(), // TODO: ViewCount aus Statistik-Model ziehen
			];
		}, $domains);

		//Test Response erstellen
		$response = new JsonResponse($payload);

		// Use the JSON_PRETTY_PRINT
		$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );

		//und raus...
		return $response;
	}

	/**
	 * @Route("/api/domains/{domainId}/consents", name="app_user_api_domain_consent_list", methods={"GET"})
	 * @param string $domainId
	 * @return JsonResponse
	 */
	public function getDomainConsent(string $domainId): JsonResponse
	{
		//Domain rausholen
		$domain = Domain::find($domainId);

		//User fürs Verhzeichnis
		$user = User::loggedInUser();

		//Falls invalid -> raus mit error
		if(empty($domain))
		{
			$response = new JsonResponse(["error" => "Invalid Domain"], 404);
		}
		//domain ok - dann daten holen...
		else{
			$protocolEntries = $this->getProtokollDataJson($domain);

			//Test Response erstellen
			$response = new JsonResponse($protocolEntries);

			// Use the JSON_PRETTY_PRINT
			$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );
		}

		//und raus...
		return $response;
	}

	/**
	 * @Route("/api/domains/{domainId}/consents/{ucid}", name="app_user_api_domain_consent_list_by_ucid", methods={"GET"})
	 * @param string $domainId
	 * @param string $ucid
	 * @return JsonResponse
	 */
	public function getDomainConsentByConsentId(string $domainId, string $ucid): JsonResponse
	{
		//Domain rausholen
		$domain = Domain::find($domainId);

		//User fürs Verhzeichnis
		$user = User::loggedInUser();

		//Falls invalid -> raus mit error
		if(empty($domain))
		{
			$response = new JsonResponse(["error" => "Invalid Domain"], 404);
		}
		//domain ok - dann daten holen...
		else{
			$protocolEntries = $this->getProtokollDataJson($domain, $ucid);

			//Test Response erstellen
			$response = new JsonResponse($protocolEntries);

			// Use the JSON_PRETTY_PRINT
			$response->setEncodingOptions( $response->getEncodingOptions() | JSON_PRETTY_PRINT );
		}

		//und raus...
		return $response;
	}


	/**
	 * @param Domain $domain
	 * @param string $ucid
	 * @return array<string,mixed>[]
	 */
	private function getProtokollDataJson(Domain $domain, string $ucid = ''): array
	{
		$ucid = trim($ucid);
		$protocolEntries = array();

		try {
			if (strlen($ucid) > 0) {
				//Search
				$protocol = Protocol::byDomain($domain);
				$protocolEntries = array_reverse($protocol->filterByConsentId($ucid));
			} else {
				$protocol = Protocol::byDomain($domain);
				$protocolEntries = iterator_to_array($protocol->all(true));
			}
		}
		catch (RuntimeException $exception) {
			$protocolEntries = array("error"=>'Logfile could not be read, check permissions.');
		}

		$pagination = new Pagination($protocolEntries, 20);
		if ($pagination->getCurrentPage() > 1 && $pagination->getItemCount() == 0) {
			$protocolEntries = array("error"=>'No items.');
		}

		$user = $domain->getUser();

		// TODO: Auch für MongoDB umsetzen
		//Verzeichnis der Logfiles und Archive
		$verz = Utils::getVarDir()."/user_data/".$user->getId()."/domains/".$domain->getId()."/";

		// Alles aus dem Verzeichnis holen an Dateien
		$files = scandir($verz, SCANDIR_SORT_ASCENDING);

		// wenn es welche gibt... könnte ja sein dass nicht...
		if(!empty($files))
		{
			//Dateien durchloopen
			foreach ($files as $file) {
				//Es handelt sich um eine consent Log Datei und ein Archiv
				if (stristr($file,"cm-consent") and stristr($file,".gz")) {
					// Nur dann auslesen wenn zip File jünger als 2 Tage...
					if (filemtime($verz.$file) > time()-(60*60*24*2)) {
						// Alle Zeilen holen
						$lines = gzfile($verz . $file);
						if (!empty($lines)) {
							// Zeilen loopen
							foreach ($lines as $jitem) {
								$item = json_decode($jitem,true);
								if (is_array($item)) {
									// Filtern nach UCID-Präfix
									if ($ucid !== '') {
										if (stripos($item['ucid'], $ucid) !== 0) {
											continue;
										}
									}
									// Anhängen
									$protocolEntries[]=$item;
								}
							}
						}
					}
				}
			}
		}

		// durchgehen und die purposes und embeddings anhängen...
		foreach ($protocolEntries as $k => $v)
		{
			$ubp = array();
			foreach (($v["purposes"] ?? []) as $pk => $pv)
			{
				if(!empty($pv))
				{
					$subPurpose = Purpose::find($pv);
					$ubp[$pv] = $subPurpose->getName();
					$protocolEntries[$k]["purposes"]=$ubp;
				}
			}

			$ube = array();
			foreach (($v["embeddings"] ?? []) as $pk => $pv)
			{
				if(!empty($pv)) {
					$embeddingOne = Embedding::find($pv);
					if(!empty($embeddingOne))
					{
						$ube[$pv] = $embeddingOne->getName();
						$protocolEntries[$k]["embeddings"] = $ube;
					}
				}
			}
		}
		return $protocolEntries;
	}

	/**
	 * @param Request $request
	 * @throws UnauthorizedException Bei ungültigem API-Schlüssel oder sonstigen Hawk-Fehlern
	 */
	public static function authorizeRequest(Request $request): void
	{
		//und korrekt sein...
		if ($user = User::loginByGeneralApiKey((string)$request->query->get('apiKey'))) {
			Domain::switchToUser($user);
			Domain::setDomainForRequest((string)$request->attributes->get('domainId'));
		}
		else {
			throw new UnauthorizedException('Invalid API key.');
		}
	}
}
