<?php
namespace App\Event;

use App\Model\User;
use Symfony\Contracts\EventDispatcher\Event;

/**
 * Event bei der Bearbeitung der Accountdaten eines Users
 *
 * Wird über `self::NAME_PRE_SAVE` vor dem Speichern und mit `self::NAME_POST_SAVE` nach dem
 * erfolgreichen Speichern verteilt.
 *
 * Mit `$this->isAccountInfoChanged()` und `$this->isPasswordChanged()` kann abgefragt werden, ob
 * (potentiell) die Stammdaten und/oder das Passwort geändert wurden.
 */
final class AccountUpdateEvent extends Event
{
	const NAME_PRE_SAVE = 'ccm19.account.update';
	const NAME_POST_SAVE = 'ccm19.account.updated';

	/** @var string CAUSE_SELFSERVICE Ausgelöst durch eigene Bearbeitung im Account-Menüpunkt */
	const CAUSE_SELFSERVICE = 'selfservice';
	/** @var string CAUSE_PWRECOVERY Ausgelöst durch Passwort-Vergessen-Funktion */
	const CAUSE_PWRECOVERY = 'pwrecovery';
	/** @var string CAUSE_ADMIN Ausgelöst durch Bearbeitung in der (Sub-)Userverwaltung */
	const CAUSE_ADMIN = 'admin';
	/** @var string CAUSE_API Ausgelöst durch API */
	const CAUSE_API = 'api';
	/** @var string CAUSE_CMDLINE Ausgelöst durch Kommandozeile */
	const CAUSE_CMDLINE = 'cmdline';

	/** @var User $user */
	private $user;
	/** @var string $cause */
	private $cause;
	/** @var bool $accountInfoChanged */
	private $accountInfoChanged = false;
	/** @var bool $passwordChanged */
	private $passwordChanged = false;

	public function __construct(User $user, string $cause)
	{
		$this->user = $user;
		$this->cause = $cause;
	}

	/**
	 * Benutzer, dessen Account-Daten aktualisiert wurden
	 */
	public function getUser(): User {
		return $this->user;
	}

	/**
	 * Anlass der Account-Datenänderung
	 *
	 * siehe die `CAUSE_*`-Konstanten in dieser Klasse.
	 */
	public function getCause(): string {
		return $this->cause;
	}

	public function isAccountInfoChanged(): bool {
		return $this->accountInfoChanged;
	}

	public function setAccountInfoChanged(): self {
		$this->accountInfoChanged = true;
		return $this;
	}

	public function isPasswordChanged(): bool {
		return $this->passwordChanged;
	}

	public function setPasswordChanged(): self {
		$this->passwordChanged = true;
		return $this;
	}
}
