<?php

namespace App\Event;

use App\Model\Model;
use LogicException;
use Symfony\Contracts\EventDispatcher\Event;

final class DomainClonePrepareEvent extends Event
{
	const NAME = 'ccm19.domain.clone.prepare';

	/** @var class-string<Model>[] $domainSpecificModels */
	private $domainSpecificModels;

	public function __construct()
	{
		$this->domainSpecificModels = [];
	}

	/**
	 * This method collects models whose data should be copied when a domain is cloned.
	 *
	 * Consider that the basic copy mechanism doesn't update model references.
	 * For example, if a plugin implements the two models `A` and `B` and one references
	 * the other, the objects of the referenced model should be copied by the plugin.
	 * This can be achieved by providing an event handler for `DomainCloneReferencesEvent`.
	 *
	 * @param class-string<Model> $className
	 * @return $this
	 * @throws LogicException if the class `$className` doesn't exist or isn't a domain specific model.
	 * @see DomainCloneReferencesEvent
	 */
	public function addDomainSpecificModel(string $className): self
	{
		if (!is_a($className, 'App\Model\Model', true)) {
			throw new LogicException("Class $className either doesn't exist or isn't of type App\\Model\\Model.");
		}
		elseif (!$className::isDomainSpecificModel()) {
			throw new LogicException("Class $className isn't a domain specific model.");
		}

		$this->domainSpecificModels[$className] = $className;
		return $this;
	}

	/**
	 * @return class-string<Model>[]
	 */
	public function getDomainSpecificModels(): array
	{
		return $this->domainSpecificModels;
	}
}
