<?php
namespace App\Event;

use App\Model\User;

/**
 * Event während des (regulären) Logins eines Nutzers.
 *
 * Account und Passwort wurden bereit auf Gültigkeit geprüft.
 *
 * Der Login-Versuch kann hier noch abgelehnt oder mit
 * 2FA erweitert werden.
 *
 * @phan-inherits FilterEvent<array<string,mixed>>
 * @extends FilterEvent<array<string,mixed>>
 */
class LoginEvent extends FilterEvent
{
	const NAME = 'ccm19.user.login';

	/** @var User $user */
	private $user;

	public function __construct(User $user, bool $allow = true)
	{
		parent::__construct(['disposition'=>($allow ? 'allow' : 'reject')]);
		$this->user = $user;
	}

	/**
	 * Benutzer, der angemeldet wird
	 */
	public function getUser(): User {
		return $this->user;
	}

	/**
	 * True, wenn die Anmeldung aktuell akzeptiert wird
	 */
	public function isDispositionAllow(): bool {
		return $this->getReturn()['disposition'] === 'allow';
	}

	/**
	 * True, wenn die Anmeldung aktuell abgelehnt wird
	 */
	public function isDispositionReject(): bool {
		return $this->getReturn()['disposition'] === 'reject';
	}

	/**
	 * True, wenn aktuell 2-Faktor-Authentifizierung nötig ist
	 */
	public function isDispositionTwoFactor(): bool {
		return $this->getReturn()['disposition'] === '2fa';
	}

	/**
	 * Begründung der Ablehnung
	 *
	 * @return string|null
	 */
	public function getRejectReason(): ?string {
		return $this->getReturn()['reason'] ?? null;
	}

	/**
	 * 2FA-Methoden
	 *
	 * @return array[]
	 */
	public function getSecondFactorAuths(): array {
		return array_values($this->getReturn()['2fa'] ?? []);
	}

	/**
	 * Anmeldung akzeptieren, auch wenn ein anderes Modul bereits abgelehnt oder 2FA aktiviert hat
	 *
	 * @return $this
	 */
	public function forceAllow(): self {
		return $this->setReturn(['disposition'=>'allow']);
	}

	/**
	 * Anmeldung ablehnen
	 *
	 * @param string|null $reason Begründung (optional)
	 * @return $this
	 */
	public function reject(?string $reason = null): self {
		return $this->setReturn(['disposition'=>'reject', 'reason'=>$reason]);
	}

	/**
	 * 2-Faktor-Authentifizierung hinzufügen
	 *
	 * Wenn die Anmeldung aktuell akzeptiert würde, wird auf 2FA umgestellt.
	 * Ist bereits eine 2FA aktiv, werden dem Nutzer beide zur Auswahl angeboten.
	 * Wird die Anmeldung aktuell abgelehnt, bleibt sie abgelehnt.
	 *
	 * @param string $name Anzeige-Name für die Auswahl bei mehreren 2FA-Methoden
	 * @param string $route Route, auf der die Anmeldung abgeschlossen werden kann
	 * @param array $routeParams Parameter für die Route (optional)
	 * @return $this
	 */
	public function addSecondFactorAuth(string $name, string $route, array $routeParams = []): self {
		if ($this->isDispositionAllow()) {
			return $this->forceSetSecondFactorAuth($name, $route, $routeParams);
		}
		elseif ($this->isDispositionTwoFactor()) {
			$authArray = $this->getReturn()['2fa'] ?? [];
			$authArray[$name] = ['name' => $name, 'route' => $route, 'routeParams' => $routeParams];
			return $this->setReturn([
				'disposition' => '2fa',
				'2fa' => $authArray,
			]);
		}
		else {
			return $this;
		}
	}

	/**
	 * 2-Faktor-Authentifizierung erzwingen
	 *
	 * Unabhängig davon, wie mit der Anmeldung vorher umgegangen wurde,
	 * wird danach nur diese 2FA-Methode verwendet.
	 *
	 * @param string $name Anzeige-Name für die Auswahl bei mehreren 2FA-Methoden
	 * @param string $route Route, auf der die Anmeldung abgeschlossen werden kann
	 * @param array $routeParams Parameter für die Route (optional)
	 * @return $this
	 */
	public function forceSetSecondFactorAuth(string $name, string $route, array $routeParams = []): self {
		return $this->setReturn([
			'disposition' => '2fa',
			'2fa' => [$name => ['name' => $name, 'route' => $route, 'routeParams' => $routeParams]],
		]);
	}
}
