<?php
namespace App\Event;

use App\Utils;
use Symfony\Contracts\EventDispatcher\Event;
use Twig\Loader\LoaderInterface;

final class TemplateRenderEvent extends Event
{
	const PREFIX = 'ccm19.template.render.';

	/** @var string $view */
	private $view;
	/** @var array<string,mixed> $parameters */
	private $parameters;

	public static function nameForView(string $view): string
	{
		return self::PREFIX.$view;
	}

	/**
	 * @param string $view
	 * @param array<string,mixed> $parameters
	 */
	public function __construct(string $view, array $parameters)
	{
		$this->view = $view;
		$this->parameters = $parameters;
	}

	/**
	 * Name des Views/des Templates
	 */
	public function getView(): string
	{
		return $this->view;
	}

	/**
	 * Name des zu rendernden Views/des Templates ändern
	 *
	 * Achtung: Das überschreibt u.u. Änderungen von anderen Plugins.
	 * Im Zweifel lieber extendTemplate() verwenden!
	 *
	 * @return $this
	 */
	public function setView(string $view): self
	{
		$this->view = $view;
		return $this;
	}

	private function quote(string $string): string
	{
		return strtr($string, ["'"=>"\\'", '"'=>'\"', "\\"=>"\\\\"]);
	}

	/**
	 * Erweitert das View über Blöcke aus dem
	 * angegeben View.
	 *
	 * @return $this
	 */
	public function extendTemplate(string $view): self
	{
		global $kernel;
		$dir = Utils::getCacheDir() . '/templates/';
		$filename = hash('sha256', ($this->getView() . '//' . $view)) . '.twig';
		$newView = '@temp/' . $filename;
		$path = $dir . $filename;

		/** @var LoaderInterface $loader */
		$loader = $kernel->getContainer()->get('twig')->getLoader();
		$source = $loader->getSourceContext($view);
		$sourcePath = $source->getPath();

		if (!file_exists($path) or ($sourcePath and @filemtime($sourcePath) > @filemtime($path))) {
			$source = $loader->getSourceContext($view);

			file_put_contents($path, '{% extends "'.$this->quote($this->getView()).'" %}'.$source->getCode());
		}
		$this->setView($newView);
		return $this;
	}

	/**
	 * View-Parameter
	 * @return array<string,mixed>
	 */
	public function getParameters(): array
	{
		return $this->parameters;
	}

	/**
	 * Ersetzen aller View-Parameter
	 * @param array<string,mixed> $parameters
	 * @return $this
	 */
	public function replaceParameters(array $parameters): self
	{
		$this->parameters = $parameters;
		return $this;
	}

	/**
	 * Parameter holen
	 * @param string $key
	 * @param mixed $default
	 * @return mixed
	 */
	public function get(string $key, $default=null)
	{
		return $this->parameters[$key] ?? $default;
	}

	/**
	 * Parameter setzen
	 * @param string $key
	 * @param mixed $value
	 * @return $this
	 */
	public function set(string $key, $value): self
	{
		$this->parameters[$key] = $value;
		return $this;
	}

	/**
	 * Mehrere Parameter auf einmal setzen
	 * @param array<string,mixed> $kv
	 * @return $this
	 */
	public function setMulti(array $kv): self
	{
		$this->parameters = array_merge($this->parameters, $kv);
		return $this;
	}
}
