<?php

namespace App\Event;

use App\Model\Domain;
use App\Model\LocaleBase;
use App\Model\Theme;
use App\Utils;
use LogicException;
use Symfony\Component\Routing\Exception\RouteNotFoundException;
use Symfony\Contracts\EventDispatcher\Event;

final class WidgetPluginsInitEvent extends Event
{
	const NAME = 'ccm19.widgetPlugins.init';

	/** @var Domain $domain */
	private $domain;
	/** @var Theme $theme */
	private $theme;
	/** @var LocaleBase|null $locale */
	private $locale;

	/** @var array $apiParams */
	private $apiParams;
	/** @var string $baseUrl */
	private $baseUrl;
	/** @var int $widgetLastModifiedTimestamp */
	private $widgetLastModifiedTimestamp;

	/** @var array[] $plugins */
	private $plugins;

	public function __construct(Domain $domain, Theme $theme, ?LocaleBase $locale, array $apiParams, string $baseUrl, int $widgetLastModifiedTimestamp)
	{
		$this->domain = $domain;
		$this->theme = $theme;
		$this->locale = $locale;

		$this->apiParams = $apiParams + [
			'theme' => $theme->getId(),
			'lang' => $locale ? $locale->getName() : 'de_DE', // Fallback, will be replaced dynamically
		];
		$this->baseUrl = rtrim($baseUrl, '/');
		$this->widgetLastModifiedTimestamp = $widgetLastModifiedTimestamp;

		$this->plugins = [];
	}

	/**
	 * @return Domain
	 */
	public function getDomain(): Domain
	{
		return $this->domain;
	}

	/**
	 * @return Theme
	 */
	public function getTheme(): Theme
	{
		return $this->theme;
	}

	/**
	 * Locale, if set explicitly.
	 *
	 * NULL if using automatic language detection at Javascript runtime.
	 *
	 * @return LocaleBase|null
	 */
	public function getLocale(): ?LocaleBase
	{
		return $this->locale;
	}

	/**
	 * @return string
	 * @throws LogicException if method isn't called from a proper plugin controller method.
	 */
	private function assertPluginNamespace(): string
	{
		$namespace = array_reduce(
			debug_backtrace(~DEBUG_BACKTRACE_PROVIDE_OBJECT & DEBUG_BACKTRACE_IGNORE_ARGS),
			function ($namespace, array $trace) {
				return $namespace ?? (preg_match('~^Plugins\\\\([^\\\\]+)~', $trace['class'] ?? '', $match) ? $match[1] : $namespace);
			}
		);

		if (!$namespace) {
			throw new LogicException('Couldn\'t determine plugin class name, call from plugin controller method only.');
		}
		elseif (array_key_exists($namespace, $this->plugins) == false) {
			// Reserved plugin properties: instance, loaded
			$this->plugins[$namespace] = [
				'name' => $namespace,
				'script' => null,
			];
		}

		return $namespace;
	}

	/**
	 * @param string $route
	 * @param array $parameters
	 * @param bool $appendVersion Adds `v={timestamp}` to the query string if `true`.
	 * @return $this
	 * @throws RouteNotFoundException
	 */
	public function setPluginScript(string $route, array $parameters=[], bool $appendVersion=true): self
	{
		$namespace = $this->assertPluginNamespace();
		$parameters = $appendVersion
			? $this->apiParams + ['v'=>$this->widgetLastModifiedTimestamp] + $parameters
			: $this->apiParams + $parameters;

		$this->plugins[$namespace]['script'] = $this->baseUrl . Utils::getRouter()->generate($route, $parameters);
		return $this;
	}

	/**
	 * This method returns an associative multidimensional array with its keys representing the respective plugin name.
	 * @return array[]
	 */
	public function getPluginInfo(): array
	{
		return $this->plugins;
	}
}
