<?php

namespace App\EventSubscriber;

use App\Component\Api\HostingApi;
use App\Component\Api\ClusterApi as ClusterApiComponent;
use App\Controller\AdminController;
use App\Controller\api\GeneralAPI;
use App\Controller\api\ClusterAPI;
use App\Controller\DomainManager;
use App\Controller\ExternalController;
use App\Controller\GeneralApiController;
use App\Controller\HostingApiController;
use App\Controller\HostingController;
use App\Controller\LicenseManager;
use App\Controller\Login;
use App\Controller\Setup;
use App\Controller\UniversalController;
use App\Controller\UnrestrictedController;
use App\Model\Domain;
use App\Model\User;
use App\Model\Index\DomainIndex;
use App\Utils;
use Dflydev\Hawk\Server\UnauthorizedException;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Bundle\WebProfilerBundle\Controller\ProfilerController;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Event\ControllerEvent;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

class AuthenticationHandler extends AbstractController implements EventSubscriberInterface
{
	/** @var bool $apiRequestAuthenticated */
	private $apiRequestAuthenticated;

	/** @var bool $domainAutodetected */
	private $domainAutodetected;

	/** @var ClusterAPI|GeneralApiController|HostingApiController|null $apiRequestController */
	private $apiRequestController;

	/** @var Response|null $response */
	private $response;

	public function __construct()
	{
		$this->apiRequestAuthenticated = false;
		$this->domainAutodetected = false;
		$this->apiRequestController = null;
		$this->response = null;
	}

	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			KernelEvents::CONTROLLER => ['onKernelController', 9999],
			KernelEvents::RESPONSE => ['onKernelResponse', 9999],
		];
	}

	/**
	 * Diese Methode gibt die generierte Response zurück, sofern die Anfrage nicht zur Bearbeitung freigegeben wurde.
	 * @return Response
	 */
	public function onUnauthorizedRequest(): Response
	{
		return $this->response;
	}

	/**
	 * Diese Methode überprüft, ob der Seitenbesucher eingeloggt ist und unternimmt u. U. entsprechende Maßnahmen.
	 *
	 * @param ControllerEvent $event
	 * @phan-suppress PhanTypeArraySuspicious
	 */
	public function onKernelController(ControllerEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}

		$authorize = true;
		$controller = $event->getController()[0];

		// Symfony Profiler
		if ($controller instanceof ProfilerController) {
			return;
		}

		$request = $event->getRequest();
		$route = $request->attributes->get('_route');
		$stateless = $request->attributes->getBoolean('_stateless');
		$user = ($stateless) ? null : User::loggedInUser();
		$method = $request->getMethod();

		$request->attributes->set('area', 'backend');

		// Alle externen Anfragen zulassen, sofern ein gültiger API-Schlüssel beiliegt
		if ($controller instanceof ExternalController) {
			$request->attributes->set('area', 'frontend');
			$apiKey = (string)$request->query->get('apiKey');

			// Fallback: In alten Systemen ohne API-Key und nur einem Benutzer, diesen Benutzer verwenden
			if (empty($apiKey) && Utils::isBaseEdition() && User::count() == 1) {
				/** @var User $user */
				$user = array_values(User::all())[0];
				$apiKey = $user->getApiKey();
			}

			// Domainchecks nur bei gültigem Benutzer durchführen
			if (User::setApiUser($apiKey)) {
				// Aktiven Benutzer und Domaindatenbank umschalten
				$user = User::loggedInUser();
				assert($user !== null);
				Domain::switchToUser($user);

				$domainId = (string)$request->query->get('domain');

				// Domain-Autoerkennung
				if (!$domainId and Utils::isExtendedEdition()) {
					$this->domainAutodetected = true;
					$domainName = Utils::getOriginDomain($request, false) ?? '';
					$domainId = DomainIndex::findDomainId($domainName);
					// Domain erstellen, wenn nicht bekannt, Feature aktiv und noch genug Domains im Account
					if (!$domainId && $domainName && $user->isAutoAddDomains() && ($user->getMaxDomainCount() == -1 || $user->getActualDomainCount() < $user->getMaxDomainCount())) {
						$domain = Domain::createForUser($user)
							->setName($domainName)
							->setFrontendWidgetEnabled(false)
							->setManagementStructurePreLegalConsent(true)
							->setManagementStructure(Domain::MANAGEMENT_STRUCTURE_EMBEDDING)
						;
						$domain->save();
						$domainId = $domain->getId();
						DomainIndex::rebuild();
						$user->updateDomainInfo();
						$user->save();
					}
				}

				// Fallback: Erste für alte Systeme nachgepflegte Domain verwenden
				if (empty($domainId) && Utils::isBaseEdition() && Domain::count() == 1) {
					/** @var Domain $domain */
					$domain = array_values(Domain::all())[0];
					$domainId = $domain->getId();
				}

				if (Domain::setDomainForRequest($domainId)) {
					return;
				}
				else {
					$this->response = new JsonResponse([
						'error' => '400 Bad Request',
						'description' => 'invalid domain',
					], 400);
					$this->response->headers->set('X-Robots-Tag', 'noindex, nofollow');
					$authorize = false;
				}
			}
			else {
				$this->response = new JsonResponse([
					'error' => '400 Bad Request',
					'description' => 'invalid api key',
				], 400);
				$this->response->headers->set('X-Robots-Tag', 'noindex, nofollow');
				// Der API-Key wird zufällig generiert. Ein ungültiger API-Key wird also
				// mit extrem hoher Wahrscheinlichkeit nie gültig => ein paar Minuten die
				// Fehlermeldung cachen ist kein Problem.
				Utils::resetCacheControl($this->response);
				$this->response->setCache(['max_age'=>300, 'public'=>true]);
				$authorize = false;
			}
		}

		// Zugriffe auf den Login- und Setup-Bereich immer zulassen
		else if ($controller instanceof Login || $controller instanceof Setup) {
			$authorize = true;
		}

		// Zugriffe auf unbeschränkte Bereiche zulassen
		else if ($controller instanceof UnrestrictedController) {
			$request->attributes->set('area', 'unrestricted');
			return;
		}

		// Zugriffe auf Cluster-API
		else if ($controller instanceof ClusterAPI) {
			$request->attributes->set('area', 'api');
			$this->apiRequestController = $controller;
			$this->apiRequestAuthenticated = false;

			try {
				ClusterApiComponent::getInstance()->authorizeRequest($request);
				$this->apiRequestAuthenticated = true;
				return;
			}
			catch (UnauthorizedException $exception) {
				$this->response = $this->json([
					'error' => 'Unauthorized: '.$exception->getMessage(),
				], 401);
				$authorize = false;
			}
		}
		// Zugriffe auf API der Hostingversion
		else if ($controller instanceof HostingApiController) {
			$request->attributes->set('area', 'api');
			$this->apiRequestController = $controller;
			$this->apiRequestAuthenticated = false;

			try {
				HostingApi::getInstance()->authorizeRequest($request);
				$this->apiRequestAuthenticated = true;
				return;
			}
			catch (UnauthorizedException $exception) {
				$this->response = $this->json([
					'error' => 'Unauthorized: '.$exception->getMessage(),
				], 401);
				$authorize = false;
			}
		}

		else if ($controller instanceof GeneralApiController) {
			$request->attributes->set('area', 'api');
			$this->apiRequestController = $controller;
			$this->apiRequestAuthenticated = false;

			try {
				GeneralApi::authorizeRequest($request);
				$this->apiRequestAuthenticated = true;
				return;
			}
			catch (UnauthorizedException $exception) {
				$this->response = $this->json([
					'error' => 'Unauthorized: '.$exception->getMessage(),
				], 401);
				$authorize = false;
			}
		}

		else if ($user) {
			// Session-Timestamp aktualisieren
			$session = Utils::getSession();
			$session->set('last_timestamp', (int)time());

			$isAdmin = $user->getRole() == User::ROLE_ADMIN;
			$hasAdminAccess = $isAdmin || $user->isManager();

			// Administratoren bei invalider Lizenz zur Lizenzverwaltung leiten; Kunden ausloggen
			if (Utils::isLicensed() == false) {
				if ($isAdmin || $hasAdminAccess && Utils::mayUserAccessRoute($user, 'app_license_manager_index', 'GET')) {
					if ($controller instanceof LicenseManager == false) {
						$this->response = $this->redirectToRoute('app_license_manager_index');
						$authorize = false;
					}
				}
				else {
					User::logout();
					$this->addFlash('alert', 'License has expired.');
					$this->response = $this->redirectToRoute('app_login');
					$authorize = false;
				}
			}

			// Administratoren den Zugriff auf den Update-Bereich/Systemeinstellungen generell gewähren
			else if ($controller instanceof AdminController) {
				if (!($isAdmin || $hasAdminAccess && Utils::mayUserAccessRoute($user, $route, $method))) {
					$this->response = $this->redirectToRoute('app_root');
					$authorize = false;
				}
			}
			// Universelle Bereiche, wie z. B. die Kontoeinstellungen
			else if ($controller instanceof UniversalController) {
				if (Utils::mayUserAccessRoute($user, $route, $method) === false) {
					if (Utils::mayUserAccessRoute($user, $route, 'GET')) {
						$this->addFlash('alert', 'You only have read access to this section.');
					}
					$this->response = $this->redirectToRoute('app_root');
					$authorize = false;
				}
			}
			else if ($controller instanceof DomainManager) {
				if (Utils::mayUserAccessRoute($user, $route, $method) === false) {
					if ($user->isSubUser() and Utils::mayUserAccessRoute($user, $route, 'GET') === true) {
						$this->addFlash('alert', 'You only have read access to this section.');
					}
					$this->response = (Utils::isExtendedEdition() and $hasAdminAccess)
						? $this->redirectToRoute('app_hosting_dashboard')
						: $this->redirectToRoute('app_root');
					$authorize = false;
				}
			}
			else if (Utils::isExtendedEdition()) {
				$wrongControllerType = ($hasAdminAccess xor $controller instanceof HostingController);
				if ((Utils::mayUserAccessRoute($user, $route, $method) === false) or $wrongControllerType) {
					if (!$wrongControllerType and Utils::mayUserAccessRoute($user, $route, 'GET')) {
						$this->addFlash('alert', 'You only have read access to this section.');
					}
					$this->response = $hasAdminAccess
						? $this->redirectToRoute('app_hosting_dashboard')
						: $this->redirectToRoute('app_root');
					$authorize = false;
				}
			}
			// Zugriff auf Hostingbereiche in der Basisversion verwehren
			else if ($controller instanceof HostingController) {
				$this->response = $this->redirectToRoute('app_root');
				$authorize = false;
			}
		}
		// Zugriff verwehren, sofern sich der Besucher weder im Login- noch im Setup-Bereich befindet
		else if ($controller instanceof Login == false && $controller instanceof Setup == false) {
			$this->response = $this->redirectToLogin($request);
			$authorize = false;
		}

		if ($authorize == false) {
			$event->setController([$this, 'onUnauthorizedRequest']);
		}
	}

	public function onKernelResponse(ResponseEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}

		/** @var Response $response */
		$response = $event->getResponse();

		/** @var Request $request */
		$request = $event->getRequest();

		if ($this->domainAutodetected) {
			$vary = $response->getVary();
			if (!in_array('Origin', $vary)) {
				$vary[] = 'Origin';
			}
			if (!$request->headers->has('Origin')) {
				if (!in_array('Referer', $vary)) {
					$vary[] = 'Referer';
				}
			}
			$response->setVary($vary);
		}

		if ($this->apiRequestController && $this->apiRequestAuthenticated) {
			if ($this->apiRequestController instanceof HostingApiController) {
				HostingApi::getInstance()->signResponse($response);
			}
			elseif ($this->apiRequestController instanceof ClusterAPI) {
				ClusterApiComponent::getInstance()->signResponse($response);
			}
		}
	}

	/**
	 * Weiterleiten zum Login
	 *
	 * @param Request $request
	 * @return RedirectResponse
	 */
	private function redirectToLogin(Request $request): RedirectResponse
	{
		$route = $request->attributes->get('_route', 'app_root');
		$routeParams = $request->attributes->get('_route_params', []);

		if ($request->getMethod() === 'GET' and !in_array($route, ['app_root', 'app_hosting_dashboard', 'app_logout', 'app_logout_post', 'app_login', 'app_login_post'])) {
			return $this->redirectToRoute('app_login', [
				'backTo'=>Utils::encodeReturnTo((string)$route, (array)$routeParams),
			]);
		}
		else {
			return $this->redirectToRoute('app_login', [], 303);
		}
	}
}
