<?php

namespace App\EventSubscriber;
use App\Config;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Session\Session;
use Symfony\Component\HttpFoundation\Session\Storage\MockArraySessionStorage;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\KernelEvents;

class BootHandler implements EventSubscriberInterface
{
	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			// Run after \Symfony\Component\HttpKernel\EventListener\RouterListener::onKernelRequest
			KernelEvents::REQUEST => ['onKernelRequest', 31],
		];
	}

	/**
	 * Am Anfang einer Anfrage wird die Session gestartet.
	 * @param RequestEvent $event
	 */
	public function onKernelRequest(RequestEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}

		global $kernel;
		$container = $kernel->getContainer();

		/** @var Request $request */
		$request = $event->getRequest();

		// Basis-URL vervollständigen wenn nötig
		if (empty($_SERVER['HTTP_HOST']) or empty($_SERVER['SCRIPT_NAME'])) {
			$config = Config::getInstance();
			if ($baseUrl = $config->getBaseUrl()) {
				$baseUrlParts = parse_url($baseUrl);
				$path = $baseUrlParts['path'] ?? '';

				$_SERVER['SCRIPT_NAME'] = stripos($path, 'index.php') !== false ? $path : rtrim($path, '/').'/index.php';
				$request->server->set('SCRIPT_NAME', $_SERVER['SCRIPT_NAME']);
				$_SERVER['HTTP_HOST'] = $baseUrlParts['host'];
				$request->server->set('HTTP_HOST', $_SERVER['HTTP_HOST']);
				if (!empty($baseUrlParts['port'])) {
					$_SERVER['HTTP_HOST'] .= ':'.$baseUrlParts['port'];
				}
				$request->server->set('HTTP_HOST', $_SERVER['HTTP_HOST']);
				if (strtolower($baseUrlParts['scheme']) === 'https') {
					$_SERVER['HTTPS'] = 'on';
					$request->server->set('HTTPS', 'on');
				}
			}
		}

		// Hotfix zur Erkennung von HTTPS über Proxies, wie CloudFlare etc.
		// Damit wird 'X-Forwarded-Proto' immer akzeptiert, solange keine eigenen
		// TRUSTED_PROXIES in .env.local gesetzt wurden.
		if (!isset($_ENV['TRUSTED_PROXIES']) or $_ENV['TRUSTED_PROXIES'] === '127.0.0.0/8,::1/128') {
			Request::setTrustedProxies(['127.0.0.1', $request->server->get('REMOTE_ADDR')], Request::HEADER_X_FORWARDED_PROTO);
		}

		// Routen bestimmen, für die keine Session geöffnet werden muss
		$path = $request->getPathInfo();
		$noSession = (
			$_ENV['APP_ENV'] == 'test' or
			$request->attributes->getBoolean('_stateless')
		);

		if ($noSession and !$request->hasSession(true)) {
			// Wenn keine Session benötigt wird, eine Dummy-Session vorbereiten
			$request->setSession(new Session(new MockArraySessionStorage()));
		}
		else {
			// Sicherheits-Parameter fine-tunen
			// @phan-suppress-next-line PhanTypeMismatchArgumentInternal
			ini_set('session.cookie_secure', $request->isSecure());
			ini_set('session.cookie_path', rtrim($request->getBasePath(), '/').'/');
		}
	}
}
