<?php

namespace App\EventSubscriber;

use App\Component\Api\ClusterApi;
use App\Component\Cache\Cache;
use App\Controller\AbstractController;
use App\Controller\DomainDependantController;
use App\Controller\Hosting\SystemThemeController;
use App\Controller\HostingController;
use App\Controller\UserLocale as UserLocaleController;
use App\Controller\UserThemeManager;
use App\Event\CacheResetEvent;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\LegalText;
use App\Model\Locale;
use App\Model\Model;
use App\Model\Placeholder;
use App\Model\Theme;
use App\Model\User;
use App\Model\UserLocale;
use App\Utils;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ControllerEvent;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

class CacheHandler implements EventSubscriberInterface
{
	/** @var AbstractController|null $controller */
	private $controller;

	public function __construct()
	{
		$this->controller = null;
	}

	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			KernelEvents::CONTROLLER => ['onKernelController', 8000],
			KernelEvents::RESPONSE => ['onKernelResponse', 8000],
		];
	}

	/**
	 * @phan-suppress PhanTypeArraySuspicious
	 */
	public function onKernelController(ControllerEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}

		$this->controller = $event->getController()[0];

		// Reset class names so that only models modified by the controller are tracked.
		Model::clearWrittenModelsClassNames();
	}

	/**
	 * Prüft, ob es sich um einen Controller handelt, der potentiell Aktionen durchführt,
	 * die ein Leeren des Frontend-Caches erforderlich machen.
	 *
	 * @param object|null $controller
	 * @return bool
	 */
	private function isRelevantController($controller): bool
	{
		return $controller !== null && (
			$controller instanceof DomainDependantController
			or $controller instanceof SystemThemeController
			or $controller instanceof UserLocaleController
			or $controller instanceof UserThemeManager
			or true // Ignore the controller type, only check for relevant models
		);
	}

	/**
	 * Hier wird der Widget-Cache einer Domain geleert, falls
	 * domainspezifische Einstellungen bearbeitet wurden.
	 * @return void
	 */
	public function onKernelResponse(ResponseEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}

		$request = $event->getRequest();

		if (
			in_array($request->getMethod(), ['POST', 'PUT', 'PATCH', 'DELETE'])
			and $this->isRelevantController($this->controller)
		) {
			// Gather available models that are considered relevant for cache invalidation
			$cacheResetEvent = new CacheResetEvent();
			Utils::dispatchEvent(CacheResetEvent::NAME, $cacheResetEvent);

			// May switch clearing from domain cache to user cache
			$clearCacheScope = $request->attributes->getAlpha('clearCacheScope');

			// Trigger cache refresh after update to the user
			if ($clearCacheScope == 'user') {
				$cacheResetEvent->addRelevantModel(User::class);
			}

			// Check if any model relevant to app.js was changed
			$relevantModels = array_intersect([
				Domain::class,
				Embedding::class,
				LegalText::class,
				Locale::class,
				Placeholder::class,
				UserLocale::class,
				Theme::class,
			] + $cacheResetEvent->getRelevantModels(), Model::getWrittenModelsClassNames());

			if ($relevantModels) {
				$user = User::loggedInUser();
				$domain = Domain::activeDomain();
				$owner = $user ? $user->getAccountOwner() : null;

				if ($this->controller instanceof DomainDependantController and $domain and $clearCacheScope != 'user') {
					$domain->touchFrontendWidgetCacheTimestamp()->save();
					Cache::clear($domain, true);

					// Broadcast clear cache request to other peers of this cluster
					ClusterApi::getInstance()->callRemoteApiBroadcast('app_cluster_api_domain_cache_clear', [
						'userId' => $owner->getId(),
						'domainId' => $domain->getId(),
					], 'DELETE', ['apiKey' => $owner->getApiKey(), 'keepMetadata'=>true]);
				} elseif ($this->controller instanceof HostingController) {
					try {
						Domain::touchAllFrontendWidgetCacheTimestampsGlobally();
					} finally {
						Cache::clearAll();

						// Broadcast clear cache request to other peers of this cluster
						ClusterApi::getInstance()->callRemoteApiBroadcast('app_cluster_api_frontend_cache_clear', [], 'DELETE');
					}
				} elseif ($user) {
					foreach (Domain::all() as $domain) {
						$domain->touchFrontendWidgetCacheTimestamp();
					}
					Domain::flush();
					Cache::clearUser($owner);

					// Broadcast clear cache request to other peers of this cluster
					ClusterApi::getInstance()->callRemoteApiBroadcast('app_cluster_api_user_cache_clear', [
						'userId' => $owner->getId(),
					], 'DELETE', ['apiKey' => $owner->getApiKey()]);
				}
			}
		}

		// Store pending responses
		Cache::storeResponseExecuteQueue();
	}
}
