<?php

namespace App\EventSubscriber;

use App\Utils;
use Symfony\Bundle\FrameworkBundle\Command\CacheClearCommand;
use Symfony\Component\Console\ConsoleEvents;
use Symfony\Component\Console\Event\ConsoleCommandEvent;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

class CommandSanityHandler implements EventSubscriberInterface
{
	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			ConsoleEvents::COMMAND => ['onCommand', 9999],
		];
	}

	private function getEffectiveUid(): int
	{
		if (function_exists('posix_geteuid')) {
			return (int)posix_geteuid();
		} else {
			return (int)getmyuid();
		}
	}

	private function getEffectiveGid(): int
	{
		if (function_exists('posix_getegid')) {
			return (int)posix_getegid();
		} else {
			return (int)getmygid();
		}
	}

	/**
	 * @param string $path
	 * @return array{uid:int, gid:int}
	 */
	private function statFile(string $path): array {
		$result = @stat($path);
		return ($result === null || $result === false) ? ['uid'=>0, 'gid'=>0] : $result;
	}

	/**
	 * Überprüft ob der aktuelle System-Benutzer der ist, dem die Dateien von CCM19
	 * gehören, um Dateirechte-Probleme zu vermeiden.
	 *
	 * Wenn der Prozess Root-Rechte hat, wird die EUID/EGID gewechselt, ansonsten
	 * eine Sicherheitsabfrage angezeigt.
	 *
	 * @param ConsoleCommandEvent $event
	 */
	public function onCommand(ConsoleCommandEvent $event): void
	{
		$input = $event->getInput();
		$output = $event->getOutput();
		$command = $event->getCommand();

		if (!empty($_SERVER['CCM19_NO_SANITY_CHECK'])) {
			return;
		}

		// Ignoriere --help
		if ($input->hasParameterOption(['--help', '-h'], true)) {
			return;
		}

		$io = new SymfonyStyle($input, $output);

		$cacheFile = Utils::getVarDir().'/cache/CACHEDIR.TAG';
		$envFile = Utils::getBaseDir().'/.env.local';

		$stat = $this->statFile($cacheFile);
		$stat2 = $this->statFile($envFile);

		// Wenn die Datei Root gehört (oder wir auf einem Windows-System laufen), können wir auch nichts mehr tun.
		if ($stat['uid'] === 0) {
			return;
		}

		if ($stat2 and $stat and ($stat['uid'] !== $stat2['uid'] or $stat['gid'] !== $stat2['gid'])) {
			$io->warning('Inconsistent filesystem permissions (`cache/CACHEDIR.TAG` and `.env.local` are not owned by the same user and group). This might cause problems.');
			sleep(1);
		}

		// Effektive Gruppen-ID wechseln sofern möglich
		if ($stat['gid'] !== 0 and $this->getEffectiveGid() !== $stat['gid'] and function_exists('posix_setegid')) {
			@posix_setegid($stat['gid']);
		}

		// Effektive User-ID wechseln bzw. Benutzer auf Probleme hinweisen
		$euid = $this->getEffectiveUid();
		$currentName = addcslashes(function_exists('posix_getpwuid') ? posix_getpwuid($euid)['name'] ?? (string)$euid : $euid, "\0..\37\"");
		$ownerName = addcslashes(function_exists('posix_getpwuid') ? posix_getpwuid($stat['uid'])['name'] ?? (string)$stat['uid'] : $stat['uid'], "\0..\37\"");
		$name = posix_getpwuid($stat['uid'])['name'] ?? (string)$stat['uid'];
		if ($euid !== $stat['uid']) {
			if (function_exists('posix_seteuid') and @posix_seteuid($stat['uid'])) {
				$name = posix_getpwuid($stat['uid'])['name'] ?? (string)$stat['uid'];
				$io->note('Acting as user "'.$ownerName.'" instead of "'.$currentName.'".');
			} else {
				// Spezialfall cache:clear
				if ($command instanceof CacheClearCommand) {
					if ($input->hasParameterOption(['--no-warmup'], true)) {
						return;
					} else {
						$io->note('CCM19 seems to be managed by system user "'.$ownerName.'" but you are logged in as "'.$currentName.'". Skipping cache warmup.');
						$input->setOption('no-warmup', true);
						// BEGIN WORKAROUND: Symfony überschreibt hier gesetzte Optionen, also hacky manuell korrigieren
						$argv = $_SERVER['argv'];
						array_shift($argv);
						$argv[] = '--no-warmup';
						$reflObj = new \ReflectionObject($input);
						$reflMet = $reflObj->getMethod('setTokens');
						$reflMet->setAccessible(true);
						$reflMet->invoke($input, $argv);
						// END WORKAROUND
					}
				}
				else {
					$io->warning('CCM19 seems to be managed by system user "'.$ownerName.'" but you are logged in as "'.$currentName.'". This may break things.');
					if (!$io->confirm('Do you still want to continue?', false)) {
						$event->disableCommand();
						return;
					}
				}
			}
		}
	}
}
