<?php

namespace App\EventSubscriber;

use App\Controller\DomainDependantController;
use App\Model\Domain;
use App\Model\User;
use App\Model\Locale;
use App\Model\LegalText;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\Theme;
use App\Utils;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\ControllerEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\Routing\RequestContext;

class DomainSelectionHandler extends AbstractController implements EventSubscriberInterface
{
	/** @var RequestContext|null $requestContext */
	private $requestContext;

	public function __construct(RequestContext $requestContext = null)
	{
		$this->requestContext = $requestContext;
	}

	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			KernelEvents::CONTROLLER => ['onKernelController', 9990],
		];
	}

	/**
	 * Setzt die Domain in der Session auf die in der URL und setzt die Domain
	 * als Standardparameter fürs Routing.
	 *
	 * @param ControllerEvent $event
	 * @phan-suppress PhanTypeArraySuspicious
	 */
	public function onKernelController(ControllerEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}

		$controller = $event->getController()[0];
		$request = $event->getRequest();

		if ($controller instanceof DomainDependantController) {

			//Falls es sich um einen Subuser handelt, dem dann die Domain zuweisen damit die Daten bearbeitet werden können...
			$user = User::loggedInUser();
			$activeDoms = $user->getDomActive();

			//Domainid holen
			$domainId = $request->attributes->get('_domainId');

			//check ob Rechte an der Domain bestehen bei Subusern...
			if ($user->isSubUser()) {
				$domRights = false;
				foreach ($activeDoms as $k=>$v)
				{
					if($k == $domainId)
					{
						$domRights = true;
					}
				};
			}
			//Hauptuser - dann immer volle Rechte...
			else{
				$domRights = true;
			}

			$domain = Domain::find((string)$domainId);
			$session = $request->getSession();
			$lastDomain = $session->get('domain', '');

			//In der Base Edition gibts nur 1 Domain - daher immer true - sonst machts ja keinen Sinn... steht ja auch so in allen Accounts..
			if (Utils::isBaseEdition()) {
				$domRights = true;
			}

			// Wenn falsche Domain in URL entweder zur einzig richtigen weiterleiten oder zur Domainauswahl
			if (!$domain || $domRights==false) {
				if (Utils::isExtendedEdition()) {
					$this->addFlash('danger', ('No Valid Domain for you!'));
					$event->setController([$this, 'redirectToDomainSelection']);
				}
				elseif (Domain::count() > 0) {
					$domain = current(Domain::all());
					Domain::select($domain->getId());
					$event->setController([$this, 'redirectToDomain']);
				}
			}
			elseif ($lastDomain !== $domainId) {
				$request->getSession()->set('domain', $domainId);

				// Models neu laden, die vielleicht schon in BootHandler für die falsche Domain geöffnet wurden aber nur dann wenn nicht default - dann läuft grad Setup
				if ($lastDomain !== '' && $domainId !=='default') {
					Locale::reloadRepository();
					LegalText::reloadRepository();
					Embedding::reloadRepository();
					EmbeddingAsset::reloadRepository();
					Theme::reloadRepository();
				}
			}
		}

		$stateless = $request->attributes->getBoolean('_stateless');
		if (!$stateless) {
			try {
				$domainId = $request->getSession()->get('domain');
			} catch (\Exception $ignore) {
				$domainId = '';
			}
		} elseif (!isset($domainId)) {
			$domainId = '';
		}

		// _domainId bei Routen-Erstellung vorgeben (für Redirects und Routen in Twig-Templates)
		$this->requestContext->setParameter('_domainId', ($domainId != "") ? $domainId : 'default');
	}

	/**
	 * Weiterleiten zur Route mit richtiger Domain
	 *
	 * @see \App\Component\DomainDependantRoutingLoader::load()
	 * @param Request $request
	 * @return RedirectResponse
	 */
	public function redirectToDomain(Request $request): RedirectResponse
	{
		$route = $request->attributes->get('_route', 'app_dashboard');
		$routeParams = $request->attributes->get('_route_params', []);
		if (strrpos($route, '.nodomain') == strlen($route)-9) {
			$route = substr($route, 0, -9);
		}

		$domain = Domain::activeDomain();
		if ($domain == null) {
			if (Utils::isExtendedEdition()) {
				return $this->redirectToRoute('app_domains_index', [], 303);
			}
			else {
				$domain = array_values(Domain::all())[0];
			}
		}
		$routeParams['_domainId'] = $domain->getId();

		return $this->redirectToRoute($route, $routeParams, 307);
	}

	/**
	 * Weiterleiten zur Domainauswahl
	 *
	 * @param Request $request
	 * @return RedirectResponse
	 */
	public function redirectToDomainSelection(Request $request): RedirectResponse
	{
		$route = 'app_domains_index';
		$routeParams = [];

		return $this->redirectToRoute($route, $routeParams, 303);
	}
}
