<?php

namespace App\EventSubscriber;

use App\Utils;
use App\Component\Mail\MailMessage;
use App\Model\User;
use App\Model\Domain;
use ErrorException;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ExceptionEvent;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\Mailer\MailerInterface;
use Throwable;

class ExceptionHandler implements EventSubscriberInterface
{
	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		$result = [
			KernelEvents::EXCEPTION => ['onKernelException', 1],
		];
		if (($_ENV['APP_ENV'] ?? 'prod') === 'dev') {
			$result[KernelEvents::RESPONSE] = ['onKernelResponse', 2];
		}
		return $result;
	}

	/**
	 * Bei Exceptions eine E-Mail versenden, wenn die Environment-Variable
	 * APP_EXCEPTION_MAIL auf eine E-Mail-Adresse gesetzt ist.
	 *
	 * @param ExceptionEvent $event
	 */
	public function onKernelException(ExceptionEvent $event): void
	{
		global $kernel;

		if (empty($_ENV['APP_EXCEPTION_MAIL'])) {
			return;
		}

		try {
			$exception = $event->getThrowable();

			if ($exception instanceof NotFoundHttpException) {
				// 404 ist nicht interessant.
				return;
			}

			/** @var MailerInterface $mailer */
			$mailer = $kernel->getContainer()->get('mailer');

			$root = $kernel->getProjectDir().'/';
			$file = preg_replace('~^'.preg_quote($root, '~').'~', '', $exception->getFile());
			$line = $exception->getLine();

			if ($exception instanceof ErrorException) {
				$msg = trim(explode("\n", $exception->getMessage(), 2)[0]);
				if (strlen($msg) > 100) {
					$msg = substr($msg, 0, 97).'...';
				}

				$subject = "[CCM19] [Exception] $msg at $file:$line";
				$body = (string)$exception."\n";
			} else {
				$subject = '[CCM19] [Exception] '.get_class($exception)." at $file:$line";
				$body = (string)$exception."\n";
			}

			$user = User::loggedInUser();
			if ($user) {
				$body .= "\nUser: ".$user->getUsername()."\n";

				$domain = Domain::activeDomain();
				if ($domain) {
					$body .= "Domain: ".trim($domain->getName())."\n";
				}
			}

			// Pfade im Text kürzen
			$body = preg_replace('~(\s)'.preg_quote($root, '~').'~', '$1', $body);

			// Footer hinzufügen
			$body .= "\n-- \nAutomatic mail sent by your CCM19 installation at ".$_SERVER['HTTP_HOST'].".\n";

			$mail = new MailMessage($subject);

			$mail->autoSubmitted();
			$mail->listId('exceptions');
			$mail->to($_ENV['APP_EXCEPTION_MAIL']);
			$mail->text($body);
			$mail->voidReplyTo();

			$mailer->send($mail);
		}
		catch (Throwable $e) {
			// ignore
		}
	}

	/**
	 * Injecten des Cache-Clear-Buttons in Exception-Debug-Seiten
	 *
	 * Keine schöne Lösung, aber es funktioniert.
	 */
	public function onKernelResponse(ResponseEvent $event): void
	{
		// Nur auf Sub-Requests arbeiten (Exception-Seiten in Symfony werden als Subrequest verarbeitet)
		if ($event->isMainRequest()) {
			return;
		}
		$response = $event->getResponse();
		if ($response->getStatusCode() === 500 and !Utils::isCCMCloud() and !Utils::isClickskeksCloud()) {
			$text = $response->getContent();
			if (strpos($text, '<div class="exception-summary') !== false) {
				$request = $event->getRequest();
				$token = Utils::getRescueToken();
				$url = $request->getBasePath() . '/rescue.php?token='.$token;
				$response->setContent(str_replace('</body>', '<form action="'.htmlspecialchars($url).'" method="POST" style="text-align:center"><button type="submit" name="action" value="clear_caches">Clear all caches</button></form></body>', $text));
			}
		}
	}
}
