<?php

namespace App\EventSubscriber;

use App\Component\ConfigImporterContext;
use App\Component\LocaleService;
use App\Config;
use App\Exception\LocaleAlreadyExistsException;
use App\Model\Model;
use App\Model\Cookie;
use App\Model\Domain;
use App\Model\Embedding;
use App\Model\EmbeddingAsset;
use App\Model\FoundCookie;
use App\Model\FoundObject;
use App\Model\FoundScript;
use App\Model\FoundStorage;
use App\Model\LegalText;
use App\Model\LocaleBase;
use App\Model\Locale;
use App\Model\Protocol;
use App\Model\Purpose;
use App\Model\UserLocale;
use App\Model\SubGroups;
use App\Model\SystemMigrationState;
use App\Model\SystemTheme;
use App\Model\Theme;
use App\Model\User;
use App\Model\UserMigrationState;
use App\Model\UserTheme;
use App\Model\Index\DomainIndex;
use App\Utils;
use Exception;
use MongoDB\Driver\Exception\WriteException;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\ControllerEvent;
use Symfony\Component\HttpKernel\KernelEvents;

class MigrationHandler implements EventSubscriberInterface
{
	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			KernelEvents::CONTROLLER => ['onKernelController', 8000],
		];
	}

	public function updateRoutines(Request $request): void
	{
		$user = User::loggedInUser();
		$config = Config::getInstance();
		$purposes = Purpose::all();

		// == Migration: IPv6-Notation mit eckigen Klammern auflösen, um mit IpUtils::checkIp kompatibel zu sein ==
		self::performSystemMigration('updateForCallCounterIpAddressRanges', function () use ($config) {
			$original = $config->getExemptAddresses();
			$processed = array_map(function (string $address) {
				return str_replace(['[', ']'], '', $address);
			}, $original);

			// Daten speichern, falls sich etwas geändert hat
			if (array_diff($processed, $original)) {
				$config->setExemptAddresses($processed)->flush();
			}
		});

		self::performSystemMigrationOnce('revertSettingsIconColor', function () {
			foreach (SystemTheme::all() as $theme) {
				if ($theme->getSettingsIconBackgroundColor() == '#10a2d2') {
					$theme->setSettingsIconBackgroundColor('#003366');
				}
			}
			SystemTheme::flush();
		});

		self::performSystemMigrationOnce('updateClientInfo', function () use ($user) {
			// Remember the current domain context
			$requestDomainId = Domain::getDomainIdForRequest();

			try {
				foreach (User::getClients() as $client) {
					// Switch the user context
					User::setApiUser($client->getApiKey());
					Model::closeUserModels();
					Domain::switchToUser($client);

					$client
						->updateDomainInfo()
						->save();
				}
			} finally {
				// Switch back to the original user/domain context
				Domain::setDomainForRequest($requestDomainId);
				User::setApiUser($user ? $user->getApiKey() : '');
				Model::closeUserModels();
			}
		});

		// Updatevorgang nur bei aktivem Benutzer durchlaufen, um benutzerspezifische Konfigurationen zu gewährleisten.
		if ($user == null) {
			return;
		}

		// == Migration: SingleUser -> MultiUser ==
		// Verhindern, dass Neuinstallationen "aktualisiert" werden
		if (Utils::isBaseEdition() && $config->get('multiUser', false) == false && User::count() == 1) {
			self::performUserMigration("multiUser", function () use ($user) {
				$context = new ConfigImporterContext($user);

				if (Domain::count() == 0) {
					$domain = Domain::createForUser($user, false)
						->setManagementStructurePreLegalConsent(true)
						->setManagementStructure(Domain::MANAGEMENT_STRUCTURE_EMBEDDING)
					;
					$domain->save();

					$context->setDomain($domain);
				}

				Cookie::migratePreMultiUserConfig($context);
				FoundCookie::migratePreMultiUserConfig($context);
				FoundObject::migratePreMultiUserConfig($context);
				FoundScript::migratePreMultiUserConfig($context);
				FoundStorage::migratePreMultiUserConfig($context);
				LegalText::migratePreMultiUserConfig($context);
				Locale::migratePreMultiUserConfig($context);
				Protocol::migratePreMultiUserConfig($context);
				Theme::migratePreMultiUserConfig($context);

				Cookie::migratePreMultiDomainConfig($context);
				FoundCookie::migratePreMultiDomainConfig($context);
				FoundObject::migratePreMultiDomainConfig($context);
				FoundScript::migratePreMultiDomainConfig($context);
				FoundStorage::migratePreMultiDomainConfig($context);
				LegalText::migratePreMultiDomainConfig($context);
				Locale::migratePreMultiDomainConfig($context);
				Protocol::migratePreMultiDomainConfig($context);
				Theme::migratePreMultiDomainConfig($context);
			});
		}

		// == Migration: Globale Admin-E-Mail -> Admin-User-E-Mail ==
		self::performUserMigration("adminEmail", function () use ($user, $config) {
			if ($user->getRole() === User::ROLE_ADMIN and $user->getEmailAddress() == '') {
				$user->setEmailAddress($config->get('email'))->save();
			}
			$user->save();
		});

		// == Migration: Gruppen für den aktuellen Benutzer ==
		self::performUserMigration("legacySubGroups", function () {
			SubGroups::migrateLegacyRepository();
		});

		// == Migration: LegalTexts ==
		self::performDomainMigrations("legalTexts", function (Domain $domain) use ($config) {
			LegalText::upgradeFromLocaleIds();
			// Migration der unübersetzten Texte in erste Übersetzung
			if (!$config->get('defaultLocale')) {
				$locales = Locale::all();
				if ($locales) {
					reset($locales);
					$locale = current($locales);
					$config->set('defaultLocale', $locale->getId());
					$config->flush();
				}
			}
		});

		// == Migration: Embedding Default Locale ==
		self::performDomainMigrations("defaultLocaleEmbeddings", function (Domain $domain) {
			$locales = LocaleService::all();
			if ($locales) {
				reset($locales);
				$locale = current($locales);
				foreach (Embedding::all() as $embedding) {
					$embedding->fillTranslationsFromDefault($locale);
				}
				Embedding::flush();
				foreach (EmbeddingAsset::all() as $asset) {
					$asset->fillTranslationsFromDefault($locale);
				}
				EmbeddingAsset::flush();
			}
		});

		// == Migration: Fix Language Codes ==
		self::performDomainMigrations("fixLanguageCodes", function (Domain $domain) {
			// Portugiesisch korrigieren
			$pt_locale = Locale::find('pt_pt');
			if ($pt_locale) {
				$pt_locale->setName('pt_PT');
				$pt_locale->save();
			}
			// Tschechisch korrigieren
			$cz_locale = Locale::find('cz_CZ');
			if ($cz_locale) {
				$cz_locale->setName('cs');
				$cz_locale->save();
				$cs_locale = $cz_locale;
				$cz_locale = Locale::createNonPersistent();
				$cz_locale->setName('cz_CZ');

				$legaltext = LegalText::byLocale($cz_locale);
				if ($legaltext) {
					$legaltext->setLocale($cs_locale);
					$legaltext->save();
				}
				foreach (Embedding::all() as $embedding) {
					foreach (['TranslatedName', 'TranslatedDescription', 'TranslatedPrivacyPolicyUrl', 'TranslatedLifetime', 'TranslatedCollectedDataInfo', 'TranslatedPurposeOfDataCollection', 'TranslatedLegalBasis', 'TranslatedPlaceOfProcessing'] as $key) {
						$value = call_user_func([$embedding, 'get'.$key], $cz_locale, false);
						if ($value !== '') {
							call_user_func([$embedding, 'set'.$key], $cs_locale, $value);
						}
					}
				}
				Embedding::flush();
				foreach (EmbeddingAsset::all() as $asset) {
					foreach (['TranslatedDescription', 'TranslatedLifetime'] as $key) {
						$value = call_user_func([$asset, 'get'.$key], $cz_locale, false);
						if ($value !== '') {
							call_user_func([$asset, 'set'.$key], $cs_locale, $value);
						}
					}
				}
				EmbeddingAsset::flush();
				Locale::reload();
			}
			$cz_locale = UserLocale::find('cz_CZ');
			if ($cz_locale) {
				$cz_locale->rename('cs');
				$cz_locale->save();
				$cs_locale = $cz_locale;
				$cz_locale = UserLocale::createNonPersistent('cz_CZ');

				$legaltext = LegalText::byLocale($cz_locale);
				if ($legaltext) {
					$legaltext->setLocale($cs_locale);
					$legaltext->save();
				}
				foreach (Embedding::all() as $embedding) {
					foreach (['TranslatedName', 'TranslatedDescription', 'TranslatedPrivacyPolicyUrl', 'TranslatedLifetime', 'TranslatedCollectedDataInfo', 'TranslatedPurposeOfDataCollection', 'TranslatedLegalBasis', 'TranslatedPlaceOfProcessing'] as $key) {
						$value = call_user_func([$embedding, 'get'.$key], $cz_locale, false);
						if ($value !== '') {
							call_user_func([$embedding, 'set'.$key], $cs_locale, $value);
						}
					}
				}
				Embedding::flush();
				foreach (EmbeddingAsset::all() as $asset) {
					foreach (['TranslatedDescription', 'TranslatedLifetime'] as $key) {
						$value = call_user_func([$asset, 'get'.$key], $cz_locale, false);
						if ($value !== '') {
							call_user_func([$asset, 'set'.$key], $cs_locale, $value);
						}
					}
				}
				EmbeddingAsset::flush();
				UserLocale::reload();
			}
		});

		if (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_FILE) {
			self::performDomainMigrations("defaultThemeColors", function (Domain $domain) {
				// Bisher nicht barrierefreie Standard-Theme-Farben ändern
				$theme = Theme::find('c6557a2');
				if ($theme) {
					if ($theme->getOkButtonColor() == '#ffffff' and $theme->getOkButtonBackgroundColor() == '#5fb376') {
						$theme->setOkButtonBorderColor('#498248')->setOkButtonBackgroundColor('#417c51');
					}
					if ($theme->getDeclineButtonTextColor() == '#9c9c9c' and $theme->getDeclineButtonBackgroundColor() == '#f7f5f5') {
						$theme->setDeclineButtonTextColor('#6a6a6a');
					}
					if ($theme->getMoreInfoButtonColor() == '#9c9c9c' and $theme->getMoreInfoButtonBackgroundColor() == '#f7f5f5') {
						$theme->setMoreInfoButtonColor('#6a6a6a');
					}
					if ($theme->getInfoTextColor() == '#787878' and $theme->getWindowBackgroundColor() == '#ffffff') {
						$theme->setInfoTextColor('#767676');
					}
				}
				$theme = Theme::find('5af884c');
				if ($theme) {
					if ($theme->getDeclineButtonTextColor() == '#e6e6e6' and $theme->getDeclineButtonBackgroundColor() == '#6b6b6b') {
						$theme->setDeclineButtonBackgroundColor('#606060');
					}
					if ($theme->getMoreInfoButtonColor() == '#e6e6e6' and $theme->getMoreInfoButtonBackgroundColor() == '#6b6b6b') {
						$theme->setMoreInfoButtonBackgroundColor('#606060');
					}
				}
				Theme::flush();

			});
		}

		self::performUserMigrationOnce('revertSettingsIconColor', function () {
			foreach (UserTheme::all() as $theme) {
				if ($theme->getSettingsIconBackgroundColor() == '#10a2d2') {
					$theme->setSettingsIconBackgroundColor('#003366');
				}
			}
			UserTheme::flush();
		});

		self::performDomainMigrationsOnce('revertSettingsIconColorForDomains', function (Domain $domain) {
			foreach (Theme::all() as $theme) {
				if ($theme->getSettingsIconBackgroundColor() == '#10a2d2') {
					$theme->setSettingsIconBackgroundColor('#003366');
				}
			}
			Theme::flush();
		});

		// == Migration: Neue Übersetzungen in Locales ==

		/**
		 * Hilfsfunktion für die Migration von Locales
		 * @param LocaleBase $locale
		 * @param array<string,mixed> $template
		 * @param string $keyPrefix
		 */
		$complementLocaleRecursive = static function (LocaleBase $locale, array $template, string $keyPrefix='') use (&$complementLocaleRecursive): void {
			foreach ($template as $key => $value) {
				$fullKey = $keyPrefix . $key;
				if (is_array($value) and !isset($value[0])) {
					$complementLocaleRecursive($locale, $value, $fullKey.'.');
				} elseif ($value !== null and $value !== '') {
					if ($locale->get($fullKey) === null) {
						$locale->setIfMissing($fullKey, $value);
					}
				}
			}
		};

		/**
		 * Hilfsfunktion für die Überarbeitung der Standard-Bannertexte
		 * @param LocaleBase $locale
		 * @param array<string,mixed> $template
		 */
		$updateBannerTexts = static function (LocaleBase $locale, array $template) use ($purposes): void {
			if (!isset($template['widget_introduction'])) {
				return;
			}
			// Alte Standardtexte (erkannt über Hash) durch neuen Text ersetzen, wenn sie nicht angepasst wurden.
			switch (sha1($locale->getWidgetIntroduction())) {
				case '0ca4fcb3f723c075113c6b14efec325315fa4c0b': // de_DE 2023-06
				case '4aace560fc38290391502adea4a8af92166c1536': // cs 2023-06
				case 'c65f48d6d6aa82a6cdf56d136e6d33e5aa81d9ca': // cs 2023-01
				case 'f42c0ae64a01d1d907691034237ccada6c6105a4': // el 2023-06
				case '7bc150a975cf56400d54f4ebde4088471e2e101d': // el 2023-01
				case '8722001cbdb3d7a7c4d66ba70d74e5eb2c1e8976': // en_US 2023-06
				case '056cd33ef0dc5302821f432d9bb51b307252cf61': // en_US 2023-01
				case '548fc3a5e3bf57e618f2db9069f28faef7caf2f3': // es_ES 2023-06
				case '8533b59415cedd3e6bf4e4df3632f477081d53be': // es_ES 2023-01
				case '664ed4aea4a5afadcc4871ecc5fe8599a4d225a4': // fi 2023-06
				case 'ba25705ea48247ead6c5767bc6b8be25240fd910': // fi 2023-01
				case '334c7f88f652e5a4486ed86929aee36403ca0738': // fr_FR 2023-06
				case 'a358d686d0c2dfdaf2bcfa531108f09b09997c44': // ga 2023-06
				case '4304a50e25c72c8fcaa4a14f6d4226f0c73b701d': // hr 2023-06
				case '29d19767a76144629deac162dad2f7d3437dda1f': // hr 2023-01
				case '47c49bc0326cef287880c8c24b656e1b0259b33d': // hu 2023-06
				case '71dc7f3ddf85e28aaa34040ef65d8907329a7ae7': // hu 2023-01
				case 'ff9fb6608903b34d4ae4d66e698df433b216b78c': // is 2023-06
				case '038cd652012e5d7c26f2edfc52570ecd38aa8513': // is 2023-01
				case 'c1e349ceab3739d0cc381187ecbbc51576e0f8a4': // it_IT 2023-06
				case '97ee7df91c716a9e8e5c6052dde968c521f2c667': // ja 2023-06
				case '7acd1db03918b228bad533c466b9918e7a21713e': // ja 2023-01
				case 'd1020a20f0b3b3da18d97287714d1017d5d54615': // nl_NL 2023-06
				case 'e6a25206e9874254d0935742fe65ea1ff6353ed8': // nl_NL 2023-01
				case 'a80d9e517f41f7169a81845d6afb07f549f9bf6e': // no 2023-06
				case '4142ade60da651accfcd1c5643a12693c29652f1': // no 2023-01
				case '5178ff0f8470e617745123b7b6c32fd1c651e6c2': // pl_PL 2023-06
				case 'ab286396d67307e12545247a0ac419c25fb1805d': // pl_PL 2023-01
				case 'de154319000b955422be17dc87f5ed1832983348': // pt_PT 2023-06
				case '9f00e76684a2222d15b1ce71f20062feb03f364e': // pt_PT 2023-01
				case 'f7998138e859d41e90b70a161b3c0aa81dca868a': // ro 2023-06
				case '0da59b042adba8bf9ae11204beec305786e01ac7': // ro 2023-01
				case 'fd1d0376fe71f80c3abbe053e229bece40869edc': // ru_RU 2023-06
				case 'a71cfb0f168df5dace1313a29e865ec6ec1db009': // sk 2023-06
				case '4cfaf754119b99674cc5b5557388571e4afc6a94': // sk 2023-01
				case '622361feacdb4e39129994e90c9b4bb563c4ca05': // sl 2023-06
				case 'dc990426a49e4c7c754a191926003134feb76443': // sl 2023-01
				case '45b0e23b0998ec2533a9ae20efa53fbbef22ebe1': // sr 2023-06
				case 'c70f2e22c5c8df12c7973deedf420c35a3157127': // sr 2023-01
				case '811c23af4e9c6f91f0c47fccfdc325c853f9efb1': // sv 2023-06
				case '3a4cc92f99bd5f068b1917a1f2cea0fc96e6f922': // tr 2023-06
				case 'cc7c598790cbe944e502e31b0eb9236a375239cd': // tr 2023-01
				case 'f6c09e118af5dd328c033305a5e8624fe6e85a4d': // zh 2023-06
				case '8e0046b850fc672cea0a19012ec7956b94dec4d8': // zh 2023-01
				// Revert the introduction back to its state before the addition of the TCF text (removing the partner count)
				case '72e847d43761f91455140fce1b07e5cfe5595455': // de_DE 0c57c98
				case 'd79ad27f77c9cf787af3e99785e054d3b2ae8b94': // de_DE 687bc00
				case 'e1d284e624d9084d6f81570e7a8ea3fb57201d8d': // de_DE 8b7f17a
				case '2fa21134a99e9cfd8899e9393cb18fe944f676b7': // cs 687bc00
				case '38f07e390587395db54be809f43a7136c8e53d6d': // el 687bc00
				case 'ff55b1b03a97de858ccb69acc6c78edbbe44ecb9': // en_US 0c57c98
				case '820b019c417b8c052839be1edab24e54c33ca371': // en_US 687bc00
				case '1c14807736eb9162efdd4ce72ed9a61fc471d1a3': // es_ES 687bc00
				case '24c428497391daa6cefd4486aaf2f230b92b2be8': // fi 687bc00
				case 'b1825b07f818de4c130e28525d420bd993245a6b': // fr_FR 687bc00
				case '1e3ecfeace385ed5cb4b62da92af9fbead15e915': // ga 687bc00
				case '6c33d0e7f038524f910e82f4c0af6f76ba0fb5ad': // hr 687bc00
				case '9b1b1307c57c3c684c2298cdd79f5161045d732c': // hu 687bc00
				case 'fd0866998e5fd99223d8c07a7c802dad2b990e91': // is 687bc00
				case 'bee3c0d1da0312955da8012645621be169c44c2c': // it_IT 687bc00
				case '8c59ac5e4c339f6ddb0e6ee3fdf36e6d480d1101': // ja 687bc00
				case 'b47da5c8e04c2163c480dfd1a0a9dfd17fdd5005': // nl_NL 687bc00
				case 'c83753fb94214fb61af25e10d8ec82d58f93723c': // no 687bc00
				case 'f4f782aff09259c942003c787dcb0ab12100295c': // pl_PL 687bc00
				case 'efd5ef83235c5b79d8533317aa6ae7392625ffa5': // pt_PT 687bc00
				case 'ada8982d35c50d9727c401888c9c8199495265cd': // ro 687bc00
				case '8f519a6cb7e8005133b141e164295cd8c214ee41': // ru_RU 687bc00
				case '85a40ffdc1154c7e073caa271e715841e6ad23a0': // sk 687bc00
				case '5858a149d2c33ef890bb9216f169a40f3c8875ab': // sl 687bc00
				case '7f35acf6c702d4f7c2c4c3b9b8e3792a5462838c': // sr 687bc00
				case '0302208686867b49dbff5383973535378f9dfc54': // sv 687bc00
				case 'd9164944a73f07a069308ec332f87904144f5664': // tr 687bc00
				case '721d0d571615a1b451d15a57b458cbe5d8fd6fca': // zh 687bc00
				case '05cfdb48b009d7fa535b605bc6bbd6919248d292': // bg 687bc00
					$locale->setWidgetIntroduction($template['widget_introduction']);
					break;
				default:
					break;
			}

			foreach ($template['purpose_description'] ?? [] as $purposeId => $translation) {
				if (is_string($translation) and strlen($translation) > 0 and $purpose = $purposes[$purposeId] ?? null) {
					switch (sha1($locale->getPurposeDescription($purpose))) {
						case '5026594abeb216e113b991a17434eab224c6928f': // de_DE[41ba25c] 39c2744,500144d,3865430
						case 'abbe436439d6a8218731ef7647c62f068f1eff29': // de_DE[cdcbd7c] 39c2744,500144d,3865430
						case '67cc04786ec0cc8d6907a53d8f84acda00e867b5': // de_DE[a717ff5] 39c2744,500144d,3865430
						case '28e2ef63347cd70c6d5893f63f4b320bf539fd2e': // de_DE[7c19e32] 39c2744,500144d,3865430
						case '011470659ef51c5b1dd34f502b9c08b3c1f8bf3a': // de_DE[6cd2721] 39c2744,500144d,9e3365f
						case '9070ae7861314799ab2aa1b64539a331615ca566': // de_DE[6cd2721] 3865430
						case '7c53ecbb921754d90a8ed1177589214f1c07ed95': // de_DE[15c61c3] 39c2744
							$locale->setPurposeDescription($purpose, $translation);
							break;
						default:
							break;
					}
				}
			}
		};

		self::performUserMigration("frontendUserLocales", function () use ($complementLocaleRecursive, $updateBannerTexts) {
			$templates = UserLocale::readSkeletonData();
			$existingLocales = UserLocale::all();

			// Hole vorhandene UserLocales und erzeuge neue für Übersetzungen, die durch ein Update hinzugekommen sind
			$locales = array_combine(
				array_keys($templates),
				array_map(function (string $localeName) use ($existingLocales) {
					try {
						$locale = $existingLocales[$localeName]
							?? UserLocale::create($localeName)
								->setEnabled(false)
						;
					} catch (LocaleAlreadyExistsException $exception) {
						$locale = null;
					}

					return $locale;
				}, array_keys($templates))
			);

			// LocaleAlreadyExistsException wird u. U. ausgelöst; leere Einträge herausfiltern
			$locales = array_filter($locales);

			foreach ($locales as $locale) {
				$template = $templates[$locale->getName()] ?? [];
				$complementLocaleRecursive($locale, $template);
				$updateBannerTexts($locale, $template);
			}

			UserLocale::flush();
		});

		self::performDomainMigrations("frontendLocales", function (Domain $domain) use ($complementLocaleRecursive, $updateBannerTexts) {
			// Standardwerte für neu hinzugekommene Übersetzungen
			$templates = UserLocale::readSkeletonData();

			foreach (Locale::all() as $locale) {
				$template = $templates[$locale->getName()] ?? [];
				$complementLocaleRecursive($locale, $template);
				$updateBannerTexts($locale, $template);
			}

			Locale::flush();
		});

		// == Migration: Base-URL befüllen, wenn leer ==
		self::performUserMigration("baseUrl", function () use ($config, $request) {
			if ($config->getBaseUrl() === '') {
				$scheme = $config->isForceHttpsConnection() ? 'https://' : $request->getScheme().'://';
				$currentHostUrl = $scheme.$request->getHttpHost();

				$router = Utils::getRouter();

				$config->setBaseUrl($currentHostUrl.$router->generate('app_root'));
				$config->flush();
			}
		});

		self::performDomainMigrations("domainIndex", function (Domain $domain) {
			// Domain-Index erstellen, falls noch nicht passiert
			if (!DomainIndex::persistedRepositorySize()) {
				DomainIndex::rebuild();
			}
		});

		self::performUserMigrationOnce('updateDomainInfo', function () use ($user) {
			$user
				->updateDomainInfo()
				->save();
		}, '2023.03.29');
	}

	/**
	 * Führt eine Migration im Systemkontext durch, falls anstehend.
	 *
	 * Die übergebene Funktion wird nur aufgerufen, wenn die Migration seit
	 * dem letzten CCM19-Update noch nicht durchgeführt wurde.
	 *
	 * @param string $id
	 * @param callable $f Funktion, die die eigentliche Migration erledigt
	 * @param string $afterVersion Migration ausführen, sofern nach Version `$afterVersion` noch nicht geschehen.
	 * @param bool $once
	 * @return bool Ob die Migration durchgeführt wurde
	 */
	private static function performSystemMigration(string $id, callable $f, string $afterVersion='', bool $once=false): bool
	{
		$state = SystemMigrationState::byId($id);
		if ($state->isStale($afterVersion, $once)) {
			$state->touch();
			try {
				$state->save();
			} catch (WriteException $e) {
				// E11000: duplicate key -> ein anderer Request hat gleichzeitig gespeichert.
				if ($e->getCode() == 11000) {
					return false;
				}
				throw $e;
			}
			try {
				$f();
			} catch (Exception $e) {
				$state->reset()->save();
				throw $e;
			}
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Führt eine Migration im Systemkontext einmalig durch.
	 *
	 * @param string $id
	 * @param callable $f Funktion, die die eigentliche Migration erledigt
	 * @param string $afterVersion Migration ausführen, sofern nach Version `$afterVersion` noch nicht geschehen.
	 * @return bool Ob die Migration durchgeführt wurde
	 */
	private static function performSystemMigrationOnce(string $id, callable $f, string $afterVersion=''): bool
	{
		return self::performSystemMigration($id, $f, $afterVersion, true);
	}

	/**
	 * Führt eine Migration für den aktuellen User durch, falls anstehend.
	 *
	 * Die übergebene Funktion wird nur aufgerufen, wenn diese seit dem letzten
	 * CCM19-Update nicht schon für den aktuellen User durchgeführt wurde.
	 *
	 * @param string $id
	 * @param callable():void $f Funktion/Closure das die eigentliche Migration erledigt
	 * @param string $afterVersion Migration ausführen, sofern nach Version `$afterVersion` noch nicht geschehen.
	 * @param bool $once
	 * @return bool Ob der Migrationscode durchgeführt wurde
	 */
	private static function performUserMigration(string $id, callable $f, string $afterVersion='', bool $once=false): bool
	{
		$state = UserMigrationState::byId($id);
		if ($state->isStale($afterVersion, $once)) {
			$state->touch();
			try {
				$state->save();
			} catch (WriteException $e) {
				// E11000: duplicate key -> ein anderer Request hat gleichzeitig gespeichert.
				if ($e->getCode() == 11000) {
					return false;
				}
				throw $e;
			}
			try {
				$f();
			} catch (Exception $e) {
				$state->reset()->save();
				throw $e;
			}
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Führt eine Migration für den aktuellen User einmalig durch.
	 *
	 * @param string $id
	 * @param callable $f Funktion, die die eigentliche Migration erledigt
	 * @param string $afterVersion Migration ausführen, sofern nach Version `$afterVersion` noch nicht geschehen.
	 * @return bool Ob die Migration durchgeführt wurde
	 */
	private static function performUserMigrationOnce(string $id, callable $f, string $afterVersion=''): bool
	{
		return self::performUserMigration($id, $f, $afterVersion, true);
	}

	/**
	 * Führt eine Migration für alle Domains des aktuellen Users durch, falls anstehend.
	 *
	 * Die übergebene Funktion wird - falls nicht schon seit dem letzten Update
	 * erledigt - für jede Domain des aktuellen Users aufgerufen.
	 *
	 * Es wird vor jedem Aufruf in den Kontext der Domain gewechselt und alle
	 * offenen domainspezifischen Models werden geschlossen.
	 *
	 * Nachdem diese Funktion zurückkehrt ist eine eventuell vorher aktive
	 * Domain auch wieder aktiv und domainspezifische Models sind geschlossen.
	 *
	 * @param string $id
	 * @param callable(Domain):void $f Funktion/Closure das die eigentliche Migration erledigt
	 * @param string $afterVersion Migration ausführen, sofern nach Version `$afterVersion` noch nicht geschehen.
	 * @param bool $once
	 * @return bool Ob der Migrationscode durchgeführt wurde
	 */
	private static function performDomainMigrations(string $id, callable $f, string $afterVersion='', bool $once=false): bool
	{
		$state = UserMigrationState::byId($id);
		if (Domain::count() > 0 and $state->isStale($afterVersion, $once)) {
			$state->touch();
			try {
				$state->save();
			} catch (WriteException $e) {
				// E11000: duplicate key -> ein anderer Request hat gleichzeitig gespeichert.
				if ($e->getCode() == 11000) {
					return false;
				}
				throw $e;
			}
			try {
				$requestDomainId = Domain::getDomainIdForRequest();
				foreach (Domain::all() as $domain) {
					Domain::setDomainForRequest($domain->getId());
					Model::closeDomainModels();
					$f($domain);
				}
				Domain::setDomainForRequest($requestDomainId);
				Model::closeDomainModels();
			} catch (Exception $e) {
				$state->reset()->save();
				throw $e;
			}
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Führt eine Migration für alle Domains des aktuellen Users einmalig durch.
	 *
	 * @param string $id
	 * @param callable $f Funktion, die die eigentliche Migration erledigt
	 * @param string $afterVersion Migration ausführen, sofern nach Version `$afterVersion` noch nicht geschehen.
	 * @return bool Ob der Migrationscode durchgeführt wurde
	 * @see self::performDomainMigrations()
	 */
	private static function performDomainMigrationsOnce(string $id, callable $f, string $afterVersion=''): bool
	{
		return self::performDomainMigrations($id, $f, $afterVersion, true);
	}

	/**
	 * Am Anfang einer Anfrage wird die Session gestartet.
	 * @param ControllerEvent $event
	 */
	public function onKernelController(ControllerEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}
		self::updateRoutines($event->getRequest());
	}
}
