<?php

namespace App\EventSubscriber;

use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Contracts\EventDispatcher\Event;
use App\Model\User;
use App\Model\Domain;
use App\Model\Protocol;
use App\Model\AccessCountJournal;
use App\Model\Statistic;
use App\Event\ProtocolRotateEvent;
use App\Event\StatisticsDisplayEvent;
use App\Event\StatisticsCronJobEvent;
use App\Utils;
use DateTime;

class StatisticsUpdateHandler implements EventSubscriberInterface
{
	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			ProtocolRotateEvent::NAME => ['onProtocolRotate', 10],
			StatisticsDisplayEvent::NAME => ['onStatisticsDisplay', 10],
			StatisticsCronJobEvent::NAME => ['onStatisticsCronJob', 10],
		];
	}

	private function updateCallCount(User $user, Domain $domain): void
	{

		$journal = AccessCountJournal::byDomainAndKind($domain, 'call');
		$stat = null;
		$callCountIncrement = 0;
		foreach ($journal->iterateEntries() as $date => $count) {
			$stat = Statistic::findOrCreate($date);
			$stat->incrementCallCount($count);
			$stat->save();
			$callCountIncrement += $count;
		}
		if (Utils::isExtendedEdition() || Utils::isMultiDomainEdition()) {
			User::beginWrite();
			$user
				->resetCallCountForNextMonth()
				->incrementCallCount($callCountIncrement)
				->save();
		}
	}

	private function updateViewCount(Domain $domain): void
	{
		$journal = AccessCountJournal::byDomainAndKind($domain, 'view');
		$stat = null;
		foreach ($journal->iterateEntries() as $date => $count) {
			$stat = Statistic::findOrCreate($date);
			$stat->incrementViewCount($count);
		}
	}

	public function onProtocolRotate(Event $event): void
	{
		$user = User::loggedInUser();
		$domain = Domain::activeDomain();
		if (!$domain or !$user) {
			return;
		}
		$this->updateStatisticsFull($user, $domain, true);
	}

	public function onStatisticsDisplay(Event $event): void
	{
		$user = User::loggedInUser();
		$domain = Domain::activeDomain();
		if (!$domain or !$user) {
			return;
		}
		$this->updateStatisticsFull($user, $domain);
	}

	public function onStatisticsCronJob(StatisticsCronJobEvent $event): void
	{
		$user = User::loggedInUser();
		$domain = Domain::activeDomain();
		if (!$domain or !$user) {
			return;
		}
		if ($event->isFullUpdateRequested()) {
			$this->updateStatisticsFull($user, $domain);
		} else {
			$this->updateCallCount($user, $domain);
			Statistic::beginWrite();
			$this->updateViewCount($domain);
			Statistic::flush();
		}
	}

	public function updateStatisticsFull(User $user, Domain $domain, bool $archiveCounts=false): void
	{
		$lastUpdate = $domain->getLastStatisticsUpdateTimestamp();
		$domain->touchLastStatisticsUpdateTimestamp()->save();

		if ($lastUpdate) {
			// Immer ganze Tage verarbeiten
			$lastUpdate->modify('00:00:00');
		} else {
			// wenn noch nie ausgeführt, bis zu 3 Tage verarbeiten
			$lastUpdate = new DateTime('- 2 days midnight');
		}

		$protocol = Protocol::byDomain($domain);
		$currentDate = null;
		// Nach Unique-ID gruppieren; immer neuesten Eintrag nehmen
		/** @var array<string,array> $byUcid */
		$byUcid = [];
		/** @var array<string,array<string,array<string,int>>> $sumsPerDay */
		$sumsPerDay = [];
		$date = null;

		foreach ($protocol->allFromDate($lastUpdate->getTimestamp()) as $item) {
			$date = date("Y-m-d", $item['timestamp'] ?? 0);
			$ucid = $item['ucid'];
			unset($item['ucid']);
			unset($item['manipulationPrevention']);
			$item['date'] = $date;

			if ($currentDate !== $date) {
				$sumsPerDay[$date] = [
					'purposes'=>[],
					'embeddings'=>[]
				];
				$currentDate = $date;
			}

			if (isset($byUcid[$ucid])) {
				$olditem = $byUcid[$ucid];
				$item['first_date'] = $olditem['first_date'] ?? $olditem['date'];
			} else {
				$olditem = null;
			}

			// Purposes zählen
			if (isset($item['purposes']) && is_array($item['purposes'])) {
				if ($olditem && isset($olditem['purposes'])) {
					foreach ($olditem['purposes'] as $purposeId) {
						if (!isset($sumsPerDay[$date]['purposes'][$purposeId])) {
							$sumsPerDay[$date]['purposes'][$purposeId] = 0;
						}
						// @phan-suppress-next-line PhanTypeArraySuspiciousNullable
						$sumsPerDay[$date]['purposes'][$purposeId]--;
					}
				}
				foreach ($item['purposes'] as $purposeId) {
					if (!isset($sumsPerDay[$date]['purposes'][$purposeId])) {
						$sumsPerDay[$date]['purposes'][$purposeId] = 0;
					}
					// @phan-suppress-next-line PhanTypeArraySuspiciousNullable
					$sumsPerDay[$date]['purposes'][$purposeId]++;
				}
			}

			// Embeddings zählen
			if (isset($item['embeddings']) && is_array($item['embeddings'])) {
				if ($olditem && isset($olditem['embeddings'])) {
					foreach ($olditem['embeddings'] as $embeddingId) {
						if (!isset($sumsPerDay[$date]['embeddings'][$embeddingId])) {
							$sumsPerDay[$date]['embeddings'][$embeddingId] = 0;
						}
						// @phan-suppress-next-line PhanTypeArraySuspiciousNullable
						$sumsPerDay[$date]['embeddings'][$embeddingId]--;
					}
				}
				foreach ($item['embeddings'] as $embeddingId) {
					if (!isset($sumsPerDay[$date]['embeddings'][$embeddingId])) {
						$sumsPerDay[$date]['embeddings'][$embeddingId] = 0;
					}
					// @phan-suppress-next-line PhanTypeArraySuspiciousNullable
					$sumsPerDay[$date]['embeddings'][$embeddingId]++;
				}
			}
			$byUcid[$ucid] = $item;
		}
		// Speicher freigeben
		foreach ($byUcid as $ucid => &$item) {
			$item = null;
		}
		unset($item);
		$lastDate = $date;

		// Statistiken speichern
		Statistic::beginWrite();
		foreach ($sumsPerDay as $date => $row) {
			$stat = Statistic::findOrCreate($date);

			foreach ($row['purposes'] as $purposeId => $count) {
				$stat->setPurposeAcceptCount($purposeId, $count);
			}
			foreach ($row['embeddings'] as $embeddingId => $count) {
				$stat->setEmbeddingAcceptCount($embeddingId, $count);
			}

			if ($archiveCounts and $date == $lastDate) {
				$stat->archivePurposeEmbeddingCounts();
			}
		}
		$this->updateViewCount($domain);
		Statistic::flush();

		$this->updateCallCount($user, $domain);
	}

}
