<?php

namespace App\EventSubscriber;

use App\Component\LocaleService;
use App\Component\Menu\Manager as MenuManager;
use App\Controller\AdminController;
use App\Controller\DomainDependantController;
use App\Controller\ExternalController;
use App\Kernel;
use App\Model\Country;
use App\Model\Domain;
use App\Model\BackendLanguage;
use App\Model\PluginState;
use App\Model\User;
use App\Config;
use App\Utils;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\ControllerEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Twig\Environment;

class TwigGlobalsHandler extends AbstractController implements EventSubscriberInterface
{
	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			KernelEvents::CONTROLLER => ['onKernelController', 1000],
		];
	}

	/**
	 * Diese Methode setzt ein paar häufig verwendete Templatevariablen an zentraler Stelle.
	 *
	 * Wichtig: Sollte nach dem AuthenticationHandler laufen, um den aktiven Benutzer korrekt zu bestimmen.
	 *
	 * @param ControllerEvent $event
	 * @phan-suppress PhanTypeArraySuspicious
	 */
	public function onKernelController(ControllerEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}

		$config = Config::getInstance();

		/** @var AbstractController $controller */
		$controller = $event->getController()[0];

		/** @var Kernel $kernel */
		global $kernel;

		/** @var Request $request */
		$request = $event->getRequest();

		/** @var Environment $twig */
		$twig = $kernel->getContainer()->get('twig');

		$isExternalRequest = $controller instanceof ExternalController;

		/** @var User|null $user */
		$user = User::loggedInUser();
		$owner = $user ? $user->getAccountOwner() : null;

		/** @var Domain|null $domain */
		$domain = $user ? Domain::activeDomain() : null;

		if (!$request->attributes->getBoolean('_stateless')) {
			try {
				$session = $request->getSession();
				$updateNotifier = (bool)$session->get('update_notifier', false);
			} catch (\Exception $ignore) {
				$updateNotifier = false;
			}

			$twig->addGlobal('update_notifier', $updateNotifier);
		}

		$twig->addGlobal('user', $user);
		$twig->addGlobal('_isAdmin', $user ? $user->getRole() == User::ROLE_ADMIN : false);
		$twig->addGlobal('_isClient', $user ? $user->getRole() == User::ROLE_CLIENT : false);
		$twig->addGlobal('_appId', substr(hash_hmac('sha256', Utils::getBaseDir(), Utils::getLicenseKey()), 1, 16));
		$twig->addGlobal('_appName', $config->getAppName());
		$twig->addGlobal('_domain', $domain);
		$twig->addGlobal('_edition', [
			'basic' => Utils::isBaseEdition(),
			'extended' => Utils::isExtendedEdition(),
			'multi' => Utils::isMultiDomainEdition(),
		]);
		$twig->addGlobal('_hostingInterface', Utils::hostingInterface());
		$twig->addGlobal('_config', $config);
		$twig->addGlobal('_themeBack_customCss', (string)(Utils::isExtendedEdition() ? $config->getBackendCustomCss() : ''));
		$request->attributes->set('cspOrigins', $config->getBackendCustomCssOrigins());
		$twig->addGlobal('_isFrontendActive',true);
		$twig->addGlobal('_backgroundImageLogin', Utils::getBackgroundImageReal());

		//Translate nur anzeigen, wenn dev
		$twig->addGlobal('_isDev', $_ENV['APP_ENV'] == 'dev');
		$twig->addGlobal('_isCcm19Cloud', Utils::isCCMCloud());
		$twig->addGlobal('_isClickskeksCloud', Utils::isClickskeksCloud());
		$twig->addGlobal('_isCloud', Utils::isCCMCloud() || Utils::isClickskeksCloud());
		$twig->addGlobal('_hasCrawler', Utils::isCCMCloud() || Utils::isClickskeksCloud() || Utils::hasCrawler());

		$twig->addGlobal('_isFreeClient', Utils::isFreePlan($owner));

		if (is_object($domain)) {
			$twig->addGlobal('_isModernManagement',$domain->getManagementStructure());
			$twig->addGlobal('_isFrontendActive',$domain->isFrontendWidgetEnabled());
		}
		$twig->addGlobal('_domainDependant', $controller instanceof DomainDependantController);
		$twig->addGlobal('_adminController', $controller instanceof AdminController);
		$twig->addGlobal('_externalRequest', $controller instanceof ExternalController);

		// Versions-ID/-Name
		$twig->addGlobal('_versionName', Utils::getVersionName());
		$twig->addGlobal('_versionId', Utils::getVersionId());
		$twig->addGlobal('_enableHelp', Utils::isHelpEnabled());
		$twig->addGlobal('_docsUrl', Utils::getDocsUrl());

		// aktuelle Route
		$route = $request->attributes->get('_route', 'app_root');
		$twig->addGlobal('_route', $route);
		$twig->addGlobal('_route_params', $request->attributes->get('_route_params', []));
		$twig->addGlobal('_writeAccess', Utils::mayUserAccessRoute($user, $route, 'POST'));

		$twig->addGlobal('_docUrl', Utils::getDocUrl($route));
		$twig->addGlobal('_domainSwitchRoute', ($controller instanceof DomainDependantController) ? $route : 'app_dashboard');

		// Minify-Einstellung
		$twig->addGlobal('_forceDisableMinification', ($domain and $domain->hasCodeMinification() == false));

		if ($isExternalRequest == false) {
			// Token, damit 2FA bei Session-Relogin nicht nötig ist
			$origUser = (User::originalLoggedInUser() ?: User::loggedInUser());
			if ($origUser and $origUser->getSessionResumeToken()) {
				$twig->addGlobal('_resumeToken', $origUser->getSessionResumeToken());
			}

			/** @var Domain[] $domains */
			$domains = $user ? Domain::all() : [];

			// Sortiere Domains dem Namen nach
			usort($domains, function (Domain $a, Domain $b) {
				return strnatcasecmp($a->getName(), $b->getName());
			});

			if ($user) {
				$frontendL10n = LocaleService::findFuzzy($request->getLocale()) ?? LocaleService::getDefault();
			} else {
				$frontendL10n = null;
			}

			$twig->addGlobal('_domains', $domains);
			$twig->addGlobal('_originalUser', User::originalLoggedInUser());
			$twig->addGlobal('_localeLanguages', BackendLanguage::allActiveSorted($request));
			$twig->addGlobal('_frontendL10n', $frontendL10n);
			$twig->addGlobal('_countries', Country::all());
			$twig->addGlobal('_preferredColorScheme', $origUser ? $origUser->getPreferredColorScheme() : null);

			// Update-Status
			$twig->addGlobal('_updateInfo', UpdateNotifyHandler::getCachedData());

			/** @var MenuManager $menuManager */
			$menuManager = $kernel->getContainer()->get('menu.manager');

			$twig->addGlobal('_metaSidebar', $menuManager->isRouteInMetaMenu($request->attributes->get('_route')));

			// Aktive App bestimmen
			$apps = $user ? $menuManager->getApps($user) : [];
			$appName = 'CCM19';
			$activeApp = $apps['CCM19'] ?? null;
			if (strpos($route, 'app_') !== 0) {
				foreach (PluginState::all() as $plugin) {
					if ($plugin->getType() === 'ccm19-app' and strpos($route, $plugin->getRoutePrefix()) === 0) {
						$appName = $plugin->getName();
						if (isset($apps[$appName])) {
							$activeApp = $apps[$appName];
						}
						break;
					}
				}
			}

			// Menüs
			$twig->addGlobal('_showAppBar', count($apps) > 1);
			$twig->addGlobal('_activeApp', $activeApp);
			$twig->addGlobal('_apps', $user ? $menuManager->getApps($user) : []);
			$metaMenu = $user ? $menuManager->getMetaMenu($user, $appName) : [];
			$menu = $user ? $menuManager->getMenu($user, $appName) : [];
			if (!$menu) {
				$menu = $metaMenu;
			}
			$twig->addGlobal('_metaMenu', $metaMenu);
			$twig->addGlobal('_menu', $menu);
		}
	}
}
