<?php

namespace App\EventSubscriber;

use App\Config;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

class UnloadHandler implements EventSubscriberInterface
{
	/**
	 * @override
	 * @return array<string,mixed>
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			KernelEvents::RESPONSE => ['onKernelResponse', 50],
		];
	}

	/**
	 * Am Ende einer Anfrage wird die Session geschlossen, die Konfiguration geschrieben
	 * und ggf. Standard-Header nachgetragen.
	 *
	 * @param ResponseEvent $event
	 */
	public function onKernelResponse(ResponseEvent $event): void
	{
		if ($event->isMainRequest() == false) {
			return;
		}

		/** @var Config $config */
		$config = Config::getInstance();
		$config->flush();

		$request = $event->getRequest();
		$area = $request->attributes->get('area');
		if ($request->hasSession(true)) {
			$session = $request->getSession();
			if ($session and $session->isStarted()) {
				$session->save();
			}
		}

		$response = $event->getResponse();
		if (!$response->headers->has('Content-Security-Policy')) {
			$cspOrigins = $request->attributes->get('cspOrigins', []);

			if ($area === 'frontend' or $area === 'api') {
				$response->headers->set('Content-Security-Policy',
					"default-src 'self'; " .
					"connect-src 'self'; " .
					"script-src 'self' 'unsafe-inline'; " .
					"style-src 'self' 'unsafe-inline'; " .
					"img-src 'self' data:; " .
					"frame-ancestors 'self'; " .
					"form-action 'self'"
				);
			}
			elseif ($response->getStatusCode() >= 400) {
				$src = implode(' ', $cspOrigins);
				$response->headers->set('Content-Security-Policy',
					"default-src 'self'; " .
					"connect-src 'self'; " .
					"script-src 'self' 'unsafe-inline'; " .
					"style-src 'self' 'unsafe-inline' $src; " .
					"img-src 'self' data: $src; " .
					"font-src 'self' data: $src; " .
					"frame-ancestors 'self'; " .
					"form-action 'self'"
				);
			} else {
				$imgSrc = implode(' ', array_unique(array_merge(['data:', 'https://www.ccm19.de'], $cspOrigins)));
				$styleSrc = implode(' ', $cspOrigins);

				$response->headers->set('Content-Security-Policy',
					"default-src 'self'; " .
					"connect-src 'self' https://update.ccm19.de; " .
					"script-src 'self' 'unsafe-inline'; " .
					"style-src 'self' 'unsafe-inline' $styleSrc; " .
					"font-src 'self' $styleSrc; " .
					"img-src 'self' $imgSrc; " .
					"frame-ancestors 'self'; " .
					"form-action 'self' https:"
				);
			}
			$response->headers->set('X-Frame-Options', 'sameorigin');
		}
		if (!$response->headers->has('X-Content-Type-Options')) {
			$response->headers->set('X-Content-Type-Options', 'nosniff');
		}
		if (!$response->headers->has('Permissions-Policy')) {
			$response->headers->set('Permissions-Policy', 'interest-cohort=(), geolocation=(), encrypted-media=(), camera=(), microphone=()');
		}
	}
}
