<?php

namespace App;

use App\Component\Cache\Cache;
use App\Component\LocaleService;
use App\Component\Plugin\Manager as PluginManager;
use App\Model\AccessCountJournal;
use App\Model\Country;
use App\Model\Domain;
use App\Model\GeoIp;
use App\Model\Index\DomainIndex;
use App\Model\ThemeBase;
use App\Model\User;
use Composer\Autoload\ClassLoader;
use Exception;
use Symfony\Bundle\FrameworkBundle\Kernel\MicroKernelTrait;
use Symfony\Component\Config\Loader\LoaderInterface;
use Symfony\Component\Config\Resource\FileResource;
use Symfony\Component\DependencyInjection\ContainerBuilder;
use Symfony\Component\HttpFoundation\IpUtils;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Kernel as BaseKernel;
use Symfony\Component\Routing\Loader\Configurator\RoutingConfigurator;
use Symfony\Component\DependencyInjection\Compiler\CompilerPassInterface;

class Kernel extends BaseKernel implements CompilerPassInterface
{
	use MicroKernelTrait;

	private const CONFIG_EXTS = '.{php,xml,yaml,yml}';
	/** @var string $cacheDir */
	private $cacheDir;

	/** @var ClassLoader $autoloader */
	private $autoloader;

	public function __construct(string $environment, bool $debug)
	{
		if ($environment === 'prod' and function_exists('error_reporting')) {
			# Only report errors, no warnings, notices or deprecations
			@error_reporting(E_ERROR|E_PARSE|E_CORE_ERROR|E_COMPILE_ERROR|E_USER_ERROR|E_RECOVERABLE_ERROR);
		}

		parent::__construct($environment, $debug);

		$this->autoloader = require dirname(__DIR__).'/vendor/autoload.php';

		if ($trustedProxies = $_SERVER['TRUSTED_PROXIES'] ?? $_ENV['TRUSTED_PROXIES'] ?? false) {
			Request::setTrustedProxies(explode(',', $trustedProxies), Request::HEADER_X_FORWARDED_FOR
				| Request::HEADER_X_FORWARDED_HOST
				| Request::HEADER_X_FORWARDED_PORT
				| Request::HEADER_X_FORWARDED_PROTO
			);
		}

		if ($trustedHosts = $_SERVER['TRUSTED_HOSTS'] ?? $_ENV['TRUSTED_HOSTS'] ?? false) {
			Request::setTrustedHosts([$trustedHosts]);
		}

		if (isset($_SERVER['HTTP_HOST'])) {
			self::ccm19BlockWhenInUpdate();
			self::handleCustomRouting();
		}

		global $kernel;
		$kernel = $this;
	}

	private function reconstructUri(array $parsedUrl): string
	{
		$scheme   = isset($parsedUrl['scheme']) ? $parsedUrl['scheme'] . ':' : '';
		if (isset($parsedUrl['host']) or isset($parsedUrl['user'])) {
			$scheme .= '//';
		}
		$host     = isset($parsedUrl['host']) ? $parsedUrl['host'] : '';
		$port     = isset($parsedUrl['port']) ? ':' . $parsedUrl['port'] : '';
		$user     = isset($parsedUrl['user']) ? rawurlencode($parsedUrl['user']) : '';
		$pass     = isset($parsedUrl['pass']) ? ':' . rawurlencode($parsedUrl['pass']) : '';
		$pass     = ($user || $pass) ? "$pass@" : '';
		$path     = isset($parsedUrl['path']) ? $parsedUrl['path'] : '';
		$query    = isset($parsedUrl['query']) ? '?' . $parsedUrl['query'] : '';
		$fragment = isset($parsedUrl['fragment']) ? '#' . $parsedUrl['fragment'] : '';

		return "$scheme$user$pass$host$port$path$query$fragment";
	}

	/**
	 * Wandelt SwiftMailer-URIs in Symfony-Mailer-DNSs um
	 */
	private function convertMailerDsn(string $swiftUrl): string
	{
		$parts = parse_url($swiftUrl) ?: [];
		$scheme = $parts['scheme'] ?? '';
		$host = $parts['host'] ?? null;
		$port = $parts['port'] ?? null;
		$user = $parts['user'] ?? null;
		$pass = $parts['pass'] ?? null;
		$path = $parts['path'] ?? null;
		$query = [];
		parse_str($parts['query'] ?? '', $query);

		if ($scheme === 'null') {
			return 'null://null';
		} else if ($scheme === 'sendmail') {
			if (!$host or $host === 'localhost') {
				$host = 'default';
			}
		} else if ($scheme === 'smtp') {
			if (($query['encryption'] ?? '') === 'ssl') {
				unset($query['encryption']);
				$scheme = 'smtps';
			}
			if (!isset($query['verify_peer'])) {
				$query['verify_peer'] = '0';
			}
			if ($user === null and isset($query['username'])) {
				$user = $query['username'];
				$pass = $query['password'] ?? null;
				unset($query['username']);
				unset($query['password']);
			}
		}
		return $this->reconstructUri([
			'scheme' => $scheme,
			'host' => $host,
			'port' => $port,
			'user' => $user,
			'pass' => $pass,
			'path' => $path,
			'query' => $query ? http_build_query($query) : null,
		]);
	}

	/**
	 * @override
	 */
	public function boot(): void
	{
		$hadAlreadyBooted = $this->booted;
		if (!$hadAlreadyBooted) {
			$this->loadCachedResourcePreBoot();
		}

		// Kompatibilität zwischen SwiftMailer und Symfony/Mailer
		if (!isset($_ENV['MAILER_DSN'])) {
			$_ENV['MAILER_DSN'] = $this->convertMailerDsn((string)($_ENV['MAILER_URL'] ?? 'null://null'));
		}

		// Sicherstellen, dass Twig-Templates-Verzeichnis existiert, wenn
		// das Cache-Verzeichnis selbst existiert.
		$cacheDir = $this->getKernelParameters()['kernel.cache_dir'];
		if (!file_exists($cacheDir.'/templates')) {
			@mkdir($cacheDir.'/templates');
		}

		// Plugin-Autoloader-Pfade laden
		if (!$hadAlreadyBooted) {
			$paths = @json_decode(file_get_contents($cacheDir.'/plugins-autoload.json'), true);
			if (is_array($paths)) {
				foreach ($paths as $prefix => $path) {
					if (($prefix[0] ?? '') === '@') {
						// Plugin autoload.php laden
						if (file_exists($path)) {
							include_once($path);
						}
					} else {
						// Plugin-Pfad zu Autoloader hinzufügen
						$this->autoloader->addPsr4($prefix, $path);
					}
				}
			}
		}

		parent::boot();

		if (!$hadAlreadyBooted) {
			$this->loadCachedResourcePostBoot();
		}
	}

	public function registerBundles(): iterable
	{
		$contents = require $this->getProjectDir().'/config/bundles.php';
		foreach ($contents as $class => $envs) {
			if ($envs[$this->environment] ?? $envs['all'] ?? false) {
				yield new $class();
			}
		}
	}

	public function getProjectDir(): string
	{
		return \dirname(__DIR__);
	}

	private function configureRedis(ContainerBuilder $container, string $redis_dsn) {
		$parts = parse_url($redis_dsn);
		$hosts = $parts['host'];
		if (!empty($parts['port'])) {
			$hosts .= ':'.$parts['port'];
		}
		$query = [];
		parse_str($parts['query'] ?? '', $query);
		$options = [];

		// Bei Replikation URI parsen und Daten für die einzelnen Server konstruieren
		if (strpos($hosts, ',') !== false or array_key_exists('replication', $query)) {
			if (!isset($query['username']) and isset($parts['user']) and $parts['user'] !== '') {
				$query['username'] = $parts['user'];
			}
			if (!isset($query['password']) and isset($parts['pass'])) {
				$query['password'] = $parts['pass'];
			}
			$path = null;
			if (!isset($query['database']) and isset($parts['path']) and preg_match('/^\/(\d+)(\/.*)?/', $parts['path'], $path)) {
				$query['database'] = $path[1];
				$query['path'] = $path[2];
			}

			$replication = $query['replication'] ?? 'predis';
			$options['replication'] = $replication;
			$options['parameters'] = $query;
			unset($query['replication']);
			if (isset($query['service'])) {
				$options['service'] = $query['service'];
				unset($query['service']);
			}

			$parameters = [];

			foreach (explode(',', $hosts) as $i => $host) {
				$data = parse_url($parts['scheme'].'://'.$host);
				if (isset($query['sentinelPassword'])) {
					$data['password'] = $query['sentinelPassword'];
				}

				if ($replication === 'predis') {
					$data['role'] = ($i === 0) ? 'master' : 'slave';
				}

				if (stripos($data['host'], '%2f') === 0) {
					$data['scheme'] = 'unix';
					$data['path'] = urldecode($data['host']);
					unset($data['host']);
				}
				$parameters[] = $data;
			}
		} else {
			// Bei "normalem" Redis die URI direkt an Predis geben.
			$parameters = $redis_dsn;
		}
		$container->setParameter('app.redis.parameters', $parameters);
		$container->setParameter('app.redis.options', $options);
	}

	protected function configureContainer(ContainerBuilder $container, LoaderInterface $loader): void
	{
		$projectDir = $this->getProjectDir();
		$container->addResource(new FileResource($projectDir.'/config/bundles.php'));
		$container->setParameter('container.dumper.inline_class_loader', true);
		$confDir = $projectDir.'/config';

		switch ($_ENV['APP_SESSION_HANDLER'] ?? 'php') {
			case 'redis':
				if (class_exists('Redis')) {
					$container->setParameter('app.session.handler', \Symfony\Component\HttpFoundation\Session\Storage\Handler\RedisSessionHandler::class);
					break;
				}
			case 'php':
			default:
				$container->setParameter('app.session.handler', null);
		}

		$this->configureRedis($container, self::getRedisDsnFromEnvironment());

		// Plugins vorbereiten
		$forcedPlugins = [
			// TODO Parse composer.json files and determine forced plugins automatically
			'Ccm19ManagerConstraints' => 'ccm19/manager_constraints',
			'Ccm19Payment' => 'ccm19/payment',
		];
		$autoloaderPaths = [];
		$translationPaths = [];
		$templatePaths = [];
		$configPaths = [];
		$cachePath = $this->getKernelParameters()['kernel.cache_dir'];
		// Erzwungene Plugins aktivieren
		foreach ($forcedPlugins as $dirname => $pluginuid) {
			if (\App\Model\PluginState::findByUID($pluginuid) === null and file_exists($projectDir.'/plugins/'.$dirname)) {
				$state = \App\Model\PluginState::create();
				$state->setName($dirname);
				$state->setVersion('1.0');
				$state->setDirname($dirname);
				$state->setPluginUID($pluginuid);
				$state->setActivatedForAllAccounts(true);
				$state->save();
			}
		}
		// Aktivierte Plugins einbinden
		$pluginManager = new PluginManager();
		$plugins = $pluginManager->configuredPlugins();
		foreach ($plugins as $plugin) {
			$pluginState = $plugin->getState();
			$dirname = $plugin->getDirname();
			if ($dirname === '') {
				continue;
			}

			$pluginPath = $projectDir.'/plugins/'.$plugin->getDirname();
			$pluginConfigPath = $pluginPath.'/config';
			$pluginSrcPath = $pluginPath.'/src';
			$pluginTemplatesPath = $pluginPath.'/templates';
			$pluginTranslationPath = $pluginPath.'/translations';

			// Wenn Plugin Up-/Downgrade
			if ($pluginState->getVersion() !== $plugin->getVersion()) {
				$pluginState->setVersion($plugin->getVersion())->save();
				// TODO: Update-Routinen ermöglichen
			}

			// Service-Definitionen laden
			if (file_exists($pluginConfigPath)) {
				$configPaths[] = [$dirname, $pluginConfigPath.'/{services}'.self::CONFIG_EXTS, 'glob'];
			}

			// Wenn ein src-Verzeichnis existiert:
			if (file_exists($pluginSrcPath)) {
				// Wenn keine Service-Definitionen vorhanden, automatisch mit Standardannahmen bauen
				if (!glob($pluginConfigPath.'/services'.self::CONFIG_EXTS, GLOB_BRACE)) {
					$template = file_get_contents($confDir.'/plugin.default.services.yaml');
					$data = strtr($template, [
						'\'../' => '\'%kernel.project_dir%/plugins/'.$dirname.'/',
						'%NAMESPACE%' => $pluginState->getNamespace(),
						'%UID%' => addcslashes($pluginState->getPluginUID(), '"\'\\'),
						'%DIRNAME%' => addcslashes($dirname, '"\'\\')
					]);
					if (file_exists($pluginSrcPath.'/Controller')) {
						$template .= "\n" . file_get_contents($confDir.'/plugin.default.controller.services.yaml');
					}
					$tmpConfFile = $cachePath.'/plugin-'.$pluginState->getId().'.yaml';
					file_put_contents($tmpConfFile, $data, LOCK_EX);
					$configPaths[] = [$dirname, $tmpConfFile, 'yaml'];
				}

				// src in Autoloader aufnehmen
				$namespace = $pluginState->getNamespace();
				$this->autoloader->addPsr4($namespace, dirname(__DIR__) . '/plugins/'.$dirname.'/src/');
				$autoloaderPaths[$namespace] = dirname(__DIR__) . '/plugins/'.$dirname.'/src/';
			}

			// Wenn vendor/autoload.php existiert:
			if (file_exists($pluginPath.'/vendor/autoload.php')) {
				require_once($pluginPath.'/vendor/autoload.php');
				$autoloaderPaths['@'.$dirname.'/vendor'] = $pluginPath.'/vendor/autoload.php';
			}

			// Templates verwenden, wenn vorhanden
			if (file_exists($pluginTemplatesPath)) {
				$templatePaths[$pluginState->getTemplateNamespace()] = $pluginTemplatesPath;
			}

			// Übersetzungsverzeichnis verwenden, wenn vorhanden
			if (file_exists($pluginTranslationPath)) {
				$translationPaths[] = $pluginTranslationPath;
			}
		}

		// Autoloader-Pfade speichern
		file_put_contents($cachePath.'/plugins-autoload.json', json_encode($autoloaderPaths), LOCK_EX);

		// Übersetzungspfade von Plugins hinzufügen
		$projectDirPattern = '~^'.preg_quote($projectDir, '~').'~';
		$container->loadFromExtension('framework', [
			'translator' => [
				'paths' => array_map(function ($path) use ($projectDirPattern) {
					return preg_replace($projectDirPattern, '%kernel.project_dir%', $path);
				}, $translationPaths),
			],
		]);

		// Templates-Pfade von Plugins hinzufügen
		$container->loadFromExtension('twig', [
			'paths' => array_flip(array_map(function ($path) use ($projectDirPattern) {
				return preg_replace($projectDirPattern, '%kernel.project_dir%', $path);
			}, $templatePaths)),
		]);

		// Packages und Services laden
		$loader->load($confDir.'/{packages}/*'.self::CONFIG_EXTS, 'glob');
		$loader->load($confDir.'/{packages}/'.$this->environment.'/**/*'.self::CONFIG_EXTS, 'glob');
		$loader->load($confDir.'/{services}'.self::CONFIG_EXTS, 'glob');
		$loader->load($confDir.'/{services}_'.$this->environment.self::CONFIG_EXTS, 'glob');

		// Plugins laden
		foreach ($configPaths as list($pluginDirname, $path, $type)) {
			try {
				$loader->load($path, $type);
			} catch (Exception $e) {
				// TODO: (sinnvolles) Error-Handling
				throw $e;
			}
		}

	}

	private function configureRoutes(RoutingConfigurator $routes): void
	{
		$projectDir = $this->getProjectDir();
		$confDir = $this->getConfigDir();

		$routes->import($confDir.'/{routes}/'.$this->environment.'/**/*'.self::CONFIG_EXTS, 'glob');
		$routes->import($confDir.'/{routes}/*'.self::CONFIG_EXTS, 'glob');
		$routes->import($confDir.'/{routes}'.self::CONFIG_EXTS, 'glob');

		$plugins = \App\Model\PluginState::all();
		foreach ($plugins as $plugin) {
			$pluginPath = $projectDir.'/plugins/'.$plugin->getDirname().'/src';
			if (file_exists($pluginPath)) {
				$routes->import($pluginPath, 'annotation');
			}


		}
	}

	/**
	 * Whitelabeling: Übersetzungen injecten
	 */
	public function process(ContainerBuilder $container): void
	{
		@mkdir(Utils::getCacheDir() . '/templates/');
		$store = Utils::getRepositoryStore();
		if (Utils::isExtendedEdition()) {
			$varDir = rtrim(Utils::getBundledVarDir(), '/\\');
			@mkdir($varDir . '/translations');
			$translator = $container->findDefinition('translator');
			$whitelabelDir = $varDir . '/translations';
			$whitelabelTemplate = $varDir . '/translations/messages.%LANG%';

			$options = $translator->getArgument(4);
			foreach ($options['resource_files'] as $lang => &$files) {
				switch ($store) {
					case Utils::REPOSITORY_STORE_MONGODB:
						$whitelabelFile = str_replace('%LANG%', $lang, $whitelabelTemplate).'.database';
						if (!file_exists($whitelabelFile)) {
							touch($whitelabelFile);
						}
						$files[] = $whitelabelFile;

						$whitelabelFile = str_replace('%LANG%', $lang, $whitelabelTemplate).'.yaml';
						if (file_exists($whitelabelFile)) {
							$files[] = $whitelabelFile;
						}
						break;
					case Utils::REPOSITORY_STORE_FILE:
					default:
						$whitelabelFile = str_replace('%LANG%', $lang, $whitelabelTemplate).'.yaml';
						if (!file_exists($whitelabelFile)) {
							touch($whitelabelFile);
						}
						$files[] = $whitelabelFile;

						$whitelabelFile = str_replace('%LANG%', $lang, $whitelabelTemplate).'.database';
						if (file_exists($whitelabelFile)) {
							$files[] = $whitelabelFile;
						}
						break;
				}
			}
			$options['scanned_directories'][] = $whitelabelDir;
			$options['cache_vary']['scanned_directories'][] = Utils::buildRelativePath($whitelabelDir, Utils::getBaseDir());

			$translator->replaceArgument(4, $options);
		}

		switch ($store) {
			case Utils::REPOSITORY_STORE_MONGODB:
				$container->setAlias('Symfony\Component\Translation\Loader\LoaderInterface', 'App\Component\Translation\MongoDbLoader');
				$container->setAlias('Symfony\Component\Translation\Dumper\DumperInterface', 'App\Component\Translation\MongoDbDumper');
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$container->setAlias('Symfony\Component\Translation\Loader\LoaderInterface', 'translation.loader.yml');
				$container->setAlias('Symfony\Component\Translation\Dumper\DumperInterface', 'translation.dumper.yaml');
				break;
		}
	}

	/**
	 * Pfad zur CACHEDIR.TAG
	 */
	public function getCacheTagFile(): string {
		return dirname(rtrim(Utils::getCacheDir(), '/')).'/CACHEDIR.TAG';
	}

	/**
	 * @override
	 * Cache-Unterverzeichnis nach Änderungstimestamp von CACHEDIR.TAG
	 */
	public function getCacheDir(): string
	{
		if ($this->cacheDir === null) {
			// Cache-Verzeichnis festlegen und Cache-Tag erstellen
			$cacheRoot = $this->getProjectDir().'/var/cache';
			$cacheCheckFile = $cacheRoot.'/CACHEDIR.TAG';
			if (!file_exists($cacheCheckFile)) {
				file_put_contents($cacheCheckFile, "Signature: 8a477f597d28d172789f06886806bc55\n# This file is a cache directory tag created by CCM19.\n# For information about cache directory tags, see:\n#       http://www.brynosaurus.com/cachedir/\n", LOCK_EX);
				clearstatcache();
			}
			$this->cacheDir = $cacheRoot.'/'.$this->environment.'_'.date("YmdHis", filemtime($cacheCheckFile));
		}

		return $this->cacheDir;
	}

	/**
	 * Änderungsdatum für Cache ändern
	 */
	public function touch(): void
    {
		$tagfile = $this->getCacheTagFile();
		@touch($tagfile);
	}

	/**
	 * Ermittelt, welche IP-Adressen/User-Agents von der Call-Count-Zählung ausgeschlossen werden sollen
	 *
	 * Die Information wird für 60 Sekunden gecached.
	 */
	private function getCountExemptionInfo(): array
	{
		$cachePath = $this->getCacheDir() . '/count-exempt-info.php';
		if (file_exists($cachePath) and (time() - filemtime($cachePath)) < 60) {
			return include($cachePath);
		} else {
			$config = Config::getInstance();
			$data = [
				'addresses' => $config->getExemptAddresses(),
				'userAgents' => $config->getExemptUserAgents()
			];

			$tmpPath = $cachePath.'.'.rand(0, 65535).'.tmp';
			$f = @fopen($tmpPath, 'xb');
			if ($f) {
				$phpData = "<?php\nreturn ".var_export($data, true).";\n";
				if (fwrite($f, $phpData) === strlen($phpData)) {
					rename($tmpPath, $cachePath);
				} else {
					unlink($tmpPath);
				}
				fclose($f);
			}

			return $data;
		}
	}

	private function loadCachedResourcePreBoot(): void
	{
		// Create from globals because with Symfony 5.x the request isn't globally accessible anymore.
		$request = Request::createFromGlobals();

		// Was script invoked by bin/console?
		if (!$request) {
			return;
		}

		// Main-Javascript
		if (preg_match('~/(app|ccm19)\.js\?~', $request->getRequestUri())) {
			$apiKey = (string)$request->query->get('apiKey');
			$domainId = (string)$request->query->get('domain');
			if ($apiKey && $domainId) {
				// GeoIP-Überprüfung, wenn aktiviert
				if (Cache::isDomainEuOnly($apiKey, $domainId, $request->query->get('theme'))) {
					$countryCode = $request->headers->get("X-Country", $request->headers->get("CF-IPCountry", ''));
					if ($countryCode) {
						$country = Country::find($countryCode);
					} else {
						$geoip = GeoIp::find($request->getClientIp());
						$country = $geoip ? $geoip->getCountry() : null;
					}
					$noConsentRequired = $country ? !$country->hasGdprOrSimilarLaw() : false;
				}
				else {
					$noConsentRequired = false;
				}

				// Language detection if not set in the query string
				$lang = $request->query->get('lang');
				if (!is_string($lang)) {
					$lang = null;
				}

				// Load cache content
				Cache::loadAndSendResponse($request, $apiKey, $domainId, [
					'lang' => $lang,
					'noConsentRequired' => $noConsentRequired,
				], function ($meta) use ($request) {
					$userAgent = $request->headers->get('User-Agent', '');
					$exempt = $this->getCountExemptionInfo();
					$isIpAddressMatch = IpUtils::checkIp($request->getClientIp(), $exempt['addresses']);
					$isUserAgentMatch = !Utils::all($exempt['userAgents'], function ($token) use ($userAgent) {
						return strpos($userAgent, $token) === false;
					});

					// Aufrufzähler inkrementieren
					if (!$isIpAddressMatch and !$isUserAgentMatch) {
						AccessCountJournal::byUserIdDomainIdAndKind($meta['userId'], $meta['domainId'], 'call')->increment();
					}
				});
			}
		}

		// Widget
		elseif (preg_match('~/(?:widget)(?=\?)~', $request->getRequestUri(), $match)) {
			$apiKey = (string)$request->query->get('apiKey');
			$domainId = (string)$request->query->get('domain');
			if ($apiKey && $domainId) {
				$callback = function ($meta) use ($request) {
					if ($request->headers->get('X-CCM19State') == 'new-view') {
						AccessCountJournal::byUserIdDomainIdAndKind($meta['userId'], $meta['domainId'], 'view')->increment();
					}
				};

				// GeoIP-Überprüfung, wenn aktiviert
				if (Cache::isDomainWidgetVaryCountry($apiKey, $domainId, $request->query->get('theme'))) {
					$countryCode = $request->headers->get("X-Country", $request->headers->get("CF-IPCountry", ''));
					if ($countryCode) {
						$country = Country::find($countryCode);
					} else {
						$geoip = GeoIp::find($request->getClientIp());
						$country = $geoip ? $geoip->getCountry() : null;
						$countryCode = $geoip ? $geoip->getCountryCode() : '';
					}
					if (!$country or $country->hasGdpr()) {
						$targetRegion = 'EU';
					}
					elseif (in_array($countryCode, ThemeBase::RESTRICTABLE_COUNTRIES)) {
						$targetRegion = $countryCode;
					}
					else {
						$targetRegion = implode(',', array_merge(
							["!EU"],
							array_map(function ($code) { return "!".$code; }, ThemeBase::RESTRICTABLE_COUNTRIES)
						));
					}
				}
				else {
					$targetRegion = null;
				}

				Cache::loadAndSendResponse($request, $apiKey, $domainId, [ 'targetRegion' => $targetRegion ], $callback);
			}
		}

		// CSS/Details-Window
		elseif (preg_match('~/(?:(?:app|ccm19)\.css|widget/details)(?=\?)~', $request->getRequestUri(), $match)) {
			$apiKey = (string)$request->query->get('apiKey');
			$domainId = (string)$request->query->get('domain');
			if ($apiKey && $domainId) {
				Cache::loadAndSendResponse($request, $apiKey, $domainId, []);
			}
		}

		elseif ($apiKey = $request->query->get('apiKey') and $domainId = $request->query->get('domain')) {
			Cache::loadAndSendResponse($request, $apiKey, $domainId);
		}
	}

	private function loadCachedResourcePostBoot(): void
	{
		/** @var RequestStack $requestStack */
		$requestStack = $this->getContainer()->get('request_stack');

		// Create from globals because with Symfony 5.x the request isn't globally accessible anymore.
		$request = $requestStack->getCurrentRequest() ?? Request::createFromGlobals();

		// Was script invoked by bin/console?
		if (!$request) {
			return;
		}

		if (preg_match('~/(app|ccm19)\.js\?~', $request->getRequestUri())) {
			$apiKey = (string)$request->query->get('apiKey');

			// Fallback: In alten Systemen ohne API-Key und nur einem Benutzer, diesen Benutzer verwenden
			if (empty($apiKey) && Utils::isBaseEdition() && User::count() == 1) {
				$apiKey = array_values(User::all())[0]->getApiKey();
			}

			// Prüfe auf gültigen Benutzer
			if (User::setApiUser($apiKey)) {
				// Aktiven Benutzer und Domaindatenbank umschalten
				$user = User::loggedInUser();
				assert($user !== null);
				Domain::switchToUser($user);

				$domainId = (string)$request->query->get('domain');

				// Domain-Autoerkennung
				if (!$domainId and Utils::isExtendedEdition()) {
					$domainName = Utils::getOriginDomain($request, false) ?? '';
					$domainId = DomainIndex::findDomainId($domainName);
				}

				// Prüfe auf gültige Domain
				if (Domain::setDomainForRequest($domainId)) {
					$domain = Domain::activeDomain();

					// Verwende angeforderte Sprache, sofern vorhanden, andernfalls rate das Locale
					$localeName = (string)$request->query->get('lang', '');
					$localeSpecified = strlen($localeName) > 0;

					$locales = LocaleService::enabledLocales();
					if ($localeSpecified) {
						$locale = LocaleService::find($localeName);

						// Wenn ein nicht-existentes Locale angegeben wird, Fallback zum Standard
						if (!$locale) {
							$locale = User::loggedInUser()->getDefaultLocale();
						}
					} else {
						$locale = null;
					}

					// Im weiteren Verlauf unbenötigte Variablen aus dem Scope entfernen
					unset($localeName);

					// GeoIP-Überprüfung, wenn aktiviert
					if (Cache::isDomainEuOnly($apiKey, $domainId, $request->query->get('theme'))) {
						$geoIp = GeoIp::find($request->getClientIp());
						$noConsentRequired = $geoIp && !$geoIp->isInEu();
					}
					else {
						$noConsentRequired = false;
					}

					Cache::loadAndSendResponse($request, $apiKey, $domainId, [
						'lang' => ($locale) ? $locale->getName() : null,
						'noConsentRequired' => $noConsentRequired,
					], function ($meta) use ($domain, $request) {
						$userAgent = $request->headers->get('User-Agent', '');

						$exempt = $this->getCountExemptionInfo();
						$isIpAddressMatch = IpUtils::checkIp($request->getClientIp(), $exempt['addresses']);
						$isUserAgentMatch = !Utils::all($exempt['userAgents'], function ($token) use ($userAgent) {
							return strpos($userAgent, $token) === false;
						});

						// Aufrufzähler inkrementieren
						if ($domain and !$isIpAddressMatch and !$isUserAgentMatch) {
							AccessCountJournal::byDomainAndKind($domain, 'call')->increment();
						}
					});
				}
			}
		}
	}

	/**
	 * Programmablauf blockieren, wenn ein Update gerade läuft und der Request nicht Teil des Update-Prozesses ist.
	 * Wenn der Prozess nach anderthalb Sekunden noch nicht freigegeben wurde, wird er stattdessen abgebrochen.
	 * @return void
	 */
	private static function ccm19BlockWhenInUpdate(): void
	{
		$updateCheckFile = dirname(__DIR__) . '/var/update.lock';
		if (!file_exists($updateCheckFile)) {
			return;
		}

		$f = @fopen($updateCheckFile, 'r');
		if (!$f) {
			return;
		}

		// insg. 1,5 s versuchen zu verzögern
		for ($i = 0; $i < 15; ++$i) {
			$wouldBlock = false;
			if (@flock($f, LOCK_SH | LOCK_NB, $wouldBlock)) {
				flock($f, LOCK_UN);
			}
			if (!$wouldBlock) {
				clearstatcache();
				return;
			}
			// Warte 100 ms
			usleep(100000);
		}

		$urlPath = ($_SERVER['REQUEST_URI'] ?? $_SERVER['REDIRECT_URL']);
		if (strpos($urlPath, '/update/') === false)  {
			http_response_code(503);
			header('Refresh: 5');
			if (strpos($urlPath, '.js') !== false) {
				header('Content-Type: application/javascript');
				header('Cache-Control: no-cache, no-store');
				echo "console.info('[CCM19] Update in progress. Retry in a few seconds.');";
			} else {
				header('Content-Type: text/plain');
				header('Cache-Control: no-cache, no-store');
				echo "Update in progress. Retry in a few seconds.";
			}
			exit();
		}
		clearstatcache();
	}

	private static function handleCustomRouting(): void
	{
		// Special Cloud Urls wg. Altlasten
		if (($_ENV['APP_CCM_CLOUD'] ?? null) == '1' && $_SERVER['HTTP_HOST'] != '5f3c395.ccm19.de') {
			$_SERVER['REQUEST_URI'] = preg_replace('~^/app/public(/|$)~i', '/', $_SERVER['REQUEST_URI']);
		}

		// Special CK Cloud
		if (($_ENV['APP_CCM_CLOUD_CL'] ?? null) == '1') {
			$matches = null;
			if (!preg_match("~/([0-9a-f]{2})/([0-9a-f]{2})/(\\1\\2[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})/([^?]*).*~", $_SERVER['REQUEST_URI'], $matches)) {
				preg_match("~/()()([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})/([^?]*).*~", $_SERVER['REQUEST_URI'], $matches);
			}

			if ($matches) {
				$baseUrl = substr($_SERVER['REQUEST_URI'], 0, strlen($_SERVER['REQUEST_URI']) - strlen($matches[0]));
				$legacyId = $matches[3];
				$subpath = $matches[4];
				$mapping = include(dirname(__DIR__).'/ck-legacy.php');

				$mappingItem = $mapping[$legacyId] ?? null;
				$apiKey = $mappingItem['apiKey'] ?? '';
				$domainId = $mappingItem['domainId'] ?? '';

				$_GET['apiKey'] = $_REQUEST['apiKey'] = $apiKey;
				$_GET['domain'] = $_REQUEST['domain'] = $domainId;

				if ($subpath === 'bundle.js') {
					$_SERVER['REQUEST_URI'] = $baseUrl."/app.js?";
				} elseif ($subpath === 'disclaimer.js') {
					header("Content-Type: application/javascript; charset=ascii");
					header("Cache-Control: public, max-age=86400");
					echo <<<'EOF'
(function () {
	"use strict";
	var script = document.getElementById("clickskeks-disclaimer-script");
	if (!script) script = document.currentScript;
	if (!script) return;
	var div = document.createElement('div');
	div.className = 'clickskeks-cookie-declaration';
	script.parentNode.insertBefore(div, script);
	window.addEventListener('DOMContentLoaded', function () {
		Clickskeks.buildDisclaimerWithClass("clickskeks-cookie-declaration");
	});
})();
EOF;
					exit();
				} elseif ($subpath === 'bundle-async.js') {
					header("Content-Type: application/javascript; charset=ascii");
					header("Cache-Control: public, max-age=86400");
					echo <<<'EOF'
(function (window) {
	"use strict";

	console.warn("Clickskeks: Deprecated file 'bundle-async.js' loaded. Please check if your website is caching an old version of Clickskeks.");

	if (!('CCM' in window)) {
		var script = document.currentScript;
		if (!script) return;
		script.src = script.src.replace("bundle-async.js", "bundle.js");

		var newScript = document.createElement('script');
		newScript.src = script.src.replace("bundle-async.js", "bundle.js");
		script.parentNode.appendChild(newScript);
	}
})(window);
EOF;
					exit();
				} elseif (substr($subpath, 0, 7) === 'config-' and substr($subpath, -5) === '.json') {
					http_response_code(404);
					header("Content-Type: application/javascript; charset=ascii");
					header("Cache-Control: public, max-age=86400");
					header("Access-Control-Allow-Origin: *");
					echo '{ "error":"deprecated file loaded" }';
					exit();
				}
			}
		}
	}

	/**
	 * @return string
	 */
	private static function getRedisDsnFromEnvironment(): string
	{
		// Rückwärtskompatibilität Redis
		if (!isset($_ENV['REDIS_DSN'])) {
			if (($_ENV['REDIS_PASSWORD'] ?? '') !== '') {
				$_ENV['REDIS_DSN'] = 'redis://:'.urlencode($_ENV['REDIS_PASSWORD']).'@'.(string)($_ENV['REDIS_HOST'] ?? 'localhost').':'.(string)($_ENV['REDIS_PORT'] ?? '6379');
			} else {
				$_ENV['REDIS_DSN'] = 'redis://'.(string)($_ENV['REDIS_HOST'] ?? 'localhost').':'.(string)($_ENV['REDIS_PORT'] ?? '6379');
			}
		}
		return $_ENV['REDIS_DSN'];
	}
}
