<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Utils;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\HttpFoundation\Request;

/**
 * Übersetzungen für die Administration
 */
class BackendLanguage extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-languages-settings';

	/**
	 * @return self[]
	 */
	public static function all(): array
	{
		// Remove redundant languages
		$languages = array_reduce(parent::all(), function (array $languages, self $language) {
			$languages[$language->getLongCode()] = $language;
			return $languages;
		}, []);
		return array_combine(
			array_map(function (self $language) { return $language->getId(); }, $languages),
			$languages
		);
	}

	/**
	 * @param Language $lang
	 * @return self|null
	 * @phan-side-effect-free
	 */
	public static function byLanguage(Language $lang): ?self
	{
		return self::byLongCode($lang->getId());
	}

	/**
	 * @param string $code
	 * @return self|null
	 * @phan-side-effect-free
	 */
	public static function byLongCode(string $code): ?self
	{
		$code = strtolower($code);
		return array_reduce(self::all(), function ($carry, $lang) use ($code) {
			return $carry ?? (strtolower($lang->getLongCode()) == $code ? $lang : $carry);
		}, null);
	}

	/**
	 * @param string $code
	 * @return self|null
	 * @phan-side-effect-free
	 */
	public static function byFuzzyLangCode(string $code): ?self
	{
		$code = strtr($code, ['-' => '_']);
		$parts = explode('_', $code, 4);
		$short = strtolower($parts[0]);
		// Script rausstrippen und Land bestimmen
		$country = '';
		if (isset($parts[1]) and strlen($parts[1]) === 4) {
			$country = strtoupper($parts[2] ?? '');
		} else {
			$country = strtoupper($parts[1] ?? '');
		}

		$long = ($country != '') ? $short.'_'.$country : $short;

		$result = null;
		$matchQuality = 0;
		foreach (self::all() as $lang) {
			if ($lang->getShortCode() == $short) {
				if ($lang->getLongCode() == $long) {
					$result = $lang;
					$matchQuality = 3;
				}
				elseif ($matchQuality < 3 and ($assumedCountry = $lang->getAssumedCountry()) and $assumedCountry->getCode() == $country) {
					$result = $lang;
					$matchQuality = 2;
				}
				elseif (!$country and $matchQuality < 2) {
					$result = $lang;
					$matchQuality = 1;
				}
			}
		}
		return $result;
	}

	/**
	 *
	 */
	public static function updateAvailableLanguages()
	{
		$knownLanguages = Language::all();
		$filesystem = new Filesystem();
		$baseDir = Utils::getBaseDir();

		// Für alle Sprachen überprüfen, ob sie übersetzt sind
		foreach ($knownLanguages as $language) {
			$short = $language->getShortCode();
			$long = $language->getId();

			$prefix = $baseDir . "/translations/" . "messages.";
			$isTranslated = ($filesystem->exists($prefix . $short . '.yaml') or $filesystem->exists($prefix . $long . '.yaml'));

			$backendLang = self::byLanguage($language);
			if ($isTranslated) {
				// Wenn übersetzt, aber noch nicht in der Backend-Liste, hinzufügen
				if (!$backendLang) {
					$backendLang = self::create();
					$backendLang->setLanguage($language);
					$backendLang->setActive($language->isActiveByDefault());
					$backendLang->setDefault(false);
				}
			}
			// Wenn nicht übersetzt, aber in der Backend-Liste, rausnehmen
			elseif ($backendLang) {
				self::delete($backendLang->getId(), false);
			}
		}

		$langStandard = self::findDefault();
		// Wenn keine Sprache Standard, dann Deutsch nehmen
		if (!$langStandard) {
			$newStandard = self::byLongCode("de_DE");
			$newStandard->setDefault(true);
		}

		// Änderungen committen
		self::flush();
	}

	/**
	 * @return self|null
	 * @phan-side-effect-free
	 */
	public static function findDefault(): ?self
	{
		$name = true;
		return array_reduce(self::all(), function ($carry, $lang) use ($name) {
			return $carry ?? ($lang->isDefault() ? $lang : $carry);
		}, null);
	}

	/**
	 * @return self[]
	 */
	public static function allActive(): array
	{
		//setup - falls leer erstmal befüllen...
		if(empty(self::all()))
		{
			self::updateAvailableLanguages();
		}

		return array_filter(self::all(), function ($item) {
			return $item->isActive();
		});
	}

	/**
	 * Gibt die Liste der Backend-Sprachen sortiert aus.
	 *
	 * Deutsch und Englisch sind immer die ersten beiden Einträge. Danach wird (wenn `$request`
	 * angegeben ist) nach der Sprachpräferenz des Browsers, und schließlich nach der alphabetischen
	 * Reihenfolge des Sprachcodes sortiert.
	 *
	 * @param Request|null $request Request zur personalisierten Sortierung
	 * @return self[]
	 */
	public static function allActiveSorted(?Request $request=null): array
	{
		$list = self::allActive();
		// Präferenzliste ermitteln
		$preferredLanguages = ["de_DE", "en_US"];
		if ($request) {
			$filterLanguages = ['de', 'en', 'de_DE', 'en_US'];
			$preferredLanguages = array_merge(
				$preferredLanguages,
				array_filter(
					$request->getLanguages(),
					static function ($lang) use ($filterLanguages) {
						return !in_array($lang, $filterLanguages);
					}
				)
			);
		}
		// Nach Präferenzliste sortieren mit Fallback auf Sortierung nach Sprachcode
		usort($list, function (BackendLanguage $a, BackendLanguage $b) use ($preferredLanguages) {
			$aCode = $a->getLongCode();
			$bCode = $b->getLongCode();
			$aIndex = array_search($aCode, $preferredLanguages, true);
			$bIndex = array_search($bCode, $preferredLanguages, true);
			if ($aIndex === false) {
				$aIndex = count($preferredLanguages);
			}
			if ($bIndex === false) {
				$bIndex = count($preferredLanguages);
			}

			return $aIndex <=> $bIndex ?: $aCode <=> $bCode;
		});
		return $list;
	}

	/**
	 * @return Language
	 * @phan-suppress PhanTypeMismatchReturnNullable
	 */
	public function getLanguage(): Language
	{
		return Language::find($this->getLongCode());
	}

	/**
	 * @return string
	 */
	public function getEnglishName(): string
	{
		return $this->getLanguage()->getEnglishName();
	}

	/**
	 * @return bool
	 */
	public function isActive(): bool
	{
		return (bool)$this->get('active');
	}

	/**
	 * @return string
	 */
	public function getGermanName(): string
	{
		return $this->getLanguage()->getGermanName();
	}

	/**
	 * @return string
	 */
	public function getDisplayName(string $user_language): string
	{
		return $this->getLanguage()->getDisplayName($user_language);
	}

	/**
	 * @return string
	 */
	public function getCode(): string
	{
		return (string)$this->getId();
	}

	/**
	 * @return string
	 */
	public function getShortCode(): string
	{
		return explode('_', (string)$this->get("longCode"), 2)[0];
	}

	/**
	 * @return string
	 */
	public function getLongCode(): string
	{
		return (string)$this->get("longCode");
	}

	/**
	 * @return bool
	 */
	public function isDefault(): bool
	{
		return $this->get("standard") ?? false;
	}


	/**
	 * @return string
	 */
	public function getIsoCode(): string
	{
		return str_replace('_', '-', $this->getLongCode());
	}

	/**
	 * @param Language $language
	 * @return $this
	 */
	public function setLanguage(Language $language): self
	{
		$this->set('longCode', $language->getId());
		return $this;
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setDefault(bool $value): self
	{
		$this->set('standard', $value);
		return $this;
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setActive(bool $value): self
	{
		$this->set('active', $value);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return  $this->getLanguage()->getName();
	}

	/**
	 * Bestimmt ein Land zu dieser Sprache
	 */
	public function getAssumedCountry(): ?Country
	{
		return  $this->getLanguage()->getAssumedCountry();
	}
}
