<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Utils;
use DateTimeImmutable;

class Cookie extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-cookies';

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = true;

	/** @var bool $hasLegallyRelevantChange */
	protected $hasLegallyRelevantChange = false;

	/** @var bool $wantsMongoDbMtimeIndex */
	protected static $wantsMongoDbMtimeIndex = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Diese Methode ergänzt die abzuleitende Methode um Sortierung der Cookies nach Vertreiber und Name.
	 * @return static[]
	 */
	public static function all(): array
	{
		/** @var static[] $cookies */
		$cookies = parent::all();

		// ORDER BY vendor, name
		uasort($cookies, function ($a, $b) {
			/**
			 * @var self $a
			 * @var self $b
			 */
			return ($vendorCmp = strnatcasecmp($a->getCookieVendor(), $b->getCookieVendor())) == 0
				? strnatcasecmp($a->getName(), $b->getName())
				: $vendorCmp;
		});

		return $cookies;
	}

	/**
	 * @param Purpose $purpose
	 * @return static[]
	 */
	public static function byPurpose(Purpose $purpose): array
	{
		return array_filter(static::all(), function (Cookie $obj) use ($purpose) {
			return $obj->getPurpose()->getId() == $purpose->getId();
		});
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @param mixed $data
	 * @return string
	 */
	public static function getNameFromGet($data): string
	{
		return (string) $data;
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setName(string $name): self
	{
		if ($name !== $this->get('name')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('name', $name);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCodeSnippet(): string
	{
		return (string)$this->get('code');
	}

	/**
	 * @param string $code
	 * @return $this
	 */
	public function setCodeSnippet(string $code): self
	{
		$this->set('code', $code);
		return $this;
	}

	/**
	* @return string[]
	*/
	public function getCodeBlock(): array
	{
		$value = $this->get('code_block');
		if (is_string($value)) {
			$value = array_filter(explode("\n", trim($value)), function ($line) { return trim($line) !== ''; });
		}
		return (array)$value;
	}

	/**
	* @param string[] $code
	* @return $this
	*/
	public function setCodeBlock(array $code): self
	{
		$this->set('code_block', $code);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getDescription(): string
	{
		return (string)$this->get('description');
	}

	/**
	 * @param string $description
	 * @return $this
	 */
	public function setDescription(string $description): self
	{
		if ($description !== $this->get('description')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('description', $description);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedDescription(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_description.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_description.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getDescription();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $description
	 * @return $this
	 */
	public function setTranslatedDescription(LocaleBase $locale, string $description): self
	{
		$this->set('translated_description.'.$locale->getName(), $description);
		return $this;
	}

	/**
	 * @return Purpose|null
	 */
	public function getPurpose(): ?Purpose
	{
		return Purpose::find((string)$this->get('purpose'));
	}

	/**
	 * @param Purpose $purpose
	 * @return $this
	 */
	public function setPurpose(Purpose $purpose): self
	{
		$purposeId = $purpose->getId();
		if ($purposeId !== $this->get('purpose')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('purpose', $purposeId);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPrivacyPolicyUrl(): string
	{
		return (string)$this->get('privacy_policy_url');
	}

	/**
	 * @param string $url
	 * @return $this
	 */
	public function setPrivacyPolicyUrl(string $url): self
	{
		$url = trim($url);
		if ($url !== $this->get('privacy_policy_url')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('privacy_policy_url', $url);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPrivacyPolicyUrl(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_privacy_policy_url.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_privacy_policy_url.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getPrivacyPolicyUrl();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $url
	 * @return $this
	 */
	public function setTranslatedPrivacyPolicyUrl(LocaleBase $locale, string $url): self
	{
		$this->set('translated_privacy_policy_url.'.$locale->getName(), trim($url));
		return $this;
	}


	/**
	 * @return string
	 */
	public function getCookieVendor(): string
	{
		return (string)$this->get('cookie_vendor');
	}

	/**
	 * @param string $vendor
	 * @return $this
	 */
	public function setCookieVendor(string $vendor): self
	{
		if ($vendor !== $this->get('cookie_vendor')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('cookie_vendor', $vendor);
		return $this;
	}

	/**
	 * Maximale Lebenszeit des Cookies in Stunden, oder
	 * $fallback wenn die Zeit nicht eindeutig bestimmbar ist.
	 *
	 * @param mixed $sessionFallback
	 * @param mixed $fallback
	 * @return int|mixed
	 */
	public function getLifetimeHours($sessionFallback=null, $fallback=null)
	{
		return Utils::parseLifetimeToHours($this->getLifetime(), $sessionFallback, $fallback);
	}

	/**
	 * @return string
	 */
	public function getLifetime(): string
	{
		return (string)$this->get('lifetime');
	}

	/**
	 * @param string $lifetime
	 * @return $this
	 */
	public function setLifetime(string $lifetime): self
	{
		if ($lifetime !== $this->get('lifetime')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('lifetime', $lifetime);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedLifetime(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_lifetime.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_lifetime.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getLifetime();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $lifetime
	 * @return $this
	 */
	public function setTranslatedLifetime(LocaleBase $locale, string $lifetime): self
	{
		$this->set('translated_lifetime.'.$locale->getName(), $lifetime);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getHttpCookieName(): string
	{
		return (string)$this->get('http_cookie_name');
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setHttpCookieName(string $name): self
	{
		if ($name !== $this->get('http_cookie_name')) {
			$this->makingLegallyRelevantChange();
		}

		$this->set('http_cookie_name', $name);
		return $this;
	}

	/**
	 * Überprüfen, ob ein Cookie-Name zu diesem Cookie passt
	 */
	public function matchName(string $name): bool
	{
		return $this->getHttpCookieName() == $name;
	}

	/**
	 * @return bool
	 */
	public function isActive(): bool
	{
		return (bool)$this->get('active');
	}

	/**
	 * @param bool $status
	 * @return $this
	 */
	public function setActive(bool $status): self
	{
		$this->set('active', $status);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCollectedDataInfo(): string
	{
		return (string)$this->get('collected_data_info');
	}

	/**
	 * @param string $info
	 * @return $this
	 */
	public function setCollectedDataInfo(string $info): self
	{
		if ($info !== $this->get('collected_data_info')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('collected_data_info', $info);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedCollectedDataInfo(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_collected_data_info.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_collected_data_info.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getCollectedDataInfo();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $info
	 * @return $this
	 */
	public function setTranslatedCollectedDataInfo(LocaleBase $locale, string $info): self
	{
		$key = 'translated_collected_data_info.'.$locale->getName();
		if ($info !== $this->get($key) and $info !== $this->getCollectedDataInfo()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $info);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPurposeOfDataCollection(): string
	{
		return (string)$this->get('purpose_of_data_collection');
	}

	/**
	 * @param string $purpose
	 * @return $this
	 */
	public function setPurposeOfDataCollection(string $purpose): self
	{
		if ($purpose !== $this->get('purpose_of_data_collection')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('purpose_of_data_collection', $purpose);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPurposeOfDataCollection(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_purpose_of_data_collection.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_purpose_of_data_collection.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getPurposeOfDataCollection();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $purpose
	 * @return $this
	 */
	public function setTranslatedPurposeOfDataCollection(LocaleBase $locale, string $purpose): self
	{
		$key = 'translated_purpose_of_data_collection.'.$locale->getName();
		if ($purpose !== $this->get($key) and $purpose !== $this->getPurposeOfDataCollection()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $purpose);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getLegalBasis(): string
	{
		return (string)$this->get('legal_basis');
	}

	/**
	 * @param string $basis
	 * @return $this
	 */
	public function setLegalBasis(string $basis): self
	{
		if ($basis !== $this->get('legal_basis')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('legal_basis', $basis);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedLegalBasis(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_legal_basis.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_legal_basis.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getLegalBasis();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $basis
	 * @return $this
	 */
	public function setTranslatedLegalBasis(LocaleBase $locale, string $basis): self
	{
		$key = 'translated_legal_basis.'.$locale->getName();
		if ($basis !== $this->get($key) and $basis !== $this->getLegalBasis()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $basis);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPlaceOfProcessing(): string
	{
		return (string)$this->get('place_of_processing');
	}

	/**
	 * @param string $place
	 * @return $this
	 */
	public function setPlaceOfProcessing(string $place): self
	{
		if ($place !== $this->get('place_of_processing')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('place_of_processing', $place);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPlaceOfProcessing(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_place_of_processing.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_place_of_processing.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getPlaceOfProcessing();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $place
	 * @return $this
	 */
	public function setTranslatedPlaceOfProcessing(LocaleBase $locale, string $place): self
	{
		$key = 'translated_place_of_processing.'.$locale->getName();
		if ($place !== $this->get($key) and $place !== $this->getPlaceOfProcessing()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $place);
		return $this;
	}

	/**
	 * Zeitpunkt der letzten für den Consent rechtlich relevanten Änderung (sofern bereits erfasst)
	 *
	 * @return DateTimeImmutable|null
	 */
	public function getLegalLastModified(): ?DateTimeImmutable
	{
		$ts = $this->get('tsLegal');

		try {
			return $ts ? new DateTimeImmutable((string)$ts) : null;
		} catch (\Exception $e) {
			return null;
		}
	}

	/**
	 * Manuell den Zeitstempel der letzten rechtlich relevanten Änderung aktualisieren.
	 *
	 * Passiert automatisch wenn ein relevantes Feld geändert wird.
	 *
	 * @return $this
	 */
	public function updateLegalLastModified(): self
	{
		$this->hasLegallyRelevantChange = true;
		$this->set('tsLegal', (new DateTimeImmutable('now'))->format('Y-m-d H:i:s'));
		return $this;
	}

	/**
	 * Ob in der aktuellen Instanz eine rechtlich relevante Änderung gemacht wurde
	 *
	 * @return bool
	 */
	public function hasLegallyRelevantChange(): bool
	{
		return $this->hasLegallyRelevantChange;
	}

	/**
	 * Den Zeitstempel der letzten rechtlich relevanten Änderung aktualisieren, falls noch nicht
	 * passiert.
	 *
	 * Intern verwendet.
	 */
	protected function makingLegallyRelevantChange(): void
	{
		if (!$this->hasLegallyRelevantChange) {
			$this->updateLegalLastModified();
		}
	}

	/**
	 * Setzt alle übersetzten Texte, die in $locale noch nie vorher gesetzt
	 * wurden, auf den nicht-übersetzten Wert.
	 *
	 * @param LocaleBase $locale Zu füllendes Locale
	 * @return $this
	 */
	public function fillTranslationsFromDefault(LocaleBase $locale): self
	{
		$this->setTranslatedCollectedDataInfo($locale, $this->getTranslatedCollectedDataInfo($locale));
		$this->setTranslatedDescription($locale, $this->getTranslatedDescription($locale));
		$this->setTranslatedLegalBasis($locale, $this->getTranslatedLegalBasis($locale));
		$this->setTranslatedLifetime($locale, $this->getTranslatedLifetime($locale));
		$this->setTranslatedPlaceOfProcessing($locale, $this->getTranslatedPlaceOfProcessing($locale));
		$this->setTranslatedPrivacyPolicyUrl($locale, $this->getTranslatedPrivacyPolicyUrl($locale));
		$this->setTranslatedPurposeOfDataCollection($locale, $this->getTranslatedPurposeOfDataCollection($locale));
		return $this;
	}

}
