<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;

class Country extends ImmutableModel
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-countries.json';

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @return string
	 */
	public function getCode(): string
	{
		return $this->getId();
	}

	/**
	 * @return string
	 */
	public function getAlpha3Code(): string
	{
		return (string)$this->get('code3');
	}

	/**
	 * @return int
	 */
	public function getNumericCode(): int
	{
		return (int)$this->get('num_code');
	}

	/**
	 * Unicode-Flagge in UTF-8
	 */
	public function getUnicodeFlag(): string
	{
		$code = strtoupper($this->getCode());
		if ($code == 'ZZ') {
			return '';
		} else {
			return implode('', ["\xF0\x9F\x87", chr(101+ord($code[0])), "\xF0\x9F\x87", chr(101+ord($code[1]))]);
		}
	}

	/**
	 * Bestimmt eine Sprache zu diesem Land
	 */
	public function getAssumedLanguage(): ?Language
	{
		$code = $this->getCode();
		switch ($code) {
			case 'JP':
				$langCode = 'ja';
				break;
			case 'CN':
				$langCode = 'zh';
				break;
			case 'US':
			case 'GB':
			case 'CA':
			case 'AU':
				$langCode = 'en';
				break;
			case 'AT':
			case 'BE':
			case 'CH':
				$langCode = 'de';
				break;
			default:
				$langCode = strtolower($code);
		}

		$language = Language::find($langCode.'_'.$code);
		if (!$language) {
			$language = Language::find($langCode);
		}

		return $language;
	}

	/**
	 * Ob dieses Land in der EU ist
	 */
	public function isInEu(): bool
	{
		$code = $this->getCode();
		switch ($code) {
			// Mitglieder
			case 'BE':
			case 'BG':
			case 'DK':
			case 'DE':
			case 'EE':
			case 'FI':
			case 'FR':
			case 'GR':
			case 'IE':
			case 'IT':
			case 'HR':
			case 'LV':
			case 'LT':
			case 'LU':
			case 'MT':
			case 'NL':
			case 'AT':
			case 'PL':
			case 'PT':
			case 'RO':
			case 'SE':
			case 'SK':
			case 'SI':
			case 'ES':
			case 'CZ':
			case 'HU':
			case 'CY':
				return true;
			default:
				return false;
		}
	}

	/**
	 * Ob dieses Land unter die DSGVO fällt
	 */
	public function hasGdpr(): bool
	{
		// Alle EU-Länder fallen unter die DSGVO
		if ($this->isInEu()) {
			return true;
		}
		$code = $this->getCode();
		switch ($code) {
			/// Nicht in der EU, aber im EWR: DSGVO gilt auch
			case 'LI':
			case 'IS':
			case 'NO':
				return true;
			default:
				return false;
		}
	}

	/**
	 * Ob dieses Land unter die DSGVO oder ähnliche Gesetze fällt
	 *
	 * Neben EU-Ländern insbesondere Länder, für die es einen Angemessenheitsbeschluss der EU gibt.
	 */
	public function hasGdprOrSimilarLaw(): bool
	{
		if ($this->hasGdpr()) {
			return true;
		}
		$code = $this->getCode();
		switch ($code) {
			/// Unbekanntes Land (z.B. Intranet-IP). Lieber annehmen, dass die DSGVO gilt
			case 'ZZ':
				return true;
			/// Ähnliche Gesetze: Angemessenheitsbeschlüsse existieren
			case 'CH': // Schweiz
			case 'AD': // Andorra
			case 'LI': // Liechtenstein
			case 'FO': // Färöer-Inseln
			case 'GG': // Guernsey
			case 'JE': // Jersey
			case 'IM': // Isle of Man
			case 'AR': // Argentinien
			case 'JP': // Japan
			case 'NZ': // Neuseeland
			case 'KR': // Südkorea
			case 'UY': // Uruguay
				return true;
			/// Noch nicht dabei:
			case 'MC': // Monaco: Nur eingeschränkter Angemessenheitsbeschluss für Finanzdaten
			case 'SM': // San Marino: Noch kein Angemessenheitsbeschluss
			case 'VA': // Vatikan
			default:
				return false;
		}
	}
}
