<?php

namespace App\Model;

use App\Component\ConfigImporterContext;
use App\Component\Config\Store\MongoDb;
use App\Component\Curl;
use App\Component\Media\MediaFileService;
use App\Component\Tcf\v2\Model\Google\AdTechProvider;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Component\Tcf\v2\Model\Stack as TcfStack;
use App\Component\Config\RepositoryInterface;
use App\Event\DomainClonePrepareEvent;
use App\Event\DomainCloneReferencesEvent;
use App\Event\DomainDeletePrepareEvent;
use App\Utils;
use DateTime;
use DateTimeInterface;
use DateTimeZone;
use Exception;
use InvalidArgumentException;
use LogicException;
use Symfony\Component\DomCrawler\Crawler;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\Routing\Exception\RouteNotFoundException;
use RuntimeException;

class Domain extends Model
{
	const MANAGEMENT_STRUCTURE_COOKIE = 'cookie';
	const MANAGEMENT_STRUCTURE_EMBEDDING = 'embedding';

	const ALLOWED_MANAGEMENT_STRUCTURE_VALUES = [
		self::MANAGEMENT_STRUCTURE_COOKIE,
		self::MANAGEMENT_STRUCTURE_EMBEDDING,
	];

	const CONSENT_STORAGE_COOKIE = 'cookie';
	const CONSENT_STORAGE_LOCAL_STORAGE = 'localStorage';
	const CONSENT_STORAGE_SESSION_STORAGE = 'sessionStorage';

	const ALLOWED_CONSENT_STORAGE_VALUES = [
		self::CONSENT_STORAGE_COOKIE,
		self::CONSENT_STORAGE_LOCAL_STORAGE,
		self::CONSENT_STORAGE_SESSION_STORAGE,
	];

	const IFRAME_BLOCK_MODE_BLACKLIST = 'blacklist';
	const IFRAME_BLOCK_MODE_WHITELIST = 'whitelist';

	const ALLOWED_IFRAME_BLOCK_MODE_VALUES = [
		self::IFRAME_BLOCK_MODE_BLACKLIST,
		self::IFRAME_BLOCK_MODE_WHITELIST,
	];

	/** @var self|null $requestDomain */
	private static $requestDomain = null;

	/** @var User|null $switchedToUser */
	private static $switchedToUser = null;

	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-domains';

	/** @var bool $userSpecificRepository */
	protected static $userSpecificRepository = true;

	/** @var bool $accountSingleton */
	protected static $accountSingleton = true;

	/** @var User|null $user */
	private $user;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
		$this->getUser();
	}

	/**
	 * Diese Methode wird ausgeführt, sobald das Repository durch die Basisklasse Model geladen wurde.
	 */
	protected static final function onRepositoryLoaded(): void
	{
		$user = User::loggedInUser();
		if (!$user || $user->getId() != self::getRepositoryContextUserId()) {
			return;
		}

		// Ensure that at least a single domain exists for users who aren't able to create one themselves
		if ((Utils::isExtendedEdition() && $user->isAdministrator() || Utils::isBaseEdition()) && self::count() == 0) {
			Domain::createForUser($user)
				->setName('')
				->setManagementStructurePreLegalConsent(true)
				->setManagementStructure(Domain::MANAGEMENT_STRUCTURE_EMBEDDING)
				->setWhitelabel(false)
			;
			self::flush();
		}
	}

	/**
	 * Übernimmt Eigenschaften des ursprünglichen Objekts aus dem Repository.
	 * Fordert exklusiven Lock an und gibt abschließend wieder frei.
	 */
	public function __clone()
	{
		// Generiere neue ID und stelle ursprüngliche Domain in $this->source bereit
		parent::__clone();

		// Generiere aufsteigende Nummerierung für neuen Namen
		for ($i = 2;; $i++) {
			if (!self::byName($newDomainName = "{$this->source->getName()} #$i")) {
				break;
			}
		}

		// Neue Ausangswerte vorgeben und Domain speichern
		$this
			->setName($newDomainName)
			->setCrossDomainSharing(false)
			->setFrontendWidgetEnabled(false)
			->setShareDomainNames([])
			->setViewCount(0)
			->setWhitelabel(false)
			->save();

		// The user who initiated the cloning process
		$initiator = User::loggedInUser();
		assert($initiator instanceof User);

		// Grant access to subuser
		if ($initiator->isSubUser()) {
			$initiator
				->grantDomainAccess($this)
				->save();
		}

		// The user who owns the original and the new domain (i.e. the account owner)
		$user = $this->getUser();
		assert($user instanceof User);

		// Domainspezifische Repositories (Embeddings, LegalTexts, Themes etc.) kopieren
		if (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB) {
			// Models sammeln, deren Objekte kopiert werden sollen
			$domainClonePrepareEvent = new DomainClonePrepareEvent();
			Utils::dispatchEvent(DomainClonePrepareEvent::NAME, $domainClonePrepareEvent);

			$domainSpecificModels = [
				Cookie::class,
				Embedding::class,
				EmbeddingAsset::class,
				LegalText::class,
				Locale::class,
				Placeholder::class,
				Script::class,
				Theme::class,
			] + $domainClonePrepareEvent->getDomainSpecificModels();

			// Mapping der Originale und Klone
			$idMappings = array_fill_keys($domainSpecificModels, []);

			// Sicherstellen, dass MediaFileService auf den Datenbereich der originalen Domain zugreift
			Domain::setDomainForRequest($this->source->getId());

			// Mediendateien auslesen
			$themeMedia = array_reduce(Theme::all(), function (array $media, Theme $theme) {
				$media[$theme->getId()] = [
					'logo' => $theme->getLogo(),
					'settingsIcon' => $theme->getSettingsIcon(),
				];
				return $media;
			}, []);

			// Relevante Repositories der originalen Domain einlesen
			$modelsDump = array_combine($domainSpecificModels, array_map(function (string $className) use ($user) {
				/** @var class-string<Model> $className Kein Objekt, sondern der vollständige Klassenpfad */

				// Wechsle in das Repository des Originals
				$className::switchRepositoryContext($user, $this->source);

				// Daten kumulieren
				return array_map(function (Model $object) {
					return $object->toArray();
				}, call_user_func([$className, 'all']));
			}, $domainSpecificModels));

			// Repositories ausschließen, die besonders behandelt werden müssen
			$automaticMigrationModels = array_diff($domainSpecificModels, [
				EmbeddingAsset::class,
			]);
			$idMappings = array_intersect_key($idMappings, array_flip($automaticMigrationModels));

			// Sicherstellen, dass MediaFileService im Datenbereich der neuen Domain arbeitet
			Domain::setDomainForRequest($this->getId());

			// EmbeddingAssets importieren und mappen
			EmbeddingAsset::switchRepositoryContext($user, $this);
			/** @var EmbeddingAsset[] $embeddingAssets */
			$embeddingAssets = array_combine(
				array_keys($modelsDump[EmbeddingAsset::class]),
				array_map(function ($data) {
					$asset = EmbeddingAsset::create();
					foreach ($data as $property => $value) {
						$asset->set($property, $value);
					}
					return $asset;
				}, $modelsDump[EmbeddingAsset::class])
			);
			EmbeddingAsset::flush();

			/** @var Model|string $className Kein Objekt, sondern der vollständige Klassenpfad */
			foreach ($automaticMigrationModels as $className) {
				// Wechsle in das Repository des Klons
				$className::switchRepositoryContext($user, $this);

				// Schreiblock anfordern, damit Skeletondaten beim ersten "set()" nicht erneut importiert werden
				$className::beginWrite();

				// Skeletondaten entfernen
				$className::clear(false);

				// Repository datensatzweise füllen
				foreach ($modelsDump[$className] as $id => $data) {
					switch ($className) {
						case Locale::class:
							$factoryParams = [$data['name']];
							break;
						default:
							$factoryParams = [];
							break;
					}

					if ($className::isModelWithCustomIds()) {
						$factoryParams = [$id];
					}

					if ($className == Embedding::class) {
						// Übertragene EmbeddingAssets im Embedding festhalten
						$data['assets'] = array_map(function ($originalAssetId) use ($embeddingAssets) {
							return $embeddingAssets[$originalAssetId]->getId();
						}, $data['assets'] ?? []);
					}

					// Neuen Datensatz erzeugen
					$object = call_user_func_array([$className, 'create'], $factoryParams);
					$idMappings[$className][$id] = $object->getId();

					// Eigenschaften übernehmen
					foreach ($data as $property => $value) {
						$object->set($property, $value);
					}

					// Mediendaten erst hier importieren, damit die Eigenschaften nicht überschrieben werden
					if ($object instanceof Theme) {
						if ($themeMedia[$id]['logo']) {
							$object->setLogo($themeMedia[$id]['logo']);
						}
						if ($themeMedia[$id]['settingsIcon']) {
							$object->setSettingsIcon($themeMedia[$id]['settingsIcon']);
						}
					}
				}

				// Kopierte Datensätze schreiben
				$className::flush();
			}

			// Referenzierte Objekte kopieren
			Utils::dispatchEvent(DomainCloneReferencesEvent::NAME, new DomainCloneReferencesEvent($this->source, $this, $idMappings));
		}
		else { // Utils::REPOSITORY_STORE_FILE
			$srcDir = $this->source->configDir();
			$destDir = $this->configDir();

			// Kopiere Dateien der ursprünglichen Domain
			Utils::copyDirectory($srcDir, $destDir);

			// Entferne Statistiken, Protokolle etc.
			$globDomainPath = rtrim(addcslashes($destDir, '{}[]*?\\'), '/').'/';
			/** @var string[] $incoherentFiles */
			$incoherentFiles = array_reduce([
				'cm-{consent-protocol,found,scripts,statistic}*',
			], function (array $pathnames, string $globFiles) use ($globDomainPath) {
				return array_merge($pathnames, glob($globDomainPath.$globFiles, GLOB_BRACE));
			}, []);

			foreach ($incoherentFiles as $pathname) {
				@unlink($pathname);
			}
		}
	}

	/**
	 * @return static[]
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 * Overrides die normale all() Methode - hier dann nur die die der Subuser darf - WENN es ein Subuser ist...
	 */
	public static function all(): array
	{
		$user = self::$switchedToUser ?? User::loggedInUser();
		$domains = parent::all();

		// Unterbenutzer haben nur Zugriff auf festgelegte Domains
		if ($user && $user->isSubUser()) {
			$accessibleDomains = array_keys($user->getDomActive());
			/* @phan-suppress-next-line PhanTypeMismatchArgument */
			$domains = array_filter($domains, function (self $domain) use ($accessibleDomains) {
				return in_array($domain->getId(), $accessibleDomains);
			});
		}

		return $domains;
	}

	/**
	 * @return static[]
	 * @param array $filter MongoDB-Query-Filter
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 * @override
	 */
	protected static function filterBy(array $filter): array
	{
		$user = self::$switchedToUser ?? User::loggedInUser();
		if ($user && $user->isSubUser()) {
			/** @var static[] $entities */
			$entities = [];
			$accessibleDomains = array_keys($user->getDomActive());

			foreach (self::getRepository()->toArrayFiltered($filter) as $id => $data) {
				if (in_array($id, $accessibleDomains)) {
					$entities[$id] = new static($id);
				}
			}

			return $entities;
		} else {
			return parent::filterBy($filter);
		}
	}

	/**
	 * Zählt alle Domains, auf die der aktuelle Benutzer Zugriff hat.
	 * @return int
	 */
	public static function count(): int
	{
		$user = self::$switchedToUser ?? User::loggedInUser();

		// Unterbenutzer haben nur Zugriff auf festgelegte Domains
		if ($user && $user->isSubUser()) {
			return count(static::all());
		} else {
			return parent::count();
		}
	}

	/**
	 * Zählt die Domains unter Einbezug der Domainnamen für das
	 * Consent-Sharing.
	 *
	 * Eine Domain mit drei Namen zählt hier z.B. als drei Domains.
	 * @phan-side-effect-free
	 */
	public static function countIncludingShareNames(): int
	{
		return array_reduce(self::all(), function ($carry, $domain) {
			return $carry + $domain->getDomainNamesCount();
		}, 0);
	}

	/**
	 * @param string $id
	 * @return static|null
	 * @override
	 */
	public static function find(string $id): ?Model
	{
		$user = self::$switchedToUser ?? User::loggedInUser();
		$id = self::canonicalizeId((string)$user->getId(), $id);

		// Unterbenutzer haben nur Zugriff auf festgelegte Domains
		if ($user->isSubUser()) {
			$accessibleDomains = $user->getDomActive();
			if (!isset($accessibleDomains[$id])) {
				return null;
			}
		}

		return parent::find($id);
	}

	/**
	 * ID in kanonische Form bringen
	 *
	 * Derzeit genutzt, um 7-stellige IDs nach einem Import in MongoDB weiterverwenden zu können.
	 */
	public static function canonicalizeId(string $userId, string $domainId): string
	{
		if (strlen($domainId) == 7 and (strlen($userId) & 1) == 0 and Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB
		) {
			$prefix = substr($userId, 0, -8);
			$userIdPart = substr($userId, -8);
			if (ctype_xdigit($domainId) and ctype_xdigit($userIdPart)) {
				$binUser = hexdec($userIdPart);
				$binDomain = hexdec($domainId.'0');
				return $prefix.str_pad(dechex($binUser ^ $binDomain), 8, '0', STR_PAD_LEFT);
			}
		}
		return $domainId;
	}

	/**
	 * MongoDB-ObjectId in 7-stellige Form bringen.
	 *
	 * @param string $userId MongoDB-ObjectId
	 * @param string $domainId MongoDB-ObjectId
	 * @return string Eine 7-stellige ID oder $domainId, wenn ein ungültiges Format übergeben wird.
	 */
	public static function convertObjectIdToJsonId(string $userId, string $domainId): string
	{
		if (strlen($userId) == 24 && strlen($domainId) == 24) {
			$userSuffix = hexdec(substr($userId, -8));
			$domainSuffix = hexdec(substr($domainId, -8));
			return str_pad(substr(dechex($userSuffix ^ $domainSuffix), 0, -1), 7, '0', STR_PAD_LEFT);
		}
		return $domainId;
	}

	/**
	 * @param string $name
	 * @return self|null
	 * @phan-side-effect-free
	 */
	public static function byName(string $name): ?self
	{
		$name = mb_strtolower($name);
		return array_reduce(self::all(), function ($carry, $domain) use ($name) {
			return $carry ?? (mb_strtolower($domain->getName()) == $name ? $domain : $carry);
		}, null);
	}

	/**
	 * Findet die Domain, die am besten zum Domainnamen passt
	 * @param string $name
	 * @return self|null
	 * @phan-side-effect-free
	 */
	public static function byMatchedDomain(string $name): ?self
	{
		$name = mb_strtolower($name);
		$result = array_reduce(self::all(), function ($carry, $domain) use ($name) {
			list($carryDomain, $carryPrio) = $carry;
			$prio = $domain->matchesDomainName($name);
			if ($prio > $carryPrio) {
				return [$domain, $prio];
			} else {
				return $carry;
			}
		}, [null, 0]);
		return $result[0];
	}

	/**
	 * Wechselt zum Domain-Repository eines anderen Users
	 *
	 * @param User $user
	 * @param bool $flush Ob Änderungen im vorherigen Repository gespeichert werden sollen
	 * @return void
	 */
	public static function switchToUser(User $user, bool $flush=false)
	{
		if ($flush && static::$repository) {
			static::flush();
		}
		static::$repository = static::openRepository($user);
		self::$switchedToUser = $user;
	}

	/**
	 * @param string|null $id
	 * @return static
	 * @throws LogicException Ist für Model Domain nicht verfügbar.
	 * @see self::createForUser
	 * @override @internal
	 */
	public static function create(?string $id=null): Model
	{
		throw new LogicException('Use Domain::createForUser to add a new domain');
	}

	/**
	 * @param User $user Use User::loggedInUser()
	 * @param bool $useSkeleton
	 * @param string|null $domainId [optional] If specified, the ID will be used for the new object.
	 * @param bool $noImport
	 * @return static
	 */
	public static function createForUser(User $user, bool $useSkeleton=true, ?string $domainId=null, bool $noImport=true): self
	{
		// Sicherstellen, dass es sich um den Hauptbenutzer handelt
		$user = $user->getAccountOwner();

		// TODO: Da ein beliebiger Benutzer übergeben werden kann, sollte sichergestellt werden,
		// TODO: dass das korrekte Domain-Repo geladen ist. Alternativ Parameter $user auflösen
		// TODO: und User::loggedInUser() intern aufrufen.

		/** @var self $domain */
		$domain = parent::create($domainId);
		assert($domain instanceof Domain);
		$domain
			->setUser($user)
			->setEvalInlineScripts(false)
			->setIframeBlockMode(self::IFRAME_BLOCK_MODE_BLACKLIST)
			->setWidgetCachingEnabled(true)
			->save() // Repo zwischenspeichern und Lock aufgeben, um Deadlock zu vermeiden
		;

		if ($useSkeleton) {
			$context = new ConfigImporterContext(null, $domain);

			Cookie::importSkeleton($context);
			LegalText::importSkeleton($context);
			Theme::importSkeleton($context);
			Placeholder::importSkeleton($context);
		}

		// System- und User-Themes für Benutzer der Hostingversion in neue Domain integrieren
		if (Utils::isExtendedEdition()) {
			// In Theme-Kontext der neuen Domain wechseln
			Theme::switchRepository($user, $domain);

			// System-Themes importieren
			foreach (SystemTheme::all() as $systemTheme) {
				$theme = Theme::create()
					->setSystemTheme($systemTheme)
					->setSettingsIconBackgroundColor($systemTheme->getPublishedSettingsIconBackgroundColor() ?? $systemTheme->getSettingsIconBackgroundColor())
					->setSettingsIconForegroundColor($systemTheme->getPublishedSettingsIconForegroundColor() ?? $systemTheme->getSettingsIconForegroundColor())
				;

				if ($systemTheme->isDefaultTheme()) {
					$theme->makeDefault();
				}
			}

			//Beim Import muss hier nicht drüber gelaufen werden.
			if($noImport==true)
			{
				// User-Themes importieren
				foreach (UserTheme::all() as $userTheme) {
					$theme = Theme::create()
						->setUserTheme($userTheme)
						->setSettingsIconBackgroundColor($userTheme->getPublishedSettingsIconBackgroundColor() ?? $userTheme->getSettingsIconBackgroundColor())
						->setSettingsIconForegroundColor($userTheme->getPublishedSettingsIconForegroundColor() ?? $userTheme->getSettingsIconForegroundColor())
					;

					if ($userTheme->isDefaultTheme()) {
						$theme->makeDefault();
					}
				}
			}


			// Themes schreiben
			Theme::flush();
		}

		// Repo ggf. neuladen und exklusives Lock anfordern
		self::beginWrite();

		return $domain;
	}

	/**
	 * Diese Methode wählt die aktuelle Domain zur Bearbeitung im Backend aus; ähnlich wie User::login, nur für Domains.
	 * @param string $id
	 * @return bool
	 */
	public static function select(string $id): bool
	{
		$domain = self::find($id);
		$session = Utils::getSession();
		$session->set('domain', $domain ? $domain->getId() : '');

		return (bool)$domain;
	}

	/**
	 * @return array
	 */
	public function checkEmbedding(): array
	{
		$domName = $this->getName();
		$return = array();
		$return['position']=false;
		$return['loaded']=false;
		$position = 999;

		//
		//curl ini
		$curl = new Curl();
		$params = array();
		//Daten holen
		$curl->get($domName, $params);
		if ($curl->curl_error) {
			throw new RuntimeException("An error has occurred in cURL module. (Code: {$curl->getErrorCode()}) " . $curl->getErrorMessage());
		}
		$response = ($curl->getResponse());
		$curl->close();
		//print_r($response);exit();
		$crawler = new Crawler($response);

		$crawler = $crawler->filter('script')->extract(['src']);
		$k = 0;

		foreach ($crawler as $domElement) {
			$k++;
			//print_r($domElement."\n");
			if (strpos($domElement, '/ccm19.js?') || strpos($domElement, '/app.js?apiKey=')) {
				$position = $k;
				$ccm19Link = $domElement;
			}
		}
		$curl = new Curl();

		if(!empty($ccm19Link))
		{
			//Checken ob die ccm19 Resource auch korrekt geladen wird...
			$curl->get(trim($ccm19Link), $params);
			if ($curl->curl_error) {
				throw new RuntimeException("An error has occurred in cURL module. (Code: {$curl->getErrorCode()}) " . $curl->getErrorMessage());
			}
			$response = ($curl->getResponse());
			if (stristr($response,"ccm.insertCookieDeclaration.bind"))
			{
				$return['loaded'] = true;
			}
		}

		if( $position == 1)
		{
			$return['position']=true;
		}
		//
		return $return;
	}

	/**
	 * Diese Methode liefert die aktuell verwendete Domain oder null, falls keine aktiv ist.
	 * Für externe Aufrufe wird, bei gültigem API-Schlüssel, die übergebene Domain-ID verwendet.
	 * @return self|null
	 */
	public static function activeDomain(): ?self
	{
		if (self::$requestDomain) {
			return self::$requestDomain;
		}
		$session = Utils::getSession();

		$user = User::loggedInUser();
		if (!$user) {
			return null;
		}

		if (Utils::isBaseEdition() && empty($session->get('domain'))) {
			if (self::count() == 0) {
				$domain = self::createForUser($user)
					->setManagementStructurePreLegalConsent(true)
					->setManagementStructure(Domain::MANAGEMENT_STRUCTURE_EMBEDDING)
				;
				self::flush();
			}
			else {
				$domain = array_values(self::all())[0];
			}
			self::select($domain->getId());
		}
		return self::find((string)$session->get('domain', ''));
	}

	/**
	 * Schlägt für existierende Domains fehl, wenn der aktive Benutzer keine Zugriffsrechte hat.
	 * @param string $id
	 * @return bool
	 */
	public static function setDomainForRequest(string $id): bool
	{
		if ($id === '') {
			return (bool)(self::$requestDomain = null);
		} else {
			return (bool)(self::$requestDomain = self::find($id));
		}
	}

	/**
	 * @return string
	 */
	public static function getDomainIdForRequest(): string
	{
		return (string)(self::$requestDomain ? self::$requestDomain->getId() : '');
	}

	/**
	 * Diese Methode nennt das domainspezifische Konfigurationsverzeichnis.
	 * @param bool $rel Wenn true, wird der Pfad relativ zum Verzeichnis var/ zurückgegeben, ansonsten als abosluter
	 *                  Pfad.
	 * @return string
	 * @throws LogicException
	 */
	public function configDir(bool $rel = false): string
	{
		$user = $this->getUser();

		if ($user == null) {
			throw new LogicException('User referenced by domain does not exist.');
		}

		return $user->configDir($rel).'/domains/'.$this->getId();
	}

	/**
	 * @param string $id
	 * @param bool $autocommit
	 */
	public static function delete(string $id, bool $autocommit = true): void
	{
		/** @var Domain|null $domain */
		$domain = Domain::find($id);

		if ($domain == null) {
			return;
		}

		$user = $domain->getUser();
		assert($user instanceof User);

		// Domainspezifische Repositories (Embeddings, LegalTexts, Themes etc.) bereinigen
		if (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB) {
			// Models sammeln, deren Objekte gelöscht werden sollen
			$domainDeletePrepareEvent = new DomainDeletePrepareEvent();
			Utils::dispatchEvent(DomainDeletePrepareEvent::NAME, $domainDeletePrepareEvent);

			$domainSpecificModels = [
				Cookie::class,
				Embedding::class,
				EmbeddingAsset::class,
				LegalText::class,
				Locale::class,
				Placeholder::class,
				Script::class,
				Theme::class,
			] + $domainDeletePrepareEvent->getDomainSpecificModels();

			// Sicherstellen, dass MediaFileService im korrekten Datenbereich arbeitet
			Domain::setDomainForRequest($domain->getId());

			/** @var class-string<Model> $className Kein Objekt, sondern der vollständige Klassenpfad */
			foreach ($domainSpecificModels as $className) {
				// Wechsle in das korrekte Repository
				$className::switchRepositoryContext($user, $domain);

				// Mediendateien entfernen
				if ($className == Theme::class) {
					foreach (Theme::all() as $theme) {
						$theme
							->setLogo(null)
							->setSettingsIcon(null)
						;
					}
				}

				// Daten entfernen
				$className::clear();
			}
		}

		Utils::removeDirectory($domain->configDir());
		parent::delete($id, true);

		// Für den Fall, dass die derzeit aktive Domain gelöscht wurde, Auswahl nach Löschvorgang aufheben
		self::select('');
	}

	/**
	 * Diese Methode nennt das domainspezifische Konfigurationsverzeichnis.
	 * @param bool $rel Wenn true, wird der Pfad relativ zum Verzeichnis var/ zurückgegeben, ansonsten als abosluter
	 *                  Pfad.
	 * @return string
	 * @throws LogicException
	 */
	public function logoDir(bool $rel = false): string
	{
		return $this->configDir($rel).'/logos';
	}

	/**
	 * Diese Methode nennt das domainspezifische Uploadverzeichnis, in dem Konfigurations-Icons gepseichert werden.
	 * @param bool $rel Wenn true, wird der Pfad relativ zum Verzeichnis var/ zurückgegeben, ansonsten als abosluter
	 *                  Pfad.
	 * @return string
	 * @throws LogicException
	 */
	public function settingsIconDir(bool $rel = false): string
	{
		return $this->configDir($rel).'/settings_icons';
	}

	/**
	 * Diese Methode holt einen Screenshot der aktuellen Domain und hinterlegt ihn im var Verzeichnis der jeweiligen
	 * Domain... Falls die Datei schon existiert, wird sie einfach ausgegeben
	 * @param string $domName
	 * @param bool $refresh
	 * @return File|null
	 */
	public function getScreenShot(string $domName, bool $refresh): ?File
	{
		$domNamePath = preg_replace('/[^a-z0-9]/', '', $domName);
		$imgPathname = Utils::getCacheDir()."/".$domNamePath.".jpg";
		$imageName = $domNamePath.".jpg";

		$service = MediaFileService::getInstance();
		$repo = $service->getDomainRepository('screenshots');
		$identifier = $this->get('ScreenshotIdentifier');

		if ($identifier) {
			// Falls kein Refresh nötig und Datei existiert, diese holen
			if (!$refresh) {
				$file = $repo->getFile($identifier);
				if ($file !== null) {
					return $file;
				}
			}
		}


		//gibts nicht oder alt - dann neu holen von Screenshot Server...
		$curl = new curl();
		$curl->get('https://cookiecheck.ccm19.de/screenshot.php?' . http_build_query([
			'url' => $domName,
			'apikey' => 'gq5garger',
		], '', '&'));

		// Abbrechen, falls eine unerwartete Antwort kommt
		if (!$curl->isSuccess() || strlen((string)$curl->response) == 0) {
			return null;
		}

		// Screenshot in temporäre Datei schreiben
		$writeOk = file_put_contents($imgPathname, (string)$curl->response, LOCK_EX) !== false;

		if (!$writeOk) {
			return null;
		}

		//Daten ins Repo eintragen und Cache File löschen
		$file = new File($imgPathname);
		$identifier  = $repo->replaceFile($domNamePath, $file, (string)$domNamePath);
		unlink($imgPathname);

		//Identifier speichern damit man den Screenshot nicht immer vom Scanner holen muss
		$this->set("ScreenshotIdentifier",$identifier);
		$this->save();

		return $repo->getFile($identifier);
	}


	/**
	 * @return string[]
	 */
	public function getActivatedPluginIds(): array
	{
		return (array)($this->get('ActivatedPlugins') ?: []);
	}

	/**
	 * @param string $id
	 * @return $this
	 */
	public function activatePlugin(string $id)
	{
		$allPlugins = $this->getActivatedPluginIds();
		$allPlugins[$id] = $id;
		$this->set('ActivatedPlugins', $allPlugins);
		return $this;
	}

	/**
	 * @param string $id
	 * @return $this
	 */
	public function deactivatePlugin(string $id)
	{
		$allPlugins = $this->getActivatedPluginIds();
		unset($allPlugins[$id]);
		$this->set('ActivatedPlugins', $allPlugins);
		return $this;
	}

	/**
	 * @return string
	 */
	private function getUserId(): string
	{
		return (string)$this->get('userId');
	}

	/**
	 * @param string $id
	 * @return $this
	 */
	private function setUserId(string $id): self
	{
		$this->set('userId', $id);
		return $this;
	}

	/**
	 * @param array $cssUrl
	 * @return $this
	 */
	public function addExtraCSS(array $cssUrl): self
	{
		$cssData = $this->get('ExtraCSS');
		if(!is_array($cssData))
		{
			$cssData = array();
		}
		if(!is_array($cssUrl))
		{
			$cssUrl = array();
		}
		$cssData = array_merge($cssData,$cssUrl);
		$this->set('ExtraCSS', $cssData);
		return $this;
	}

	/**
	 * @return array
	 */
	public function getExtraCSS(): array
	{
		$urls = array();
		$cssData = $this->get('ExtraCSS');

		if(is_array($cssData)) {
			foreach ($cssData as $k => $v) {
				try {
					$urls[] = Utils::getFullUrl($v);
				} catch (RouteNotFoundException $e) {
					unset($cssData[$k]);
				}
			}
		}

		return (array)$urls;
	}

	/**
	 * @param array $jsUrl
	 * @return $this
	 */
	public function addExtraJS(array $jsUrl): self
	{
		$jsData = $this->getExtraJS();
		if(!is_array($jsData))
		{
			$jsData = array();
		}
		if(!is_array($jsUrl))
		{
			$jsUrl = array();
		}
		$this->set('ExtraJS', $jsUrl);
		return $this;
	}

	/**
	 * @return array
	 */
	public function getExtraJS(): array
	{
		$urls = array();
		$jsData = $this->get('ExtraJS');

		if(is_array($jsData))
		{
			foreach($jsData as $k=>$v)
			{
				try {
					$urls[] = Utils::getFullUrl($v);
				}
				catch (RouteNotFoundException $e) {
					unset($jsData[$k]);
				}
			}
		}


		return (array)$urls ?? array();
	}

	/**
	 * Diese Methode gibt den Benutzer zurück, dem diese Domain zugewiesen ist.
	 * @return User|null
	 */
	public function getUser(): ?User
	{
		if ($this->user == null) {
			$this->user = User::byId($this->getUserId());

			if ($this->user != null) {
				$this->user = $this->user->getAccountOwner();
			}
		}

		return $this->user;
	}

	/**
	 * @param User $user
	 * @return $this
	 */
	private function setUser(User $user): self
	{
		$this->setUserId($user->getId());
		return $this;
	}

    /**
     * @return bool
     */
    public function isOnBoarding(): bool
    {
        return (bool)$this->get('OnBoarding') ?? FALSE;
    }

    /**
     * @param bool $OnBoarding
     * @return $this
     */
    public function setOnBoarding(bool $OnBoarding): self
    {
        $this->set('OnBoarding', $OnBoarding);
        return $this;
    }

	/**
	 * @return bool
	 */
	public function hasWhitelabel(): bool
	{
		return (bool)$this->get('whitelabel');
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setWhitelabel(bool $value): self
	{
		$this->set('whitelabel', $value);
		return $this;
	}

	/**
	 * Zählt die vorhandenen Instanzen dieses Models.
	 * @return int
	 */
	public static function whitelabelCountForAllDomains(): int
	{
		return array_reduce(self::all(), function ($count, $domain) {
			/** @var self $domain */
			return $domain->hasWhitelabel() ? $count + 1 : $count;
		}, 0);
	}

	/**
	 * Zählt die Domains, in denen die Ausgabe des Frontend-Widgets aktiviert ist.
	 * @return int
	 */
	public static function frontendWidgetCountForAllDomains(): int
	{
		return array_reduce(self::all(), function ($count, $domain) {
			/** @var Domain $domain */
			return $domain->isFrontendWidgetEnabled() ? $count + 1 : $count;
		}, 0);
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setName(string $name): self
	{
		$this->set('name', $name);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getUrlHost(): string
	{
		return parse_url($this->getName(), PHP_URL_HOST) ?: $this->getName();
	}

	/**
	 * @return string
	 */
	public function getUrlPath(): string
	{
		return $this->get('urlPath', '/');
	}

	/**
	 * @param string $path
	 * @return $this
	 */
	public function setUrlPath(string $path): self
	{
		$this->set('urlPath', '/'.ltrim($path, '/'));
		return $this;
	}

	/**
	 * @return string
	 */
	public function getUrlScheme(): string
	{
		return $this->get('urlScheme', 'https').'://';
	}

	/**
	 * @param string $scheme https|http
	 * @return $this
	 */
	public function setUrlScheme(string $scheme): self
	{
		$allowedValues = ['http', 'https'];

		if (in_array($scheme, $allowedValues) == false) {
			$scheme = 'https';
		}

		$this->set('urlScheme', $scheme);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getUrl(): string
	{
		return $this->getUrlScheme().$this->getUrlHost().$this->getUrlPath();
	}

	/**
	 * @return bool
	 */
	public function isIframesBlocked(): bool
	{
		return (bool)$this->get('blockIframes');
	}

	/**
	 * @param bool $block
	 * @return $this
	 */
	public function setIframesBlocked(bool $block): self
	{
		$this->set('blockIframes', $block);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isRememberIframeConsentPerDomain(): bool
	{
		return (bool)$this->get('rememberIframeConsentPerDomain');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setRememberIframeConsentPerDomain(bool $state): self
	{
		$this->set('rememberIframeConsentPerDomain', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isAppendConsentSwitchToIframes(): bool
	{
		return (bool)$this->get('appendConsentSwitchToIframes');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setAppendConsentSwitchToIframes(bool $state): self
	{
		$this->set('appendConsentSwitchToIframes', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isEvalInlineScripts(): bool
	{
		return (bool)$this->get('evalInlineScripts', true);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setEvalInlineScripts(bool $state): self
	{
		$this->set('evalInlineScripts', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isScriptsTrack(): bool
	{
		return (bool)$this->get('scriptsTrack');
	}

	/**
	 * @param bool $block
	 * @return $this
	 */
	public function setScriptsTrack(bool $block): self
	{
		$this->set('scriptsTrack', $block);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isNewScriptsBlocked(): bool
	{
		return (bool)$this->get('blockScripts');
	}

	/**
	 * @param bool $block
	 * @return $this
	 */
	public function setNewScriptsBlocked(bool $block): self
	{
		$this->set('blockScripts', $block);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isSameDomainScriptsBlocked(): bool
	{
		return (bool)($this->get('blockSameDomainScripts') ?? true);
	}

	/**
	 * @param bool $block
	 * @return $this
	 */
	public function setSameDomainScriptsBlocked(bool $block): self
	{
		$this->set('blockSameDomainScripts', $block);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isInlineScriptsBlocked(): bool
	{
		return (bool)($this->get('blockInlineScripts') ?? true);
	}

	/**
	 * @param bool $block
	 * @return $this
	 */
	public function setInlineScriptsBlocked(bool $block): self
	{
		$this->set('blockInlineScripts', $block);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isActivatePermaScan(): bool
	{
		return (bool)($this->get('activatePermaScan') ?? true);
	}

	/**
	 * @param bool $block
	 * @return $this
	 */
	public function setActivatePermaScan(bool $block): self
	{
		$this->set('activatePermaScan', $block);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getAllowedScriptMarkers(): array
	{
		return (array)$this->get('allowedScriptMarkers');
	}

	/**
	 * @param string[] $list
	 * @return $this
	 */
	public function setAllowedScriptMarkers(array $list): self
	{
		$this->set('allowedScriptMarkers', $list);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getDeniedScriptMarkers(): array
	{
		return (array)$this->get('deniedScriptMarkers');
	}

	/**
	 * @param string[] $list
	 * @return $this
	 */
	public function setDeniedScriptMarkers(array $list): self
	{
		$this->set('deniedScriptMarkers', $list);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getIframeMarkers(): array
	{
		return (array)$this->get('iframeMarkers', $this->get('allowedIframeMarkers'));
	}

	/**
	 * @param string[] $list
	 * @return $this
	 */
	public function setIframeMarkers(array $list): self
	{
		$this->set('iframeMarkers', $list);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getIframeBlockMode(): string
	{
		return $this->get('iframeBlockMode', 'whitelist');
	}

	/**
	 * @param string $mode
	 * @return $this
	 */
	public function setIframeBlockMode(string $mode): self
	{
		$this->set('iframeBlockMode', $mode);
		return $this;
	}


	/**
	 * @return bool
	 */
	public function isFrontendLegalEnabled(): bool
	{
		return (bool)$this->get('frontendLegalEnabled', false);
	}

	/**
	 * @param bool $enabled
	 * @return $this
	 */
	public function setFrontendLegalEnabled(bool $enabled): self
	{
		$this->set('frontendLegalEnabled', $enabled);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isFrontendWidgetEnabled(): bool
	{
		return (bool)$this->get('frontendWidgetEnabled', true);
	}

	/**
	 * @param bool $enabled
	 * @return $this
	 */
	public function setFrontendWidgetEnabled(bool $enabled): self
	{
		$this->set('frontendWidgetEnabled', $enabled);
		return $this;
	}

	/**
	 * Diese Methode liefert den letzten Zeitpunkt einer Anpassung an der
	 * Systemkonfiguration durch eine beliebige Komponente, die relevant
	 * für die Cache-Invalidierung des Frontend-Widgets ist.
	 * @return int
	 */
	public function getFrontendWidgetCacheTimestamp(): int
	{
		return (int)$this->get('frontendWidgetCacheTimestamp');
	}

	/**
	 * @param int $time
	 * @return $this
	 */
	private function setFrontendWidgetCacheTimestamp(int $time): self
	{
		$this->set('frontendWidgetCacheTimestamp', $time);
		return $this;
	}

	/**
	 * Diese Methode wird von beliebigen Komponenten aufgerufen, um die
	 * Cache-Invalidierung für das Frontend-Widget auszulösen.
	 * @return $this
	 */
	public function touchFrontendWidgetCacheTimestamp(): self
	{
		return $this->setFrontendWidgetCacheTimestamp(time());
	}

	/**
	 * Löst die Cache-Invalidierung für *alle* Domains *aller* User aus.
	 *
	 * Eventuell gehen dabei ungespeicherte Änderungen im Domain-Model verloren!
	 */
	public static function touchAllFrontendWidgetCacheTimestampsGlobally(): void
	{
		$repo = self::getRepository();
		if ($repo instanceof MongoDb) {
			// Für MongoDB: Optimiertes Bulk-Update-Query
			$repo->performCustomBulkUpdate(
				[ 'frontendWidgetCacheTimestamp' => [ '$exists' => true ] ],
				[ '$set' => ['frontendWidgetCacheTimestamp' => time()] ],
				true
			);
		} else {
			// Für JSON: Alle Clients und Domains in Schleifen durchgehen
			$user = User::loggedInUser();
			$clients = User::getClients();

			try {
				array_walk($clients, function (User $user) {
					Domain::switchToUser($user, false);

					foreach (Domain::all() as $domain) {
						$domain->touchFrontendWidgetCacheTimestamp();
					}

					Domain::flush();
				});
			} finally {
				if ($user) {
					Domain::switchToUser($user, false);
				}
			}
		}
	}

	/**
	 * @return bool
	 */
	public function isWidgetCachingEnabled(): bool
	{
		return (bool)$this->get('widgetCachingEnabled', true);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setWidgetCachingEnabled(bool $state): self
	{
		$this->set('widgetCachingEnabled', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isPartOfAdAndTrackingFreeNetwork(): bool
	{
		return (bool)$this->get('partOfAdAndTrackingFreeNetwork');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setPartOfAdAndTrackingFreeNetwork(bool $state): self
	{
		$this->set('partOfAdAndTrackingFreeNetwork', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isManagementStructurePreLegalConsent(): bool
	{
		if (Utils::isClickskeksCloud()) {
			return true;
		}

		return (bool)$this->get('managementStructurePreLegalConsent');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setManagementStructurePreLegalConsent(bool $state): self
	{
		$this->set('managementStructurePreLegalConsent', $state);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getManagementStructure(): string
	{
		if (Utils::isClickskeksCloud()) {
			return self::MANAGEMENT_STRUCTURE_EMBEDDING;
		}

		return (string)$this->get('managementStructure', self::MANAGEMENT_STRUCTURE_COOKIE);
	}

	/**
	 * @param string $type
	 * @return $this
	 * @throws LogicException if invalid management structure is supplied.
	 */
	public function setManagementStructure(string $type): self
	{
		if (in_array($type, self::ALLOWED_MANAGEMENT_STRUCTURE_VALUES) == false) {
			throw new LogicException('Invalid value for storage type, allowed values: '.implode(', ', self::ALLOWED_MANAGEMENT_STRUCTURE_VALUES));
		}

		$this->set('managementStructure', $type);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function getDeleteUnknownCookies(): bool
	{
		return (bool)$this->get('deleteUnknownCookies');
	}

	/**
	 * @param bool $delete
	 * @return $this
	 */
	public function setDeleteUnknownCookies(bool $delete): self
	{
		$this->set('deleteUnknownCookies', $delete);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function getDeleteUnknownCookiesForceReload(): bool
	{
		// Funktion auf unbestimmte Zeit deaktiviert
		return false;
		return (bool)$this->get('deleteUnknownCookiesForceReload');
	}

	/**
	 * @param bool $force
	 * @return $this
	 */
	public function setDeleteUnknownCookiesForceReload(bool $force): self
	{
		$this->set('deleteUnknownCookiesForceReload', $force);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isLoadTechnicallyNecessaryEmbeddingsWithoutConsent(): bool
	{
		return (bool)$this->get('loadTechnicallyNecessaryEmbeddingsWithoutConsent', true);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setLoadTechnicallyNecessaryEmbeddingsWithoutConsent(bool $state): self
	{
		$this->set('loadTechnicallyNecessaryEmbeddingsWithoutConsent', $state);
		return $this;
	}

	/**
	 * Letzter Zeitpunkt eines Statistik-Updates
	 * @return DateTime|null
	 */
	public function getLastStatisticsUpdateTimestamp(): ?DateTime
	{
		$value = $this->get('statisticsUpdateTimestamp');
		if ($value) {
			return new DateTime((string)$value);
		} else {
			return null;
		}
	}

	/**
	 * @param DateTimeInterface $ts
	 * @return $this
	 */
	private function setLastStatisticsUpdateTimestamp(DateTimeInterface $ts): self
	{
		$this->set('statisticsUpdateTimestamp', $ts->format('Y-m-d H:i:s'), false);
		return $this;
	}

	/**
	 * @return $this
	 */
	public function touchLastStatisticsUpdateTimestamp(): self
	{
		return $this->setLastStatisticsUpdateTimestamp(new DateTime());
	}

	/**
	 * @return DateTime
	 * @throws \Exception
	 */
	public function getLastScanTimestamp(): DateTime
	{
		$value = $this->get('lastScanDate');
		if ($value) {
			$value = new DateTime((string)$value);
		} else if(Utils::isClickskeksCloud())
		{
			$value = new DateTime(date('Y-m-d', strtotime( '-' . random_int(5, 17).' days')));
			$this->setLastScanTimestamp($value)->save();
		}
		else {
			$value = new DateTime('2000-01-01');
		}

		return $value;
	}

	/**
	 * @param DateTimeInterface $ts
	 * @return $this
	 */
	public function setLastScanTimestamp(DateTimeInterface $ts): self
	{
		$this->set('lastScanDate', $ts->format('Y-m-d'), false);
		return $this;
	}


	public function getRealCrawlCount(): int
	{
		return (int)$this->get('realCrawlCount', 0);
	}

	/**
	 * @param int $realCrawlCount
	 *
	 * @return $this
	 */
	public function setRealCrawlCount(int $realCrawlCount): self
	{
		$this->set('realCrawlCount', $realCrawlCount);
		return $this;
	}

	/**
	 * @deprecated
	 * @return bool
	 */
	public function hasRecordedViewCounts(): bool
	{
		return $this->get('viewCount') !== null;
	}

	/**
	 * @deprecated
	 * @return int
	 */
	public function getViewCount(): int
	{
		return (int)$this->get('viewCount');
	}

	/**
	 * @deprecated
	 * @param int $value
	 * @return $this
	 */
	public function setViewCount(int $value): self
	{
		$this->set('viewCount', $value);
		return $this;
	}

	/**
	 * @deprecated
	 * @param int $diff
	 * @return $this
	 */
	public function incrementViewCount(int $diff = 1): self
	{
		$this->increment('viewCount', $diff);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentStorage(): string
	{
		return (string)$this->get('consentStorage', 'localStorage');
	}

	/**
	 * @param string $storageType
	 * @return $this
	 */
	public function setConsentStorage(string $storageType): self
	{
		$this->set('consentStorage', $storageType);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentStorageNamespaceUrlParameter(): string
	{
		return (string)$this->get('consentStorageNamespaceUrlParameter');
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setConsentStorageNamespaceUrlParameter(string $name): self
	{
		$this->set('consentStorageNamespaceUrlParameter', $name);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getConsentCookieLifetime(): int
	{
		return (int)$this->get('consentCookieLifetime', 365);
	}

	/**
	 * @param int $cookieLifetime
	 * @return $this
	 */
	public function setConsentCookieLifetime(int $cookieLifetime): self
	{
		$this->set('consentCookieLifetime', max(0, $cookieLifetime));
		return $this;
	}

	/**
	 * @return int
	 */
	public function getConsentLifetime(): int
	{
		$result = (int)$this->get('consentLifetime', 0);
		// Bei TCF Maximum von 13 Monaten erzwingen:
		if ($this->isTcfVendorsEnabled()) {
			if ($result == 0 or $result > 390) {
				$result = 390;
			}
		}
		return $result;
	}

	/**
	 * @param int $cookieLifetime
	 * @return $this
	 */
	public function setConsentLifetime(int $cookieLifetime): self
	{
		$this->set('consentLifetime', max(0, $cookieLifetime));
		return $this;
	}

	/**
	 * @return int|null
	 */
	public function getConsentLifetimeForPartialConsent(): ?int
	{
		$result = $this->get('consentLifetimeForPartialConsent');
		// Bei TCF Maximum von 13 Monaten erzwingen:
		if ($this->isTcfVendorsEnabled()) {
			if ($result === 0 or $result > 390) {
				$result = 390;
			}
		}
		return $result === null ? null : (int)$result;
	}

	/**
	 * @param int|null $lifetime
	 * @return $this
	 */
	public function setConsentLifetimeForPartialConsent(?int $lifetime): self
	{
		// Allowed values: null,0,3,4,5,…
		if ($lifetime !== null) {
			$lifetime = $lifetime > 0 ? max(3, $lifetime) : 0;
		}
		$this->set('consentLifetimeForPartialConsent', $lifetime);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getShareDomainNames(): array
	{
		return (array)$this->get('shareDomainNames', []);
	}

	/**
	 * @return int Anzahl dieser Domain zugeordneter Domainnamen, ausgenommen IP-Adressen (Minimum 1).
	 */
	public function getDomainNamesCount(): int
	{
		$mainName = mb_strtolower(rtrim($this->getName(), '.'));
		return array_reduce($this->getShareDomainNames(), function ($carry, $domainName) use ($mainName) {
			return $carry + (int)(
				filter_var($domainName, FILTER_VALIDATE_IP) === false
				and (
					$domainName[0] !== '['
					or
					filter_var(substr($domainName, 1, -1), FILTER_VALIDATE_IP, FILTER_FLAG_IPV6) === false
				)
				and $mainName !== mb_strtolower($domainName)
			);
		}, 0) + 1;
	}

	/**
	 * @param string[] $domainNames
	 * @return $this
	 */
	public function setShareDomainNames(array $domainNames): self
	{
		$this->set('shareDomainNames', $domainNames);
		return $this;
	}

	/**
	 * @param string $domainName
	 * @return int je höher, desto besser passt die Domain. 0 wenn kein Match.
	 */
	public function matchesDomainName(string $domainName): int
	{
		$domainName = mb_strtolower(rtrim($domainName, '.'));
		$mainHost = mb_strtolower(rtrim($this->getName(), '.'));
		if ($mainHost === $domainName) {
			return strlen($domainName);
		}
		foreach ($this->getShareDomainNames() as $domain)
		{
			if ($domainName === mb_strtolower($domain) or preg_match('~\.'.preg_quote($domain, '~').'$~i', $domainName)) {
				return strlen($domain);
			}
		}
		return 0;
	}

	/**
	 * @return bool
	 */
	public function hasCrossDomainSharing(): bool
	{
		return (bool)$this->get('crossDomainSharing', false);
	}

	/**
	 * @param bool $activate
	 * @return $this
	 */
	public function setCrossDomainSharing(bool $activate): self
	{
		$this->set('crossDomainSharing', $activate);
		return $this;
	}


	/**
	 * @return string
	 */
	public function getconsentCookieReset(): string
	{
		return (string)$this->get('consentCookieReset', false);
	}

	/**
	 * @param string $consentCookieReset
	 * @return $this
	 */
	public function setconsentCookieReset(string $consentCookieReset): self
	{
		$this->set('consentCookieReset', $consentCookieReset);
		return $this;
	}

	/**
	 * Ob eine relevante Änderung an einem Embedding/Cookie
	 * erteilten Consent anullieren soll.
	 *
	 * @return bool
	 */
	public function hasConsentResetOnRelevantChange(): bool
	{
		return (bool)$this->get('consentResetOnChange', true);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setConsentResetOnRelevantChange(bool $value): self
	{
		$this->set('consentResetOnChange', $value);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function hasSecureConsentCookie(): bool
	{
		return (bool)$this->get('secureConsentCookie', true);
	}

	/**
	 * @param bool $activate
	 * @return $this
	 */
	public function setSecureConsentCookie(bool $activate): self
	{
		$this->set('secureConsentCookie', $activate);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function hasCodeMinification(): bool
	{
		return (bool)$this->get('minify', true);
	}

	/**
	 * @param bool $activate
	 * @return $this
	 */
	public function setCodeMinification(bool $activate): self
	{
		$this->set('minify', $activate);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function hasForcedCssImportant(): bool
	{
		return (bool)$this->get('cssImportant', false);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setForcedCssImportant(bool $state): self
	{
		$this->set('cssImportant', $state);
		return $this;
	}

	/**
	 * This method returns whether the domain applies the optimization for the LCP.
	 * @return bool
	 */
	public function isLcpOptimizationEnabled(): bool
	{
		return (bool)$this->get('lcpOptimization', false);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setLcpOptimizationEnabled(bool $state): self
	{
		$this->set('lcpOptimization', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function getConsentTablePrependIntroText(): bool
	{
		return (bool)$this->get('consentTablePrependIntroText', true);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setConsentTablePrependIntroText(bool $value): self
	{
		$this->set('consentTablePrependIntroText', $value);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function getConsentTablePrependConsentChangeLink(): bool
	{
		return (bool)$this->get('consentTablePrependConsentChangeLink', true);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setConsentTablePrependConsentChangeLink(bool $value): self
	{
		$this->set('consentTablePrependConsentChangeLink', $value);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getConsentTableHeadingStartLevel(): int
	{
		return (int)$this->get('consentTableHeadingStartLevel', 3);
	}

	/**
	 * @param int $level
	 * @return $this
	 */
	public function setConsentTableHeadingStartLevel(int $level): self
	{
		$this->set('consentTableHeadingStartLevel', $level);
		return $this;
	}

	/**
	 * Hauptland der Domain oder null, wenn das Land des Users verwendet werden soll
	 */
	public function getCountry(): ?Country
	{
		$code = $this->getCountryCode();
		return $code ? Country::find($code) : null;
	}

	public function getCountryCode(): ?string
	{
		return $this->get('country', null);
	}

	/**
	 * @param Country|null $country
	 * @return $this
	 */
	public function setCountry(?Country $country): self
	{
		$this->set('country', $country ? $country->getCode() : null);
		return $this;
	}

	/**
	 * @return TcfVendor[]
	 */
	public function getTcfVendorBlacklist(): array
	{
		$vendors = array_filter(array_map(function ($id) {
			return TcfVendor::find((string)$id);
		}, $this->get('tcfVendorBlacklist', [])));

		// Anbieter anhand der ID indexieren
		$vendors = array_combine(array_map(function (TcfVendor $x) { return $x->getId(); }, $vendors), $vendors);

		// Anbieter nach ID sortieren
		ksort($vendors);

		/** @var TcfVendor|null $mostRecentVendor */
		$mostRecentVendor = $this->getMostRecentTcfVendor() ?? (end($vendors) ?: null);
		$unusedVendorIds = $this->getUnusedTcfVendorIds();

		// Blacklist dynamisch erweitern, wenn eingestellt und neue Anbieter vorhanden
		if ($this->isNewTcfVendorsActive() == false && ($mostRecentVendor || $unusedVendorIds)) {
			$additionalBlacklistItems = array_filter(TcfVendor::all(),
				function (TcfVendor $vendor) use ($mostRecentVendor, $unusedVendorIds) {
					$id = $vendor->getId();
					return $mostRecentVendor && $id > $mostRecentVendor->getId() || in_array($id, $unusedVendorIds);
				}
			);
		}
		else {
			$additionalBlacklistItems = [];
		}

		return array_values(array_merge($vendors, $additionalBlacklistItems));
	}

	/**
	 * @param TcfVendor[] $vendors
	 * @return self
	 */
	public function setTcfVendorBlacklist(array $vendors): self
	{
		$vendorIds = array_map(function (TcfVendor $vendor) {
			return (int)$vendor->getId();
		}, $vendors);
		$vendorIds = array_values(array_unique($vendorIds, SORT_NUMERIC));

		$this->set('tcfVendorBlacklist', $vendorIds);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isTcfVendorsEnabled(): bool
	{
		return (bool)$this->get('tcfVendorsEnabled', false);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setTcfVendorsEnabled(bool $state): self
	{
		$this->set('tcfVendorsEnabled', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isNewTcfVendorsActive(): bool
	{
		// Funktion auf unbestimmte Zeit deaktiviert
		return false;
		return (bool)$this->get('newTcfVendorsActive', true);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setNewTcfVendorsActive(bool $state): self
	{
		$this->set('newTcfVendorsActive', $state);
		return $this;
	}

	/**
	 * @return TcfVendor|null
	 */
	public function getMostRecentTcfVendor(): ?TcfVendor
	{
		return TcfVendor::find((string)$this->get('mostRecentTcfVendor'));
	}

	/**
	 * @param TcfVendor $vendor
	 * @return $this
	 */
	public function setMostRecentTcfVendor(TcfVendor $vendor): self
	{
		$this->set('mostRecentTcfVendor', (int)$vendor->getId());
		return $this;
	}

	/**
	 * @return int[]
	 */
	public function getUnusedTcfVendorIds(): array
	{
		return $this->get('unusedTcfVendorIds', []);
	}

	/**
	 * @param int[] $ids
	 * @return $this
	 */
	public function setUnusedTcfVendorIds(array $ids): self
	{
		$this->set('unusedTcfVendorIds', array_values(array_filter($ids, 'is_int')));
		return $this;
	}

	/**
	 * @return TcfSpecialFeature[]
	 */
	public function getExplicitTcfSpecialFeatures(): array
	{
		return array_values(array_filter(array_map(function ($id) {
			return TcfSpecialFeature::find((string)$id);
		}, $this->get('tcfSpecialFeatures', []))));
	}

	/**
	 * @param TcfSpecialFeature[] $features
	 * @return self
	 */
	public function setExplicitTcfSpecialFeatures(array $features): self
	{
		$featureIds = array_map(function (TcfSpecialFeature $feature) {
			return (int)$feature->getId();
		}, $features);
		$featureIds = array_values(array_unique($featureIds, SORT_NUMERIC));

		$this->set('tcfSpecialFeatures', $featureIds);
		return $this;
	}

	/**
	 * Liefert die TCF-Stacks zusammen mit übrig bleibenden Purposes und Special Features für diese Domain
	 *
	 * Dies sind alle TCF-relevanten Dinge, die sich aus aktivierten Embeddings und global für die Domain
	 * aktivierten TCF-Einstellungen ergeben.
	 *
	 * Nach Änderungen an Embeddings oder TCF-Einstellungen muss `updateTcfStacks()` ausgeführt werden,
	 * um diese Liste zu aktualisieren.
	 * @return (TcfStack|TcfPurpose|TcfSpecialFeature)[]
	 */
	public function getTcfStacks(): array
	{
		return array_filter(array_merge(
			array_map(function ($id) {
				return TcfStack::find($id);
			}, $this->get('tcfStacks.stacks', [])),
			array_map(function ($id) {
				return TcfPurpose::find($id);
			}, $this->get('tcfStacks.purposes', [])),
			array_map(function ($id) {
				return TcfSpecialFeature::find($id);
			}, $this->get('tcfStacks.specialFeatures', []))
		));
	}

	/**
	 * @param (TcfStack|TcfPurpose|TcfSpecialFeature)[] $items
	 * @return $this
	 */
	public function setTcfStacks(array $items): self
	{
		$stackIds = $purposeIds = $specialFeatureIds = [];
		foreach ($items as $item) {
			if ($item instanceof TcfStack) {
				$stackIds[] = $item->getId();
			}
			elseif ($item instanceof TcfPurpose) {
				$purposeIds[] = $item->getId();
			}
			elseif ($item instanceof TcfSpecialFeature) {
				$specialFeatureIds[] = $item->getId();
			}
			else {
				throw new InvalidArgumentException(sprintf('"%s" is no valid tcf stack, purpose or special feature', addslashes((string)$item)));
			}
			$this->set('tcfStacks', [
				'stacks' => $stackIds,
				'purposes' => $purposeIds,
				'specialFeatures' => $specialFeatureIds,
			]);
		}
		return $this;
	}

	/**
	 * Aktualisiert die Liste der TCF-Stacks auf Basis der aktuell aktiven Embeddings und TCF-Einstellungen
	 * @see TcfStack::findMinimalSet()
	 * @return $this
	 */
	public function updateTcfStacks(): self
	{
		$purposes = $specialFeatures = [];

		foreach (Embedding::all() as $embedding) {
			if (!$embedding->isActive() or $embedding->getTcfVendor() === null) {
				continue;
			}
			foreach ($embedding->getTcfPurposes() as $p) {
				$purposes[$p->getId()] = $p;
			}
			foreach ($embedding->getTcfSpecialFeatures() as $sf) {
				$specialFeatures[$sf->getId()] = $sf;
			}
		}

		foreach ($this->getExplicitTcfSpecialFeatures() as $sf) {
			$specialFeatures[$sf->getId()] = $sf;
		}

		$blacklist = $this->get('tcfVendorBlacklist', []);
		foreach (TcfVendor::all() as $vendor) {
			if (in_array($vendor->getId(), $blacklist)) {
				continue;
			}
			foreach ($vendor->getPurposes() as $p) {
				$purposes[$p->getId()] = $p;
			}
		}

		return $this->setTcfStacks(TcfStack::findMinimalSet($purposes, $specialFeatures));
	}

	public function hasGoogleAdditionalConsentMode(): bool
	{
		return (bool)$this->get('googleAdditionalConsentMode', false);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setGoogleAdditionalConsentMode(bool $value): self
	{
		$this->set('googleAdditionalConsentMode', $value);
		return $this;
	}

	public function isTcfAdTagsEnabled(): bool
	{
		return (bool)$this->get('tcfAdTagsEnabled', false);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setTcfAdTagsEnabled(bool $value): self
	{
		$this->set('tcfAdTagsEnabled', $value);
		return $this;
	}

	public function hasTcfAdTagsAutoAppend(): bool
	{
		return (bool)$this->get('tcfAdTagsAutoAppend', false);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setTcfAdTagsAutoAppend(bool $value): self
	{
		$this->set('tcfAdTagsAutoAppend', $value);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isTcfLegitimateInterestEnabled(): bool
	{
		return (bool)$this->get('tcfLegitimateInterest', false);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setTcfLegitimateInterestEnabled(bool $state): self
	{
		$this->set('tcfLegitimateInterest', $state);
		return $this;
	}

	/**
	 * @return AdTechProvider[]
	 */
	public function getGoogleAdditionalConsentWhitelist(): array
	{
		$providers = array_filter(array_map(function ($id) {
			return AdTechProvider::find((string)$id);
		}, $this->get('googleAdditionalConsentWhitelist', [])));

		return array_values($providers);
	}

	/**
	 * @param AdTechProvider[] $providers
	 * @return self
	 */
	public function setGoogleAdditionalConsentWhitelist(array $providers): self
	{
		$providerIds = array_map(function (AdTechProvider $provider) {
			return (int)$provider->getId();
		}, $providers);
		$providerIds = array_values(array_unique($providerIds, SORT_NUMERIC));

		$this->set('googleAdditionalConsentWhitelist', $providerIds);
		return $this;
	}

	/**
	 * Ob für diese Domain einen "Ablehnen"-Button nötig ist, wenn
	 * ein "Alles Akzeptieren"-Button existiert
	 *
	 * Standardmäßig aktiv. Der Nutzer muss das explizit deaktivieren.
	 *
	 * @return bool
	 */
	public function isDeclineButtonEnforced(): bool
	{
		return (bool)$this->get('declineButtonEnforced', true);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setDeclineButtonEnforced(bool $state): self
	{
		$this->set('declineButtonEnforced', $state);
		return $this;
	}

	/**
	 * Ob die Domain bereits eingerichtet wurde
	 *
	 * Das ist True, wenn noch kein Impressum und keine
	 * Datenschutzerklärung hinterlegt ist.
	 *
	 * @return bool
	 */
	public function isUnconfigured(): bool
	{
		$activeDomain = Domain::activeDomain();
		Domain::select($this->getId());
		LegalText::reloadRepository();

		$unconfigured = true;

		foreach (LegalText::all() as $legal) {
			if ($legal->getImprintScheme() === 'link' or strlen($legal->getImprint()) >= 28) {
				$unconfigured = false;
			}
			if ($legal->getPrivacyScheme() === 'link' or strlen($legal->getPrivacyPolicy()) >= 40) {
				$unconfigured = false;
			}
		}

		if ($activeDomain) {
			Domain::select($activeDomain->getId());
			LegalText::reloadRepository();
		}
		return $unconfigured;
	}

	/**
	 * Ob der Banner nur aktiv sein soll, wenn die Domain auch zu einem der eingetragenen Domainnamen passt
	 */
	public function isOnlyActiveOnMatchingDomain(): bool
	{
		return (bool)$this->get('onlyActiveOnMatchingDomain', false);
	}

	/**
	 * @param bool $state
	 * @return self
	 */
	public function setOnlyActiveOnMatchingDomain(bool $state): self
	{
		$this->set('onlyActiveOnMatchingDomain', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isAllowEmbeddingConsentViaJsApi(): bool
	{
		return (bool)$this->get('embeddingConsentViaJsApi');
	}

	/**
	 * @param bool $state
	 * @return self
	 */
	public function setAllowEmbeddingConsentViaJsApi(bool $state): self
	{
		$this->set('embeddingConsentViaJsApi', $state);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getWidgetDelayInMilliseconds(): int
	{
		return (int)$this->get('widgetDelayInMilliseconds', 0);
	}

	/**
	 * @param int $milliseconds
	 * @return $this
	 */
	public function setWidgetDelayInMilliseconds(int $milliseconds): self
	{
		$this->set('widgetDelayInMilliseconds', max(0, $milliseconds));
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getWidgetExclusionUrls(): array
	{
		return (array)$this->get('widgetExclusionUrls', []);
	}

	/**
	 * @param string[] $urls
	 * @return $this
	 */
	public function setWidgetExclusionUrls(array $urls): self
	{
		$this->set('widgetExclusionUrls', $urls);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getDomainExternalCrawlerFrequency(): int
	{
		$fallBack = 0;
		//Wenn die Domain gecrawlt werden kann und darf, dann auf 8 setzen falls es keinen Eintrag gibt
		if($this->getDoCrawl())
		{
			$fallBack = 8;
		}
		return (int)$this->get('domainExternalCrawlerFrequency', $fallBack);
	}

	/**
	 * @param int $domainExternalCrawlerFrequency
	 * @return $this
	 */
	public function setDomainExternalCrawlerFrequency(int $domainExternalCrawlerFrequency): self
	{
		$this->set('domainExternalCrawlerFrequency', $domainExternalCrawlerFrequency);
		return $this;
	}

	/**
	 * Beschreibung: Wieviel Crawl Budget steht zur Verfügung pro Domain, umgerechnet auf die Anzahl der verfügbaren
	 * Domains ZU nutzen bei Neuerstellung einer Domain
	 * @return int
	 *
	 */
	public static function getVerfuegbaresCrawlBudgetProDomain() :int
	{
		$user = User::loggedInUser();
		if($user!== null) {

			//das verfügbare Rest Crawlbudget
			$verfuegbaresCrawlBudget = self::getRestCrawlBudget();

			if($verfuegbaresCrawlBudget=="-1")
			{
				return -1;
			}

			//Verrechnen mit restlichen Domains um einen akzeptablen Mittelwert zu bekommen, falls es einen gibt, sonst eben 0
			return max( 0, round($verfuegbaresCrawlBudget / ($user->getMaxDomainCount() - $user->getActualDomainCount())));
		}
		//oder 1 zurückgeben
		return 1;
	}

	/**
	 * Beschreibung: Wie viel Crawl Budget haben wir noch im Account gesamt
	 * @return int
	 *
	 */
	public static function getRestCrawlBudget(): int
	{
		$user = User::loggedInUser();
		if($user!== null) {
			$verfuegbaresCrawlBudget = 0;
			$domains = Domain::all();
			foreach ($domains as $domain) {
				$verfuegbaresCrawlBudget += $domain->getDomainActualMaxCrawlBudget();
			}
			//Account CrawlBudget
			$accountCrawlBudget =  $user->getAccountFeatureMaxCrawlBudget();

			//Wenn -1 aus dem Account kommt, dann auf -1 setzen
			if( $user->getAccountFeatureMaxCrawlBudget()===-1)
			{
				return -1;
			}

			//Abzgl. des schon verbrauchten aus anderen Domains
			return max(0, $accountCrawlBudget - $verfuegbaresCrawlBudget);
		}
		//oder 1 zurückgeben
		return 1;
	}

	/**
	 * @return int
	 */
	public function getDomainMaxCrawlBudget(): int
	{
		return (int)$this->get('domainMaxCrawlBudget', 0);
	}

	public function getDomainActualMaxCrawlBudget(): int
	{
		return (int)$this->get('domainMaxCrawlBudget', 0);
	}

	/**
	 * @param int $domainMaxCrawlBudget
	 *
	 * @return $this
	 */
	public function setDomainMaxCrawlBudget(int $domainMaxCrawlBudget): self
	{
		$this->set('domainMaxCrawlBudget', $domainMaxCrawlBudget);
		return $this;
	}

	/**
	 * @return DateTimeZone|null
	 */
	public function getTimeZone(): ?DateTimeZone
	{
		try {
			return new DateTimeZone((string)$this->get('timeZone'));
		} catch (Exception $exception) {
			return null;
		}
	}

	/**
	 * @param DateTimeZone|null $timeZone
	 * @return $this
	 */
	public function setTimeZone(?DateTimeZone $timeZone): self
	{
		$this->set('timeZone', $timeZone ? $timeZone->getName() : null);
		return $this;
	}

	/**
	 * Ob erweiterte GoogleTagManager-Events ausgelöst werden sollen
	 * @return bool
	 */
	public function isTriggeringExtGtmEvents(): bool
	{
		return (bool)$this->get('triggeringExtGtmEvents', true);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setTriggeringExtGtmEvents(bool $value): self
	{
		$this->set('triggeringExtGtmEvents', $value);
		return $this;
	}

	/**
	 * Ob Skripte erkannt werden sollen, die CMP-spezifischen HTML-Code zum Nachladen aufweisen.
	 * @return bool
	 */
	public function isDetectCloakedScripts(): bool
	{
		return (bool)$this->get('detectCloakedScripts');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setDetectCloakedScripts(bool $state): self
	{
		$this->set('detectCloakedScripts', $state);
		return $this;
	}

	/**
	 * Ob das Frontend-Widget bei automatischer Spracherkennung vorzugsweise in der durch den Browser
	 * vorgegebenen Sprache anstelle der durch das lang-Attribut definierten Sprache geladen werden soll.
	 * @return bool
	 */
	public function isPrioritizeLanguageHeaderOverLangAttribute(): bool
	{
		return (bool)$this->get('prioritizeLanguageHeaderOverLangAttribute');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setPrioritizeLanguageHeaderOverLangAttribute(bool $state): self
	{
		$this->set('prioritizeLanguageHeaderOverLangAttribute', $state);
		return $this;
	}

    /**
     * @param bool $getInt
     * @return $this
     */
    public function setDoCrawl(bool $getInt): Domain
    {
        $this->set('doCrawl', $getInt);
        return $this;
    }

    /**
     * @return boolean
     */
    public function getDoCrawl(): bool
    {
		$user = $this->getUser();
		return ($user && $user->isCrawlAllDomains()) && $this->get('doCrawl', true);
    }

	/**
	 * @param bool $getInt
	 * @return $this
	 */
	public function setActivateEmbeddingsFoundByCrawler(bool $getInt): Domain
	{
		$this->set('activateEmbeddingsFoundByCrawler', $getInt);
		return $this;
	}

	/**
	 * @return boolean
	 */
	public function isActivateEmbeddingsFoundByCrawler(): bool
	{
		//Wenn CrawlAllDomains auf true gesetzt ist, dann darf immer gecrawlt werden
		$user = $this->getUser();
		return ($user && $user->isAllowAutoActivationOfNewEmbeddings()) && $this->get('activateEmbeddingsFoundByCrawler', true);
	}


    /**
     * @param $user
     * @return boolean
     */
    public function getDoCrawlApi($user): bool
    {
        //Wenn CrawlAllDomains auf true gesetzt ist, dann darf immer gecrawlt werden
        if($user->isCrawlAllDomains()=== true)
        {
            return (bool)$this->get('doCrawl', true);
        }
        return false;
    }

    /**
     * @param string $domainHtPassWd
     * @return $this
     */
    public function setDomainHtPassWd(string $domainHtPassWd): Domain
    {
        $this->set('domainHtPassWd', $domainHtPassWd);
        return $this;
    }

    /**
     * @return string
     */
    public function getDomainHtPassWd(): string
    {
        return (string)$this->get('domainHtPassWd', '');
    }


	/**
	 * @param string $domainFilterForCrawler
	 * @return $this
	 */
	public function setDomainFilterForCrawler(string $domainFilterForCrawler): Domain
	{
		$this->set('domainFilterForCrawler', $domainFilterForCrawler);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getDomainFilterForCrawler(): string
	{
		return (string)$this->get('domainFilterForCrawler', '');
	}

	/**
	 * @param string $domainFilterForCrawler
	 * @return $this
	 */
	public function setDomainIncludeFilterForCrawler(string $domainFilterForCrawler): Domain
	{
		$this->set('domainIncludeFilterForCrawler', $domainFilterForCrawler);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getDomainIncludeFilterForCrawler(): string
	{
		return (string)$this->get('domainIncludeFilterForCrawler', '');
	}

	/**
	 * @param array $countPages
	 * @return $this
	 */
	public function setCountPages(array $countPages): Domain
	{
		$this->set('countPages', $countPages);
		return $this;
	}

	/**
	 * @return array
	 */
	public function getCountPages(): array
	{
		return (array)$this->get('countPages', array());
	}

	/**
	 * @param string $cms
	 * @return $this
	 */
	public function setCMS(string $cms): Domain
	{
		$this->set('cms', $cms);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCMS(): string
	{
		return ucfirst((string)$this->get('cms', "Standard"));
	}

	/**
	 * Gibt den spezifischen Link zur Dokumentation des jeweiligen CMS zurück
	 * @return string|null
	 *
	 */
	public function getCmsDocumentationLink(): ?string
	{
		$cms = $this->getCMS();

		$links = array(
			"Standard"=> "https://docs.ccm19.com/integrationen/ccm19-standard/",
			"Typo3"=> "https://docs.ccm19.com/integrationen/ccm19-typo3/",
			"Shopware"=> "https://docs.ccm19.com/integrationen/ccm19-shopware/",
			"Shopify"=> "https://docs.ccm19.com/integrationen/ccm19-shpfy/",
			"Wordpress"=> "https://docs.ccm19.com/integrationen/ccm19-wordpress/",
			"Elementor"=> "https://docs.ccm19.com/integrationen/ccm19-wordpress/",
			"Wix"=> "https://docs.ccm19.com/integrationen/ccm19-wix/",
			"Magento"=> "https://docs.ccm19.com/integrationen/ccm19-magento/",
			"Joomla"=> "https://docs.ccm19.com/integrationen/ccm19-joomla/",
			"Drupal"=> "https://docs.ccm19.com/integrationen/ccm19-drupal/",
			"Jimdo"=> "https://docs.ccm19.com/integrationen/ccm19-jimdo/",
			"Ionos"=> "https://docs.ccm19.com/integrationen/ccm19-IONOS%20Baukasten/",
			"Epages"=> "https://docs.ccm19.com/integrationen/ccm19-Strato%20epages/",
			"Gambio"=> "https://docs.ccm19.com/integrationen/ccm19-gambio/",
			"Nopcommerce"=> "https://docs.ccm19.com/integrationen/ccm19-nopcommerce/",
			"Oxid"=> "https://docs.ccm19.com/integrationen/ccm19-OXID%20eShop/"
		);

		return $links[$cms] ?? $links["Standard"];
	}

	/**
	 * @param string $version "2.2"
	 * @return $this
	 */
	public function setTcfCompatVersion(string $version)
	{
		$this->set('tcfCompatVersion', $version);
		return $this;
	}

	/**
	 * Reserviert für TCF-Kompatibilitätsmodi
	 * @return string "2.2"
	 */
	public function getTcfCompatVersion(): string
	{
		return (string)$this->get('tcfCompatVersion', "2.2");
	}

	/**
	 * Rechtsgrundlagen ('eu-gdpr', 'ch-dsg', 'us-ccpa')
	 *
	 * NULL wenn Standardauswahl des Benutzers genommen werden soll.
	 * Wenn $fallback TRUE ist, wird bei NULL die Standardauswahl des Benutzers ausgegeben.
	 *
	 * @return string[]|null
	 */
	public function getLegalBases(): ?array
	{
		$result = $this->get('legalBases');
		if ($result === null) {
			return null;
		} else {
			return (array)$result;
		}
	}

	/**
	 * Effektiv gültige Rechtsgrundlagen ('eu-gdpr', 'ch-dsg', 'us-ccpa')
	 *
	 * Wenn keine spezielle Konfiguration für die Domain gesetzt ist, wird die Standardauswahl des
	 * Owners genommen.
	 *
	 * @return string[]
	 */
	public function getEffectiveLegalBases(): array
	{
		$result = $this->getLegalBases();
		if ($result === null) {
			$owner = $this->getUser();
			if ($owner) {
				return $owner->getDefaultLegalBases();
			}
			return [];
		} else {
			return $result;
		}
	}

	/**
	 * @param string[]|null $values
	 * @return $this
	 */
	public function setLegalBases(?array $values): self
	{
		$this->set('legalBases', $values);
		return $this;
	}

	/**
	 * @param string|null $name
	 * @return $this
	 */
	public function setShareGroupName(?string $name)
	{
		$this->set('shareGroupName', $name);
		return $this;
	}

	/**
	 * Name der Gruppe für die Consent-Teilung
	 *
	 * Kein automatisches Fallback bei Nicht-Angabe. Das Fallback muss im Controller
	 * umgesetzt werden.
	 * @return string|null
	 */
	public function getShareGroupName(): ?string
	{
		$result = $this->get('shareGroupName', '');
		if ($result === null or $result === '') {
			return null;
		} else {
			return (string)$result;
		}
	}

	/**
	 * @param string|null $url
	 * @return $this
	 */
	public function setShareGroupUrl(?string $url)
	{
		$this->set('shareGroupUrl', $url);
		return $this;
	}

	/**
	 * URL der Gruppe für die Consent-Teilung
	 *
	 * Kein automatisches Fallback bei Nicht-Angabe. Das Fallback muss im Controller
	 * umgesetzt werden.
	 * @return string|null
	 */
	public function getShareGroupUrl(): ?string
	{
		$result = $this->get('shareGroupUrl', '');
		if ($result === null or $result === '') {
			return null;
		} else {
			return (string)$result;
		}
	}
}
