<?php

namespace App\Model;

use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Component\Tcf\v2\Model\Feature as TcfFeature;
use App\Component\Tcf\v2\Model\SpecialPurpose as TcfSpecialPurpose;
use App\Component\Tcf\v2\Model\Translation\Feature as TranslatedTcfFeature;
use App\Component\Tcf\v2\Model\Translation\Purpose as TranslatedTcfPurpose;
use App\Component\Tcf\v2\Model\Translation\SpecialFeature as TranslatedTcfSpecialFeature;
use App\Component\Tcf\v2\Model\Translation\SpecialPurpose as TranslatedTcfSpecialPurpose;
use App\Component\Config\RepositoryInterface;

/**
 * Domain-spezifische Embeddings
 */
class Embedding extends EmbeddingBase
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Liste aller Einbettungen mit dem Zweck `$purpose`
	 *
	 * @param Purpose $purpose
	 * @return static[]
	 */
	public static function byPurpose(Purpose $purpose): array
	{
		return array_filter(static::all(), function (Embedding $obj) use ($purpose) {
			return $obj->getPurpose()->getId() == $purpose->getId();
		});
	}

	/**
	 * @return EmbeddingAsset[]
	 */
	public function getAssets(): array
	{
		if (empty($this->assets)) {
			$this->assets = array_values(array_filter(array_map(function ($id) {
				return EmbeddingAsset::find($id, $this);
			}, $this->get('assets', []))));
		}

		return $this->assets;
	}

	/**
	 * @param string $assetId
	 * @return static
	 */
	public function removeAsset(string $assetId): EmbeddingBase
	{
		parent::removeAsset($assetId);
		EmbeddingAsset::delete($assetId, false);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedName(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_name.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_name.'.$locale->getId());
		}
		if ($useFallback and ($result === null or $result === '')) {
			return $this->getName();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $label
	 * @return $this
	 */
	public function setTranslatedName(LocaleBase $locale, string $label): self
	{
		$key = 'translated_name.'.$locale->getName();
		if ($label !== $this->get($key) and $label !== $this->getName()) {
			$this->makingLegallyRelevantChange();
		}

		$this->set($key, $label);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedDescription(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_description.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_description.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getDescription();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $description
	 * @return $this
	 */
	public function setTranslatedDescription(LocaleBase $locale, string $description): self
	{
		$key = 'translated_description.'.$locale->getName();
		if ($description !== $this->get($key) and $description !== $this->getDescription()) {
			$this->makingLegallyRelevantChange();
		}

		$this->set($key, $description);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPrivacyPolicyUrl(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_privacy_policy_url.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_privacy_policy_url.'.$locale->getId());
		}
		if ($useFallback and ($result === null or $result === '')) {
			return $this->getPrivacyPolicyUrl();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $url
	 * @return $this
	 */
	public function setTranslatedPrivacyPolicyUrl(LocaleBase $locale, string $url): self
	{
		$key = 'translated_privacy_policy_url.'.$locale->getName();
		$url = trim($url);
		if ($url !== $this->get($key) and $url !== $this->getPrivacyPolicyUrl()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $url);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @return TcfPurpose[]|TranslatedTcfPurpose[]
	 */
	public function getTranslatedTcfPurposes(LocaleBase $locale): array
	{
		$langCode = preg_replace('~[-_].+~', '', $locale->getName());
		return array_map(function (TcfPurpose $purpose) use ($langCode) {
			return $purpose->getTranslation($langCode);
		}, $this->getTcfPurposes());
	}

	/**
	 * @param LocaleBase $locale
	 * @return TcfSpecialPurpose[]|TranslatedTcfSpecialPurpose[]
	 */
	public function getTranslatedTcfSpecialPurposes(LocaleBase $locale): array
	{
		$langCode = preg_replace('~[-_].+~', '', $locale->getName());
		return array_map(function (TcfSpecialPurpose $purpose) use ($langCode) {
			return $purpose->getTranslation($langCode);
		}, $this->getTcfSpecialPurposes());
	}

	/**
	 * @param LocaleBase $locale
	 * @return TcfFeature[]|TranslatedTcfFeature[]
	 */
	public function getTranslatedTcfFeatures(LocaleBase $locale): array
	{
		$langCode = preg_replace('~[-_].+~', '', $locale->getName());
		return array_map(function (TcfFeature $feature) use ($langCode) {
			return $feature->getTranslation($langCode);
		}, $this->getTcfFeatures());
	}

	/**
	 * @param LocaleBase $locale
	 * @return TcfSpecialFeature[]|TranslatedTcfSpecialFeature[]
	 */
	public function getTranslatedTcfSpecialFeatures(LocaleBase $locale): array
	{
		$langCode = preg_replace('~[-_].+~', '', $locale->getName());
		return array_map(function (TcfSpecialFeature $feature) use ($langCode) {
			return $feature->getTranslation($langCode);
		}, $this->getTcfSpecialFeatures());
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 * @deprecated
	 */
	public function getTranslatedLifetime(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_lifetime.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_lifetime.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getLifetime();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $lifetime
	 * @return $this
	 * @deprecated
	 */
	public function setTranslatedLifetime(LocaleBase $locale, string $lifetime): self
	{
		$this->set('translated_lifetime.'.$locale->getName(), $lifetime);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedCollectedDataInfo(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_collected_data_info.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_collected_data_info.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getCollectedDataInfo();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $info
	 * @return $this
	 */
	public function setTranslatedCollectedDataInfo(LocaleBase $locale, string $info): self
	{
		$key = 'translated_collected_data_info.'.$locale->getName();
		if ($info !== $this->get($key) and $info !== $this->getCollectedDataInfo()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $info);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPurposeOfDataCollection(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_purpose_of_data_collection.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_purpose_of_data_collection.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getPurposeOfDataCollection();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $purpose
	 * @return $this
	 */
	public function setTranslatedPurposeOfDataCollection(LocaleBase $locale, string $purpose): self
	{
		$key = 'translated_purpose_of_data_collection.'.$locale->getName();
		if ($purpose !== $this->get($key) and $purpose !== $this->getPurposeOfDataCollection()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $purpose);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedLegalBasis(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_legal_basis.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_legal_basis.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getLegalBasis();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $basis
	 * @return $this
	 */
	public function setTranslatedLegalBasis(LocaleBase $locale, string $basis): self
	{
		$key = 'translated_legal_basis.'.$locale->getName();
		if ($basis !== $this->get($key) and $basis !== $this->getLegalBasis()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $basis);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedLegalBasisSupplement(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_legal_basis_supplement.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_legal_basis_supplement.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getLegalBasisSupplement();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $basis
	 * @return $this
	 */
	public function setTranslatedLegalBasisSupplement(LocaleBase $locale, string $basis): self
	{
		$key = 'translated_legal_basis_supplement.'.$locale->getName();
		if ($basis !== $this->get($key) and $basis !== $this->getLegalBasisSupplement()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $basis);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPlaceOfProcessing(LocaleBase $locale, $useFallback=true): string
	{
		$result = $this->get('translated_place_of_processing.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_place_of_processing.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getPlaceOfProcessing();
		}
		return (string)$result;
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $place
	 * @return $this
	 */
	public function setTranslatedPlaceOfProcessing(LocaleBase $locale, string $place): self
	{
		$key = 'translated_place_of_processing.'.$locale->getName();
		if ($place !== $this->get($key) and $place !== $this->getPlaceOfProcessing()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('translated_place_of_processing.'.$locale->getName(), $place);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentTitle(LocaleBase $locale, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_title.{$locale->getName()}");
		return strlen($value) > 0 ? $value : ($useFallback ? $locale->getBlockedContentTitle() : '');
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentTitle(LocaleBase $locale, string $value): self
	{
		$value = trim($value);
		$currentValue = $this->getTranslatedBlockedContentTitle($locale, false);
		$defaultValue = $locale->getBlockedContentTitle();
		if ($value != $currentValue && ($currentValue != '' || $value != $defaultValue)) {
			$this->makingLegallyRelevantChange();
		}
		$this->set("translated_blocked_content_title.{$locale->getName()}", $value);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentText(LocaleBase $locale, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_text.{$locale->getName()}");
		return strlen($value) > 0 ? $value : ($useFallback ? $locale->getBlockedContentText() : '');
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentText(LocaleBase $locale, string $value): self
	{
		$value = trim($value);
		$currentValue = $this->getTranslatedBlockedContentText($locale, false);
		$defaultValue = $locale->getBlockedContentText();
		if ($value != $currentValue && ($currentValue != '' || $value != $defaultValue)) {
			$this->makingLegallyRelevantChange();
		}
		$this->set("translated_blocked_content_text.{$locale->getName()}", $value);
		return $this;
	}

	/**
	 * @param LocaleBase $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentButtonText(LocaleBase $locale, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_button_text.{$locale->getName()}");
		return strlen($value) > 0 ? $value : ($useFallback ? $locale->getBlockedContentButtonText() : '');
	}

	/**
	 * @param LocaleBase $locale
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentButtonText(LocaleBase $locale, string $value): self
	{
		$value = trim($value);
		$currentValue = $this->getTranslatedBlockedContentButtonText($locale, false);
		$defaultValue = $locale->getBlockedContentButtonText();
		if ($value != $currentValue && ($currentValue != '' || $value != $defaultValue)) {
			$this->makingLegallyRelevantChange();
		}
		$this->set("translated_blocked_content_button_text.{$locale->getName()}", $value);
		return $this;
	}

	/**
	 * Setzt alle übersetzten Texte, die in $locale noch nie vorher gesetzt
	 * wurden, auf den nicht-übersetzten Wert.
	 *
	 * @param LocaleBase $locale Zu füllendes Locale
	 * @return $this
	 */
	public function fillTranslationsFromDefault(LocaleBase $locale): self
	{
		$this->setTranslatedCollectedDataInfo($locale, $this->getTranslatedCollectedDataInfo($locale));
		$this->setTranslatedDescription($locale, $this->getTranslatedDescription($locale));
		$this->setTranslatedLegalBasis($locale, $this->getTranslatedLegalBasis($locale));
		$this->setTranslatedLifetime($locale, $this->getTranslatedLifetime($locale));
		$this->setTranslatedPlaceOfProcessing($locale, $this->getTranslatedPlaceOfProcessing($locale));
		$this->setTranslatedPrivacyPolicyUrl($locale, $this->getTranslatedPrivacyPolicyUrl($locale));
		$this->setTranslatedPurposeOfDataCollection($locale, $this->getTranslatedPurposeOfDataCollection($locale));
		$this->setTranslatedPurposeOfDataCollection($locale, $this->getTranslatedPurposeOfDataCollection($locale));
		$this->setTranslatedBlockedContentTitle($locale, $this->getTranslatedBlockedContentTitle($locale));
		$this->setTranslatedBlockedContentText($locale, $this->getTranslatedBlockedContentText($locale));
		$this->setTranslatedBlockedContentButtonText($locale, $this->getTranslatedBlockedContentButtonText($locale));
		$this->setTranslatedLegalBasisSupplement($locale, $this->getTranslatedLegalBasisSupplement($locale));
		return $this;
	}

	/**
	 * Listet die Drittanbieter/Partner, die aktive Embeddings bereitstellen
	 *
	 * Dabei werden Embeddings mit Anbieter "{Firma}" und Duplikate ausgelassen.
	 *
	 * @param bool $includeTcf Ob Embedding-Anbieter mit verknüpftem TCF-Vendor auch gelistet werden sollen
	 * @return string[]
	 */
	public static function listActiveThirdPartyVendors(bool $includeTcf = true): array
	{
		$vendors = [];
		foreach (self::all() as $embedding) {
			if ($embedding->isActive()) {
				$vendor = $embedding->getEmbeddingVendor();
				$vendorKey = mb_convert_case(trim($embedding->getEmbeddingVendor()), MB_CASE_LOWER);
				if ($vendorKey != '{firma}' and ($includeTcf or !(bool)$embedding->get('tcf_vendor', null))) {
					$vendors[$vendorKey] = $vendor;
				}
			}
		}
		return array_values($vendors);
	}

}
