<?php

namespace App\Model;

use App\Utils;

abstract class EmbeddingAssetBase extends Model
{
	const STORAGE_TYPE_COOKIE = 'cookie';
	const STORAGE_TYPE_LOCAL_STORAGE = 'localStorage';
	const STORAGE_TYPE_SESSION_STORAGE = 'sessionStorage';
	const STORAGE_TYPE_INDEXED_DB = 'indexedDb';
	const STORAGE_TYPE_WEB_SQL = 'webSql';
	const STORAGE_TYPE_OTHER = 'other';

	const ALLOWED_STORAGE_TYPE_VALUES = [
		self::STORAGE_TYPE_COOKIE,
		self::STORAGE_TYPE_LOCAL_STORAGE,
		self::STORAGE_TYPE_SESSION_STORAGE,
		self::STORAGE_TYPE_INDEXED_DB,
		self::STORAGE_TYPE_WEB_SQL,
		self::STORAGE_TYPE_OTHER,
	];

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-embedding-assets';

	/** @var EmbeddingBase|null $associatedEmbedding */
	protected $associatedEmbedding;

	/** @var bool $wantsMongoDbMtimeIndex */
	protected static $wantsMongoDbMtimeIndex = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * @param string $id
	 * @param EmbeddingBase|null $_embedding (intern verwendet)
	 * @return static|null
	 */
	public static function find(string $id, ?EmbeddingBase $_embedding = null): ?Model
	{
		$result = parent::find($id);
		if ($result) {
			assert($result instanceof EmbeddingAssetBase);
			$result->associatedEmbedding = $_embedding;
		}
		return $result;
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setName(string $text): self
	{
		if ($text !== $this->get('name')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('name', $text);
		return $this;
	}

	/**
	 * Überprüfen, ob ein Cookie/...-Name zu diesem Asset passt
	 */
	public function matchName(string $name): bool
	{
		return self::matchAssetName($this->getName(), $name, $this->isDynamic());
	}

	/**
	 * @return bool
	 */
	public function isDynamic(): bool
	{
		return (bool)$this->get('dynamic');
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setDynamic(bool $value): self
	{
		$this->set('dynamic', $value);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getStorageType(): string
	{
		return (string)$this->get('storageType');
	}

	/**
	 * @param string $code
	 * @return $this
	 */
	public function setStorageType(string $code): self
	{
		if ($code !== $this->get('storageType')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('storageType', $code);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getLifetime(): string
	{
		return (string)$this->get('lifetime');
	}

	/**
	 * @param string $lifetime
	 * @return $this
	 */
	public function setLifetime(string $lifetime): self
	{
		if ($lifetime !== $this->get('lifetime')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('lifetime', $lifetime);
		return $this;
	}

	/**
	 * Maximale Lebenszeit des Cookies in Stunden, oder
	 * $fallback wenn die Zeit nicht eindeutig bestimmbar ist.
	 *
	 * @param mixed $sessionFallback
	 * @param mixed $fallback
	 * @return int|mixed
	 */
	public function getLifetimeHours($sessionFallback=null, $fallback=null)
	{
		return Utils::parseLifetimeToHours($this->getLifetime(), $sessionFallback, $fallback);
	}

	/**
	 * @return string
	 */
	public function getDescription(): string
	{
		return (string)$this->get('description');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setDescription(string $text): self
	{
		if ($text !== $this->get('description')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('description', $text);
		return $this;
	}

	protected function makingLegallyRelevantChange(): void
	{
		if ($this->associatedEmbedding) {
			$this->associatedEmbedding->makingLegallyRelevantChange();
		}
	}

	public function assumeAssociatedEmbedding(?EmbeddingBase $embedding): void
	{
		$this->associatedEmbedding = $embedding;
	}

	public static function matchAssetName(string $pattern, string $name, bool $dynamic): bool
	{
		$name = strtolower(trim($name));
		$pattern = strtolower(trim($pattern));

		if (strpos($pattern, ' ') !== false) {
			$patterns = [$pattern, explode(' ', $pattern, 2)[0]];
		} else {
			$patterns = [$pattern];
		}

		foreach ($patterns as $pattern) {
			if (!$dynamic) {
				if ($pattern === $name) {
					return true;
				}
			}
			else {
				if (strpos($pattern, '[') !== false) {
					$pattern = preg_replace('/(?<!\\\\)([][])/', '\\\\$1', $pattern);
				}
				if (fnmatch($pattern, substr($name, 0, 4095))) {
					return true;
				}
			}
		}
		return false;
	}
}
