<?php

namespace App\Model;

use App\Component\Tcf\v2\Model\Feature as TcfFeature;
use App\Component\Tcf\v2\Model\SpecialPurpose as TcfSpecialPurpose;
use App\Component\Tcf\v2\Model\Vendor as TcfVendor;
use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Model\Regulations\Article;
use App\Model\Regulations\Regulation;
use App\Utils;

use DateTimeImmutable;
use DateTimeInterface;
use InvalidArgumentException;

/**
 * Basis-Klasse für Einbettungen, unabhängig vom Speicherort
 */
abstract class EmbeddingBase extends Model
{
	const SOURCE_UNKNOWN = '';
	const SOURCE_LOCAL = 'local';
	const SOURCE_SYSTEM_TEMPLATE = 'system';
	const SOURCE_USER_TEMPLATE = 'user';
	const SOURCE_GLOBAL_DB_LEGACY = 'db';
	const SOURCE_GLOBAL_DB = 'db2';

	const EXCLUDE_URLS_MODE_BLACKLIST = 'blacklist';
	const EXCLUDE_URLS_MODE_WHITELIST = 'whitelist';

	const ALLOWED_EXCLUDE_URLS_MODE_VALUES = [
		self::EXCLUDE_URLS_MODE_BLACKLIST,
		self::EXCLUDE_URLS_MODE_WHITELIST,
	];

	const GCM_AD_STORAGE = 'ad_storage';
	const GCM_ANALYTICS_STORAGE = 'analytics_storage';
	const GCM_FUNCTIONALITY_STORAGE = 'functionality_storage';
	const GCM_PERSONALIZATION_STORAGE = 'personalization_storage';
	const GCM_SECURITY_STORAGE = 'security_storage';
	const GCM_AD_USER_DATA = 'ad_user_data';
	const GCM_AD_PERSONALIZATION = 'ad_personalization';

	const ALLOWED_GCM_VALUES = [
		self::GCM_AD_STORAGE,
		self::GCM_ANALYTICS_STORAGE,
		self::GCM_FUNCTIONALITY_STORAGE,
		self::GCM_PERSONALIZATION_STORAGE,
		self::GCM_SECURITY_STORAGE,
		self::GCM_AD_USER_DATA,
		self::GCM_AD_PERSONALIZATION,
	];

	public const FIELD_CODE_BLOCK = 'code_block';
	public const FIELD_CODE_SNIPPET = 'code_snippet';
	public const FIELD_PURPOSE = 'purpose';
	public const FIELD_EMBEDDING_NAME = 'name';
	public const FIELD_EMBEDDING_ACTIVE = 'embedding_active';
	public const FIELD_EMBEDDING_VENDOR = 'embedding_vendor';
	public const FIELD_SCRIPT_LOADER = 'script_loader_group';
	public const FIELD_STRUCTURED_LEGAL_BASIS = 'structured_legal_basis';
	public const FIELD_IFRAMES_BLOCK = 'iframe_blocking_strings';
	public const FIELD_IFRAMES_HIDE = 'iframes_hide';
	public const FIELD_IFRAMES_LOAD_EVENT = 'iframes_load_event';
	public const FIELD_EXCLUDE_URLS_MODE = 'exclude_urls_mode';
	public const FIELD_EXCLUDE_URLS = 'exclude_urls';



	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-embeddings';

	/** @var EmbeddingAssetBase[] $assets */
	protected $assets;

	/** @var bool $hasLegallyRelevantChange */
	protected $hasLegallyRelevantChange = false;

	/** @var bool $wantsMongoDbMtimeIndex */
	protected static $wantsMongoDbMtimeIndex = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
		$this->assets = [];
	}

	/**
	 * Derzeit Noop (bis auf String-Umwandlung).
	 *
	 * @param mixed $data
	 * @return string
	 */
	public static function decodeNameFromQueryString($data): string
	{
		return (string) $data;
	}

	/**
	 * Diese Methode ergänzt die abzuleitende Methode um Sortierung der Embeddings nach Vertreiber und Name.
	 * @return static[]
	 */
	public static function all(): array
	{
		/** @var static[] $embeddings */
		$embeddings = parent::all();

		// ORDER BY vendor, name
		uasort($embeddings, function ($a, $b) {
			/**
			 * @var self $a
			 * @var self $b
			 */
			return ($vendorCmp = strnatcasecmp($a->getEmbeddingVendor(), $b->getEmbeddingVendor())) == 0
				? strnatcasecmp($a->getName(), $b->getName())
				: $vendorCmp;
		});

		return $embeddings;
	}

	/**
	 * Diese Methode ergänzt die abzuleitende Methode um Sortierung der Embeddings nach Purpose und Name.
	 * @return static[]
	 */
	public static function sortAllByPurpose(): array
	{
		/** @var static[] $embeddings */
		$embeddings = parent::all();

		// ORDER BY vendor, name
		uasort($embeddings, function ($a, $b) {
			/**
			 * @var self $a
			 * @var self $b
			 */
			return ($vendorCmp = strnatcasecmp($a->getPurpose()->getName(), $b->getPurpose()->getName())) == 0
				? strnatcasecmp($a->getName(), $b->getName())
				: $vendorCmp;
		});

		return $embeddings;
	}

	/**
	 * Diese Methode liefert alle Embeddings, die technisch notwendig sind.
	 * @return static[]
	 */
	public static function mandatory(): array
	{
		return array_filter(self::all(), function (self $embedding) {
			return $embedding->isMandatory();
		});
	}

	/**
	 * Diese Methode liefert alle Embeddings, die nicht technisch notwendig sind.
	 * @return static[]
	 */
	public static function optional(): array
	{
		return array_filter(self::all(), function (self $embedding) {
			return $embedding->isOptional();
		});
	}

	/**
	 * @return bool
	 */
	public function isMandatory(): bool
	{
		return $this->getPurpose() && $this->getPurpose()->isMandatory();
	}

	/**
	 * @return bool
	 */
	public function isOptional(): bool
	{
		return $this->isMandatory() == false;
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setName(string $name): self
	{
		if ($name !== $this->get('name')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('name', $name);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCodeSnippet(): string
	{
		return (string)$this->get('code');
	}

	/**
	 * @param string $code
	 * @return $this
	 */
	public function setCodeSnippet(string $code): self
	{
		$this->set('code', $code);
		return $this;
	}

	/**
	* @return string[]
	*/
	public function getCodeBlock(): array
	{
		$value = $this->get('code_block');
		if (is_string($value)) {
			$value = array_filter(explode("\n", trim($value)), function ($line) { return trim($line) !== ''; });
		}
		return (array)$value;
	}

	/**
	* @param string[] $code
	* @return $this
	*/
	public function setCodeBlock(array $code): self
	{
		$this->set('code_block', $code);
		return $this;
	}

	/**
	 * URLs bei denen das Embedding nicht geladen werden soll
	 *
	 * @return string[]
	 */
	public function getExcludeUrls(): array
	{
		return (array)$this->get('excludeUrls');
	}

	/**
	 * @param string[] $excludeUrls
	 * @return $this
	 */
	public function setExcludeUrls(array $excludeUrls): self
	{
		$this->set('excludeUrls', $excludeUrls);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getExcludeUrlsMode(): string
	{
		return (string)$this->get('excludeUrlsMode', self::EXCLUDE_URLS_MODE_BLACKLIST);
	}

	/**
	 * @param string $mode
	 * @return $this
	 * @throws InvalidArgumentException if an invalid mode is used.
	 */
	public function setExcludeUrlsMode(string $mode): self
	{
		if (in_array($mode, self::ALLOWED_EXCLUDE_URLS_MODE_VALUES) == false) {
			throw new InvalidArgumentException('Invalid excludeUrlsMode, allowed values: '.implode(', ', self::ALLOWED_EXCLUDE_URLS_MODE_VALUES));
		}

		$this->set('excludeUrlsMode', $mode);
		return $this;
	}

	/**
	 * @return EmbeddingAssetBase[]
	 */
	abstract public function getAssets(): array;

	/**
	 * Liste aller Assets, die den angegebenen Storage-Type haben
	 *
	 * @return EmbeddingAssetBase[]
	 */
	public function getAssetsWithType(string $storageType): array
	{
		return array_filter($this->getAssets(), function (EmbeddingAssetBase $asset) use ($storageType) {
			return $asset->getStorageType() == $storageType;
		});
	}

	/**
	 * @param EmbeddingAssetBase[] $assets
	 * @return $this
	 */
	public function setAssets(array $assets): self
	{
		$this->assets = $assets;

		foreach ($assets as $asset) {
			$asset->assumeAssociatedEmbedding($this);
		}

		$ids = array_map(function ($item) {
			return $item->getId();
		}, $assets);

		// Hinzufügen von Assets ist relevant
		if (array_diff($ids, $this->get('assets', []))) {
			$this->makingLegallyRelevantChange();
		}

		$this->set('assets', $ids);
		return $this;
	}

	/**
	 * @param string $id
	 * @return EmbeddingAssetBase|null
	 */
	public function getAsset(string $id): ?EmbeddingAssetBase
	{
		return array_reduce($this->getAssets(), function ($carry, EmbeddingAssetBase $asset) use ($id) {
			return $carry ?? ($asset->getId() == $id ? $asset : $carry);
		}, null);
	}

	/**
	 * @param EmbeddingAssetBase $target
	 * @return bool
	 */
	public function hasAsset(EmbeddingAssetBase $target): bool
	{
		return array_reduce($this->getAssets(), function ($hasAsset, EmbeddingAssetBase $asset) use ($target) {
			return $hasAsset || $asset->getId() == $target->getId();
		}, false);
	}

	/**
	 * @param EmbeddingAssetBase $asset
	 * @return $this
	 */
	public function addAsset(EmbeddingAssetBase $asset): self
	{
		if ($this->hasAsset($asset) == false) {
			$this->assets[] = $asset;
			$asset->assumeAssociatedEmbedding($this);
			$this->appendToArray('assets', $asset->getId());
			$this->makingLegallyRelevantChange();
		}

		return $this;
	}

	/**
	 * @param string $assetId
	 * @return $this
	 */
	public function removeAsset(string $assetId): self
	{
		$this->assets = array_filter($this->assets, function (EmbeddingAssetBase $asset) use ($assetId) {
			return $asset->getId() != $assetId;
		});

		$this->removeFromArray('assets', $assetId);

		return $this;
	}

	/**
	 * @return string
	 */
	public function getDescription(): string
	{
		return (string)$this->get('description');
	}

	/**
	 * @param string $description
	 * @return $this
	 */
	public function setDescription(string $description): self
	{
		if ($description !== $this->get('description')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('description', $description);
		return $this;
	}

	/**
	 * @return Purpose|null
	 */
	public function getPurpose(): ?Purpose
	{
		return Purpose::find((string)$this->get('purpose'));
	}

	/**
	 * @param Purpose $purpose
	 * @return $this
	 */
	public function setPurpose(Purpose $purpose): self
	{
		$purposeId = $purpose->getId();
		if ($purposeId !== $this->get('purpose')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('purpose', $purposeId);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPrivacyPolicyUrl(): string
	{
		return (string)$this->get('privacy_policy_url');
	}

	/**
	 * @param string $url
	 * @return $this
	 */
	public function setPrivacyPolicyUrl(string $url): self
	{
		$url = trim($url);
		if ($url !== $this->get('privacy_policy_url')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('privacy_policy_url', $url);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getEmbeddingVendor(): string
	{
		return (string)$this->get('embedding_vendor');
	}

	/**
	 * @param string $vendor
	 * @return $this
	 */
	public function setEmbeddingVendor(string $vendor): self
	{
		if ($vendor !== $this->get('embedding_vendor')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('embedding_vendor', $vendor);
		return $this;
	}

	public function hasVendorGoogle(): bool
	{
		return stripos($this->getEmbeddingVendor(), 'google') !== false;
	}

	/**
	 * @return TcfVendor|null
	 */
	public function getTcfVendor(): ?TcfVendor
	{
		return TcfVendor::find((string)$this->get('tcf_vendor'));
	}

	/**
	 * @param TcfVendor|null $vendor
	 * @return $this
	 */
	public function setTcfVendor(?TcfVendor $vendor): self
	{
		$vendorId = $vendor ? $vendor->getId() : null;
		if ($vendorId !== $this->get('tcf_vendor')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('tcf_vendor', $vendorId);
		return $this;
	}

	/**
	 * @return TcfPurpose[]
	 */
	public function getTcfPurposes(): array
	{
		return array_filter(array_map(function ($id) {
			return TcfPurpose::find($id);
		}, $this->get('tcf_purposes', [])));
	}

	/**
	 * @param TcfPurpose[] $purposes
	 * @return $this
	 */
	public function setTcfPurposes(array $purposes): self
	{
		$purposeIds = array_map(function (TcfPurpose $purpose) {
			return $purpose->getId();
		}, $purposes);

		// Hinzufügen von Zwecken ist relevant
		if (array_diff($purposeIds, (array)$this->get('tcf_purposes'))) {
			$this->makingLegallyRelevantChange();
		}

		$this->set('tcf_purposes', $purposeIds);
		return $this;
	}

	/**
	 * @return TcfSpecialPurpose[]
	 */
	public function getTcfSpecialPurposes(): array
	{
		$vendor = $this->getTcfVendor();
		return $vendor ? $vendor->getSpecialPurposes() : [];
	}

	/**
	 * @return TcfFeature[]
	 */
	public function getTcfFeatures(): array
	{
		$vendor = $this->getTcfVendor();
		return $vendor ? $vendor->getFeatures() : [];
	}

	/**
	 * @return TcfSpecialFeature[]
	 */
	public function getTcfSpecialFeatures(): array
	{
		return array_filter(array_map(function ($id) {
			return TcfSpecialFeature::find($id);
		}, $this->get('tcf_specialFeatures', [])));
	}

	/**
	 * @param TcfSpecialFeature[] $features
	 * @return $this
	 */
	public function setTcfSpecialFeatures(array $features): self
	{
		$featureIds = array_map(function (TcfSpecialFeature $feature) {
			return $feature->getId();
		}, $features);

		// Hinzufügen von Special Features ist relevant
		if (array_diff($featureIds, (array)$this->get('tcf_specialFeatures'))) {
			$this->makingLegallyRelevantChange();
		}

		$this->set('tcf_specialFeatures', $featureIds);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getGoogleConsentMode(): array
	{
		return (array)$this->get('gcm', []);
	}

	/**
	 * @param string[] $types
	 * @return $this
	 */
	public function setGoogleConsentMode(array $types): self
	{
		// Hinzufügen von Special Features ist relevant
		if (array_diff($types, (array)$this->get('gcm'))) {
			$this->makingLegallyRelevantChange();
		}

		$this->set('gcm', $types);
		return $this;
	}

	/**
	 * Maximale Lebenszeit des Embeddings in Stunden, oder
	 * $fallback wenn die Zeit nicht eindeutig bestimmbar ist.
	 *
	 * @param mixed $sessionFallback
	 * @param mixed $fallback
	 * @return int|mixed
	 * @deprecated
	 */
	public function getLifetimeHours($sessionFallback=null, $fallback=null)
	{
		return Utils::parseLifetimeToHours($this->getLifetime(), $sessionFallback, $fallback);
	}

	/**
	 * @return string
	 * @deprecated
	 */
	public function getLifetime(): string
	{
		return (string)$this->get('lifetime');
	}

	/**
	 * @param string $lifetime
	 * @return $this
	 * @deprecated
	 */
	public function setLifetime(string $lifetime): self
	{
		$this->set('lifetime', $lifetime);
		return $this;
	}

	/**
	 * @return string
	 * @deprecated
	 */
	public function getHttpEmbeddingName(): string
	{
		return (string)$this->get('http_embedding_name');
	}

	/**
	 * @param string $name
	 * @return $this
	 * @deprecated
	 */
	public function setHttpEmbeddingName(string $name): self
	{
		$this->set('http_embedding_name', $name);
		return $this;
	}

	public function setFoundUrl(string $name): self
	{
		$this->set('foundUrl', $name);
		return $this;
	}

	public function getFoundUrl(): string
	{
		return (string)$this->get('foundUrl');
	}

	public function setFoundTime(int $name): self
	{
		$this->set('foundTime', $name);
		return $this;
	}

	public function getFoundTime(): int
	{
		return (int)$this->get('foundTime');
	}


	/**
	 * @return bool
	 */
	public function isActive(): bool
	{
		return (bool)$this->get('active');
	}

	/**
	 * @param bool $status
	 * @return $this
	 */
	public function setActive(bool $status): self
	{
		if ($status != $this->isActive()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('active', $status);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCollectedDataInfo(): string
	{
		return (string)$this->get('collected_data_info');
	}

	/**
	 * @param string $info
	 * @return $this
	 */
	public function setCollectedDataInfo(string $info): self
	{
		if ($info !== $this->get('collected_data_info')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('collected_data_info', $info);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPurposeOfDataCollection(): string
	{
		return (string)$this->get('purpose_of_data_collection');
	}

	/**
	 * @param string $purpose
	 * @return $this
	 */
	public function setPurposeOfDataCollection(string $purpose): self
	{
		if ($purpose !== $this->get('purpose_of_data_collection')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('purpose_of_data_collection', $purpose);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isLegalBasisStructured(): bool
	{
		return (bool)$this->get('legal_basis_is_structured', false);
	}

	/**
	 * @param bool $structured
	 * @return $this
	 */
	public function setLegalBasisStructured(bool $structured): self
	{
		if ($structured !== $this->get('legal_basis_is_structured', false)) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('legal_basis_is_structured', $structured);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getLegalBasis(): string
	{
		return (string)$this->get('legal_basis');
	}

	/**
	 * @param string $basis
	 * @return $this
	 */
	public function setLegalBasis(string $basis): self
	{
		if ($basis !== $this->get('legal_basis')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('legal_basis', $basis);
		return $this;
	}

	/**
	 * @return array<string,Article[]>
	 */
	public function getStructuredLegalBasis(): array
	{
		$data = (array)$this->get('structured_legal_basis');

		foreach ($data as $name => &$items) {
			$regulation = Regulation::find($name);
			if (!$regulation) { $items = []; }
			$items = array_filter(array_map(function ($id) use ($regulation): ?Article {
				return $regulation->getArticle($id);
			}, (array)$items));
		}

		// Add missing regulations
		$regulationIds = array_keys(Regulation::rawAll());
		$data += array_fill_keys($regulationIds, []);

		return $data;
	}

	/**
	 * @return array<string,int[]>
	 */
	public function getStructuredLegalBasisRaw(): array
	{
		return (array)$this->get('structured_legal_basis');
	}

	/**
	 * @param array<string,(int|Article)[]> $basis
	 * @return $this
	 */
	public function setStructuredLegalBasis(array $basis): self
	{
		$basis = array_map(function ($articles) {
			return array_filter(array_map(function ($article) {
				if (is_object($article)) {
					return $article->getId();
				} else {
					return (int)$article;
				}
			}, (array)$articles));
		}, $basis);

		if ($basis !== $this->get('structured_legal_basis')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('structured_legal_basis', $basis);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getLegalBasisSupplement(): string
	{
		return (string)$this->get('legal_basis_supplement');
	}

	/**
	 * @param string $basis
	 * @return $this
	 */
	public function setLegalBasisSupplement(string $basis): self
	{
		if ($basis !== $this->get('legal_basis_supplement')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('legal_basis_supplement', $basis);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPlaceOfProcessing(): string
	{
		return (string)$this->get('place_of_processing');
	}

	/**
	 * @param string $place
	 * @return $this
	 */
	public function setPlaceOfProcessing(string $place): self
	{
		if ($place !== $this->get('place_of_processing')) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('place_of_processing', $place);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getScriptLoaderGroup(): string
	{
		return $this->get('scriptLoaderGroup', '');
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setScriptLoaderGroup(string $name): self
	{
		$this->set('scriptLoaderGroup', $name);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getIframeBlockingStrings(): array
	{
		return array_values(array_filter($this->get('iframeBlockingStrings', [])));
	}

	/**
	 * @param string[] $strings
	 * @return $this
	 */
	public function setIframeBlockingStrings(array $strings): self
	{
		$this->set('iframeBlockingStrings', array_values(array_filter(array_map(function ($x) { return trim($x); }, $strings))));
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isIframesHiddenUntilConsent(): bool
	{
		return (bool)$this->get('hideIframesUntilConsent', false);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setIframesHiddenUntilConsent(bool $state): self
	{
		$this->set('hideIframesUntilConsent', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isDispatchLoadEventForIframes(): bool
	{
		return (bool)$this->get('dispatchLoadEventForIframes', true);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setDispatchLoadEventForIframes(bool $state): self
	{
		$this->set('dispatchLoadEventForIframes', $state);
		return $this;
	}

	public function getSource(): string
	{
		return (string)($this->get('source') ?? self::SOURCE_UNKNOWN);
	}

	/**
	 * @param string $source
	 * @return $this
	 */
	public function setSource(string $source): self
	{
		$this->set('source', $source);
		return $this;
	}

	/**
	 * @return string|int|null
	 */
	public function getSourceId()
	{
		return $this->get('source_id');
	}

	/**
	 * @param string|int|null $id
	 * @return $this
	 */
	public function setSourceId($id): self
	{
		$this->set('source_id', $id);
		return $this;
	}

	public function isImported(): bool
	{
		$source = $this->getSource();
		return ($source !== self::SOURCE_UNKNOWN and $source !== self::SOURCE_LOCAL and $this->getSourceTimestamp());
	}

	public function getSourceTimestamp(): ?DateTimeImmutable
	{
		$ts = $this->get('source_ts');

		try {
			return $ts ? new DateTimeImmutable((string)$ts) : null;
		} catch (\Exception $e) {
			return null;
		}
	}

	/**
	 * @param DateTimeInterface|null $date
	 * @return $this
	 */
	public function setSourceTimestamp(?DateTimeInterface $date): self
	{
		$this->set('source_ts', $date ? $date->format('Y-m-d H:i:s') : null);
		return $this;
	}

	public function isAutoUpdated(): bool
	{
		return (bool)$this->get('autoUpdate', false);
	}

	/**
	 * @param bool $active
	 * @return $this
	 */
	public function setAutoUpdated(bool $active): self
	{
		$this->set('autoUpdate', $active);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getUpdateParameterBlacklist(): array
	{
		// By default, lock fields that were omitted from updates before this feature was introduced
		$fallback = [self::FIELD_CODE_SNIPPET];
		return (array)$this->get('updateParameterBlacklist', $fallback);
	}

	/**
	 * @param string[] $parameters
	 * @return $this
	 */
	public function setUpdateParameterBlacklist(array $parameters): self
	{
		$parameters = array_values(array_map('strval', $parameters));
		$this->set('updateParameterBlacklist', $parameters);
		return $this;
	}

	/**
	 * @param string $parameter
	 * @return bool
	 */
	public function isUpdateParameterAllowed(string $parameter): bool
	{
		return !in_array($parameter, $this->getUpdateParameterBlacklist());
	}

	/**
	 * @param string $parameter
	 * @return bool
	 */
	public function isUpdateParameterLocked(string $parameter): bool
	{
		return !$this->isUpdateParameterAllowed($parameter);
	}

	public function getLastModified(): ?DateTimeImmutable
	{
		$ts = $this->get('ts');

		try {
			return $ts ? new DateTimeImmutable((string)$ts) : null;
		} catch (\Exception $e) {
			return null;
		}
	}

	/**
	 * @param DateTimeInterface|null $date
	 * @return $this
	 */
	public function setLastModified(?DateTimeInterface $date): self
	{
		$this->set('ts', $date ? $date->format('Y-m-d H:i:s') : null);
		return $this;
	}

	/**
	 * @return $this
	 */
	public function updateLastModified(): self
	{
		$this->setLastModified(new DateTimeImmutable('now'));
		return $this;
	}

	/**
	 * Zeitpunkt der letzten für den Consent rechtlich relevanten Änderung (sofern bereits erfasst)
	 *
	 * @return DateTimeImmutable|null
	 */
	public function getLegalLastModified(): ?DateTimeImmutable
	{
		$ts = $this->get('tsLegal');

		try {
			return $ts ? new DateTimeImmutable((string)$ts) : null;
		} catch (\Exception $e) {
			return null;
		}
	}

	/**
	 * Manuell den Zeitstempel der letzten rechtlich relevanten Änderung aktualisieren.
	 *
	 * Passiert automatisch wenn ein relevantes Feld geändert wird.
	 *
	 * @return $this
	 */
	public function updateLegalLastModified(): self
	{
		$this->hasLegallyRelevantChange = true;
		$this->set('tsLegal', (new DateTimeImmutable('now'))->format('Y-m-d H:i:s'));
		return $this;
	}

	/**
	 * Ob in der aktuellen Instanz eine rechtlich relevante Änderung gemacht wurde
	 *
	 * @return bool
	 */
	public function hasLegallyRelevantChange(): bool
	{
		return $this->hasLegallyRelevantChange;
	}

	/**
	 * Den Zeitstempel der letzten rechtlich relevanten Änderung aktualisieren, falls noch nicht
	 * passiert.
	 *
	 * Vor allem intern verwendet.
	 */
	public function makingLegallyRelevantChange(): void
	{
		if (!$this->hasLegallyRelevantChange) {
			$this->updateLegalLastModified();
		}
	}
}
