<?php

namespace App\Model;

use App\Exception\ConfigNotReadableException;
use App\Component\Config\ModelInterface;
use App\Component\Config\RepositoryInterface;
use App\Component\Config\Store\JsonConfig;
use App\Utils;
use LogicException;
use ReflectionClass;

/**
 * Basisklasse für JSON-Models
 * @phan-immutable
 * @phpstan-consistent-constructor
 */
abstract class ImmutableModel implements ModelInterface
{
	/** @var string $id */
	private $id;

	/** @var string|null $repositoryFilename */
	protected static $repositoryFilename = null;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		$this->id = $id;
	}

	/**
	 * Diese Methode stellt das zugrundeliegende Repository bereit.
	 * @return RepositoryInterface
	 * @phan-suppress PhanUndeclaredStaticProperty
	 * @phan-suppress PhanUndeclaredStaticMethod
	 */
	protected static function getRepository(): RepositoryInterface
	{
		if (static::$repository === null) {
			static::$repository = static::openRepository();

			if (method_exists(static::class, 'onRepositoryLoaded')) {
				static::onRepositoryLoaded();
			}
		}

		return static::$repository;
	}

	/**
	 * Öffnet das Repository vollständig neu
	 *
	 * @phan-suppress PhanUndeclaredStaticProperty
	 */
	public static function reloadRepository(): void
	{
		static::$repository = static::openRepository();
	}

	/**
	 * Verzeichnis-Name bzw. -Pfad für das share-Verzeichnis relativ zum Base-Dir
	 *
	 * Bei Plugin-Models wird das share-Verzeichnis im Plugin zurückgegeben, falls es existiert,
	 * ansonsten das globals share-Verzeichnis.
	 *
	 * @return string
	 * @phan-side-effect-free
	 */
	protected static function repoRelDir(): string
	{
		// Plugins verwenden ihre eigenen `share`-Verzeichnisse, falls vorhanden
		if (strpos(static::class, 'Plugins\\') === 0) {
			$basePath = Utils::getBaseDir() . DIRECTORY_SEPARATOR;
			$class = new ReflectionClass(static::class);
			$filePath = $class->getFileName();

			$pluginDir = implode(DIRECTORY_SEPARATOR, array_slice(explode(DIRECTORY_SEPARATOR, substr($filePath, strlen($basePath)), 3), 0, 2));
			$shareDir = $pluginDir . DIRECTORY_SEPARATOR . 'share';
			if (file_exists($basePath . $shareDir)) {
				return $shareDir;
			}
		}

		// Ansonsten: globales `share`-Verzeichnis
		return 'share';
	}

	/**
	 * Diese Methode stellt die Konfiguration des aktuellen Models in Form eines neuen JsonConfig-Objekts bereit.
	 *
	 * @param User|null $user (ignoriert)
	 * @param Domain|null $domain (ignoriert)
	 * @return RepositoryInterface
	 */
	protected static function openRepository(?User $user=null, ?Domain $domain=null): RepositoryInterface
	{
		$repoFilename = static::$repositoryFilename;

		// Gewährleisten, dass dem Model ein Dateiname als Repository beiliegt
		if (empty($repoFilename)) {
			throw new LogicException('No repository defined for '.static::class);
		}

		return new JsonConfig($repoFilename, false, static::repoRelDir(), null);
	}

	/**
	 * @return int
	 */
	public static final function lastModifiedTimestamp(): int
	{
		return self::getRepository()->getLastModifiedTimestamp();
	}

	/**
	 * @param string $id
	 * @return bool
	 */
	public static function exists(string $id): bool
	{
		return is_array(self::getRepository()->get($id));
	}

	/**
	 * @param string $id
	 * @return static|null
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function find(string $id): ?self
	{
		return self::exists($id) ? new static($id) : null;
	}

	/**
	 * @return static[]
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function all(): array
	{
		/** @var static[] $entities */
		$entities = [];

		foreach (self::getRepository()->toArray() as $id => $data) {
			$entities[$id] = new static($id);
		}

		return $entities;
	}

	/**
	 * @return static[]
	 * @param array $filter MongoDB-Query-Filter
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function filterBy(array $filter): array
	{
		/** @var static[] $entities */
		$entities = [];

		foreach (self::getRepository()->toArrayFiltered($filter) as $id => $data) {
			$entities[$id] = new static($id);
		}

		return $entities;
	}

	/**
	 * Zählt die vorhandenen Instanzen dieses Models.
	 * @return int
	 */
	public static function count(): int
	{
		return count(self::getRepository()->toArray());
	}

	/**
	 * Gibt ein zufälliges Element zurück
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 * @return ?static
	 */
	public static function random(): ?self
	{
		$array = self::getRepository()->toArray();
		if (!$array) {
			return null;
		}
		$id = array_rand($array);
		return new static($id);
	}

	/**
	 * Gibt das erstbeste Element zurück.
	 *
	 * Verwendbar wenn nur ein beliebiges, aber nicht wirklich
	 * zufälliges Element gebraucht wird.
	 *
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 * @return ?static
	 */
	public static function any(): ?self
	{
		$array = self::getRepository()->toArray();
		if (!$array) {
			return null;
		}
		$id = key($array);
		return new static($id);
	}

	/**
	 * Größe des Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes
	 */
	public static function persistedRepositorySize(): int
	{
		return static::getRepository()->getPersistedSize();
	}

	/**
	 * Wrapper für den Zugriff auf `JsonConfig::get(string $identifier, $fallback = null)`.
	 * @param string $property
	 * @param mixed|null $fallback
	 * @return mixed
	 * @see JsonConfig::get()
	 */
	public final function get(string $property, $fallback = null)
	{
		$data =  self::getRepository()->get($this->getId().'.'.$property, $fallback);
		return $data;
	}

	/**
	 * Liest alle Einträge neu von der Festplatte ein.
	 * @throws ConfigNotReadableException
	 */
	public static function reload(): void
	{
		self::getRepository()->reload();
	}

	/**
	 * @return string
	 */
	public final function getId(): string
	{
		return $this->id;
	}

	/**
	 * Debug-Ausgabe für var_dump() erweitern
	 * @return array
	 */
	public function __debugInfo()
	{
		$data = ['id' => $this->getId()];
		foreach ((array)self::getRepository()->get($this->getId(), []) as $key => $value) {
			$data['.'.$key] = $value;
		}
		return $data;
	}
}
