<?php

namespace App\Model\Index;

use App\Model\Model;
use App\Model\Domain;
use App\Component\Config\RepositoryInterface;
use LogicException;

class DomainIndex extends Model
{

	/** @var self|null $requestDomain */
	private static $requestDomain = null;

	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-domain-index';

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = true;

	/** @var bool $userSpecificRepository */
	protected static $userSpecificRepository = true;

	/** @var bool $accountSingleton */
	protected static $accountSingleton = true;

	/** @var bool $useRepositoryLocking */
	protected static $useRepositoryLocking = false;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * @override
	 * @return static[]
	 */
	public static function all(): array
	{
		/** @var static[] $entities */
		$entities = [];
		self::collectRecursive($entities, self::getRepository()->toArray(), []);
		return $entities;
	}

	private static function collectRecursive(array &$array, array $root, array $parts): void
	{
		foreach ($root as $part => $node) {
			if (in_array(substr($part, 0, 1), ['@', '_'])) {
				continue;
			}
			$nodeparts = $parts;
			$nodeparts[] = $part;

			if (isset($node['@id']) || isset($node['_id'])) {
				$array[str_replace('.:', ':', implode('.', array_reverse($nodeparts)))] = new static(implode('.', $nodeparts));
			}
			self::collectRecursive($array, $node, $nodeparts);
		}
	}

	/**
	 * @override
	 * @param string $domain DNS-Name
	 * @return static|null
	 */
	public static function find(string $domain): ?Model
	{
		$parts = self::dnsNameToIndexKeyParts($domain);

		$id_parts = [];
		$node = self::getRepository()->toArray();
		foreach ($parts as $part) {
			if (in_array(substr($part, 0, 1), ['@', '_'])) {
				break;
			}
			if (isset($node[$part])) {
				$node = $node[$part];
				$id_parts[] = $part;
			}
			elseif (($node['@sub'] ?? $node['_sub'] ?? false) or $part === 'www') {
				break;
			}
			else {
				$node = null;
				break;
			}
		}
		if ($node && (isset($node['@id']) || isset($node['_id']))) {
			return new static(implode('.', $id_parts));
		} else {
			return null;
		}
	}

	/**
	 * @param string $domain DNS-Name
	 * @return string Domain-ID oder leerer String, wenn nicht gefunden
	 */
	public static function findDomainId(string $domain): string
	{
		$indexEntry = self::find($domain);
		return (string)(($indexEntry) ? $indexEntry->getDomainId() : '');
	}

	/**
	 * @param string|null $id
	 * @return static
	 * @throws LogicException Ist für DomainIndex nicht verfügbar.
	 * @see self::rebuild()
	 * @override @internal
	 */
	public static function create(?string $id=null): Model
	{
		throw new LogicException('Use DomainIndex::rebuild() to recreate the domain index');
	}

	/**
	 * Domain-Index neuaufbauen
	 * @return void
	 */
	public static function rebuild(): void
	{
		try {
			self::clear(false);
			foreach (Domain::all() as $domain) {
				self::addToIndex($domain->getName(), $domain, false);

				foreach ($domain->getShareDomainNames() as $name) {
					self::addToIndex($name, $domain, true);
				}
			}
			self::flush();
		} catch (\MongoDB\Driver\Exception\BulkWriteException $e) {
			$result = $e->getWriteResult();
			$errors = $result->getWriteErrors();
			$wcError = $result->getWriteConcernError();
			if ($wcError) {
				$errors[] = $wcError;
			}

			if ($errors) {
				if (in_array($errors[0]->getCode(), [112, 11000])) {
					// WriteConflict und DuplicateKey ignorieren
					// ein anderer Request hat anscheinend parallel den Index schon neu aufgebaut.
					return;
				}
			}
			throw $e;
		}
	}

	protected static function addToIndex(string $dnsName, Domain $domain, bool $catchSubdomains): void
	{
		if (function_exists('idn_to_ascii')) {
			$dnsName2 = idn_to_ascii($dnsName);
			if ($dnsName2 === $dnsName) {
				$dnsName2 = idn_to_utf8($dnsName);
			}
			if ($dnsName2 !== $dnsName) {
				self::addIndexEntry($dnsName2, $domain, $catchSubdomains);
			}
		}
		self::addIndexEntry($dnsName, $domain, $catchSubdomains);
	}

	private static function addIndexEntry(string $dnsName, Domain $domain, bool $catchSubdomains): void
	{
		$key = self::dnsNameToIndexKey($dnsName);
		if (!$key) {
			return;
		}
		$item = new self($key);
		$item->setDomain($domain);
		$item->setCatchingSubdomains($catchSubdomains);
	}

	/**
	 * @return string
	 */
	private function getDomainId(): string
	{
		// Bezeichner gewechselt, da "_id" in MongoDB reserviert ist
		// @see https://docs.mongodb.com/manual/core/document/#field-names
		$jsonLegacyField = $this->get('_id');
		return (string)$this->get('@id', $jsonLegacyField);
	}

	/**
	 * @param string $id
	 * @return $this
	 */
	private function setDomainId(string $id): self
	{
		$this->set('@id', $id);
		return $this;
	}

	/**
	 * @return Domain|null
	 */
	public function getDomain(): ?Domain
	{
		return Domain::find($this->getDomainId());
	}

	/**
	 * @param Domain $domain
	 * @return $this
	 */
	private function setDomain(Domain $domain): self
	{
		$this->setDomainId($domain->getId());
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isCatchingSubdomains(): bool
	{
		// Der Konsistenz halber…
		$jsonLegacyField = $this->get('_sub');
		return (bool)$this->get('@sub', $jsonLegacyField);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setCatchingSubdomains(bool $value): self
	{
		$this->set('@sub', $value);
		return $this;
	}

	private static function dnsNameToIndexKeyParts(string $domain): array
	{
		return array_reverse(explode('.', trim(mb_strtolower(str_replace(':', '.:', $domain), 'utf-8'), ". \t\n\r\0")));
	}

	private static function dnsNameToIndexKey(string $domain): string
	{
		return implode('.', self::dnsNameToIndexKeyParts($domain));
	}
}
