<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use Twig\Markup;

class Language extends ImmutableModel
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-languages.json';

	/**
	 * @return static[]
	 */
	public static function activeLanguages(): array
	{
		return array_filter(self::all(), static function (self $language) {
			return $language->isActiveByDefault();
		});
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @return string
	 */
	public function getEnglishName(): string
	{
		return (string)$this->get('name_en');
	}

	/**
	 * @return bool
	 */
	public function isActiveByDefault(): bool
	{
		return (bool)$this->get('active');
	}

	/**
	 * @return string
	 */
	public function getGermanName(): string
	{
		return (string)$this->get('name_de');
	}

	/**
	 * @return string
	 */
	public function getDisplayName(string $user_language): string
	{
		if (strpos($user_language, 'de') === 0) {
			$loc_name = $this->getGermanName();
		} else {
			$loc_name = $this->getEnglishName();
		}

		$name = $this->getName();
		if ($name !== $loc_name and strpos($loc_name, '(') === false) {
			return $loc_name . ' ('.$name.')';
		} else {
			return $loc_name;
		}
	}

	/**
	 * @return Markup
	 */
	public function getDisplayNameHtml(string $user_language): Markup
	{
		if (strpos($user_language, 'de') === 0) {
			$loc_name = '<span lang="de">'.htmlspecialchars($this->getGermanName()).'</span>';
		} else {
			$loc_name = '<span lang="en">'.htmlspecialchars($this->getEnglishName()).'</span>';
		}

		$name = $this->getName();
		if ($name !== $loc_name and strpos($loc_name, '(') === false) {
			$loc_name .= ' (<span lang="'.$this->getShortCode().'">'.htmlspecialchars($name).'</span>)';
		}
		return new Markup($loc_name, 'utf-8');
	}

	/**
	 * @return string
	 */
	public function getCode(): string
	{
		return (string)$this->getId();
	}

	/**
	 * @return string
	 */
	public function getShortCode(): string
	{
		return explode('_', $this->getCode(), 2)[0];
	}

	/**
	 * @return string
	 */
	public function getIsoCode(): string
	{
		return str_replace('_', '-', $this->getId());
	}

	/**
	 * Bestimmt ein Land zu dieser Sprache
	 */
	public function getAssumedCountry(): ?Country
	{
		$countryCode = (string)$this->get('assumed_country');
		$parts = explode('_', $this->getCode());
		if (!$countryCode) {
			if (count($parts) > 1) {
				$countryCode = $parts[1];
			}
		}
		// Wenn ein Land im Code steht, nimm das
		$country = null;
		if ($countryCode) {
			$country = Country::find($countryCode);
		}
		// Ansonsten rate anhand der Sprache
		if (!$country) {
			$country = Country::find(strtoupper($parts[0]));
		}
		return $country;
	}
}
