<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Component\LocaleService;

class LegalText extends Model
{
	const DISPLAY_MODE_DIRECT = 'direct';
	const DISPLAY_MODE_LINK = 'link';

	const ALLOWED_DISPLAY_MODE_VALUES = [
		self::DISPLAY_MODE_DIRECT,
		self::DISPLAY_MODE_LINK,
	];

	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-legal-texts';

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = true;

	/** @var bool $wantsMongoDbMtimeIndex */
	protected static $wantsMongoDbMtimeIndex = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	public static function byLocale(LocaleBase $locale): ?self
	{
		return array_reduce(self::all(), function ($carry, $legalText) use ($locale) {
			/**
			 * @var LegalText $legalText
			 * @var LocaleBase|null $currentLocale
			 */
			$currentLocale = $legalText->getLocale();

			if ($currentLocale === null) {
				return $carry;
			}

			// Fallback for deprecated locales identified by ID instead of their name
			if ($carry && $currentLocale->getId() == $locale->getId()) {
				return $legalText;
			}

			return $carry ?? ($currentLocale->getName() == $locale->getName() ? $legalText : $carry);
		}, null);
	}

	/**
	 * @return self[]
	 */
	public static function all(): array
	{
		$legalTexts =  parent::all();

		$locales = LocaleService::all();
		/** @var LocaleBase[] $locales */
		$locales = array_combine(
			array_map(function (LocaleBase $locale) { return $locale->getName(); }, $locales),
			$locales
		);

		// Keep all objects in mind to remove duplicate entries for a single language
		$allObjectIds = array_values(array_map(function (self $item) { return $item->getId(); }, $legalTexts));

		// Guarantee distinct instances for every locale
		$legalTexts = array_reduce($legalTexts, function (array $legalTexts, self $legalText) use ($locales) {
			$currentLocale = $legalText->getLocale();
			$localeName = $currentLocale ? $currentLocale->getName() : '';
			if ($currentLocale and $locale = $locales[$localeName] ?? null) {
				if (isset($legalTexts[$localeName])) {
					// Fallback for deprecated locales identified by ID instead of their name
					if ($currentLocale->getId() == $locale->getId()) {
						$legalTexts[$localeName] = $legalText;
					}
				}
				elseif ($currentLocale->getName() == $locale->getName()) {
					$legalTexts[$localeName] = $legalText;
				}
			}

			return $legalTexts;
		}, []);

		$removableObjects = array_values(array_diff($allObjectIds, array_map(function (self $item) {
			return $item->getId();
		}, $legalTexts)));

		// Remove redundant, not necessarily identical objects
		if ($removableObjects && count($allObjectIds) > count($removableObjects)) {
			foreach ($removableObjects as $objectId) {
				self::delete($objectId, false);
			}
			self::flush();
		}

		return array_combine(
			array_map(function (self $legalText) { return $legalText->getId(); }, $legalTexts),
			$legalTexts
		);
	}

	/**
	 * @return self[]
	 */
	public static function havingLocaleEnabled(): array
	{
		$localeNames = array_map(function (LocaleBase $locale) {
			return $locale->getName();
		}, LocaleService::enabledLocales());

		return array_filter(self::all(), function (self $legalText) use ($localeNames) {
			$locale = $legalText->getLocale();
			return $locale && in_array($locale->getName(), $localeNames);
		});
	}

	/**
	 * @return LocaleBase|null
	 */
	public function getLocale(): ?LocaleBase
	{
		$localeName = $this->get('locale');
		if (!$localeName) {
			return Locale::byId((string)$this->get('localeId'));
		}
		$result = Locale::find($localeName);
		if (!$result) {
			$result = UserLocale::find($localeName);
		}
		return $result;
	}

	/**
	 * @param LocaleBase $locale
	 * @return $this
	 */
	public function setLocale(LocaleBase $locale): self
	{
		$this->set('locale', $locale->getName());
		return $this;
	}

	/**
	 * @return string
	 */
	public function getImprint(): string
	{
		return (string)$this->get('imprint');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setImprint(string $text): self
	{
		$this->set('imprint', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getImprintScheme(): string
	{
		return (string)$this->get('imprintScheme') ?? "link";
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setImprintScheme(string $text): self
	{
		$this->set('imprintScheme', $text);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowWidgetOnImprintPage(): bool
	{
		return (bool)$this->get('showWidgetOnImprintPage');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowWidgetOnImprintPage(bool $state): self
	{
		$this->set('showWidgetOnImprintPage', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function istargetBlank(): bool
	{
		return (bool)$this->get('targetBlank') ?? false;
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function settargetBlank(bool $state): self
	{
		$this->set('targetBlank', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function istargetBlankPrivacy(): bool
	{
		return (bool)$this->get('targetBlankPrivacy') ?? false;
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function settargetBlankPrivacy(bool $state): self
	{
		$this->set('targetBlankPrivacy', $state);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPrivacyScheme(): string
	{
		return (string)$this->get('privacyScheme');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setPrivacyScheme(string $text): self
	{
		$this->set('privacyScheme', $text);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowWidgetOnPrivacyPage(): bool
	{
		return (bool)$this->get('showWidgetOnPrivacyPage');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowWidgetOnPrivacyPage(bool $state): self
	{
		$this->set('showWidgetOnPrivacyPage', $state);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getImprintUrl(): string
	{
		return (string)$this->get('imprintUrl');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setImprintUrl(string $text): self
	{
		$this->set('imprintUrl', $text);
		return $this;
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setPrivacyUrl(string $text): self
	{
		$this->set('privacyUrl', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPrivacyUrl(): string
	{
		return (string)$this->get('privacyUrl');
	}

	/**
	 * @return string
	 */
	public function getPrivacyPolicy(): string
	{
		return (string)$this->get('privacyPolicy');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setPrivacyPolicy(string $text): self
	{
		$this->set('privacyPolicy', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getAccessibilityScheme(): string
	{
		return (string)$this->get('accessibilityScheme');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setAccessibilityScheme(string $text): self
	{
		$this->set('accessibilityScheme', $text);
		return $this;
	}


	/**
	 * @param string $text
	 * @return $this
	 */
	public function setAccessibilityUrl(string $text): self
	{
		$this->set('accessibilityUrl', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getAccessibilityUrl(): string
	{
		return (string)$this->get('accessibilityUrl');
	}


	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowAccessibilityInFrontendWidget(bool $state): self
	{
		$this->set('showAccessibilityInFrontendWidget', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowAccessibilityInFrontendWidget(): bool
	{
		return $this->get('showAccessibilityInFrontendWidget', (bool)$this->get('setEnableFrontendWidgetAccessibility'));
	}

	/**
	 * @return string
	 */
	public function getAccessibility(): string
	{
		return (string)$this->get('accessibility');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setAccessibility(string $text): self
	{
		$this->set('accessibility', $text);
		return $this;
	}

	public static function upgradeFromLocaleIds(): void
	{
		$all = self::all();

		if (empty($all)) {
			return; // no texts -> nothing to do
		}

		$first = reset($all);
		if (!($first->get('locale'))) {
			foreach($all as $item) {
				$locale = $item->getLocale();
				if ($locale) {
					$item->setLocale($locale);
				}
			}
			self::flush();
		}
	}
}
