<?php

namespace App\Model;

use App\Exception\LocaleAlreadyExistsException;
use App\Component\Config\RepositoryInterface;
use LogicException;

class Locale extends LocaleBase
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * @param string $id
	 * @return self|null
	 */
	public static function byId(string $id): ?self
	{
		return parent::find($id);
	}

	/**
	 * @override
	 * @return Language|null
	 */
	public function getLanguage(): ?Language
	{
		$id = $this->get('name');
		if (strpos($id, '_') !== false) {
			return Language::find($id) ?? Language::find(explode('_', $id, 2)[0]);
		} else {
			return Language::find($id);
		}
	}

	/**
	 * @param string $localeName
	 * @return bool
	 */
	public static function exists(string $localeName): bool
	{
		return count(static::filterBy(['name' => $localeName])) !== 0;
	}

	/**
	 * @param string $localeName
	 * @return static|null
	 * @phan-return static|null
	 */
	public static function find(string $localeName): ?Model
	{
		return current(static::filterBy(['name' => $localeName])) ?: null;
	}

	/**
	 * @param ?string $localeName
	 * @return static
	 * @throws LocaleAlreadyExistsException
	 * @throws LogicException
	 * @override
	 * @phan-assert string $localeName
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function create(?string $localeName = null): Model
	{
		if ($localeName === null) {
			throw new LogicException('Locale name may not be null.');
		}
		elseif (!Language::exists($localeName)) {
			throw new LogicException('Invalid locale name, language not found.');
		}

		self::beginWrite();

		if (self::exists($localeName)) {
			self::reload();
			throw new LocaleAlreadyExistsException($localeName);
		}

		/** @var self $locale */
		$locale = parent::create();
		assert($locale instanceof Locale);
		$locale->setName($localeName);
		self::flush();

		if (LegalText::byLocale($locale) == null) {
			LegalText::create()->setLocale($locale)->save();
		}

		self::beginWrite();
		return $locale;
	}

	/**
	 * @override
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setName(string $name): self
	{
		$this->set('name', $name);
		return $this;
	}
}
