<?php

namespace App\Model;

abstract class LocaleBase extends Model
{
	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-locales';

	/** @var bool $wantsMongoDbMtimeIndex */
	protected static $wantsMongoDbMtimeIndex = true;

	/**
	 * @return static[]
	 */
	public static function enabledLocales(): array
	{
		return static::filterBy(['$or'=> [
			['enabled' => true],
			['enabled' => ['$exists'=>false]],
		]]);
	}

	/**
	 * @param static[] $locales
	 * @return static[]
	 */
	public static function sortByEnabledAndName(array $locales): array
	{
		// ORDER BY enabled, name
		uasort($locales, function (self $a, self $b) {
			if ($a->isEnabled() xor $b->isEnabled()) {
				return $a->isEnabled() ? -1 : 1;
			}
			else {
				return strnatcasecmp($a->getName(), $b->getName());
			}
		});

		return $locales;
	}

	/**
	 * @return Language|null
	 */
	public abstract function getLanguage(): ?Language;

	/**
	 * @return string
	 */
	public abstract function getName(): string;

	/**
	 * @return bool
	 */
	public function isEnabled(): bool
	{
		// Fallback ist true, um alte Versionen zu unterstützen
		return (bool)$this->get('enabled', true);
	}

	/**
	 * @param bool $enabled
	 * @return $this
	 */
	public function setEnabled(bool $enabled): self
	{
		$this->set('enabled', $enabled);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getWidgetTitle(): string
	{
		return (string)$this->get('widget_title');
	}

	/**
	 * @param string $title
	 * @return $this
	 */
	public function setWidgetTitle(string $title): self
	{
		$this->set('widget_title', $title);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getWidgetIntroduction(): string
	{
		return (string)$this->get('widget_introduction');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setWidgetIntroduction(string $text): self
	{
		$this->set('widget_introduction', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getWidgetConsentButtonText(): string
	{
		return (string)$this->get('widget_consent_button_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setWidgetConsentButtonText(string $text): self
	{
		$this->set('widget_consent_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getWidgetDeclineButtonText(): string
	{
		return (string)$this->get('widget_decline_button_text', 'Ablehnen');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setWidgetDeclineButtonText(string $text): self
	{
		$this->set('widget_decline_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getWidgetSettingsButtonText(): string
	{
		return (string)$this->get('widget_settings_button_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setWidgetSettingsButtonText(string $text): self
	{
		$this->set('widget_settings_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCategoriesAcceptAllButtonText(): string
	{
		$language = $this->getLanguage();
		$fallback = $language && $language->getCode() == 'de_DE' ? 'Alle akzeptieren' : 'Accept all';
		return (string)$this->get('categories_accept_all_button_text', $fallback);
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCategoriesAcceptAllButtonText(string $text): self
	{
		$this->set('categories_accept_all_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCheckAllButtonText(): string
	{
		$language = $this->getLanguage();
		$fallback = $language && $language->getCode() == 'de_DE' ? 'Alle auswählen' : 'Check all';
		return (string)$this->get('check_all_button_text', $fallback);
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCheckAllButtonText(string $text): self
	{
		$this->set('check_all_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getUncheckAllButtonText(): string
	{
		$language = $this->getLanguage();
		$fallback = $language && $language->getCode() == 'de_DE' ? 'Alle abwählen' : 'Uncheck all';
		return (string)$this->get('uncheck_all_button_text', $fallback);
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setUncheckAllButtonText(string $text): self
	{
		$this->set('uncheck_all_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getStatusActiveText(): string
	{
		$language = $this->getLanguage();
		$fallback = $language && $language->getCode() == 'de_DE' ? 'Aktiv' : 'Active';
		return (string)$this->get('status_active_text', $fallback);
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setStatusActiveText(string $text): self
	{
		$this->set('status_active_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getStatusInactiveText(): string
	{
		$language = $this->getLanguage();
		$fallback = $language && $language->getCode() == 'de_DE' ? 'Inaktiv' : 'Inactive';
		return (string)$this->get('status_inactive_text', $fallback);
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setStatusInactiveText(string $text): self
	{
		$this->set('status_inactive_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getStatusPartialText(): string
	{
		$language = $this->getLanguage();
		$fallback = $language && $language->getCode() == 'de_DE' ? 'Teilweise' : 'Partial';
		return (string)$this->get('status_partial_text', $fallback);
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setStatusPartialText(string $text): self
	{
		$this->set('status_partial_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getControlPanelTitle(): string
	{
		return (string)$this->get('control_panel_title');
	}

	/**
	 * @param string $title
	 * @return $this
	 */
	public function setControlPanelTitle(string $title): self
	{
		$this->set('control_panel_title', $title);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getControlPanelCancelButtonText(): string
	{
		return (string)$this->get('control_panel_cancel_button_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setControlPanelCancelButtonText(string $text): self
	{
		$this->set('control_panel_cancel_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getControlPanelSaveButtonText(): string
	{
		return (string)$this->get('control_panel_save_button_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setControlPanelSaveButtonText(string $text): self
	{
		$this->set('control_panel_save_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getDetailsTitle(): string
	{
		return (string)$this->get('details_title');
	}

	/**
	 * @param string $title
	 * @return $this
	 */
	public function setDetailsTitle(string $title): self
	{
		$this->set('details_title', $title);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getDetailsCloseButtonText(): string
	{
		return (string)$this->get('details_close_button_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setDetailsCloseButtonText(string $text): self
	{
		$this->set('details_close_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getImprintText(): string
	{
		return (string)$this->get('imprint_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setImprintText(string $text): self
	{
		$this->set('imprint_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPrivacyPolicyText(): string
	{
		return (string)$this->get('privacy_policy_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setAccessibilityText(string $text): self
	{
		$this->set('accessibility_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getAccessibilityText(): string
	{
		return (string)$this->get('accessibility_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setPrivacyPolicyText(string $text): self
	{
		$this->set('privacy_policy_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCookieVendorText(): string
	{
		return (string)$this->get('cookie_vendor_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCookieVendorText(string $text): self
	{
		$this->set('cookie_vendor_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCookieDescriptionText(): string
	{
		return (string)$this->get('cookie_description_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCookieDescriptionText(string $text): self
	{
		$this->set('cookie_description_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCookiePrivacyPolicyUrlText(): string
	{
		return (string)$this->get('cookie_privacy_policy_url_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCookiePrivacyPolicyUrlText(string $text): self
	{
		$this->set('cookie_privacy_policy_url_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCookieLifetimeText(): string
	{
		return (string)$this->get('cookie_lifetime_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCookieLifetimeText(string $text): self
	{
		$this->set('cookie_lifetime_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCookieCollectedDataInfoText(): string
	{
		return (string)$this->get('cookie_collected_data_info_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCookieCollectedDataInfoText(string $text): self
	{
		$this->set('cookie_collected_data_info_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCookiePurposeOfDataCollectionText(): string
	{
		return (string)$this->get('cookie_purpose_of_data_collection_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCookiePurposeOfDataCollectionText(string $text): self
	{
		$this->set('cookie_purpose_of_data_collection_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCookieLegalBasisText(): string
	{
		return (string)$this->get('cookie_legal_basis_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCookieLegalBasisText(string $text): self
	{
		$this->set('cookie_legal_basis_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCookiePlaceOfProcessingText(): string
	{
		return (string)$this->get('cookie_place_of_processing_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setCookiePlaceOfProcessingText(string $text): self
	{
		$this->set('cookie_place_of_processing_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getBlockedContentTitle(): string
	{
		return (string)$this->get('blocked_content_title');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setBlockedContentTitle(string $text): self
	{
		$this->set('blocked_content_title', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getBlockedContentText(): string
	{
		return (string)$this->get('blocked_content_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setBlockedContentText(string $text): self
	{
		$this->set('blocked_content_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getBlockedContentButtonText(): string
	{
		return (string)$this->get('blocked_content_button_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setBlockedContentButtonText(string $text): self
	{
		$this->set('blocked_content_button_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getBlockedContentShowExternalContent(): string
	{
		return (string)$this->get('blocked_content_show_external_content');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setBlockedContentShowExternalContent(string $text): self
	{
		$this->set('blocked_content_show_external_content', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getBlockedContentAllowEmbedding(): string
	{
		return (string)$this->get('blocked_content_allow_embedding');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setBlockedContentAllowEmbedding(string $text): self
	{
		$this->set('blocked_content_allow_embedding', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getBlockedContentProviderDetails(): string
	{
		return (string)$this->get('blocked_content_provider_details');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setBlockedContentProviderDetails(string $text): self
	{
		$this->set('blocked_content_provider_details', $text);
		return $this;
	}

	/**
	 * @param Purpose $purpose
	 * @return string
	 */
	public function getPurposeName(Purpose $purpose): string
	{
		$result = (string)$this->get('purpose_name.'.$purpose->getId());
		if ($result === '') {
			$result = (string)$purpose->getTranslatedName($this);
		}

		return $result;
	}

	/**
	 * @param Purpose $purpose
	 * @param string $text
	 * @return $this
	 */
	public function setPurposeName(Purpose $purpose, string $text): self
	{
		$this->set('purpose_name.'.$purpose->getId(), $text);
		return $this;
	}

	/**
	 * @param Purpose $purpose
	 * @return string
	 */
	public function getPurposeDescription(Purpose $purpose): string
	{
		return (string)$this->get('purpose_description.'.$purpose->getId()) ?: $purpose->getTranslatedDescription($this);
	}

	/**
	 * @param Purpose $purpose
	 * @param string $text
	 * @return $this
	 */
	public function setPurposeDescription(Purpose $purpose, string $text): self
	{
		$this->set('purpose_description.'.$purpose->getId(), $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getChangeConsentLabel(): string
	{
		return (string)$this->get('change_consent_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setChangeConsentLabel(string $text): self
	{
		$this->set('change_consent_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getAssetNameLabel(): string
	{
		return (string)$this->get('asset_name_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setAssetNameLabel(string $text): self
	{
		$this->set('asset_name_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getAssetLifetimeLabel(): string
	{
		return (string)$this->get('asset_lifetime_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setAssetLifetimeLabel(string $text): self
	{
		$this->set('asset_lifetime_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getAssetDescriptionLabel(): string
	{
		return (string)$this->get('asset_description_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setAssetDescriptionLabel(string $text): self
	{
		$this->set('asset_description_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentStateLabel(): string
	{
		return (string)$this->get('consent_state_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentStateLabel(string $text): self
	{
		$this->set('consent_state_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentStateNone(): string
	{
		return (string)$this->get('consent_state_none');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentStateNone(string $text): self
	{
		$this->set('consent_state_none', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentStateCustom(): string
	{
		return (string)$this->get('consent_state_custom');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentStateCustom(string $text): self
	{
		$this->set('consent_state_custom', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentStateAccepted(): string
	{
		return (string)$this->get('consent_state_accepted');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentStateAccepted(string $text): self
	{
		$this->set('consent_state_accepted', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentStateDeclined(): string
	{
		return (string)$this->get('consent_state_declined');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentStateDeclined(string $text): self
	{
		$this->set('consent_state_declined', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentIdLabel(): string
	{
		return (string)$this->get('consent_id_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentIdLabel(string $text): self
	{
		$this->set('consent_id_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getConsentDateLabel(): string
	{
		return (string)$this->get('consent_date_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentDateLabel(string $text): self
	{
		$this->set('consent_date_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfVendorGrantedPermissionsText(): string
	{
		return (string)$this->get('tcf_vendor_granted_permissions_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfVendorGrantedPermissionsText(string $text): self
	{
		$this->set('tcf_vendor_granted_permissions_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfPurposesLabel(): string
	{
		return (string)$this->get('tcf_purposes_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfPurposesLabel(string $text): self
	{
		$this->set('tcf_purposes_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfSpecialPurposesLabel(): string
	{
		return (string)$this->get('tcf_special_purposes_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfSpecialPurposesLabel(string $text): self
	{
		$this->set('tcf_special_purposes_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfFeaturesLabel(): string
	{
		return (string)$this->get('tcf_features_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfFeaturesLabel(string $text): self
	{
		$this->set('tcf_features_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfSpecialFeaturesLabel(): string
	{
		return (string)$this->get('tcf_special_features_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfSpecialFeaturesLabel(string $text): self
	{
		$this->set('tcf_special_features_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfConsentLabel(): string
	{
		return (string)$this->get('tcf_consent_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfConsentLabel(string $text): self
	{
		$this->set('tcf_consent_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfLegitimateInterestLabel(): string
	{
		return (string)$this->get('tcf_legitimate_interest_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfLegitimateInterestLabel(string $text): self
	{
		$this->set('tcf_legitimate_interest_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfOptedInRemark(): string
	{
		return (string)$this->get('tcf_opted_in_remark');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfOptedInRemark(string $text): self
	{
		$this->set('tcf_opted_in_remark', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfThirdPartiesLabel(): string
	{
		return (string)$this->get('tcf_third_parties_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfThirdPartiesLabel(string $text): self
	{
		$this->set('tcf_third_parties_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getNonTcfPurposesLabel(): string
	{
		return (string)$this->get('non_tcf_purposes_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setNonTcfPurposesLabel(string $text): self
	{
		$this->set('non_tcf_purposes_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getNonTcfVendorsLabel(): string
	{
		return (string)$this->get('non_tcf_vendors_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setNonTcfVendorsLabel(string $text): self
	{
		$this->set('non_tcf_vendors_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getMoreInformationLabel(): string
	{
		return (string)$this->get('more_information_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setMoreInformationLabel(string $text): self
	{
		$this->set('more_information_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfIllustrationsLabel(): string
	{
		return (string)$this->get('tcf_illustrations_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfIllustrationsLabel(string $text): self
	{
		$this->set('tcf_illustrations_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfVendorUsesNonCookieStorageText(): string
	{
		return (string)$this->get('tcf_vendor_uses_non_cookie_storage_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfVendorUsesNonCookieStorageText(string $text): self
	{
		$this->set('tcf_vendor_uses_non_cookie_storage_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfSpecialFeaturesHandledGloballyText(): string
	{
		return (string)$this->get('tcf_special_features_handled_globally_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfSpecialFeaturesHandledGloballyText(string $text): self
	{
		$this->set('tcf_special_features_handled_globally_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfWidgetIntroduction(): string
	{
		return (string)$this->get('tcf_widget_introduction');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfWidgetIntroduction(string $text): self
	{
		$this->set('tcf_widget_introduction', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfStacksListText(): string
	{
		return (string)$this->get('tcf_stacks_list_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfStacksListText(string $text): self
	{
		$this->set('tcf_stacks_list_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfLegitimateInterestText(): string
	{
		return (string)$this->get('tcf_legitimate_interest_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfLegitimateInterestText(string $text): self
	{
		$this->set('tcf_legitimate_interest_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfCookieRefreshText(): string
	{
		return (string)$this->get('tcf_cookie_refresh_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfCookieRefreshText(string $text): self
	{
		$this->set('tcf_cookie_refresh_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfExtendedDisclosureText(): string
	{
		return (string)$this->get('tcf_extended_disclosure_text');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfExtendedDisclosureText(string $text): self
	{
		$this->set('tcf_extended_disclosure_text', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfAssetTypeLabel(): string
	{
		return (string)$this->get('tcf_asset_type_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfAssetTypeLabel(string $text): self
	{
		$this->set('tcf_asset_type_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getTcfAssetPurposesLabel(): string
	{
		return (string)$this->get('tcf_asset_purposes_label');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfAssetPurposesLabel(string $text): self
	{
		$this->set('tcf_asset_purposes_label', $text);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getMorePurposeDetailsButtonText(): string
	{
		return (string)$this->get('more_purpose_details_button_text', '?');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setMorePurposeDetailsButtonText(string $text): self
	{
		$this->set('more_purpose_details_button_text', $text);
		return $this;
	}

	public function getMorePurposeDetailsButtonTitle(): string
	{
		return (string)$this->get('more_purpose_details_button_title');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setMorePurposeDetailsButtonTitle(string $text): self
	{
		$this->set('more_purpose_details_button_title', $text);
		return $this;
	}

	public function getTcfDataRetentionText(): string
	{
		return (string)$this->get('tcf_data_retention_text', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfDataRetentionText(string $text): self
	{
		$this->set('tcf_data_retention_text', $text);
		return $this;
	}

	public function getTcfDataRetentionSessionText(): string
	{
		return (string)$this->get('tcf_data_retention_session_text', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfDataRetentionSessionText(string $text): self
	{
		$this->set('tcf_data_retention_session_text', $text);
		return $this;
	}

	public function getTcfDataCategoriesLabel(): string
	{
		return (string)$this->get('tcf_data_categories_label', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfDataCategoriesLabel(string $text): self
	{
		$this->set('tcf_data_categories_label', $text);
		return $this;
	}

	public function getTcfPurposeVendorListLabel(): string
	{
		return (string)$this->get('tcf_purpose_vendor_list_label', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfPurposeVendorListLabel(string $text): self
	{
		$this->set('tcf_purpose_vendor_list_label', $text);
		return $this;
	}

	public function getTcfConsentStorageText(): string
	{
		return (string)$this->get('tcf_consent_storage_text', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setTcfConsentStorageText(string $text): self
	{
		$this->set('tcf_consent_storage_text', $text);
		return $this;
	}

	public function getWithdrawAllText(): string
	{
		return (string)$this->get('withdraw_all_text', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setWithdrawAllText(string $text): self
	{
		$this->set('withdraw_all_text', $text);
		return $this;
	}

	public function getConsentSharingText(): string
	{
		return (string)$this->get('consent_sharing_text', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentSharingText(string $text): self
	{
		$this->set('consent_sharing_text', $text);
		return $this;
	}

	public function getConsentNoSharingText(): string
	{
		return (string)$this->get('consent_no_sharing_text', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setConsentNoSharingText(string $text): self
	{
		$this->set('consent_no_sharing_text', $text);
		return $this;
	}

	public function getLanguageChooserUsage(): string
	{
		return (string)$this->get('language_chooser_usage', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setLanguageChooserUsage(string $text): self
	{
		$this->set('language_chooser_usage', $text);
		return $this;
	}

	public function getEmbeddingPrivacyPolicyDesc(): string
	{
		return (string)$this->get('embedding_privacy_policy_desc', '');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setEmbeddingPrivacyPolicyDesc(string $text): self
	{
		$this->set('embedding_privacy_policy_desc', $text);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getUpdateParameterBlacklist(): array
	{
		return (array)$this->get('updateParameterBlacklist');
	}

	/**
	 * @param string[] $parameters
	 * @return $this
	 */
	public function setUpdateParameterBlacklist(array $parameters): self
	{
		$parameters = array_values(array_map('strval', $parameters));
		$this->set('updateParameterBlacklist', $parameters);
		return $this;
	}

	/**
	 * @param string $parameter
	 * @return bool
	 */
	public function isUpdateParameterAllowed(string $parameter): bool
	{
		return !in_array($parameter, $this->getUpdateParameterBlacklist());
	}

	/**
	 * @param string $parameter
	 * @return bool
	 */
	public function isUpdateParameterLocked(string $parameter): bool
	{
		return !$this->isUpdateParameterAllowed($parameter);
	}

}
