<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Exception\ConfigNotWritableException;
use App\Exception\UniqueConstraintException;
use LogicException;

class LocaleOverride extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-locale-overrides';

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = true;

	/** @var bool $useRepositorySkeleton */
	protected static $useRepositorySkeleton = false;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * @param string|null $fieldName
	 * @return static
	 * @throws LogicException
	 * @throws UniqueConstraintException
	 * @override
	 * @phan-assert string $fieldName
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function create(?string $fieldName=null): Model
	{
		if ($fieldName === null) {
			throw new LogicException('Field name may not be null.');
		}

		self::beginWrite();

		if (self::exists($fieldName)) {
			throw new UniqueConstraintException("Object for field $fieldName already exists.");
		}

		return new static(static::getRepository()->escapeKey($fieldName));
	}

	/**
	 * @param string|null $fieldName
	 * @return static
	 * @throws LogicException
	 * @override
	 * @phan-assert string $fieldName
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function createNonPersistent(?string $fieldName=null): Model
	{
		if ($fieldName === null) {
			throw new LogicException('Field name may not be null.');
		}

		self::getRepository()->makeVolatile();

		return new static(static::getRepository()->escapeKey($fieldName));
	}

	/**
	 * Diese Methode löscht den durch das aktuelle Objekt referenzierten Datensatz aus der Konfiguration.
	 * @param string $fieldName
	 * @param bool $autocommit
	 * @throws ConfigNotWritableException
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function delete(string $fieldName, bool $autocommit=true): void
	{
		parent::delete(static::getRepository()->escapeKey($fieldName), $autocommit);
	}

	/**
	 * @param string $fieldName
	 * @return bool
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function exists(string $fieldName): bool
	{
		return parent::exists(static::getRepository()->escapeKey($fieldName));
	}

	/**
	 * @param string $fieldName
	 * @return static|null
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function find(string $fieldName): ?Model
	{
		return self::exists($fieldName) ? new static(static::getRepository()->escapeKey($fieldName)) : null;
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return static::getRepository()->unescapeKey($this->getId());
	}

	/**
	 * @return string[]
	 */
	public function getTranslations(): array
	{
		return (array)$this->get('translations');
	}

	/**
	 * @param string[] $translations
	 * @return $this
	 */
	private function setTranslations(array $translations): self
	{
		// Strip invalid language codes and enforce string type
		$translations = array_map('strval', array_intersect_key($translations, Language::all()));
		$this->set('translations', $translations);
		return $this;
	}

	/**
	 * @param Language $language
	 * @return string|null
	 */
	public function getTranslation(Language $language): ?string
	{
		return $this->getTranslations()[$language->getCode()] ?? null;
	}

	/**
	 * @param Language $language
	 * @param string $text
	 * @return $this
	 */
	public function setTranslation(Language $language, string $text): self
	{
		$translations = $this->getTranslations();
		$translations[$language->getCode()] = $text;
		uksort($translations, 'strnatcasecmp');
		return $this->setTranslations($translations);
	}

	/**
	 * @return Language[]
	 */
	public function getLanguages(): array
	{
		return array_intersect_key(Language::all(), $this->getTranslations());
	}
}
