<?php

namespace App\Model;

use App\Component\LogStream\RepositoryInterface;
use App\Component\LogStream\Store\JsonStore;
use App\Component\LogStream\Store\MongoDb;
use App\Component\LogStream\Store\RedisStore;
use MongoDB\BSON\ObjectId;
use App\Utils;
use DateTime;
use Iterator;

/** Login-Versuche der letzten 5 Minuten */
class LoginAttemptsLog
{
	const TIMEWINDOW_MINUTES = 5;
	private static $file = 'log/cm-login-attempts';
	private static $collection = 'cm-login-attempts';

	private $filter;
	private $kind;

	/** @var RepositoryInterface $repository */
	private $repository;

	/**
	 * @param string $kind 'login'/'recovery'
	 */
	private function __construct(string $kind)
	{
		$this->kind = $kind;
		$repositoryStore = Utils::getLoginAttemptsLogRepositoryStore();
		$start = (new Datetime('-'.self::TIMEWINDOW_MINUTES.' minutes'))->getTimestamp();
		switch ($repositoryStore) {
			case Utils::REPOSITORY_STORE_MONGODB:
				$repository = new MongoDb(null, null, self::$collection);
				$this->filter = [
					'_id' => [
						'$gte' => new ObjectId(substr(str_pad(dechex($start), 8, '0', STR_PAD_LEFT), -8, 8).'0000000000000000'),
					],
					'kind' => $kind,
				];
				break;
			case Utils::REPOSITORY_STORE_REDIS:
				// Store kann gelöscht werden nach 3 * TIMEWINDOW (15 Minuten) ohne Schreibzugriffe
				$fullExpiry = 3 * 60 * self::TIMEWINDOW_MINUTES;
				$repository = new RedisStore(null, null, self::$collection.':'.$kind, 'ts', $fullExpiry);
				$this->filter = [
					'ts' => [
						'$gte' => $start,
					],
				];
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$path = self::$file;
				$repository = new JsonStore(null, null, $path);
				$this->filter = ['ts'=>['$gte'=>$start], 'kind'=>$kind];
				break;
		}

		$this->repository = $repository;
	}

	/**
	 * @param string $kind 'login'/'recovery'
	 */
	public static function forKind($kind): self
	{
		return new self($kind);
	}

	/**
	 * @return Iterator<array>
	 */
	public function all(): Iterator
	{
		if ($this->filter) {
			return $this->repository->filter($this->filter, false);
		} else {
			return $this->repository->all(false);
		}
	}

	public function count(): int
	{
		return $this->repository->count();
	}

	/**
	 * @param string $ip IP-Adresse
	 */
	public function append(string $ip): void
	{
		$timestamp = new DateTime('now');
		$this->repository->append([
			'ts' => $timestamp->getTimestamp(),
			'ip' => $ip,
			'kind' => $this->kind,
		]);
	}

	public function countAttempts(string $ip): int
	{
		return iterator_count($this->repository->filter($this->filter + ['ip'=>$ip]));
	}

	public function cleanup(): void
	{
		if ($this->repository->getPersistedSize() > 10000) {
			$this->repository->trimBefore($this->filter);
		}
	}

	public static function clear(string $kind): void
	{
		$self = new static($kind);
		$self->repository->clear();
	}
}
