<?php

namespace App\Model;

use App\Utils;
use LogicException;

abstract class MigrationStateBase extends Model
{
	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-migrationstate';

	/** @var bool $useRepositorySkeleton */
	protected static $useRepositorySkeleton = false;

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * @param string $id
	 * @return static
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function byId(string $id): self
	{
		if (static::class == self::class) {
			throw new LogicException('Can\'t instantiate abstract class, use subclass instead.');
		}

		return new static($id);
	}

	private function getVersionName(): string
	{
		return (string)$this->get('version');
	}

	/**
	 * @param string $afterVersion Prüfen, ob nach Version `$afterVersion` bereits durchgeführt
	 * @param bool $once
	 * @return bool
	 */
	public function isStale(string $afterVersion='', bool $once=false): bool
	{
		$stateVersion = $this->getVersionName();
		$systemVersion = Utils::getVersionName();

		// Migration states now use the version name instead of the version ID to determine
		// whether they should be run.
		// Consider that the version name may not have propagated to all replica sets yet.
		// It could also be the first request after the update and switch from ID to name.
		if (strlen($stateVersion) == 40) {
			return !$once;
		}
		elseif ($afterVersion) {
			return version_compare(strtolower($systemVersion), strtolower($afterVersion), '>') && (
				$stateVersion == ''
				|| ($once && version_compare(strtolower($afterVersion), strtolower($stateVersion), '>'))
				|| (!$once && version_compare(strtolower($systemVersion), strtolower($stateVersion), '>'))
			);
		}

		return $stateVersion == '' || !$once && version_compare(strtolower($systemVersion), strtolower($stateVersion), '>');
	}

	/**
	 * @param string $version
	 * @return $this
	 */
	private function setVersionName(string $version): self
	{
		$this->set('version', $version);
		return $this;
	}

	/**
	 * @return $this
	 */
	public function touch(): self
	{
		return $this->setVersionName(Utils::getVersionName());
	}

	/**
	 * @return $this
	 */
	public function reset(): self
	{
		return $this->setVersionName('');
	}
}
