<?php

namespace App\Model;

use App\Component\Config\Store\MongoDb;
use App\Component\ConfigImporterContext;
use App\Exception\ConfigNotReadableException;
use App\Exception\ConfigNotWritableException;
use App\Component\Config\MutableModelInterface;
use App\Component\Config\RepositoryInterface;
use App\Component\Config\Store\JsonConfig;
use App\Utils;
use LogicException;
use RuntimeException;
use ReflectionClass;

/**
 * Basisklasse für JSON-Models
 *
 * @phpstan-consistent-constructor
 */
abstract class Model implements MutableModelInterface
{
	/** @var string $id */
	private $id;

	/** @var static|null $source */
	protected $source;

	/** @var string|null $contextDomainId */
	protected static $contextDomainId = null;
	/** @var string|null $contextUserId */
	protected static $contextUserId = null;

	/** @var string|null $repositoryFilename */
	protected static $repositoryFilename = null;

	/** @var string|null $forceRepositoryStore [Utils::REPOSITORY_STORE_*] */
	protected static $forceRepositoryStore = null;

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = false;

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = false;

	/** @var bool $userSpecificRepository */
	protected static $userSpecificRepository = false;

	/** @var bool $accountSingleton Immer das Repository des Hauptbenutzers verwenden; auch für Unterbenutzer. */
	protected static $accountSingleton = false;

	/** @var bool $useRepositoryLocking */
	protected static $useRepositoryLocking = true;

	/** @var bool $useRepositorySkeleton */
	protected static $useRepositorySkeleton = true;

	/** @var string[] $requiredProjectionKeys */
	protected static $requiredProjectionKeys = [];

	/** @var bool $wantsMongoDbMtimeIndex */
	protected static $wantsMongoDbMtimeIndex = false;

	/**
	 * @var array<class-string<Model>,bool> $writtenModels Models, in denen relevante Bearbeitungen stattgefunden haben (true = bereits geflusht).
	 */
	private static $writtenModels = [];
	/**
	 * @var array<class-string<Model>,true> $openModels
	 */
	private static $openModels = [];

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		$this->id = $id;
		$this->source = null;
	}

	/**
	 * Beim Klonen wird eine neue ID generiert, um einen dedizierten Bereich
	 * im Repository für das Objekt zu schaffen.
	 * Es wird ein exklusiver Lock auf dem Repository erzeugt.
	 */
	public function __clone()
	{
		static::beginWrite();
		$destId = self::getRepository()->generateUniqueObjectId();

		$this->copyAndSwitch($destId);
	}

	/**
	 * Kopiert die internen Eigenschaften zu einer neuen ID und lässt
	 * das Model auf den neuen Eintrag zeigen.
	 */
	protected function copyAndSwitch(string $destId): void
	{
		$this->source = self::find($this->getId());

		$this->id = $destId;

		// Eigenschaften übernehmen
		foreach ($this->source->toArray() as $property => $value) {
			$this->set($property, $value);
		}
	}

	/**
	 * Diese Methode stellt das zugrundeliegende Repository des aktuellen Models bereit.
	 * @return RepositoryInterface
	 * @phan-suppress PhanUndeclaredStaticProperty
	 * @phan-suppress PhanUndeclaredStaticMethod
	 */
	protected static function getRepository(): RepositoryInterface
	{
		if (empty(static::$repository)) {
			static::$repository = static::openRepository();

			if (method_exists(static::class, 'onRepositoryLoaded')) {
				static::onRepositoryLoaded();
			}
		}

		return static::$repository;
	}

	/**
	 * Öffnet das Repository vollständig neu
	 *
	 * Dabei werden Wechsel des aktiven Benutzers und der aktiven Domain berücksichtigt
	 * und alle ungespeicherten Änderungen gehen verloren.
	 *
	 * @phan-suppress PhanUndeclaredStaticProperty
	 */
	public static function reloadRepository(): void
	{
		if (!empty(static::$repository)) {
			if (static::$repository->isWriteLocked()) {
				static::reload();
			}
		}

		static::$repository = static::openRepository();
	}

	/**
	 * @return string|null
	 */
	public static function getRepositoryContextDomainId(): ?string
	{
		return static::$contextDomainId;
	}

	/**
	 * @return string|null
	 */
	public static function getRepositoryContextUserId(): ?string
	{
		return static::$contextUserId;
	}

	public static final function isUserSpecificModel(): bool
	{
		return static::$userSpecificRepository ?? false;
	}

	public static final function isDomainSpecificModel(): bool
	{
		return static::$domainSpecificRepository ?? false;
	}

	public static final function getRepositoryName(): string
	{
		return static::$repositoryFilename;
	}

	public static final function isModelWithCustomIds(): bool
	{
		return static::$repositoryWithCustomIds;
	}

	public static final function getRepositoryStoreType(): string
	{
		return static::$forceRepositoryStore ?? Utils::getRepositoryStore();
	}

	/**
	 * Wechselt das Repository in einen bestimmten Benutzer-Domain-Kontext.
	 *
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @param bool $flush Ob bisherige Änderungen am Repository gespeichert werden sollen
	 * @phan-suppress PhanUndeclaredStaticProperty
	 */
	public static function switchRepositoryContext(?User $user, ?Domain $domain=null, bool $flush=false): void
	{
		if ($flush) {
			static::flush();
		}
		if ($user === null) {
			static::$repository = null;
			static::$contextDomainId = null;
			static::$contextUserId = null;
			static::flagClosed();
		} else {
			static::$repository = static::openRepository($user, $domain);
		}
	}

	/**
	 * Diese Methode schließt alle Repositories, die `$domainSpecificRepository` oder `$userSpecificRepository` gesetzt haben.
	 * Nicht gespeicherte Änderungen gehen verloren.
	 * @return void
	 */
	public static final function closeUserModels(): void
	{
		/** @var class-string<static> $class */
		foreach (self::getOpenModelsClassNames() as $class) {
			if ($class::isDomainSpecificModel() or $class::isUserSpecificModel()) {
				$class::switchRepositoryContext(null);
			}
		}
	}

	/**
	 * Diese Methode schließt alle Repositories, die `$domainSpecificRepository` gesetzt haben.
	 * Nicht gespeicherte Änderungen gehen verloren.
	 * @return void
	 */
	public static final function closeDomainModels(): void
	{
		/** @var class-string<static> $class */
		foreach (self::getOpenModelsClassNames() as $class) {
			if ($class::isDomainSpecificModel()) {
				$class::switchRepositoryContext(null);
			}
		}
	}

	/**
	 * Diese Methode schließt alle Repositories.
	 *
	 * Das wird z.B. in Tests benötigt, die mit verschiedenen JSON-Dateien arbeiten.
	 * Nicht gespeicherte Änderungen gehen verloren.
	 */
	public static final function closeAllModels(): void
	{
		/** @var class-string<static> $class */
		foreach (self::getOpenModelsClassNames() as $class) {
			$class::switchRepositoryContext(null);
		}
	}

	/**
	 * Diese Methode stellt die Konfiguration des aktuellen Models in Form eines neuen JsonConfig-Objekts bereit.
	 *
	 * @param User|null $user
	 * @param Domain|null $domain
	 * @return RepositoryInterface
	 */
	protected static function openRepository(?User $user=null, ?Domain $domain=null): RepositoryInterface
	{
		$repoFilename = static::$repositoryFilename;
		$repositoryStore = static::getRepositoryStoreType();
		$domainSpecificRepository = (bool)(static::$domainSpecificRepository ?? false);
		$userSpecificRepository = (bool)(static::$userSpecificRepository ?? false);

		// Gewährleisten, dass dem Model ein Dateiname als Repository beiliegt
		if (empty($repoFilename)) {
			throw new LogicException('No repository defined for '.static::class);
		}

		if ($domainSpecificRepository) {
			if (static::class == Domain::class) {
				throw new LogicException('Domain model repository may not be domain specific');
			}

			if ($domain === null) {
				$domain = Domain::activeDomain();
			}

			if ($domain) {
				if ($repositoryStore == Utils::REPOSITORY_STORE_FILE) {
					$domainConfigDir = $domain->configDir();
					if (is_dir($domainConfigDir) == false && @mkdir($domainConfigDir, 0777, true) == false || is_writable($domainConfigDir) == false) {
						throw new RuntimeException('Could not access domain config folder '.$domainConfigDir.', check permissions.');
					}
				}
			}
			else {
				throw new LogicException('Cannot access domain repository '.static::class.' if no domain is selected.');
			}

			if ($repositoryStore == Utils::REPOSITORY_STORE_FILE) {
				$repoFilename = $domain->configDir(true).'/'.static::$repositoryFilename;
			}

			static::$contextDomainId = $domain->getId();
			static::$contextUserId = $domain->getUser() ? $domain->getUser()->getId() : null;
		}
		else if ($userSpecificRepository) {
			if (static::class == User::class) {
				throw new LogicException('User model repository may not be user specific');
			}

			if ($user === null) {
				$user = User::loggedInUser();
			}

			if ($user) {
				if (static::$accountSingleton) {
					$user = $user->getAccountOwner();
				}

				if ($repositoryStore == Utils::REPOSITORY_STORE_FILE) {
					$userConfigDir = $user->configDir();
					if (is_dir($userConfigDir) == false && @mkdir($userConfigDir) == false || is_writable($userConfigDir) == false) {
						throw new RuntimeException('Could not access user config folder '.$userConfigDir.', check permissions.');
					}
				}
			}
			else {
				throw new LogicException('Cannot access user repository '.static::class.' if no user is logged in.');
			}

			if ($repositoryStore == Utils::REPOSITORY_STORE_FILE) {
				$repoFilename = $user->configDir(true).'/'.static::$repositoryFilename;
			}

			static::$contextDomainId = null;
			static::$contextUserId = $user->getId();
		}

		switch ($repositoryStore) {
			case Utils::REPOSITORY_STORE_MONGODB:
				$repository = new MongoDb($repoFilename, $user, $domain, static::$repositoryWithCustomIds, static::$useRepositorySkeleton ? static::templateRelDir() : null, static::$wantsMongoDbMtimeIndex);
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$repository = new JsonConfig($repoFilename.'.json', static::$useRepositoryLocking, 'var', static::$useRepositorySkeleton ? static::templateRelDir() : null);
				break;
		}

		static::flagOpen();
		return $repository;
	}

	/**
	 * Diese Methode lädt eine Konfigurationsdatei in ein benutzerspezifisches Verzeichnis und löscht optional die
	 * originale Datei.
	 * @param ConfigImporterContext $context
	 * @param string $path Vollständiger Pfad
	 * @param bool $keepOriginal
	 * @param bool $overwrite
	 */
	private static function importConfigFile(ConfigImporterContext $context, string $path, bool $keepOriginal = true, bool $overwrite = false): void
	{
		$sourceFile = $path;
		$repoFilename = '';

		if ((bool)(static::$domainSpecificRepository ?? false)) {
			if ($context->hasDomain() == false) {
				return;
			}

			$repoFilename = $context->getDomain()->configDir().'/'.static::$repositoryFilename;
		}
		else if ((bool)(static::$userSpecificRepository ?? false)) {
			if ($context->hasUser() == false) {
				return;
			}

			$repoFilename = $context->getUser()->configDir().'/'.static::$repositoryFilename;
		}
		else {
			return;
		}

		$repoDir = dirname($repoFilename);
		if (is_dir($repoDir) == false && @mkdir($repoDir, 0777, true) == false || is_writable($repoDir) == false) {
			throw new RuntimeException('Cannot access config folder '.$repoDir.', check permissions.');
		}

		// Wenn globale Konfiguration nicht oder lokale bereits existiert, Vorgang abbrechen
		if (!$overwrite and (is_file($sourceFile) == false or is_file($repoFilename))) {
			return;
		}
		// Konfigurationsdatei migrieren, alte Datei u. U. löschen
		else if (@copy($sourceFile, $repoFilename)) {
			@chmod($repoFilename, 0666);

			if ($keepOriginal == false) {
				@unlink($sourceFile);
			}
		}
	}

	/**
	 * Verzeichnis-Name bzw. -Pfad für das share-Verzeichnis relativ zum Base-Dir
	 *
	 * Bei Plugin-Models wird das share-Verzeichnis im Plugin zurückgegeben, falls es existiert,
	 * ansonsten das globals share-Verzeichnis.
	 *
	 * @return string
	 * @phan-side-effect-free
	 */
	protected static function templateRelDir(): string
	{
		// Plugins verwenden ihre eigenen `share`-Verzeichnisse, falls vorhanden
		if (strpos(static::class, 'Plugins\\') === 0) {
			$basePath = Utils::getBaseDir() . DIRECTORY_SEPARATOR;
			$class = new ReflectionClass(static::class);
			$filePath = $class->getFileName();

			$pluginDir = implode(DIRECTORY_SEPARATOR, array_slice(explode(DIRECTORY_SEPARATOR, substr($filePath, strlen($basePath)), 3), 0, 2));
			$shareDir = $pluginDir . DIRECTORY_SEPARATOR . 'share';
			if (file_exists($basePath . $shareDir)) {
				return $shareDir;
			}
		}

		// Ansonsten: globales `share`-Verzeichnis
		return 'share';
	}

	/**
	 * Diese Methode liest die Daten eines bestimmten Skeletons aus um damit z.B. für Themes neue Daten in bestehende UserDaten einlesen zu können.
	 * Die Daten werden als Array zur Verfügung gestellt damit sich da gut durch iterieren läßt
	 * @return array<string,mixed>
	 * @throws LogicException if no skeleton is defined for the model.
	 */
	public static function readSkeletonData(): array
	{
		$filename = static::$repositoryFilename . '.json';
		$path = Utils::getBaseDir() . DIRECTORY_SEPARATOR . static::templateRelDir();
		if (static::$domainSpecificRepository or static::$userSpecificRepository) {
			$path .= DIRECTORY_SEPARATOR . 'skeleton';

			// Read template files from share/skeleton/clickskeks/ if available
			if (Utils::isClickskeksCloud()) {
				if (file_exists($path.DIRECTORY_SEPARATOR.'clickskeks'.DIRECTORY_SEPARATOR.$filename)) {
					$path .= DIRECTORY_SEPARATOR.'clickskeks';
				}
			}
		}
		$path .= DIRECTORY_SEPARATOR . $filename;

		$result = is_file($path) ? json_decode(file_get_contents($path), true) : null;
		if (is_array($result)) {
			return $result;
		} else {
			throw new LogicException('No skeleton data defined for '.static::class);
		}
	}

	/**
	 * Diese Methode lädt die für das aufgerufene Model vorgesehene Konfigurationsdatei aus dem Vorlagenverzeichnis und
	 * speichert diese für den übergebenen Benutzer in seinem Konfigurationsverzeichnis.
	 * @param ConfigImporterContext $context
	 * @param bool $overwrite
	 */
	public static function importSkeleton(ConfigImporterContext $context, bool $overwrite = false): void
	{
		$path = Utils::getBaseDir() . DIRECTORY_SEPARATOR . static::templateRelDir();
		if (static::$domainSpecificRepository or static::$userSpecificRepository) {
			$path .= DIRECTORY_SEPARATOR . 'skeleton';
		}
		$path .= DIRECTORY_SEPARATOR . static::$repositoryFilename;
		self::importConfigFile($context, $path, true, $overwrite);
	}

	/**
	 * Diese Methode lädt die Konfigurationsdatei des aufgerufenen Models in den nutzerspezifischen Konfigurationsraum.
	 * Hierbei handelt es sich um eine Migrationsroutine alter Versionen, die als Single-User-Version liefen und alle
	 * Konfigurationsdateien direkt im Verzeichnis var/ speicherten.
	 * @param ConfigImporterContext $context
	 */
	public static function migratePreMultiUserConfig(ConfigImporterContext $context): void
	{
		self::importConfigFile($context, Utils::getVarDir().DIRECTORY_SEPARATOR.static::$repositoryFilename, false);
	}

	/**
	 * Diese Methode lädt die Konfigurationsdatei des aufgerufenen Models in den domainspezifischen Konfigurationsraum.
	 * Hierbei handelt es sich um eine Migrationsroutine alter Versionen, die als Single-Domain-Version liefen und
	 * Konfigurationsdateien ggf. im Verzeichnis var/user_data/... speicherten.
	 * @param ConfigImporterContext $context
	 * @throws LogicException
	 */
	public static function migratePreMultiDomainConfig(ConfigImporterContext $context): void
	{
		if ($context->hasUser() == false) {
			throw new LogicException('Import context is missing a user');
		}

		self::importConfigFile($context, Utils::getVarDir().DIRECTORY_SEPARATOR.$context->getUser()->configDir(true).'/'.static::$repositoryFilename, false);
	}

	/**
	 * @return int
	 */
	public static final function lastModifiedTimestamp(): int
	{
		return self::getRepository()->getLastModifiedTimestamp();
	}

	/**
	 * @param string $id
	 * @return bool
	 */
	public static function exists(string $id): bool
	{
		return is_array(self::getRepository()->get($id));
	}

	/**
	 * @param string $id
	 * @return static|null
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function find(string $id): ?self
	{
		return self::exists($id) ? new static($id) : null;
	}

	/**
	 * @return static[]
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function all(): array
	{
		/** @var static[] $entities */
		$entities = [];

		foreach (self::getRepository()->toArray() as $id => $data) {
			$entities[$id] = new static($id);
		}

		return $entities;
	}

	/**
	 * @return static[]
	 * @param array $filter MongoDB-Query-Filter
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	protected static function filterBy(array $filter): array
	{
		/** @var static[] $entities */
		$entities = [];

		foreach (self::getRepository()->toArrayFiltered($filter) as $id => $data) {
			$entities[$id] = new static($id);
		}

		return $entities;
	}

	/**
	 * Diese Methode sorgt dafür, dass das Repository nur noch bestimmte Dokumentfelder lädt.
	 * Speicherintensive Vorgänge kommen so u. U. mit einem größeren Datenumfang zurecht.
	 * Da Dokumente i. d. R. unvollständig geladen werden, sind bis zu einem `reload()` keine
	 * weiteren Schreibvorgänge für das Model zulässig.
	 * Das Repository wird auf den letzten Stand zurückgesetzt.
	 * @param array<string,bool> $projection
	 * @return void
	 */
	public static function setProjection(array $projection): void
	{
		if ($inclusionProjection = array_filter($projection)) {
			$projection = array_fill_keys(static::$requiredProjectionKeys, true) + $inclusionProjection;
		}
		elseif ($projection) {
			$projection = array_diff_key($projection, array_flip(static::$requiredProjectionKeys));
		}

		self::reload();
		$repository = self::getRepository();
		$repository->makeVolatile();
		$repository->setProjection($projection);
	}

	/**
	 * Diese Methode erstellt ein Objekt dieser Klasse mit generierter oder vordefinierter ID.
	 * @param string|null $id if `null`, generates a new ID; uses supplied ID otherwise.
	 * @return static
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function create(?string $id=null): self
	{
		if ($id === null) {
			$id = self::getRepository()->generateUniqueObjectId();
		}
		else {
			self::getRepository()->reserveObjectId($id);
		}

		return new static($id);
	}

	/**
	 * @param string|null $id
	 * @return static
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function createNonPersistent(?string $id=null): self
	{
		self::getRepository()->makeVolatile();

		if ($id === null) {
			$id = self::getRepository()->generateUniqueObjectId();
		}
		else {
			self::getRepository()->reserveObjectId($id);
		}

		return new static($id);
	}

	/**
	 * Zählt die vorhandenen Instanzen dieses Models.
	 * @return int
	 */
	public static function count(): int
	{
		return self::getRepository()->count();
	}

	/**
	 * Gibt an, ob das Repository leer ist. Diese Methode kann performanter
	 * sein als der Vergleich `count() == 0`.
	 * @return bool
	 */
	public static function isEmpty(): bool
	{
		return self::getRepository()->isEmpty();
	}

	/**
	 * Gibt ein zufälliges Element zurück
	 * @return ?static
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function random(): ?self
	{
		$array = self::getRepository()->toArray();
		if (!$array) {
			return null;
		}
		$id = array_rand($array);
		return new static($id);
	}

	/**
	 * Gibt das erste Element aus dem Repo zurück
	 * @return ?static
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function getFirst(): ?self
	{
		$array = self::getRepository()->toArray();
		if (!$array) {
			return null;
		}
		$id = array_key_first($array);
		return new static($id);
	}

	/**
	 * Diese Methode löscht den durch das aktuelle Objekt referenzierten Datensatz aus der Konfiguration.
	 * @param string $id
	 * @param bool $autocommit
	 * @throws ConfigNotWritableException
	 */
	public static function delete(string $id, bool $autocommit=true): void
	{
		$repository = self::getRepository();
		static::primeWritten();
		$repository->unset($id);
		if ($autocommit) {
			static::flush();
		}
	}

	/**
	 * Größe des Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes
	 */
	public static function persistedRepositorySize(): int
	{
		return static::getRepository()->getPersistedSize();
	}

	/**
	 * Diese Methode löscht alle Datensätze aus einer Konfiguration
	 *
	 * @param bool $autocommit
	 */
	public static function clear(bool $autocommit=true): void
	{
		$repository = self::getRepository();
		static::primeWritten();
		$repository->clear();
		if ($autocommit) {
			static::flush();
		}
	}

	/**
	 * Wrapper für den Zugriff auf `RepositoryInterface::get()`.
	 * @param string $property
	 * @param mixed|null $fallback
	 * @return mixed
	 * @see RepositoryInterface::get()
	 */
	public final function get(string $property, $fallback = null)
	{
		$data = self::getRepository()->get($this->getId().'.'.$property, $fallback);
		return $data;
	}

	/**
	 * Wrapper für den Zugriff auf `RepositoryInterface::set()`.
	 * @param string $property
	 * @param mixed $data
	 * @param bool $mayAffectCache
	 * @see RepositoryInterface::set()
	 */
	public final function set(string $property, $data, bool $mayAffectCache=true): void
	{
		self::getRepository()->set($this->getId().'.'.$property, $data);
		if ($mayAffectCache) {
			static::primeWritten();
		}
	}

	/**
	 * Wrapper für den Zugriff auf `RepositoryInterface::setBinary()`.
	 * @param string $property
	 * @param string $data
	 * @param bool $mayAffectCache
	 * @see RepositoryInterface::set()
	 */
	public final function setBinary(string $property, $data, bool $mayAffectCache=true): void
	{
		self::getRepository()->setBinary($this->getId().'.'.$property, $data);
		if ($mayAffectCache) {
			static::primeWritten();
		}
	}

	/**
	 * Wrapper für den Zugriff auf `RepositoryInterface::setIfMissing()`.
	 * @param string $property
	 * @param mixed $data
	 * @param bool $mayAffectCache
	 * @see RepositoryInterface::setIfMissing()
	 */
	public final function setIfMissing(string $property, $data, bool $mayAffectCache=true): void
	{
		self::getRepository()->setIfMissing($this->getId().'.'.$property, $data);
		if ($mayAffectCache) {
			static::primeWritten();
		}
	}

	/**
	 * Wrapper für den Zugriff auf `RepositoryInterface::increment()`.
	 * @param string $property
	 * @param int|float $amount
	 * @param bool $mayAffectCache
	 * @see RepositoryInterface::increment()
	 */
	public final function increment(string $property, $amount, bool $mayAffectCache=true): void
	{
		self::getRepository()->increment($this->getId().'.'.$property, $amount);
		if ($mayAffectCache) {
			static::primeWritten();
		}
	}

	/**
	 * Wrapper für den Zugriff auf `RepositoryInterface::appendToArray()`.
	 * @param string $property
	 * @param mixed $item
	 * @param bool $noDuplicate Duplikate vermeiden
	 * @see RepositoryInterface::appendToArray()
	 */
	public function appendToArray(string $property, $item, bool $noDuplicate=true, bool $mayAffectCache=true): void
	{
		self::getRepository()->appendToArray($this->getId().'.'.$property, $item, $noDuplicate);
		if ($mayAffectCache) {
			static::primeWritten();
		}
	}

	/**
	 * Wrapper für den Zugriff auf `RepositoryInterface::removeFromArray()`.
	 * @param string $property
	 * @param mixed $item
	 * @see RepositoryInterface::removeFromArray()
	 */
	public function removeFromArray(string $property, $item, bool $mayAffectCache=true): void
	{
		self::getRepository()->removeFromArray($this->getId().'.'.$property, $item);
		if ($mayAffectCache) {
			static::primeWritten();
		}
	}

	/**
	 * Diese Methode schreibt alle durch Setter-Methoden vorgenommenen
	 * Änderungen in die Konfiguration auf der Platte.
	 *
	 * Alternativ kann auch stattdessen Model::flush() nach dem Verändern
	 * mehrerer Objekte des gleichen Models verwendet werden.
	 *
	 * @return void
	 * @throws ConfigNotWritableException
	 */
	public final function save(): void
	{
		static::flush();
	}

	/**
	 * Diese Methode schreibt alle durch Setter-Methoden vorgenommenen
	 * Änderungen im Model in die Konfiguration auf der Platte.
	 *
	 * @return void
	 * @throws ConfigNotWritableException
	 */
	public static function flush(): void
	{
		$repository = self::getRepository();

		$wasDirty = $repository->isDirty();
		$repository->flush();

		if ($wasDirty) {
			static::flagWrittenIfPrimed();
		} else {
			static::unprimeWritten();
		}
	}

	/**
	 * Liest alle Einträge neu von der Festplatte ein.
	 * Sämtliche Änderungen gehen verloren.
	 * @throws ConfigNotReadableException
	 */
	public static function reload(): void
	{
		self::getRepository()->reload();
		self::unprimeWritten();
	}

	/**
	 * Sperrt diese Model um Schreiben.
	 *
	 * Danach kann bis zu flush(), reload() oder save() kein parallel
	 * laufender Prozess auf das Model zugreifen.
	 */
	public static function beginWrite(): void
	{
		self::getRepository()->beginWrite();
	}

	/**
	 * @return string
	 */
	public final function getId(): string
	{
		return $this->id;
	}

	/**
	 * IDs sind nicht modifizierbar.
	 */
	protected final function setId(): void {}

	/**
	 * @return array<string,mixed>
	 */
	protected final function toArray(): array
	{
		return (array)self::getRepository()->get($this->getId(), []);
	}

	/**
	 * Debug-Ausgabe für var_dump() erweitern
	 * @return array<string,mixed>
	 */
	public function __debugInfo()
	{
		$data = ['id' => $this->getId()];
		foreach ($this->toArray() as $key => $value) {
			$data['.'.$key] = $value;
		}
		return $data;
	}

	/**
	 * Liste der Model-Klassen in denen während dieses Requests Änderungen persistiert wurden.
	 *
	 * Models an denen ausschließlich Änderungen mit `$mayAffectCache=false` durchgeführt wurden,
	 * werden nicht gelistet.
	 *
	 * @return string[]
	 * @phan-return class-string<Model>[]
	 */
	public final static function getWrittenModelsClassNames(): array
	{
		return array_keys(array_filter(self::$writtenModels));
	}

	/**
	 * Leert die Liste der modifizierten Model-Klassen.
	 *
	 * @see self::getWrittenModelsClassNames()
	 * @see self::isFlaggedWritten()
	 */
	public final static function clearWrittenModelsClassNames(): void
	{
		self::$writtenModels = [];
	}

	/** @internal */
	protected final static function primeWritten(): void
	{
		self::$writtenModels[static::class] = self::$writtenModels[static::class] ?? false;
	}

	/** @internal */
	protected final static function unprimeWritten(): void
	{
		if ((self::$writtenModels[static::class] ?? null) === false) {
			unset(self::$writtenModels[static::class]);
		}
	}

	/** @internal */
	protected final static function flagWrittenIfPrimed(): void
	{
		if (isset(self::$writtenModels[static::class])) {
			self::$writtenModels[static::class] = true;
		}
	}

	/**
	 * Gibt an, ob an diesem Model während dieses Requests Änderungen persistiert werden.
	 *
	 * Ausschließlich Änderungen mit `$mayAffectCache=false` werden ignoriert.
	 *
	 * @return bool
	 */
	protected final static function isFlaggedWritten(): bool
	{
		return self::$writtenModels[static::class] ?? false;
	}

	/**
	 * @return string[]
	 * @phan-return class-string<Model>[]
	 */
	public final static function getOpenModelsClassNames(): array
	{
		return array_keys(self::$openModels);
	}

	/** @internal */
	protected final static function flagOpen(): void
	{
		self::$openModels[static::class] = true;
	}

	/** @internal */
	protected final static function flagClosed(): void
	{
		unset(self::$openModels[static::class]);
	}
}
