<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Exception\ConfigNotWritableException;

class Placeholder extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-placeholders';

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = true;

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = true;

	/** @var bool $wantsMongoDbMtimeIndex */
	protected static $wantsMongoDbMtimeIndex = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Diese Methode erstellt ein Objekt dieser Klasse
	 * @param string|null $id Platzhalter-Name
	 * @return static
	 * @throws \LogicException
	 */
	public static function create(?string $id=null): Model
	{
		if ($id === null) {
			throw new \LogicException('cannot use create() with Placeholder model');
		}
		self::beginWrite();
		$id = trim($id, '{}');
		return new static(static::getRepository()->escapeKey($id));
	}

	/**
	 * Diese Methode löscht den durch das aktuelle Objekt referenzierten Datensatz aus der Konfiguration.
	 * @param string $id
	 * @param bool $autocommit
	 * @throws ConfigNotWritableException
	 */
	public static function delete(string $id, bool $autocommit=true): void
	{
		$id = trim($id, '{}');
		parent::delete(static::getRepository()->escapeKey($id), $autocommit);
	}

	/**
	 * @param string $id
	 * @return bool
	 */
	public static function exists(string $id): bool
	{
		$id = trim($id, '{}');
		return parent::exists(static::getRepository()->escapeKey($id));
	}

	/**
	 * @param string $id
	 * @return static|null
	 */
	public static function find(string $id): ?Model
	{
		$id = trim($id, '{}');
		return self::exists($id) ? new static(static::getRepository()->escapeKey($id)) : null;
	}

	/**
	 * @param string $text
	 * @param bool $htmlSafe Ob der Platzhalter-Wert durch `htmlspecialchars` läuft.
	 * @return string
	 */
	public static function applyAll(string $text, bool $htmlSafe=true): string
	{
		foreach (Placeholder::all() as $placeholder) {
			$text = $placeholder->applyTo($text, $htmlSafe);
		}
		return $text;
	}

	/**
	 * @param string $text
	 * @param bool $htmlSafe Ob der Platzhalter-Wert durch `htmlspecialchars` läuft.
	 * @return string
	 */
	public function applyTo(string $text, bool $htmlSafe=true): string
	{
		return str_ireplace($this->getName(), $htmlSafe ? htmlspecialchars($this->getValue()) : $this->getValue(), $text);
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return '{'.static::getRepository()->unescapeKey($this->getId()).'}';
	}

	/**
	 * @return string
	 */
	public function getValue(): string
	{
		return (string)$this->get('value');
	}

	/**
	 * @param string $text
	 * @return $this
	 */
	public function setValue(string $text): self
	{
		$this->set('value', $text);
		return $this;
	}
}
