<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Utils;
use DateInterval;
use DateTime;

class PluginState extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-plugins';

	/** @var string $forceRepositoryStore */
	protected static $forceRepositoryStore = Utils::REPOSITORY_STORE_FILE;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	public static function findByUID(string $uid): ?self
	{
		foreach (self::all() as $state) {
			if ($state->getPluginUID() == $uid) {
				return $state;
			}
		}
		return null;
	}

	/**
	 * Gibt den PluginState des Plugins zurück, zu dem die übergebene Klasse gehört
	 *
	 * @param class-string $class
	 * @return self|null
	 */
	public static function findByClassName(string $class): ?self
	{
		foreach (self::all() as $state) {
			if (strpos($class, $state->getNamespace()) === 0) {
				return $state;
			}
		}
		return null;
	}

	public function getName(): string
	{
		return (string)$this->get("name",'');
	}

	/**
	 * @return $this
	 */
	public function setName(string $value): self
	{
		$this->set("name", $value);
		return $this;
	}

	public function getType(): string
	{
		return (string)$this->get("type",'ccm19-plugin');
	}

	/**
	 * @return $this
	 */
	public function setType(string $value): self
	{
		$this->set("type", $value);
		return $this;
	}

	public function getVersion(): string
	{
		return (string)$this->get("version",'');
	}

	/**
	 * @return $this
	 */
	public function setVersion(string $value): self
	{
		$this->set("version", $value);
		return $this;
	}

	/**
	 * @param bool $value
	 *
	 * @return $this
	 */
	public function setAllowTrial(bool $value): self
	{
		$this->set("allowTrial",$value);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isAllowTrial(): bool
	{
		return (bool)$this->get("allowTrial",false);
	}

	public function getPluginUID(): string
	{
		return (string)$this->get("pluginUID","");
	}

	/**
	 * @return $this
	 */
	public function setPluginUID(string $uid): self
	{
		$this->set("pluginUID",$uid);
		return $this;
	}

	public function getIcon(): string
	{
		return (string)$this->get("icon", 'fa-plug');
	}

	/**
	 * @return $this
	 */
	public function setIcon(string $value): self
	{
		$this->set("icon", $value);
		return $this;
	}

	public function isActivatablePerDomain(): bool
	{
		return (bool)($this->get('mayActivatePerDomain') ?? false);
	}

	/**
	 * @return $this
	 */
	public function setActivatablePerDomain(bool $value): self
	{
		$this->set("mayActivatePerDomain",$value);
		return $this;
	}

	public function isActivatedForAllAccounts(): bool
	{
		return (bool)$this->get("activateForAllAccounts",false);
	}

	/**
	 * @return $this
	 */
	public function setActivatedForAllAccounts(bool $value): self
	{
		$this->set("activateForAllAccounts",$value);
		return $this;
	}

	public function isActivatedForAllAccountsOnTheirOwn(): bool
	{
		return (bool)$this->get("activateForAllAccountsOnTheirOwn",false);
	}

	/**
	 * @return $this
	 */
	public function setActivatedForAllAccountsOnTheirOwn(bool $value): self
	{
		$this->set("activateForAllAccountsOnTheirOwn",$value);
		return $this;
	}

	public function setFullAccessForAllAccounts(bool $value): self
	{
		$this->set("fullAccessForAllAccounts",$value);
		return $this;
	}

	public function isFullAccessForAllAccounts(): bool
	{
		//Fallback - Plugin wurde noch nie gespeichert, true damit es immer für alle aktiv ist wenn nichts dazu gespeichert wurde
		if(!$this->getSavedAccess())
		{
			return true;
		}
		return (bool)$this->get("fullAccessForAllAccounts",false);
	}


	public function setFullAccessCallCount(int $value): self
	{
		$this->set("FullAccessCallCount",$value);
		return $this;
	}

	public function getFullAccessCallCount(): int
	{
		return (int)$this->get("FullAccessCallCount","");
	}

	public function setFullAccessDomCount(int $value): self
	{
		$this->set("FullAccessDomCount",$value);
		return $this;
	}

	public function getFullAccessDomCount(): int
	{
		return (int)$this->get("FullAccessDomCount","");
	}

	public function setTestAccessCallCount(int $value): self
	{
		$this->set("testAccessCallCount",$value);
		return $this;
	}

	public function getTestAccessCallCount(): int
	{
		return (int)$this->get("testAccessCallCount","");
	}

	public function setTestAccessDomCount(int $value): self
	{
		$this->set("testAccessDomCount",$value);
		return $this;
	}

	public function getTestAccessDomCount(): int
	{
		return (int)$this->get("testAccessDomCount","");
	}


	public function setTestAccessTime(int $value): self
	{
		$this->set("testAccessTime",$value);
		return $this;
	}

	public function getTestAccessTime(): int
	{
		return (int)$this->get("testAccessTime","");
	}


	public function setSavedAccess(bool $value): self
	{
		$this->set("saveAccess",$value);
		return $this;
	}

	public function getSavedAccess(): bool
	{
		return (bool)$this->get("saveAccess",false);
	}


	/**
	 * Name des Verzeichnisses im Plugins-Ordner
	 * @return string
	 */
	public function getDirname(): string
	{
		return (string)$this->get("dirname","");
	}

	/**
	 * @return $this
	 */
	public function setDirname(string $value): self
	{
		$this->set("dirname", $value);
		return $this;
	}

	/** PHP-Namespace für Klassen dieses Plugins */
	public function getNamespace(): string
	{
		$dirname = $this->getDirname();
		if ($dirname === '') {
			$dirname = '__Unknown';
		}
		return "Plugins\\".strtr($dirname, ['.'=>'_', '\\'=>'_', '/'=>'_', '-'=>'_']).'\\';
	}

	/** Twig-Namespace für Templates dieses Plugins */
	public function getTemplateNamespace(): string
	{
		$dirname = $this->getDirname();
		if ($dirname === '') {
			$dirname = '__Unknown';
		}
		return "plugin:".$dirname;
	}

	/** Präfix für Routen-Namen dieses Plugins  */
	public function getRoutePrefix(): string
	{
		$uid = $this->getPluginUid();
		if ($uid === '') {
			$uid = '_unknown_';
		}
		return "plugin_".strtr($uid, ['\\'=>'_', '/'=>'_', '-'=>'_']).'_';
	}

	/**
	 * Ob das Plugin von einem Benutzer verwendet werden darf
	 * In der Agency-Edition können Rechte pro User gesetzt werden.
	 */
	public function isAllowedForUser(User $user): bool {
		$mainUser = $user->getAccountOwner();

		// Subuser-Rechte berücksichtigen
		if ($mainUser->getId() !== $user->getId()) {
			$rights = Utils::getGroupRights($user);
			//Grundsätzlich prüfen ob der Subuser hier diese Route überhaupt aufrufen darf, definiert in der SubUser Rechte Verwaltung
			if (($rights[$this->getRoutePrefix()]['read'] ?? false) == false) {
				//Keine Rechte, dann direkt raus ohne weitere Prüfung
				return false;
			}
		}

		// Rechte des Hauptbenutzers prüfen
		if (Utils::isExtendedEdition() and $mainUser->isClient()) {
			$key = self::buildUserRightsKey($this->getPluginUID());

			//Es besteht kein individueller Zugriff, dann prüfen ob der Tarif des Users den Zugriff erlaubt
			if(!$mainUser->hasAccountFeature($key))
			{
				//Hier checken ob der Tarif des Users freigeschaltet ist für den Zugriff
				if(!$this->checkFullAccessPerCount($mainUser))
				{
					return $this->checkTestAccessPerCount($mainUser);
				}
				return true;
			}
			//Wenn kein Zugriff grundsätzlich besteht, dann ohnehin raus
			return true;
		}

		//Grundsätzlich true, da es hier nur admin sein kann
		return true;
	}

	/**
	 * Checken, ob über spezielle Einstellungen im Plugin Zugriff besteht
	 * @param \App\Model\User $mainUser
	 * @return bool
	 */
	public function checkFullAccessPerCount(User $mainUser): bool
	{
		return ($mainUser->getMaxCallCount() >= $this->getFullAccessCallCount() && $this->getFullAccessCallCount() > 0) && ($mainUser->getMaxDomainCount() >= $this->getFullAccessDomCount() && $this->getFullAccessDomCount() > 0);
	}

	/**
	 * Checken, ob ein Testzugriff besteht
	 * @param \App\Model\User $mainUser
	 *
	 * @return bool
	 */
	public function checkTestAccessPerCount(User $mainUser): bool
	{
		return ($mainUser->getMaxCallCount() <= $this->getTestAccessCallCount() && $this->getTestAccessCallCount() > 0) && ($mainUser->getMaxDomainCount() <= $this->getTestAccessDomCount() && $this->getTestAccessDomCount() > 0);
	}

	public function isAllowedForCurrentUser(): bool {
		$user = User::loggedInUser();
		if ($user) {
			return $this->isAllowedForUser($user);
		}
		return false;
	}

	/**
	 * Ob das Plugin in der Domain aktiv ist
	 */
	public function isActiveForDomain(Domain $domain): bool {
		$owner = $domain->getUser();
		if (!$owner) {
			return false;
		}

		// Wenn der Domain-Besitzer das Plugin nicht verwenden darf,
		// ist es auch nicht aktiv.
		if (!$this->isAllowedForUser($owner)) {
			return false;
		}

		// Aktiv, wenn immer für alle Accounts aktiv
		if ($this->isActivatedForAllAccounts()) {
			return true;
		}
		// Wenn aktivierbar, Domain-Einstellungen prüfen
		elseif ($this->isActivatedForAllAccountsOnTheirOwn()) {
			$plugins = $domain->getActivatedPluginIds();
			return in_array($this->getPluginUID(), $plugins);
		}
		else {
			return false;
		}
	}

	public function isActiveForCurrentDomain(): bool {
		$domain = Domain::activeDomain();
		if ($domain) {
			return $this->isActiveForDomain($domain);
		} else {
			return $this->isActivatedForAllAccounts();
		}
	}

	public function cleanupCompat(): void
	{
		if ($this->get('img')) {
			$this->set('img', '');
		}
	}

	/** Schlüssel im Rechte-Array von Agency-Usern */
	public function getUserRightsKey(): string {
		return self::buildUserRightsKey($this->getPluginUID());
	}

	/** Schlüssel im Rechte-Array von Subgruppen */
	public function getSubgroupRightsKey(): string {
		return self::buildSubgroupRightsKey($this->getPluginUID());
	}

	public static function buildUserRightsKey(string $uid): string
	{
		// Sonderfälle für alte Plugins
		$specialCases = [
			'ccm19/abtests' => 'canUseABTests',
			'ccm19/statistics' => 'canUseAnalytics'
		];
		// Generischen Namen aus Plugin-Name/ID bauen.
		return $specialCases[$uid] ?? 'canUsePlugin/'.$uid;
	}

	/**
	 * @param string $uid
	 *
	 * @return string
	 */
	public static function buildSubgroupRightsKey(string $uid): string
	{
		// Sonderfälle für alte Plugins
		$specialCases = [
			'ccm19/abtests' => '202w33b',
			'ccm19/statistics' => '102w33b'
		];
		// Generischen Namen aus Plugin-Name/ID bauen.
		return $specialCases[$uid] ?? 'plugin/'.$uid;
	}

	/**
	 * This method checks if a trial period is possible for the account owner.
	 * @param User|null $user If omitted, the logged in user will be queried. (will check the account owner)
	 * @return bool
	 */
	public function trialIsPossible(User $user=null): bool
	{
		$user = $user ?? User::loggedInUser();
		$owner = $user ? $user->getAccountOwner() : null;
		return $this->isAllowTrial() // If the plugin allows for a trial period
			&& !$this->isFullAccessForAllAccounts() // If full access is granted to all accounts, there is no trial period
			&& $owner // If no user is logged in, there is no trial period
			&& !$owner->hasAccountFeature(self::buildUserRightsKey($this->getPluginUID())) // If the user already has access to the plugin, there is no trial period
			&& !$this->checkFullAccessPerCount($owner) // If full access is granted based on the plan parameters, there is no trial period
			&& $this->checkTestAccessPerCount($owner); // Should a trial period be granted based on the plan parameters?
	}

	/**
     * @return bool
     */
    public function trialIsExpired(): bool
    {
		return $this->trialIsExpiredForUser(User::loggedInUser());
	}

	/**
	 * @param User|null $user
	 * @return bool
	 */
	public function trialIsExpiredForUser(?User $user): bool
	{
		if (!$this->trialIsPossible($user)) {
			// If full access is not granted to all accounts and no user is logged in,
			// the trial period is considered expired.
			return !$this->isFullAccessForAllAccounts() && $user === null;
		}

		// If a trial period is possible, a user must be logged in.
		assert($user instanceof User);
		$user = $user->getAccountOwner();

		//TrialData
		$trialTime = $this->getTestAccessTime();
		$name = $this->getPluginUID();

		//Trial Daten des Users
		$trialPeriods = $user->getTrialPeriods();

		//Wenn trialPeriods leer ist, dann setzen und dann passt das ja auf jeden Fall, weil die Periode dann startet
		if(empty($trialPeriods[$name]))
		{
			$trialPeriods[$name] = time()+($trialTime*86400);
			$user->setTrialPeriods($trialPeriods)->save();
			return false;
		}

		//Die Trialperiode ist noch nicht abgelaufen
		if($trialPeriods[$name] > time())
		{
			return false;
		}

		//Base - immer expired
		return true;
	}

	/**
	 * This method checks whether the trial period is active for the account owner.
	 * @return bool
	 */
	public function trialIsRunning(): bool
	{
		$user = User::loggedInUser();
		return $this->trialIsPossible($user) && !$this->trialIsExpired();
	}

	/**
	 * @return bool
	 */
	public function trialIsRunningOrExpired(): bool
	{
		return $this->trialIsRunning() || $this->trialIsExpired();
	}

	/**
	 * @return DateInterval|null The remaining time of the trial period, or null if the trial period is not running.
	 */
	public function trialTimeRemaining(): ?DateInterval
	{
		$user = User::loggedInUser();
		$owner = $user ? $user->getAccountOwner() : null;

		if ($owner && $this->trialIsRunning()) {
			// Calculate the remaining time by subtracting the current time from the expiration time
			$expires = new DateTime('@' . $owner->getTrialPeriods()[$this->getPluginUID()]);
			return (new DateTime('now'))->diff($expires);
		}
		else {
			return null;
		}
	}

	/**
	 * @return int|null The number of days remaining in the trial period, or null if the trial period is not running.
	 */
	public function trialDaysRemaining(): ?int
	{
		$remaining = $this->trialTimeRemaining();
		if ($remaining) {
			return $remaining->invert ? 0 : $remaining->days;
		}
		else {
			return null;
		}
	}
}
