<?php

namespace App\Model;

use App\Component\ConfigImporterContext;
use App\Component\LogStream\RepositoryInterface;
use App\Component\LogStream\Store\JsonStore;
use App\Component\LogStream\Store\MongoDb;
use App\Component\LogStream\Store\RedisStore;
use App\Exception\LogNotReadableException;
use App\Exception\LogNotWritableException;
use App\Exception\LogNotCreatableException;
use App\Kernel;
use App\Utils;
use DateTimeInterface;
use Iterator;
use LimitIterator;
use LogicException;
use MongoDB\BSON\ObjectId;
use RuntimeException;

class Protocol
{
	public const CLICKED_BUTTON_ACCEPT_ALL = 'acceptAll';
	public const CLICKED_BUTTON_DECLINE = 'decline';
	public const CLICKED_BUTTON_SAVE = 'save';

	/** @var string $name */
	private static $name = 'cm-consent-protocol';

	/** @var RepositoryInterface $repository */
	private $repository;

	/**
	 * @param Domain $domain
	 * @throws LogNotCreatableException Wenn Logdatei nicht geschrieben werden kann.
	 */
	private function __construct(Domain $domain)
	{
		$repositoryStore = Utils::getProtocolRepositoryStore();
		switch ($repositoryStore) {
			case Utils::REPOSITORY_STORE_MONGODB:
				$repository = new MongoDb($domain->getUser(), $domain, self::$name);
				break;
			case Utils::REPOSITORY_STORE_REDIS:
				$repository = new RedisStore($domain->getUser(), $domain, self::$name);
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$repository = new JsonStore($domain->getUser(), $domain, self::$name);
				break;
		}

		$this->repository = $repository;
	}

	/**
	 * @param Domain $domain
	 * @return self
	 * @throws LogNotWritableException Wenn nicht auf Logdatei zugegriffen werden kann.
	 */
	public static function byDomain(Domain $domain): self
	{
		return new self($domain);
	}

	/**
	 * Diese Methode durchsucht das Protokoll anhand einer Consent-ID. Sucht wie Git.
	 *
	 * Änderungen eines Consents werden zu einem Eintrag zusammengefasst.
	 *
	 * @param string $ucid
	 * @return array[]
	 * @throws RuntimeException Wenn nicht auf Logdatei zugegriffen werden kann.
	 */
	public function filterByConsentId(string $ucid): array
	{
		$matches = iterator_to_array(new LimitIterator($this->repository->filter(
			[
				'consent'=>true,
				'ucid'=>['$regex' => '^'.preg_quote($ucid), '$options'=>'i'],
			],
			true
		), 0, 12000));

		$uniqueGroup = count(array_reduce($matches, function (array $groups, array $match) {
			$groups[$match['ucid']] = $match['ucid'];
			return $groups;
		}, [])) < 2;

		return $uniqueGroup ? $matches : [];
	}

	/**
	 * Diese Methode durchsucht das Protokoll anhand von Datum und Consent-ID. Sucht wie Git.
	 *
	 * Änderungen eines Consents werden NICHT zu einem Eintrag zusammengefasst.
	 *
	 * @param DateTimeInterface $startDate Startzeitpunkt für die Suche
	 * @param DateTimeInterface|null $endDate Endzeitpunkt (optional, nicht-inklusiv)
	 * @param string $ucid UCID bzw. Präfix nach dem gesucht werden soll
	 * @param bool|null $consented Filtern nach consent true/false (optional)
	 * @param bool $reversed true für antichronologische Sortierung
	 * @return Iterator<array>
	 * @throws LogNotReadableException
	 */
	public function filter(DateTimeInterface $startDate, ?DateTimeInterface $endDate=null, string $ucid='', ?bool $consented=null, bool $reversed = false): Iterator
	{
		$filter = [
			'timestamp'=>['$gte' => $startDate->getTimestamp()]
		];
		if ($endDate !== null) {
			$filter['timestamp']['$lt'] = $endDate->getTimestamp();
		}

		if (Utils::getProtocolRepositoryStore() === Utils::REPOSITORY_STORE_MONGODB) {
			// Optimierung: Vorfiltern über die Objekt-IDs ist effizienter
			$filter['_id'] = [ '$gte' => new ObjectId(substr(str_pad(dechex($startDate->getTimestamp()-5), 8, '0', STR_PAD_LEFT), -8, 8).'0000000000000000') ];
			if ($endDate !== null) {
				$filter['_id']['$lte'] = new ObjectId(substr(str_pad(dechex($endDate->getTimestamp()+5), 8, '0', STR_PAD_LEFT), -8, 8).'FFFFFFFFFFFFFFFF');
			}
		}

		if ($ucid !== '') {
			$filter['ucid'] = ['$regex' => '^'.preg_quote($ucid), '$options'=>'i'];
		}
		if ($consented !== null) {
			$filter['consent'] = $consented;
		}

		return $this->repository->filter($filter, $reversed);
	}

	/**
	 * Diese Methode gibt den Dateipfad der Konfigurationsdatei zurück.
	 * @param Domain $domain
	 * @return string
	 */
	public static function staticPathname(Domain $domain): string
	{
		return $domain->configDir().'/'.self::$name.'.log';
	}

	/**
	 * @param ConfigImporterContext $context
	 * @phan-suppress PhanTypeMismatchArgumentNullable
	 */
	public static function migratePreMultiUserConfig(ConfigImporterContext $context): void
	{
		if ($context->hasDomain() == false) {
			return;
		}

		/** @var Kernel $kernel */
		global $kernel;

		$sourceFile = Utils::getVarDir() . DIRECTORY_SEPARATOR . self::$name.'.log';
		$destFile = self::staticPathname($context->getDomain());

		if (is_file($sourceFile) == false || is_file($destFile)) {
			return;
		}
		else if (@copy($sourceFile, $destFile)) {
			@chmod($destFile, 0666);
			@unlink($sourceFile);
		}
	}

	/**
	 * @param ConfigImporterContext $context
	 * @throws LogicException
	 * @phan-suppress PhanTypeMismatchArgumentNullable
	 */
	public static function migratePreMultiDomainConfig(ConfigImporterContext $context): void
	{
		if ($context->hasUser() == false || $context->hasDomain() == false) {
			return;
		}

		/** @var Kernel $kernel */
		global $kernel;

		$sourceFile = $context->getUser()->configDir().'/'.self::$name.'.log';
		$destFile = self::staticPathname($context->getDomain());

		if (is_file($sourceFile) == false || is_file($destFile)) {
			return;
		}
		else if (@copy($sourceFile, $destFile)) {
			@chmod($destFile, 0666);
			@unlink($sourceFile);
		}
	}

	/**
	 * @param bool $reversed TRUE für umgekehrt-chronologische Sortierung
	 * @return Iterator<array>
	 * @throws LogNotReadableException
	 */
	public function all(bool $reversed=false): Iterator
	{
		return $this->repository->all($reversed);
	}

	/**
	 * @param bool $reversed TRUE für umgekehrt-chronologische Sortierung
	 * @return Iterator<array>
	 * @throws LogNotReadableException
	 */
	public function allToday(bool $reversed=false): Iterator
	{
		$timestamp = strtotime('today midnight');
		return $this->repository->filter(['timestamp'=>['$gte' => $timestamp]], $reversed);
	}

	/**
	 * @param int $timestamp timestamp ab wann die Daten ausgegeben werden sollen...
	 * @param bool $reversed TRUE für umgekehrt-chronologische Sortierung
	 * @return Iterator<array>
	 */
	public function allFromDate(int $timestamp, bool $reversed = false): Iterator
	{
		return $this->repository->filter(['timestamp'=>['$gte' => $timestamp]], $reversed);
	}

	/**
	 * @param $reversed TRUE für umgekehrt-chronologische Sortierung
	 * @return Iterator<array>
	 * @throws LogNotReadableException
	 */
	public function allConsented(bool $reversed=false): Iterator
	{
		return $this->repository->filter(['consent'=>true], $reversed);
	}

	public function first(): ?array
	{
		$iterator = $this->repository->all(false);
		$iterator->rewind();
		return ($iterator->valid()) ? $iterator->current() : null;
	}

	/**
	 * Protokoll rotieren/archivieren
	 *
	 * Vor dem eigentlichen Prozess wird ein ProtocolRotateEvent ausgelöst.
	 * @see ProtocolRotateEvent
	 */
	public function rotateProtocol(): void
	{
		$this->repository->rotateProtocol();
	}

	public function count(): int
	{
		return $this->repository->count();
	}

	/**
	 * @param array $data
	 */
	public function append(array $data): void
	{
		$this->repository->append($data);
	}

	/**
	 * Löscht alle Einträge, die vor $date angelegt wurden
	 *
	 * @param DateTimeInterface $date
	 * @return int Anzahl gelöschter Einträge
	 */
	public function trimBefore(DateTimeInterface $date): int
	{
		$timestamp = $date->getTimestamp();

		return $this->repository->trimBefore(['timestamp' => ['$gte' => $timestamp]]);
	}

	/**
	 * Größe des (nicht-archivierten) Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes (-1 wenn nicht bestimmbar)
	 */
	public function persistedRepositorySize(): int
	{
		return $this->repository->getPersistedSize();
	}

	/**
	 * Erstellt Einträge im Protokoll. Sollte nur zum Import/Migrieren verwendet werden.
	 *
	 * Bestehende Daten werden nicht überschrieben, aber wenn Einträge aus der
	 * Vergangenheit in eine Domain mit bestehenden Einträgen importiert werden,
	 * kann das die Suchfunktion verwirren.
	 *
	 * Die Einträge MÜSSEN mit höchstens 1s abweichung aufsteigend nach dem Timestamp sortiert sein.
	 *
	 * @param array[]|\Traversable<array> $items (aufsteigend sortiert nach Timestamp)
	 * @return int Anzahl geschriebener Einträge
	 * @throws LogNotWritableException
	 */
	public function writeBulk(iterable $items): int
	{
		$repo = $this->repository;
		$total = 0;
		if ($repo instanceof MongoDb) {
			foreach (Utils::chunk(100, $items) as $chunk) {
				foreach ($chunk as &$item) {
					if (isset($item['timestamp'])) {
						$id = new ObjectId();
						$item['_id'] = new ObjectId(substr(str_pad(dechex($item['timestamp']), 8, '0', STR_PAD_LEFT), -8, 8).substr((string)$id, 8));
					}
					$total++;
				}
				unset($item);
				$repo->appendMulti($chunk);
			}
		} else {
			foreach (Utils::chunk(100, $items) as $chunk) {
				$repo->appendMulti($chunk);
				$total += count($chunk);
			}
		}
		return $total;
	}

	/**
	 * @param string $button
	 * @return bool
	 */
	public static function isClickedButtonValid(string $button): bool
	{
		return in_array($button, [
			self::CLICKED_BUTTON_ACCEPT_ALL,
			self::CLICKED_BUTTON_DECLINE,
			self::CLICKED_BUTTON_SAVE,
			'opened', // bounce
		]);
	}
}
