<?php

namespace App\Model;

use App\Component\LogStream\RepositoryInterface;
use App\Component\LogStream\Store\GzJsonStore;
use App\Component\LogStream\Store\MongoDb;
use App\Exception\LogNotReadableException;
use App\Exception\LogNotWritableException;
use App\Exception\LogNotDeletableException;
use App\Utils;
use DateTimeInterface;
use DateTimeImmutable;
use Iterator;
use IteratorIterator;
use MongoDB\BSON\ObjectId;

class ProtocolArchive
{
	private static $name = 'cm-consent-protocol';

	/** @var array[] $repositoryNames  */
	private $repositoryNames;
	/** @var string $repositoryStoreKind */
	private $repositoryStoreKind;
	/** @var string|null $customArchiveRoot */
	private $customArchiveRoot;

	private $domain;

	/**
	 * @param Domain $domain
	 */
	private function __construct(Domain $domain)
	{
		$this->domain = $domain;

		$this->customArchiveRoot = Utils::getCustomArchiveRoot();
		$this->repositoryStoreKind = Utils::getArchiveRepositoryStore();
		switch ($this->repositoryStoreKind) {
			case Utils::REPOSITORY_STORE_MONGODB:
				// Bei MongoDB ist das Archiv in einer einzelnen komprimierten Collection
				$name = self::$name.'.archive';
				$this->repositoryNames = [ [new DateTimeImmutable('2010-01-01'), new DateTimeImmutable('now'), $name] ];
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				// Sammle die Namen aller passenden .log.gz-Dateien
				$names = GzJsonStore::listCollections($domain->getUser(), $domain, self::$name, $this->customArchiveRoot);
				$this->calculateGzJsonMetadata($names);
				break;
		}
	}

	/**
	 * Erstellt eine Liste von Mindest-Startzeitpunkt mit Dateinamen
	 * damit die Suche nicht immer alle Dateien durchgehen muss
	 *
	 * @param string[] $names
	 */
	private function calculateGzJsonMetadata(array $names): void
	{
		$repositoryNames = [];
		sort($names, SORT_STRING);
		$prevEndDate = new DateTimeImmutable('2010-01-01');
		foreach ($names as $name) {
			$endDate = DateTimeImmutable::createFromFormat('Ymd-His', substr(explode('~', $name, 2)[1], 0, 15));
			$repositoryNames[] = [$prevEndDate, $endDate, $name];
			$prevEndDate = $endDate;
		}
		$this->repositoryNames = $repositoryNames;
	}

	/**
	 * @param Domain $domain
	 * @return self
	 * @throws LogNotWritableException Wenn nicht auf Logdatei zugegriffen werden kann.
	 */
	public static function byDomain(Domain $domain): self
	{
		return new self($domain);
	}

	/**
	 * @param string $name
	 */
	private function openRepository($name): RepositoryInterface
	{
		switch ($this->repositoryStoreKind) {
			case Utils::REPOSITORY_STORE_MONGODB:
				return new MongoDb($this->domain->getUser(), $this->domain, $name, false);
			case Utils::REPOSITORY_STORE_FILE:
			default:
				return new GzJsonStore($this->domain->getUser(), $this->domain, $name, $this->customArchiveRoot);
		}
	}

	/**
	 * @param DateTimeInterface $startDate Startzeitpunkt für die Suche
	 * @param DateTimeInterface|null $endDate Endzeitpunkt (optional, nicht-inklusiv)
	 * @param string $ucid UCID bzw. Präfix nach dem gesucht werden soll
	 * @param bool|null $consented Filtern nach consent true/false (optional)
	 * @param bool $reversed True für antichronologische Sortierung
	 * @param bool $includeActiveProtocol Auch das aktive Protokoll durchsuchen
	 * @return Iterator<array>
	 * @throws LogNotReadableException
	 */
	public function filter(
		DateTimeInterface $startDate, ?DateTimeInterface $endDate=null, string $ucid='',
		?bool $consented=null, bool $reversed=false, bool $includeActiveProtocol=false
	): Iterator {
		$repos = array_map(function ($item) { return $item[2]; }, array_filter($this->repositoryNames, function ($item) use ($startDate, $endDate) {
			return ($item[1] >= $startDate and ($endDate === null or $item[0] <= $endDate));
		}));
		assert(is_array($repos));

		$filter = [
			'timestamp'=>['$gte' => $startDate->getTimestamp()]
		];
		if ($endDate !== null) {
			$filter['timestamp']['$lt'] = $endDate->getTimestamp();
		}

		if ($this->repositoryStoreKind === Utils::REPOSITORY_STORE_MONGODB) {
			// Optimierung: Vorfiltern über die Objekt-IDs ist effizienter
			$filter['_id'] = [ '$gte' => new ObjectId(substr(str_pad(dechex($startDate->getTimestamp()-5), 8, '0', STR_PAD_LEFT), -8, 8).'0000000000000000') ];
			if ($endDate !== null) {
				$filter['_id']['$lte'] = new ObjectId(substr(str_pad(dechex($endDate->getTimestamp()+5), 8, '0', STR_PAD_LEFT), -8, 8).'FFFFFFFFFFFFFFFF');
			}
		}

		if ($ucid !== '') {
			$filter['ucid'] = ['$regex' => '^'.preg_quote($ucid), '$options'=>'i'];
		}
		if ($consented !== null) {
			$filter['consent'] = $consented;
		}

		if ($reversed and $includeActiveProtocol) {
			// Das aktive Protokoll muss nicht durchgegangen werden, wenn der Datumsbereich nur im Archiv liegen kann
			if (!$repos or end($repos) === end($this->repositoryNames)[2]) {
				yield from Protocol::byDomain($this->domain)->filter($startDate, $endDate, $ucid, $consented, $reversed);

			}
		}

		if ($reversed) {
			$repos = array_reverse($repos, true);
		}

		foreach ($repos as $name) {
			$repo = $this->openRepository($name);
			yield from $repo->filter($filter, $reversed);
			unset($repo);
		}

		if (!$reversed and $includeActiveProtocol) {
			// Das aktive Protokoll muss nicht durchgegangen werden, wenn der Datumsbereich nur im Archiv liegen kann
			if (!$repos or end($repos) === end($this->repositoryNames)[2]) {
				yield from Protocol::byDomain($this->domain)->filter($startDate, $endDate, $ucid, $consented, $reversed);
			}
		}
	}

	/**
	 * @param DateTimeInterface $endDate Endzeitpunkt bis zu welchem gelöscht werden soll (nicht-inklusiv)
	 * @return void
	 * @throws LogNotDeletableException
	 */
	public function deleteAllBefore(DateTimeInterface $endDate): void {
		// Bestimmen, welche Repositories berücksichtigt werden müssen
		$datedRepos = array_map(function ($item) { return [$item[1], $item[2]]; }, array_filter($this->repositoryNames, function ($item) use ($endDate) {
			return ($item[0] <= $endDate);
		}));

		$filter = [
			'timestamp'=>['$lt' => $endDate->getTimestamp()]
		];

		if ($this->repositoryStoreKind === Utils::REPOSITORY_STORE_MONGODB) {
			// Optimierung: Vorfiltern über die Objekt-IDs ist effizienter
			$filter['_id'] = [ '$lte' => new ObjectId(substr(str_pad(dechex($endDate->getTimestamp()+5), 8, '0', STR_PAD_LEFT), -8, 8).'FFFFFFFFFFFFFFFF') ];
		}

		foreach ($datedRepos as $entry) {
			list($repoEndDate, $name) = $entry;
			$repo = $this->openRepository($name);
			// Vollständig veraltete Archive direkt komplett löschen
			if ($repoEndDate < $endDate) {
				$repo->clear();
			} else {
				$repo->trimBefore($filter);
			}

		}
	}

	/**
	 * Erstellt Einträge im Archiv. Sollte nur zum Import/Rotieren verwendet werden.
	 *
	 * Bestehende Daten werden nicht überschrieben, aber wenn Einträge weit aus der
	 * Vergangenheit in eine Domain mit bestehenden Archiv-Einträgen importiert werden,
	 * kann das die Suchfunktion verwirren.
	 *
	 * Die Einträge MÜSSEN mit höchstens 1s abweichung aufsteigend nach dem Timestamp sortiert sein.
	 *
	 * @param array[]|\Traversable<array> $items (aufsteigend sortiert nach Timestamp)
	 * @return int Anzahl geschriebener Einträge
	 * @throws LogNotWritableException
	 */
	public function writeBulk(iterable $items): int
	{
		$total = 0;
		if ($this->repositoryStoreKind === Utils::REPOSITORY_STORE_MONGODB) {
			$repo = null;
			foreach ($this->repositoryNames as $entry) {
				$repo = $this->openRepository($entry[2]);
				break;
			}
			assert($repo !== null);

			foreach (Utils::chunk(100, $items) as $chunk) {
				foreach ($chunk as &$item) {
					$id = new ObjectId();
					if (isset($item['timestamp'])) {
						$item['_id'] = new ObjectId(substr(str_pad(dechex($item['timestamp']), 8, '0', STR_PAD_LEFT), -8, 8).substr((string)$id, 8));
					}
					$total++;
				}
				unset($item);
				$repo->appendMulti($chunk);
			}
		} else {
			$iter = new IteratorIterator($items);
			do {
				$name = 'tmp-'.self::$name . '~' . rand(1000000, 9999999);
				$repo = $this->openRepository($name);
				assert($repo instanceof GzJsonStore);
				$repo->createOrOverwrite();
				$lastTimestamp = 1546297200;
				$counter = 0;
				$continue = false;
				foreach ($iter as $item) {
					$counter++;
					$total++;
					$lastTimestamp = max($item['timestamp'] ?? $lastTimestamp, $lastTimestamp);
					$repo->append($item);

					if ($counter >= 200000) {
						$continue = true;
						break;
					}
				}
				$tmppath = $repo->getPathname();
				$repo->close();
				if ($counter > 0) {
					$pathinfo = pathinfo($tmppath);
					$i = 0;
					do {
						$destPathname = $pathinfo['dirname'].'/'.self::$name.date('~Ymd-His', $lastTimestamp+1).(($i == 0) ? '' : '-'.$i).'.'.($pathinfo['extension'] ?? '');
						$i++;
					} while (file_exists($destPathname));
					rename($tmppath, $destPathname);
				} else {
					@unlink($tmppath);
				}
			} while ($continue);
		}
		return $total;
	}
}
