<?php

namespace App\Model\Regulations;

use App\Component\EmbeddingDb\Repository;
use \DateTimeInterface;
use \LogicException;

abstract class Model
{
	/** @var string $id */
	private $id;
	/** @var array $dataset */
	private $dataset;

	/** @var string|null $repositoryFilename */
	protected static $repositoryFilename = null;

	/**
	 * @param array $dataset
	 */
	protected function __construct(string $id, array $dataset)
	{
		$this->id = $id;
		$this->dataset = $dataset;
	}

	public function asAssociativeArray(): array
	{
		return $this->dataset;
	}

	protected function getTranslated(string $key, string $langCode, string $fallback): string
	{
		if (strpos($langCode, '_') !== false) {
			$langCode = explode('_', $langCode)[0];
		} elseif (strpos($langCode, '-') !== false) {
			$langCode = explode('-', $langCode)[0];
		}

		$candidates = (array)$this->get($key, []);
		if (count($candidates) === 0) {
			return $fallback;
		}
		elseif (count($candidates) === 1) {
			return reset($candidates) ?? $fallback;
		}
		else {
			return $candidates[$langCode] ?? $candidates['en'] ?? $candidates['de'] ?? $fallback;
		}
	}

	/**
	 * Diese Methode stellt die Konfiguration des aktuellen Models bereit.
	 * @return Repository
	 * @phan-suppress PhanUndeclaredStaticProperty
	 * @phan-suppress PhanUndeclaredStaticMethod
	 */
	protected static function getRepository(): Repository
	{
		if (static::$repository === null) {
			static::$repository = static::openRepository();
		}
		return static::$repository;
	}

	/**
	 * Diese Methode stellt die Konfiguration des aktuellen Models bereit und lädt dieses ggf. neu.
	 *
	 * @return Repository
	 */
	protected static function openRepository(): Repository
	{
		$repoFilename = static::$repositoryFilename;

		// Gewährleisten, dass dem Model ein Dateiname als Repository beiliegt
		if (empty($repoFilename)) {
			throw new LogicException('No repository defined for '.static::class);
		}

		return Repository::load($repoFilename);
	}

	/**
	 * Größe des Repositories auf der Festplatte in Bytes, ohne ungespeicherte Änderungen
	 *
	 * @return int Größe in Bytes
	 */
	public static function persistedRepositorySize(): int
	{
		return static::getRepository()->getPersistedSize();
	}

	/**
	 * @return DateTimeInterface
	 */
	public static function lastModified(): DateTimeInterface
	{
		return static::getRepository()->getLastModified();
	}

	/**
	 * @param string $property
	 * @param mixed|null $fallback
	 * @return mixed
	 */
	protected final function get(string $property, $fallback = null)
	{
		return isset($this->dataset[$property]) ? $this->dataset[$property] : $fallback;
	}

	/**
	 * @return string
	 */
	public function getId(): string
	{
		return (string)$this->id;
	}
}

