<?php

namespace App\Model\Regulations;

use App\Component\Config\ModelInterface;
use App\Component\EmbeddingDb\Repository;
use App\Exception\ConfigNotReadableException;
use App\Model\Country;

class Regulation extends Model implements ModelInterface
{
	protected static $repositoryFilename = 'regulations';

	/** @var Repository|null $repository */
	protected static $repository = null;

	public function getName(string $langCode = 'en'): string
	{
		return (string)$this->getTranslated('name', $langCode, '');
	}

	public function getDescription(string $langCode = 'en'): string
	{
		return (string)$this->getTranslated('desc', $langCode, '');
	}

	public function getRegionCode(): string {
		return strtoupper(explode('-', $this->getId())[0]);
	}

	public function getRegionName(): string {
		$code = $this->getRegionCode();
		if ($code == "EU") {
			return "EU";
		} else {
			$country = Country::find($code);
			if ($country) {
				return $country->getName();
			} else {
				return $code;
			}
		}
	}

	/**
	 * @return Article|null
	 */
	public function getArticle(int $id): ?Article
	{
		$id = (string)$id;
		$articles = $this->get('articles', []);
		$data = $articles[$id] ?? null;
		if ($data) {
			return new Article($id, $data);
		} else {
			return null;
		}
	}

	/**
	 * @return Article[]
	 */
	public function getArticles(): array
	{
		$articles = $this->get('articles', []);
		return array_map(function ($key, $value): Article {
			return new Article($key, $value);
		}, array_keys($articles), $articles);
	}

	/**
	 * @return array<int,Article[]>
	 */
	public function getArticlesByGroups(): array
	{
		$articles = $this->get('articles', []);

		$result = [];
		foreach ($articles as $articleId => $article) {
			$group = $article['group'] ?? 0;
			if (!isset($result[$group])) {
				$result[$group] = [];
			}
			$result[$group][$articleId] = new Article($articleId, $article);
		}
		ksort($result, SORT_NUMERIC);
		return $result;
	}

	/**
	 * @param string $id
	 * @return bool
	 */
	public static function exists(string $id): bool
	{
		return $id ? is_array(self::getRepository()->get($id)) : false;
	}

	/**
	 * @param string $id
	 * @return static|null
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function find(string $id): ?self
	{
		return self::exists($id) ? new static($id, self::getRepository()->get($id)) : null;
	}

	/**
	 * @return static[]
	 * @phan-suppress PhanTypeInstantiateAbstractStatic
	 */
	public static function all(): array
	{
		/** @var static[] $entities */
		$entities = [];

		foreach (self::getRepository()->toArray() as $id => $data) {
			$entities[$id] = new static($id, $data);
		}

		return $entities;
	}

	/**
	 * @return array
	 */
	public static function rawAll(): array
	{
		return self::getRepository()->toArray();
	}

	/**
	 * Zählt die vorhandenen Instanzen dieses Models.
	 * @return int
	 */
	public static function count(): int
	{
		return count(self::rawAll());
	}

	/**
	 * Liest alle Einträge neu von der Festplatte oder ggf. dem Internet ein.
	 * @throws ConfigNotReadableException
	 */
	public static function reload(): void
	{
		self::getRepository()->reload();
	}

	public function __toString(): string
	{
		return sprintf('[Regulation %s: "%s"]', $this->getId(), $this->getName());
	}
}
