<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Exception\ConfigNotWritableException;

class Script extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-scripts';

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = true;

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = true;

	const CURRENT_HASH_VERSION = 2;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * @return static[]
	 * @phan-suppress PhanTypeMismatchArgument
	 */
	public static function all($ignoreHashVersion=false): array
	{
		if ($ignoreHashVersion) {
			return parent::all();
		}
		else {
			return array_filter(parent::all(), function (Script $obj) {
				return $obj->getHashVersion() == self::CURRENT_HASH_VERSION;
			});
		}
	}

	/**
	 * Zählt die vorhandenen Instanzen dieses Models.
	 * @return int
	 */
	public static function count($ignoreHashVersion=false): int
	{
		if ($ignoreHashVersion) {
			return parent::count();
		}
		return count(array_filter(self::getRepository()->toArray(), function ($item) {
			return (int)($item['hashVersion'] ?? 1) == self::CURRENT_HASH_VERSION;
		}));
	}

	/**
	 * @return static[]
	 */
	public static function byEnabled(bool $enabled=true): array
	{
		return array_filter(static::all(), function (Script $obj) use ($enabled) {
			return $obj->getEnableScript() == $enabled;
		});
	}

	/**
	 * @return string
	 */
	public function getContent(): string
	{
		return (string)$this->get('content');
	}

	/**
	 * @param string $url
	 * @return $this
	 */
	public function setContent(string $url): self
	{
		$this->setBinary('content', $url);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getFoundOnPages(): array
	{
		return $this->get('found_on_pages') ?? [];
	}

	/**
	 * @param string[] $pages
	 * @return $this
	 */
	public function setFoundOnPages(array $pages): self
	{
		$this->set('found_on_pages', $pages);
		return $this;
	}

	/**
	 * @param string $page
	 * @return $this
	 */
	public function addFoundOnPage(string $page): self
	{
		$this->appendToArray('found_on_pages', $page);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getCookies(): array
	{
		return $this->get('cookies') ?? [];
	}

	/**
	 * @param string[] $cookies
	 * @return $this
	 */
	public function setCookies(array $cookies): self
	{
		$this->set('cookies', $cookies);
		return $this;
	}

	/**
	 * @param string[] $cookies
	 * @return $this
	 */
	public function addCookies(array $cookies): self
	{
		$array = $this->get('cookies');
		if (!$array) {
			$array = [];
		}
		$array = array_unique(array_merge(
			$array,
			$cookies
		));

		$this->set('cookies', $array);
		return $this;
	}

	/**
	 * Diese Methode erstellt ein Objekt dieser Klasse
	 * @param string|null $id
	 * @return static
	 * @throws \LogicException
	 */
	public static function create(?string $id=null): Model
	{
		if ($id === null) {
			throw new \LogicException('cannot use create() with Script model');
		}
		self::beginWrite();
		$obj = new static(static::getRepository()->escapeKey($id));
		$obj->setHashVersion();
		return $obj;
	}

	/**
	 * Diese Methode löscht den referenzierten Datensatz aus der Konfiguration.
	 * @param string $id
	 * @param bool $autocommit
	 * @throws ConfigNotWritableException
	 */
	public static function delete(string $id, bool $autocommit=true): void
	{
		parent::delete(static::getRepository()->escapeKey($id), $autocommit);
	}
		/**
	 * @param string $id
	 * @return bool
	 */
	public static function exists(string $id): bool
	{
		return parent::exists(static::getRepository()->escapeKey($id));
	}

	/**
	 * @param string $id
	 * @return static|null
	 */
	public static function find(string $id): ?Model
	{
		return self::exists($id) ? new static(static::getRepository()->escapeKey($id)) : null;
	}

	/**
	 * @return bool
	 */
	public function getEnableScript(): bool
	{
		return (bool)$this->get('enableScript');
	}

	/**
	 * @param bool $enableScript
	 * @return $this
	 */
	public function setEnableScript(bool $enableScript): self
	{
		$this->set('enableScript', $enableScript);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getHashVersion(): int
	{
		return (int)($this->get('hashVersion') ?? 1);
	}

	/**
	 * @param int $version
	 * @return $this
	 */
	public function setHashVersion(int $version=2): self
	{
		$this->set('hashVersion', $version);
		return $this;
	}

}
