<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Component\Config\Store\JsonConfig;
use App\Component\Plugin\Manager as PluginManager;

class SubGroups extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-sub-groups';

	/** @var bool $userSpecificRepository */
	protected static $userSpecificRepository = true;

	/** @var bool $accountSingleton */
	protected static $accountSingleton = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * @return string
	 */
	public function getGroupname(): string
	{
		return (string)$this->get('groupname');
	}


	/**
	 * @param string $groupname
	 * @return $this
	 */
	public function setGroupname(string $groupname): self
	{
		$this->set('groupname', $groupname);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isSubAdmin(): bool
	{
		return (bool)$this->get('isSubAdmin');
	}

	/**
	 * @param bool $active
	 * @return $this
	 */
	public function setSubAdmin(bool $active): self
	{
		$this->set('isSubAdmin', $active);
		return $this;
	}

	/**
	 * @return array<string,(bool|string)>
	 */
	public function getGroupRights(): array
	{
		return (array)$this->get('groupRights',[]);
	}

	/**
	 * @param array<string,bool> $rights
	 * @return $this
	 */
	public function setGroupRights(array $rights): self
	{
		$this->set('groupRights', $rights);
		return $this;
	}

	/**
	 * @param $name
	 * @param $arguments
	 * @return array|string|bool|$this
	 */
	public function __call($name, $arguments)
	{
		return $this->get($name,"");
	}

	/**
	 * @param array $postData
	 * @return $this
	 */
	public function setArray(array $postData): self
	{
		foreach($postData as $k=>$v)
		{
			$this->set($k, $v);
		}
		return $this;
	}

	/**
	 * @return bool
	 * @see User::getAccountOwner()
	 */
	public static function createSubAdminStarterGroup(): bool
	{
		$groups = array_filter(self::all(), function (self $group) {
			return $group->isSubAdmin();
		});

		if(empty($groups))
		{
			//Alle Menüpunkte
			$menu = MenuData::all();
			$rights = array();
			foreach ($menu as $menuitem)
			{
				$rights[$menuitem->getId()]="on";
			}

			//Alle Plugins
			$pluginManager = new PluginManager();
			$user = User::loggedInUser();
			$owner = $user->getAccountOwner();

			$plugins = array_filter(
				$pluginManager->configuredPlugins(),
				function ($plugin) use ($owner) {
					return $plugin->getState()->isAllowedForUser($owner) and !$plugin->isHidden();
				}
			);

			foreach ($plugins as $plugin) {
				$rights[$plugin->getSubgroupRightsKey()] = "on";
			}

			//Setzen und Speichern...
			$newgroup = self::create()
				->setGroupname('SubAdmin')
				->setSubAdmin(true)
			;

			$newgroup->set("groupRights",(object)$rights);
			$newgroup->flush();

			return true;
		}
		return false;
	}

	/**
	 * Diese Methode prüft, ob das "accountSingleton"- und "userSpecific"-Repository
	 * Einträge enthält. Sofern keine Einträge existieren, wird versucht, Einträge aus
	 * dem alten Repository (var/cm-sub-groups.json) einzulesen und zu migrieren.
	 */
	public static function migrateLegacyRepository(): void
	{
		$user = User::loggedInUser();
		if (!$user || self::count() > 0) {
			return;
		}

		$user = $user->getAccountOwner();

		try {
			$legacyRepo = new JsonConfig(self::$repositoryFilename.'.json', false);
		} catch (\Exception $exception) {
			echo $exception->getMessage();
			exit;
		}

		$migratedGroups = array_values(array_map(function (array $data) {
			unset($data['subUserId']);

			$group = self::create();
			$group->setArray($data);

			return $group;
		}, array_filter($legacyRepo->toArray(), function ($data) use ($user) {
			return ($data['subUserId'] ?? '') == $user->getId();
		})));

		if (count($migratedGroups) > 0) {
			self::flush();
		}
	}
}
