<?php

namespace App\Model;

use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Component\Tcf\v2\Model\Feature as TcfFeature;
use App\Component\Tcf\v2\Model\SpecialPurpose as TcfSpecialPurpose;
use App\Component\Tcf\v2\Model\Translation\Feature as TranslatedTcfFeature;
use App\Component\Tcf\v2\Model\Translation\Purpose as TranslatedTcfPurpose;
use App\Component\Tcf\v2\Model\Translation\SpecialFeature as TranslatedTcfSpecialFeature;
use App\Component\Tcf\v2\Model\Translation\SpecialPurpose as TranslatedTcfSpecialPurpose;
use App\Component\Config\RepositoryInterface;
use App\Config;

/**
 * Systemweite Embedding-Vorlagen
 */
class SystemEmbedding extends EmbeddingBase
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Diese Methode sucht nach Embedding-Vorlagen und gibt ein Ergebnis-Array zurück.
	 * @param string $search Suchbegriff
	 * @return SystemEmbedding[]
	 */
	public static function search(string $search): array
	{
		$languages = Config::getInstance()->getSystemEmbeddingLanguages();

		return array_filter(self::all(), function (SystemEmbedding $e) use ($search, $languages) {
			// Grundfelder des Embeddings durchgehen
			if (stripos($e->getName(), $search) !== false or
				stripos(implode("\n", $e->getCodeBlock()), $search) !== false or
				stripos($e->getDescription(), $search) !== false or
				stripos($e->getEmbeddingVendor(), $search) !== false or
				stripos($e->getCollectedDataInfo(), $search) !== false or
				stripos($e->getPurposeOfDataCollection(), $search) !== false or
				stripos($e->getLegalBasis(), $search) !== false or
				stripos($e->getPlaceOfProcessing(), $search) !== false or
				stripos($e->getPrivacyPolicyUrl(), $search) !== false
			) {
				return true;
			}

			// Assets durchgehen
			foreach ($e->getAssets() as $a) {
				if (stripos($a->getName(), $search) !== false or
					stripos($a->getDescription(), $search) !== false or
					stripos($a->getLifetime(), $search) !== false
				) {
					return true;
				}
			}

			// Übersetzungen durchgehen
			foreach ($languages as $lang) {
				if (stripos($e->getTranslatedName($lang, false), $search) !== false or
					stripos($e->getTranslatedDescription($lang, false), $search) !== false or
					stripos($e->getTranslatedCollectedDataInfo($lang, false), $search) !== false or
					stripos($e->getTranslatedLegalBasis($lang, false), $search) !== false or
					stripos($e->getTranslatedPlaceOfProcessing($lang, false), $search) !== false or
					stripos($e->getTranslatedPrivacyPolicyUrl($lang, false), $search) !== false or
					stripos($e->getTranslatedPurposeOfDataCollection($lang, false), $search) !== false
				) {
					return true;
				}

				// Übersetzungen der Assets durchgehen
				foreach ($e->getAssets() as $a) {
					if (stripos($a->getTranslatedDescription($lang), $search) !== false or
						stripos($a->getTranslatedLifetime($lang), $search) !== false
					) {
						return true;
					}
				}

				return false;
			}
		});
	}

	/**
	 * @return SystemEmbeddingAsset[]
	 */
	public function getAssets(): array
	{
		if (empty($this->assets)) {
			$this->assets = array_values(array_filter(array_map(function ($id) {
				return SystemEmbeddingAsset::find($id);
			}, $this->get('assets', []))));
		}

		return $this->assets;
	}

	/**
	 * @param string $assetId
	 * @return static
	 */
	public function removeAsset(string $assetId): EmbeddingBase
	{
		parent::removeAsset($assetId);
		SystemEmbeddingAsset::delete($assetId, false);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedName(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_name.'.$language->getCode());
		if ($useFallback and ($result === null or $result === '')) {
			return $this->getName();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $label
	 * @return $this
	 */
	public function setTranslatedName(Language $language, string $label): self
	{
		$this->set('translated_name.'.$language->getCode(), $label);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedDescription(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_description.'.$language->getCode());
		if ($useFallback and $result === null) {
			return $this->getDescription();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $description
	 * @return $this
	 */
	public function setTranslatedDescription(Language $language, string $description): self
	{
		$this->set('translated_description.'.$language->getCode(), $description);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPrivacyPolicyUrl(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_privacy_policy_url.'.$language->getCode());
		if ($useFallback and $result === null) {
			return $this->getPrivacyPolicyUrl();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $url
	 * @return $this
	 */
	public function setTranslatedPrivacyPolicyUrl(Language $language, string $url): self
	{
		$this->set('translated_privacy_policy_url.'.$language->getCode(), trim($url));
		return $this;
	}

	/**
	 * @param Language $language
	 * @return TcfPurpose[]|TranslatedTcfPurpose[]
	 */
	public function getTranslatedTcfPurposes(Language $language): array
	{
		$langCode = $language->getShortCode();
		return array_map(function (TcfPurpose $purpose) use ($langCode) {
			return $purpose->getTranslation($langCode);
		}, $this->getTcfPurposes());
	}

	/**
	 * @param Language $language
	 * @return TcfSpecialPurpose[]|TranslatedTcfSpecialPurpose[]
	 */
	public function getTranslatedTcfSpecialPurposes(Language $language): array
	{
		$langCode = $language->getShortCode();
		return array_map(function (TcfSpecialPurpose $purpose) use ($langCode) {
			return $purpose->getTranslation($langCode);
		}, $this->getTcfSpecialPurposes());
	}

	/**
	 * @param Language $language
	 * @return TcfFeature[]|TranslatedTcfFeature[]
	 */
	public function getTranslatedTcfFeatures(Language $language): array
	{
		$langCode = $language->getShortCode();
		return array_map(function (TcfFeature $feature) use ($langCode) {
			return $feature->getTranslation($langCode);
		}, $this->getTcfFeatures());
	}

	/**
	 * @param Language $language
	 * @return TcfSpecialFeature[]|TranslatedTcfSpecialFeature[]
	 */
	public function getTranslatedTcfSpecialFeatures(Language $language): array
	{
		$langCode = $language->getShortCode();
		return array_map(function (TcfSpecialFeature $feature) use ($langCode) {
			return $feature->getTranslation($langCode);
		}, $this->getTcfSpecialFeatures());
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @deprecated
	 * @return string
	 */
	public function getTranslatedLifetime(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_lifetime.'.$language->getCode());
		if ($useFallback and $result === null) {
			return $this->getLifetime();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $lifetime
	 * @return $this
	 */
	public function setTranslatedLifetime(Language $language, string $lifetime): self
	{
		$this->set('translated_lifetime.'.$language->getCode(), $lifetime);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedCollectedDataInfo(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_collected_data_info.'.$language->getCode());
		if ($useFallback and $result === null) {
			return $this->getCollectedDataInfo();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $info
	 * @return $this
	 */
	public function setTranslatedCollectedDataInfo(Language $language, string $info): self
	{
		$this->set('translated_collected_data_info.'.$language->getCode(), $info);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPurposeOfDataCollection(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_purpose_of_data_collection.'.$language->getCode());
		if ($useFallback and $result === null) {
			return $this->getPurposeOfDataCollection();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $purpose
	 * @return $this
	 */
	public function setTranslatedPurposeOfDataCollection(Language $language, string $purpose): self
	{
		$this->set('translated_purpose_of_data_collection.'.$language->getCode(), $purpose);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedLegalBasis(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_legal_basis.'.$language->getCode());
		if ($useFallback and $result === null) {
			return $this->getLegalBasis();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $basis
	 * @return $this
	 */
	public function setTranslatedLegalBasis(Language $language, string $basis): self
	{
		$this->set('translated_legal_basis.'.$language->getCode(), $basis);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedLegalBasisSupplement(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_legal_basis_supplement.'.$language->getCode());
		if ($useFallback and $result === null) {
			return $this->getLegalBasisSupplement();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $basis
	 * @return $this
	 */
	public function setTranslatedLegalBasisSupplement(Language $language, string $basis): self
	{
		$key = 'translated_legal_basis_supplement.'.$language->getCode();
		$this->set($key, $basis);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPlaceOfProcessing(Language $language, $useFallback=true): string
	{
		$result = $this->get('translated_place_of_processing.'.$language->getCode());
		if ($useFallback and $result === null) {
			return $this->getPlaceOfProcessing();
		}
		return (string)$result;
	}

	/**
	 * @param Language $language
	 * @param string $place
	 * @return $this
	 */
	public function setTranslatedPlaceOfProcessing(Language $language, string $place): self
	{
		$this->set('translated_place_of_processing.'.$language->getCode(), $place);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentTitle(Language $language, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_title.{$language->getCode()}");
		if (strlen($value) == 0 && $useFallback and $locale = UserLocale::findByLanguage($language)) {
			$value = $locale->getBlockedContentTitle();
		}
		return $value;
	}

	/**
	 * @param Language $language
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentTitle(Language $language, string $value): self
	{
		$value = trim($value);
		$this->set("translated_blocked_content_title.{$language->getCode()}", $value);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentText(Language $language, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_text.{$language->getCode()}");
		if (strlen($value) == 0 && $useFallback and $locale = UserLocale::findByLanguage($language)) {
			$value = $locale->getBlockedContentText();
		}
		return $value;
	}

	/**
	 * @param Language $language
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentText(Language $language, string $value): self
	{
		$value = trim($value);
		$this->set("translated_blocked_content_text.{$language->getCode()}", $value);
		return $this;
	}

	/**
	 * @param Language $language
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentButtonText(Language $language, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_button_text.{$language->getCode()}");
		if (strlen($value) == 0 && $useFallback and $locale = UserLocale::findByLanguage($language)) {
			$value = $locale->getBlockedContentButtonText();
		}
		return $value;
	}

	/**
	 * @param Language $language
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentButtonText(Language $language, string $value): self
	{
		$value = trim($value);
		$this->set("translated_blocked_content_button_text.{$language->getCode()}", $value);
		return $this;
	}

	/**
	 * Setzt alle übersetzten Texte, die in $language noch nie vorher gesetzt
	 * wurden, auf den nicht-übersetzten Wert.
	 *
	 * @param Language $language Zu füllendes Language
	 * @return $this
	 */
	public function fillTranslationsFromDefault(Language $language): self
	{
		$this->setTranslatedCollectedDataInfo($language, $this->getTranslatedCollectedDataInfo($language));
		$this->setTranslatedDescription($language, $this->getTranslatedDescription($language));
		$this->setTranslatedLegalBasis($language, $this->getTranslatedLegalBasis($language));
		$this->setTranslatedLifetime($language, $this->getTranslatedLifetime($language));
		$this->setTranslatedPlaceOfProcessing($language, $this->getTranslatedPlaceOfProcessing($language));
		$this->setTranslatedPrivacyPolicyUrl($language, $this->getTranslatedPrivacyPolicyUrl($language));
		$this->setTranslatedPurposeOfDataCollection($language, $this->getTranslatedPurposeOfDataCollection($language));
		$this->setTranslatedBlockedContentTitle($language, $this->getTranslatedBlockedContentTitle($language));
		$this->setTranslatedBlockedContentText($language, $this->getTranslatedBlockedContentText($language));
		$this->setTranslatedBlockedContentButtonText($language, $this->getTranslatedBlockedContentButtonText($language));
		$this->setTranslatedLegalBasisSupplement($language, $this->getTranslatedLegalBasisSupplement($language));
		return $this;
	}

}
