<?php

namespace App\Model;

use App\Component\LoggerInterface;
use App\Component\LogStream\RepositoryInterface;
use App\Component\LogStream\Store\JsonStore;
use App\Component\LogStream\Store\MongoDb;
use MongoDB\BSON\ObjectId;
use App\Utils;
use DateTime;
use Iterator;
use Throwable;

class SystemLog implements LoggerInterface
{
	/** @var string $filestemTemplate */
	private static $filestemTemplate = 'log/cm-system-%s';

	/** @var string $collection */
	private static $collection = 'cm-system-log';

	/** @var array $filter */
	private $filter;

	/** @var RepositoryInterface $repository */
	private $repository;

	/**
	 * @param DateTime|null $date
	 */
	public function __construct(?DateTime $date)
	{
		if ($date === null) {
			$date = new DateTime('now');
		}

		$repositoryStore = Utils::getRepositoryStore();
		switch ($repositoryStore) {
			case Utils::REPOSITORY_STORE_MONGODB:
				$repository = new MongoDb(null, null, self::$collection);
				$start = (clone $date)->modify('first day of this month midnight')->getTimestamp();
				$end = (clone $date)->modify('first day of next month midnight')->getTimestamp();
				$this->filter = [
					'_id' => [
						'$gte' => new ObjectId(substr(str_pad(dechex($start), 8, '0', STR_PAD_LEFT), -8, 8).'0000000000000000'),
						'$lt' => new ObjectId(substr(str_pad(dechex($end), 8, '0', STR_PAD_LEFT), -8, 8).'FFFFFFFFFFFFFFFF'),
					]
				];
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$path = sprintf(self::$filestemTemplate, $date->format('Y-m'));
				$repository = new JsonStore(null, null, $path);
				$this->filter = [];
				break;
		}

		$this->repository = $repository;
	}

	/**
	 * @param DateTime $date
	 * @return self
	 */
	public static function byDate(DateTime $date): self
	{
		return new self($date);
	}

	public static function listDateRanges(): array
	{
		$repositoryStore = Utils::getRepositoryStore();
		switch ($repositoryStore) {
			case Utils::REPOSITORY_STORE_MONGODB:
				$model = self::current();
				assert($model->repository instanceof MongoDb);
				$cursor = $model->repository->aggregate([
					['$group' => [
						'_id' => [
							'year' => ['$year' => '$_id'],
							'month' => ['$month' => '$_id'],
						],
					]],
					['$sort' => [
						'_id.year' => -1,
						'_id.month' => -1
					]],
				]);

				return array_map(function ($document) {

					return new DateTime(sprintf('%d-%d-01 12:00:00',
						$document['_id']['year'],
						$document['_id']['month']
					));
				}, iterator_to_array($cursor, false));
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$result = [];
				foreach (glob(Utils::getBundledVarDir().'/log/cm-system-????-??.log') as $file)
				{
					$date = substr(basename($file), 10, -4). ' 12:00:00';
					$result[] = new Datetime($date);
				}
				return array_reverse($result);
		}
	}

	/**
	 * Löscht einen Datumsbereich (zur Zeit einen Monat) aus dem Log.
	 *
	 * WARNUNG: Egal, wo in dem Monat das Datum liegt, es wird immer ein ganzer Monat gelöscht!
	 *
	 * @param DateTime $dateInRange Datum im zu löschenden Bereich (z.B. von `listDateRanges()`)
	 */
	public static function deleteDateRange(DateTime $dateInRange): void
	{
		$repositoryStore = Utils::getRepositoryStore();
		switch ($repositoryStore) {
			case Utils::REPOSITORY_STORE_MONGODB:
				$instance = (new static($dateInRange));
				assert($instance->repository instanceof MongoDb);
				$instance->repository->deleteByFilter($instance->filter);
				break;
			case Utils::REPOSITORY_STORE_FILE:
			default:
				$instance = (new static($dateInRange));
				$instance->repository->clear();
				break;
		}
	}

	/**
	 * @return self
	 */
	public static function current(): self
	{
		return new self(null);
	}

	/**
	 * @return Iterator<array>
	 */
	public function all(): Iterator
	{
		if ($this->filter) {
			return $this->repository->filter($this->filter, true);
		} else {
			return $this->repository->all(true);
		}
	}

	/**
	 * @param string $search Substring-Suchbegriff
	 * @return Iterator
	 */
	public function search(string $search): Iterator
	{
		$filter = ['msg'=>['$regex'=>preg_quote($search)]];
		if ($this->filter) {
			$filter = [ '$and' => [ $this->filter, $filter ] ];
		}
		return $this->repository->filter($filter, true);
	}


	public function count(): int
	{
		return $this->repository->count();
	}

	/**
	 * @param string $severity Schweregrad (Warnung/Fehler/etc.)
	 * @param string $component Betroffene Komponente
	 * @param string $message Nachricht
	 */
	private function append(string $severity, string $component, string $message): void
	{
		$timestamp = new DateTime('now');
		$this->repository->append([
			'ts'=>$timestamp->format(DateTime::ISO8601),
			'cm'=>$component,
			'sev'=>$severity,
			'msg'=> $message,
		]);
	}

	/**
	 * @param string $component Betroffene Komponente
	 * @param string $message Nachricht
	 */
	public function logError(string $component, string $message): void
	{
		$this->append('ERROR', $component, $message);
	}

	/**
	 * @param string $component Betroffene Komponente
	 * @param string $message Nachricht
	 */
	public function logWarning(string $component, string $message): void
	{
		$this->append('WARNING', $component, $message);
	}

	/**
	 * @param string $component Betroffene Komponente
	 * @param string $message Nachricht
	 */
	public function logInfo(string $component, string $message): void
	{
		$this->append('INFO', $component, $message);
	}

	/**
	 * @param string $component Betroffene Komponente
	 * @param string $message Nachricht
	 */
	public function logSuccess(string $component, string $message): void
	{
		$this->append('SUCCESS', $component, $message);
	}

	public function logException(string $component, Throwable $exception, string $message=""): void
	{
		$exceptionMessage = array_reverse(explode('\\', get_class($exception)))[0];
		if ($exception->getCode()) {
			$exceptionMessage .= '['.$exception->getCode().']';
		}
		$exceptionMessage .= ': '. $exception->getMessage();

		$fullMessage = ($message == '') ? $exceptionMessage : $message . ' ' . $exceptionMessage;

		$this->append('ERROR', $component, $fullMessage);
	}

}
