<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Component\Media\MediaFileService;
use App\Utils;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\HttpFoundation\ParameterBag;

class Theme extends ThemeBase
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var bool $domainSpecificRepository */
	protected static $domainSpecificRepository = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Diese Methode wird ausgeführt, sobald das Repository durch die Basisklasse Model geladen wurde.
	 */
	protected static function onRepositoryLoaded(): void
	{
		$domain = Domain::activeDomain();

		if (User::count() == 1) {
			// Updateroutine: Themelogos aus Single-User-Env in domainspezifische Konfigurationsstruktur migrieren
			foreach (self::all() as $theme) {
				/** @var self $theme */
				$basename = $theme->getLogoIdentifier();
				if (empty($basename)) {
					continue;
				}

				$sourceFile = Utils::getBaseDir() . '/public/img/logos/'.$basename;
				if (is_file($sourceFile) == false) {
					continue;
				}

				if ($theme->getLogo() === null) {
					$theme->setLogo(new File($sourceFile));
				}
			}
		}
	}

	/**
	 * Schließt das aktive Theme-Repository.
	 * @param bool $flush
	 */
	public static function closeRepository(bool $flush = false): void
	{
		if ($flush && static::$repository) {
			static::flush();
		}
		static::$repository = null;
	}

	/**
	 * Wechselt zum Theme-Repository einer bestimmten Benutzer-Domain-Kombination.
	 *
	 * @param User $user
	 * @param Domain $domain
	 * @param bool $flush Ob Änderungen im vorherigen Repository gespeichert werden sollen
	 */
	public static function switchRepository(User $user, Domain $domain, bool $flush = false): void
	{
		self::closeRepository($flush);
		static::$repository = static::openRepository($user, $domain);
	}

	/**
	 * @return static[]
	 */
	public static function systemThemes(): array
	{
		return array_filter(self::all(), function ($theme) {
			/** @var self $theme */
			return $theme->getSystemTheme() !== null;
		}, 0);
	}

	/**
	 * @param SystemTheme $systemTheme
	 * @return static|null
	 */
	public static function bySystemTheme(SystemTheme $systemTheme): ?self
	{
		return array_reduce(self::systemThemes(), function ($theme, $currentTheme) use ($systemTheme) {
			/**
			 * @var self|null $theme
			 * @var self $currentTheme
			 * @var SystemTheme|null $currentSystemTheme
			 */
			$currentSystemTheme = $currentTheme->getSystemTheme();

			/** @var self $currentTheme */
			return $theme ?? ($currentSystemTheme && $currentSystemTheme->getId() == $systemTheme->getId() ? $currentTheme : $theme);
		}, null);
	}

	/**
	 * @return SystemTheme|null
	 */
	public function getSystemTheme(): ?SystemTheme
	{
		return SystemTheme::find((string)$this->get('systemThemeId'));
	}

	/**
	 * Diese Methode übernimmt sämtliche Eigenschaften des gegebenen System-Themes.
	 * @param SystemTheme|null $theme
	 * @return $this
	 */
	public function setSystemTheme(?SystemTheme $theme): self
	{
		if ($theme) {
			$this->setUserTheme(null);
			$this->importTheme($theme);
		}

		$this->set('systemThemeId', $theme ? $theme->getId() : null);
		return $this;
	}

	/**
	 * @return bool
	 */
	private function referencesSystemTheme(): bool
	{
		return (bool)$this->get('systemThemeId');
	}

	/**
	 * @return static[]
	 */
	public static function userThemes(): array
	{
		return array_filter(self::all(), function ($theme) {
			/** @var self $theme */
			return $theme->getUserTheme() !== null;
		}, 0);
	}

	/**
	 * @param UserTheme $userTheme
	 * @return static|null
	 */
	public static function byUserTheme(UserTheme $userTheme): ?self
	{
		return array_reduce(self::userThemes(), function ($theme, $currentTheme) use ($userTheme) {
			/**
			 * @var self|null $theme
			 * @var self $currentTheme
			 * @var UserTheme|null $currentUserTheme
			 */
			$currentUserTheme = $currentTheme->getUserTheme();

			/** @var self $currentTheme */
			return $theme ?? ($currentUserTheme && $currentUserTheme->getId() == $userTheme->getId() ? $currentTheme : $theme);
		}, null);
	}

	/**
	 * @return UserTheme|null
	 */
	public function getUserTheme(): ?UserTheme
	{
		return UserTheme::find((string)$this->get('userThemeId'));
	}

	/**
	 * Diese Methode übernimmt sämtliche Eigenschaften des gegebenen User-Themes.
	 * @param UserTheme|null $theme
	 * @return $this
	 */
	public function setUserTheme(?UserTheme $theme): self
	{
		if ($theme) {
			$this->setSystemTheme(null);
			$this->importTheme($theme);
		}

		$this->set('userThemeId', $theme ? $theme->getId() : null);
		return $this;
	}

	/**
	 * @return bool
	 */
	private function referencesUserTheme(): bool
	{
		return (bool)$this->get('userThemeId');
	}

	/**
	 * @return bool Whether `systemThemeId` or `userThemeId` points to a non-existing object.
	 */
	public function isDangling(): bool
	{
		return $this->referencesSystemTheme() && $this->getSystemTheme() === null
			|| $this->referencesUserTheme() && $this->getUserTheme() === null;
	}

	/**
	 * @return SystemTheme|UserTheme|null
	 */
	public function getParentTheme(): ?ThemeBase
	{
		return $this->getSystemTheme() ?? $this->getUserTheme();
	}

	/**
	 * Diese Methode verarbeitet die URL-Parameter `preview`, `systemTheme`, `userTheme` und `theme`.
	 *
	 * Als Ergebnis wird entweder das angeforderte `Theme` geliefert oder `null`, falls es nicht
	 * gefunden wird.
	 *
	 * Falls ein `SystemTheme` oder `UserTheme` im `preview`-Modus angefordert wird, liefert die
	 * Methode ein temporäres `Theme`, das die Werte der Vorlage enthält.
	 *
	 * Falls ein `SystemTheme` oder `UserTheme` im Live-Modus angefordert wird, liefert die
	 * Methode das referenzierende `Theme`.
	 *
	 * **Achtung:** Bei `preview=true` wird das Repository `volatile`.
	 *
	 * @param ParameterBag $query
	 * @param array $apiParams [Referenz]
	 * @return static|null
	 */
	public static function deriveFromQuery(ParameterBag $query, array &$apiParams=[]): ?self
	{
		if ($query->getBoolean('preview')) {
			// In preview mode, it is not desirable to fall back on the active theme
			return self::makePreviewTheme($query, $apiParams);
		}
		elseif ($query->has('systemTheme') and $systemTheme = SystemTheme::find($query->getAlnum('systemTheme'))) {
			$theme = self::bySystemTheme($systemTheme);
		}
		elseif ($query->has('userTheme') and $userTheme = UserTheme::find($query->getAlnum('userTheme'))) {
			$theme = self::byUserTheme($userTheme);
		}
		elseif ($query->has('theme')) {
			$theme = self::find($query->getAlnum('theme'));
		}

		// In live mode, fall back on the active theme
		$theme = $theme ?? self::activeTheme();

		if ($theme) {
			$apiParams['theme'] = $theme->getId();
		}

		return $theme;
	}

	/**
	 * @param ParameterBag $query
	 * @param array $apiParams [Referenz]
	 * @return static|null
	 */
	private static function makePreviewTheme(ParameterBag $query, array &$apiParams=[]): ?self
	{
		$preview = self::createNonPersistent();

		if ($query->has('systemTheme') and $systemTheme = SystemTheme::find($query->getAlnum('systemTheme'))) {
			$apiParams['systemTheme'] = $systemTheme->getId();
			$preview->setSystemTheme($systemTheme);
		}
		elseif ($query->has('userTheme') and $userTheme = UserTheme::find($query->getAlnum('userTheme'))) {
			$apiParams['userTheme'] = $userTheme->getId();
			$preview->setUserTheme($userTheme);
		}
		elseif ($query->has('theme') and $theme = self::find($query->getAlnum('theme'))) {
			$apiParams['theme'] = $theme->getId();
			$preview = $theme;
		}
		else {
			// In preview mode, it is not desirable to fall back on the active theme
			$preview = null;
		}

		$apiParams['preview'] = 'true';

		return $preview;
	}

	/**
	 * @return self[]
	 */
	public static function allOrderedByTypeAndName(): array
	{
		$themes = self::all();

		// ORDER BY systemTheme, userTheme, name
		usort($themes, function (self $a, self $b) {
			if ($a->getSystemTheme() xor $b->getSystemTheme()) {
				return $a->getSystemTheme() ? -1 : 1;
			} elseif ($a->getUserTheme() xor $b->getUserTheme()) {
				return $a->getUserTheme() ? -1 : 1;
			} else {
				return strnatcasecmp($a->getName(), $b->getName());
			}
		});

		return $themes;
	}

	/**
	 * @return bool
	 */
	public function getIsDeleted(): bool
	{
		return (bool)$this->get('isdeleted');
	}

	/**
	 * @param bool $deleted
	 * @return $this
	 */
	public function setIsDeleted(bool $deleted = true): self
	{
		$this->set('isdeleted', $deleted);
		return $this;
	}


	/**
	 * @return string
	 */
	public function getLogoIdentifier(): string
	{
		return (string)$this->get('logoFilename');
	}

	/**
	 * @param string $logoFilename
	 * @return $this
	 */
	protected function setLogoIdentifier(string $logoFilename): self
	{
		$this->set('logoFilename', $logoFilename);
		return $this;
	}

	public function getLogo(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getDomainRepository('logos');
		$identifier = $this->getLogoIdentifier();
		return ($identifier == '') ? null : $repo->getFile($identifier);
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setLogo(?File $file): self
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getDomainRepository('logos');
		$prevIdentifier = $this->getLogoIdentifier();

		$identifier = $repo->replaceFile($prevIdentifier, $file, (string)$this->getId());
		$this->setLogoIdentifier($identifier);

		return $this;
	}

	/**
	 * @return bool
	 */
	public function isLogoDisplayed(): bool
	{
		return (bool)$this->get('logoDisplayed');
	}

	/**
	 * @param bool $logoDisplayed
	 * @return $this
	 */
	public function setLogoDisplayed(bool $logoDisplayed): self
	{
		$this->set('logoDisplayed', $logoDisplayed);
		return $this;
	}

	public function getSettingsIcon(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getDomainRepository('settings_icons');
		$identifier = $this->getSettingsIconIdentifier();
		return ($identifier == '') ? null : $repo->getFile($identifier);
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setSettingsIcon(?File $file): ThemeBase
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getDomainRepository('settings_icons');
		$prevIdentifier = $this->getSettingsIconIdentifier();

		$identifier = $repo->replaceFile($prevIdentifier, $file, (string)$this->getId());
		$this->setSettingsIconIdentifier($identifier);

		return $this;
	}

	/**
	 * @return File|null
	 */
	public function getSettingsIconFromParentTheme(): ?File
	{
		/** @var UserTheme|SystemTheme $parentTheme */
		$parentTheme = $this->getParentTheme();
		return $parentTheme ? $parentTheme->getPublishedSettingsIcon() : null;
	}

	/**
	 * @return string
	 */
	public function getSettingsIconBackgroundColor(): string
	{
		/** @var string|null $color */
		$color = $this->get('settingsIconBackgroundColor');

		// Check for `null` because the old implementation would use that value to fall back on the parent's value
		if ($color === null) {
			if ($parentTheme = $this->getParentTheme()) {
				$color = $parentTheme->getPublishedSettingsIconBackgroundColor() ?? $parentTheme->getSettingsIconBackgroundColor();
			}
			else {
				$color = parent::getSettingsIconBackgroundColor();
			}
		}

		assert(is_string($color));
		return $color;
	}

	/**
	 * @return string
	 */
	public function getSettingsIconForegroundColor(): string
	{
		/** @var string|null $color */
		$color = $this->get('settingsIconForegroundColor');

		// Check for `null` because the old implementation would use that value to fall back on the parent's value
		if ($color === null) {
			if ($parentTheme = $this->getParentTheme()) {
				$color = $parentTheme->getPublishedSettingsIconForegroundColor() ?? $parentTheme->getSettingsIconForegroundColor();
			}
			else {
				$color = parent::getSettingsIconForegroundColor();
			}
		}

		assert(is_string($color));
		return $color;
	}

	/**
	 * @return bool
	 */
	public function getWhitelabel(): bool
	{
		return (bool)$this->get('whitelabel');
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setWhitelabel(bool $value): self
	{
		$this->set('whitelabel', $value);
		return $this;
	}

	private function domainHasOptionalEmbeddings(Domain $domain): bool
	{
		if ($domain->getManagementStructure() == Domain::MANAGEMENT_STRUCTURE_EMBEDDING) {
			return count(array_filter(Embedding::optional(), function ($e) { return $e->isActive(); })) > 0;
		} else {
			// Fallback für alte Cookie-Infrastruktur
			return true;
		}
	}

	/**
	 * @override
	 * @return bool
	 */
	public function hasDeclineButton(): bool
	{
		$domain = Domain::activeDomain();
		if ($domain and $domain->isDeclineButtonEnforced() and $this->hasAcceptAllButton() and $this->domainHasOptionalEmbeddings($domain)) {
			return true;
		}
		return parent::hasDeclineButton();
	}

	/**
	 * @override
	 * @param string[] $wantedButtons
	 * @return string[]
	 */
	protected function listMandatoryButtons(array $wantedButtons): array
	{
		$values = parent::listMandatoryButtons($wantedButtons);

		// Ablehnen-Button erzwingen, wenn "Alles akzeptieren" drin und nicht explizit erlaubt
		$domain = Domain::activeDomain();
		if ($domain and $domain->isDeclineButtonEnforced() and in_array('accept', $wantedButtons) and $this->domainHasOptionalEmbeddings($domain)) {
			$values[] = 'decline';
		}

		return $values;
	}
}
