<?php

namespace App\Model;

use App\Utils;
use LogicException;
use Symfony\Component\HttpFoundation\File\File;

abstract class ThemeBase extends Model
{
	const SETTINGS_ICON_TARGET_MAIN = 'main';
	const SETTINGS_ICON_TARGET_SETTINGS = 'purpose';

	const ALLOWED_SETTINGS_ICON_TARGET_VALUES = [
		self::SETTINGS_ICON_TARGET_MAIN,
		self::SETTINGS_ICON_TARGET_SETTINGS,
	];

	const WIDGET_BUTTON_ACCEPT = 'accept';
	const WIDGET_BUTTON_DECLINE = 'decline';
	const WIDGET_BUTTON_SETTINGS = 'cpanel';

	const ALLOWED_WIDGET_BUTTON_VALUES = [
		self::WIDGET_BUTTON_ACCEPT,
		self::WIDGET_BUTTON_DECLINE,
		self::WIDGET_BUTTON_SETTINGS,
	];

	const WIDGET_POSITION_TOP = 'top';
	const WIDGET_POSITION_CENTER = 'center';
	const WIDGET_POSITION_BOTTOM = 'bottom';
	const WIDGET_POSITION_BOTTOM_LEFT = 'bottomleft';
	const WIDGET_POSITION_BOTTOM_RIGHT = 'bottomright';

	const ALLOWED_WIDGET_POSITION_VALUES = [
		self::WIDGET_POSITION_TOP,
		self::WIDGET_POSITION_CENTER,
		self::WIDGET_POSITION_BOTTOM,
		self::WIDGET_POSITION_BOTTOM_LEFT,
		self::WIDGET_POSITION_BOTTOM_RIGHT,
	];

	/** @var string[] Ländercodes (neben "EU" und "world") für die Einschränkungen
	 *                festgelegt werden können. (Aktuell nur für den Ablehnen-Button)
	 */
	const RESTRICTABLE_COUNTRIES = ["CH"];

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-themes';

	/** @var bool $wantsMongoDbMtimeIndex */
	protected static $wantsMongoDbMtimeIndex = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * @return static|null
	 */
	public static function activeTheme(): ?self
	{
		return array_reduce(static::all(), function ($theme, $currentTheme) {
			/** @var self $currentTheme */
			return $theme ?? ($currentTheme->isDefaultTheme() ? $currentTheme : $theme);
		}, null);
	}

	/**
	 * @param ThemeBase $theme
	 * @return $this
	 */
	public function importTheme(ThemeBase $theme): self
	{
		if ($buttons = $theme->getWidgetButtons()) {
			$this->setWidgetButtons($buttons);
		}

		// Don't copy, for example, from UserTheme to Theme as there is a fallback procedure in place.
		if ($theme instanceof $this) {
			$this
				->setSettingsIcon($theme->getSettingsIcon())
				->setSettingsIconBackgroundColor($theme->getSettingsIconBackgroundColor())
				->setSettingsIconForegroundColor($theme->getSettingsIconForegroundColor())
			;
		}

		return $this
			->touchLastModifiedTimestamp()
			->setName($theme->getName())
			->setBaseLayout($theme->getBaseLayout())
			->setPosition($theme->getPosition())
			->setSelectType($theme->getSelectType())
			->setShowPurposesInMainWindow($theme->isShowPurposesInMainWindow())
			->setBlocking($theme->isBlocking())
			->setShowEmbeddingsInControlPanel($theme->isShowEmbeddingsInControlPanel())
			->setShowAcceptAllButtonInControlPanel($theme->isShowAcceptAllButtonInControlPanel())
			->setShowWithdrawAllButtonInControlPanel($theme->isShowWithdrawAllButtonInControlPanel())
			->setShowCheckAndUncheckAllButtonInControlPanel($theme->isShowCheckAndUncheckAllButtonInControlPanel())
			->setShowExtraImprintFooter($theme->isShowExtraImprintFooter())
			->setSaveButtonInDetailsWindow($theme->hasSaveButtonInDetailsWindow())
			->setSettingsIconEnabled($theme->isSettingsIconEnabled())
			->setSettingsIconTarget($theme->getSettingsIconTarget())
			->setComplyingDoNotTrack($theme->isComplyingDoNotTrack())
			->setOnlyInEu($theme->getOnlyInEu())
			->setManipulationPrevention($theme->getManipulationPrevention())
			->setCloseButton($theme->hasCloseButton())
			->setDeclineButtonRestrictions($theme->getDeclineButtonRestrictions())
			->setOkButtonBackgroundColor($theme->getOkButtonBackgroundColor())
			->setOkButtonColor($theme->getOkButtonColor())
			->setOkButtonBorderColor($theme->getOkButtonBorderColor())
			->setDeclineButtonBackgroundColor($theme->getDeclineButtonBackgroundColor())
			->setDeclineButtonTextColor($theme->getDeclineButtonTextColor())
			->setDeclineButtonBorderColor($theme->getDeclineButtonBorderColor())
			->setMoreInfoButtonBackgroundColor($theme->getMoreInfoButtonBackgroundColor())
			->setMoreInfoButtonColor($theme->getMoreInfoButtonColor())
			->setMoreInfoButtonBorderColor($theme->getMoreInfoButtonBorderColor())
			->setWindowBackgroundColor($theme->getWindowBackgroundColor())
			->setInfotextColor($theme->getInfotextColor())
			->setInfotextFontFamily($theme->getInfotextFontFamily())
			->setInfotextFontSize($theme->getInfotextFontSize())
			->setWindowBorderColor($theme->getWindowBorderColor())
			->setLinkInInfotextColor($theme->getLinkInInfotextColor())
			->setIframeBlockerBackgroundColor($theme->getIframeBlockerBackgroundColor())
			->setIframeBlockerForegroundColor($theme->getIframeBlockerForegroundColor())
			->setIframeBlockerButtonBackgroundColor($theme->getIframeBlockerButtonBackgroundColor())
			->setIframeBlockerButtonForegroundColor($theme->getIframeBlockerButtonForegroundColor())
			->setCustomCss($theme->getCustomCss())
			->setCustomCssTemplated($theme->isCustomCssTemplated())
			->setCustomCssForIframeBlocker($theme->getCustomCssForIframeBlocker())
			->setCustomCssForIframeBlockerTemplated($theme->isCustomCssForIframeBlockerTemplated())
			->setBannerHeightMobile($theme->getBannerHeightMobile())
		;
	}

	/**
	 * @param bool $default
	 */
	private function setDefaultTheme(bool $default): void
	{
		$this->set('isDefault', $default);
	}

	/**
	 * @return bool
	 */
	public function isDefaultTheme(): bool
	{
		return (bool)$this->get('isDefault');
	}

	public function makeDefault(): void
	{
		/** @var self[] $themes */
		$themes = self::all();

		foreach ($themes as $theme) {
			$theme->setDefaultTheme(false);
		}
		$this->setDefaultTheme(true);

		self::flush();
	}

	/**
	 * @return int
	 */
	public function getLastModifiedTimestamp(): int
	{
		return (int)$this->get('lastModifiedTimestamp');
	}

	/**
	 * @param int $time
	 * @return $this
	 */
	public function setLastModifiedTimestamp(int $time): self
	{
		$this->set('lastModifiedTimestamp', $time);
		return $this;
	}

	/**
	 * @return $this
	 */
	public function touchLastModifiedTimestamp(): self
	{
		// Am elegantesten wäre die Implementierung in einer überschriebenen set-Methode; die ist aber final...
		return $this->setLastModifiedTimestamp(time());
	}

	/**
	 * @return string
	 */
	public function getName(): string
	{
		return (string)$this->get('name');
	}

	/**
	 * @param string $name
	 * @return $this
	 */
	public function setName(string $name): self
	{
		$this->set('name', $name);
		return $this;
	}

	/**
	 * @return BaseLayout
	 */
	public function getBaseLayout(): BaseLayout
	{
		return BaseLayout::find((string)$this->get('baseLayout')) ?? BaseLayout::default();
	}

	/**
	 * @param BaseLayout $baseLayout
	 * @return $this
	 */
	public function setBaseLayout(BaseLayout $baseLayout): self
	{
		$this->set('baseLayout', $baseLayout->getId());
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPosition(): string
	{
		return (string)$this->get('position');
	}

	/**
	 * @param string $position
	 * @return $this
	 */
	public function setPosition(string $position): self
	{
		$position = $position ?: 'center';

		if (in_array($position, self::ALLOWED_WIDGET_POSITION_VALUES) == false) {
			throw new LogicException('Invalid value for position, allowed values: '.implode(', ', self::ALLOWED_WIDGET_POSITION_VALUES));
		}

		$this->set('position', $position);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowPurposesInMainWindow(): bool
	{
		return (bool)$this->get('showPurposesInMainWindow');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowPurposesInMainWindow(bool $state): self
	{
		$this->set('showPurposesInMainWindow', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isBlocking(): bool
	{
		return (bool)$this->get('blocking');
	}

	/**
	 * @param bool $blocking
	 * @return $this
	 */
	public function setBlocking(bool $blocking): self
	{
		$this->set('blocking', $blocking);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isSettingsIconEnabled(): bool
	{
		return (bool)$this->get('settingsIconEnabled');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setSettingsIconEnabled(bool $state): self
	{
		$this->set('settingsIconEnabled', $state);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getSettingsIconTarget(): string
	{
		return (string)$this->get('settingsIconTarget', 'purpose');
	}

	/**
	 * @param string $target
	 * @return $this
	 */
	public function setSettingsIconTarget(string $target): self
	{
		$allowedValues = [
			'main',
			'purpose',
		];

		if (in_array($target, $allowedValues) == false) {
			throw new LogicException('Invalid value for settings icon target, allowed values: '.implode(', ', $allowedValues));
		}

		$this->set('settingsIconTarget', $target);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getSettingsIconIdentifier(): string
	{
		return (string)$this->get('settingsIconFilename', '');
	}

	/**
	 * @param string $filename
	 * @return $this
	 */
	protected function setSettingsIconIdentifier(string $filename): self
	{
		$this->set('settingsIconFilename', $filename);
		return $this;
	}

	/**
	 * @return File|null
	 */
	abstract public function getSettingsIcon(): ?File;

	/**
	 * @param File|null $file
	 * @return $this
	 */
	abstract public function setSettingsIcon(?File $file): self;

	/**
	 * @return string
	 */
	public function getSettingsIconBackgroundColor(): string
	{
		return $this->get('settingsIconBackgroundColor') ?? (Utils::isClickskeksCloud() ? '#4cb474' : '#003366');
	}

	/**
	 * @param string $color
	 * @return $this
	 */
	public function setSettingsIconBackgroundColor(string $color): self
	{
		$this->set('settingsIconBackgroundColor', $color);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getSettingsIconForegroundColor(): string
	{
		return $this->get('settingsIconForegroundColor') ?? '#ffffff';
	}

	/**
	 * @param string $color
	 * @return $this
	 */
	public function setSettingsIconForegroundColor(string $color): self
	{
		$this->set('settingsIconForegroundColor', $color);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function hasCloseButton(): bool
	{
		return (bool)$this->get('showCloseButton', false);
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setCloseButton(bool $state): self
	{
		$this->set('showCloseButton', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function hasAcceptAllButton(): bool
	{
		$buttons = $this->getWidgetButtons();
		return $buttons ? in_array('accept', $buttons) : true;
	}

	/**
	 * @return bool
	 */
	public function hasDeclineButton(): bool
	{
		$buttons = $this->getWidgetButtons();
		return $buttons ? in_array('decline', $buttons) : (bool)$this->get('showDeclineButton');
	}

	/**
	 * @return bool
	 */
	public function hasRegularControlPanelButton(): bool
	{
		$buttons = $this->getWidgetButtons();
		return $buttons ? in_array('cpanel', $buttons) : true;
	}

	/**
	 * @return string[]
	 */
	public function getWidgetButtons(): array
	{
		return $this->get('widgetButtons', ['accept', 'cpanel', 'decline']);
	}

	/**
	 * @param string[] $wantedButtons
	 * @return string[]
	 */
	protected function listMandatoryButtons(array $wantedButtons): array
	{
		return [];
	}

	public static function allRestrictRegions(): array
	{
		return array_merge(["EU"], self::RESTRICTABLE_COUNTRIES, ["world"]);
	}

	/**
	 * @return string[] Liste der Regionen in denen der Button angezeigt wird
	 */
	public function getDeclineButtonRestrictions(): array
	{
		$result = (array)$this->get('declineButtonRestrictions', []);
		$result = array_intersect(self::allRestrictRegions(), $result);
		if (!$result) {
			$result = $this->allRestrictRegions();
		}
		return $result;
	}

	public function hasDeclineButtonRestrictions(): bool
	{
		return ($this->get('declineButtonRestrictions', []) != []);
	}

	public function isDeclineButtonShownInCH(): bool
	{
		$restrictions = $this->getDeclineButtonRestrictions();
		return (!$restrictions or in_array('CH', $restrictions));
	}

	public function isDeclineButtonShownElsewhere(): bool
	{
		$restrictions = $this->getDeclineButtonRestrictions();
		return (!$restrictions or in_array('world', $restrictions));
	}

	/**
	 * @param string[] $values ['EU', 'CH', 'world']
	 * @return $this
	 */
	public function setDeclineButtonRestrictions(array $values): self
	{
		if (Utils::all($this->allRestrictRegions(), function ($item) use ($values) { return in_array($item, $values); })) {
			$values = [];
		}
		$values = array_intersect(self::allRestrictRegions(), $values);
		$this->set('declineButtonRestrictions', $values);
		return $this;
	}

	/**
	 * @param string[] $buttons
	 * @return $this
	 */
	public function setWidgetButtons(array $buttons): self
	{
		$mandatoryValues = $this->listMandatoryButtons($buttons);

		// Ungültige Werte und Duplikate entfernen; erstes Vorkommen für Reihenfolge bewahren
		$buttons = array_reduce($buttons, function ($buttons, $value) {
			if (in_array($value, self::ALLOWED_WIDGET_BUTTON_VALUES, true)) {
				$buttons[$value] = $value;
			}
			return $buttons;
		}, []);

		// Pflichtwerte ergänzen
		array_walk($mandatoryValues, function ($value) use (&$buttons) {
			if (in_array($value, $buttons, true) == false) {
				$buttons[] = $value;
			}
		});

		$this->set('widgetButtons', array_values($buttons));
		return $this;
	}

	/**
	 * @return string
	 */
	public function getSelectType(): string
	{
		return (string)$this->get('selectType', 'checkbox');
	}

	/**
	 * @param string $type
	 * @return $this
	 */
	public function setSelectType(string $type): self
	{
		$allowedValues = [
			'checkbox',
			'switch',
		];

		if (in_array($type, $allowedValues) == false) {
			throw new LogicException('Invalid value for select type, allowed values: '.implode(', ', $allowedValues));
		}

		$this->set('selectType', $type);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isComplyingDoNotTrack(): bool
	{
		return (bool)$this->get('complyingDoNotTrack');
	}

	/**
	 * @param bool $complyingDoNotTrack
	 * @return $this
	 */
	public function setComplyingDoNotTrack(bool $complyingDoNotTrack): self
	{
		$this->set('complyingDoNotTrack', $complyingDoNotTrack);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowEmbeddingsInControlPanel(): bool
	{
		return (bool)$this->get('showEmbeddingsInControlPanel');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowEmbeddingsInControlPanel(bool $state): self
	{
		$this->set('showEmbeddingsInControlPanel', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowAcceptAllButtonInControlPanel(): bool
	{
		return (bool)$this->get('showAcceptAllButtonInControlPanel');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowAcceptAllButtonInControlPanel(bool $state): self
	{
		$this->set('showAcceptAllButtonInControlPanel', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowWithdrawAllButtonInControlPanel(): bool
	{
		return (bool)$this->get('showWithdrawAllButtonInControlPanel');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowWithdrawAllButtonInControlPanel(bool $state): self
	{
		$this->set('showWithdrawAllButtonInControlPanel', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowCheckAndUncheckAllButtonInControlPanel(): bool
	{
		return (bool)$this->get('showCheckAndUncheckAllButtonInControlPanel');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setShowCheckAndUncheckAllButtonInControlPanel(bool $state): self
	{
		$this->set('showCheckAndUncheckAllButtonInControlPanel', $state);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isShowExtraImprintFooter(): bool
	{
		return (bool)$this->get('showExtraImprintFooter');
	}

	/**
	 * @param bool $showExtraImprintFooter
	 * @return $this
	 */
	public function setShowExtraImprintFooter(bool $showExtraImprintFooter): self
	{
		$this->set('showExtraImprintFooter', $showExtraImprintFooter);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function hasSaveButtonInDetailsWindow(): bool
	{
		return (bool)$this->get('saveButtonInDetailsWindow');
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setSaveButtonInDetailsWindow(bool $state): self
	{
		$this->set('saveButtonInDetailsWindow', $state);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getOkButtonBackgroundColor(): string
	{
		return (string)$this->get('okButtonBackgroundColor');
	}

	/**
	 * @param string $okButtonBackgroundColor
	 * @return $this
	 */
	public function setOkButtonBackgroundColor(string $okButtonBackgroundColor): self
	{
		$this->set('okButtonBackgroundColor', $okButtonBackgroundColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getMoreInfoButtonBackgroundColor(): string
	{
		return (string)$this->get('moreInfoButtonBackgroundColor');
	}

	/**
	 * @param string $moreInfoButtonBackgroundColor
	 * @return $this
	 */
	public function setMoreInfoButtonBackgroundColor(string $moreInfoButtonBackgroundColor): self
	{
		$this->set('moreInfoButtonBackgroundColor', $moreInfoButtonBackgroundColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getWindowBackgroundColor(): string
	{
		return (string)$this->get('windowBackgroundColor');
	}

	/**
	 * @param string $windowBackgroundColor
	 * @return $this
	 */
	public function setWindowBackgroundColor(string $windowBackgroundColor): self
	{
		$this->set('windowBackgroundColor', $windowBackgroundColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getWindowBorderColor(): string
	{
		return (string)$this->get('windowBorderColor');
	}

	/**
	 * @param string $windowBorderColor
	 * @return $this
	 */
	public function setWindowBorderColor(string $windowBorderColor): self
	{
		$this->set('windowBorderColor', $windowBorderColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getLinkInInfotextColor(): string
	{
		return (string)$this->get('linkInInfotextColor');
	}

	/**
	 * @param string $linkInInfotextColor
	 * @return $this
	 */
	public function setLinkInInfotextColor(string $linkInInfotextColor): self
	{
		$this->set('linkInInfotextColor', $linkInInfotextColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getOkButtonColor(): string
	{
		return (string)$this->get('okButtonColor');
	}

	/**
	 * @param string $okButtonColor
	 * @return $this
	 */
	public function setOkButtonColor(string $okButtonColor): self
	{
		$this->set('okButtonColor', $okButtonColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getMoreInfoButtonColor(): string
	{
		return (string)$this->get('moreInfoButtonColor');
	}

	/**
	 * @param string $moreInfoButtonColor
	 * @return $this
	 */
	public function setMoreInfoButtonColor(string $moreInfoButtonColor): self
	{
		$this->set('moreInfoButtonColor', $moreInfoButtonColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getOkButtonBorderColor(): string
	{
		return (string)$this->get('okButtonBorderColor');
	}

	/**
	 * @param string $okButtonBorderColor
	 * @return $this
	 */
	public function setOkButtonBorderColor(string $okButtonBorderColor): self
	{
		$this->set('okButtonBorderColor', $okButtonBorderColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getMoreInfoButtonBorderColor(): string
	{
		return (string)$this->get('moreInfoButtonBorderColor');
	}

	/**
	 * @param string $moreInfoButtonBorderColor
	 * @return $this
	 */
	public function setMoreInfoButtonBorderColor(string $moreInfoButtonBorderColor): self
	{
		$this->set('moreInfoButtonBorderColor', $moreInfoButtonBorderColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getDeclineButtonBackgroundColor(): string
	{
		return (string)$this->get('declineButtonBackgroundColor', $this->getMoreInfoButtonBackgroundColor());
	}

	/**
	 * @param string $color
	 * @return $this
	 */
	public function setDeclineButtonBackgroundColor(string $color): self
	{
		$this->set('declineButtonBackgroundColor', $color);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getDeclineButtonTextColor(): string
	{
		return (string)$this->get('declineButtonTextColor', $this->getMoreInfoButtonColor());
	}

	/**
	 * @param string $color
	 * @return $this
	 */
	public function setDeclineButtonTextColor(string $color): self
	{
		$this->set('declineButtonTextColor', $color);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getDeclineButtonBorderColor(): string
	{
		return (string)$this->get('declineButtonBorderColor', $this->getMoreInfoButtonBorderColor());
	}

	/**
	 * @param string $color
	 * @return $this
	 */
	public function setDeclineButtonBorderColor(string $color): self
	{
		$this->set('declineButtonBorderColor', $color);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getInfotextColor(): string
	{
		return (string)$this->get('infotextColor');
	}

	/**
	 * @param string $infotextColor
	 * @return $this
	 */
	public function setInfotextColor(string $infotextColor): self
	{
		$this->set('infotextColor', $infotextColor);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getInfotextFontFamily(): string
	{
		return (string)$this->get('infotextFontFamily');
	}

	/**
	 * @param string $infotextFontFamily
	 * @return $this
	 */
	public function setInfotextFontFamily(string $infotextFontFamily): self
	{
		$this->set('infotextFontFamily', $infotextFontFamily);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getInfotextFontSize(): string
	{
		return (string)$this->get('infotextFontSize');
	}

	/**
	 * @param string $infotextFontSize
	 * @return $this
	 */
	public function setInfotextFontSize(string $infotextFontSize): self
	{
		$this->set('infotextFontSize', $infotextFontSize);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getIframeBlockerBackgroundColor(): string
	{
		return (string)$this->get('iframeBlockerBackgroundColor', '#111111');
	}

	/**
	 * @param string $hexCode
	 * @return $this
	 */
	public function setIframeBlockerBackgroundColor(string $hexCode): self
	{
		$this->set('iframeBlockerBackgroundColor', $hexCode);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getIframeBlockerForegroundColor(): string
	{
		return (string)$this->get('iframeBlockerForegroundColor', '#f8f8f8');
	}

	/**
	 * @param string $hexCode
	 * @return $this
	 */
	public function setIframeBlockerForegroundColor(string $hexCode): self
	{
		$this->set('iframeBlockerForegroundColor', $hexCode);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getIframeBlockerButtonBackgroundColor(): string
	{
		return (string)$this->get('iframeBlockerButtonBackgroundColor', $this->getOkButtonBackgroundColor());
	}

	/**
	 * @param string $hexCode
	 * @return $this
	 */
	public function setIframeBlockerButtonBackgroundColor(string $hexCode): self
	{
		$this->set('iframeBlockerButtonBackgroundColor', $hexCode);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getIframeBlockerButtonForegroundColor(): string
	{
		return (string)$this->get('iframeBlockerButtonForegroundColor', '#ffffff');
	}

	/**
	 * @param string $hexCode
	 * @return $this
	 */
	public function setIframeBlockerButtonForegroundColor(string $hexCode): self
	{
		$this->set('iframeBlockerButtonForegroundColor', $hexCode);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getCustomCss(): string
	{
		return (string)$this->get('customCss');
	}

	/**
	 * @param string $customCss
	 * @return $this
	 */
	public function setCustomCss(string $customCss): self
	{
		$origins = Utils::extractCSPOrigins($customCss);
		$this->set('customCssOrigins', $origins);

		$this->set('customCss', $customCss);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isCustomCssTemplated(): bool
	{
		return (bool)$this->get('customCssTemplated', false);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setCustomCssTemplated(bool $value): self
	{
		$this->set('customCssTemplated', $value);
		return $this;
	}

	/**
	 * Externe Origins die im CSS gefunden wurden.
	 *
	 * Dienen dem Einfügen in eine Content-Security-Policy
	 *
	 * @return string[]
	 */
	public function getCustomCssOrigins(): array
	{
		return (array)$this->get('customCssOrigins', []);
	}

	/**
	 * @return string
	 */
	public function getCustomCssForIframeBlocker(): string
	{
		return (string)$this->get('customCssForIframeBlocker');
	}

	/**
	 * @param string $customCss
	 * @return $this
	 */
	public function setCustomCssForIframeBlocker(string $customCss): self
	{
		$origins = Utils::extractCSPOrigins($customCss);
		$this->set('customCssForIframeBlockerOrigins', $origins);
		$this->set('customCssForIframeBlocker', $customCss);
		return $this;
	}

	/**
	 * Externe Origins die im Iframe-Blocker-CSS gefunden wurden.
	 *
	 * Dienen dem Einfügen in eine Content-Security-Policy
	 *
	 * @return string[]
	 */
	public function getCustomCssForIframeBlockerOrigins(): array
	{
		return (array)$this->get('customCssForIframeBlockerOrigins', []);
	}

	/**
	 * @return bool
	 */
	public function isCustomCssForIframeBlockerTemplated(): bool
	{
		return (bool)$this->get('customCssForIframeBlockerTemplated', false);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setCustomCssForIframeBlockerTemplated(bool $value): self
	{
		$this->set('customCssForIframeBlockerTemplated', $value);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function getOnlyInEu(): bool
	{
		return (bool)$this->get('onlyInEu');
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setOnlyInEu(bool $value): self
	{
		$this->set('onlyInEu', $value);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function getManipulationPrevention(): bool
	{
		return (bool)$this->get('manipulationPrevention', true);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setManipulationPrevention(bool $value): self
	{
		$this->set('manipulationPrevention', $value);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getBannerHeightMobile(): int
	{
		return (int)($this->get('bannerHeightMobile') ?? 100);
	}

	/**
	 * @param int $bannerHeightMobile
	 * @return $this
	 */
	public function setBannerHeightMobile(int $bannerHeightMobile): self
	{
		if ($bannerHeightMobile < 60 && $bannerHeightMobile > 0) {
			$bannerHeightMobile = 60;
		}
		if ($bannerHeightMobile > 100 || $bannerHeightMobile <= 0) {
			$bannerHeightMobile = 100;
		}
		$this->set('bannerHeightMobile', $bannerHeightMobile);
		return $this;
	}
}
