<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Utils;
use Google\Cloud\Core\Exception\GoogleException;
use Google\Cloud\Core\Exception\ServiceException;
use Google\Cloud\Translate\V2\TranslateClient;
use RuntimeException;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Yaml\Yaml;

class TransLate extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-translate';

	/**
	 * @param array $postData
	 * @return $this
	 */
	public function setArray(array $postData): self
	{
		foreach($postData as $k=>$v)
		{
			$this->set($k, $v);
		}
		return $this;
	}

	/**
	 * @return array
	 */
	public static function setGoogleLangs():array
	{
		return array("hr","sr","fa","ga","hi","is");
	}

	/**
	 * @param string $deeplApiKey
	 * @param string $targetLang Sprachkürzel (de, en, …)
	 * @return void
	 */
	public static function translateJsonsAndCo(string $deeplApiKey, string $targetLang): void
	{
		//die Purposes
		self::translatePurposes($deeplApiKey, $targetLang);

		//die Skeleton
		self::translateSkeletonUserLocale($deeplApiKey, $targetLang);
		self::translateClickskeksSkeletonUserLocale($deeplApiKey, $targetLang);
	}

	/**
	 * @param string $deeplApiKey
	 * @param string $localeName Locale name (de_DE, en_US, fi, ja, …)
	 * @param string|null $filename
	 * @return void
	 * @throws RuntimeException if the language is not found or the skeleton file doesn't exist.
	 */
	private static function translateSkeletonUserLocale(string $deeplApiKey, string $localeName, ?string $filename=null): void
	{
		if (!Language::exists($localeName)) {
			throw new RuntimeException("Cannot translate user locale skeleton, language $localeName not found.");
		}

		$targetLang = self::makeDeeplTargetLang($localeName);

		$filename = $filename === null ? '/share/skeleton/cm-user-locales.json' : ('/' . ltrim($filename, '/'));
		$filename = Utils::getBaseDir() . $filename;

		if (!is_file($filename)) {
			throw new RuntimeException("Cannot translate user locale skeleton, file $filename not found.");
		}

		$data = file_get_contents($filename);
		$userLocales = json_decode($data,true);

		if (!is_array($userLocales)) {
			return;
		}
		elseif (!isset($userLocales[$localeName]) || !is_array($userLocales[$localeName])) {
			$userLocales[$localeName] = [];
		}

		$base = $userLocales['de_DE'];
		$filesystem = new Filesystem();

		foreach($base as $k=>$item)
		{
			if(is_array($item))
			{
				foreach($item as $subk=>$subitem)
				{
					if (!isset($userLocales[$localeName][$k]) || !is_array($userLocales[$localeName][$k])) {
						$userLocales[$localeName][$k] = [];
					}

					if (is_string($subitem)) {
						// Überspringe, falls bereits übersetzt
						$translation = $userLocales[$localeName][$k][$subk] ?? null;
						if (is_string($translation) && !self::isTranslationMessagePending($translation)) {
							continue;
						}
						$body = '&text='.rawurlencode($subitem);
						$message = self::curlFromDeepl($deeplApiKey,$body,$targetLang);
						$userLocales[$localeName][$k][$subk] = $message['translations'][0]['text'];
					}
					elseif (!isset($userLocales[$localeName][$k][$subk])) {
						$userLocales[$localeName][$k][$subk] = $subitem;
					}
				}
			}
			elseif (is_string($item)) {
				// Überspringe, falls bereits übersetzt
				$translation = $userLocales[$localeName][$k] ?? null;
				if (is_string($translation) && !self::isTranslationMessagePending($translation)) {
					continue;
				}
				$body = '&text='.rawurlencode($item);
				$message = self::curlFromDeepl($deeplApiKey,$body,$targetLang);
				$userLocales[$localeName][$k] = $message['translations'][0]['text'];
			}
			elseif (!isset($userLocales[$localeName][$k])) {
				// Wert `false` für bestimmte Eigenschaften statt der Übernahme des Basiswerts erzwingen
				if (in_array($k, [
					'enabled',
				])) {
					$item = false;
				}
				$userLocales[$localeName][$k] = $item;
			}
		}

		// Sort by key to generate reliably clean diffs
		array_walk($userLocales, function (array &$locale) {
			uksort($locale, function (string $a, string $b) { return strnatcasecmp($a, $b); });
		});

		uksort($userLocales, function (string $a, string $b) { return strnatcasecmp($a, $b); });

		$userLocalesJson = json_encode($userLocales, JSON_UNESCAPED_SLASHES|JSON_UNESCAPED_UNICODE|JSON_PRETTY_PRINT);
		$filesystem->dumpFile($filename,$userLocalesJson);
	}

	/**
	 * @param $deeplApiKey
	 * @param string $targetLang
	 * @return void
	 */
	private static function translateClickskeksSkeletonUserLocale($deeplApiKey, string $targetLang): void
	{
		self::translateSkeletonUserLocale($deeplApiKey, $targetLang, '/share/skeleton/clickskeks/cm-user-locales.json');
	}

	/**
	 * @return void
	 */
	public static function complementClickskeksSkeleton(): void
	{
		// UserLocale-Vorlage laden und erweitern
		$base = json_decode((string)file_get_contents(Utils::getBaseDir() . '/share/skeleton/cm-user-locales.json'), true);
		$pathname = Utils::getBaseDir() . '/share/skeleton/clickskeks/cm-user-locales.json';
		$skeleton = json_decode((string)file_get_contents($pathname), true) ?? [];

		// Kopiere die deutschen Spracheinträge der CCM19-Vorlage in die Clickskeks-Vorlage
		if (isset($base['de_DE']) && is_array($base['de_DE'])) {
			$skeleton['de_DE'] = Utils::complementArray($skeleton['de_DE'] ?? [], $base['de_DE']);
			$json = json_encode($skeleton, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE | JSON_PRETTY_PRINT);
			file_put_contents($pathname, $json, LOCK_EX);
		}
	}

	/**
	 * @param string $deeplApiKey
	 * @param string $localeName Locale name (de_DE, en_US, fi, ja, …)
	 * @return void
	 */
	public static function translatePurposes(string $deeplApiKey, string $localeName): void
	{
		if (!Language::exists($localeName)) {
			throw new RuntimeException("Cannot translate user locale skeleton, language $localeName not found.");
		}

		$targetLang = self::makeDeeplTargetLang($localeName);

		//Purposes
		$filesystem = new Filesystem();
		$filename = Utils::getBaseDir()."/share/cm-purposes.json";
		$data = file_get_contents($filename);
		$purposes = json_decode($data,true);
		assert(is_array($purposes));

		foreach($purposes as $k=>$v)
		{
			if (!self::isTranslationMessagePending((string)($v['translated_name'][$localeName] ?? ''))) {
				continue;
			}
			$body = '&text='.rawurlencode($v['translated_name']['de_DE']);
			$dataTrans = self::curlFromDeepl($deeplApiKey,$body,$targetLang);
			$purposes[$k]['translated_name'][$localeName] = (string)($dataTrans['translations'][0]['text'] ?? '');
		}

		// Sort by key to generate reliably clean diffs
		array_walk($purposes, function (array &$purpose) {
			if (isset($purpose['translated_name']) && is_array($purpose['translated_name'])) {
				uksort($purpose['translated_name'], function (string $a, string $b) { return strnatcasecmp($a, $b); });
			}
		});

		$purposeJson = json_encode($purposes, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE | JSON_PRETTY_PRINT);
		$filesystem->dumpFile($filename,$purposeJson);
	}

	/**
	 * @param string $deeplApiKey
	 * @param string $localeName Locale name (de_DE, en_US, fi, ja, …)
	 */
	public static function translateJPlugins(string $deeplApiKey, string $localeName): void
	{
		foreach (scandir(Utils::getBaseDir()."/plugins/", SCANDIR_SORT_NONE) as $plugin) {
			if ($plugin === '.' or $plugin === '..') {
				continue;
			}
			$pluginsdir = "/plugins/".$plugin;
			self::translateNow($deeplApiKey, $localeName, $pluginsdir);
		}
	}

	/**
	 * @param string $message
	 * @return bool
	 */
	public static function isTranslationMessagePending(string $message): bool
	{
		return strlen($message) == 0 || mb_substr($message, 0, 2) == '__';
	}

	/**
	 * @return string
	 */
	public static function getDeeplApiKey(): string
	{
		$all = self::all();
		$deeplApiKey = "";

		if(!empty(array_key_first($all)))
		{
			$deeplKey = $all[array_key_first($all)];
			$deeplApiKey = $deeplKey->get("deeplAPIKey");
		}
		return is_string($deeplApiKey) ? $deeplApiKey : '';
	}

	/**
	 * @return string
	 */
	public static function getGoogleApiKey(): string
	{
		$all = self::all();
		$googleApiKey = "";

		if(!empty(array_key_first($all)))
		{
			$deeplKey = $all[array_key_first($all)];
			$googleApiKey = $deeplKey->get("googleAPIKey");
		}
		return is_string($googleApiKey) ? $googleApiKey : '';
	}

	/**
	 * @param array $translatedData
	 * @param string $lang Sprachkürzel (de, en, …)
	 */
	private static function saveTranslatedData(array $translatedData, string $lang, string $pluginsdir=''): void
	{

		$filesystem = new Filesystem();
		$filename = Utils::getBaseDir().$pluginsdir."/translations/"."messages.".strtolower($lang).".yaml";
		$yaml = Yaml::dump($translatedData);
		$yaml = str_ireplace("{  }","",$yaml);
		$yaml = str_ireplace('<nt>',"",$yaml);
		$yaml = str_ireplace("</nt>","",$yaml);
		$yaml = preg_replace('%<code class="notranslate">([^<>]*)</code>%', '$1', $yaml);
		$filesystem->dumpFile($filename, $yaml);
	}


	/**
	 * @param string $lang
	 * @return array
	 */
	public static function getBaseTranslateData(string $lang, string $pluginsdir=''): array
	{
		$filesystem = new Filesystem();
		$filename = Utils::getBaseDir().$pluginsdir."/translations/"."messages.".strtolower($lang).".yaml";
		//print_r($filename);
		$value = [];

		if($filesystem->exists($filename)) {
			$value = Yaml::parseFile($filename);
		}

		if(empty($value)) {
			$value = [];
		}

		return $value;
	}

	/**
	 * @param array $base
	 * @param array $target
	 * @return array
	 */
	public static function getNotTranslatedData(array $base, array $target) : array
	{
		// Consider missing messages in target translation
		$newMessages = array_diff_key($base, $target);

		// Consider pending messages, which are indicated by two leading underscores
		$pendingMessages = array_intersect_key(
			$base,
			array_filter(array_intersect_key($target, $base), [self::class, 'isTranslationMessagePending'])
		);

		return array_merge($pendingMessages, $newMessages);
	}

	/**
	 * @param string $deeplApiKey
	 * @param string $localeName Locale name (de_DE, en_US, fi, ja, …)
	 * @param string $pluginsdir
	 * @return void
	 */
	public static function translateNow(string $deeplApiKey, string $localeName, string $pluginsdir=''): void
	{
		if (!Language::exists($localeName)) {
			throw new RuntimeException("Cannot translate user locale skeleton, language $localeName not found.");
		}

		$languageCode = explode('_', $localeName, 2)[0];
		$targetLang = self::makeDeeplTargetLang($localeName);

		//Die DE Daten
		$valueBase = self::getBaseTranslateData('de', $pluginsdir);
		#print_r($valueBase);exit();

		//Die TargetDaten
		$valueTarget = self::getBaseTranslateData($languageCode, $pluginsdir);
		#print_r($valueTarget);exit();
		//Nur die nicht übersetzten Teile
		$notTranslated = self::getNotTranslatedData($valueBase, $valueTarget);
		//print_r($notTranslated);exit();

		// Remove empty items
		/** @var string[] $notTranslated */
		$notTranslated = array_filter(array_map('trim', array_map('strval', $notTranslated)), static function (string $text) {
			return strlen($text) > 0;
		});

		/**
		 * Up to 50 texts can be sent for translation in one request.
		 * @see https://www.deepl.com/docs-api/translate-text/
		 * @var int $batchSize
		 */
		$batchSize = min(count($notTranslated), 45);

		/**
		 * Jetzt die Daten in Chunks aufteilen
		 * Je 5 zusammenfassen - evtl. später...
		 * und 50 Blöcke erstellen und die übersetzen lassen...
		 */
		$translated = array();

		for ($i = 0; $i < count($notTranslated); $i += $batchSize) {
			$batch = array_slice($notTranslated, $i, $batchSize, true);
			$body = '&' . implode('&', array_map(static function (string $text) {
				// Use percent encoded spaces
				return http_build_query(['text'=>$text], '', '&', PHP_QUERY_RFC3986);
			}, $batch));
			$transKeys = array_keys($batch);
			$translated = array_merge($translated, self::getFromDeepl($deeplApiKey,$body,$targetLang,$transKeys));
		}

		// Upsert translated messages to preserve the position of pending messages
		$translated = array_merge($valueTarget, $translated);

		//print_r($translated);
		//Daten speichern
		self::saveTranslatedData($translated, $languageCode, $pluginsdir);
	}

	/**
	 * @param string $text
	 * @return string|string[]|null
	 */
	public static function replacePlaceholder($text)
	{
		/**
		$suchmuster = '/(&.*?;)/i';
		$ersetzung = '<nt>$1</nt>';
		$text =  preg_replace($suchmuster, $ersetzung, $text);
		**/

		//einmal die mit :
		$suchmuster = '/("?:[a-zA-Z0-9_-]*"?)/i';
		$ersetzung = '<nt>$1</nt>';
		$text =  preg_replace($suchmuster, $ersetzung, $text);

		//UNd einmal die mit %%
		$suchmuster = '/(%.*?%)/i';
		$ersetzung = '<nt>$1</nt>';
		$text = preg_replace($suchmuster, $ersetzung, $text);

		return $text;
	}

	/**
	 * @param string[] $body
	 * @param string $targetLang
	 * @return array{source:string,input:string,text:string}[]|null
	 * @throws RuntimeException
	 */
	private static function getFromGoogle(array $body, string $targetLang): ?array
	{
		if (!$targetLang or !$body) {
			return null;
		}

		try {
			$translate = new TranslateClient([
				'key' => self::getGoogleApiKey()
			]);

			$result = $translate->translateBatch($body, [
				'target' => $targetLang,
				'source' => "de"
			]);
		} catch (GoogleException|ServiceException $exception) {
			throw new RuntimeException($exception->getMessage(), $exception->getCode());
		}

		return $result;
	}

	/**
	 * @param string $deeplApiKey
	 * @param string $body
	 * @param string $targetLang
	 * @param int[]|string[] $transKeys
	 * @return array<string, string>
	 *  https://www.deepl.com/docs-api/translating-text/multiple-sentences/
	 */
	public static function getFromDeepl($deeplApiKey, $body, $targetLang, $transKeys): array
	{
		//Daten holen
		$responseArray = self::curlFromDeepl($deeplApiKey,$body,$targetLang);

		if(empty($responseArray['translations']))
		{
			return [];
		}
		$translatedData = $responseArray['translations'];
		$data = array();

		//print_r($transKeys);
		//print_r($translatedData);

		//die richtigen Keys zusortieren
		foreach($transKeys as $count => $realKey)
		{
			$data[$realKey] = $translatedData[$count]['text'];
		}

		//Rückgabe
		return (empty($data)) ? [] : $data;
	}

	/**
	 * @param string $deeplApiKey
	 * @param string $body
	 * @param string $targetLang
	 * @return array
	 */
	public static function curlFromDeepl($deeplApiKey,$body,$targetLang)
	{
		//Hier checken ob es eine nur von G unterstützte Sprache ist...
		$googleLangs = self::setGoogleLangs();
		//$targetLang = "en";
		//print_r($body);//exit();

		if (in_array(strtolower($targetLang), $googleLangs)) {
			// Convert DeepL language code to Google language code; ignore country code
			// https://cloud.google.com/translate/docs/languages
			$targetLang = strtolower($targetLang);

			if($targetLang=="sr")
			{
				$targetLang = "sr_Latn";
			}
			$body = rawurldecode($body);
			$body = strtr($body, ['<nt>'=>'<code class="notranslate">', '</nt>' =>'</code>']);
			$bodyArray = explode('&text=',$body);
			unset($bodyArray[0]);
			$bodyArray2 = array_values($bodyArray);
			$translation = [
				'translations' => self::getFromGoogle($bodyArray2,$targetLang)
			];
			return $translation;
		}
		else{

			$targetLang = str_replace('_', '-', $targetLang);
			$url = "https://api.deepl.com/v2/translate?auth_key=".$deeplApiKey."&target_lang=".$targetLang."&source_lang=de&preserve_formatting=1&tag_handling=xml&ignore_tags=nt";

			//Daten holen
			$curl = curl_init();
			curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
			curl_setopt($curl, CURLOPT_POST, true);
			curl_setopt($curl, CURLOPT_URL, $url);
			curl_setopt($curl, CURLOPT_POSTFIELDS, $body);
			curl_setopt($curl, CURLOPT_HTTPHEADER, array('Content-Type: application/x-www-form-urlencoded'));
			$response = curl_exec($curl);
			$translation = json_decode($response, true);

			if (!is_array($translation) || !isset($translation['translations'])) {
				throw new RuntimeException($translation['message'] ?? 'Unknown error while fetching translation from DeepL');
			}

			return $translation;

		}


	}

	/**
	 * @param string $localeName
	 * @return string
	 * @see https://www.deepl.com/docs-api/translate-text/
	 */
	private static function makeDeeplTargetLang(string $localeName): string
	{
		$variants = [
			'en_GB', 'en_US',
			'pt_BR', 'pt_PT',
		];

		if (in_array($localeName, $variants)) {
			$targetLang = strtoupper(str_replace('_', '-', $localeName));
		} else {
			$targetLang = strtoupper(explode('_', $localeName, 2)[0]);
		}

		$defaults = ['CZ'=>'CS', 'EN'=>'EN-US', 'PT'=>'PT-PT', 'NO'=>'NB'];
		if (isset($defaults[$targetLang])) {
			$targetLang = $defaults[$targetLang];
		}

		return $targetLang;
	}
}
