<?php

namespace App\Model;

use App\Config;
use App\Component\LocaleService;
use App\Event\LoginEvent;
use App\Event\UserDeletePrepareEvent;
use App\Exception\ApiKeyAlreadyExistsException;
use App\Exception\ConfigNotReadableException;
use App\Exception\ConfigNotWritableException;
use App\Exception\SecondFactorRequiredException;
use App\Exception\UserAlreadyExistsException;
use App\Exception\UserIsDeactivatedException;
use App\Component\Config\RepositoryInterface;
use App\Component\Media\MediaFileService;
use App\Model\Index\DomainIndex;
use App\Utils;
use Exception;
use Symfony\Component\HttpFoundation\File\File;
use DateTime;
use DateTimeInterface;
use DateTimeZone;
use LogicException;

class User extends Model
{
	// Zeit in Sekunden für die nach einem erfolglosen Login kein Login funktioniert
	const LOGIN_BLOCKED_DURATION = 3;

	// Zeit in Sekunden für die ein 2-Faktor-Login in der Schwebe sein darf
	const LOGIN_2FA_TIMEOUT = 900;

	// Rollen, die ein User haben kann
	const ROLE_ADMIN = 'admin';
	const ROLE_CLIENT = 'client';
	const ROLE_MANAGER = 'manager';

	// Mögliche Farbschemata
	const COLOR_SCHEMES = ['light', 'dark'];

	private const PASSWORD_ALGORITHM = PASSWORD_BCRYPT;
	private const PASSWORD_COST = 10;

	/** @var self|null $apiUser */
	private static $apiUser = null;

	/** @var self|null $generalApiUser */
	private static $generalApiUser = null;

	/** @var self|null $userOverride  */
	private static $userOverride = null;

	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-users';

	/** @var string[] $requiredProjectionKeys */
	protected static $requiredProjectionKeys = [
		'DomActive',
		'role',
		'subUserOf',
	];

	/** @var array<string, self> $nameLookupCache */
	private static $nameLookupCache = [];

	/** @var array<string, bool> $accountFeatures */
	private $accountFeatures;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Diese Methode wird ausgeführt, sobald das Repository durch die Basisklasse Model geladen wurde.
	 */
	protected static final function onRepositoryLoaded(): void
	{
		self::migrateFromConfig();
	}

	/**
	 * Diese Methode migriert den Datenbestand an Benutzern aus der Systemkonfiguration in das eigene Modul.
	 * @throws ConfigNotWritableException
	 */
	private static function migrateFromConfig(): void
	{
		$config = Config::getInstance();

		if ((bool)$config->get('usersMigratedIntoOwnModule')) {
			return;
		}

		$users = $config->get('users');
		if (is_array($users) == false) {
			return;
		}

		foreach ($users as $userData) {
			$username = trim($userData['username'] ?? '');
			$passwordHash = trim($userData['password_hash'] ?? '');

			if (empty($username) || empty($passwordHash)) {
				continue;
			}

			try {
				$user = self::create($username)->setPasswordHash($passwordHash);
			}
			catch (UserAlreadyExistsException $exception) {
				continue;
			}
		}

		// Migrierte Benutzer auf die Platte schreiben und Migration als durchgeführt festhalten
		self::flush();
		$config->unset('users');
		$config->set('usersMigratedIntoOwnModule', true);
	}

	/**
	 * @param string $id
	 * @return static|null
	 */
	public static function byId(string $id): ?self
	{
		return parent::find($id);
	}

	/**
	 * @param string $key
	 * @return self|null
	 */
	public static function byApiKey(string $key): ?self
	{
		return current(self::filterBy(['apiKey' => $key])) ?: null;
	}

	/**
	 * @param string $key
	 * @return self|null
	 */
	private static function byGeneralApiKey(string $key): ?self
	{
		//bei leerem oder zu kurzem Key - nullen...
		if (strlen($key)<45){
			return null;
		}
		return current(self::filterBy(['generalApiKey'=>$key])) ?: null;
	}

	/**
	 * @param string $role
	 * @return self[]
	 */
	public static function byRole(string $role): array
	{
		$users = self::filterBy(['role' => $role]);
		$user = self::loggedInUser();

		if (!$user) {
			return $users;
		}

		if ($role == self::ROLE_CLIENT) {
			// Users such as managers may have limited access to clients
			// Explicitly check the user role because clients could already iterate over all clients
			if ($user->isManager() && $user->hasAccessToAllClients() == false) {
				$allowedClientGroups = $user->getClientGroups();
				$users = array_filter($users, static function (User $client) use ($allowedClientGroups) {
					return (bool)array_intersect_key($allowedClientGroups, $client->getClientGroups());
				});
			}
		}

		return $users;
	}

	/**
	 * Diese Methode liefert die Hauptaccounts aller Kunden; Unterbenutzer sind nicht enthalten.
	 * @return self[]
	 */
	public static function getClients(): array
	{
		return array_values(array_filter(self::byRole(self::ROLE_CLIENT), function (self $client) {
			return $client->isAccountOwner();
		}));
	}

	/**
	 * Diese Methode liefert alle Benutzer, die dem Administrator unterstehen
	 * und (teilweisen) Zugriff auf die Administration haben.
	 * @return self[]
	 */
	public static function getManagers(): array
	{
		return array_values(self::byRole(self::ROLE_MANAGER));
	}

	//
	/**
	 * Diese Methode liefert die Clienten einer Agentur
	 * @return self[]
	 */
	public static function allOfAgency(string $id): array
	{
		return array_values(array_filter(self::getClients(), function (self $client) use ($id) {
			return $client->getAgencyId() === $id;
		}));
	}

	/**
	 * Diese Methode liefert den aktuell eingeloggten Benutzer oder null, falls keiner existiert.
	 * Bei externen Aufrufen wird ggf. der API-Schlüssel berücksichtigt und der entsprechende Benutzer verwendet.
	 * @return self|null
	 * @phan-return static|null
	 */
	public static function loggedInUser(): ?self
	{
		if (self::$userOverride) {
			return self::$userOverride;
		}
		if (self::$apiUser) {
			return self::$apiUser;
		}
		elseif (self::$generalApiUser) {
			return self::$generalApiUser;
		}

		try {
			$session = Utils::getSession();
			$id = (string)$session->get('userId', '');
			$last_ts = $session->get('last_timestamp', 0);
		} catch (\Exception $ignore) {
			// Wenn der Session-Speicher nicht funktioniert, als ausgeloggt betrachten
			$id = '';
			$last_ts = 0;
		}

		// Als nicht mehr eingeloggt betrachten, wenn länger inaktiv als erlaubt
		if ($id !== '' and !empty($_ENV['APP_MAX_SESSION_IDLE']) and (time() - $last_ts > (int)$_ENV['APP_MAX_SESSION_IDLE'])) {
			return null;
		}

		return ($id !== '') ? self::byId($id) : null;
	}

	/**
	 * Erzwingt einen bestimmten Benutzer beim Aufruf von `self::loggedInUser()`.
	 * @param User|null $user
	 */
	private static function setUserOverride(?User $user): void
	{
		self::$userOverride = $user;
	}

	/**
	 * @return self|null
	 */
	private static function getUserOverride(): ?self
	{
		return self::$userOverride;
	}

	/**
	 * Diese Methode liefert den Benutzer für den ein 2FA-Login im Gange ist oder null, falls kein 2FA-Login läuft.
	 *
	 * @return self|null
	 * @phan-return static|null
	 */
	public static function loggingInUser(): ?self
	{
		try {
			$session = Utils::getSession();
			$id = (string)$session->get('2faUserId', '');
			$last_ts = $session->get('2faTimestamp', 0);
		} catch (\Exception $ignore) {
			$id = '';
			$last_ts = 0;
		}

		// Als nicht mehr in der Schwebe betrachten, wenn länger inaktiv als erlaubt
		if ($id !== '' and (time() - $last_ts > self::LOGIN_2FA_TIMEOUT)) {
			return null;
		}

		return ($id !== '') ? self::byId($id) : null;
	}

	/**
	 * Diese Methode liefert die nutzbaren 2FA-Methoden, falls ein 2FA-Login läuft.
	 *
	 * @return array[]
	 */
	public static function currentTwoFactorAuths(): array
	{
		$session = Utils::getSession();
		$methods = (array)$session->get('2faMethods', []);

		// Als nicht mehr in der Schwebe betrachten, wenn länger inaktiv als erlaubt
		$last_ts = $session->get('2faTimestamp', 0);
		if ($methods and time() - $last_ts > self::LOGIN_2FA_TIMEOUT) {
			$methods = [];
		}

		return $methods;
	}

	/**
	 * Diese Methode liefert den ursprünglich eingeloggten Benutzer, wenn aus
	 * dem Admin-Konto in einen Kunden gewechselt wurde und ansonsten null.
	 * @return self|null
	 * @phan-return static|null
	 */
	public static function originalLoggedInUser(): ?self
	{
		try {
			$session = Utils::getSession();
			$id = (string)$session->get('originalUserId', '');
		} catch (\Exception $ignore) {
			$id = '';
		}
		return ($id !== '') ? self::byId($id) : null;
	}

	/**
	 * @param string $key
	 * @return bool
	 */
	public static function setApiUser(string $key): bool
	{
		return (bool)(self::$apiUser = self::byApiKey($key));
	}

	/**
	 * Gibt den Nutzer zurück, der über `setApiUser` gesetzt wurde
	 * und ansonsten `null`.
	 *
	 * @return User|null
	 */
	public static function getCurrentApiUser(): ?User
	{
		return self::$apiUser;
	}

	/**
	 * Diese Methode ermittelt, ob bereits ein Benutzer mit demselben Namen existiert.
	 * @param string $username
	 * @return bool
	 */
	public static function exists(string $username): bool
	{
		if (isset(self::$nameLookupCache[$username])) {
			return true;
		}

		return array_reduce(self::all(), function ($userExists, $user) use ($username) {
			/** @var self $user */
			return $userExists || mb_strtolower($user->getUsername()) === mb_strtolower($username);
		}, false);
	}

	/**
	 * Diese Methode gibt den zu $username passenden Benutzer zurück.
	 * @param string $username
	 * @return static|null
	 * @override
	 * @phan-return static|null
	 */
	public static function find(string $username): ?Model
	{
		/** @var User|null $result */
		$result = self::$nameLookupCache[$username] ?? null;

		// Wenn nicht im Cache: suchen
		if ($result === null) {
			$result = array_reduce(self::all(), function ($foundUser, $user) use ($username) {
				/** @var self $user */
				/** @var self|null $foundUser */
				if ($foundUser) {
					return $foundUser;
				}

				return mb_strtolower($user->getUsername()) === mb_strtolower($username)
					? $user
					: null;
			}, null);

			// Ergebnis im Cache speichern
			if ($result) {
				self::$nameLookupCache[$username] = $result;
			}
		}

		return $result;
	}

	/**
	 * Diese Methode sucht nach Kunden und gibt ein Ergebnis-Array zurük.
	 * @param string $search Suchbegriff
	 * @param string $role Benutzerrolle
	 * @return User[]
	 */
	public static function search(string $search, string $role): array
	{
		return self::searchInArray($search, self::byRole($role));
	}

	/**
	 * Diese Methode filtert die Benutzer nach dem Suchbegriff.
	 * @param string $search Suchbegriff
	 * @param User[] $users Benutzer, die gefiltert werden sollen
	 * @return User[]
	 */
	public static function searchInArray(string $search, array $users): array
	{
		return array_filter($users, function (self $user) use ($search) {
			return stristr($user->getUsername(), $search)
				|| 	stristr($user->getEmailAddress()	,$search)
				|| 	stristr($user->getFirstName()		,$search)
				|| 	stristr($user->getLastName()		,$search)
				|| 	stristr($user->getCompany()			,$search)
				|| 	stristr($user->getAdminInfo()		,$search)
				||  stristr((string)$user->getAgencyId(), $search)
				||  stristr((string)$user->getAgencyAffiliateId(), $search)
				|| 	stristr($user->getSubUserOf()		,$search)
				|| 	stristr($user->getId()				,$search)
				|| 	$user->getApiKey() === $search
				|| array_reduce($user->getDomList(), function (bool $hasMatch, string $domain) use ($search) {
					return $hasMatch || stripos($domain, $search) !== false;
				}, false)
			;
		});
	}

	/**
	 * Gibt einen zufälligen Kunden zurück, bzw. den Hauptuser in der Basic-Edition
	 * @return ?static
	 */
	public static function randomClient(): ?self
	{
		if (Utils::isBaseEdition()) {
			return current(self::byRole(User::ROLE_ADMIN)) ?: null;
		}

		$array = self::getClients();
		if (!$array) {
			return null;
		}
		$key = array_rand($array);
		return $array[$key];
	}

	/**
	 * Diese Methode fügt der Konfiguration einen neuen Benutzer hinzu, allerdings wird die Anpassung nicht permanent
	 * geschrieben. Zu einem späteren Zeitpunkt kann dies mit User::flush nachgeholt werden.
	 * @param string|null $username
	 * @return self
	 * @throws LogicException
	 * @throws UserAlreadyExistsException
	 * @override
	 * @phan-return static
	 * @phan-assert string $username
	 * @phan-assert-true-condition $username
	 */
	public static function create(?string $username = null): Model
	{
		if (empty($username)) {
			throw new LogicException('Username may not be empty.');
		}

		self::beginWrite();
		if (self::exists($username)) {
			throw new UserAlreadyExistsException($username);
		}

		/** @var self $user */
		$user = parent::create();
		assert($user instanceof User);
		$user
			->setUsername($username)
			->updateCreateTimeStamp()
			->generateApiKey()
			->generateGeneralApiKey();

		//Default Sprache setzen nach BackendLang
		$defaultLang = BackendLanguage::findDefault();
		$user->setDefaultLocaleString($defaultLang ? $defaultLang->getLongCode() : "de_DE");

		return $user;
	}

	/**
	 * @param string $id
	 * @param bool $autocommit
	 */
	public static function delete(string $id, bool $autocommit=true): void
	{
		/** @var self|null $user */
		$user = self::byId($id);
		if (!$user) {
			return;
		}

		// Benutzerspezifische Repositories (Domains, Groups, UserThemes etc.) bereinigen
		// Unterbenutzer werden derzeit durch Controller separat entfernt (kein Bulk-Löschen für Hauptbenutzer)
		if (Utils::getRepositoryStore() == Utils::REPOSITORY_STORE_MONGODB) {
			// Sicherstellen, dass MediaFileService im korrekten Datenbereich arbeitet
			$oldUserOverride = self::getUserOverride();
			self::setUserOverride($user);

			// Account-Singletons nur für den Hauptbenutzer bereinigen
			if ($user->isAccountOwner()) {
				Domain::switchToUser($user);

				foreach (Domain::all() as $domain) {
					Domain::delete($domain->getId(), false);
				}
				Domain::flush();

				// Models sammeln, deren Objekte gelöscht werden sollen
				$userDeletePrepareEvent = new UserDeletePrepareEvent();
				Utils::dispatchEvent(UserDeletePrepareEvent::NAME, $userDeletePrepareEvent);

				$userSpecificModels = [
					DomainIndex::class,
					UserMigrationState::class,
					SubGroups::class,
					UserLocale::class,
					UserTheme::class,
				] + $userDeletePrepareEvent->getUserSpecificModels();

				/** @var class-string<Model> $className Kein Objekt, sondern der vollständige Klassenpfad */
				foreach ($userSpecificModels as $className) {
					// Wechsle in das korrekte Repository
					$className::switchRepositoryContext($user);

					// Daten entfernen
					$className::clear();
				}
			}

			// Mediendateien entfernen
			$user
				->setUserImage(null)
			;

			// Benutzer für den restlichen Programmablauf wiederherstellen
			self::setUserOverride($oldUserOverride);
		}

		Utils::removeDirectory($user->configDir());
		parent::delete($id, true); // Ignore $autocommit
	}

	/**
	 * Diese Methode loggt einen API-Benutzer ohne Session ins System ein.
	 * @param string $apiKey
	 * @return self|null
	 * @throws UserIsDeactivatedException
	 */
	public static function loginByGeneralApiKey(string $apiKey): ?self
	{
		self::$generalApiUser = null;
		$user = self::byGeneralApiKey($apiKey);

		if ($user) {
			if ($user->isActive() == false && $user->getRole() == User::ROLE_CLIENT) {
				throw new UserIsDeactivatedException($user->getUsername());
			}
			else {
				self::$generalApiUser = $user;
			}
		}

		return self::$generalApiUser;
	}

	/**
	 * Diese Methode loggt einen Benutzer ins System ein.
	 * @param string $username
	 * @param string $password
	 * @return self|null
	 * @throws ConfigNotWritableException
	 * @throws UserIsDeactivatedException
	 * @throws SecondFactorRequiredException
	 */
	public static function login(
		string $username,
		#[\SensitiveParameter]
		string $password
	): ?self {
		$user = self::find($username);

		if ($user) {
			$isBlocked = $user->isBlocked();
			//User ist inaktiv und ein CLient -> raus...
			if ($user->isActive() == false && $user->getRole() == User::ROLE_CLIENT) {
				throw new UserIsDeactivatedException($user->getUsername());
			}
			//Ist ein Subuser aber Hauptuser Account ist inaktiv - dann raus....
			elseif ($user->isSubUser() && $user->getAccountOwner()->isActive() == false) {
				throw new UserIsDeactivatedException($user->getUsername());
			}
			elseif ($user->checkPassword($password)) {
				if ($isBlocked) {
					return null;
				}

				// Passwort-Hash ggf. aktualisieren
				if (password_needs_rehash($user->getPasswordHash(), self::PASSWORD_ALGORITHM, [
					'cost' => self::PASSWORD_COST,
				])) {
					$user->setPassword($password)->save();
				}

				self::finalizeLogin($user);

				return $user;
			}
			else {
				$user
					->touchFailedLoginTimestamp()
					->save()
				;
			}
		}
		return null;
	}

	/**
	 * Diese Methode loggt einen Kunden ins System ein, ohne das Passwort zu prüfen.
	 *
	 * Ein passwortloser Login für einen Admin-Account schlägt immer fehl.
	 *
	 * @param string $username
	 * @return self|null
	 * @throws ConfigNotWritableException
	 * @throws UserIsDeactivatedException
	 * @throws SecondFactorRequiredException
	 */
	public static function loginPasswordless(string $username): ?self
	{
		$user = self::find($username);

		if ($user) {
			// Passwortlosen Login anderen Benutzern als Clients grundsätzlich nicht gestatten
			if ($user->getRole() !== User::ROLE_CLIENT) {
				return null;
			}

			$isBlocked = $user->isBlocked();

			// User ist inaktiv -> raus...
			if ($user->isActive() == false) {
				throw new UserIsDeactivatedException($user->getUsername());
			}
			// Ist ein Subuser aber Hauptuser Account ist inaktiv - dann raus....
			elseif ($user->isSubUser() && $user->getAccountOwner()->isActive() == false) {
				throw new UserIsDeactivatedException($user->getUsername());
			}
			else {
				if ($isBlocked) {
					return null;
				}
				self::finalizeLogin($user);
				return $user;
			}
		}
		return null;
	}

	/**
	 * Letzter Schritt im Login-Prozess nach dem Überprüfen
	 * der Credentials.
	 *
	 * Dispatcht das `LoginEvent` und wenn kein Modul noch
	 * 2FA erfordert oder den Login noch abbricht, wird
	 * die Session für den User vorbereitet und der Zeitpunkt
	 * des letzten Logins gesetzt.
	 *
	 * @param User $user Der einzuloggende Benutzer
	 * @return void
	 * @throws UserIsDeactivatedException
	 * @throws SecondFactorRequiredException
	 */
	private static function finalizeLogin(User $user): void
	{
		// Session-Resume-Token ggf. setzen
		if ($user->getSessionResumeToken() === null) {
			$user->generateSessionResumeToken();
		}

		$user
			->updateLastLogin()
			->save()
		;

		$event = new LoginEvent($user);
		Utils::dispatchEvent(LoginEvent::NAME, $event);

		if ($event->isDispositionAllow()) {
			$user->switchTo();
		} elseif ($event->isDispositionReject()) {
			throw new UserIsDeactivatedException($user->getUsername(), $event->getRejectReason());
		} else {
			$user->setLoginInProgress($event->getSecondFactorAuths());
			throw new SecondFactorRequiredException($user, $event->getSecondFactorAuths());
		}
	}

	/**
	 * Vermerkt den Benutzer in der Session als "Login im Gange" für 2-Faktor-Authentifizierung
	 * @param array[] $secondFactorMethods
	 * @return self
	 */
	public function setLoginInProgress(array $secondFactorMethods = []): self
	{
		$session = Utils::getSession();
		$session->set('2faUserId', $this->getId());
		$session->set('2faTimestamp', (int)time());
		$session->set('2faMethods', $secondFactorMethods);
		return $this;
	}

	/**
	 * Diese Methode loggt einen Benutzer ins System ein, ohne Passwortabfrage.
	 * @param bool $keepPreviousUserInfo Ob gespeichert werden soll, wer ursprünglich eingeloggt war, um zurückwechseln zu können.
	 * @return self
	 * @throws ConfigNotWritableException
	 */
	public function switchTo(bool $keepPreviousUserInfo=false): self
	{
		$prevUser = User::loggedInUser();
		$session = Utils::getSession();
		$locale = $session->get('locale');
		$session->clear();
		$session->set('2faUserId', null);
		$session->set('userId', $this->getId());
		$session->set('login_timestamp', (int)time());
		$session->set('last_timestamp', (int)time());
		if ($locale) {
			$session->set('locale', $locale);
		}
		if ($keepPreviousUserInfo && $prevUser->getId() !== $this->getId()) {
			$session->set('originalUserId', $prevUser->getId());
		}
		return $this;
	}

	/**
	 * Liest alle Einträge neu von der Festplatte ein.
	 * Sämtliche Änderungen gehen verloren.
	 * @throws ConfigNotReadableException
	 * @override
	 */
	public static function reload(): void
	{
		parent::reload();
		self::$nameLookupCache = [];
	}

	/**
	 * Diese Methode schreibt alle durch Setter-Methoden vorgenommenen
	 * Änderungen im Model in die Konfiguration auf der Platte.
	 * @return void
	 * @throws ConfigNotWritableException
	 * @override
	 */
	public static function flush(): void
	{
		parent::flush();
		self::$nameLookupCache = [];
	}

	/**
	 * Diese Methode loggt den aktiven Benutzer aus dem System aus.
	 */
	public static function logout(): void
	{
		$session = Utils::getSession();
		$session->set('2faUserId', null);
		$session->set('2faTimestamp', null);
		$session->set('2faMethods', null);
		$session->set('userId', null);
		$session->set('login_timestamp', null);
		$session->set('last_timestamp', 0);
		$session->set('csrf_token', null);
		$session->set('info', null);
		$session->getFlashBag()->clear();
		$session->clear();

	}

	/**
	 * Diese Methode nennt das benutzerspezifische Konfigurationsverzeichnis.
	 * @param bool $rel Wenn true, wird der Pfad relativ zum Verzeichnis var/ zurückgegeben, ansonsten als abosluter Pfad.
	 * @return string
	 */
	public function configDir(bool $rel = false): string
	{
		$relPath = 'user_data/'.$this->getId();
		if ($rel) {
			return $relPath;
		}

		return Utils::getBundledVarDir() . DIRECTORY_SEPARATOR . $relPath;
	}

	/**
	 * Dies ist eine Sammelmethode, die diverse Statistiken ermittelt und festhält.
	 * @return $this
	 */
	public function updateDomainInfo(): self
	{
		return $this
			->setActualDomainCount(Domain::countIncludingShareNames())
			->setActualWLCount(Domain::whitelabelCountForAllDomains())
			->setDomList(array_values(array_map(function (Domain $domain) {
				return $domain->getUrlHost() . ($domain->getUrlPath() == '/' ? '' : $domain->getUrlPath());
			}, Domain::all())))
		;
	}

	/**
	 * @return string
	 */
	public function getUsername(): string
	{
		return (string)$this->get('username');
	}


	/**
	 * @param string $username
	 * @return $this
	 */
	public function setUsername(string $username): self
	{
		$this->set('username', $username);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getUserImageIdentifier(): string
	{
		return (string)$this->get('userImageFilename');
	}

	public function getUserImage(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getSubuserRepository('images', $this);
		$identifier = $this->getUserImageIdentifier();
		return ($identifier == '') ? null : $repo->getFile($identifier);
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setUserImage(?File $file): self
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getSubuserRepository('images', $this);
		$prevIdentifier = $this->getUserImageIdentifier();

		$identifier = $repo->replaceFile($prevIdentifier, $file, 'avatar');

		$this->set('userImageFilename', $identifier);

		return $this;
	}

	/**
	 * Holt die Userid des Users dem der Account gehört...
	 * @return string
	 */
	public function getSubUserOf(): string
	{
		return (string)$this->get('subUserOf');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setSubUserOf(string $value): self
	{
		$this->set('subUserOf', $value);
		return $this;
	}

	public function subUserHasRights(): bool
	{
		return true;
	}

	/**
	 * @return self|$this
	 */
	public function getAccountOwner(): self
	{
		return self::byId($this->getSubUserOf()) ?? $this;
	}

	/**
	 * @return bool
	 */
	public function isAccountOwner(): bool
	{
		return $this->getId() == $this->getAccountOwner()->getId();
	}

	/**
	 * @return bool
	 */
	public function isAdministrator(): bool
	{
		return $this->getRole() == self::ROLE_ADMIN;
	}

	/**
	 * @return bool
	 */
	public function isClient(): bool
	{
		return $this->getRole() == self::ROLE_CLIENT;
	}

	/**
	 * @return bool
	 */
	public function isManager(): bool
	{
		return $this->getRole() == self::ROLE_MANAGER;
	}

	/**
	 * @return bool
	 */
	public function isSubUser(): bool
	{
		return (bool)$this->getSubUserOf();
	}

	/**
	 * @return bool
	 */
	public function hasAllPrivileges(): bool
	{
		return $this->get('allPrivileges') || $this->isAdministrator();
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setAllPrivileges(bool $state): self
	{
		$this->set('allPrivileges', $state);
		return $this;
	}

	/**
	 * @return string[]
	 */
	public function getPrivileges(): array
	{
		return (array)$this->get('privileges');
	}

	/**
	 * @param string[] $privileges
	 * @return $this
	 */
	public function setPrivileges(array $privileges): self
	{
		$privileges = array_unique(array_filter(array_map('trim', array_map('strval', $privileges))));
		natcasesort($privileges);
		$this->set('privileges', array_values($privileges));
		return $this;
	}

	/**
	 * Hilfsfunktion, die `getPrivileges` kapselt.
	 * @param string $name
	 * @return bool
	 */
	public function hasPrivilege(string $name): bool
	{
		return in_array($name, $this->getPrivileges());
	}

	/**
	 * Hilfsfunktion, die `getPrivileges` und `setPrivileges` kapselt.
	 * @param string $name
	 * @return $this
	 */
	public function grantPrivilege(string $name): self
	{
		$this->setPrivileges(array_merge($this->getPrivileges(), [$name]));
		return $this;
	}

	/**
	 * Hilfsfunktion, die `getPrivileges` und `setPrivileges` kapselt.
	 * @param string $name
	 * @return $this
	 */
	public function revokePrivilege(string $name): self
	{
		$this->setPrivileges(array_keys(array_diff_key(array_flip($this->getPrivileges()), [$name=>null])));
		return $this;
	}

	/**
	 * @return bool
	 */
	public function hasAccessToAllClients(): bool
	{
		return $this->get('accessToAllClients') || $this->isAdministrator();
	}

	/**
	 * @param bool $state
	 * @return $this
	 */
	public function setAccessToAllClients(bool $state): self
	{
		$this->set('accessToAllClients', $state);
		return $this;
	}

	/**
	 * This is a helper method that wraps logic. The user must either be assigned any of
	 * the client groups assigned to `$client` or have access to all clients.
	 *
	 * @param self $client
	 * @return bool
	 */
	public function mayAccessClient(self $client): bool
	{
		if (!$client->isClient()) {
			throw new LogicException("Client expected, got {$client->getRole()}.");
		}
		return $this->hasAccessToAllClients() || array_intersect_key($this->getAccessibleClientGroups(), $client->getClientGroups());
	}

	/**
	 * @return ClientGroup[]
	 */
	public function getClientGroups(): array
	{
		return array_intersect_key(ClientGroup::all(), array_flip((array)$this->getAccountOwner()->get('clientGroups')));
	}

	/**
	 * @param ClientGroup[] $groups
	 * @return $this
	 */
	public function setClientGroups(array $groups): self
	{
		if ($this->isAccountOwner()) {
			$groupIds = array_map(function (ClientGroup $group) { return $group->getId(); }, $groups);
			natcasesort($groupIds);
			$this->set('clientGroups', array_keys(array_flip($groupIds)));
		}
		return $this;
	}

	/**
	 * @param ClientGroup $group
	 * @return bool
	 */
	public function hasClientGroup(ClientGroup $group): bool
	{
		return isset($this->getClientGroups()[$group->getId()]);
	}

	/**
	 * @param ClientGroup $group
	 * @return self
	 */
	public function assignClientGroup(ClientGroup $group): self
	{
		$this->setClientGroups(array_merge($this->getClientGroups(), [$group]));
		return $this;
	}

	/**
	 * @param ClientGroup $group
	 * @return self
	 */
	public function removeClientGroup(ClientGroup $group): self
	{
		$this->setClientGroups(array_diff_key($this->getClientGroups(), [$group->getId()=>null]));
		return $this;
	}

	/**
	 * This is a helper method that wraps logic. The user must either be assigned the
	 * specified client group or have access to all clients.
	 *
	 * The result containing a client group does not mean that it would be reported
	 * when calling `getClientGroups`.
	 *
	 * @return ClientGroup[]
	 */
	public function getAccessibleClientGroups(): array
	{
		return $this->hasAccessToAllClients() ? ClientGroup::all() : $this->getClientGroups();
	}

	/**
	 * This is a helper method that wraps logic. The user must either be assigned the
	 * specified client group or have access to all clients.
	 *
	 * The return value `true` does not mean that the user would report the client group
	 * when calling `getClientGroups`.
	 *
	 * @param ClientGroup $group
	 * @return bool
	 */
	public function mayAccessClientGroup(ClientGroup $group): bool
	{
		return $this->hasAccessToAllClients() || isset($this->getClientGroups()[$group->getId()]);
	}

	/**
	 * @return self[]
	 */
	public function getSubusers(): array
	{
		return self::filterBy(['subUserOf'=>$this->getId()]);
	}

	/**
	 * Manuelles Setzen des API-Keys (nur für Import)
	 *
	 * @param string $apiKey
	 * @return $this
	 * @throws ApiKeyAlreadyExistsException
	 */
	public function setApiKey(string $apiKey): self
	{
		$user = self::byApiKey($apiKey);
		if ($user && $user->getId() != $this->getId()) {
			throw new ApiKeyAlreadyExistsException($apiKey);
		}

		$this->set('apiKey', $apiKey);
		return $this;
	}

	/**
	 * Manuelles Setzen des General-API-Keys (nur für Import)
	 * @param string $apiKey
	 * @return $this
	 * @throws ApiKeyAlreadyExistsException
	 */
	public function setGeneralApiKey(string $apiKey): self
	{
		if ($apiKey !== '') {
			$user = self::byGeneralApiKey($apiKey);
			if ($user && $user->getId() != $this->getId()) {
				throw new ApiKeyAlreadyExistsException('general:'.$apiKey);
			}
		}

		$this->set('generalApiKey', $apiKey);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getCreateTimeStamp(): int
	{
		return (int)$this->get('createtimestamp');
	}

	/**
	 * @param int|DateTimeInterface $timestamp
	 * @return $this
	 */
	public function setCreateTimeStamp($timestamp): self
	{
		if (!is_int($timestamp)) {
			$timestamp = $timestamp->getTimestamp();
		}
		$this->set('createtimestamp', $timestamp);
		return $this;
	}

	/**
	 * @return $this
	 */
	private function updateCreateTimeStamp(): self
	{
		$this->set('createtimestamp', time());
		return $this;
	}

	//setLastLogin
	/**
	 * @return int
	 */
	public function getLastLogin(): int
	{
		return (int)$this->get('lastlogin');
	}

	/**
	 * @param int|DateTimeInterface $timestamp
	 * @return $this
	 */
	public function setLastLogin($timestamp): self
	{
		if (!is_int($timestamp)) {
			$timestamp = $timestamp->getTimestamp();
		}
		$this->set('lastlogin', $timestamp);
		return $this;
	}

	/**
	 * @return $this
	 */
	public function updateLastLogin(): self
	{
		$this->set('lastlogin', time());
		return $this;
	}

	/**
	 * @return string
	 */
	public function getRole(): string
	{
		$role = (string)$this->get('role');
		// Updateroutine: Benutzerrolle Administrator in alten Versionen nachtragen
		if ($role === '' and self::count() == 1) {
			$this->setRole(self::ROLE_ADMIN)->save();
			return self::ROLE_ADMIN;
		}
		return $role;
	}

	/**
	 * @param string $role
	 * @return $this
	 */
	public function setRole(string $role): self
	{
		$this->set('role', $role);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getApiKey(): string
	{
		// Idealerweise würde direkt der API-Schlüssel des Hauptbenutzers zurückgegeben,
		// hat jedoch gravierende Auswirkung auf die Performance bei großer Benutzerdatenbank.
		$key = (string)$this->get('apiKey');

		// Updateroutine: API-Schlüssel für Benutzer generieren
		if ($key === '') {
			$this->generateApiKey();
			return $this->get('apiKey');
		}

		return $key;
	}

	/**
	 * @return string
	 */
	public function getGeneralApiKey(): string
	{
		return (string)$this->get('generalApiKey');
	}

	/**
	 * Diese Methode generiert einen unter allen Benutzern eindeutigen API-Schlüssel.
	 * @return $this
	 */
	public function generateApiKey(): self
	{
		// Keine API-Schlüssel für Unterbenutzer erzeugen
		if ($this->isSubUser()) {
			return $this;
		}

		$apiKey = null;
		do {
			$apiKey = bin2hex(random_bytes(24));
		} while (static::getRepository()->toArrayFiltered(['apiKey' => $apiKey]));

		$this->set('apiKey', $apiKey);
		return $this;
	}

	/**
	 * Diese Methode generiert einen unter allen Benutzern eindeutigen API-Schlüssel für die General Domain Api.
	 * @return $this
	 */
	public function generateGeneralApiKey(): self
	{
		$apiKey = null;
		do {
			$apiKey = bin2hex(random_bytes(24));
		} while (static::getRepository()->toArrayFiltered(['generalApiKey' => $apiKey]));

		$this->set('generalApiKey', $apiKey);
		return $this;
	}


	/**
	 * @param string $password
	 * @return $this
	 */
	public function setPassword(string $password): self
	{
		$passwordHash = password_hash($password, self::PASSWORD_ALGORITHM, [
			'cost' => self::PASSWORD_COST,
		]);

		return $this->setPasswordHash($passwordHash);
	}

	/**
	 * @param string $password
	 * @return bool
	 */
	public function checkPassword(string $password): bool
	{
		return password_verify($password, $this->getPasswordHash());
	}

	/**
	 * @return string
	 */
	public function getPasswordHash(): string
	{
		return (string)$this->get('passwordHash');
	}

	/**
	 * Diese Methode überschreibt den Passwort-Hash des Benutzers. Aufgrund bestimmter Anforderungen
	 * an die Hosting-API ist die Methode `public`. Nach Möglichkeit sollte vorzugsweise ein Passwort
	 * in Klartext über die Methode `setPassword` eingepflegt werden.
	 *
	 * @see setPassword
	 *
	 * @param string $hash
	 * @return $this
	 */
	public function setPasswordHash(string $hash): self
	{
		$this->set('passwordHash', $hash);
		return $this;
	}

	/**
	 * Entfernt das Passwort so, dass kein Login mit Passwort mehr möglich ist.
	 *
	 * Der Passwort-Hash wird dafür auf einen String gesetzt, der nie ein gültiger
	 * Passwort-Hash sein kann. Jeder Login schlägt damit automatisch fehl.
	 *
	 * @return $this
	 */
	public function unsetPassword(): self
	{
		$this->set('passwordHash', 'xNONEx');
		return $this;
	}

	/**
	 * Ob ein Passwort gesetzt ist. Wenn FALSE, ist ein Login nur über andere
	 * Methoden, wie das SSO-Plugin möglich.
	 *
	 * @return bool
	 */
	public function hasPassword(): bool
	{
		return !hash_equals('xNONEx', $this->get('passwordHash', 'xNONEx'));
	}

	/**
	 * Ob das Passwort gesperrt ist. Es kann dann nicht vom Benutzer selbst geändert werden.
	 *
	 * @return bool
	 */
	public function isPasswordLocked(): bool
	{
		return (bool)$this->get('passwordLocked', false);
	}

	/**
	 * Wird $locked auf TRUE gesetzt, kann der Benutzer sein eigenes Passwort nicht ändern
	 *
	 * @param bool $locked
	 * @return $this
	 */
	public function setPasswordLocked(bool $locked): self
	{
		$this->set('passwordLocked', $locked);
		return $this;
	}

	/**
	 * @return int Unix-Timestamp
	 */
	public function getFailedLoginTimestamp(): int
	{
		return (int)$this->get('failedLoginTimestamp');
	}

	/**
	 * @return $this
	 */
	public function touchFailedLoginTimestamp(): self
	{
		$this->set('failedLoginTimestamp', time());
		return $this;
	}

	/**
	 * Prüft, ob ein Login für diesen User kurzzeitig fehlschlagen soll
	 *
	 * @return bool
	 */
	public function isBlocked(): bool
	{
		$time = $this->getFailedLoginTimestamp();

		// Wenn zu schnell Passwörter durchprobiert werden, blockieren
		return (time() - $time < self::LOGIN_BLOCKED_DURATION);
	}

	/**
	 * @return string
	 */
	public function getFirstName(): string
	{
		return (string)$this->get('firstName');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setFirstName(string $value): self
	{
		$this->set('firstName', $value);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getLastName(): string
	{
		return (string)$this->get('lastName');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setLastName(string $value): self
	{
		$this->set('lastName', $value);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getEmailAddress(): string
	{
		return (string)$this->get('emailAddress');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setEmailAddress(string $value): self
	{
		$this->set('emailAddress', $value);
		return $this;
	}

	/**
	 * @param bool $fallback Bei TRUE Fallback auf die Haupt-E-Mail-Adresse
	 * @return string
	 */
	public function getInvoiceEmailAddress(bool $fallback=false): string
	{
		$result = (string)$this->get('invoiceEmailAddress');
		if ($fallback and $result == '') {
			$result = $this->getEmailAddress();
		}
		return $result;
	}

	/**
	 * Rechnungs-E-Mail-Adresse
	 * @param string|null $value
	 * @return $this
	 */
	public function setInvoiceEmailAddress(?string $value): self
	{
		if ($value === '') {
			$value = null;
		}
		$this->set('invoiceEmailAddress', $value);
		return $this;
	}

	/**
	 * Mahnungs-E-Mail-Adresse
	 * @param bool $fallback Bei TRUE Fallback auf die Haupt-E-Mail-Adresse
	 * @return string
	 */
	public function getDunningEmailAddress(bool $fallback=false): string
	{
		$result = (string)$this->get('dunningEmailAddress');
		if ($fallback and $result == '') {
			$result = $this->getEmailAddress();
		}
		return $result;
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setDunningEmailAddress(?string $value): self
	{
		if ($value === '') {
			$value = null;
		}
		$this->set('dunningEmailAddress', $value);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getStatus(): string
	{
		return (string)$this->get('status');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setStatus(string $value): self
	{
		$this->set('status', $value);
		return $this;
	}
	/**
	 * @return string
	 */
	public function getToken(): string
	{
		return (string)$this->get('token');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setToken(string $value): self
	{
		$this->set('token', $value);
		return $this;
	}


	/**
	 * @return string
	 */
	public function getCompany(): string
	{
		return (string)$this->get('company');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setCompany(string $value): self
	{
		$this->set('company', $value);
		return $this;
	}

	/**
	 * @return array
	 */
	public function getDomActive(): array
	{
		return (array)$this->get('DomActive');
	}

	/**
	 * @param array $value
	 * @return $this
	 */
	public function setDomActive($value): self
	{
		$this->set('DomActive', (array)$value);
		return $this;
	}

	/**
	 * @param Domain $domain
	 * @return $this
	 */
	public function grantDomainAccess(Domain $domain): self
	{
		if ($this->isSubUser()) {
			$domainAccess = $this->getDomActive();
			$domainAccess[$domain->getId()] = 'on';
			$this->setDomActive($domainAccess);
		}

		return $this;
	}

	/**
	 * Set von IDs der Gruppen, in denen dieser Subuser Mitglied ist
	 * @return array<string, bool>
	 */
	public function getGroupActive(): array
	{
		return (array)$this->get('GroupActive');
	}

	/**
	 * Gruppen, in denen dieser Subuser Mitglied ist
	 * @return SubGroups[]
	 */
	public function getMemberGroups(): array
	{
		$membership = $this->getGroupActive();
		return array_filter(
			SubGroups::all(),
			function ($group) use ($membership) {
				return !empty($membership[$group->getId()]);
			}
		);
	}

	/**
	 * @param array $value
	 * @return $this
	 */
	public function setGroupActive($value): self
	{
		$this->set('GroupActive', (array)$value);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isOnBoarding(): bool
	{
		return (bool)($this->get('OnBoarding') ?? !$this->get('NoOnboarding', true));
	}

	/**
	 * @param bool $OnBoarding
	 * @return $this
	 */
	public function setOnBoarding(bool $OnBoarding): self
	{
		$this->set('OnBoarding', $OnBoarding);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isActive(): bool
	{
		if ($this->isAdministrator()) {
			return true;
		}
		return (bool)$this->get('active');
	}

	/**
	 * @param bool $active
	 * @return $this
	 */
	public function setActive(bool $active): self
	{
		$this->set('active', $active);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getMaxWLCount(): int
	{
		return (int)$this->get('maxWLCount');
	}

	/**
	 * @param int $value
	 * @return $this
	 */
	public function setMaxWLCount($value): self
	{
		$this->set('maxWLCount', (int)$value);
		return $this;
	}

	/**
	 * @return array
	 */
	public function getDomList(): array
	{
		return (array)$this->get('domlist');
	}

	/**
	 * @param array $value
	 * @return $this
	 */
	public function setDomList($value): self
	{
		$this->set('domlist', (array)$value);
		return $this;
	}

	/**
	 * Metadaten zur Zuordnung des Nutzers in Drittsystemen
	 *
	 * Wird nur in der Hosting-API gesetzt und gelesen.
	 * @return array<string|int,mixed>
	 */
	public function getExternalMetadata(): array
	{
		return (array)$this->get('externalMetadata', []);
	}

	/**
	 * @see self::getExternalMetadata()
	 * @param array<string|int,mixed> $value
	 * @return $this
	 */
	public function setExternalMetadata($value): self
	{
		$this->set('externalMetadata', (array)$value);
		return $this;
	}

	/**
	 * @return array
	 */
	public function getDomCKList(): array
	{
		return (array)$this->get('domcklist');
	}

	/**
	 * @param array $value
	 * @return $this
	 */
	public function setDomCKList($value): self
	{
		$this->set('domcklist', (array)$value);
		return $this;
	}


	/**
	 * @return array
	 */
	public function getDomCK30DayFreeList(): array
	{
		return (array)$this->get('DomCK30DayFreeList');
	}

	/**
	 * @param array $value
	 * @return $this
	 */
	public function setDomCK30DayFreeList($value): self
	{
		$this->set('DomCK30DayFreeList', (array)$value);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getActualWLCount(): int
	{
		return (int)$this->get('actualWLCount');
	}

	/**
	 * @param int $value
	 * @return $this
	 */
	public function setActualWLCount($value): self
	{
		$this->set('actualWLCount', (int)$value);
		return $this;
	}

	/**
	 * @return string
	 */
	public function getPasswordRecoveryKey(): string
	{
		return (string)$this->get('passwordRecoveryKey');
	}

	/**
	 * @return DateTimeInterface
	 */
	public function getPasswordRecoveryTime(): DateTimeInterface
	{
		$value = $this->get('passwordRecoveryTime');
		if (!$value) {
			$value = '1970-01-01 00:00:00';
		}
		return new DateTime((string)$value);
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setPasswordRecoveryKey(?string $value): self
	{
		if ($value === null) {
			$this->set('passwordRecoveryKey', null);
			$this->set('passwordRecoveryTime', null);
		}
		else {
			$this->set('passwordRecoveryKey', $value);
			$this->set('passwordRecoveryTime', (new DateTime('now'))->format('Y-m-d H:i:s'));
		}
		return $this;
	}

	/**
	 * @return LocaleBase
	 */
	public function getDefaultLocale(): LocaleBase
	{
		$locale = LocaleService::find((string)$this->get('defaultLocale'));

		// Determine default locale if it is not set or if no corresponding user locale exists
		if ($locale === null || !($locale instanceof UserLocale) && !UserLocale::exists($locale->getName())) {
			// Store locale names as keys to avoid duplicates and to keep the order
			$localePriorityList = [];

			if ($request = Utils::getCurrentRequest()) {
				$localePriorityList[$request->getLocale()] = null;
			}

			if ($defaultBackendLanguage = BackendLanguage::findDefault()) {
				$localePriorityList[$defaultBackendLanguage->getLanguage()->getCode()] = null;
			}

			$localePriorityList += array_flip(LocaleService::DEFAULT_PRIORITY);

			// Add language codes to the list for fuzzy matching
			$localePriorityList = array_reduce(array_keys($localePriorityList), static function ($carry, $item) {
				$carry[$item] = null;
				$carry[explode('_', $item, 2)[0]] = null;
				return $carry;
			}, []);

			// Convert keys to the order of the priority list
			$localePriorityList = array_keys($localePriorityList);

			// Get enabled user locales or, alternatively, all user locales
			$userLocales = UserLocale::enabledLocales() ?: UserLocale::all();
			assert(count($userLocales) > 0);

			// Find the best matching locale and set it as the default locale
			$locale = LocaleService::anyLocaleFuzzy($localePriorityList, $userLocales);
			$this->setDefaultLocale($locale)->save();

			// Switch to domain locale if available
			$locale = LocaleService::find($locale->getName());
		}

		assert($locale instanceof LocaleBase);
		return $locale;
	}

	/**
	 * @param LocaleBase $locale
	 * @return $this
	 */
	public function setDefaultLocale(LocaleBase $locale): self
	{
		$this->set('defaultLocale', $locale->getName());
		return $this;
	}

	/**
	 * @param string $localeLongCode
	 * @return $this
	 */
	public function setDefaultLocaleString(string $localeLongCode): self
	{
		$this->set('defaultLocale', $localeLongCode);
		return $this;
	}

	/**
	 * @return int
	 */
	public function getMaxDomainCount(): int
	{
		return (int)$this->get('maxDomainCount');
	}

	/**
	 * @param int $value
	 * @return $this
	 */
	public function setMaxDomainCount(int $value): self
	{
		if ($value < -1) {
			throw new \OutOfRangeException("$value is no valid maximum domain count");
		}
		$this->set('maxDomainCount', $value);
		return $this;
	}
	//MaxCallCount
	/**
	 * @return int
	 */
	public function getMaxCallCount(): int
	{
		return (int)$this->get('MaxCallCount');
	}

	/**
	 * @param int $value
	 * @return $this
	 */
	public function setActualDomainCount(int $value): self
	{
		$this->set('actualdomaincount', $value);
		return $this;
	}
	/**
	 * @return int
	 */
	public function getActualDomainCount(): int
	{
		return (int)$this->get('actualdomaincount');
	}

	/**
	 * This method returns the number of domains that can still be added to the account.
	 * The result is based on the account owner's domain limit.
	 * @return int number of available domains or -1 if unlimited
	 */
	public function getAvailableDomainCount(): int
	{
		$owner = $this->getAccountOwner();
		$domainCount = $owner->getActualDomainCount();
		$domainMaxCount = $owner->getMaxDomainCount();
		return ($domainMaxCount > -1) ? max(0, $domainMaxCount - $domainCount) : -1;
	}

	/**
	 * This method determines whether the domain limit has been reached.
	 * The result is based on the account owner's domain limit.
	 * @return bool
	 */
	public function isDomainLimitReached(): bool
	{
		return $this->getAvailableDomainCount() == 0;
	}

	/**
	 * This method checks whether the user has enough free domain slots to add the given number of domains.
	 * The result is based on the account owner's domain limit.
	 * @param int $quantity
	 * @return bool
	 */
	public function hasFreeDomainSlots(int $quantity): bool
	{
		$available = $this->getAvailableDomainCount();
		return $available == -1 || $quantity <= $available;
	}

	/**
	 * @return int
	 */
	public function getActualCallCount(): int
	{
		return (int)$this->get('actualCallCount');
	}

	/**
	 * @param int $count
	 * @return $this
	 */
	public function setCallCount(int $count): self
	{
		$this->set('actualCallCount', $count);
		return $this;
	}

	//MaxCallCount
	/**
	 * @return string
	 */
	public function getActualTarif(): string
	{
		return (string)$this->get('actualTarif');
	}

	/**
	 * @param string $value
	 * @return $this
	 */
	public function setActualTarif(string $value): self
	{
		$this->set('actualTarif', $value);
		return $this;
	}


	/**
	 * @return $this
	 */
	public function incrementCallCount(int $diff=1): self
	{
		$this->increment('actualCallCount', $diff);
		return $this;
	}

	/**
	 * Gibt zurück ob der User die maximale Anzahl an Aufrufen erreicht hat.
	 * Zur Sicherheit gibt es ein Überlauf von 10 % - damit die Banner nicht sofort abgeschaltet werden.
	 * Das ist ein Kulanz-Feature.
	 *
	 * @return bool
	 */
	public function callCountReached(): bool
	{
		//FIX alte Versionen - ist eine Multidomain aber kein MaxCallCount Wert - dann muss es eine alte Version sein, bei den base zählt eh nicht...
		if(Utils::isMultiDomainEdition() && ($this->getMaxCallCount() != -1 && $this->getMaxCallCount() < 20000))
		{
			$this->setMaxCallCount(-1);
			$this->save();
		}
		$callCount = $this->getActualCallCount();
		$callMaxCount = $this->getMaxCallCount();

		//10 % Überlauf Schutz aus Kulanz
		return ($callMaxCount > -1) && ($callCount >= ($callMaxCount*1.1));
	}

	/**
	 * Diese Methode setzt den Aufrufszähler zurück auf 0, wenn ein neuer Monat begonnen hat.
	 * @return $this
	 */
	public function resetCallCountForNextMonth(): self
	{
		$previousMonth = (new DateTime('@'.max(0, (int)$this->get('callCountMonthTimestamp'))))->setTimezone(new DateTimeZone('Europe/Berlin'));
		$currentMonth = new DateTime(date('Y-m-01 00:00:00'), new DateTimeZone('Europe/Berlin'));

		if ($currentMonth != $previousMonth) {
			$this->set('callCountMonthTimestamp', $currentMonth->getTimestamp());
			$this
				->setCallCount(0)
				->setLastCallReportQuota(0)
			;
		}

		return $this;
	}

	/**
	 * @return int
	 */
	public function getLastCallReportQuota(): int
	{
		return (int)$this->get('lastCallReportQuota');
	}

	/**
	 * @param int $quota
	 * @return $this
	 */
	public function setLastCallReportQuota(int $quota): self
	{
		$this->set('lastCallReportQuota', (int)min(max(0, $quota), 100));
		return $this;
	}

	/**
	 * @param int $value
	 * @return $this
	 */
	public function setMaxCallCount(int $value): self
	{
		if ($value < -1) {
			throw new \OutOfRangeException("$value is no valid maximum domain count");
		}
		$this->set('MaxCallCount', $value);
		return $this;
	}

	/**
	 * @return bool
	 */
	public function isAutoAddDomains(): bool
	{
		return (bool)$this->get('autoAddDomains', false);
	}

	/**
	 * @param bool $value
	 * @return $this
	 */
	public function setAutoAddDomains(bool $value): self
	{
		$this->set('autoAddDomains', $value);
		return $this;
	}

	/**
	 * Bevorzugtes Farbschema, falls der User eines gewählt hat.
	 *
	 * Aktuell sind 'dark' und 'light' definiert.
	 *
	 * @see User::COLOR_SCHEMES
	 * @return string|null
	 */
	public function getPreferredColorScheme(): ?string
	{
		return $this->get('preferredColorScheme') ?? null;
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setPreferredColorScheme(?string $value): User
	{
		if ($value !== null and !in_array($value, self::COLOR_SCHEMES)) {
			$value = null;
		}
		$this->set('preferredColorScheme', $value);
		return $this;
	}

	/**
	 * Id der Agentur, wenn vorhanden
	 */
	public function getAgencyAffiliateId(): ?string
	{
		return $this->get('agencyAffiliateId') ?? null;
	}

	/**
	 * Id der Agentur, wenn vorhanden
	 */
	public function setAgencyAffiliateId(?string $value): User
	{
		$this->set('agencyAffiliateId', $value);
		return $this;
	}

	/**
	 * Id der Agentur, wenn vorhanden
	 */
	public function getAgencyId(): ?string
	{
		return $this->get('agencyId') ?? null;
	}

	/**
	 * Id der Agentur, wenn vorhanden
	 */
	public function setAgencyId(?string $value): User
	{
		$this->set('agencyId', $value);
		return $this;
	}

	/**
	 * adminInfo, wenn vorhanden
	 * @return string
	 */
	public function getAdminInfo(): string
	{
		return (string)$this->get('adminInfo');
	}

	/**
	 * adminInfo, wenn vorhanden
	 * @return $this
	 */
	public function setAdminInfo(string $value): self
	{
		$this->set('adminInfo', $value);
		return $this;
	}

	/**
	 * VatId
	 * @return string|null
	 */
	public function getVatId(): ?string
	{
		return $this->get('vatId') ?? null;
	}

	/**
	 * @param string|null $value
	 * @return $this
	 * VatId
	 */
	public function setVatId(?string $value): User
	{
		$this->set('vatId', $value);
		return $this;
	}

	/**
	 * @return string|null
	 */
	public function getStrNr(): ?string
	{
		return $this->get('strNr') ?? null;
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setStrNr(?string $value): User
	{
		$this->set('strNr', $value);
		return $this;
	}

	/**
	 * @return string|null
	 */
	public function getPlz(): ?string
	{
		return $this->get('plz') ?? null;
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setPlz(?string $value): User
	{
		$this->set('plz', $value);
		return $this;
	}


	/**
	 * @return string|null
	 */
	public function getCity(): ?string
	{
		return $this->get('city') ?? null;
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setCity(?string $value): User
	{
		$this->set('city', $value);
		return $this;
	}

	/**
	 * @return string|null
	 */
	public function getPayment(): ?string
	{
		return $this->get('payment') ?? 3;
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setPayment(?string $value): User
	{
		$this->set('payment', $value);
		return $this;
	}

	/**
	 * @return array|null
	 */
	public function getKontoDaten(): ?array
	{
		return $this->get('kontodaten') ?? array();
	}

	/**
	 * @param array|null $value
	 * @return $this
	 * VatId
	 */
	public function setKontoDaten(?array $value): User
	{
		$this->set('kontodaten', $value);
		return $this;
	}

	/**
	 * Letztes Versionsupdate, über das der Nutzer informiert wurde
	 */
	public function getLastNotifiedVersion(): ?string
	{
		return $this->get('lastNotifiedVersion') ?? null;
	}

	/**
	 * @param string|null $value
	 * @return $this
	 */
	public function setLastNotifiedVersion(?string $value): self
	{
		$this->set('lastNotifiedVersion', $value);
		return $this;
	}

	/**
	 * Land des Nutzers
	 */
	public function getCountry(): ?Country
	{
		return Country::find($this->getCountryCode());
	}

	public function getCountryCode(): string
	{
		return $this->get('country', 'DE');
	}

	/**
	 * @param string $country
	 * @return $this
	 */
	private function setCountryCode(string $country): self
	{
		$this->set('country', $country);
		return $this;
	}

	/**
	 * @param Country|null $country
	 * @return $this
	 */
	public function setCountry(?Country $country): self
	{
		if (!$country) {
			$country = Country::find('ZZ');
			assert($country instanceof Country);
		}
		$this->setCountryCode($country->getCode());
		return $this;
	}

	/**
	 * Standard-Hauptland der angebotenen Dienste
	 *
	 * Bei fehlender Auswahl wird das Land des Users zurückgegeben.
	 */
	public function getDefaultServiceCountry(): ?Country
	{
		$code = $this->getDefaultServiceCountryCode();
		return $code ? Country::find($code) : $this->getCountry();
	}

	public function getDefaultServiceCountryCode(): ?string
	{
		return $this->get('defaultServiceCountry', null);
	}

	/**
	 * @param Country|null $country
	 * @return $this
	 */
	public function setDefaultServiceCountry(?Country $country): self
	{
		$this->set('defaultServiceCountry', $country ? $country->getCode() : null);
		return $this;
	}

	/**
	 * Standard-Rechtsgrundlagen ('eu-gdpr', 'ch-dsg', 'us-ccpa')
	 *
	 * @return string[]
	 */
	public function getDefaultLegalBases(): array
	{
		return (array)$this->get('defaultLegalBases', ['eu-gdpr', 'ch-dsg']);
	}

	/**
	 * @param string[] $values
	 * @return $this
	 */
	public function setDefaultLegalBases(array $values): self
	{
		$this->set('defaultLegalBases', $values);
		return $this;
	}

	public function getTwoFactorAuthParameters(string $pluginId): array
	{
		return $this->get('2fa.'.$pluginId, []) ?? [];
	}

	/**
	 * @param string $pluginId
	 * @param array $parameters
	 * @return $this
	 */
	public function setTwoFactorAuthParameters(string $pluginId, array $parameters): self
	{
		$this->set('2fa.'.$pluginId, $parameters);
		return $this;
	}

	/**
	 * Nur für Import/Export
	 * @return array<string,array>
	 */
	public function getAllTwoFactorAuthParameters(): array
	{
		return $this->get('2fa', []);
	}

	/**
	 * Nur für Import/Export
	 * @param array<string,array> $data
	 * @return $this
	 */
	public function setAllTwoFactorAuthParameters(array $data): self
	{
		$this->set('2fa', $data);
		return $this;
	}

	/**
	 * Entfernt alle 2FA-Einstellungen des Nutzers
	 *
	 * @return $this
	 */
	public function clearAllTwoFactorAuthParameters(): self
	{
		$this->set('2fa', []);
		return $this;
	}

	/**
	 * Prüft, ob irgendein 2FA-Plugin Einstellungen hinterlegt hat
	 */
	public function hasTwoFactorAuthParameters(): bool
	{
		$array = $this->get('2fa', []) ?? [];
		return Utils::any($array, function ($item) { return (bool)$item; });
	}

	/**
	 * Token, mit dem eine Session ohne erneute 2FA wiederhergestellt werden kann
	 */
	public function getSessionResumeToken(): ?string
	{
		return $this->get('resumeToken', null);
	}

	/**
	 * @param string|null $token
	 * @return $this
	 */
	public function setSessionResumeToken(?string $token): self
	{
		$this->set('resumeToken', $token);
		return $this;
	}

	/**
	 * @return $this
	 */
	public function generateSessionResumeToken(): self
	{
		$this->set('resumeToken', base64_encode(random_bytes(63)));
		return $this;
	}

	/**
	 * @return bool[]
	 */
	public function getAccountFeatures(): array
	{
		if ($this->accountFeatures === null) {
			$this->accountFeatures = $this->getAccountOwner()->getSettings();
			// Guarantee all boolean values
			$this->accountFeatures = array_filter($this->accountFeatures, static function ($x) { return is_bool($x); });
		}
		return $this->accountFeatures;
	}

	/**
	 * @param string $feature
	 * @return bool
	 */
	public function hasAccountFeature(string $feature): bool
	{
		$featureValue = $this->getAccountFeatures()[$feature] ?? null;
		return is_bool($featureValue) ? $featureValue : false;
	}

	/**
	 * @return int
	 */
	public function getAccountFeatureMaxCrawlBudget(): int
	{
		return $this->getSettings()['maxCrawlBudget'] ?? 1;
	}

	/**
	 * @return int
	 */
	public function getAccountFeatureExternalCrawlerFrequency(): int
	{
		return $this->getSettings()['externalCrawlerFrequency'] ?? 0;
	}

	/**
	 * @param string $feature
	 * @return bool
	 */
	public function isAccountFeatureConfigured(string $feature): bool
	{
		return array_key_exists($feature, $this->getAccountFeatures());
	}

	/**
	 * @return array<string,bool|int>
	 */
	public function getSettings(): array
	{
		$result = (array)$this->get('Settings');

		/**
		 * Fallback für alte Accounts damit die auch die korrekten Settings haben, und Grandfather: alles auf true...
		 * aber nur wenn keine Settings vorhanden sind...
		 **/
		if (!$result) {
			$result = self::buildDefaultSettings(true);
		}

		return $result;
	}

	/**
	 * Setzt Rechte für den Kunden anhand eines Tarif-Namens
	 *
	 * @param string $value Tarif
	 */
	public function setUserTarifSettings(string $value): void
	{
		$userNewSettings = self::getSettings();

		if ($value=="starter")
		{
			$userNewSettings['moreThan2Lang'] = false;
			$userNewSettings['canUseAnalytics'] = false;
			$userNewSettings['canUseABTests'] = false;
			$userNewSettings['canUseScripts'] = false;
			$userNewSettings['canUseIframes'] = true;
			$userNewSettings['canUseConsentSharing'] = false;
			$userNewSettings['canUseIndCSS'] = false;
			$userNewSettings['canUseIframeCSS'] = false;
			$userNewSettings['canUseIAB'] = false;
			$userNewSettings['canUseUserGroupRights'] = false;
			$userNewSettings['canUseDomAPI'] = false;
			$userNewSettings['canUseExternalCrawler'] = true;
			$userNewSettings['externalCrawlerFrequency'] = 0;
			$userNewSettings['maxCrawlBudget'] = 1;
			$userNewSettings['canUsePlugin/ccm19/batch_tag_manager'] = false;
			$userNewSettings['canUsePlugin/ccm19/script_placeholder'] = false;

			$this->setSettings($userNewSettings);
		}

		if ($value=="base")
		{
			$userNewSettings['moreThan2Lang'] = true;
			$userNewSettings['canUseAnalytics'] = true;
			$userNewSettings['canUseABTests'] = true;
			$userNewSettings['canUseScripts'] = true;
			$userNewSettings['canUseIframes'] = true;
			$userNewSettings['canUseConsentSharing'] = false;
			$userNewSettings['canUseIndCSS'] = true;
			$userNewSettings['canUseIframeCSS'] = true;
			$userNewSettings['canUseIAB'] = true;
			$userNewSettings['canUseUserGroupRights'] = true;
			$userNewSettings['canUseDomAPI'] = true;
			$userNewSettings['canUseMultiDom'] = false;
			$userNewSettings['canUsePlugin/ccm19/batch_tag_manager'] = false;
			$userNewSettings['canUsePlugin/ccm19/script_placeholder'] = false;
			//Crawler ist hier keine Option

			$this->setSettings($userNewSettings);
		}

		// Business auch alles - es sind ja die Agency Versionseinstellungen die in der EP anders sind
		// und die verfügbaren Domains ....
		if ($value=="business")
		{
			$this->setSettings(self::buildDefaultSettings(true));
		}

		//EP dann alles ON und raus...
		if ($value=="enterprise")
		{
			$this->setSettings(self::buildDefaultSettings(true));
		}
	}

	/**
	 * @param array<string,bool|int> $settings
	 */
	public function setSettings($settings): self
	{
		if(is_array($settings))
		{
			uksort($settings, 'strnatcasecmp');
			$this->set('Settings',$settings);

			if ($this->isAccountOwner()) {
				$this->accountFeatures = $settings;
			}
		}
		return $this;
	}

	/**
	 * Gibt alle standardmäßig vorhandenen Rechte zurück
	 *
	 * @param bool $on Ob alle Rechte auf aktiv gesetzt sein sollen
	 * @return array<string,bool|int>
	 */
	public static function buildDefaultSettings(bool $on=false): array
	{
		return [
			'moreThan2Lang' => $on,
			'canUseAnalytics' =>  $on,
			'canUseABTests' =>  $on,
			'canUseScripts' =>  $on,
			'canUseIframes' =>  $on,
			'canUseConsentSharing' =>  $on,
			'canUseOwnLogo' =>  $on,
			'canUseIndCSS' =>  $on,
			'canUseIframeCSS' =>  $on,
			'canUseMultiDom' =>  $on,
			'canUseIAB' =>  $on,
			'canUseUserGroupRights' =>  $on,
			'canUseDomAPI' =>  $on,
			'canUseExternalCrawler' =>  false,
			'externalCrawlerFrequency'	=> 0,
			'maxCrawlBudget'	=> 1,
		];
	}

	/**
	 * @return array<string,bool>
	 */
	public function getMailingPreferences(): array
	{
		return (array)$this->get('mailPref', []);
	}

	/**
	 * @param array<string,bool> $preferences
	 * @return $this
	 */
	public function setMailingPreferences(array $preferences): self
	{
		$this->set('mailPref', $preferences);
		return $this;
	}

	/**
	 * @param string $key
	 * @param bool $default
	 * @return bool
	 */
	public function isMailingSubscribed(string $key, bool $default = false): bool
	{
		return (bool)$this->get('mailPref.'.static::getRepository()->escapeKey($key), $default);
	}

	/**
	 * @param string $key
	 * @return $this
	 */
	public function unsubscribeMailing(string $key): self
	{
		$this->set('mailPref.'.static::getRepository()->escapeKey($key), false);
		return $this;
	}

	public function hasMandatoryBillingAddress(): bool
	{
		return (Utils::isCCMCloud() || Utils::isClickskeksCloud()) && $this->isClient() && !$this->isSubUser() && ($this->getMaxCallCount() > 5000 || Utils::isClickskeksCloud());
	}

	/**
	 * Ob der Banner bei neuen Domains nur aktiv sein soll, wenn die Domain auch zu einem der eingetragenen Domainnamen passt
	 */
	public function isNewDomainOnlyActiveOnMatchingDomain(): bool
	{
		return (bool)$this->get('newDomainOnlyActiveOnMatchingDomain', false);
	}

	/**
	 * @param bool $state
	 * @return self
	 */
	public function setNewDomainOnlyActiveOnMatchingDomain(bool $state): self
	{
		$this->set('newDomainOnlyActiveOnMatchingDomain', $state);
		return $this;
	}

	/**
	 * Ob importierte Einbindungen automatisch aktualisiert werden dürfen.
	 */
	public function isAllowAutoUpdateOfEmbeddings(): bool
	{
		return (bool)$this->get('allowAutoUpdateOfEmbeddings', true);
	}

	/**
	 * @param bool $state
	 * @return self
	 */
	public function setAllowAutoUpdateOfEmbeddings(bool $state): self
	{
		$this->set('allowAutoUpdateOfEmbeddings', $state);
		return $this;
	}

	/**
	 * Accountweite Erweiterung der Consent-Teilung für die Benutzersignale "Alles akzeptieren" und "Ablehnen".
	 * @return bool
	 */
	public function isSendFullConsentAndRejectionAcrossDomains(): bool
	{
		return (bool)$this->get('sendFullConsentAndRejectionAcrossDomains');
	}

	/**
	 * @param bool $state
	 * @return self
	 */
	public function setSendFullConsentAndRejectionAcrossDomains(bool $state): self
	{
		$this->set('sendFullConsentAndRejectionAcrossDomains', $state);
		return $this;
	}

	/**
	 * Liefert den Standardwert für die Zeitzone von Domains.
	 * Der Wert wird aus dem Hauptbenutzer ausgelesen.
	 * @return DateTimeZone
	 */
	public function getDefaultTimeZone(): DateTimeZone
	{
		try {
			return new DateTimeZone((string)$this->getAccountOwner()->get('defaultTimeZone'));
		} catch (Exception $exception) {
			return new DateTimeZone('Europe/Berlin');
		}
	}

	/**
	 * @param DateTimeZone $timeZone
	 * @return $this
	 */
	public function setDefaultTimeZone(DateTimeZone $timeZone): self
	{
		$this->set('defaultTimeZone', $timeZone->getName());
		return $this;
	}

    /**
     * @param bool $state
     * @return self
     */
    public function setActivateCrossDomainThemes(bool $state): self
    {
        $this->set('activateCrossDomainThemes', $state);
        return $this;
    }
    /**
     * Ob Cross DOmain Themes aktiviert sind und Menüpunkt angezeigt werden soll
     */
    public function getActivateCrossDomainThemes(): bool
    {
        return (bool)$this->get('activateCrossDomainThemes', $this->getFalseIfUserNew());
    }

    /**
     * @param bool $state
     * @return $this
     */
    public function setActivateCrossDomainLocales(bool $state): self
    {
        $this->set('activateCrossDomainLocales', $state);
        return $this;
    }

    /**
     * Ob Cross Domain Locales aktiviert sind und Menüpunkt angezeigt werden soll
     */
    public function getActivateCrossDomainLocales(): bool
    {
        return (bool)$this->get('activateCrossDomainLocales', $this->getFalseIfUserNew());
    }

    /**
     * @param bool $state
     * @return $this
     */
    public function setActivateCrossDomainGeneralAPI(bool $state): self
    {
        $this->set('activateCrossDomainGeneralAPI', $state);
        return $this;
    }

    /**
     * Ob Cross Domain Locales aktiviert sind und Menüpunkt angezeigt werden soll
     */
    public function getActivateCrossDomainGeneralAPI(): bool
    {
        return (bool)$this->get('activateCrossDomainGeneralAPI', $this->getFalseIfUserNew());
    }

    /**
     * @param bool $state
     * @return $this
     */
    public function setActivateCrossDomainUserGroup(bool $state): self
    {
        $this->set('activateCrossDomainUserGroup', $state);
        return $this;
    }

    /**
     * Ob Cross Domain Locales aktiviert sind und Menüpunkt angezeigt werden soll
     */
    public function getActivateCrossDomainUserGroup(): bool
    {
        return (bool)$this->get('activateCrossDomainUserGroup', $this->getFalseIfUserNew());
    }

    /**
     * @param bool $state
     * @return $this
     */
    public function setActivateCrossDomainIab(bool $state): self
    {
        $this->set('activateCrossDomainIab', $state);
        return $this;
    }

    /**
     * Ob Cross Domain Locales aktiviert sind und Menüpunkt angezeigt werden soll
     */
    public function getActivateCrossDomainIab(): bool
    {
        return (bool)$this->get('activateCrossDomainIab', $this->getFalseIfUserNew());
    }

    /**
     * @param bool $state
     * @return $this
     */
    public function setActivateCrossDomainAccessibility(bool $state): self
    {
        $this->set('activateCrossDomainAccessibility', $state);
        return $this;
    }

    /**
     * Ob Cross Domain Locales aktiviert sind und Menüpunkt angezeigt werden soll
     */
    public function getActivateCrossDomainAccessibility(): bool
    {
        return (bool)$this->get('activateCrossDomainAccessibility', $this->getFalseIfUserNew());
    }

    /**
     * @param bool $state
     * @return $this
     */
    public function setCrawlAllDomains(bool $state): self
    {
        $this->set('crawlAllDomains', $state);
        return $this;
    }

    /**
     * Sollen alle Domains auf regulärer Basis gecrawlt werden?
     */
    public function isCrawlAllDomains(): bool
    {
        return (bool)$this->get('crawlAllDomains', false);
    }

	/**
	 * Sollen neue Einbindungen / Cookies, die vom externen Crawler gefunden wurden, direkt aktiviert werden??
	 */
	public function isAllowAutoActivationOfNewEmbeddings(): bool
	{
		return (bool)$this->get('allowAutoActivationOfNewEmbeddings', false);
	}

	/**
	 * Sollen neue Einbindungen / Cookies, die vom externen Crawler gefunden wurden, direkt aktiviert werden??
	 * @param bool $state
	 * @return $this
	 */
	public function setAllowAutoActivationOfNewEmbeddings(bool $state): self
	{
		$this->set('allowAutoActivationOfNewEmbeddings', $state);
		return $this;
	}

    /**
     * Gibt True zurück, wenn der User vor einem bestimmten Stichtag erzeugt wurde
     * Stichtag ist der 22.02.2023
     * @return bool
     */
    public function getTrueIfUserNew(): bool
    {
        return $this->getCreateTimeStamp() >= 1677074454;
    }

    /**
     *  Gibt False zurück, wenn der User vor einem bestimmten Stichtag erzeugt wurde
     * Stichtag ist der 22.02.2023
     * @return bool
     */
    public function getFalseIfUserNew(): bool
    {
        return $this->getCreateTimeStamp() < 1677074454;
    }

	/**
	 * Liste der Locales für User-Embeddings
	 *
	 * @return UserLocale[]
	 */
	public function getUserEmbeddingLocales(): array
	{
		$result = $this->get('userEmbeddingLocales', ['de_DE', 'en_US']);

		return array_filter(array_combine($result, array_map(function ($code) {
			return UserLocale::find($code);
		}, $result)));
	}

	/**
	 * @param UserLocale[] $locales
	 * @return $this
	 */
	public function setUserEmbeddingLocales(array $locales): self
	{
		$this->set('userEmbeddingLocales', array_map(function (UserLocale $l) {
			return $l->getName();
		}, $locales));
		return $this;
	}

    /**
     * @return array<string,int>
     */
    public function getTrialPeriods(): array
    {
        return (array)$this->get('trialPeriods', []);
    }

    /**
     * @param array<string,int> $trialPeriods
     * @return $this
     */
    public function setTrialPeriods(array $trialPeriods): self
    {
        $this->set('trialPeriods', $trialPeriods);
        return $this;
    }
}
