<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Exception\ConfigNotWritableException;
use DateTimeInterface;
use DateTime;

/// Zustand der userspezifischen Cronjobs
class UserCronState extends Model
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-user-cronstate';

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = true;

	/** @var bool $userSpecificRepository */
	protected static $userSpecificRepository = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Diese Methode erstellt ein Objekt dieser Klasse
	 * @param string|null $namespace Cronjob-Namensraum (lcfirst($classname))
	 * @return static
	 * @throws \LogicException
	 */
	public static function create(?string $namespace=null): Model
	{
		if ($namespace === null) {
			throw new \LogicException('cannot use create() with UserCronState model');
		}
		self::beginWrite();
		return new static(static::getRepository()->escapeKey($namespace));
	}

	/**
	 * Diese Methode löscht einen Datensatz aus der Konfiguration.
	 * @param string $namespace
	 * @param bool $autocommit
	 * @throws ConfigNotWritableException
	 */
	public static function delete(string $namespace, bool $autocommit=true): void
	{
		parent::delete(static::getRepository()->escapeKey($namespace), $autocommit);
	}

	/**
	 * @param string $namespace
	 * @return bool
	 */
	public static function exists(string $namespace): bool
	{
		return parent::exists(static::getRepository()->escapeKey($namespace));
	}

	/**
	 * @param string $namespace
	 * @return static|null
	 */
	public static function find(string $namespace): ?Model
	{
		return self::exists($namespace) ? new static(static::getRepository()->escapeKey($namespace)) : null;
	}

	/**
	 * @param string $namespace
	 * @return static
	 */
	public static function findOrCreate(string $namespace): self
	{
		return self::exists($namespace) ? new static(static::getRepository()->escapeKey($namespace)) : self::create($namespace);
	}

	/**
	 * @return DateTime|null
	 */
	public function getLastExecutionTime(): ?DateTime
	{
		$ts = $this->get('lastExecutionTimestamp');
		if ($ts) {
			$dt = new DateTime();
			$dt->setTimestamp($ts);
			return $dt;
		} else {
			return null;
		}
	}

	/**
	 * @param DateTimeInterface $ts
	 * @return $this
	 */
	public function setLastExecutionTime(DateTimeInterface $ts): self
	{
		$this->set('lastExecutionTimestamp', (int)$ts->getTimestamp());
		return $this;
	}

	/**
	 * @param string $path '.'-separierter Config-Name
	 * @return mixed
	 */
	public function getExtra(string $path)
	{
		return $this->get($path);
	}

	/**
	 * @param string $path '.'-separierter Config-Name
	 * @param mixed $value
	 * @return $this
	 */
	public function setExtra(string $path, $value): self
	{
		$this->set($path, $value);
		return $this;
	}

}
