<?php

namespace App\Model;

use App\Component\Tcf\v2\Model\Purpose as TcfPurpose;
use App\Component\Tcf\v2\Model\SpecialFeature as TcfSpecialFeature;
use App\Component\Tcf\v2\Model\Feature as TcfFeature;
use App\Component\Tcf\v2\Model\SpecialPurpose as TcfSpecialPurpose;
use App\Component\Tcf\v2\Model\Translation\Feature as TranslatedTcfFeature;
use App\Component\Tcf\v2\Model\Translation\Purpose as TranslatedTcfPurpose;
use App\Component\Tcf\v2\Model\Translation\SpecialFeature as TranslatedTcfSpecialFeature;
use App\Component\Tcf\v2\Model\Translation\SpecialPurpose as TranslatedTcfSpecialPurpose;
use App\Component\Config\RepositoryInterface;

/**
 * Benutzerweite Embedding-Vorlagen
 */
class UserEmbedding extends EmbeddingBase
{
	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-user-embeddings';

	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var bool $userSpecificRepository */
	protected static $userSpecificRepository = true;

	/** @var bool $accountSingleton */
	protected static $accountSingleton = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Diese Methode sucht nach Embedding-Vorlagen und gibt ein Ergebnis-Array zurück.
	 * @param string $search Suchbegriff
	 * @return UserEmbedding[]
	 */
	public static function search(string $search): array
	{
		$languages = User::loggedInUser()->getUserEmbeddingLocales();

		return array_filter(self::all(), function (UserEmbedding $e) use ($search, $languages) {
			// Grundfelder des Embeddings durchgehen
			if (stripos($e->getName(), $search) !== false or
				stripos(implode("\n", $e->getCodeBlock()), $search) !== false or
				stripos($e->getDescription(), $search) !== false or
				stripos($e->getEmbeddingVendor(), $search) !== false or
				stripos($e->getCollectedDataInfo(), $search) !== false or
				stripos($e->getPurposeOfDataCollection(), $search) !== false or
				stripos($e->getLegalBasis(), $search) !== false or
				stripos($e->getPlaceOfProcessing(), $search) !== false or
				stripos($e->getPrivacyPolicyUrl(), $search) !== false
			) {
				return true;
			}

			// Assets durchgehen
			foreach ($e->getAssets() as $a) {
				if (stripos($a->getName(), $search) !== false or
					stripos($a->getDescription(), $search) !== false or
					stripos($a->getLifetime(), $search) !== false
				) {
					return true;
				}
			}

			// Übersetzungen durchgehen
			foreach ($languages as $lang) {
				if (stripos($e->getTranslatedName($lang, false), $search) !== false or
					stripos($e->getTranslatedDescription($lang, false), $search) !== false or
					stripos($e->getTranslatedCollectedDataInfo($lang, false), $search) !== false or
					stripos($e->getTranslatedLegalBasis($lang, false), $search) !== false or
					stripos($e->getTranslatedPlaceOfProcessing($lang, false), $search) !== false or
					stripos($e->getTranslatedPrivacyPolicyUrl($lang, false), $search) !== false or
					stripos($e->getTranslatedPurposeOfDataCollection($lang, false), $search) !== false
				) {
					return true;
				}

				// Übersetzungen der Assets durchgehen
				foreach ($e->getAssets() as $a) {
					if (stripos($a->getTranslatedDescription($lang), $search) !== false or
						stripos($a->getTranslatedLifetime($lang), $search) !== false
					) {
						return true;
					}
				}

				return false;
			}
		});
	}

	/**
	 * @return UserEmbeddingAsset[]
	 */
	public function getAssets(): array
	{
		if (empty($this->assets)) {
			$this->assets = array_values(array_filter(array_map(function ($id) {
				return UserEmbeddingAsset::find($id);
			}, $this->get('assets', []))));
		}

		return $this->assets;
	}

	/**
	 * @param string $assetId
	 * @return static
	 */
	public function removeAsset(string $assetId): EmbeddingBase
	{
		parent::removeAsset($assetId);
		UserEmbeddingAsset::delete($assetId, false);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedName(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_name.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_name.'.$locale->getId());
		}
		if ($useFallback and ($result === null or $result === '')) {
			return $this->getName();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $label
	 * @return $this
	 */
	public function setTranslatedName(UserLocale $locale, string $label): self
	{
		$key = 'translated_name.'.$locale->getName();
		if ($label !== $this->get($key) and $label !== $this->getName()) {
			$this->makingLegallyRelevantChange();
		}

		$this->set($key, $label);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedDescription(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_description.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_description.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getDescription();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $description
	 * @return $this
	 */
	public function setTranslatedDescription(UserLocale $locale, string $description): self
	{
		$key = 'translated_description.'.$locale->getName();
		if ($description !== $this->get($key) and $description !== $this->getDescription()) {
			$this->makingLegallyRelevantChange();
		}

		$this->set($key, $description);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPrivacyPolicyUrl(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_privacy_policy_url.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_privacy_policy_url.'.$locale->getId());
		}
		if ($useFallback and ($result === null or $result === '')) {
			return $this->getPrivacyPolicyUrl();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $url
	 * @return $this
	 */
	public function setTranslatedPrivacyPolicyUrl(UserLocale $locale, string $url): self
	{
		$key = 'translated_privacy_policy_url.'.$locale->getName();
		$url = trim($url);
		if ($url !== $this->get($key) and $url !== $this->getPrivacyPolicyUrl()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $url);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @return TcfPurpose[]|TranslatedTcfPurpose[]
	 */
	public function getTranslatedTcfPurposes(UserLocale $locale): array
	{
		$langCode = preg_replace('~[-_].+~', '', $locale->getName());
		return array_map(function (TcfPurpose $purpose) use ($langCode) {
			return $purpose->getTranslation($langCode);
		}, $this->getTcfPurposes());
	}

	/**
	 * @param UserLocale $locale
	 * @return TcfSpecialPurpose[]|TranslatedTcfSpecialPurpose[]
	 */
	public function getTranslatedTcfSpecialPurposes(UserLocale $locale): array
	{
		$langCode = preg_replace('~[-_].+~', '', $locale->getName());
		return array_map(function (TcfSpecialPurpose $purpose) use ($langCode) {
			return $purpose->getTranslation($langCode);
		}, $this->getTcfSpecialPurposes());
	}

	/**
	 * @param UserLocale $locale
	 * @return TcfFeature[]|TranslatedTcfFeature[]
	 */
	public function getTranslatedTcfFeatures(UserLocale $locale): array
	{
		$langCode = preg_replace('~[-_].+~', '', $locale->getName());
		return array_map(function (TcfFeature $feature) use ($langCode) {
			return $feature->getTranslation($langCode);
		}, $this->getTcfFeatures());
	}

	/**
	 * @param UserLocale $locale
	 * @return TcfSpecialFeature[]|TranslatedTcfSpecialFeature[]
	 */
	public function getTranslatedTcfSpecialFeatures(UserLocale $locale): array
	{
		$langCode = preg_replace('~[-_].+~', '', $locale->getName());
		return array_map(function (TcfSpecialFeature $feature) use ($langCode) {
			return $feature->getTranslation($langCode);
		}, $this->getTcfSpecialFeatures());
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 * @deprecated
	 */
	public function getTranslatedLifetime(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_lifetime.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_lifetime.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getLifetime();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $lifetime
	 * @return $this
	 * @deprecated
	 */
	public function setTranslatedLifetime(UserLocale $locale, string $lifetime): self
	{
		$this->set('translated_lifetime.'.$locale->getName(), $lifetime);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedCollectedDataInfo(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_collected_data_info.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_collected_data_info.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getCollectedDataInfo();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $info
	 * @return $this
	 */
	public function setTranslatedCollectedDataInfo(UserLocale $locale, string $info): self
	{
		$key = 'translated_collected_data_info.'.$locale->getName();
		if ($info !== $this->get($key) and $info !== $this->getCollectedDataInfo()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $info);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPurposeOfDataCollection(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_purpose_of_data_collection.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_purpose_of_data_collection.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getPurposeOfDataCollection();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $purpose
	 * @return $this
	 */
	public function setTranslatedPurposeOfDataCollection(UserLocale $locale, string $purpose): self
	{
		$key = 'translated_purpose_of_data_collection.'.$locale->getName();
		if ($purpose !== $this->get($key) and $purpose !== $this->getPurposeOfDataCollection()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $purpose);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedLegalBasis(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_legal_basis.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_legal_basis.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getLegalBasis();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $basis
	 * @return $this
	 */
	public function setTranslatedLegalBasis(UserLocale $locale, string $basis): self
	{
		$key = 'translated_legal_basis.'.$locale->getName();
		if ($basis !== $this->get($key) and $basis !== $this->getLegalBasis()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $basis);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedLegalBasisSupplement(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_legal_basis_supplement.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_legal_basis_supplement.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getLegalBasisSupplement();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $basis
	 * @return $this
	 */
	public function setTranslatedLegalBasisSupplement(UserLocale $locale, string $basis): self
	{
		$key = 'translated_legal_basis_supplement.'.$locale->getName();
		if ($basis !== $this->get($key) and $basis !== $this->getLegalBasisSupplement()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set($key, $basis);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedPlaceOfProcessing(UserLocale $locale, $useFallback=true): string
	{
		$result = $this->get('translated_place_of_processing.'.$locale->getName());
		if ($result === null) {
			$result = $this->get('translated_place_of_processing.'.$locale->getId());
		}
		if ($useFallback and $result === null) {
			return $this->getPlaceOfProcessing();
		}
		return (string)$result;
	}

	/**
	 * @param UserLocale $locale
	 * @param string $place
	 * @return $this
	 */
	public function setTranslatedPlaceOfProcessing(UserLocale $locale, string $place): self
	{
		$key = 'translated_place_of_processing.'.$locale->getName();
		if ($place !== $this->get($key) and $place !== $this->getPlaceOfProcessing()) {
			$this->makingLegallyRelevantChange();
		}
		$this->set('translated_place_of_processing.'.$locale->getName(), $place);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentTitle(UserLocale $locale, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_title.{$locale->getName()}");
		return strlen($value) > 0 ? $value : ($useFallback ? $locale->getBlockedContentTitle() : '');
	}

	/**
	 * @param UserLocale $locale
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentTitle(UserLocale $locale, string $value): self
	{
		$value = trim($value);
		$currentValue = $this->getTranslatedBlockedContentTitle($locale, false);
		$defaultValue = $locale->getBlockedContentTitle();
		if ($value != $currentValue && ($currentValue != '' || $value != $defaultValue)) {
			$this->makingLegallyRelevantChange();
		}
		$this->set("translated_blocked_content_title.{$locale->getName()}", $value);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentText(UserLocale $locale, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_text.{$locale->getName()}");
		return strlen($value) > 0 ? $value : ($useFallback ? $locale->getBlockedContentText() : '');
	}

	/**
	 * @param UserLocale $locale
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentText(UserLocale $locale, string $value): self
	{
		$value = trim($value);
		$currentValue = $this->getTranslatedBlockedContentText($locale, false);
		$defaultValue = $locale->getBlockedContentText();
		if ($value != $currentValue && ($currentValue != '' || $value != $defaultValue)) {
			$this->makingLegallyRelevantChange();
		}
		$this->set("translated_blocked_content_text.{$locale->getName()}", $value);
		return $this;
	}

	/**
	 * @param UserLocale $locale
	 * @param bool $useFallback
	 * @return string
	 */
	public function getTranslatedBlockedContentButtonText(UserLocale $locale, bool $useFallback=true): string
	{
		$value = (string)$this->get("translated_blocked_content_button_text.{$locale->getName()}");
		return strlen($value) > 0 ? $value : ($useFallback ? $locale->getBlockedContentButtonText() : '');
	}

	/**
	 * @param UserLocale $locale
	 * @param string $value
	 * @return $this
	 */
	public function setTranslatedBlockedContentButtonText(UserLocale $locale, string $value): self
	{
		$value = trim($value);
		$currentValue = $this->getTranslatedBlockedContentButtonText($locale, false);
		$defaultValue = $locale->getBlockedContentButtonText();
		if ($value != $currentValue && ($currentValue != '' || $value != $defaultValue)) {
			$this->makingLegallyRelevantChange();
		}
		$this->set("translated_blocked_content_button_text.{$locale->getName()}", $value);
		return $this;
	}

	/**
	 * Setzt alle übersetzten Texte, die in $locale noch nie vorher gesetzt
	 * wurden, auf den nicht-übersetzten Wert.
	 *
	 * @param UserLocale $locale Zu füllendes Locale
	 * @return $this
	 */
	public function fillTranslationsFromDefault(UserLocale $locale): self
	{
		$this->setTranslatedCollectedDataInfo($locale, $this->getTranslatedCollectedDataInfo($locale));
		$this->setTranslatedDescription($locale, $this->getTranslatedDescription($locale));
		$this->setTranslatedLegalBasis($locale, $this->getTranslatedLegalBasis($locale));
		$this->setTranslatedLifetime($locale, $this->getTranslatedLifetime($locale));
		$this->setTranslatedPlaceOfProcessing($locale, $this->getTranslatedPlaceOfProcessing($locale));
		$this->setTranslatedPrivacyPolicyUrl($locale, $this->getTranslatedPrivacyPolicyUrl($locale));
		$this->setTranslatedPurposeOfDataCollection($locale, $this->getTranslatedPurposeOfDataCollection($locale));
		$this->setTranslatedBlockedContentTitle($locale, $this->getTranslatedBlockedContentTitle($locale));
		$this->setTranslatedBlockedContentText($locale, $this->getTranslatedBlockedContentText($locale));
		$this->setTranslatedBlockedContentButtonText($locale, $this->getTranslatedBlockedContentButtonText($locale));
		$this->setTranslatedLegalBasisSupplement($locale, $this->getTranslatedLegalBasisSupplement($locale));
		return $this;
	}

}
