<?php

namespace App\Model;

use App\Component\LocaleOverrideHelper;
use App\Exception\LocaleAlreadyExistsException;
use App\Component\Config\RepositoryInterface;
use LogicException;

class UserLocale extends LocaleBase
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var string $repositoryFilename */
	protected static $repositoryFilename = 'cm-user-locales';

	/** @var bool $repositoryWithCustomIds */
	protected static $repositoryWithCustomIds = true;

	/** @var bool $userSpecificRepository */
	protected static $userSpecificRepository = true;

	/** @var bool $accountSingleton */
	protected static $accountSingleton = true;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	public static function findByLanguage(Language $language): ?self
	{
		return self::find($language->getId());
	}

	/**
	 * @override
	 * @return string
	 */
	public function getName(): string
	{
		return $this->getId();
	}

	/**
	 * @override
	 * @return Language|null
	 */
	public function getLanguage(): ?Language
	{
		$id = $this->getId();
		if (strpos($id, '_') !== false) {
			return Language::find($id) ?? Language::find(explode('_', $id, 2)[0]);
		} else {
			return Language::find($id);
		}
	}

	/**
	 * @param string|null $localeName
	 * @return static
	 * @throws LocaleAlreadyExistsException
	 * @throws LogicException
	 * @override
	 * @phan-assert string $localeName
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function create(?string $localeName=null): Model
	{
		$language = Language::find($localeName ?? '');

		if ($localeName === null) {
			throw new LogicException('Locale name may not be null.');
		}
		else if (!$language) {
			throw new LogicException('Invalid locale name, language not found.');
		}

		self::beginWrite();

		if (self::findByLanguage($language)) {
			throw new LocaleAlreadyExistsException($language->getCode());
		}

		/** @var self $locale */
		$locale = new static($language->getId());
		assert($locale instanceof UserLocale);
		return $locale;
	}

	/**
	 * @param string|null $localeName
	 * @return static
	 * @throws LogicException
	 * @override
	 * @phan-suppress PhanParamSignatureMismatch
	 */
	public static function createNonPersistent(?string $localeName=null): Model
	{
		if ($localeName === null) {
			throw new LogicException('Locale name may not be null.');
		}

		self::getRepository()->makeVolatile();
		/** @var self $locale */
		$locale = new static($localeName);
		assert($locale instanceof UserLocale);
		return $locale;

	}

	protected static final function onRepositoryLoaded(): void
	{
		if (self::getRepository()->isTemplateLoaded()) {
			LocaleOverrideHelper::applyOverridesToUserLocales(LocaleOverride::all());
		}
	}

	/**
	 * Verschiebt die Daten des Models zu einer neuen ID.
	 *
	 * Es wird ein exklusiver Lock auf dem Repository erzeugt.
	 *
	 * @param string $code
	 * @return $this
	 */
	public function rename($code): self
	{
		if ($code === $this->getId()) {
			return $this;
		}

		self::beginWrite();
		$source = self::find($this->getId());
		$this->copyAndSwitch($code);
		self::delete($source->getId());
		return $this;
	}
}
