<?php

namespace App\Model;

use App\Component\Config\RepositoryInterface;
use App\Component\Media\MediaFileService;
use Symfony\Component\HttpFoundation\File\File;

class UserTheme extends ThemeBase
{
	/** @var RepositoryInterface|null $repository */
	protected static $repository = null;

	/** @var bool $userSpecificRepository */
	protected static $userSpecificRepository = true;

	/** @var bool $accountSingleton */
	protected static $accountSingleton = true;

	/** @var bool $useRepositorySkeleton */
	protected static $useRepositorySkeleton = false;

	/**
	 * @param string $id
	 */
	protected function __construct(string $id)
	{
		parent::__construct($id);
	}

	/**
	 * Diese Methode iteriert über alle User-Themes für jede Domain des aktiven Benutzers.
	 * Aktualisierungen im Theme-Repo werden automatisch geschrieben; in Callback nicht nötig.
	 * @param callable $callback function (Theme $deployedTheme, UserTheme $userTheme)
	 * @param callable|null $onSuccess
	 * @param callable|null $onError The Exception object is provided in the first parameter of the callback function.
	 */
	public static function forEachDeployedTheme(callable $callback, ?callable $onSuccess=null, ?callable $onError=null): void
	{
		$userThemes = UserTheme::all();
		$user = User::loggedInUser();

		if (!$user) {
			return;
		}

		try {
			foreach (Domain::all() as $domain) {
				// Themes pro Domain durchgehen und Repos speichern, um Themes zu aktualisieren
				Theme::switchRepository($user, $domain, true);

				foreach ($userThemes as $userTheme) {
					// Finde das Theme, das das aktuelle User-Theme referenziert; erstelle notfalls ein neues
					$theme = Theme::byUserTheme($userTheme) ?? Theme::create()
						->setUserTheme($userTheme)
						->setSettingsIconBackgroundColor($userTheme->getPublishedSettingsIconBackgroundColor() ?? $userTheme->getSettingsIconBackgroundColor())
						->setSettingsIconForegroundColor($userTheme->getPublishedSettingsIconForegroundColor() ?? $userTheme->getSettingsIconForegroundColor())
					;

					$callback($theme, $userTheme);
				}
			}

			if ($onSuccess) {
				$onSuccess();
			}
		} catch (\Exception $exception) {
			if ($onError) {
				$onError($exception);
			}
		} finally {
			// Theme-Repo speichern und schließen
			Theme::closeRepository(true);
		}
	}

	/**
	 * @return File|null
	 */
	public function getSettingsIcon(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getUserRepository('settings_icons');
		$identifier = $this->getSettingsIconIdentifier();
		return ($identifier == '') ? null : $repo->getFile($identifier);
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setSettingsIcon(?File $file): ThemeBase
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getUserRepository('settings_icons');
		$prevIdentifier = $this->getSettingsIconIdentifier();

		$identifier = $repo->replaceFile($prevIdentifier, $file, $this->getId());
		$this->setSettingsIconIdentifier($identifier);

		return $this;
	}

	/**
	 * @return string
	 */
	public function getPublishedSettingsIconIdentifier(): string
	{
		return (string)$this->get('publishedSettingsIconId', '');
	}

	/**
	 * @param string $mediaId
	 * @return $this
	 */
	protected function setPublishedSettingsIconIdentifier(string $mediaId): self
	{
		$this->set('publishedSettingsIconId', $mediaId);
		return $this;
	}

	/**
	 * @return File|null
	 */
	public function getPublishedSettingsIcon(): ?File
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getUserRepository('settings_icons');
		$identifier = $this->getPublishedSettingsIconIdentifier();
		return ($identifier == '') ? null : $repo->getFile($identifier);
	}

	/**
	 * @param File|null $file
	 * @return $this
	 */
	public function setPublishedSettingsIcon(?File $file): self
	{
		$service = MediaFileService::getInstance();
		$repo = $service->getUserRepository('settings_icons');
		$prevIdentifier = $this->getPublishedSettingsIconIdentifier();

		$identifier = $repo->replaceFile($prevIdentifier, $file, $this->getId().'-published');
		$this->setPublishedSettingsIconIdentifier($identifier);

		return $this;
	}

	/**
	 * @return string|null
	 */
	public function getPublishedSettingsIconBackgroundColor(): ?string
	{
		return $this->get('publishedSettingsIconBackgroundColor');
	}

	/**
	 * @param string $color
	 * @return $this
	 */
	public function setPublishedSettingsIconBackgroundColor(string $color): self
	{
		$this->set('publishedSettingsIconBackgroundColor', $color);
		return $this;
	}

	/**
	 * @return string|null
	 */
	public function getPublishedSettingsIconForegroundColor(): ?string
	{
		return $this->get('publishedSettingsIconForegroundColor');
	}

	/**
	 * @param string $color
	 * @return $this
	 */
	public function setPublishedSettingsIconForegroundColor(string $color): self
	{
		$this->set('publishedSettingsIconForegroundColor', $color);
		return $this;
	}
}
