<?php

namespace App\Twig;

use App\Component\Color;
use App\Model\User;
use App\Model\Placeholder;
use App\Utils;
use Collator;
use NumberFormatter;
use Symfony\Component\HttpFoundation\File\File;
use Twig\Environment;
use Twig\Extension\AbstractExtension;
use Twig\Markup;
use Twig\TwigFunction;
use Twig\TwigFilter;
use UnexpectedValueException;

class AppExtension extends AbstractExtension
{
	/** @var Collator[] $collators */
	private $collators = [];

	/** @var NumberFormatter[] $numberFormatters */
	private $numberFormatters = [];

	/**
	 * @return TwigFunction[]
	 */
	public function getFunctions(): array
	{
		return [
			new TwigFunction('in_route', [$this, 'inRoute'], ['needs_environment' => true]),
			new TwigFunction('may_access', [$this, 'mayAccess'], ['needs_environment' => true]),
			new TwigFunction('may_use_feature', [$this, 'mayUseFeature'], ['needs_environment' => true]),
			new TwigFunction('hover', [$this, 'hoverColor']),
			new TwigFunction('alpha', [$this, 'alpha']),
			new TwigFunction('darken', [$this, 'darkenColor']),
			new TwigFunction('lighten', [$this, 'lightenColor']),
			new TwigFunction('tint', [$this, 'tintColor']),
			new TwigFunction('selectColor', [$this, 'selectColor']),
			new TwigFunction('contrastColor', [$this, 'contrastColor']),
			new TwigFunction('checked', [$this, 'checked'], ['is_safe' => ['html']]),
			new TwigFunction('selected', [$this, 'selected'], ['is_safe' => ['html']]),
			new TwigFunction('required', [$this, 'required'], ['is_safe' => ['html']]),
			new TwigFunction('disabled', [$this, 'disabled'], ['is_safe' => ['html']]),
			new TwigFunction('readOnly', [$this, 'readOnly'], ['is_safe' => ['html']]),
			new TwigFunction('classIf', [$this, 'classIf'], ['is_safe' => ['html']]),
			new TwigFunction('preserve_links', [$this, 'preserveLinks'], ['is_safe' => ['html']]),
			new TwigFunction('wrap_sentences', [$this, 'wrapSentences'], ['is_safe' => ['html']]),
			new TwigFunction('wrap_words', [$this, 'wrapWords'], ['is_safe' => ['html']]),
			new TwigFunction('plugin_asset', [$this, 'pluginAsset'], ['needs_environment' => true]),
			new TwigFunction('rescue_token', [$this, 'rescueToken']),
		];
	}

	/**
	 * @return TwigFilter[]
	 */
	public function getFilters(): array
	{
		return [
			new TwigFilter('hover', [$this, 'hoverColor']),
			new TwigFilter('alpha', [$this, 'alpha']),
			new TwigFilter('darken', [$this, 'darkenColor']),
			new TwigFilter('lighten', [$this, 'lightenColor']),
			new TwigFilter('tint', [$this, 'tintColor']),
			new TwigFilter('trunc_url', [$this, 'truncUrl']),
			new TwigFilter('compare', [$this, 'compare']),
			new TwigFilter('format_bytes', [$this, 'formatBytes']),
			new TwigFilter('format_number', [$this, 'formatNumber'], ['needs_environment' => true]),
			new TwigFilter('replace_placeholders', [$this, 'replacePlaceholders'], ['pre_escape' => 'html', 'is_safe' => ['html']]),
			new TwigFilter('checked', [$this, 'checked'], ['is_safe' => ['html']]),
			new TwigFilter('selected', [$this, 'selected'], ['is_safe' => ['html']]),
			new TwigFilter('required', [$this, 'required'], ['is_safe' => ['html']]),
			new TwigFilter('disabled', [$this, 'disabled'], ['is_safe' => ['html']]),
			new TwigFilter('readOnly', [$this, 'readOnly'], ['is_safe' => ['html']]),
			new TwigFilter('classIf', [$this, 'classIf'], ['is_safe' => ['html']]),
			new TwigFilter('preserve_links', [$this, 'preserveLinks'], ['is_safe' => ['html']]),
			new TwigFilter('wrap_sentences', [$this, 'wrapSentences'], ['is_safe' => ['html']]),
			new TwigFilter('wrap_words', [$this, 'wrapWords'], ['is_safe' => ['html']]),
			new TwigFilter('base64_data_url', [$this, 'base64DataUrl']),
			new TwigFilter('image_dimensions', [$this, 'imageDimensions'], ['is_safe' => ['html']]),
			new TwigFilter('interval_lang', [$this, 'humanReadableInterval']),
			new TwigFilter('intl_sort', [$this, 'intlSort'], ['needs_environment' => true]),
		];
	}

	/**
	 * Prüft, ob eine Route aktuell aktiv ist
	 *
	 * Usage: `{% if in_route("foobar_*") %}`
	 *
	 * @param Environment $env
	 * @param string $name Name bzw. Pattern mit `*`/`?`-Platzhaltern
	 * @return bool
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function inRoute(Environment $env, $name)
	{
		$vars = $env->getGlobals();
		$app = $vars['app'];
		$route = $app->getRequest()->attributes->get('_route');

		return fnmatch($name, $route);
	}

	/**
	 * Prüft, ob ein User auf eine Route zugreifen darf
	 *
	 * Usage: `{% if may_access("foobar_index") %}`, `{% if may_access("foobar_index", "POST") %}`
	 *
	 * @param Environment $env
	 * @param string $route
	 * @param string $method
	 * @param User|null $user
	 * @return bool
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function mayAccess(Environment $env, $route, $method='GET', ?User $user=null)
	{
		if ($user === null) {
			$user = $env->getGlobals()['user'];
		}
		return Utils::mayUserAccessRoute($user, (string)$route, (string)$method);
	}

	/**
	 * Prüft, ob ein User ein Feature nutzen darf
	 *
	 * Usage: `{% if may_use_feature("foobar") %}`
	 *
	 * @param Environment $env
	 * @param string $feature
	 * @param User|null $user
	 * @return bool
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function mayUseFeature(Environment $env, string $feature, ?User $user=null): bool
	{
		if ($user === null) {
			$user = $env->getGlobals()['user'];
		}
		return $user->hasAccountFeature($feature);
	}

	/**
	 * Ermittelt das Größenverhältnis von `$a` zu `$b`.
	 *
	 * Usage: `{% for domain in domains|sort((a,b) => a.name|compare(b.name)) %}`
	 *
	 * @param int|float|string $a
	 * @param int|float|string $b
	 * @return int Returns a value similar to the spaceship operator.
	 */
	public function compare($a, $b): int
	{
		if (is_string($a) && is_string($b)) {
			return strnatcasecmp($a, $b);
		}
		else {
			return $a <=> $b;
		}
	}

	/**
	 * Formatiert Bytes in GiB/MiB/KiB/B
	 *
	 * Usage: `{{ variable|format_bytes }}`
	 *
	 * @param int|string $bytes
	 * @return string
	 */
	public function formatBytes($bytes)
	{
		if ($bytes > 900 * 1024 * 1024) {
			return number_format($bytes / 1024 / 1024 / 1024, 1, ',', '&thinsp;') .' GiB';
		}
		if ($bytes > 900 * 1024) {
			return number_format($bytes / 1024 / 1024, 1, ',', '&thinsp;') .' MiB';
		}
		if ($bytes > 900) {
			return number_format($bytes / 1024, 1, ',', '&thinsp;') .' KiB';
		}
		return number_format($bytes, 0, ',', '&thinsp;') .' B';
	}

	/**
	 * Formatiert eine Zahl (unter Berücksichtigung eines Locales)
	 *
	 * Usage: `{{ numbervar|format_number }}`, `{{ numbervar|format_number(2) }}`,
	 *        `{{ numbervar|format_number(3, "de_DE") }}`
	 *
	 * Abweichende Implementierung, die anstelle des `format_number`-Filters der `twig/intl-extra`-Erweiterung dient.
	 * @param Environment $env
	 * @param float|int $number
	 * @param int $decimals
	 * @param string|null $locale Nach welcher Locale-Regel formatiert werden soll (NULL für aktuelles Backend-Locale)
	 * @return string
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function formatNumber(Environment $env, $number, int $decimals=0, ?string $locale=null): string
	{
		if (is_string($number) and !is_numeric($number)) {
			return $number;
		}
		if ($locale === null) {
			$vars = $env->getGlobals();
			$locale = $vars['_locale'];
		}

		if (class_exists('NumberFormatter')) {
			if (!isset($this->numberFormatters[$locale])) {
				$this->numberFormatters[$locale] = NumberFormatter::create($locale, NumberFormatter::DEFAULT_STYLE);
			}
			$formatter = $this->numberFormatters[$locale];
		} else {
			$formatter = null;
		}

		if ($formatter) {
			$result = $formatter->format(round($number, $decimals));
			if ($result === false) {
				$result = (string)$number;
			}
		} else {
			// Use the decimal comma as a fallback because it's more common, for example, in the European region.
			$result = number_format($number, $decimals, ',', '.');
		}

		return $result;
	}

	/**
	 * @param string $string
	 * @param int $maxlen
	 * @param string $delim
	 * @param string $ellipsis
	 * @return string
	 */
	private function truncMiddle($string, $maxlen, $delim, $ellipsis='…')
	{
		$result = [];
		$parts = explode($delim, $string);
		$suffix = array_pop($parts);
		$len = strlen($suffix);
		$used_parts = 0;
		
		if ($parts and $parts[0] == '') {
			$result[] = array_shift($parts) . $delim;
			$len += strlen($delim);
		}

		$part = '';
		foreach ($parts as $part) {
			if ($len+strlen($part)+strlen($delim) < $maxlen-1) {
				$result[] = $part . $delim;
				$len += strlen($part)+strlen($delim);
				$used_parts ++;
			} else {
				break;
			}
		}
		
		if ($used_parts == count($parts)) {
			return $string;
		} else {
			$finalPrefix = implode('', $result);
			$finalSuffix = $delim . $suffix;
			if (strlen($finalPrefix)+strlen($finalSuffix) + strlen($ellipsis) >= $maxlen) {
				return $finalPrefix . $ellipsis . $finalSuffix;
			}
			else {
				$restlen = $maxlen - (strlen($finalPrefix)+strlen($finalSuffix) + strlen($ellipsis));
				return $finalPrefix . substr($part, 0, $restlen) . $ellipsis . $finalSuffix;
			}
		}
	}

	/**
	 * Kürzt eine URL zur Anzeige möglichst so, dass Host- und Dateiname sichtbar bleiben
	 *
	 * Usage: `{{ urlvar|trunc_url }}`, `{{ urlvar|trunc_url(80) }}`
	 *
	 * @param string $url URL
	 * @param int $maxlen Maximale Stringlänge
	 * @return string
	 */
	public function truncUrl($url='',$maxlen=60)
	{
		$urlparts = parse_url($url);
		if ($urlparts === false) {
			return '[invalid url]';
		}
		$url = '';
		if (isset($urlparts['scheme'])) {
			$url .= $urlparts['scheme'].':';
		}
		if (isset($urlparts['host'])) {
			$url .= '//'.$urlparts['host'];
			if (isset($urlparts['port'])) {
				$url .= ':'.$urlparts['port'];
			}
		}
		if (isset($urlparts['path'])) {
			$maxpathlen = $maxlen - strlen($url) - (isset($urlparts['query']) ? 3 : 0);
			$url .= $this->truncMiddle($urlparts['path'], $maxpathlen, '/');
		}
		if (isset($urlparts['query'])) {
			$url .= '?';
			if (strlen($urlparts['query']) + strlen($url) < $maxlen) {
				$url .= $urlparts['query'];
			} else {
				$maxquerylen = $maxlen - strlen($url) - 1;
				if ($maxquerylen > 1) {
					$url .= substr($urlparts['query'], 0, $maxlen - strlen($url) - 1).'…';
				} else {
					$url .= '…';
				}
			}
		}
		return $url;
	}

	/**
	 * Dunkelt den Farbwert ab
	 *
	 * Usage: `{{ colorvar|darken(30) }}`, `{{ darken(colorvar, 30) }}`
	 *
	 * @param string $hexCode Ein Farbcode, z.B. #000000 oder #FFFFFF.
	 * @param float $amount Ein Wert zwischen 0 und 100. Keine Änderung bei 0, komplett abgedunkelt bei 100.
	 * @return string
	 */
	public function darkenColor($hexCode, float $amount): string
	{
		$amount = min(max(0, $amount), 100);

		try {
			$color = Color::fromCssString((string)$hexCode)->darken($amount / 100);
		} catch (UnexpectedValueException $e) {
			return (string)$hexCode;
		}

		return (string)$color;
	}

	/**
	 * Hellt den Farbwert auf
	 *
	 * Usage: `{{ colorvar|lighten(30) }}`, `{{ lighten(colorvar, 30) }}`
	 *
	 * @param string $hexCode Ein Farbcode, z.B. #000000 oder #FFFFFF.
	 * @param float $amount Ein Wert zwischen 0 und 100. Keine Änderung bei 0, komplett aufgehellt bei 100.
	 * @return string
	 */
	public function lightenColor($hexCode, float $amount): string
	{
		$amount = min(max(0, $amount), 100);

		try {
			$color = Color::fromCssString((string)$hexCode)->lighten($amount / 100);
		} catch (UnexpectedValueException $e) {
			return (string)$hexCode;
		}

		return (string)$color;
	}

	/**
	 * Gleicht Farbton und Sättigung an.
	 *
	 * Usage: `{{ colorA|tintColor(colorB) }}`, `{{ tintColor(colorA, colorB) }}`
	 *
	 * @param string $color Ein Farbcode, z.B. #000000 oder #FFFFFF.
	 * @param string $tint Ein Farbcode, z.B. #000000 oder #FFFFFF.
	 * @return string
	 */
	public function tintColor(string $color,  string $tint): string
	{
		try {
			$color = Color::fromCssString($color)->tint(Color::fromCssString($tint));
		} catch (UnexpectedValueException $e) {}

		return (string)$color;
	}

	/**
	 * Hellt den Farbwert auf, wenn die Helligkeit unter einen bestimmten Punkt
	 * fällt; andernfalls wird er abgedunkelt.
	 *
	 * Usage: `{{ colorvar|hoverColor }}`, `{{ colorvar|hoverColor(30) }}`
	 *
	 * @param string $hexCode Ein Farbcode, z.B. #000000 oder #FFFFFF.
	 * @param float $amount Ein Wert zwischen -100 und 100. Keine Änderung bei 0, komplett aufgehellt/abgedunkelt bei 100.
	 * @param float $breakpoint Ab welcher Lightness der Wechsel von Aufhellen zu Abdunkeln stattfinden soll
	 * @return string
	 */
	public function hoverColor($hexCode, float $amount = 15, float $breakpoint = 0.4): string
	{
		$amount = min(max(-100, $amount), 100);
		try {
			$color = Color::fromCssString((string)$hexCode);
		} catch (UnexpectedValueException $e) {
			return (string)$hexCode;
		}

		if ($amount < 0) {
			return $color->getLightness() < $breakpoint
				? $this->darkenColor($hexCode, -$amount)
				: $this->lightenColor($hexCode, -$amount);
		} else {
			return $color->getLightness() < $breakpoint
				? $this->lightenColor($hexCode, $amount)
				: $this->darkenColor($hexCode, $amount);
		}
	}

	/**
	 * Fügt einem Hex-Farbcode einen Alphawert hinzu und gibt einen rgba-String zurück.
	 *
	 * Usage: `{{ colorvar|alpha(0.5) }}`
	 *
	 * @param string $hexCode Ein Farbcode, z.B. #000000 oder #FFFFFF.
	 * @param float $alpha Ein Wert zwischen 0 und 1.
	 * @return string
	 */
	public function alpha($hexCode, float $alpha): string
	{
		$alpha = (float)min(max(0, $alpha), 1);
		try {
			$color = Color::fromCssString((string)$hexCode);
		} catch (UnexpectedValueException $e) {
			return $hexCode;
		}

		return $color->setAlpha($alpha);
	}

	/**
	 * Wählt die dunkle Vordergrundfarbe, wenn der Hintergrund hell ist und ansonsten die
	 * helle Vordergrundfarbe.
	 *
	 * Usage: `{{ selectColor(bgColor, darkFgColor, lightFgColor) }}`
	 *
	 * @param string $reference Farbcode des Hintergrunds
	 * @param string $dark Dunkle Vordergrundfarbe
	 * @param string $light Helle Vordergrundfarbe
	 * @param float $breakpoint Ab welcher Lightness der Wechsel stattfinden soll
	 * @return string
	 */
	public function selectColor($reference, $dark, $light, float $breakpoint = 0.47): string
	{
		try {
			$color = Color::fromCssString((string)$reference);
		} catch (UnexpectedValueException $e) {
			return $reference;
		}

		return $color->getLightness() < $breakpoint
			? (string)$light
			: (string)$dark;
	}

	/**
	 * Wählt `$preference`, wenn das Kontrastverhältnis zu `$test` akzeptabel ist.
	 * Alternativ wird bei besserem Kontrastverhältnis `$fallback` gewählt.
	 *
	 * Usage: `{{ contrastColor(buttonBgColor, windowBgColor, buttonFgColor) }}`
	 *
	 * @param Color|string $preference
	 * @param Color|string $test
	 * @param Color|string $fallback
	 * @return string
	 */
	public function contrastColor($preference, $test, $fallback): string
	{
		try {
			$preference = Color::fromCssString((string)$preference);
			$test = Color::fromCssString((string)$test);
			$fallback = Color::fromCssString((string)$fallback);
		} catch (UnexpectedValueException $e) {
			return (string)$preference;
		}

		$threshold = 1.5;

		return $preference->contrastRatio($test) < $threshold && $fallback->contrastRatio($test) >= $threshold
			? (string)$fallback
			: (string)$preference;
	}

	/**
	 * CCM19-Platzhalter in HTML-Code ersetzen
	 *
	 * Usage: `{{ variable|replace_placeholders }}`
	 *
	 * @param string|Markup $content HTML-Code in dem Placeholder ersetzt werden sollen
	 * @return Markup
	 */
	public function replacePlaceholders($content)
	{
		$content = Placeholder::applyAll((string)$content);
		return new Markup($content, 'UTF-8');
	}

	/**
	 * Das HTML-Attribut `checked` generieren, wenn eine Bedingung erfüllt ist
	 *
	 * Usage: `{{ variable|checked }}`, `{{ checked(var_a == var_b) }}`
	 *
	 * @param bool $expression
	 * @return string
	 */
	public function checked($expression)
	{
		return $expression ? ' checked="checked"' : '';
	}

	/**
	 * Das HTML-Attribut `disabled` generieren, wenn eine Bedingung erfüllt ist
	 *
	 * Usage: `{{ variable|disabled }}`, `{{ disabled(var_a == var_b) }}`
	 *
	 * @param bool $expression
	 * @return string
	 */
	public function disabled($expression)
	{
		return $expression ? ' disabled="disabled"' : '';
	}

	/**
	 * Das HTML-Attribut `readonly` generieren, wenn eine Bedingung erfüllt ist
	 *
	 * Usage: `{{ variable|readOnly }}`, `{{ readOnly(var_a == var_b) }}`
	 *
	 * @param bool $expression
	 * @return string
	 */
	public function readOnly(bool $expression): string
	{
		return $expression ? ' readonly="readonly"' : '';
	}

	/**
	 * Klassennamen in den finalen `string` übernehmen, wenn eine jeweilige Bedingung erfüllt ist.
	 *
	 * Usage: `{{ {'classA': false, 'classB': true}|classIf }}`, `{{ classIf({'classA': false, 'classB': true}) }}`
	 *
	 * @param array<string,bool> $classNames
	 * @return string
	 */
	public function classIf(array $classNames): string
	{
		return rtrim(' ' . implode(' ', array_map('htmlspecialchars', array_keys(array_filter($classNames)))));
	}

	/**
	 * Escaped HTML mit Ausnahme von HTML-Links (`<a></a>`)
	 *
	 * Links mit `target` erhalten `rel="noopener"`.
	 *
	 * Usage: `{{ variable|preserve_links }}`
	 *
	 * @param string $htmlCode
	 * @param bool $forceTargetBlank Ob immer `target="_blank"` ergänzt werden soll
	 * @return string
	 */
	public function preserveLinks(string $htmlCode, bool $forceTargetBlank=true): string
	{
		$links = [];

		// Extract links and escape their text contents
		$htmlCode = preg_replace_callback(
			'~(?<openingTag><a(?=\s|>)[^<>]*>)(?<text>[^<>]+)</a>~i',
			function ($match) use (&$links, $forceTargetBlank) {
				$openingTag = $match['openingTag'];
				$target = preg_match('~\starget=(?<quote>"|\')(?<target>[^<>]*?)\g{quote}~i', $openingTag, $targetMatch)
					? $targetMatch['target']
					: '_self';

				// Apply target attribute
				if ($forceTargetBlank) {
					$openingTag = preg_replace('~\starget=("|\')[^<>]*?\g{-1}~i', ' target="_blank"', $openingTag, -1, $count);
					if ($count == 0) {
						$openingTag = str_replace('>', ' target="_blank">', $openingTag);
					}
					$target = '_blank';
				}

				// Apply rel attribute
				if ($target != '_self') {
					$openingTag = preg_replace('~\srel=("|\')[^<>]*?\g{-1}~i', ' rel="noopener"', $openingTag, -1, $count);
					if ($count == 0) {
						$openingTag = str_replace('>', ' rel="noopener">', $openingTag);
					}
				}

				$placeholder = uniqid('#preserveLink-');
				$links[$placeholder] = $openingTag.htmlspecialchars($match['text']).'</a>';
				return $placeholder;
			},
			$htmlCode
		);

		// Escape everything else inside the text
		$htmlCode = htmlspecialchars($htmlCode);

		// Restore links
		foreach ($links as $placeholder => $link) {
			$htmlCode = str_replace($placeholder, $link, $htmlCode);
		}

		return $htmlCode;
	}

	/**
	 * This method masks HTML elements to expose the full text content for processing.
	 * Masks start with `@mask--` followed by a unique identifier and end with a semicolon.
	 * After processing, the masked content is restored to its previous state.
	 *
	 * @param string $text HTML text
	 * @param callable $wrapFn Wrapping function that processes the masked text
	 * @return string
	 */
	private static function wrapHTMLText(string $text, callable $wrapFn): string
	{
		// Stashed content that is to be restored later
		$contentStack = [];

		/**
		 * Pushes content to the stack and returns its unique identifier.
		 * @param string $content
		 * @return string
		 */
		$maskContent = function (string $content) use (&$contentStack): string {
			$needle = '@mask--'.uniqid().';';
			$contentStack[$needle] = $content;
			return $needle;
		};

		// Mask block-level HTML tags to expose full sentences for processing
		$blockTags = ['div', 'p'];
		$text = preg_replace_callback(
			'~<(' . implode('|', $blockTags) . ')(?=[\s/>])[^>]*>|</(' . implode('|', $blockTags) . ')>~',
			static function (array $match) use ($maskContent) {
				return $maskContent($match[0]);
			},
			$text
		);

		// Mask the remaining HTML elements including their content (i.e. sentence chunks)
		$text = preg_replace_callback(
			'~<(?<tag>[^\s/>]+)[^>]*(?(?<=/)>|>[^<]*(?R)*[^<]*)(?(?<=/>)|</\g{tag}>)~',
			static function (array $match) use ($maskContent) {
				return $maskContent($match[0]);
			},
			$text
		);

		// Mask abbreviations that are separated by whitespace
		$text = preg_replace_callback(
			'~\b([a-z]{1,3}\.\s?){2,}~i',
			static function (array $match) use ($maskContent) {
				return $maskContent($match[0]);
			},
			$text
		);

		// Call the individual wrapping function
		$text = $wrapFn($text);

		// Restore masked text to its previous state
		return str_replace(array_reverse(array_keys($contentStack)), array_reverse($contentStack), $text);
	}

	/**
	 * This method splits a text into sentences and puts each sentence into its own element.
	 *
	 * Usage: `{{ variable|wrap_sentences('span', 'classA classB') }}`, `{{ wrap_sentences(variable, 'span', 'classA classB') }}`
	 *
	 * @param string $text The text to split into sentences
	 * @param string $htmlTag The HTML tag to wrap the sentences in
	 * @param string $className The class name to apply to the sentences
	 * @return string
	 */
	public function wrapSentences(string $text, string $htmlTag='span', string $className=''): string
	{
		return self::wrapHTMLText($text, static function (string $text) use ($className, $htmlTag): string {
			$classAttribute = $className ? ' class="' . htmlspecialchars($className) . '"' : '';
			// Split sentences into separate span elements taking masked content (@) into account
			return preg_replace_callback(
				'~(?(DEFINE)(?<mask>@mask--[^;]+;))((?:(?&mask)\s*)*+)((?:[^.!?]++.)+?)((?(?=\s)\s++(?!\d)|(?=$|(?&mask))))~',
				static function (array $match) use ($classAttribute, $htmlTag): string{
					// The subroutine definition causes additional groups to be created; remove them
					array_splice($match, 1, -3);
					// Explicitly reindex the array
					$match = array_values($match);
					// Leading masks (1) + sentence (2) + trailing whitespace (3)
					return "{$match[1]}<$htmlTag$classAttribute>{$match[2]}</$htmlTag>{$match[3]}";
				},
				$text
			);
		});
	}

	/**
	 * This method splits a text into words and puts each word into its own element.
	 *
	 * Usage: `{{ variable|wrap_words('span', 'classA classB') }}`, `{{ wrap_words(variable, 'span', 'classA classB') }}`
	 *
	 * @param string $text The text to split into words
	 * @param string $htmlTag The HTML tag to wrap the words in
	 * @param string $className The class name to apply to the words
	 * @return string
	 */
	public function wrapWords(string $text, string $htmlTag='span', string $className=''): string
	{
		return self::wrapHTMLText($text, static function (string $text) use ($className, $htmlTag): string {
			$classAttribute = $className ? ' class="' . htmlspecialchars($className) . '"' : '';
			// Split sentences into separate span elements taking masked content (@) into account
			return preg_replace_callback(
				'~(?(DEFINE)(?<mask>@mask--[^;]+;))((?:(?&mask)\s*)*+)(\S++)(\s*+)~',
				static function (array $match) use ($classAttribute, $htmlTag): string{
					// The subroutine definition causes additional groups to be created; remove them
					array_splice($match, 1, -3);
					// Explicitly reindex the array
					$match = array_values($match);
					// Leading masks (1) + word (2) + trailing whitespace (3)
					return "{$match[1]}<$htmlTag$classAttribute>{$match[2]}</$htmlTag>{$match[3]}";
				},
				$text
			);
		});
	}

	/**
	 * Kodiert eine Datei in eine `data:`-URL
	 *
	 * Usage: `{{ file|base64_data_url }}`
	 *
	 * @param File|null $file
	 * @return string
	 */
	public function base64DataUrl(?File $file): string
	{
		return $file ? "data:{$file->getMimeType()};base64,".base64_encode(file_get_contents((string)$file)) : '';
	}

	/**
	 * Generiert die HTML-Attribute `height="yyy" width="xxx"` für eine Bilddatei
	 *
	 * Usage: `{{ file|image_dimensions }}`
	 *
	 * @param File|null $file
	 * @return string
	 */
	public function imageDimensions(?File $file): string
	{
		return rtrim($file ? ' '.(@getimagesize((string)$file)[3] ?? '') : '', ' ');
	}

	/**
	 * Das HTML-Attribut `required` generieren, wenn eine Bedingung erfüllt ist
	 *
	 * Usage: `{{ variable|required }}`, `{{ selected(var_a == var_b) }}`
	 *
	 * @param bool $expression
	 * @return string
	 */
	public function required(bool $expression): string
	{
		return $expression ? ' required="required"' : '';
	}

	/**
	 * Das HTML-Attribut `selected` generieren, wenn eine Bedingung erfüllt ist
	 *
	 * Usage: `{{ variable|selected }}`, `{{ variable|selected(false) }}`, `{{ selected(var_a == var_b) }}`
	 *
	 * @param mixed $expression Zu überprüfender Ausdruck
	 * @param mixed $test Womit verglichen werden soll (Default: truthy)
	 * @param bool $strict Wenn TRUE: Strikter Vergleich mit Typübereinstimmung
	 * @return string
	 */
	public function selected($expression, $test=true, bool $strict=false): string
	{
		return ($strict ? $expression === $test : $expression == $test) ? ' selected="selected"' : '';
	}

	/**
	 * Eine Lebenszeit-Angabe in Sekunden für Menschen lesbar formatieren
	 *
	 * Usage: `{{ variable|interval_lang("de") }}`
	 *
	 * @param int $seconds
	 * @param string $langCode Sprachcode (z.B. de, de_DE oder de-DE)
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function humanReadableInterval(int $seconds, $langCode)
	{
		$langCode = strtolower(explode('-', str_replace('_', '-', $langCode))[0]);

		$wordMap = [
			'de' => [
				'1s'=>'%d Sekunde', 's'=>'%d Sekunden',
				'1min'=>'%d Minute', 'min'=>'%d Minuten',
				'1h'=>'%d Stunde', 'h'=>'%d Stunden',
				'1d'=>'%d Tag', 'd'=>'%d Tage',
				'1w'=>'%d Woche', 'w'=>'%d Wochen',
				'1mon'=>'%d Monat', 'mon'=>'%d Monate',
				'1y'=>'%d Jahr', 'y'=>'%d Jahre',
				'0' => 'bis zum Ende der Browser-Sitzung',
			],
			'en' => [
				'1s'=>'%d second', 's'=>'%d seconds',
				'1min'=>'%d minute', 'min'=>'%d minutes',
				'1h'=>'%d hour', 'h'=>'%d hours',
				'1d'=>'%d day', 'd'=>'%d days',
				'1w'=>'%d week', 'w'=>'%d weeks',
				'1mon'=>'%d month', 'mon'=>'%d months',
				'1y'=>'%d year', 'y'=>'%d years',
				'0' => 'until end of browser session',
			],
		];
		$words = $wordMap[$langCode] ?? $wordMap['en'];

		if ($seconds <= 0) {
			return $words['0'];
		}

		$result = [];
		if ($seconds >= 30000000) {
			$years = round($seconds/31536000, 0, PHP_ROUND_HALF_DOWN);
			$seconds -= 31536000*$years;
			if ($seconds < -2470000) {
				$seconds += 31536000;
				$years -= 1;
			}
			$result[] = sprintf($words[$years.'y'] ?? $words['y'], $years);
		}
		if ($seconds >= 2470000) {
			$months = ceil($seconds/2592000);
			$seconds = 0;
			$result[] = sprintf($words[$months.'mon'] ?? $words['mon'], $months);
		}
		if ($seconds >= 604800) {
			$weeks = ceil($seconds/604800);
			$seconds = 0;
			$result[] = sprintf($words[$weeks.'w'] ?? $words['w'], $weeks);
		}
		if ($seconds >= 86399) {
			$days = round($seconds/86400, 0, PHP_ROUND_HALF_DOWN);
			$seconds -= 86400*$days;
			if ($seconds < -3600) {
				$seconds += 86400;
				$days -= 1;
			}
			$result[] = sprintf($words[$days.'d'] ?? $words['d'], $days);
		}
		if ($seconds >= 3599) {
			$hours = round($seconds/3600, 0, PHP_ROUND_HALF_DOWN);
			$seconds -= 3600*$hours;
			if ($seconds < -300) {
				$seconds += 3600;
				$hours -= 1;
			}
			$result[] = sprintf($words[$hours.'h'] ?? $words['h'], $hours);
		}
		if ($seconds >= 59) {
			$minutes = ceil($seconds/60);
			$seconds = 0;
			$result[] = sprintf($words[$minutes.'min'] ?? $words['min'], $minutes);
		}
		if ($seconds >= 1) {
			$result[] = sprintf($words[$seconds.'s'] ?? $words['s'], $seconds);
		}
		return implode(' ', $result);
	}

	/**
	 * URL zum Asset eines Plugins generieren
	 *
	 * Usage: `{{ plugin_asset("plugin-uid", "subpath.png") }}`
	 *
	 * @param Environment $env
	 * @param string $pluginUid
	 * @param string $path
	 * @return string
	 * @phan-suppress PhanAccessMethodInternal
	 * @phan-side-effect-free
	 */
	public function pluginAsset(Environment $env, $pluginUid, $path)
	{
		$func = $env->getFunction('path');
		if (!$func) {
			return 'about:blank';
		}
		$callable = $func->getCallable();
		return $callable('app_plugin_assets', ['pluginuid'=>$pluginUid, 'path'=>$path]);
	}

	/**
	 * Sortiert unter Berücksichtigung eines Locales
	 *
	 * Usage: `{{ variable|intl_sort }}`, `{{ variable|intl_sort("de_DE") }}`, `{{ variable|intl_sort("de_DE", (x) => x.name) }}`
	 *
	 * @template T
	 * @param Environment $env
	 * @param array<T> $array
	 * @param string|null $locale Nach welchem Locale sortiert werden soll (NULL für aktuelles Backend-Locale)
	 * @param null|callable(T):mixed $keyFunction Funktion, die aus einem Element den Wert holt, nach dem sortiert werden soll (`(x) => x.name` o.ä.)
	 * @return list<T>
	 * @phan-side-effect-free
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function intlSort(Environment $env, $array, $locale=null, $keyFunction=null)
	{
		if ($locale === null) {
			$vars = $env->getGlobals();
			$locale = $vars['_locale'];
		}
		if (class_exists('Collator')) {
			if (!isset($this->collators[$locale])) {
				$this->collators[$locale] = Collator::create($locale);
			}
			$collator = $this->collators[$locale];
		} else {
			$collator = null;
		}

		if ($keyFunction) {
			$mapped = array_map($keyFunction, $array);
			if ($collator) {
				$collator->asort($mapped);
			} else {
				asort($mapped);
			}
			$result = [];
			foreach ($mapped as $key => $_) {
				$result[] = $array[$key];
			}
			return $result;
		} else {
			if ($collator) {
				$collator->sort($array);
			} else {
				sort($array);
			}
			return $array;
		}
	}

	/**
	 * Token, um die rescue.php für die nächsten 10 Minuten öffnen zu können
	 *
	 * Usage: `{{ rescue_token() }}`
	 *
	 * @return string
	 * @phan-side-effect-free
	 */
	public function rescueToken(): string
	{
		return Utils::getRescueToken();
	}
}
