<?php

namespace App\Twig;

use App\Event\TemplateResolveEvent;
use App\Event\TemplateRenderEvent;
use App\Utils;
use Twig\Template;

class Environment extends \Twig\Environment
{
	/** @var array<string,string> $templateNameOverrideCache */
	private $templateNameOverrideCache = [];

	/** {@inheritdoc}
	 * @override
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function loadTemplate(string $cls, string $name, int $index = null): Template
	{
		$name = $this->resolveTemplateName($name);

		// Consider that the resolved template name might result in a different class name when compiled
		$cls = $this->getTemplateClass($name, $index);

		return parent::loadTemplate($cls, $name, $index);
	}

	/** {@inheritdoc}
	 * @override
	 * @phan-suppress PhanAccessMethodInternal
	 */
	public function getTemplateClass(string $name, int $index = null): string
	{
		if (substr($name, 0, 7) === '@@orig/') {
			$name = substr($name, 7);
		}
		return parent::getTemplateClass($name, $index);
	}

	/**
	 * @param string $name
	 * @return string
	 */
	private function resolveTemplateName(string $name): string
	{
		if (substr($name, 0, 7) === '@@orig/') {
			return substr($name, 7);
		}
		if (isset($this->templateNameOverrideCache[$name])) {
			return $this->templateNameOverrideCache[$name];
		}

		// Event erzeugen und dispatchen
		$event = new TemplateResolveEvent($name);
		Utils::dispatchEvent(TemplateResolveEvent::nameForView($name), $event);

		// Mögliche Änderungen durch Resolve-Event-Handler berücksichtigen
		$newName = $event->getView();
		$this->templateNameOverrideCache[$name] = $newName;

		return $newName;
	}

	/**
	 * {@inheritdoc}
	 * @param string $name
	 * @param array<string,mixed> $context
	 * @override
	 *
	 * Löst ein `TemplateRenderEvent` aus.
	 */
	public function render($name, array $context = []): string
	{
		// Event erzeugen und dispatchen
		$event = new TemplateRenderEvent($name, $context);
		Utils::dispatchEvent(TemplateRenderEvent::nameForView($name), $event);

		return parent::render($event->getView(), $event->getParameters());
	}

	/**
	 * {@inheritdoc}
	 * @override
	 * @param string $name
	 * @param array<string,mixed> $context
	 *
	 * Löst ein `TemplateRenderEvent` aus.
	 */
	public function display($name, array $context = []): void
	{
		// Event erzeugen und dispatchen
		$event = new TemplateRenderEvent($name, $context);
		Utils::dispatchEvent(TemplateRenderEvent::nameForView($name), $event);

		// Mögliche Änderungen durch Render-Event-Handler berücksichtigen
		parent::display($event->getView(), $event->getParameters());
	}
}
