<?php

namespace App;

use App\Event\FilterEvent;
use App\Exception\FilesizeLimitExceededException;
use App\Exception\MimeTypeNotSupportedException;
use App\Exception\TempFileNotWritableException;
use App\Model\Domain;
use App\Model\MenuData;
use App\Model\PluginState;
use App\Model\SubGroups;
use App\Model\User;
use DirectoryIterator;
use Exception;
use InvalidArgumentException;
use LogicException;
use RuntimeException;
use Symfony\Component\Dotenv\Dotenv;
use Symfony\Component\EventDispatcher\GenericEvent;
use Symfony\Component\HttpFoundation\Exception\SessionNotFoundException;
use Symfony\Component\HttpFoundation\File\Exception\FileException;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\HttpFoundation\IpUtils;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\Session\Session;
use Symfony\Component\HttpKernel\EventListener\SessionListener;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;
use Symfony\Component\Routing\Router;
use Symfony\Contracts\Translation\TranslatorInterface;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;
use Symfony\Contracts\EventDispatcher\Event;

class Utils
{
	const RANDOM_STRING_CHARSET = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';

	const REPOSITORY_STORE_FILE = 'json';
	const REPOSITORY_STORE_MONGODB = 'mongodb';
	const REPOSITORY_STORE_REDIS = 'redis';

	const CRON_MODE_FRONTEND = 'frontend';
	const CRON_MODE_SERVER = 'server';

	const EVENT_FILTER_USER_RIGHTS = 'ccm19.routeAccess.user.rights';

	/** @var array<string,array<string,bool>> $cachedGroupRights */
	private static $cachedGroupRights = [
		'' => ['app_login' => true, 'app_login_' => true, 'app_root' => true]
	];

	/**
	 * Diese Methode generiert eine zufällige Zeichenkette bestimmter Länge und gegebenem Zeichensatz.
	 * @param int $length
	 * @param string $charset
	 * @return string
	 * @throws \Exception if it was not possible to gather sufficient entropy.
	 * @phan-side-effect-free
	 */
	public static function randomString(int $length = 16, string $charset = self::RANDOM_STRING_CHARSET): string
	{
		$charsetLength = mb_strlen($charset);

		$chars = [];
		for ($i = 0; $i < $length; $i++) {
			$chars[] = mb_substr($charset, random_int(0, $charsetLength - 1), 1);
		}

		return implode('', $chars);
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getLicenseKey(): string
	{
		return (string)Config::getInstance()->get('licensekey');
	}

	/**
	 * Key für globale Cron-Aufrufe, direkt vom Lizenzschlüssel abgeleitet
	 *
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getCronKey(): string
	{
		return substr(hash_hmac('sha256', 'cron', self::getLicenseKey()), 0, 16);
	}

	/**
	 * Token, um die rescue.php für die nächsten 10 Minuten öffnen zu können
	 *
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getRescueToken(): string
	{
		$baseKey = hash_pbkdf2("sha256", $_SERVER['APP_SECRET'], 'rescue', 1000);
		return substr(hash_hmac('sha256', (string)(int)(time()/600), $baseKey), 0, 24);
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function hasLicenseKey(): bool
	{
		return strlen(self::getLicenseKey()) > 0;
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isLicensed(): bool
	{
		return preg_match('~^[a-f0-9]{10,}$~', self::getLicenseKey()) == 1;
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getUpdateChannel(): string
	{
		return (string)($_ENV['APP_UPDATE_CHANNEL'] ?? 'stable');
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isCCMCloud(): bool
	{
		return (bool)($_ENV['APP_CCM_CLOUD'] ?? false);
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isClickskeksCloud(): bool
	{
		return (bool)($_ENV['APP_CLICKSKEKS_CLOUD'] ?? false);
	}

	/**
	 * @param User|null $user If omitted, the currently logged in user is used.
	 * @return bool Returns `true` if the user is using the free plan in the CCM19 cloud, `false` otherwise.
	 */
	public static function isFreePlan(?User $user=null): bool
	{
		// If the user is omitted, check the currently logged in user
		if ($user === null) {
			$user = User::loggedInUser();
		}

		// If the user is not a client, the free plan does not apply
		if ($user === null || !$user->isClient()) {
			return false;
		}

		// Read plan details from the account owner
		$user = $user->getAccountOwner();

		return self::isCCMCloud() && $user->getMaxCallCount() >= 0 && $user->getMaxCallCount() <= 5000;
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getRepositoryStore(): string
	{
		$store = $_ENV['APP_REPOSITORY_STORE'] ?? '';
		$allowedValues = [
			self::REPOSITORY_STORE_FILE,
			self::REPOSITORY_STORE_MONGODB,
		];

		return in_array($store, $allowedValues) ? $store : self::REPOSITORY_STORE_FILE;
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getProtocolRepositoryStore(): string
	{
		$store = $_ENV['APP_PROTOCOL_STORE'] ?? $_ENV['APP_REPOSITORY_STORE'] ?? '';
		$allowedValues = [
			self::REPOSITORY_STORE_FILE,
			self::REPOSITORY_STORE_MONGODB,
			self::REPOSITORY_STORE_REDIS,
		];

		return in_array($store, $allowedValues) ? $store : self::getRepositoryStore();
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getArchiveRepositoryStore(): string
	{
		$store = $_ENV['APP_PROTOCOL_ARCHIVE_STORE'] ?? $_ENV['APP_PROTOCOL_STORE'] ?? $_ENV['APP_REPOSITORY_STORE'] ?? '';
		$allowedValues = [
			self::REPOSITORY_STORE_FILE,
			self::REPOSITORY_STORE_MONGODB,
		];

		return in_array($store, $allowedValues) ? $store : self::getRepositoryStore();
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getLoginAttemptsLogRepositoryStore(): string
	{
		$store = $_ENV['APP_LOGIN_ATTEMPTS_LOG_STORE'] ?? $_ENV['APP_REPOSITORY_STORE'] ?? '';
		$allowedValues = [
			self::REPOSITORY_STORE_FILE,
			self::REPOSITORY_STORE_MONGODB,
			self::REPOSITORY_STORE_REDIS,
		];

		return in_array($store, $allowedValues) ? $store : self::getRepositoryStore();
	}

	/**
	 * @return string|null
	 * @phan-side-effect-free
	 */
	public static function getCustomArchiveRoot(): ?string
	{
		$value = $_ENV['APP_PROTOCOL_ARCHIVE_ROOT'] ?? null;
		if ($value === null or trim($value) === '') {
			return null;
		}
		return $value;
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function hasWhitelabelLicense(): bool
	{
		/** @var User|null $user */
		$user = User::loggedInUser();
		/** @var Domain|null $domain */
		$domain = null;

		if ($user) {
			$domain = Domain::activeDomain();
		}

		return self::isExtendedEdition()
			? ($user->isAdministrator() || $user->isManager()) || ($domain && $domain->hasWhitelabel())
			: (bool)Config::getInstance()->get('whitelabel');
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getPoweredByText(): string
	{
		$fallback = 'Powered by CCM19';
		return self::isExtendedEdition() ? (string)(Config::getInstance()->get('poweredByText') ?? $fallback) : $fallback;
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getPoweredByTooltip(): string
	{
		$fallback = 'Cookie Consent Tool / Cookie Consent Manager';
		return self::isExtendedEdition() ? (string)(Config::getInstance()->get('poweredByTooltip') ?? $fallback) : $fallback;
	}

	/**
	 * @return string
	 * @phan-suppress PhanTypeInvalidDimOffset
	 */
	public static function getPoweredByUrl(): string
	{
		$url = Config::getInstance()->get('poweredByUrl');

		// Migration des Konfigurationsinhalts alter Versionen
		if (preg_match('~<a\s[^>]*(?<=\s)href="(?<url>[^"]+)~', (string)$url, $match)) {
			$url = $match['url'];
			Config::getInstance()->set('poweredByUrl', $url);
		}

		$fallback = self::getCmpServiceUrl('/');
		return self::isExtendedEdition() ? (string)($url ?? $fallback) : $fallback;
	}

	/**
	 * @return string
	 */
	public static function getDocsUrl(): string
	{
		$url = Config::getInstance()->get('docsurl');
		$fallback = 'https://www.docs.ccm19.de/';
		if(self::isClickskeksCloud())
		{
			$fallback = 'https://www.docs.clickskeks.de/';
		}
		return (string)($url ?? $fallback);
	}

	/**
	 * Diese Methode liefert eine URL zur Service-Website, die zur gegebenen Umgebung passt.
	 * Gemeint sind damit die Infoseiten (Shops) in Anbindung an CCM19 und Clickskeks.
	 *
	 * Die Basis-URL lässt sich mittels Umgebungsvariable `APP_SERVICE_BASE_URL` überschreiben.
	 * Das ist nützlich, um z. B. Funktionalität in einer Testinstallation zu prüfen.
	 *
	 * @param string $requestUri
	 * @param array $params
	 * @return string
	 */
	public static function getCmpServiceUrl(string $requestUri, array $params=[]): string
	{
		if (preg_match('~^https?://[^/]~', $_ENV['APP_SERVICE_BASE_URL'] ?? '')) {
			$baseUrl = rtrim($_ENV['APP_SERVICE_BASE_URL'], '/') . '/';
		}
		elseif (self::isClickskeksCloud()) {
			$baseUrl = 'https://service.clickskeks.de/';
		}
		else {
			$baseUrl = 'https://www.ccm19.de/';
		}

		return $baseUrl . ltrim($requestUri, '/') . ($params ? '?' . http_build_query($params, '', '&') : '');
	}

	/**
	 * @return bool
	 */
	public static function isHelpEnabled(): bool
	{
		return (bool)(Config::getInstance()->get('enablehelp') ?? true);
	}

	/**
	 * @return array<string,mixed>
	 */
	public static function getBackgroundImageReal(): array
	{
		$backgroundImage = [
			"src" => self::getBackgroundImage()['src'],
			"unsplashHref" => self::getUnsplashHref(),
			"unsplashName" => self::getUnsplashName(),
			"headlineColor" => self::getHeadlineColor(),
			"unsplashLinkColor" =>self::getUnsplashLinkColor(),
			"unsplashLinkBackgroundColor" =>self::getUnsplashLinkBackgroundColor(),
			"unsplashLinkBackgroundColorOpacity" =>self::getUnsplashLinkBackgroundColorOpacity(),
			"portalName" =>self::getPortalName(),
			"portalURL" => self::getPortalURL(),
		];
		return $backgroundImage;
	}

	/**
	 * @return string
	 */
	public static function getPortalURL(): string
	{
		$data = self::getBackgroundImage();
		return (string)(Config::getInstance()->get('portalURL') ?? $data['portalURL']);
	}

	/**
	 * @return string
	 */
	public static function getPortalName(): string
	{
		$data = self::getBackgroundImage();
		return (string)(Config::getInstance()->get('portalName') ?? $data['portalName']);
	}

	/**
	 * @return string
	 */
	public static function getUnsplashLinkBackgroundColorOpacity(): string
	{
		$data = self::getBackgroundImage();
		return (string)(Config::getInstance()->get('unsplashLinkBackgroundColorOpacity') ?? $data['unsplashLinkBackgroundColorOpacity']);
	}

	/**
	 * @return string
	 */
	public static function getUnsplashLinkBackgroundColor(): string
	{
		$data = self::getBackgroundImage();
		return (string)(Config::getInstance()->get('unsplashLinkBackgroundColor') ?? $data['unsplashLinkBackgroundColor']);
	}

	/**
	 * @return string
	 */
	public static function getUnsplashLinkColor(): string
	{
		$data = self::getBackgroundImage();
		return (string)(Config::getInstance()->get('unsplashLinkColor') ?? $data['unsplashLinkColor']);
	}

	/**
	 * @return string
	 */
	public static function getHeadlineColor(): string
	{
		$data = self::getBackgroundImage();
		return (string)(Config::getInstance()->get('headlineColor') ?? $data['headlineColor']);
	}

	/**
	 * Gibt im WL Agency Falle die ID eines WL Embeddings zurück das statt des CCM19 Embeddings verwendet werden soll
	 * @return string
	 */
	public static function getWhiteLabelEmbedding(): string
	{
		return (string)(Config::getInstance()->get('wlEmbedding') ?? "");
	}

	/**
	 * @return string
	 */
	public static function getUnsplashName(): string
	{
		$data = self::getBackgroundImage();
		return (string)(Config::getInstance()->get('unsplashName') ?? $data['unsplashName']);
	}
	/**
	 * @return string
	 */
	public static function getUnsplashHref(): string
	{
		$data = self::getBackgroundImage();
		return (string)(Config::getInstance()->get('unsplashHref') ?? $data['unsplashHref']);
	}

	/**
	 * Diese Methode dekodiert eine Anfrage vom Typ `application/json`.
	 * @param Request $request
	 * @param mixed|JsonResponse $data @phan-output-reference Das Ergebnis der Methode wird in der Referenz $data
	 *                                 hinterlegt. Bei Erfolg ist das dekodierte Ergebnis enthalten, andernfalls ein
	 *                                 JsonResponse-Objekt, das die Fehlermeldung bezeichnet.
	 * @param bool $assoc   Wird intern an json_decode übergeben.
	 * @param int  $depth   Wird intern an json_decode übergeben.
	 * @param int  $options Wird intern an json_decode übergeben.
	 * @return bool Gibt true zurück, wenn das JSON-Objekt des Requests erfolgreich dekodiert wurde, andernfalls false.
	 */
	public static function decodeJsonRequest(Request $request, &$data, bool $assoc = true, int $depth = 5, int $options = 0): bool
	{
		if ($request->getContentType() == 'json') {
			try {
				$json = $request->getContent();
				$data = @json_decode($json, $assoc, $depth, $options);
				if ($data === null && json_last_error() != JSON_ERROR_NONE) {
					throw new RuntimeException(json_last_error_msg(), json_last_error());
				}

				return true;
			}
			catch (RuntimeException $e) {
				$data = new JsonResponse(['success'=>false, 'message'=>$e->getMessage()], 400);
			}
			catch (Exception $e) {
				$data = new JsonResponse(['success'=>false, 'message'=>$e->getMessage()], 500);
			}
		}
		else {
			$data = new JsonResponse(['success'=>false, 'message'=>'Invalid payload. Make sure to send a JSON encoded object and set Content-Type header to application/json.'], 400);
		}

		return false;
	}

	/**
	 * This method decodes an image from base64 and provides it as a temporary file.
	 * When the file is not needed anymore, unlink it.
	 * @param string $data base64 encoded image
	 * @param int $maxSize maximum file size in bytes; unlimited if 0.
	 * @param string[] $mimeTypes supported image file types
	 * @return File
	 * @throws FilesizeLimitExceededException
	 * @throws InvalidArgumentException if $data wasn't base64 encoded.
	 * @throws LogicException if a mime type is not mappable. (only during development)
	 * @throws MimeTypeNotSupportedException
	 * @throws TempFileNotWritableException
	 */
	public static function decodeBase64Image(string $data, int $maxSize=0, array $mimeTypes=['image/jpeg', 'image/png']): File
	{
		$extensions = [
			'image/gif' => '.gif',
			'image/jpeg' => '.jpg',
			'image/png' => '.png',
			'image/svg+xml' => '.svg',
		];

		if ($unsupportedMimeTypes = array_diff($mimeTypes, array_keys($extensions))) {
			throw new LogicException('Unsupported mime type(s): '.implode(', ', $unsupportedMimeTypes));
		}

		$mimeTypes = array_intersect_key($extensions, array_flip($mimeTypes));

		if (($data = base64_decode($data, true)) !== false) {
			if ($maxSize > 0 && strlen($data) > $maxSize) {
				throw new FilesizeLimitExceededException();
			}
			elseif ($tempFilename = tempnam(sys_get_temp_dir(), 'ccm')) {
				file_put_contents($tempFilename, $data);
				$file = new File($tempFilename);

				if (array_key_exists($file->getMimeType(), $mimeTypes)) {
					$newTempFilename = $tempFilename.$mimeTypes[$file->getMimeType()];

					try {
						$file = $file->move(sys_get_temp_dir(), basename($newTempFilename));
					}
					catch (FileException $exception) {
						unset($file);
						unlink($tempFilename);
						throw new TempFileNotWritableException('Error while moving image.');
					}

					return $file;
				}
				else {
					unset($file);
					unlink($tempFilename);
					throw new MimeTypeNotSupportedException(
						'Invalid file, valid image formats: '.implode(', ', array_keys($mimeTypes))
					);
				}
			}
			else {
				throw new TempFileNotWritableException('Can\'t store image, writing of temporary file failed.');
			}
		}
		else {
			throw new InvalidArgumentException('Invalid file, expecting base64 encoded string.');
		}
	}

	/**
	 * This method validates `$subnet` and removes unnecessary characters like brackets.
	 * IP addresses and subnets, i.e. IP address ranges, are considered valid.
	 * @param string $subnet IP address or subnet
	 * @return string|null Returns sanitized `string` if it's a valid IP address, `null` otherwise.
	 */
	public static function sanitizeIpAddress(string $subnet): ?string
	{
		$subnet = str_replace([' ', "\t", ',', '[', ']', '"', "'", '$', '\\'], '', trim($subnet));

		if (strpos($subnet, '/')) {
			[$addr, $cidr] = explode('/', $subnet, 2);
		}
		else {
			$addr = $subnet;
			$cidr = null;
		}

		if (filter_var($addr, FILTER_VALIDATE_IP) !== false && $cidr === null || ctype_digit($cidr ?? '')) {
			$cidr = $cidr === null ? null : max(0, min((int)$cidr, strpos($addr, ':') === false ? 32 : 128));
			return $addr.($cidr === null ? '' : '/'.$cidr);
		}
		else {
			return null;
		}
	}

	/**
	 * @param string[] $subnets IP addresses or subnets
	 * @return string[] Returns an `array` that contains every valid IP address out of `$subnets`.
	 * @see self::sanitizeIpAddress()
	 */
	public static function sanitizeIpAddresses(array $subnets): array
	{
		return array_values(array_filter(array_map([self::class, 'sanitizeIpAddress'], $subnets)));
	}

	/**
	 * @param string[] $items
	 * @param int $limit [Referenz] Ausgangswert wird beim Abarbeiten von `$items` aktualisiert.
	 * @return string[]
	 */
	public static function reduceDomainNamesAndIpAddresses(array $items, int &$limit=2147483647): array
	{
		$domainNames = [];
		$ipAddresses = [];
		foreach ($items as $domainName) {
			$domainName = strtolower(trim(trim($domainName), '.'));
			if (!$domainName) {
				continue;
			}

			// Sonderfall: Klammern um IPv6-Adresse setzen
			if (filter_var($domainName, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6) !== false) {
				$domainName = '['.$domainName.']';
			}

			// Nochmal mit parse_url sicherstellen, dass nur ein Hostname drin steht
			$url = (strpos($domainName, '//') === false) ? 'http://'.$domainName : $domainName;
			$domainName = @trim(parse_url($url, PHP_URL_HOST), '.');

			// IP-Adressen getrennt verarbeiten
			if (filter_var($domainName, FILTER_VALIDATE_IP) !== false) {
				$ipAddresses[$domainName] = $domainName;
				continue;
			}
			if (substr($domainName, 0, 1) == '[' and filter_var(substr($domainName, 1, -1), FILTER_VALIDATE_IP, FILTER_FLAG_IPV6) !== false) {
				$ipAddresses[$domainName] = $domainName;
				continue;
			}

			foreach ($domainNames as $i => $otherName) {
				if ($otherName == $domainName or strpos($domainName, '.'.$otherName) !== false) {
					$domainName = '';
					break;
				}
				elseif (strpos($otherName, '.'.$domainName) !== false) {
					if ($limit >= 0) {
						unset($domainNames[$i]);
						$limit += 1;
					}
				}
			}
			if ($domainName) {
				$limit -= 1;

				if ($limit < 0) {
					continue;
				}

				$domainNames[] = $domainName;
			}
		}

		return array_values(array_merge($domainNames, $ipAddresses));
	}

	/**
	 * @param User $user Subuser
	 * @return array<string, array<string, bool>> [route=>['read'=>bool, 'write'=>bool]]
	 * @phan-side-effect-free
	 */
	public static function getGroupRights(User $user)
	{
		$baseItems = [
			"app_login" => ['read' => true, 'write' => true],
			"app_root" => ['read' => true, 'write' => true],
			"app_language_switch" => ['read' => true, 'write' => true],
		];

		$menRights = $baseItems;

		$owner = $user->getAccountOwner();
		$loggedInUser = User::loggedInUser();
		if (!$loggedInUser or $loggedInUser->getId() !== $owner->getId()) {
			SubGroups::switchRepositoryContext($owner, null);
		}

		try {
			//Das dürfen aber nur die Gruppen des aktuellen USers sein...
			$subgroups = $user->getMemberGroups();

			foreach ($subgroups as $group) {

				foreach ($group->getGroupRights() as $menk=>$menOn) {
					if ((bool)$menOn) {
						// Read-Only oder Read-Write checken
						$readOnly = (substr($menk, -3) === '.ro');
						if ($readOnly) {
							$menk = substr($menk, 0, -3);
						}
						$routePrefix = null;
						// Plugin-Routen
						if (strpos($menk, 'plugin/') === 0) {
							$plugin = PluginState::findByUID(substr($menk, 7));
							$routePrefix = $plugin->getRoutePrefix();
						}
						// Restliche Routen
						else {
							$menItem = MenuData::find($menk);
							if ($menItem) {
								$routePrefix = $menItem->getRoutePrefix();
							}
						}
						if ($routePrefix !== null) {
							// Rechte setzen
							if (!isset($menRights[$routePrefix])) {
								$menRights[$routePrefix] = [];
							}
							$menRights[$routePrefix]['read'] = true;
							if (!$readOnly) {
								$menRights[$routePrefix]['write'] = true;
							}
						}
					}
				}
			}
		}
		finally {
			if (!$loggedInUser or $loggedInUser->getId() !== $owner->getId()) {
				SubGroups::switchRepositoryContext($loggedInUser, null);
			}
		}


		return $menRights;
	}

	/**
	 * @param bool|array<string,bool> $right
	 * @param string $method
	 * @return bool
	 */
	private static function _checkAccessRight($right, $method): bool
	{
		$method = strtoupper($method);
		if (is_array($right)) {
			if ($method === 'GET' or $method === 'HEAD' or $method === 'OPTIONS' or $method === 'PROPFIND') {
				return $right['read'] ?? $right['write'] ?? false;
			} else {
				return $right['write'] ?? false;
			}
		}
		else {
			return (bool)$right;
		}
	}

	/**
	 * Zugriffskontroll-Check
	 *
	 * Gibt zurück, ob der User eine bestimmte Route/einen Menüpunkt nutzen darf.
	 * Nur anwendbar für Routen der Backend-Oberfläche.
	 *
	 * @param User|null $user
	 * @param string $route
	 * @param string $method GET/POST/...
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function mayUserAccessRoute(?User $user, string $route, string $method): bool
	{
		$userId = $user ? $user->getId() : '';

		if (isset(self::$cachedGroupRights[$userId])) {
			/** @var array<string,(bool|array<string,bool>)> $rights */
			$rights = self::$cachedGroupRights[$userId];
		} elseif ($user === null) {
			// Wenn nicht angemeldet -> nur Login/Recovery
			return ($route === 'app_root' or $route === 'app_login' or strpos($route, 'app_login_') === 0);
		} else {
			if ($user->isSubUser()) {
				$filterable = true;

				// Wenn Subuser, dann Gruppenrechte auslesen
				$rights = array_filter(
						self::getGroupRights($user),
						function ($key) { return (strpos($key, 'app_hosting_') !== 0); },
						ARRAY_FILTER_USE_KEY
				);
				$rights['app_domain_selector'] = self::hasMultiDomainSupport();
				$rights['app_root'] = true;
				$rights['app_dashboard'] = true;
				$rights['app_logout'] = true;
			} elseif ($user->isManager() && !$user->hasAllPrivileges()) {
				$filterable = true;

				$rights = array_fill_keys($user->getPrivileges(), true);
				$rights['app_root'] = true;
				$rights['app_account'] = true;
				$rights['app_hosting_dashboard'] = true;
				$rights['app_language_switch'] = true;
				$rights['app_logout'] = true;
				$rights['app_tcf_'] = true;
				$rights['plugin_ccm19_singlesignon_action'] = true;
				$rights['plugin_ccm19_twofactorauth_'] = true;
			} else {
				$filterable = !$user->isAdministrator();

				// auf Owner (client) / admin / manager (allPrivileges) setzen
				$hasMultiDomainAccess = self::hasMultiDomainAccess($user);
				$hasDomainAccess = self::hasDomainAccess($user);
				$hasHostingAccess = self::isExtendedEdition() && ($user->isAdministrator() || $user->isManager());

				$rights = [
					'app_account' => true,
					'app_apikey' => $hasMultiDomainAccess,
					'app_barrierefreiheit' => $hasDomainAccess,
					'app_datenschutz' => $hasDomainAccess,
					'app_declarationtable' => $hasDomainAccess,
					'app_domains_' => $hasMultiDomainAccess,
					'app_domain_selector' => $hasMultiDomainAccess,
					'app_embeddings_' => $hasDomainAccess,
					'app_hosting_' => $hasHostingAccess,
					'app_impressum' => $hasDomainAccess,
					'app_onboarding' => $hasDomainAccess,
					'app_page_check' => $hasDomainAccess,
					'app_placeholder_' => $hasDomainAccess,
					'app_protokoll' => $hasDomainAccess,
					'app_scripts_' => $hasDomainAccess,
					'app_settings_consentstorage' => $hasDomainAccess,
					'app_settings_frontend' => $hasDomainAccess,
					'app_setting_locale_' => $hasDomainAccess,
					'app_sub_group_' => $hasMultiDomainAccess,
					'app_tcf_xhr_' => $hasDomainAccess || $hasHostingAccess,
					'app_tcf_vendorlist' => $hasDomainAccess || $hasHostingAccess,
					'app_tcf_' => $hasDomainAccess || $hasHostingAccess,
					'app_theme_' => $hasDomainAccess,
					'app_user_locale_' => $hasMultiDomainAccess,
					'app_user_embeddings_' => $hasMultiDomainAccess,
					'app_user_theme_' => $hasMultiDomainAccess,
					'app_user_user_' => $hasMultiDomainAccess,
					'app_' => true,
					'plugin_' => true,
					'meta__' => true,
				];

				// Alle Plugins für nicht-Subuser grundsätzlich erlauben (wird ggf. durch Feature-Check überschrieben)
				foreach (PluginState::all() as $plugin) {
					$rights[$plugin->getRoutePrefix()] = true;
				}
			}

			//Spezialfall page_check / Permanent Crawler inaktiv? Dann Menüpunkt deaktivieren
			//Macht auch Sinn, wenn die Instanz keinen Crawler hat, aber trotzdem deaktiviert ist. Dann muss da auch nichts angezeigt werden
			$domain = Domain::activeDomain();
			if($domain!==null && $domain->isActivatePermaScan()===false)
			{
				$rights['app_page_check'] = false;
			}

			/**
 			 * Die Zugriffe nochmal überarbeiten in Hinsicht auf die Settings aus den Account-Settings
			 */
			if (self::hasDomainAccess($user)) {
				$rights = self::filterAccessRightsAccount($rights, $user);
			}

			//Hier Spezialcheck... für die Versionen - ist der Zugriff auf bestimmte Menüpunkte erlaubt oder nicht...
			$featureRoutes = [
				'canUseDomAPI' => [
					'app_apikey',
				],
				'canUseIAB' => [
					'app_tcf_',
				],
				'canUseUserGroupRights' => [
					'app_sub_group_',
					'app_user_user_',
				],
			];

			if (self::isExtendedEdition() and $user->isClient()) {
				// Plugins als Account-Features eintragen
				foreach (PluginState::all() as $plugin) {
					$featureRoutes[$plugin->getUserRightsKey()] = [ $plugin->getRoutePrefix() ];
				}
			}

			// Prüfe grundsätzlichen Zugriff ($rights) plus individuelle Account-Features
			// Plugins pro Haupt-User werden hier auch verwaltet
			array_walk($featureRoutes, function ($routes, $feature) use ($user, &$rights) {
				foreach ($routes as $route) {
					// Fehlender grundsätzlicher Zugriff kann nicht durch das Account-Feature überschrieben werden.
					// Bei undefiniertem Account-Feature wird nichts geändert, als Fallback für alle Editionen.
					if ($user->isAccountFeatureConfigured($feature) and !$user->hasAccountFeature($feature)) {
						$rights[$route] = false;
					}
					elseif (strpos($feature, 'canUsePlugin/') === 0 and !$user->hasAccountFeature($feature)) {
						$rights[$route] = false;
					}
				}
			});

            //Hier nochmal überschreiben falls der Zugriff durch Tarifrechte gesetzt wird
            if (self::isExtendedEdition() and $user->isClient()) {
                // Plugins als Account-Features eintragen
                foreach (PluginState::all() as $plugin) {
                    //Hier checken, ob der Tarif des Users freigeschaltet ist für den Zugriff
                    if($plugin->checkFullAccessPerCount($user))
					{
                        $rights[$plugin->getRoutePrefix()] = true;
                    }
                    //und hier checken ob Zugriff für den Testzeitraum besteht
                    if($plugin->checkTestAccessPerCount($user))
					{
                        $rights[$plugin->getRoutePrefix()] = true;
                    }
                }
            }

			if ($filterable) {
				$rights = self::filterWithEvent(self::EVENT_FILTER_USER_RIGHTS, ['rights'=>$rights, 'user'=>$user])['rights'];
			}

			self::$cachedGroupRights[$userId] = $rights;
		}

		// Prüfen, ob exakt diese Route mit Rechten belegt ist
		if (isset($rights[$route])) {
			return self::_checkAccessRight($rights[$route], $method);
		}

		// Nach Unterstrichen die übergeordneten Bereiche durchgehen, bis einer
		// Rechte hat (app_domains_index → app_domains_, app_)
		while (strlen($route) > 1 && ($i = strrpos($route, '_', -2)) !== false) {
			$route = substr($route, 0, $i+1);
			if (isset($rights[$route])) {
				return self::_checkAccessRight($rights[$route], $method);
			}
			if (isset($rights[substr($route, 0, -1)])) {
				return self::_checkAccessRight($rights[substr($route, 0, -1)], $method);
			}
		}
		return false;
	}

    /**
     * Account Settings Rechte überprüfen für den jeweiligen Menüpunkt
     * @param array<string,bool> $rights
     * @param User $user
     * @return array<string,bool>
     */
    public static function filterAccessRightsAccount($rights, User $user)
    {
        if(($rights['app_user_theme_']  ?? false) === true)
        {
            $rights['app_user_theme_'] = $user->getActivateCrossDomainThemes();
        }
        if( ($rights['app_user_locale_'] ?? false) === true)
        {
            $rights['app_user_locale_'] = $user->getActivateCrossDomainLocales();
        }
        if(($rights['app_apikey']  ?? false) === true)
        {
            $rights['app_apikey'] = $user->getActivateCrossDomainGeneralAPI();
        }
        if(($rights['app_sub_group_'] ?? false) === true)
        {
            $rights['app_sub_group_'] = $user->getActivateCrossDomainUserGroup();
        }
        if(($rights['app_user_user_'] ?? false) === true)
        {
            $rights['app_user_user_'] = $user->getActivateCrossDomainUserGroup();
        }
        if(($rights['app_tcf_'] ?? false) === true)
        {
            $rights['app_tcf_'] = $user->getActivateCrossDomainIab();
        }
        if(($rights['app_barrierefreiheit'] ?? false) === true)
        {
            $rights['app_barrierefreiheit'] = $user->getActivateCrossDomainAccessibility();
        }

        return $rights;
    }

	/**
	 * Kodiert den 'backTo'-Parameter für den Rücksprung nach einem Login
	 * @param string $route
	 * @param array<string,mixed> $routeParams
	 */
	public static function encodeReturnTo(string $route, array $routeParams = []): string
	{
		return strtr(base64_encode(json_encode([$route, $routeParams])), ['+'=>'-', '/'=>'_', '='=>'']);
	}

	/**
	 * @param string $routeName
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getDocUrl($routeName)
	{
		//print_r($routeName);
		$arrayLinks = array(
			"app_domains_index"			=>"/verwaltung/domainverwaltung/",
			"app_user_theme_index" 		=> "/agency-version/themes/",
			# TODO "app_user_locale_list" 		=> "/agency-version/…/",
			"app_theme_edit" 			=> "/funktionen/themes/",
			"app_theme_new" 			=> "/funktionen/themes/",
			"app_account"				=>"/verwaltung/kontoeinstellungen/",
			"app_dashboard" 			=> "/funktionen/dashboard/",
			"app_theme_list" 			=> "/funktionen/themes/",
			"app_embeddings_list" 		=> "/funktionen/cookies-und-andere/",
			"app_embeddings_edit" 		=> "/funktionen/cookies-und-andere/",
			"app_embeddings_new" 		=> "/funktionen/cookies-und-andere/",
			"app_embeddings_from_db"	=> "/funktionen/cookies-und-andere/",
			"app_scripts_list"	 		=> "/system-und-co/dev-einstellungen/",
			"app_setting_locale_list"	=> "/funktionen/uebersetzungen/",
			"app_setting_locale_edit"   => "/funktionen/uebersetzungen/",
			"app_impressum"				=> "/funktionen/impressum/",
			"app_datenschutz"			=> "/funktionen/datenschutz/",
			"app_placeholder_list"		=> "/funktionen/platzhalter/",
			"app_protokoll"				=> "/logs/consent-protokoll/",
			"app_settings_frontend"		=> "/system-und-co/frontend-einstellungen/",
			"app_settings_consentstorage"=> "/system-und-co/consent-speicherung/",
			"app_page_check"			=> "/funktionen/pagecheck/",
			"app_iframes_list"			=> "/funktionen/iframes/",
			"app_declarationtable"		=> "/funktionen/cookietabelle/",
			"app_barrierefreiheit"		=> "/funktionen/barrierefreiheit/",
			"plugin_ccm19_statistics_index" => "/plugins/erweiterte_statistik/",
			"plugin_ccm19_abtests_index" => "/plugins/abtests/",
			"app_ccm19payment_admin" 		=> "/verwaltung/tarife/",
			"app_user_locale_list" 		=> "/verwaltung/standardtexte/",
			"app_apikey"				=> "/verwaltung/domainapi/",
			"app_license_manager_index"  => "/verwaltung/lizenzverwaltung/",
			"app_systemlog"				=> "/logs/system-protokoll/",
			"plugin_ccm19_imex_multi_index_get" => "/plugins/importexport/",
			"app_setting_email"=> "/system-und-co/email/",
			"app_settings_cache" => "/system-und-co/caching/",
			"app_settings_network" => "/system-und-co/netzwerk/",
			"app_plugins"=> "/system-und-co/verfuegbare-plugins/",
			"app_update"=> "/system-und-co/update/"

		);

		if (!empty($arrayLinks[$routeName]))
		{
			$link = $arrayLinks[$routeName];
		}
		else{
			$link = "/";
		}

		return  $link;
	}

	/**
	 * @return string[]
	 * @phan-side-effect-free
	 */
	public static function getBackgroundImage(): array
	{
		$backgroundImages = [
			array(
				"src" => "img/yonatan-anugerah-3k0-K12TgLc-unsplash1.jpg",
				"unsplashHref" =>"https://unsplash.com/@yonatanugerah",
				"unsplashName" => "yonatan anugerah",
				"headlineColor" => "#333",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => "0.8",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/vincent-guth-Sdc35GnHgvI-unsplash1.jpg",
				"unsplashHref" =>"https://unsplash.com/@vingtcent",
				"unsplashName" => "Vincent Guth",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#fff",
				"unsplashLinkBackgroundColor" => "none",
				"unsplashLinkBackgroundColorOpacity" => "1",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/marc-schulte-3z8yZYTyM6w-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@marc_schulte",
				"unsplashName" => "Marc Schulte",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".7",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/ce-kirsche.jpg",
				"unsplashHref" =>"https://unsplash.com/@carsten_euwens",
				"unsplashName" => "Carsten Euwens",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".7",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/sven-van-der-pluijm-W7EgqYJdQOE-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@svenson",
				"unsplashName" => "Sven van der Pluijm",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".7",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			)
			,array(
				"src" => "img/pietro-de-grandi-T7K4aEPoGGk-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@peter_mc_greats",
				"unsplashName" => "Pietro De Grandi",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".7",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/mateo-giraud-wtBex4wQw60-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@mateo_giraud",
				"unsplashName" => "Mateo Giraud",
				"headlineColor" => "#000",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/dillon-austin-xI-57L4_P1U-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@dillonaustin",
				"unsplashName" => "Dillon Austin ",
				"headlineColor" => "#000",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/denys-nevozhai-duo-xV0TU7s-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@dnevozhai",
				"unsplashName" => "Denys Nevozhai",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/bruno-kelzer-rhod8LQEeic-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@bruno_kelzer",
				"unsplashName" => "Bruno Kelzer",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/george-desipris-WpLm6hix8i8-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@desipris",
				"unsplashName" => "GEORGE DESIPRIS",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/lukasz-szmigiel-ps2daRcXYes-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@szmigieldesign",
				"unsplashName" => "Lukasz Szmigiel",
				"headlineColor" => "#000",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/aaron-burden-ejUU3U7Rlns-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/@aaronburden",
				"unsplashName" => "Aaron Burden",
				"headlineColor" => "#000",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/nick-fewings-IDU1v-rDmQ4-unsplashs.jpg",
				"unsplashHref" =>"https://unsplash.com/@jannerboy62",
				"unsplashName" => "Nick Fewings",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".4",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/john-ehrlich-45G8U6cx5sM-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/de/@tallpines",
				"unsplashName" => "john ehrlich",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/luca-baggio-eKU3JGNCCMg-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/de/@luca42",
				"unsplashName" => "Luca Baggio",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			),
			array(
				"src" => "img/freestocks-ZArDeAtxj0Q-unsplash.jpg",
				"unsplashHref" =>"https://unsplash.com/de/@freestocks",
				"unsplashName" => "freestocks",
				"headlineColor" => "#fff",
				"unsplashLinkColor" => "#000",
				"unsplashLinkBackgroundColor" => "#fff",
				"unsplashLinkBackgroundColorOpacity" => ".6",
				"portalName" =>"Unsplash",
				"portalURL" => "https://unsplash.com/",
			)
		];

		return $backgroundImages[16];

	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function generateUniqueCookieId(): string
	{
		// alternativ bin2hex(random_bytes(n))
		return hash('sha256', uniqid(microtime(), true));
	}

	/**
	 * @param string $ucid
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function validateUniqueCookieId(string $ucid): bool
	{
		// Match either legacy SHA-256 hash or new client-generated 64-bit FNV-1a hash
		return (bool)preg_match('~^([0-9a-f]{64}|[0-9a-f]{16})$~', $ucid);
	}

	/**
	 * @param string $value
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function validateEmailAddress(string $value): bool
	{
		// Kleine Schreibweise der E-Mail-Adresse validieren
		$value = mb_strtolower($value);

		// General Email Regex (RFC 5322 Official Standard) - https://emailregex.com/
		$pattern = '/^(?:[a-z0-9!#$%&\'*+\/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&\'*+\/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])$/';

		return (bool)preg_match($pattern, $value);
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function hostingInterface(): bool
	{
		return self::isExtendedEdition() && ($user = User::loggedInUser()) && ($user->isAdministrator() || $user->isManager());
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getEdition(): string
	{
		return (string)($_ENV['APP_EDITION'] ?? '');
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isExtendedEdition(): bool
	{
		return self::getEdition() == 'extended';
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isMultiDomainEdition(): bool
	{
		return self::getEdition() == 'multi';
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isBaseEdition(): bool
	{
		$isAnyOtherEdition = self::isExtendedEdition() || self::isMultiDomainEdition();

		return self::getEdition() == 'base' || $isAnyOtherEdition == false;
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function hasMultiDomainSupport(): bool
	{
		return self::isExtendedEdition()
			|| self::isMultiDomainEdition()
		;
	}

	/**
	 * @param User $user
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function hasMultiDomainAccess(User $user): bool
	{
		return self::hasMultiDomainSupport() && self::hasDomainAccess($user);
	}

	/**
	 * @param User $user
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function hasDomainAccess(User $user): bool
	{
		return (self::isExtendedEdition() xor $user->isAdministrator())
			&& !$user->isManager() // Managers don't have access to domains in the extended edition
		;
	}

	/**
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isAutoUpdateEnabled(): bool
	{
		return (bool)Config::getInstance()->get('enableAutoUpdate');
	}

	/**
	 * @param string $path
	 * @throws RuntimeException Wenn das angegebene Verzeichnis nicht existiert und nicht angelegt werden konnte.
	 */
	public static function assertDirectoryWritable(string $path): void
	{
		if (is_dir($path) == false && @mkdir($path, 0777, true) == false || is_writable($path) == false) {
			throw new RuntimeException('Cannot access directory '.$path.', check permissions.');
		}
	}

	/**
	 * @return Kernel
	 * @throws LogicException if the kernel hasn't been initialized. This should only ever happen during development.
	 */
	public static function getKernel(): Kernel
	{
		global $kernel;
		if (!$kernel instanceof Kernel) {
			throw new LogicException('Kernel not initialized');
		}
		return $kernel;
	}

	/**
	 * Gets a service.
	 * @param string $id The service identifier
	 * @return object|null
	 */
	public static function getService(string $id): ?object
	{
		return self::getKernel()->getContainer()->get($id);
	}

	/**
	 * Gibt den RequestStack zurück.
	 * @return RequestStack
	 */
	public static function getRequestStack(): RequestStack
	{
		/** @var RequestStack $requestStack */
		$requestStack = self::getService('request_stack');
		assert($requestStack instanceof RequestStack);
		return $requestStack;
	}

	/**
	 * Gibt den aktuellen Request zurück.
	 * @return Request|null
	 */
	public static function getCurrentRequest(): ?Request
	{
		return self::getRequestStack()->getCurrentRequest();
	}

	/**
	 * Gibt den äußersten Request zurück.
	 * @return Request|null
	 */
	public static function getMainRequest(): ?Request
	{
		return self::getRequestStack()->getMainRequest();
	}

	/**
	 * Gibt die aktuelle Session zurück.
	 * @return Session
	 * @throws SessionNotFoundException if no session is currently available.
	 */
	public static function getSession(): Session
	{
		$requestStack = self::getRequestStack();

		$session = $requestStack->getCurrentRequest()
			? $requestStack->getCurrentRequest()->getSession()
			: $requestStack->getSession();

		assert($session instanceof Session);
		return $session;
	}

	/**
	 * Liefert den Router-Dienst.
	 * @return Router
	 * @phan-side-effect-free
	 */
	public static function getRouter(): Router
	{
		$router = self::getService('router');
		assert($router instanceof Router);
		return $router;
	}

	/**
	 * Liefert das Übersetzungsmodul.
	 * @return TranslatorInterface
	 * @phan-side-effect-free
	 */
	public static function getTranslator(): TranslatorInterface
	{
		$translator = self::getService('translator');
		assert($translator instanceof TranslatorInterface);
		return $translator;
	}

	/**
	 * Helper function that can add flash messages outside of controllers.
	 * @param string $type
	 * @param string $message
	 * @return void
	 */
	public static function addFlash(string $type, string $message): void
	{
		self::getSession()->getFlashBag()->add($type, $message);
	}

	/**
	 * Übernimmt die übergebenen Daten in die .env.local
	 * @param array<string,string|int|float|null> $data
	 * @return void
	 */
	public static function setLocalEnvironment(array $data): void
	{
		$path = static::getBaseDir() . DIRECTORY_SEPARATOR . '.env.local';

		// Daten aus bestehender Datei holen
		if (file_exists($path)) {
			$dotenv = new Dotenv();
			$existing_items = $dotenv->parse(file_get_contents($path), '.env.local');
			$data = array_filter(array_merge($existing_items, $data), function ($value) { return $value !== null; });
		}

		// Zeilen für neue Datei erstellen
		$items = array_map(function ($key, $value) {
			$key = preg_replace('/[^a-zA-Z0-9_]/', 'X', $key);
			if (! preg_match('/^[a-zA-Z0-9,._-]+$/', $value)) {
				$value = escapeshellarg($value);
			}
			return "$key=$value";
		}, array_keys($data), array_values($data));

		// Datei atomar schreiben
		$result = file_put_contents($path.'.tmp', implode("\n", $items)."\n", LOCK_EX);
		if ($result !== false) {
			rename($path.'.tmp', $path);
		}
	}

	/**
	 * Basis-Verzeichnis der CCM19-Installation
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getBaseDir(): string
	{
		return self::getKernel()->getProjectDir();
	}

	/**
	 * Verzeichnispfad für variable (und ggf. temporäre) Daten
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getVarDir(): string
	{
		return $_ENV['APP_OVERRIDE_VAR_DIR'] ?? static::getBaseDir() . DIRECTORY_SEPARATOR . 'var';
	}

	public static function buildRelativePath(string $path, string $base): string
	{
		if (strpos($path, $base.DIRECTORY_SEPARATOR) === 0) {
			return ltrim(substr($path, strlen($base) + 1), '/'.DIRECTORY_SEPARATOR);
		}
		if (strpos($path, '/') !== 0 and strpos($path, DIRECTORY_SEPARATOR) !== 0) {
			return $path;
		}

		$path = str_replace('/', DIRECTORY_SEPARATOR, $path);
		$base = str_replace('/', DIRECTORY_SEPARATOR, $base);
		if (strpos($path, $base.DIRECTORY_SEPARATOR) === 0) {
			return ltrim(substr($path, strlen($base) + 1), DIRECTORY_SEPARATOR);
		}

		$pathComponents = explode(DIRECTORY_SEPARATOR, trim($path, DIRECTORY_SEPARATOR));
		$baseComponents = explode(DIRECTORY_SEPARATOR, trim($base, DIRECTORY_SEPARATOR));

		if ($pathComponents[0] !== $baseComponents[0]) {
			// Aufgeben: kein gemeinsamer Pfadbestandteil -> absoluten Pfad liefern
			return $path;
		}

		// Gemeinsamen Präfix abschneiden
		while ($pathComponents) {
			if ($pathComponents[0] === ($baseComponents[0] ?? '')) {
				array_shift($pathComponents);
				array_shift($baseComponents);
			}
			elseif ($pathComponents[0] === '') {
				array_shift($pathComponents);
			}
			elseif ($baseComponents[0] === '') {
				array_shift($baseComponents);
			}
			else {
				break;
			}
		}

		// Resultierenden relativen Pfad konstruieren
		$result = array_merge(array_fill(0, count($baseComponents), '..'), $pathComponents);
		return implode(DIRECTORY_SEPARATOR, $result);
	}

	/**
	 * Verzeichnispfad für variable Daten, die ggf. von mehreren Instanzen in einem Cluster geteilt werden.
	 *
	 * Im Normalfall gleiche Pfad wie bei `Utils::getVarDir()`. Bei Installations-Varianten mit Cluster-Dateisystem
	 * kann es auch `var/config-bundle` sein.
	 *
	 * @return string Pfad
	 * @phan-side-effect-free
	 */
	public static function getBundledVarDir(): string
	{
		$subdir = $_ENV['APP_BUNDLE_VAR'] ?? '';
		if ($subdir === '1') {
			$subdir = 'config-bundle';
		}
		if ($subdir != '') {
			return self::getVarDir() . DIRECTORY_SEPARATOR . $subdir;
		} else {
			return self::getVarDir();
		}
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getShareDir(): string
	{
		return static::getBaseDir() . DIRECTORY_SEPARATOR . 'share';
	}

	/**
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getCacheDir(): string
	{
		return self::getKernel()->getCacheDir();
	}

	/**
	 * Diese Methode löscht ein komplettes Verzeichnis. Dabei werden sämtliche Dateien und Ordner rekursiv entfernt.
	 *
	 * Es wird nicht rekursiv durch symbolische Links auf Verzeichnisse entfernt.
	 * With great power comes great responsibility.
	 *
	 * @param string $path Pfad zum Verzeichnis
	 */
	public static function removeDirectory(string $path): void
	{
		self::clearDirectory($path);

		@rmdir($path);
	}

	/**
	 * Diese Methode löscht die Inhalte eines kompletten Verzeichnisses.
	 *
	 * Dabei werden sämtliche Dateien und Unterordner rekursiv entfernt, aber nicht das Verzeichnis selbst.
	 *
	 * Es wird nicht rekursiv durch symbolische Links auf Verzeichnisse entfernt.
	 * With great power comes great responsibility.
	 *
	 * @param string $path Pfad zum Verzeichnis
	 */
	public static function clearDirectory(string $path): void
	{
		try {
			/** @var DirectoryIterator $it */
			$it = new DirectoryIterator($path);
		}
		catch (RuntimeException $exception) {
			return;
		}

		foreach (new DirectoryIterator($path) as $it) {
			if ($it->isDot()) {
				continue;
			}
			elseif (!$it->isLink() and $it->isDir()) {
				self::removeDirectory($it->getPathname());
			}
			else {
				@unlink($it->getPathname());
			}
		}
	}

	/**
	 * Diese Methode kopiert ein komplettes Verzeichnis...
	 * Ref: https://www.php.net/manual/de/function.copy.php#91010
	 * @param string $src
	 * @param string $dst
	 */
	public static function copyDirectory(string $src, string $dst):void
	{
		$dir = opendir($src);
		@mkdir($dst);
		while(false !== ( $file = readdir($dir)) ) {
			if (( $file != '.' ) && ( $file != '..' )) {
				if ( is_dir($src . '/' . $file) ) {
					self::copyDirectory($src . '/' . $file,$dst . '/' . $file);
				}
				else {
					copy($src . '/' . $file,$dst . '/' . $file);
				}
			}
		}
		closedir($dir);
	}

	/**
	 * Diese Methode gibt die aktuelle Versions-ID zurück.
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getVersionId(): string
	{
		$filename = self::getBaseDir().'/config/version-id.txt';
		return trim(@file_get_contents($filename));
	}

	/**
	 * Diese Methode gibt die aktuelle menschenlesbare Versionsnummer zurück.
	 * @return string
	 * @phan-side-effect-free
	 */
	public static function getVersionName(): string
	{
		$filename = self::getBaseDir().'/config/version-name.txt';
		return trim(@file_get_contents($filename));
	}

	/**
	 * Diese Methode gibt die Generation der aktuellen Version zurück.
	 * @return int
	 * @phan-side-effect-free
	 */
	public static function getVersionGeneration(): int
	{
		$filename = self::getBaseDir().'/config/version-generation.txt';
		return (int)(trim(@file_get_contents($filename) ?: '1' ));
	}

	/**
	 * Diese Methode gibt den aktiven Cron-Modus zurück.
	 *
	 * 'frontend': Automatisch von Frontend-Requests ausgelöst.
	 * 'server': Von einem "echten" Cronjob/Timer auf dem Server ausgelöst.
	 *
	 * @return string Utils::CRON_MODE_FRONTEND|Utils::CRON_MODE_SERVER
	 * @phan-side-effect-free
	 */
	public static function getCronMode(): string
	{
		return (string)($_ENV['APP_CRON_MODE'] ?? self::CRON_MODE_FRONTEND);
	}

	/**
	 * Setzt bestehende Cache-Control-Header zurück und verhindert, dass
	 * der SessionListener die Cache-Control überschreibt.
	 * @param ?Response $response
	 * @return void
	 */
	public static function resetCacheControl(?Response $response): void
	{
		if (headers_sent() == false) {
			// Setzen des Headers in AbstractSessionHandler::open() annullieren
			header_remove('Cache-Control');
		}
		if ($response) {
			// Ggf. schon gesetztes Cache-Control aus $response entfernen
			$response->headers->remove('Cache-Control');
			// Setzen des Headers in AbstractSessionListener::onKernelResponse() annullieren
			$response->headers->set(SessionListener::NO_AUTO_CACHE_CONTROL_HEADER, '1');
		}
	}

	/**
	 * Prüft, ob der Request höchstwahrscheinlich von einem Bot stammt.
	 * @param Request $request
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isBotRequest(Request $request): bool
	{
		$useragent = $request->headers->get('User-Agent', '');
		if (preg_match('/(bot\b|BOT\b|Bot\b|BingPreview| \(\+http)/', $useragent)) {
			return true;
		}
		return false;
	}

	/**
	 * Gibt eine Liste aller IPv4/IPv6-Adressen zu einem Hostnamen zurück
	 * @param string $host
	 * @return string[]
	 * @phan-side-effect-free
	 */
	public static function getIPAddresses(string $host): array
	{
		$addresses = [];

		// Wenn Socket-Extension vorhanden, verwenden.
		if (function_exists('socket_addrinfo_lookup')) {
			foreach (socket_addrinfo_lookup($host, 'https') ?: [] as $addr) {
				$addrinfo = (socket_addrinfo_explain($addr));
				if (!empty($addrinfo['ai_addr']['sin_addr'])) {
					$addresses[] = $addrinfo['ai_addr']['sin_addr'];
				}
				elseif (!empty($addrinfo['ai_addr']['sin6_addr'])) {
					$addresses[] = $addrinfo['ai_addr']['sin6_addr'];
				}
			}
		}
		// Ansonsten mit gethostbynamel und dns_get_record behelfen
		else {
			$addresses = array_merge(
				gethostbynamel($host) ?: [],
				array_map(function ($record)  {
					return $record['ipv6'];
				},
				dns_get_record($host, DNS_AAAA) ?: [])
			);
		}

		$addresses = array_unique($addresses);

		return $addresses;
	}

	/**
	 * Prüft ob der Hostname eine öffentliche IP-Adresse zugeordnet hat
	 * @param string $host
	 * @return bool
	 * @phan-side-effect-free
	 */
	public static function isPublicHost(string $host): bool
	{
		$subnets = ['127.0.0.0/8', '10.0.0.0/8', '100.64.0.0/10', '172.16.0.0/12', '192.168.0.0/16', '::1/128', 'fc00::/7', 'fd00::/8'];

		$ip = filter_var($host, FILTER_VALIDATE_IP);
		if ($ip) {
			$addresses = [$ip];
		} else {
			$addresses = self::getIPAddresses($host);
		}

		foreach ($addresses as $addr) {
			foreach ($subnets as $subnet) {
				if (IpUtils::checkIp($addr, $subnet)) {
					return false;
				}
			}
		}

		return (bool)$addresses;
	}

	/**
	 * Gibt die volle URL für eine Route aus (Standard: Root/Startseite)
	 *
	 * @param string $route
	 * @param array<string,mixed> $parameters
	 * @throws \Symfony\Component\Routing\Exception\RouteNotFoundException
	 * @throws \Symfony\Component\Routing\Exception\MissingMandatoryParametersException
	 * @phan-side-effect-free
	 */
	public static function getFullUrl(string $route='app_root', array $parameters=[]): string
	{
		return self::getRouter()->generate($route, $parameters, UrlGeneratorInterface::ABSOLUTE_URL);
	}

	/**
	 * @param string $name
	 * @param Event|null $event
	 */
	public static function dispatchEvent(string $name, ?Event $event = null): void
	{
		/** @var EventDispatcherInterface $dispatcher */
		$dispatcher = self::getService('event_dispatcher');
		assert($dispatcher instanceof EventDispatcherInterface);
		$dispatcher->dispatch($event ?? new Event(), $name);
	}

	/**
	 * @template T
	 * @param string $name
	 * @param FilterEvent<T>|T $data
	 * @return T
	 */
	public static function filterWithEvent(string $name, $data)
	{
		if ($data instanceof FilterEvent) {
			$event = $data;
		} else {
			$event = new FilterEvent($data);
		}
		self::dispatchEvent($name, $event);
		return $event->getReturn();
	}

	/**
	 * @param string $name
	 * @param mixed $subject
	 * @param array $arguments
	 */
	public static function notifyWithEvent(string $name, $subject, array $arguments=[]): void
	{
		self::dispatchEvent($name, new GenericEvent($subject, $arguments));
	}

	/**
	 * Gibt die Domain zurück, von der der auf diese URL verwiesen wurde, soweit bestimmbar.
	 *
	 * Verwendet den Origin-Header, wenn vorhanden, ansonsten Referer.
	 *
	 * @param Request $request
	 * @param bool $removeWWW www. entfernen
	 * @return string|null
	 * @phan-side-effect-free
	 */
	public static function getOriginDomain(Request $request, bool $removeWWW=true): ?string
	{
		$url = $request->headers->get('Origin');
		if (!$url) {
			$url = $request->headers->get('Referer');
		}
		if (!$url) {
			return null;
		}
		$host = parse_url(trim($url), PHP_URL_HOST);
		if ($host) {
			if ($removeWWW) {
				$host = preg_replace('/^www[0-9]?\./', '', $host);
			}
			return strtolower(rtrim($host, '.'));
		}
		else {
			return null;
		}
	}

	/**
	 * Eine Lebenszeit-Angabe ("1 Tag" usw.) parsen und aufgerundet in Stunden zurückgeben.
	 *
	 * $sessionFallback wird zurückgegeben, wenn "Sitzung"/"Session" statt einer Zeitangabe
	 * verwendet wird und wenn die Zeit nicht eindeutig bestimmbar ist, $fallback.
	 *
	 * @param string $lifetimeString
	 * @param mixed $sessionFallback
	 * @param mixed $fallback
	 * @return int|mixed
	 * @phan-side-effect-free
	 */
	public static function parseLifetimeToHours(string $lifetimeString, $sessionFallback=null, $fallback=null)
	{
		$lifetime = null;
		$match = [];

		preg_match('/^\s*(?<number>\.[0-9]+|[0-9]+(?:\.[0-9*]*)?)\s*(?<unit>[^\s\d.]*)/ui', $lifetimeString, $match);

		if ($match and $match['number'] != '') {
			$number = (float)$match['number'];
			switch (strtolower($match['unit'])) {
				case 's':
				case 'sek':
				case 'sec':
				case 'second':
				case 'seconds':
				case 'sekunde':
				case 'sekunden':
					$lifetime = $number / 3600;
					break;
				case 'm':
				case 'min':
				case 'minute':
				case 'minutes':
				case 'minuten':
					$lifetime = $number / 60;
					break;
				case 'h':
				case 'hour':
				case 'hours':
				case 'std':
				case 'st':
				case 'stdn':
				case 'stunde':
				case 'stunden':
					$lifetime = $number;
					break;
				case 'd':
				case 'day':
				case 'days':
				case 't':
				case 'tag':
				case 'tage':
					$lifetime = $number*24;
					break;
				case 'w':
				case 'week':
				case 'weeks':
				case 'woche':
				case 'wochen':
					$lifetime = $number*24*7;
					break;
				case 'mon':
				case 'month':
				case 'months':
				case 'monat':
				case 'monate':
					$lifetime = $number*24*30.5;
					break;
				case 'y':
				case 'year':
				case 'years':
				case 'jahr':
				case 'jahre':
					$lifetime = $number*24*365;
					break;
				case 'dec':
				case 'decade':
				case 'decades':
				case 'jahrzehnt':
				case 'jahrzehnte':
					$lifetime = $number*24*365*10;
					break;
			}
		}

		if ($lifetime === null) {
			if (preg_match('/(sitzung|session)/i', $lifetimeString)) {
				return $sessionFallback;
			} else {
				return $fallback;
			}
		} else {
			return (int)ceil($lifetime);
		}
	}

	/**
	 * Liefert TRUE, wenn die MongoDB-PHP-Extension und die MongoDB-Library installiert sind.
	 * @phan-side-effect-free
	 */
	public static function mongoDbSupportAvailable(): bool
	{
		return (class_exists('\MongoDB\Driver\Manager') and class_exists('\MongoDB\Client'));
	}

	/**
	 * @template T
	 * @param array<T>|\Traversable<T> $iterable
	 * @param callable(T):bool|null $test
	 * @return bool
	 * @phan-suppress PhanTemplateTypeNotUsedInFunctionReturn
	 */
	public static function all(iterable $iterable, ?callable $test=null)
	{
		if ($test === null) {
			foreach ($iterable as $value) {
				if (!$value) {
					return false;
				}
			}
		}
		else {
			foreach ($iterable as $value) {
				if ($test($value) == false) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * @template T
	 * @param array<T>|\Traversable<T> $iterable
	 * @param callable(T):bool|null $test
	 * @return bool
	 * @phan-suppress PhanTemplateTypeNotUsedInFunctionReturn
	 */
	public static function any(iterable $iterable, ?callable $test=null)
	{
		if ($test === null) {
			foreach ($iterable as $value) {
				if ($value) {
					return true;
				}
			}
		}
		else {
			foreach ($iterable as $value) {
				if ($test($value)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Teilt die Ergebnisse eines Iterators in Array-Blöcke einer bestimmten Größe
	 *
	 * z.B. Utils::chunk(3, [1,2,3,4,5,6,7]) -> [1,2,3], [4,5,6], [7]
	 *
	 * @template T
	 * @param int $chunkSize
	 * @param iterable<T> $iterable
	 * @return \Iterator<T[]>
	 */
	public static function chunk(int $chunkSize, iterable $iterable)
	{
		$chunk = [];
		foreach ($iterable as $item) {
			$chunk[] = $item;
			if (count($chunk) >= $chunkSize) {
				yield $chunk;
				$chunk = [];
			}
		}
		if ($chunk) {
			yield $chunk;
		}
	}

	/**
	 * This method is more or less an implementation of `array_merge_recursive` with a few key differences:
	 *
	 * - Numeric keys are merged properly, just as the `+` operator would do.
	 * - If a key already exists, the left-hand side value is preserved. (as the `+` operator does)
	 * - If the right value is an array and the left value isn't, the array replaces the left value.
	 *
	 * @param array $target
	 * @param array $base
	 * @return array Returns an array that contains all keys from $target and $base mirroring multidimensional entries as necessary.
	 */
	public static function complementArray(array $target, array $base): array
	{
		$target += array_diff_key($base, $target);

		foreach ($base as $key => $value) {
			if (is_array($value)) {
				$target[$key] = self::complementArray(is_array($target[$key]) ? $target[$key] : [], $value);
			}
		}

		return $target;
	}

	/**
	 * This method implements a sorting algorithm that considers both numeric and natural order.
	 * If two values are numeric (hexadecimal or decimal), they are compared using numeric order.
	 * Otherwise, they are compared as strings using natural order.
	 * Keys are preserved.
	 * @param array $array
	 * @param string|int|null $sortColumn The column to sort by. If null, the array is sorted by values.
	 * @param string $sortDirection 'asc' or 'desc'
	 * @param bool $caseSensitive
	 * @return array
	 */
	public static function sortNatNum(array $array, $sortColumn=null, string $sortDirection='asc', bool $caseSensitive=false): array
	{
		uasort($array, static function ($a, $b) use ($caseSensitive, $sortColumn, $sortDirection) {
			if ($sortColumn !== null) {
				$a = $a[$sortColumn] ?? '';
				$b = $b[$sortColumn] ?? '';
			}

			// Prepare values for comparison
			$a = $caseSensitive ? (string)$a : mb_strtolower((string)$a);
			$b = $caseSensitive ? (string)$b : mb_strtolower((string)$b);

			// Treat hexadecimal values as numbers
			if (ctype_xdigit($a) && ctype_xdigit($b)) {
				$result = $a <=> $b;
			}
			else {
				// Use strnatcmp instead of strnatcasecmp because the values are already normalized
				$result = strnatcmp($a, $b);
			}

			return $sortDirection === 'asc' ? $result : -$result;
		});

		return $array;
	}

	/**
	 * Parsen der CSP-Origins aus CSS-Code
	 *
	 * @param string $cssCode CSS-Code
	 * @return string[] Alle externen Origins, die im CSS referenziert werden
	 */
	public static function extractCSPOrigins(string $cssCode): array
	{
		// url(…)
		$matches = [];
		preg_match_all('%\burl\(\s*([a-z]{3,6}://[^/);\'"\s]+|[a-z]{3,6}:(?!/)|//[^/);\'"\s]+)[^)]*\)%i', $cssCode, $matches, PREG_PATTERN_ORDER);
		$origins = $matches[1];

		// url('…')
		$matches = [];
		preg_match_all('%\burl\(\s*(?P<delim>["\'])(?P<url>[a-z]{3,6}://[^/);\'"\s]+|[a-z]{3,6}:(?!/)|//[^/);\'"\s]+).*?(?P=delim)%i', $cssCode, $matches, PREG_PATTERN_ORDER);
		$origins = array_merge($origins, $matches['url']);

		// @import '…'
		$matches = [];
		preg_match_all('%@import\s+(?P<delim>["\'])(?P<url>[a-z]{3,6}://[^/);\'"\s]+|[a-z]{3,6}:(?!/)|//[^/);\'"\s]+)[^)]*(?P=delim)%i', $cssCode, $matches, PREG_PATTERN_ORDER);
		$origins = array_merge($origins, $matches['url']);

		return array_unique(array_map(function ($origin) {
			if (substr($origin, 0, 2) == '//') {
				return substr($origin, 2);
			} else {
				return $origin;
			}
		}, $origins));
	}

	/**
	 * Gibt zurück, ob die Instanz den externen Crawler nutzen darf
	 * @return bool
	 */
	public static function hasCrawler(): bool
	{
		return (bool)($_ENV['APP_HAS_CRAWLER'] ?? false);
	}
}
